=head1 NAME

Bio::Narcisse::BaseAdaptor - Base Adaptor for Narcisse adaptors

=head1 SYNOPSIS

   	Base adaptor provides generic methods to retrieve data from Narcisse Tabbed files

=head1 DESCRIPTION

Fetch functions go by the convention of

    @object_array = @{$adaptor->fetch_all_by_XXXX($arguments_for_XXXX)};

sometimes it returns an array ref denoted by the 'all' in the name of the
method, sometimes an individual object. For example

    $gene = $gene_adaptor->fetch_by_stable_id($stable_id);

or

    @fp  = @{$simple_feature_adaptor->fetch_all_by_Slice($slice)};

=cut

package Bio::Narcisse::BaseAdaptor;

use strict;
use Data::Dumper;

use Bio::Narcisse::Utils::NarcisseDataReader;
use General;

=head2 new

  Title		: new
  Usage    	: $adaptor = new AdaptorInheritedFromBaseAdaptor($dbobj);
  Function	: Creates a new BaseAdaptor object.  The intent is that this
               constructor would be called by an inherited superclass either
               automatically or through $self->SUPER::new in an overridden
               new method.
  Returns	: Bio::Narcisse::BaseAdaptor
  Args		: Bio::Narcisse::DBConnection $dbobj

=cut

sub new {
	my ($class,$dbobj) = @_;

  	my $self = {};
  	bless $self,$class;

 	 if( !defined $dbobj || !ref $dbobj ) {
	    throw("Don't have a db [$dbobj] for new adaptor");
	 }
	 if($dbobj->isa('Bio::Narcisse::DBAdaptor')){
	   $self->db($dbobj);
	   $self->dbc($dbobj->dbc);
	 }
	 elsif( ref($dbobj) =~ /DBAdaptor$/){
	   $self->db($dbobj);
	   $self->dbc($dbobj->dbc);
	 }
	 elsif( ref($dbobj) =~ /DBConnection$/){
	   $self->dbc($dbobj);
	 }
	 else{
	   throw("Don't have a DBAdaptor [$dbobj] for new adaptor");
	 }

	 return $self;


}

=head2 db

  Title		: db
  Usage		: $db = $adaptor->db();
  Function	: Getter/Setter for the DatabaseConnection that this adaptor is
               using.
  Returns	: Bio::Narcisse::DBAdaptor
  Args		: (optional) Bio::Narcisse::DBAdaptor $obj
               the database this adaptor is using.

=cut
sub db{
  my $self = shift;
  $self->{'db'} = shift if(@_);

  return $self->{'db'};

}

=head2 dbc

  Title		: dbc
  Usage	    : $dbc = $adaptor->dbc();
  Function	: Getter/Setter for the DBConnection that this adaptor is
               using.
  Returns	: Bio::Narcisse::DBConnection
  Args		: (optional) Bio::Narcisse::DBConnection $obj
               the database this adaptor is using.

=cut
sub dbc{
  my $self = shift;
  $self->{'dbc'} = shift if(@_);

  return $self->{'dbc'};
}


=head2 features_path

  Title		: features_path
  Usage	    : $features_path = $adaptor->features_path();
  Function	: Getter/Setter for the path to access the features data the adaptor is managing
  Returns	: Path of features path
  Args		: (optional) Features path

=cut
sub features_path{

	my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_gene_path} = $arg );
  $self->{_gene_path};

}

=head2 files_extension

  Title		: files_extension
  Usage	    : $files_extension = $adaptor->files_extension();
  Function	: Getter/Setter for the extension of files the adaptor is handling
  Returns	: Extension
  Args		: (optional) Extention

=cut
sub files_extension{

	my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_files_extension} = $arg );
  $self->{_files_extension};

}

=head2 get_features_in_region

  Title		: get_features_in_region
  Usage	    : $ra_features = $adaptor->get_features_in_region($file, $start, $end, $constraint, $value);
  Function	: Returns an ref on an array holding all the features stored in $file, located in the region
  			  delimited by $start and $end. We can limit the results returned by giving
  			  a constraint on a column, with $column_constraint (column on which we put
  			  a contraint) and $value_constraint (contraint that must be satisfied)
  Returns	: Array ref of generic features (hashes)
  Args		: $file - file in which we will look for features
  			  $start - begining of region of interest
  			  $end - end of region of interest
  			  $column_constraint - name of the column we put a contraint on
  			  $value_constraint - value of the column to be considered as a "valid" feature

  Note      : all feature which START position is inside the the region will be returned.

=cut

sub get_features_in_region
{
# warn "API : BaseAdaptor::get_features_in_region";
	my ($class, $file, $start, $end, $column_constraint, $value_constraint) = @_;

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader(); # marche aussi, un peu moins rapide

	my %h_where = ( RANGE => { start => $start, end => $end } );
	if( defined $column_constraint )
	{
		$h_where{ VALUE } = { $column_constraint => $value_constraint };
	}

	$o_ndr->Prepare( -from  => $file,
	                 -where => \%h_where,
	                 -case_sensitive => &FALSE );
	my $ra_row = $o_ndr->FetchallArrayref();

	return $ra_row;

#   	$class->Bio::Narcisse::Utils::ScriptUtils::openFileAndReadHeader($file);
# 
#   	my $keys_pos = $class->{'keys_pos'};
#   	my $pos_keys = $class->{'pos_keys'};
#   	my $fh = $class->file_handler;
# 
#   	my $index = 0;
#   	$index = Bio::Narcisse::Utils::ScriptUtils::getLocationFromIndex($file."_index", $start) if (-e $file."_index");
# 
# 
#   	# Then we move to the correct location in the file
#   	seek ($fh, $index, 0);
# 
# 	my @a_features;
#   	while (my $line = <$fh>){
# 
#   		chomp $line;
#   		next if ($line =~ /^#/);
# 
#   		my @a_elements = split("\t", $line);
#   		my $start_pos_feature = $a_elements[ $keys_pos->{'start'} ];
#   		my $end_pos_feature = $a_elements[ $keys_pos->{'end'} ];
# 
# # warn Dumper {	file => $file,
# # 				fea_start => $start_pos_feature, fea_end => $end_pos_feature,
# # 				start => $start, end => $end
# # 			};
# #
#   		next if ($end_pos_feature < $start);
#   		last if ($end < $start_pos_feature );
# 
#   		if ($column_constraint && $value_constraint){
#   			next if ($a_elements[ $keys_pos->{$column_constraint} ] !~ /^$value_constraint$/);
#   		}
# 
#   		my $rh_current_feature = {};
#   		for (my $i = 0; $i <= $#a_elements; $i++){
#   			$rh_current_feature->{ $pos_keys->{$i} } = $a_elements[$i];
#   		}
# 
#   		push @a_features, $rh_current_feature;
# 
#   	}
# 
# warn Dumper \@a_features;
# 	return \@a_features;
}

=head2 generic_fetch

  Title		: generic_fetch
  Usage	    : $rh_feature = $adaptor->generic_fetch($file, $column, $element);
  Function	: Finds and returns a feature which has $element in is column $column in file $file
  Returns	: Hash ref representing a feature
  Args		: $file - file in which we will look for the feature
  			  $column - column contraint
  			  $element - value of the contraint

=cut

sub generic_fetch{

	my ($self, $file, $column, $element) = @_;

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr->Prepare( -from => $file,
					 -where => { VALUE => { $column => $element } },
					 -case_sensitive => &FALSE
				);
	my $ra_row = $o_ndr->FetchallArrayref();
# warn Dumper { file => $file, col => $column, elt => $element, res => $ra_ };

	return $$ra_row[ 0 ];
}


=head2 generic_multiple_fetch

  Prerequiste : Not in EnsEMBL API.
  Title		: generic_multiple_fetch
  Usage	    : $ra_features = $adaptor->generic_multiple_fetch( $column, $element, $chrom );
  Function	: Finds and returns features whose has $element in their column $column in file $file
  Returns	: Ref on an array holding hash ref representing features
  Args		: $file - file in which we will look for features
  			  $column - column contraint
  			  $element - value of the contraint

	Note	: apparemment pas utilise <-- Seb.
=cut
sub generic_multiple_fetch{

	my ($self, $file, $column, $element) = @_;

	throw( "method not implemented because it does not appear to be used." );

#   	Bio::Narcisse::Utils::ScriptUtils::openFileAndReadHeader($self, $file);
#   	my $fh = $self->file_handler;
# 
#   	# Then we move to the correct location in the file
#   	seek ($fh, 0, 0);
# 
#   	my $rh_keys_pos = $self->{'keys_pos'};
#   	my $rh_pos_keys = $self->{'pos_keys'};
# 
#   	my $found = 0;
#   	my @a_features = ();
# 
#   	while (my $line = <$fh>){
# 
#   		my $rh_feature = {};
# 
#   		chomp $line;
#   		next if ($line =~ /^#/);
# 
#   		my @a_elements = split("\t", $line);
# 
#   		my $requested_column = $a_elements[ $rh_keys_pos->{$column} ] if (defined $column);
# 
#  		# The second test corresponds to the search of a transcript from a gene id.
#   		next if ($element && ($requested_column !~ /^$element$/ && $requested_column !~ /^($element)t\d+$/));
# 
#   		for (my $i = 0; $i <= $#a_elements; $i++){
#   			$rh_feature->{$rh_pos_keys->{$i}} = $a_elements[$i];
#   		}
# 
#   		push @a_features, $rh_feature;
# 
#   	}
# 
#   	return \@a_features;

}

=head2 fetch_columns

  Title		: fetch_columns
  Usage	    : $rh_columns = $adaptor->fetch_columns($file, $column, $element, $column1, $column2, ...);
  Function	: Finds and returns values for requested columns ($column1, ..., $columnN) that matches the contraint on $column
  Returns	: Hash ref with key = column_name and value = column content
  Args		: $file - file in which we will look for features
  			  $column - column contraint
  			  $element - value of the contraint
  			  $columnN - name of the column we want the value

=cut
sub fetch_columns{

	my ($self, $file, $column, $element) = splice(@_, 0, 4);
	my $ra_requested_columns_names = \@_;

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

	# clause select pour maintenir la coherence avec la doc --> TODO a voir si c'est necessaire parce que ca ralentit le bouzin
	# actuellement (02/2009) la clause select est ignoree.
	$o_ndr->Prepare( -select => $ra_requested_columns_names,
	                 -from => $file,
	                 -where => { VALUE => { $column => $element } },
	                 -case_sensitive => &FALSE );
	my $rh_row = $o_ndr->FetchallHashref( 'element' );

	# renvoie le plus souvent un seul resultat, mais pas toujours !
	my @a_rh_row = values %$rh_row;

	return $a_rh_row[0];


# 	$self->Bio::Narcisse::Utils::ScriptUtils::openFileAndReadHeader($file);
# 	my $fh = $self->file_handler;
#
# 	seek ($fh, 0, 0);
#
#   	my $rh_keys_pos = $self->{'keys_pos'};
#
#   	while (my $line = <$fh>){
#   		chomp $line;
#   		next if ($line =~ /^#/);
#
#   		my @a_elements = split("\t", $line);
#   		my $column_value = $a_elements[$rh_keys_pos->{$column}];
#
#   		next if ($column_value !~ /$element/);
#
#   		my %h_requested_values;
#   		foreach my $requested_colum (@$ra_requested_columns_names){
#   			$h_requested_values{$requested_colum} = $a_elements[ $rh_keys_pos->{$requested_colum} ];
#   		}
#  		return \%h_requested_values;
#
#   	}
#
#   	return undef;

}

sub _objs_from_hashes {
  my $self = shift;

  throw("abstract method _objs_from_hashes not defined by implementing"
             . " subclass of BaseFeatureAdaptor");
}


1;
