=pod
=head1 NAME

Bio::Narcisse::BaseFeatureAdaptor - An Abstract Base class for all
FeatureAdaptors

=head1  SYNOPSIS

Abstract class - should not be instantiated.  Implementation of
abstract methods must be performed by subclasses.

=head1 DESCRIPTION

This is a base adaptor for feature adaptors. This base class is simply a way
of eliminating code duplication through the implementation of methods
common to all feature adaptors.

=head1 METHODS

=cut

package Bio::Narcisse::BaseFeatureAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::BaseAdaptor;
use Bio::Narcisse::Utils::Exception qw( throw );

use base qw( Bio::Narcisse::BaseAdaptor );


=head2 new

  Function	: Constructor which just initializes internal cache structures
  Returns 	: Bio::Narcisse::BaseFeatureAdaptor
  Args		: list of args @args
               Superclass constructor arguments

=cut


=head2 generic_features_translation

  Title		: generic_features_translation
  Usage	    : $ra_obj_feature = $adaptor->generic_features_translation ($ra_hashes_features, $region_name);
  Function	: Generic method used to translate features represented by hashes into the right objets
  Prerequisite : each feature adaptor should implement a "_objs_from_hashes" method
  Returns	: Ref on an array holding object features
  Args		: $ra_features - ref on an array holding hashes ref representing generic features
  			  $seq_region_name - name of the region the features belong to

=cut
sub generic_features_translation
{
	my $self = shift;
	my( $ra_features, $seq_region_name ) = @_;

	my $cs = $self->db->default_CoordSystem;

	my $sla = $self->db->get_SliceAdaptor;
	my $slice = Bio::EnsEMBL::Slice->new(   -COORD_SYSTEM => $cs,
											-SEQ_REGION_NAME => $seq_region_name,
											-START => 1,
											-END => $sla->karyotype->{$seq_region_name}->{length},
											-ADAPTOR => $sla
										);

	return $self->_objs_from_hashes($ra_features, $slice);
}


=head2 fetch_feature

  Title		: fetch_feature
  Usage	    : $obj_feature = $adaptor->fetch_feature($column, $element, $path, $extension);
  Function	: Generic method used to find and return a feature as object, seeking into all files
  			  with an extension $extension in the folder specified in $path
  Returns	: Feature object
  Args		: $column - column constraint
  			  $element - value constraint
  			  $path - folder in which the search will be performed
  			  $extension - extension of the files we must search into

=cut
sub fetch_feature{

	my ($self, $column, $element, $ra_files ) = @_;

#	my @a_files = glob ($path."*.$extension");
# 	# on pourrait aussi faire comme ca, on verra plus tard
# 	my @a_files = $self->conf()->GetFeatureFiles( $self->feature_key);
	foreach my $filename ( @$ra_files )
	{
		# $rh_feature est une ref de hash sur une ligne de fichier
		my $rh_feature = $self->generic_fetch( $filename ,$column, $element );
		if( defined $rh_feature )
		{
			my ($seq_region_name) = ( $filename =~ /.+\/chromosome(.+)\..+$/);
			my $ra_features = $self->generic_features_translation( [$rh_feature], $seq_region_name );
			return $$ra_features[0];
		}
	}

	return undef;
}


sub _GetFeatureKey
{
	my $self = shift;

	throw LipmError::ObjectException( $self, 'abstract method' );

	return;
}

=head2 fetch_all_by_Slice

  Title		: fetch_all_by_Slice
  Function	: Retrieves all features for the Slice. Each FeatureAdaptor should implement
  			  _objs_from_hashes method, in order to translate hashes into objects of the type they deal with.
  Returns	: listref Bio::EnsEMBL::Feature on slice
  Args		: Bio::EnsEMBL::Slice $slice

 Note : <seb>J'ai des doutes sur cette fonction :
       ExonAdaptor->fetch_all_by_slice ne fonctionne pas. -> je surcharge dans la classe
	   GeneAdaptor->fetch_all_by_slice ne fonctionne pas. -> je surcharge dans la classe

	   --> 31-10-2010 correction de cette fonction pour VariationFeatureAdaptor.
	   </seb>

=cut

sub fetch_all_by_Slice
{
	my $self = shift;
	my( $slice, $logic_names ) = @_;

  	my $start = $slice->start;
  	my $end   = $slice->end;

	my $chrom       = $slice->seq_region_name;
	my $feature_key = $self->_GetFeatureKey();
	my $o_species   = $self->db()->conf();

	my %h_args = ( -chrom => $chrom, key => $feature_key );
	my $chrom_file  = $o_species->GetFeatureFiles( %h_args );

	my $ra_features = $self->get_features_in_region( $chrom_file, $start, $end );

	return $self->_objs_from_hashes( $ra_features, $slice );
}


=head2 function fetch_by_keyword

 Usage     : $obj_feature = $adaptor->fetch_by_keyword( $keyword );
 Function  : Generic method used to find and return a feature as object, seeking into all files
             with an extension $extension in the folder specified in $path
 Returns   : Feature object
 Args      :
     $keyword, string, what we seek.

=cut

sub fetch_by_keyword
{
	my $self = shift;
	my( $keyword ) = @_;


	# en developpement
# #	my @a_files = glob ($path."*.$extension");
# # 	# on pourrait aussi faire comme ca, on verra plus tard
# # 	my @a_files = $self->conf()->GetFeatureFiles( $self->feature_key);
# 	foreach my $filename ( @$ra_files )
# 	{
# 		# $rh_feature est une ref de hash sur une ligne de fichier
# 		my $rh_feature = $self->generic_fetch( $filename ,$column, $element );
# 		if( defined $rh_feature )
# 		{
# 			my ($seq_region_name) = ( $filename =~ /.+\/chromosome(.+)\..+$/);
# 			my $ra_features = $self->generic_features_translation( [$rh_feature], $seq_region_name );
# 			return $$ra_features[0];
# 		}
# 	}

	return undef;
} #fetch_by_keyword


1;
