=head1 NAME - Bio::Narcisse::DBAdaptor

=head1 SYNOPSIS

    $db = new Bio::Narcisse::DBAdaptor( -dbname => "chick" );

    $gene_adaptor = $db->get_GeneAdaptor();

    $gene = $gene_adaptor()->fetch_by_stable_id($stable_id);

    $slice = $db->get_SliceAdaptor()->fetch_by_chr_start_end('X', 1, 10000);

=head1 DESCRIPTION


This class provides access to Narcisse configuration. When an instance is created, this class looks for
the "NARCISSE_CFG" environment variable, which gives the location of Narcisse configuration file. The
content of this file is then stored in a hash. It also allows to load all possible adaptors in the
Registry.


=cut


package Bio::Narcisse::DBAdaptor;

use strict;
use warnings;

use Data::Dumper;

use base qw( Bio::EnsEMBL::DBSQL::DBAdaptor );

use Bio::Narcisse::Utils::Exception qw( throw );
use Bio::Narcisse::Utils::ScriptUtils qw( LoadData );
use Bio::Narcisse::Utils::NarcisseDataReader;
# use Bio::EnsEMBL::Utils::Argument qw(rearrange);
use Bio::EnsEMBL::Utils::ConfigRegistry;
use Bio::Narcisse::DBConnection;
use Bio::EnsEMBL::CoordSystem;

use ParamParser;
use Configuration::InstanceObserver;
use NARCISSE_utils;	#&NONE

my $reg = "Bio::EnsEMBL::Registry";

#--------------------------------------------------
#	Constants
#--------------------------------------------------
# use constant DATA_PATH_KEY => "ROOT_DATA";
# use constant ROOT_PATH_KEY => "ROOT";
# use constant CFG_PATH_KEY => "CFG";
# use constant SPECIES_LIST => "SPECIES_LIST";
# use constant NARCISSE_VERSION => "NARCISSE_VERSION";

=head1 SUBROUTINES

=cut


=head2 function new

 Title        : new
 Usage        : my $dba = new Bio::Narcisse::DBAdaptor( -dbname => $species,
                                                        -cfg => $path_to_cfg );
 Prerequisite : none
 Function     : Constructs a new Object DBAdaptor with information contained in Narcisse
                 configuration file.
 Returns      : a ref to the object DBAdaptor
 Args         : $species, string, the species short name
                $path_to_cfg, string, path to narcisse.cfg file. [$ENV{NARCISSE_CFG}]

 Attributes   :

 	non present into native DBAdaptor :
#  		_species_long
		_version


=cut

sub new
{
	my $class = shift;
	my %h_args = ( -o_inst_obs => undef,
					-o_species => undef,
	               @_ );

	my ( $species, $userName, $password, $con ) =
		@h_args{ qw( -dbname -username -password -dbconn ) };

	my $self = {};
	bless $self,$class;

	if(defined($con)){
		$self->dbc($con);
	}
	else{
		# cette classe ne sert a rien (pour le moment), toutes les infos sont bidons (Seb)
		$self->dbc( new Bio::Narcisse::DBConnection( %h_args ) );
	}


	my( $short, $long, $version );

	my( $o_species, $o_inst_obs ) = @h_args{ qw( -o_species -o_inst_obs ) };
	use Carp;
	if( defined $o_species )
	{
		$self->conf( $o_species );
	}
	else
	{
		carp "OBSOLETE : use a SpeciesOberserver";

		$o_species = New SpeciesObserver( species => $species );
		$self->conf( $o_species );
	}

	$short = $o_species->GetShortName();
	$long  = $o_species->GetLongName();
	$version = $o_species->GetInstanceObserver()->GetVersion();

	$self->species( $short );

# 	# Genome long name
# 	$self->species_long( $long );

	# Narcisse instance version
	$self->version( $version );

	# Default coordinate system, used for compatibility reasons. In Narcisse, the only
	# coordinate system used  is the "chromosome"
	my $default_cs = Bio::EnsEMBL::CoordSystem->new( -NAME => "chromosome", -RANK => 1, -VERSION => $self->version);
	$self->default_CoordSystem( $default_cs );

	# End of the registry initialization part. Loads available adaptors
	$self = Bio::EnsEMBL::Utils::ConfigRegistry::gen_load($self);

	return $self;
}


=head2 function species

 Title        : species
 Usage        : my $species = $dba->species;
 Prerequisite : none
 Function     : Returns the species linked to the DBAdaptor
 Returns      : species name
 Args         : (optional) The species name

=cut

sub species {
  my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_species} = $arg );
  $self->{_species};
}

# commentee parce que pas d'origine dans DBAdaptor
# et on peut l'avoir en passant par conf()->GetLongName()
# =head2 function species_long
# 
#  Title        : species_long
#  Usage        : my $species_long = $dba->species_long;
#  Prerequisite : none
#  Function     : Returns the species' long name linked to the DBAdaptor
#  Returns      : species long name
#  Args         : (optional) The species long name
# 
# =cut
# sub species_long {
#   my ($self, $arg ) = @_;
#   ( defined $arg ) &&
#     ( $self->{_species_long} = $arg );
#   $self->{_species_long};
# }
# 

=head2 function version

 Title        : version
 Usage        : my $version = $dba->version;
 Prerequisite : none
 Function     : Returns the version of Narcisse
 Returns      : version of Narcisse
 Args         : (optional) Version of Narcisse

=cut
sub version {
  my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_version} = $arg );
  $self->{_version};
}

=head2 function conf_file

 Title        : conf_file
 Usage        : my $conf_file = $dba->conf_file;
 Prerequisite : none
 Function     : Returns the path leading to Narcisse configuration file (defined by
 				the environment variable $NARCISSE_CFG).
 Returns      : Path to Narcisse configuration file
 Args         : (optional) Path to Narcisse configuration file

=cut

sub conf_file
{
	my $self = shift;
	my ( $arg ) = @_;

  ( defined $arg ) &&
    ( $self->{_conf_file} = $arg );

	return $self->{_conf_file};
}


=head2 function conf

 Title        : conf
 Usage        : my $conf = $dba->conf;
 Prerequisite : none
 Function     : Returns the SpeciesObserver object relative to this species.
 Returns      : SpeciesObserver object
 Args         : (optional) SpeciesObserver object

=cut
sub conf
{
	my $self = shift;
	my ( $arg ) = @_;

  ( defined $arg ) &&
    ( $self->{_narcisse_conf} = $arg );

	return $self->{_narcisse_conf};
}

=head2 function group

 Title        : group
 Usage        : my $group = $dba->group;
 Prerequisite : none
 Function     : Returns the group of this adaptor (a DBAdaptor is in the "Core" group)
 Returns      : Group of the DBAdpator
 Args         : (optional) Group of the DBAdpator

=cut
sub group
{
	my $self = shift;
	my ( $arg ) = @_;

  ( defined $arg ) &&
    ( $self->{_group} = $arg );

	return $self->{_group};
}

=head2 function dbc

 Title        : dbc
 Usage        : my $dbc = $dba->dbc;
 Prerequisite : none
 Function     : Returns the DBConnection linked to this adaptor. A DBConnection is
 				linked to an adaptor for compatibility reasons with Ensembl API.
 Returns      : Bio::Narcisse::DBConnection object
 Args         : (optional) DBConnection object

=cut
sub dbc{
  my $self  = shift;

  if(@_){
    my $arg = shift;
    if(defined($arg)){
      if(!$arg->isa('Bio::Narcisse::DBConnection')){
	throw("$arg is no a DBConnection\n");
      }
    }
    $self->{_dbc} = $arg;
  }
  return $self->{_dbc};
}


=head2 function default_CoordSystem

 Title        : default_CoordSystem
 Usage        : my $cs = $dba->default_CoordSystem;
 Prerequisite : none
 Function     : Returns the default coordinate system associated with this adaptor.
 				In Narcisse, we use only one type of coordinate system (chromosome)
 Returns      : Bio::EnsEMBL::CoordSystem object
 Args         : (optional) Bio::EnsEMBL::CoordSystem

=cut
sub default_CoordSystem{

	my ($self, $arg ) = @_;
  	( defined $arg ) &&
    	( $self->{_default_CoordSystem} = $arg );
  	$self->{_default_CoordSystem};
}


=head2 get_available_adaptors

  Title		: get_available_adaptors
  Usage		: my %pairs = %{$dba->get_available_adaptors()};
  Function	: gets a hash of the available adaptors
  Returns	: reference to a hash

=cut
sub get_available_adaptors{
  	my %h_pairs =  (

    map( { $_ => "Bio::Narcisse::${_}Adaptor" } qw(
      Exon	                   Gene
      RepeatFeature        Sequence
      Slice                Transcript
      Translation              Attribute
    ) ));
#     map( { $_ => "Bio::Narcisse::${_}Adaptor" } qw(
#       Exon	                   Gene                 MiscFeature
#       ProteinFeature           RepeatFeature        Sequence
#       SimpleFeature            Slice                Transcript
#       Translation              Attribute
#     ) ));

  return (\%h_pairs);
}


use vars '$AUTOLOAD';

sub  AUTOLOAD{
  my ($self,@args) = @_;

  my $type;
  if($AUTOLOAD =~ /^.*::get_(\w+)Adaptor$/){
    $type = $1;
  }
  elsif($AUTOLOAD =~ /^.*::get_(\w+)$/){
    $type = $1;
  }
  else{
    Bio::Narcisse::Utils::Exception::throw("Could not work out type for $AUTOLOAD \n");
  }

	# note : $reg = "Bio::EnsEMBL::Registry"
	my $ret = $reg->get_adaptor( $self->species(), $self->group(), $type );

  # $ret est un objet Bio::Narcisse::SequenceAdaptor ( ou bien fonction du type ? )
  if($ret){
    return $ret;
  }
  else{
    Bio::Narcisse::Utils::Exception::throw( "Couldnt get adaptor $type for ".$self->species." ".$self->group."\n");
    Bio::Narcisse::Utils::Exception::warning("Could not find $type adaptor in the registry for ".$self->species." ".$self->group."\n");
    return $ret;
  }
  die("No such method: $AUTOLOAD\n");
}


sub DESTROY {

} # required due to AUTOLOAD

1;