=head1 NAME

Bio::Narcisse::ExonAdaptor - An adaptor responsible for the retrieval and
storage of exon objects

=head1 SYNOPSIS

  $exon_adaptor = $database_adaptor->get_ExonAdaptor();
  $exon = $exon_adaptor->fetch_by_dbID($dbID);

=head1 DESCRIPTION

The ExonAdaptor is responsible for retrieving and storing Exon objects Narcisse files.
Most of the ExonAdaptor functionality is inherited from the
B<Bio::Narcisse::BaseFeatureAdaptor> class.

=cut



package Bio::Narcisse::ExonAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::SliceAdaptor;
use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::Narcisse::DBAdaptor;
use Bio::EnsEMBL::Exon;
use Bio::Narcisse::Utils::Exception qw( throw );

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_ExonAdaptor();
  Function	: Creates a new ExonAdaptor object.
  Returns	: Bio::Narcisse::ExonAdaptor
  Args		: None

=cut


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $exon = $exon_adaptor->fetch_by_dbID('s19g8t1e2');
  Function	: Retrieves an Exon from the files via its db id
  Returns	: Bio::EnsEMBL::Exon
  Args		: string $db_id
               the db id of the exon to retrieve

=cut
sub fetch_by_dbID
{
	my $self = shift;
	my( $db_id ) = @_;

	my @a_files = $self->db()->conf()->GetExonFiles();
	my $o_exon = $self->fetch_feature( 'element', $db_id, \@a_files );

	return $o_exon;
}


=head2 fetch_by_stable_id

  Title		: fetch_by_stable_id
  Usage		: $exon = $exon_adaptor->fetch_by_stable_id('ENSE0000988221');
  Function	: Retrieves an Exon from the files via its stable id
  Returns	: Bio::EnsEMBL::Exon
  Args		: string $stable_id
               the stable id of the exon to retrieve

=cut
sub fetch_by_stable_id
{
	my $self = shift;
	my( $stable_id ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetExonFiles();
	my $o_exon = $self->fetch_feature( 'symbol', $stable_id, \@a_files );

	return $o_exon;
}


=head2 fetch_all_by_Transcript

  Title		: fetch_all_by_Transcript
  Function	: Retrieves all Exons for the Transcript
  Returns	: listref Bio::EnsEMBL::Exon on Transcript slice 
  Args		: Bio::EnsEMBL::Transcript $transcript

=cut
sub fetch_all_by_Transcript
{
	my $self = shift;
	my( $transcript ) = @_;

	my $tslice = $transcript->slice();
	if( !defined $tslice )
	{
		throw("Transcript must have attached slice to retrieve exons.");
	}

	my $o_dba = $self->db();

	# use a small slice the same size as the transcript
	my $slice = $o_dba->get_SliceAdaptor()->fetch_by_Feature( $transcript );

	my $chrom   = $slice->seq_region_name;
	my $chromosome_file = $o_dba->conf()->GetExonFiles( -chrom  => $chrom );

	my $ra_exons = $self->get_features_in_region($chromosome_file, $slice->start, $slice->end, 'parent', $transcript->dbID);

	my @a_objects = @{$self->_objs_from_hashes($ra_exons, $slice)};

	@a_objects = reverse @a_objects if ($transcript->strand == -1);

	return \@a_objects;
}

sub _GetFeatureKey
{
	my $self = shift;

	return 'EXON';
}


# =head2 fetch_all_by_Slice

  # Title		: fetch_all_by_Slice
  # Function	: Read Exon file and get features that belong to the slice.
  # Returns	: listref Bio::EnsEMBL::Feature on slice
  # Args		: Bio::EnsEMBL::Slice $slice

 # Note : <seb>on pourra peut-etre utiliser la fonction de la mere et ne passer qu'un mot clef.
		# la fonction de la mere contenant GetFeatureFiles.
		# A voir plus tard.
	   # </seb>

# =cut

# sub fetch_all_by_Slice
# {
	# my $self = shift;
	# my( $slice, $logic_names ) = @_;

  	# my $start = $slice->start;
  	# my $end = $slice->end;

	# my $o_dba   = $self->db();
	# my $species = $o_dba->species();
	# my $chrom   = $slice->seq_region_name;
	# my $chromosome_file = $o_dba->conf()->GetExonFiles( -chrom  => $chrom );
	# my @a_features = @{$self->get_features_in_region($chromosome_file, $start, $end)};

	# return $self->_objs_from_hashes(\@a_features, $slice);

# }


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::Exon objects
  Returns	: listref Bio::EnsEMBL::Exon
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut
sub _objs_from_hashes {
	
	my ($self, $tab_of_hash, $slice) = @_;
	
	my @a_exons;
	foreach my $feature (@$tab_of_hash){
		push @a_exons, Bio::EnsEMBL::Exon->new( '-start'  		=>  $feature->{'start'},
											    '-end'			=>  $feature->{'end'},
											    '-strand'		=>  $feature->{'strand'},
											    '-adaptor'		=>  $self,
											    '-slice'		=>  $slice,
											    '-dbID'			=>  $feature->{'element'},
											    '-stable_id'	=>  $feature->{'symbol'},
											    '-created_date'	=>  undef,
											    '-modified_date'=>  undef,
											    '-phase'        =>  0,
											    '-end_phase'    =>  0
											);
	}

	return \@a_exons;
}
1;
