=head1 NAME

Bio::Narcisse::GeneAdaptor - Database adaptor for the retrieval and
storage of Gene objects

=head1 SYNOPSIS

  use Bio::Narcisse::DBAdaptor;

  $dba = Bio::Narcisse::DBAdaptor->new(...);

  $gene_adaptor = $dba->get_GeneAdaptor();

  $gene = $gene_adaptor->fetch_by_dbID('XM_429624');

  $gene = $gene_adaptor->fetch_by_stable_id('LOC419094');

  @genes = @{$gene_adaptor->fetch_all_by_external_name('BRCA2')};

  $slice_adaptor = $db->get_SliceAdaptor;
  $slice = $slice_adaptor->fetch_by_region('chromosome', '1', 1, 1000000);
  @genes = @{$gene_adaptor->fetch_all_by_Slice($slice)};

=head1 DESCRIPTION

This is an adaptor for the retrieval and storage of gene objects.

=cut


package Bio::Narcisse::GeneAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::SliceAdaptor;
use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::Narcisse::DBAdaptor;
use Bio::EnsEMBL::Gene;
use Bio::Narcisse::Utils::Exception qw( throw );

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_GeneAdaptor();
  Function	: Creates a new GeneAdaptor object.
  Returns	: Bio::Narcisse::GeneAdaptor
  Args		: None

=cut


sub _GetFeatureKey
{
	my $self = shift;

	return 'GENE';
}

# =head2 fetch_all_by_Slice

  # Title		: fetch_all_by_Slice
  # Function	: Retrieves all Genes for the Slice
  # Returns	: listref Bio::EnsEMBL::Gene on slice
  # Args		: Bio::EnsEMBL::Slice $slice
  			  # $load_transcripts - flag indicating if the method must load transcripts for each gene

# =cut

# sub fetch_all_by_Slice
# {
	# my $self = shift;
	# my( $slice, $load_transcripts ) = @_;

	# # mother copy-paste like
  	# my $start = $slice->start;
  	# my $end = $slice->end;

	# my $o_dba   = $self->db();
	# my $species = $o_dba->species();
	# my $chrom   = $slice->seq_region_name;
	# my $chromosome_file = $o_dba->conf()->GetGeneFiles( -chrom  => $chrom );

	# my @a_features = @{$self->get_features_in_region($chromosome_file, $start, $end)};

	# my $ra_genes = $self->_objs_from_hashes(\@a_features, $slice);
	# # end of mother copy-paste like

	# if (defined($load_transcripts) and $load_transcripts == 1){

		# # Else the transcripts have to be loaded and added to their respective gene.
		# my $transcript_adaptor = $self->db->get_TranscriptAdaptor();
		# foreach my $gene (@$ra_genes){
			# my $ra_transcripts = $transcript_adaptor->fetch_all_by_Gene($gene);

			# # we add transcripts to the current gene
			# foreach my $transcript (@$ra_transcripts){
				# $gene->add_Transcript($transcript);
			# }
		# }
	# }

	# return $ra_genes;
# }


=head2 fetch_by_stable_id

  Title		: fetch_by_stable_id
  Usage		: $gene = $gene_adaptor->fetch_by_stable_id('ENSGXXXXXXXXXX');
  Function	: Retrieves a Gene from the files via its stable id which is locus_tag
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $stable_id
               the stable id of the gene to retrieve

=cut
sub fetch_by_stable_id
{
	my $self = shift;
	my( $stable_id ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetGeneFiles();
	my $o_gene = $self->fetch_feature( 'locus_tag', $stable_id, \@a_files );

	return $o_gene;
}


=head2 fetch_by_translation_stable_id

  Usage		: $gene = $gene_adaptor->fetch_by_translation_stable_id( $prot_id );
  Function	: Retrieves a Gene from the files via a translation stable_id,
              which in Narcisse is the protein_id of transcript files.
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $prot_id
               the stable id of the gene to retrieve

=cut

sub fetch_by_translation_stable_id
{
	my ($self, $prot_id) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_translation_id( $prot_id );
	return undef  unless( defined $o_transcript );

	# obtenir le gene
	my $tr_id = $o_transcript->dbID();
	my $o_gene = $self->fetch_by_transcript_id( $tr_id );

	return $o_gene;
}


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $gene = $gene_adaptor->fetch_by_dbID('s1g16');
  Function	: Retrieves a Gene from the files via its internal id
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $dbID
               the internal id of the gene to retrieve

=cut

sub fetch_by_dbID
{
	my $self    = shift;
	my( $dbID ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetGeneFiles();
	my $o_gene = $self->fetch_feature( 'element', $dbID, \@a_files );

	return $o_gene;
}


=head2 fetch_by_symbol

  Title		: fetch_by_symbol
  Usage		: $gene = $gene_adaptor->fetch_by_symbol('cdc25a');
  Function	: Retrieves a Gene from the files via its symbol
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $symbol
              the $symbol of the gene to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_symbol
{
	my $self = shift;
	my( $symbol ) = @_;

	my $o_dba     = $self->db();
	my $o_species = $o_dba->conf();
	my @a_files   = $o_species->GetGeneFiles();
	my $o_gene    = $self->fetch_feature( 'symbol', $symbol, \@a_files );

	return $o_gene;
}



=head2 fetch_by_transcript_accession

  Title		: fetch_by_symbol
  Usage		: $gene = $gene_adaptor->fetch_by_transcript_accession( $x );
  Function	: Retrieves a Gene from the files via its accession
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $accession
              the $symbol of the gene to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_transcript_accession
{
	my $self = shift;
	my( $accession ) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_accession( $accession );
	return undef  unless( defined $o_transcript );

	# obtenir le gene
	my $tr_id = $o_transcript->dbID();
	my $o_gene = $self->fetch_by_transcript_id( $tr_id );

	return $o_gene;
}


sub fetch_by_exon_stable_id {

	my ($self, $stable_id) = @_;

	throw("Not implemented!");

}


=head2 fetch_by_transcript_id

  Title		: fetch_by_transcript_id
  Usage		: $gene = $gene_adaptor->fetch_by_transcript_id('s1g4t1');
  Function	: Retrieves a Gene from the files via his transcript db id
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $db_id
               the db id of the transcript

=cut
sub fetch_by_transcript_id {

	my ($self, $db_id) = @_;

	#Dans narcisse: transcript db id = gene db id + 't' + num transcript
	$db_id =~ s/t\d+$//;

	return $self->fetch_by_dbID($db_id);

}


=head2 fetch_by_transcript_stable_id

  Title		: fetch_by_transcript_stable_id
  Usage		: $gene = $gene_adaptor->fetch_by_transcript_stable_id('ENSG00000146556#1');
  Function	: Retrieves a Gene from the files via his transcript stable id
  Returns	: Bio::EnsEMBL::Gene
  Args		: string $trans_stable_id
               the stable id of the transcript

=cut
sub fetch_by_transcript_stable_id {
    my ($self, $trans_stable_id) = @_;

	#Dans narcisse: transcript stable id = locus tag
	# c'est donc le  locus tag de son gene + eventuellement #numTranscript
	$trans_stable_id =~ s/#.+$//;

	return $self->fetch_by_stable_id($trans_stable_id);
}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::Gene objects
  Returns	: listref Bio::EnsEMBL::Gene
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut
sub _objs_from_hashes {

	my ($self, $tab_of_hash, $slice) = @_;

	my @a_genes;
	foreach my $feature (@$tab_of_hash){
		push @a_genes, Bio::EnsEMBL::Gene->new( '-start'  		=> $feature->{'start'},
											  '-end'			=> $feature->{'end'},
											  '-strand'			=> $feature->{'strand'},
											  '-adaptor'		=> $self,
											  '-slice'			=>  $slice,
											  '-dbID'			=>  $feature->{'element'},
											  '-stable_id'		=>  $feature->{'locus_tag'},
											  '-external_name' =>  $feature->{'symbol'},
											  '-created_date'	=>  undef,
											  '-modified_date'	=>  undef
											);

	}

	return \@a_genes;
}


1;
