=head1 NAME

Bio::Narcisse::RepeatFeatureAdaptor

=head1 SYNOPSIS

$rfa = $database_adaptor->get_RepeatFeatureAdaptor();

my $repeat = $rfa->fetch_by_dbID(1234);
my @repeats = @{$rfa->fetch_all_by_Slice($slice)};

=head1 DESCRIPTION

This is an adaptor for the retrieval of RepeatFeature objects
from the database.  Most of the implementation is in the
superclass BaseFeatureAdaptor.

=head1 METHODS

=cut



package Bio::Narcisse::RepeatFeatureAdaptor;

use strict;
use warnings;

use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::EnsEMBL::RepeatFeature;
use Bio::EnsEMBL::RepeatConsensus;
use Bio::Narcisse::Utils::Exception qw(throw warning);

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_RepeatFeatureAdaptor();
  Function	: Creates a new RepeatFeatureAdaptor object.
  Returns	: Bio::Narcisse::RepeatFeatureAdaptor
  Args		: None

=cut


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $repeat = $repeat_adaptor->fetch_by_dbID('s7r269428');
  Function	: Retrieves a repeat feature from the files via its db id
  Returns	: Bio::EnsEMBL::RepeatFeature
  Args		: string $db_id
               the db id of the repeat feature to retrieve

=cut

sub fetch_by_dbID
{
	my $self = shift;
	my( $db_id ) = @_;

	my $o_dba    = $self->db();
	my @a_files  = $o_dba->conf()->GetRepeatFiles();
	my $o_repeat = $self->fetch_feature( 'element', $db_id, \@a_files );

	return $o_repeat;
}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::RepeatFeature objects
  Returns	: listref Bio::EnsEMBL::RepeatFeature
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut
sub _objs_from_hashes{
	
	my ($self, $tab_of_hash, $slice) = @_;
	
	my @a_repeats;
	foreach my $feature (@$tab_of_hash){
		
		my $rc = Bio::EnsEMBL::RepeatConsensus->new_fast
          ({'dbID'             => $feature->{'element'},
            'adaptor'          => undef,
            'name'             => $feature->{'symbol'},
	        'repeat_type'      => $feature->{'type'}});
            
		push @a_repeats, Bio::EnsEMBL::RepeatFeature->new_fast
      		( { 'hstart'         =>  $feature->{'start'},
          		'hend'           =>  $feature->{'end'},
          		'hstrand'        =>  $feature->{'strand'},
          		'adaptor'       =>  $self,
          		'slice'         =>  $slice,
          		'repeat_consensus' => $rc,
          		'dbID'          =>  $feature->{'element'} });

  	}

  	return \@a_repeats;
}

1;