=head1 NAME

Bio::Narcisse::SequenceAdaptor - produce sequence strings from locations

=head1 SYNOPSIS

$seq_adptr = $database_adaptor->get_SequenceAdaptor();
$dna = ${$seq_adptr->fetch_by_Slice_start_end_strand($slice, 1, 1000, -1);}

=head1 DESCRIPTION

An adaptor for the retrieval of DNA sequence from Narcisses files.

=cut

package Bio::Narcisse::SequenceAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::EnsEMBL::DBSQL::BaseAdaptor;
use Bio::Narcisse::Utils::Exception qw(throw);
use Bio::EnsEMBL::Utils::Sequence  qw(reverse_comp);
use Bio::EnsEMBL::Utils::Cache;


use base qw( Bio::Narcisse::BaseAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_SequenceAdaptor();
  Function	: Creates a new SequenceAdaptor object.
  Returns	: Bio::Narcisse::SequenceAdaptor
  Args		: None

=cut


=head2 fasta_path

  Title		: fasta_path
  Usage    	: my $fasta_path = $sequence_adaptor->fasta_path;
  Function	: Getter/Setter for the path to access fasta files containing chromosomes sequences
  Returns	: Path to Fasta files
  Args		: (optional) Fasta path

=cut
sub fasta_path{

	my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_fasta_path} = $arg );
  $self->{_fasta_path};

}


=head2 fetch_by_Slice_start_end_strand

  Title		: fetch_by_Slice_start_end_strand
  Usage		: $dna = $seq_adptr->fetch_by_Slice_start_end_strand($slice, 1, 1000, -1);
  Function	: retrieves from fasta files the sequence for this slice
  Returns 	: string
  Args		: 	- Bio::EnsEMBL::Slice slice
               The slice from which you want the sequence
            	- (optional) int startBasePair
               The start base pair relative to the start of the slice. Default 1
            	- (optional) int endBasePair
               The end base pair. Default = the length of the slice
            	- (optional) int strand (1, -1). Default = 1

=cut

sub fetch_by_Slice_start_end_strand
{
	my ( $self, $slice, $start, $end, $strand ) = @_;

	if(!ref($slice) || !$slice->isa("Bio::EnsEMBL::Slice"))
	{
		throw("Slice argument is required.");
	}

	if( !defined $end ) # appel Slice->seq
	{
		$start = $slice->start();
		$end   = $slice->end();
	}
	elsif( !defined $start )
	{
		my @a_ = caller();
		warn "cas non identifie, merci de noter un commentaire dans le code\n" . Dumper \@a_;
		$start = 1;
	}
	else # appel depuis Slice->subseq
	{ 1; }
# 	$start = 1	unless defined $start;
# 	$end   = $slice->end() - $slice->start() + 1	unless defined $end;

	throw("Start must be less than or equal to end.") if($start > $end);

	$strand ||= 1;

# warn Dumper { start => $start, end => $end, strand => $strand };
	my $chrom   = $slice->seq_region_name;
	my $o_dba   = $self->db();
	my $chrom_file = $o_dba->conf()->GetChromFastaFile( chrom => $chrom );

	my $seq = ${Bio::Narcisse::Utils::ScriptUtils::ExtractSeqInFasta(["$start..$end"], $chrom_file)};

# print STDERR "start: $start, end: $end, strand : $strand\n$chrom_file\n$seq\n";
	reverse_comp(\$seq) if($strand == -1);

	return \$seq;
}

1;
