=head1 NAME

Bio::Narcisse::SliceAdaptor - A file aware adaptor responsible for
the creation of Slice objects.

=head1 SYNOPSIS

  use Bio::Narcisse::DBAdaptor;
  use Bio::EnsEMBL::Utils::Slice qw(split_Slices);

  $db = Bio::Narcisse::DBAdaptor->new(...);

  $slice_adaptor = $db->get_SliceAdaptor();

  # get a slice on the entire chromosome X
  $chr_slice = $slice_adaptor->fetch_by_region('chromosome','X');

=head1 DESCRIPTION

This module is responsible for fetching Slices representing genomic regions
from Narcisse files.  A Details on how slices can be used are in the
Bio::EnsEMBL::Slice module.

=head1 METHODS

=cut



package Bio::Narcisse::SliceAdaptor;

use strict;
use warnings;

use Data::Dumper;
use File::Spec;

use Bio::EnsEMBL::DBSQL::SliceAdaptor;
use Bio::EnsEMBL::Slice;
use Bio::EnsEMBL::Mapper;
use Bio::EnsEMBL::CoordSystem;
use Bio::Narcisse::Utils::ScriptUtils qw( get_karyotype );
use Bio::Narcisse::Utils::NarcisseDataReader;

# use Bio::EnsEMBL::Utils::Exception qw(throw warning);
use Bio::Narcisse::Utils::Exception qw(throw warning);
use Bio::Narcisse::BaseAdaptor;

use base qw( Bio::EnsEMBL::DBSQL::SliceAdaptor Bio::Narcisse::BaseAdaptor );

#----------------------------------------------------------------------------
#                      Constants
#----------------------------------------------------------------------------
use constant KARYOTYPE_FILE_NAME => "karyotype";


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_SliceAdaptor();
  Function	: Creates a new SliceAdaptor object.
  Returns	: Bio::Narcisse::SliceAdaptor
  Args		: None

=cut
sub new
{
	my $caller = shift;

	my $class = ref($caller) || $caller;
	my $self = $class->SUPER::new(@_);

	my $o_spe_obs = $self->db()->conf();
	my $filepath  = $o_spe_obs->GetKaryotypeFile();

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr->Prepare( -from   => $filepath );

	my $rh_ = $o_ndr->FetchallHashref( 'chr' );
	##### removed by MANU to make use of the NarcisseDataReader's cache $o_ndr->Finish();

	$self->karyotype( $rh_ );

	return $self;
}


=head2 karyotype

  Title		: karyotype
  Usage    	: my $karyotype = $sequence_adaptor->karyotype;
  Function	: Getter/Setter for information about karyotype of each species
  			  in Narcisse (list of chromosomes, length of each chromosome etc)
  Returns	: a hash looks like
            	# $chr_name => { chrName => $chrName, length => $length, centro => $centro }
  Args		: (optional) hash holding information about chromosomes of each species.

=cut
sub karyotype {
  my ($self, $arg ) = @_;
  ( defined $arg ) &&
    ( $self->{_karyotype} = $arg );

  $self->{_karyotype};
}



=head2 Function fetch_by_region

 Usage		: $slice = $slice_adaptor->fetch_by_region('chromosome', 'X');
 Function	: Retrieves a slice on the requested region.  At a minimum the
  			  name the name of the seq_region to fetch must be provided.
			  Narcisse only deals with chromosome regions for now. $coord_system_name must be
			  equals to 'chromosome'

 Returns	: Bio::EnsEMBL::Slice or undef
 Args		: - string $coord_system_name. The name of the coordinate system of the slice
			  - string $seq_region_name. The name of the sequence region that the slice
			    will be created on.
			  - int $start (optional, default = 1). The start of the slice on the sequence region
			  - int $end (optional, default = seq_region length). The end of the slice on the sequence region
 			  - int $strand (optional, default = 1). The orientation of the slice on the sequence region
			  - string $version (not used for now in Narcisse context)

=cut
sub fetch_by_region {
  my ($self, $coord_system_name, $seq_region_name,
      $start, $end, $strand, $version) = @_;

  $start  = 1 if (!defined($start) or $start eq '' );
  $strand = 1 if (!defined($strand));

  throw('seq_region_name argument is required') if (!defined($seq_region_name) );
  $coord_system_name eq 'chromosome' || throw("Coordinate system must be \"chromosome\" with Narcisse");

  &throw("Unknown slice: $coord_system_name $seq_region_name") if (! defined $self->karyotype->{$seq_region_name});

  # Si la fin n'est pas renseignée, on charge par défaut la fin du chromosome
  $end = $self->karyotype->{$seq_region_name}->{length} if (!defined ($end) or $end eq '' );

  my $cs = $self->db->default_CoordSystem;
	my $o_slice = Bio::EnsEMBL::Slice->new(  -COORD_SYSTEM => $cs,
										-SEQ_REGION_NAME => $seq_region_name,
										-START => $start,
										-END => $end,
										-STRAND => $strand,
										-ADAPTOR => $self
									);
  return $o_slice;
}


=head2 fetch_all

  Title		: fetch_all
  Usage		: @chromos = @{$slice_adaptor->fetch_all('chromosome')};

  Function	: Retrieves slices of all seq_regions for a given coordinate
               system.  This is analagous to the methods fetch_all which were
               formerly on the ChromosomeAdaptor, RawContigAdaptor and
               CloneAdaptor classes.  Slices fetched span the entire
               seq_regions and are on the forward strand.
               If the coordinate system with the provided name and version
               does not exist an empty list is returned.
               If the coordinate system name provided is 'toplevel', all
               non-redundant toplevel slices are returned.

               Retrieved slices can be broken into smaller slices using the
               Bio::EnsEMBL::Utils::Slice module.

  Returns	: listref of Bio::EnsEMBL::Slices
  Args		: - string $coord_system_name. The name of the coordinate system to retrieve slices of.
   			  - string $coord_system_version (not use yet in Narcisse)

=cut

sub fetch_all {
  my $self = shift;
  my $cs_name = shift;
  my $cs_version = shift || '';

  throw "Narcisse only supports \"chromosome\" coordinate system" if ($cs_name ne "chromosome");

  my @a_slices;

  my $cs = $self->db->default_CoordSystem;

  my @chromosomes = keys(%{$self->karyotype});
  foreach my $chr (@chromosomes){
  		push @a_slices, Bio::EnsEMBL::Slice->new(  -COORD_SYSTEM => $cs,
  											  -SEQ_REGION_NAME => $chr,
  											  -START => 1,
  											  -END => $self->karyotype->{$chr}->{length},
  											  -ADAPTOR => $self
  								  			);
  }

  return \@a_slices;

}

=head2 fetch_by_Feature

  Title		: fetch_by_feature
  Example	: $slice = $slice_adaptor->fetch_by_Feature($feat, 100);
  Function	: Retrieves a slice around a specific feature.  All this really
               does is return a resized version of the slice that the feature
               is already on. Note that slices returned from this method
               are always on the forward strand of the seq_region regardless of
               the strandedness of the feature passed in.
  Returns	: Bio::EnsEMBL::Slice
  Args		: - Bio::EnsEMBL::Feature $feat. The feature to fetch the slice around
              - int size (optional)
               The desired number of flanking basepairs around the feature.
               The size may also be provided as a percentage of the feature
               size such as 200% or 80.5%.

=cut
sub fetch_by_Feature{

	my ($self, $feature, $size) = @_;

	$size ||= 0;

	if(!ref($feature) || !$feature->isa('Bio::EnsEMBL::Feature')) {
    	throw('Feature argument expected.');
  	}

  	my $slice = $feature->slice();
  	if(!$slice || !$slice->isa('Bio::EnsEMBL::Slice')) {
    	throw('Feature must be attached to a valid slice.');
  	}

  	my $fstart = $feature->start();
 	my $fend   = $feature->end();
  	if(!defined($fstart) || !defined($fend)) {
    	throw('Feature must have defined start and end.');
  	}

  	my $slice_strand = $slice->strand();

  	## Size may be stored as a %age of the length of the feature
  	## Size = 100% gives no context
  	## Size = 200% gives context - 50% the size of the feature either side of
  	## feature

  	$size = int( ($1-100)/200 * ($fend-$fstart+1) ) if( $size =~/([\d+\.]+)%/ );

  	#TODO réflechir à une méthode propre pour la récupération de la table des codons

  	# TODO - Verifier que l'on recupere bien le bon slice (pas de contexte)

	#return a new slice covering the region of the feature
  	my $S = Bio::EnsEMBL::Slice->new (	-seq_region_name   => $slice->seq_region_name,
     									-seq_region_length => $slice->seq_region_length,
     									-coord_system      => $slice->coord_system,
     									-start             => $fstart - $size,
     									-end               => $fend + $size,
     									-strand            => 1,
     									-adaptor           => $self);

  	$S->{'_raw_feature_strand'}  = $feature->strand * $slice_strand if $feature->can('strand');
  	return $S;

}

1;
