=head1 NAME

Bio::Narcisse::TranscriptAdaptor - An adaptor allowing the storage and retrieval of Transcripts

=head1 SYNOPSIS

  $db = Bio::Narcisse::DBAdaptor->new(...);
  $slice_adaptor = $db->get_SliceAdaptor();

  $transcript_adaptor = $db->get_TranscriptAdaptor();

  $transcript = $transcript_adaptor->fetch_by_dbID('LOC532606#1');

  $transcript = $transcript_adaptor->fetch_by_stable_id('LOC532606#1');

  $slice = $slice_adaptor->fetch_by_region('chromosome', '3', 1, 1000000);
  @transcripts = @{$transcript_adaptor->fetch_all_by_Slice($slice)};

=head1 DESCRIPTION

This adaptor provides a means to retrieve and store information related to
Transcripts. Primarily this involves the retrieval
Bio::EnsEMBL::Transcript objects from Narcisse files.

See Bio::EnsEMBL::Transcript for details of the Transcript class.

=cut


package Bio::Narcisse::TranscriptAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::EnsEMBL::Gene;
use Bio::EnsEMBL::Exon;
use Bio::EnsEMBL::Transcript;
use Bio::EnsEMBL::Translation;
use Bio::EnsEMBL::Utils::Exception qw( throw warning );
use LipmError::ParameterException;

use Bio::Narcisse::Transcript;

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_TranscriptAdaptor();
  Function	: Creates a new TranscriptAdaptor object.
  Returns	: Bio::Narcisse::TranscriptAdaptor
  Args		: None

=cut

sub _GetFeatureKey
{
	my $self = shift;

	return 'TRANSCRIPT';
}


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: my $tr = $tr_adaptor->fetch_by_dbID('s1g4t1');
  Function	: Retrieves a transcript via its db id.
  Returns 	: Bio::EnsEMBL::Transcript
  Args		: - String $id. The stable id of the transcript to retrieve

=cut
sub fetch_by_dbID
{
	my $self  = shift;
	my( $id ) = @_;

	my $o_dba    = $self->db();
	my $species  = $o_dba->species();
	my @a_files  = $o_dba->conf()->GetTranscriptFiles();
	my $o_transcript = $self->fetch_feature( 'element', $id, \@a_files );

	return $o_transcript;
}


=head2 fetch_by_stable_id


  Usage    : my $tr = $tr_adaptor->fetch_by_stable_id('ENSG00000146556#1');
  Function : Retrieves a transcript via its stable id,
             which is its 'locus tag' eventually followed by #num_transcript
  Returns  : Bio::EnsEMBL::Transcript
  Args     : - String $stable_id. The stable id of the transcript to retrieve

=cut

sub fetch_by_stable_id
{
	my $self  = shift;
	my( $stable_id ) = @_;

	my $o_dba    = $self->db();
	my $species  = $o_dba->species();
	my @a_files  = $o_dba->conf()->GetTranscriptFiles();
	my $o_transcript = $self->fetch_feature( 'locus_tag', $stable_id, \@a_files );

	return $o_transcript;
}


=head2 fetch_all_by_Slice

  Title		: fetch_all_by_Slice
  Usage		: my @transcripts = @{ $tr_adaptor->fetch_all_by_Slice($slice) };
  Function	: Overrides superclass method to optionally load exons
               immediately rather than lazy-loading them later.
  Returns	: Listref of Bio::EnsEMBL::Transcript objects
  Args		: - Bio::EnsEMBL::Slice $slice. The slice to fetch transcripts on
  			  - (optional) Boolean $load_exons.  If true, exons will be loaded immediately rather than
               lazy loaded later
  Arg [3]    : (Not used yet in Narcisse)

=cut

sub fetch_all_by_Slice
{
	my $self = shift;
	my( $slice, $load_exons, $logic_name ) = @_;

	my $ra_transcripts = $self->SUPER::fetch_all_by_Slice( $slice );
	

	if( defined( $load_exons ) && ( 1 == $load_exons ) )
	{
		# Else exons have to be loaded and added to their respective transcript.
		my $exon_adaptor = $self->db->get_ExonAdaptor();
		foreach my $transcript (@$ra_transcripts){
			my $ra_exons = $exon_adaptor->fetch_all_by_Transcript($transcript);

			# we add exons to the current transcript
			foreach my $exon (@$ra_exons){
				$transcript->add_Exon($exon);
			}

		}
	}
	return $ra_transcripts;

}


=head2 fetch_all_by_Gene

  Title		: fetch_all_by_Gene
  Usage		: my @transcripts = $tr_adaptor->fetch_all_by_Gene($gene);
  Function	: Retrieves Transcript objects for given gene. Puts Genes slice
              in each Transcript.
  Returns	: Listref of Bio::EnsEMBL::Transcript objects
  Args		: Bio::EnsEMBL::Gene $gene. The gene to fetch transcripts of

=cut
sub fetch_all_by_Gene {
	my $self = shift;
  	my $gene = shift;
  return undef unless( $gene );	# ajout Seb L

	my $gslice = $gene->slice();

	if (!$gslice)
	{
		throw("Gene must have attached slice to retrieve transcripts.");
	}

	my $o_dba = $self->db();
	my $slice;

	if ($gene->start() < 1 || $gene->end() > $gslice->length() )
	{
		$slice = $o_dba->get_SliceAdaptor()->fetch_by_Feature( $gene );
	}
	else
	{	$slice = $gslice;	}

	my $chrom   = $slice->seq_region_name;

	# $o_dba->conf() est un object SpeciesObserver
	my $chromosome_file = $o_dba->conf()->GetTranscriptFiles( -chrom => $chrom );
  	my $ra_transcripts = $self->get_features_in_region($chromosome_file, $slice->start, $slice->end, 'parent', $gene->dbID);

	my $rao_ = $self->_objs_from_hashes($ra_transcripts, $slice);
	return $rao_;
}


=head2 fetch_by_translation_id

  Title		: fetch_by_translation_id
  Usage		: my $tr = $tr_adaptor->fetch_by_translation_id('ENSP00000348979');
  Function	: Retrieves a Transcript object using the db identifier of
               its translation.
  Args		: String $stable_id. The db identifier of the translation of the transcript to retrieve
  TODO  : devrait etre reserve a id = element_id

=cut

sub fetch_by_translation_id
{
	my $self  = shift;
	my( $id ) = @_;

	if( !defined $id )
	{
		my $str = join ', ', caller();
		throw LipmError( -text => "id is not defined\n" . $str );
	}

	my $o_dba    = $self->db();
	my $species  = $o_dba->species();
	my @a_files  = $o_dba->conf()->GetTranscriptFiles();
	my $o_transcript = $self->fetch_feature( 'protein_id', $id, \@a_files );

	return $o_transcript;

}


=head2 fetch_by_translation_stable_id

  Title		: fetch_by_translation_stable_id
  Usage		: my $tr = $tr_adaptor->fetch_by_translation_stable_id('ENSP00000348979');
  Function	: Retrieves a Transcript object using the stable identifier of
               its translation. In Narcisse, we use the same identifier for stable_id
               and db_id for a translation. So this method just calls fetch_by_translation_id.
  Returns	: Bio::EnsEMBL::Transcript
  Args		: String $stable_id. The stable identifier of the translation of the transcript to retrieve

  TODO : devrait etre reserve a id = num. accession

=cut

sub fetch_by_translation_stable_id{
	my ($self,$stable_id) = @_;

	# pour le moment, correspond exactement à fetch_by_translation_id, puisque nos traductions n'ont qu'un seul identifiant.
	return $self->fetch_by_translation_id($stable_id);

}


=head2 fetch_by_symbol

  Title		: fetch_by_symbol
  Usage		: $o_tr = $tr_adaptor->fetch_by_symbol( $symbol );
  Function	: Retrieves a Transcript from the files via its symbol
  Returns	: Bio::EnsEMBL::Transcript
  Args		: string $symbol, the $symbol of the transcript to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_symbol
{
	my $self = shift;
	my( $symbol ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetTranscriptFiles();
	my $o_gene = $self->fetch_feature( 'symbol', $symbol, \@a_files );

	return $o_gene;
}


=head2 fetch_by_accession

  Title		: fetch_by_accession
  Usage		: $o_tr = $tr_adaptor->fetch_by_accession( $accession );
  Function	: Retrieves a Transcript from the files via its accession
  Returns	: Bio::EnsEMBL::Transcript
  Args		: string $accession, the $accession of the transcript to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_accession
{
	my $self = shift;
	my( $accession ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetTranscriptFiles();
	my $o_gene = $self->fetch_feature( 'accession', $accession, \@a_files );

	return $o_gene;
}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::Transcript objects
  Returns	: listref Bio::EnsEMBL::Transcript
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut
sub _objs_from_hashes {

	my ($self, $tab_of_hash, $slice) = @_;
	my @a_transcripts;
	foreach my $feature (@$tab_of_hash){
# use Data::Dumper;
# print Dumper [ $feature ];

	my @a_tmp = split( /#/, $feature->{'locus_tag'} );
	my $locus_tag = $a_tmp[ 0 ];

# 		my $transcript = Bio::EnsEMBL::Transcript->new(
		my $transcript = Bio::Narcisse::Transcript->new(
							'-start'         	=>  $feature->{'start'},
							'-end'           	=>  $feature->{'end'},
							'-strand'        	=>  $feature->{'strand'},
							'-adaptor'       	=>  $self,
							'-slice'         	=>  $slice,
							'-dbID'          	=>  $feature->{'element'},
							'-stable_id'     	=>  $locus_tag,
							'-created_date'  	=>  undef,
							'-modified_date' 	=>  undef,
							'-external_name' 	=>  $feature->{'protein_id'},
							'-external_db'   	=>  "genbank prot",
							'-display_xref' 	=> $feature->{'accession'},
							'nc_consistency' 	=> $feature->{ consistency }
						);

		#$transcript->cdna_coding_start($feature->{'start_translation'}+1);
		#$transcript->cdna_coding_end($feature->{'stop_translation'}+1);
		#Modified by Thomas 28/11/2007
		$transcript->cdna_coding_start($feature->{'start_translation'});
		$transcript->cdna_coding_end($feature->{'stop_translation'});


# 		$transcript->{ nc_consistency } = $feature->{'consistency'};
		push @a_transcripts, $transcript

	}
	return \@a_transcripts;
}

1;
