=head1 NAME

Bio::Narcisse::TranslationAdaptor - Provides a means to fetch
Translation objects from Narcisse files.

=head1 DESCRIPTION

Bio::Narcisse::TranslationAdaptor - Provides a means to fetch
Translation objects from Narcisse files.

Translation objects only truly make sense in the context of their transcripts
so the recommended means to retrieve Translations is by retrieving the
Transcript object first, and then fetching the Translation.

=head1 SYNOPSIS

  my $db = Bio::Narcisse::DBAdaptor->new(...);

  my $transcript_adaptor = $db->get_TranscriptAdaptor();
  my $translation_adaptor = $db->get_TranslationAdaptor();
  my $transcript = $transcript_adaptor->fetch_by_dbID('LOC532606#5');
  my $translation = $translation_adaptor->fetch_by_Transcript($transcript);

  print "Translation Start Site: " .
        $translation->start_Exon()->stable_id(), " ", $translation->start();
  print "Translation Stop: " .
        $translation->end_Exon()->stable_id(), " ", $translation->end();

=head1 METHODS

=cut


package Bio::Narcisse::TranslationAdaptor;

use strict;

use Data::Dumper;

use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::EnsEMBL::Translation;
use Bio::Narcisse::Utils::Exception qw( throw warning );

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_TranslationAdaptor();
  Function	: Creates a new TranslationAdaptor object.
  Returns	: Bio::Narcisse::TranslationAdaptor
  Args		: None

=cut

sub _GetFeatureKey
{
	my $self = shift;

	return 'PROT';
}


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $translation = $translation_adaptor->fetch_by_dbID(ENSP00000317668);
  Function	: This fetches a Translation object via its internal id.
  Returns	: Bio::EnsEMBL::Translation or undef if the translation is not
               found.
  Args		: String $dbID. The internal identifier of the Translation to obtain

=cut

sub fetch_by_dbID {
   my ($self,$dbID, $transcript) = @_;

   if($transcript) {
     deprecate("Use of fetch_by_dbID with a Transcript argument is deprecated."
               . "Use fetch_by_Transcript instead." );
   }

   if(!$dbID) {
     throw("dbID argument is required");
   }

   my $transcript_adaptor = $self->db()->get_TranscriptAdaptor();
   $transcript = $transcript_adaptor->fetch_by_translation_id($dbID);

   return undef if(!$transcript);

   return $self->fetch_by_Transcript($transcript);
}

=head2 fetch_by_stable_id

  Title		: fetch_by_stable_id
  Usage		: $translation = $translation_adaptor->fetch_by_stable_id(ENSP00000317668);
  Function	: This fetches a Translation object via its stable id. As in Narcisse, db id and
  			  stable_id are the same, this method only calls fetch_by_dbID
  Returns	: Bio::EnsEMBL::Translation or undef if the translation is not
               found.
  Args		: String $stable_id. The internal identifier of the Translation to obtain

=cut

sub fetch_by_stable_id {
   my ($self,$stable_id, $transcript) = @_;

   return $self->fetch_by_dbID($stable_id, $transcript);
}

=head2 fetch_by_Transcript

  Title		: fetch_by_Transcript
  Usage		: $tl = $translation_adaptor->fetch_by_Transcript($transcript);
  Function	: Retrieves a Translation via its associated transcript.
               If the Translation is not found, undef is returned.
  Returns	: Bio::EnsEMBL::Translation or undef
  Args		: Bio::EnsEMBL::Transcript $transcript


=cut

sub fetch_by_Transcript
{
	my $self = shift;
	my( $o_transcript ) = @_;

	return undef unless( $o_transcript );	# ajout Seb L

	#fichier du transcript
	my $o_dba = $self->db();
	my $species = $o_dba->species();
	my $chrom   = $o_transcript->slice()->seq_region_name;
	my $chromosome_file = $o_dba->conf()->GetTranscriptFiles( -chrom => $chrom );

  # on ne se sert plus que de protein_id
  # recuperer partiellement l'enregistrement ralentit la recherche et n'est plus implemente.
  # les colonnes a recuperer sont ignorees.
  my $rh_prot_infos = $self->fetch_columns( $chromosome_file, 'element', $o_transcript->dbID, 'protein_id', 'start_translation', 'stop_translation', 'consistency' );

# 	my $cds_start = $rh_prot_infos->{'start_translation'};
# 	my $cds_end = $rh_prot_infos->{'stop_translation'};
	my ( $start_exon, $end_exon );
	my $ra_exons  = $o_transcript->get_all_Exons();

	my $cds_start = $o_transcript->cdna_coding_start;
	my $cds_end   = $o_transcript->cdna_coding_end;
	if( $cds_end < $cds_start or $cds_start < 0 or $cds_end < 0 )
	{ # devrait etre une erreur je pense
# 		warn "aaaaaaaaaaaaaaaaaaaaaargh !";
		$start_exon = $$ra_exons[0];
		$end_exon = $$ra_exons[-1];
	}
	else
	{
		#In Narcisse, start_translation is the position of the first base of the start codon
		# relatively to the transcript.
		# In ensembl, it is relative to the first coding exon.
		#So, I remove the length of all 5' UTR exons.
		
		my( $start_found, $end_found ) = (0,0);

		# je suppose que les exons arrivent dans le bon ordre !
		foreach my $exon ( @$ra_exons )
		{
			my $ex_len = $exon->length();

			# start
			if( 0 == $start_found )
			{
				if( (0 <= $cds_start) && ($cds_start <= $ex_len) )
				{	$start_exon = $exon; $start_found = 1;	}
				else
				{	$cds_start -= $ex_len;	}
			}

			# end
			if( 0 == $end_found )
			{
				if( (0 <= $cds_end) && ($cds_end <= $ex_len) )
				{	$end_exon = $exon; $end_found = 1;	}
				else
				{	$cds_end -= $ex_len;	}
			}
		}
	}

	my $translation = Bio::EnsEMBL::Translation->new
	(
# 		-dbID => $$rh_prot_infos{'element'},
		-dbID => $$rh_prot_infos{'protein_id'},
		-stable_id => $$rh_prot_infos{'protein_id'},
		-adaptor => $self,
		-seq_start  => $cds_start,
		-seq_end    => $cds_end,
		-start_exon => $start_exon,	# doit contenir le start
		-end_exon   => $end_exon # doit contenir le stop

	);

  return $translation;
}

# Ajout pour nos besoins
sub fetch_longest_by_Gene {
  my ( $self, $gene ) = @_;

  return undef unless( $gene );	# ajout Seb L

  my @translations = @{$self->fetch_all_by_Transcript_list($gene->get_all_Transcripts)};

  my $longest_translation = shift (@translations);
  foreach my $translation (@translations){
  	if (length($translation->seq) > length($longest_translation->seq)){
  		$longest_translation = $translation;
  	}
  }

  return $longest_translation;
}


=head2 fetch_all_by_Transcript_list

  Title		: fetch_all_by_Transcript_list
  Usage		: @translations = @{$tla->fetch_all_by_Transcript_list([$t1,$t2]);
  Function	: Fetches all translations associated with the list of transcripts
               passed to this method.  The passed transcripts will also have
               their translation set by this method.
  Returns	: Reference to list of Bio::EnsEMBL::Translations
  Args		: reference to list of Bio::EnsEMBL::Transcripts $transcripts
               The list of $transcripts to obtain Translation object for.

=cut
sub fetch_all_by_Transcript_list {
  my ($self,$ra_transcripts) = @_;

  if(!defined($ra_transcripts) || ref($ra_transcripts) ne 'ARRAY') {
    throw("reference to list of Transcripts argument is required");
  }

  return [] if(!@$ra_transcripts);

  my @a_out;

  foreach my $transcript (@$ra_transcripts){

  	my $current_translation = $self->fetch_by_Transcript($transcript);
  	$transcript->translation($current_translation);
  	push @a_out, $current_translation;

  }

  return \@a_out;
}



#sub fetch_all_by_Slice{
#
#	my ($self, $slice)  = @_;
#
#  	return $self->SUPER::fetch_all_by_Slice($slice);
#
##  	my $start = $slice->start;
##  	my $end = $slice->end;
##  	my $chromosome_file = $self->features_path.$self->db->conf->Get(&MASK_FILE_CHROMOSOME).$slice->seq_region_name.".".$self->files_extension;
##
##	my @a_features = @{$self->get_features_in_region($chromosome_file, $start, $end)};
##
##	return $self->_objs_from_hashes(\@a_features, $slice);
#
#}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::Translation objects
  Returns	: listref Bio::EnsEMBL::Translation
  Args		: - $tab_of_hash - array ref holding features as hashes
  			  - $slice - Bio::EnsEMBL::Slice

=cut
sub _objs_from_hashes{

	my ($self, $tab_of_hash, $slice) = @_;

	my @a_translations;
	foreach my $feature (@$tab_of_hash){
		push @a_translations, Bio::EnsEMBL::Translation->new
   						(
					     -dbID => $feature->{'element'},
					     -stable_id => $feature->{'protein_id'},
					     -adaptor => $self,
					     -seq_start => $feature->{'start'},
					     -seq_end => $feature->{'end'},
					     -start_exon => undef,
					     -end_exon => undef
   						);
	}
	return \@a_translations;
}


1;
