#
# 	sebastien.letort@toulouse.inra.fr
#	Created: July 31, 2008
#   revised by emmanuel.courcelle@toulouse.inra.fr - Oct 2008
#   modified by thomas.faraut@toulouse.inra.fr - march 2009
#	$Id: NarcisseDataReader.pm 2115 2008-10-14 07:24:38Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

	NarcisseDataReader - Class to read data file of the Narcisse project

=head1  SYNOPSIS

	# code

=head1 DESCRIPTION

	This class is used to access the datafiles of the Narcisse project.
	1- all read file are maintained in a cache memory
	2- results are stored into array and hash format
	3- This class try to optimized read and access to NarcisseData because DB "can't be optimized" (sic.)

=cut

package Bio::Narcisse::Utils::NarcisseDataReader;

use strict;
use warnings;

use base qw( LipmObject );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;
use IO::File;


use LipmError::IOException;
use LipmError::ParameterException;

use General;

#Testing for the presence of the Inline module, if available ($USE_INLINE_C ==1)
# our NARCISSE_inlineC module will be loaded, and associated methods used
# if not, perl implementation of the same routines will be used
eval "require Inline";
our $USE_INLINE_C = ( $@ ne "" ) ? 0 : 1;
#our $USE_INLINE_C = 0;

#Trick to enable to choose bewteen inlineC implementation or perl implementation at execution time
#according to option set by nc_configure.pl
#eval "require InlineClib::NO_INLINE";
#$USE_INLINE_C = ( $@ ne "" ) ?  1 : 0;

if ($USE_INLINE_C)
{
   eval "require NARCISSE_inlineC";
}
#use NARCISSE_inlineC;

use constant MAX_CHROM_SIZE => 4000000000;

# Column separator
our $SEP = "\t";

# A cache for file content, common to all objects of this class
# The key is the file name (see the switch -from in method Prepare)
# The value is an array of hashes, 1 hash / line

our %H_DATA_CACHE = ();

=head2 method _Init

 Usage        : my $o_x = New NarcisseDataReader();
 Prerequisite : none
 Fonction     : Constructor
 Args         : none
 Attibuts     : The attributes are only protected because they are needed in NarcisseDataReader::ChromosomeFeatures
  		_file       		: filename from which we want data
  		_ra_data    		: ref list on row ref **row ref can be HASH or ARRAY, in this class HASH**
  		_rh_headers 		: keys are column names, values are column index
  		_ra_headers 		: list of columns names
  		_ra_wanted_headers 	: list of column names to retrieve 
  		_ra_sorted_keys 	: when data are retrieved in a hash, we need to give a way to retrieve the order of rows
  		_ra_condition_values : value constraint associated to the _ra_column_names
  		_ra_column_names 	: list of column names which have a constraint
 
  		_storage_mode 		: how each line of the data file is stored: as a HASH or as an ARRAY (default : HASH)
  		_return_type  		: indicates the return type for the data: as an array or as a hash (default : HASH)

=cut
sub _Init
{
	my $self = shift;

	$self->SUPER::_Init( @_ );

	$self->_Set( _storage_mode => 'HASH',
				 _return_type  => 'HASH',
				 _rh_return    => {},
				 _ra_return    => [],
				 _file         => '',
				 _rh_headers   => {},
				 _ra_headers   => [],
				 _ra_data      => undef,
				 _key_name     => 'element',
				 _ra_wanted_headers => [],
				 _ra_sorted_keys    => [],
				 _case_sensitive    => &TRUE
			);

	$self->__InitConditionsToDefault();

	$self->_Set(
				_ra_condition_values => [],
				_b_condition_value  => &FALSE,
				_ra_column_names     => [],

				_use_cache => 1   #We are always using the cache
		);

	return;
}
sub GetHeaders
{
	my $self = shift;
	return $self->_Get('_ra_headers');
}

#---------------------------------
#    Reading the file
#---------------------------------

=head2 Procedure Prepare

 Usage        : $o_obj->Prepare( %h_args );
 Procedure    : intialize attributes (in particular reset _ra_return _rh_return and _ra_sorted_keys)
                (see NarcisseDataReader::prepare, the difference here is that there is no cache and no VALUE condition)
 Args         : NAMED ARGUMENTS:
               	-from   => $file     The file name
               	-where  => $rh_cond  A condition, expressed as a hash:
                    { VALUE => { col1_name => val1, col2_name => val2, ... },
               		  RANGE => { start => x, end => x, start_label => 'q_start', end_label => 'q_end' }
               		}
               	-case_sensitive => $bool, set to &FALSE if you want the API to be insensitive.
               		Only used with -where => { VALUE } condition.
 Returns	  : none
TODO : There is no need to keep here in prepare the RANGE condition

=cut
sub Prepare
{
	my $self = shift;
	my %h_args = (
        -from   => '-from option',	# dummy name, useful for error message generation
        -where  => {},
        -select => [],  
        -force_reading  => 0,
        -case_sensitive => &TRUE,
        @_ );
	
	my $file = $h_args{ -from };

	$self->_Set( _file => $file,
				 _case_sensitive => $h_args{ -case_sensitive } );

	#Clearing the return structures we are preparing the object for new data to retrieve
	$self->_Set( '_ra_return' 		=> [],
	             '_rh_return' 		=> {},
	             '_ra_sorted_keys'  => []
	            );
	$self->__InitConditionsToDefault();

	my $rh_where = $h_args{ -where };

	# VALUE condition
	if( defined( $$rh_where{ VALUE } ) )
	{
		my $rh_conditions = $$rh_where{ VALUE };
		my @a_col_names   = keys %$rh_conditions;
		my @a_col_values  = map( "\Q$_\E", values %$rh_conditions );
		$self->_Set( '_b_condition_value' => &TRUE,
					 _ra_column_names 	  => \@a_col_names,
					 _ra_condition_values => \@a_col_values,
					 _rh_value_condition  =>  $rh_conditions
					 );
		$self->{ _rh_where_condition }{ value } = &TRUE;
    }

	# RANGE condition
	my $rh_cond = $self->_Get( '_rh_range_condition' );
	if ( defined( $$rh_where{ RANGE } ) )# Modify default values according to the arguments
	{
		my $rh_range = $$rh_where{ RANGE };

		# Setting the start and end conditions, start and end have to be non-zero values (if no default values are used)
		$$rh_cond{ start }{ value } = $$rh_range{ start } || 0;
		$$rh_cond{ end   }{ value } = $$rh_range{ end   } || &MAX_CHROM_SIZE;

		$$rh_cond{ start }{ label } = $$rh_range{ start_label } || 'start';
		$$rh_cond{ end   }{ label } = $$rh_range{ end_label   } || 'end';

		$self->{ _rh_where_condition }{ range } = &TRUE;
	}

	# Invalid start-end
	if( $$rh_cond{ start }{ value } > $$rh_cond{ end }{ value } )
	{
		my $msg = "( start < end ) is not respected reading  $file \n";
		$msg   .= "start=$$rh_cond{ start }{ value },";
		$msg   .= " end=$$rh_cond{ end }{ value } using default values instead";
		warn($msg);
		$self->_Trace( $msg, '.', 'ID' );

		#We are using default values instead
		$self->{ _rh_where_condition }{ range } = &FALSE;
		$$rh_cond{ start }{ value } = 0;
		$$rh_cond{ end }{ value }   = &MAX_CHROM_SIZE;
	}

	my $bool_use_cache = $self->_Get( '_use_cache' );
	if ( $bool_use_cache )
	{
		delete $H_DATA_CACHE{ $file }  if ( 1 == $h_args{ -force_reading } );

		if( !defined $H_DATA_CACHE{ $file } )
		{
			$H_DATA_CACHE{ $file } = $self->_ReadAllDataFromFile( $file );
		}

		# Retrieve data from the cache
		$self->_Set( _ra_data    => $H_DATA_CACHE{$file}{ _ra_data },
				 	 _ra_headers => $H_DATA_CACHE{$file}{ _ra_headers },
				 	 _rh_headers => $H_DATA_CACHE{$file}{ _rh_headers } );
	}

	# Selecting specific fields to retrieve
	my $ra_select = $h_args{ -select };
	if( 0 != ( scalar @$ra_select ) )
	{
		$self->_Set( _ra_wanted_headers => $ra_select );
	}
	# Selecting all fields
	else 
	{
		$self->_Set( _ra_wanted_headers => $self->_Get( '_ra_headers' ) );
	}

	return;
}


=head2 Procedure __InitConditionsToDefault

 Usage        : $o_obj->__InitConditionsToDefault();
 Procedure    : Set _rh_value_condition, _rh_range_condition, and _rh_where_condition
                with their default values.
 Args         : none
 Returns 	  : none

=cut

sub __InitConditionsToDefault
{
	my $self = shift;

	my $rh_range_condition = {
			start => { value => 0,               label => 'start' },
			end   => { value => &MAX_CHROM_SIZE, label => 'end'},
		};

	$self->_Set(
			_rh_value_condition => {},
			_rh_range_condition => $rh_range_condition,
			_rh_where_condition => { range => &FALSE, value => &FALSE }
		);

	return;
}


=head2 Procedure _ReadAllDataFromFile

 Usage        : $o_obj->_ReadAllDataFromFile( $file );
 Procedure    : Calls _ReadFileWrapper and retrieves the header fields from the first line
 				Each line is stored in an array
 Args         : The file name, absolute path, as a string
 Returns 	  : a reference on a hash containing the relevant information for subsequent data extraction
                _ra_data    : a reference on the array containing all the data lines
                _ra_headers : a reference on a table containing the header information
                _rh_headers : a reference on a hash containing the column numbers corresponding to the header fields

=cut

sub _ReadAllDataFromFile
{
	my $self    = shift;
	my( $file ) = @_;

	my $ra_data = $self->_ReadFileWrapper( $file );

	my $ra_header = shift @$ra_data;

	#Its not the header its the index line, we skip it
	#warn "KOUKOU " . Dumper ( $ra_header );
	if( $$ra_header[0] =~ /\@/ )
	{
		$ra_header = shift @{$ra_data};
	}

	$self->_RetrieveHeaderFields( $ra_header );

	# Keep track of the three arrays in the cache
	my %h_cache_line = (
	                     _ra_data       => $ra_data,
	                     _ra_headers    => $self->_Get( '_ra_headers' ),
	                     _rh_headers    => $self->_Get( '_rh_headers' )
	                    );

	return  \%h_cache_line;

}# _ReadDataFromFile


=head2 Procedure __ReadFileWrapper

 Usage        : $o_obj->__ReadFileWrapper( $file );
 Procedure    : A wrapper to ReadFile enabling to switch easily between _ReadFile and NARCISSE_inlineC::ReadFile
 Args         : A file name
 Returns 	  : same as __ReadFile
=cut

sub _ReadFileWrapper
{
	my $self 	= shift;
	my $file = shift;
	my $ra_data ;
	
	if ( $USE_INLINE_C )
	{
		$ra_data = &NARCISSE_inlineC::ReadFile( $file );
	}
	else
	{
		$ra_data = $self->_ReadFile( $file );
	}
	
	return $ra_data;
}# _ReadFile_Wrapper

=head2 Procedure _ReadFile

 Usage        : $o_obj->_ReadFile( $file );
 Procedure    : Reads the Narcisse formated file and stores the lines
 				in an array of array ref
 				Each line is transformed into an array using the split function,
 				the returned arrays contiains reference on those line arrays
 Args         : A file name
 Returns 	  : a reference on the data array containing the whole file
=cut
sub _ReadFile
{
	my $self 	= shift;
	my $file = shift;

	my $fh_file = new IO::File( $file );
	if( !defined $fh_file )
	{
		throw LipmError::IOException( $file, 'f', -depth => 1 );
	}
	
	my @a_data = ();
	
	while( my $line=<$fh_file> )
	{
		next unless ($line =~ /\S+/);
		
		chomp($line);

		my @a_row = split( /$SEP/, $line );

		push( @a_data, \@a_row );
	}
	$fh_file->close();

	return \@a_data;
}# _ReadAndFormatData

=head2 Procedure _RetrieveHeaderFields

 Usage        : $o_obj->_RetrieveHeaderFields( $header_line );
 Procedure    : Retrieve the header fields from the first line store them in _ra_headers and _rh_headers
 Args         : The header line as a string
 Returns 	  : none
=cut

sub _RetrieveHeaderFields
{
	my $self = shift;
	my ( $ra_header ) = @_;

	my @a_headers = @$ra_header;
	$a_headers[0]  =~ s/^#//;

	my %h_headers;
	@h_headers{ @a_headers } = (0..$#a_headers);

	$self->_Set( _ra_headers => \@a_headers,
				 _rh_headers => \%h_headers );

	return;
}

#----------------------------------
#    Retrieving data
#---------------------------------

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallHashref( 'key' );
 Function     : Return the data read in a hash, the key is the value found in column 'key'
 Returns      : a reference on the hash containing the data 
 Args         : key_name : header field used as a the hash key

=cut

sub FetchallHashref
{
	my $self = shift;
	my ( $key_name ) = @_;

	my $ra_headers  = $self->_Get( '_ra_headers' );
	$key_name = $$ra_headers[0]	unless defined( $key_name );

	$self->_Set( '_return_type' => 'HASH' );
	$self->_Set( '_key_name' => $key_name );

	$self->__Fetchall( );

	my $rh_data = $self->_Get( '_rh_return' );

	return $rh_data;
}# FetchallHashref

=head2 function FetchallArrayref

 Usage        : $ra_data = $o_xx->FetchallArrayref();
 Function     : Return the data read in an array
 Returns      : a reference on the array containing the data 
 Args         : none

=cut
sub FetchallArrayref
{
	my $self = shift;

	$self->_Set( '_return_type' => 'ARRAY' );
	$self->__Fetchall();

	my $ra_data = $self->_Get( '_ra_return' );

	return $ra_data;
}# FetchallArrayref


=head2 Procedure __Fetchall

 Usage        : $self->__Fetchall
 Procedure    : Retrieves the data and stores it in a structure specified by _return_type
 				Each data line is transformed in a hash with the header fields as keys.
 				When a condition has been specified by prepare, only those lines verifying the condition are returned
                This method is designed for all Narcisse data files that are not of the type chromosome feature
 Args         : none
 Returns      : none

=cut

sub __Fetchall
{
	my $self = shift;

	#input and ouput data structures
	my @a_attr= qw( _ra_data _rh_return _ra_return _return_type );
	my ( $ra_data, $rh_return, $ra_return, $return_type) = $self->_Get( @a_attr );

	#Hash specific values
	my ( $key_name_col, $ra_sorted_keys);
	if ( $return_type eq 'HASH' )# Doesn't need to be specified if return type is an ARRAY
	{
		( $key_name_col, $ra_sorted_keys) = $self->_GetHashSpecificValues();
	}

	#Fields selection, only for hash storage mode
	my @a_wanted_headers = $self->_GetWantedHeaders();
	my @a_wanted_col = $self->_GetWantedColIndex();

	#Condition specific values (if any)
	my( $rh_where_condition, $case_sensitive ) =
						$self->_Get( qw( _rh_where_condition _case_sensitive ) );
	my $b_value_conditions_exist = $rh_where_condition->{ value };
	my $b_range_conditions_exist = $rh_where_condition->{ range };

	my ( $ra_col_names_number, $ra_vals ) =$self->_GetCondSpecificValues();
	my $m =scalar @$ra_col_names_number; #should be equal to @$ra_vals

	#Loop control values stored in scalars for performance issue
	my( $start_col, $end_col, $start_value, $end_value )
									= $self->_GetLoopControlValues();

	#my $start_index = $self->_GetFirstRelevantLineIndex();
	#my $i  = $start_index;
	my $i = -1;
	
	# if $i is still not initialized (other return_types, of first call of FetchLineNext)
	if ( $i == -1 )
	{
		my $start_index = $self->_GetFirstRelevantLineIndex();
		$i  = $start_index;
	}
		
	while( my $ra_row = $ra_data->[$i++] )
	{
		if ( $b_range_conditions_exist )
		{
			next if( $$ra_row[ $end_col ] < $start_value );
			last if ( $end_value < $$ra_row[ $start_col ] );
		}

		if ( $b_value_conditions_exist )
		{
			next if ( 0 == $self->_ValidateValueConditions(
									$ra_row, $ra_col_names_number,
									$ra_vals, $m, $case_sensitive ) );
		}
		my %h_row;
		@h_row{ @a_wanted_headers } = @{$ra_row}[ @a_wanted_col ];

		if ( $return_type eq 'ARRAY' )
		{
			push @{$ra_return}, \%h_row;
		}
		else
		{
			my $key_val = $$ra_row[ $key_name_col ];
			$$rh_return{ $key_val } = \%h_row;
		}
	}

	return;
}# __Fetchall

=head2 Function _GetFirstRelevantLineIndex

 Usage        : $self->__GetFirstRelevantLineIndex( );
 Function     : Returns the index of the first relevant data line according to the range conditions
 Args         : none
=cut
sub _GetFirstRelevantLineIndex
{
	my $self = shift;
	
	my( $ra_data, $rh_headers, $rh_where_condition, $rh_range_condition ) = $self->_Get( qw( _ra_data _rh_headers _rh_where_condition _rh_range_condition ) );
	
	my $n = scalar @$ra_data;
	
	#If there is no data or no range conditions there is no need to dichotomize :-)
	return 0 unless ( $n > 0 && $rh_where_condition->{ range } eq &TRUE);
	
	my $start   = $$rh_range_condition{ start }{ value };
	my $end     = $$rh_range_condition{ end   }{ value };

	my $s_label = $$rh_range_condition{ start }{ label };
	my $start_i = $$rh_headers{ $s_label };

	my $e_label = $$rh_range_condition{ end   }{ label };
	my $end_i   = $$rh_headers{ $e_label };

	#We just dichotomize to the first relevant line
	
	my ( $left, $right ) = ( 0, $n-1 );
	while ( $left != $right )
	{
		my $middle = ($left + $right)/ 2;
		my $floor_middle = floor( $middle );
		if ( $start < $$ra_data[ $floor_middle ][ $end_i ] )
		{
			$right = $floor_middle;
		}
		else
		{
			$left = ceil ( $middle );
		}
	}
	
	return $left;
}

=head2 Procedure _GetCondSpecificValues

 Usage        : $self->_GetCondSpecificValues
 Procedure    : returns the necessary information for testing the VALUE condition specified by the -where argument
                a reference on an array containing the column indexes of the tested fields
                a reference on an array containing the values requested for the aforementioned fields
 Args         : none
 Returns      : two references on arrays 
=cut
sub _GetCondSpecificValues
{
	my $self = shift;
	
	my @a_attr = qw( _ra_column_names _ra_condition_values _rh_headers );
	my ( $ra_col_names, $ra_vals, $rh_headers ) = $self->_Get( @a_attr );
	my @col_names_number = @{ $rh_headers } { @$ra_col_names };
	my $m =scalar @col_names_number;
		
	return ( \@col_names_number, $ra_vals );
}


=head2 Procedure _ValidateValueConditions

 Usage        : $self->_ValidateValueConditions( )
 Procedure    : iterates among all the condition values to check them, returns 1 if at least one is verified
 Args         :
     $ra_row  => a reference on the record (an array)
     $ra_col_names_number => a reference on the column indexes of the tested fields
     $ra_vals =>  a reference on the values that requested for the corresponding fields
     $m       => the number of contitions to be tested
     $case_sensitive => boolean, if false, will use 'i' flag on regexp.
 Returns      : a boolean flag : 0 if the conditino is not fullfilled

=cut

sub _ValidateValueConditions
{
	my $self = shift;
	my( $ra_row, $ra_col_names_number, $ra_vals, $m, $case_sensitive ) = @_;

	my @a_col_names_number = @$ra_col_names_number;

	#Testing the value condition
	my $flag = 0;
	if( &FALSE == $case_sensitive )
	{
		for( my $j=0; $j < $m; ++$j )
		{
			if( $$ra_row[ $a_col_names_number[ $j ] ] =~ /^$$ra_vals[ $j ]$/i )
			{
				++$flag;
				last;   # exit from foreach
			}
		}
	}
	else
	{
		for( my $j=0; $j < $m; ++$j )
		{
			if( $$ra_row[ $a_col_names_number[ $j ] ] =~ /^$$ra_vals[ $j ]$/ )
			{
				++$flag;
				last;   # exit from foreach
			}
		}
	}

	return $flag;
}

=head2 Procedure __GetHashSpecificValues

 Usage        : $self->__GetHashSpecificValues( )
 Procedure    : Constructs the hash specific values              
 Args         : none
 Returns      : $key_name_col => the key_name index
                $ra_sorted_keys => a reference on the array that will store the keys according to the already sorted file order 
=cut

sub _GetHashSpecificValues
{
	my $self = shift;
	
	my @a_attr = qw( _rh_headers _key_name );
	my ( $rh_headers, $key_name ) = $self->_Get( @a_attr );
	my $ra_sorted_keys = $self->_Get( '_ra_sorted_keys' );
					
	return  ( $$rh_headers{ $key_name }, $ra_sorted_keys ) ;
}# _GetHashSpecificValues


=head2 Procedure __GetWantedColIndex

 Usage        : $self->__GetWantedColIndex( )
 Procedure    : returns the colums selected by the -select argument     
 Args         : none
 Returns      : an array of column indexes
=cut

sub _GetWantedColIndex
{
	my $self = shift;
	my @a_attr = qw( _rh_headers  _ra_wanted_headers);
	my ( $rh_headers, $ra_wanted_headers ) = $self->_Get( @a_attr );
	
	my @a_wanted_col = @{$rh_headers}{ @$ra_wanted_headers };
	
	return @a_wanted_col;
}# _GetWantedColIndex


=head2 Procedure __GetWantedHeaders

 Usage        : $self->__GetWantedHeaders( )
 Procedure    : returns the field names of selected colums
 Args         : none
 Returns      : an array of column identifiers
=cut

sub _GetWantedHeaders
{
	my $self = shift;
	
	my @a_attr = qw(  _ra_wanted_headers);
	my ( $ra_wanted_headers ) = $self->_Get( @a_attr );
	
	return @$ra_wanted_headers;
}# _GetWantedHeaders


=head2 Procedure _GetLoopControlValues

 Usage        : my( $start_col, $end_col, $start_value, $end_value )
                                   = $self->_GetLoopControlValues();
 Procedure    : Constructs the loop control values needed to iterate among the records
                These are stored in scalar variables prior the while loop for performance issue
 Args         : none
 Returns      :
	$start_col   => the column index of the start field
	$end_col     => the column index of the end field
	$start_value => the start condition
	$end_value   => the end condition

=cut

sub _GetLoopControlValues
{
	my $self = shift;

	my @a_attr = qw( _rh_headers _rh_where_condition _rh_range_condition );
	my ($rh_headers, $rh_where_condition, $rh_cond ) = $self->_Get( @a_attr );

	return (0,0) unless ( $rh_where_condition->{ range } eq &TRUE );

	my $start_label = $$rh_cond{ start }{ label };
	my $end_label   = $$rh_cond{ end }{ label };

	if( !defined $$rh_headers{ $start_label } )
	{
		my $file = $self->_Get( '_file' );
		my $debug_msg = " 'start' is not a key of the h_headers hash reading $file\n";
		my $msg_simple = " something wrong with the file format ";
		throw LipmError( -text => $debug_msg, -user_msg => $msg_simple );
	}

	my $start_col   = $$rh_headers{ $start_label };
	my $end_col     = $$rh_headers{ $end_label };
	my $end_value   = $$rh_cond{ end }{ value };	
	my $start_value = $$rh_cond{ start }{ value };

	# start and end are not grouped due to retro compatibility
	return ( $start_col, $end_col, $start_value, $end_value );
	
}# _GetLoopControlValues



=head2 Procedure Finish

 Title        : Finish
 Usage        : $o_xx->Finish( none );
 Procedure    : Empty the memory cache
 Args         : none

=cut
sub Finish
{
	my $self = shift;	
	
    delete $H_DATA_CACHE{ $self->_Get( '_file' )  };
	
	return;
}

1;
