#
# 	sebastien.letort@toulouse.inra.fr
#	Created: July 31, 2008
#   revised by emmanuel.courcelle@toulouse.inra.fr - Oct 2008
#   modified by thomas.faraut@toulouse.inra.fr - march 2009
#	$Id: ChromosomeFeatures.pm 2115 2008-10-14 07:24:38Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

	ChromosomeFeatures - Class to read chromosome data file of the Narcisse project

=head1  SYNOPSIS

	my $o_ndr = 	New Bio::Narcisse::Utils::NarcisseDataReader::ChromosomeFeatures( );		
	
	#Default start and en lables are 'start' and 'end'
	my %h_range_param = ( start => 1, start_label => 'q_start', end => 10000000, end_label => 'q_end' );
			
	$o_ndr->Prepare( 
	                 -from   => $file,
	                 -where  => { RANGE => \%h_range_param, VALUE => {  col1_name => val1, col2_name => val2, ... } },
	               );	  
    
    #$rh_return is a reference on a hash with the following keys : rh_return ra_return rh_headers ra_sorted_keys          
	my $rh_return = $o_ndr->FetchallRequestedData( key_name => 'q_prot', return_type => 'HASH' );
	
	The $rh_return is well adapted for subsequent use with NarcisseData
	
	

=head1 DESCRIPTION

	This class is used to access the datafiles of the Narcisse project.
	1- The file is read according to the range condition given in prepare, the index line (starting with @) if used if available
	2- results are returned into array or hash format	

=cut

package Bio::Narcisse::Utils::NarcisseDataReader::ChromosomeFeatures;

use strict;
use warnings;

use base qw( Bio::Narcisse::Utils::NarcisseDataReader );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;
use LipmError::IOException;
use LipmError::ParameterException;

use General;

use constant MAX_CHROM_SIZE => 4000000000;
use constant ARRAY => 0;
use constant HASH  => 1;
use constant SLICE_UNIT =>  100000; # 100kb is the unit slice size 

# Column separator
our $SEP = "\t";

# A cache for file content, common to all objects of this class
# The key is the file name (see the switch -from in method Prepare)
# The value is an array of hashes, 1 hash / line

#our %H_DATA_CACHE = ();

=head2 method _Init

 Usage        : see Bio::Narcisse::Utils::NarcisseDataReader
 Attibuts     : Additional attributes of ChromosomeFeatures
	_rh_range_condition : range constraints on the data to retrieve (the range condition is always set for ChromosomeFeatures)

=cut
sub _Init
{
	my $self = shift;
	my %h_args = ( use_cache => 0,
					@_
					);
		
	$self->SUPER::_Init( @_ );
			
	$self->_Set(  _use_cache => $h_args{ use_cache } ); 	
					
	return $self;
}

#----------------------------------
#          Reading the file
#---------------------------------

=head2 Procedure Prepare

 Usage        : $o_obj->Prepare( %h_args );
 Procedure    : see NarcisseDataReader::Prepare
                                     
=cut
sub Prepare
{
	my $self = shift;
	my %h_args = (
        -from   => '-from option',	# dummy name, useful for error message generation
        -where  => {},
        -select => [],  
        -force_reading => 0,
        @_ );
        
    $h_args{ -select } = []; # Always empty here because we always get the entire line        
        
	$self->SUPER::Prepare( %h_args );
				
	return;
}

=head2 Procedure __ReadelectedDataFromFile

 Usage        : $o_obj->_ReadDataFromFile( s);
 Procedure    : Reads the file header and calls _ReadAndFormatData
 				All the lines are stored in an array
 				The method then stores, the following attributes
 				_ra_return 		: an array of the array refs (only available when the return type is an array)
 				_rh_return 		: a hash with the array refs (only available when return type is a hash)
 				_ra_sorted_keys	: an array with the keys ordered as in the file	
                _ra_headers : a reference on a table containing the header information
                _rh_headers : a reference on a hash containing the column numbers corresponding to the header fields 
 				
 				(There is a possibility to use and Inline::C implementation of the _ReadAndFormatData method)				
 Args         : none
 Returns 	  : none
               
=cut
sub __ReadSelectedDataFromFile
{
	my  $self    = shift;
	
	my ( $file, $return_type, $key_name ) = $self->_Get( qw( _file _return_type _key_name ) ); 

	my $mode = ( $return_type eq 'ARRAY' ) ? &ARRAY : &HASH;
					
	#$rh_return_struct = $self->_ReadAndFormatData( $file, $start_label, $start, $end_label, $end, $mode, $key_name );	
	#Uncomment the following lines, and comment out the previous line, if you want to use the Inline::C implementation of ReadAndFormatData
	
	my $rh_range_cond  = $self->_Get( '_rh_range_condition' );
	my $rh_value_cond  = $self->_Get( '_rh_value_condition' );
	my %h_args = ( 
			   file			=> $file,	
			   start 		=> $$rh_range_cond{ start }{ value },
               end   		=> $$rh_range_cond{ end }{ value },
               start_label 	=> $$rh_range_cond{ start }{ label },
               end_label  	=> $$rh_range_cond{ end }{ label },
               mode			=> $mode,
               keyname 		=> $key_name,
               VALUE 		=> $rh_value_cond
              );
		
	my $rh_return_struct = $self->_ReadAndFormatDataWrapper( \%h_args );
							
	$self->_SetHeader( $$rh_return_struct{ _ra_headers } );
		
	$self->_Set(	%$rh_return_struct,
	                _ra_headers    	=> $self->_Get( '_ra_headers' ),
	                _rh_headers    	=> $self->_Get( '_rh_headers' )
	                    );	
	return ;

=head2 Procedure _SetHeader

 Usage        : $o_obj->__RetrievHeaderFields( $header_line );
 Procedure    : Retrieve the header fields from the first line stroe them in _ra_headers and _rh_headers
 Args         : The header line as a string
 Returns 	  : none
=cut	
sub _SetHeader
{
	my $self = shift;
	my $ra_header = shift;
	
	my @a_headers = @$ra_header;

	my %h_headers;
	@h_headers{ @a_headers } = (0..$#a_headers);
	$self->_Set( __ra_headers => \@a_headers,
					 _rh_headers => \%h_headers );
}	

}# _ReadDataFromFile

=head2 Procedure _ReadAndFormatDataWrapper

 Usage        : $o_obj->_ReadAndFormatDataWrapper(\%h_args );
 Procedure    : A wrapper to _ReadAndFormatData enabling to switch easily between _ReadAndFormatData and NARCISSE_inlineC::_ReadAndFormatData
 Args         : A file name
 Returns 	  : same as _ReadAndFormatData
=cut
sub _ReadAndFormatDataWrapper
{
	my $self 	= shift;
	my $rh_return;
		
	if ( $Bio::Narcisse::Utils::NarcisseDataReader::USE_INLINE_C )
	{
		$rh_return = &NARCISSE_inlineC::ReadAndFormatData( @_ );
	}
	else
	{
		$rh_return = $self->_ReadAndFormatData( @_);
	}
	
	return $rh_return;
}# _ReadFile_Wrapper

=head2 Procedure _ReadAndFormatData

 Usage        : $o_obj->_ReadAndFormatData( \%h_args );
 Procedure    : Reads the Narcisse formated file and stores the lines in an array of array ref
 				Each line is transformed into an array using the split function.	  
 Args         : 
 				file 		: A file name
 				start_label : string - name of the column to use as start posotions -
 				start		: the start of the desired slice
 				end_label	: string - name of the column to use as end posotions -
 				mode		: ARRAY|HASH (0|1) the desired returned type
 				Optional
 				key_name	: the name of the column to use as the returned hash primary key
 Returns 	  : a reference on a hash with the following keys:
 				  _ra_return 		: an array of the array refs
 				  _rh_return 		: a hash with the array refs
 				  _ra_headers		: an array with the field names
 				  _ra_sorted_keys	: an array with the keys ordered as in the file
 
=cut
sub _ReadAndFormatData
{
	my $self = shift;
	my $rh_args = shift;
	my %h_args = ( file			=> undef,
				   start 		=> undef,
				   end			=> undef,
				   start_label	=> 'start',
				   end_label	=> 'end',
				   mode			=> &HASH,
				   keyname		=> "_none",
				   VALUE		=> {},		# This parameter is useless and there only for compatibility reason 
				   							# with the signature of NARCISSE_inlineC::ReadAndFormatData 	
					%$rh_args
				  );
	
	my ( $file, $start, $end, $start_label, $end_label, $mode, $key_name) = @h_args{ qw( file start end start_label end_label mode keyname)};
	my $b_hash_return_type = ( $mode eq &HASH ) ? 1 : 0;
		
	if ( $mode eq &HASH && $key_name eq "_none" )
	{
		die "If hash return type is requested a valid key_name must be provided";
	}	
	
	my $fh_file = new IO::File( $file );
	if( !defined $fh_file )
	{
		throw LipmError::IOException( $file, 'f', -depth => 1 );
	}
	
	my $header = <$fh_file> ;
	chomp($header);
	my $relevant_offset = 0;
	my $data_offset = tell( $fh_file ); #where the actual data start
	if ($header =~ s/^\@//) #We are dealing with an indexed file
	{
		my $index_line = $header;
		$relevant_offset = __GetFilePosFromIndexline($index_line, $start, $end);		
		$header =<$fh_file>;
		chomp($header);	
		$data_offset = tell( $fh_file );	
	} 	
	$header =~ s/^\#//;
	
	my @a_headers = split(/$SEP/,$header);
	
	# Added to enable a correct behavior of _GetCondSpecificValues borrowed from NarcisseDataReader
	$self->	_RetrieveHeaderFields( \@a_headers );
	
	my $key_index = 0;
	if ( $b_hash_return_type )
	{
		$key_index = __GetHeaderIndex( \@a_headers, $key_name);
	}	

	my( $rh_cond, $case_sensitive ) = $self->_Get( qw( _rh_range_condition _case_sensitive ) );
	my $start_value = $start;
	my $end_value   = $end;
	my $start_index = __GetHeaderIndex( \@a_headers, $start_label );
	my $end_index   = __GetHeaderIndex( \@a_headers, $end_label );
	
	#Condition specific values (if any)
	my $b_conditions_exist = $self->_Get( '_b_condition_value' );
	my ( $ra_col_names_number, $ra_vals ) = $self->_GetCondSpecificValues();
	my $m =scalar @$ra_col_names_number; #should be equal to @$ra_vals

	my @a_data = ();
	my %h_data = ();
	my @a_sorted_keys = ();

	if ( $relevant_offset >= 0 ) # $relevant_offset == -1 means we are outside the chromosome range for this file
	{
		seek $fh_file, ($data_offset + $relevant_offset), 0;
		while( my $line=<$fh_file> )
		{
			chomp($line);

			my @a_row = split( /$SEP/, $line );
			next if ( $a_row[ $end_index ] < $start_value );
			last if ( $a_row[ $start_index ] > $end_value );

			if ( $b_conditions_exist )
    		{
    			next if ( 0 == $self->_ValidateValueConditions(
    										\@a_row, $ra_col_names_number,
    										$ra_vals, $m, $case_sensitive ) );
    		}

			if ( $b_hash_return_type )
			{
				my $key_value  = $a_row[ $key_index ];
				push @a_sorted_keys, $key_value;
				$h_data{ $key_value } = \@a_row;
			}
			else
			{
				push( @a_data, \@a_row );
			}
		}
	}
	$fh_file->close();

	my $ra_sorted_keys = \@a_sorted_keys;
	my $rh_data = \%h_data;
	my $ra_data = \@a_data;
	
	my %h_return;
	$h_return{ _ra_return }      = $ra_data ;
	$h_return{ _rh_return }      = $rh_data ;
	$h_return{ _ra_headers }     = \@a_headers ;
	$h_return{ _ra_sorted_keys } = $ra_sorted_keys;

	return \%h_return;

	sub __GetHeaderIndex
	{
		my $ra_header = shift;
		my $field_key = shift;
		
		my @a_headers = @$ra_header;

		my $i =0;
		while ( $i <= $#a_headers && $a_headers[ $i ] ne $field_key ) { $i++; };
		my $key_index = $i;
		die "key name absent from header " if ( $key_index > $#a_headers);

		return $key_index;
	}

	sub __GetFilePosFromIndexline
	{
		my ( $index_line, $start, $end  ) = @_;

		my ( $chrom_start, $chrom_end, $slice_size_base, $nb_pos, $positions) = split (";",$index_line);

		my @indexes = split(":",$positions );
		my $start_offset = $start - $chrom_start;
		
		my $relevant_offset;
		if ( $start > $chrom_end || $end < $chrom_start ) # nothing to read
		{
			return -1;
		}
		elsif ( $start_offset > ($slice_size_base*&SLICE_UNIT) ) # If not we return 0
		{
			my $index = int ( $start_offset / ($slice_size_base*&SLICE_UNIT) ) - 1; 
			$index =  $#indexes if ( $index > $#indexes ); # in case of (we have limited indeed the number of indexing cells to NUM_INDEX_SLICE = 40, change this ?)
			$relevant_offset=$indexes[ $index ];		
		}
		else
		{
			$relevant_offset = 0;
		}

		return $relevant_offset;
	}

}# _ReadAndFormatData

#----------------------------------
#    Retrieving data
#---------------------------------

=head2 function FetchallRequestedData

 Usage        : $rh_data = $o_obj->FetchallHashref('key');
 Function     : Fetch the data according to the specified range condition and return type
 				Calls __ReadDataFromFile which does the job.
 Args         : NAMED ARGUMENTS:
 				return_type => 'HASH'|'ARRAY' 
 				key_name   	=> column used as a the hash key
 Returns      : a reference on the hash containing with the following keys
 				  ra_return 		: an array of the array refs (empty except when the return type is an array)
 				  rh_return 		: a hash with the arraya refs (empty except when return type is a hash)
 				  ra_headers		: an array with the field names
 				  ra_sorted_keys	: an array with the keys ordered as in the file
 

=cut
sub FetchallRequestedData
{
	my $self = shift;
	my %h_args = ( 	
					return_type  => 'ARRAY',
				  	key_name	  => '__none',
				  	@_
				  );	

	my ( $return_type, $key_name ) = @h_args{ qw( return_type key_name )};

	if ($key_name eq '__none' && $return_type eq 'HASH')
	{
		my $msg = " defined when a HASH return type is wanted";
		die with LipmError::ParameterException( 'key_name', 'value', -should_be => $msg);
	}
		
	$self->_Set( '_return_type' => $return_type );		
	$self->_Set( '_key_name' 	=> $key_name );
		
	my $bool_use_cache = $self->_Get( '_use_cache' );
	if ( $bool_use_cache )
	{		
		$self->__Fetchall();			
	}
	else
	{
		$self->__ReadSelectedDataFromFile();	
	}		
			
	my @a_return_attributes      = qw( _rh_return _ra_return _rh_headers _ra_sorted_keys ); 
	my @a_return_attribute_names = qw( rh_data ra_data rh_headers ra_sorted_keys ); 
	my %rh_return;

	@rh_return{ @a_return_attribute_names } = $self->_Get( @a_return_attributes );
	
	return \%rh_return;	
}

=head2 Procedure _Fetchall

 Usage        : $self->__Fetchall
 Procedure    : Retrieves the data and stores according to (1)  _return_type, (2) _storage_mode and 
                (3) optional additional conditions on values 
 Args         : none
 Returns      : none
TODO see wether this method could be replaced by NarcisseDataReader::_Fetchall
=cut
sub __Fetchall
{
	my $self = shift;
				
	#input and ouput data structures
	my @a_attr= qw( _ra_data  _ra_return _rh_return _storage_mode _return_type);
	my ( $ra_data, $ra_return, $rh_return, $storage_mode, $return_type) = $self->_Get( @a_attr );
	
	#Hash specific values
	my ( $key_name_col, $ra_sorted_keys);
	if ( $return_type eq 'HASH' )# Does'nt need to be specified if return type is an ARRAY
	{
		( $key_name_col, $ra_sorted_keys) = $self->_GetHashSpecificValues();			
	} 

	#Condition specific values (if any)
	my( $rh_where_condition, $case_sensitive ) =
						$self->_Get( qw( _rh_where_condition _case_sensitive ) );
	my $b_conditions_exist = $rh_where_condition->{ value };
	my ( $ra_col_names_number, $ra_vals ) =$self->_GetCondSpecificValues();
	my $m =scalar @$ra_col_names_number; #should be equal to @$ra_vals
	
	#Loop control values stored in scalars for performance issue
	my( $start_col, $end_col, $start_value, $end_value )
									= $self->_GetLoopControlValues();

	my $start_index = $self->_GetFirstRelevantLineIndex();
	my $i  = $start_index;
	while( my $ra_row = $ra_data->[$i++] )
	{
		next if( $$ra_row[ $end_col ] < $start_value );
		last if( $end_value < $$ra_row[ $start_col ] );

		if ( $b_conditions_exist )
		{
			next if ( 0 == $self->_ValidateValueConditions(
									$ra_row, $ra_col_names_number,
									$ra_vals, $m, $case_sensitive ) );
		}

		if ( $return_type eq 'ARRAY' )
		{
			push @{$ra_return}, $ra_row;
		}
		else
		{
			my $key_val = $$ra_row[ $key_name_col ];
			push ( @$ra_sorted_keys, $key_val );
			$$rh_return{ $key_val } = $ra_row;
		}
	}

	return;
}# _Fetchall

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallHashref( 'key' );
 Function     : Return the data read in a hash, the key is the value found in column 'key'
 Returns      : a reference on the hash containing the data 
 Args         : key_name : header field used as a the hash key

=cut
sub FetchallHashref
{
	my $self = shift;
	my ( $key_name ) = @_;
				 
	(defined $key_name)
		or die "A key name must be provided for FetchallHashref";

	$self->FetchallRequestedData( return_type => 'HASH', key_name => $key_name );

	my $rh_data = $self->_Get( '_rh_return' );
		
	return $rh_data;
}# FetchallHashref

=head2 function FetchallArrayref

 Usage        : $ra_data = $o_xx->FetchallArrayref();
 Function     : Return the data read in an array
 Returns      : a reference on the array containing the data 
 Args         : none

=cut
sub FetchallArrayref
{
	my $self = shift;
						
	$self->_Set( '_return_type' => 'ARRAY' ); 		   
	$self->FetchallRequestedData(  );

	my $ra_data = $self->_Get( '_ra_return' );
		
	return $ra_data;
}# FetchallArrayref


=head2 Procedure __ValidateValueConditions

 Usage        : $self->__ValidateValueConditions( )
 Procedure    : iterates among all the condition values to check them, returns 1 if at least one is verified
 Args         : 	$ra_row => a reference on the record (an array)
 				 	$ra_col_names_number => a reference on the column indexes of the tested fields
 				 	$ra_vals  =>  a reference on the values that requested for the corresponding fields
 				 	$m	=> the number of contitions to be tested
 Returns      : a boolean flag : 0 if the conditino is not fullfilled 
DEPRECATED
#=cut
sub __ValidateValueConditions
{
	my 	( $ra_row, $ra_col_names_number, $ra_vals, $m ) = @_;
	
	my @a_col_names_number = @$ra_col_names_number;
	
	#Testing the value condition
    my $flag = 0;        
    for( my $j=0; $j < $m; ++$j )
    {			
 		if( $$ra_row[ $a_col_names_number[ $j ] ] =~ /^$$ra_vals[ $j ]/ )
		{
			++$flag;
			last;   # exit from foreach
		}
	}
	
	return $flag;
}

=cut

1;
