=head1 NAME

Bio::Narcisse::VariationAdaptor - Database adaptor for the retrieval and
storage of VariationFeature objects ( which are uniquely SNPs for Narcisse ).

=head1 SYNOPSIS

  use Bio::Narcisse::DBAdaptor;

  $dba = Bio::Narcisse::DBAdaptor->new(...);

  $VariationFeature_adaptor = $dba->get_VariationAdaptor();

  $VariationFeature = $VariationFeature_adaptor->fetch_by_name('rs666');

  $VariationFeature = $VariationFeature_adaptor->fetch_by_stable_id('LOC419094');

  @VariationFeatures = @{$VariationFeature_adaptor->fetch_all_by_external_name('BRCA2')};

  $slice_adaptor = $db->get_SliceAdaptor;
  $slice = $slice_adaptor->fetch_by_region('chromosome', '1', 1, 1000000);
  @VariationFeatures = @{$VariationFeature_adaptor->fetch_all_by_Slice($slice)};

=head1 DESCRIPTION

This is an adaptor for the retrieval and storage of VariationFeature objects.

=cut


package Bio::Narcisse::VariationAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::SliceAdaptor;
use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::Narcisse::DBAdaptor;
use Bio::EnsEMBL::VariationFeature;
use Bio::Narcisse::Utils::Exception qw( throw );

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_VariationAdaptor();
  Function	: Creates a new VariationAdaptor object.
  Returns	: Bio::Narcisse::VariationAdaptor
  Args		: None

=cut


=head2 fetch_all_by_Slice

  Title		: fetch_all_by_Slice
  Function	: Retrieves all VariationFeatures for the Slice
  Returns	: listref Bio::EnsEMBL::VariationFeature on slice
  Args		: Bio::EnsEMBL::Slice $slice
  			  $load_transcripts - flag indicating if the method must load transcripts for each VariationFeature

=cut

sub fetch_all_by_Slice
{
	my $self = shift;
	my( $slice, $load_transcripts ) = @_;

	# mother copy-paste like
  	my $start = $slice->start;
  	my $end = $slice->end;

	my $o_dba   = $self->db();
	my $species = $o_dba->species();
	my $chrom   = $slice->seq_region_name;
	my $chromosome_file = $o_dba->conf()->GetVariationFeatureFiles( -chrom  => $chrom );

	my @a_features = @{$self->get_features_in_region($chromosome_file, $start, $end)};

	my $ra_VariationFeatures = $self->_objs_from_hashes(\@a_features, $slice);
	# end of mother copy-paste like

	if (defined($load_transcripts) and $load_transcripts == 1){

		# Else the transcripts have to be loaded and added to their respective VariationFeature.
		my $transcript_adaptor = $self->db->get_TranscriptAdaptor();
		foreach my $VariationFeature (@$ra_VariationFeatures){
			my $ra_transcripts = $transcript_adaptor->fetch_all_by_VariationFeature($VariationFeature);

			# we add transcripts to the current VariationFeature
			foreach my $transcript (@$ra_transcripts){
				$VariationFeature->add_Transcript($transcript);
			}
		}
	}

	return $ra_VariationFeatures;
}


=head2 fetch_by_stable_id

  Title		: fetch_by_stable_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_stable_id('ENSGXXXXXXXXXX');
  Function	: Retrieves a VariationFeature from the files via its stable id which is locus_tag
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $stable_id
               the stable id of the VariationFeature to retrieve

=cut
sub fetch_by_stable_id
{
	my $self = shift;
	my( $stable_id ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'locus_tag', $stable_id, \@a_files );

	return $o_VariationFeature;
}


=head2 fetch_by_translation_stable_id

  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_translation_stable_id( $prot_id );
  Function	: Retrieves a VariationFeature from the files via a translation stable_id,
              which in Narcisse is the protein_id of transcript files.
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $prot_id
               the stable id of the VariationFeature to retrieve

=cut

sub fetch_by_translation_stable_id
{
	my ($self, $prot_id) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_translation_id( $prot_id );
	return undef  unless( defined $o_transcript );

	# obtenir le VariationFeature
	my $tr_id = $o_transcript->dbID();
	my $o_VariationFeature = $self->fetch_by_transcript_id( $tr_id );

	return $o_VariationFeature;
}


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_dbID('s1g16');
  Function	: Retrieves a VariationFeature from the files via its internal id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $dbID
               the internal id of the VariationFeature to retrieve

=cut

sub fetch_by_dbID
{
	my $self    = shift;
	my( $dbID ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'element', $dbID, \@a_files );

	return $o_VariationFeature;
}


=head2 fetch_by_symbol

  Title		: fetch_by_symbol
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_symbol('cdc25a');
  Function	: Retrieves a VariationFeature from the files via its symbol
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $symbol
              the $symbol of the VariationFeature to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_symbol
{
	my $self = shift;
	my( $symbol ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'symbol', $symbol, \@a_files );

	return $o_VariationFeature;
}



=head2 fetch_by_transcript_accession

  Title		: fetch_by_symbol
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_accession( $x );
  Function	: Retrieves a VariationFeature from the files via its accession
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $accession
              the $symbol of the VariationFeature to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_transcript_accession
{
	my $self = shift;
	my( $accession ) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_accession( $accession );
	return undef  unless( defined $o_transcript );

	# obtenir le VariationFeature
	my $tr_id = $o_transcript->dbID();
	my $o_VariationFeature = $self->fetch_by_transcript_id( $tr_id );

	return $o_VariationFeature;
}


sub fetch_by_exon_stable_id {

	my ($self, $stable_id) = @_;

	throw("Not implemented!");

}


=head2 fetch_by_transcript_id

  Title		: fetch_by_transcript_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_id('s1g4t1');
  Function	: Retrieves a VariationFeature from the files via his transcript db id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $db_id
               the db id of the transcript

=cut
sub fetch_by_transcript_id {

	my ($self, $db_id) = @_;

	#Dans narcisse: transcript db id = VariationFeature db id + 't' + num transcript
	$db_id =~ s/t\d+$//;

	return $self->fetch_by_dbID($db_id);

}


=head2 fetch_by_transcript_stable_id

  Title		: fetch_by_transcript_stable_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_stable_id('ENSG00000146556#1');
  Function	: Retrieves a VariationFeature from the files via his transcript stable id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $trans_stable_id
               the stable id of the transcript

=cut
sub fetch_by_transcript_stable_id {
    my ($self, $trans_stable_id) = @_;

	#Dans narcisse: transcript stable id = locus tag
	# c'est donc le  locus tag de son VariationFeature + eventuellement #numTranscript
	$trans_stable_id =~ s/#.+$//;

	return $self->fetch_by_stable_id($trans_stable_id);
}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of VariationFeatureric features given as hash refs into Bio::EnsEMBL::VariationFeature objects
  Returns	: listref Bio::EnsEMBL::VariationFeature
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut
sub _objs_from_hashes {

	my ($self, $tab_of_hash, $slice) = @_;

	my @a_VariationFeatures;
	foreach my $feature (@$tab_of_hash){
		push @a_VariationFeatures, Bio::EnsEMBL::VariationFeature->new( '-start'  		=> $feature->{'start'},
											  '-end'			=> $feature->{'end'},
											  '-strand'			=> $feature->{'strand'},
											  '-adaptor'		=> $self,
											  '-slice'			=>  $slice,
											  '-dbID'			=>  $feature->{'element'},
											  '-stable_id'		=>  $feature->{'locus_tag'},
											  '-external_name' =>  $feature->{'symbol'},
											  '-created_date'	=>  undef,
											  '-modified_date'	=>  undef
											);

	}

	return \@a_VariationFeatures;
}


1;
