=head1 NAME

Bio::Narcisse::VariationFeatureAdaptor - Database adaptor for the retrieval and
storage of VariationFeature objects ( which are uniquely SNPs for Narcisse ).

=head1 SYNOPSIS

  use Bio::Narcisse::DBAdaptor;

  $dba = Bio::Narcisse::DBAdaptor->new(...);

  ### not yet implemented, no use ?

  # fetch all genome hits for a particular variation
  $v = $va->fetch_by_name('rs56');

  foreach $vf (@{$vfa->fetch_all_by_Variation($v)}) {
    print $vf->seq_region_name(), $vf->seq_region_start(), '-',
          $vf->seq_region_end(),"\n";
  }

  ### not present in EnsEMBL API
  # fetch all genome hits for a particular variation
  my $ra_vf = $o_vfa->fetch_all_by


=head1 DESCRIPTION

This is an adaptor for the retrieval and storage of VariationFeature objects.

=cut


package Bio::Narcisse::VariationFeatureAdaptor;

use strict;
use warnings;

use Data::Dumper;

use Bio::Narcisse::SliceAdaptor;
use Bio::Narcisse::BaseFeatureAdaptor;
use Bio::Narcisse::DBAdaptor;
use Bio::EnsEMBL::VariationFeature;
use Bio::Narcisse::Utils::Exception qw( throw );

use base qw( Bio::Narcisse::BaseFeatureAdaptor );


=head2 new

  Title		: new
  Usage    	: $adaptor = $dba->get_VariationFeatureAdaptor();
  Function	: Creates a new VariationFeatureAdaptor object.
  Returns	: Bio::Narcisse::VariationFeatureAdaptor
  Args		: None

=cut


=head2 _GetFeatureKey

  Prerequiste: Not in EnsEMBL API.
  Usage      : my $key = $self->_GetFeatureKey();
  Description: return the cfg key for this feature.
  Return     : string.
  Caller     : Bio::Narcisse::BaseFeatureAdaptor

=cut

sub _GetFeatureKey
{
	my $self = shift;

	return 'SNP';
}


=head2 fetch_all_by_Variation

  Prerequite : NOT COMPATIBLE WITH EnsEMBL API
               WHICH USE A VARIATION OBJECT AS ARG 1
  Arg [1]    : $rsid
  Example    : my $ra_vfs = $vfa->fetch_all_by_Variation( $rsid );
  Description: Retrieves all variation features for a given rsid.  Most
               rsids should only hit the genome once and only a return
               a single variation feature.
               In Narcisse, we only tolerate 1 variation feature per chromosome
  Returntype : reference to list Bio::EnsEMBL::Variation::VariationFeature
  Exceptions : throw on bad argument
  Caller     : general
  Status     : developpement in progress

=cut

sub fetch_all_by_Variation
{
	my $self = shift;
	my( $rsid ) = @_;

	$rsid =~ s/^rs//;

#	ce qu'il faudrait faire pour coller au plus prêt de l'API
#	mais generic_fetch ne renvoie qu'un enregistrement
#	return $self->generic_fetch( $col, $var );
#
#	Je ne peux meme pas utiliser generic_fetch qui renvoie uniquement
#	le premier enregistrement.

	#my $ra_features_rows = $self->generic_fetch( $col, $var );
	# ... a voir au besoin
	# pour que cette fonction soit efficace,
	# 
	
}


=head2 _objs_from_hashes

  Title		: _objs_from_hashes
  Function	: Translate a list of generic features given as hash refs into Bio::EnsEMBL::VariationFeature objects
  Returns	: listref Bio::EnsEMBL::VariationFeature
  Args		: $tab_of_hash - array ref holding features as hashes
  			  Bio::EnsEMBL::Slice - $slice

=cut

sub _objs_from_hashes
{
	my( $self, $ra_hashes, $slice ) = @_;

	my @a_VariationFeatures;

	foreach my $feature ( @$ra_hashes )
	{
		my $o_vf = Bio::EnsEMBL::VariationFeature->new(
						'-start'        => $feature->{'start'},
						'-end'			=> $feature->{'end'},
#						'-strand'		=> $feature->{'strand'},
						'-adaptor'		=> $self,
						'-slice'		=> $slice,
						'-dbID'			=> $feature->{'element'}
						#les autres attributs seront mis si besoin
					);
		push( @a_VariationFeatures, $o_vf );
	}

	return \@a_VariationFeatures;
}

1;

__END__
=head2 fetch_by_stable_id

  Title		: fetch_by_stable_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_stable_id('ENSGXXXXXXXXXX');
  Function	: Retrieves a VariationFeature from the files via its stable id which is locus_tag
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $stable_id
               the stable id of the VariationFeature to retrieve

=cut
sub fetch_by_stable_id
{
	my $self = shift;
	my( $stable_id ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'locus_tag', $stable_id, \@a_files );

	return $o_VariationFeature;
}


=head2 fetch_by_translation_stable_id

  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_translation_stable_id( $prot_id );
  Function	: Retrieves a VariationFeature from the files via a translation stable_id,
              which in Narcisse is the protein_id of transcript files.
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $prot_id
               the stable id of the VariationFeature to retrieve

=cut

sub fetch_by_translation_stable_id
{
	my ($self, $prot_id) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_translation_id( $prot_id );
	return undef  unless( defined $o_transcript );

	# obtenir le VariationFeature
	my $tr_id = $o_transcript->dbID();
	my $o_VariationFeature = $self->fetch_by_transcript_id( $tr_id );

	return $o_VariationFeature;
}


=head2 fetch_by_dbID

  Title		: fetch_by_dbID
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_dbID('s1g16');
  Function	: Retrieves a VariationFeature from the files via its internal id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $dbID
               the internal id of the VariationFeature to retrieve

=cut

sub fetch_by_dbID
{
	my $self    = shift;
	my( $dbID ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'element', $dbID, \@a_files );

	return $o_VariationFeature;
}


=head2 fetch_by_symbol

  Title		: fetch_by_symbol
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_symbol('cdc25a');
  Function	: Retrieves a VariationFeature from the files via its symbol
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $symbol
              the $symbol of the VariationFeature to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_symbol
{
	my $self = shift;
	my( $symbol ) = @_;

	my $o_dba = $self->db();
	my @a_files = $o_dba->conf()->GetVariationFeatureFiles();
	my $o_VariationFeature = $self->fetch_feature( 'symbol', $symbol, \@a_files );

	return $o_VariationFeature;
}



=head2 fetch_by_transcript_accession

  Title		: fetch_by_symbol
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_accession( $x );
  Function	: Retrieves a VariationFeature from the files via its accession
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $accession
              the $symbol of the VariationFeature to retrieve
 Note       : Not part of API ensEMBL.

=cut

sub fetch_by_transcript_accession
{
	my $self = shift;
	my( $accession ) = @_;

	my $o_dba = $self->db();

	# identifier le transcript
	my $o_tra = $o_dba->get_TranscriptAdaptor();
	my $o_transcript = $o_tra->fetch_by_accession( $accession );
	return undef  unless( defined $o_transcript );

	# obtenir le VariationFeature
	my $tr_id = $o_transcript->dbID();
	my $o_VariationFeature = $self->fetch_by_transcript_id( $tr_id );

	return $o_VariationFeature;
}


sub fetch_by_exon_stable_id {

	my ($self, $stable_id) = @_;

	throw("Not implemented!");

}


=head2 fetch_by_transcript_id

  Title		: fetch_by_transcript_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_id('s1g4t1');
  Function	: Retrieves a VariationFeature from the files via his transcript db id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $db_id
               the db id of the transcript

=cut
sub fetch_by_transcript_id {

	my ($self, $db_id) = @_;

	#Dans narcisse: transcript db id = VariationFeature db id + 't' + num transcript
	$db_id =~ s/t\d+$//;

	return $self->fetch_by_dbID($db_id);

}


=head2 fetch_by_transcript_stable_id

  Title		: fetch_by_transcript_stable_id
  Usage		: $VariationFeature = $VariationFeature_adaptor->fetch_by_transcript_stable_id('ENSG00000146556#1');
  Function	: Retrieves a VariationFeature from the files via his transcript stable id
  Returns	: Bio::EnsEMBL::VariationFeature
  Args		: string $trans_stable_id
               the stable id of the transcript

=cut
sub fetch_by_transcript_stable_id {
    my ($self, $trans_stable_id) = @_;

	#Dans narcisse: transcript stable id = locus tag
	# c'est donc le  locus tag de son VariationFeature + eventuellement #numTranscript
	$trans_stable_id =~ s/#.+$//;

	return $self->fetch_by_stable_id($trans_stable_id);
}


1;
