#!/usr/bin/perl
#
# $Id$
#
# Copyright INRA/CNRS

# Emmanuel.Permal@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 

 ncs_AnalyseDsData.pl - a script to analyse Ds Data

=head1  SYNOPSIS

 perl ncs_AnalyseDsData.pl @args

 @args:
 [mandatory]
  --species : the species name
 
=head1 DESCRIPTION

 ncs_ncs_AnalyseDsData.pl  = Using Narcisse API analyse Ds data

example :

perl ncs_AnalyseDsData.pl --species=my_species 

=cut

=head1 SUBROUTINES

MAIN


=cut

use strict;
use warnings;

#Standard modules
use Error qw( :try );
use Data::Dumper;
use File::Basename;
use File::Path;
use File::Spec;
use Cwd;
use Cwd 'abs_path';
use IO::File;
use Pod::Usage;
use FindBin;
use File::Copy qw(move copy);
use File::Path qw( mkpath rmtree ); 
use File::Temp qw(tempfile tempdir);
use Env;

#Narcisse modules 
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

#Narcisse
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;
use Runner;
use NARCISSE_utils;
use NarcisseServiceFactory;
use NarcisseService;
use NarcisseService::GrepBased;

#Analysis module
use DsDataAnalyser;

#Constants
use constant NARCISSE_CFG => 'cfg';

&TryMain(0);

MAIN:
{ 
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = ( @_ );
		warn $o_err-> stacktrace(); 
	}
	otherwise
	{
		my ( $o_err ) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";
		print STDERR $str;		    
	};
	
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{ 
	#Init Parameters
	print "Initialize parameters\n";
	my $o_param = &GetParameters();
	
	#Perform analysis
	
	my $o_dsa = $o_param -> Get ('o_dsa');
	
	##BuildParaclusters
	print "Building paraclusters\n";
	my $cutoff = $o_param -> Get ('cutoff');
	my $rh_para_cluster = $o_dsa -> BuildParaCluster ( $cutoff);
	#~print Dumper (%$rh_para_cluster);
	
	##Ds File classification
	my $rh_ds_hash = $o_dsa ->GetRefDsHash();
	my %h_classif =();
	foreach my $pair_id ( sort keys %$rh_ds_hash)
	{
		my $ds_class = $o_dsa -> DefineDsClass ($pair_id);
		my $exon_class = $o_dsa -> DefineExonClass ($pair_id);
		my $orientation = $o_dsa -> DefineOrientation ($pair_id);
		my $phy_link = $o_dsa -> DefinePhysicalLink ($pair_id);
		my $gene_space = $o_dsa -> GetGeneDistance ($pair_id);
		my $id1 = $rh_ds_hash -> {$pair_id} -> {seq1};
		my $id2 = $rh_ds_hash -> {$pair_id} -> {seq2};
		my $para_class = $o_dsa -> DefineParaclusterClass ($pair_id);
		my $para_clusterID = "NA";
		if ($para_class eq "single")
		{
			$para_clusterID = $o_dsa -> GetParaclusterFromId ($id1);
		}
		
		#print "PAIR $pair_id ds_class $ds_class ExonClass $exon_class Orient $orientation Link $phy_link Gene Sapec $gene_space Para $para_class Parsid $para_clusterID\n";
		
		$h_classif{$pair_id} = { 	
									id1 => $id1,
									id2 => $id2,
									dS => $rh_ds_hash -> {$pair_id} -> {dS},
									ds_class => $ds_class,
									exon_class => $exon_class,
									orientation => $orientation,
									chr_link => $phy_link,
									gene_span => $gene_space,
									para_class => $para_class,
									paracluster_id => $para_clusterID
								}
	}
	
	#~print Dumper (%h_classif);
	
	#Define  Repeat type
	
	#Print output
	my $output = $o_param -> Get('out_classif');
	print "Printing result for class in file $output\n";
	&PrintClassifResults(\%h_classif, $output);
	
	my $output_para = $o_param -> Get('out_para');
	print "Printing result for class in file $output_para\n";
	&PrintParaclusterResults ($rh_para_cluster, $output_para, $o_dsa);
		
}
=head2 Procedure PrintClassifResults

 Title        : PrintClassifResults
 Usage        : &PrintClassifResults;
 Procedure    : 
 Args         : 
 Returns	  : 

=cut

sub PrintClassifResults
{
	my ($rh_ds_classif, $output_file) = (@_);
	my $fo_ds = &GetStreamOut($output_file);
	my @a_headers = qw(cds_id1 cds_id2 dS ds_class exon_class orientation chr_link gene_span para_class paracluster_id);
	print $fo_ds join("\t", @a_headers)."\n";
	
	foreach my $selected_pair (keys %$rh_ds_classif)
	{
		my $outString = $rh_ds_classif -> {$selected_pair } ->{id1} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{id2} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{dS} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{ds_class} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{exon_class} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{orientation} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{chr_link} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{gene_span} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{para_class} . "\t";
		$outString .= $rh_ds_classif -> {$selected_pair } ->{paracluster_id} . "\n";
		
		print $fo_ds $outString;

	}
	return;
}

=head2 Procedure PrintParaclusterResults

 Title        : PrintParaclusterResults
 Usage        : &PrintParaclusterResults;
 Procedure    : 
 Args         : 
 Returns	  : 

=cut

sub PrintParaclusterResults
{
	my ($rh_para, $output_file, $o_dsa) = (@_);
	my $fo_ds = &GetStreamOut($output_file);
	my @a_headers = qw(paracluster_id  chr family size members mb_index);
	print $fo_ds join("\t", @a_headers)."\n";
	foreach my $para_id (keys %$rh_para)
	{
		my $ra_members = $rh_para -> {$para_id} -> {ra_gene};
		my $chr = $rh_para -> {$para_id} -> {chr};
		my @a_cds_members = ();
	
		foreach my $mb_index (@$ra_members)
		{
			my $cds_id = $o_dsa -> GetProteinIDFromIndex($chr, $mb_index);
			push (@a_cds_members, $cds_id);
		}
		
		my $size = scalar @a_cds_members;	
		my $outString = $para_id . "\t";
		$outString .= $chr ."\t";
		$outString .= $rh_para -> {$para_id} -> {fam} ."\t";
		$outString .= $size ."\t";
		$outString .= join (",", @a_cds_members)."\t";
		$outString .= join (",", @$ra_members)."\n";
		
		print $fo_ds $outString;
		
	}
	return;
}


=head2 Procedure GetParameters

 Title        : GetParameters
 Usage        : &GetParameters;
 Procedure    : Reads parameters from command line
 Args         : none
 Returns	  : $o_param object

=cut

sub GetParameters
{
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'species=s', 'cfg=s','prefix=s', 'ds=s', 'gene=s', 'blocks=s', 'families=s','cutoff=i', 'help');
	
	if ( !$o_param -> AssertDefined('species') || !$o_param -> AssertDefined('ds') 
	|| !$o_param -> AssertDefined('gene') || !$o_param -> AssertDefined('cutoff'))
	{
		warn "\nERROR - species and ds and gene and blocks parameters are mandatory\n";
		&Usage();
		exit 1;
	}
	
	#Setting default values
	$o_param -> SetUnlessDefined ('prefix', 'stdout');
	$o_param -> SetUnlessDefined ('blocks', undef);
	$o_param -> SetUnlessDefined ('cutoff', 10);
	$o_param -> SetUnlessDefined ('families', undef);
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param -> Get(&NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	$o_param -> SetUnlessDefined ('cfg', $ENV{NARCISSE_CFG});

	my $o_instanceObserver = New InstanceObserver( -cfg => $o_param -> Get ('cfg'));
	$o_param -> Set ('o_io', $o_instanceObserver);
	
	my $o_dsa = New DsDataAnalyser (	species => $o_param -> Get ('species'), 
										ds_file => $o_param -> Get ('ds'),
										gene_file => $o_param -> Get ('gene'),
										block_file => $o_param -> Get ('blocks'),
										fam_file => $o_param -> Get ('families'));
	$o_param -> Set ('o_dsa' , $o_dsa);
	
	my $prefix = $o_param -> Get ('prefix');
	my $output_classif = my $output_para = "";
	
	if ($prefix ne 'stdout')
	{
		$output_classif = $prefix.".annotated";
		$output_para = $prefix. ".paracluster";
	}
	$o_param -> Set ('out_classif', $output_classif);
	$o_param -> Set ('out_para', $output_para);
	
	return $o_param;
}

=head2 Procedure Usage

 Title        : Usage
 Usage        : &Usage;
 Procedure    : Displays script usage
 Args         : none

=cut

sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe [--cfg], [--species], [--ds], [--gene], [--block], [--families], [--cutoff], [--prefix]

[mandatory]
    --species=      : The species name (Use quotes - For example: "Sus scrofa")
   
[options]
    --cfg=			: The narcisse config file if NARCISSE_CFG not in PATH
    --ds=      		: The file with ds data
    --gene=         : The file with gene information
    --families=     : The fiel with families
    --block=		: The file with Dominant/sensitive blocks information
    --cutoff=       : The cutoff to build paraclusters
    --prefix=		: The name of the prefix for the output files displaying selected couple of proteins
					  If not specified, results will be displayed on stdout.
    
END

;
}
