#!/usr/bin/perl

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

 ncs_Blast.pl

=head1 SYNOPSIS

 perl ncs_Blast.pl @args

 @args:
  --

=head1 DESCRIPTION

 # description

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 05 March 2009
 $Id: ncs_Blast.pl 3835 2013-06-03 15:03:33Z  $

=cut


BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);

	# pour Appli
	unshift @INC, "$dirprg/../../multi/web_services/playmoby/lib/";

	# bien mettre ../lib apres playmoby sinon pb avec ParamParser.
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

#---------------------------------------------------------------
package MultiService::Blast;

use strict;
use warnings;

use base qw( MultiService );

use Data::Dumper;
use File::Copy;

use NarcisseService::UsingExternalProgram;
use Configuration::NarcisseObserver;

#params
use constant INFILE  => 'input_file';

#options
use constant PROG        => 'prog';
use constant EXPECT      => 'expect';
use constant SPECIES     => 'species';
use constant MORE_PARAMS => 'params';


=head2 Procedure

 Title        : _InitAttributes
 Usage        : $o_xx->_InitAttributes( $o_param );
 Prerequisite : none
 Procedure    : check that mandatory params are sets and set InstanceObservers List Attribute.
 Args         : $o_param, ParamParser object
 Attributes   :
     __params, string, list of parameters to add to the command line.
     __expect, float, value of expect parameter.
     __species, string, short species name.
     __prog, string, program name

=cut

sub _InitAttributes
{
	my $self = shift;
	my ( $o_param ) = @_;

	my( $o_inst_obs, $species ) = $self->__FindInstanceObserver( $o_param );
	$self->_SetInstanceObserversListAttribute( [ $o_inst_obs ] );

	$self->_Set( __species => $species )  if( defined $species );

	my @a_keys  = ( &PROG, &EXPECT, &MORE_PARAMS );
	my( $prog, $expect, $params ) = $o_param->Get( @a_keys );
	my $outfile = $o_param->Get( &MultiService::OUT_FILE );

	$self->_Set( __prog    => $prog,
	             __expect  => $expect,
	             __params  => $params,
	             __outfile => $outfile );

	return;
} # _InitAttributes


=head2 Function __FindInstancesObserver

 Usage        : my ( $o_inst_obs, $species ) = $o_x->__FindInstanceObserver( $o_param )
 Function     : If instance is specified and not species, use it.
                If species is specied and not instance, search its instance.
                if both are specified, check that species belong to the instance,
                    if not throw error message.
                If none is specified, use multi instance which is a NarcisseObserver.
 Returns      : ( $o_inst_obs, $species ), a list of 2 elements.
     $o_inst_obs, the InstanceObserver object to use.
     $species, the short species name or undef.
 Args         : $o_param, the ParamParser object.
 Error        : ObjectException.

=cut

sub __FindInstanceObserver
{
	my $self = shift;
	my( $o_param ) = @_;

	my @a_keys = ( &MultiService::INSTANCES, &SPECIES );
	my ( $str_instance, $str_species ) = $o_param->Get( @a_keys );

	my $species_flag  = &General::IsNotNull( $str_species );
	my $instance_flag = &General::IsNotNull( $str_instance );

	my( $species, $o_inst_obs );
	my $o_nar_obs = New NarcisseObserver( -o_logger => $self->_GetLogger() );

	if ( &FALSE == $species_flag || ( 'all'  eq $str_species ) )
	{
		# idee ci dessous plus maintenue : maintenant all == MULTI => fichier pal
		# 'all' c'est une liste d'instance,
		# 'MULTI' c'est UNE instance qui regroupe les donnees des autres cf fichiers pal.
		if( ( &FALSE == $instance_flag ) || ( 'all'  eq $str_instance  ) )
		{
			$o_inst_obs = $o_nar_obs; # faudra le renommer pour $o_conf si ca marche ! TODO
		}
		else
		{ # instance + NIL
			my $ra_inst_names = $self->_FetchArgs( $str_instance );
			my $rao_inst_obs = $o_nar_obs->FindInstanceObservers( ra_instances => $ra_inst_names );
			$o_inst_obs   = $$rao_inst_obs[0];

			if( !defined $o_inst_obs )
			{
				my $msg = "No instance found, check your parameters, \n";
				$msg   .= "$str_instance does not contain valid instance name.";
				throw LipmError::ObjectException( $self, '!build', -text => $msg );
			}
		}
	}
	else # species defined
	{
		my $rao_inst_obs;

		if( &FALSE == $instance_flag )
		{ # NIL + species
			$rao_inst_obs = $o_nar_obs->FindInstanceObservers( ra_species => [ $str_species ] );
			$self->__AssertOnlyOneInstance( $rao_inst_obs );
		}
		else
		{ # instance + species
			my $ra_inst_names = $self->_FetchArgs( $str_instance );
			$rao_inst_obs = $o_nar_obs->FindInstanceObservers( ra_instances => $ra_inst_names,
																ra_species => [ $str_species ] );
		}

		$o_inst_obs = $$rao_inst_obs[0];

		if( !defined $o_inst_obs )
		{
			my $msg = "No instance found, check your parameters, \n";
			$msg   .= "maybe '$str_species' does not belong to the instance '$str_instance'.\n";
			throw LipmError::ObjectException( $self, '!build', -text => $msg );
		}

		$species = $o_inst_obs->GetSpeciesShortName( $str_species );
	}

	return ( $o_inst_obs, $species );

	# arg : $rao_inst_obs, the list of InstanceObservers
	sub __AssertOnlyOneInstance
	{
		my $self = shift;
		my( $rao_inst_obs ) = @_;

		if( scalar( @$rao_inst_obs ) == 0 )
		{
			my $msg = "Your species name does not belong to any Narcisse Instance.";
			throw LipmError::ObjectException( $self, '!build',
										-text => $msg, -user_msg => $msg );
		}
		elsif( scalar( @$rao_inst_obs ) > 1 )
		{
			my $msg = "Your species name correspond to more than one Narcisse Instance.";
			$msg   .= "\nYou have to specify which one to use.";
			throw LipmError::ObjectException( $self, '!build',
										-text => $msg, -user_msg => $msg );
		}

		return;
	} #__AssertOnlyOneInstance

} #__FindInstanceObserver


=head2 _FetchAndSetMandatories

 Usage        : $o_x->_FetchAndSetMandatories( $o_param )
 Function     : Read mandatories attributes and throw Error if they are missing.
                Then set attributes with values.
 Args         :
     $o_param
 Error        : none.

=cut

sub _FetchAndSetMandatories
{
	my $self   = shift;
	my( $o_param ) = @_;

	# out_file
	$self->SUPER::_FetchAndSetMandatories( $o_param );

	# infile
	my $infile = $o_param->Get( &INFILE );
	if( $infile eq '' or !-e -r $infile )
	{
		my %h_msg = ( -user_msg => "You need to specify an input file." );
		throw LipmError::ParameterException( &INFILE, 'missing', %h_msg );
	}

	$self->_Set( __seq_filename => $infile );

	return;
}


=head2 Function _BuildFileAttribute

 Usage        : my $scalar = $o_x->_BuildFileAttribute( $outfile )
 Function     : This file do nothing, return the output filename $outfile.
 Returns      : $scalar, string : $outfile.
 Args         :
     $outfile, string, the output filename.
 Error        : none

=cut

sub _BuildFileAttribute
{
	my $self = shift;
	my( $outfile ) = @_ ;
	
	return $outfile;
}


=head2 _BuildNarcisseService

 Usage        : my $o_ns = $o_x->_BuildNarcisseService( %h_args ).
 Prerequiste  : none
 Function     : This function must build and return a NarcisseService object.
 Returns      : $o_ns, NarcisseService object
 Args         : %h_params,
     o_instance_obs => $o_inst_obs,
 Error        : none.

=cut

sub _BuildNarcisseService
{
	my $self   = shift;
	my %h_args = ( o_instance_obs => undef,
	               @_ );

	# je ne verifie pas que les params sont definis, je fais confiance.
	my( $o_inst_obs ) = @h_args{ qw( o_instance_obs ) };

	my @a_attr = qw( __seq_filename __species __outfile);
	my( $seq, $species, $outfile ) = $self->_Get( @a_attr );
	@a_attr = qw( __prog __expect __params );
	my( $prog, $expect, $params ) = $self->_Get( @a_attr );

	my %h_params = ( seq      => $seq,
					 -outfile => $outfile,
					 -o_instance_obs => $o_inst_obs,
					 -o_logger       => $self->_GetLogger(),
	                 @_ );

	#les options, on ajoute que si on a une valeur pour ne pas ecraser la valeur par defaut
	$h_params{ -prog }        = $prog    if( &General::IsNotNull( $prog ) );
	$h_params{ -species }     = $species if( &General::IsNotNull( $species ) );
	$h_params{ -expect }      = $expect  if( &General::IsNotNull( $expect ) );
	$h_params{ -more_params } = $params  if( &General::IsNotNull( $params ) );

	my $o_ns = New NarcisseService::Blast( %h_params );

	return $o_ns;
}


=head2 _PrintHash

 Usage        : $o_x->_PrintHash( $rh_results, $ra_results )
 Function     : move the temporary output filename to output file.
 Args         :
     $rh_results, hash ref, key = 'data', value = the temporary filepath
     $ra_results, ignored here
 Error        : none.

=cut

sub _PrintHash
{
	my $self = shift;
	my( $rh_results ) = @_;

	# on ne fait rien, le fichier est deja a sa place.
# 	my $outfile = $self->_GetFileAttribute();
# 	
# 	my $ra_res_file = $$rh_results{ data };
# 	my $res_file = $$ra_res_file[0];
# 	
# 	# deplacer $res_file dans $outfile - File::Copy
# 	move( $res_file, $outfile );
	
	return;
}


=head2 function _GetHelpParams

 Usage        : $string = $o_xx->__GetHelpParams();
 Prerequisite : none
 Function     : none
 Returns      : $string, description of parameters
 Args         : none

=cut

sub _GetHelpParams
{
	my $self   = shift;

	my $string = $self->SUPER::_GetHelpParams();

	my $format = '%-20s';

	my $p_infile   = sprintf( $format, &INFILE );

	$string .= '
    --' . $p_infile . ' : string, the filename of the fasta file with the sequence to search.';

	return $string;
}


=head2 function _GetHelpOptions

 Usage        : $string = $o_xx->__GetHelpOptions();
 Prerequisite : none
 Function     : none
 Returns      : $string, description of options
 Args         : none

=cut

sub _GetHelpOptions
{
	my $self   = shift;

	my $string = $self->SUPER::_GetHelpOptions();

	# instance passe d'option a parametre
#	$string =~ s/\n(\s+--instances.+\.)//;

	my $format = '%-20s';

#	my $p_instance = sprintf( $format, &MultiService::INSTANCES );

	my $p_prog   = sprintf( $format, &PROG );
	my @a_progs = &NarcisseService::Blast::O_PROGS_PROT->ListItems();
	push( @a_progs, &NarcisseService::Blast::O_PROGS_NUC->ListItems() );

	my $p_species= sprintf( $format, &SPECIES );
	my $p_expect = sprintf( $format, &EXPECT );
	my $p_params = sprintf( $format, &MORE_PARAMS );

 #   --' . $p_instance . ' : string, instance name or coma separated list of instance name.
	$string .= '
    --' . $p_prog . " : string, name of the blast program to use. ( @a_progs )";
    $string .= '
    --' . $p_species. ' : string, species name (short or long).
    --' . $p_expect . ' : float, expectation value
    --' . $p_params . ' : string, any parameters you want to add to the program';

	return $string;
}

1;
#---------------------------------------------------------------

use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Cwd;


# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;
use MultiService;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3835 $ =~ /\d+/g); $r[0]};

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;


MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{	warn( $o_err->Explain(), "\n" );	}

		print '*' . $o_err->GetUserMessage() . "*\n";
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $service_name = 'NarcisseBlast';
	my $cmd = &Cwd::abs_path( $0 );

	# 0- hash of general info
	my %h_general = (
			name	=>	$service_name,
			descr	=>	"Perfom a blastall program against Narcisse data.",
			authors	=>	'sebastien.letort@toulouse.inra.fr',
			doc 	=>	'no doc yet',
			version =>	'1.0',
			cmd     => $cmd
 		);

	# inputs
	my %h_inputs = (
 			'FastaSequence'	=>	{
				descr   	 =>	"A fasta file containing one or more sequence.",
				namespace	 =>	'Narcisse',
				type_biomoby =>	'FASTA',
				datatype	 =>	'Sequence',
				cmd	         =>	'--' . &INFILE . '=$value'
			}
		);

	# ouputs
	my %h_outputs = (
			'BlastResult' => {
				descr   	=>	"A standard Blast result.",
				namespace	=>	'Narcisse',
				type_biomoby=>	'NCBI_BLAST_Text',
				datatype	=>	'text',
				cmd     	=>	'--' . &MultiService::OUT_FILE . '=$value'
			}
		);

	# params
	#les differents programmes
	my %h_progs;
	foreach my $prog ( &NarcisseService::Blast::O_PROGS_PROT->ListItems(), &NarcisseService::Blast::O_PROGS_NUC->ListItems() )
	{	$h_progs{ $prog } = $prog;	}




	# copier-coller de MultiService::ThroughSpecies::_GetCommonParams
	# a factoriser => TODO
	# pour obtenir la liste des instances dispo
	# j'ai besoin d'un objet NarcisseObserver
	my @a_names = ();

	my $o_nar_obs = New NarcisseObserver();
	my $ra_io     = $o_nar_obs->FindInstanceObservers() ;
	foreach my $o_io ( @$ra_io )
	{
		my @a_long     = $o_io->GetSpeciesLongName();
		my @a_no_space = ();
		foreach my $long ( @a_long )
		{
			$long =~ s/\s/_/g;
			push( @a_no_space, $long );
		}
		push( @a_names, @a_no_space );
	}

	my %h_names = map { $_ => '' } @a_names;
	$h_names{ all } = 'All the species';




	my %h_parameters = (
			&SPECIES	=>	{
				name        =>	&SPECIES,
				datatype    =>	'Choice',
				type_biomoby=>	'String',
				enum        =>	\%h_names,
				default 	=>	'all',
				descr       =>	"the name of a species of narcisse DB. You can also use a coma separated list."
			},
			&PROG	=>	{
				datatype   	=>	'Choice',
				type_biomoby=>	'String',
				enum    	=>	\%h_progs,
				default 	=>	'blastp',
				descr   	=>	"name of the blast program to run."
			},
			&EXPECT	=>	{
				datatype   	=>	'float',
				type_biomoby=>	'Float',
				default 	=>	'0.001',
				descr   	=>	"Expectation value (E) [Real]."
			},
			&MORE_PARAMS	=>	{
				datatype   	=>	'string',
				type_biomoby=>	'String',
				default 	=>	'',
				cmd			=>	'--params=\'$value\'',
				descr   	=>	"Any additionnal parameters to add to the command line."
			}
		);


	my %h_params = ( rh_descr  => { -general => \%h_general,
									-outputs => \%h_outputs,
									-inputs  => \%h_inputs,
									-params  => \%h_parameters },
# 	                rh_help    => { params   => $params_str,
# 	                           		options  => $options_str,
# 	                           		base     => MultiService::GetBasicHelp($0)},
					prog_name  => $0,
	                ra_options => [ &INFILE . '=s', # bizarre de decrire le param ici et dans la classe, faudra revoir ca !
	                                &PROG . '=s',
	                                &EXPECT . '=s',
	                                &SPECIES . '=s',
	                                &MORE_PARAMS . '=s' ],
	                log_file   => $service_name . '.log' );

	my $o_service = New MultiService::Blast( %h_params );
	$o_service->Launch()  if( defined $o_service );

	return;
}


__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
