#!/usr/bin/perl
#
# $Id$
#
# Copyright INRA/CNRS

# Emmanuel.Permal@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 

 ncs_ComputeDs.pl - a script to compute Ds for a couple of sequences

=head1  SYNOPSIS

 perl ncs_ComputeDs.pl @args

 @args:
 [mandatory]
  --protein : file with amino acid fasta sequence
  --yn00 : yn00.ctl file
  --species : the species name
 [option]
 --cfg=			: The narcisse config file if NARCISSE_CFG not in your environment PATH
    --s_cutoff=     : The value of the cutoff applied on synonymous sites according to yn00 result (default = 100)
                      Set s_cutoff to 0 to avoid filtering data on S value 
    --maskCpG       : Add this option if you want to mask CpG isalnds in sequences and add the following parameters
					-minlen=     : The minimum sequence length used by EMBOSS newcpgreport program (default = 150)
					-window=     : The window size used by EMBOSS newcpgreport program (default = 100)
    --output=		: The name of the output file displaying Ds, Dn and Dn/Ds values. 
					  If not specified, results will be displayed on stdout.
	--help=         : This help
 

=head1 DESCRIPTION

 ncs_ComputeDs.pl  = Compute Ds values between sequences
This program first does a 2 by 2 sequence alignement with amino acid sequences using clustalw, 
and then translate this alignement in nucleotidique alignment using BioPerl aa_to_dna_aln. Finally the program
computes the Ds between sequences using codeml.

example :

perl ncs_ComputeDs.pl --protein=protein.list --species=my_species --codeml=codeml.ctl

=cut

=head1 SUBROUTINES

MAIN


=cut

use strict;
use warnings;

#Standard modules
use Error qw( :try );
use Data::Dumper;
use File::Basename;
use File::Path;
use File::Spec;
use Cwd;
use Cwd 'abs_path';
use IO::File;
use Pod::Usage;
use FindBin;
use File::Copy qw(move copy);
use File::Path qw( mkpath rmtree ); 
use File::Temp qw(tempfile tempdir);
use Env;

#Narcisse modules 
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

#BioPerl
use Bio::Seq;
use Bio::SeqIO;
use Bio::Align::Utilities qw(aa_to_dna_aln);
use Bio::AlignIO;

#Narcisse
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;
use Runner;
use NARCISSE_utils;
use NarcisseServiceFactory;
use NarcisseService;
use NarcisseService::GrepBased;
use DBAccess;
use Bio::Narcisse::DBAdaptor;

#Constants
use constant FASTA => 'fasta';
use constant NARCISSE_CFG => 'cfg';

&TryMain(0);

MAIN:
{ 
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = ( @_ );
		warn $o_err-> stacktrace(); 
	}
	otherwise
	{
		my ( $o_err ) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";
		print STDERR $str;		    
	};
	
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main{ 
	
	#Init Parameters
	my $o_param = &InitParameters();

	#Getting Protein IDs from file
	my $ra_couple_IDs = &GetProteinIDs($o_param);
	
	#Build a list of all IDs from couples
	my $ra_IDlist = &MakeIDlist ($ra_couple_IDs);
	
	#Create and cd to work directory
	my ($save_dir, $wrk_dir) = &CreateTempDir("computeDs");
	
	#Do the job
	my ($ds_status, $rh_results, $ra_selected) = &ComputeDs($ra_couple_IDs, $ra_IDlist, $o_param);
	
	#Writing output file
	my $out_file_name = $o_param -> Get( 'output');
	&WriteOutputFile($rh_results, $out_file_name);
	
	unless ($out_file_name eq 'stdout')
	{
		#Get back file to the beginning directory
		&CopyFile($out_file_name, $save_dir);
		print "\nPlease see results in file = $out_file_name\n";
	}
	
	#Cleaning temporary directories
	&CleaningDirectory($wrk_dir, $save_dir);
	
	
	return $ds_status;	
}

=head1 SUBROUTINES

=cut

=head2 Procedure GetDataAccess

 Title        : GetDataAccess
 Usage        : &GetDataAccess ($o_param);
 Procedure    : Builds a DBAccess object
 Args         : $o_param
 Returns     : $o_dba, a dbaccess object
 
=cut

sub GetDataAccess
{
	my ($o_param) = (@_);
	
	#Creating an access to data
	my $o_io = $o_param -> Get ('o_io');
	my $species = $o_param -> Get ('species');
	my $o_dba = New DBAccess (species => $species, -o_inst_obs => $o_io, -rep => &TRUE);

	return $o_dba;
}

=head2 Procedure ComputeDs

 Title        : ComputeDs
 Usage        : &ComputeDs($ra_couple_IDs, $ra_IDlist, $o_param);
 Procedure    : Access data via a DBAccess object
				Then check if the CpG masking option is on
				For each selected couple of CDS
				Creates local temporary ids for CDS (avoid problems with headres maximal size for clustal and yn00)
				Runs clustalw on protein sequences and returns a bioperl Align object
				Builds a hash of CDS sequences correspondind to current protein couple masked or not according to maskCpG option
				Translates Bioperl AA alignment into a Bioperl Nt alignement using hash with CDS sequences
				Runs Yn00 on Phylip formatted Nt alignment and returns a hash with results (ds, dn and dnds for all couple)
 Args         : $ra_IDs: a ref to an array with all protein IDs
				$ra_couple_IDs : a ref to an array with all protein couples
				$o_param: a ParamParser object containing all parameters
 Return       : $rh_results: a hash with a couple $Ids as keys and $S, $dn, $ds, $dn_ds as values
 
=cut

sub ComputeDs
{
	my ($ra_couple_IDs, $ra_IDlist, $o_param) = (@_);
	
	print "\nComputing Ds values";
	print " for " . scalar @$ra_IDlist . " CDS and " . scalar @$ra_couple_IDs . " couples\n";
	
	#Creating an access to data	
	my $o_dba = $o_param -> Get ('o_dba');
	
	#Need masking
	my $mask_cpg = $o_param -> Get ('maskCpG');
	
	#Results will be in a hash
	my %h_results = ();;
	my $status = 0;
	#~my $debug = 0;
	
	#Selected genes
	my @a_selected = ();
	
	while (scalar (@$ra_couple_IDs) > 0)
	{
		my $ra_ID_pair = shift(@$ra_couple_IDs);
		
		#Create temporary IDs to fit with yn00 conventions (less than 10 characters)
		my $rh_headers = &CreateTempHeaders ($ra_ID_pair);
		
		#Align two sequnces using clustalw	
		my ($o_AA_alignment, $clustal_status) = &RunClustal($rh_headers, $o_dba);	
		if ($clustal_status != 0 || !defined ($o_AA_alignment))
		{
			$status = 1;
			next;
		}
		#Building a hash with CDS sequnces masked or not
		my $rh_cds_seq = ();
		
		if ($mask_cpg)
		{
			
			$rh_cds_seq = &BuildCdsMaskedSeqHash ($o_dba, $rh_headers, $o_param);
		}
		else
		{
			$rh_cds_seq = &BuildCdsSeqHash ($o_dba, $rh_headers);
		}
		
		#Translating AA alignment into DNA alignment
		my ($o_DNA_alignment, $aln_status) = &GetDNAAlignment($o_AA_alignment, $rh_cds_seq);
		if ($aln_status != 0)
		{
			$status = 1;
			next;
		}
		
		#Computes dn, ds, dn/ds, S
		my ($rh_resYn00, $yn00_status) = &RunYn00 ($o_DNA_alignment, $o_param);
		if ($yn00_status != 0)
		{
			$status = 1;
			next;
		}
		
		next if (!defined $rh_resYn00);
		#Here results are stored into a hash
		my $seq1 = $ra_ID_pair->[0];
		my $seq2 = $ra_ID_pair->[1];
		my $para_id = &BuildParaID ( $seq1, $seq2);
		$h_results { $para_id } = { seq1 => $seq1,
									seq2 => $seq2,
									dN => $rh_resYn00 -> {'dN'},
									dS => $rh_resYn00 -> {'dS'},
									dN_dS => $rh_resYn00 -> {'dn_ds'},
									S => $rh_resYn00 -> {'S'},
									};
									
		push (@a_selected, @$ra_ID_pair);
		#~$debug += 1;
		#~last if ($debug == 10);
	}		
	my @a_selected_uniq = &Uniq (@a_selected);
	
	return ($status, \%h_results, \@a_selected_uniq);
}

=head2 Procedure CreateTempHeaders

 Title        : CreateTempHeaders
 Usage        : &CreateTempHeaders($ra_IDs);
 Procedure    : Creates a hash with headers and their corresponding temporary headers
 Args         : $ra_IDs
 Returns      : A hash of correspondance between header and temporary header
 
=cut

sub CreateTempHeaders
{
	my ($ra_IDs) = (@_);
	my $prefix_id = "tmp"; 
	my $index = 0;
	my %h_temp_ID = ();
	foreach my $id (@$ra_IDs)
	{
		$h_temp_ID {$id} = $prefix_id."_".$index;
		$index += 1;
	}
	
	return \%h_temp_ID;
}

=head2 Procedure PrintFasta

 Title        : PrintFasta
 Usage        : &PrintFasta ( $ra_headers, $rh_data, $type);
 Procedure    : Print data in a fasta formatted file
 Args         : $ra_headers, $rh_data, $type where $ra_headers is ref to headers, $rh_data is a ref to data sequences $type is either "cds" or "protein",
				and $type, the type of sequence cds or protein
 Returns 	  : A file name

=cut

sub PrintFasta
{
	my ( $rh_headers, $o_dba, $type) = @_;
	my $file_name = "input_".$type.".fasta";
	my $fh_output = &GetStreamOut($file_name);
	my $data = "";	
	foreach my $h ( keys (%$rh_headers) )
	{
		if ($type eq 'prot')
		{
			$data = $o_dba -> GetProteinSequence($h);
		}
		elsif ($type eq 'cds')
		{
			$data = $o_dba -> GetCDSSequence($h);
		}
		
		$data =~ s/(.{60})/$1\n/g;
		my $tempID = $rh_headers -> { $h };
		print $fh_output ">$tempID\n";
		print $fh_output "$data\n";
	}
	warn "An error occured while creating $file_name\n" unless (-e $file_name) ;	
	return $file_name;
}

=head2 Procedure CopyFile

 Title        : CopyFile
 Usage        : &CopyFile($file, $path);
 Procedure    : Copies a $file to the location $path
 Args         : $file to copy
				$path the destination of the file $file
 
=cut

sub CopyFile
{
	my ($file, $path) = (@_);
	
	copy($file, $path);
	
	return;
}

=head2 Procedure CleaningDirectory

 Title        : CleaningDirectory
 Usage        : &CleaningDirectory($wkdir, $sav_dir);
 Procedure    : Cleans useless temporary files at the end of program
 Args         : $wkdir : the current working direactory to remove
				$sav_dir : the directory to go to then remove $wkdir
 
=cut

sub CleaningDirectory
{
	my ($wkdir, $sav_dir) = (@_);
	chdir $sav_dir;
	rmtree($wkdir);
	return;
}

=head2 Procedure WriteOutputFile


 Title        : WriteOutputFile
 Usage        : &WriteOutputFile($rh_data, $output_file_name);
 Procedure    : Write an output file with values
 Args         : A reference ton a hash containing data and the output file name
 
=cut

sub WriteOutputFile
{
	my ($rh_data, $output_file_name) = (@_);
	
	print "\n*****Writing results in $output_file_name*****\n";
	my $fh_output= &GetStreamOut( "$output_file_name" );
	
	#Here read ref to hash parameters
	my @a_fields = qw (seq1 seq2 S dN dS dN/dS );
	print $fh_output join("\t", @a_fields)."\n";
	
	foreach my $ID (keys %{$rh_data})
	{	
		my $outString = $rh_data->{$ID} -> {'seq1'} . "\t";
		$outString .= $rh_data->{$ID} -> {'seq2'} . "\t";
		$outString .= $rh_data->{$ID} -> {'S'} . "\t";
		$outString .= $rh_data->{$ID} -> {'dN'} . "\t";
		$outString .= $rh_data->{$ID} -> {'dS'} . "\t";
		$outString .= $rh_data->{$ID} -> {'dN_dS'} . "\n";
		
		#Debug
		#print "STDOUT DS $outString \n";
		print $fh_output $outString;
	}
	return;
}

#~=head2 Procedure GetProteinInfo
#~
 #~Title        : GetProteinInfo
 #~Usage        : &GetProteinInfo();
 #~Procedure    : Get protein information
 #~Args         : 
 #~
#~=cut
#~
#~sub GetProteinInfo
#~{
	#~my ($o_param, $ra_ids) = (@_);
	#~
	#~my $o_io = $o_param -> Get ('o_io');
	#~my $o_spo = $o_io -> GetSpeciesObserver( $o_param -> Get ('species'));
	#~my @transcript_file = $o_spo ->GetTranscriptFiles();
	#~my $tr_file_name = "temp_all_trancript.txt";
	#~my $filelist = join (" ", @transcript_file);
	#~system `cat $filelist > $tr_file_name`;
	#~
	#~my $rh_prot_info = &ParseTranscriptFile( $tr_file_name, $ra_ids);
	#~return $rh_prot_info;
#~}


#~=head2 Procedure ParseTranscriptFile
#~
 #~Title        : ParseTranscriptFile
 #~Usage        : &ParseTranscriptFile();
 #~Procedure    : Parses the transcript file to get information on protein, chromosome, coordinates, exon number and gene name
 #~Args         : 
 #~
#~=cut
#~
#~sub ParseTranscriptFile
#~{
	#~my ($t_file, $ra_ids) = (@_);
	#~
	#~my $fi_file = &GetStreamIn ($t_file);
	#~
	#~my $chrom = my $gene_id = my $start = my $end = my $strand = my $nbexon = my $protein_id = "";
	#~my %h_prot_info = ();
	#~
	#~while (my $line = <$fi_file>)
	#~{
		#~my $index = 0;
		#~foreach my $id (@$ra_ids)
		#~{
			#~if ($line =~ /$id/)
			#~{
				#~
				#~my @a_line = split ("\t", $line);
				#~$chrom = $a_line[0];
				#~$gene_id = $a_line[4];
				#~$start = $a_line[6];
				#~$end = $a_line[7];
				#~$strand = $a_line[8];
				#~$nbexon = $a_line[9];
				#~$protein_id = $id;
				#~#DEBUG
				#~#print "ID $id gene id $gene_id\n";
				#~last;	
			#~}
			#~$index += 1;
		#~}
		#~
		#~$h_prot_info {$protein_id} = {chrom => $chrom,
										#~gene_id => $gene_id,
										#~start => $start,
										#~end => $end,
										#~strand => $strand,
										#~nb_exon => $nbexon};
										#~
		#~splice @$ra_ids, $index, 1;
		#~last if (scalar @$ra_ids == 0);	
	#~}
#~
	#~return \%h_prot_info;
#~}



=head2 Procedure CreateTempDir


 Title        : CreateTempDir
 Usage        : &CreateTempDir($prefix);
 Procedure    : Creates a working temp directory and cd into it
 Args         : $prefix of the temp dir
 Returns      : the $working_dir the created working directory path
              : the $sav_dir the current directory
 
=cut

sub CreateTempDir
{
	my ($prefix) =(@_);
	my $sav_dir = cwd;
	my $working_dir = File::Temp::tempdir( $prefix."_XXXX", CLEANUP => 0, TMPDIR => 1 );
	chdir($working_dir);
	#~print "Creating the work directory $working_dir\n";
	return $sav_dir, $working_dir;
}

=head2 Procedure BuildParaID

 Title        : BuildParaID
 Usage        : &BuildParaID ($seq1, $seq2);
 Procedure    : Creates an ID for the two compared sequences
 Args         : The two locustags
 Comment      : This method is here so that the ID could be changed easily on something else
 Return       : An ID for the couple $seq1, $seq2

=cut

sub BuildParaID
{
	my ($seq1, $seq2) = (@_);
	return $seq1. "_". $seq2;
}

=head2 Procedure RunClustal

 Title        : RunClustal
 Usage        : &RunClustal($ra_IDs, $o_dba);
 Procedure    : Does a clustalw alignment for two amino acid sequences
 Args         : A ref to an array of two IDs and a object DBAccess
 Returns	  : A Bio::SimpleAlign object alignment and a success/failed statuss
 
=cut

sub RunClustal
{
	my ($rh_headers, $o_dba) = @_;
	
	#Create work directory
	my ($cwd, $wkd) = &CreateTempDir("clustal");
	
	my $ra_IDs = \(keys (%$rh_headers));
	#First create fasta temporary file with two sequences
	my $tmp_file_out = "output_clustal.aln";
	my $tmp_file_in = &PrintFasta ($rh_headers, $o_dba, 'prot');
	
	#Create log file
	my ($fo_clustal, $log_file_name) = &CreateLogFile ("clustal");
	my $log_string = "START CLUSTAL FOR SEQUENCES: \t". (keys ($rh_headers))[0]. " AND:\t" . (keys ($rh_headers))[1]. "\n";
	&LogMsg ($fo_clustal, $log_string);
	
	#Then run clustal
	my $cmd_clustalw = "clustalw -INFILE=$tmp_file_in -TYPE=PROTEIN -MATRIX=BLOSUM -OUTFILE=$tmp_file_out -QUIET"; 
	&LogMsg ($fo_clustal, "COMMAND : $cmd_clustalw\n");
	my ($status, $killed) = RunExt( -cmd => $cmd_clustalw, -stdout  => ">>$log_file_name", -stderr  => ">>$log_file_name");

	#Create a simple Align Object out of it

	my $AA_aln_io = Bio::AlignIO -> new(	-file   => $tmp_file_out,
											-format => "clustalw" );
	my $o_AA_aln = $AA_aln_io -> next_aln;
	&LogMsg ($fo_clustal, "BIOPERL SIMPLE ALIGN OBJECT CREATED\n");
	
	if ($status == 0)
	{
		&CleaningDirectory ($cwd, $wkd);
	}	
	else
	{
		$log_string ="WARNING: A problem occured in clustal execution - see $wkd/$log_file_name file \n";
		warn $log_string ;
		&LogMsg ($fo_clustal,$log_string);
	}
	
	if ($killed !=0)
	{
		$log_string = "WARNING: clustal job has been killed - see $wkd/$log_file_name file \n";
		warn $log_string;
		&LogMsg ($fo_clustal, $log_string);
	}
	
	&LogMsg ($fo_clustal, "END OF CLUSTAL ALIGNEMENT\n");
	$fo_clustal -> close();
	
	return $o_AA_aln, $status;
}

=head2 Procedure GetDNAAlignment

 Title        : GetDNAAlignment
 Usage        : &GetDNAAlignment ($AA_aln, $ra_cds_seq);
 Procedure    : Translates an aa alignment into an dna alignment
 Args         : A reference to a hash containing two Bio::Seq objects and an AA alignment of two sequences
 Retruns      : A Bio::SimpleAlign object alignment and a success/failed status
=cut

sub GetDNAAlignment
{
	my ($AA_aln, $rh_cds_seq) = (@_);
	my $aln_status = 0; #everything is OK
	#print "Translating AA alignement to DNA alignment\n";
	
	my $o_DNA_aln = aa_to_dna_aln ($AA_aln, $rh_cds_seq);
	
	if (defined $o_DNA_aln)
	{
		$aln_status = 0;
	}
	else
	{
		$aln_status = 1;
		warn "WARNING: A problem occured in aa_to_dna_aln execution\n";
	}
	return $o_DNA_aln, $aln_status;
}

=head2 Procedure RunYn00

 Title        : RunYn00
 Usage        : &RunYn00 ($o_DNA_aln, $o_param);
 Procedure    : Runs yn00 and returns results for dn, ds, dnds value via a hash
 Args         : A Bio::AlignIO DNA alignment and $o_param
 Retruns      : a ref to an hash with results and a success/failed status;
=cut

sub RunYn00
{
	my ($o_DNA_aln, $o_param) = (@_);
	
	#Creating work directory
	my ($cwd, $wkd) = &CreateTempDir('yn00');
	
	my ($fo_log_yn00, $log_file_name) = &CreateLogFile ("yn00");
	&LogMsg ($fo_log_yn00, "Getting Yn00 parameters and writing de phylip formmated temporary file \n");
	#Getting paramters for yn00
	my $yn00_path = $o_param -> Get ('yn00');
	my $rh_yn00_param = $o_param -> Get ('yn00_param');
	my $yn00_input = $rh_yn00_param -> {'input'};
	my $rh_resYn00 = ();
	
	#First we need a phylip format file
	my $yn00_file_tmp = "input_yn00_tmp.phy";
	my $out = Bio::AlignIO->new(-file => ">$yn00_file_tmp",
								-format => 'phylip');
	$out->write_aln($o_DNA_aln);

	&LogMsg ($fo_log_yn00, "START EDITING FILE: \t $yn00_input to add Intervealed information in first line \n");
	
	#Then edit first line of the file to specify that the format is intervealed	
	my $fi_yn00 = &GetStreamIn($yn00_file_tmp) or Carp::croak("Cannot open $yn00_file_tmp file");
	my $first_line = <$fi_yn00>;
	chomp ($first_line);
	$first_line .= " I\n";
	
	my $fo_yn00 = &GetStreamOut($yn00_input) or Carp::croak("Cannot open $yn00_input file");
	print $fo_yn00 $first_line;
	while (my $line = <$fi_yn00>)
	{
		print $fo_yn00 $line;
	}
	$fi_yn00 -> close();
	$fo_yn00 -> close();
	
	&LogMsg ($fo_log_yn00,"END EDITING. \n");

	#print "Running yn00\n";
	#Then run yn00 
	&LogMsg ($fo_log_yn00,"START RUNNING YN00 WITH: \t $yn00_path FILE\n");
	
	my $cmd_yn00 = "yn00 $yn00_path "; 
	my ($status, $killed) = RunExt( -cmd => $cmd_yn00, -stdout  => ">>$log_file_name", -stderr  => ">>$log_file_name");

	&LogMsg ($fo_log_yn00,"END YN00\n");
	
	#If everything OK remove dir -> to be replaced
	if ($killed !=0)
	{
		my $log_msg = "WARNING: yn00 job has been killed - see $wkd/$log_file_name file \n";
		warn $log_msg;
		&LogMsg ($fo_log_yn00, $log_msg);
	}
	
	if ($status == 0 ) 
	{
		#Then parse rst file for results
		my $yn00_output = $rh_yn00_param -> {'output'};
		my $cutoff = $o_param -> Get ('s_cutoff');
		&LogMsg ($fo_log_yn00, "START PARSING Yn00 OUTPUT FILE $yn00_output.\n");	
			
		$rh_resYn00 = &ParseYn00Results($yn00_output, $cutoff);
		
		&LogMsg ($fo_log_yn00, "END PARSING.\n");
		&CleaningDirectory ($cwd, $wkd);
	}	
	else
	{
		my $log_msg = "WARNING: A problem occured in yn00 execution - see $wkd/$log_file_name file \n";
		warn $log_msg ;
		&LogMsg ($fo_log_yn00, $log_msg);
	}
	
	$fo_log_yn00 -> close();
	return $rh_resYn00, $status;
}

=head2 Procedure ParseYn00Results

 Title        : ParseYn00Results
 Usage        : &ParseYn00Results($file, $cutoff);
 Procedure    : Parses output file from yn00 file that contains all results and apply filter on synonymous sites if needed
 Args         : $file to parse and $cutoff for number of synonymous sites
 Returns      : $dn, $ds, $dn_ds values

=cut

sub ParseYn00Results
{
	my ($file_2_parse, $cutoff) = (@_);
	
	my $fi_file = &GetStreamIn( $file_2_parse ) or Carp::croak("Cannot open $file_2_parse file");
	my $dn = my $ds = my $dn_ds = my $Scount = undef;
	
	while (my $line = <$fi_file>)
	{
		if ( $line =~ /\S/)
		{
			if ($line =~ /\d/)
			{
				if ($line =~ /.*\+\-.*\+\-.*/)
				{		
					my @a_results = split(" ", $line);
					$dn = $a_results[7]; 
					$ds = $a_results[10]; 
					$dn_ds = $a_results[6];
					$Scount = $a_results[2];
					
					if ($Scount < $cutoff)
					{
						$dn = $ds = $dn_ds = $Scount = "NA";
						warn "WARNING: Number of synonymous site is $Scount is below treshold $cutoff\n";
						return undef;	
					}
					if ($dn_ds =~ /99.0000/)
					{
						#This means infinite and we have chosen to tag it as NAN
						$dn_ds = "NAN";
						warn "WARNING: dn/ds value is infinite. So we replace it with NAN \n";
					}
					next;	
				}
			}		
		}
	}
	
	#Uncomment to debug
	#print "S => $Scount, dN => $dn, dS => $ds, dn_ds => $dn_ds\n\n";
	
	my %h_resYn00 = ('S' => $Scount, 'dN' => $dn, 'dS' => $ds, 'dn_ds' => $dn_ds);

	return \%h_resYn00;
}


=head2 Procedure BuildCdsSeqHash

 Title        : BuildCdsSeqHash
 Usage        : &BuildCdsSeqHash($o_dba, $rh_headers);
 Procedure    : Gets DNA sequences from a hash and outputs a ref to an hash with ids and Bio::Seq objects
 Args         : A ref to a hash with cds sequences, and a ref to cds ids
 Returns      : A ref to a hash with cds sequences only

=cut

sub BuildCdsSeqHash
{
	my ($o_dba, $rh_headers) = @_;
	
	my %h_cds_seqs = ();
	my $rh_sequences = $o_dba -> GetRefSequenceHash();
	my @a_couple = keys (%$rh_headers);
	
	foreach my $seqID (@a_couple)
	{
		my $tempID = $rh_headers -> { $seqID };
		$h_cds_seqs { $tempID } = Bio::PrimarySeq->new ( 	-seq => $rh_sequences -> { $seqID } -> { cds },
														    -id  => $tempID);	
	}

	return \%h_cds_seqs;
}

=head2 Procedure BuildCdsMaskedSeqHash

 Title        : BuildCdsMaskedSeqHash
 Usage        : &BuildCdsMaskedSeqHash($rh_sequences);
 Procedure    : Gets DNA sequences from a odba, mask CpG islands and outputs a ref to an hash with ids and Bio::Seq objects
 Args         : A ref to a hash with cds sequences, and a ref to cds ids
 Returns      : A ref to a hash with masked cds sequences only

=cut

sub BuildCdsMaskedSeqHash
{
	my ($o_dba, $rh_headers, $o_param) = @_;
	
	my %h_cds_seqs = ();
	my @a_couple = keys (%$rh_headers);
	
	my $minlen = $o_param -> Get ('minlen');
	my $window = $o_param -> Get ('window');
	
	my $rh_sequences = $o_dba -> MaskCpG (\@a_couple, $minlen, $window);
	
	foreach my $seqID (@a_couple)
	{
		my $tempID = $rh_headers -> { $seqID };
		$h_cds_seqs { $tempID } = Bio::PrimarySeq->new ( 	-seq => $rh_sequences -> { $seqID },
															-id  => $tempID);	
	}
	
	return \%h_cds_seqs;
}


=head2 Procedure GetProteinIDs

 Title        : GetProteinIDs
 Usage        : &GetProteinIDs($o_param);
 Procedure    : Gets Protein IDs couples from input file and returns a ref to an array with all ref to array with couple members
 Args         : $o_param
 Returns      : a ref to an array that contains ref to array

=cut

sub GetProteinIDs
{
	my ( $o_param ) = (@_);
	print "\n****Reading the list of protein IDs****\n\n";
	my $fi_list = &GetStreamIn( $o_param -> Get ( 'protein' ) );
	my @a_tags = ();
	my $header = <$fi_list>; #first line
	
	while ( my $line = <$fi_list>)
	{
		chomp ($line);
		my @a_pair = split ("\t", $line); 
		push (@a_tags, \@a_pair);
	}
	my $ra_check = $a_tags[0];
	return \@a_tags;
}

=head2 Procedure InitParameters

 Title        : InitParameters
 Usage        : &InitParameters;
 Procedure    : Reads parameters from command line and parses yn00.ctl file 
 Args         : none
 Returns	  : $o_param object

=cut

sub InitParameters
{
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'protein=s', 'cfg=s','species=s', 'yn00=s', 
	's_cutoff=i', 'output=s','maskCpG','minlen=i', 'window=i', 'help');

	#Output file if defined or stdout as default
	$o_param -> SetUnlessDefined ('output', 'stdout');
	
	#Check the Narcisse cfg file of the instance
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param -> Get(&NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	$o_param -> SetUnlessDefined ('cfg', $ENV{NARCISSE_CFG});
	
	#Check mandatory parameters e.g. a file with couple of proteins (one per line), the species name, the configuration file of yn00.
	if ( !$o_param -> AssertDefined('protein') && !$o_param -> AssertDefined('species') && !$o_param -> AssertDefined('yn00'))
	{
		warn "ERROR - Species must be defined and Protein file, yn00.ctl file must not be empty \n\n";
		exit 1;
	}
	
	my $protein_path = abs_path ($o_param -> Get ('protein'));
	
	if ($o_param -> AssertFileExists ('protein'))
	{
		$o_param -> Set ('protein', $protein_path);
	}
	
	my $yn00_path = abs_path($o_param -> Get ('yn00')); #yn00 needs absolute path
	
	if ($o_param -> AssertFileExists ('yn00'))
	{
		$o_param -> Set ('yn00', $yn00_path);
	}
	
	#Setting yn00 parameters
	my $rh_yn00 = &BuildYn00Hash($o_param);
	$o_param -> Set ('yn00_param', $rh_yn00);
	
	#Setting Instance Observer
	my $o_instanceObserver = New InstanceObserver( -cfg => $o_param -> Get ('cfg'));
	$o_param -> Set ('o_io', $o_instanceObserver);
	
	#Getting Sequences from database
	my $o_dbaccess = &GetDataAccess( $o_param);
	$o_param -> Set ('o_dba',  $o_dbaccess);
	
	#Synonymous cutoff
	print "\n****Recalling options****\n\n";
	$o_param -> SetUnlessDefined ( 's_cutoff', 100);
	my $s = $o_param -> Get ('s_cutoff');
	if ($s == 0)
	{
		print "Cutoff on the synonymous sites S value: OFF\n";
	}
	else
	{
		print "Cutoff on the synonymous sites S value: ON with value = $s\n";
	}
	
	#CpG masking
	if ($o_param -> IsDefined('maskCpG'))
	{
		$o_param -> Set ('maskCpG', &TRUE);
		$o_param -> SetUnlessDefined ( 'minlen', 150);
		$o_param -> SetUnlessDefined ( 'window', 100);
		print "CpG Masking: ON \n";
	}
	else
	{
		print "CpG Masking: OFF\n";
		$o_param -> Set ('maskCpG', &FALSE);
	}
	
	print "\n****Checking environment****\n\n";
	exit 1 if (&CheckEnvPrograms);

	return $o_param;
}

=head2 Procedure CheckEnvPrograms

 Title        : CheckEnvPrograms
 Usage        : &CheckEnvPrograms;
 Procedure    : Checks if the programs yn00, newcpgreport, maskseq, clustalw are installed and prints corresponding messages
 Args         : none
 Returns      : $status 0 if everything is ok and 1 if their is a program missing

=cut

sub CheckEnvPrograms {
    
    my @a_progs = qw (clustalw yn00 newcpgreport maskseq);
    my $status = 0;
    foreach my $pgm (@a_progs)
    {
		print "Checking " . $pgm . " availability: ";
		if (grep { -x "$_/$pgm"}split /:/,$ENV{PATH})
		{
			print "OK...\n";
		}
		else
		{
			print "Not in PATH - Check if $pgm is installed...\n";
			$status = 1;
		}
	}
	return $status;
}

=head2 Procedure MakeIDlist

 Title        : MakeIDlist
 Usage        : &MakeIDlist ($ra_couple_ID);
 Procedure    : Takes a reference to an array couple IDs and returns a reference to an array with single IDs 
 Args         : none

=cut

sub MakeIDlist
{
	my ($ra_couple_IDs) = (@_);
	my @a_IDlist = ();
	foreach my $couple (@$ra_couple_IDs)
	{
		push (@a_IDlist, @$couple);
	}
	my @a_uniq_IDlist = &Uniq (@a_IDlist);
	return \@a_uniq_IDlist;	
}

=head2 Procedure Uniq

 Title        : Uniq
 Usage        : &Uniq;
 Procedure    : Remove redundancy in an array
 Args         : an array with redundant elements
 Returns      : an array with uniq elements
 Note 		  : Formule magique trouvée sur le web

=cut

sub Uniq {
    return keys %{{ map { $_ => 1 } @_ }};
}

=head2 Procedure LogMsg

 Title        : LogMsg
 Usage        : &LogMsg ($file_handler, $msg);
 Procedure    : Write in log file
 Args         : $file_handler : a file handler
				$msg : a message to write in the log file

=cut

sub LogMsg 
{
    my( $file_handler, $msg) = (@_);
    
    my $tag_time = "TIME AT THIS STEP: \t" . &NARCISSE_utils::GetTime()."\n";
    print $file_handler $tag_time;
    print $file_handler $msg;
    return;
}

=head2 Procedure CreateLogFile

 Title        : CreateLogFile
 Usage        : &CreateLogFile;
 Procedure    : Create log file
 Args         : $prefix
 Returns      : $filehandler file handler and log file name $log_file_name

=cut

sub CreateLogFile 
{
	my ($prefix) = (@_);
	my $log_file_name = $prefix . ".log"; 
	my $fo_log = &GetStreamOut($log_file_name) or Carp::croak("Cannot open $log_file_name file");
	return $fo_log, $log_file_name;
}

=head2 Procedure BuildYn00Hash

 Title        : BuildYn00Hash
 Usage        : &BuildYn00Hash($o_param);
 Procedure    : Gets yn00 parameters and outputs a ref to hash containing yn00 parameters from yn00.ctl file
 Args         : $o_param
 Returns      : A reference to a hash containing data and the output file name
 
=cut

sub BuildYn00Hash
{
	my ($o_param) = (@_);
	
	my $ctl_file = $o_param -> Get ('yn00');	
	my $fh_ctl = &GetStreamIn($ctl_file);
	my %h_yn00 = ();
	my $inputFile= my $outputFile = "";
	
	foreach my $line ( <$fh_ctl> )
	{	
		if ( $line =~ /\S/)
		{
			chomp($line);
			my $strip_line = &Strip($line);
			my @a_args = split ("=", $strip_line);
			if ($a_args[0] =~ /seqfile/)
			{
				$inputFile= (split("\\*", $a_args[1]))[0];
			}
			elsif ($a_args[0] =~/outfile/)
			{
				$outputFile = (split("\\*", $a_args[1]))[0];
				last;
			}
		}
	}
	
	%h_yn00 = (input => $inputFile, output => $outputFile);

	return \%h_yn00;
}

=head2 Procedure Strip

 Title        : Strip
 Usage        : &Strip($string);
 Procedure    : Removes whites spaces
 Args         : a string
 Return       : A trim string

=cut

sub Strip
{
	my ($string ) = (@_);
	$string =~ s/\s//g;
	return $string;
}

=head2 Procedure Usage

 Title        : Usage
 Usage        : &Usage;
 Procedure    : Displays script usage
 Args         : none

=cut

sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe [--protein], [--cfg], [--species], [--s_cutoff], [--yn00], [--output], [--help]

Requirements:   yn00 from PAML package
				clustalw program
				newcpgreport from EMBOSS package
				maskseq from EMBOSS package
				BioPerl

[mandatory]
    --protein=  	: The file that lists all suitable protein couples selected by ncs_PrepareInput.pl 
					  (One Ensembl or Narcisse ID per line. For example: ENSSSCG00000015179)
    --species=      : The species name (Use quotes - For example: "Sus scrofa")
    --yn00=    	    : The path to the yn00.ctl file
[options]
    --cfg=			: The narcisse config file if NARCISSE_CFG not in your environment PATH
    --s_cutoff=     : The value of the cutoff applied on number of synonymous sites (S) according to yn00 result
                      If not defined Set s_cutoff to 0 to avoid filtering data on S value
                      If defined Set s_cutoff to 100
                      --s=        : If you want to set to another value than 0 or 100
    --maskCpG       : Add this option if you want to mask CpG isalnds in sequences and add the following parameters
					--minlen=     : The minimum sequence length used by EMBOSS newcpgreport program (default = 150)
					--window=     : The window size used by EMBOSS newcpgreport program (default = 100)
    --output=		: The name of the output file displaying Ds, Dn and Dn/Ds values. 
					  If not specified, results will be displayed on stdout.
	--help=         : This help
END
;
}
