#!/usr/bin/perl
#
# $Id$
#
# Copyright INRA/CNRS

# Emmanuel.Permal@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 

 ncs_GetGeneInformation.pl - a script to get some Gene information to use with Ds data analysis

=head1  SYNOPSIS

 perl ncs_GetGeneInformation.pl @args

 @args:
 [mandatory]
  --output : the file name
 
=head1 DESCRIPTION

 ncs_GetGeneInformation.pl  = Each line will display the protein id, the gene id, the start, the end, the strand, 
 the chromosome and the amount of exon
 
example :

perl ncs_GetGeneInformation.pl --species='my species' --output=genes.tsv

=cut

=head1 SUBROUTINES

MAIN


=cut

use strict;
use warnings;

#Standard modules
use Error qw( :try );
use Data::Dumper;
use File::Basename;
use File::Path;
use File::Spec;
use Cwd;
use Cwd 'abs_path';
use IO::File;
use Pod::Usage;
use FindBin;
use File::Copy qw(move copy);
use File::Path qw( mkpath rmtree ); 
use File::Temp qw(tempfile tempdir);
use Env;

#Narcisse modules 
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

#Narcisse
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;
use Runner;
use NARCISSE_utils;
use NarcisseServiceFactory;
use NarcisseService;
use NarcisseService::GrepBased;

#Constants
use constant NARCISSE_CFG => 'cfg';

&TryMain(0);

MAIN:
{ 
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = ( @_ );
		warn $o_err-> stacktrace(); 
	}
	otherwise
	{
		my ( $o_err ) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";
		print STDERR $str;		    
	};
	
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{ 
	#Init Parameters
	print "***** Initialize parameters *****\n";
	my $o_param = &InitParameters();
	
	#Get Protein information
	my $rh_data = &GetProteinInfo ($o_param);

	#Compute the mean of all proteins
	my $size_avg = &ComputeProtSizeMean ($rh_data);
	
	#Write Gene information in file
	&WriteGeneFile ($rh_data, $o_param);
	
	#Print average
	print "\n\n***** Average of protein size: $size_avg nt *****\n";
	
}

=head2 Procedure WriteGeneFile

 Title        : WriteGeneFile
 Usage        : &WriteGeneFile();
 Procedure    : Write an output file with gene information
 Args         : 
 
=cut

sub WriteGeneFile
{
	my ($rh_prot_chr, $o_param) = (@_);
	
	my $genefilename = $o_param -> Get ('output');
	print "\n*****Writing gene information in $genefilename\n";
	
	my $fo_gene= &GetStreamOut( "$genefilename" );

	#Here read ref to hash parameters
	my @a_fields = qw (ProteinID GeneID index chr start end strand nbexon );
	print $fo_gene join("\t", @a_fields)."\n";
	my $nb_gene = 0;
	
	foreach my $chr (sort (keys (%$rh_prot_chr)))
	{
		my $rh_prot_info =$rh_prot_chr -> {$chr};
		#~
		foreach my $index (  sort {$a <=> $b} (keys (%$rh_prot_info)))
		{	
			my $outString = $rh_prot_info -> {$index} -> {'protein_id'} . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'gene_id'} . "\t";
			$outString .= $index . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'chrom'} . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'start'} . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'end'} . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'strand'} . "\t";
			$outString .= $rh_prot_info -> {$index} -> {'nb_exon'} . "\n";
			
			#Debug
			#print "STDOUT GENE $outString \n";
			print $fo_gene $outString;
		}
		
		$nb_gene += scalar keys (%$rh_prot_info);
	}
	
	print "\n\n***** Number of genes: $nb_gene *****\n";
	return;
}

=head2 Procedure GetProteinInfo

 Title        : GetProteinInfo
 Usage        : &GetProteinInfo();
 Procedure    : Get protein information
 Args         : 
 
=cut

sub GetProteinInfo
{
	my ($o_param) = (@_);
	
	my $o_io = $o_param -> Get ('o_io');
	my $o_spo = $o_io -> GetSpeciesObserver( $o_param -> Get ('species'));
	my @transcript_file = $o_spo ->GetTranscriptFiles();
	my $path_to_rep_prot = $o_spo -> GetProtFastaFile (-rep => &TRUE);
	#Get representative protein ids
	my $ra_rep_cds_ids = &GetRepresentative ($path_to_rep_prot);
	
	#Get gene information from transcrits file
	my %h_prot_info = ();
	foreach my $t_file (@transcript_file)
	{
		my $chr = &GetChrName ($t_file);
		$h_prot_info { $chr } = &ParseTranscriptFile( $t_file, $ra_rep_cds_ids);
	}
	return \%h_prot_info;
}

=head2 Procedure GetRepresentative

 Title        :GetRepresentative
 Usage        : &GetRepresentative();
 Procedure    : Get the name of the chromosome from file name
 Args         : $file_name
 
=cut

sub GetRepresentative
{
	my ($file_name) = (@_);
	
	my $fi_stream = &GetStreamIn ($file_name);
	my @a_rep_cds_ids =();
	foreach my $line (<$fi_stream>)
	{
		
		if ($line =~ />/)
		{
			chomp ($line);
			my @a_header = split (" ", $line);
			my $prot_id = $a_header[0];
			$prot_id =~ s/>prot://;
			push (@a_rep_cds_ids, $prot_id);
			
			#~print "PROT $prot_id\n";
		}
	}
	
	$fi_stream -> close();
	return \@a_rep_cds_ids;	
}

=head2 Procedure GetChrName

 Title        :GetChrName
 Usage        : &GetChrName();
 Procedure    : Get the name of the chromosome from file name
 Args         : $file_name
 
=cut

sub GetChrName
{
	my ($file_name) = (@_);
	
	my $prefix = (split ("\\.", $file_name))[0];
	$prefix =~ s/chromosome//;
	my $chr = (split ("\\/", $prefix))[-1];
	return $chr;	
}

=head2 Procedure ComputeProtSizeMean

 Title        :ComputeProtSizeMean
 Usage        : &ComputeProtSizeMean();
 Procedure    : Get 
 Args         : 
 
=cut

sub ComputeProtSizeMean
{
	my ($rh_data) = (@_);
	
	my $sum = 0;
	my $nbid = 0;
	foreach my $chr (keys (%$rh_data))
	{
		my $rh_prot_info = $rh_data -> {$chr};
		
		foreach my $id (keys (%$rh_prot_info))
		{
			my $end = $rh_prot_info -> {$id} -> {'end'};
			my $start = $rh_prot_info -> {$id} -> {'start'};
		
			{
				no warnings 'numeric'; #Removing warning here
				$sum += $end - $start + 1;
			}
		}
		$nbid += scalar keys (%$rh_prot_info);
	}
	
	my $mean = $sum / $nbid ;
	return $mean;
}


=head2 Procedure ParseTranscriptFile

 Title        : ParseTranscriptFile
 Usage        : &ParseTranscriptFile();
 Procedure    : Parses the transcript file to get information on protein, chromosome, coordinates, exon number and gene name
 Args         : 
 
=cut

sub ParseTranscriptFile
{
	my ($t_file, $ra_rep_cds_ids) = (@_);
	
	my $fi_file = &GetStreamIn ($t_file);
	
	print "Reading and parsing transcripts file $t_file\n";
	my $header_line = <$fi_file>;
	chomp ($header_line);
	my @a_headers = split ("\t", $header_line);
	
	my %h_prot_info = ();
	my $index = 0; #the file is ordered by start
	
	while (my $line = <$fi_file>)
	{
		my %h_line = ();
		chomp($line);
		my @a_values = split ("\t", $line);
		@h_line{@a_headers} = @a_values;
		$index += 1;
		
		#Here Check if the transcript is alone on prefix in id
		my $prot_id = $h_line {'protein_id'};
		
		my $isArep = &IsARepresentativeId($prot_id, $ra_rep_cds_ids);
		
		next if (!$isArep );
		#Return boolean condition to add new member in hash
		
		$h_prot_info {$index} = {	protein_id => $h_line {'protein_id'},
									chrom => $h_line {'chrom'},
									gene_id => $h_line {'symbol'},
									start => $h_line {'start'},
									end => $h_line {'end'},
									strand => $h_line {'strand'},
									nb_exon => $h_line {'nb_exons'},													
								};									
	}
	
	$fi_file -> close();

	return \%h_prot_info;
}

=head2 Procedure IsARepresentativeId

 Title        : IsARepresentativeId
 Usage        : &IsARepresentativeId();
 Procedure    : 
 Args         : $prot, $ra_rep_cds_id
 
=cut

sub IsARepresentativeId
{
	my ($prot_id, $ra_rep_id) = (@_);
	
	my @a_match = grep (/$prot_id/, @$ra_rep_id);
	
	if (scalar @a_match > 0)
	{
		return 1;
	}
	else
	{
		return 0;
	}

}



=head2 Procedure InitParameters

 Title        : InitParameters
 Usage        : &InitParameters;
 Procedure    : Reads parameters from command line
 Args         : none
 Returns	  : $o_param object

=cut

sub InitParameters
{
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'output=s', 'cfg=s', 'species=s','help');
	
	if ( !$o_param -> AssertDefined('species'))
	{
		warn "\nERROR - input file and/or ctl file as well as species name nare mandatory\n";
		&Usage();
		exit 1;
	}
	
	#Setting default values
	$o_param -> SetUnlessDefined ('output', 'stdout');
	
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param -> Get(&NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	$o_param -> SetUnlessDefined ('cfg', $ENV{NARCISSE_CFG});
	
	#Setting Instance Observer
	my $o_instanceObserver = New InstanceObserver( -cfg => $o_param -> Get ('cfg'));
	$o_param -> Set ('o_io', $o_instanceObserver);
	
	return $o_param;
}

=head2 Procedure Usage

 Title        : Usage
 Usage        : &Usage;
 Procedure    : Displays script usage
 Args         : none

=cut

sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe  [--species], [--cfg], [--output]

[mandatory]
    --species=      : The species name (Use quotes - For example: "Sus scrofa")
[options]
	--cfg=			: The narcisse config file if NARCISSE_CFG not in your environment PATH
    --output=		: The name of the output file displaying paraloop command lines.
					  If not specified, results will be displayed on stdout.
END

;
}
