#!/usr/bin/perl

=pod
=head1 

 ncs_GlintFish.pl - a script to select unique probes for a subsequent FISH analysis to identify a specific genome region

=head1  SYNOPSIS

 perl ncs_GlintFish.pl @args

 @args:
 [mandatory]
  --species : string, organism short or long name 
  --region : string, the region of interest, format chr:start-end 
 [optional]
  --target : string, path directory  (default = .)
  --chroms : string, a comma-separated list of chromosomes (ex: 1,3,4 or 1-3,5)
  --product_size_range : string, probe size range, PRIMER3 OPTION size range of the pcr product (ex 5000-8000)
  
  --verbose : flag (default: false)
  --debug  : flag (default: false)
  --opt_size : integer, optimal size for pcr primers (default = 20)
  --opt_tm : integer, optimal value for tm (default = 60)
  --probe_max_gc_percent : float, maximal GC% for probe (default=0.6)
  

=head1 DESCRIPTION
 ncs_GlintFish.pl = Create unique pcr probes to identify a specific genome region

This program first identifies suitable regions (with a limited number of repeated sequences), 
and launches primer3 on these sequences to identify long-range pcr primers. Finally the program
checks the specificity of the proposed probes

example :

perl ncs_GlintFish.pl


=cut
=head1 SUBROUTINES

MAIN
DoYourJob
GetRegionInfos
GetParameters
FindCfgFile

=cut

use strict;
use warnings;

use Cwd qw( abs_path getcwd );
use Carp;
use Data::Dumper;

use File::Spec;
use Error qw( :try );
use File::Temp qw( tempdir );
use Pod::Usage;
use FindBin qw( $Bin );

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);

	# bien mettre ../lib apres playmoby sinon pb avec ParamParser.
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;

use GlintFish;

#Som program specific constants

use constant SPECIES => 'species';
use constant REGION  => 'region' ;
use constant TARGET  => 'target' ;
use constant PARAMS  => 'params' ;
use constant CHROMS => 'chromosomes';
use constant GENOME => 'genome';
use constant NARCISSE_CFG => 'cfg';
use constant NO_CLEAN   => 'no_clean';
use constant GLINTFISH => 'GLINTFISH';
use constant GLINTFISH_PRIMER3 => 'GLINTFISH_PRIMER3_PRIMER';
use constant OUTPUT_PREFIX => 'prefix';

use constant FALSE => 0;
use constant DEBUG => 1;

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{	warn( $o_err->Explain(), "\n" );	}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}

sub Main
{
	my $rh_params = &GetParameters();
	my ( $species, $region, $chroms, $genome, $rh_param ) = @{ $rh_params }{ ( &SPECIES, &REGION, &CHROMS, &GENOME, &PARAMS ) };
		
	if ( defined $genome )
	{
		&DetectGenomeProbes( &SPECIES => $species,  &PARAMS => $rh_params );
		
	}
	elsif ( defined $chroms )
	{
		&DetectChromosomeProbes( &SPECIES => $species,  &CHROMS => $chroms, &PARAMS => $rh_params );		
	}
	elsif ( defined $region )
	{
		&DetectRegionProbes( -prefix => $rh_params->{ &OUTPUT_PREFIX }, &SPECIES => $species, &REGION => $region, &PARAMS => $rh_params );
	}	
	else
	{
		print STDERR "One of genome, chromosomes, region must be defined \n";	
	}
}

=head2 Procedure DetectGenomeProbes

 Title        : DetectGenomeProbes
 Usage        : DetectGenomeProbes();
 Function     : Detect the probes for all chromosomes
 Returns      : none
 Args         : species => the name of the species
                params  => a reference on a hash of parameters
=cut

sub DetectGenomeProbes
{
	my %h_args = @_;
	
	my ( $species, $rh_param ) = @h_args{ ( &SPECIES, &PARAMS ) };	
	
	my $o_instanceObserver = New InstanceObserver( -cfg => $rh_param->{ &NARCISSE_CFG } );
	my $o_speciesObserver = $o_instanceObserver->GetSpeciesObserver( $species );
	
	my $rh_chromsinfos = $o_speciesObserver->GetChromsInfo();
	
	foreach my $chrom ( keys %$rh_chromsinfos )
	{
		my ( $start, $end )	= ( 1, $rh_chromsinfos->{$chrom}{ length } );
		my $region = "$chrom:$start-$end";
		my $output_prefix = "$species$chrom"."_gf";
		print STDERR "Detecting probes in chrom $region\n";
		&DetectRegionProbes( &SPECIES => $species, 
		                     &REGION => $region, 
		                     &PARAMS => $rh_param, 
		                     -prefix => $output_prefix 
		                   );		
	}
	
}

=head2 Procedure DetectChromosomeProbes

 Title        : DetectChromosomeProbes
 Usage        : DetectChromosomeProbes();
 Function     : Detect the probes for a list of chromosomes
 Returns      : none
 Args         : species => the name of the species
                chromosomes => the list of chromosomes as a string
                params  => a reference on a hash of parameters
=cut
sub DetectChromosomeProbes
{
	my %h_args = @_;
	
	my ( $species, $chroms, $rh_param ) = @h_args{ ( &SPECIES, &CHROMS, &PARAMS ) };	
		
	my @a_chromosomes = &GetChromosomes( $chroms );
	
	my $o_instanceObserver = New InstanceObserver( -cfg => $rh_param->{ &NARCISSE_CFG } );
	my $o_speciesObserver = $o_instanceObserver->GetSpeciesObserver( $species );
	
	my $rh_chromsinfos = $o_speciesObserver->GetChromsInfo();
	
	foreach  my $chrom ( @a_chromosomes )
	{	
		my ( $start, $end )	= ( 1, $rh_chromsinfos->{$chrom}{ length } );
		my $region = "$chrom:$start-$end";
		my $output_prefix = "$species$chrom"."_gf";
		print STDERR "Detecting probes in chrom $region\n";
		&DetectRegionProbes( &SPECIES => $species, 
		                     &REGION => $region, 
		                     &PARAMS => $rh_param, 
		                     -prefix => $output_prefix 
		                   );			
	}
	
}

=head2 Procedure DetectRegionProbes

 Title        : DetectRegionProbes
 Usage        : DetectRegionProbes();
 Function     : Detect the probes for a specific region
 Returns      : none
 Args         : species => the name of the species
                params  => a reference on a hash of parameters
=cut
sub DetectRegionProbes
{
	my %h_args = ( -prefix => 'glintFish',
				   @_
				  );
	
	my ( $species, $region, $rh_param ) = @h_args{ ( &SPECIES, &REGION, &PARAMS ) }; 
	
	my $rh_region = &GetRegionInfos( $region );
	my ( $chrom, $start, $end ) = @{$rh_region}{ qw( chrom start end )};
		
	my $o_instanceObserver = New InstanceObserver( -cfg => $rh_param->{ &NARCISSE_CFG }  );	
	
	my $o_probeSelection = New GlintFish(	chr          => $chrom, 
										    start        => $start, 
										    end          => $end,
										    species      => $species,
										    instanceObs  => $o_instanceObserver,
										    -target      => $rh_param->{ target },
										    -no_clean    => $rh_param->{ no_clean },
										    -prefix      => $h_args{ -prefix },
										    -param       => $rh_param
								          );
					          
	my $error_in_region_param = $o_probeSelection -> DisplayErrors;

	if ($error_in_region_param eq "000"){
		
		$o_probeSelection->Run();
		$o_probeSelection->ProbeOutput();	
		$o_probeSelection->ProbeOutputNarcisseFormat();	
	}
	else
	{
		print $error_in_region_param;
	}					    
}


=head2 function GetChromosomes

 Title        : GetChromosomes
 Usage        : GetChromosomes();
 Function     : Creates an array of chroms from the list of chromosomes (format : a comma-separated list of chromosomes )
 Returns      : a reference on an array of chromosomes
 Args         : string, the list of chromosome provided in argument 
=cut
sub GetChromosomes
{
	my ( $chroms )  = @_;	
	
	my @a_range_chroms = split( ":", $chroms );
	my @a_AllChroms = ();
	
	foreach my $range_chrom ( @a_range_chroms )
	{
		if ( $range_chrom =~/(\d+)-(\d+)/ )
		{
			my ( $first, $last ) = ( $1, $2);			
			my @a_chroms = ($first..$last);
			push @a_AllChroms, @a_chroms;
		}
		else
		{
			push @a_AllChroms, $range_chrom;
		}
	}
	
	return @a_AllChroms;
}

=head2 function GetRegionInfos

 Title        : GetRegionInfos
 Usage        : GetRegionInfos();
 Function     : Parse the region infos
 Returns      : a hash with corresponding information
 Args         : string,  format chr:start-end 
=cut
sub GetRegionInfos
{
	my ( $region ) = @_;
	
	my %h_region = ();
	my $ok = 0;
	
	if ( $region =~ /^([XY\d]+):(\d+)\-(\d+)$/ )
	{
		@h_region{qw( chrom start end)} = ($1,$2,$3) ;
		$ok = 1 if ( $h_region{ start } < $h_region{ end } );
	}
	elsif ( $region =~/^([XY\d]+):$/ )
	{
		@h_region{qw( chrom start end)} = ($1,undef,undef) ;
		$ok = 1
	}
	$ok or die 'Wrong region format, please see usage';
	
	return \%h_region;
}

sub GetParameters
{
	&ParamParser::SetDefaultBehaviour( 'use_substitution_table' );
	&ParamParser::SetDefaultUsage( sub { pod2usage(1) } );
	
	my @a_glintfish_params = ( 'probe_max_gc_percent=s','conseq_unfreq=i', 'conseq_freq=i', 'count_cutoff=i', 'condemned_min_size=i', 'suitable_min_size=i',  'probe_nb=i', 'select_threshold=i');
	my @a_primer3_params = ('product_size_range=s', 'opt_size=i','opt_tm=i' );
	my @a_cmdline_params= ( 'debug', 'force', 'select', 'versus', 'verbose', 'unmasked' );
		
	my @a_keys   = qw( help );
	push( @a_keys, &SPECIES . '=s', &REGION . '=s', &TARGET . '=s', &CHROMS ."=s", &GENOME, &NARCISSE_CFG . "=s", &NO_CLEAN );
	push( @a_keys, &OUTPUT_PREFIX . '=s' );
	push( @a_keys, @a_glintfish_params );
	push( @a_keys, @a_primer3_params);
	push( @a_keys, @a_cmdline_params );
	push( @a_keys, 'evaluate=s' );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );
	
	#Species obligatory parameter
	$o_param->AssertDefined( &SPECIES );
	#$o_param->AssertDefined( &REGION );
	#$o_param->SetUnlessDefined( &NARCISSE_CFG => undef);
	$o_param->SetUnlessDefined( &NO_CLEAN  => 0 );
	$o_param->SetUnlessDefined( &OUTPUT_PREFIX => 'glintFish' );
	
	#Target directory
	if ( $o_param->IsDefined( &TARGET ) )
	{
		$o_param->AssertDirExists( &TARGET );
	}
	
	foreach my $glintfish_param ( @a_glintfish_params )
	{
		$glintfish_param =~ s/\=[is]//;		
		if ( $o_param->IsDefined( $glintfish_param ) )	
		{
			my $newParamName = &GLINTFISH."_\U$glintfish_param";
			$o_param->Set( $newParamName, $o_param->Get( $glintfish_param ) );
			$o_param->Delete(  $glintfish_param );
		}
	}
	
	foreach my $primer3_param ( @a_primer3_params )
	{
		$primer3_param =~ s/\=[is]//;
		if ( $o_param->IsDefined( $primer3_param ) )	
		{
			my $newParamName = &GLINTFISH_PRIMER3."_\U$primer3_param";
			$o_param->Set( $newParamName, $o_param->Get( $primer3_param ));
			$o_param->Delete( $primer3_param );
		}
	}
	
	
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param->Get( &NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	
	#print STDERR "$ENV{NARCISSE_CFG} \n";
	
	$o_param->SetUnlessDefined( 'force', &FALSE );
	$o_param->SetUnlessDefined( 'verbose', &FALSE );
	$o_param->SetUnlessDefined( 'debug', &FALSE );
		
	#Returning a hash
	my $rh_params = {};
	$o_param->Dump( 'HASH', $rh_params );
	
	return $rh_params;
}
