#!/usr/bin/perl
#
# $Id$
#
# Copyright INRA/CNRS

# Emmanuel.Permal@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 

 ncs_PrepareInput.pl - a script to prepare input for Ds computing using Narcisse API

=head1  SYNOPSIS

 perl ncs_PrepareInput.pl @args

 @args:
 [mandatory]
  --species : the species name
 
=head1 DESCRIPTION

 ncs_PrepareInput.pl  = Using Narcisse API select protein data from blast file for dS computing

example :

perl ncs_PrepareInput.pl --species=my_species 

=cut

=head1 SUBROUTINES

MAIN


=cut

use strict;
use warnings;

#Standard modules
use Error qw( :try );
use Data::Dumper;
use File::Basename;
use File::Path;
use File::Spec;
use Cwd;
use Cwd 'abs_path';
use IO::File;
use Pod::Usage;
use FindBin;
use File::Copy qw(move copy);
use File::Path qw( mkpath rmtree ); 
use File::Temp qw(tempfile tempdir);
use Env;

#Narcisse modules 
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

#Narcisse
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;
use Runner;
use NARCISSE_utils;
use NarcisseServiceFactory;
use NarcisseService;
use NarcisseService::GrepBased;

#Constants
use constant NARCISSE_CFG => 'cfg';

&TryMain(0);

MAIN:
{ 
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = ( @_ );
		warn $o_err-> stacktrace(); 
	}
	otherwise
	{
		my ( $o_err ) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";
		print STDERR $str;		    
	};
	
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{ 
	#Init Parameters
	print "Initialize parameters\n";
	my $o_param = &InitParameters();
	
	#Filter proteins (identity, coverage, protein AA length)
	my @a_selectedpairs = &FiltersProteinOnHsp ($o_param);
	
	##Assign pair to family
	#&AssignFamilies();
	
	&PrintOutput ($o_param, \@a_selectedpairs);
		
}

=head1 SUBROUTINES

=cut

=head2 Procedure FiltersProteinOnHsp

 Title        : FiltersProteinOnHsp
 Usage        : &FiltersProteinOnHsp;
 Procedure    : Reads blastp line and apply filter
 Args         : $o_param
 Returns	  : returns an array of references to array that contain gene pairs

=cut

sub FiltersProteinOnHsp
{
	my ($o_param) = (@_);
	
	#Get represntative protein info
	my $rh_rep_ids = &GetRepIDs( $o_param );
	
	#Table with selcetd protein IDs
	my @a_selected=();
	
	#Get cutoff values
	my $ident_cutoff = $o_param -> Get ('ident');
	my $cov_cutoff = $o_param -> Get ('cov');
	my $len_cutoff = $o_param -> Get ('length');
	
	#Get data file
	print "Opening file\n";
	my $fh_input = &GetStreamIn($o_param -> Get ('blast'));
	
	#Retrieve sel-comparison file
	print "Reading and parsing file\n";
	my $header_line = <$fh_input>;
	chomp ($header_line);
	my @a_headers = split ("\t", substr ($header_line, 1));
	
	#my $count = 0; #Uncomment to debug
	while (my $line = <$fh_input>)
	{
		my %h_line= ();
		chomp ($line);
		
		my @a_values = split ("\t", $line);
		
		@h_line{@a_headers} = @a_values;
		
		#We do not need protein against itself couple
		next if (&CheckSelfProtein (\%h_line));
		
		#Check if representative
		next if (&CheckRepresentative (\%h_line, $rh_rep_ids));
		
		#Protein needs to pass identity filter
		next if (&IdentityFilter ($ident_cutoff, \%h_line)) ;
	
		#Protein needs to pass coverage filter
		next if (&CoverageFilter ($cov_cutoff, \%h_line));
		
		#Protein needs to pass length filter
		next if (&LengthFilter ($len_cutoff, \%h_line));

		#Check if the reciprocal pair has already been selected
		next if (&CheckDuplicates (\%h_line, \@a_selected));
		
		#My line has passed all filters I want to the gene pair IDs
		my $ra_genepair = &GetGenePair (\%h_line);
	
		#Here do mapping
		push (@a_selected, $ra_genepair);
		
		#$count ++; #Uncomment to debug
		#last if ($count == 10); #Uncomment to debug
	}
	
	return @a_selected;
}

=head2 Procedure PrintOutput

 Title        : PrintOutput
 Usage        : &PrintOutput($o_param, $ra_pairs);
 Procedure    : prints output file wih one couple of protein per line separated by a tabulation
 Args         : $o_param and a reference to an array with the selected couple ids
 Returns	  : none

=cut

sub PrintOutput
{
	my ($o_param, $ra_pairs) = (@_);
	
	my $fh_out = &GetStreamOut ($o_param -> Get ('output'));
	my @a_fields = qw ( ID1 ID2);
	print $fh_out join("\t", @a_fields)."\n";

	foreach my $ra_gene_pair (@$ra_pairs)
	{
		print $fh_out join("\t", @$ra_gene_pair)."\n";	
	}
}

=head2 Procedure CheckDuplicates

 Title        : CheckDuplicates
 Usage        : &CheckDuplicates($line, $ra_selected);
 Procedure    : check if the reciprocal gene pair has been alreday selected
 Args         : $blast_line and a ref to an array of selected ids
 Returns	  : 0 or 1

=cut

sub CheckDuplicates
{
	my ($rh_line, $ra_selected) = (@_);
	
	my $firstID = $rh_line -> {'q_name'};
	my $secondID = $rh_line -> {'s_name'};
	
	foreach my $ra_gp (@$ra_selected)
	{
		my $gp1 = $ra_gp -> [0];		
		my $gp2 = $ra_gp -> [1];

		if ($firstID eq $gp1)
		{
			if ($secondID eq $gp2)
			{
				return 1; #True = This pair is already selected
			}
		}
		if ($firstID eq $gp2)
		{
			if ($secondID eq $gp1)
			{
				return 1; #True = This pair is already selected
			}
		}
	}	
	return 0; #False = This pair is not in the table of selected pairs
}

=head2 Procedure CheckSelfProtein

 Title        : CheckSelfProtein
 Usage        : &CheckSelfProtein ($line);
 Procedure    : Reads blastp line and returns 0 or 1 depending if the protein has matched against itself
 Args         : $blast_line
 Returns	  : 0 or 1

=cut

sub CheckSelfProtein
{
	my ($rh_line) = (@_);
	
	my $firstID = $rh_line -> {'q_name'};
	my $secondID = $rh_line -> {'s_name'};

	if ($firstID eq $secondID)
	{ 
		return 1; #true = they are the same
	}	
	return 0; #false = they are not the same
}

=head2 Procedure CheckRepresentative

 Title        : CheckRepresentative
 Usage        : &CheckRepresentative ($line);
 Procedure    : Reads blastp line and returns 0 or 1 depending if the protein has matched against itself
 Args         : $blast_line
 Returns	  : 0 or 1

=cut

sub CheckRepresentative
{
	my ($rh_line, $rh_repids) = (@_);
	
	my @a_IDs = @$rh_line { qw (q_name s_name) };
	
	foreach my $id (@a_IDs)
	{
		return 1 if ( ! (exists $rh_repids -> {$id})); #true = not representative
	}
	return 0; #false = representative
}

=head2 Procedure GetGenePair

 Title        : GetGenePair
 Usage        : &GetGenePair ($line);
 Procedure    : Reads blastp line and returns a gene pair
 Args         : $blast_line
 Returns	  : a ref to an array that contains the two gene ids

=cut

sub GetGenePair
{
	my ($rh_line) = (@_);
	my @a_gene_pair =();
	my $gene1= $rh_line -> {'q_name'};
	my $gene2 = $rh_line -> {'s_name'};
	
	push (@a_gene_pair, $gene1);
	push (@a_gene_pair, $gene2);
	
	return \@a_gene_pair;
}

=head2 Procedure GetRepIDs

 Title        : GetRepIDs
 Usage        : &GetRepIDs ($line);
 Procedure    : Reads rep file and returns a list of rep IDs
 Args         : $o_param
 Returns	  : a ref to an array that contains the rep protein ids

=cut

sub GetRepIDs
{
	my ($o_param) = (@_);
	my %h_rep_ids =();
	
	my $file2parse = $o_param -> Get ('rep');
	my $fi_rep = &GetStreamIn ($file2parse);
	
	foreach my $line (<$fi_rep>)
	{
		if ($line =~ />/)
		{
			my $id	 = (split(" ", $line))[0];
			$id =~ s/>prot://;
			$h_rep_ids{$id} =(); #bidouille
		}
	}	
	
	return \%h_rep_ids;
}

=head2 Procedure IdentityFilter

 Title        : IdentityFilter
 Usage        : &IdentityFilter($cutoff, $line);
 Procedure    : Reads blastp line and apply filter
 Args         : $blast_line and cutoff
 Returns	  : empty $line if below cutoff, the $line give as argument if not

=cut

sub IdentityFilter
{
	my ($cutoff, $rh_line) = (@_);
	
	my $p_ident = $rh_line -> {'pident'};
	if ($p_ident < $cutoff)
	{
		return 1; #true = it is below cutoff
	}
	return 0 #false  = it is above cutoff
}

=head2 Procedure CoverageFilter

 Title        : CoverageFilter
 Usage        : &CoverageFilter($cutoff, $line);
 Procedure    : Reads blastp line and apply filter
 Args         : $blast_line and cutoff
 Returns	  : empty $line if below cutoff, the $line give as argument if not

=cut

sub CoverageFilter
{
	my ($cutoff, $rh_line) = (@_);
	
	my $q_cov = $rh_line -> {'q_cov'};
	my $s_cov = $rh_line -> {'s_cov'};
	
	if ($q_cov < $cutoff)
	{
		return 1; #True : below cov cutoff
	}
		
	if ($s_cov < $cutoff)
	{
		return 1; #True : below cov cutoff
	}
	return 0; #False : above cov cutoff
}

=head2 Procedure LengthFilter

 Title        : LengthFilter
 Usage        : &LengthFilter($cutoff, $line);
 Procedure    : Reads blastp line and apply filter
 Args         : $blast_line and cutoff
 Returns	  : empty $line if below cutoff, the $line give as argument if not

=cut

sub LengthFilter
{
	my ($cutoff, $rh_line) = (@_);
	
	my $q_len = $rh_line -> {'q_len'};
	my $s_len = $rh_line -> {'s_len'};
	return 1 if ($q_len < $cutoff); #True : below cov cutoff	
	return 1 if ($s_len < $cutoff); #True : below cov cutoff
	return 0;  #False : above cov cutoff
}

=head2 Procedure InitParameters

 Title        : InitParameters
 Usage        : &InitParameters;
 Procedure    : Reads parameters from command line
 Args         : none
 Returns	  : $o_param object

=cut

sub InitParameters
{
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'species=s', 'cfg=s','output=s', 'ident=i', 'cov=i', 'length=i', 'help');
	
	if ( !$o_param -> AssertDefined('species') )
	{
		warn "\nERROR - species parameter is mandatory\n";
		&Usage();
		exit 1;
	}

	#Setting default values
	$o_param -> SetUnlessDefined ('output', 'stdout');
	$o_param -> SetUnlessDefined ('ident', 35);
	$o_param -> SetUnlessDefined ('cov', 70);
	$o_param -> SetUnlessDefined ('length', 50);
	
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param -> Get(&NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	$o_param -> SetUnlessDefined ('cfg', $ENV{NARCISSE_CFG});

	my $o_instanceObserver = New InstanceObserver( -cfg => $o_param -> Get ('cfg'));
	$o_param -> Set ('o_io', $o_instanceObserver);
	my $sp = $o_param -> Get ('species');
	my $o_speciesObserver = $o_instanceObserver -> GetSpeciesObserver( $sp);
	
	#Try with spo to get blast file path
	my $short_name = $o_speciesObserver -> GetShortName();
	my $path_data = $o_speciesObserver -> GetComparisonPath( method => 'genomic', target => $short_name);
	my $blast_file = $path_data . '/blastp_'. $short_name . '_' . $short_name . '.out.gz';
	
	if (-e $blast_file)
	{
		$o_param -> Set ('blast', $blast_file);	
	}
	else
	{
		warn "The file $blast_file does not exist\n. Exiting.\n";
		exit 0 ;
	}
	
	#Get representative file information
	my $rep_file = $o_speciesObserver -> GetProtFastaFile (-rep => &TRUE);
	$o_param -> Set ('rep', $rep_file);
	
	return $o_param;
}

=head2 Procedure Usage

 Title        : Usage
 Usage        : &Usage;
 Procedure    : Displays script usage
 Args         : none

=cut

sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe [--cfg], [--species], [--ident], [--cov], [--length], [--output]

[mandatory]
    --species=      : The species name (Use quotes - For example: "Sus scrofa")
   
[options]
    --cfg=			: The narcisse config file if NARCISSE_CFG not in PATH
    --ident=        : Cutoff for identity percentage (default = 35% - Ezawa et al. 2011)
    --cov=          : Cutoff for minimal coverage (default = 70% - Rizzon et al. 2006)
    --length=       : Cutoff for sequence length in AA (default = 50 Aa - Pan et al. 2007)
    --output=		: The name of the output file displaying selected couple of proteins
					  If not specified, results will be displayed on stdout.
    
END

;
}
