#!/usr/bin/perl
#
# $Id$
#
# Copyright INRA/CNRS

# Emmanuel.Permal@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 

 ncs_PrepareParaloop.pl - a script to prepare input for paraloop e.g. a .shell file with all perl command lines needed by paraloop

=head1  SYNOPSIS

 perl ncs_PrepareParaloop.pl @args

 @args:
 [mandatory]
  --output : the file name
 
=head1 DESCRIPTION

 ncs_PrepareParaloop.pl  = Write each need command lines in a file. Each line is for one protein couple. And will compute the Ds value.
 This file will be used by paraloop and is given as argument to paraloop via the .cfg file

example :

perl ncs_PrepareParaloop.pl --input=protein_couple.list --output=paraloop_input.shell 

=cut

=head1 SUBROUTINES

MAIN


=cut

use strict;
use warnings;

#Standard modules
use Error qw( :try );
use Data::Dumper;
use File::Basename;
use File::Path;
use File::Spec;
use Cwd;
use Cwd 'abs_path';
use IO::File;
use Pod::Usage;
use FindBin;
use File::Copy qw(move copy);
use File::Path qw( mkpath rmtree ); 
use File::Temp qw(tempfile tempdir);
use Env;

#Narcisse modules 
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

#Narcisse
use LipmError;
use General;
use ParamParser;
use Logger;
use Configuration::InstanceObserver;
use Runner;
use NARCISSE_utils;
use NarcisseServiceFactory;
use NarcisseService;
use NarcisseService::GrepBased;

#Constants
use constant NARCISSE_CFG => 'cfg';

&TryMain(0);

MAIN:
{ 
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = ( @_ );
		warn $o_err-> stacktrace(); 
	}
	otherwise
	{
		my ( $o_err ) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";
		print STDERR $str;		    
	};
	
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{ 
	#Init Parameters
	my $o_param = &InitParameters();
	
	#CreateTempDir
	my ($savedir, $wkdir) = &CreateDir ("Paraloop_files");
	
	#Split input file in multiple files
	my $ra_file_names = &SplitInputFile($o_param);
	
	#Filter proteins (identity, coverage, protein AA length)
	my $ra_comandLines = &BuildCommandLines($o_param, $ra_file_names);
	
	#Return to root directory
	chdir ($savedir);
	
	#PrintOutput
	&PrintOutput ($o_param, $ra_comandLines);	
}

=head1 SUBROUTINES

=cut

=head2 Procedure BuildCommandLines

 Title        : BuildCommandLines
 Usage        : &BuildCommandLines($o_param, $ra_files );
 Procedure    : Prepare commend lines and fill an array with them
 Args         : $o_param to get parameters, $ra_files a ref to an array with input file names
 Returns	  : $ra_cmd a ref to an array that contain all command lines

=cut

sub BuildCommandLines
{
	my ($o_param, $ra_files ) = (@_);
	
	my $species = $o_param -> Get ('species');
	my $cutoff = $o_param -> Get ('cutoff');
	my $ctlFile = $o_param -> Get ('ctl');
	my $cfg =  $o_param -> Get ('cfg');
	my $cpg_bool = $o_param -> Get ('maskCpG');
	
	my @a_cmd = ();
	my $cmd_opt = "--cfg=$cfg --species='$species' --yn00=$ctlFile --s_cutoff --s=$cutoff ";
	if ($cpg_bool)
	{
		my $minlen = $o_param -> Get ('minlen');
		my $window = $o_param -> Get ('window');
		$cmd_opt .= "--maskCpG --minlen=$minlen --window=$window ";
	}
	foreach my $input (@$ra_files)
	{
		my $output = &CreateDsOutputName ($input);
		my $cmd = "perl ncs_ComputeDs.pl --protein=$input " . $cmd_opt;
		$cmd .= "--output=$output \n";
		push (@a_cmd, $cmd);
	}
	
	return \@a_cmd;	
}

=head2 Procedure CreateDsOutputName

 Title        : CreateDsOutputName
 Usage        : &CreateDsOutputName ($input);
 Procedure    : Takes the prefix of the input file name and add a suffix to it to build the output file name
 Args         : $input the input file name
 Returns	  : $output file name

=cut

sub CreateDsOutputName
{
	my ($input) = (@_);
	my $prefix = (split ("\\.", $input))[0];
	my $output = $prefix ."_dS.txt";
	return $output;
}

=head2 Procedure PrintOutput

 Title        : PrintOutput
 Usage        : &PrintOutput($o_param, $ra_lines);
 Procedure    : Prints all command lines to an output file that will be used as input for paraloop
 Args         : $o_param and a reference to the array of commandlines
 Returns	  : none

=cut

sub PrintOutput
{
	my ($o_param, $ra_lines) = (@_);
	
	my $fh_out = &GetStreamOut ($o_param -> Get ('output'));
	
	foreach my $command (@$ra_lines)
	{
		print $fh_out $command;	
	}
	return;
}

=head2 Procedure SplitInputFile

 Title        : SplitInputFile
 Usage        : &SplitInputFile($o_param);
 Procedure    : Reads the input file with all gene pairs and outputs as many files as there were lines in it
 Args         : $o_param
 Returns	  : $ra_file_names a ref to an array that contains all the input filenames to command lines

=cut

sub SplitInputFile
{
	my ($o_param) = (@_);
	my $cwd = cwd;
	my $inputFile = $o_param -> Get ('input');
	my $prefix_path = (split ("\\/", $inputFile))[-1];
	my $prefix = $cwd."/".(split ("\\.", $prefix_path))[0];
	
	my $fh_in = &GetStreamIn ($inputFile);
	my @a_lines = <$fh_in>;
	my $nb_lines = scalar @a_lines;
	my $index = 0;
	
	my @a_file_names =();
	
	while ($index < $nb_lines)
	{
		my $line = $a_lines[$index];
		my $outputFile = $prefix . "_" . $index. ".txt";
		my $fh_out = &GetStreamOut ($outputFile);
		print $fh_out $line;
		$fh_out -> close();
		if (-e $outputFile)
		{
			push (@a_file_names, $outputFile);
		}
		else
		{
			warn "The file $outputFile has not been created!\n";
		}
		#print "INDEX $index\n";
		$index += 1;
	}
	return \@a_file_names;
}

=head2 Procedure CreateDir

 Title        : CreateDir
 Usage        : &CreateDir( $working_dir);
 Procedure    : Creates a working directory and cd into it
 Args         : $working_dir name of the dir
 Returns      : the $working_dir the created working directory path
              : the $sav_dir the current directory
 
=cut

sub CreateDir
{
	my ( $working_dir) =(@_);
	my $save_dir = cwd;
	mkdir( $working_dir );
	chdir( $working_dir );
	#~print "Creating the work directory $working_dir\n";
	return $save_dir, $working_dir;
}

=head2 Procedure InitParameters

 Title        : InitParameters
 Usage        : &InitParameters;
 Procedure    : Reads parameters from command line
 Args         : none
 Returns	  : $o_param object

=cut

sub InitParameters
{
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'input=s','output=s', 'ctl=s', 'species=s', 'cutoff', 's=i', 'maskCpG','minlen=i','window=i', 'help');
	
	if ( !$o_param -> AssertDefined('input') || !$o_param -> AssertDefined('ctl') || !$o_param -> AssertDefined('species'))
	{
		warn "\nERROR - input file and/or ctl file as well as species name nare mandatory\n";
		&Usage();
		exit 1;
	}
	my $input_path = abs_path ($o_param -> Get ('input'));
	if ($o_param -> AssertFileExists ('input'))
	{
		$o_param -> Set ('input', $input_path);
	}
	
	my $ctl_path = abs_path ($o_param -> Get ('ctl'));
	if ($o_param -> AssertFileExists ('ctl'))
	{
		$o_param -> Set ('ctl', $ctl_path);
	}

	#Setting default values
	$o_param -> SetUnlessDefined ('output', 'stdout');
	#$o_param -> SetUnlessDefined ('cutoff', 100);
	
	if ( !defined($ENV{NARCISSE_CFG}) && ! (-f $o_param -> Get(&NARCISSE_CFG )) )
	{
		die "Environment variable NARCISSE_CFG not defined and no valid cfg parameter provided"; 
	}
	$o_param -> SetUnlessDefined ('cfg', $ENV{NARCISSE_CFG});
	
	#Checking options
	
	if ($o_param -> IsDefined('cutoff')) 
	{
		$o_param -> Set ( 'cutoff', 100);	
	}
	else
	{
		$o_param -> Set ('cutoff', 0);
	}
	if ($o_param -> IsDefined('s'))
	{
		$o_param -> Set ('cutoff', $o_param ->Get ('s'));
	} 
	
	#CpG masking
	if ($o_param -> IsDefined('maskCpG'))
	{
		$o_param -> Set ('maskCpG', &TRUE);
		$o_param -> SetUnlessDefined ( 'minlen', 150);
		$o_param -> SetUnlessDefined ( 'window', 100);
	}
	else
	{
		$o_param -> Set ('maskCpG', &FALSE);
	}
	
	return $o_param;
}

=head2 Procedure Usage

 Title        : Usage
 Usage        : &Usage;
 Procedure    : Displays script usage
 Args         : none

=cut

sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe [--input], [--ctl], [--species], [--cutoff], ,[--maskCpG], [--output]

[mandatory]
    --input=      : The file name with all couple selected by ncs_PrepareInput.pl
    --species=      : The species name (Use quotes - For example: "Sus scrofa")
    --ctl=    	    : The path to the yn00.ctl file
   
[options]
	--cfg=			: The narcisse config file if NARCISSE_CFG not in your environment PATH
    --cutoff=     : The value of the cutoff applied on number of synonymous sites (S) according to yn00 result
                      If not defined Set s_cutoff to 0 to avoid filtering data on S value
                      If defined Set s_cutoff to 100
                      --s=        : If you want to set to another value than 0 or 100
    --maskCpG       : Add this option if you want to mask CpG islands in sequences and add the following parameters
					--minlen=     : The minimum sequence length used by EMBOSS newcpgreport program (default = 150)
					--window=     : The window size used by EMBOSS newcpgreport program (default = 100)
    --output=		: The name of the output file displaying paraloop command lines.
					  If not specified, results will be displayed on stdout.
END

;
}
