#!/usr/bin/perl

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

 ncs_karyotype.pl - a script to build a karyotype image.

=head1 SYNOPSIS

 perl ncs_karyotype.pl @args

 @args:
  --org : string, organism short or long name
  --elements : structures like chrom,start,end[,label]
               each structure is separated by space
               start and end unit can be base (default), kb (1_000b) or Mb (1_000kb)
  --elements_file : string, name of a file containing one structure (cf above) per line
                    note, column separator can be a tabulation.
  --output : output filename, default is karyotype.png

 example :
  to draw a human karyotype and place onto it a structure on chrom 2,
  from 1_000_000 to 1.5Mb with the label 'example' and another on chrom X
  from 500kb to 750kb

  perl ncs_karyotype --org human --elements '2,1000000,1.5Mb,example X,500kb,750kb'

 Attention : chrom et label mustn't have space or coma.

=head1 DESCRIPTION

 This script build a karyotpe image of an organism given in parameter.
 It can place onto it any element given in elements arg.

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: February 12, 2009
 $Id: ncs_karyotype.pl 3142 2009-12-22 12:38:52Z manu $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Sequence_utils;
use Args_utils;
use KaryotypeBuilder;   <--- maintenant KaryotypeMaker : code a verifier
use Configuration::InstanceObserver;
use NARCISSE_utils;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3142 $ =~ /\d+/g); $r[0]};

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{	warn( $o_err->Explain(), "\n" );	}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $o_log = New Logger( 'ncs_karyo.log', 'ICVIKVIBVIEV' );

	# Init Instance Observer
    my $o_io = New InstanceObserver();

	my $rh_params = &GetParameters($o_io);
# 	warn Dumper $rh_params;

	my $rh_padding   = { inner_height => 30, top => 30, bottom => 10 };

	my $o_karyo = New KaryotypeBuilder( o_io => $o_io,
	                                    -o_logger => $o_log,
	                                    org => $$rh_params{ org },
	                                    -rh_padding => $rh_padding,
	                                    -filter => 1 # no supp data with 1 so no displayed
	                                  );

	my $ra_elts = $$rh_params{ elements };
	foreach my $rh_elt ( @$ra_elts )
	{
		$o_karyo->Plot( @$rh_elt{ qw( chrom start end label ) } );
	}

	$o_karyo->Print( %$rh_params );

	return;
}


=head2 GetParameters

 Usage        : my $rh_params = $o_x->GetParameters($o_io)
 Function     : Read parameters from command line and return them formatted as a hash ref.
 Returns      : $rh_params, hash ref with keys org, -elements, -output
 Args         : $o_io An instance Observer

=cut

sub GetParameters
{
    my ($o_io) = @_;
    
	ParamParser::SetDefaultUsage( sub { pod2usage(1) } );

	my @a_keys   = qw( org=s elements=s elements_file=s output=s help );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );

	my %h_params = ();

	# organisme
	my $org          = $o_param->Get( 'org' );
	if( $org eq '' )
	{
		my $msg      = "--org parameter is missing";
		my $user_msg = "Can't guess which org you want, use --org to specifie one.";
		throw LipmError( -text => $msg, -user_msg => $user_msg );
	}

	$h_params{ org } = $o_io->GetSpeciesShortName( $org );
	my $msg = "$org is not in this instance";
	throw LipmError( -text => $msg )  if( $h_params{ org } eq &NONE );

	# output
	my $output          = $o_param->Get( 'output' );
	$h_params{ output } = $output  unless( $output eq '' );

	# elements
	my $str_elts  = $o_param->Get( 'elements' );
	my $file_elts = $o_param->Get( 'elements_file' );
	my $ra_elts   = &Args_utils::GetFromStringOrFile( str  => $str_elts,
	                                                  file => $file_elts,
	                                                  sep  => ' ' );
	$h_params{ elements } = &FormatElements( $ra_elts );

	return \%h_params;
}


=head2 FormatElements

 Usage        : my $ra_elements = &FormatElements( $ra_str )
 Function     : given an elements string, reformat it in perl data structure.
 Returns      : $ra_elements, ref list of ref hashes.
                ref hashes hace chrom, start, end, and label keys.
 Args         : $ra_str, a lisr ref containing element structure (cf SYNOPSIS)

=cut

sub FormatElements
{
	my( $ra_str ) = @_;

	my @a_elts = ();

	foreach my $str_elt( @$ra_str )
	{
		my( $chrom, $start, $end, $label ) = split( /,|\t/, $str_elt );

		my %h_args = ();
		@h_args{ qw( chrom start end ) } = ( $chrom, $start, $end );
		LipmError::ParameterException::Assert( %h_args );

		$h_args{ label } = $label;

		# convertion d'unite
		$h_args{ start } = &Sequence_utils::Unit2Base( $h_args{ start } );
		$h_args{ end   } = &Sequence_utils::Unit2Base( $h_args{ end } );

		# check that start < end
		my $msg = "start <= end not verified.";
		$msg   .= "$chrom : start = $h_args{ start }, end = $h_args{ end }";
		$msg   .= ", label=$label" if( defined $label );
		throw LipmError( -text => $msg )  if( $h_args{ end } < $h_args{ start });

		push( @a_elts, \%h_args );
	}

	return \@a_elts;
}



__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
