#!/usr/bin/perl
#
# $Id: ncbi-blast2clusters.pl 14 2005-12-23 16:11:46Z gouzy $
#
#KEYWORDS: HSP single-linkage clustering

=head1 NAME

ncbi-blast2clusters.pl - build families from ncbi-blast HSPs

=head1 SYNOPSIS

ncbi-blast2clusters.pl --collect collection_of_HSPs --format m9|m8  \
                       --sequence "multifastafile_query,multifastafile_subject" \
                       --dirout  directory

see the program usage for the optional parameters

=head1 DESCRIPTION

 The program takes as first input the result tabular format (ncbi-blast -m 9 or -m 8 or iANT), it gets the length of the sequences involved in the process from one or many multifasta files (--sequence)
 which must contain headers in FASTiANT or Narcisse formats (where the length of the sequence is stored in the header line).
 Then, based on different thresholds (identity percent (--identity), e-value (--evalue), percentage of the query|subject sequence length (--pcs, --pcq)
 it applies a single linkage clustering algorithm. At the end it dumps the results in a given directory (--dirout, --prefix).
 Finally, for each family (protein only) a multiple aligment can be computed by the multalin program (--multalin).
 The script can handle both dna and protein sequences (--type dna|aa).

 Jerome.Gouzy@toulouse.inra.fr, Yoann.Beausse@toulouse.inra.fr, Thomas.Faraut@toulouse.inra.fr

=cut

#ca pourrait etre du DNA, a condition de remmettre tout dans le bon sens (deja fait dans mens) ou de faire du cap3

BEGIN
{
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
    unshift @INC, "$dirprg/../API";
    unshift @INC, "$dirprg/../API/Bio/Narcisse/Utils";
}

use Getopt::Long;
use ParamParser;
use Carp;
use strict;

my $DEBUG=0;
my $UNIXGZIP="/bin/gzip";

MAIN:
{
    &ParamParser::SetDefaultUsage(\&Usage);
    my $o_param = New ParamParser('GETOPTLONG',("help:s", "collectm9=s", "collect=s", "format=s", "famfile=s", "dirout=s", "pcs=i", "pcq=i", "identity=i", "evalue=s", "sequence=s", "prefix=s", "type=s", "cfg=s","prj","multalin","multalin_dir=s","coglike","rbh","result_format=s"));
    $o_param->SetUnlessDefined('prj','narcisse');
    $o_param->SetUnlessDefined('result_format','families');
    $o_param->AssertAllowedValue('result_format','families','minimal');
    my $multalin;
	my $paraloop="";
    if ( $o_param->Get('prj') eq 'narcisse' )
    {
        use Configuration::InstanceObserver;
        use ConfigurationFactory;
        my $o_io = ConfigurationFactory::GetConfiguration();
        my $root_core = $o_io->GetCfg('ROOT_CORE');
        my $root_tmp  = $o_io->GetCfg('ROOT_TMP');
        my $root_plp  = $o_io->GetCfg('ROOT_PARALOOP_SHELL');
        $multalin = "$root_core/multalin/ma -c:$root_core/multalin/blosum62.tab -o:msf";
		$paraloop = "cd $root_tmp ; root_plp ";
    }
    else
    {
        $o_param->AssertDirExists('multalin_dir');
        my $ma_dir = $o_param->Get('multalin_dir');
        $multalin = "$ma_dir/ma -c:$ma_dir/blosum62.tab -o:msf";
        $paraloop = $o_param->Get('paraloop'); # doit contenir --profram Shell ( et --ncpu x si *System)
    }

# compatibility mode
    $o_param->SetUnlessDefined('collect',$o_param->Get('collectm9'));
    $o_param->SetUnlessDefined('format','m9');
    $o_param->SetUnlessDefined('type','aa');
    $o_param->Set('rbh',1) if ( $o_param->IsDefined('coglike') ) ;
# end

    if ( $o_param->IsDefined('cfg') )
    {
        $o_param->AssertFileExists('cfg');
        $o_param->Update($o_param->Get('cfg'),'A');
    }
    # init default values
    $o_param->SetUnlessDefined('famfile',0);
	$o_param->SetUnlessDefined('pcs',0);
	$o_param->SetUnlessDefined('pcq',0);
	$o_param->SetUnlessDefined('identity',0);
	$o_param->SetUnlessDefined('evalue',1000);
	$o_param->SetUnlessDefined('prefix','nc_F');
    # asserts
    $o_param->AssertFileExists('collect');
    $o_param->AssertAllowedValue('format','m9','m8','iant');
    $o_param->AssertAllowedValue('type','dna','aa');
	$o_param->AssertDefined('sequence');
	$o_param->AssertDirExists('dirout');

	my %h_sequences = ();
	&LoadSequenceInfo($o_param->Get('sequence'),\%h_sequences);

    # cle=sequence, valeur=id du cluster
    &LoadResultsAndClusterize(\$o_param,\%h_sequences,$o_param->Get('rbh'));

    # cle=id du cluster, valeur=liste des id de sequence
    my %h_cluseq=();
 	foreach my $seq (keys (%h_sequences))
 	{
		next if ( !defined($h_sequences{$seq}->{CLUSTERID}));
        # REMOVED BY JG and MANU 
        # Le fichier fasta a change de format on n'a jamais prot
		#next unless ($seq =~/prot/);
 		$h_cluseq{$h_sequences{$seq}->{CLUSTERID}} .= "$seq ";
 	}

	if ( $o_param->Get('result_format') eq 'minimal' )
	{
		&DumpMinimal($o_param->Get('famfile'),\%h_cluseq,\%h_sequences);
	}
	else
	{
		my ($cmdma,$nbcmd) = &DumpFamilies($o_param->Get('dirout'),$o_param->Get('famfile'),$o_param->Get('prefix'),$multalin,\%h_cluseq,\%h_sequences);
		# attention le switch --multalin implique de changer l'appel au script dans narcisse (cf serveur web displayfam)
		if ( $o_param->Get('multalin') && ($o_param->Get('type') eq 'aa') )
		{
			# si paraloop est défini alors l'utiliser
            # REMOVE paraloop for Narcisse ==> JAMAIS utilise ici
			###&RunMultalin($cmdma,$nbcmd,$o_param->Get('dirout'),$paraloop);
            &RunMultalin($cmdma,$nbcmd,$o_param->Get('dirout'));
		}
	}
}

=head2 function LoadResultsAndClusterize

 Title        : LoadResultsAndClusterize
 Usage        : &LoadResultsAndClusterize($o_param,$rh_seqclu)
 Prerequisite :
 Function     :
 Returns      :
 Args         :
 Globals      : none

=cut

sub LoadResultsAndClusterize
{
my ($o_param,$rh_sequences,$coglike) = @_;

	my $collect = $$o_param->Get('collect');
	my $dirout  = $$o_param->Get('dirout');
	my $famfile = $$o_param->Get('famfile');
	my $thr_pcs = $$o_param->Get('pcs');
	my $thr_pcq = $$o_param->Get('pcq');
	my $thr_identity = $$o_param->Get('identity');
	my $thr_evalue = $$o_param->Get('evalue');
	my $prefix = $$o_param->Get('prefix');
			
	my $cpt_hsp=0;
	my $cpt_clu=1;
	my %h_clusters;
	my %h_encours=();
	my %h_encours_subjt=();
	my $qsp_prev="";
	my $ssp_prev="";
	my ($q,$s)=('','');
	$collect = "$UNIXGZIP -cd $collect |" if ( $collect =~ /\.gz$/ );
	open(COLLECT,$collect) or croak "ERROR openfile >$collect<";
	while(my $lign=<COLLECT>)
	{
		&Trace("HSP #$cpt_hsp ".$lign) if ( ! ( ($cpt_hsp++) % 50000 ));
		chomp($lign);

        # -m 9 ncbi-blast on target databases in FASTiANT or Narcisse format
        # Query: CAD13546.1-21645-22484 | 280 | CAD13546.1-21645-22484; | PROBABLE RESPONSE REGULATOR TRANSCRIPTION REGULATOR PROTEIN
		# Query: ccaa:NP_829153 chr=1 len=844 strand=-1 sp=ccaa polymorphic outer membrane protein G family protein/autotransporter,
		my $qspan="";
		my $sspan="";
		next if ( $lign eq "" || ($lign =~ /^#/ && $lign !~ /^\# Query: /) );
		if ( $lign =~ /^\# Query: / )
        {
			my $qlen=0;
			my $qname="";
			$lign =~ s/^# Query:\s+//;
			if ( $lign =~ / \| / )	# FASTiANT format
			{
				($qname,$qlen)=split(/ \| /,$lign);
			}
			elsif ( $lign =~ /len=(\d+) / )
			{
				$qlen=$1;
				($qname) = $lign =~ /^(\S+)/;
			}
			if ( ! defined($$rh_sequences{$qname}->{LENGTH}) )
			{
				$$rh_sequences{$qname}->{LENGTH} = $qlen;
			}
			next;
        }
		my ($pci,$alilen,$mismatch,$gap,$qstart,$qend,$sstart,$send,$evalue,$bitscore,$score)=();
		my @a_F=split(/\s+/,$lign);
		if ( $lign =~ /^</ ) # format iANT
		{
			($q,$qstart,$qend,$s,$sstart,$send) = ($a_F[0],$a_F[1],$a_F[2],$a_F[4],$a_F[5],$a_F[6]);
			$s =~ s/;$//;
			$q =~ s/^<//;
			($pci)    = $lign =~ / I=(\S+)/;
			($evalue) = $lign =~ / E=(\S+)/;
			($qspan)  = $lign =~ / \%Q=(\d+)/;
			($sspan)  = $lign =~ / \%S=(\d+)/;
			($score)  = $lign =~ / S=(\d+) /;
		}
		else # ncbi-blast m8 or m9
		{
			($q,$s,$pci,$alilen,$mismatch,$gap,$qstart,$qend,$sstart,$send,$evalue,$bitscore)=@a_F;
			$q = $$rh_sequences{$q}->{SPECIES}.":$q" if ( defined($$rh_sequences{$q}->{SPECIES}) && ($q !~  /$$rh_sequences{$q}->{SPECIES}/) );
			$s = $$rh_sequences{$s}->{SPECIES}.":$s" if ( defined($$rh_sequences{$s}->{SPECIES}) && ($s !~  /$$rh_sequences{$s}->{SPECIES}/) );
			&Trace("#$q // $s#");
			my $slen = $$rh_sequences{$s}->{LENGTH};
			my $qlen = $$rh_sequences{$q}->{LENGTH};
			croak "No length for $s (len=$slen) or $q (len=$qlen)" if ( ! $slen || ! $qlen);
			$qspan = (abs($qend-$qstart)+1) * 100 / $qlen;
			$sspan = (abs($send-$sstart)+1) * 100 / $slen;
			$score = $bitscore;
		}
		$evalue = "1$evalue" if ( $evalue =~ /^e/ );
		my $self = 0;
		if ( $q eq $s )
		{
			$self = ( $qstart == $sstart && $qend == $send ) ? 1 : 0;
		}
		next if ( ($qspan < $thr_pcq)  || ($sspan < $thr_pcs)  || ($pci < $thr_identity) || ($evalue > $thr_evalue) || $self );
		&Trace("($q ($qspan\%),$s ($sspan\%),$pci,$alilen,$mismatch,$gap,$qstart,$qend,$sstart,$send,$evalue,$bitscore)");
		if ( $coglike )  # RBH
		{
			my $qsp = $$rh_sequences{$q}->{SPECIES};
			my $ssp = $$rh_sequences{$s}->{SPECIES};

			if ( $qsp_prev ne "" && ( (($qsp ne $qsp_prev) || ($ssp ne $ssp_prev)) && ($qsp_prev ne $ssp && $ssp_prev ne $qsp) ) )
			{
				# ATTENTION: si on mets les hits dans les deux sens il faux imperativement qu'ils soient consecutifs dans le fichier de collecte
				&PairwiseAnalysis($q,$s,$qsp_prev,$ssp_prev,\%h_encours,\%h_encours_subjt,$rh_sequences,\%h_clusters,\$cpt_clu);
			}
			$h_encours{$q}=1;
			$h_encours_subjt{$s}=1;
			$qsp_prev = $qsp;
			$ssp_prev = $ssp;

			if ( $score > $$rh_sequences{$q}->{$ssp}->{SCORE} )
			{
				$$rh_sequences{$q}->{$ssp}->{SCORE} = $score;
				$$rh_sequences{$q}->{$ssp}->{RBH} = $s;
			}
			if ( $score > $$rh_sequences{$s}->{$qsp}->{SCORE} )
			{
				$$rh_sequences{$s}->{$qsp}->{SCORE} = $score;
				$$rh_sequences{$s}->{$qsp}->{RBH} = $q;
			}
		}
		else
		{
			&UpdateClusters($q,$s,\$cpt_clu,$rh_sequences,\%h_clusters);
		}
	}
	close(COLLECT);
	&Trace("Cluster updated");

	if ( $coglike && ($q ne "") )
	{
		&PairwiseAnalysis($q,$s,$$rh_sequences{$q}->{SPECIES},$$rh_sequences{$s}->{SPECIES},\%h_encours,\%h_encours_subjt,$rh_sequences,\%h_clusters,\$cpt_clu);
		&Trace("COG computed");
	}


}

=head2 function Analyse minimal genome comparison

 Title        : PairwiseAnalysis
 Usage        : &PairwiseAnalysis

=cut

sub PairwiseAnalysis
{
my ($q,$s,$qsp,$ssp,$rh_query,$rh_subjt,$rh_sequences,$rh_clusters,$r_cpt_clu) = @_;

	foreach my $q (sort keys (%$rh_query))
 	{
		if ( defined($$rh_sequences{$q}->{$ssp}->{SCORE}) )
		{
			my $best = $$rh_sequences{$q}->{$ssp}->{RBH};
			my $rbh  = $$rh_sequences{$best}->{$qsp}->{RBH};
			if ( $rbh eq $q )
			{
				&UpdateClusters($q,$best,$r_cpt_clu,$rh_sequences,$rh_clusters);
  				Trace("RBH $q vs $ssp ".$$rh_sequences{$q}->{$ssp}->{RBH}." ".$$rh_sequences{$q}->{$ssp}->{SCORE});
			}
			delete($$rh_sequences{$q}->{$ssp}->{SCORE});
			delete($$rh_sequences{$q}->{$ssp}->{RBH});
		}
		delete($$rh_sequences{$q}->{$ssp});
 	}

	foreach my $s (sort keys (%$rh_subjt))
 	{
		if ( defined($$rh_sequences{$s}->{$qsp}->{SCORE}) )
		{
			delete($$rh_sequences{$s}->{$qsp}->{SCORE});
			delete($$rh_sequences{$s}->{$qsp}->{RBH});
		}
		delete($$rh_sequences{$s}->{$qsp});
	}

	%{$rh_query}=();
	%{$rh_subjt}=();
}


=head2 function DumpMinimal

 Title        : DumpMinimal
 Usage        : &DumpMinimal($dirout,$rh_cluseq,$famfile)
 Args         : $dirout,$famfile,$prefix,$multalin,$rh_cluseq,$rh_sequences
 Globals      : none

=cut


sub DumpMinimal
{
my($famfile,$rh_cluseq,$rh_sequences)=@_;

	open(INDEX,">$famfile") or croak "ERROR createfile >$famfile<";
    foreach my $clu ( sort
                      {
                          my @F=split(' ',$$rh_cluseq{$a});
                          my @G=split(' ',$$rh_cluseq{$b});
						  $#G <=> $#F;
                      }
                      keys(%$rh_cluseq) )
	{
		my @a_list = sort split(' ',$$rh_cluseq{$clu});
		print INDEX join(' ',@a_list)."\n";
	}
	close(INDEX);
}



=head2 function DumpFamilies

 Title        : DumpFamilies
 Usage        : &DumpFamilies($dirout,$rh_cluseq,$famfile)
 Prerequisite :
 Function     :
 Returns      :
 Args         : $dirout,$famfile,$prefix,$multalin,$rh_cluseq,$rh_sequences
 Globals      : none

=cut


sub DumpFamilies
{
my($dirout,$famfile,$prefix,$multalin,$rh_cluseq,$rh_sequences)=@_;
my $execdir = $dirout;

	my $cpt=1;

	my ($idx)   = ( $famfile ) ? "$famfile.index.txt" : "index.txt";
	my ($cmdma) = ( $famfile ) ? "$famfile.cmd.ma"    : "cmd.ma";

	open(INDEX,">$dirout/$idx") or croak "ERROR createfile >$dirout/$idx<";
	open(CMD,">$dirout/$cmdma") or croak "ERROR createfile >$dirout/$cmdma<";

	my $fam_fas;
	my $fam_msf;

    foreach my $clu ( sort
                      {
                          my @F=split(' ',$$rh_cluseq{$a});
                          my @G=split(' ',$$rh_cluseq{$b});
						  $#G <=> $#F;
                      }
                      keys(%$rh_cluseq) )
	{
		my $fam_raw;
		if ( $famfile )
		{
			$fam_raw = "$dirout/$famfile";
		}
		else
		{
            my $id = sprintf("$prefix%05d",$cpt);
            mkdir "$dirout/$id";
			$fam_raw = "$dirout/$id/$id";
			$execdir = "$dirout/$id";
		}

		$fam_fas = "$fam_raw.fas";
		$fam_msf = "$fam_raw.msf";
		my $species_file = "$fam_raw.species";

 		my $description = &SaveOneFamily($fam_fas,$species_file,$$rh_cluseq{$clu},$rh_sequences);
		print INDEX "$fam_fas $$rh_cluseq{$clu} // $description\n";

		my $trash;
		($trash,$fam_fas) = $fam_fas =~ /(.+)\/(.+)/;
		($trash,$fam_raw) = $fam_raw =~ /(.+)\/(.+)/;

		print CMD "cd $execdir ; ";
		print CMD "( $multalin $fam_fas > /dev/null ) ; rm -f $fam_raw.cl?\n";
		$cpt++;
	}
	close(INDEX);
	close(CMD);
	#unlink "$dirout/$idx";
    return ("$dirout/$cmdma",$cpt-1);
}


=head2 function RunMultalin

 Title        : RunMultalin
 Usage        : &RunMultalin($cmdfile)
 Prerequisite : cd in a directory specific to the process (due to paraloop lock dir)
 Function     : run multalin on a multifasta file
 Returns      : 0 if there is nothing to run, 1 after execution
 Args         : $cmdfile,$nbcmd,$paraloop
 Globals      : none

=cut

sub RunMultalin
{
my ($cmdfile,$nbcmd,$dirout,$paraloop)=@_;
my $cde = "";

	chdir $dirout;
	return 0 if ( $nbcmd == 0 );
	if ( $nbcmd == 1 || ! $paraloop )
	{
		$cde = "nice sh $cmdfile";
		`$cde`;
	}
	else
	{
		$cde = "$paraloop --input $cmdfile --output $cmdfile.out";
		`$cde`;
		# au cas ou cela ne serait pas vraiment fini (ca devrait etre regle ds les versions > 1.1)
		$cde = "$paraloop --restart PARALOOP_lock";
		`$cde`;
	}
	return 1;

    #unlink "$cmdfile";

}

=head2 function UpdateKeywords

 Title        : UpdateKeywords
 Usage        : &UpdateKeywords()
 Prerequisite :
 Function     :
 Returns      :
 Args         : $header,$rh_keywords
 Globals      : none

=cut

sub UpdateKeywords
{
my ($header,$rh_keywords)=@_;

	chomp($header);

	$header =~ s/^>\S+//;
	$header=~ s/\bof\b|\bprotein\b|\s\S\s|hypothetical|probable|putative|\b\d+\b|\w+=\S+|\([^\)]+\)//gi;
	$header=~ s/\btranscription\b/transcriptional/gi;
	foreach my $kw (split(' ',$header))
	{
		$kw = "\L$kw";
		$$rh_keywords{$kw}++;
	}
}


=head2 function SaveOneFamily

 Title        : SaveOneFamily
 Usage        : &SaveOneFamily()
 Prerequisite :
 Function     :
 Returns      :
 Args         : $fam:
                $list_seq : string which contain the list of the family members
                $rh_index
                $rh_filename
 Globals      : none

=cut

sub SaveOneFamily
{
my ($fam,$species_file,$list_seq,$rh_sequences)=@_;

    my @a_seq=split(' ',$list_seq);
    my $prev="firstfile";
	my %h_sp=();

	my %h_keywords=();

	open(FAM,">$fam") or croak "ERROR createfile >$fam<";
	foreach my $seq (@a_seq)
	{
		next if ( $seq =~ /^\s*$/ );
		&Trace("#$seq#$$rh_sequences{$seq}->{FINDEX},$$rh_sequences{$seq}->{FILENAME}");
		$h_sp{$$rh_sequences{$seq}->{SPECIES}}=1;
		if ( $$rh_sequences{$seq}->{FILENAME} ne $prev )
        {
          close(FASTA) if ( $prev ne "firstfile");
          open(FASTA,$$rh_sequences{$seq}->{FILENAME}) or croak "ERROR openfile >$$rh_sequences{$seq}->{FILENAME}< (sequence $seq)";
          $prev=$$rh_sequences{$seq}->{FILENAME};
        }
		seek (FASTA,$$rh_sequences{$seq}->{FINDEX},0);
		my $end_seq = 0;
		my $lign;
		while( ($lign=<FASTA>) && (! $end_seq) )
		{
			$lign =~ s/^>/>$$rh_sequences{$seq}->{SPECIES}:/ if ( defined($$rh_sequences{$seq}->{SPECIES}) && $lign !~ /^>$$rh_sequences{$seq}->{SPECIES}/);
			if ( $lign =~ /^>/ && $lign !~ /^>$seq/ )
			{
				$end_seq = 1;
			}
			else
			{
				$lign =~ s/\*$//;
				if ( $lign =~ /^>/)
				{
					&UpdateKeywords($lign,\%h_keywords);
				}
				print FAM $lign;
			}
		}
	}
	close(FAM);
	close(FASTA);

	my $nb = keys(%h_sp);
	open(SPECIES,">$species_file") or croak "ERROR createfile >$species_file<";
	print SPECIES $nb." ".join('#',sort keys (%h_sp))."\n";
	close(SPECIES);

	my $description="";
	my $c=0;
	foreach my $kw (sort { $h_keywords{$b} <=> $h_keywords{$a} } keys(%h_keywords))
	{
		if ($h_keywords{$kw} > 0 )
		{
			$description .= "$kw ";
		}
		$c++;
		last if ( $c == 3 );
	}
	$description =~ s/\s*$//;
	$description = "protein family" if ( $description eq "" );
	return $description;
}

=head2 function UpdateClusters

 Title        : UpdateClusters
 Usage        : &UpdateClusters()
 Prerequisite : none
 Function     : defines the cluster id, from a pair of sequence ID
 Returns      : none
 Args         : $q: query id
                $s: subject id
                $r_cpt_clu: cluster id count
                $rh_sequences{$q|$s}->{CLUSTERID}
 Globals      : none

=cut

sub UpdateClusters
{
	my($q,$s,$r_cpt_clu,$rh_sequences,$rh_clusters) = @_;



	if ( ! defined($$rh_sequences{$q}->{CLUSTERID}) && ! defined($$rh_sequences{$s}->{CLUSTERID}) )
	{
		&Trace("Case 1 (none exists) $q $s");
		$$rh_sequences{$q}->{CLUSTERID} = $$r_cpt_clu;
		$$rh_sequences{$s}->{CLUSTERID} = $$r_cpt_clu;
		$$rh_clusters{$$r_cpt_clu} = "$q $s";
		$$r_cpt_clu++;
		print STDOUT "$$r_cpt_clu $q $s\n" if ( ($$r_cpt_clu % 300) == 0 );
	}
	elsif ( $$rh_sequences{$q}->{CLUSTERID} && ! defined($$rh_sequences{$s}->{CLUSTERID}) )
	{
		&Trace("Case 2 (query is already in a cluster) $q/$$rh_sequences{$q}->{CLUSTERID} $s");
		$$rh_sequences{$s}->{CLUSTERID} = $$rh_sequences{$q}->{CLUSTERID};
		$$rh_clusters{$$rh_sequences{$q}->{CLUSTERID}} .= " $s";
	}
	elsif ( ! defined($$rh_sequences{$q}->{CLUSTERID}) && $$rh_sequences{$s}->{CLUSTERID} )
	{
		&Trace("Case 3 (subject is already in a cluster) $q $s/$$rh_sequences{$s}->{CLUSTERID}");
		$$rh_sequences{$q}->{CLUSTERID} = $$rh_sequences{$s}->{CLUSTERID};
		$$rh_clusters{$$rh_sequences{$s}->{CLUSTERID}} .= " $q";
	}
	elsif ( $$rh_sequences{$q}->{CLUSTERID} && $$rh_sequences{$s}->{CLUSTERID} && ($$rh_sequences{$q}->{CLUSTERID} != $$rh_sequences{$s}->{CLUSTERID}) )
	{
		my $q_cluid = $$rh_sequences{$q}->{CLUSTERID};
		my $s_cluid = $$rh_sequences{$s}->{CLUSTERID};
		&Trace("Case 4 (query and subject are already in a cluster) $q/$q_cluid $s/$s_cluid");
		my @a_list=(split(' ',$$rh_clusters{$q_cluid}),split(' ',$$rh_clusters{$s_cluid}));
		#foreach my $elem2 (keys(%$rh_sequences))
		foreach my $elem2 (@a_list)
		{
			#&Trace("MERGE CLU##$elem2#$$rh_sequences{$elem2}->{CLUSTERID}#$q_cluid#$s_cluid#");
			#if ( defined($$rh_sequences{$elem2}->{CLUSTERID}) && ($$rh_sequences{$elem2}->{CLUSTERID} == $s_cluid ) || ($$rh_sequences{$elem2}->{CLUSTERID} == $q_cluid) )
			{
				&Trace("AV: $elem2 $$rh_sequences{$elem2}->{CLUSTERID}");
				$$rh_sequences{$elem2}->{CLUSTERID} = $$r_cpt_clu;
				&Trace("AP: $elem2 $$rh_sequences{$elem2}->{CLUSTERID}");
			}
		}
		delete($$rh_clusters{$q_cluid});
		delete($$rh_clusters{$s_cluid});
		$$rh_clusters{$$r_cpt_clu} = join(' ',@a_list);
		$$r_cpt_clu++;
		print STDOUT "$$r_cpt_clu $q $s\n" if ( ($$r_cpt_clu % 300) == 0 );
	}
}


=head2 function LoadSequenceInfo

 Title        : LoadSequenceInfo
 Usage        : &LoadSequenceInfo()
 Prerequisite : the fasta header is expected to be either FASTiANT or Narcisse format
 Function     : Get the length of the sequences from a multifasta file
 Returns      : update hashes $rh_len, $rh_index and $rh_filename (the key is the sequence id)
                MANU - 1 key added: species:locus_tag
 Args         : $sequence: a multifasta filename
                $rh_sequences{}->{LENGTH|FINDEX|FILENAME}:   an hash to store the sequence length, the index of the sequence in the multifasta file and the name of the multifasta file
 Globals      : none

=cut

sub LoadSequenceInfo
{
	my ($seqfile,$rh_sequences) = @_;
	my $id;    # $id the main key
    my $id_lt;  # Another key (using the locus tag)
    my $id_acc; #  Another key (removing prot from the sequence identifier)
	my $len;
	my $cpt=1;

    #warn ("KOUKOU $seqfile");
    my @a_seqfile = split(/[,\s]+/,$seqfile);

    foreach my $sf (@a_seqfile)
    {       
        open(FASTA,$sf) or croak "ERROR openfile >$sf<";
        my $pos = tell(FASTA);
        my $sp = 0;
		my $len = 0;
		while(my $lign=<FASTA>)
        {
            if ( $lign !~ /^>/ )
            {
                $pos = tell(FASTA);
                next;
            }
            chomp($lign);
			$cpt++;
			if ( $lign =~ / len=(\d+) / ) # Narcisse
            {
                $len=$1;
                ($id)    = $lign =~ /^(\S+)/;
                ($sp)    = $lign =~ / sp=(\S+) /;
                ($id_lt) = $lign =~ / locus_tag=(\S+) /;
                $id_lt   = $sp.':'.$id_lt;
            }
            elsif ( $lign =~ / \| / )	# FASTiANT format
            {
                ($id,$len)=split(/ \| /,$lign);
            }
            $id =~ s/>//;
			$$rh_sequences{$id}->{SPECIES} = $sp  if ( $sp  ); # je veux garder l'espece
			
			#$id = "$sp:$id" if ( $sp && ($id ~ /$sp/) );     
			# MANU - 23/4/2012 - espece chick, id = prot:chicken_s6g180c2r15 - chick est dans chicken donc ca ne passe pas !
			#                    La condition est donc: "Commencer par le nom de l'espece"
			$id = "$sp:$id" if ( $sp && ($id !~ /^$sp/) );     
			$$rh_sequences{$id}->{LENGTH} = $len  if ( $len );
            $$rh_sequences{$id}->{SPECIES} = $sp  if ( $sp  );
			$$rh_sequences{$id}->{FINDEX} = $pos;
            $$rh_sequences{$id}->{FILENAME} = $sf if ( $sf  );
            if ($id_lt ne $id)
            {
                $$rh_sequences{$id_lt} = $$rh_sequences{$id};
            }
            #Patch to enable discordant sequence file and blast output file with regard to the "prot:" prefix
            if ( $id =~/prot\:([\w\-\.\:]+)$/ ) 
            {
            	$id_acc = $1;
            	my $sp_id_acc = "$sp:$id_acc";
            	$$rh_sequences{$id_acc} = $$rh_sequences{$id};
            	$$rh_sequences{$sp_id_acc} = $$rh_sequences{$id};
            }                 
        }
        close(FASTA);
    }
 	Trace("Loaded sequences: $cpt\n");
}

=head2 function Trace

 Title        : Trace
 Usage        : &Trace($msg)
 Prerequisite : none
 Function     : Print $msg of the program when $DEBUG is true
 Globals      : $DEBUG

=cut

sub Trace
{
	my ($msg) = @_;

	if ( $DEBUG )
	{
		print STDERR "$msg\n";
	}
}

=head2 function Usage

 Title        : Usage
 Usage        : &Usage()
 Prerequisite : none
 Function     : Print the usage of the program

=cut

sub Usage
{
my ($o_param) = @_;
print STDERR "
$0
  --help
Mandatory
  --sequence       quoted string of multifasta files in FASTiANT|Narcisse format
  --collect        filename containing a collection of HSPs
  --format         m9|m8|iant
  --dirout         directoryname for outputs

Optional
  --cfg            filename (configuration file format)
  --pcs            integer [".$o_param->Get('pcs')."]
  --pcq            integer [".$o_param->Get('pcq')."]
  --identity       integer [".$o_param->Get('identity')."]
  --evalue         float   [".$o_param->Get('evalue')."]

  --famfile        string  [".$o_param->Get('famfile')."]
  --prefix         string  [".$o_param->Get('prefix')."]
  --multalin       [0]  switch to activate the computation of the muliple alignments
  --prj            narcisse|iant [".$o_param->Get('prj')."]
  --rbh            [0]	flag to switch in a clustering mode similar to the COG protocol
  --result_format  families|minimal  [".$o_param->Get('result_format')."]	dump each family files or a minimal file of related sequences

  --multalin_dir   directoryname (required if not --prj narcisse)

Warning:
	when rbh mode is active, only one blast is required (gen1 vs gen2),
	but if you want to provide blast results in both orientation (gen1 vs gen2 AND gen2 vs gen1)
	then the two blast result files must be consecutive in the HSP collection file

";

}
