/*
   $Id: Comparison.h 750 2014-03-03 12:35:35Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __COMPARISON_H__
#define __COMPARISON_H__

#include <memory>
using namespace std;

#include "Parameters.h"
#include "QueryBuffer.h"
#include "QueryBufferDouble.h"
#include "QueryBufferBank.h"
#include "CompactTree.h"
#include "GlintEngineFreq.h"

/** \brief the Comparison hierarchy 
 *         - These classes represent the comparison process
 *         - They are non copyable
 *         - The Run() method starts the comparison process
 *         - The classes:
 *             - ComparisonBase   : The base class, abstract.
 *             - Comparison       : The 'normal' comparison class
 *             - ComparisonSelf   : Used with self comparisons (-S switch)
 *             - ComparisonGenome : Used with Genome comparisons (-G switch)                   
*/

/** \brief The base abstract object
*/
class ComparisonBase: private NonCopyable {
	public:
		ComparisonBase(); 
		virtual ~ComparisonBase(){};
		virtual void Run();
		
	protected:
		Parameters&  _parameters;
		CompactTree& _ctree;			    

		// initializer functions - WARNING ! You'll HAVE to delete the returned objects ! 
		GlintEngine*         _InitGlintEngine( QueryBuffer* q, Display& d )      { return new GlintEngine(q,d); };
		GlintEngineDouble*   _InitGlintEngine( QueryBufferDouble* q, Display& d) 
			{ 
				if ( _parameters.IsePCR() )
				{
					return new GlintEngineDoublePCR(q,d);
				}
				else
				{
					return new GlintEngineDouble(q,d); 
				}
			}; 
						
		auto_ptr<Display>    _InitDisplay();
		virtual QueryBuffer* _InitQueryBuffer() = 0;

	private:
		const Mask& __mask;
		bool  __ValidQueryFile();
};

class ComparisonGenome: public ComparisonBase 
{
	protected:
		QueryBuffer* _InitQueryBuffer() { return new QueryBufferBank(_parameters.GetQueryFilename().GetName(),_parameters.GetQueryFilename().GetCompression()); };
};

class ComparisonSelf: public ComparisonBase 
{
	protected:
		QueryBuffer* _InitQueryBuffer() { return new QueryBufferSelf(); };
		void Run();

};	

class Comparison: public ComparisonBase 
{
	protected:
		QueryBuffer* _InitQueryBuffer();
		GlintEngineBase* _InitGlintEngine();
};

#endif
