/*
   $Id: word.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _WORD_H
#define _WORD_H

#include <cstring>
#include <string>
#include "glint.h"
#include "masque.h"

using std::string;

/*

   WordDnaCode: All Dna Word classes derive from this tiny object
                Our genetic code is encapsulated here (...)

*/

class WordDnaCode {
    public:
	    static int code(char c);
	    static char decode(int c);
	    static bool IsCode(char c) {return (code(c) < N);};
	    enum {A,C,G,T,N};

		/* After ToHigherCase(true), every letter is converted to upper case before being pushed
		 * After ToHigherCase(false) (default), the letters are not converted
	    */
		static void ToUpperCase(bool to_up) { h_flg = to_up ? 0xDF : 0xFF; };
	
    private:
		static uint8_t h_flg;   /* may be $FF (do not change) or $DF (change to upper) */
};

/*

    Word: Deriving from WordDnacode
          This object is used to manipulate the characters read by Window
          There are 2 versions of the implementation:
                -Built upon a 32 bit memory word, for 32 bits architecture
                -Built upon a 64 bit memory word, more performant on 64 bits architecture

          Both versions have the same interface, but their implementations are completely different
          This was implemented with #define rather than with inheritance, for performance reasons.
          (we avoid the use of virtual functions in such a crticial part of the application)

*/

class DinucleotideComplexity;
class Window;

class Word: public WordDnaCode {
    public:

	    /*
	
	    The constructors
	
	    1/ From an uint8_t: initialize a word to 0 from a length information.
	    2/ From a char *: The size is taken from the char * length
	    3/ From another Word: the system's constructor is OK here
	
	    */
	
	    Word(uint8_t lg);
	    Word(const char* w);
	
	    /*
	    Some accessors
	
	    1/ GetSize returns the size of the Word (number of bytes)
	    2/ GetLength returns the length of the word (number of char)
	    3/ Getc returns the character whose position is passed by the parameter (0 = First character)
	    */
	
	    char GetSize() const { return size;};
	    char GetLength() const {return length;};
	    char Getc(int i) const;
	
	    /* The Pushc, Popc, Getc functions
	
	    Pushc: insert a new letter in the word by pushing everything to the left and adding with OR
	    Since two uint32_t are used, a shift and push mechanism is needed: last character in r[0] (2 MSB)
	    is extracted before shifting everything to the left, and added (ORed) to r[1]
	    Popc:  The reverse of Pushc
	
	    */
	
	    void Pushc(char c);
	    char Popc();
	
	    /*
	
	            The Conversion functions:
	                1/ string Convert():             convert a word to a string
	                2/ string Convert(uint32_t,length): convert an uint32_t to a 'chaine', the length is passed by parameter
	                          This latter function is static, because it is not related to any object
	    */
	
	    string Convert() const;
	    static string Convert(uint32_t l,int length);
	
	    /*
	        The hashing functions:
	        Hash            ==> Very simple, used for black_listing some words
	    */
	
	    int Hash(int sz) const;
	
	    /*
	             The Complement, ReverseComplement functions
	
	             Those functions return a new word, which is the DNA complement or reverse complement of the current word.
	             They return a Word*
	             The coding of the nucleotides A:00 T:11 C:01 G:10 was devised in order to enable such an easy complementation
	    */
	
	    Word* Complement() const;
	    Word* ReverseComplement() const;
	
	    /*
	       Some operators, very useful for using with containers
	    */
	
	    friend bool operator==(const Word&, const Word&);
	    friend bool operator<(const Word&, const Word&);
	
	    /* The friend classes */
	
	    friend     class DinucleotideComplexity;
	    friend     class Window;

#define REGISTER_SIZE 64

    private:
		uint64_t r;        /* A 64 bits word, allowing for masks as long as 32 characters */
		uint64_t m;        /* If less than 32 letters are wished a mask must be used for the remaining bits */
	    uint8_t size;        /* size = The size of the word in bytes */
	    uint8_t length;      /* length = The number of characters of the word */
	    void InitMask() { /* Called by the constructors */
	        m=0;
	        for (int i=0;i<size;i++) {         /* The mask is 1 for every bit except for the discarded bits : i.e  00011111111 */
	            m <<=1;
	            m|=1;
	        }
	    };
};

/* INLINE FUNCTIONS */
/* The constructors */

inline Word::Word(uint8_t lg): r(0),size(lg*2),length(lg) {InitMask();}   /* The parameter is a size information, init to 0 */
inline Word::Word(const char* w)
{
    length = strlen(w);
    size = 2 * length;                 /* 2 bits / letter */
    InitMask();
    for(int i=0;i<length;i++)
    {
        Pushc(w[i]);
    };
}

/* The accessors */

inline char Word::Getc(int i) const
{
    int shift_count=size-2*(i+1);
    uint64_t mask = 3<<shift_count;
    uint64_t rvl = mask&r;
    rvl >>= shift_count;
    return (char)(rvl&3);
}

/* The Pushc, Popc functions */

inline void Word::Pushc(char c) {
    r<<=2;
    r|=code(c);
}

inline char Word::Popc() {
    int b=r&3;
    r>>=2;
    return decode(b);
}

/* The Conversion functions */

inline string Word::Convert() const
{
    string s="";
    for (int k=0;k<length;k++) {
        char c=Getc(k);
        s += decode(c);
    }
    return s;
}

inline string Word::Convert(uint32_t l,int length)
{
    int m=3;        // A mask
    string s="";
    for (int k=0;k<length;k++,l>>=2) {
        uint32_t c=m & l;
        s += decode(c);
    }
    return s;
}

/* The hashing functions */

inline int Word::Hash(int sz) const {  /* A hashing function */
    return (r&m) % sz;
}

inline Word* Word::Complement() const {
    Word *w=new Word(size);
    w->r=(~r)&m;
    return w;
}

inline Word* Word::ReverseComplement() const {
    Word *rw = new Word(length);
    Word *w  = Complement();
    int cnt = length;
    while (cnt--)
    {
        rw->Pushc(w->Popc());
    };
    delete w;
    return rw;
}

/* The operator==, <, != */

inline bool operator==(const Word& a, const Word& b)
{
    return ((a.r&a.m)==(b.r&b.m));
}
inline bool operator<(const Word& a, const Word& b)
{
    return ((a.r&a.m) < (b.r&b.m));
}

inline bool operator!=(const Word& a, const Word& b) { return !operator==(a,b);}

/*
   The WordDnacode methods

*/

inline int WordDnaCode::code(char c){
	c &= h_flg;		// Converting to uppercase... or not !
    switch (c)
    {
        case 'A' : return A;
        case 'C' : return C;
        case 'G' : return G;
        case 'T' : return T;
        default :  return N;
    }
}
inline char WordDnaCode::decode(int c){
    switch (c)
    {
        case A : return 'A';
        case C : return 'C';
        case G : return 'G';
        case T : return 'T';
        default :  return 'N';
    }
}


#endif
