#
# $Id: AnnotatedSeqFile.pm 1159 2007-12-28 16:11:54Z gouzy $
#

package AnnotatedSeqFile;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeqFile - provides methods to read/write chromosome annotations

=head1 SYNOPSIS

do not use directly, instantiate child classes instead

	AnnotatedSeqFile::Genbank : protein-coding gene = {mRNA,CDS}
	AnnotatedSeqFile::GFF3    : protein-coding gene = {mRNA<CDS}
	AnnotatedSeqFile::TIGRXML : protein-coding gene = {mRNA,CDS}
	AnnotatedSeqFile::Ensembl
	AnnotatedSeqFile::Fasta

=head1 DESCRIPTION

=cut

use Carp;
use LipmError::ParameterException;
use NarcisseDataWriter::OnDisk;
use AnnotatedSeqFileWriter::Narcisse;
use General;
use Error qw(:try);
use Locus;
use Locus::Gene;
use Locus::Mrna;
use Locus::Rna;
use Locus::Cds;
use Locus::Repeat;
use Data::Dumper;

# a mettre dans un dumper xml
use FileHandle;
use XML::Twig;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1159 $ =~ /\d+/g); $r[0]};
}

# TODO - UTILISER _Get et _Set pour ACEEDER AUX MEMBRES PRIVES OU PROTEGES !!!

=head2 procedure _Init

 Title        : Constructor
 Usage        : Use child classes instead	(AnnotatedSeqFile::GenBank, AnnotatedSeqFile::GFF3, etc.)
 Prerequisite : none
 Function     : Initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         : see child classes SYNOPSIS
 Globals      : none

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$$self{'_debug'} = &FALSE;	
	#$$self{'_debug'} = &TRUE;
	
	my $infile      = $h_param{'infile'};
	my $informat    = $h_param{'informat'};
	my $outformat   = $h_param{'outformat'};
	my $tmpdir      = $h_param{'tmpdir'};

	$outformat = 'NARCISSE' if ( ! defined($outformat) );
	$informat  = 'GENBANK'  if ( ! defined($informat) );
	$tmpdir    = '.'        if ( ! defined($tmpdir) );
	
	my $o_annotated_writer = $self -> AnnotatedSeqFileWriterFactory( $outformat );
	$self -> { __o_annotated_writer } = $o_annotated_writer;

	if ( $informat !~ /FASTA|ENSEMBL/ && (! defined($infile) || ! -e $infile || -z $infile) )
	{
		Carp::croak("Fatal error in constructor: $infile not defined or does not exist or empty\n");
	}
	if ( ! -d $tmpdir )
	{
		Carp::croak("Fatal error in constructor: the outdirectory $tmpdir must exist\n");
	}

	foreach my $required ('chr', 'species', 'kingdom')
	{
		if ( !defined($h_param{$required}) )
		{
			Carp::croak("Fatal error in constructor: parameter $required must be provided to the constructor\n");
		}
		$$self{'__'.$required}= $h_param{$required};
	}


	# Sequence file
	if ( defined($h_param{seq_file}) && $h_param{seq_file} ne '' )
	{
		$$self{'__seq_file'} = $h_param{seq_file};
	}
	else
	{
		$$self{'__seq_file'} = undef;
	}
	
	# Assembly
	if ( defined($h_param{file_assembly}) && $h_param{file_assembly} ne "")
	{
		$$self{'__file_assembly'}  = $h_param{file_assembly};
		$$self{'__assembly_label'} = ( ! defined($h_param{assembly_label}) ) ? "reference" : $h_param{assembly_label};
		$$self{'__rh_contigs'}     = $self->LoadContigPath($$self{'__file_assembly'},$$self{'__assembly_label'},$$self{'__chr'});
        $$self{'__assembly_needed'}= &TRUE;
	}
    else
    {
        $$self{'__assembly_needed'} = &FALSE;
        $$self{'__rh_contigs'}      = undef;
    }
    
    # TODO -- REMPLACER __listofseqfiles PAR UNE REF DE ARRAY
    #         ATTENTION CE TRUC EST UTILISE PAR LES CLASSES DERIVEES !!! (devait etre protected)
	$$self{'__listofseqfiles'}="";
	
	# See AddGene...CDS or AddGene...Exon
	$$self{ '__cds_overlap_allowed'} = $h_param{'cds_overlap_allowed'};
	
	# See GetNarcisseId() in this object AND in the Locus::XXX objects
	$$self{ '__narcisse_id'} = $self->GetSpecies().'_s'.$self->GetChromosomeLabel();
    
    # An array of Locus::Gene objects and an index to those objects
	$$self{'__a_genes'}   = [];
	
	# An index to the Locus::Gene objects
	# key = locus tag
	# val = Index of the FIRST gene found with this locus tag in the array of genes
	$$self{'__idx_genes'} = {};
	
	# An index of mrnas objects:
	# key = locus tag
	# val = (ind_g,ind_r)(ind_g,ind_r)...
	#       where ind_g is the index of the gene parent of the mrna, ind_r is the index of the mrna found in the gene
	#       There may be several mrnas with the same locus tag
	$$self{'__idx_mrnas'} = {};

	# An array of Locus::Rna objects
	$$self{'__a_ncrnas'}  = [];
	
	# An array of Locus:Repeat objects
	$$self{'__a_repeats'} = [];

	$$self{'__infile'}= $infile;
	$$self{'__tmpdir'}= $tmpdir;
	$$self{'__outfmt'}= $outformat;
	$$self{'__consistency_limit'}= $h_param{'consistency_limit'} if ( defined($h_param{'consistency_limit'}) && $h_param{'consistency_limit'} =~ /^\d+$/ );
	
	if ( defined $h_param{'table'} && $h_param{'table'} =~ /^\d+$/ )
	{
		$$self{'__codon_usage_table'} = $h_param{'table'};
	}
	elsif ( $self->GetKingdom() !~ /eukar/i )
	{
		$$self{'__codon_usage_table'} = 11
	}
	
	# The default is selected by bioperl
	else
	{
		$$self{'__codon_usage_table'} = undef;
	}
	
	$$self{__length}             = 0;
	
	# Will be used by the methods __LoadOneSeq and GetGenomicSequence
	my $seq = "";
	$$self { __genomic_sequence } = \$seq;

	return;
}

=head2 AnnotatedSeqFileWriterFactory

 Title     : AnnotatedSeqFileWriterFactory
 Usage     : my $o_annot = &AnnotatedSeqFileWriterFactory( $outfile );
 Function  : Use $outformat to choose an AnnotatedSeqFileWriter to build
 Args     : $outfile 
 Return   : the built object
=cut
sub AnnotatedSeqFileWriterFactory
{
	my $self = shift;
	my ($outformat) = @_;
	
	if ( $outformat eq 'NARCISSE' )
	{
		return New AnnotatedSeqFileWriter::Narcisse( $self );
	}
	else
	{
		throw LipmError::ParameterException ("$outformat NOT SUPPORTED (only NARCISSE is supported yet)", 'value');
	}
}

=head2 Some accessors
=cut
sub GetFormat()
{
	throw LipmError::ObjectException( 'AnnotatedSeqFile', 'abstract method' );
}
sub GetDebug
{
	my $self = shift;
	return $self -> { _debug };
}
sub GetKingdom
{
	my $self = shift;
	return $self -> { __kingdom };
}
sub GetSpecies
{
	my $self = shift;
	return $self -> { __species };
}
sub GetChromosomeLabel
{
	my $self = shift;
	return $self -> { __chr };
}
sub GetCodonTable
{
	my $self = shift;
	return $self -> { __codon_usage_table };
}
sub GetContigInfo
{
	my $self = shift;
	my ($contig) = @_;
	if (defined $self -> { __rh_contigs } )
	{
		return $self -> { __rh_contigs } -> { $contig };
	}
	else
	{
		return undef;
	}
}
sub GetNarcisseId
{
	my $self = shift;
	return $self -> { __narcisse_id };
}

=head2 SetDebug/UnsetDebug
=cut
sub SetDebug
{
	my $self = shift;
	$self -> { _debug } = &TRUE;
}
sub UnsetDebug
{
	my $self = shift;
	$self -> { _debug } = &FALSE;
}

=head2 function AddGene

 Title        : AddGene
 Usage        : $o_annot -> AddGene($o_gene)
 Function     : Add a new protein coding gene to this object:
                    - update the index
                    - update the gene parent and the gene narcisseId
 Returns      : the added gene
 Args         : $o_gene: a Locus::Gene object

=cut

sub AddGene
{
	my($self,$o_gene) = @_;
	throw LipmError::ParameterException( 'o_gene', 'type', -text => 'should by a Locus::Gene object instead of a ' . ref($o_gene) ) unless ref $o_gene eq 'Locus::Gene';

	# Put the gene inside the __a_genes array, and update the index	
	my $ra_genes = $self -> { __a_genes };
	push @$ra_genes, $o_gene;
	
	my $locus_tag = $o_gene->GetLocusTag();

	my $rh_index = $self -> { __idx_genes };
	if ( !defined $rh_index -> { $locus_tag } )
	{
		$rh_index -> { $locus_tag } = [];
	}
	push ( @{$rh_index -> { $locus_tag }}, $o_gene );

	# Set the gene parent and compute the Narcisse Id
	$o_gene -> SetParent ( $self );

	return $o_gene;
}

=head2 function AddGeneMRNA

 Title        : AddGeneMRNA
 Usage        : &AddGeneMRNA($o_mrna,$locus_tag)
 Prerequisite : none
 Function     : Based on the locus_tag and the coordinates, search the parent gene in the database. 
                If found, add a new transcript to the gene and keep this transcript to the mrnas index
 Returns      : $o_locus if OK, undef if the gene is not found
 Args         : $o_mrna   : A Locus::Mrna object
                $locus_tag: the locus_tag value of its parent (gene)
 Globals      : none

=cut

sub AddGeneMRNA
{
	my($self,$o_mrna,$locus_tag) = @_;
	
	my $o_gene = $self -> __SearchGene( $o_mrna, $locus_tag );
	if ( ! defined $o_gene )
	{
		print STDERR "AnnotatedSeqFile::AddGeneMRNA: Non fatal error: unable to find >$locus_tag<\n";
		return undef;
	}
	
	$o_gene -> AddmRNA( $o_mrna );
	
	my $rh_index       = $self -> { __idx_mrnas };
	my $mrna_locus_tag = $o_mrna -> GetLocusTag();
	if ( !defined $rh_index -> { $mrna_locus_tag } )
	{
		$rh_index -> { $mrna_locus_tag } = [];
	}
	push ( @{$rh_index -> { $mrna_locus_tag }}, $o_mrna );
	return $o_mrna;
}

=head2 function AddGeneCDS

 Title        : AddGeneCDS
 Usage        : $o_annot -> AddGeneCDS($o_cds,$locus_tag)
 Prerequisite : The "parent" of the CDS must be the gene identified by $locus_tag (as in Genbank formatted files)
 Function     : look for a Mrna surrounding the CDS and call AddCDS on it
                If not found, try to create a new Mrna, then call AddGeneMRNA and call AddCDS to the new Mrna
 Returns      : $o_cds The added CDS, undef if the gene was not found
 Args         : $o_cds  : a Locus::CDS object
                $locus_tag: the locus_tag value of the gene parent
 Globals      : none

=cut

sub AddGeneCDS
{
	my($self,$o_cds,$locus_tag) = @_;
	
	my $i = 1;
	while ( &TRUE )
	{
		my $o_mrna = $self -> __SearchMRNA($o_cds,$locus_tag,$i);
		if (defined $o_mrna)
		{
			# We must use an auxiliary variable because it is forbidden to last,next,return in try/catch constructs. Pity.
			my $ok=0;
			try
			{
				$o_mrna -> AddCDS ( $o_cds );
				$ok=1;
				#print "OK\n";
			}
			
			# If there is an error - meaning there is an overlap between two CDS, the behaviour here is
			#                        to create ANOTHER CDS. Not sure it is a good idea, however
			catch Error with
			{
				my $o_err = shift;
				#warn " WARNING - " . $o_err -> text . "\n";
				#warn "           gene = $locus_tag\n";
				#warn "           mrna = " . $o_mrna -> GetLocusTag();
				#warn "           cds  = " . $o_cds  -> GetLocusTag();
				$i++;
			};
			return $o_cds if ($ok==1);
			next
		}
		else
		{
			my $o_gene = $self -> __SearchGene ( $o_cds, $locus_tag );
			return undef unless defined $o_gene;

			my %h_param;
			($h_param{ start}, $h_param{end}) = $o_gene -> GetEndsContig();			
			$o_mrna = New Locus::Mrna ( $o_cds, %h_param );
			my $o_added_mrna = $self -> AddGeneMRNA ( $o_mrna, $locus_tag );
			return undef unless defined $o_added_mrna;
			
			$o_added_mrna -> AddCDS ( $o_cds );
			warn "CREATED A NEW MRNA FROM A CDS (".$o_cds->GetLocusTag().")" if ($self -> GetDebug());

			return $o_cds;
		}
	}
}

=head2 function AddGeneMRNACDS

 Title        : AddGeneMRNACDS
 Usage        : $o_annot -> AddGeneMRNACDS($o_locus,$locus_tag)
 Prerequisite : The parent of the CDS must be the transcript OR the Gene (as in GFF3 formatted files)
 Function     : Add a CDS or add a region to an existing CDS, looking for an mRNA as parent, or calling AddGeneCDS if not found
 Returns      : $o_cds or undef if the transcript cannot be found
 Args         : $o_locus: a locus object (may be a CDS or a region)
                $locus_tag: the locus_tag value of its parent (transcript)
 Globals      : none

=cut

sub AddGeneMRNACDS
{
	my($self,$o_locus,$locus_tag) = @_;

	my $o_mrna = $self -> __SearchMRNA ( $o_locus, $locus_tag );
	if ( defined $o_mrna )
	{
		try
		{
			$o_mrna -> AddCDS ( $o_locus );
		}
			
		# If there is an error - meaning there is an overlap between two CDS, the behaviour here is 
		#                        controlled by the parameter CDS_OVERLAP_ALLOWED: we ignore OR throw the error
		catch Error with
		{
			my $o_err = shift;
			if ( $self->{ '__cds_overlap_allowed' } eq '0' )
			{
				my $text = $o_err -> text() . "\n";
				$text .= "May be you could try CDS_OVERLAP_ALLOWED=1 in the parameter file ?";
				throw Error::Simple ( $text );
			}
			else
			{
				my $msg = "WARNING - mrna " . $o_mrna->GetLocusTag() . " - Overlap detected in the CDS - Check this mrna\n";
				warn $msg;
			}
		};
		return $o_locus;
	}
	else
	{
		my $o_cds = $self -> AddGeneCDS ( $o_locus, $locus_tag );
		if ( ! defined $o_cds )
		{
			print STDERR "AnnotateSeqfile::AddGeneMRNACDS: Non fatal error: unable to find locus >$locus_tag< (it SHOULD BE a gene or an mRNA)\n";
		}
		return $o_cds;
	}
}

=head2 function AddGeneMRNAExon

 Title        : AddGeneMRNAExon
 Usage        : &AddGeneMRNAExon($o_locus,$locus_tag)
 Prerequisite : The parent of the Exon must be the transcript (as in GFF3 formatted files)
 Function     : Add an exon to an existing MRNA 
                Based on the locus_tag, search the transcript in the database. If found, add a new Exon to the gene (the parent of the transcript)
 Returns      : $o_mrna the transcript, undef if not found
 Args         : $o_locus: a locus object
                $locus_tag: the locus_tag value of its parent (transcript)
 Globals      : none

=cut

sub AddGeneMRNAExon
{
	my($self,$o_locus,$locus_tag) = @_;

	my $o_mrna = $self -> __SearchMRNA ( $o_locus, $locus_tag );
	if ( defined $o_mrna )
	{
		try
		{
			$o_mrna -> AddRegion ( $o_locus -> GetEndsContig() );
		}
			
		# If there is an error - meaning there is an overlap between two CDS, the behaviour here is 
		#                        controlled by the parameter CDS_OVERLAP_ALLOWED: we ignore OR throw the error
		catch Error with
		{
			my $o_err = shift;
			if ( $self->{ '__cds_overlap_allowed' } eq '0' )
			{
				throw $o_err;
			}
			else
			{
				my $msg = "WARNING - mrna " . $o_mrna->GetLocusTag() . " - Overlap detected in the CDS - Check this mrna\n";
				warn $msg;
			}

		};
		return $o_mrna;
	}
	else
	{
		print STDERR "AnnotateSeqfile::AddGeneMRNAExon: Non fatal error: unable to find locus >$locus_tag< (maybe the corresponding transcript not yet loaded)\n";
		return undef;
	}
}

=head2 function AddRna

 Title        : AddRna
 Usage        : $o_annot -> AddGene($o_ncnra)
 Function     : Add a new non-protein coding gene
 Returns      : the index in the array
 Args         : $o_ncrna: a Locus::Rna object

=cut

sub AddRna
{
	my($self,$o_ncrna) = @_;
	throw LipmError::ParameterException( 'o_ncrna', 'type', -text => 'should by a Locus::RNA object instead of a ' . ref($o_ncrna) ) unless ref $o_ncrna eq 'Locus::Rna';

	my $ra_ncrnas = $self -> { __a_ncrnas };
	push @$ra_ncrnas, $o_ncrna;

	# Set the parent and compute the Narcisse Id
	$o_ncrna -> SetParent ( $self );
	
	return $#$ra_ncrnas;
}

=head2 function AddRepeat

 Title        : AddRepeat
 Usage        : $o_annot -> AddRepeat($o_repeat)
 Function     : Add a new repeat
 Returns      : the index in which the repeat has been inserted
 Args         : $o_repeat: a Locus::Repeat object

=cut

sub AddRepeat
{
	my($self,$o_repeat) = @_;
	throw LipmError::ParameterException( 'o_repeat', 'type', -text => 'should by a Locus::Repeat object instead of a ' . ref($o_repeat) ) unless ref $o_repeat eq 'Locus::Repeat';

	my $ra_repeats = $self -> { __a_repeats };
	push @$ra_repeats, $o_repeat;

	# Set the parent and compute the Narcisse Id
	$o_repeat -> SetParent ( $self );
	
	return $#$ra_repeats;
}

=head2 function GetNbRepeats

 Title        : GetNbRepeats
 Usage        : my $nb_rep = &GetNbRepeats()
 Prerequisite : none
 Function     : returns the number of repeats loaded
 Returns      : the number of repeats loaded
 Args         : none
 Globals      : none

=cut

sub GetNbRepeats
{
	my ($self) = @_;
	my $ra_repeats = $self -> { __a_repeats };
	return $#$ra_repeats + 1;
}

=head2 function GetNbRNAs

 Title        : GetNbRNAs
 Usage        : my $nb_ncrnas = &GetNbRNAs()
 Function     : returns the number of non protein coding RNAs loaded
 Returns      : the number of non protein coding RNAs
 Args         : none
 Globals      : none

=cut

sub GetNbRNAs
{
	my ($self) = @_;
	my $ra_ncrnas = $self -> { __a_ncrnas };
	return scalar( @$ra_ncrnas );
}

=head2 function GetNbGenes

 Title        : GetNbGenes
 Usage        : my $nb_genes = &GetNbGenes()
 Function     : returns the number of protein-coding genes
 Returns      : the number of protein coding genes
 Args         : none
 Globals      : none

=cut

sub GetNbGenes
{
	my ($self) = @_;
	my $ra_genes = $self -> { __a_genes };
	return scalar( @$ra_genes );
}

=head2 function GetNbProteinSequences

 Title        : GetNbProteinSequences
 Usage        : my $nb_prot = $o_annot -> GetNbProteinSequences();
 Function     : returns the number of proteins translated from the protein-coding genes
                Warning: Some genes are NOT translated because of consistency problems
 Returns      : the number of protein sequences
 Args         : none
 Globals      : none

=cut

sub GetNbProteinSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_annotated_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'prot' );
}

=head2 function GetNbGeneSequences

 Title        : GetNbGeneSequences
 Usage        : my $nb_genes = $o_annot -> GetNbGeneSequences();
 Function     : returns the number of gene sequences kept in the database
 Returns      : the number of gene sequences
 Args         : none
 Globals      : none

=cut

sub GetNbGeneSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_annotated_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'gene' );
}

=head2 function GetNbCdsSequences

 Title        : GetNbCdsSequences
 Usage        : my $nb_cds = $o_annot -> GetNbCdsSequences();
 Function     : returns the number of cds sequences kept in the database
 Returns      : the number of cds sequences
 Args         : none
 Globals      : none

=cut

sub GetNbCdsSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_annotated_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'cds' );

}

=head2 function GetNbTranscriptSequences

 Title        : GetNbTranscriptSequences
 Usage        : my $nb_prot = $o_annot -> GetNbTranscriptSequences();
 Function     : returns the number of trascripts kept in the database
 Returns      : the number of transcripts
 Args         : none
 Globals      : none

=cut

sub GetNbTranscriptSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_annotated_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'mrna' );
}

=head2 function GetNbTranscripts

 Title        : GetNbTranscripts
 Usage        : my $nb_prot = &GetNbTranscripts()
 Function     : returns the number of transcripts 
 Returns      : the number of transcripts
 Args         : none
 Globals      : none

=cut

sub GetNbTranscripts
{
	my ($self) = @_;
	my $nb_mrnas = 0;
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene (@$ra_genes)
	{
		my $ra_mrnas = $o_gene -> GetMrnas();
		$nb_mrnas += @$ra_mrnas;
	}
	return $nb_mrnas;
}

=head2 function GetNbExons

 Title        : GetNbExons
 Usage        : my $nb_prot = &GetNbExons()
 Function     : returns the number of exons found - This is more than the number of proteins
 Returns      : the number of exons
 Args         : none
 Globals      : none

=cut

sub GetNbExons
{
	my ($self) = @_;
	my $nb_exons = 0;
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene (@$ra_genes)
	{
		my $ra_mrnas = $o_gene -> GetMrnas();
		foreach my $o_mrna (@$ra_mrnas)
		{
			my @a_regions = $o_mrna -> GetRegions();
			$nb_exons += @a_regions;
		}
	}
	return $nb_exons;
}

=head2 GetGenes,GetRepeats,GetRnas

 Title    : GetGenes,GetRepeats,GetRnas
 Function : Accessor to the array of genes/repeats/nc_rnas
 Usage    : my $ra_genes = $o_annot -> GetGenes();
 Arg      : None
 Return   : A ref to an array of Gene/Repeat/RNA objects
 
=cut
sub GetGenes
{
	my $self = shift;
	return $self -> { __a_genes };
}
sub GetRepeats
{
	my $self = shift;
	return $self -> { __a_repeats };
}
sub GetRnas
{
	my $self = shift;
	return $self -> { __a_ncrnas };
}

=head2 function GetGene

 Title        : GetGene
 Usage        : my $o_gene = $o_annot->GetGene( $index );
 Function     : Return the gene found at the index $index
 Returns      : The gene
 Args         : The index
 TODO - A SUPPRIMER ?

=cut

sub GetGene
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_genes } -> [ $index ];
}
sub GetRepeat
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_repeats } -> [ $index ];
}
sub GetNcrna
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_ncrnas } -> [ $index ];
}

=head2 procedure Print

 Title        : Print
 Usage        : &Print($level);
 Prerequisite : none
 Function     : Print the data structure
 Returns      : the number of protein coding genes
 Args         : $level: if 'full' then a full report of the datastructure is printed else only a short summary
 Globals      : none

=cut

sub Print
{
	my $self = shift;
	my ($level) = @_;

	$level = '' if ( !defined($level));
	print STDOUT "chr      :\t".$self->{'__chr'}."\n";
	print STDOUT "infile   :\t".(defined ($self->{'__infile'}) ? $self->{'__infile'} : "NO ANNOTATION FILE DEFINED")."\n";
	print STDOUT "informat :\t".$self->GetFormat()."\n";
	print STDOUT "tmpdir   :\t".$self->{'__tmpdir'}."\n";
	print STDOUT "outformat:\t".$self->{'__outfmt'}."\n";
	if ( defined $self->GetCodonTable() )
	{
		print STDOUT "Usage table:\t". $self->GetCodonTable() . "\n";
	}
	else
	{
		print STDOUT "Usage table:\tNOT DEFINED (default choosen by bioperl)\n";
	}

	my $nb_genes = $self->GetNbGenes();
	print STDOUT $nb_genes." protein coding genes.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_genes; $i++)
	{
		$$self{'__a_genes'}[$i]->Print();
	}
	
	my $nb_ncrnas = $self->GetNbRNAs();
	print STDOUT $nb_ncrnas." non protein coding genes.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_ncrnas; $i++)
    {
		$$self{'__a_ncrnas'}[$i]->Print();
    }
    
    my $nb_repeats = $self-> GetNbRepeats();
	print STDOUT $nb_repeats." repeat regions.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_repeats; $i++)
    {
		$$self{'__a_repeats'}[$i]->Print();
    }
}

=head2 procedure Load

 Title        : Load
 Usage        : Virtual method, die when called
 Prerequisite : none
 Function     : Ensure that the class AnnotatedSeqFile is not directly instantiated
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my ($self) = @_;

	Carp::croak("Fatal Error: Load is a virtual method => do not instantiate AnnotatedSeqFile class, use child classes instead");
}

=head2 procedure _Synchronize

 Title        : _Synchronize
 Usage        : $o_annot -> _Synchronize();
 Prerequisite : The data file must have been completely read
 Function     : When the number of exons in the mRNA is lower than in the CDS, it means either that only the CDS are described in the file (generally GFF)
                or that a strong discrepancy exists between mRNA and CDS. In that case the CDS is considered as the reference and the CDS locus 
                replaces the mRNA 
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub _Synchronize
{
	my $self = shift;
	
	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene ( @$ra_genes )
	{
		$o_gene -> SyncCDS2Mrna( $self -> GetDebug() );
	}
}

=head2 Accessor GetLength

 Title        : GetLength
 Usage        : my $len = $o_obj -> GetLength()
 Prerequisite : Dump must have been called, else the dummy length 0 is returned
 Args         : nothing
 Return       : The chromosome length
 
=cut

sub GetLength
{
	my $self = shift;
	return $self->{__length};
}
 
=head2 procedure Dump

 Title        : Dump
 Usage        : $o_obj -> Dump(%h_param)
 Prerequisite : DumpFastaDb should have been called first to set the consistency status of the cds
				(see Locus::Cds::GetTranslation for the details)
 Procedure    : Call the Dump procedure from the writer object
 Args         : %h_param: parameters of dumping method, see the objects derived from AnnotatedseqfileWriter for the details
 Globals      : none

=cut

sub Dump
{
	my $self = shift;
	my $o_annotated_writer = $self -> { __o_annotated_writer };
	$o_annotated_writer -> Dump ( @_ );
}

=head2 procedure DumpFastaDb

 Title     : DumpFastaDb
 Usage     : $o_obj -> DumpFastaDb(%h_param)
 Procedure : Load the fasta sequence, then call the DumpFastaDb procedure from the writer obejct
 Returns   : none
 Args      : %h_param: parameters of dumping method, see the objects derived from AnnotatedseqfileWriter for the details
 Globals   : none

=cut

sub DumpFastaDb
{
	my $self = shift;
	my (%h_param) = @_;

	$self -> __LoadOneSeq ( $h_param{sequence} );
	
	my $o_annotated_writer = $self -> { __o_annotated_writer };
	$o_annotated_writer -> DumpFastaDb ( @_ );
	
}


=head2 procedure AssertConsistencyLimit

 Title     : AssertConsistencyLimit
 Usage     : $o_obj -> AssertConsistencyLimit()
 Procedure : Croak if the nb of inconsistent cds is under the consistency limit
 Args      : none
 Globals   : none
 TODO      : UN check SERAIT UN PEU MIEUX !!!

=cut
sub AssertConsistencyLimit
{
	my $self = shift;
	
	my $chr = $self -> GetChromosomeLabel();
	
	# Consistency limit	
	my $consistency_limit=$$self{'__consistency_limit'};
	if ( defined($consistency_limit) && $consistency_limit =~ /^\d+$/ )
	{
		my $nb_protein    = $self -> GetNbProteinSequences();
		my $nb_transcript = $self -> GetNbTranscriptSequences();
		my $nb_gene       = $self -> GetNbGenes();
		if ( $nb_transcript != 0 )
		{
			my $consistency   = sprintf("%04.1f",100.0 * $nb_protein / $nb_transcript);
			
			if ($nb_protein < $nb_transcript * $consistency_limit/100.0)
			{
	            my $msg = "Fatal error: chromosome $chr: too many inconsistencies during protein extraction:\n";
	            $msg .= "nb of genes       = $nb_gene;\nnb of transcripts = $nb_transcript\nnb of proteins    = $nb_protein\nconsistency       = $consistency% (consistency_limit = $consistency_limit\%)\n";
	            $msg .= "You should: 1) check the translation table (parameter TABLE_$chr in your species.cfg\n";
	            $msg .= "            2) check the annotation file format (is it REALLY correctly formatted ?\n";
	            $msg .= "            3) At a last resort, increase the parameter CONSISTENCY_LIMIT_$chr in your species.cfg file\n";
				Carp::croak($msg);
			}
			else
			{
				print "Consistency check OK: $nb_protein proteins for $nb_transcript transcripts ($nb_gene genes) \nconsistency = $consistency% (consistency limit = $consistency_limit%)\n\n";
			}
		}
	}
	else
	{
		print "Warning: consistency checks skipped, consistency_limit is not initialized or is not an integer\%)\n";
	}
}

=head2 procedure LoadContigPath

 Title        : LoadContigPath
 Usage        : &LoadContigPath($file_assembly,$grouplabel,$chr)
 Prerequisite : none
 Function     : Load the tiling/golden path
 Returns      : return a hash encoding the tiling path
 Args         : $file_assembly    : file containing the golden path
                $grouplabel: identifier of the required assembly (for human the file contains Celera and reference assembly)
				$chr       : chromosome ID
 Globals      : none

 rh_contigs description:
                key = feature name
                val = A hash
                      keys = feature_id (part number in agp specification, here it is a chromosome),
                             chr_start,chr_end,orientation (object beg and end, orientation in agp specification)
	                         feature_type (component type in agp specification),
	                         feature_start,feature_end (component beg and end in agp specification)
	                         chromosome (irrelevant in agp specification)
	                         feature_name (component_id in agp specification)
=cut

sub LoadContigPath
{
	my ( $self, $file_assembly, $grouplabel, $chr)=@_;

	my $rh_contigs;
	if ( $file_assembly =~ /\.agp/ )
	{ 
		$rh_contigs = $self->_LoadAgpFile( $file_assembly, $chr );
	}
	elsif ( $file_assembly =~ /\.md/ )
	{
		$rh_contigs = $self->_LoadSeqContigMd( $file_assembly, $grouplabel, $chr );	
		if( 0 == keys %$rh_contigs )
		{
			my $msg = "Your ASSEMBLY ($grouplabel) was probably not found";
			$msg   .= " in your ASSEMBLY_FILE ($file_assembly).\n";
			$msg   .= "Anyway, I'm quite sure that an error will occur.\n";
			warn $msg;
		}
	}
	else
	{
		throw LipmError::ParameterException ("The ASSEMBLY_FILE parameter is not correct: should be a .agp or .md file",'value');
	}

	# __StupidDump( $rh_contigs );
	#print STDERR Dumper $rh_contigs;

	return $rh_contigs;
}

sub __StupidDump
{
	my $rh_anything = shift;
	
	foreach my $key (sort keys %$rh_anything)
	{
		print "$key\t"; 
		if ( &UNIVERSAL::isa( $rh_anything->{$key}, "HASH") )
		{
			foreach my $elt (sort keys %{$rh_anything->{$key}})
			{
				print $rh_anything->{$key}{$elt}."\t";			
			}
		}
		elsif ( &UNIVERSAL::isa( $rh_anything->{$key}, "ARRAY") )
		{
			print join("\t",@{$rh_anything->{$key}});
		}
		else
		{
			print "$rh_anything->{$key}";
		}
		print "\n";
	}	
}

=head2 procedure _LoadSeqContigMd

 Title        : _LoadSeqContigMd
 Usage        : &_LoadSeqContigMd($file_assembly,$grouplabel,$chr)
 Prerequisite : none
 Function     : Load the tiling/golden path
 Returns      : return a hash encoding the tiling path
 Args         : $file_assembly: file containing the golden path
                $grouplabel: identifier of the required assembly (for human the file contains Celera and reference assembly
				$chr: chromosome ID
 Globals      : none

=cut

sub _LoadSeqContigMd
{
	my($self,$file_assembly,$grouplabel,$chr)=@_;

	my %contig;
	my $assembly;
	# JER 20070201 : my @intitule = ("taxid","chr","from","to","orientation","ac_version","id","type","assembly","weight");
	my @intitule = ("tax_id","chromosome","chr_start","chr_stop","orientation","feature_name","feature_id","feature_type","group_label","weight");
	my $randomchrom=($chr=~/_r$/) ? 1 : 0;
	my $artificial_offset=0;
	my $nuclbuffersize=600;

	$file_assembly = &AnnotatedSeqFile::GetStream($file_assembly);
	open(CONTIG,$file_assembly) or	Carp::croak("LoadContigPath/Error: unable to read the contig path file >$file_assembly<\n");
	while ( my $line = <CONTIG> )
	{
		next if ( $line =~ /^#/ );

		#chomp win compatible
		$line =~ s/\r?\n//;

		my %h_tmp = ();
		my @st = split('\t', $line);
		for ( my $i=0 ; $i<$#intitule ; $i++ )
		{
			$h_tmp{$intitule[$i]} = $st[$i];
		}
		#JER $h_tmp{feature_name} =~ s/\.\d+$//;

		if ($randomchrom) {
			my $parent_chrom=$chr;
			$parent_chrom=~s/_r$//;
			$h_tmp{chromosome}=~s/$parent_chrom\|.+/$chr/;
		}

		next if ( $h_tmp{chromosome} ne $chr );	# only data related to the current chromosome
		next if ( defined($grouplabel) && ($h_tmp{group_label} ne $grouplabel) );
		
		# ON NE SAUTE PLUS L'ORIENTATION NEGATIVE (!!!)
		###next if ( $h_tmp{orientation} =~ /^-$/);
		next if ( $h_tmp{chromosome} =~ /\|/ && !$randomchrom);

		my $chr_key_start = $h_tmp{chromosome} . "_start";
		if ( ($h_tmp{feature_name} =~ /start/ || $h_tmp{chr_start} == 1)  &&  !defined($contig{$chr_key_start}) )
		{
			$assembly = $h_tmp{group_label};
			$contig{$chr_key_start} = $h_tmp{chr_start};
		}

		next if ( $h_tmp{group_label} !~ /$assembly/ );

		if ( $h_tmp{feature_name} =~ /end/ )
		{
			my $chr_key_end = $h_tmp{chromosome} . "_end";
			$contig{$chr_key_end} = $h_tmp{chr_stop};
		}

		if ($randomchrom) {
			$h_tmp{chr_start} += $artificial_offset;
			$h_tmp{chr_stop} += $artificial_offset;
			$artificial_offset += ($h_tmp{chr_stop}-$h_tmp{chr_start}+1)+$nuclbuffersize;
		}

		$contig{$h_tmp{feature_name}} = \%h_tmp;

		#THOMAS: adding feature_start and feature_end to be compliant with agp files (ce n'est pas très propre)
		$contig{$h_tmp{feature_name}}{feature_start} = 1;
		$contig{$h_tmp{feature_name}}{feature_end} = $h_tmp{chr_stop}-$h_tmp{chr_start}+1;

		if ( !defined($contig{$h_tmp{chromosome}}) )
		{
			my @tmp;
			$contig{$h_tmp{chromosome}} = \@tmp;				
		}

		push(@{$contig{$h_tmp{chromosome}}},$h_tmp{feature_name});

	}
	close(CONTIG);

	if ($randomchrom) {
		unshift(@{$contig{$chr}},"start");
		push(@{$contig{$chr}},"end");
		my $chr_key_end = $chr . "_end";
		$contig{$chr_key_end} = $artificial_offset+$artificial_offset;
	}

	foreach my $k ( keys %contig )
	{
		next if ( $k ne $chr );
		my ($from,$to) = (99999999999,0);
		foreach my $ctg ( @{$contig{$k}})
		{
			next if ($ctg eq "start" || $ctg eq "end");
			$from = &General::Min($from,$contig{$ctg}->{chr_start});
			$to   = &General::Max($to,  $contig{$ctg}->{chr_stop}  );
		}
		my $key_start = $k."_start";
		my $key_end   = $k."_end";

		if ( !defined($contig{$key_start}) ) { $contig{$key_start} = $from; }
		if ( !defined($contig{$key_end})   ) { $contig{$key_end}   = $to;}
	}

	print Dumper(%contig) if ( $self->GetDebug() );

	return \%contig;
}

=head2 procedure LoadAgpFile

 Title        : LoadAgpFile
 Usage        : &LoadAgpFile( $agp_file, $chrom )
 Function     : Load the assembly information contained in the agp file
 Returns      : return a hash encoding the contigs and coordinates on each chromosome of the genome
 Args         : $agp_file: file containing the agp description
 				$chrom : the chromosome of interest               				
 Globals      : none

=cut
sub _LoadAgpFile
{
   my( $self, $file_assembly, $chrom )=@_;
	
	my %contig;
	
	$file_assembly = &AnnotatedSeqFile::GetStream($file_assembly);
	
	#Agp file are organized according to http://www.ncbi.nlm.nih.gov/genome/assembly/agp/AGP_Specification.shtml
	#columns 0-4 : object  object_begin  object_end  part_number feature_type 
	# feature_type != N    columns 5-8: feature_id feature_beg  feature_end orientation
	# feature_type == N   columns 5-8: gap_length  gap_type  linkage empty
	
	#To be compatible with LoadContigPath
	#my @intitule = ("tax_id","chromosome","chr_start","chr_stop","orientation","feature_name","feature_id","feature_type","group_label","weight");
	
	my @common_keys = qw( chromosome chr_start chr_stop feature_id feature_type );
	my @buffer_fields = qw( length buffer_type );
	my @contig_fields = qw( feature_name feature_start feature_end orientation );
	
	open(AGP,$file_assembly) or	Carp::croak("LoadContigPath/Error: unable to read the contig path file >$file_assembly<\n");
	
	while( my $line = <AGP>){
 		next if ( $line =~/^#/ );
		chomp($line);
		my @fields =  split('\t', $line);;
 		
 		my ($chr, $chr_start,$chr_end,$feat_id,$feat_type) = @fields[0..4];
		$chr=~s/chr//;
				
		my %fragment;
		@fragment{@common_keys} = ($chr, $chr_start, $chr_end, $feat_id, $feat_type );
		my $fragment_key;
		
		if ( $feat_type eq "N" )
		{
			my ( $length, $feat_name ) = @fields[5..6];
			$fragment_key = "$feat_name\_$feat_id";
			@{fragment}{@buffer_fields} = ( $length, $feat_name ) ;
			
			#Anonymous fragments are not taken into account
		}
		else 
		{
			my ($feat_name, $feat_start, $feat_end, $orientation) = @fields[5..$#fields];
			$fragment_key = $feat_name;
			@{fragment}{@contig_fields} = ($feat_name, $feat_start, $feat_end, $orientation);
			
			$contig{$chr}{$fragment_key}=\%fragment;
		}			
 	} 

	&_FromAgpToContigMd( $chrom, \%contig );	
	
	print Dumper($contig{$chrom}) if ( $self->GetDebug() );

	return $contig{$chrom};	
}

=head2 procedure FromAgpToContigMd

 Title        : FromAgpToContigMd
 Usage        : &FromAgpToContigMd( $rh_contigs )
 Function     : Adding extra information for compatibility reasons
 Returns      : none
 Args         : hash encoding the contigs and coordinates on each chromosome of the genome           				
 Globals      : none

 Note : seb - on pourrait simplifier l'ecriture en passant un seul parametre
		-> $rh_contigs->{$chrom}
=cut

sub _FromAgpToContigMd
{
	my ( $chrom, $rh_contigs ) = @_;
	
	my $rh_chrom_contigs = $rh_contigs->{$chrom};
	
	my @sorted_fragments = sort { $rh_chrom_contigs->{$a}{ chr_start }  <=> $rh_chrom_contigs->{$b}{ chr_start } } keys %$rh_chrom_contigs;
		
	my $chr_start = $rh_chrom_contigs->{$sorted_fragments[0]}{ chr_start };
	my $chr_end   = $rh_chrom_contigs->{$sorted_fragments[$#sorted_fragments]}{ chr_stop };
	my @fragment_names = @sorted_fragments;
	
	my $key_start = $chrom."_start";
    my $key_end   = $chrom."_end";
	my $key_chrom = $chrom;
	
	$rh_contigs->{$chrom}{ $key_start } = $chr_start;
	$rh_contigs->{$chrom}{ $key_end }   = $chr_end;
	$rh_contigs->{$chrom}{ $chrom } = \@fragment_names;	
}

=head2 procedure BuildChromosomeFile

 Title        : BuildChromosomeFile
 Usage        : BuildChromosomeFile($file_chr)
 Prerequisite : When assembly is needed, the contigs sequences ($it) must be accessible at "$$self{'__tmpdir'}/seqfastmp.$chr.$it";
 Fonction     : Pseudomolecule assembly
 Returns      : the chromosome length
 Args         : $file_chr: the filename of the fasta assembly of the chromosome sequence
 Globals      : none

=cut

sub BuildChromosomeFile
{
	my ( $self, $file_chr ) = @_;
	
	my $chrom_length;
	#my $assembly_label = ( defined $self->{'__file_assembly'} && $self->{'__file_assembly'}=~/\.agp/ ) ? 1 : 0;

	$chrom_length = $self->_BuildChromosomeFileFromSeqContig( $file_chr );
	
	#Old version made a distinction between agp and SeqContig
	#if ( $assembly_label )
	#{
    #	$chrom_length = $self->_BuildChromosomeFileFromAgp( $file_chr );
	#}
	#else
	#{
	#	$chrom_length = $self->_BuildChromosomeFileFromSeqContig( $file_chr )
	#}
	
	$self -> {__length } = $chrom_length;
	return $chrom_length;
}

#sub _BuildChromosomeFileFromAgp_REMOVED
#{
	#my ( $self, $file_chr ) = @_;
#	
	#my $rh_contigs = $$self{'__rh_contigs'};   #We are here always in a situation where we have an agp file
	#my $chr = $$self{'__chr'};
	#my $code_sp = $$self{'__species'};
	#my $tmp_sequence_dir = $$self{'__tmpdir'};
#	
	#my $nb_fasta_sequences = 0;
#	
	#if ( defined($$self{'__seq_file'}) )	#If a multifasta or single fasta file has been provided in the species.cfg file
	#{
		#($nb_fasta_sequences,$$self{'__listofseqfiles'}) = $self->ExtractFastaFiles($$self{'__seq_file'},$$self{'__tmpdir'},$$self{'__chr'});
	#}
	#my @a_seqfiles = split(' ',$$self{'__listofseqfiles'});
	#$nb_fasta_sequences = $#a_seqfiles+1;
#				
	#my $chr_seq = "";
	#foreach my $contig ( sort { $rh_contigs->{$a}{chr_start} <=> $rh_contigs->{$b}{chr_start} } keys %$rh_contigs )
	#{
			#if ( $rh_contigs->{$contig}{feature_type} eq "N" )
			#{
				#$chr_seq.= "N" x $rh_contigs->{$contig}{length};
			#}
			#else
			#{
				#my $seq_file = "$tmp_sequence_dir/seqfastmp.$chr.$rh_contigs->{$contig}{feature_name}";
				#if (!-e $seq_file )
				#{
					##We try without the fragment extension
					#my $contig_name = $rh_contigs->{$contig}{feature_name};
					#$contig_name=~s/\_\d+$//;
					#$seq_file = "$tmp_sequence_dir/seqfastmp.$chr.$contig_name";
					#print STDERR "WARNING $rh_contigs->{$contig}{feature_name} replaced by $contig_name \n";
					#if ( !-e $seq_file )
					#{
						#die "You really have a problem with $rh_contigs->{$contig}{feature_name} ";
					#}
				#}
				#my $seq_in =  Bio::SeqIO->new(-file => $seq_file, -format => 'fasta' );
				#my $o_seq = $seq_in->next_seq;
				#my ($start,$end) = @{$rh_contigs->{$contig}}{qw( feature_start feature_end )};
				#my $subseq_obj = Bio::Seq->new( -seq => $o_seq->subseq($start,$end) );			                                    				
				#my $subseq = ( $rh_contigs->{$contig}{orientation} > 0 ) ?  $subseq_obj->seq :  $subseq_obj->revcom->seq;
				#$chr_seq.= $subseq;
			#}
		#}
#	
	#my $seq_out = Bio::SeqIO->new(-file => ">$file_chr", -format => 'fasta' );
	#my $o_chrseq = Bio::Seq->new( -seq => $chr_seq,                        
                                      #-display_id => "$code_sp$chr",                                   
                                      #-alphabet => "dna" );
 #
	#$seq_out->write_seq( $o_chrseq );
#	
	#return $o_chrseq->length();
#}

=head2 procedure _BuildChromosomeFileFromSeqContig

 Title        : _BuildChromosomeFileFromSeqContig
 Usage        : _BuildChromosomeFileFromSeqContig($file_chr)
 Prerequisite : When assembly is needed, the contigs sequences ($it) must be accessible at "$$self{'__tmpdir'}/seqfastmp.$chr.$it";
 Fonction     : Pseudomolecule assembly
 Returns      : the chromosome length
 Args         : $file_chr: the filename of the fasta assembly of the chromosome sequence
 Globals      : none

=cut

sub _BuildChromosomeFileFromSeqContig
{
	my ( $self, $file_chr ) = @_;

	my $rh_contigs     = ( $self->{'__assembly_needed'} != &FALSE ) ? $self->{'__rh_contigs'} : undef;
	my $chr     = $self -> GetChromosomeLabel();
	my $code_sp = $self -> GetSpecies();

	my $file_chr_tmp="";
	my $nb_fasta_sequences = 0;

	# only one sequence file for the chromosome, extracted from the genbank file, no golden path
	# TODO - HOULALA CA DEVRAIT MARCHER QUELQUE SOIT LE TYPE DE FICHIER !
	if (  $self->GetFormat() eq 'GENBANK' &&  $self->{'__assembly_needed'}== &FALSE )	
	{
		$nb_fasta_sequences=1;
	}
	# the sequence comes either from genbank or a fasta file, a golden path is provided => maybe the assembly is needed
	else  
	{
		if ( defined($$self{'__seq_file'}) )	# from a multifasta or singlefasta file
		#if ( $$self{'__seq_file'} ne '' )	# from a multifasta or singlefasta file
		{
			($nb_fasta_sequences,$$self{'__listofseqfiles'}) = $self->ExtractFastaFiles($$self{'__seq_file'},$$self{'__tmpdir'},$chr);
		}
		my @a_seqfiles = split(' ',$$self{'__listofseqfiles'});
		$nb_fasta_sequences = $#a_seqfiles+1;
	}
	print STDERR "Nb of fasta sequences $nb_fasta_sequences \n" if ( $self->GetDebug() );

	$file_chr_tmp =  $$self{'__listofseqfiles'} if ( $nb_fasta_sequences == 1 );
	$file_chr_tmp =~ s/\s+$//;
	my $assembly_required = 0;

	if ( $self->GetFormat() eq 'GENBANK' || $self->GetFormat() eq 'GFF3' )
	{
		my ($acc) = $$self{'__listofseqfiles'} =~ /seqfastmp.$chr.(\S+)/;
		if ( defined $acc )
		{
			# even if there is only one sequence one can need to "assemble" the sequence by adding strech of 'N' at the beginning (ex chr14 human)
			# so the assembly is not required only when the single sequence does not appear in the golden path file (general case)
			$assembly_required = ( ! defined($rh_contigs->{$acc}->{chr_start}) ) ? 0 : 1;
			$assembly_required = 1 if ( $nb_fasta_sequences > 1 );
		}
	}

	if ( $assembly_required )
	{
		print STDERR ">>>>>>>>>>Assembly required<<<<<<<<<<<<<<\n" if ( $self->GetDebug() );
		my ($start_current,$end_precedent) = (0,0);

		$file_chr_tmp = "$$self{'__tmpdir'}/seqtmp.$code_sp$chr.fasta";
		my $fh_chr_tmp = new IO::File ($file_chr_tmp,'w') or Carp::croak("BuildChromosomeFile/Error, unable to create filename >$file_chr_tmp<\n");
		#open(CHR_TMP,">$file_chr_tmp") or Carp::croak("BuildChromosomeFile/Error, unable to create filename >$file_chr_tmp<\n");

		my $ra_list_contig = $rh_contigs->{$chr};
		foreach my $it (@{$ra_list_contig})
		{
			next if ( $it =~ /^start$/ );

			$start_current = ($it =~ /^end$/) ? $rh_contigs->{$chr."_end"} +1 : $rh_contigs->{$it}->{chr_start};

			my $diff = $start_current - $end_precedent;
			# seb_10/08/2011 : print $fh_chr_tmp $diff x "N"; ?
			for(my $i=1 ; $i<$diff ; $i++)
			{
				print $fh_chr_tmp "N";
			}

			if ( $it !~ /^start$/ && $it !~ /^end$/ )
			{
				my $file_contig = "$$self{'__tmpdir'}/seqfastmp.$chr.$it";
				my $seq_in =  new Bio::SeqIO (-file => $file_contig, -format => 'fasta' );
				my $o_seq = $seq_in->next_seq;
				
				#THOMAS In some situations the contig (scaffold) is not completely included in the assembly
				#       and we need to extract a subseq of this contig according to the information provided
				#       by the assembly file (relevant only for agp files but works also with ContigMd files)
				my ($start,$end) = @{$rh_contigs->{$it}}{qw( feature_start feature_end )};
				my $subseq_obj = Bio::Seq->new( -seq => $o_seq->subseq($start,$end) );			                                    				
				my $ctg_subseq = ( $rh_contigs->{$it}{orientation} eq '+' ) ?  $subseq_obj->seq :  $subseq_obj->revcom->seq;				
				#my $chr_seq = ($rh_contigs->{$it}->{orientation} eq '+') ? $o_seq -> seq : $o_seq -> revcom -> seq;
				print $fh_chr_tmp $ctg_subseq;
				
				unlink($file_contig);
			}
			$end_precedent = $rh_contigs->{$it}->{chr_stop};
			last if ( $it =~ /^end$/ );
		}
		$fh_chr_tmp -> close();
	}

	if( !defined $file_chr_tmp || '' eq $file_chr_tmp )
	{
		my $msg = "an important temporary file cannot be written,";
		$msg   .= " check your ASSEMBLY and the group_label of the FILE_ASSEMBLY.\n";
		Carp::croak( $msg );
	}

	return $self->__ReformatFasta($file_chr, "$code_sp$chr",$file_chr_tmp,! defined($$self{'__seq_file'}));
}

=head2 function __ReformatFasta

 Title        : __ReformatFasta
 Usage        : my $len = &__ReformatFasta($file_chr, $file_chr_tmp, $remove_tmp_file)
 Prerequisite : The fasta sequence in $file_chr_tmp must be on a single line
 Fonction     : Reformat a fasta file as expected by Narcisse (header = >species_code$chromosomeID\nline=60 char
 Returns      : the chromosome length
 Args         : $file_chr: the filename of the fasta assembly of the chromosome sequence
                $id: sequence id
                $file_chr_tmp: the filename of the raw fasta assembly of the chromosome sequence
                $remove_tmp_file: a flag to remove (true) or not (false) the temporary sequence
 Globals      : none

=cut

sub __ReformatFasta
{
	my ($self,$file_chr, $seqid, $file_chr_tmp, $remove_tmp_file) = @_;

	if ( $self->GetDebug() )
	{
		print STDERR "chr: $file_chr\n";
		print STDERR "tmpchr: $file_chr_tmp\n";
	}
	open(CHR,">$file_chr") or Carp::croak("__ReformatFasta/Error: unable to create filename >$file_chr<\n");
	print CHR ">$seqid\n";

	open(CHR_TMP,"$file_chr_tmp") or Carp::croak("__ReformatFasta/Error: unable to read filename >$file_chr_tmp<\n");
	my $len = 0;
	my $c = getc(CHR_TMP);
	if( !defined $c )
	{
		my $msg = "__ReformatFasta/Error: $file_chr_tmp seems empty.\n";
		Carp::croak( $msg );
	}

	seek(CHR_TMP,0,0);
	my $line = <CHR_TMP> if ( $c eq '>' );
	my $seq_chr = "";
	while ( !eof(CHR_TMP) )
	{
		read(CHR_TMP,$seq_chr,60);	# sequence is in a single line
		$len += length($seq_chr);
		print CHR "$seq_chr\n";
	}
	close(CHR_TMP);
	close(CHR);
	if ( $remove_tmp_file )
	{
		unlink("$file_chr_tmp");
	}
	return $len;
}

=head2 Function ExtractFastaFiles

 Title        : ExtractFastaFiles
 Usage        : my ($nb,$listoffiles)=&ExtractFastaFiles($file_chr, $tmp_dir, $chr)
 Prerequisite : none
 Fonction     : Extracts the sequences found in the multifasta source file, and creates in $tmp_dir 1 file / sequence
 Returns      : the number of files and the list of filenames
 Args         : $file: fasta or multifasta filename
                $tmpdir: output directory
                $chr: chromosome ID
 Globals      : none

=cut

sub ExtractFastaFiles
{
	 my ($self,$file,$tmpdir,$chr)= @_;

	my $list = "";
	my $cpt=0;

    my $file_obj = &GetStream($file);
    my $f = new IO::File($file_obj) or Carp::croak("Cannot open $file_obj\n");
    my $g = undef;
    while( my $line = <$f>)
    {
        chomp($line);
        if ($line =~ /^>(\S+)/)
        {
            my $id = $1;
			my $contig = $id;
            if ($id =~ /^gi\|.+\|.+\|(.+)\|/)
            {
                $contig = $1;
                # JER $contig =~ s/\..*//;    # NW_929744.1 => NW_929744
			}
			if ( $contig eq '' )
            {
                Carp::croak("ERROR - fasta file badly formatted\n");
            }
			my $tmp_file = $tmpdir."/seqfastmp.$chr.$contig";
			undef $g if ( defined($g) );
			print STDERR "Creates: $tmp_file\n" if ( $self->GetDebug() );
			$g = new IO::File(">$tmp_file");
			print $g ">$contig\n";
			$list .= "$tmp_file ";
			$cpt++;
        }
        else
        {
			$line =~ s/[\s\d\.]//g;
            print $g "$line";
        }
    }
	undef $g if ( defined($g) );
	undef $f;
	print STDERR "List Of Sequences: $list\n" if( $self->GetDebug() );

	return ($cpt,$list);
}

=head1 PRIVATE METHODS

=cut

=head2 function __SearchMRNA

 Title        : __SearchMRNA
 Usage        : $self->__SearchMRNA ($o_locus,$locus_tag)
 Prerequisite : none
 Function     : Search a mRNA with a given transcript id and matching the coordinates of the given locus
                Matches means that $o_locus is INCLUDED inside the retrieved mrna
 Returns      : $o_mrna: the mrna retrieved, undef if not found
 Args         : $o_locus: a Locus::Mrna object
                $locus_tag: the locus_tag value of the transcript to search
                $cnt [optional, default = 1] If specified, looks for the nth mRNA matching the given condition
 Globals      : none

=cut

sub __SearchMRNA
{
	my($self,$o_locus,$locus_tag,$cnt) = @_;
	$cnt = 1 unless defined $cnt;

	my $rh_index = $self -> { __idx_mrnas };
	if (!defined ( $rh_index -> { $locus_tag } ) )
	{
		print STDERR "Warning: unable to find the mRNA $locus_tag\n" if $self->GetDebug();
		return undef;
	}
	else
	{
		# Looping among the mrnas with same locus_tag, we return the nth mrna
		# surrounding $o_locus
		my $ra_mrnas = $rh_index -> { $locus_tag };
		my ($start,$end) = $o_locus -> GetEndsContig();
		my $i = 1;
		foreach my $o_mrna (@$ra_mrnas)
		{
			my ($mrna_start,$mrna_end) = $o_mrna -> GetEndsContig();
			if ( ($start >= $mrna_start) && ($start <= $mrna_end) && ($end >=$mrna_start) && ($end <= $mrna_end) )
			{
				#print STDOUT "Trouve: $locus, $i, ($start >= $gene_start) && ($start <= $gene_end) && ($end >=$gene_start) && ($end <= $gene_end)\n";
				if ($i == $cnt)
				{
					return $o_mrna;
				}
				else
				{
					$i++;
				}
			}
		}
		print STDERR "Warning: unable to find the mRNA $locus_tag\n";
		return undef;
	}
}	

=head2 function __SearchGene

 Title        : __SearchGene
 Usage        : $self -> __SearchGene($o_locus,$locus_tag)
 Prerequisite : none
 Function     : Search a gene with a given locus_tag which matches the coordinates of a given locus
                Matches means that $o_locus is INCLUDED inside the retrieved gene
                ===> NOOOOOOOOOO means that $o_locus and the gene OVERLAP !
 Returns      : The Locus::Gene found, undef if none found
 Args         : $o_locus  : A locus object
                $locus_tag: A locus tag
 Globals      : none

=cut

sub __SearchGene
{
	my($self,$o_locus,$locus_tag) = @_;

	my $rh_index = $self -> { __idx_genes };
	if (!defined ( $rh_index -> { $locus_tag } ) )
	{
		return undef;
	}
	else
	{
		# Looping among the genes with same locus_tag, we return the first gene
		# surrounding $o_locus
		my $ra_genes = $rh_index -> { $locus_tag };
		my ($start,$end)         = $o_locus -> GetEndsContig();
		foreach my $o_gene (@$ra_genes)
		{
			my ($gene_start,$gene_end) = ($o_gene -> GetEndsContig());
			if ( ($start>=$gene_start && $start<=$gene_end) || ($end>=$gene_start && $end <= $gene_end) )
			{
				#print STDOUT "Trouve: $locus, $i, ($start >= $gene_start) && ($start <= $gene_end) && ($end >=$gene_start) && ($end <= $gene_end)\n";
				return $o_gene;
			}
		}
		return undef;
	}
}	

=head2 function GetSequenceFile, GetAnnotationFile

 Title        : GetSequenceFile, GetAnnotationFile
 Usage        : my $seq_file   = $self -> GetSequenceFile();
                my $annot_file = $self -> GetAnnotationFile();
 Function     : Accessors
 Returns      : Sequence or annotation file names
 Args         : None
 
=cut
sub GetSequenceFile
{
	my $self = shift;
	return $self -> { __seq_file };
}
sub GetAnnotationFile
{
	my $self = shift;
	return $self -> { __infile };
}

=head2 function GetGenomicSequence

 Title        : GetGenomicSequence
 Usage        : my $r_seq = $self -> GetGenomicSequence();
 Prerequisite : The function __LoadOneSeq should have been already called, ie the method DumpFastaDb should have been called
 Function     : Get a ref to the genomic sequence
 Returns      : The ref to the genomic sequence (can be huge data)
 Args         : None
 Globals      : none
 
=cut
sub GetGenomicSequence
{
	my $self = shift;
	return $self -> { __genomic_sequence };
}

=head2 function __LoadOneSeq

 Title        : __LoadOneSeq
 Usage        : $self -> LoadOneSeq($fastafile);
 Prerequisite : Only one sequence in the fasta file
 Function     : Load a sequence from a fastafile, keep the ref to the sequence inside the object
 Returns      : the sequence length
 Args         : $fastafile: a fasta file
 Globals      : none
 
=cut

sub __LoadOneSeq
{
	my $self        = shift;
	my ($fastafile) = @_;

	my $r_seq = $self -> { __genomic_sequence };
	print STDERR "Loading $fastafile\n";

	open(SEQ,$fastafile) or Carp::croak("ERROR: AnnotatedSeqFile::__LoadOneSeq Unable to read >$fastafile<\n");
	while(my $line=<SEQ>)
	{
		chomp($line);
		next if ( $line =~ /^>/ );
		$line =~ s/[0-9\s]//gi;
		$line =~ s/[^ACGTRYMKWSBDHVN]/N/gi;
		$line = "\U$line";
		$$r_seq .= $line;
	}
	close(SEQ);

	my $len = length($$r_seq);

	print STDERR "$fastafile loaded len=$len\n";
	if ( $len == 0 )
	{
		#Carp::croak("$0/LoadOneSeq: Fatal Error: failed to load $fastafile\n");
		#TF allows empty fasta sequences (compatibility for random chromosomes)
		Carp::carp("$0/LoadOneSeq: Warning: failed to load $fastafile\n");
	}
	return $len;
}


=head2 Validation

 Title    : Validation
 Usage    : $o_annot -> Validation()
 Procedure: FOR DEBUG ONLY - Execute some sanity checks on our objects
 
=cut
sub Validation
{
	my $self = shift;
	warn "Starting validation";
	
	# Verifying that EACH gene is properly indexed
	$self -> __GeneIndexationValidate();
	
	# Verifying that EACH mrna is properly indexed
	$self -> __MrnaIndexationValidate();
	
	# Verifying that starting from each cds we can go to $self, using GetParent()
	$self -> __ParentValidate();
	
	# Verifying that starting from each cds, the cds is inside the mrna who is inside the gene
	# Verifying also that the regions are inside the locus
	$self -> __BoundaryValidate();
	
	# Verifying that each object has a Narcisse Id, and that there are NO two equal NarcisseIds 
	$self -> __NarcisseIdValidate();
	warn "End of validation";
}

sub __GeneIndexationValidate
{
	my $self = shift;
	
	warn "__GeneIndexationValidate";

	my $ra_genes     = $self -> { __a_genes };
	my $rh_idx_genes = 	$self -> { __idx_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my $locus_tag = $o_gene -> GetLocusTag();
		if ( defined $rh_idx_genes -> { $locus_tag } )
		{
			my $ra_indexed_genes = $rh_idx_genes -> { $locus_tag };
			my $fnd = 0;
			foreach my $o_igene (@$ra_indexed_genes)
			{
				if ($o_gene == $o_igene) 
				{
					$fnd = 1;
					last;
				}
			}
			if ($fnd==0)
			{
				warn "AnnotatedSeqFile::__GeneIndexationValidate - The gene " . $o_gene->GetLocusTag() . " is NOT indexed, but the locus tag is indexed";
			}
		}
		else
		{
			warn "AnnotatedSeqFile::__GeneIndexationValidate - The gene " . $o_gene->GetLocusTag() . " is NOT indexed, neither its locus tag";
		}
	}
}


sub __MrnaIndexationValidate
{
	my $self = shift;
	
	warn "__MrnaIndexationValidate";

	my $ra_mrnas     = $self -> { __a_mrnas };
	my $rh_idx_mrnas = 	$self -> { __idx_mrnas };
	foreach my $o_mrna (@$ra_mrnas)
	{
		my $locus_tag = $o_mrna -> GetLocusTag();
		if ( defined $rh_idx_mrnas -> { $locus_tag } )
		{
			my $ra_indexed_mrnas = $rh_idx_mrnas -> { $locus_tag };
			my $fnd = 0;
			foreach my $o_imrna (@$ra_indexed_mrnas)
			{
				if ($o_mrna == $o_imrna) 
				{
					$fnd = 1;
					last;
				}
			}
			if ($fnd==0)
			{
				warn "AnnotatedSeqFile::__MrnaIndexationValidate - The mrna " . $o_mrna->GetLocusTag() . " is NOT indexed, but the locus tag is indexed";
			}
		}
		else
		{
			warn "AnnotatedSeqFile::__MrnaIndexationValidate - The mrna " . $o_mrna->GetLocusTag() . " is NOT indexed, neither its locus tag";
		}
	}
}					

sub __ParentValidate
{
	my $self = shift;
	
	warn "__ParentValidate";

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		if ($o_gene -> GetParent() != $self)
		{
			warn "AnnotatedSeqFile::__ParentValidate - The gene " . $o_gene->GetLocusTag() . " should have as parent the annotatedSeqfile object";
		}

		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my $o_cds  = $o_mrna -> GetCDS();
			if ($o_mrna -> GetParent() != $o_gene)
			{
				warn "AnnotatedSeqFile::__ParentValidate - The mrna " . $o_mrna->GetLocusTag() . " should have as parent the GENE " . $o_gene->GetLocusTag();
			}
			if( defined $o_cds && $o_cds->GetParent() != $o_mrna)
			{
				warn "AnnotatedSeqFile::__ParentValidate - The cds " . $o_cds->GetLocusTag() . " should have as parent the MRNA " . $o_mrna->GetLocusTag();
			}
		}
	}
}

sub __BoundaryValidate
{
	my $self = shift;
	
	warn "__BoundaryValidate";

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my ($gene_start,$gene_end) = $o_gene -> GetEnds();
		if ($gene_start < 1)
		{
			warn "AnnotatedSeqFile::__BoundaryValidate - The gene " . $o_gene->GetLocusTag() . " starts at " . $gene_start;
		};
		# I CANNOT CHECK THE END, BECAUSE I DO NOT KNOW THE CHR LENGTH
		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my ($mrna_start, $mrna_end) = $o_mrna -> GetEnds();
			if ($mrna_start < $gene_start)
			{
				warn "AnnotatedSeqFile::__BoundaryValidate - The mrna " . $o_mrna->GetLocusTag() . " starts at " . $mrna_start . " gene start = $gene_start";
			};
			if ($mrna_end > $gene_end)
			{
				warn "AnnotatedSeqFile::__BoundaryValidate - The mrna " . $o_mrna->GetLocusTag() . " ends at " . $mrna_end . " gene end  = $gene_end";
			};
			my $mrna_regions = $o_mrna -> GetRegions();
			if ($mrna_regions eq '')
			{
				warn "AnnotatedSeqFile::__BoundaryValidate - The mrna " . $o_mrna->GetLocusTag() . " has NO region defined";
			}
			else
			{
				my ($mrna_start_region,$mrna_end_region) = ($mrna_regions=~/^(\d+).+(\d+)$/);
				if ($mrna_start_region < $mrna_start)
				{
					warn "AnnotatedSeqFile::__BoundaryValidate - The mrna " . $o_mrna->GetLocusTag() . " has a region starting at " . $mrna_start_region . " mrna start = $mrna_start";
				};
				if ($mrna_end_region > $mrna_end)
				{
					warn "AnnotatedSeqFile::__BoundaryValidate - The mrna " . $o_mrna->GetLocusTag() . " has a region ending    at " . $mrna_end_region . " mrna end = $mrna_end";
				};
			}

			my $o_cds  = $o_mrna -> GetCDS();
			if( defined $o_cds )
			{
				my ($cds_start, $cds_end) = $o_cds -> GetEnds();
				if ($cds_start < $mrna_start)
				{
					warn "AnnotatedSeqFile::__BoundaryValidate - The CDS " . $o_cds->GetLocusTag() . " starts at " . $cds_start . " mrna start = $mrna_start";
				};
				if ($cds_end > $mrna_end)
				{
					warn "AnnotatedSeqFile::__BoundaryValidate - The CDS " . $o_mrna->GetLocusTag() . " ends at " . $cds_end . " mrna end  = $mrna_end";
				};
				my $cds_regions = $o_cds -> GetRegions();
				if ($cds_regions eq '')
				{
					warn "AnnotatedSeqFile::__BoundaryValidate - The CDS " . $o_cds->GetLocusTag() . " has NO region defined";
				}
				else
				{
					my ($cds_start_region,$cds_end_region) = ($cds_regions=~/^(\d+).+(\d+)$/);
					if ($cds_start_region < $cds_start)
					{
						warn "AnnotatedSeqFile::__BoundaryValidate - The cds " . $o_cds->GetLocusTag() . " has a region starting at " . $cds_start_region . " cds start = $cds_start";
					};
					if ($cds_end_region > $cds_end)
					{
						warn "AnnotatedSeqFile::__BoundaryValidate - The cds " . $o_cds->GetLocusTag() . " has a region ending    at " . $cds_end_region . " cds end = $cds_end";
					};
				}
			}
		}
	}
}
			
sub __NarcisseIdValidate
{
	my $self = shift;
	
	warn "__NarcisseIdValidate";

	my %h_narcisse_id;

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my $narcisse_id = $o_gene -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			warn "AnnotatedSeqFile::__NarcisseIdValidate - The gene " . $o_gene->GetLocusTag() . " should have a Narcisse Id";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				warn "AnnotatedSeqFile::__NarcisseIdValidate - gene " . $o_gene->GetLocusTag() . " its narcisse Id $narcisse_id is already used";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my $narcisse_id = $o_mrna -> GetNarcisseId();
			if ( ! defined $narcisse_id )
			{
				warn "AnnotatedSeqFile::__NarcisseIdValidate - The mrna " . $o_mrna->GetLocusTag() . " should have a Narcisse Id";
			}
			else
			{
				if (exists $h_narcisse_id { $narcisse_id })
				{
					warn "AnnotatedSeqFile::__NarcisseIdValidate - mrna " . $o_mrna->GetLocusTag() . " its narcisse Id $narcisse_id is already used";
				}
				$h_narcisse_id { $narcisse_id } = 1;
			}
			
			my $o_cds = $o_mrna -> GetCDS();
			if( defined $o_cds )
			{
				$narcisse_id = $o_cds -> GetNarcisseId();
				if ( ! defined $narcisse_id )
				{
					warn "AnnotatedSeqFile::__NarcisseIdValidate - The cds " . $o_cds->GetLocusTag() . " should have a Narcisse Id";
				}
				else
				{
					if (exists $h_narcisse_id { $narcisse_id })
					{
						warn "AnnotatedSeqFile::__NarcisseIdValidate - cds " . $o_cds->GetLocusTag() . " its narcisse Id $narcisse_id is already used";
					}
					$h_narcisse_id { $narcisse_id } = 1;
				}
			}
		}
	}
	
	my $ra_ncrnas = $self -> { __a_ncrnas };
	foreach my $o_ncrna (@$ra_ncrnas)
	{
		my $narcisse_id = $o_ncrna -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			warn "AnnotatedSeqFile::__NarcisseIdValidate - The ncrna " . $o_ncrna->GetLocusTag() . " should have a Narcisse Id";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				warn "AnnotatedSeqFile::__NarcisseIdValidate - ncrna " . $o_ncrna->GetLocusTag() . " its narcisse Id $narcisse_id is already used";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
	}
	
	my $ra_repeats = $self -> { __a_repeats };
	foreach my $o_repeat (@$ra_repeats)
	{
		my $narcisse_id = $o_repeat -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			warn "AnnotatedSeqFile::__NarcisseIdValidate - The repeat " . $o_repeat->GetLocusTag() . " should have a Narcisse Id";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				warn "AnnotatedSeqFile::__NarcisseIdValidate - repeat " . $o_repeat->GetLocusTag() . " its narcisse Id $narcisse_id is already used";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
	}
}

=head1 UTILITIES

=cut

=head2 function GetSeqFiles

 Title        : GetSeqFiles
 Usage        : my @a_files = &GetSeqFiles()
 Prerequisite : none
 Function     : returns the number and the list of sequence files
 Returns      : $nb_files
                @a_files
 Args         : none
 Globals      : none

=cut

#sub GetSeqFiles
#{
#my($self) = @_;

	#my @a_files=split(/[,\s]/,$$self{'__listofseqfiles'});
	#return ($#a_files+1,@a_files);
#}


=head2 function GetStream

 Title        : GetStream
 Usage        : my $str = &GetStreamoad($file)
 Prerequisite : none
 Function     : If the file ends with a known 'zipped' suffix, returns a stream on its content else the filename itself
 Returns      : a stream
 Args         : $file: a file
 Globals      : none

=cut

sub GetStream
{
my($filename) = @_;

	Carp::croak("GetStream Error: the file >$filename< should exists\n") if ( ! -e $filename );
	if ( $filename =~ /\.gz$/ )
	{
		my @a_which = `which gzip`;
		if ( $#a_which != -1 )
		{
			chomp(@a_which);
			$filename = "$a_which[0] -cd $filename|";
		}
	}
	return $filename;
}

sub GetFileDate
{
my($filename) = @_;
my(@date)=();
my($t,$modif,$jour,$mois,$an);

	($t,$t,$t,$t,$t,$t,$t,$t,$t,$modif,$t,$t,$t) = stat "$filename";
	($t,$t,$t,$jour,$mois,$an,$t,$t,$t) = localtime($modif);
	$an+=1900;
	$mois = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec') [$mois];
	if ( $an =~ /1970/ )
	{
			return "";
	}
	return  "$jour-$mois-$an";
}


sub GetRunDate2
{
my(@date)=();
my($new);

	my($t1,$t2,$t3,$dd,$mm,$yy,$t4,$t5,$t6) = localtime(time);
	$yy+=1900;
	$new = sprintf("%4d%02d%02d",$yy,$mm+1,$dd);
	return ("$new");
}


sub GetRunDate
{
my(@date)=();
my($new);

        @date=split(' ',`date`);

        $new = sprintf("%02d-%3s-%4d",$date[2],$date[1],$date[$#date]);

        return ("$new");
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut



1;
