#
# $Id: Genbank.pm 292 2007-02-19 14:30:09Z faraut $
#
 
package AnnotatedSeqFile::Ensembl;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeqFile::Ensembl - Use Ensembl API to retrieve remote data

=head1 SYNOPSIS

   my %h_param=('seq'          => "sequence_file",        # Optional: either a single sequence file or a multifasta 
                                                          # file (which will be assembled using the seqctg) 
                'chr'          => "chromosome_id",        # Mandatory
                'species'      => "short_species_name",   # Mandatory
                'kingdom'      => "species kingdom",      # Mandatory: either bacteria or eukaryota
                'outformat'    => "output format",        # default=NARCISSE
                'tmpdir'       => "output directory"      # default=.

				
   my $o_ens = New AnnotatedSeqFile::Genbank(%h_param);

   see the description of other public methods in the SYNOPSIS of AnnotatedSeqFile.pm

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeqFile;
use Data::Dumper;
use Bio::EnsEMBL::Registry;
@AnnotatedSeqFile::Ensembl::ISA = qw(AnnotatedSeqFile);


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 292 $ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_ens = New AnnotatedSeqFile::Ensembl(...);
 Prerequisite : none
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         : see SYNOPSIS
 Globals      : none

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	#die "SORRY - Ensembl DATA FORMAT IS NOT SUPPORTED FOR THE 2.1 RELEASE OF NARCISSE\n";

	$self->SUPER::_Init(%h_param);

	my $seq_file = $self -> GetSequenceFile();
	if ( $seq_file eq '' || ! -r $seq_file || -z $seq_file )
	{
		throw LipmError::ParameterException ("Sequence file ($seq_file) not found or empty", 'value');
	}
	
	$self -> { __org_ensembl } = $h_param{ org_ensembl };
	$self -> { __ensembl_host} = $h_param{ ensembl_host };
	$self -> { __ensembl_user} = $h_param{ ensembl_user };

	#$$self{'__seq'}=$h_param{seq} 		if ( defined($h_param{seq}));
	#$$self{'__informat'}='ENSEMBL';
	#$$self{'__listofseqfiles'}="";
	
	#if ( ! defined($$self{'__seq'}) || ! -e $$self{'__seq'} || -z $$self{'__seq'}  )
	#{
	#	Carp::croak("Fatal error in constructor: $$self{'__seq'} not defined or not exists or empty\n");
	#}
	#else
	#{
	#	my $nb_fasta_sequences=0;
	#	($nb_fasta_sequences,$$self{'__listofseqfiles'}) = $self->ExtractFastaFiles($$self{'__seq'},$$self{'__tmpdir'},$$self{'__chr'});
	#}

}
sub GetFormat
{
	return 'ENSEMBL';
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_ens->Load();
 Prerequisite : the infile must be in genbank format and should contain unique "locus_tag"
 Function     : - Load the protein-coding genes, the ncRNA and the repeat for a given chromosome from EnsEMBL databases.
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my ($self) = @_;

	my $chr     = $self -> { __chr };
	my $species = $self -> { __species };
	my $org     = $self -> { __org_ensembl };
	my $host    = $self -> { __ensembl_host};
	my $user    = $self -> { __ensembl_user};

	my $registry = 'Bio::EnsEMBL::Registry';
	
	# connexion EnsEMBL.
	$registry->load_registry_from_db ( -host => $host, -user => $user);

	# Slice ( = chromosome )
	my $chr_tmp = ( $chr =~ /_r/ ) ? "${chr}andom" : $chr;	# the id has been shortened for random chromosomes
	#my $slice_adaptor = $registry->get_adaptor($species, 'Core', 'Slice');
	my $slice_adaptor = $registry->get_adaptor($org, 'Core', 'Slice');
	my $chr_slice = $slice_adaptor->fetch_by_region('chromosome', $chr_tmp);
	
	#if ( ! defined($$self{'__seq'}) )
	#{
	#	# save the accession sequence
	#	my $seqfile = "$$self{'__tmpdir'}/seqfastmp.$$self{'__chr'}";
#
#		open(FASTA,">$seqfile") or Carp::croak("Error: failed to create filename >$seqfile<");
#		print FASTA ">$species.$chr\n";
#		print FASTA $chr_slice->seq;
#		close(FASTA);
#		$$self{'__listofseqfiles'} .= "$seqfile ";
#	}
	
	# Genes
	#	my $gene_adaptor = $registry->get_adaptor($species, 'Core', 'Gene');
	my $gene_adaptor = $registry->get_adaptor($org, 'Core', 'Gene');
	my $ra_genes = $gene_adaptor->fetch_all_by_Slice($chr_slice);
	my $nb_genes = @$ra_genes;
	print "Downloaded $nb_genes genes (coding or not coding)\n";
	my $cnt_genes = 0;
	foreach my $o_gene (@$ra_genes)
	{
		$cnt_genes++;
		if ( $cnt_genes % 100 == 0 )
		{
			print "$cnt_genes" . '/' . "$nb_genes" . "\n";
		}
		
		my $biotype = $o_gene->biotype;
		
		#print $o_gene->display_id." biotype: ".$biotype." status:".$o_gene->start."..".$o_gene->end."\n";
		
		next if ($biotype =~ /pseudogene/);			
			
		my $coding = ($biotype eq 'protein_coding') ? 1 : 0;		
		my %h_param = ();
		$h_param{start} = $o_gene->start;
		$h_param{end} = $o_gene->end;
		$h_param{strand} = $o_gene->strand;
		$h_param{regions} = "$h_param{start}..$h_param{end}";
		$h_param{locus_tag} = $o_gene->stable_id;

		#warn "KOUKOU ".$h_param{locus_tag};

		# Adding a non coding RNA
		if ( ! $coding )
		{
			$h_param{type} = $biotype;
			
			my $o_rna = New Locus::Rna(%h_param);
			$self->AddRna($o_rna);
			next;
		}

		$h_param{type} = "gene";
		
		# Nom du gene, qui sera indique dans la colonne 'symbol' du fichier de genes. 
		# Si le HGNC symbol est renseigne on l'utilise en nom de gene. Sinon on utilise le 'RefSeq DNA'.
		# si aucun des 2 n'est renseigne, on reutilise l'identifiant EnsEMBL.
		$h_param{gene_name} = &get_GeneXref($o_gene, 'HGNC Symbol');
		$h_param{gene_name} = &get_GeneXref($o_gene, 'RefSeq DNA') if (! defined $h_param{gene_name} || $h_param{gene_name} eq '');
		$h_param{gene_name} = $o_gene->display_id if (! defined $h_param{gene_name} || $h_param{gene_name} eq '');
		$h_param{accession} = $h_param{locus_tag};
		my $o_locus_gene = New Locus::Gene(%h_param);
		$self->AddGene($o_locus_gene);
		
		# transcripts
		my $ra_transcripts = $o_gene->get_all_Transcripts;
		foreach my $o_transcript (@$ra_transcripts)
		{	
			
			#print $o_gene -> display_id . " transcript: " . $o_transcript -> display_id . "\n";

			my %h_transcript_param = ();
			
			$h_transcript_param{start}     = $o_transcript->start;
			$h_transcript_param{end}       = $o_transcript->end;
			$h_transcript_param{strand}    = $o_transcript->strand;
			$h_transcript_param{type}      = 'mrna';
			$h_transcript_param{gene_name} = $h_param{gene_name};
			$h_transcript_param{locus_tag} = $o_transcript->stable_id;
			$h_transcript_param{accession} = $h_transcript_param{locus_tag};
			$h_transcript_param{parent}    = $h_param{locus_tag};
			
			my $ra_exons = 	$o_transcript->get_all_Exons;
			my @a_exons_regions;
			foreach my $exon (@$ra_exons)
			{
				push @a_exons_regions, $exon->start."..".$exon->end;
			}
			
			if ($o_transcript->strand == 1)
			{
				$h_transcript_param{regions} = join(",", @a_exons_regions);
			}
			else
			{
				$h_transcript_param{regions} = join(",", reverse(@a_exons_regions));
			}
			
			$h_transcript_param{regions} = $h_param{regions} if ($h_transcript_param{regions} eq '');
			$h_transcript_param{nb_exons} = @$ra_exons;

			#$h_transcript_param{cds_start} = $o_transcript->cdna_coding_start-1;
			#$h_transcript_param{cds_end} = $o_transcript->cdna_coding_end-1;
			#Modified by Thomas 28/11/2007
			$h_transcript_param{cds_start} = $o_transcript->cdna_coding_start;
			$h_transcript_param{cds_end} = $o_transcript->cdna_coding_end;
			$h_transcript_param{product} = $o_gene->description if (defined $o_gene->description);
			
			# ajoute par manu (mars 2010)
			my $translated = $o_transcript->translation;
			$h_transcript_param{ensembl_product} = $translated->stable_id if defined $translated;
			#$h_transcript_param{ensembl_product} = $o_transcript->translation->stable_id;
			
			# Ajout du mRNA
			my $o_trans = New Locus::Mrna(%h_transcript_param);
			$self->AddGeneMRNA($o_trans, $h_transcript_param{parent});
			
			# CDS start and stop
			my %h_cds_param = ();
			my $ra_cdsexons = $o_transcript->get_all_translateable_Exons;
			my @a_cdsexons_regions;
			foreach my $cdsexon (@$ra_cdsexons)
			{
				push @a_cdsexons_regions, $cdsexon->start."..".$cdsexon->end;
			}
			if ($o_transcript->strand == 1)
			{
				$h_cds_param{regions} = join(",", @a_cdsexons_regions);
			}
			else
			{
				$h_cds_param{regions} = join(",", reverse(@a_cdsexons_regions));
			}
			if ( ! defined($h_cds_param{regions}) )
			{
				$h_cds_param{regions} = "$h_cds_param{start}..$h_cds_param{end}";
			}
			
			$h_cds_param{start} = $o_transcript->coding_region_start;
			$h_cds_param{end} = $o_transcript->coding_region_end;
			$h_cds_param{strand} = $h_transcript_param{strand};
			$h_cds_param{parent} = $h_transcript_param{locus_tag};
			# le locus_tag du cds est l'identifiant de la proteine resultante
			$h_cds_param{locus_tag} = $h_transcript_param{ensembl_product};
			$h_cds_param{ensembl_product} = $h_transcript_param{ensembl_product};
			$h_cds_param{type} = 'cds';
			
			my $o_cds = New Locus::Cds(%h_cds_param);
			$self->AddGeneCDS($o_cds, $h_transcript_param{parent});
		}
	}
	
	# Repeats
	my @a_repeats = @{ $chr_slice->get_all_RepeatFeatures() };
	my $nb_repeats = @a_repeats;
	print "Downloaded $nb_repeats repeats\n";
	my $cnt_repeats = 0;
	foreach my $o_repeat (@a_repeats) 
	{
		$cnt_repeats++;
		if ( $cnt_repeats % 100 == 0 )
		{
			print "$cnt_repeats" . '/' . "$nb_repeats" . "\n";
		}

		#print $o_repeat->display_id."\n";

		my %h_repeat_param = ();
		$h_repeat_param{locus_tag} = $o_repeat->display_id;
		$h_repeat_param{start} = $o_repeat->start;
		$h_repeat_param{end} = $o_repeat->end;
		$h_repeat_param{strand} = $o_repeat->strand;
		$h_repeat_param{type} = 'repeat';

		#warn "KOUKOU ".$h_repeat_param{locus_tag};
	
		my $o_repeat = New Locus::Repeat(%h_repeat_param);
		$self->AddRepeat($o_repeat);
	}
}

=head2 procedure get_GeneXref

 Title        : get_GeneXref
 Usage        : get_GeneXref($o_gene);
 Function     : returns the id of the given gene in $db_name. If there is no id for this database, the function returns undef.
 Returns      : identifier of given gene. undef if the gene does not have a known identifier for this database
 Args         : Bio::EnsEMBL::Gene gene
 				$db_name Name of the external database.
 Globals      : none

=cut

sub get_GeneXref{

	my $o_gene = shift;
	my $db_name = shift;
	
	my $ra_dbentries = $o_gene->get_all_DBLinks;
		foreach my $db_entry (@$ra_dbentries){
			if ( $db_entry->db_display_name eq $db_name ){
				return $db_entry->display_id;
		}
	}
	
	return undef;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
