#
# $Id: GFF3.pm 1154 2007-12-28 16:00:56Z gouzy $
#

package AnnotatedSeqFile::GFF3;
 
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFile::GFF3 - parse a GFF3 formatted file

=head1 SYNOPSIS

   my %h_param=('infile'       => "filename",             # Mandatory: GFF3 formatted file
                'seq'          => "sequence_file",        # Mandatory: either a single sequence file or a multifasta 
                                                          # file (which will be assembled using the seqctg) 
                'chr'          => "chromosome_id",        # Mandatory
                'species'      => "short_species_name",   # Mandatory
                'kingdom'      => "species kingdom",      # Mandatory: either bacteria or eukaryota
                'seqctg'       => "golden_path_description_file"
                'agp_assembly' => "group_label_in_seqctg_will_be_used_for_the_assembly"	# default=reference
                'outformat'    => "output format",        # default=NARCISSE
                'tmpdir'       => "output directory"      # default=.

				
   my $o_gff3 = New AnnotatedSeqFile::GFF3(%h_param);

   Warnings: the current version of the parser does not manage 
   - embedded fasta sequence => you must provide both a gff3 file (infile) and a sequence file (seq)
   - exon features

   see the description of other public methods in the SYNOPSIS of AnnotatedSeqFile.pm


=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeqFile;
use Data::Dumper;
use File::Spec;

use base qw(AnnotatedSeqFile);


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1154 $ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gff3 = New AnnotatedSeqFile::GFF3(...);
 Prerequisite : none
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         : see SYNOPSIS
 Globals      : none

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$self->SUPER::_Init(%h_param);

	my $seq_file = $self -> GetSequenceFile();
	if ( !defined($seq_file) || $seq_file eq '' || ! -e $seq_file || -z $seq_file )
	{
		throw LipmError::ParameterException('SEQUENCE FILE', 'missing');
	}
	else
	{
		my $nb_fasta_sequences=0;
		($nb_fasta_sequences,$$self{'__listofseqfiles'}) = $self->ExtractFastaFiles($seq_file,$$self{'__tmpdir'},$$self{'__chr'});
	}
}
sub GetFormat
{
	return 'GFF3';
}
=head2 procedure Load

 Title        : Load
 Usage        : $o_gff3->Load();
 Prerequisite : the infile must be in GFF3 format and should contain unique IDs
 Function     : Load the protein-coding genes, the ncRNA and the repeat contained in a GFF3 formatted file
                Warning: only the features gene|RNA|repeat_region|CDS are loaded
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my ($self) = @_;
		
	my $filename = &AnnotatedSeqFile::GetStream($self->{'__infile'});
	
	# To ensure that the gene is defined before the corresponding mRNA and mRNA before CDS,
	# we split the gff3 file in 4 temporary gff3 files
	my $f_gff3 = new IO::File("$filename") or Carp::croak("unable to read >$filename<");
	
	my (%h_f_gff3,%h_fh_gff3);
	my @a_codes = qw( gene mRNA CDS misc );
	my $tmp_dir = $self -> { __tmpdir };
	foreach my $code ( @a_codes )
	{
		$h_f_gff3  { $code } = File::Spec->catfile( $tmp_dir, "import." . $code . ".gff3" );
		$h_fh_gff3 { $code } = new IO::File($h_f_gff3{$code},"w") or Carp::croak("unable to write >$h_f_gff3{$code}<");
	}
	while( my $line = <$f_gff3>)
	{
		chomp($line);
		next if ( $line =~ /^\#/ || $line eq '');
		my ($chr,$src,$code,$start,$end,$score,$strand,$phase,$comment) = split(/\t/,$line);
		if ( defined($$self{'__file_assembly'}) ) # TF Implementing assembly for the GFF3 format
		{
			next if (! $self->_WithinAssembly($chr,$start,$end) ); #Checking if the contig, more precisely the corresponding annotation is within the assembly
			                                                       #Here $chr is in fact a contig (or a scaffold)		
		} 
		else # No assemblly file is provided, the script is working chr per chr (or scaffold per scaffold)
		{
			next if ( $chr ne $self->{'__chr'} );
		}
		###next if ( !defined($comment) || $comment !~ /ID=/ );
		next if ( $code !~ /gene|RNA|repeat_region|CDS|exon/);
		$code  = 'misc' unless (defined $h_f_gff3{$code});
		my $fh = $h_fh_gff3{$code};
		print $fh "$line\n";
	}
	foreach my $code ( @a_codes )
	{
		$h_fh_gff3 { $code }->close();
	}
	$f_gff3->close;
	print STDERR "Filtering GFF3 files finished\n" if ( $self -> GetDebug() );

	# Read the files we just wrote, in the appropriate order
	foreach my $key ( @a_codes )
	{
		my $fh_gff3 = new IO::File($h_f_gff3{$key},"r") or Carp::croak("unable to read >$h_f_gff3{$key}<");
		while( my $line = <$fh_gff3>)
		{
			chomp($line);
			my %h_param=();
			my ($chr,$src,$code,$start,$end,$score,$strand,$phase,$comment) = split(/\t/,$line);
			$h_param{start}  = $start;
			$h_param{end}    = $end;
			$h_param{strand} = ( $strand eq '+' ) ? 1 : -1;
			
			# MANU-Thomas - The hash h_contig_info is passed to the Locus object, it will add the offset if necessary
            $h_param{rh_contig_info} = $self -> GetContigInfo($chr);
            
			my @a_com=split(';',$comment);
			
			# seulement pour le PATCH AFFREUX (fev 2013)
			my ($transcript_id,$protein_id);
			
			for (my $i=0;$i<=$#a_com;$i++)
			{	
                my $pair = $a_com[$i];
                $pair =~ s/^\s+|\s+$//g ; # MANU Oct 2009 - Remove heading and trailing blanks
                next if ($pair eq "");
				my ($key,$value) = split('=',$pair,2);
				$h_param{locus_tag} = $value if ($key eq 'ID');
				$h_param{parent}    = $value if ( $key eq 'Parent');
				$h_param{gene_name} .= ",$value" if ($key =~ /Alias|Name/ && 
													 (! defined($h_param{gene_name}) || 
													 (($h_param{gene_name} !~ /$value/) || ($value !~ /$h_param{gene_name}/)))
													);
				$h_param{product} .= ",$value" if   ($key =~ /product|Note/ && 
													 (! defined($h_param{product}) || 
													 (($h_param{product} !~ /$value/) || ($value !~ /$h_param{product}/)))
													);
				# seulement pour le PATCH AFFREUX (fev 2013)
				$transcript_id = $value if ($key eq 'transcript_id');
				$protein_id    = $value if ($key eq 'protein_id'); 
			}
			$h_param{gene_name} =~ s/^,// if ( defined($h_param{gene_name}) ) ;
			$h_param{product} =~ s/^,// if ( defined($h_param{product}) ) ;
				
			# CDS or exon do not need an ID
			if ( $code =~ /gene|\w+RNA|repeat_region/ )
			{
				if (! defined $h_param{locus_tag})
				{
					print STDERR "Warning: anonymous $code, no ID value in the feature field, skipped $line\n";
				}
			}
			if ( $code =~ /gene|mRNA|CDS|exon/ )
			{
				$h_param{type} = "\L$code";
				$h_param{accession} = $h_param{locus_tag};
				
				# CECI EST UN PATCH AFFREUX POUR UTILISER LES SORTIES BioFileConverter AVEC EnsEMBL !
				# manu 18 Fev 2013
				$h_param{accession} = $transcript_id if ( $code eq 'mRNA' && $transcript_id ne '' );
				$h_param{accession} = $protein_id    if ( $code eq 'CDS' && $protein_id ne '');

				if ( $code eq 'gene' )
				{
					$h_param{regions} = "$h_param{start}..$h_param{end}";
					my $o_item = New Locus::Gene(%h_param);
					$self->AddGene($o_item);
				}
				else
				{
					if ( ! defined($h_param{parent}) )
					{
						print STDERR "Warning: orphan $code, no Parent value in the feature field, skipped $line\n";
						next;
					}
					# parent = gene
					if ( $code eq 'mRNA' )
					{
						my $o_mrna = New Locus::Mrna(%h_param);
						my $o_added = $self->AddGeneMRNA($o_mrna,$h_param{parent});
						if (! defined $o_added )
						{
							warn "Recovery mode: creation of the gene $h_param{parent} which is the Parent of the current mRNA ($h_param{locus_tag})\n";
							my $o_gene = New Locus::Gene(%h_param,'type'=>'gene','locus_tag'=>$h_param{parent},'regions'=>"$h_param{start}..$h_param{end}",'accession'=>$h_param{parent});
							$self->AddGene($o_gene);
							$self->AddGeneMRNA($o_mrna,$h_param{parent});
						}
					}

					# parent = mRNA (may be several mRNA)
					elsif ( $code eq 'CDS' )
					{
						my @a_parent = split(',',$h_param{parent});
						foreach my $parent (@a_parent)
						{
							$h_param{parent} = $parent;
							# The locus_tag defaults to the parent's locus_tag
							$h_param{locus_tag}=$h_param{parent} unless (defined $h_param{locus_tag});
							my $o_cds = New Locus::Cds(%h_param);
							$self->AddGeneMRNACDS($o_cds,$parent);
						}
					}
					elsif ( $code eq 'exon' )
					{
						my @a_parent = split(',',$h_param{parent});
						foreach my $parent (@a_parent)
						{
							# The locus_tag defaults to the parent's locus_tag
							$h_param{locus_tag}=$h_param{parent} unless (defined $h_param{locus_tag});
							my $o_exon = New Locus(%h_param);
							$self->AddGeneMRNAExon($o_exon,$parent);
						}
					}
				}
			}
			elsif ( $code =~ /(\w+RNA)$/ )
			{
				my $id = $1;
				$h_param{type} = $id;
				# CECI EST ENCORE UN PATCH AFFREUX POUR UTILISER LES SORTIES BioFileConverter AVEC EnsEMBL !
				$h_param{locus_tag} = $h_param{gene_name} if ( exists $h_param{gene_name} );

				my $o_item = New Locus::Rna(%h_param);
				$self->AddRna($o_item);
			}
			elsif ( $code =~ /repeat_region/ )
			{
				$h_param{type} = 'repeat';
				my $o_item = New Locus::Repeat(%h_param);
				$self->AddRepeat($o_item);
			}
			
			foreach my $key (keys(%h_param))
			{
				delete($h_param{$key});
			}
		}
		undef $f_gff3;
	}
	$self->_Synchronize();
}

=head2 function 

 Title        : _WithinAssembly
 Usage        : $o_gff3->_WithinAssembly($ctg,$start,$end);
 Prerequisite : an assembly file is provided  ( __file_assembly is defined )
 Function     : Test if the corresponding contig, or the corresponding annotation is included in the assembly as defined by the assembly file
 Returns      : 0|1
 Args         : $ctg, $start, $end : contig name, start and end of the annotation within the contig
 Globals      : none

=cut
sub _WithinAssembly
{
	my $self = shift;
	my ($ctg,$start,$end) = @_;
	
	my $IsWithinAssembly = 1;
	
	#The contig is not part of the assembly
	$IsWithinAssembly = 0 if (!defined($$self{'__rh_contigs'}->{$ctg}));
	
	#The annotation is not within the current assembly
	my ($f_start, $f_end) = ($$self{'__rh_contigs'}->{$ctg}{feature_start}, $$self{'__rh_contigs'}->{$ctg}{feature_end});
	$IsWithinAssembly = 0 if ( $end < $f_start || $start > $f_end );
	
	return $IsWithinAssembly;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
