#
# $Id: Genbank.pm 1127 2007-12-13 09:17:01Z tfaraut $
#
 
package AnnotatedSeqFile::Genbank;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeqFile::Genbank - parse a Genbank formatted file

=head1 SYNOPSIS

   my %h_param=('infile'       => "filename",             # Mandatory: GenBank formatted file
                'seq'          => "sequence_file",        # Optional: either a single sequence file or a multifasta 
                                                          # file (which will be assembled using the file_assembly) 
                'chr'          => "chromosome_id",        # Mandatory
                'species'      => "short_species_name",   # Mandatory
                'kingdom'      => "species kingdom",      # Mandatory: either bacteria or eukaryota
                'file_assembly'       => "golden_path_description_file"
                'assembly_label' => "group_label_in_file_assembly_will_be_used_for_the_assembly"	# default=reference
                'outformat'    => "output format",        # default=NARCISSE
                'tmpdir'       => "output directory"      # default=.

				
   my $o_gbk = New AnnotatedSeqFile::Genbank(%h_param);

   see the description of other public methods in the SYNOPSIS of AnnotatedSeqFile.pm

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeqFile;
use Data::Dumper;
@AnnotatedSeqFile::Genbank::ISA = qw(AnnotatedSeqFile);


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1127 $ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gbk = New AnnotatedSeqFile::Genbank(...);
 Prerequisite : none
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         : see SYNOPSIS
 Globals      : none

=cut

#sub _Init
#{
#	my ($self,%h_param) = @_;
#
#	$self->SUPER::_Init(%h_param);
#}
sub GetFormat
{
	return 'GENBANK';
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_gbk->Load();
 Prerequisite : the infile must be in genbank format and should contain unique "locus_tag"
 Function     : - Load the protein-coding genes, the ncRNA and the repeat contained in a Genbank formatted file
                - Extract sequence(s) files when the sequence file is not provided
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my ($self) = @_;

	my $filename   = &AnnotatedSeqFile::GetStream($self->{'__infile'});

	my $seqstream  = new Bio::SeqIO(-format => "genbank", -file => $filename);
	while( my $seq = $seqstream->next_seq )
	{
        my $version;
		  
        if (! defined($seq->accession_number()) )
        {
            Carp::croak("The accession number is not defined in $filename");
        }
		else
		{
			$version = $seq->accession_number();
		}
        
        if (defined $seq->version())
        {
            $version .= '.' .  $seq->version();
        }

		#TF next contig if the current contig is not part of the assembly (enables to treat separatly random contigs)
		next if (defined($$self{'__file_assembly'}) && !defined($$self{'__rh_contigs'}->{$version}));

		# set the offset if golden path data are loaded
		#my $offset = ( $$self{'__assembly_needed'} != 0 && defined($$self{'__rh_contigs'}->{$version}) ) ? $$self{'__rh_contigs'}->{$version}->{chr_start} : 1;
		#$offset--;
		
        # MANU ON ANNULE L'OFFSET C'EST DANS Locus MAINTENANT
        #$offset = 0;
        
		# If no sequence file provided, the sequence is embedded in the gbk file: save it as a fasta file
		if ( !defined $self -> GetSequenceFile() || $self -> GetSequenceFile() eq '' )
		{
			# save the accession sequence
			my $seqfile = "$$self{'__tmpdir'}/seqfastmp.$$self{'__chr'}.$version";

			open(FASTA,">$seqfile") or Carp::croak("Error: failed to create filename >$seqfile<");
			print FASTA ">$version\n";
			print FASTA $seq->seq;
			close(FASTA);
			$$self{'__listofseqfiles'} .= "$seqfile ";
		}

		foreach my $f ( $seq->all_SeqFeatures() )
	  	{		
			my $primary_tag = $f->primary_tag();

			next if ( $primary_tag !~ /gene|CDS|mRNA|repeat_region|\w+RNA/ );
			my %h_param = ();
			$h_param{start} = ($f->location->start);
			$h_param{end}   = ($f->location->end);
			$h_param{strand}= $f->strand;
            
            # MANU - The hasht h_contig_info is passed to the Locus object, it will add the offset if necessary
            $h_param{rh_contig_info} = $self -> GetContigInfo($version);
            
			if ( !defined($h_param{strand}) || $h_param{strand} eq '')
			{
				carp("Warning: consistency issue at [start=$h_param{start}; end=$h_param{end}; strand=$h_param{strand}], check the genbank file\n");
				next;
			}
			my %h_tags=();
			foreach my $tag ($f->all_tags())
			{
				$h_tags{$tag} = join(';',$f->each_tag_value($tag));
			}
			if ( $primary_tag eq "repeat_region" )
			{	
				$h_param{locus_tag} = $h_tags{'rpt_family'};
				$h_param{locus_tag} = "rep".$self->GetNbRepeats() if ( !defined($h_param{locus_tag}) || $h_param{locus_tag} eq '' );
				$h_param{type} = 'repeat';
				my $o_item = New Locus::Repeat(%h_param);
				$self->AddRepeat($o_item);
			}
			elsif ( $primary_tag =~  /gene|mRNA|CDS/ )
			{
# can use ID tag ?
				$h_param{locus_tag} = $h_tags{locus_tag} if ( defined($h_tags{locus_tag}) );
				$h_param{gene_name} = $h_tags{gene} if ( defined($h_tags{gene}) );
				if ( !defined($h_param{locus_tag}) || $h_param{locus_tag} eq '' )
				{
					if ( defined($h_param{gene_name}) )
					{
						$h_param{locus_tag} = $h_param{gene_name};
					}
					else
					{
						$h_param{locus_tag} = sprintf("gene%05d",$self->GetNbGenes()+1);
					}
				}
				$h_param{product} = $h_tags{product};

				my $o_item;
				if ( $primary_tag eq "gene" )
				{
					$h_param{accession} = $h_param{locus_tag};
					$h_param{type} = 'gene';
					$o_item = New Locus::Gene(%h_param);

					# NOTE - IS IT REALLY USEFUL ?
					$o_item -> AddRegion ( $h_param{start},$h_param{end} );
					$self->AddGene($o_item);
#warn Dumper { 1 => $o_item };
				}
	 			elsif ( $primary_tag eq "CDS" )
				{
					$h_param{accession} = ( defined($h_tags{protein_id}) ) ? $h_tags{protein_id} : $h_param{locus_tag};
					# JG 06/03/08 h_param{accession} =~ s/\.\d+$// if  ( defined($h_param{accession}) );	# a voir
					$h_param{type} = 'cds';
					$o_item = New Locus::Cds(%h_param);
					#warn "KOUKOU cds = " . $o_item->GetLocusTag() . " gene = " . $h_param{locus_tag};
					foreach my $loc ($f->location->each_Location)
					{
						$o_item -> AddRegion ($loc->start,$loc->end);
					}

					# NOTE - IS IT REALLY USEFUL ?
					if ( $o_item -> GetRegions().'' eq '')
					{
						$o_item -> AddRegion ( $h_param{start},$h_param{end} );
					}
					$self->AddGeneCDS($o_item,$h_param{locus_tag});
#warn Dumper { 2 => $o_item };
				}
				elsif ( $primary_tag eq "mRNA" )
				{
					$h_param{accession} = ( defined($h_tags{transcript_id}) ) ? $h_tags{transcript_id} : $h_param{locus_tag};
					# JG 06/03/08 $h_param{accession} =~ s/\.\d+$//;  # a voir
					$h_param{type} = 'mrna';
					$o_item = New Locus::Mrna(%h_param);
					foreach my $loc ($f->location->each_Location)
					{
						$o_item -> AddRegion ($loc->start,$loc->end);
					}
					# NOTE - IS IT REALLY USEFUL ?
					if ( $o_item -> GetRegions().'' eq '')
					{
						$o_item -> AddRegion ( $h_param{start},$h_param{end} );
					}
					$self->AddGeneMRNA($o_item,$h_param{locus_tag});
				}
			}
			elsif ( $primary_tag =~ /(\w+RNA)$/ )
			{
				my $id = $1;
				$h_param{type} = $id;
				$h_param{locus_tag} = $h_tags{locus_tag};
				$h_param{locus_tag} = "$id".$self->GetNbRNAs() if ( !defined($h_param{locus_tag}) || $h_param{locus_tag} eq "" );
				my $o_item = New Locus::Rna(%h_param);
				$self->AddRna($o_item);
			}
			%h_param = ();
			%h_tags = ();
		}
	}
	$self->_Synchronize();
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
