#
# $Id: Genetic.pm 717 2007-07-10 07:43:48Z gouzy $
#

package AnnotatedSeqFile::Genetic;
 
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeqFile::Genetic - parse a Genetic data file

=head1 SYNOPSIS

   my %h_param=('seq'          => "sequence file",
   				'chr'          => "chromosome_id",        # Mandatory
                'species'      => "short_species_name",   # Mandatory
                'kingdom'      => "species kingdom",      # Mandatory: either bacteria or eukaryota
                'outformat'    => "output format",        # default=NARCISSE
                'tmpdir'       => "output directory"      # default=.
				"expansion_factor_cm2bp" => "convert cM in a scale usable by narcisse" # default=1000 
				'deltacm'      => "length of a maker on the map (after cM expansion)", # default=100
				);

   my $o_map = New AnnotatedSeqFile::Genetic(%h_param);


=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeqFile;
use Data::Dumper;
use IO::File;
use NARCISSE_utils;
use Locus;
use Locus::Marker;
use base qw(AnnotatedSeqFile);

# TODO - There should be TWO classes deriving from AnnotatedSeqfile: Genomic and Genetic
#        This is not correct to derive Genetic from AnnotatedSeqFile, as many AnnotatedSeqfile features are NOT used in Genetic
#

use constant EXPANSION_FACTOR_CM2BP => 1000;	# to force the narcisse scale to be Kb (1 cM =~ 1 Kb, je sais c'est faux)
use constant DELTACM => 100;					# size of the maker on the map, it depends on the value of EXPANSION_FACTOR_CM2BP


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 717 $); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_fasta = New AnnotatedSeqFile::Genetic(...);
 Prerequisite : none
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         : see SYNOPSIS
 Globals      : none

=cut


sub _Init
{
	my ($self,%r_param) = @_;

	$self->SUPER::_Init(%r_param);

	$$self{'__deltacm'}=( defined($r_param{deltacm}) ) ? $r_param{deltacm} : &DELTACM;	
	$$self{'__expansion_factor_cm2bp'}= ( defined($r_param{expansion_factor_cm2bp}) ) ? $r_param{expansion_factor_cm2bp} : &EXPANSION_FACTOR_CM2BP;

    # An array of Locus::Markers objects
	$$self{'__a_markers'}   = [];
	
}
sub GetFormat
{
	return 'GENETIC';
}
=head2 function AddMarker

 Title        : AddMarker
 Usage        : $o_annot -> AddMarker($o_marker)
 Function     : Add a new marker
 Returns      : the index of the marker
 Args         : $o_marker: a Locus::Marker object

=cut

sub AddMarker
{
	my($self,$o_marker) = @_;
	throw LipmError::ParameterException( 'o_marker', 'type', -text => 'should by a Locus::Marker object' ) unless ref $o_marker eq 'Locus::Marker';

	my $ra_markers = $self -> { __a_markers };
	push @$ra_markers, $o_marker;

	# Set the parent and compute the Narcisse Id
	$o_marker -> SetParent ( $self );
	
	return $#$ra_markers;
}

=head2 GetMarkers,GetNbMarkers

 Title    : GetMarkers
 Function : Accessor to the array of markers
 Usage    : my $ra_markers = $o_annot -> GetMarkers();
 Arg      : None
 Return   : A ref to an array of markers
 
=cut
sub GetMarkers
{
	my $self = shift;
	return $self -> { __a_markers };
}
sub GetNbMarkers
{
	my ($self) = @_;
	my $ra_markers = $self -> { __a_markers };
	return $#$ra_markers + 1;
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_fasta->Load();
 Prerequisite : none
 Function     : Do nothing as a fasta file does not contain annotation
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my($self) = @_;

	my $len = 0;
 	
	my $mapfile = new IO::File($self->{'__infile'},'r');
	my $header = <$mapfile>;
	chomp($header);
	my @a_line = split('\t',$header);
	my ($idxchr,$idxcm,$idxmarker,$idxlocustag,$idxfprimer,$idxrprimer)=(-1,-1,-1,-1,-1,-1);
	for(my $i=0;$i<=$#a_line;$i++)
	{
		$idxchr = $i if ( $a_line[$i] =~ /^#*chr/);
		$idxcm  = $i if ( $a_line[$i] eq 'cM' || $a_line[$i] eq 'location' || $a_line[$i] eq 'region');
		$idxmarker    = $i if ( $a_line[$i] eq 'symbol');
		$idxlocustag  = $i if ( $a_line[$i] eq 'locus_tag');
	}
	my $max = -1;
	my $deltacm = $self->{'__deltacm'};
	my $fact = $self->{'__expansion_factor_cm2bp'};
	print "deltacm = $deltacm\nfact = $fact\n";
	while(my $line=<$mapfile>)
	{
		chomp($line);
		my @a_line = split('\t',$line);
		next if ( $a_line[$idxchr] ne $self->{'__chr'} );
		my $pos = $a_line[$idxcm];
		my $symbol = $a_line[$idxmarker];
		my $locus_tag = $a_line[$idxlocustag];
		my %h_param=();
		next if ( !defined($pos) || $pos eq '' );
		#$h_param{type} = 'marker';
		$h_param{type} = $a_line[3];
		$h_param{strand} = 1;
		if ( $pos =~ /\.\./ )
		{
			my ($start,$end)=$pos=~ /(\d+)\.\.(\d+)/;
			next if ($start < 0 || $end < 0);
			$start *= $fact;
			$end *= $fact;
			$h_param{start} = $start;
			$h_param{end} = $end;
		}
		else
		{
			next if ( $pos < 0 );
			$pos = ($pos * $fact );
			$h_param{start} = $pos;
			$h_param{end}   = $pos+$deltacm;
		}
		$h_param{locus_tag}= $locus_tag;
		$h_param{symbol}   = $symbol;
		$h_param{gene_name}= $symbol;
		$h_param{accession}= $locus_tag;
		$h_param{regions}  = "$h_param{start}..$h_param{end}";
		$h_param{forward_primer}  = $a_line[5];
		$h_param{reverse_primer}  = $a_line[6];
		$h_param{cM}       = $a_line[$idxcm];
		$h_param{length}   = $a_line[4];
		$max = $pos if ( $pos > $max );
		#print Dumper(%h_param);
		my $o_marker = New Locus::Marker(%h_param);
		$self->AddMarker($o_marker);
	}
	$mapfile->close;
	$max = int($max+$deltacm*2);
	#print STDERR "max=$max\n";
	my $seqfile = "$$self{'__tmpdir'}/seqfastmp.$$self{'__chr'}";
	my $f_fasta = new IO::File("$seqfile","w") or Carp::croak("Error: failed to create filename >$seqfile<");
	print $f_fasta ">$self->{'__chr'}\n";
	print $f_fasta 'N' x $max;
	print $f_fasta "\n";
	$f_fasta->close;
	$$self{'__listofseqfiles'} .= "$seqfile ";
}

=head2 procedure Print

 Title        : Print
 Usage        : &Print($level);
 Prerequisite : none
 Function     : Print the data structure
 Returns      : the number of protein coding genes
 Args         : $level: if 'full' then a full report of the datastructure is printed else only a short summary
 Globals      : none

=cut

sub Print
{
	my $self = shift;
	my ($level) = @_;

	$level = '' if ( !defined($level));
	print STDOUT "chr      :\t".$self->{'__chr'}."\n";
	print STDOUT "infile   :\t".(defined ($self->{'__infile'}) ? $self->{'__infile'} : "NO ANNOTATION FILE DEFINED")."\n";
	print STDOUT "informat :\t".$self->GetFormat()."\n";
	print STDOUT "tmpdir   :\t".$self->{'__tmpdir'}."\n";
	print STDOUT "outformat:\t".$self->{'__outfmt'}."\n";
	my $nb_markers = $self->GetNbMarkers();
	print STDOUT $nb_markers." markers found\n";
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
