#
# $Id: Gene.pm 3308 2010-04-21 15:56:05Z manu $
#

package AnnotatedSeqFileWriter::Narcisse;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::Narcisse - Dumping AnnotatedSeqFile objects for Narcisse

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::Narcisse( o_annot => $o_annot );

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter, it dumps data from AnnotatedSeqFile to Narcisse files

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Bio::Seq;
use Data::Dumper;
use General;

use Locus;
use base qw( AnnotatedSeqFileWriter );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 3308 $ =~ /\d+/g); $r[0]};
}

=head2 procedure Dump

 Title        : Dump
 Usage        : $o_annot_writer -> Dump( %h_params );
 Procedure    : Dump data to Narcisse-formatted files
 
 Args         : %h_params : a hash with named parameters
                Mandatory: 'repeat'    => "repeat_filename"
                Mandatory: 'rna'       => "rna_filename"
                Mandatory: 'exon'      => "exon_filename"
                Mandatory: 'gene'      => "gene_filename"
                Mandatory: 'transcript'=> "transcript_filename"
                Mandatory: 'protein_id'=> "MRNA|CDS"

=cut
sub Dump
{
	my $self = shift;
	
	my (%h_param) = @_;
	
	# Check the parameters
	# exon is no more required
	foreach my $file ('repeat','rna','gene','transcript','protein_id')
	{
		if ( !defined($h_param{$file}) )
		{
			Carp::croak("Fatal error: Unable to dump Narcisse file, filename for $file is not defined\n");
			return;
		}
	}

	# @a_h_xxx   = The headers
	# @a_d_xxx   = The data
	my @a_h_exon       = qw( chrom element parent symbol start end strand );
	my @a_h_gene       = qw( chrom element locus_tag symbol start end strand );
	my @a_h_transcript = qw( chrom element parent locus_tag symbol accession start end strand nb_exons start_translation stop_translation protein_id regions consistency );
    my @a_h_repeat     = qw( chrom element symbol start end strand type );
    my @a_h_rna        = qw( chrom element symbol start end strand type );
    my @a_h_marker     = qw( chromosome locus_tag symbol type length forward_primer reverse_primer cM );
    my @a_h_prot       = qw( chrom element locus_tag gene_name protein_id len strand consistency product iprscan );
    my (@a_d_exon, @a_d_gene, @a_d_transcript, @a_d_repeats, @a_d_ncrnas, @a_d_marker, @a_d_prot);

	my $o_annot = $self -> _Get ( '_o_annot' );
	my $chr     = $o_annot -> GetChromosomeLabel();

	# The Genetic subclass: we work with markers, NOT with genes. However, the output files are called gene, transcript, exon... because 
	# the GUI does not make the difference
	if ( ref $o_annot eq "AnnotatedSeqFile::Genetic" )
	{
		my $ra_markers = $o_annot -> GetMarkers();
		foreach my $o_marker ( @$ra_markers )
		{
			# For the gene file
			my $uid          = $o_marker -> GetNarcisseId();
			my $locus_tag    = $o_marker -> GetLocusTag();
			my $gene_name    = $o_marker -> GetGeneName();
			my ($start,$end) = $o_marker -> GetEnds();
			my $strand       = $o_marker -> GetStrand();
			my %h_data;
			@h_data{ @a_h_gene } = ( $chr, $uid, $locus_tag, $gene_name, $start, $end, $strand );
			push @a_d_gene, \%h_data;
			
			# For the transcript file
			$uid       = $o_marker -> GetNarcisseTranscriptId();
			my $parent = $o_marker -> GetNarcisseTranscriptParent();
			my ($nb_exons, $start_trans,$stop_trans)=(1, -1, -1);
			my $region = $o_marker -> GetRegions();
			my $consistency = 'C';
			my %h_datat;
			@h_datat{ @a_h_transcript } = ( $chr, $uid, $parent, $locus_tag, $gene_name, $locus_tag, $start, $end, $strand, $nb_exons, $start_trans,$stop_trans,$locus_tag,$region,$consistency);
			push @a_d_transcript, \%h_datat;
			
			# For the exon file
			$uid    = $o_marker -> GetNarcisseExonId();
			$parent = $o_marker -> GetNarcisseExonParent();
			my %h_datae;
			@h_datae{ @a_h_exon } = ( $chr, $uid, $parent, $gene_name, $start, $end, $strand );
			push @a_d_exon, \%h_datae;
			
			# For the marker file
			my %h_datam;
			my $length = $o_marker -> GetLength();
			my $type   = $o_marker -> GetType();
			my $forward_primer = $o_marker -> GetForwardPrimer();
			my $reverse_primer = $o_marker -> GetReversePrimer();
			my $cM             = $o_marker -> GetcM();
			@h_datam{ @a_h_marker } = ( $chr, $locus_tag, $gene_name, $type, $length, $forward_primer, $reverse_primer, $cM );
			push @a_d_marker, \%h_datam;
		}
	}
	
	# the other subclasses: we work with genes
	else
	{
		my $protein_id = $h_param{'protein_id'};
		my $ra_genes   = $o_annot -> GetGenes();
		foreach my $o_gene ( @$ra_genes )
		{
			# We skip the pseudogenes
			if ( ! $o_gene->IsProteinCoding() )
			{
				#print STDERR "dump mode=$mode: Warning: ".$self->{__locus_tag}." is not a protein coding gene (no CDS), removed...\n";
				next;
			}
			
			# For the gene file
			my $uid          = $o_gene -> GetNarcisseId();
			my $locus_tag    = $o_gene -> GetLocusTag();
			my $gene_name    = $o_gene -> GetGeneName();
			my ($start,$end) = $o_gene -> GetEnds();
			my $strand       = $o_gene -> GetStrand();
			my %h_data;
			@h_data{ @a_h_gene } = ( $chr, $uid, $locus_tag, $gene_name, $start, $end, $strand );
			push @a_d_gene, \%h_data;
			
			# For the transcript and exon files
			my ($ra_d_t,$ra_d_p) = $self -> __GetMrnas ( $o_gene, $protein_id );
			push @a_d_transcript, @$ra_d_t;
			push @a_d_prot, @$ra_d_p;
			if ( defined $h_param{exon} )
			{
				my $ra_d_e = $self -> __GetExons ( $o_gene );
				push @a_d_exon, @$ra_d_e;
			}
		}
	}

	my $o_gene_ndw = New NarcisseDataWriter::OnDisk( $h_param{ gene }, \@a_h_gene );
	$o_gene_ndw -> Update ( \@a_d_gene );
	$o_gene_ndw = undef;
	
	my $o_transcript_ndw = New NarcisseDataWriter::OnDisk( $h_param{ transcript }, \@a_h_transcript );
	$o_transcript_ndw -> Update ( \@a_d_transcript );
	$o_transcript_ndw = undef;
	
	if ( defined $h_param{exon} )
	{
		my $o_exon_ndw = New NarcisseDataWriter::OnDisk( $h_param{ exon }, \@a_h_exon );
		$o_exon_ndw -> Update ( \@a_d_exon );
		$o_exon_ndw = undef;
	}
	
	my @a_h_locus = qw( chrom element symbol start end strand type );
	my $ra_repeats = $o_annot -> GetRepeats();
	foreach my $o_repeat ( @$ra_repeats )
	{
		push @a_d_repeats, $self -> __GetLocus( $o_repeat );
	}
	
	my $o_repeat_ndw = New NarcisseDataWriter::OnDisk( $h_param{ repeat }, \@a_h_locus );
	$o_repeat_ndw -> Update ( \@a_d_repeats );
	$o_repeat_ndw = undef;

	my $ra_ncrnas = $o_annot -> GetRnas();
	foreach my $o_ncrna ( @$ra_ncrnas )
	{
		push @a_d_ncrnas, $self -> __GetLocus ( $o_ncrna );
	}
	
	my $o_ncrna_ndw = New NarcisseDataWriter::OnDisk( $h_param{ rna }, \@a_h_locus );
	$o_ncrna_ndw -> Update ( \@a_d_ncrnas );
	$o_ncrna_ndw = undef;
	
	if ( defined $h_param{ marker } )
	{
		my $o_mrkr_ndw = New NarcisseDataWriter::OnDisk( $h_param{ marker }, \@a_h_marker, { sort_numerical => &FALSE }, [ 'locus_tag' ] );
		$o_mrkr_ndw -> Update ( \@a_d_marker );
		$o_mrkr_ndw = undef;
	}
	
	if ( defined $h_param{ prot } )
	{
		my $o_mrkr_ndw = New NarcisseDataWriter::OnDisk( $h_param{ prot }, \@a_h_prot, { sort_numerical => &FALSE }, [ 'element' ] );
		$o_mrkr_ndw -> Update ( \@a_d_prot );
		$o_mrkr_ndw = undef;
	}
}

=head2 procedure __GetLocus

 Title    : __GetLocus
 Usage    : $self -> __GetLocus( $o_locus )
 Function : Retrieve data corresponding to a generic locus
 
 Args     : $o_locus A Locus object
 Return   : A ref to hash: the keys of the hash are those useful for the Narcisse files

=cut
sub __GetLocus
{
	my $self     = shift;
	my ($o_locus) = @_;
	
	my $chr          = $o_locus -> GetChromosomeLabel();
	my $uid          = $o_locus -> GetNarcisseId();
	my $locus_tag    = $o_locus -> GetLocusTag();
	my ($start,$end) = $o_locus -> GetEnds();
	my $strand       = $o_locus -> GetStrand();
	my $type         = $o_locus -> GetType();

	my @a_h_locus = qw( chrom element symbol start end strand type );
	my %h_data;
	@h_data{ @a_h_locus } = ( $chr, $uid, $locus_tag, $start, $end, $strand, $type );
	
	return \%h_data;
}

=head2 procedure __GetMrnas

 Title    : __GetMrnas
 Usage    : my $ra_mrnas = $self -> __GetMrnas( $o_gene )
 Function : Retrieve from data relative to the mrnas and to the proteins
 
 Args     : $o_gene A gene object
            $protein_id How to set the protein id ? From CDS or from MRNA 
 Return   : Two ra_arrays of hashes: 
				-1 row / mrna, the keys of the hash are those useful for the Narcisse transcript file
				-1 row / prot, the keys of the hash are those useful for the Narcisse prot file

=cut
sub __GetMrnas
{
	my $self     = shift;
	my ($o_gene, $protein_id) = @_;
	
	my $chr = $o_gene -> GetChromosomeLabel();
	my $uid = $o_gene -> GetNarcisseId();

	my @a_h_mrna = qw( chrom element parent locus_tag symbol accession start end strand nb_exons start_translation stop_translation protein_id regions consistency );
	my @a_h_prot = qw( chrom element locus_tag gene_name protein_id len strand consistency product iprscan );
	my (@a_d_rmna,@a_d_prot);

	my $ra_mrnas = $o_gene -> GetMrnas();
	foreach my $o_mrna ( @$ra_mrnas )
	{
		my $locus_id = $o_mrna -> GetNarcisseLocusId();
		my $o_cds    = $o_mrna -> GetCDS();
		if ( ! defined($o_cds) )
		{
			warn "WARNING - dump Mrnas: $locus_id, Inconsistency between mRNA and CDS detected, removed...\n";
			next;
		}

		my $exonmrna = $o_mrna->GetRegions();
		my $exoncds  = $o_cds->GetRegions();
		if ( $exonmrna eq '' || $exoncds eq '' )
		{
			printf STDERR "mRNA %s: $exonmrna\n",$o_mrna->GetLocus();
			printf STDERR "CDS  %s: $exoncds\n",$o_cds->GetLocus();
		}

		my ($start_trans,$stop_trans,$nb_exons)=();
		
		# ACCES A UN MEMBRE PRIVE DE Locus::Mrna PAS BEAU - MANU
		# JG/API currently only defined in Ensembl.pm
		if ( defined($o_mrna) && defined($o_mrna->{'__cds_start'}) ) 
		{
			($start_trans,$stop_trans,$nb_exons) = ( $o_mrna->{'__cds_start'}, $o_mrna->{'__cds_end'}, $o_mrna->{'__nb_exons'});
		}

		# that is why I must keep the older way to get the cds_start
		else 
		{
			($start_trans,$stop_trans,$nb_exons) = $o_mrna -> GetTranslationStartStop();
		}
		
		# The protein id should be the cds accession, unless we want the mrna accession
		my $prot_id = ($protein_id eq 'MRNA') ? $o_mrna->GetAccession() : $o_cds -> GetAccession();
		
		# The accession is the mrna accession
		my $accession = $o_mrna->GetAccession();
		
		# The gene name is the mrna gene name
		my $gene_name = $o_mrna -> GetGeneName();

		# The transcript and protein id
		my $tuid        = $o_mrna -> GetNarcisseId();
		my $puid        = $tuid;
		$puid =~ s/t(\d+)$/p$1/;
		my $consistency = $o_cds -> GetConsistency();
		
		my ($start,$end) = $o_mrna -> GetEnds();
		my $strand       = $o_mrna -> GetStrand();
		my $regions      = $o_mrna -> GetRegions();

		my %h_rmna_row;
		@h_rmna_row { @a_h_mrna } = ($chr,$tuid,$uid,$locus_id,$gene_name,$accession,$start,$end,$strand,$nb_exons,$start_trans,$stop_trans,$prot_id,$regions,$consistency);
		push @a_d_rmna, \%h_rmna_row;

		
		# No protein if inconsistent
		if ($consistency ne 'I')
		{
			my %h_prot_row;
			my $len     = $o_cds -> GetProteinSequenceLength();
			my $product = $o_cds -> GetProduct();
			@h_prot_row { @a_h_prot } = ($chr,$puid,$locus_id,$gene_name,$prot_id,$len,$strand,$consistency,$product,'');
			push @a_d_prot, \%h_prot_row;
		}
	}
	return (\@a_d_rmna,\@a_d_prot);
}

=head2 procedure __GetExons

 Title    : __GetExons
 Usage    : my $ra_exons = $self -> __GetExons( $o_gene )
 Function : Retrieve data relative to the exons
 
 Args     : $o_gene A gene object
 Return   : An ra_array of hashes: 1 row / exons, the keys of the hash are those useful for the Narcisse transcript file

=cut
sub __GetExons
{
	my $self     = shift;
	my ($o_gene) = @_;

	my $chr = $o_gene -> GetChromosomeLabel();
	my $uid = $o_gene -> GetNarcisseId();

	my @a_h_exon = qw( chrom element parent symbol start end strand );
	my @a_data;
	my $ra_mrnas = $o_gene -> GetMrnas();
	foreach my $o_mrna (@$ra_mrnas)
	{
		my $locus_id = $o_mrna -> GetNarcisseLocusId();
		my $tuid     = $o_mrna -> GetNarcisseId();
		my @a_exon   = $o_mrna->GetRegions();
		
		my $j=1;
		foreach my $exon (@a_exon)
		{
			my $locuscds_id = sprintf("${locus_id}_%d",$j);
			my ($start,$end) = split(/\.\./,$exon);
			my $luid = $tuid .'e' . $j++;
			my %h_row;
			@h_row{ @a_h_exon } = ( $chr, $luid, $tuid, $locuscds_id, $start, $end, $o_mrna->GetStrand() );
			push @a_data, \%h_row;
		}
	}
	return \@a_data;
}

=head2 procedure DumpFastaDb

 Title        : DumpFastaDb
 Usage        : $o_annot_writer -> DumpFastaDb ( %h_param )
 Procedure    : Dump the genes objects to 4 fasta-formatted files
 Args         : Mandatory: 'gene' => "gene_fastadb_filename"
                Mandatory: 'prot' => "protein_fastadb_filename"
                Mandatory: 'cds'  => "cds_fastadb_filename"
                Mandatory: 'mrna' => "cds_transcript_filename"
                Mandatory: 'protein_id'=> "MRNA|CDS"
 Globals      : none

=cut

sub DumpFastaDb
{
	my $self = shift;
	my (%h_param) = @_;
	
	my $rh_dumped_sequences = $self -> _Get ( '_rh_fasta_sequences' );
	my $o_annot  = $self -> _Get ( '_o_annot' );
	my $ra_genes = $o_annot -> GetGenes();
	my $protein_id = $h_param{'protein_id'};
	foreach my $type ('prot','cds','gene','mrna')
	{
		Carp::croak("Fatal error: Unable to dump Narcisse file, filename for $type is not defined\n") unless (defined $h_param{$type} );
		
		my $fh_out = new IO::File($h_param{$type},"a") or Carp::croak("Error: unable to create >$h_param{$type}<\n");
		foreach my $o_gene (@$ra_genes)
		{
			$rh_dumped_sequences -> { $type } += $self -> __DumpFastaFeature ( $fh_out,$o_gene,$type, $protein_id );
		}
		$fh_out -> close();
	}
}

=head2 Function __DumpFastaFeature

 Title        : __DumpFastaFeature
 Function     : Print on the $fh_out filehandle the fasta sequences corresponding to the CDS, the proteins, the mrna or the gene
				For mod = prot (dumping proteins), we call the GetTranslation function from Locus::Cds, MAY BE MODIFYING THE CONSISTENCY FLAG
                Warning: only protein coding genes are considered
 Usage        : print $o_gene->DumpFastaNarcisse($fh_out,$r_seq,$mode)
 Returns      : The number of printed sequences
 Args         : $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
                $mode      : prot|mrna|cds|gene
                $protein_id: CDS|MRNA
 Globals      : none

=cut

sub __DumpFastaFeature
{
	my $self           = shift;
	my($fh_out, $o_gene, $mode, $protein_id) = @_;

	my $o_annot = $self -> _Get ( '_o_annot' );
	my $chr     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	my $kingdom = $o_annot -> GetKingdom();
	my $r_seq   = $o_annot -> GetGenomicSequence();
	
	my $nb_obj = 0;
	
	my $code = $mode;
	if ( $mode =~ /prot|cds/ )
	{
		if ( ! $o_gene->IsProteinCoding() )
		{
			warn "dump mode=$mode: Warning: ".$o_gene->GetLocusTag()." is not a protein coding gene (no CDS), skipped...\n";
			return 0;
		}
		my $ra_mrna = $o_gene -> GetMrnas();
		foreach my $o_mrna (@$ra_mrna)
		{
			my $o_cds       = $o_mrna -> GetCDS();
			# JG/API defini seulement dans le parseur Ensembl => je ne pense pas qu'il y ait eu des tests de non regression, dont je garde egalement l'ancien code
			# TODO M'ETONNERAIT QUE CA MARCHE
			my $data_source_ensembl = ( defined($o_mrna) && $o_mrna->Get('cds_start') ne '' ) ? 1 : 0; 

			next unless defined $o_cds;
			if ( $o_cds->GetRegions() eq '')
			{
				warn "No region for cds ".$o_cds->GetAccession()."\n";
				next;
			}
			#my %h_param     = $o_cds->GetAll('locus_tag','product','gene_name','accession');
			#my $locus_tag = $o_cds -> GetLocusTag();
			
			# NOTE - The Locus tag found in the fast file is the GENE locus tag !
			my $locus_tag = $o_gene ->GetLocusTag();

			my $gene_name = $o_cds -> GetGeneName();
			my $accession = ($protein_id eq 'MRNA') ? $o_mrna -> GetAccession() : $o_cds -> GetAccession();
			my $product   = $o_cds -> GetProduct();

			# JG changement suite au merge API my ($len,$sseq) = $o_gene->GetSequence($r_seq,$h_param{strand},$exoncds);
			my ($len,$sseq);
			if ( $data_source_ensembl )
			{
				($len,$sseq) = $o_mrna->GetSequence($r_seq);
			}
			else
			{
				#($len,$sseq) = $o_gene->GetSequence($r_seq,$h_param{strand},$exoncds);
				($len,$sseq) = $o_cds->GetSequence($r_seq);
			}
			next if ( $len == 0 );
			my $strand      = $o_cds -> GetStrand();
			
			#my $acc = $h_param{accession}; # $r_cds->GetAccession();
			#$acc = $h_param{locus_tag} if ( $acc eq '');
			
			if ( $mode eq 'cds' )
			{
				my $gn = ( $gene_name ne '')  ? " gn=$gene_name" : "";
				$sseq  =~ s/(\w{60})/$1\n/g;
				print $fh_out ">$code:$accession locus_tag=$locus_tag$gn len=$len chr=$chr strand=$strand sp=$sp $product\n$sseq\n"; 
				$nb_obj++;
			}
			else
			{
				my $seq_prot = $o_cds -> GetTranslation ( $r_seq );
                my $len_prot = length $seq_prot;
				
				if ( $len_prot > 0 )
				{
	                # The translation is correct: format the proteic sequence for a fasta file
	                $seq_prot =~ s/(\w{60})/$1\n/g;
	                my $gn = ( $gene_name ne '')  ? " gn=$gene_name" : "";
	                
					print $fh_out ">$code:$accession locus_tag=$locus_tag$gn len=$len_prot chr=$chr strand=$strand sp=$sp $product\n$seq_prot\n";
					$nb_obj++;
				}
			}
        }
    }
	elsif ( $mode =~ /mrna/ )
	{
		my $ra_mrnas = $o_gene -> { __a_mrnas };
		foreach my $o_mrna ( @$ra_mrnas )
		{
			#my %h_param    = $o_mrna -> GetAll('locus_tag','product','gene_name','accession');
			my $locus_tag = $o_mrna -> GetLocusTag();
			my $gene_name = $o_mrna -> GetGeneName();
			my $accession = $o_mrna -> GetAccession();
			my $product   = $o_mrna -> GetProduct();

			my ($len,$sseq) = $o_mrna -> GetSequence($r_seq);
			my $strand      = $o_mrna -> GetStrand();
			next if ( $len == 0 );
			$sseq  =~ s/(\w{60})/$1\n/g;
			my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
			print $fh_out ">$code:$accession locus_tag=$locus_tag$gn len=$len chr=$chr strand=$strand sp=$sp $product\n$sseq\n";
			$nb_obj++;
		}
	}
	elsif ( $mode =~ /gene/ )
	{
		#my %h_param = $o_gene->GetAll('locus_tag','product','gene_name','accession');
		my $locus_tag = $o_gene -> GetLocusTag();
		my $gene_name = $o_gene -> GetGeneName();
		my $accession = $o_gene -> GetAccession();
		my $product   = $o_gene -> GetProduct();
		
		my ($len,$sseq) = $o_gene->GetSequence($r_seq);
		my $strand      = $o_gene -> GetStrand();
		my ($start,$end) = $o_gene->GetEnds();

		return if ( $len == 0);
		$sseq  =~ s/(\w{60})/$1\n/g;
		my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
		print $fh_out ">gene:$accession locus_tag=$locus_tag$gn len=$len chr=$chr strand=$strand sp=$sp range=${start}-${end} $product\n$sseq\n";
		$nb_obj ++;
	}
	
	return $nb_obj;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


