# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Box -

=head1 SYNOPSIS



=head1 DESCRIPTION



B<CONSTRUCTOR PARAMETERS>



=head1 SUBROUTINES



=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package Box;

use strict;
use warnings;

use base qw( LipmObject );

use LipmError::ObjectException;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

# Global class attributes
our $DEFAULT_TOP_PADDING	= 0;
our $DEFAULT_RIGHT_PADDING	= 0;
our $DEFAULT_BOTTOM_PADDING	= 0;
our $DEFAULT_LEFT_PADDING	= 0;


##################################################
###########			Constructor		 #############
##################################################
=head2  _Init

	Title        : 	_Init
	Usage        : 	_Init(%h_param)
	Prerequisite : 	none
	Procedure    : 	Initialization
	Returns      : 	reference of the blessed object
				   	else
					undef if the object can't be initialized
	Args         :	X0,Y0	=> coordinates of the box top-left corner
					X1,Y1	=> coordinates of the box bottom-right corner
	Globals	  	 : 	object attributes
	Errors		 : 	LipmError::ObjectException

=cut
sub _Init
{
	my $self = shift;

	# 0) Set default parameters
	my %h_param = (
				# Mandatory parameters
				X0	=> undef,
				Y0	=> undef,
				X1	=> undef,
				Y1	=> undef,

				# Optional parameters
				padding_top 	=> $DEFAULT_TOP_PADDING,
				padding_right	=> $DEFAULT_RIGHT_PADDING,
				padding_bottom	=> $DEFAULT_BOTTOM_PADDING,
				padding_left	=> $DEFAULT_LEFT_PADDING,

				@_);
	$self->SUPER::_Init( %h_param );


	# 1) Test mandatory parameters
	foreach my $required ( qw( X0 Y0 X1 Y1 ) )
	{
		if( !defined $h_param{ $required } )
		{
			my $pack = __PACKAGE__;
			my $dbg = "Parameter '$required' missing for $pack initialization.";

			throw LipmError::ObjectException( $pack, '!build', -text => $dbg );
		}
	}

	# 2) Store object attributes
	my @a_attributes = ('X0','Y0','X1','Y1','padding_top','padding_right','padding_bottom','padding_left');
	@$self{@a_attributes} = @h_param{@a_attributes};

	my $msg = "a Box was built from ($self->{ X0 },$self->{ Y0 })";
	$msg   .= " to ($self->{ X1 },$self->{ Y1 })\n";
	$self->_Trace( $msg, '.', 'IBV' );

	return $self;
}

=head2  SetDefaultPaddings
	Title        :	SetDefaultPaddings
	Usage        :	&SetDefaultPaddings(top,right,bottom,left)
					&Box::SetDefaultPaddings(top,right,bottom,left)
	Prerequisite :	none
	Procedure    :	Class method to change default paddings attributes.
	Returns      :	none
	Args         :	top,right,bottom,left (in pixel)
	Globals      :	$DEFAULT_TOP_PADDING, $DEFAULT_RIGHT_PADDING, $DEFAULT_BOTTOM_PADDING, $DEFAULT_LEFT_PADDING

=cut
sub SetDefaultPaddings
{
	my ($top,$right,$bottom,$left) = (shift,shift,shift,shift) or return;

	$DEFAULT_TOP_PADDING	= int($top)		if ($top >= 0);
	$DEFAULT_RIGHT_PADDING	= int($right)	if ($right >= 0);
	$DEFAULT_BOTTOM_PADDING	= int($bottom)	if ($bottom >= 0);
	$DEFAULT_LEFT_PADDING	= int($left)	if ($left >= 0);

	return;
}

=head2  GetAttributes
	Title        :	GetAttributes
	Usage        :	$self->GetAttributes()
	Prerequisite :	This object should be initialized
	Function     :	Get the attributes of this object to make a copy
	Returns      : 	%h_attributes
	Args         : 	none
	Globals      : 	none

=cut
sub GetAttributes
{
	my $self = shift;

	my %h_attributes = ();
	my @a_attributes = ('X0','Y0','X1','Y1','padding_top','padding_right','padding_bottom','padding_left');

	@h_attributes{@a_attributes} = @$self{@a_attributes};

	return %h_attributes;
}

=head2  GetTopCornerCoordinates
	Title        :	GetTopCornerCoordinates
	Usage        :	GetTopCornerCoordinates
	Prerequisite :	none
	Function     :	Get the origin of the frame
	Returns      :	(X0,Y0)
	Args         :	none
	Globals      :	X0,Y0

=cut
sub GetTopCornerCoordinates
{
	my $self = shift;
	return ($self->{X0},$self->{Y0});
}

=head2  GetBottomCornerCoordinates
	Title        :	GetBottomCornerCoordinates
	Usage        :	GetBottomCornerCoordinates
	Prerequisite :	none
	Function     :	Get the origin of the frame
	Returns      :	(X0,Y0)
	Args         :	none
	Globals      :	X0,Y0

=cut
sub GetBottomCornerCoordinates
{
	my $self = shift;
	return ($self->{X1},$self->{Y1});
}

=head2  GetBoxArea
	Title        :	GetBoxArea
	Usage        : %h_param = $self->GetBoxArea()
	Prerequisite :	none
	Function     :
	Returns      : 	%h_param
	Args         : 	none
	Globals      : 	X0,Y0,X1,Y1

=cut
sub GetBoxArea
{
	my $self = shift;


	my %h_param = (	X0 => $self->{X0},
					Y0 => $self->{Y0},
					X1 => $self->{X1},
					Y1 => $self->{Y1}
					);

	return %h_param;
}

=head2  GetDisplayedArea
	Title        :	GetDisplayedArea
	Usage        :	%h_param = $self->GetDisplayedArea()
	Prerequisite :	none
	Function     :
	Returns      : 	%h_param
	Args         : 	none
	Globals      : 	X0,Y0,X1,Y1

=cut
sub GetDisplayedArea
{
	my $self = shift;
	my $orientation = (scalar @_ ) ? shift : "+";

	my $X0 = $self->{X0}+$self->{padding_left};
	my $X1 = $self->{X1}-$self->{padding_right};
	my ($Y0,$Y1);

	if ($orientation eq "+")
	{
		$Y0 = $self->{Y0}+$self->{padding_top};
		$Y1 = $self->{Y1}-$self->{padding_bottom};

	} else
	{
		$Y0 = $self->{Y1}-$self->{padding_bottom};
		$Y1 = $self->{Y0}+$self->{padding_top};
	}

	my %h_param = (
		 			X0 => $X0,
					Y0 => $Y0,
					X1 => $X1,
					Y1 => $Y1
					);
	return %h_param;
}

=head2  GetBounds
	Title        :	GetBounds
	Usage        :	$self->GetBounds()
	Prerequisite :	none
	Function     :	Get the coordinates of the bounds of the box
	Returns      : 	(X0,Y0,X1,Y1)
	Args         : 	none
	Globals      : 	X0,Y0,X1,Y1
=cut
sub GetBounds
{
	my $self = shift;

	return ($self->{X0},$self->{Y0},$self->{X1},$self->{Y1});
}

=head2  GetPaddings
	Title        :	GetPaddings
	Usage        :	$self->GetPaddings()
	Prerequisite :	none
	Function     :	Get paddings values
	Returns      : 	%paddings
	Args         : 	none
	Globals      : 	none

=cut
sub GetPaddings
{
	my $self = shift;

	my %h_paddings = (	padding_top 	=> $self->{padding_top},
						padding_right	=> $self->{padding_right},
						padding_bottom	=> $self->{padding_bottom},
						padding_left	=> $self->{padding_left}
						);

	return %h_paddings;
}

=head2  GetJointBoxParameters
	Title        :	GetJointBoxParameters
	Usage        :	&Box::GetJointBoxParameters($first_box,$second_box [,$axis])
	Prerequisite :	none
	Function     :	Joint two boxes and return parameters to build the resulting box.
By default, this method joined the two boxes, bearing in mind two axes (X0 will be the smaller of two boxes ...)
It can bind two boxes according to a single axis ( 'x' or 'y'), in this case, the first box is the box under reference.
	Returns      : 	%h_param : new Box constructor parameters.
	Args         : 	$first_box	: first Box object to join
					$second_box	: second Box object to join
					$axis		: (x or y)
	Globals      : 	none

=cut
sub GetJointBoxParameters
{
	my ($first_box,$second_box,$axis) = (shift,shift,shift);

	my ($X0,$X1,$Y0,$Y1);

	#======================================
	# Testing mandatory parameters
	throw LipmError::ParameterException('box1', 'missing') if ( !defined $first_box || ref($first_box) !~ /Box/  );
	throw LipmError::ParameterException('box2', 'missing') if ( !defined $second_box || ref($second_box) !~ /Box/  );
	$axis = '' if (!defined $axis);


	#======================================
	# Horizontal joining
	if ($axis eq 'x')
	{
		$X0 = ($first_box->{X0} < $second_box->{X0}) ? $first_box->{X0} : $second_box->{X0};
		$X1 = ($first_box->{X1} > $second_box->{X1}) ? $first_box->{X1} : $second_box->{X1};
		$Y0 = $first_box->{Y0};
		$Y1 = $first_box->{Y1};
	}
	#======================================
	# Vertical joining
	elsif ($axis eq 'y')
	{
		$X0 = $first_box->{X0};
		$X1 = $first_box->{X1};
		$Y0 = ($first_box->{Y0} < $second_box->{Y0}) ? $first_box->{Y0} : $second_box->{Y0};
		$Y1 = ($first_box->{Y1} > $second_box->{Y1}) ? $first_box->{Y1} : $second_box->{Y1};
	}
	#======================================
	# Full joining
	else
	{
		$X0 = ($first_box->{X0} < $second_box->{X0}) ? $first_box->{X0} : $second_box->{X0};
		$X1 = ($first_box->{X1} > $second_box->{X1}) ? $first_box->{X1} : $second_box->{X1};
		$Y0 = ($first_box->{Y0} < $second_box->{Y0}) ? $first_box->{Y0} : $second_box->{Y0};
		$Y1 = ($first_box->{Y1} > $second_box->{Y1}) ? $first_box->{Y1} : $second_box->{Y1};
	}


	#======================================
	# Return new Box parameters
	my %h_param = (X0=>$X0, X1=>$X1, Y0=>$Y0, Y1=>$Y1);

	return %h_param;
}

=head2  Height
 	Title		 :	Height
 	Usage		 :	$self -> Height()
 	Prerequisite :	instanciate the object
 	Function	 :	Get the height of the frame (pixel)
 	Returns	  	 :	integer
 	Args		 :	none
 	Globals	  	 :	Y0,Y1

=cut
sub Height
{
	my $self = shift;
	my $Y0 = $self -> {Y0};
	my $Y1 = $self -> {Y1};

	return abs($Y1 - $Y0);
}

=head2  Width
	Title		 :	Width
	Usage		 :	$self -> Width()
	Prerequisite :	instanciate the object
	Function	 : 	Get the width of the frame (pixel)
	Returns	  	 :	integer
	Args		 :	none
	Globals	  	 :	X0,X1

=cut
sub Width
{
	my $self = shift;
	my $X0 = $self -> {X0};
	my $X1 = $self -> {X1};

	return abs($X1 - $X0);
}

sub GetX0
{
	my $self = shift;
	return $self -> {X0};
}

sub GetX1
{
	my $self = shift;
	return $self -> {X1};
}

sub GetY0
{
	my $self = shift;
	return $self -> {Y0};
}

sub GetY1
{
	my $self = shift;
	return $self -> {Y1};
}

=head2  CenterX
	Title		 :	CenterX
	Usage		 :	$self -> CenterX()
	Function	 : 	returns the center of the box in x coordinates
	Returns	  	 :	integer
=cut
sub CenterX
{
	my $self = shift;

	return int(($self -> {X0}+$self -> {X1})/2);
}

=head2  CenterY
	Title		 :	CenterY
	Usage		 :	$self -> CenterY()
	Function	 : 	returns the center of the box in x coordinates
	Returns	  	 :	integer
=cut
sub CenterY
{
	my $self = shift;

	return int(($self -> {Y0}+$self -> {Y1})/2);
}

=head2 function

 Title        : GetCoords
 Usage        : ( $x2,$y2 ) = $o_xx->GetCoords( $x1,$y1 );
 Prerequisite : none
 Function     : ($x1,$y1) is a point related to the box
                (0,0) is the top left corner
 Returns      : ( $x2,$y2 ) is the coords related to the whatever the box is inside
 Args         : $x1,$y1

=cut

sub GetCoords
{
	my $self = shift;
	my ( $x,$y ) = @_;

	my %h_coords = $self->GetDisplayedArea();

	return ( $h_coords{ X0 }+$x, $h_coords{ Y0 }+$y );
}



1;
__END__

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

