#
# 	sebastien.letort@toulouse.inra.fr
#	Created: January 30, 2009
#	$Id$
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package BoxesAdministrator::Karyotype;

=pod

=head1 NAME

 BoxAdministrator::Karyotype - create and manage boxes for a karyotype

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 All boxes have the same width.
 They are placed in a grid. Inside a row, all boxes have the same height.
 Height of a row depends on the length of the longuest chrom.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( BoxesAdministrator );

use LipmError::ParameterException;
use POSIX; # ceil func

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New BoxAdministrator::Karyotype( x, -o_logger => $o_logger );
 Prerequisite : none
 Fonction     : initialise an instance of BoxAdministrator::Karyotype class.
 Returns      : a valid BoxAdministrator::Karyotype object
 Args         : x
 Attributes   :

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger   => undef,
	               -rh_padding => {},
	               row        => undef,
	               col        => undef,
	               ra_chroms   => undef,
	               boxes_width => undef,
	               height      => undef,
	               @_ );

	# logger, padding
	$self->SUPER::_Init(@_);

	# mandatories
	LipmError::ParameterException::Assert( boxes_width=> $h_args{ boxes_width },
	                                       height     => $h_args{ height },
	                                       ra_chroms  => $h_args{ ra_chroms } );

	# toutes les boites auront la meme largeur
	$self->_Set( '__boxes_width', $h_args{ boxes_width } );

	# height est la hauteur de la boite du plus gd chrom
	# les autres hauteurs sont proportionnelles a elle
	my $max_len = &__GetMaxChromLength( $h_args{ ra_chroms } );
	$self->_Set( '__height_ratio', $h_args{ height } / $max_len );

	# par ce que les boites ne font rien
	$self->_Set( __transformers => {} );

	# build boxes
	my ( $nb_rows, $nb_cols );
	my $ra_chroms = $h_args{ ra_chroms };
	my $nb_chroms = scalar( @$ra_chroms );
	if( defined $h_args{ row } )
	{
		$nb_rows = $h_args{ row };
		$nb_cols = ceil( $nb_chroms / $nb_rows );
	}
	elsif( defined $h_args{ col } )
	{
		$nb_cols = $h_args{ col };
		$nb_rows = ceil( $nb_chroms / $nb_cols );
	}
	else
	{
		my $msg = "No row nor col args defined";
		throw LipmError::ParameterException( 'row and col', 'missing',
		                                     -text => $msg, -depth => 1 );
	}
	$self->__InitBoxes( $ra_chroms, $nb_rows, $nb_cols );

# warn Dumper $self;
	return $self;
}


=head2 Procedure _SetPadding

 Usage        : $o_xx->_SetPadding( %h_args );
 Prerequisite : Should be called by _Init only _ pour ca qu'elle est incluse !
                j'ai fait une fonction parce que je pense qu'elle pourrait
 Procedure    : none
 Args         : %h_args

=cut

sub _SetPadding
{
	my $self   = shift;
	my %h_args = ( top    => 1,
				   bottom => 1,
				   right  => 1,
				   left   => 1,
				   inner_width  => 1,
				   inner_height => 1,
				   @_ );

	$self->_Set( '_rh_padding', \%h_args );

	return;
}


=head2 Procedure __InitBoxes

 Usage        : $o_xx->__InitBoxes( $ra_chroms, $nb_rows, $nb_cols );
 Prerequisite : none
 Procedure    : Initialise les boites, une boite par chromosome en plusieurs rangees 
                Associe chaque boite a un FrameTransformer pour reutiliser le code d'Olivier
 Args         : $ra_chroms, ordered chroms
                $nb_rows, int
                $nb_cols, int


=cut

sub __InitBoxes
{
	my $self = shift;
	my( $ra_chroms, $nb_rows, $nb_cols ) = @_;

	my $nb_chroms = scalar( @$ra_chroms );

	# x,y positions
	my $rh_padding = $self->_Get( '_rh_padding' );
	my $y = $$rh_padding{ top };

	# largeur fixee pour la classe
	my $width = $self->_Get( '__boxes_width' );

	for( my $j=0; $j<$nb_rows; ++$j )
	{
		my $first = $j*$nb_cols;
		my $last  = $first + $nb_cols - 1;
		$last     = $nb_chroms-1 if( $nb_chroms <= $last );

		my @a_row_chroms = @$ra_chroms[ $first..$last ];
		my $max_length = &__GetMaxChromLength( \@a_row_chroms );

		my $height = ceil( $max_length * $self->_Get( '__height_ratio' ) );
		my $x = $$rh_padding{ left };

		for( my $i=0; $i<@a_row_chroms; ++$i )
		{
			my $rh_chrom = $a_row_chroms[$i];
			$self->_BuildBox( $$rh_chrom{ chr }, $x,$y, $width,$height );
			$self->__AssociateFrameTransformerToBox( @$rh_chrom{ qw( chr length ) } );

			$x += ( $width + $$rh_padding{ inner_width } );
		}
		$y += ( $height + $$rh_padding{ inner_height } );
	}

	return;
}


=head2 function __GetMaxChromLength

 Usage        : $max = &__GetMaxChromLength( $ra_chroms );
 Prerequisite : none
 Function     :
 Returns      : $max
 Args         : $ra_chroms

=cut

sub __GetMaxChromLength
{
	my ( $ra_chroms ) = @_;

	my $max = -99;
	foreach my $rh_chrom ( @$ra_chroms )
	{
		my $len = $$rh_chrom{ length };
		$max = $len if( $max < $len );
	}

	return $max;
}


# astuce pour faire le travail que devraient faire les objets Box
sub __AssociateFrameTransformerToBox
{
	my $self = shift;
	my ( $name, $length ) = @_;

	my $rh_ft = $self->_Get( '__transformers' );

	my $o_box = $self->GetBox( $name );
	my %h_coords = $o_box->GetDisplayedArea();

	my $taille_du_chrom_en_pixel = $self->_Get( '__height_ratio' ) * $length;
	my $o_ft = New FrameTransformer::1D( %h_coords,
	                                     x0 => 1, y0 => 1, # oblige de les mettre, meme si je ne sais pas tres bien ce que c'est !
	                                     x1 => 1, y1 => $length,
	                                     Y1 => $h_coords{ Y0 } + $taille_du_chrom_en_pixel );
	$$rh_ft{ $name } = $o_ft;

	return;
}

# # methode pour pas a avoir a reecrire toutes les classes d'Olivier
# # je fais passer la classe pour un objet Planner::ChromosomeMap
# sub GetPlannerInfos
# {
# 	return ();
# }

sub GetFrameTransfo
{
	my $self = shift;
	
	return $self->_Get( '__transformers' );
}
1;
