# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

a Narcisse object to manage a chromosome slice : data and representation.

=head1 DESCRIPTION

ChromosomeSlice manage ElementsSet inherited objects which compose a chromosomme representation (Genes, Rna, Cyto...).


B<Attributes :>

 - organism			=> 	short name of the organism
 - chromosome
 - start
 - end
 - graph_orientation=> 	vertical orientation of the representation. Positive orientation show data from top to bottom. ('-' or '+')
 - h_elements		=> 	hash of all elements which are initialized for chromosome representation
 - representation_type => flag to know which elements need to be load and display. Values : full,basic (default full)


=head1 SYNOPSIS

	my %h_param =  ( organism => 'organism',
					 chromosome => 'chromosome',
					 start => 'start',
					 end => 'end'
					);

	my $chromosome 	= ChromosomeSlice -> New ('/root/data/', \%h_param);
	my $image 		= Image -> New (width,height);

	# To display data :
	$chromosome -> Display($image);


=head1 SUBROUTINES

Display
GetEnd
GetMap
GetRange
GetStart
IsDataLoad

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
Letort Sebastien : sebastien.letort@toulouse.inra.fr

=cut

package ChromosomeSlice;

use strict;
use warnings;

use POSIX;
use Time::HiRes qw( gettimeofday tv_interval);

use base qw( LipmObject );

use Scalar::Util qw( blessed );
use Error qw(:try);
use Data::Dumper;

use Image;
use General;
use Box::ChromosomeSliceBox;
use ElementsSet;
use ElementsSet::Cyto;
use ElementsSet::Gene;
use ElementsSet::Contig;
use ElementsSet::Clone;
use ElementsSet::Ocean;
use ElementsSet::Personal;
use ElementsSet::Repeat;
use ElementsSet::Rna;
use ElementsSet::Snp;
use ElementsSet::MultiGraph;
use LipmError::ObjectException;
use LipmError::ParameterException;

use SuppDataLoader;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

use constant ALL  => "All";
use constant SUPP => "Supp";
use constant CORE => "Core";

# Default parameters
# If you change this, change ALSO @A_DEFAULT_FEATURES in ChromosomeMap
my @A_DEFAULT_FEATURES = qw( Gene Cyto Rna Repeat Personal Ocean Snp Contig Super Clone );

# Multigraph does not have to draw a legend (for a specific feature)
# if the legend has previously been drawn.
# This information is stored into this private class variable
# It was decided to show the legend only in reference if possible
# To respect this, reference ChromSlice will be created first
our %_H_MULTIGRAPH_LEGEND = ();

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object

 Args         : %h_param a hash of parameters
 					mandatory
 						o_species       => a SpeciesObserver
#                       o_io            => The instance observer
#                       organism        => short name of the organism
  						chromosome		=> chromosome name or number
  						o_param_graph   => ParamParser object, with Graphics.cfg data

  					optional
  						feature_list 	=> a list of features to display
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 1)
  						exon_bool			=> are the exon displayed  (default 0)
						graph_orientation 	=> the orientation (default '+')
						is_reference 		=> Is it the reference species (default 0)
						graduation_label	=> title label of the graduation (default undef)
						graduation_scale	=> exponent ten value for corresponding to the graduation scale : 1,3,6 (default undef)

 Globals	  : none

=cut
sub _Init
{
	my $self = shift;
	my %h_param = (
				# Mandatory parameters
# 				organism			=> undef,
				chromosome			=> undef,
			    start				=> 0,
				end					=> 1,
				textual             => 0,
				# Optional parameters with default value
				feature_list		=> \@A_DEFAULT_FEATURES,

				o_species           => undef,
# 				o_io                => undef,
                o_planner			=> undef,
				o_box               => undef,
				o_image             => undef,
				-o_textual          => undef,
				graph_orientation	=> '+',
				graduation_label	=> undef,
				graduation_scale	=> undef,
				representation_type => 'full',
				__full_screeen_mode => 0,
				use_cache			=> 0,
                -o_logger => undef,
				@_ );
	$self->SUPER::_Init( @_ ); # only set logger

	#-------------------------------------
	# 0) Checking mandatory parameters
	my @mandatory_attributes = qw( o_species o_planner chromosome o_image );
	foreach my $mandatory_attribute ( @mandatory_attributes ) 
	{
		(defined $h_param{ $mandatory_attribute })
			or throw LipmError::ParameterException($mandatory_attribute,'missing');
	}
	throw LipmError::ParameterException('Image object', 'missing') if (ref($h_param{o_image}) !~ /(Image)|(GD::Image)/  );

	#-------------------------------------
	# 1) Set object attributes
	my @a_keys        = keys( %h_param );
	@$self{ @a_keys } = @h_param{ @a_keys };
	
	$self -> { o_textual } = $h_param { -o_textual };
		
	#-------------------------------------
	# 2) Transfer some planner attributes	
	my $o_planner = $self->{ o_planner };	
	my @a_graph_attributes = qw( graduation_scale feature_list gene_structure gene o_param_graph o_maparea );
	@{$self}{ @a_graph_attributes } =  $o_planner->_Get( @a_graph_attributes );
	
	# This does not make sens in textual mode
	if ( $self -> {textual} == 0 )
	{
		#-------------------------------------
		# 3) Initialize some attributes.	
		# Graduation scale and labels
		# __SetGraduationScale is now in planner
		#$self->__SetGraduationScale() if (! defined $self->{graduation_scale} ); 
		$self->__SetGraduationLabel() if (! defined $self->{graduation_label} );
	
		# Creating the dna cylinders
		$self->__SetDNACylinders();
	
		#Sets the box frame transformation object for subsequent display
		$self-> __SetFrame();
	}
								
	#-------------------------------------
	# 4) Data initalization
	# Set the hash h_elements
	$self->{h_elements} = {};	
				
	#Some parameters from h_param might have been changed, so we take these changes into account
	my %h_new_param = ( -o_logger => $h_param{ -o_logger } );
	
	#The organism specific parameters
	my @a_org_parameters = qw( chromosome start end );
	@h_new_param{ @a_org_parameters } = @$self{ @a_org_parameters };
		
	#Some graphical parameters
	my @a_graph_parameters =  qw( o_box graph_orientation gene_structure gene o_param_graph o_maparea );
	@h_new_param{ @a_graph_parameters } = @$self{ @a_graph_parameters };

	#Some technical parameters
	my @a_tech_parameters =  qw( o_species use_cache );
	@h_new_param{ @a_tech_parameters } = @$self{ @a_tech_parameters };
	
	# Adding a reference on chromosome slice object for elements 
	$h_new_param{ o_chrom_slice } = $self;

	# Store each feature in an ElementSet object
	my $ra_list = $self->{ feature_list };
	foreach my $element( @$ra_list )
	{
		try
		{
            #warn "creation d'un ES::$element de $h_new_param{ start } - $h_new_param{ end }";
			$self->{h_elements}{$element} = "ElementsSet::$element"->New( %h_new_param );
		}
		catch LipmError with
		{
			my $E = shift;
			my $msg = $E->Explain(). " - element = $element\n";
			$self->_Trace( $msg, '.', 'IE' );
		};
	}

	# Supplementary data
	if( $self->{ o_planner }->IncludeSuppData() )
	{
		try
		{
			foreach my $where ( qw( SESSION INSTANCE ) )
			{
                # warn "creation d'un donnee supp de $h_new_param{ start } - $h_new_param{ end }";
				my $rh_supp_elements = $self->LoadSupplementaryData( \%h_new_param, $where );

				my $rh_elts = $self->{h_elements};
				$self->{h_elements} = { %$rh_elts, %$rh_supp_elements };
			}
		}
		catch LipmError with
		{
			my $o_err = shift;

			$self->_Trace( $o_err->Explain(), '.', 'IE' );

			$o_err->throw() if( &TRUE == $o_err->IsFatal() );
		};
	}

	return $self;
}

#------------------------------------------------#
#        Features selection and methods          #
#------------------------------------------------#
#Obsolote methods (now in ChromosomMap)
=head2 Function __SetSelectedFeatures

 Title        : __SetSelectedFeatures
 Usage        : $o_xx->__SetSelectedFeatures();
 Procedure    : Sets the feature list according to the selected features given as arguments
 Args         : none
 
#=cut
sub __SetSelectedFeatures
{
	my $self = shift;
	
	my @a_feature_list = ();
	my %h_features_batch = () ;						
	foreach my $feat ( @{ $self->{ selected_features } } )
	{
		if ( $feat eq &ALL || $feat eq &SUPP || $feat eq &CORE )
		{
			$h_features_batch{ $feat }++;
		} else
		{
			push @a_feature_list, $feat;
		}				
	}
	# We overwrite the feature list if we have either All features or Core features selected
	if (defined $h_features_batch{ &ALL } || defined $h_features_batch{ &CORE } )
	{
			@a_feature_list = @A_CORE_FEATURES;
	}		
	$self->{ _features_batch } = \%h_features_batch;
	$self->{ feature_list } = \@a_feature_list;	
}
=cut

=head2 Function LoadSupplementaryData

 Usage        : my $rh_elements = $o_xx->LoadSupplementaryData( $rh_params, $where );
 Prerequisite : none
 Function     : browse session and instance dir to look for supplementary data.
 Args         :
     $rh_params, hash ref onto ElementSet::Multigraph arguments.
         arguments that are common to all ElementSet of this object.
         this hash ref will be passed to SuppElementSet constructor
     $where, keyword 'SESSION' or 'INSTANCE', depending on wheter you want to search.
 Return       : $rh_elements, a hash ref of ElementsSet::MultiGraph objects.
                              one object per 

=cut

sub LoadSupplementaryData
{
	my $self = shift;
	my ( $rh_params, $where ) = @_;


	my $o_species = $self->_Get( 'o_species' );
	my $supp_path = $o_species->GetSuppDataPath( $where );

	my $o_loader  = New SuppDataLoader( o_species      => $o_species,
										ra_directories => [ $supp_path ],
										-o_logger => $self->GetLogger() );
	my $rao_elts_set = $o_loader->BuildFeatures( %$rh_params );

	my %h_elts_set = map{ $_->_Get( '_feature' ) => $_ } @$rao_elts_set;

	return \%h_elts_set;
}


=head2 Procedure __ExploreSuppDirectory

 Title        : __ExploreSuppDirectory
 Usage        : $o_xx->__ExploreSuppDirectory( $rh_params );
 Prerequisite : none
 Procedure    : browse a supp dir to look for supplementary data.
 Args         : $rh_params, hash ref onto ElementSet::Multigraph arguments.
                arguments that are common to all ElementSet of this object.
                this hash ref will be passed to SuppElementSet constructor.
                $graph_dir, string, the current directory name.
                $data_path, string, the path to data (ROOT_DATA or ROOT_SESSDIR)

=cut

sub __ExploreSuppDirectory
{
	my $self = shift;
	my ( $rh_params, $graph_class, $graph_path ) = @_;

# 	return unless( $graph_dir =~ /^supp_(.+)/ );
# 	my $graph_class    = &General::Camelize( $1 );
# 
# 	my $graph_path = File::Spec->catfile( $datapath, $graph_dir );
# 
# 	return unless( -d $graph_path );
	
	#on a un repertoire candidat : chaque sous-dossier est une 'feature'

	return;
}


=head2 Function __IncludeSuppData

 Title        : __IncludeSuppData
 Usage        : $o_xx->__IncludeSuppData();
 Procedure    : Returns true (1) is supplementary data have to be displayed
 Args         : none
 
#=cut
sub __IncludeSuppData
{
	my $self = shift;
	
	my $bool_include_supp = 0;
	my $rh_features_batch = $self->_Get( '_features_batch' );
	
	if ( defined $rh_features_batch->{ &ALL } || defined $rh_features_batch->{ &SUPP })
	{
		$bool_include_supp = 1;
	}
	return 	$bool_include_supp;
}
=cut
#------------------------------------------------#
#            Setting the frame                   #
#------------------------------------------------#

=head2 Function __SetFrame

 Title        : __SetFrame
 Usage        : $o_xx->__SetFrame();
 Prerequisite : none
 Procedure    : Sets the frame according to the orientation, the box and the range (Called by the constructor)
 Args         : none
 
=cut
sub __SetFrame
{
	my $self = shift;
	
	my $o_box = $self->{ o_box };
	$self->{ o_frame } = FrameTransformer::1D->New( $o_box->GetDisplayedArea( $self->{ graph_orientation } ), y0 => $self->GetStart(), y1 => $self->GetEnd());	

	return;
}

#------------------------------------------------#
#                  Display methods               #
#------------------------------------------------#

=head2 DrawCenterIndicatorAndLabel

 Title        : DrawCenterIndicatorAndLabel
 Usage        : $self -> DrawCenterIndicatorAndLabel()
 Procedure    : Draws the center Indicator (triangle) and the label name
 Returns      : none
 Args         : none

=cut
sub DrawCenterIndicatorAndLabel
{
	my $self = shift;
	
	#Nothing to do in the general case, see the child Reference
	
	return 
}

=head2 Display

 Title        : Display
 Usage        : $self -> Display(%h_param)
 Prerequisite : Data should be initialized
 Procedure    : 1) Call DisplayDNA
 				2) Call DisplayGraduation
 				3) Call Display method of each ElementsSet object
 Returns      : none
 Args         : %h_param {
 					box   => Box object
 					representation_type  => optional, 'basic' or 'full'
 						 }
 Globals	  : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			representation_type => $self->{representation_type},
			@_
	);

	return if ( defined $h_param{ dont_display_anything } && 1 == $h_param{ dont_display_anything });
	
	my $o_box = $self->{ o_box };

	# General parameters
	my $o_image				= $self -> { o_image };
	my $o_textual           = $self -> { o_textual };
	my $representation_type = $h_param{representation_type};
	my $side 				= $o_box -> GetJustification();
	
	# Textual representation		
	if (defined $o_textual)
	{
		#======================================
		# ElementsSet data representations.
		my $rh_elements = $self->{h_elements};
		foreach my $o_element ( values %$rh_elements )
		{
			if ( $o_element->isa( 'ElementsSet' ) && $o_element->IsEmpty() == &FALSE )
			{
				$o_element->DisplayText( o_textual => $o_textual );
			}
		}
	}
	
	# Graphics representation
	else
	{

		#From now on all the ElemenSets will be displayed using the frame change defined in the next line
		$o_image -> SetFrameChange( $self->{ o_frame } );
			
		#======================================
		# 1) DNA cylinder.	
		$self -> DisplayDNA( image=>$o_image, side=>$side );
	
		# !! Procedure break if the representation_type is set to 'basic' !!
		return if ( $self->{representation_type} eq 'basic' );
	
		#======================================
		# 2) Graduations.
		#
		$self -> DisplayGraduation( image=>$o_image, side=>$side );	
	
		#======================================
		# 3) ElementsSet data representations.
		my $rh_elements = $self->{h_elements};
		foreach my $o_element ( values %$rh_elements )
		{
			if ( $o_element->isa( 'ElementsSet' ) && $o_element->IsEmpty() == &FALSE )
			{
				$o_element->Display( image=>$o_image, side=>$side, o_box => $o_box);
			}
		}
	
		#=====================================
		# 4) Add links between features areas
		$self->MakeLinksOnMap();
	}
		
	return;
}

=head2 DisplayDNA

 Title        : DisplayDNA
 Usage        : $self -> DisplayDNA(%h_param)
 Prerequisite : Data should be initialized
 Procedure    : Draw the cylinder element which represent the DNA
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					side  => optional, 'right' or 'left'
 						 }
 Globals	  : none

=cut
sub DisplayDNA
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			side	=> 'right',
			@_
			);
	
	#======================================
	# Testing mandatory parameters
	throw LipmError::ParameterException('image_object', 'missing') if (!defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );

	# Parameters
	my $o_image	= $h_param{image};
	my $side	= $h_param{side};

	#If the organism is central we duplicate the cylinder
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> DisplayDNA(%h_param, side => 'left');
	}

	my $layer = $Planner::A_IMG_LIST[0];

	#======================================
	# Graphics parameters

	#TODO we should probably change the name of the colorAllocate function since the color ist most probably already created
	# Color
	my $cylinder_color 	= $o_image -> colorAllocate( 'black', $layer );

	$o_image -> SetCurrentLayer( $layer );
	$o_image -> SetAbsoluteFrame();
	
	my %dna_cyl = %{ $self->GetDNACylinder( $side ) };

	my $cylinder_height= $self->GetDNACylinderHeight();

	# Draw DNA representation (cylinder)
	$o_image -> verticalCylinder( @dna_cyl{qw(X0 Y0 X1 Y1)}, $cylinder_height, $cylinder_color, $layer );

	#We go back to relative frame
	$o_image -> SetRelativeFrame();

	return;
}

=head2 DisplayGraduation

 Title        : DisplayGraduation
 Usage        : $self -> DisplayGraduation( image=>$o_image, side=>$side, display_unit=>$bool);
 Prerequisite : none
 Procedure    : Draw graduation of the representation
 Returns      : none
 Args         : image	-> An Image object
 				side	-> Optional parameter, set the position of the representation. 'left' or 'right' (default = 'right')
 Globals      : none

=cut
sub DisplayGraduation
{
	my $self    = shift;
	my %h_param = (
			image	=> undef,
			side	=> 'right',
			@_
			);

	#======================================
	# Testing mandatory parameters
	throw LipmError::ParameterException('image_object', 'missing') if (!defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );

	# Other parameters
	my $o_image	= $h_param{image};
	my $side	= $h_param{side};

	if ($side eq 'center')
	{
		$side = 'right';
	}

	my $graph_orientation = $self->{graph_orientation};
	my $graduation_scale  = $self->GetGraduationScale() or return;	# This method returns 0 if their is a problem on range. If it returns 0, the procedure is breaking.

	#======================================
	# Graphics parameters
	my $x_grad 			 = $self->{ o_param_graph }->Get( 'x_grad' );
	my $width_grad 		 = $self->{ o_param_graph }->Get( 'width_grad' );
	my $width_grad_big 	 = $self->{ o_param_graph }->Get( 'width_grad_big' );
	my $space_grad_label = $self->{ o_param_graph }->Get( 'space_grad_label' );
	my $color_grad 		 = $self->{ o_param_graph }->Get( 'color_grad','black' );
	my $color_index 	 = $o_image->colorAllocate($color_grad);

	# Image parameters :
	my $nucleic_min  = $self->{start};
	my $nucleic_max  = $self->{end};
	my $font 		 = $o_image->gdTinyFont;

	my %grad_param = $self->GetGraduationParameters( $side );

	my %h_area=$self ->{ o_box }->GetBoxArea();
	#TODO should probably put this into general parameters
	my $y_min = $h_area{Y0} + 5 ;
	my $y_max = $h_area{Y1} - 5;

	#======================================
	# Calculate x positions :
	$x_grad = $grad_param{X0};

	my $label_justification = ($side eq 'left') ? 'right' : 'left';
	my $x_simple_dash = $grad_param{X_SimpleDash};
	my $x_big_dash	  = $grad_param{X_BigDash};
	my $x_label		  = $grad_param{X_Label};

	my $layer = $Planner::A_IMG_LIST[0];
	$o_image -> SetCurrentLayer( $layer );
	my $o_layer = $o_image -> GetCurrentLayer();
	
	#======================================
	# Draw main lines :
	
	$o_layer -> line($x_grad,$y_min, $x_grad,$y_max, $color_index);
	$o_layer -> line($x_grad-$width_grad,$y_min,$x_grad+$width_grad,$y_min,$color_index);
	$o_layer -> line($x_grad-$width_grad,$y_max,$x_grad+$width_grad,$y_max,$color_index);

	#======================================
	# Draw the center indicator and the main label
	
	$self->DrawCenterIndicatorAndLabel( image => $o_image, side => $side );
	
	#======================================
	# Draw graduations if $representation_type set to 'full':
	return if ($self->{representation_type} eq 'basic');
	my $grad_unit = 10**(3*int(log($graduation_scale)/log(1000)));

		# 1) Calculate the first graduation
	my $first_grad = floor($nucleic_min/$graduation_scale);
	my $last_grad  = $nucleic_max/$graduation_scale;
			
		# 2) Calculate step between big graduations.
	my $grad_nb = int($last_grad - $first_grad);
	my $big_grad_step = 1;

	return if ($grad_nb < 1);
	if ($grad_nb >= 3 && $grad_nb < 10)
	{
		$big_grad_step = 2;
	}
	elsif ($grad_nb >= 10 && $grad_nb < 100)
	{
		$big_grad_step = 5;
	}
	elsif ($grad_nb >= 100)
	{
		$big_grad_step = 10;
	}

		# 3) Draw graduations.
	my $frame = $o_image -> GetFrame();	
				
	my $small_grad = 0;
	$first_grad -= $first_grad%10;
		
	for (my $grad = $first_grad; $grad <= $last_grad; $grad++,$small_grad++)
	{
		my $y_grad = $grad*$graduation_scale;
		next if ($y_grad < $nucleic_min or $y_grad > $nucleic_max);

		# Small graduation.
		$o_image -> line($x_grad,$y_grad,$x_simple_dash,$y_grad,$color_index);

		# Big graduation.
		if ($small_grad%$big_grad_step == 0)
		{
			my $big_grad_step_label=$grad*$graduation_scale/$grad_unit;
			$o_image -> line($x_grad,$y_grad,$x_big_dash,$y_grad,$color_index);
			$o_image -> stringVerticalAligned($font, $x_label, $y_grad, $big_grad_step_label, $color_index, side => $label_justification);
		}
	}

	return;
}

#------------------------------------------------#
#                Graphic parameters              #
#------------------------------------------------#

=head2 GetGraphicParameter

 Title        : GetGraphicParameter
 Usage        : $o_x->GetGraphicParameter (parameter_label [, default_value])
 Prerequisite : none
 Function     : Return the value of the graphical parameter named 'parameter_label'.
                If it isn't set and 'default_value' is defined,
                    this parameter value is set to 'default_value'
 Returns      : parameter value
 Args         : parameter_label -> name of the parameter in the config file
                default_value	-> value by default for this parameter.
                    It is advisable to fill it.
                    Do not use it anymore, policy seemed to be changed.
 Globals      : %H_GRAPHIC_PARAM _no more !

=cut
sub GetGraphicParameter
{
	my $self = shift;
	my ( $label, $default_value ) = @_;

	my $value = $default_value;

	my $o_param_graph = $self->{ o_param_graph };
	$value = $o_param_graph->Get( $label )	if( $o_param_graph->IsDefined( $label ) );

	if( !defined $value or $value eq '' )
	{
		my ( $pack, $file, $line ) = caller();
		my $msg = "Check your graphics.cfg file, it seems not to have $label.";
		$msg   .= " ( $file, $line )\n";
		$self->_Trace ( $msg, '.', 'IEV', $pack );
	}

	return $value;
}

#------------------------------------------------#
#             DNA cylinder           #
#------------------------------------------------#

=head2 Function __SetDNACylinders

 Title        : __SetDNACylinders
 Usage        : $o_xx->__SetDNACylinders();
 Prerequisite : none
 Procedure    : Sets the left and right position of the DNA cylinder according to the orientation and the box
 Args         : none
 
=cut
sub __SetDNACylinders
{
	my $self = shift;
		
	# x position of the box for a negative strand gene
	my $x_genes = $self->{ o_param_graph }->Get( 'x_genes' );
	# width of the gene representation box
	my $width_genes = $self->{ o_param_graph }->Get( 'width_genes' );
	
	my $cylinder_height 		= $width_genes/3;
			
	$self->{ __dna_cylinder_height } = $cylinder_height + 3;
	
	my %h_area=$self ->{ o_box }->GetBoxArea();

	my ($X0,$X1,$Y0,$Y1, $X);
	my $side;

	$Y0 = $h_area{Y0};
	$Y1 = $h_area{Y1};
	
	# DNA cylinder for a right side of the border between the two species
	$side = "right";
	$X0 = $h_area{X0} + $x_genes ;
	$X1 = $X0 + $width_genes*2 +1;		
	$X = $X0;
	
	$self->{ __dna_cylinders }{$side} = {X0 => $X0, X1 => $X1, Y0 => $Y0, Y1 => $Y1, X => $X };
	
	# DNA cylinder for a left side of the border between the two species
	$side = "left";
	$X1 = $h_area{X1} - $x_genes ;
	$X0 = $X1 - $width_genes*2  -1;	
	$X = $X1;
	
	$self->{ __dna_cylinders }{$side} = {X0 => $X0, X1 => $X1, Y0 => $Y0, Y1 => $Y1, X => $X  };
	
}

=head2 Function GetDNACylinder

 Title        : GetDNACylinders
 Usage        : $o_xx->GetDNACylinders();
 Procedure    : Returns a reference on a hash containing the dna cylinders coordinates
 Retruns      : a refence on a hash with the box coordinates

=cut
sub GetDNACylinders
{
	my $self = shift;
	
	return $self->{ __dna_cylinders};

}

=head2 Function GetDNACylinder

 Title        : GetDNACylinder
 Usage        : $o_xx->GetDNACylinder();
 Procedure    : Returns a reference on a hash containing the dna cylinder coordinates
 Retruns      : a refence on a hash with the box coordinates

=cut
sub GetDNACylinder
{
	my $self = shift;
	my $side = shift;
	
	return $self->{ __dna_cylinders }{ $side };

}

=head2 Function GetDNACylinderHeight

 Title        : GetDNACylinderHeight
 Usage        : $o_xx->GetDNACylinderHeight();
 Prerequisite : none
 Procedure    : Returns the height of the dna cylinder
 Returns      : the height of the cylinder

=cut
sub GetDNACylinderHeight
{
	my $self = shift;
	
	return $self->{ __dna_cylinder_height };
}

#------------------------------------------------#
#                  Accessors                     #
#------------------------------------------------#

=head2 Function GetNamePosition

 Title        : GetNamePosition
 Usage        : $o_xx->GetNamePosition($side);
 Prerequisite : none
 Procedure    : Returns the position of the name label according to the orienttion and the box
 Args         : none
 Retruns      : an X position in pixel

=cut
sub GetNamePosition
{
	my $self = shift;
	my $side = shift;

	my $x_genes_label 			= $self->{ o_param_graph }->Get( 'x_genes_label' );		# left alignment of locus text
	my $width_genes_label 		= $self->{ o_param_graph }->Get( 'width_genes_label' );	# max width for gene labels
		
	my %h_area=$self ->{ o_box }->GetBoxArea();
		
	my $X;

	if ( $side eq "right")
	{
		$X = $h_area{X0} + $x_genes_label;
	}
	else
	{
		$X = $h_area{X1} - $x_genes_label;
	}

	return $X;
}

=head2 GetStart
 Title        : GetStart
 Usage        : $object -> GetStart()
 Prerequisite : data must be initialized
 Function     : Get the start of the first element, this is the minimum position.
 Returns      : integer
 Args         : none
 Globals      : data, sort_elements

=cut
sub GetStart
{
	my $self = shift;
	return $self->{start} if (defined $self->{start});
}

=head2 GetEnd
 Title        : GetEnd
 Usage        : $object -> GetEnd()
 Prerequisite : data must be initialized
 Function     : Get the end of the last element, this is the maximum position.
 Returns      : integer
 Args         : none
 Globals      : data, sort_elements

=cut
sub GetEnd
{
	my $self = shift;
	return $self->{end} if (defined $self->{end});
}

=head2 GetRangeSize
 Title        : GetRangeSize
 Usage        : $object -> GetRangeSize()
 Prerequisite : data must be initialized
 Function     : Get the range between the first and the last locus.
 Returns      : integer
 				undef if there is a problem on data set
 Args         : none
 Globals      : none

=cut
sub GetRangeSize
{
	my $self = shift;
	my $range = ($self->GetEnd() - $self->GetStart());

	return 0 if ($range < 0);		# Return 0 if the range is negative (problem on initialization parameters)
	return $range;
}

#------------------------------------------------#
#          Managing the graduation               #
#------------------------------------------------#

=head2 Function GetGraduationParameters

 Title        : GetGraduationParameters
 Usage        : $o_xx->GetGraduationParameters();
 Prerequisite : none
 Procedure    : Returns the left and right position of the DNA cyliinder accoing to the orienttion and the box
 Args         : none
 Retruns      : a refence on a hash with lft and right coordinate
 TODO         : This method should only return the gradution parameter which should be set elsewhere
=cut
sub GetGraduationParameters
{
	my $self = shift;
	my $side = shift;
	
	my $x_grad 			 = $self->{ o_param_graph }->Get( 'x_grad' );
	my $width_grad 		 = $self->{ o_param_graph }->Get( 'width_grad' );
	my $width_grad_big 	 = $self->{ o_param_graph }->Get( 'width_grad_big' );
	my $space_grad_label = $self->{ o_param_graph }->Get( 'space_grad_label' );
		
	my %h_area=$self ->{ o_box }->GetBoxArea();
	my ($X0, $X_SimpleDash, $X_BigDash, $X_Label);

	my $x_simple_dash= ($side eq 'left') ? $x_grad-$width_grad : $x_grad+$width_grad;
	my $x_big_dash	 = ($side eq 'left') ? $x_grad-$width_grad_big : $x_grad+$width_grad_big;
	my $x_label		 = ($side eq 'left') ? $x_grad-$space_grad_label : $x_grad+$space_grad_label;
	
	if ( $side eq "right")
	{
		$X0 = $h_area{X0} + $x_grad;
		$X_SimpleDash = $X0 + $width_grad;
		$X_BigDash = $X0 + $width_grad_big;
		$X_Label = $X0 + $space_grad_label
	}
	else
	{
		$X0 = $h_area{X1} - $x_grad;
		$X_SimpleDash = $X0 - $width_grad;
		$X_BigDash = $X0 - $width_grad_big;	
		$X_Label = $X0 - $space_grad_label
	}	
	
	my %grad_parameter =( X0 => $X0, X_SimpleDash => $X_SimpleDash, X_BigDash => $X_BigDash, X_Label => $X_Label);
		
	return %grad_parameter;
}

=head2 GetGraduationScale
 Title        : GetGraduationScale
 Usage        : $self -> GetGraduationScale()
 Prerequisite : none
 Function     : Return graduation_scale value.
 Returns      : graduation_scale
 Args         : none
 Globals      : none

=cut
sub GetGraduationScale
{
	my $self = shift;

	return $self->{ graduation_scale };
}

=head2 GetGraduationLabel
 Title        : GetGraduationLabel
 Usage        : $self -> GetGraduationLabel()
 Prerequisite : none
 Function     : Return graduation_label value.
 				If it's not defined, call __SetGraduationLabel() and returns its value.
 Returns      : graduation_label
 Args         : none
 Globals      : $GRADUATION_LABEL

=cut
sub GetGraduationLabel
{
	my $self = shift;

	return $self->{graduation_label};
}

=head2 __SetGraduationScale
 Title        : __SetGraduationScale
 Usage        : $self -> __SetGraduationScale()
 Prerequisite : none
 Procedure    : Set the attribute graduation_scale value. It's a 10 factor
 Returns      : none
 Args         : graduation_scale -> optional
 Globals      : none

=cut
sub __SetGraduationScale
{
	my $self = shift;

	# Calculate the scale.
	my $range = $self -> GetRangeSize();

	my $factor1 = ceil(log($range/200*5)/log(10));
	my $factor2 = ceil(log($range/200)/log(10));

	$self->{ graduation_scale } = ($factor1 < $factor2) ? 5*10**$factor1 : 10**$factor2;

	return;
}

=head2 __SetGraduationLabel
 Title        : __SetGraduationLabel
 Usage        : $self -> __SetGraduationLabel([graduation_label][graduation_scale])
 Prerequisite : none
 Procedure    : Set the attribute graduation_label value.
 				Values could be : cR, bp , kb , Mb
 Returns      : none
 Args         : graduation_label -> optional
 				graduation_scale -> optional
 Globals      : none

=cut
sub __SetGraduationLabel
{
	my $self = shift;
	my $parameter = (defined $_[0] && $_[0] =~ /^cR|bp|[kM]b|1|3|6$/) ? shift : $self->GetGraduationScale();
	my $graduation_label = undef;

	# Test the optional parameter and set $graduation_label if it's properly set.
	if ($parameter =~ /^cR|bp|[kM]b$/)
	{
		$graduation_label = $parameter;
	}
	else
	{
		my $grad_scale = $self->GetGraduationScale;#int(log($parameter)/log(1000));
		if ($grad_scale < 10**3)
		{
			$graduation_label = 'bp';
		}
		elsif ($grad_scale < 10**6)
		{
			$graduation_label = 'kb';
		}
		else
		{
			$graduation_label = 'Mb';
		}
	}

	$self->{graduation_label} = $graduation_label;

	return;
}

=head2 IsDataLoad
 Title        : IsDataLoad
 Usage        : $self -> IsDataLoad()
 Prerequisite : none
 Function     : Test data initialization
 Returns      : 1 or 0
 Args         : none
 Globals      : none

=cut
sub IsDataLoad
{
	my $self = shift;

	my $return = ( scalar( keys(%{$self->{h_elements}})) == 0 ) ? 0 : 1 ;
	return $return;
}

=head2 procedure __SetReferenceCenterPos

 Title        : __SetReferenceCenterPos
 Usage        : $o_chrom_slice -> __SetReferenceCenterPos($pos)
 Function     : Change a range in order to center the view on the new position
 Returns      : none
 Args         : a valid position on the chromosome

=cut
sub __SetOrganismCenterPos
{
	my $self = shift;
	my $pos = shift;

	my $range_size = $self->GetRangeSize();
	my $new_start  = int($pos - $range_size/2);
	my $new_end    = int($pos + $range_size/2-1);
	
	$self->__SetNewRange($new_start,$new_end);

	return;
}

=head2 __SetNewRange

 Title        : SetNewRange
 Usage        : $self -> SetNewRange(25,100);
 Procedure    : modifies the new range of the chromosome slice
 Args         : $start, $end = two integers

=cut
sub __SetNewRange
{
	my $self = shift;
	my ( $start, $end ) = 	(shift, shift);

	$self->{ start } = $start;
	$self->{ end }   = $end;
}

=head2 MakeLinksOnMap
 Title        : MakeLinksOnMap
 Usage        : $self -> MakeLinksOnMap()
 Prerequisite : none
 Procedure    :
 Returns      :
 Args         : none
 Globals      : none

=cut
sub MakeLinksOnMap
{
	my $self = shift;

	return;
}

=head2 Infos
 Title        : Infos
 Usage        : $self -> Infos()
 Prerequisite : none
 Procedure    :
 Returns      :
 Args         : none
 Globals      : none

=cut
sub Infos
{
	my $self = shift;
	
	return "--------".join("\t",@{$self}{qw(organism chromosome start end)});	
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
