# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

a Narcisse object to manage chromosome data and representation.

=head1 DESCRIPTION

ChromosomeSlice manage ElementsSet inherited objects which compose a chromome representation (Genes, Rna, Cyto...).


B<Attributes :>

 - organism			=> 	short name of the organism
 - chromosome
 - start
 - end
 - graph_orientation=> 	vertical orientation of the representation. Positive orientation show data from top to bottom. ('-' or '+')
 - h_elements		=> 	hash of all elements which are initialized for chromosome representation
 - representation_type => flag to know which elements need to be load and display. Values : full,basic (default full)


=head1 SYNOPSIS

	my %h_param =  ( organism => 'organism',
					 chromosome => 'chromosome',
					 start => 'start',
					 end => 'end'
					);

	my $chromosome 	= ChromosomeSlice -> New ('/root/data/', \%h_param);
	my $image 		= Image -> New (width,height);

	# To display data :
	$chromosome -> Display($image);


=head1 SUBROUTINES

Display
GetEnd
GetMap
GetRange
GetStart

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
Letort Sebastien : sebastien.letort@toulouse.inra.fr

=cut

package ChromosomeSlice::Reference;

use strict;
use warnings;

use POSIX;
use base qw( ChromosomeSlice );

use Scalar::Util qw( blessed );
use Error qw( :try );
use Data::Dumper;

use Image;
use General;
use Box::ChromosomeSliceBox;
use ElementsSet;
use ElementsSet::Cyto;
use ElementsSet::Gene;
use ElementsSet::Ocean;
use ElementsSet::Personal;
use ElementsSet::Repeat;
use ElementsSet::Rna;
use ElementsSet::Snp;
use ElementsSet::MultiGraph;
use LipmError::ObjectException;
use LipmError::ParameterException;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

# Default parameters => c'est deja defini dans la mere, on devrait pouvoir l'utiliser.
my @A_DEFAULT_FEATURES = qw( Gene Cyto Rna Repeat Personal Ocean Snp );

# Multigraph does not have to draw a legend (for a specific feature)
#	if the legend has previously been drawn.
# This information is stored into this private class variable
# /!\ We decided to show the legend only in reference if possible
#	To respect this wish, reference ChromSlice have to been created first
our %_H_MULTIGRAPH_LEGEND = ();

=head2 _Init

 Title        : _Init
 Usage        : _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object

 Args         : %h_param a hash of parameters
 					mandatory
  						organism  		=> short name of the organism
  						chromosome		=> chromosome name or number
  						o_param_graph         => ParamParser object, with Graphics.cfg data

  					optional
  						feature_list 	=> a list of features to display
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 1)
  						exon_bool			=> are the exon displayed  (default 0)
						graph_orientation 	=> the orientation (default '+')
						is_reference 		=> Is it the reference species (default 0)
						graduation_label	=> title label of the graduation (default undef)
						graduation_scale	=> exponent ten value for corresponding to the graduation scale : 1,3,6 (default undef)

 Globals	  : none

=cut
sub _Init
{
	my $self = shift;
	my %h_param = (				
				# Optional parameters with default value
				feature_list		=> \@A_DEFAULT_FEATURES,
				bool_is_reference	=> 1,
				graph_orientation	=> '+',
				tracks 				=> {},
				__full_screeen_mode => 0,
				@_ );
	
	#1) First call the parent constructor
	$self -> SUPER::_Init( %h_param );
							
	#======================================
	# 2) Initialize undefined attributes.
	#TODO check if this step should always and only be performed in the reference case
	
	$self->__SetGraduationLabel() if ( !defined $self->{graduation_label} );
											
	return $self;
}


=head2 __SetGraduationLabel
 Title        : __SetGraduationLabel
 Usage        : $self -> __SetGraduationLabel([graduation_label][graduation_scale])
 Prerequisite : none
 Procedure    : Set the attribute graduation_label value.
 				Values could be : cR, bp , kb , Mb
 Returns      : none
 Args         : graduation_label -> optional
 				graduation_scale -> optional
 Globals      : none

=cut
sub __SetGraduationLabel
{
	my $self = shift;
	my $parameter = (defined $_[0] && $_[0] =~ /^cR|bp|[kM]b|1|3|6$/) ? shift : $self->GetGraduationScale();
	my $graduation_label = undef;

	# Test the optional parameter and set $graduation_label if it's properly set.
	if ($parameter =~ /^cR|bp|[kM]b$/)
	{
		$graduation_label = $parameter;
	}
	else
	{
		my $grad_scale = $self->GetGraduationScale;#int(log($parameter)/log(1000));
		if ($grad_scale < 10**3)
		{
			$graduation_label = 'bp';
		}
		elsif ($grad_scale < 10**6)
		{
			$graduation_label = 'kb';
		}
		else
		{
			$graduation_label = 'Mb';
		}
	}

	$self->{graduation_label} = $graduation_label;

	return;
}

=head2 DrawCenterIndicatorAndLabel

 Title        : DrawCenterIndicatorAndLabel
 Usage        : $self -> DrawCenterIndicatorAndLabel()
 Procedure    : Draws the center Indicator (triangle) and the label name
 Returns      : none
 Args         : none

=cut
sub DrawCenterIndicatorAndLabel
{
	my $self = shift;
	my %h_param = (
			image	=> undef,
			side	=> 'right',
			@_
			);
	
	my $o_image	= $h_param{image};
	my $side	= $h_param{side};

	if ($side eq 'center')
	{
		$side = 'right';
	}
		
	
	# Image parameters :
	my $nucleic_min  = $self->{start};
	my $nucleic_max  = $self->{end};

	my %grad_param = $self->GetGraduationParameters( $side );

	my %h_area=$self ->{ o_box }->GetBoxArea();
		 		 
	my $y_min = $h_area{Y0} + 3;
	my $y_max = $h_area{Y1} - 5;
	
	my $x_grad 	= $grad_param{X0};
	my $x_label = $grad_param{X_Label};
	
	my $color_grad 		 = $self->{ o_param_graph }->Get( 'color_grad','black' );
	my $color_index 	 = $o_image->colorAllocate($color_grad);
	
	# Label
	my $label_justification = ($side eq 'left') ? 'right' : 'left';
	my $graduation_label = $self->GetGraduationLabel();
	my $font_label = $o_image->gdSmallFont;	
	
	$o_image -> SetAbsoluteFrame();
	
	$o_image -> stringVerticalAligned($font_label, $x_label, $y_min , $graduation_label, $color_index, side => $label_justification);

	$o_image -> SetRelativeFrame();

	# Center triangle
	my $middle = ($nucleic_min + $nucleic_max)/2;
	$o_image -> filledEquilateralTriangle($x_grad,$middle,10,$label_justification,$color_index);	
	
	return ;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
