# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

a Narcisse object to manage chromosome data and representation.

=head1 DESCRIPTION

ChromosomeSlice manage ElementsSet inherited objects which compose a chromome representation (Genes, Rna, Cyto...).


B<Attributes :>

 - organism			=> 	short name of the organism
 - chromosome
 - start
 - end
 - graph_orientation=> 	vertical orientation of the representation. Positive orientation show data from top to bottom. ('-' or '+')
 - h_elements		=> 	hash of all elements which are initialized for chromosome representation
 - representation_type => flag to know which elements need to be load and display. Values : full,basic (default full)


=head1 SYNOPSIS

	my %h_param =  ( organism => 'organism',
					 chromosome => 'chromosome',
					 start => 'start',
					 end => 'end'
					);

	my $chromosome 	= ChromosomeSlice -> New ('/root/data/', \%h_param);
	my $image 		= Image -> New (width,height);

	# To display data :
	$chromosome -> Display($image);


=head1 SUBROUTINES

Display
GetEnd
GetMap
GetRange
GetStart
IsDataLoad

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
Letort Sebastien : sebastien.letort@toulouse.inra.fr

=cut

package ChromosomeSlice::Target;

use strict;
use warnings;

use POSIX;
use base qw( ChromosomeSlice );

use Scalar::Util qw( blessed );
use Error qw( :try );
use Data::Dumper;

use Image;
use General;
use Box::ChromosomeSliceBox;
use ElementsSet;
use ElementsSet::Cyto;
use ElementsSet::Gene;
use ElementsSet::Ocean;
use ElementsSet::Personal;
use ElementsSet::Repeat;
use ElementsSet::Rna;
use ElementsSet::Snp;
use ElementsSet::MultiGraph;
use LipmError::ObjectException;
use LipmError::ParameterException;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

# Default parameters => c'est deja defini dans la mere, on devrait pouvoir l'utiliser.
my @A_DEFAULT_FEATURES = qw( Gene Cyto Rna Repeat Personal Ocean Snp );

# Multigraph does not have to draw a legend (for a specific feature)
#	if the legend has previously been drawn.
# This information is stored into this private class variable
# /!\ We decided to show the legend only in reference if possible
#	To respect this wish, reference ChromSlice have to been created first
our %_H_MULTIGRAPH_LEGEND = ();

=head2 _Init

 Title        : _Init
 Usage        : _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object

 Args         : %h_param a hash of parameters
 					mandatory
                        o_io            => the Instance Observer
  						organism  		=> short name of the organism
  						chromosome		=> chromosome name or number
  						o_param         => ParamParser object, with Graphics.cfg data

  					optional
  						feature_list 	=> a list of features to display
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 1)
  						exon_bool			=> are the exon displayed  (default 0)
						graph_orientation 	=> the orientation (default '+')
						is_reference 		=> Is it the reference species (default 0)
						graduation_label	=> title label of the graduation (default undef)
						graduation_scale	=> exponent ten value for corresponding to the graduation scale : 1,3,6 (default undef)

 Globals	  : none

=cut
sub _Init
{
	my $self = shift;
	my %h_param = (
				# Mandatory parameters
                o_io    => undef,
				o_synt  => undef,
				synt_id => undef,	
				# image_height is quite a strange mandatory parameter but is used to 
				# decide if we are in a full screen mode	
				image_height		=> undef,   
				
				# Optional parameters with default value
				
				feature_list		=> \@A_DEFAULT_FEATURES,
				bool_is_reference	=> 0,
				representation_type => 'full',
				tracks 				=> {},
				__full_screeen_mode => 0,				
				@_
				);

	#1) First call the parent constructor	
	$self -> SUPER::_Init( %h_param );
			
	my @mandatory_attributes = qw( o_io o_planner o_synt synt_id image_height);
	#0 Checking additional mandatory parameters
	foreach my $mandatory_attribute ( @mandatory_attributes ) 
	{
		(defined $h_param{ $mandatory_attribute })
			or throw LipmError::ParameterException($mandatory_attribute,'missing');
	}
	
	my $o_planner =  $self->{ o_planner };	
	my $o_synteny = $h_param{ o_synt };
		
	(blessed($o_synteny) && $o_synteny->isa("ElementsSet::Synt"))
		or throw LipmError::ParameterException('object synt ','missing');
			
	# Test 'representation_type' parameter, if we won't display all data, we don't load them.
	return $self if ($self->{representation_type} eq 'basic');
																
	#Checking if the Slice is a "full-screen" slice
	#TODO This will have to be changed. Remark : the data are loaded before the range correction is made (it used to work ???) Thomas
	if ( $self->{ o_box }->Height() == $self->{ image_height } && 0)
	{
		$self->{ __full_screeen_mode } = 1;			
		my $target_center = $o_synteny->GetTargetProjectedPos( $self->{ synt_id }, $o_planner->GetReferenceCenterPos() );
		$self->__SetOrganismCenterPos( $target_center );					
	};		
	
	return $self;
}

=head2 procedure __SetReferenceCenterPos

 Title        : __SetReferenceCenterPos
 Usage        : $o_planner -> __SetReferenceCenterPos($pos)
 Function     : Change a range in order to center the view on the new position
 Returns      : none
 Args         : a valid position on the chromosome

=cut
sub __SetOrganismCenterPos
{
	my $self = shift;
	my $pos = shift;

	my $range_size = $self->GetRangeSize();
	my $new_start  = int($pos - $range_size/2);
	my $new_end    = int($pos + $range_size/2-1);
	
	$self->_SetNewRange($new_start,$new_end);

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
