#
# $Id: Locus.pm 1428 2008-03-06 14:00:09Z faraut $
#

package ChromosomeSliceBigMac;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ChromosomeSliceBigMac - different methods for chromosomes ???????????

=head1 SYNOPSIS


=head1 DESCRIPTION

=cut

use Carp;
use Data::Dumper;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1428 $ =~ /\d+/g); $r[0]};
}

use Image;
use ElementsSet::Prbh;
use ElementsSet::Hsp;
use ChromosomeSlice::Target;

use constant OCCURENCE_CUTOFF => 4;
use constant SELDOM => "Seldom";
use constant FREQUENT => "Frequent";

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_xxx = New ChromosomeSliceBigMac(reference	=> "human", target => "chick", program => "glint", level => 4, view_level => 4);
 Prerequisite : none
 Returns      : none
 Args         : Mandotory: o_io       The Instance Observer  
                Mandatory: reference  the reference orgamnism
                Mandatory: target     the target orgamnism
				Mandatory: program    the program used for the comparison
 Globals      : none

=cut
sub _Init
{
	my $self = shift;

	my %h_param = (

        #Comparative Specific attributes
        organism		=> undef,
        chromosome		=> undef,
        start			=> undef,
        end				=> undef,
        target			=> undef,
        program			=> undef,
        level			=> undef,
        view_level		=> undef,		

        #Additional technical attributes
        o_io            => undef,
        -o_textual      => undef, # Necessary to generate Textual instead of graphical information
        -o_logger       => undef, # Useful at debug time
        o_planner		=> undef, # Necessary to provide some information stored by Planner (o_maparea, o_param_graph, ....)
        o_image			=> undef, # Necessary to access to the frame transformation object		
        @_ );

    # Initializing Logger
    $self->SUPER::_Init( %h_param );

	#0 Checking mandatory parameters
	# mandatory parameters
	my @mandatory_parameters = qw( o_io organism chromosome start end target program level view_level o_planner o_image );
	foreach my $required ( @mandatory_parameters )
	{
		( defined( $h_param{ $required } ) )
			or throw LipmError::ParameterException($required,'missing');
	}

	@{$self}{ @mandatory_parameters } = @h_param{ @mandatory_parameters };
	$self -> { o_synteny } = $self->__PrepareCorrespondingSyntenies();
	$self -> { o_textual } = $h_param { -o_textual };

	#Set target chromosome type (seldom or frequent)
    $self->__SetTargetChromosomeType( );	

	return $self;
}

=head2 procedure __PrepareCorrespondingSyntenies

 Title        : __PrepareCorrespondingSyntenies
 Usage		  : $o_synt	= $self->__PrepareCorrespondingSyntenies();
 Function     : Construct the synteny object corresponding to the target and the reference chromosome slice
                and organize the target slices
 Returns      : an ElemenSet::Synt object
 Args         : none

=cut
sub __PrepareCorrespondingSyntenies
{
	my $self = shift;
	
	my $o_planner = $self->_Get( 'o_planner' );
	my $target    = $self->_Get( 'target');
	
	my @a_synteny_attributes  = qw( organism chromosome start end program level view_level );
	my ( $organism, $chromosome, $start, $end, $program, $level, $view_level ) = $self->_Get( @a_synteny_attributes );
	#Some graphical parameters
	my @a_graph_parameters =  qw( o_maparea o_param_graph  );
	my ( $o_maparea, $o_param_graph ) =  $o_planner->_Get( @a_graph_parameters );

	my $o_species = New SpeciesObserver( species => $o_planner->GetRefOrg(),
										 -o_inst_obs => $self->_Get( 'o_io' ) );
    # warn "un ES::Synt cree pour " . $o_planner->GetTargetOrganism( $target ) . " $chromosome(${start}::${end})";
	my $o_synteny = ElementsSet::Synt -> New(

        #Synteny specific attributes
        chromosome      =>  $chromosome,
        start           =>  $start,
        end             =>  $end,
        program			=>  $program,
        level           =>  $level ,
        view_level      =>  $view_level, 
        target          =>  $o_planner->GetTargetOrganism( $target ),

		#Additional technical attributes
        o_species       =>  $o_species,
        o_maparea		=> $o_maparea,
        o_param_graph	=> $o_param_graph

        );

	 # StoreTargetSlicesInfos records in addition the occurence of each target chromosome
	 $o_synteny -> StoreTargetSlicesInfos(	image         => $self->{ o_image }, 
            								reference_box => $o_planner->GetBox(  ), 
            								target_box    => $o_planner->GetBox( $target ) 
            							);

	return $o_synteny;
		
}

=head2 procedure __SetTargetChromosomeType

 Title        : __SetTargetChromosomeType
 Usage		  : $o_xxx	= $self->__SetTargetChromosomeType();
 Function     : sets the target chromosome type according to target_chrom_occ-hash of the o_synteny object
 Returns      : none
 Args         : none

=cut
sub __SetTargetChromosomeType
{
	my $self = shift;
	
	my $o_synteny = $self->{ o_synteny };
	
	my $rh_chrom_occ = $o_synteny->_Get( 'target_chrom_occ' );
	
	my %h_chrom_type;
	foreach my $chrom (keys %$rh_chrom_occ)
	{
		$h_chrom_type{ $chrom } = ( $rh_chrom_occ->{ $chrom } > &OCCURENCE_CUTOFF ) ? &FREQUENT : &SELDOM;
	}
	
	$self->_Set( '_chrom_type' => \%h_chrom_type);
}

=head2 procedure Display

 Title        : Display
 Usage		  : $o_xxx	= $self->Display();
 Function     : Display the target slices
 Args         : none
 Returns      : a reference on a hash contining the exact coordinates of the target slices

=cut
sub Display
{
	my $self = shift;
	
	my $o_synteny = $self->_Get( 'o_synteny' );
	my $o_image   = $self->_Get( 'o_image' );
    my $o_planner = $self->_Get( 'o_planner' );
    my $o_textual = $self->_Get( 'o_textual' );

    return if ( $o_synteny->IsEmpty() ); # nothing to display
    
    my $rh_ImageParameters = $o_planner->GetImageParameters();
     
    my $o_ref_box = $o_planner->GetBox();  						# The box (rectangular area) of the reference genome 
     
    my %h_target_slice_boxes;  
    # For each synteny, draw the ChromosomeSlice object and reciprocal best hits links:
	
	my $rh_chrom_type = $self->_Get( '_chrom_type' );
	
    foreach my $target_slice ( reverse($o_synteny->GetSortElements()) )
    {
		my %h_TargetSliceInfos = $o_synteny->GetTargetSliceParameters( $target_slice ); 
		my ( $s_chrom ) = $o_synteny->GetFields( $target_slice, qw( s_chrom ) );
		my $use_cache = ( $rh_chrom_type->{$s_chrom} eq &FREQUENT ) ? 1 : 0;
				
		my $o_tgt_slice_box = $o_synteny->GetTargetSliceBox( $target_slice );
		my $o_slice = ChromosomeSlice::Target->New(

            				#Target chromosome slice specific attributes
            				%h_TargetSliceInfos,
            				use_cache 			=> $use_cache,
            				#Additional technical attributes
            				o_io                => $self->_Get('o_io'),
           	 				o_planner			=> $o_planner,
            				image_height		=> $rh_ImageParameters->{ height },
            				o_box 				=> $o_tgt_slice_box,
            				o_image             => $o_image,
            				-o_textual          => $o_textual,
            				-o_logger 			=> $self->_GetLogger() 
            				);

		# Test if the targets ChromosomeSlice object is correctly initialized.
		next if ( !defined $o_slice );
		next if ( $o_slice->GetRangeSize() == 0);
						
		$o_slice -> Display ();

		# Display Synteny data (arrows and links).
		$o_synteny -> DisplaySlice (o_image 		=> $o_image,
									-o_textual      => $o_textual,
									reference_box 	=> $o_ref_box,
									target_box 		=> $o_tgt_slice_box,
									level_id		=> $target_slice
									) ;

		my $slice_id = $o_synteny->GetElementField($target_slice, "slice_id" );
		$h_target_slice_boxes{$slice_id} = $o_tgt_slice_box;
	}
	
	$self->{ __target_slices_boxes } = \%h_target_slice_boxes;
	
	return \%h_target_slice_boxes;
}

=head2 procedure DisplayPrbh

 Title        : DisplayPrbh
 Usage		  : $o_xxx	= $self->DisplayPrbh();
 Function     : Display the prbh links
 Args         : the reference box and the target box
 Returns      : none

=cut
sub DisplayPrbh
{
	my $self = shift;
	my %h_args = ( reference_box => undef,
				   target_box => undef,
				   @_
				   );
				   
	#0 Checking mandatory args
	my @mandatory_args = qw( reference_box target_box );
	foreach my $required ( @mandatory_args )
	{
		( defined( $h_args{ $required } ) )
			or throw LipmError::ParameterException($required,'missing');
	}			   
	
	my 	$o_ref_box  	= $h_args{ reference_box };
	my 	$o_target_box   = $h_args{ target_box };
				   
	my $o_planner = $self->_Get( 'o_planner' );
	my $o_image   = $self->_Get( 'o_image' );
	my $o_textual = $self->_Get( 'o_textual');
	  
	my @a_prbh_attributes = qw ( reference chromosome start end o_maparea o_param_graph);
	my ( $organism, $chromosome, $start, $end, $o_maparea, $o_param_graph ) =  $o_planner->_Get( @a_prbh_attributes );

	my $o_species = New SpeciesObserver( species => $organism,
										 -o_inst_obs => $self->_Get( 'o_io') );
	my $o_prbh = ElementsSet::Prbh->New(       
        		organism 		=> $organism,
        		chromosome		=> $chromosome,
        		start			=> $start,
        		end				=> $end,      
        		target 			=> $self->_Get( 'target'),
        		
        		o_species		=> $o_species,
        		o_maparea		=> $o_maparea,
        		o_param_graph	=> $o_param_graph
        );
			
	$o_prbh->Display(o_image       => $o_image,
					 -o_textual     => $o_textual,
					 reference_box => $o_ref_box, 
					 target_box    => $o_target_box
					 );
		
}

=head2 procedure DisplayHsp

 Title        : DisplayHsp
 Usage		  : $o_xxx	= $self->DisplayHsp();
 Function     : Display the hsp
 Args         : the reference box and the target box
 Returns      : none

=cut
sub DisplayHsp
{
	my $self = shift;
	my %h_args = ( 
        reference_box 	=> undef,
        target_box 	=> undef,
        @_
        );
	
	#0 Checking mandatory args
	my @mandatory_args = qw( reference_box target_box );
	foreach my $required ( @mandatory_args )
	{
		( defined( $h_args{ $required } ) )
			or throw LipmError::ParameterException($required,'missing');
	}
	
	my 	$o_ref_box 	= 	$h_args{ reference_box };
	my 	$o_target_box   = 	$h_args{ target_box };
	my 	$onlinecompare	= 	$h_args{ onlinecompare };
	 	
	my $o_planner = $self->_Get( 'o_planner' );
	my $o_image   = $self->_Get( 'o_image' );
	my $o_synteny = $self->_Get( 'o_synteny' );
    my $level     = $self->_Get('level');
    
	my $rh_target_slice_boxes = $self->{ __target_slices_boxes };
	
	my @a_hsp_attributes = qw ( reference chromosome start end o_maparea o_param_graph );
	my ( $organism, $chromosome, $start, $end, $o_maparea, $o_param_graph  ) =  $o_planner->_Get( @a_hsp_attributes );

	my $o_species = New SpeciesObserver( species => $organism,
										 -o_inst_obs => $self->_Get('o_io') );
	my $o_hsp = ElementsSet::Hsp->New(
        		chromosome	=> $chromosome,
        		start		=> $start,
        		end		    => $end,   
                level       => $level,
        		OnlineCompare  	=> $o_planner->_Get( 'OnlineCompare'),
        		o_synteny      	=> $o_synteny,

				o_species  => $o_species,
        		o_maparea	=> $o_maparea,
        		o_param_graph	=> $o_param_graph,
                -o_logger => $self->GetLogger()
        );
     
     $o_hsp  -> Display (	image => $o_image,
 							reference_box => $o_ref_box,
							target_box 	  => $o_target_box,
							target_slice_boxes => $rh_target_slice_boxes
						);

}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
