#
# 	Sebastien.Letort@toulouse.inra.fr
#	Created: March 26, 2007
#	$Id$
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package Circos;

use strict;
use warnings;

use base qw( LipmObject );

=pod

=head1 NAME

Circos

=head1 SYNOPSIS

	%files = (circos_dir => '/mnt/data2/www-narcisse/Dev/seb2/Core/circos/',
			conf_dir => '/mnt/data2/www-narcisse/Dev/seb2/Core/circos/etc/',
			data_dir => '/mnt/data2/www-narcisse/Dev/seb2/Core/circos/data/',
			img_dir => './',
			karyotype => 'bacc1-karyotype.txt',
			chrs_links => 'bacc1_links.txt',
			);
	$o_circos = New Circos(%files);
	if ($o_circos->BuildImage())
	{
		print "Roger, it's OK\n";
	}

=head1 DESCRIPTION

	With this class you can generate Circos images,
	you can manage configuration file to represent what you want

=head1 SUBROUTINES

 GetImageFilename
	Dump
	ComputeMap
	BuildImage

	_ParseTemplate
	_Init

	__GenerateImage
	__WriteConfFile
	__ChecksArguments

=cut

use IO::File;
use Data::Dumper;
use File::Spec;

# for TRUE and FALSE
use General qw(:DEFAULT ShowDebug);
# error management
use LipmError::ParameterException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        :	_Init
 Function     :	Constructor, initialize the object, do not use directly, use New method instead
 Usage        :	my $o_circos = New Circos( %h_param);
              	or in daughter class : $self->SUPER::_Init(%h_param);
 Prerequisite :	circos program must be installed somewhere reachable
            	colors file contains a list of alias of colors with the form "color_name = R,G,B"
            	fonts file contains the list of alias of font with the form "font_name = path_to_font"
            	chrs file contains the elements of configuration concerning the chromosomes representation.
            	links file contains the default value for the elements of configuration concerning the links representation
            	karyotype file contains the list of all chromosomes with the form "chrs_name	start	end	region_name	color". Where color must be defined.
            	chrs_link file contains the list of links between chromosomes with the form "link_name	chrs1	start	end". As a link is between 2 regions, a second line is needed like thie "link_name	chrs2	start	end".
            	ticks file contains the default value for the elements of configuration concerning the ticks representation
            	if the name of the files are not absolute path, we add the directory as a prefixe

 Returns      :	none
 Args         :	%h_param,	a hash of parameters, here the details of keys :
              	circos_bin : the path to the circos program
              	circos_dir : the path to the install dir of the circos program
              	karyotype : name/path of the karyotype file
              	chrs_link : name/path of the chrs_link file
              	data_dir : path to the data directory
              	conf_dir : path to the configuration directory
              	colors : name/path of the colors file
              	fonts : name/path of the fonts file
              	chrs : name/path of the chrs file
              	ticks : name/path of the ticks file
              	links : name/path of the links file
              	img_dir : directory where the generated image will be stored
              	img_file : name of the output file
              	conf_output : name of the generated configuration file
              	conf_file : name of the file used as a template for generating the configuration file
              	image_size : size of the generated image
 Error        :	ParameterException
 Globals      :	none

=cut

sub _Init
{
	my $self = shift;
	my %h_param = ( #attributs obligatoire
				karyotype => '',
				chrs_links => '',

				#attributs facultatifs
				circos_dir => '',
				data_dir => '',
				conf_dir => '',

				#attributs avec une valeur par defaut
				circos_bin => 'bin/circos',
				colors		=> 'colors.conf',
				fonts		=> 'fonts.conf',
				chrs		=> 'chrs.conf',
				ticks		=> 'ticks.conf',
				links		=> 'links.conf',
				conf_file	=> 'template.conf',
				conf_output	=> 'output.conf',
				img_dir		=> 'image',
				img_file 	=> 'output', # file format is defined into output.conf.
				image_size	=> 800,
				imagemap 	=> &FALSE,@_);

	# 0- initialisation de qqs attributs
	$self->{_radius} = $h_param{image_size} / 2;
	$self->{_conf_output} = $h_param{conf_output};
	$self->{_img_file} = $h_param{img_file};
	$self->{_img_path} = $h_param{img_dir};
	$self->{_imagemap} = $h_param{ imagemap };
	$self->{_maps} = undef;

	# 1- verifier le programme
	$self->__ChecksArguments($h_param{circos_dir}, $h_param{circos_bin},
								'_circos', 'exe');

	# 2- verifier les fichiers de donnees


	# 3- verifier les fichier de conf
	my %h_conf = (	_colors => $h_param{colors},
					_fonts => $h_param{fonts},
					_chrs => $h_param{chrs},
					_ticks => $h_param{ticks},
					_links => $h_param{links},
					_conf_file => $h_param{conf_file}
				);
	while( my ($key, $val) = each( %h_conf ) )
	{
		$self->__ChecksArguments($h_param{conf_dir}, $val, $key);
	}

	return;
}

=head2 Procedure __ChecksArguments

	Title      :	__ChecksArguments
	Usage      :	__ChecksArguments($dir, $file, $key[, $type])
	Prerequiste:	$dir must end by '/' or beeing empty
	Procedure  :	checks that $file or $dir/$file used to build image exists.
                    If necessary, completes the path with $dir
	            	stores the complete path into $self->{$key}
	Returns    :	nothing
	Args       :	$dir,	a directory that can be used to recover the path to the file
                	$file,	the file of interest
                	$key,	the key of the hash corresponding to file
                	$type,	the type of the file to test, only used to specify an exe test.
 	Globals    :	none

=cut

sub __ChecksArguments
{
	my $self = shift;
	my ($dir, $file, $key, $type) = @_;
	$type = 'file' if ( !defined($type) );

	if(!defined($file) or ($file eq '') )
	{	throw LipmError::ParameterException($file, 'missing');	}

	# si le chemin n'est pas absolu, on le recompose
	$file = File::Spec->catfile( $dir, $file )	if ( $file !~ /^\// );

	# fichier existe ?
	if ( !-e $file )
	{	throw LipmError::ParameterException($file, 'value');	}

	# fichier executable ?
	if ( ($type eq 'exe') and (!-x $file) )
	{	throw LipmError::ParameterException($file, 'value');	}

	# the file is ok to be used !
	$self->{$key} = $file;
	return;
}


=head2 Procedure BuildImage

	Title      :	BuildImage
	Usage      :	BuildImage();
	Prerequiste:	none
	Procedure  :	Build the circos image if everything is ok
	Returns    :	true if succeed, false otherwise
	Args       :	none
 	Globals    :	none

=cut

sub BuildImage
{
	my $self = shift;

	# 1- write the configuration file
	$self->__WriteConfFile();

	# 2- generate the image
	$self->__GenerateImage();

	return;
}


=head2 function __WriteConfFile

	Title      :	__WriteConfFile
	Usage      :	__WriteConfFile();
	Prerequiste:	none
	Function   :	parse the conf file if it's a template and writes a valid config file
	Returns    :	true if succeed, false otherwise
	Args       :	none
	Error      :	LipmError::ParameterException
 	Globals    :	none

=cut

sub __WriteConfFile
{
	my $self = shift;

	# open the output file
	my $fh = new IO::File( $self->{_conf_output}, "w" )
			or throw LipmError::IOException( $self->{_conf_output}, 'f' );

	# parsing the template file
	my $conf_file = $self->_ParseTemplate( $self->{_conf_file} );

	#writing the output file
	print $fh $conf_file;

	$fh->close;

	return &TRUE;
}


=head2 function __GenerateImage

	Title      :	__GenerateImage
	Usage      :	__GenerateImage();
	Prerequiste:	none
	Function   :	launch the perl command to run circos program
	Returns    :	true if succeed, false otherwise
	Args       :	none
	Error      :
 	Globals    :	none

=cut

sub __GenerateImage
{
	my $self = shift;
	# je mets 'perl' devant car le pragma change souvent
	my $perl_cmd = "perl $self->{_circos} -conf $self->{_conf_output}";
#warn $perl_cmd;
	$self->_Trace( "!-$perl_cmd-!\n", '.', 'IcV', __PACKAGE__ );
# warn Dumper { imagemap => $self->{ _imagemap } };

	my @a_rumors = `$perl_cmd`;
	if ( @a_rumors == 0)
	{
		warn "CIRCOS RETURNED SOMETHING WRONG!";
		warn $perl_cmd;
		throw LipmError(-text => "la commande $perl_cmd ne s'est pas executee normalement" );
	}
	else
	{
		return &TRUE;
	}
#	my $pb = ( 0 == @a_rumors ) ? 0 : 1; # si rien sur la sortie std.
#warn Dumper $pb;
	# il faut trouver un moyen de lire ce que la commande ecrit !
	# la ca ne marche pas
# 	my @a_pb = grep(/fatal error/, @a_rumors );
# 	$pb |= (0 != scalar( @a_pb ) ); # erreur a l'execution.

	#aucun SystemException n'existe encore
# 	throw Error::SystemException($perl_cmd, "@rumors");



#	throw LipmError(-text => "la commande $perl_cmd ne s'est pas executee normalement" )
#		if( 1 == $pb  );	# si rien sur la sortie std.
# 	$self->PrintMap( @a_rumors )	if( &TRUE == $self->{ _imagemap } );


#	return &TRUE;
}


# =head2 procedure PrintMap
# 
# 	Title      : PrintMap
# 	Usage      :	PrintMap( @a_lines );
# 	Prerequiste:	not available before version 0.3x
# 	Procedure  :	this function parse the entry to find the coords
# 	            		of the polygon that represents the chromosomes.
# 	            	then it simply print them out
# 	Args       :	@a_lines, an array of strings, the output of the circos prog.
# 
# =cut
# 
# sub PrintMap
# {
# 	my $self = shift;
# 	my ( @a_lines ) = @_;
# 
# 	my $filename = $self->{ _img_file } . '.map';
# 	my $f_map = new IO::File( $filename, 'w' )
# 			or throw LipmError::IOException( $filename, 'f' );
# 	foreach my $line ( @a_lines )
# 	{
# 		next if( $line !~ /^imagemap\sslice\sideogram/ );
# 		my ( $key ) = $line =~ /ideogram\s-\s((0-)?\S+)\s/;
# 
# 		my ( $coords ) = $line =~ /vertices=([\d,-\.;]+)\s/;
# 		$coords =~ s/;/,/g;
# 
# 		print $f_map "$key=$coords\n";
# 	}
# 
# 	return;
# }
# 

=head2 procedure Dump

	Title      :	Dump
	Usage      :	Dump();
	Prerequiste:	none
	Procedure  :	print all attributes to STDERR output _mostly for debug purpose
	Returns    :	nothing
	Args       :	none
 	Globals    :	none

=cut

sub Dump
{
	my $self = shift;

	foreach my $key ( keys(%$self) )
	{
		print STDERR "$key => $self->{$key}\n"	if( $key ne '_err' );
		$self->DumpErrors()	if( $key eq '_err' );
	}

	return;
}


=head2 function _ParseTemplate

	Title       :	_ParseTemplate
	Usage       :	$file_parsed = $self->_ParseTemplate($filepath)
	Prerequiste :	file must not contain any variable named 'warning'
                	Variables that can be substitued are 'var~xxx~'
	function    :	open and read the file passed as param and substitute variables.
	Returns     :	return a string which is nothing more than the file with perl evaluation of variables
	Args        :	$filepath	the file to read
	Error       :	LipmError::ParameterException
	Globals     :	none

=cut

sub _ParseTemplate
{
	my $self = shift;
	# file to parse
	my ($file) = @_;

	my $page = "";
	my $fh = new IO::File( $file, 'r' )
				or throw LipmError::IOException( $file, 'f' );
# 	throw LipmError::ParameterException("#$file#", 'value')	if (!defined($fh));


	#just collect each line of the file
	$page = join( '', <$fh> );
# 	while(my $line = <$fh>)
# 	{	$page .= $line;	}
	$fh->close;

# print STDERR "ce que contient $file :\n$page\n";
	# evaluate all variable 'var~xxx~' of the file
	# where xxx is an attribute's name

# 	$page =~ s/var~([^~\s]+)~/$self->{"_$1"}/g;

	while( $page =~ /var~([^~\s]+)~/ )
	{
# 		print STDERR "$1 = " . $self->{"_$1"} . "\n";
		my $pp = $self->{"_$1"};
		if( !defined( $pp ) )
		{
			my $msg = "The template.conf file has a template for $1 which is not in the object.\n";
			throw LipmError::ParameterException( $1, 'missing', -text => $msg, -user_msg=> $msg );
		}
		my $nb = $page =~ s/var~$1~/$pp/g;
# 		print STDERR "$nb substitution effectuee\n";
	}

	return $page;
}


=head2 GetImageFilename

 Usage        : my $filepath = $o_x->GetImageFilename( $ext )
 Function     : return the path corresponding to "_img_path / _img_file . $ext".
 Returns      : $filepath, string
 Args         : $ext, string, any extension you wish none by default.
 Error        : none

 Note : en fait elle devrait s'appeler GetImageFilepath
=cut

sub GetImageFilename
{
	my $self   = shift;
	my( $ext ) = @_;

	my $filename = $self->_Get( '_img_file' );
	$filename   .= ".$ext" if( defined( $ext ) );

	return $filename;
}


=head2 GetImageFilepath

 Usage        : my $filepath = $o_x->GetImageFilepath( $ext )
 Function     : return the path corresponding to "_img_path / _img_file . $ext".
 Returns      : $filepath, string
 Args         : $ext, string, any extension you wish none by default.
 Error        : none

=cut

sub GetImageFilepath
{
	my $self = shift;
	my( $ext ) = @_;

	my @a_attr   = qw( _img_path _img_file );
	my $filepath = File::Spec->catfile( $self->_Get( @a_attr ) );
	$filepath   .= ".$ext" if( defined( $ext ) );

	return $filepath;
} #GetImageFilepath



=head2 GetImageBasename

 Usage    : my $img_file = $o_x->GetImageBasename()
 Function : accessor to _img_file.
 Returns  : $img_file
 Args     : none
 Error    : none

=cut

sub GetImageBasename
{
	my $self = shift;

	return $self->_Get( '_img_file' );
} #GetImageBasename


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
