
#
# 	Sebastien.Letort@toulouse.inra.fr
#	Created: March 26, 2007
#	Last Updated: May 10, 2007
#	$Id$
#


# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Circos::Narcisse;

=pod

=head1 NAME

CircosNarcisse

=head1 SYNOPSIS

	This class generate Circos images for Narcisse.

=head1 DESCRIPTION

	With this class you can generate Circos images,
	you can managed configuration file to represent what you want
	It is a specification of the Circos class.

	- __level	:	level of the syntheny
	- __method	:	method used to compute syntheny
	- __org_ref	:	short name of the reference org.
	- _auto_hide	:	true if we only display chrs with links
	- _img_file	:	img filename
	- _conf_file	:	URI of the template configuration file
	- _conf_output	:	name of the configuration file
	- __local_conf_dir	:	URI of the session, where to put conf file
	- _data_dir	:	URI of the folder containing narcisse data
	- _chrs_list	:	string containing all chrs IDs coma separated
	- _chrs_links	:	URI of the output file containing the links between chrs
	- _karyotype	:	URI of the data file containing the karyotype
	- __rh_orgs	:	the organisms, keys are orgs' name values are hashs which keys are chrs name and values are hashes with 3 keys
		-- org_name --> chrs_name
			---> show	:	true if should be displayed
			---> color	:	color alias name of this chrs
			---> length	:	length of this chrs
		-- max_length	:	length of the longer chrs
		-- nb_bases	:	nb of bases in the entiere representation
	- _ticks_rules	:	string representing ticks rules for circos configuration file
	- exist	:	boolean, true if the image has already been calculated
	- nb_links_max :	constant, the max nb of lines that can contains the links file

	-_links	:	URI of the links configuration file
	- _font	:	URI of the folder containing font
	- _circos	:	URI of the folder containing the executable file
	- _img_path	:	URI of the folder containing
	- _chrs	:	URI of the chrs configuration file
	- _colors	:	URI of the colors configuration file
	- _radius	:	radius of the circle representing chrs.
	- _unite	:	nb of bases per Unit.


=head1 SUBROUTINES

	GetValueInUnit
	GetOrderOfMagnitude
	GetColorAlias
	GetLegendInfos
	LargeArrondi
	Show
	#inherited
	GetImageFilename
	Dump
	ComputeMap
	BuildImage

	_Init
	_ComputeImgName
	#inherited
	_ParseTemplate

	__GetColorsRules
	__WriteConfFile
	__PurgeFileNameList
	__GetDirectoriesElements
	__ConcatenateLinks
	__GetInputPathPart
	__OpenFile
	__WriteLinksFile
	__WriteKaryotypeFile
	__GetColorParams
	__SetOrgsAttribute
	__AffectIfCorrectValues
	__CheckParameters
	__SetTicksRules
	__SetMixedInfos
	__SetChrsList

=cut

use warnings;
use strict;

use IO::File;
use IO::Dir;
use Digest::MD5;	# to reduce the length of the img filename
use Data::Dumper;
use Cwd 'abs_path';

# heritage
use base qw(Circos);

#pour TRUE et FALSE
use General;

use DataReader::Whole;

# error management
use Error qw(:try);
use LipmError::ParameterException;
use LipmError::IOException;

use NARCISSE_utils;
use Sequence_utils;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};
}

# les constantes
	#the name of the directory where the circos package is installed
use constant NUANCES	=>	40;
use constant PACK_NAME	=>	"Circos::Narcisse";

use constant IMG_EXT   => "png";

# we limit the number of org to display because of software limitation
# we need at least 40 shades (one per chrs) by color, and one color by org
# as circos doesn't use imageTrueColor, we are limited to 256 colors.
# so 256 / 40 ~ 6
# but nowaday we are unable to generate exactly 40 shades _it can be changed
# as we generate around 44 shades we must limit the number of org to 5
use constant MAX_ORG	=>	5;
use constant RA_COLORS	=>	['blue', 'green', 'pink', 'grey'];
use constant COLOR_REF	=>	'red';

# to correct the bug of chr name : mouse_un_r => pb caused by '_un_'
# '_' is prohibited because if the chr as already a _ inside (like Un_r)
#		the pattern _xx_ will be intercepted, and provoque an error
# '-' is prohibited as it is a special character of karyotype file
# '' is forbidden because we do a split onto this separator.
# '*' is forbidden because circos seems to eval the string
# '#' ';'  '%' idem
use constant NAME_SEP => '-';

# the prefix that starts the organism of reference into rh_orgs
# $__PREFIX_REF est une variable de classe privee, invisible de l'exterieur
my $__PREFIX_REF	=	'0-';

=head2 function GetPrefixRef

 Title        : GetPrefixRef
 Usage        : $prefix = &Circos::Narcisse::GetPrefixRef();
 Prerequisite : it's a static function
 Function     : accessor to the prefix
 Returns      : $prefix, the string that must prefix the reference organism.
 Args         : none

=cut

sub GetPrefixRef
{
	return $__PREFIX_REF;
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_circos = New CircosNarcisse(%h_param);
 Prerequisite : circos must be install into the "path_to_Core"/CIRCOS_DIR directory
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Returns      : none
 Args         :	%h_param,	a hash of parameters, here the details of keys :
                    o_species, SpeciesObserver of the reference
                    circos_dir, the path to the directory where circos is installed
                    output_dir, the path to output directory (=sessdir ?)
                    level, the synteny level
                    method, name of the program used to compute synteny
                    rh_orgs, a reference to a hash whose keys are the short names of organisms. 
                    The values are hash which keys are 'long' to get the long name, 'short' to get the prefixed short name, 
                    and 'chrs' to get a ref to the chrs hash whose keys are chrs names and values are length of chrs.
#               		karyotype, the name of the karyotype file.
              		conf_output_dir, le repertoire ou sera genere le fichier de configuration.
              		auto_hide, boolean, if true unlinked chromosomes will not be displayed.
 Error        :	LipmError::ParameterException
 Globals      : none

=cut

sub _Init
{
	my $self = shift;
	my %h_param = (	# required
		circos_dir => '',
		data_dir => '',
		output_dir => '',
		method => '',
		rh_orgs => {},

		# optional
		karyotype => './karyotype.txt',
		level => 2,
		conf_output_dir => '',
		auto_hide => '0',	#besoin d'affecter ici ?,
		image_size => 600,
		nb_links_max => 10000,
		imagemap => &FALSE,
		chroms_url =>'',
		@_);

	# 0- check parameters
	$self->__CheckParameters( %h_param );

	# 1- get org reference => set __rh_orgs and __org_ref
	$self->__SetOrgsAttribute( $h_param{rh_orgs} );
# ShowDebug( %{ $self->{ __rh_orgs } });
# ShowDebug( %{ $h_param{rh_orgs} }, %{ $self->{ __rh_orgs } });

	# 2- initialisation of attributes
	@$self{ qw( _data_dir _karyotype nb_links_max __chroms_url ) } =
	@h_param{ qw( data_dir karyotype nb_links_max chroms_url ) };

	$self->_Set( __level  => $h_param{ level },
				 __method => $h_param{ method } );

	# - those needed by the circos template.conf
	$self->{_chrs_list}	= "";
	$self->{ _unite }	= 5;	# valeur a  peu pres bonne. pb c'est pas dynamique
	$self->{_ticks_rules} = "";
	$self->{  _highlights_perso } = "";

	# 2.1- conditionnal affectation
	$self->__AffectIfCorrectValues( $h_param{auto_hide},
							$h_param{conf_output_dir}, $h_param{output_dir} );

	# 2.2- set a name to the image
	my $img_file = $self->_ComputeImgName( $h_param{rh_orgs} ); # . '.png';
# 	my $img_file = $self->_ComputeImgName( $h_param{rh_orgs} ) . '.svg';

	# 3- call mummy
	my %h_mother_param = (
					circos_dir => $h_param{circos_dir},
					conf_dir => $h_param{circos_dir} . '/etc/Narcisse/',
					data_dir => $h_param{data_dir},
					img_dir => $h_param{output_dir},
					img_file => $img_file,
					colors => "colors.conf",
					fonts => $h_param{circos_dir} . '/fonts/',
					image_size => $h_param{image_size},
					imagemap => $h_param{ imagemap }
				);
	$self->SUPER::_Init( %h_mother_param );
# ShowDebug( %{ $self->{ __rh_orgs } });

	my( $org_ref, $local_conf ) = @$self{ qw( __org_ref __local_conf_dir ) };
	my( $lvl, $prog ) = $self->_Get( qw( __level __method ) );
	my $links_file = join( '_', $local_conf . $org_ref, $prog, $lvl, 'links.txt' );
	my $o_species  = New SpeciesObserver( species => $self->{ __org_ref } );
	$self->_Set( 
        __o_ref     => $o_species,
        _chrs_links => $links_file,
        __o_io      => $o_species->GetInstanceObserver());

	if( -e $self->GetImageFilepath( &IMG_EXT ) )
	{	$self->{exist} = &TRUE;	return;	}
	$self->{exist} = &FALSE;

	# 4- generate karyotype file
	$self->__WriteKaryotypeFile();
# ShowDebug( %{ $self->{ __rh_orgs } });

	# - generate chrs_links file
	$self->__WriteLinksFile( $h_param{rh_orgs} );
# ShowDebug( %{ $self->{ __rh_orgs } });

	# - generate perso file for highlights
	$self->__WritePersoFile();
# ShowDebug( %{ $self->{ __rh_orgs } });

	# set the list of displayed chrs
	$self->__SetChrsList();
# ShowDebug( %{ $self->{ __rh_orgs } });

	# set
	$self->__SetMixedInfos();

	return;
}


=head2 Procedure __CheckParameters

 Title      :	__CheckParameters
 Usage      :	$self->__CheckParameters( %h_param );
 Prerequiste:	none
 Procedure  :	record an error (and stop the prog)
            	if one of the tested arguments is empty
 Returns    :	nothing
 Args       :	%h_param, 	we test the following key
 Error      :	LipmError::ParameterException
 Globals    :	none

=cut

sub __CheckParameters
{
	my $self = shift;
	my (%h_param) = @_;

	# TODO ParameterException::Assert
	foreach my $key ( qw( circos_dir data_dir output_dir
							method ) )
	{
		if( $h_param{ $key } eq '' )
		{	throw LipmError::ParameterException( $key, 'missing');	}
	}

	# non empty or overloaded hash
	my $nb_orgs = scalar( keys(%{ $h_param{rh_orgs} }) );
	if( ( $nb_orgs == 0 ) or ( &MAX_ORG < $nb_orgs ) )
	{
		throw LipmError::ParameterException('rh_orgs', 'value',
								-should_be => "0 < X <= " . &MAX_ORG);
	}

	return;
}

=head2 Procedure __AffectIfCorrectValues

	Title      :	__AffectIfCorrectValues
	Usage      :	$self->__AffectIfCorrectValues( $auto_hide, $conf_out, $out );
	Prerequiste:	none
	Procedure  :	set those arguments to attributes only if they have "valid" value
	Returns    :	nothing
	Args       :	$auto_hide,	if equals "1", unlinked chromosomes will not be displayed
	            	$conf_out, directory where to found configuration file
	            	$out, output dir
 	Globals    :	none

=cut

sub __AffectIfCorrectValues
{
	my $self = shift;
	my ($auto_hide, $conf_out, $out) = @_;

	$self->{_auto_hide} = ($auto_hide eq "1") ? &TRUE : &FALSE;

	#verifier que conf_output_dir exist ?
	$self->{__local_conf_dir} = ($conf_out ne '') ? $conf_out : $out;
	$self->{__local_conf_dir} .= '/';

	return;
}


=head2 Procedure __SetOrgsAttribute

 Title      :	__SetOrgsAttribute
 Usage      :	$self->__SetOrgsAttribute( $rh_orgs );
 Prerequiste:	none
 Procedure  :	reform the hash containing chrs information
                    and store it as an attribute __rh_orgs.
                we add color representation and a boolean show.
                By default none of the chrs will be displayed.
 Returns    :	nothing
 Args       :	$rh_orgs, a ref to an hash that contains org's info cf doc of Init.
 Globals    :	none

 Note : a refaire proprement avec des sous fonctions, et refaire __GetColorParams avec.

=cut

sub __SetOrgsAttribute
{
	my $self = shift;
	my ($rh_orgs) = @_;

	# 1- identifier la ref
	my ( $key_ref ) = grep( /^$__PREFIX_REF/, keys %$rh_orgs );
	my $short = $self->{ __org_ref } = $$rh_orgs{ $key_ref }{ short };

		# il faut fusionner les hashs de chrs (0-xx et xx) si on veut les liens internes
		my $rh_chrs_ref = $$rh_orgs{ $key_ref }{ chrs };
		my %h_chrs_fusion;
		while( my ( $name, $rh_chrs_infos ) = each( %$rh_chrs_ref ) )
		{	
			$h_chrs_fusion{ $name } = $$rh_chrs_infos{ length };
		}

	if( defined( $$rh_orgs{ $short } ) )
	{
		my $rh_chrs = $$rh_orgs{ $short }{ chrs };
		while( my ( $name, $rh_chrs_infos ) = each( %$rh_chrs ) )
		{	$h_chrs_fusion{ $name } = $$rh_chrs_infos{ length };	}
	} # fin if defined $$rh_orgs{ $short }

		# $i est un indice necessaire pour trouve une couleur
		my $i = 0;

		my $nb_chrs = keys( %h_chrs_fusion );
		my ($color, $shade_step) =$self->__GetColorParams( $key_ref, $nb_chrs, \$i);
		# $i = tjs 0 et $colors = COLOR_REF
		my $indice = 0;

		# - extract chrs name and length
		my %h_chrs_infos;
		my @a_sorted_chrs = &NARCISSE_utils::SortByChromID( keys %h_chrs_fusion );
		foreach my $chrs ( @a_sorted_chrs )
		{
			$indice += $shade_step;
			my $length = $h_chrs_fusion{ $chrs };
			my $chrs_name = $short . &NAME_SEP . $chrs;

			my $rh_info = { length	=> $length,
							color	=> $color.$indice,
							show	=> &FALSE,
						};

			#set the attribute
			$h_chrs_infos{ $chrs_name } = $rh_info;
		} # !chrs
		$self->{__rh_orgs}{ $key_ref } = \%h_chrs_infos;
# print STDERR Dumper $self->{__rh_orgs};


	# - parsing the organisms
	# org for 1 org and orgs for all orgs
	while( my ($key, $rh_org) = each( %$rh_orgs ) )
	{
# ShowDebug( $key );
		# je ne veux pas de doublons en cas de liens internes
# 		next	if( $key =~ $short );
		next	if( $key eq $key_ref );

		my $rh_chrs = $$rh_org{ chrs };
		my $nb_chrs = keys( %$rh_chrs );
		my ($color, $shade_step) =$self->__GetColorParams( $key, $nb_chrs, \$i);
		my $indice = 0;

		# - extract chrs name and length
		my %h_chrs_infos;
		my @a_sorted_chrs = &NARCISSE_utils::SortByChromID( keys %$rh_chrs );
		foreach my $chrs ( @a_sorted_chrs )
		{
			$indice += $shade_step;
			my $length = $$rh_chrs{$chrs}{ length };
			my $chrs_name = $key . &NAME_SEP . $chrs;

			my $rh_info = { length	=> $length,
							color	=> $color.$indice,
							show	=> &FALSE,
						};

			#set the attribute
			$h_chrs_infos{ $chrs_name } = $rh_info;
		} # !chrs
		$self->{__rh_orgs}{ $key } = \%h_chrs_infos;
	}

# warn Dumper $self->{__rh_orgs};

	if( !defined( $self->{__org_ref} ) )
	{
		my $msg = "the organism of reference must begin with " . $__PREFIX_REF;
		my $continue = "There's a pb with the argument, your reference organism should be prefixed.";
		throw LipmError::ParameterException('__org_ref', 'missing',
										-text => $msg, -onContinue => $continue );
	}

	return;
}


=head2 Function __GetColorParams

	Title      :	__GetColorParams
	Usage      :	($color, $indice, $shade_step) =
	                			$self->__GetColorParams( $org_key, $nb_chrs, $ri );
	Prerequiste:	none
	Function   :	attributes a color for each org,
	            	computes the first color indice for this org
	            	and the shade step.
	            	It also set the $self->{__org_ref} attribute.
	Returns    :	the dominant color of the organism,
	            	the lower indice of that org. and the shade step.
	Args       :	$org_name, 	the name of the current org. prefixed short
	            	$nb_chrs,	the number of chrs involved
	            	$ri, a reference to the color indice _can be changed by this function.
	Error      :	LipmError::ParameterException.
 	Globals    :	none

=cut

sub __GetColorParams
{
	my $self = shift;
	my ($org_name, $nb_chrs, $ri) = @_;

	# les constantes ne peuvent etre utilisees dans les expressions rationnelles
	my ($color);

	# i set $self->{__org_ref} as we found it
	if ($org_name =~ /^$__PREFIX_REF/)
	{
		# le prefix sert uniquement a placer la ref en tete de tri alphabetique
		# pourquoi circos fait un tri alphabetique ? parce que les hashs sont incertaines
		$color = &COLOR_REF;
	}
	else
	{	$color = &RA_COLORS->[$$ri++];	}

# detect a failure : no reference !
# ca ne marche pas, d'ailleurs, je me demande si c'est bien la qu'il faut faire ce test !
# -> aucune erreur jetee <= record ne jete rien.
# code remonte a SetOrgsAttribute
# 	if( !defined( $self->{__org_ref} ) )
# 	{
# 		my $msg = "the organism of reference must begin with " . &PREFIX_REF;
# 		record Error::ParameterException('__org_ref', 'missing',
# 										-text => $msg );
# 	}

	# -how many chrs for that organism
	if($nb_chrs >= &NUANCES)
	{
		my $msg = "too much chrs selected.";
		my $should_be = "nb_chrs < " . &NUANCES;
		my $continue = "too much chromosomes, contact your administrator";
		throw LipmError::ParameterException($org_name, 'value',
							-should_be => $should_be, -text => $msg,
							-onContinue => $continue );
	}

	# le int sert a retourner un arrondi inferieur
	my $shade_step = int(&NUANCES / ($nb_chrs+1));

	return ($color, $shade_step);
}


=head2 Function __WriteKaryotypeFile

	Title      :	__WriteKaryotypeFile
	Usage      :	$msg_err = $self->__WriteKaryotypeFile();
	Prerequiste:	none
	Function   :	Write a karyotype file from data found into $self->{__rh_orgs}
	            	A line of this file is like :
	            	$chrs_name	1	$length	$chrs_name	$color.$indice
	Returns    :	undef if everything was well, a message of error otherwise
	Args       :	none
 	Globals    :	none

=cut

sub __WriteKaryotypeFile
{
	my $self = shift;

	# - opening output files
	my $f_out	= new IO::File($self->{_karyotype}, "w")
				or throw LipmError::IOException( $self->{_karyotype}, 'f' );

	# - writing the organisms
	my $rh_orgs = $self->{ __rh_orgs };
# use Data::Dumper;
# print STDERR Dumper $rh_orgs;
	while( my ($key, $rh_chrs) = each( %$rh_orgs ) )
	{
		next	if( $key eq $self->{ __org_ref } );

		# - extract chrs infos
		while( my ($chrs_name, $rh_chrs_infos) = each( %$rh_chrs ) )
		{
			my $length = $$rh_chrs_infos{length};
			my $color = $$rh_chrs_infos{color};
			my @a_line_elt = ( 'chrs', '-', $chrs_name, $chrs_name );

			# I start at 0 but I'm not sure of that, we have to ask martin for that.
			push( @a_line_elt, ( 0, $length, $color ) );
			my $line = join("\t", @a_line_elt );
			print $f_out "$line\n";
		} # !chrs
	} # !org
	$f_out->close();

	return undef;
}


=head2 Function __WriteLinksFile

	Title      : __WriteLinksFile
	Usage      : $msg_err = $self->__WriteLinksFile($rh_orgs);
	Prerequiste: none
	Procedure  : Write a link file from data found into Narcisse tree directory
	             A link is composed of 2 lines, cf __ConcatenateLinks to know more
	Args       : $rh_orgs, the ref hash passed as an argument of the constructor.
 	Globals    : none

=cut

sub __WriteLinksFile
{
	my $self = shift;
	my ($rh_orgs) = @_;

	my $nb_lines = 0;

	# - opening output files
	my $f_out = $self->__OpenFile( $self->_Get( '_chrs_links' ), "w" );

	# - input file(s)
	my $can_be_drawn = &FALSE;	# can be drawn if we have at least one link !

	# on ne veut analyser que les fichiers pour lequel le chrs a ete selectionne
	#  -> ce sont les chrs de la reference
	my $key_ref = $__PREFIX_REF . $self->{ __org_ref };
	my $rh_org_ref = $$rh_orgs{ $key_ref };

	while( my ($key, $rh_org) = each( %$rh_orgs ) )
	{
		# pour chaque org sauf la reference
		next	if ( $key eq $key_ref );

# print STDERR Dumper [ $rh_org ];
		my $org = $$rh_org{ short };

		my @a_attr = qw( __o_ref __method __level );
		my( $o_ref, $prog, $lvl ) = $self->_Get( @a_attr );
		my %h_params = ( prog => $prog, level => $lvl,
						 target => $org, -existing_files => &TRUE );
		my @a_files_in = $o_ref->GetSyntFiles( %h_params );

		# @a_files_in contient la liste des fichiers de liens
		# pour chaque chrs de la ref lies a cet org. pas seulement ceux choisit !
		foreach my $file_in (@a_files_in)
		{
			# on ne s'interesse qu'aux fichiers de donnees
			my $chrs_nb = $o_ref->ExtractChromNameFromFilename( $file_in );
			next if( !defined( $$rh_org_ref{ chrs }{ $chrs_nb } ) );
# 			&ShowDebug( "chrs : $chrs_nb\n" );

			# on a besoin du nom prefixe pour ecrire les liens.
			my $link_found = $self->__ConcatenateLinks( $file_in, $f_out,
							$self->{ __org_ref }, $org, \$nb_lines, $rh_orgs);

			#si on a trop de lignes, on jete une erreur
			if( $nb_lines > $self->{nb_links_max} )
			{
				my $continue = "Sorry, we cannot compute the circos maps with those data.<br />";
				$continue .= "Could you please <b>remove one or several chromosomes</b>,";
				$continue .= " or retry with <b>a higher level</b> of syntheny";
				my $msg = "not enough time to compute this, Apache will kill it before it ends.";
				my $should = "nb_lines <= " . $self->{nb_links_max};
				throw LipmError::ParameterException($nb_lines, 'value',
										-txt => $msg,
										-should_be => $should,
										-onContinue => $continue);
			}

			# which chrs ref will be displayed ?
			if( $link_found == &TRUE )
			{
				$can_be_drawn = &TRUE;
				if($self->{_auto_hide} == &TRUE)
				{
					#	chrs full name
					my $chrs_name = $self->{ __org_ref } . &NAME_SEP . $chrs_nb;

	# 				$$rh_org_ref{$chrs_nb}{show} = &TRUE;
					$self->{__rh_orgs}{ $key_ref }{ $chrs_name }{show} = &TRUE;
				}
			}
		}#end @files_in
	}#@org_s
	$f_out->close();

	if( $can_be_drawn == &FALSE )
	{
		my $msg = "no link found for any org for level " . $self->{__level};
		throw LipmError::ParameterException("", 'x', -msg => $msg,
									-onContinue => "no synteny found");
	}

	return;
}


=head2 __WritePersoFile

 Usage    : $o_x->__WritePersoFile()
 Procedure: .
 Args     : none
 Error    : none.

=cut

sub __WritePersoFile
{
	my $self = shift;

	my $o_ref   = $self->_Get( '__o_ref' );
	my @a_files = $o_ref->GetPersoFiles( -existing_files => &TRUE );
	return if( 0 == @a_files );

	my $o_ndr = New DataReader::Whole();
	my @a_rows = (); # en fait ce sera un hash
	foreach my $file ( @a_files )
	{
		$o_ndr->Prepare( -from => $file );
		my $rh_data = $o_ndr->FetchallHashref( 'element', -line_type => 'HASH' );
		push( @a_rows, %$rh_data );
	}

	my %h_rows = @a_rows;
	my $short  = $o_ref->GetShortName();
	foreach my $rh_data ( values %h_rows )
	{
		$$rh_data{ chrom } = $short . '-' . $$rh_data{ chrom };
		$$rh_data{ color } = 'fill_color=' . $$rh_data{ color };
	}
	my @a_header = qw( chrom start end color );
	my $file = 'highlights.perso';
	&NARCISSE_utils::PrintNarcisseData( $file, \@a_header, \%h_rows );

	my $abs_path = &Cwd::abs_path( $file );
	my $radius = $self->{_radius};

	#MaJ de highlights_perso
	my $perso = <<CODE
	<highlight>
		file       = $abs_path
# 		ideogram   = yes
		r0         = 1.02r
		r1         = 1.05r
	</highlight>
CODE
;
	$self->{ _highlights_perso } = $perso;

	return;
} #__WritePersoFile


=head2 Function __OpenFile

	Title      :	__OpenFile
	Usage      :	$file_handle = $self->__OpenFile($filename, $mode);
	Prerequiste:	none
	Function   :	tries to open a file, and throw an error if it can't
	Returns    :	a valid file handle
	Args       :	$filename,	the file to open
	            	$mode, 	the opening mode ('r', 'w', 'a' => cf fopen)
	Error       	LipmError::ParameterException
 	Globals    :	none

=cut

sub __OpenFile
{
	my $self = shift;
	my ($filename, $mode) = @_;


	my $file = new IO::File($filename, $mode);
	if (!defined $file)
	{
		# on ne devrait pas a avoir a preciser lenom du fichier dans le message selon moi <-- Seb
		my $msg = "$filename could not be opened in '$mode' mode\n";
		throw LipmError::IOException($filename, 'file',
									-text => $msg, depth => 2);
	}

	return $file;
}


=head2 Function __ConcatenateLinks

	Title      :	__ConcatenateLinks
	Usage      :	$ok = $self->__ConcatenateLinks( $file_in, $fh_out, $q_org, $s_org, \$nb_lines );
 Prerequiste : none.
	Function   :	add to file_out links for Circos from Narcisse input files
	Returns    :	&TRUE if at least one link was written, &FALSE otherwise
	Args       :	$file_in,	data filename
                	$fh_out, a file handler, object from IO:File. Where data are written
                	$q_org,	the prefixed short name of the reference organism
                	$s_org,	the short name of the organism (target)
                	$r_nb_lines,	a ref to a scalar that count the nb of lines written
 	Globals    :	none

=cut

sub	__ConcatenateLinks
{
	my $self = shift;
	my ( $file_in, $fh_out, $q_org, $s_org, $r_nb_lines, $rh_orgs ) = @_;
	my $flag = &FALSE;

	# je charge tout le fichier en memoire
	my $o_ndr = New DataReader::Whole();
	$o_ndr->Prepare( -select => [],
						-from   => $file_in,
						-where  => {}
					);
	my $rh_data = $o_ndr->FetchallHashref('element', -line_type => 'HASH');

	foreach my $elt ( keys %$rh_data )
	{
		my $rh_line = $$rh_data{ $elt };
		my @a_q_elts = qw( chrom start end );
		my ( $q_chrs, $q_start, $q_end ) = @$rh_line{ @a_q_elts };
		my @a_s_elts = qw( s_chrom s_start s_end );
		my ( $s_chrs, $s_start, $s_end ) = @$rh_line{ @a_s_elts };

		# on passe si le chrs de la cible n'est pas voulu
		next	if( !defined $$rh_orgs{$s_org}{ chrs }{ $s_chrs } );

		#2.2- writing lines into the output file
		my $link_name = "${q_org}_${s_org}_$elt";
		my $ref_chrs_name = $q_org . &NAME_SEP . $q_chrs;
		my $tgt_chrs_name = $s_org . &NAME_SEP . $s_chrs;
		print $fh_out "$link_name\t$ref_chrs_name\t$q_start\t$q_end\n";
		print $fh_out "$link_name\t$tgt_chrs_name\t$s_start\t$s_end\n";
		$$r_nb_lines += 2;
		$flag = &TRUE;

		# which target chrs will be displayed
		my $chrs_name = $s_org . &NAME_SEP . $s_chrs;

		if($self->{_auto_hide} == &TRUE)
		{
			$self->{__rh_orgs}->{$s_org}->{$chrs_name}->{show} = &TRUE;

			if( $s_org eq $q_org ) # cas des liens internes
			{
				my $rh_org_ref = $self->{ __rh_orgs }->{$__PREFIX_REF . $q_org};
				$rh_org_ref->{ $chrs_name }->{ show } = &TRUE;
			}
		}
	} #foreach keys $rh_data

	return $flag;
}


=head2 function __WriteConfFile

	Title      :	__WriteConfFile
	Usage      :	__WriteConfFile();
	Prerequiste:	this methode overwrite the mother one to add colors rules
	Function   :	parse the conf file if it's a template and writes a valid config file
	Returns    :	true if succeed, false otherwise
	Args       :	none
 	Globals    :	none

=cut

sub __WriteConfFile
{
	my $self = shift;

	# open the output file
	my $fh = $self->__OpenFile( $self->{_conf_output}, "w" );

	# parsing the template file
	my $conf_file = $self->_ParseTemplate( $self->{_conf_file} );

	## What its mother doesn't do
	my $rules = $self->__GetColorsRules();
# warn $rules;
	# we insert those new rules just before the closing links tag
	$conf_file =~ s/<\/rules>/$rules<\/rules>/;

	#writing the output file
	print $fh $conf_file;

	$fh->close;

	return;
}


=head2 function __GetColorsRules

	Title      :	__GetColorsRules
	Usage      :	$rules = __GetColorsRules();
	Prerequiste:	none
	Function   :	add rules to make links have the color of the target organism (ie not ref)
	Returns    :	a string containing those new circos rules
	Args       :	none
 	Globals    :	none

=cut

sub __GetColorsRules
{
	my $self = shift;
	my $rules = "";
	my $rh_orgs = $self->{__rh_orgs};

	foreach my $org (keys %$rh_orgs)
	{
		next	if ($org =~ /^$__PREFIX_REF(\S+)/);

		my $rh_chrs = $$rh_orgs{$org};
		foreach my $chrs (keys %$rh_chrs)
		{
			my $color = $$rh_chrs{$chrs}->{color};
			$rules .= "
	<rule>
		importance	=	9
		condition	=	_CHR2_ eq \"$chrs\"
		color	=	$color
	</rule>
	";
		}
	}

	return $rules;
}


=head2 Function Show

	Title      : Show
	Usage      : $html = $self->Show( $url );
	Prerequiste: $self->{_img_file} should exist.
	Function   : print a HTML tag to display the image, and the associated maps
	Returns    : The html
	Args       : $url,	the base url to the image.
	TODO       : __chroms_url N'EST PLUS UTILISE, tout se trouve dans Core/circos/etc/Narcisse/chrs.conf
	             Peut-etre faudrait-il revoir cela, par exemple pour echanger target/reference selon les chrs ?                

=cut

sub Show
{
	my $self = shift;
	my ($url) = @_;
    my $rh_orgs = $self->{ __rh_orgs };
    my $o_io    = $self->_Get('__o_io');
#    use Data::Dumper;
#    warn Dumper $rh_orgs;

	my $filename = $self->GetImageFilename( &IMG_EXT );
	my $page = "<img src='$url/$filename' alt='Circos image'";

	my $map = '';
# warn $self->{ _imagemap };
	if( &TRUE == $self->{ _imagemap } )
	{
		my $map_filename = $self->GetImageFilename( 'html' );
		my $ofh_map = new IO::File( $map_filename, 'r' )
				or throw LipmError::IOException( $map_filename, 'f' );
		my @a_lines = <$ofh_map>;
		$map = join( "\n", @a_lines );
# 		$map = '<map name="map" id="circos_map">';
# 
# 		my $filename = $self->{ _img_file } . '.map';
# 		my $o_ref = $self->_Get( '__o_ref' );
# 
# 		# un polygon par ligne <-> par chrs
# 		my $o_map = New ParamParser( $filename );
# 		foreach my $chrs_name ( $o_map->GetKeys() )
# 		{
# 			# PATCH BY MANU - Calculer nom de reference et de chromosome
# 			my ($orga,$chr) = split( &NAME_SEP, $chrs_name );
#             $orga           = $o_io->GetSpeciesLongName( $orga );
# 			my $url         = $self->{ __chroms_url };
# 			$url =~ s/reference=/reference=$orga/;
# 			$url =~ s/chroms_reference=/chroms_reference=$chr/;
# 
# 			$map .= "\n<area alt='$chrs_name' title='$orga ($chr)'";
# 			$map .= " shape='poly' coords='";
# 			$map .= $o_map->Get( $chrs_name );
# 			$map .= "' href='$url' ";
#             $map .= 'onmouseover="javascript:O_CIRCOS_CHROM.MouseOver(0,0,0,0,\'' . $orga . '\',\'' . $chr . '\')"';
#             $map .= ' onmousedown="javascript:O_CIRCOS_CHROM.MouseDown(event)"';
#             $map .= " />";
# 		}
# 
# 		$map .= '</map>';
		$page .= ' usemap="#' . $self->GetImageFilename() . '"';
	}

	$page .= " id='circos_img' />";

	return $map . $page . "\n";
}


=head2 procedure __SetChrsList

	Title      :	__SetChrsList
	Usage      :	$self->__SetChrsList( );
	Prerequiste:	none
	Procedure   :	set the chrs list, it contains the ID of the chrs we want to display.
	            	The ID is $org_name_$chrs_name.
	            	If auto_hide isn't set, the default 'show' value is changed to TRUE
	            	As we want, is that case, show all chrs.
	Returns    :	nothing
	Args       :	none
	Globals    :	none

=cut

sub __SetChrsList
{
	my $self = shift;
	my @a_chrs_list = ();
# 	my %h_chrs_list = ();

	my $rh_orgs = $self->{ __rh_orgs };
# use Data::Dumper;
# print STDERR Dumper [ $rh_orgs ];
	my $show_all = !$self->{_auto_hide};

	# note : on perd l'ordre de selection des org, sauf la ref.
	foreach my $key (sort keys %$rh_orgs)
	{
		# si l'org est la ref, alors on est sur un "self" et on n'en veut pas
		next	if( $key eq $self->{ __org_ref } );

		# le cas des liens internes suppose que la liste des chrs peut contenir
		#	des doublons, pour les éviter, j'utilise un hash
		my $rh_chrs = $$rh_orgs{ $key };
		my @a_sorted_chrs = &NARCISSE_utils::SortByChromID( keys(%$rh_chrs) );
		foreach my $chrs ( @a_sorted_chrs )
		{
			# we change the show value if all should be displayed
			$$rh_chrs{$chrs}{show} = &TRUE	if( $show_all );

			# on affiche ce chrs si on doit le montrer
# 			$h_chrs_list{ $chrs } = 1	if( $$rh_chrs{$chrs}{show} == &TRUE );
			push(@a_chrs_list, $chrs)	if( $$rh_chrs{$chrs}{show} == &TRUE );
		}
	}

	$self->{_chrs_list} = join(';', @a_chrs_list);

	return;
}



=head2 method __SetMixedInfos

 Title        : __SetMixedInfos
 Usage        : $o_xx->__SetMixedInfos(   );
 Prerequisite : none
 Function     : calcul l'espacement entre les taquets
 Returns      : none
 Args         : none

=cut

sub __SetMixedInfos
{
	my $self = shift;
	my %h_infos;

	$h_infos{ max_length } = 0;
	$h_infos{ nb_bases } = 0;

	my $rh_orgs = $self->{__rh_orgs};
# use Data::Dumper;
# print STDERR Dumper [ 1084, $rh_orgs ];
	foreach my $key (keys %$rh_orgs)
	{
		my $rh_chrs = $$rh_orgs{ $key };
		foreach my $chrs (keys %$rh_chrs)
		{
			next	if( $$rh_chrs{ $chrs }->{ show } == &FALSE );

			# mixed infos
			my $length = $$rh_chrs{$chrs}->{length};
			$h_infos{ nb_bases } += $length;
			if( $h_infos{ max_length } < $length )
			{	$h_infos{ max_length } = $length;	}
		}
	}# fin foreach : max_length et nb_bases sont connus

	# conf des taquets :
	my %h_ticks_conf;
	$h_ticks_conf{nb_ticks_all_over} = 20;		# 1 taquet tous les 5%
	$h_ticks_conf{nb_ticks_min_over_the_longer} = 3;	#nb mini de grand taquets sur le + gd chrs
	$h_ticks_conf{nb_small_ticks_between_two_big} = 4;

	# value of big and small will be x*10y
	# and x is a value of @a_grads
	my ($big, $small) = (undef, undef);
	my @a_grads = (1, 2, 2.5, 4, 5, 10);

	# ~ 1 tous les 5%
	#!! si $self->{__rh_infos}->{nb_bases} == 0 Error !!
	my $ratio = $h_infos{ max_length } / $h_infos{ nb_bases };
	my $ratio_min = $h_ticks_conf{nb_ticks_min_over_the_longer} / $h_ticks_conf{nb_ticks_all_over};

	# on met les grands taquets que si on a la visibilite pour
	if ($ratio > $ratio_min)
	{
		my $N = $h_infos{ nb_bases } / $h_ticks_conf{nb_ticks_all_over};
		$big	=	LargeArrondi($N, @a_grads );
	}
	else
	{
		my $N = $h_infos{ max_length };
		$big	=	LargeArrondi($N, @a_grads );
	}
		$small = $big / $h_ticks_conf{nb_small_ticks_between_two_big};

	$self->__SetTicksRules($big, $small);

	return;
}


# return (x * 10^y) tel que x est une valeur de @a_vals
sub LargeArrondi
{
	my ($float, @a_vals) = @_;
	my $return;

	# 1- obtenir l'ordre de grandeur du nb
	my ($dixieme, $power) = GetOrderOfMagnitude( $float );

	# 2- obtenir la valeur la plus proche de notre quotient
	for(my $i=0; $i < @a_vals-1; ++$i)
	{
		if( $dixieme <= $a_vals[$i+1])
		{
			$return = $a_vals[$i];
			last;
		}
	}

	if(!defined $return)
	{
		print STDERR "Your value $dixieme does not fit in the array (@a_vals)\n";
		return undef;
	}

	return $return * 10**$power;
}


=head2 procedure __SetTicksRules

	Title      :	__SetTicksRules
	Usage      :	$self->__GetTicksRules( $big, $small );
	Prerequiste:	ticks.conf indicates that label is in 10^6 base
	Procedure   :	Set the rules used to display Ticks over the circle.
	Returns    :	nothing
	Args       :	$big, nb of bases between 2 big ticks (with label)
	            	$small, nb of bases between 2 small ticks (without label)
	Globals    :	none

=cut

sub __SetTicksRules
{
	my $self = shift;
	my ($big, $small) = @_;

	if( defined $big )
	{
		my $format = '%d';	# selon ticks.conf,la valeur reelle est divisee par 10^6¶
		if( $big >= 10**6 )
		{	$format = '%d Mb';	}
		elsif( $big >= 10**3 )
		{	$format = '%.2f Kb';	}

		$self->{_ticks_rules} .= "
	<tick>
		spacing      = $big" . "b
		size         = 15p
		thickness    = 2p
		color	     = grey42
		show_label   = no
		label_size   = 12p
		label_offset = 2p
		format       = $format
	</tick>";
	}

	if( defined $small )
	{
		$self->{_ticks_rules} .= "
	<tick>
		spacing   = $small" . "b
		size      = 5p
		thickness = 1p
		color	= grey42
		show_label = no
	</tick>";
	}

	return ;
}

=head2 Function GetLegendInfos

	Title    : GetLegendInfos
	Usage    : my %h_legend_info = $o_circos->GetLegendInfos()
	Function : Compute and return a hash containing information to build the static legend, or to pass to the dynamic legend
	Returns  : The legend information
               PROCHAINEMENT
               %h_legend_info : keys = Species short name, or (0-name for the reference)
                                val = an array, with 3 cells for each chromosome:
                                                0 -> An array, with the RGB color
                                                1 -> The chromosome size
                                                2 -> The chromosome name
               MAINTENANT
               %h_legend_info : keys = Species short name, or (0-name for the reference)
                                val = an array, with 3 cells for each chromosome:
                                                0 -> An array, with the RGB color
                                                1 -> Label (org short name + chrom name + formatted size)
                                                2 -> The chromosome name
              
	Args     : none
    Globals  : none

=cut

sub GetLegendInfos_PROCHAINEMENT
{
	my $self = shift;
	my %h_alias = $self->GetColorAlias();
	my %h_legend_data;

	my $rh_orgs = $self->{__rh_orgs};
	while( my ($org, $rh_chrs) = each( %$rh_orgs ) )
	{
		my @a_org_data;
		my @a_sorted_chrs = &NARCISSE_utils::SortByChromID( keys(%$rh_chrs) );
		foreach my $chrs ( @a_sorted_chrs )
		{
			my $rh_data = $$rh_chrs{$chrs};
			next	if( $$rh_data{show} == &FALSE );

			# Drop the prefix (ref only)
			$chrs =~ s/^$__PREFIX_REF//;
            
            # Drop the species short name
            $chrs =~ s/$org-//;

			# Keep the color
			my $alias = $$rh_data{color};
			push(@a_org_data, $h_alias{$alias});

			# on choisit l'ordre de grandeur de la taille du chrs
			#my $ra_unit = ['base', 'kb', 'Mb', 'Gb'];
			#my ($x, $unit) = &GetValueInUnit( $$rh_data{length}, '%.1f', 1000, $ra_unit );
            # Keep the chromosome size
            push(@a_org_data,$$rh_data{length});

			# Keep the chromosome name
			#my $label = "$chrs ($x $unit)";
			push(@a_org_data, $chrs);
		}
		$h_legend_data{$org} = \@a_org_data;
	}

#  use Data::Dumper;
#  print STDERR Dumper [ \%h_legend_data ];
	return %h_legend_data;
}

sub GetLegendInfos
{
	my $self = shift;
	my %h_alias = $self->GetColorAlias();
	my %h_legend_data;

	my $rh_orgs = $self->{__rh_orgs};
	while( my ($org, $rh_chrs) = each( %$rh_orgs ) )
	{
		my @a_org_data;
		my @a_sorted_chrs = &NARCISSE_utils::SortByChromID( keys(%$rh_chrs) );
		foreach my $chrs ( @a_sorted_chrs )
		{
			my $rh_data = $$rh_chrs{$chrs};
			next	if( $$rh_data{show} == &FALSE );

			# on vire le prefix du chrs.
			$chrs =~ s/^$__PREFIX_REF//;

			# change "short-chr" to "long (chr)"
			my $label = &CircosChromToNarcisseLabel( $chrs, $$rh_data{ length } );

			# stocke la couleur.
			my $alias = $$rh_data{color};
			push(@a_org_data, $h_alias{$alias}, $label );
		}
		$h_legend_data{$org} = \@a_org_data;
	}
# &General::ShowDebug( \%h_legend_data );

	return %h_legend_data;
}


=head2 GetNbBasesInWholeImage

 Usage        : my $nb = $o_x->GetNbBasesInWholeImage( none )
 Function     : ad hoc method.
                For javascript interaction, i need the number of bases involved
                into the whole representation.
                So this function return the sum of all chromosomes' length.
 Returns      : $nb, the number of base involvbed into the whole representation.
 Args         : none
 Error        : none

=cut

sub GetNbBasesInWholeImage
{
	my $self = shift;

	my $nb = 0;

	my $rh_orgs = $self->{__rh_orgs};
	foreach my $rh_chrs ( values( %$rh_orgs ) )
	{
		foreach my $rh_data ( values( %$rh_chrs ) )
		{
			next	if( $$rh_data{show} == &FALSE );
			$nb += $$rh_data{ length };
		}
	}

	return $nb;
} #GetNbBasesInWholeImage



sub GetColorAlias
{
	my $self = shift;
# 	my () = @_;

	# todo : en faire une fonction statique.
	my $fe = new IO::File( $self->{_colors}, 'r' );
	if( !$fe )
	{
		throw Error::Simple();
	}

	my %h_alias;
	while( my $line = <$fe> )
	{
		$line =~ s/\s//g;
		next	if( $line =~ /^#/ );

		my ($alias, $rgb) = split('=', $line);
		next	if( !$rgb );	# ligne sans signe =

		my @a_rgb = split(',', $rgb);
		$h_alias{$alias} = \@a_rgb;
	}

	return %h_alias;
}


#renvoie x et y tel que $float = x * 10^y
sub GetOrderOfMagnitude
{
	my ($dixieme) = @_;
	my ($step, $power) = (10, 0);

	while($dixieme > $step)
	{
		$dixieme /= $step;
		++$power;
	}

	return ($dixieme, $power);
}


# usage : ($val, $unit) = GetValueInUnit($float,
# 						"%.2f", 100, ['base', 'kb', 'Mb', 'Gb'], 1000);
sub GetValueInUnit
{
	my ($val, $format, $max, $ra_units, $step) = @_;
	$format = "%d"	unless defined( $format );
	$step = 1000	unless defined( $step );
	if( !defined ($ra_units) )
	{	$ra_units = ['unit', 'kilo', 'mega', 'giga'];	}
	$max = $step	unless defined( $max );

	my $i;
	for( $i = 0; $max < $val; ++$i )
	{	$val /= $step;	}

	return sprintf( $format, $val ), $$ra_units[$i];
}


=head2 function _ComputeImgName

 Title        : _ComputeImgName
 Usage        : $name = _ComputeImgName( $rh_orgs );
 Prerequisite : use Digest::MD5
 Function     : compute a unique name for this picture and code it with MD5
 Returns      : $name, a MD5 coded name
 Args         : $rh_orgs, the hash ref passed to the constructor
              		key => { short => xx, long => xx, chrs => { x => len_x } }
 Globals      : none

=cut

sub _ComputeImgName
{
	my $self = shift;
	my ( $rh_orgs ) = @_;

	my @a_attr = qw( __level __method );
	my( $lvl, $prog ) = $self->_Get( @a_attr );

	my $name = $lvl . '_' . $prog . '_';

	foreach my $key (sort keys %$rh_orgs)
	{
		# short name
		my $short = $$rh_orgs{ $key }{ short };

		#chrs
		my $rh_chrs = $$rh_orgs{ $key }{ chrs };
		foreach my $chrs (sort keys %$rh_chrs)
		{	$name .= $short . $chrs . '_';	}
	}

	#auto_hide
	$name .= $self->{_auto_hide};
# warn Dumper $name;

	#md5
	my $md5 = Digest::MD5->new;
	$md5->add($name);

	my $name_md5 = $md5->hexdigest;

	return 'circos_' . $name_md5;
}


=head2 CircosChromToNarcisseLabel

 Usage    : my $label = $self->CircosChromToNarcisseLabel( $circos_chrom )
 Function : change "short-chr" to "long (chr)".
 Returns  : $label, string
 Args     : $circos_chrom, a circos chrom composed with short species name and chrom label.
 Error    : none

=cut

sub CircosChromToNarcisseLabel
{
# 	my $self = shift;
	my( $circos_chrom, $chrom_length ) = @_;

	my ( $short, $chrom ) = split( &NAME_SEP, $circos_chrom );

	my $o_io      = InstanceObserver::Singleton();
# 	my $o_io      = $self->_Get( '__o_io' );
	my $o_species = $o_io->GetSpeciesObserver( $short );
	my $long      = $o_species->GetLongName();

	# on choisit l'ordre de grandeur de la taille du chrs
	my $rh_options = { -format => '%1.f' };
	my $str_length = &Sequence_utils::Base2Unit( $chrom_length, $rh_options );
# 			my $ra_unit = ['base', 'kb', 'Mb', 'Gb'];
# 			my ($x, $unit) = &GetValueInUnit( $$rh_data{length}, '%.1f', 1000, $ra_unit );

	return "$long (chr $chrom, $str_length)";
} #CircosChromToNarcisseLabel


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
