#
#       emmanuel.courcelle@toulouse.inra.fr
#       Created: April 2009
#       $Id$
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Configuration;

=pod

=head1 NAME

 Configuration - A base class, used by InstanceObserver and NarcisseObserver
                 Those classes are used to encapsulate all configuration information

=head1 SYNOPSIS

 my $o_no = New NarcisseObserver( -cfg => $cfg_file );
 my $o_io = New InstanceObserver( -cfg => $cfg_file );

=head1 DESCRIPTION

 This class reads some configuration files (narcisse.cfg, default.cfg, etc.) and provides low-level as
 well as high-level methods to retrieve this information
 You should use high-level methods when you can, low-level methods when you must

=head1 AUTHOR

	Emmanuel Courcelle : emmanuel.courcelle@toulouse.inra.fr

=cut

use warnings;
use strict;

use Data::Dumper;
use File::Spec;
use File::Basename;
use Cwd;

use base qw( LipmObject );

use General;
use NARCISSE_utils;
use NARCISSE_version;
use DataReader::Whole;

use LipmError::ObjectException;
use LipmError::ParameterException;
use LipmError::IOException;
use NarcisseError;
use Bio::Narcisse::DBAdaptor;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 2490 $ =~ /\d+/g); $r[0]};
}


=head2 Function _Init

 Title      : _Init
 Usage      : my $o_inst_obs = New xxxObserver(-cfg => $cfg_file);
 Function   : constructor
 Args       : -cfg_file (optional):       the path to narcisse.cfg - See __FindNarcisseCfg for the details
              -light (optional, boolean): default &FALSE, if &TRUE only a partial initialization is performed.
              $o_logger (optional)
=cut

sub _Init
{
	my $self = shift;
	my %h_args = (	
        -cfg => undef,
        -o_logger => undef,
        -light    => &FALSE,
        @_);

    $self->SUPER::_Init( %h_args );

	my $cfg_file = $self->__FindNarcisseCfg( $h_args{ -cfg } );

	my $cfg_path = &Cwd::abs_path( $cfg_file );
    my $light    = $h_args{-light};
    $self->_Set('_light',$light);

	my $type_init = ( $light == &TRUE ) ? 'LIGHT INITIALIZATION'
										: 'COMPLETE INITIALIZATION';
	my $msg = "Generic constructor of Configuration classes - now building ";
    $self->_Trace( $msg . ref( $self ) ." $type_init\n",".","IO");

    # This will set the _rh_cfg attribute
	$self->__InitCFG( $cfg_path,$light );

# 	my $rh_cfg = $self->_Get('_rh_cfg');

	# other attributes
	$self->_Set( '_cfg_path',$cfg_path );

    # Keep the name in a private attribute
    # NOTE If the parameter INSTANCE_NAME is empty, we try to determine the instance name from the path of the cfg file
    #      However, this is NOT liable, particularly if the instance object under the narcisse directory is in fact a symbolic link
    my $instance_name = $self->GetCfg('INSTANCE_NAME');
    if ($instance_name eq '')
    {
        my @a_dirs = File::Spec->splitdir($cfg_path);
        $instance_name = $a_dirs[-4];
    }
    $self->_Set('__instance_name',$instance_name);
	return $self;
}

=head2 Procedure DumpCfg

 Title     : DumpCfg
 Usage     : $object->DumpCfg( );
 Procedure : dump the __rh_cfg to the file whose name is passed by parameter
 Args      : $DumpFile The dump file name

=cut
sub DumpCfg
{
    my $self = shift;
    
    my ($cfg_path_dump) = @_;
    my $rh_cfg        = $self->GetConfigHash();
    #local $Data::Dumper::Purity = 1;
    my $msg = Dumper($rh_cfg);
    my $serial = new IO::File(">$cfg_path_dump") or LipmError::IOException( $cfg_path_dump,'f');
    print $serial $msg . "\n";
    $serial->close();
}

=head2 Functions GetName,GetVersion,GetNarcisseVersion,GetCfgPath

 Title      : GetName,GetKingdom,GetVersion,GetNarcisseVersion,GetCfgPath
 Usage      : my $inst_name = $object->GetName( );
 Function   : accessors for the instance name or version
 Returns    : the instance name or version
 Args       : none

=cut

sub GetName            {my $self = shift; return $self->_Get( '__instance_name' );};
sub GetVersion         {my $self = shift; return $self->GetCfg( 'INSTANCE_VERSION' );};
sub GetNarcisseVersion {my $self = shift; return $self->GetCfg( 'NARCISSE_VERSION' );};
sub GetCfgPath         {my $self = shift; return $self->_Get('_cfg_path');};


=head2 Function GetDB

 Usage      :	my $fasta_file = $o_xx->GetDB( $type_db );
 Prerequiste:	I do not check if the file exists or is readable.
 Function   :	retrieve the path to the fasta/pal merge file to use as db.
 Return     :	the path to the fasta file.
 Args       :	$type_db, string. If contains 'prot', will search for AA_file, else for AN_file
 Error      :   none

=cut

sub GetDB
{
	my $self = shift;
	my ( $type_db ) = @_;

	if ($self->_Get('_light') == &TRUE)
	{
		throw LipmError::ObjectException('InstanceObserver','!attribute',
						-text => "You cannot use this function in Light mode" );
	}

	my ( $root_data, $db ) = $self->GetCfg( qw( ROOT_DATA DB ) );
	$type_db = ( $type_db and $type_db =~ /prot/i ) ? 'prot' : 'cds' ;

	my $key = 'DB_' . uc( $type_db ) . '_MERGE';
	my ( $filename ) = $self->GetCfg( $key );

	my $path = File::Spec->catfile( $root_data, $db, $filename );

	return $path;
}


=head2 Function GetCfg

 Title      : GetCfg
 Usage      : my @a_list = $object->GetCfg( @a_params );
 Function   : High level accessor to any configuration parameter
 Args       : @a_params, a list of keys (string)
 Return     : A list of values, or a single value
 Error      : LipmError::ParameterException, if a key is missing

=cut

sub GetCfg
{
	my $self = shift;
	my ( @a_list ) = @_;

	# You could comment out the following for speed, but keep it for an easier debug
	if ($self->AreDefined( @a_list ) == &FALSE)
	{
		my $msg = "One of ".join(' ',@a_list)." parameters is missing\n";
			throw LipmError::ParameterException('GetCfg key', 'missing',
				-text => $msg, -depth => 0.5 );
	}

	my $rh_cfg = $self->_Get( '_rh_cfg' );
	my @a_return = @$rh_cfg{ @a_list };

	return wantarray ? @a_return : $a_return[0];
}


=head2 Function AreDefined

 Title      : AreDefined
 Usage      : if ($o_io->IsDefined('SOME_PARAMETER'))...
 Function   : returns true if ALL parameters are defined
 Return     : A boolean
 Args       : A list of keys
 Error      : none

=cut

sub AreDefined
{
	my $self = shift;
	my ( @a_list ) = @_;

	my $rh_cfg = $self->_Get( '_rh_cfg' );
	foreach my $key (@a_list)
	{	#defined suffirait.
		return &FALSE unless ( exists $rh_cfg->{$key} and defined $rh_cfg->{$key} );
	}
	return &TRUE;
}


=head2 Function ExtractChromNameFromFilename

 Usage        : my $chrom = $o_x->ExtractChromNameFromFilename( $filepath )
 Function     : Not clear
 Returns      : $chrom
 Args         : A file path

=cut

sub ExtractChromNameFromFilename
{
	my $self = shift;
	my( $filepath ) = @_;

	my $filename = fileparse( $filepath );
	my $mask = $self->GetCfg( qw( MASK_FILE_CHROM ) );

	$filename =~ /$mask(\w+)\./;
	my $chrom = $1;

	return $chrom;
} #ExtractChromNameFromFilename


=head2 Function GetUnixPath

 Title      : GetUnixPath
 Usage      : my @a_list = $object->GetUnixPath( @a_params );
 Function   : High level accessor to any unix path
 Args       : A list of keys
 Return     : a list of values.
 Error      : none

=cut

sub GetUnixPath
{
	my $self = shift;
	my ( @a_list ) = @_;

    # TODO - mettre les commandes unix dans un hash a part
	my $rh_unix = $self->_Get( '_rh_cfg' );
	my @a_return = @$rh_unix{ @a_list };

	return wantarray ? @a_return : $a_return[0];
}


=head2 Function GetConfigHash

 Title      : GetConfigHash
 Usage      : my $rh_cfg     = $object->GetConfigHash()
 Function   : Low level accessor to the configuration hash
 Return     : a ref to the configuration hash
 Args       : none
 NOTE       : Prefer, if possible, the high-level accessor routines

=cut

sub GetConfigHash  {my $self = shift; return $self->_Get( '_rh_cfg' );};


=head2 Function __FindNarcisseCfg

 Title      : __FindNarcisseCfg
 Usage      : my $f = $self->__FindNarcisseCfg()
 Function   : Search narcisse.cfg:
                 - If the file is specifed, check the file exists and return the complete path
                 - Elsif the env variable NARCISSE_CFG is specified, its value points to narcisse.cfg
                 - Else  walk through the hierarchy to find the file
 Return     : The complete path 
 Args       : $file_cfg (optional): the path to narcisse.cfg
 Globals    : %ENV
 Error      : Error      : throw ParameterException if the cfg file cannot be found

=cut

sub __FindNarcisseCfg
{
	my $self = shift;

	my ($file_cfg) = @_;

	if (!defined $file_cfg)
	{
		$file_cfg = '';
		# if the env variable is defined, we use this variable
		if ( defined($ENV{NARCISSE_CFG}) )
		{
			$file_cfg = $ENV{NARCISSE_CFG};
		}

		# The env variable is not defined in the cgi-bin scripts - Try to find narcisse.cfg, anyway
		else
		{
			$file_cfg = &__Walk( $0 );

			# If not found, find the file from the current directory
			$file_cfg = &__Walk( cwd )	if(! -e $file_cfg )
		}
	}

	my $cfg_path = &Cwd::abs_path( $file_cfg );
	#&Carp::carp("#$file_cfg#$cfg_path#");

	# did we find the file ?
	if ( ! -f $cfg_path or ! -r $cfg_path or -z $cfg_path )
	{
		my $msg = "cfg file => $cfg_path ";
		throw LipmError::ParameterException('cfg file', 'missing', -text => $msg );
	}

	return $cfg_path;

    # Internal function - Walk through the hierarchy, looking for narcisse.cfg    
    sub __Walk
    {
        my ($strtdir) = @_;
        my $mask_file_cfg = "Site/cfg/narcisse.cfg";
        my $file_cfg = "$strtdir/$mask_file_cfg";
        while ( ! -e $file_cfg )
        {
            last if ($strtdir eq '/' or $strtdir eq '.');
            $strtdir = dirname($strtdir);
            $file_cfg = "$strtdir/$mask_file_cfg";
            #warn "KOUKOU $strtdir $file_cfg";
            #warn "SUPER" if (-r $file_cfg);
        }
        return $file_cfg;
    }
         
}


=head2 Procedure __InitCFG

 Title    : __InitCFG
 Usage    : my $rh_cfg = $self->__InitCFG($file_cfg)
 Procedure: Completely init the rh_cfg hash from the text parameter files
            Called from the constructor
 Args     : $file_cfg = The bootstrap file
            $light    = Light initialization if &TRUE

=cut

sub __InitCFG
{
    my $self = shift;

    my ( $file_cfg, $light ) = @_;
	
    # Read narcisse.cfg
	my %h_cfg;
    $self->_Set('_rh_cfg',\%h_cfg);
	$self->__ReadFileCfg( $file_cfg,\%h_cfg );

    if ($light == &FALSE )
    {
        
        my $dir = File::Basename::dirname( $file_cfg );

        # read defaults.cfg
        my $defaults_cfg = File::Spec->catfile( $dir, 'defaults.cfg' );
        $self->__ReadFileCfgWithSubst( $defaults_cfg,\%h_cfg );
        
        # read insdef.cfg (local changes from defaults.cfg, overriding the values read from defaults.cfg)
        my $insdef_cfg = File::Spec->catfile( $dir, 'insdef.cfg' );
        $self->__ReadFileCfgWithSubst( $insdef_cfg,\%h_cfg ) if ( -r $insdef_cfg );
        
        if ($h_cfg{ NARCISSE_VERSION } ne $NARCISSE_version::NARCISSE_VERSION)
        {
            my $msg = "ERROR: Bad NARCISSE version (config $h_cfg{ NARCISSE_VERSION }, program $NARCISSE_version::NARCISSE_VERSION)";
            $msg .= "<br />";
            $msg .= "The Narcisse administrator should do:<br />";
            $msg .= "<pre>chins;\nnc_configure.pl --upgrade_instance $h_cfg{ INSTANCE_NAME }</pre>";
            throw NarcisseError( -bd => $msg, -cv => $h_cfg{ NARCISSE_VERSION }, -pv => $NARCISSE_version::NARCISSE_VERSION ); 
        }

        # Check ROOT, HTTP, ROOT_DATA
        
        if ($h_cfg{ROOT} eq "" or $h_cfg{HTTP} eq ""
            or $h_cfg{ROOT_DATA} eq "" or $h_cfg{ROOT_DATA} eq "NOT_USED"
            or $h_cfg{NARCISSE_ADM_EMAIL} eq "")
        {
            throw LipmError::ParameterException('narcisse.cfg', 'missing', -text => "Pb in narcisse.cfg - Did you run nc_configure.pl ?" ) 
        };
        
        my $o_ndr = New DataReader::Whole();

        if (-r $h_cfg{ ROOT_ZOOM_LEVEL_CFG } )
        {
            $o_ndr->Prepare( -from => $h_cfg{ROOT_ZOOM_LEVEL_CFG} );
            $h_cfg{ ZOOM_LEVEL } = $o_ndr->FetchallHashref( 'level', -line_type => 'HASH' );
        }
        
        # this file is required !
        $self->__ReadFileCfg( $h_cfg{ROOT_UNIX_CMDS_CFG},\%h_cfg) ;
        
        if (-r $h_cfg{ROOT_GUIDEF_CFG})
        {
            my %g;
            $self->__ReadFileCfg( $h_cfg{ROOT_GUIDEF_CFG}, \%g );
            $h_cfg{GUI_DEFAULTS} = \%g;
        }
        
        # If we are using a database for storing the files, we shall extract the files to the cache when necessary
        # ROOT_DATA is in this case redefined to point to the cache directory
        #if ($h_cfg{DATA_IN_DATABASE}==1)
        #{
        #    $h_cfg{ROOT_DATA} = "$h_cfg{ROOT_CACHE}/$h_cfg{DATA_CACHE}/";
        #}
    }
    
    return \%h_cfg;
}


=head2 procedure __ReadFileCfg

 Title        : __ReadFileCfg
 Usage        : __ReadFileCfg($file, \%h_var)
 Access       : Private
 Fonction     : Init the \%h_var hash from $file
                blank lines or lines starting with # or lines without = sign  are ignored
                The lines should be in the form key=value (NO SPACE AROUND THE = !)
 Args         : $file  : The file to read
                \%h_var: The hash to initialize
=cut

sub __ReadFileCfg
{
    my $self   = shift;
    my ($file, $rh_cfg) = @_;
    
	my $o_param = New ParamParser( $file );
	$o_param->Dump( 'HASH', $rh_cfg );

	return;
}


=head2 procedure __ReadFileCfgWithSubst

 Title        : __ReadFileCfgWithSubst
 Usage        : __ReadFileCfgWithSubst($file,\%h_var)
 Access       : Private
 procedure    : Init the \%h_var hash passed by parameter, from $file
                blank lines or lines starting with # or lines without = sign  are ignored
                The lines should be in the form key=value
                the Keys should be written in capitals (if not, the line is ignored)
                The file should begin with the keys:
                    ROOT=
                    HTTP=
                Then, each key/value is modified as follows:
                    1/ Keys  R_XXX: a new key ROOT_XXX is created, ROOT is prepended to the value, 
                                    UNLESS THE VALUES STARTS WITH '/': in the latter case, the value is kept unchanged
                    2/ Keys  H_XXX: a new key HTTP_XXX is created, HTTP is prepended to the value,
                                    UNLESS THE VALUE STARTS WITH http://: in the latter case, the value is kept unchanged
                    3/ Keys RH_XXX: 1/ and 2/ are applied
                    4/ Key=Val R_XXX=R_YYY/something: R_YYY is substituted with ROOT_YYY (should be already defined)
                    5/ Key=Val H_XXX=H_YYY/something: H_YYY is substituted with HTTP_YYY (should be already defined)
                    6/ Key=Val RH_XXX=RH_YYY/something 4/ and 5/ are applied
                Last, the R_XXX and H_XXX keys are removed
 Args         : $file  : The file to read
                \%h_var: The hash to initialize
 Globals      : none

=cut

sub __ReadFileCfgWithSubst
{
    my $self   = shift;
    my ($file,$rh_cfg) = @_;

    my $fd_file = new IO::File($file) or throw LipmError::IOException( "-$file-", 'f' );
    my $root = $$rh_cfg{ ROOT };
    my $http = $$rh_cfg{ HTTP };
  	while ( my $line = <$fd_file> )
	{
		if ( $line !~ /^\s*$/ && $line !~ /^#/ && $line =~ /=/)
		{
			my ($key,$value) = split (/=/, $line, 2);
			chomp($value);
            
            $rh_cfg->{$key} = $value;
            
#             $root = $value if ($key eq 'ROOT');
#             $http = $value if ($key eq 'HTTP');
            if ( $key =~ /^RH?J?_([\w-]+)/ )
            {
                if ( $root eq "" )
                {
                    throw LipmError::ParameterException("$key=$value", 'missing', -text => " Cannot create ROOT_ key because ROOT is not read ");
                }
                my $nkey = 'ROOT_'.$1;
                my $nval;
                if ( $value =~ /^RH?J?_([\w-]+)\/(.+)/ )
                {
                    my $nkv = 'ROOT_'.$1;
                    if ( !exists $rh_cfg->{$nkv} )
                    {
                        throw LipmError::ParameterException("$key=$value", 'missing', -text => " Cannot create ROOT_ value because $nkv is unknown ");
                    }
                    $nval = $rh_cfg->{$nkv}.'/'.$2;
                }
                elsif ( $value =~ /^\// )
                {
                    $nval = $value;
                }
                else
                {
                    #$nval = ($value eq '') ? $root : $root . '/' . $value;
                    $nval = $root . '/' . $value;
                }
                
                $rh_cfg->{$nkey} = $nval;
            };
            if ( $key =~ /^R?HJ?_([\w-]+)/ )
            {
                if ( $http eq "" )
                {
                    throw LipmError::ParameterException("$key=$value", 'missing', -text => " Cannot create HTTP_ key because HTTP is not read");
                }
                my $nkey = 'HTTP_'.$1;
                my $nval;
                if ( $value =~ /^R?HJ?_([\w-]+)\/(.+)/ )
                {
                    my $nkv = 'HTTP_'.$1;
                    if ( !exists $rh_cfg->{$nkv} )
                    {
                        throw LipmError::ParameterException("$key=$value", 'missing', -text => " Cannot create HTTP_ key because $nkv is unknown");
                    }
                    $nval = $rh_cfg->{$nkv}.'/'.$2;
                }
                elsif ( $value =~ /^http:\/\// )
                {
                    $nval = $value;
                }
                else
                {
                    #$nval = ($value eq '') ? $http : $http . '/' . $value;
                    $nval = $http . '/' . $value;
                }

                $rh_cfg->{$nkey} = $nval;
            }
		}
	}

    # Remove R_, H_, RH_ keys, as they are useless now
    foreach my $k (keys %$rh_cfg)
    {
        delete ($rh_cfg->{$k}) if ($k=~/^RH?_/ or $k=~/^R?H_/);
    }

    close($fd_file);
}


1;
