#
#       Sebastien.Letort@toulouse.inra.fr
#       Created: November 19, 2007
#       $Id: InstanceObserver.pm 2490 2009-04-22 14:57:49Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package InstanceObserver;

=pod

=head1 NAME

 InstanceObserver - This class maintains the configuration information for a specific Narcisse instance
                    Several accessors are defined here: high level accessors and low level accessors.
                    The rule is: Use high level accessors when you can (nearly always), 
                    low level accessors when there is no choice (nearly never)
                    
                    You can build InstanceObserver is 'light' mode, this is ONLY USED by nc_configure.pl, to write the narcisse.cfg configuration file

=head1 SYNOPSIS

 my $o_io = New InstanceObserver( -cfg => $cfg_file );
 OR
 my $o_io = InstanceObserver::Singleton();
 
 The second form builds a UNIQUE object through the whole application (pattern Singleton)

 Does this instance have this species:
    my $boolean = $o_inst_obs->HaveSpecies( 'aspergillus' );
    
 General information:
    my $version = $o_io->GetVersion();
    see also GetName, GetNarcisseVersion

 High level accessors:
    SetSessiondirectory, GetSessionDirectory: useful to set the session directory, and to compute the *REAL* values of keys,
    which are defined using the pattern D_SESSDIR. cf. Configuration::__ReadFileCfgWithSubst for the details, cf. also default.cfg
 
 Working with species:
    HaveSpecies         : returns true if the species is in this instance (may be regex in long name)
    ExtractMySpecies    : for each name in input, returns the short name
                          skip the name if not found (never returns none)
                          Use HaveSpecies
                          Return all short names if nothing specified
    GetSpeciesShortName : for each name in input, returns the short name, or none if not found
                          Return all short names if nothing specified
    GetSpeciesLongName  : for each name in input, returns the long name, or none if not found
                          Return all long names if nothing specified
 
    GetSpeciesObserver  : From a long or short name, return the corresponding species observer
                          Throw an exception if not found
                          
 Low Level accessors - DO NOT USE unless required to do so
    GetConfigHash, GetGuiDefHash: access to the internal data structures
    GetSpeciesInfo      : for each short name specified, return the hash corresponding 
                          Return '' for an unknown name (no error)

=head1 DESCRIPTION

	This class gives access to information of a Narcisse

=head1 TODO
	faire une classe SpeciesObserver (?) avec les méthodes DBAdaptor,
	GetKaryotype, ...


=head1 AUTHOR

	Sebastien Letort : sebastien.letort@toulouse.inra.fr

=cut

use warnings;
use strict;

use Data::Dumper;
use File::Spec;
use File::Basename;
use File::Path;
use Cwd;
use Carp;

use base qw( Configuration );

use General;
use NARCISSE_utils;
use SpeciesObserver;
use Bio::Narcisse::DBAdaptor;
use DataReader::Whole;

use constant NONE => 'none';

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 2490 $ =~ /\d+/g); $r[0]};
}

#	----	----	----	----	----	----	----	----	----

=head2 Function _Init

 Title      : _Init
 Usage      : my $o_inst_obs = New InstanceObserver(-cfg => $cfg_file);
 Function   : constructor.
 Args       : $cfg_file (optional): the path to narcisse.cfg - See __FindNarcisseCfg for the details
              -light (optional):    See Configuration 
              $o_logger (optional)
 Attributes :
     _rh_cfg,
     __rh_genomes,
     __rh_genomes_by_long
     __nb_species, integer, number of species managed by this object.
=cut

sub _Init
{
	my $self = shift;
	my %h_args = (	
        -cfg      => undef,
        -light    => &FALSE,
        -o_logger => undef,
        @_);

	$self->SUPER::_Init( %h_args );

	if ($self->GetName() eq 'multi')
	{
		my $msg = "ERROR - Please run chins some_instance_other_than_multi\n";
		throw LipmError::ParameterException('Name', 'value', -text => $msg )
	}

	my $rh_cfg    = $self->_Get('_rh_cfg');
    
    #Unless in $light mode, we read the species.list file inorder to know the species list
    my $light = $h_args{-light};
    if ($light == &FALSE)
    {
        my $species_list = $self->GetSpeciesListFilepath('INSTANCE');
        if ( -r $species_list )
        {
			my $o_ndr = New DataReader::Whole();
            $o_ndr->Prepare(
                -from => $species_list,
                -force_reading => &TRUE
                );
            $rh_cfg -> { GENOME } = $o_ndr->FetchallHashref( 'short', -line_type => 'HASH' );
        }
    }        

	my $rh_genomes = $rh_cfg -> { GENOME };

	# other attributes
	$self->_Set( __rh_genomes => $rh_genomes,
				 __rh_genomes_by_long => undef,
				 __InitedURL          => &FALSE,
                 __sessdir            => 'D_SESSDIR',
                 __light              => $light
			);

	return $self;
}


=head2 Singleton

 Usage    : $o_x->Singleton( %h_params )
 Function : Constructor using singleton pattern.
 Args     :
     -cfg      => $cfg, string, the path to the cfg file
     -light    => $bool, boolean, ?
     -o_logger => $o_logger, Logger object
 Error    : none.
 Note     : Perl 5.10 introduce a state variable
            http://fr.wikipedia.org/wiki/Singleton_(patron_de_conception)#Perl
 
=cut

my $SINGLETON;
sub Singleton
{
	if( !defined $SINGLETON )
	{
		$SINGLETON = New InstanceObserver( @_ );
	}

	return $SINGLETON;
} #Singleton


=head2 Procedure InitURL

 Title     : InitURL
 Usage     : $o_io->IniURL();
 Procedure : Initialize the field 'URL' in each species hash (cf. $self->{__rh_cfg}->{GENOME})
             If already done, don't redo it
 Args      : none

=cut

sub InitURL
{
    my $self = shift;
    if ($self->_Get('__InitedURL')==&FALSE)
    {
        $self->__AddToCfgGenome( '', 'URL' );
        $self->__AddToCfgGenome( '', 'SYNT_URL' );
        foreach my $species ($self->GetSpeciesShortName())
        {
            $self->__AddToCfgGenome( $species, 'URL' );
            $self->__AddToCfgGenome( $species, 'SYNT_URL' );
        }
        $self->_Set('__InitedURL',&TRUE);
    }
	return;
}


=head2 UpdateGenomesListWithSessionData

 Usage        : $o_x->UpdateGenomesListWithSessionData()
 Function     : update __rh_genomes with genomes uploaded by the user and stored in sessdir.
 Args         : none
 Error        : none

=cut
sub UpdateGenomesListWithSessionData
{
	my $self   = shift;

	my $o_ndr = New DataReader::Whole();

	my $species_file = $self->GetSpeciesListFilepath( 'SESSION' );

	my $rh_session_genomes;
	if( -r $species_file )
	{
		$o_ndr->Prepare(
			-from => $species_file,
			-force_reading => &TRUE
			);
		$rh_session_genomes = $o_ndr->FetchallHashref( 'short', -line_type => 'HASH' );
	}

	my $rh_genomes = $self->_Get( '__rh_genomes' );

	my @a_keys = keys %$rh_session_genomes;
	@$rh_genomes{ @a_keys } = @$rh_session_genomes{ @a_keys };

	$self->_Set( __rh_genomes => $rh_genomes );
    #warn Dumper $rh_genomes;
	return;
} #UpdateGenomesListWithSessionData

=head2 UpdateGenomesListWithoutSessionData

 Usage        : $o_x->UpdateGenomesListWithSessionData()
 Function     : update __rh_genomes removing genomes living in the session directory
 Args         : none
 Error        : none

=cut
sub UpdateGenomesListWithoutSessionData
{
	my $self   = shift;

    my $rh_genomes = $self->_Get( '__rh_genomes' );
    foreach my $key (keys %$rh_genomes)
    {
        my $rh_genome = $rh_genomes->{$key};
        delete $rh_genomes->{$key} if ($rh_genome->{ where } eq 'SESSION');
    }
    return;
} #UpdateGenomesListWithoutSessionData

=head2 Procedure __AddToCfgGenome

 Title        : __AddToCfgGenome
 Usage        : $self->__AddToCfgGenome( $species, $key );
 Procedure    : If $species is '', read the data/url.txt file and copy it to __rh_cfg->URL
                Else, read the data/$species/url.txt file, and copy it to __rh_cfg->$species->URL 
                NOTE - The string 'URL' is passed by the $key parameter, it can be URL or SYNT_URL
 Args         : $species, $key (URL, SYNT_URL)

=cut

sub __AddToCfgGenome
{
    my $self = shift;
    
    my ( $species, $key ) = @_;
    my $url_file;
    my $rh_cfg       = $self->_Get('_rh_cfg');
    my $rh_config;
    
    if ($species eq '')
    {
        $url_file  = $self->GetCfg('ROOT_' . $key . '_TXT');
        $rh_config = $rh_cfg;  
    }
    else
    {
        my $o_so = $self->GetSpeciesObserver( $species );
        
        my $species_info = $self->GetSpeciesInfo( $species );
        my $species_path = $o_so->GetDataPath( 'INSTANCE' );
        $url_file        = File::Spec->catfile($species_path, $self->GetCfg($key.'_TXT'));
        $rh_config       = $species_info;
    }
    
    if( -e $url_file )
    {
        my @a_lines = ();
        
        # no ParamParser cause i want a list !
        # 			&_ReadFileCfg( $url_file, $rh_tmp );
        my $fh_in = new IO::File( $url_file, 'r' )
        or throw LipmError::IOException( "*$url_file*", 'f' );
        while( my $line = <$fh_in> )
        {
            next	if( $line =~ /^#/ );
            next	if( $line =~ /^\s*$/ );
            my ( $text, $tmp ) = split( '=', $line, 2 );
            my ( $url, $key )  = split( ' ', $tmp );
            
            my %h_tmp = ();
            @h_tmp{ qw( text url key ) } = ( $text, $url, $key );
            
            push( @a_lines, \%h_tmp );
        }
        
        $rh_config->{ $key } = \@a_lines;
    }
    return;
}

=head2 procedure SetSessionDirectory, function GetSessionDirectory

 Usage        : $o_inst_obs->SetSessionDirectory('WBxxxxx');
                $o_inst_obs->SetSessionDirectory('/tmp/toto');
                my $sessdir = $o_inst_obs->GetSessionDirectory();
 Procedure    : 2 cases :
     1- arg is a directory ( does not start with '/' )
         __sessdir is set and all D_SESSDIR parameters are substituted in
         __rh_cfg with the new sessdir.
     2- arg is an absolute path ( does start with '/' )
         __sessdir is set and that's it.
 Args         : $sessdir or sess_path

=cut

sub SetSessionDirectory
{
	my $self = shift;
	my ($sessdir) = @_;

	if ($self->_Get('_light') == &TRUE)
	{
		throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" )
	}

    # /some/absolute/path/to/sessdir ==> sessdir
	if( $sessdir =~ /^\// )
	{
		my @a_dirs = File::Spec->splitdir( $sessdir );
		$sessdir = $a_dirs[-1];
	}
    
    # compare the current sessdir with the new one - If equal, there is nothing to do
    # TODO - Revoir cette optimisation merdouilleuse
    my $old_sessdir = $self->_Get('__sessdir');
    #if ( $old_sessdir ne $sessdir )
    #{
        # Remove the genomes living inside the session directory
        $self->UpdateGenomesListWithoutSessionData();

        # Substitute the OLD sessdir for the NEW sessdir everywhere
        if ( $old_sessdir ne $sessdir )
        {
	        my $rh_cfg = $self->_Get('_rh_cfg');
	        foreach my $key (keys %$rh_cfg)
	        {
	            $rh_cfg->{$key} =~ s/$old_sessdir/$sessdir/g;
	        }
	        $self->_Set( __sessdir => $sessdir );
		}

        # MaJ rh_genomes with session data.
        $self->UpdateGenomesListWithSessionData();
    #}
    return;
}

sub GetSessionDirectory 
{
    my $self = shift; 
    if ($self->_Get('_light') == &TRUE)
    {
        throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" ) 
    }
    
    # If __sessdir is still at its initial value, there is NO SESSION: return ''
    return ($self->_Get('__sessdir') ne 'D_SESSDIR') ? $self->_Get('__sessdir') : '';
};

sub GetRootSessDir
{
    my $self = shift; return $self->_Get('_rh_cfg')->{ ROOT_SESSDIR };
};

sub GetUrlSessDir
{
    my $self = shift; return $self->_Get('_rh_cfg')->{ HTTP_SESSDIR };
};

=head2 Functions GetKingdom

 Title      : GetKingdom
 Usage      : my $inst_name = $object->GetKindom( );
 Function   : accessors for the instance kingdom
 Returns    : the instance name or version
 Args       : none

=cut

sub GetKingdom
{
    my $self    = shift;
    
    my $kingdom = $self->GetCfg( 'KINGDOM' );
	if ( $kingdom ne "bacteria"  && $kingdom ne "eukaryote" )
	{
        throw LipmError::ParameterException( 'KINGDOM', 'value', -text=> "KINGDOM must be bacteria or eukaryote in narcisse.cfg" );
	}
    return $kingdom;
}

=head2 Function GetGuidef

 Title      : GetGuidef
 Usage      : my @a_list = $object->GetGuiDef( @a_params );
 Function   : Accessor to any gui default 
 Return     : a list of values.
 Args       : A list of keys

=cut

sub GetGuidef
{
	my $self = shift;
	my ( @a_list ) = @_;

	my $rh_guidef = $self->_Get( '_rh_cfg' )->{ GUI_DEFAULTS };
	my @a_return = @$rh_guidef{ @a_list };

	return wantarray ? $a_return[0] : @a_return;
}

=head2 Function GetGuiDefHash

 Title      : GetGuiDefHash()
 Usage      : my $rh_guidef  = $object->GetGuiDefHash()
 Function   : return the guidef hash
 Return     : a ref to the guidef hash
 Args       : none
 Error      : none
 NOTE       : Prefer, if possible, the high-level accessor routines

=cut

sub GetGuiDefHash  {my $self = shift; return $self->_Get( '_rh_cfg' )->{ GUI_DEFAULTS };};

=head2 Function GetNbSpecies

 Usage        : my $nb = $o_x->GetNbSpecies( none )
 Function     : return the nb of species (keys) of __rh_genomes.
 Returns      : $nb, integer
 Args         : none
 Error        : none

=cut

sub GetNbSpecies
{
	my $self = shift;

	my $rh_genomes = $self->_Get( '__rh_genomes' );

	return scalar( keys %$rh_genomes );
} #GetNbSpecies


=head2 Function ExtractMySpecies

 Usage      : my $ra_spe = $object->ExtractMySpecies( $ra_species );
 Prerequiste: none
 Function   : extract the short species names from $ra_species, which are in the instance
              Return all species if $ra_species is empty.
 Returns    : $ra_spe, ref list of short species name.
 Args       : $ra_species, a ref list of any string. (short/long species names)

=cut

sub ExtractMySpecies
{
	my $self = shift;
    if ($self->_Get('_light') == &TRUE)
    {
        throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" ) 
    }

	my ( $ra_species ) = @_;

	my @a_species = ();

	my $rh_genomes = $self->_Get( '__rh_genomes' );
	if( 0 == scalar( @$ra_species ) )
	{
		@a_species = keys %$rh_genomes;
	}
	else
	{
		foreach my $species ( @$ra_species )
		{
			# une liste avec des undef dedans, ca peut arriver !
			next if( !defined $species );
			if( $self->HaveSpecies( $species ) )
			{
				my $short = $self->GetSpeciesShortName( $species );
				push( @a_species, $short );
			}
		}
	}

	return \@a_species;
}


=head2 Function GetDB

 Usage      :	my $fasta_file = $o_xx->GetDB( $type_db, $species );
 Prerequiste:	I do not check if the file exist or is readable.
 Function   :	retrieve the path to the fasta file to use as db.
             	depending if we use a merged one or not, regarding the species.
 Return     :	the path to the fasta file.
 Args       :	$type_db, string. If contains 'prot', will search for AA_file, else for AN_file
             	$species, short name of the wanted species, optional.
                          If not specified, returns the merged fasta files
 Error      :    none

 NOTE       : fonction en sursis, a revoir avec les services Blast et PsiBlast.

=cut

sub GetDB
{
	my $self = shift;
	my ( $type_db, $species ) = @_;

	if ($self->_Get('_light') == &TRUE)
	{
		throw LipmError::ObjectException('InstanceObserver','!attribute',
						-text => "You cannot use this function in Light mode" );
	}

	my $path;
	if( IsNotNull( $species ) and $species ne 'none' )	# and HaveSpecies ?
	{
		my ( $root_data, $db ) = $self->GetCfg( qw( ROOT_DATA DB ) );
		$type_db = ( $type_db and $type_db =~ /prot/i ) ? 'prot' : 'cds' ;

		my @a_path_elts = ( $root_data, $species, $db, $type_db );
		my $fasta_file  = $species . '_' . $type_db . '.fasta';
		$path = File::Spec->catfile( @a_path_elts, $fasta_file );
	}
	else
	{
		$path = $self->SUPER::GetDB( $type_db );
	}

	return $path;
}

=head2 Function IsSpeciesShortName
 Usage    : if ($object->IsSpeciesShortName( $name )) {print "$name is a short name";};
 Function : returns TRUE if the name is a short name, FALSE if it is not
 Return   : A boolean
 
=cut

sub IsSpeciesShortName
{
	my $self = shift;
	my ( $name ) = @_;
    
    my $rh_genomes = $self->_Get('__rh_genomes');
    return defined $rh_genomes->{$name};
}

=head2 Function IsSpeciesLongName
 Usage    : if ($object->IsSpeciesLongName( $name )) {print "$name is a long name";};
 Function : returns TRUE if the name is a long name, FALSE if it is not
            NOTE - If IsSpeciesLongName returns &FALSE, may be GetSpeciesLongName or GetSepeciesShortName will return something
                   This is because those functions try to select the long name narrowest to the parameter
                   A contrario, IsSpeciesLongName performs an exact match
 Return   : A boolean
 
=cut

sub IsSpeciesLongName
{
	my $self = shift;
	my ( $name ) = @_;
    
    $self->__InitGenomesByLongIfNecessary();
	my $rh_genomes_by_long = $self->_Get('__rh_genomes_by_long');

    my $long_key = $name;
    $long_key    =~ s/[ _]//g;
    $long_key = uc $long_key;
    
    return defined $rh_genomes_by_long->{$long_key};
}
    
=head2 function HaveSpecies

 Usage        : my $bool = $object->HaveSpecies( $species, $regexp_flag );
 Prerequisite : none
 Function     : check if the species belong to the instance
                you can search for long or short name
                Return &TRUE if ONE species belongs to the instance
 Returns      : &TRUE or &FALSE 
 Args         : @species, string, list of short or long species name

=cut

sub HaveSpecies
{
	my $self = shift;
	my( @a_species ) = @_;
    
    foreach my $species (@a_species)
    {
        return &TRUE if ( $self->GetSpeciesShortName( $species ) ne &NONE );
    }
    return &FALSE;
}

# sub HasSpecies
# {
	# # changement de nom en cours.
	# &HaveSpecies( @_ );
# }

# 
	# if ($self->_Get('_light') == &TRUE)
	# {
		# throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" )
	# }
# 
# 
	# $regexp_flag = &FALSE unless defined $regexp_flag;
# 
	# my $rh_genomes = $self->_Get( '__rh_genomes' );
# 
	# if( &FALSE == $regexp_flag )
	# {
		# # short name are also keys
		# return &TRUE  if( defined $$rh_genomes{ $species } );
# 
		# foreach my $rh_genome ( values %$rh_genomes )
		# {
			# return &TRUE  if( $$rh_genome{ long } eq $species );
		# }
	# }
	# else
	# {
		# foreach my $rh_genome ( values %$rh_genomes )
		# {
			# return &TRUE  if( $$rh_genome{ long } =~ /$species/ );
		# }
	# }
# 
	# return &FALSE;
# }
# sub HaveSpecies
# {
	# my $self = shift;
	# my( $species, $regexp_flag ) = @_;
# 
	# if ($self->_Get('_light') == &TRUE)
	# {
		# throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" )
	# }
# 
# 
	# $regexp_flag = &FALSE unless defined $regexp_flag;
# 
	# my $rh_genomes = $self->_Get( '__rh_genomes' );
# 
	# if( &FALSE == $regexp_flag )
	# {
		# # short name are also keys
		# return &TRUE  if( defined $$rh_genomes{ $species } );
# 
		# foreach my $rh_genome ( values %$rh_genomes )
		# {
			# return &TRUE  if( $$rh_genome{ long } eq $species );
		# }
	# }
	# else
	# {
		# foreach my $rh_genome ( values %$rh_genomes )
		# {
			# return &TRUE  if( $$rh_genome{ long } =~ /$species/ );
		# }
	# }
# 
	# return &FALSE;
# }
    
=head2 Function GetSpeciesShortName

 Usage    : my $short_name    = $object->GetSpeciesShortName( $name );
            my @a_short_names = $object->GetSpeciesShortName( @a_names );
            my @a_short_names = $object->GetSpeciesShortName();
 Function : retrieve the short name(s), from short or long name(s)
            If no argument or if 'all' is the ONLY name provided, retrieve ALL short names 
 Return   : 1/ a string, the short name of the species (may be &NONE)
            2/ an array, the short names of the species
            3/ an array, the short names of all the species
 Args     : undef, $name, @a_names 0,1 or several names
 Error    : a ParameterException is thrown if the number of arguments does not fit with the context

=cut

sub GetSpeciesShortName
{
	my $self = shift;
	my ( @a_names ) = @_;

    #pop(@a_names) if (@a_names==1 && $a_names[0] eq 'all');
    # PATCH CRADO TEMPORAIRE - POUR ncs_Blast.pl
    return &NONE if (@a_names==1 && (!defined $a_names[0] || $a_names[0] eq 'all') );
    
	if ($self->_Get('_light') == &TRUE)
	{
		throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" )
	}

	$self->__InitGenomesByLongIfNecessary();

	# We may use this function in scalar context only if 1 name specified
	throw LipmError::ParameterException('a_names', 'type', -text => "Only 1 name must be specified in scalar context " )
	if (!wantarray and @a_names!=1);

	my $rh_genomes         = $self->_Get('__rh_genomes');
	my $rh_genomes_by_long = $self->_Get('__rh_genomes_by_long');


	# No parameter selected = return all species short names
	if (@a_names == 0)
	{
		return keys %$rh_genomes;
	}

	# One or several parameters selected = convert names and return the result
	my @a_short;
	foreach my $name (@a_names)
	{
		# We may have a list of undef as parameter
		next unless (defined $name);

		my $long_key = $name;
		$long_key    =~ s/[ _]//g;
        $long_key = uc $long_key;

		# if it is a short name
		if ($self->IsSpeciesShortName($name))
		{
			push(@a_short,$name);
		}

		# if it is a long name: try to replace
		elsif (defined $rh_genomes_by_long->{$long_key})
		{
			push(@a_short,$rh_genomes_by_long->{$long_key});
		}

		# if it is nothing, we still try to find something
		else
		{
			push(@a_short,$self->__FindNearestSpecies($long_key));
		}
	}

	# return a string or an array
	return wantarray ? @a_short : $a_short[0];
}

=head2 Function _FindNearestSpecies
 Usage    : my $short_name    = $self->__FindNearestSpecies( $name );
 Prerequisites : $name is a UPPER CASE string, without any blank or '_'
 Function : Loop through all "long keys" (ie uppercased/space deleted long names) and
            return the FIRST key matching with $name
            Return &NONE is nothing is matching
 Return   : 1/ a short name, &NONE is nothing found
 Args     : $name a fuzzy long name

=cut
sub __FindNearestSpecies
{
    my $self = shift;
    
    my ($name) = @_;
    my $rh_genomes_by_long = $self->_Get('__rh_genomes_by_long');
    foreach my $long_key (sort keys %$rh_genomes_by_long)
    {
        return $rh_genomes_by_long->{$long_key} if $long_key =~ /$name/;
    }
    return &NONE;
}

=head2 Function GetSpeciesShortNamesForUpgrade

 Usage    : my @a_short_names = $object->GetSpeciesShortNamesForUpgrade();
 Function : retrieve the short names, excluding the species marked with upgrade == -1
            If there is NO upgrade field, call GetspeciesShortNames
            This is useful for nc_configure.pl/nc_install.pl
 Return   : an array
 Args     : None

=cut
sub GetSpeciesShortNamesForUpgrade
{
	my $self = shift;

	if ($self->_Get('_light') == &TRUE)
	{
		throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" )
	}

	my $rh_genomes         = $self->_Get('__rh_genomes');
    my @a_short_names;
    foreach my $short ( keys %$rh_genomes )
    {
        return $self->GetSpeciesShortName() unless (exists $rh_genomes->{ $short }->{ upgrade });
        if ( $rh_genomes->{ $short }->{ upgrade } >= 0 )
        {
            push (@a_short_names,$short);
        }
    }
    
    return @a_short_names;
}

=head2 Function GetSpeciesLongName

 Usage    : my $long_name    = $object->GetSpeciesLongName( $name );
            my @a_long_names = $object->GetSpeciesLongName( @a_names );
            my @a_long_names = $object->GetSpeciesLongName();
 Function : retrieve the long name(s), from short or long name(s)
            If no argument or if 'all' is the ONLY name provided, retrieve ALL short names
 Return   : 1/ a string, the long name of the species (may be &NONE)
            2/ an array, the long names of the species
            3/ an array, the long names of all the species
 Args     : undef, $name, @a_names 0,1 or several names
 Error    : a ParameterException is thrown if the number of arguments does not fit with the context

=cut

sub GetSpeciesLongName
{
	my $self = shift;
	my ( @a_names ) = @_;
    if ($self->_Get('_light') == &TRUE)
    {
        throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" ) 
    }
    
    #pop(@a_names) if (@a_names==1 && $a_names[0] eq 'all');
    # PATCH CRADO TEMPORAIRE - POUR ncs_Blast.pl
    return &NONE if (@a_names==1 && (!defined $a_names[0] || $a_names[0] eq 'all') );

    $self->__InitGenomesByLongIfNecessary();
    
    # We may use this function in scalar context only if 1 name specified
    throw LipmError::ParameterException('a_names', 'type', -text => "Only 1 name must be specified in scalar context " ) 
    if (!wantarray and @a_names!=1);
	
    my $rh_genomes         = $self->_Get('__rh_genomes');
    my $rh_genomes_by_long = $self->_Get('__rh_genomes_by_long');

    # No parameter selected = return all species long names
    if (@a_names == 0)
    {
        my @a_rvl = map { $rh_genomes->{$_}{long} } keys %$rh_genomes;
        return @a_rvl;
    }
    
    # One or several parameters selected = convert names and return the result
    my @a_long;
    foreach my $name (@a_names)
    {
        # We may have a list of undef as parameter 
        next unless (defined $name);
        
        # if it is a long name, return it
        if ($self->IsSpeciesLongName ( $name ) )
        {
            push(@a_long,$name);
        }
        
        # if it is a short name, return the corresponding long name
        elsif ( $self->IsSpeciesShortName ( $name ) )
        {
            push(@a_long, $rh_genomes->{$name}->{long});
        }
        
        # else try to find a short name from this name (using regex match), and if found return the corresponding long name
        else
        {
            my $short_name = $self->GetSpeciesShortName( $name );
            push ( @a_long, ($short_name eq &NONE) ? $short_name : $rh_genomes->{$short_name}->{long});
        }
    }
    
    # return a string or an array
    return wantarray ? @a_long : $a_long[0];
}


=head2 Function GetSpeciesObserver

 Usage        : my $o_sp_obs = $o_x->GetSpeciesObserver( $species )
 Function     : Build a species observer object for the given species.
 Returns      : $o_sp_obs, a SpeciesObserver object.
 Args         : $species, string, short or long species name.
 Error        : LipmError::ParameterException

=cut

sub GetSpeciesObserver
{
	my $self = shift;
	my( $species ) = @_;

	my $o_sp_obs = New SpeciesObserver( species     => $species,
										-o_inst_obs => $self,
										-o_logger   => $self->_GetLogger() );

	return $o_sp_obs;
} #GetSpeciesObserver

=head2 Function GetSpeciesInfo

 Usage    : my $rh_info   = $object->GetSpeciesInfo( $name );
            my @a_rh_info = $object->GetSpeciesInfo( @a_names );
            my @a_rh_info = $object->GetSpeciesInfo();
 Function : retrieve the hash corresponding to the short names passed by parameters
            The (keys,values) found in this hash give all the information we have about this species
            If no argument, retrieve ALL hashes
 Return   : 1/ a scalar, the info hashes of the species
            2/ an array, the info hashes of the species
            3/ an array, the info hashes of all the species
 Args     : undef, $name, @a_names 0,1 or several SHORT names (no check)
 Error    : a ParameterException is thrown if the number of arguments does not fit with the context
            If some species is not found, '' is returned (no error thrown)

=cut

sub GetSpeciesInfo
{
	my $self = shift;
	my ( @a_names ) = @_;

    if ($self->_Get('_light') == &TRUE)
    {
        throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" ) 
    }

	# We may use this function in scalar context only if 1 name specified
	throw LipmError::ParameterException('a_names', 'type', -text => "Only 1 name must be specified in scalar context " )
	if (!wantarray and @a_names!=1);

	my $rh_genomes = $self->_Get('__rh_genomes');

	# No parameter selected = return all species
	if (@a_names == 0)
	{
		@a_names = keys %$rh_genomes;
	}

	my @a_infos = ();
	foreach my $name ( @a_names )
	{
		if( defined $$rh_genomes{ $name } )
		{
			push( @a_infos, $$rh_genomes{ $name } );
		}
		else
		{
			my $msg = "You probably asked for information about a species";
			$msg   .= " ($name) that is not in this instance.\n";
			throw LipmError::ParameterException('@a_names','value',-text=>$msg);
		}
	}

	# return a string or an array
	return wantarray ? @a_infos : $a_infos[0];
}


=head2 Function GetDataPaths

 Usage        : my $path = $o_x->GetDataPaths( $where )
 Function     : read.
 Returns      : An array containing all the paths (typically the session path, then the instance path)
 Args         : $where, coma separated list of keywords
                Allowed keywords => INSTANCE,SESSION

=cut

sub GetDataPaths
{
	my $self = shift;
	my( $where ) = @_;

	my @a_paths = ();

	my @a_keywords = split( /,/, $where );
	foreach my $keyword ( @a_keywords )
	{
        # Ignore the session keywords if the session is not set
        next if ( $keyword eq 'SESSION' && $self->GetSessionDirectory() eq '');
        
		my $key = ( $keyword eq 'SESSION' ) ? 'ROOT_LOCAL_DATA' : 'ROOT_DATA';

		push( @a_paths, $self->GetCfg( $key ) );
	}

	return @a_paths;
} #GetDataPaths


=head2 Function GetSpeciesListFilepath

 Usage        : my $path = $o_x->GetSpeciesListFilepath( $where )
 Function     : return the path to a species.list file
                the instance one or the session one regarding $where.
 Returns      : $path, string, path to a species.list file.
 Args         : $where, string : INSTANCE or SESSION
 Error        : none

=cut

sub GetSpeciesListFilepath
{
	my $self = shift;
	my( $where ) = @_;

	my $key = ( $where eq 'SESSION' ) ? 'ROOT_PERSONAL' : 'ROOT_SITE';

	my @a_keys = ( $key,'SPECIES_LIST' );
	my @a_dirs = $self->GetCfg( @a_keys );
	my $path = File::Spec->catfile( @a_dirs );

	return $path;
} #GetSpeciesListFilepath


=head2 Function GetSpeciesCfgDirpath

 Usage        : my $path = $o_x->GetSpeciesCfgDirpath( $where )
 Function     : return the path to the species directory.
                the instance one or the session one regarding $where.
 Returns      : $path, string, path to a species.list file.
 Args         : $where, string : INSTANCE or SESSION
 Error        : none

=cut

sub GetSpeciesCfgDirpath
{
	my $self = shift;
	my( $where ) = @_;

	my $key = ( $where eq 'SESSION' ) ? 'ROOT_LOCAL_SPECIES' : 'ROOT_SPECIES';
    return $self->GetCfg( $key );
} #GetSpeciesCfgDirpath


=head2 Function GetComparativeDirpaths

 Usage        : my $path = $o_x->GetComparativeDirpaths( $where )
 Function     : read.
 Returns      : @a_paths, a list of pathes ('.../data/comparative' etc.) 
 Args         : $where, coma separated list of keywords
                Allowed keywords => INSTANCE,SESSION
 Error        : none ?

=cut

sub GetComparativeDirpaths
{
	my $self = shift;
	my( $where ) = @_;

	my @a_paths = ();

	my @a_keywords = split( /,/, $where );
	foreach my $keyword ( @a_keywords )
	{
        # Ignore the session keywords if the session is not set
        next if ( $keyword eq 'SESSION' && $self->GetSessionDirectory() eq '');
        
		my $key = ( $keyword eq 'SESSION' ) ? 'ROOT_LOCAL_DATA' : 'ROOT_DATA';

		my( $path, $comp_dir ) = $self->GetCfg( $key, 'COMPARATIVE' );

		push( @a_paths, File::Spec->catdir( $path, $comp_dir ) );
	}

	return @a_paths;
} #GetComparativeDirpaths


=head2 Function GetComparisonPaths

 Usage        : my $path = $o_x->GetComparisonPaths( $where )
 Function     : return a list of comparison paths depending on $where value.
 Returns      : @a_path, array, a set of paths
 Args         : $where, coma separated list of keywords
                Allowed keywords => INSTANCE,SESSION
 Error        : none ?

=cut

sub GetComparisonPaths
{
	my $self = shift;
	my( $where ) = @_;

	my @a_paths = ();

	my @a_keywords = split( /,/, $where );
	foreach my $keyword ( @a_keywords )
	{
        # Ignore the session keywords if the session is not set
        next if ( $keyword eq 'SESSION' && $self->GetSessionDirectory() eq '');
        
		my $key = ( $keyword eq 'SESSION' ) ? 'ROOT_LOCAL_DATA' : 'ROOT_DATA';

		my( $path, $comp_dir ) = $self->GetCfg( $key, 'COMPARISON' );

		push( @a_paths, File::Spec->catdir( $path, $comp_dir ) );
	}

	return @a_paths;
} #GetComparisonPaths


=head2 GetLogsDirpath

 Usage        : my $dirpath = $o_x->GetLogsDirpath( $where )
 Function     : .
 Returns      : $dirpath
 Args         : $where, string, ['SESSION'] *or* 'INSTANCE'
 Error        : none

=cut

sub GetLogsDirpath
{
	my $self = shift;
	my( $where ) = @_;

	my $key = ( $where eq 'SESSION' ) ? 'ROOT_LOCAL_LOGS' : 'ROOT_LOGS';
    return $self->GetCfg( $key );
} #GetLogsDirpath


=head2 GetSyntenyListFilepath

 Usage        : my $filepath = $o_x->GetSyntenyListFilepath( none )
 Function     : accessor to the synteny.list file.
 Returns      : The complete path to the file
 Args         : none
 Error        : none

 Note    : A voir si cette fonction reste utile malgre la presence de GetSyntenyListFilepaths

=cut

sub GetSyntenyListFilepath
{
	my $self = shift;
	my( $where ) = @_; # a voir si j'en ai besoin.

	# pour la retro compatibilite
	$where = 'INSTANCE'  unless defined( $where );

	my $key = ( $where eq 'SESSION' ) ? 'ROOT_PERSONAL' : 'ROOT_SITE';

	my( $path, $file ) = $self->GetCfg( $key, 'SYNTENY_LIST' );

	my $filepath = File::Spec->catfile( $path, $file );

	return $filepath;
} #GetSyntenyListFilepath


=head2 GetSyntenyListFilepaths

 Usage        : my $filepath = $o_x->GetSyntenyListFilepaths( none )
 Function     : accessor to the synteny.list file.
 Returns      : an array with all the synteny.list filepaths
 Args         : none
 Error        : none

=cut

sub GetSyntenyListFilepaths
{
	my $self = shift;
	my( $where ) = @_; # a voir si j'en ai besoin.

	$where = 'SESSION,INSTANCE'  unless defined( $where );
	my @a_filepaths = ();

	foreach my $keyword( split( /,/, $where ) )
	{
        # Ignore the session keywords if the session is not set
        next if ( $keyword eq 'SESSION' && $self->GetSessionDirectory() eq '');
        
		my $key = ( $keyword eq 'SESSION' ) ? 'ROOT_PERSONAL' : 'ROOT_SITE';

		my( $path, $file ) = $self->GetCfg( $key, 'SYNTENY_LIST' );

		my $filepath = File::Spec->catfile( $path, $file );
		push( @a_filepaths, $filepath ) if( -e $filepath );
	}

	return @a_filepaths;
} #GetSyntenyListFilepaths



=head1 Private methods

=head2 procedure __InitGenomesByLongIfNecessary

 Usage     : $self->__InitGenomesByLongIfNecessary()
 Procedure : If necessary only, initialize the hash __rh_genomes_by_long
             keys:   the long species name after removing any space
                     TODO - Hope that we'll never have 2 species whose long names differ ONLY by spaces....
             values: the short species name
 Args      : none

=cut

sub __InitGenomesByLongIfNecessary
{
    my $self = shift;

    if ($self->_Get('_light') == &TRUE)
    {
        throw LipmError::ObjectException('InstanceObserver','!attribute', -text => "You cannot use this function in Light mode" ) 
    }

    unless (defined $self->_Get('__rh_genomes_by_long'))
    {
        my $rh_cfg     = $self->_Get('_rh_cfg');
        my $rh_genomes = $rh_cfg->{GENOME};
        my %h_genomes_by_long;
        foreach my $key (keys %$rh_genomes)
        {
			#warn "$key ";
			# TODO - Vilain patch, ca peut arriver dans le cas de commentaires lus par ndr, 
			#        ou autres raisons qui font que le fichier species.list est defectueux
			#        L'utilisation des nouveaux objets NarcisseDataWriter pourrait arranger les choses ?
			if (! defined $rh_genomes->{$key}->{long})
			{
				delete $rh_genomes->{ $key };
				#warn "$key VIRE";
			}
			else
			{
	            my $long_key = $rh_genomes->{$key}->{long};
	            $long_key =~ s/ //g;
	            $long_key = uc $long_key;
	            $h_genomes_by_long{$long_key}=$key;
			}
        }
        $self -> _Set('__rh_genomes_by_long',\%h_genomes_by_long);
    }
}


1;
