#
#       Sebastien.Letort@toulouse.inra.fr
#       Created: November 19, 2007
#       $Id: NarcisseObserver.pm 2433 2009-03-26 09:59:35Z sletort $
#

package NarcisseObserver;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

 NarcisseObserver - Access to global configuration information about all instances integrated to a Narcisse installation

=head1 DESCRIPTION

 NarcisseObserver inherits from Configuration. It is supposed that we are living inside the instance 'multi'.
 The parameter 'NARCISSE_INSTANCE_ROOTS' should be set, it gives the list of recorded instances. The method
 GetInstanceObservers returns all the recorded instances

=head1 SUBROUTINES
#
# 	GetInstancesObserver
# 	GetInstanceObserver		# be carefull cf note
 GetInstancesNames
# 	__GetInstances
#
=head1 AUTHOR

	Sebastien Letort : sebastien.letort@toulouse.inra.fr

=cut

use warnings;
use strict;

use Data::Dumper;
use IO::Dir;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 2433 $ =~ /\d+/g); $r[0]};
}

use base qw( Configuration );
use Error qw( :try );
use LipmError;
use General;
use Configuration::InstanceObserver;
use File::Basename;

use constant ALL => 'all';


=head2 Function _Init

 Title      : _Init
 Usage      : $obj = New NarcisseObserver(-cfg => $cfg_file, -o_logger = $o_logger);
 Prerequisite: The parameter called NARCISSE_INSTANCE_ROOTS must be defined
               Those directories are siblings of the 'multi' directory (they can be symbolic links)
 Function   : constructor.
 Args       : $cfg_file (optional): the path to narcisse.cfg - See __FindNarcisseCfg for the details
              $o_logger (optional)

=cut

sub _Init
{
	my $self = shift;
	my %h_args = (	-cfg      => undef,
					-o_logger => undef,
					@_ );

	$self->SUPER::_Init( %h_args );

	if ($self->GetName() ne 'multi')
	{
		my $msg = "ERROR - Please run chins multi\n";
		throw LipmError::ParameterException('Name', 'value', -text => $msg )
	}

	my $cfg_path      = $self->_Get('_cfg_path');
	my @a_dirs        = File::Spec->splitdir( $cfg_path );
	my $narcisse_root = File::Spec->catfile(@a_dirs[0..@a_dirs-5]);
	my @a_instance_roots  = split(',',$self->GetCfg('NARCISSE_INSTANCE_ROOTS'));
	$self->_Set( '__ra_instance_roots', \@a_instance_roots );

	# Create a hash of instance observers (key = root)
	my %h_instances       = ();
	foreach my $instance_name (@a_instance_roots)
	{
		my $cfg_file = File::Spec->catfile($narcisse_root,$instance_name,'Site','cfg','narcisse.cfg');
        try
        {
            my $o_io     = New InstanceObserver( -cfg => $cfg_file, -o_logger => $h_args { -o_logger } );
            $h_instances{$instance_name} = $o_io;
        }
        catch LipmError with
        {
            my $o_err = shift;
            warn "WARNING - ERROR WITH INSTANCE $instance_name -> IGNORED " . $o_err->Explain() . "\n";
            warn "-------------------\n";
        }
        catch NarcisseError with
        {
            my $o_err = shift;
            warn "WARNING - ERROR WITH INSTANCE $instance_name -> IGNORED (configuration version = " . $o_err->cv . ")\n";
        };
	}

	$self->_Set('__rh_instances',\%h_instances);

	return;
} # _Init


=head2 Function GetInstanceObservers

 Usage    : my @a_o_io = $object->GetInstanceObservers()
 Function : retrieve the hash of InstanceObserver objects known by this object
 Return   : A ref to a hash: key = the root directory, value = $o_io 
 Args     : none 

=cut
sub GetInstanceObservers
{
    my $self = shift;
    return $self->_Get('__rh_instances')
}


=head2 function FindInstanceObservers

 Usage        : $ra_list = $o_xx->FindInstanceObservers( %h_args  );
 Function     : Get a list of InstanceObserver objects regarding args.
 Returns      : a ref to a list of InstanceObserver objects
 Args         :
     ra_species => $ra_species, a ref to a list of strings: the short or long species name
                   If $ra_species is specified, we return only the InstanceObservers with this species.
     ra_instances => $ra_instances, a ref to a list of strings: the names of the instances you want.

     # not yet implemented <- on en a pas eu besoin encore.
     $regexp_flag, boolean, if set to &TRUE, will search $species as a regexp in long names.

=cut

sub FindInstanceObservers
{
	my $self = shift;
	my %h_args = (	ra_species   => [],
					ra_instances => [],
					#regexp
					@_ );

	my $rh_io = $self->GetInstanceObservers();

	# filter from instance names
	my $ra_instances = $h_args{ ra_instances };
	my @ao_io = ();

	if( 0 == @$ra_instances )
	{
		@ao_io = values %$rh_io;
	}
	elsif( 1 == @$ra_instances && $$ra_instances[0] eq 'all' )
	{
		@ao_io = values %$rh_io;
	}
	else
	{
		foreach my $name ( @$ra_instances )
		{
			if( defined( $$rh_io{ $name } ) )
			{
				push( @ao_io, $$rh_io{ $name } );
			}
		}
	}

	# filter from species names
	my $ra_species = $h_args{ ra_species };
    return \@ao_io  if ( 0 == @$ra_species );

	my @ao_io_final = ();
	foreach my $o_io ( @ao_io )
	{
		foreach my $species ( @$ra_species )
		{
# 			if( $o_io->HaveSpecies( $species, $regexp_flag ) )
			if( ( &ALL eq $species ) or ( $o_io->HaveSpecies( $species ) ) )
			{
				push( @ao_io_final, $o_io );
				last; # sort de la boucle species
			}
		}
	}

	return \@ao_io_final;
}


=head2 function GetInstancesDescr

 Title        : GetInstancesDescr
 Usage        : %h_list = &GetInstancesDescr();
 Prerequisite : none
 Function     : give an access to all instances name and description.
 Returns      : %h_list, keys are name of instances and value their description
 Args         : none
 Globals      : none
 TODO - Terminer cette fonction ou la supprimer, a mon avis elle n'a rien a faire ici

=cut

sub GetInstancesDescr
{
	my $self = shift;

	my %h_list;
	foreach my $name ( $self->GetInstancesNames() )
	{
		#utiliser le fichier abstract.inc et appeler une methode d'instanceObserver.
		$h_list{ $name } = "coming soon a brief description of the instance";
	}

	return %h_list;
}


=head2 GetInstancesNames

 Usage        : my @a_names = $o_x->GetInstancesNames( none )
 Function     : Give the list of all instance name.
 Returns      : @a_names, the list of all instance names.
 Args         : none

=cut

sub GetInstancesNames
{
	my $self = shift;

	my $rh_instances = $self->_Get( '__rh_instances' );

	# trie ou pas ?
	my @a_names = sort keys %$rh_instances;

	return @a_names;
} #GetInstancesNames


=head2 GetDB

 Usage        : my $filepath = $o_x->GetDB( $type_db )
 Function     : call mother and change .fasta to .pal.
 Returns      : absolute path, $filepath to the db file, pal file.
 Args         : $type_db, string : 'prot' or 'nuc'.
 Error        : none

=cut

sub GetDB
{
	my $self = shift;

	my $filepath = $self->SUPER::GetDB( @_ );

	$filepath =~ s/\.fasta$/.pal/;

	return $filepath;
} #GetDB



1;
