#
# 	emmanuelle.permal@toulouse.inra.fr
#	Created: 24 04 2013
#	$Id: DBAccess.pm 
#
# Copyright INRA/CNRS

# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package DBAccess;

=pod

=head1 NAME

DBAccess - An object that gives access to protein and cds sequences

=head1  SYNOPSIS

my $o_dba = New DBAccess ( id => $id);

=head1 DESCRIPTION

This class allows the access to CDS and protein sequences given their protein ID
For example: PAC:19644189.CDS.1 for Arabidopsis thaliana

=cut

use strict;
use warnings;
use Data::Dumper;
use base qw( LipmObject );
use Bio::Seq;
use Bio::SeqIO;
use Bio::DB::Fasta;
use FastaReader;
use General;
use FindBin;
use Cwd;
use File::Path qw( rmtree ); 
use Cwd 'abs_path';
use Runner;
my $DEBUG = 0;

=head2 method _Init

 Title        : _Init
 Usage        : my $o_dba = New DBAccess(species => $species, -o_inst_obs => $o_io);
 Prerequisite : none
 Fonction     : Init
 Returns      : $self
 Args         : $id une ide de type : "PAC:19651494.CDS.1"

     
=cut

sub _Init
{
	my $self = shift;
	
	my %h_args = (  
					sequences => undef,
					species => undef,
					-o_inst_obs => undef,
					-rep => &FALSE,
					@_);

	$self -> SUPER::_Init( @_ );
	
	my ($species, $o_io, $bool_rep ) = @h_args{ qw( species -o_inst_obs -rep) };
	my $o_spo = $o_io -> GetSpeciesObserver( $species);
	
	#Get path to data via species observer
	my $path_to_prot = "";
	if ($bool_rep == &TRUE)
	{
		$path_to_prot = $o_spo -> GetProtFastaFile (-rep => &TRUE);
	}
	else
	{
		$path_to_prot = $o_spo -> GetProtFastaFile ();
	}
	my $path_to_cds = $o_spo -> GetCDSFastaFile();
	
	#Create working dir to index files
	my ($cwd, $wkd) = $self -> _CreateWorkDir("DBAccess");
	
	#Creta symbolink link in working directory
	my $cds_file_name = $self -> _CreateSymLink($path_to_cds);
	my $protein_file_name = $self -> _CreateSymLink($path_to_prot);
	
	#Index files and get db
	my $o_db_cds = my $o_db_prot = undef;
	
	if ( -e $cds_file_name )
	{
		$o_db_cds = $self -> _GetNewDBFasta ($cds_file_name);
	}
	else
	{
		warn "the file $cds_file_name does not exist\nExiting program\n";
		exit(0);	
	}
	
	if ( -e $protein_file_name )
	{
		$o_db_prot = $self -> _GetNewDBFasta ($protein_file_name);
	}
	else
	{
		warn "the file $protein_file_name does not exist\nExiting program\n";
		exit(0);
	}
	
	#Get sequences
	my $rh_sequences = $self -> _FillSequenceHash($o_db_cds, $o_db_prot);
	
	$self -> _Set (	sequences => $rh_sequences);
	
	$self -> _RemoveWorkDir($cwd, $wkd);
	return $self;
} # _Init

=head2 Procedure _GetNewDBFasta

 Usage        : $self -> _GetNewDBFasta($file_name);
 Procedure    : Create a new Bio::DB::Fasta object from a $file
 Args         : A file name from which the db object will be created
 Returns      : A Bio::DB::Fasta object

=cut

sub _GetNewDBFasta
{
	my $self = shift;
	my ($file_name) = (@_);
	my $o_db = Bio::DB::Fasta->new($file_name);
	return $o_db;
}

=head2 Procedure _GetSeqbyDBFasta

 Usage        : $self -> _GetSeqbyDBFasta($o_db, $id);
 Procedure    : Extract a sequence corresponding to an id
 Args         : $o_db a Bio::DB::Fasta object, $id the id given in the constructor
 Returns      : The sequence corresponding to the $db and the $id

=cut

sub _GetSeqbyDBFasta
{
	my $self = shift;
	my ($o_db, $id) = (@_);
	my $o_seq = $o_db -> get_Seq_by_id($id);
	my $seq   = $o_seq -> seq();
	
	return $seq;
}

=head2 Procedure _GetSequenceHash

 Usage        : $self -> _GetSequenceHash;
 Procedure    : Gets hash with all sequences and returns it
 Args         : none
 Returns      : %h_sequences

=cut

sub _GetSequenceHash
{
	my $self = shift;
	my $rh_sequences = $self -> _Get ('sequences');
	return %$rh_sequences;
}

=head2 Procedure _FillSequenceHash

 Usage        : $self -> _FillSequenceHash();
 Procedure    : Fills a hash with protein $id and their corresponding sequences in AA and nt
 Args         : $o_db_cds, $o_db_protein : Bio::DB::Fasta objects containing respectiveley cds and protein sequnces
 Returns      : A hash with filled with cds and protein sequences

=cut

sub _FillSequenceHash
{
	my $self = shift;
	my ($o_db_cds, $o_db_protein) = (@_);
	my %h_sequences = ();
	
	my @a_protein_ids = $o_db_protein -> ids;
	
	foreach my $id (@a_protein_ids)
	{
		my $protein_seq = $self -> _GetSeqbyDBFasta ($o_db_protein, $id);
		my $trim_id = $self-> _TrimId ($id, 'prot');
		$h_sequences{$trim_id} = {protein=> $protein_seq,
									cds => undef};
		
	}
	my @a_cds_ids = $o_db_cds -> ids;
	
	foreach my $id (@a_cds_ids)
	{
		my $cds_seq = $self -> _GetSeqbyDBFasta ($o_db_cds, $id);
		my $trim_id = $self-> _TrimId ($id,'cds');
		$h_sequences{$trim_id}->{ cds } = $cds_seq;
	}
		
	return \%h_sequences;
}

=head2 Procedure _TrimId

 Usage        : $self -> _TrimId($id, $type);
 Procedure    : Remove the tag before ":" in the $id
 Args         : $id the ID to shorten
 Returns      : The trimed id

=cut

sub _TrimId
{
	my $self = shift;
	my ( $id, $type )= (@_);
	
	$type .= ":";
	$id =~ s/$type//;
	
	return $id;
}

=head2 Procedure MaskCpG

 Usage        : $o_dbaccess -> MaskCpG ($ra_id, $minlen, $window);
 Procedure    : Mask CpG Island using Emboss
 Args         : $ra_ids: A reference to the array containing id of sequnces to be masked
				$minlen: The minimum length of the sequence (optional parameter)
				$window: The window to compute CpG (optional parameter)
 Returns      : A hash with masked sequences

=cut

sub MaskCpG
{
	my $self = shift;
	
	my ($ra_ids, $minlen, $window) = (@_);
	my %h_results = ();
	
	#Create a work directory
	my ($save_dir, $working_dir) = $self -> _CreateWorkDir("CpGreport");
	
	#Writting fasta file
	my $rh_filenames = $self -> _WriteCpGFasta ($ra_ids);
	
	#For each id check CpG
	foreach my $id (keys (%$rh_filenames))
	{
		my $file_name = $rh_filenames -> {$id};
		
		#Run Emboss newcpgreport
		my $res_cpg_file = $self -> _RunNewCpGReport($file_name, $minlen, $window);
	
		#Parse newcpgreport Output
		my $rh_cpg = $self -> _ParseCpGReport($res_cpg_file);

		#Mask sequence with CpG region if needed
		my $seq = $self -> _RunMaskSeq($file_name, $rh_cpg);
		
		#Store results in hash
		$h_results { $id } = $seq;
	}
	
	#Clean work directory
	$self -> _RemoveWorkDir($save_dir, $working_dir);
	
	return \%h_results;
}

=head2 Procedure _RunMaskSeq

 Usage        : $o_dbaccess -> _RunMaskSeq ($filename, $ra_cpg);
 Prerequisite : maskseq has to be in path
 Procedure    : Run maskseq with results from newcpgreport
 Args         : $filename, the name of the file to be masked
				$ra_cpg, the ref to the array with cpg island coordinates to mask
 Returns      : a sequnce id and the associated sequence
 Note         : maskseq only runs one sequence 

=cut

sub _RunMaskSeq
{
	my $self = shift;
	
	my ($input, $rh_cpg) = (@_);
	my $output = $input .".masked";
	my $seqio = undef;
	
	#Getting ID
	my $id = (keys (%$rh_cpg))[0];
	#Getting region to mask
	my $ra_cpg = $rh_cpg -> {$id};
	my $cpg_nb = scalar @$ra_cpg;
	
	if ($cpg_nb > 0)
	{
		my $region = join(",", @$ra_cpg);
		#Run maskseq
		my $cmd = "maskseq -auto -sequence $input -reg $region -outseq $output"; 
		my ($status, $killed) = RunExt( -cmd => $cmd );
	
		#Get the sequence if cpg island
		$seqio = Bio::SeqIO -> new(-file => "$output", '-format' => 'Fasta');

	}
	else
	{
		#Get the sequence if no Cpg Island
		$seqio = Bio::SeqIO -> new(-file => "$input", '-format' => 'Fasta');
	}
	my $o_seq = $seqio->next_seq;
	my $seq = $o_seq -> seq;
	
	return $id, $seq;
}

=head2 Procedure _RunNewCpGReport

 Usage        : $o_dbaccess -> _RunNewCpGReport ($filename, $minlen, $window);
 Prerequisite : newcpgreport has to be in PATH
 Procedure    : Run newcpgreport
 Args         : $filename, the name of the input file for newcpgreport 
				$minlen, the minimum length of a sequence
				$window, the window size to compute CpG
 Returns      : the name of the output file

=cut

sub _RunNewCpGReport
{
	my $self = shift;
	
	my ($input, $minlen, $window) = (@_);
	
	#Setting default EMBOSS default values
	$minlen = 200 if (!defined $minlen);
	$window = 100 if (!defined $window);
	
	#Running
	my $output = ( split ( "\\.", $input ) )[0] . ".newreportcpg";
	my $cmd = "newcpgreport -auto -minlen $minlen -window $window --sequence $input -outfile $output"; 
	my ($status, $killed) = RunExt( -cmd => $cmd );

	return $output;
}

=head2 Procedure _ParseCpGReport

 Usage        : $o_dbaccess -> _ParseCpGReport($filetoparse);
 Procedure    : Run newcpgreport program
 Args         : The output file of newcpgreport to parse
 Returns      : $rh_cpg a ref to a hash where each $id has a corresponding refrence to an array with cpg island coordinates

=cut

sub _ParseCpGReport
{
	my $self = shift;
	
	my ( $file2parse ) = (@_);
	my $fi_cpg = &GetStreamIn ($file2parse);
	my %h_cpg = ();
	my @a_cpg = ();
	my $id ="";
	foreach my $line (<$fi_cpg>)
	{
		if ($line =~ /^ID/)
		{
			$id = ( split (" ", $line))[1];
			@a_cpg = ();	
		}
		if ($line =~ /^FT/)
		{
			if ($line =~ /CpG island/)
			{
				my $coords = (split (" ", $line))[3];
				$coords =~ s/\.\./-/;
				push (@a_cpg, $coords);
			}
		}
		if ($line =~ /^\/\//)
		{
			$h_cpg { $id } = \@a_cpg;	
		}	
	}
	return \%h_cpg;
}

=head2 Procedure _WriteCpGFasta

 Title        : _WriteCpGFasta
 Usage        : &_WriteCpGFasta ($ra_ids);
 Procedure    : Print data in a fasta format file
 Args         : a reference to an array with id
 Returns 	  : A ref to a hash full of file names

=cut

sub _WriteCpGFasta
{
	my $self = shift;
	my ($ra_ids) = (@_);
	my $index = 0;
	
	my %h_files = ();
	foreach my $id (@$ra_ids)
	{
		my $file_name = "input_newcpg_" . $index . ".fasta";
		my $fo_cpg  = &GetStreamOut ($file_name);
		my $seq = $self -> GetCDSSequence ($id);
		my $fasta_tag = ">" . $id;
		$seq =~ s/(.{60})/$1\n/g;
		print $fo_cpg "$fasta_tag\n";
		print $fo_cpg "$seq\n";
		$fo_cpg -> close();	
		$index += 1;
		
		$h_files {$id} = $file_name;
	}	
	
	return \%h_files;
}

=head2 Procedure GetRefSequenceHash
 
 Usage        : $o_dbaccess -> GetRefSequenceHash();
 Procedure    : Get the ref to the hash containing all sequences
 Args         : none
 Returns      : A reference to a hash
 
=cut
sub GetRefSequenceHash
{
	my $self = shift;
	return $self -> _Get ('sequences');
}

=head2 Procedure GetCDSSequence

 Usage        : $o_dbaccess-> GetCDSSequence($id);
 Procedure    : Get the cds sequence from object $o_DBaccess
 Args         : $id the cds id
 Returns      : The CDS sequence corresponding to the $id given to build $o_DBaccess

=cut

sub GetCDSSequence
{
	my $self = shift;
	my ( $id )= (@_);
	my %h_sequences= $self -> _GetSequenceHash;
	my $cds_seq = $h_sequences{$id}{'cds'};
	if ( !defined $cds_seq)
	{
		$cds_seq = "NA";
	}	
	return $cds_seq;
}

=head2 Procedure GetProteinSequence

 Usage        : $o_DBaccess-> GetProteinSequence($id);
 Procedure    : Get the protein sequence from object $o_DBaccess
 Args         : $id the protein id
 Returns      : The protein sequence corresponding to the $id given to build $o_DBaccess

=cut

sub GetProteinSequence
{
	my $self = shift;
	my ( $id )= (@_);
	my %h_sequences = $self -> _GetSequenceHash;
	my $protein_seq = $h_sequences{$id}{'protein'};
	if ( !defined $protein_seq)
	{
		$protein_seq = "NA";
	}	
	return $protein_seq;
}

=head2 Procedure _CreateWorkDir

 Usage        : $self -> _CreateWorkDir($prefix);
 Procedure    : Creates a temporary working directory to avoid index problems with Bio::DB::Fasta
 Args         : $prefix a prefix
 Returns      : the $working_dir and the $currentdirectory
 
=cut

sub _CreateWorkDir
{
	my $self = shift;
	my ($prefix) = (@_);
	
	my $sav_dir = cwd;
	my $working_dir = File::Temp::tempdir( $prefix."_XXXX", CLEANUP => 0, TMPDIR => 1 );
	chdir($working_dir);
	
	return $sav_dir, $working_dir;
}

=head2 Procedure _RemoveWorkDir

 Usage        : $self -> _RemoveWorkDir($prefix);
 Procedure    : Removes temporary working directory 
 Args         : $prefix a prefix
 Returns      : none
 
=cut

sub _RemoveWorkDir
{
	my $self = shift;
	my ($sav_dir, $working_dir) = (@_);
	chdir $sav_dir;
	rmtree($working_dir);
	return ;
}

=head2 Procedure _CreateSymLink

 Usage        : $self -> _CreateSymLink($path_to_old_file);
 Prerequisite : 
 Procedure    : Create a symbolic link to $path_to_old_file
 Args         : $path_to_old_file the path to the file to be linked
 returns      : The name of the symbolic linked file without the path in it

=cut

sub _CreateSymLink
{
	my $self = shift;
	my ($path_to_old_file) = (@_);	
	my $new_file = (split("\\/", $path_to_old_file))[-1];
	my $symlink_exists=eval {symlink("",""); 1};
	if ($symlink_exists) 
	{
		symlink ($path_to_old_file, $new_file);
	}
	else
	{
		print "Linking db file $path_to_old_file failed\n";
		exit(1);
	}
	
	return $new_file;	
}

1;
