#
# 	sebastien.letort@toulouse.inra.fr
#	Created: July 31, 2008
#   revised by emmanuel.courcelle@toulouse.inra.fr - Oct 2008
#   modified by thomas.faraut@toulouse.inra.fr - march 2009
#	$Id: ChromosomeFeatures.pm 2115 2008-10-14 07:24:38Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

	DataReader::Region - Subclass of DataReader, useful to read "indexed" tsv files

=head1  SYNOPSIS

	my $o_reader = New DataReader::Region();
	#Default start/end labels are 'start' and 'end'
	$o_ndr->Prepare( 
	                 -from   => $file,
	                 -where  => { RANGE => { start => 1, start_label => 'q_start', end => 10000000, end_label => 'q_end' } ,
	                              VALUE => { col1_name => val1, col2_name => val2 } 
	                            }
	               );	  

=head1 DESCRIPTION

    The file is read when you call the FetchallArrayref or FetchallHashref functions. Thus, subsequent calls of those
    functions, after different calls of Prepare, will start new I/O
    However, there is no cache as in DataReader::Whole, so there should be less memory consumption.
    Furthermore, if the function FetchallLineByLine is used, a parameter function ("line handler") is called at each line, 
    which may avoid using huge (for huge files) data structures for storing the results.
	    
=head1 FILE FORMAT

    See the FILE FORMAT secification in the DataReader.pm file
	This object manages some sort of indexed files, embedding the "index" in the first line (before the labels).
	Thinking of the file as the description of features along a chromosome, this line cuts the chromosome in slices,
	and gives for each slice the position of the first line corresponding to this slice.
	It is thus easy to use this information to read only parts of the file, when a RANGE condition is specified.
    The index line starts with a @ character, and is formatted as follows:
@start;end;slice_size_base;nb_pos;pos1:pos2:pos3:pos4:...

    start = The LOWEST position specified in the start column
    end   = The HIGHEST position specified in the end column
    slice_size = The size of a slice
    nb_pos     = The number of slices in the file
    pos1:...   = A list of positions, there should be nb_pos in the list
    

=cut

package DataReader::Region;

use strict;
use warnings;

use base qw( DataReader );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;
use LipmError::IOException;
use LipmError::ParameterException;

use General;

use constant MAX_CHROM_SIZE => 4000000000;
use constant ARRAY => 0;
use constant HASH  => 1;
use constant SLICE_UNIT =>  100000; # 100kb is the unit slice size 

# Column separator
our $SEP = "\t";

=head2 Procedure Prepare

 Usage        : $o_obj->Prepare( %h_args );
 Procedure    : Call SUPER::Prepare, then force the headers if necessary
 Args         : see SUPER::Prepare
=cut
sub Prepare
{
	my $self = shift;
	my %h_args = (
        -from   => '-from option',	# dummy name, useful for error message generation
        -where  => {},
        -select => [],  
        -force_reading => 0,
        @_
        );

	$self->SUPER::Prepare( %h_args );
	
	# Forcing the header 
	my $ra_headers = $self->_Get('_force_headers');
	$self->_SetHeaderField( $ra_headers );
}
#----------------------------------
#          Reading the file
#---------------------------------

=head2 Procedure __ReadelectedDataFromFile

 Usage        : $o_obj->_ReadDataFromFile( s);
 Procedure    : Reads the file header and calls _ReadAndFormatData
 				All the lines are stored in an array
 				The method then stores, the following attributes
 				_ra_return 		: an array of the array refs (only available when the return type is an array)
 				_rh_return 		: a hash with the array refs (only available when return type is a hash)
 				(REMOVED) _ra_sorted_keys	: an array with the keys ordered as in the file	
                _ra_headers : a reference on a table containing the header information
                _rh_headers : a reference on a hash containing the column numbers corresponding to the header fields 
 				
 				(There is a possibility to use and Inline::C implementation of the _ReadAndFormatData method)				
 Args         : none
 Returns 	  : none
               
=cut
sub __ReadSelectedDataFromFile
{
	my  $self    = shift;
	
	my ( $file, $return_type, $line_type, $key_name, $rf_line_handler ) = $self->_Get( qw( _file _return_type _line_type _key_name _line_handler ) ); 

	if ( $return_type eq 'HASH' && $key_name eq "_none" )
	{
		die "If hash return type is requested a valid key_name must be provided";
	}	
	
	my $rh_return_struct = $self -> __ReadAndFormatDataWrapper( $rf_line_handler );
							
	$self->_SetHeaderField( $$rh_return_struct{ _ra_headers } );

	$self->_Set(	%$rh_return_struct,
	                _ra_headers    	=> $self->_Get( '_ra_headers' ),
	                _rh_headers    	=> $self->_Get( '_rh_headers' )
	            );	
	return ;
}

=head2 Procedure __ReadAndFormatDataWrapper

 Usage        : $o_obj->__ReadAndFormatDataWrapper( \%h_args );
                $o_obj->__ReadAndFormatDataWrapper( \%h_args, \&LineHandler );
 Procedure    : A wrapper to _ReadAndFormatData enabling to switch easily between __ReadAndFormatData and DataReader_inlineC::__ReadAndFormatData
                Reads the Narcisse formated file and stores the lines in an array of array ref
 				Each line is transformed into an array using the split function.	  
 Args         : $self, using following protected fields will be used:
	 				-file 		: A file name OR an already opened file handler
	 				start_label : string - name of the column to use as start posotions -
	 				start		: the start of the desired slice
	 				end_label	: string - name of the column to use as end posotions -
	 				end         : the end of the desired slice
	 				-return_type : 'ARRAY'|'HASH' the desired returned type
	 				-line_type   : 'ARRAY'|'HASH' the desired line type
	 				Optional
	 				-key_name	: (required for HASH return type) the name of the column to use as the primary key
	 				-ra_wanted_headers   : (for HASH line type) ref to array - the list of wanted headers - If empty, all headers are wanted
	 				-VALUE       : A ref to a hash (see DataReader::Prepare for the details)
	 				-case_sensitive: bool (see DataReader::Prepare for the details)
	 				-keep_order  : 0/1
	 			&LineHandler (optional): 
		 			If specified, this function will be called for each line, passing the line (an $ra_ or a $rh_) by parameters
		 			This is probably slower (pure perl only), but consumes much less memory
 Returns 	  : a reference on a hash with the following keys:
 				  _ra_return 		: an array of the array refs
 				  _rh_return 		: a hash with the array refs (one of both is empty, depending on the return_type)
 				  _ra_headers		: an array with the field names
 				  (REMOVED) _ra_sorted_keys	: an array with the keys ordered as in the file

=cut
sub __ReadAndFormatDataWrapper
{
	my $self = shift;
	my ($rf_line_handler) = @_;
				  
	my $b_file_name = &FALSE;
	my $file = $self -> _Get ( '_file' );
	my $fh_file;
	if ( ref($file) eq '')
	{
		$fh_file = GetStreamIn ( $file );
		$b_file_name = &TRUE;
	}
	else
	{
		$fh_file = $file;
	}
	$self -> _Set ( '_fh_file', $fh_file);
	
	my $rh_return;
	if ( &DataReader::IsInlineCEnabled() && ! defined($rf_line_handler))
	{
		$rh_return = &DataReader_inlineC::__ReadAndFormatData( $self,$fh_file );
	}
	else
	{
		$rh_return = $self->__ReadAndFormatData( $fh_file,$rf_line_handler );
	}
	
	if ($b_file_name == &TRUE)
	{
		$fh_file -> close;
		$self -> _Set ( '_fh_file', undef);
	}
	
	return $rh_return;
}# _ReadFile_Wrapper

=head2 Procedure __ReadAndFormatData

 Usage        : &DataReader::Region::->__ReadAndFormatData($fh_file,\%h_args, \&rf_line_handler );
 Procedure    : The pure perl version of data file reading
 Args         : $fh_file: An opened file handle
                \&rf_line_handler: See __ReadAndFormatDataWrapper
 Returns 	  : See __ReadAndFormatDataWrapper

=cut

sub __ReadAndFormatData
{
	my $self = shift;
	my ($fh_file,$rf_line_handler) = @_;
	my ( $return_type, $line_type, $key_name, $case_sensitive, $keep_order, $indexed, $rh_values, $rh_range, $ra_headers, $ra_wanted_headers) = $self -> _Get ( qw(_return_type _line_type _key_name _case_sensitive _keep_order _indexed _rh_value_condition _rh_range_condition _ra_headers _ra_select)  );
	my $f_hash_return_type = ( $return_type eq 'HASH' );
	my $l_hash_type        = ( $line_type eq 'HASH' );
	my $start       = $rh_range -> {start}{value};
	my $start_label = $rh_range -> {start}{label};
	my $end         = $rh_range -> {end}{value};
	my $end_label   = $rh_range -> {end}{label};	
	my $data_avail;
	( $data_avail,  $ra_headers ) = $self -> __GetHeaderOffset ( $fh_file, $start, $end, $ra_headers, $indexed );

	my $key_index = 0;
	if ( $f_hash_return_type )
	{
		$key_index = $self -> _GetHeaderColumn( $ra_headers, $key_name);
	}	

	my ($start_col, $end_col) = $self -> _GetHeaderColumn( $ra_headers, $start_label, $end_label );

	# The wanted columns, if any
	my @a_wanted_col;
	if ( @$ra_wanted_headers == 0 )
	{
		@$ra_wanted_headers = @$ra_headers;
	}
	@a_wanted_col = $self -> _GetHeaderColumn( $ra_headers, @$ra_wanted_headers );
	
	# The VALUE parameter (if any)
	my ( @a_value_col ) = $self -> _GetHeaderColumn( $ra_headers, keys %$rh_values );
	my ( @a_value_val ) = values %$rh_values;
	
	# Reading the data
	my @a_data = ();
	my %h_data = ();
	my @a_sorted_keys = ();

#	my $NB_LINE=0;
	if ( $data_avail == &TRUE ) # If false we are outside the range for this file
	{
		while ( my $line = $self -> __NextLine ( $fh_file ) )
		{
			chomp($line);
			#warn $line;
#			$NB_LINE++;
			last if ($line =~ /^\// );		# Leaving if $line starts with /
			my @a_row = split  /$SEP/, $line ;
			next if ( $a_row[ $end_col ]   < $start );
			last if ( $a_row[ $start_col ] > $end );
			if ( @a_value_col != 0 )
    		{
    			next if ( 0 == $self -> _ValidateValueConditions(\@a_row, \@a_value_col,\@a_value_val,$case_sensitive ) );
    		}

			if ( $l_hash_type )
			{
				my %h_row;
				@h_row{ @$ra_wanted_headers } = @{a_row}[ @a_wanted_col ];
				if ( defined $rf_line_handler )
				{
					&$rf_line_handler($self,\%h_row);
				}
				else
				{
					if ( $f_hash_return_type )
					{
						my $key_value  = $a_row[ $key_index ];
						push @a_sorted_keys, $key_value if ( $keep_order );
						$h_data{ $key_value } = \%h_row;
					}
					else
					{
						push( @a_data, \%h_row );
					}
				}
			}
			else
			{
				if ( defined $rf_line_handler )
				{
					&$rf_line_handler($self,\@a_row);
				}
				else
				{
					if ( $f_hash_return_type )
					{
						my $key_value  = $a_row[ $key_index ];
						push @a_sorted_keys, $key_value if ( $keep_order );
						$h_data{ $key_value } = \@a_row;
					}
					else
					{
						push( @a_data, \@a_row );
					}
				}
			}
		}
	}
#	warn "KOUKOU $NB_LINE";

	my %h_return;
	$h_return{ _ra_return }      = \@a_data;
	$h_return{ _rh_return }      = \%h_data;
	$h_return{ _ra_headers }     = $ra_headers ;
	$h_return{ _ra_sorted_keys } = \@a_sorted_keys;

	return \%h_return;

}# _ReadAndFormatData
	

=head2 Function __GetHeaderOffset
 Usage      : my ( $fh_file, $relevant_offset,  $ra_headers ) = $self -> __GetHeaderOffset ( $file, $start, $next );
 Prerequisite: $fh_file must be an already opened file
 Function   : Read the index line and the header, position $fh_file to the beginning of the data section
 Args       : $fh_file - An already opened file handler
              $start - The start of desired slice
              $end   - The end of desired slice
              $ra_headers, $indexed - Constructor parameters -force_headers, -indexed

 Return     : $data_available : (&FALSE if no data in the wanted range)
              $ra_header      : A ref to the array of headers
=cut
sub __GetHeaderOffset
{
	my $self = shift;
	my ($fh_file,$start,$end,$ra_headers, $indexed) = @_;
	
	my $relevant_offset = 0;

	# If the headers are in the file
	if ( ! defined $ra_headers )
	{
		my $line    = $self -> __NextLine ( $fh_file, &TRUE );
		chomp($line);
			
		if ( $line =~ s/^\@// ) # THIS IS AN INDEX !
		{
			$relevant_offset = $self -> __GetFilePosFromIndexline($line, $start, $end);
			$line = $self -> __NextLine ( $fh_file );  # Next line is the header
			chomp $line;
			if ( $relevant_offset > 0 )
			{
				seek $fh_file, $relevant_offset, &SEEK_CUR;
			}
		}
		$line =~ s/^#//;
		my @a_headers = split /$SEP/,$line;
		return (($relevant_offset!=-1), \@a_headers);
	}
	
	# If the headers are passed to Prepare
	else
	{
		if ( $indexed )
		{
			my $line    = $self -> __NextLine ( $fh_file, &TRUE );
			chomp($line);
	
			if ( $line =~ s/^\@// ) # THIS IS AN INDEX !
			{
				$relevant_offset = $self -> __GetFilePosFromIndexline($line, $start, $end);
			}
			if ( $relevant_offset > 0 )
			{
				seek $fh_file, $relevant_offset, &SEEK_CUR;
			}
		}
		return (($relevant_offset!=-1), $ra_headers);
	}
}

=head2 Function __GetFilePosFromIndexline

 Usage    : my $relevant_offset = &__GetFilePosFromIndexline($line, $start, $end);
 Function : Parse the index line and compute the offset from $start and the index line passed by parameter
 Args     : $line The index line
            $start The start position
            $end   The end position
 Return   : The relevant index, or -1 if nothing to read
 
=cut
sub __GetFilePosFromIndexline
{
	my $self = shift;
	my ( $index_line, $start, $end  ) = @_;

	my ( $chrom_start, $chrom_end, $slice_size_base, $nb_pos, $positions) = split ";",$index_line;
	my @a_indexes = split ":",$positions;
	#warn $#a_indexes;
	my $start_offset = $start - $chrom_start;
	my $relevant_offset;
	if ( $start > $chrom_end || $end < $chrom_start ) # nothing to read
	{
		$relevant_offset = -1;
	}
	elsif ( $start_offset > ($slice_size_base*&SLICE_UNIT) ) # If not we return 0
	{
		#my $index = int ( $start_offset / ($slice_size_base*&SLICE_UNIT) ) - 1; 
		my $index = int ( $start_offset / ($slice_size_base*&SLICE_UNIT) ); 
		#warn $index;
		$index =  $#a_indexes if ( $index > $#a_indexes ); # in case of (we have limited indeed the number of indexing cells to NUM_INDEX_SLICE = 40, change this ?)
		$relevant_offset=$a_indexes[ $index ];		
	}
	else
	{
		$relevant_offset = 0;
	}

	return $relevant_offset;
}

=head2 Function __NextLine

 Usage        : $self -> __NextLine ( $fh_file, $flgskip_slash );
 Function     : Reads the next line of the file handler passed to parametr, skipping comments or empty lines
 Prerequisite : $fh_file must be opened for read
 Args         : $fh_file 
                $flg_skip_slash If &TRUE (default &FALSE), lines starting with '/' are skipped
                The default is to return when a line starting with '/' is met
                This is useful to let concatenate several files together
 Returns 	  : The line
 
=cut
sub __NextLine
{
	my $self = shift;
	my ( $fh_file, $flg_skip_slash ) = @_;
	$flg_skip_slash = &FALSE unless defined $flg_skip_slash;
	while ( my $line = <$fh_file> )
	{
		next if ( $line eq "\n");
		next if ( $line =~ /^;/ );
		next if ( $line =~ /\// && $flg_skip_slash);
		return $line;
	}
}

#----------------------------------
#    Retrieving data
#---------------------------------

=head2 function __FetchRequestedData

 Usage        : $rh_data = $o_obj->__FetchHashref('key');
 Function     : Fetch the data according to the specified range condition and return type
 				Calls __ReadDataFromFile which does the job.
 Args         : NAMED ARGUMENTS:
 				return_type  => 'HASH'|'ARRAY' 
 				key_name   	 => column used as a the hash key
 				line_handler => Ref to a function called for each read line
 Returns      : a reference on the hash containing with the following keys
 				  ra_return 		: an array of the array refs (empty except when the return type is an array)
 				  rh_return 		: a hash with the arraya refs (empty except when return type is a hash)
 				  ra_headers		: an array with the field names
 				  ra_sorted_keys	: an array with the keys ordered as in the file (empty unless -keep_order is &TRUE)
 
=cut
sub __FetchRequestedData
{
	my $self = shift;
	my %h_args = ( 	
					return_type  => 'ARRAY',
				  	key_name	  => '__none',
				  	line_handler => undef,
				  	@_
				  );	

	my ( $return_type, $key_name, $rf_line_handler ) = @h_args{ qw( return_type key_name line_handler )};

	if ($key_name eq '__none' && $return_type eq 'HASH')
	{
		my $msg = " defined when a HASH return type is wanted";
		die with LipmError::ParameterException( 'key_name', 'value', -should_be => $msg);
	}
		
	$self->_Set( '_return_type'  => $return_type );		
	$self->_Set( '_key_name' 	 => $key_name );
	$self->_Set( '_line_handler' => $rf_line_handler );
		
	$self->__ReadSelectedDataFromFile( );	
			
	my @a_return_attributes      = qw( _rh_return _ra_return _rh_headers _ra_sorted_keys ); 
	my @a_return_attribute_names = qw( rh_data ra_data rh_headers ra_sorted_keys ); 
	#my @a_return_attributes      = qw( _rh_return _ra_return _rh_headers ); 
	#my @a_return_attribute_names = qw( rh_data ra_data rh_headers ); 
	my %rh_return;

	@rh_return{ @a_return_attribute_names } = $self->_Get( @a_return_attributes );
	
	return \%rh_return;	
}

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallHashref( 'key' );
                $rh_data = $o_obj->FetchallHashref( 'key', -line_type => 'HASH', -keep_order => &TRUE );
 Function     : Return the data read in a hash, the key is the value found in column 'key'
 Returns      : a reference on the hash containing the data 
 Args         : $key_name : header field used as a the hash key
                -line_type  [HASH/ARRAY, default ARRAY]    : The line type to use
                -keep_order [&FALSE/&TRUE, default &FALSE] : If set, we keep the list of sorted keys, ie same order as in the file

=cut
sub FetchallHashref
{
	my $self     = shift;
	my $key_name = shift;
	my %h_args = (
	    -line_type  => 'ARRAY',
	    -keep_order => &FALSE, 
        @_
    );
    $self -> _Set ( _line_type  => $h_args{-line_type});
    $self -> _Set ( _keep_order => $h_args{-keep_order});
    $self -> _AssertLineType();
 				 
	(defined $key_name)
		or croak "ERROR - A key name must be provided for FetchallHashref";

	$self->__FetchRequestedData( return_type => 'HASH', key_name => $key_name );
	return $self->_Get( '_rh_return' );
}# FetchallHashref

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallLineByLine( \&LineHandler )
                $rh_data = $o_obj->FetchallLineByLine( \&LineHandler, -line_type => 'HASH' );

 Function     : Call LineHandler at each line - This is probably slowest, but consumes less memory thant FetchallHashref or FetchallArrayref
 Returns      : a reference on the hash containing the data... except that it does not contain anything
 Args         : \&LineHandler  The LineHandler, the line ( a $ra_ or $rh_, following the return_type) is passed by parameter to the handler
                -line_type [HASH/ARRAY, default ARRAY] : The line type to use

=cut
sub FetchallLineByLine
{
	my $self            = shift;
	my $rh_line_handler = shift;
	my %h_args = (
	    -line_type => 'ARRAY',
        @_
    );
    $self -> _Set ( _line_type => $h_args{-line_type});
    $self -> _AssertLineType();

	if ( ! defined $rh_line_handler || ref($rh_line_handler) ne "CODE" )
	{
		croak "ERROR - The line handler is NOT or BADLY defined !";
	}
	$self->__FetchRequestedData( return_type => 'HANDLER', line_handler => $rh_line_handler );
	return $self->_Get( '_rh_return' );
}	

=head2 Function Tell
 Usage        : my $position = $o_reader -> Tell();
 Prerequisite : The file must be already opened
 Fonction     : Returns the current file positions - This is convenient for use with FetchallLineByLine, as the line handler
                can use this information to index data, for instance
 Args         : none
 Return       : The result of 
=cut

sub Tell
{
	my $self = shift;
	my $fh_file = $self -> _Get('_fh_file');
	if ( defined $fh_file )
	{
		return tell $fh_file;
	}
	else
	{
		croak "ERROR - The file is NOT opened, you cannot call Tell now !";
	}
}
	
	
=head2 function FetchallArrayref

 Usage        : $ra_data = $o_obj->FetchallArrayref();
                $ra_data = $o_obj->FetchallArrayref(-line_type => 'HASH' );
 Function     : Return the data read in an array
 Returns      : a reference on the array containing the data 
 Args         : -line_type [HASH/ARRAY, default ARRAY] : The line type to use

=cut
sub FetchallArrayref
{
	my $self = shift;
	my %h_args = (
	    -line_type => 'ARRAY',
        @_
    );
    $self -> _Set ( _line_type => $h_args{-line_type});
    $self -> _AssertLineType();
						
	$self->_Set( '_return_type' => 'ARRAY' ); 		   
	$self->__FetchRequestedData();
	return $self->_Get( '_ra_return' );
}# FetchallArrayref

1;
