#
# 	sebastien.letort@toulouse.inra.fr
#	Created: July 31, 2008
#   revised by emmanuel.courcelle@toulouse.inra.fr - Oct 2008
#   modified by thomas.faraut@toulouse.inra.fr - march 2009
#   stolen to the Narcisse project by manu - Aug 2011
#	$Id$
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

	DataReader::Whole - Subclass of DataReader, this class maintains a cache, minimizing I/O

=head1  SYNOPSIS

	my $o_reader = New DataReader::Region();
	#Default start/end labels are 'start' and 'end'
	$o_ndr->Prepare( -from   => $file );

=head1 DESCRIPTION

	The whole file is read when Prepare is called. If Prepare is called several times with different parameters but with the same file,
	the file is NOT reread, because the Whole file (thus the module name) is read and stored in a cache
	This can be good for I/O minimization, but may consume a lot of memory for long files.

=head1 FILE FORMAT

    See DataReader.pm documentation. The indexation lines (@) are ignored if present

=cut

package DataReader::Whole;

use strict;
use warnings;

use base qw( DataReader );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;
use IO::File;


use LipmError::IOException;
use LipmError::ParameterException;

use General;

# A cache for file content, common to all objects of this class
# The key is the file name (see the switch -from in method Prepare)
# The value is an array of hashes, 1 hash / line

our %H_DATA_CACHE = ();

#---------------------------------
#    Reading the file
#---------------------------------

=head2 Procedure Prepare

 Usage        : $o_obj->Prepare( %h_args );
 Procedure    : then Read the file, filling the cache if necessary, then call the mother Prepare
 Args         : NAMED ARGUMENTS:
               	-from   => $file     The file name
               	-from_key  => Defaults to $file is $file is a filename, required if $file is a file handler
               	-where  => $rh_cond  A condition, expressed as a hash:
                    { VALUE => { col1_name => val1, col2_name => val2, ... },
               		  RANGE => { start => x, end => x, start_label => 'q_start', end_label => 'q_end' }
               		}
               	-case_sensitive => bool, set to &FALSE if you want the API to be insensitive.
                                   Only used with -where => { VALUE } condition.
                -force_reading  => bool, default false, clear the cache and reread the file if true
=cut
sub Prepare
{
	my $self = shift;
	my %h_args = (
        -from   => '-from option',	# dummy name, useful for error message generation
        -from_key => undef,
        -where  => {},
        -select => [],  
        -force_reading => 0,
        @_
        );

	# Fill the cache, only if necessary
	my $file = $h_args { -from };
	my $key  = $h_args { -from_key };
	if ( !defined $key )
	{
		if ( ref($file) eq '' )
		{
			$key = $file;
			$self -> _Set ( '_key', $key );
		}
		else
		{
			croak "ERROR - -from is a file handler, thus -key is REQUIRED !";
		}
	}
		
	delete $H_DATA_CACHE{ $key }  if ( 1 == $h_args{ -force_reading } );
	
	if( !defined $H_DATA_CACHE{ $key } )
	{
		my $ra_headers = $self -> _Get ( '_force_headers' );
		$H_DATA_CACHE{ $key } = $self->__ReadAllDataFromFile( $file, $ra_headers );
	}
	else
	{
		# Retrieve data from the cache
		$self->_Set(_ra_data        => $H_DATA_CACHE{$key}{ _ra_data },
					_ra_headers     => $H_DATA_CACHE{$key}{ _ra_headers },
					_rh_headers     => $H_DATA_CACHE{$key}{ _rh_headers },
					_ra_sorted_keys => [] );
	}

	$self->SUPER::Prepare( %h_args );

}

=head2 Procedure __ReadAllDataFromFile

 Usage        : $o_obj->__ReadAllDataFromFile( $file );
 Procedure    : Calls __ReadFileWrapper and retrieves the header fields from the first line
 				Each line is stored in an array
 Args         : $file The file name (string) OR an already opened file handler
                $ra_headers If defined, the force_header switch of Prepare
 Returns 	  : a reference on a hash containing the relevant information for subsequent data extraction
                _ra_data    : a reference on the array containing all the data lines
                _ra_headers : a reference on a table containing the header information
                _rh_headers : a reference on a hash containing the column numbers corresponding to the header fields

=cut

sub __ReadAllDataFromFile
{
	my $self    = shift;
	my( $file, $ra_headers ) = @_;

	my $ra_data = $self->__ReadFileWrapper( $file );
	
	# Skip the index line, if any
	if ( @$ra_data > 0 && $ra_data->[0]->[0] =~ /^\@/ )
	{
		shift @{$ra_data};
	}
	
	# If $ra_headers is defined, there are NO headers in the file
	if ( !defined $ra_headers )
	{
		if ( @$ra_data > 0 )
		{
			$ra_headers = shift @$ra_data;
		}
		else
		{
			$ra_headers = [];
		}
	}
	$self->_SetHeaderField( $ra_headers );

	$self -> _Set ( _ra_data => $ra_data );

	# Keep track of the three arrays in the cache
	my %h_cache_line = (
	                     _ra_data       => $ra_data,
	                     _ra_headers    => $self->_Get( '_ra_headers' ),
	                     _rh_headers    => $self->_Get( '_rh_headers' )
	                    );

	return  \%h_cache_line;

} # _ReadDataFromFile


=head2 Procedure __ReadFileWrapper

 Usage        : $o_obj->__ReadFileWrapper( $file );
 Procedure    : A wrapper to ReadFile enabling to switch easily between _ReadFile and NARCISSE_inlineC::ReadFile
 Args         : A file name OR an already opened file handler
 Returns 	  : same as __ReadFile
=cut

sub __ReadFileWrapper
{
	my $self   = shift;
	my ($file) = @_;

	my $ra_data ;
	my $fh_file;
	
	# If $file is not a reference, it should be a string (filename)
	if (ref($file) eq '')
	{
		$fh_file = GetStreamIn ( $file );
	}
	else
	{
		$fh_file = $file;
	}
	
	if ( &DataReader::IsInlineCEnabled() )
	{
		$ra_data = &DataReader_inlineC::__ReadFile( $self, $fh_file );
	}
	else
	{
		$ra_data = $self->__ReadFile( $fh_file );
	}

	return $ra_data;
}

=head2 Procedure __ReadFile

 Usage        : $o_obj->__ReadFile( $fh_file );
 Procedure    : Reads the Narcisse formated file and stores the lines
 				in an array of array ref
 				Each line is transformed into an array using the split function,
 				the returned arrays contiains reference on those line arrays
 Args         : A file handle (return from IO::File, or from GetStreamIn, etc.)
 Returns 	  : a reference on the data array containing the whole file
=cut
sub __ReadFile
{
	my $self   = shift;
	my ($fh_file) = @_;
	my @a_data = ();
	my $flg_starting = &TRUE;
	
	while( my $line=<$fh_file> )
	{
		chomp($line);
		next if ( $line =~ /^;/ ); # Comments = Start with ;
		next if ( $line eq  ""  ); # Ignore blank lines
		next if ( $line =~ /^\// && $flg_starting );	# ignore /path/to/file if FIRST line
		last if ( $line =~ /^\// );						# exit when /path/to/file if not FIRST line
		my @a_row = split( /$DataReader::SEP/, $line );
		push( @a_data, \@a_row );
		$flg_starting = &FALSE;
	}
	return \@a_data;
}

#----------------------------------
#    Retrieving data
#---------------------------------

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallHashref( 'key' );
                $rh_data = $o_obj->FetchallHashref( 'key', -line_type => 'HASH', -keep_order => &TRUE );
 Function     : Return the data read in a hash, the key is the value found in column 'key'
 Returns      : a reference on the hash containing the data 
 Args         : key_name : header field used as a the hash key
                -line_type [HASH/ARRAY, default ARRAY] : The line type to use

=cut

sub FetchallHashref
{
	my $self     = shift;
	my $key_name = shift;
	my %h_args = (
	    -line_type  => 'ARRAY',
 	    -keep_order => &FALSE, 
        @_
    );
    $self -> _Set ( _line_type => $h_args{-line_type});
    $self -> _Set ( _keep_order => $h_args{-keep_order});
    $self -> _AssertLineType();
 
	my $ra_headers  = $self->_Get( '_ra_headers' );
	(defined $key_name)
		or croak "ERROR - A key name must be provided for FetchallHashref";

	$self->_Set( '_return_type' => 'HASH' );
	$self->_Set( '_key_name' => $key_name );

	$self->__Fetchall( );

	my $rh_data = $self->_Get( '_rh_return' );

	return $rh_data;
}# FetchallHashref

=head2 function FetchallArrayref

 Usage        : $ra_data = $o_obj->FetchallArrayref();
                $ra_data = $o_obj->FetchallArrayref(-line_type => 'HASH' );

 Function     : Return the data read in an array
 Returns      : a reference on the array containing the data 
 Args         : -line_type [HASH/ARRAY, default ARRAY] : The line type to use

=cut
sub FetchallArrayref
{
	my $self = shift;
	my %h_args = (
	    -line_type => 'ARRAY',
        @_
    );
    $self -> _Set ( _line_type => $h_args{-line_type});
    $self -> _AssertLineType();

	$self->_Set( '_return_type' => 'ARRAY' );
	$self->__Fetchall();

	my $ra_data = $self->_Get( '_ra_return' );

	return $ra_data;
}# FetchallArrayref


=head2 Procedure __Fetchall

 Usage        : $self->__Fetchall
 Procedure    : Retrieves the data and stores it in a structure specified by _return_type
 				Each data line is transformed in a hash with the header fields as keys.
 				When a condition has been specified by prepare, only those lines verifying the condition are returned
                This method is designed for all Narcisse data files that are not of the type chromosome feature
 Args         : none
 Returns      : none

=cut

sub __Fetchall
{
	my $self = shift;

	#input and ouput data structures
	my @a_attr= qw( _ra_data _rh_return _ra_return _ra_sorted_keys _return_type _line_type _keep_order );
	my ( $ra_data, $rh_return, $ra_return, $ra_sorted_keys, $return_type, $line_type, $keep_order ) = $self->_Get( @a_attr );
	
	my $f_hash_return_type = ( $return_type eq 'HASH') ? 1 : 0;
	my $l_hash_type        = ( $line_type eq 'ARRAY')  ? 1 : 0;

	#Hash specific values
	# (REMOVED $ra_sorted_keys !) my ( $key_name_col, $ra_sorted_keys)=('','');
	# if ( $f_hash_return_type )# Doesn't need to be specified if return type is an ARRAY
	# {
	#	( $key_name_col, $ra_sorted_keys) = $self->_GetHashSpecificValues();
	#	
	#}
	my $key_name_col = '';
	if ( $f_hash_return_type )
	{
		$key_name_col = $self -> _GetHeaderColumn ( $self -> _Get ( '_ra_headers'), $self -> _Get ( '_key_name' ) );
	}

	# Fields selection, only for hash storage mode
	my $ra_headers        = $self->_Get ( '_ra_headers' );
	my $ra_wanted_headers = $self->_Get ( '_ra_select' );
	if ( @$ra_wanted_headers == 0 )
	{
		$ra_wanted_headers = $ra_headers;
	}
	my @a_wanted_col = $self -> _GetHeaderColumn( $ra_headers, @$ra_wanted_headers );

	# Condition specific values (if any)
	my( $rh_where_condition, $case_sensitive ) =
						$self->_Get( qw( _rh_where_condition _case_sensitive ) );
	my $b_value_conditions_exist = $rh_where_condition->{ value };
	my $b_range_conditions_exist = $rh_where_condition->{ range };
	

	my ( $ra_col_names_number, $ra_vals ) =$self->_GetCondSpecificValues();
	
	#Loop control values stored in scalars for performance issue
	my ( $start_col, $end_col, $start_value, $end_value ) = (0,0,0,0);
	if ( $b_range_conditions_exist )
	{
		my $rh_condition = $self->_Get( '_rh_range_condition' );
		my $start_label  = $$rh_condition{ start }{ label };
		my $end_label    = $$rh_condition{ end }{ label };
		$start_value     = $$rh_condition{ start }{ value };
		$end_value       = $$rh_condition{ end }{ value };
		my $ra_headers   = $self->_Get( '_ra_headers' );
	   ($start_col, $end_col) = $self -> _GetHeaderColumn( $ra_headers, $start_label, $end_label );
   }

	my $i = $self->__GetFirstRelevantLineIndex();
	while( my $ra_row = $ra_data->[$i++] )
	{
		if ( $b_range_conditions_exist )
		{
			next if( $$ra_row[ $end_col ] < $start_value );
			last if ( $end_value < $$ra_row[ $start_col ] );
		}

		if ( $b_value_conditions_exist )
		{
			next if ( 0 == $self -> _ValidateValueConditions($ra_row, $ra_col_names_number, $ra_vals, $case_sensitive) );
		}

		# 4 possibilities: line = ARRAY/HASH, while file = ARRAY/HASH
		if ( $l_hash_type )
		{
			if ( !$f_hash_return_type )
			{
				push @{$ra_return}, $ra_row;
			}
			else
			{
				my $key_val = $$ra_row[ $key_name_col ];
				push (@{$ra_sorted_keys}, $key_val) if ( $keep_order );
				$$rh_return{ $key_val } = $ra_row;
			}
		}
		else
		{
			my %h_row;
			@h_row{ @$ra_wanted_headers } = @{$ra_row}[ @a_wanted_col ];
			if ( !$f_hash_return_type )
			{
				push @{$ra_return}, \%h_row;
			}
			else
			{
				my $key_val = $$ra_row[ $key_name_col ];
				push (@${ra_sorted_keys}, $key_val) if ( $keep_order );
				$$rh_return{ $key_val } = \%h_row;
			}
		}
	}
	return;
}# __Fetchall


=head2 Function __GetFirstRelevantLineIndex

 Usage        : $self->__GetFirstRelevantLineIndex( );
 Function     : Returns the index of the first relevant data line according to the range conditions
 Prerequisite : The file must be loaded in memory, as an array of arrays
 Args         : none
=cut
sub __GetFirstRelevantLineIndex
{
	my $self = shift;
	
	my( $ra_data, $rh_headers, $rh_where_condition, $rh_range_condition ) = $self->_Get( qw( _ra_data _rh_headers _rh_where_condition _rh_range_condition ) );
	
	my $n = scalar @$ra_data;
	
	#If there is no data or no range conditions there is no need to dichotomize :-)
	return 0 unless ( $n > 0 && $rh_where_condition->{ range } eq &TRUE);
	
	my $start   = $$rh_range_condition{ start }{ value };
	my $end     = $$rh_range_condition{ end   }{ value };

	my $s_label = $$rh_range_condition{ start }{ label };
	my $start_i = $$rh_headers{ $s_label };

	my $e_label = $$rh_range_condition{ end   }{ label };
	my $end_i   = $$rh_headers{ $e_label };

	#We just dichotomize to the first relevant line
	
	my ( $left, $right ) = ( 0, $n-1 );
	while ( $left != $right )
	{
		my $middle = ($left + $right)/ 2;
		my $floor_middle = floor( $middle );
		if ( $start < $$ra_data[ $floor_middle ][ $end_i ] )
		{
			$right = $floor_middle;
		}
		else
		{
			$left = ceil ( $middle );
		}
	}
	
	return $left;
}

=head2 Procedure Finish

 Title        : Finish
 Usage        : $o_xx->Finish( none );
 Procedure    : Empty the memory cache corresponding to the LAST Prepared file
 Args         : none

=cut
sub Finish
{
	my $self = shift;	
	my $key  = $self->_Get( '_key' );
	
	if ( defined $key )
	{
	    delete $H_DATA_CACHE{ $self->_Get( '_key' )  };
	}
	
	return;
}

=head2 Procedure Finish

 Title        : ClearCache
 Usage        : $o_xx->ClearCache();
 Procedure    : Clear the cache for all files
 Args         : none

=cut
sub ClearCache
{
	my $self = shift;
	%H_DATA_CACHE = ();
}

1;
