#
# $Id$
#

package Dotplot;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Dotplot - different methods for dotplots

=head1 SYNOPSIS

my $o_dotplot = New Dotplot(reference	=> org_short_name, target => org_short_name , program => (prbh|glint), level => integer);

=head1 DESCRIPTION

=cut

use Carp;
use Data::Dumper;

use Image;
use NARCISSE_utils;
use Planner::GenomeMap;
use SpeciesObserver;
use General;


use base qw( LipmObject );
BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1428 $ =~ /\d+/g); $r[0]};
}

#For the where (to put the files) option
use constant CACHE   => 'cache';
use constant SESSDIR => 'sessdir';

#For the mode option
use constant DEFAULT =>  'default';
use constant WITH_PERSONAL_DATA  =>  'perso';
use constant PRINTABLE  =>  'printable';

#For the style option
use constant TRANSPARENT => 'transparent';
use constant OPAQUE 	 => 'opaque';

use constant ALL => "__all";

use constant PI => 3.14159;
use constant ALL_CHROM => 'all';

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_dotplot = New Dotplot(reference	=> "human", target => "chick", program => "glint", level => 4);
 Prerequisite : none
 Returns      : none
 Args         : Mandatory: reference  the reference orgamnism
                Mandatory: target     the target orgamnism
				Mandatory: program    the program used for the comparison
 Globals      : none

=cut
sub _Init
{
	my $self = shift;

	my %h_param = (
        o_io        => undef,
        reference	=> undef,
        target		=> undef,
        program		=> undef,
        level		=> undef,
        ref_chroms  => [],
        target_chroms => [],
        type		=> 'unset',
        -mode       => &DEFAULT,  # For a printable image we place it in sess dir and make it opaque
        -where		=> &CACHE,    # Where the dotplot files (image and coord) will be written and searched for (specified in the constructor only for debugging)
        @_
        );

	#0 Checking mandatory parameters
	# mandatory parameters
	foreach my $required ( qw( reference target program level o_io ) )
	{
		if ( ! defined($h_param{$required}) )
		{
			Carp::croak("Fatal error in constructor: $required must be provided to the constructor\n");
		}
	}

	$self->_Set( '__o_io'    => $h_param{o_io} );
	$self->_Set( '__program' => $h_param{program} );
	$self->_Set( '__level'   => $h_param{level} );

	# Complementary parameter
	$self->_Set( '__type'    => $h_param{type} );
	
	my $o_ref = New SpeciesObserver( species     => $h_param{ reference },
									 -o_inst_obs => $h_param{ o_io } );
	my $o_tgt = New SpeciesObserver( species     => $h_param{ target },
									 -o_inst_obs => $h_param{ o_io } );
	$self->_Set( __o_reference => $o_ref,
				 __o_target    => $o_tgt );

	#1 Setting the chromosome parameters
	$self->__SettingChromosomeSet(\%h_param);
	
		
	$self->__SetDotplotType();			
	$self->__SetDotplotCutoff();

	$self->__SetDotplotNameAndPath ( %h_param );
	
	$self->_Set ( '_framework' => {} );  # Structure maintaining the coordinate change information per species (reference target)
	
	$self->_Set ( '_mode' =>  $h_param{ -mode } )

}

=head2 function _SettingChromosomesSet

 Title        : _SettingChromosomesSet
 Usage        : _SettingChromosomesSet(\%h_param)
 Function     : select a set of chromosomes to be displayed according to the ref_chroms and target_chroms parameters
 Returns      : none
 Args         : a reference on the param hash

=cut
sub __SettingChromosomeSet
{
	my $self = shift;
	my $rh_param = shift;

	my( $o_io, $o_ref, $o_tgt ) = $self->_Get( qw( __o_io __o_reference __o_target ) );
	
	$self->{_ref_chrom_query} = "_unset";
	$self->{_target_chrom_query} =  "_unset";

	my $max_scaf = $o_io -> GetCfg( 'MAX_DISPLAYED_SCAFFOLDS' );
	if ( 0 != scalar grep(/all/,@{$rh_param->{ref_chroms}}) )
	{
		my $ra_chroms;
		if ($o_ref -> IsPseudo())
		{
			$ra_chroms = $o_ref -> GetChromsNames();
		}
		else
		{
			$ra_chroms = $o_ref -> GetChromsNames('SIZE',$max_scaf);
		}
		$self->{ref_chroms}=$ra_chroms;
		$self->{_ref_chroms_as_string}=&ALL_CHROM;
		$self->{_ref_chrom_query} = $rh_param->{ref_chroms}[0];		
	} else
	{
		$self->{ref_chroms}=$rh_param->{ref_chroms};
	}	

	if ( 0 != scalar grep(/all/,@{$rh_param->{target_chroms}}) )
	{
		my $ra_chroms;
		if ($o_tgt -> IsPseudo())
		{
			$ra_chroms = $o_tgt -> GetChromsNames();
		}
		else
		{
			$ra_chroms = $o_tgt -> GetChromsNames('SIZE',$max_scaf);
		}
		$self->{target_chroms}=$ra_chroms,
		$self->{_target_chroms_as_string}=&ALL_CHROM;
		$self->{_target_chrom_query} = $rh_param->{target_chroms}[0];		
	} else
	{
		$self->{target_chroms}=$rh_param->{target_chroms};
	}

	if ( $self->{_ref_chrom_query} =~ /all/ && $self->{_target_chrom_query} =~ /all/)
	{
		$self->{_query_type} = "soft"; 
	}
}

=head2 function __SetDotplotNameAndPath

 Title        : __SetDotplotNameAndPath
 Usage        : $self->__SetDotplotNameAndPath(\%h_param)
 Function     : Sets the name of the dotplot according to mode ( default, with personnal data or for a printable image)
 Returns      : none
 Args         : only optionnal arguments
 				

=cut

sub __SetDotplotNameAndPath
{
	my $self = shift;
	my %h_args = (  -mode  => &DEFAULT,
					-where => &CACHE,
					-style => &TRANSPARENT,
					@_
				);
	
	my ( $mode, $where, $style ) = @h_args{ qw( -mode -where -style) };	
	
	if ( $mode eq &WITH_PERSONAL_DATA )
	{
		$style = &TRANSPARENT;
		$where = &SESSDIR;
	}
	elsif ( $mode eq &PRINTABLE )
	{
		$style = &OPAQUE;
		$where = &SESSDIR;
	}
	$self->_Set( '_dotplot_name' => $self->ComputeDotPlotName( -style => $style, -mode => $mode ));
	$self->_Set ( '_where' => $where );

}

=head2 function ComputeDotPlotName

 Title        : ComputeDotPlotName
 Usage        : ComputeDotPlotName()
 Prerequisite :
 Function     : Computes the full basename of the dotplot
 Returns      : a string representing the name of the dotplot
 Args         : [optional] -style : transparent or opaque
                           -mode  : printable or personal 

=cut
sub ComputeDotPlotName
{
	my $self = shift;
	my %h_args =  ( -style => &TRANSPARENT,
					-mode  => &DEFAULT,
					@_
					 );	
	
	my $style = $h_args { -style };
	
	my $o_ref = $self -> _Get ( '__o_reference' );
	my $o_tgt = $self -> _Get ( '__o_target' );
	my $reference = $o_ref -> GetShortName();
	my $target    = $o_tgt -> GetShortName();
	
	my ( $program, $level)  = $self->_Get( qw(__program __level) );
	
	my $ref_chroms    = $self->__GetRefChromsAsString();
	my $target_chroms = $self->__GetTargetChromsAsString();

	my $dotplot_name = "dotplot_$style\_". $reference."_".$target."_".$ref_chroms.":".$target_chroms ."_".$level."_".$program;
	
	if ( $h_args{ -mode } eq &PRINTABLE ) #We oversimplify the name to avoid long names
	{
		$dotplot_name = "dotplot_". $reference."_".$ref_chroms."_vs_".$target."_".$target_chroms;
	}
	elsif ( $h_args{ -mode } eq &WITH_PERSONAL_DATA )
	{
		$dotplot_name.="_perso";	
	}
		
	return $dotplot_name;
}

=cut

=head2 function GetDotPlotPaths

 Title        : GetDotPlotPaths
 Usage        : ($path_dotplot,$path_coord_map,$http_dotplot)=$self->GetDotPlotPaths
 Prerequisite :
 Function     :
 Returns      : 3 strings :
 				$path_dotplot : the path to the dotplot image file
 				$path_coord_map : the path to the dotplot coord file
 				$http_dotplot : the http path to the dotplot image file
 Args         : a ParamParser object

=cut
sub GetDotPlotPaths
{
	my $self = shift;
	my $o_io      = $self->_Get( '__o_io' );
	
	my ( $root , $http );
	
	if ( $self->_Get( '_where' ) eq &CACHE )
	{
		$root = $o_io->GetCfg('ROOT_DOTPLOT_CACHE');
		$http = $o_io->GetCfg('HTTP_DOTPLOT_CACHE');
	}
	elsif ( $self->_Get( '_where' ) eq  &SESSDIR )
	{
		$root = $o_io->GetCfg('ROOT_SESSDIR');
		$http = $o_io->GetCfg('HTTP_SESSDIR');
	}
	else
	{	
		$root = $self->_Get( '_where' );
		$http = $self->_Get( '_where' );				
	}
	
	my $dotplot_name = $self->_Get( '_dotplot_name' );
	
	my $path_dotplot = "$root/$dotplot_name.png";
	my $path_coord_map = "$root/$dotplot_name.coord";
	my $http_dotplot = "$http/$dotplot_name.png";
	
	return ( $path_dotplot, $path_coord_map, $http_dotplot);
}

=head2 procedure ComputeDotPlots

 Title        : ComputeDotPlots
 Usage        : $rh_coord_map=&ComputeDotPlots()
 Prerequisite : Paramaparser object correctly initialized : see #0 Checking mandatory parameters
 Function     : Creates the dotplots in a new fashion (using the new Image object)
 Returns      : none
 Args         : none

=cut
sub ComputeDotPlots
{
 	my $self=shift;
    
    my $o_io      = $self->_Get( '__o_io' );
    
   	my $o_ref = $self -> _Get ( '__o_reference' );
	my $o_tgt = $self -> _Get ( '__o_target' );
	my $reference = $o_ref -> GetShortName();
	my $target    = $o_tgt -> GetShortName();
	my ( $program, $level ) = $self->_Get( qw( __program __level ) );
	
	my $rh_data_homologies=$self->__LoadDataHomologies();
	
	$self->_Set( '_rh_data' => $rh_data_homologies );
	
	# # Initialization of graphics parameters :
    Image->InitColorsLibrary($o_io->GetCfg('ROOT_COLORS_CFG'));
    Image->InitFontLibrary($o_io->GetCfg('ROOT_FONTS_CFG'),$o_io->GetCfg('ROOT_FONTS'));

    my $o_planner = Planner::GenomeMap -> New(
        o_io      => $o_io,
        reference => $reference,
        target	  => $target,
        level     => $level,
        program   => $program );

    $self->_Set( '_o_planner' => $o_planner);
	
	my $map_w	= $o_io->GetCfg('DOTPLOT_WIDTH');
	my $map_h	= $o_io->GetCfg('DOTPLOT_HEIGHT');
	my $margin	= $o_io->GetCfg('DOTPLOT_MARGIN');

	my $right_limit=$map_w-10;
	my $bottom_limit=$map_h-10;

	my $dotplot_layer='dotplot';

    my $o_image  = Image -> New (	
        frame_transf_type => "2D",
        width => $map_w,
        height	=> $map_h,
        layers	=> [$dotplot_layer]
        );

	$self->_Set( '__o_image' =>  $o_image );

	# Drawing the box and the chromosome pictures
	my $black	  = $o_image->colorAllocate("black",$dotplot_layer);
	my $dark_blue = $o_image->colorAllocate(0,0,150,$dotplot_layer);

	$o_planner->SetBox("dotplot_main",$margin,$right_limit,$margin,$bottom_limit);	
	$o_planner->SetBox("box_chrom_top",$margin,$right_limit,$margin/2,$margin-2);
	$o_planner->SetBox("box_chrom_left",$margin/2+5,$margin-2,$margin,$bottom_limit);
			
	my ($X0,$X1,$Y0,$Y1)=$o_planner->GetBoxCoordinates("dotplot_main");
	$o_image->rectangle( $X0,$Y0,$X1,$Y1,$black);

	my $ref_name    = $o_ref -> GetLongName();
	my $target_name = $o_tgt -> GetLongName();

	#Display reference name
	my $x_text=$margin+$o_planner->BoxMiddlex("dotplot_main")-length($ref_name)*10;
	my $y_text=25;
	$o_image->stringFT($dark_blue,$o_image -> font("georgiai"),14,0,$x_text,$y_text, $ref_name);

	#Display Target name
	$x_text=25;
	$y_text=$margin+$o_planner->BoxMiddley("dotplot_main")+length($target_name);
	$o_image->stringFT($dark_blue,$o_image -> font("georgiai"),12,PI/2,$x_text,$y_text,$target_name);

	if ($self->__GetDotPlotType() eq 'unset')
	{
 		$self->__SelectChromosomes( );	
	}

	if ($self->__NbTargetChroms()>0)
	{   
		$self->__DisplayTopChromosomes( );

		$self->__DisplayLeftChromosomes( );

 		$self->__DisplayGridPattern( );

 		$self->__DisplayConservedSegments( );
	}

	$self->__DisplayPersonnalData( );	
	
	my ( $path_dotplot, $path_coord_map, $http_dotplot ) = $self->GetDotPlotPaths();
	  
	$self->__PrintImage( image_file => $path_dotplot, coord_file => $path_coord_map );
	
    return ;

}

=head2 procedure __DisplayTopChromosomes

 Title        : __DisplayTopChromosomes
 Usage        : $rh_framework=__DisplayTopChromosomes($o_image,$rh_box,$rh_selected_chromosomes,$rh_chrom_infos)
 Function     : Constructs the top X-axis query chromosome charts and names
 Returns      : 0
 Args         : none

=cut
sub __DisplayTopChromosomes
{
	my $self = shift;
    
	my $rh_top_box 	= $self->_Get( '_o_planner' )->GetBoxHCoordinates("box_chrom_top");
	
	my $o_io = $self -> _Get( '__o_io' );
	my $o_ref= $self -> _Get( '__o_reference' );
    my $o_image = $self->_Get( '__o_image' );
 
	my %framework=();

	my $centro_size=$o_io->GetCfg('CENTRO_SIZE');
	my $chrom_thickness=$o_io->GetCfg('CHROM_THICKNESS');
	my $buffer_size=$o_io->GetCfg('CHROM_BUFFER_SIZE');

	my $font = $o_image -> gdSmallFont;
	my $black = $o_image->colorAllocate("black");
	my $grey = $o_image->colorAllocate("grey_dark");
	my $light_grey = $o_image->colorAllocate("grey_light");
	
	my $rh_chrom_infos = $o_ref -> GetChromsInfo();
	my $genome_size = 0;
	foreach my $chrom ($self->__GetRefChroms())
	{
		$genome_size+=	$rh_chrom_infos->{$chrom}{length};
	}
	my $start = $rh_top_box->{X0};
	my $end = 0;
	my $box_h_size = $rh_top_box->{X1} - $rh_top_box->{X0};
	my $box_v_size = $rh_top_box->{Y1} - $rh_top_box->{Y0};

	my $x_factor=($box_h_size-$buffer_size*($self->__NbRefChroms()-1))/$genome_size;
	$framework{factor}=$x_factor;

	my $gd_image=$o_image->GetLayer();

	foreach my $chrom (NARCISSE_utils::SortByChromID($self->__GetRefChroms()))
	{
		$framework{chrom}{$chrom}{start}=$start;
		$end=$start+$rh_chrom_infos->{$chrom}{length}*$x_factor;
		$framework{chrom}{$chrom}{end}=$end;
		$framework{chrom}{$chrom}{centro}=$start+$rh_chrom_infos->{$chrom}{centro}*$x_factor;

		my $rh_box={X0 => $framework{chrom}{$chrom}{start}, X1 => $framework{chrom}{$chrom}{end}, Y0 => $rh_top_box->{Y0}, Y1 => $rh_top_box->{Y1}};

		#Writing the chromosome name in a centered manner
		my $x_text=($rh_box->{X0}+$rh_box->{X1})/2-$font->width*length($chrom)/2;
		my $y_text=$rh_box->{Y0};
		$o_image -> absoluteString($font,$x_text,$y_text,$chrom,$black);

		$o_image -> SetFrameChange(%$rh_box,x0 => 1, x1 => $rh_chrom_infos->{$chrom}{length});

		if ( $rh_chrom_infos->{$chrom}{centro} == 1) #Information not available for centromere position
		{
			# Only a long arm
			my ($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle(1,$box_v_size-$chrom_thickness,$rh_chrom_infos->{$chrom}{length},$box_v_size,$grey);
			$framework{chrom}{$chrom}{long_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => 1, end => $rh_chrom_infos->{$chrom}{length}};
		}
		else
		{
			my $centro_start=$rh_chrom_infos->{$chrom}{centro}-$centro_size/2;
			my $centro_end=$rh_chrom_infos->{$chrom}{centro}+$centro_size/2;

			# Short arm (displayed on the left)
			my ($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle(1,$box_v_size-$chrom_thickness,$centro_start,$box_v_size,$grey);
			$framework{chrom}{$chrom}{short_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => 1, end => $centro_start};

			# Long arm (displayed on the right)
			($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle($centro_end,$box_v_size-$chrom_thickness,$rh_chrom_infos->{$chrom}{length},$box_v_size,$grey);
			$framework{chrom}{$chrom}{long_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => $centro_end, end => $rh_chrom_infos->{$chrom}{length}};
			
			#Displaying the centromere
			my $x_centro=($framework{chrom}{$chrom}{short_arm}{X1}+$framework{chrom}{$chrom}{long_arm}{X0})/2;
			my $y_centro=($framework{chrom}{$chrom}{short_arm}{Y0}+$framework{chrom}{$chrom}{long_arm}{Y1})/2;
			$gd_image->arc($x_centro,$y_centro,5,5,0,360,$black);
			$gd_image->fill($x_centro,$y_centro,$light_grey);
		}

		$end+=$buffer_size;
		$start=$end;
	}

	my $rh_framework = $self->_Get( '_framework');
	$rh_framework->{ reference } = \%framework;
	
	return 0;
}

=head2 function __DisplayLeftChromosomes
 Title        : __DisplayLeftChromosomes
 Usage        : __DisplayLeftChromosomes($o_image,$rh_box,$rh_selected_chromosomes,$rh_chrom_infos)
 Function     : Construcut the left Y-axis query chromosome charts and names
 Returns      : 0
 Args         : none

=cut
sub __DisplayLeftChromosomes
{
	my $self = shift;
    	
	my $rh_left_box = $self->_Get( '_o_planner' )->GetBoxHCoordinates("box_chrom_left");	
		
	my $o_io = $self->_Get( '__o_io' );
	my $o_tgt= $self->_Get( '__o_target');
    my $o_image = $self->_Get( '__o_image' );
	my %framework=();
	
	my $centro_size=$o_io->GetCfg('CENTRO_SIZE');
	my $chrom_thickness=$o_io->GetCfg('CHROM_THICKNESS');
	my $buffer_size=$o_io->GetCfg('CHROM_BUFFER_SIZE');

	my $font = $o_image -> gdSmallFont;
	my $black = $o_image->colorAllocate("black");
	my $grey = $o_image->colorAllocate("grey_dark");
	my $light_grey = $o_image->colorAllocate("grey_light");
	
	my $rh_chrom_infos = $o_tgt -> GetChromsInfo();
	my $genome_size = 0;
	foreach my $chrom ($self->__GetTargetChroms())
	{
		$genome_size+=	$rh_chrom_infos->{$chrom}{length};
	}

	my $start = $rh_left_box->{Y0};
	my $end = 0;
	my $box_h_size = $rh_left_box->{X1} - $rh_left_box->{X0};
	my $box_v_size = $rh_left_box->{Y1} - $rh_left_box->{Y0};

	my $y_factor=($box_v_size-$buffer_size*($self->__NbTargetChroms()-1))/$genome_size;
	$framework{factor}=$y_factor;

	my $gd_image=$o_image->GetLayer();
		
	foreach my $chrom (NARCISSE_utils::SortByChromID($self->__GetTargetChroms()))
	#foreach my $chrom ($self->__GetTargetChroms())
	{

		$framework{chrom}{$chrom}{start}=$start;
		$end=$start+$rh_chrom_infos->{$chrom}{length}*$y_factor;
		$framework{chrom}{$chrom}{end}=$end;
		$framework{chrom}{$chrom}{centro}=$start+$rh_chrom_infos->{$chrom}{centro}*$y_factor;

		my $rh_box={X0 => $rh_left_box->{X0}, X1 =>  $rh_left_box->{X1}, Y0 => $framework{chrom}{$chrom}{start}, Y1 => $framework{chrom}{$chrom}{end}};

		#Writing the chromosome name in a centered manner
		my $x_text=$rh_box->{X0};
		my $y_text=($rh_box->{Y0}+$rh_box->{Y1})/2-$font->height/2;
		$o_image -> absoluteString($font,$x_text,$y_text,$chrom,$black);

		$o_image -> SetFrameChange(%$rh_box,y0 => 1, y1 => $rh_chrom_infos->{$chrom}{length});

		if ( $rh_chrom_infos->{$chrom}{centro} == 1) #Information not available for centromere position
		{
			# Only a long arm
			my ($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle($box_h_size-$chrom_thickness,1,$box_h_size,$rh_chrom_infos->{$chrom}{length},$grey);
			$framework{chrom}{$chrom}{long_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => 1, end => $rh_chrom_infos->{$chrom}{length}};
		}
		else
		{
			my $centro_start= $rh_chrom_infos->{$chrom}{centro}-$centro_size/2;
			my $centro_end=$rh_chrom_infos->{$chrom}{centro}+$centro_size/2;

			# Short arm (displayed on the top)
			my ($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle($box_h_size-$chrom_thickness,1,$box_h_size,$centro_start,$grey);
			$framework{chrom}{$chrom}{short_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => 1, end => $centro_start};

			# Long arm (displayed on the bottom)
			($X0,$Y0,$X1,$Y1)=$o_image -> filledRectangle($box_h_size-$chrom_thickness,$centro_end,$box_h_size,$rh_chrom_infos->{$chrom}{length},$grey);
			$framework{chrom}{$chrom}{long_arm}={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, start => $centro_end, end => $rh_chrom_infos->{$chrom}{length}};

			#Displaying the centromere
			my $x_centro=($framework{chrom}{$chrom}{short_arm}{X1}+$framework{chrom}{$chrom}{long_arm}{X0})/2;
			my $y_centro=($framework{chrom}{$chrom}{short_arm}{Y1}+$framework{chrom}{$chrom}{long_arm}{Y0})/2;
			$gd_image->arc($x_centro,$y_centro,5,5,0,360,$black);
			$gd_image->fill($x_centro,$y_centro,$light_grey);

		}

		$end+=$buffer_size;
		$start=$end;
	}
	
	my $rh_framework = $self->_Get( '_framework');
	$rh_framework->{ target } = \%framework;

	return 0;
}

=head2 procedure __DisplayGridPattern

 Title        : __DisplayGridPattern
 Usage        : __DisplayGridPattern
 Prerequisite : called from DisplayDotPlot after  __DisplayTopChromosomes and __DisplayLeftChromosomes have been called
 Function     : Displays the dashed grid pattern of the genome dotplot
 Returns      : none
 Args         : exactly 4 arguments
 				$o_image :  an image object
 				$rh_box : a refecence on a { X0, Y0,...} formated box (the box where the dotplot will be displayed)
 				$rh_fwk_query_chrom : a reference on the hash as returned by __DisplayTopChromosomes
 				$rh_fwk_subject_chrom : a reference on the hash as returned by __DisplayLeftChromosomes
 Globals      : none

=cut
sub __DisplayGridPattern
{
	my $self = shift;
	
	my $rh_box 	= $self->_Get( '_o_planner' )->GetBoxHCoordinates("dotplot_main");
	
	my $o_image = $self->_Get( '__o_image' );
	
	my $rh_fwk=$self->{_framework};

	# Change image's frame to default value => next draw will be in pixel
	$o_image -> SetFrameChange();
	my $grey = $o_image->colorAllocate("grey_dark");

	#For each query chromsome on the x-axis a vertical dashed line separates them
	my $previous_chrom_end=0;
	foreach my $chrom (NARCISSE_utils::SortByChromID($self->__GetRefChroms()))
	{
		my $chrom_start=$rh_fwk->{'reference'}{chrom}{$chrom}{start};
		my $chrom_end=$rh_fwk->{'reference'}{chrom}{$chrom}{end};
		my $X0=($chrom_start+$previous_chrom_end)/2;
		$o_image ->dashedLine($X0,$rh_box->{Y0},$X0,$rh_box->{Y1},$grey) if ($previous_chrom_end!=0);
		$previous_chrom_end=$chrom_end;
	}

	#For each subject chromsome on the y-axis an horizontal dashed line separates them
	$previous_chrom_end=0;
	foreach my $chrom (NARCISSE_utils::SortByChromID($self->__GetTargetChroms()))
	#foreach my $chrom ($self->__GetTargetChroms())
	{
		my $chrom_start=$rh_fwk->{'target'}{chrom}{$chrom}{start};
		my $chrom_end=$rh_fwk->{'target'}{chrom}{$chrom}{end};
		my $Y0=($chrom_start+$previous_chrom_end)/2;
		$o_image ->dashedLine($rh_box->{X0},$Y0,$rh_box->{X1},$Y0,$grey) if ($previous_chrom_end!=0);
		$previous_chrom_end=$chrom_end;
	}

}

=head2 procedure __DisplayConservedSegments

 Title        : __DisplayConservedSegments
 Usage        : &__DisplayConservedSegments($o_image,$selected_chromosomes,$rh_fwk_query_chrom,$rh_fwk_subject_chrom,$rh_data)
 Prerequisite : __DisplayLeftChromosomes aned __DisplayTopChromosomes have been called
 Function     : Displays the conserved segments on the genome dotplot
 Returns      : none
 Args         : exactly 6 arguments
 				$o_image :  an image object
 				$rh_box : a refecence on a { X0, Y0,...} formated box (the box where the dotplot will be displayed)
 				rh_selected_chromosomes : a reference on a hash with the selected 'queyr' and 'subject' chromosomes as returned
 				                         by __SelectChromosomes
 				$rh_fwk  : a reference on the hash containg the hash returned by __DisplayTopChromosomes and the one by __DisplayLeftChromosomes
 				$rh_data ; a reference on a hash containing synteny segments as returnd by __LoadDataHomologies
 Globals      : none

=cut
sub __DisplayConservedSegments
{
	my $self = shift;
		
	my $rh_box 	= $self->_Get( '_o_planner' )->GetBoxHCoordinates("dotplot_main");
	
	my $o_image = $self->_Get( '__o_image' );
	
	$o_image -> SetFrameChange(%$rh_box);
	my $color = $o_image->colorAllocate("blue");
	my $gd_image=$o_image->GetLayer();
	
	my $rh_data = $self->_Get( '_rh_data' );
	
	my $rh_fwk=	$self->{_framework};
	my @a_fields = qw( element start end s_chrom s_start s_end sign matching);
	foreach my $chrom ($self->__GetRefChroms())
	{
		next unless (defined $rh_data->{$chrom}); # no data no conserved segments
		
		my $ref_chrom_start = $rh_fwk->{'reference'}{chrom}{$chrom}{start};
		my $ref_factor = $rh_fwk->{'reference'}{factor};

		my $rh_target_chroms = $rh_fwk->{'target'}{chrom};
		my $target_factor = $rh_fwk->{'target'}{factor};
		my $o_nd = $rh_data->{$chrom};
		my %h_labelindices = $o_nd->GetLabelIndices();
		foreach my $ra_elt ( $o_nd->GetSortedElements() )
		{
			# We inline the GetFields method for efficiency reason	
			#my ($elt_id, $start, $end, $s_chrom, $s_start, $s_end, $sign, $matching ) = $o_nd->GetFields( $match, @a_fields );
			my ($elt_id, $start, $end, $s_chrom, $s_start, $s_end, $sign, $matching ) = @{$ra_elt}[ @h_labelindices{ @a_fields } ];
			next unless (defined $rh_fwk->{'target'}{chrom}{$s_chrom});
			my $X0 = $ref_chrom_start + $start * $ref_factor;
			my $X1 = $ref_chrom_start + $end   * $ref_factor;

			my $target_chrom_start = $$rh_target_chroms{$s_chrom}{start};
			my $Y0=$target_chrom_start + $s_start * $target_factor;
			my $Y1=$target_chrom_start + $s_end * $target_factor;

			($Y0,$Y1)=($Y1,$Y0) if ($sign eq "-");
			$gd_image->line($X0,$Y0,$X1,$Y1,$color);

			#A very dirty trick to get a rectangular area
			($Y0,$Y1)=($Y1,$Y0) if ($sign eq "-");

			$rh_fwk->{'synteny'}{ $elt_id }={X0 => $X0,Y0 => $Y0,X1 => $X1,Y1 => $Y1, q_chrom =>  $chrom, 's_chrom' => $s_chrom, match => $matching, element => $elt_id};
		}

	}
}

=head2 __DisplayPersonnalData

 Usage        : $o_dotplot->__DisplayPersonnalData( )
 Prerequisite : the object image must have been created
 Function     : Prints the uploaded personal data 
 Returns      : none
 Args         : none
=cut

sub __DisplayPersonnalData
{
	my $self = shift;
		
   	my $o_species = $self->_Get( '__o_reference' );
   	my @a_perso_files = $o_species->GetPersoFiles( -existing_files => &TRUE );
	
	my @a_keys = qw( chrom start end color label );
	foreach my $perso_file ( @a_perso_files )
	{
		my ($rh_data, $ra_keys) =
				&NARCISSE_utils::ReadNarcisseTabFile( file => $perso_file );
		foreach my $rh_row ( values %$rh_data )
		{
			my( $chrom, $start, $end, $color, $label ) = @$rh_row{ @a_keys };
			$self->__PlotRef( $chrom, $start, $end, $color, $label );
		}
	}			
}

=head2 Procedure __PlotRef

 Usage        : $self->__PlotRef( )
 Prerequisite : __DisplayTopChromosomes must have been called first (it defines the _framework{'reference'} )
 Procedure    : Plots the uploaded personal data 
 Returns      : none
 Args         : self-explanatory
=cut
sub __PlotRef
{
	my $self = shift;
	my ( $chrom, $start, $end, $color, $label ) = @_;
	
	my $o_io = $self->_Get( '__o_io' );	
	
	my $rh_top_box 	= $self->_Get( '_o_planner' )->GetBoxHCoordinates("box_chrom_top");	
		
	my $o_image = $self->_Get( '__o_image');
	my $gd_image = $o_image->GetLayer();
	
	my $rh_ref_frame =	$self->{_framework}{'reference'};
	my $chrom_start = $rh_ref_frame->{chrom}{$chrom}{start};	
	my $factor = $rh_ref_frame->{ factor };
	
	my $chrom_thickness=$o_io->GetCfg('CHROM_THICKNESS');
	
	my $X0 = $chrom_start + $start*$factor;
	my $X1 = $chrom_start + $end*$factor;
	my $Y0 = $rh_top_box->{Y1} - $chrom_thickness -4;
	my $Y1 = $rh_top_box->{Y1} - $chrom_thickness -2;

	$gd_image->filledRectangle($X0, $Y0 ,$X1, $Y1, $o_image->colorAllocate($color));

}

=head2 __PrintImage

 Usage        : $self ->__PrintImage(image_file => $image_file, coord_file => $coord_file)
 Prerequisite : the object image must have been created
 Procedure    : Prints the image to a file and the corresponding coord map to the map file
 Args         : hash with
 				image_file => the image file
 				coord_file => the map file

=cut
sub  __PrintImage
{
	my $self = shift;
	my %h_args = ( 
        image_file => 'dotplot.png',
        coord_file => 'dotplot.coord',
        @_ );
     
    $self->__WriteMapCoordinates( $h_args{ coord_file }  );
	$self->_Get('__o_image') -> PrintImage( $h_args{ image_file }  );  
        
	return ;
}

=head2 function __WriteMapCoordinates

 Title        : __WriteMapCoordinates
 Usage        : $self -> __WriteMapCoordinates($query,$subject,$rh_framework)
 Prerequisite : __DisplayConservedSegments has been called
 Function     : Write in a file the necessary information for the map code
 Returns      : none
 Args         : exactly 1 argument
 				$outfile : The absolute path to the map file with coord as extension
 Globals      : none

=cut
sub __WriteMapCoordinates
{
	my $self = shift;
	my ( $outfile ) = @_;
    my $o_io      = $self->_Get( '__o_io' );
    
    my $MaxMapLines = $o_io->GetCfg( 'DOTPLOT_MAX_MAP_LINES' );
     
   	my $o_ref = $self -> _Get ( '__o_reference' );
	my $o_tgt = $self -> _Get ( '__o_target' );
	my $reference = $o_ref -> GetShortName();
	my $target    = $o_tgt -> GetShortName();
			
	my $rh_framework = $self->{_framework};

	my @coord_names=('X0','Y0','X1','Y1');
	my @arms=('short_arm','long_arm');
	my $out = new IO::File(">$outfile") or die "Cannot open $outfile for writing";

	my @species=('reference','target');
	my %species_names=('reference' => $reference,'target' => $target);
	foreach my $species (@species)
	{
		my	($query_name,$subject_name)=($species eq 'reference') ? @species_names{@species}: @species_names{reverse @species};
		foreach my $chrom (keys %{$rh_framework->{$species}{chrom}})
		{
			my @chrom_info=(quoted("DOTPLOT_CHROMOSOME_ARM"),quoted($subject_name),quoted($query_name),quoted($chrom));
			#my @chrom_info = ("DOTPLOT_CHROMOSOME_ARM", $subject_name, $query_name, $chrom);
			foreach my $arm (@arms)
			{
				next unless (defined $rh_framework->{$species}{chrom}{$chrom}{$arm});
				my @coord=@{$rh_framework->{$species}{chrom}{$chrom}{$arm}}{@coord_names};
				my @arm_range=( $rh_framework->{$species}{chrom}{$chrom}{$arm}{start},$rh_framework->{$species}{chrom}{$chrom}{$arm}{end} );

				if ($species eq 'reference')
				{
					&VerticalDilatation(\@coord);
				}
				else
				{
					&HorizontalDilatation(\@coord);
				}
				print $out join("\t",(@coord,@chrom_info,@arm_range))."\n";
			}
		}
	}

	if ($self->__GetDotPlotType() eq "one2one")
	{
		foreach my $match_id (keys %{$rh_framework->{'synteny'}})
		{
			my $match=$rh_framework->{'synteny'}{$match_id};
			next unless $match->{match}>50;
			my @coord=map( int, @$match{@coord_names} );
			my @match_info=(quoted("DOTPLOT_SYNT"),quoted($match->{element}),quoted($species_names{reference}),quoted($match->{q_chrom}),quoted($species_names{target}),quoted($match->{s_chrom}));
			#my @match_info=( "DOTPLOT_SYNT", $match->{element}, $species_names{reference}, $match->{q_chrom}, $species_names{target}, $match->{s_chrom} );
			print $out join("\t",(@coord,@match_info))."\n";
		}
	}
	else
	{
		my @query_chroms = keys %{$rh_framework->{'reference'}{chrom}};
		my @target_chroms = keys %{$rh_framework->{'target'}{chrom}};
		
		my $predicted_number = (scalar @query_chroms)*(scalar @target_chroms);
		
		if ( $predicted_number > $MaxMapLines )
		{
			warn("Too many map coordinates for the doplot boxes.... skipped \n");
			return 0;
		}
		
		#The mapping coordinates for the dotplot boxes
		foreach my $query_chrom (keys %{$rh_framework->{'reference'}{chrom}})
		{
			foreach my $subjt_chrom (keys %{$rh_framework->{'target'}{chrom}})
			{
				my @chrom_info=(quoted("DOTPLOT_CHROM2CHROM"),quoted($species_names{reference}),quoted($query_chrom),quoted($species_names{target}),quoted($subjt_chrom));
				#my @chrom_info=( "DOTPLOT_CHROM2CHROM", $species_names{reference}, $query_chrom, $species_names{target}, $subjt_chrom );
				my ($X0,$Y0)=($rh_framework->{'reference'}{chrom}{$query_chrom}{start},$rh_framework->{'target'}{chrom}{$subjt_chrom}{start});
				my ($X1,$Y1)=($rh_framework->{'reference'}{chrom}{$query_chrom}{end},$rh_framework->{'target'}{chrom}{$subjt_chrom}{end});
				my @coord=(int($X0),int($Y0),int($X1),int($Y1));
				print $out join("\t",(@coord,@chrom_info))."\n";
			}
		}
	}

	return 0;
}

=head2 function ConstructPrintableImage

 Title        : ConstructPrintableImage
 Usage        : ConstructPrintableImage()
 Prerequisite : The actual dotplot image should exist
 Function     : creates an image for printing purpose
 Returns      : the http path to the constructed image
 Args         : none

DEPRECATED

#=cut
sub ConstructPrintableImage
{
	my $self = shift;

	my ($path_dotplot,$path_coord_map,$http_dotplot)=$self->GetDotPlotPaths();

	my $o_image = Image->New($path_dotplot);
	
	warn "We are going through here \n";
	
	my ($path_printable_dotplot,$http_printable_dotplot)=$self->_ComputePrintableDotPlotPaths(); 
	
	$o_image->__PrintImage($path_printable_dotplot);
	
	return $http_printable_dotplot;							
}

=cut

=head2 Procedure HorizontalDilatation,VerticalDilatation

 Title        : HorizontalDilatation,VerticalDilatation
 Usage        : HorizontalDilatation($ra_coord)
                VerticalDilatation($ra_coord)
 Prerequisite : none
 Procedure    : add a few pixels to the area vertically
 Args (modified) : An array of integers, representing 2 points
 NOTE - This is NOT a method

=cut
sub HorizontalDilatation

{	
	my $ra_coord = shift;
	$ra_coord->[0]-=6;
	$ra_coord->[2]+=5;
}
sub VerticalDilatation
{	
	my $ra_coord = shift;

	$ra_coord->[1]-=6;
	$ra_coord->[3]+=5;
}

=head2 function quoted

 Title        : quoted
 Usage        : $quoted_string=&quoted($string)
 Prerequisite : none
 Function     : add protected quotes for the purpose of WriteMapCoordinates
 Returns      : a quoted string
 Args         : a string
 Globals      : none

=cut
sub quoted
{
	my $string =shift;

	return "\"$string\"";

}

=head2 procedure __SelectChromosomes

 Title        : __SelectChromosomes
 Usage        : $self -> __SelectChromosomes()
 Prerequisite : LoadatHomologies has been called
 Procedure    : select the chromosomes that are relevant for a dotplot display
				Change the fields target_chroms and ref_chroms according to this selection
 Args         : none
 Globals      : none

=cut
sub __SelectChromosomes
{
	my $self=shift;
	
	# synteny segments as returned by __LoadDataHomologies
	my $rh_data = $self->_Get( '_rh_data' );
	
	my %h_selected_chrom=();
	
	my $o_ref = $self -> _Get( '__o_reference' );
	my $rh_chrom_info = $o_ref -> GetChromsInfo();
	
	$h_selected_chrom{ target } = {};
	#foreach my $chrom (keys %$rh_chrom_info)
	foreach my $chrom ($self -> __GetRefChroms())
	{
		next if( !defined $$rh_data{ $chrom } );
		my %h_selected_tgt_chroms = $self->__MatchCriteria( $$rh_data{$chrom});
		my @a_keys = keys %h_selected_tgt_chroms;
		if( 0 != @a_keys )
		{
			$h_selected_chrom{ reference }{ $chrom } = 1;
			@{ $h_selected_chrom{ target } }{ @a_keys } = @h_selected_tgt_chroms{ @a_keys };
		}
	}
	
	my @ref_chroms;
	foreach my $chr ($self -> __GetRefChroms())
	{
		push @ref_chroms,$chr if (exists $h_selected_chrom{reference}{$chr});
	}
	$self -> { ref_chroms } = \@ref_chroms;
	
	my @tgt_chroms;
	foreach my $chr ($self -> __GetTargetChroms())
	{
		push @tgt_chroms,$chr if (exists $h_selected_chrom{target}{$chr});
	}
	$self -> { target_chroms } = \@tgt_chroms;

#	@{$self->{target_chroms}}= keys %{$h_selected_chrom{target}};
#	@{$self->{ref_chroms}}= keys %{$h_selected_chrom{reference}};
	
#	return 	\%h_selected_chrom;
}

=head2 function __MatchCriteria

 Title        : __MatchCriteria
 Usage        :  %selected_subj_chroms= __MatchCriteria($rh_data->{$chrom});
 Prerequisite : This function is called by __SelectChromosomes
 Function     : according to a given criteria decides which target chromosomes are selected for a given reference chromosome
 Returns      : a hash with the selected target chromosomes
 Args         : a NarcisseData object containing the synteny segments associated with a query chromosome
 Globals      : none

=cut
sub __MatchCriteria
{
	my $self=shift;
	my $o_nd =shift;
	my %target_chrom=();
	my %selected_chrom=();

	my %h_labelindices = $o_nd->GetLabelIndices();

	my @a_fields= qw( s_chrom matching );
	foreach my $ra_elt ( $o_nd->GetSortedElements() )
	{
		# We inline the GetFields method for efficiency reason	
		#my ( $s_chrom, $matching ) = $o_nd->GetFields( $ra_elt, @a_fields );
		my ( $s_chrom, $matching ) = @{$ra_elt}[ @h_labelindices{ @a_fields } ];
		$target_chrom{ $s_chrom } += $matching;
	}

	foreach my $chrom (keys %target_chrom)
	{
		$selected_chrom{$chrom}++ if ($target_chrom{$chrom}>$self->{__DotplotCutoff});
	}

	return 	%selected_chrom;
}

=head2 function __LoadDataHomologies

 Title        : __LoadDataHomologies
 Usage        : __LoadDataHomologies($orga_ref, $orga_target)
 Prerequisite : none
 Function     : Read the synteny segments from the synteny dir
 Returns      : a hash with the correspoding information where keys are chromosomes 
 				and values NarcisseData objects corresponding to the chromosomes.
 Args         :
 				$orga_ref : the first organisme short name
 				$orga_target : the second organisme short name
 				$zoom : the zoom information (0:0 or chrQ:chrS)
 				$program :  the name of the program ( glint, prbh)
 				$level : the level of conservation

=cut
sub __LoadDataHomologies
{
	my $self = shift;

	my $o_io    = $self->_Get( '__o_io' );
	
	my $o_ref = $self -> _Get ( '__o_reference' );
	my $o_tgt = $self -> _Get ( '__o_target' );
	my $reference = $o_ref -> GetShortName();
	my $target    = $o_tgt -> GetShortName();
	
	my ($program, $level ) = $self->_Get( qw( __program __level ) );
	
	my @intitule;
	my $rh_homologies;
	my $crap;

	my %h_homology;
	my %h_params = (
		prog      => $program,
		level     => $level,
		reference => $reference,
		target    => $target,
		-existing_files => &TRUE );
# on met check a FALSE pour obtenir le liste de tous les fichiers
# en contre partie, on doit tester leur existence.

	# Only one chromsome synteny file has to be read
	if( $self->__GetDotPlotType() eq "one2one" )
	{
		my $ref_chrom = $self->__GetRefChrom() ;
		my $filepath  = $o_ref->GetSyntFiles( %h_params,
											-chrom => $ref_chrom );
		$h_homology{$ref_chrom}=&NARCISSE_utils::ReadChromFeatureData( file => $filepath, return_type => 'ARRAY', use_cache => 1  );
	}
	else # We have to read all the files
	{
		my @a_files = $o_ref->GetSyntFiles( %h_params );
		foreach my $filepath ( @a_files )
		{
			next if( !-e $filepath ); # pas de fichier, donc pas de donnees
			
			my $chrom = $o_io->ExtractChromNameFromFilename( $filepath );
			$h_homology{ $chrom } = &NARCISSE_utils::ReadChromFeatureData(
											file => $filepath,
											return_type => 'ARRAY',
											use_cache => 1  );
		}
	}

	return \%h_homology;
}

=head2 function DotPlotsAvailable

 Title        : DotPlotsAvailable
 Usage        : &DotPlotsAvailable($o_param)
 Prerequisite :
 Function     : Checks if the corresponding dotplot image and map are already in the cache directory
 Returns      : a boolean 1 or 0
 Args         : a ParamParser object
 Globals      :

=cut
sub DotPlotsAvailable
{
	my $self = shift;

	my ( $path_dotplot, $path_coord_map, $http_dotplot)=$self->GetDotPlotPaths();
		
	my $bool_dotplotsavailable=(-e $path_dotplot && -e  $path_coord_map) ? 1 : 0 ;

	return $bool_dotplotsavailable;
}

=head2 Function __GetRefChroms,__GetTargetChroms

 Title    : __GetRefChroms,__GetTargetChroms
 Usage    : my @chroms = $self->__GetRefChroms()
            my @chroms = $self->__GetTargetChroms()
 Function : get the reference/target chromosomes (after selection)
 Returns  : an array of chromosome names
=cut
sub __GetRefChroms
{
	my $self = shift;

	return @{$self->{ref_chroms}};
}

=head2 Function __GetTargetChroms

	Title      :	__GetTargetChroms	
	Usage      :	@chroms = $self->__GetTargetChroms()
	Function   :	get target chromosomes
	Returns    :	an array of chromosome names
=cut
sub __GetTargetChroms
{
	my $self = shift;

	return @{$self->{target_chroms}};
}

=head2 Function __GetRefChrom,__GetTargetChrom

 Title   : __GetRefChrom,__GetTargetChrom
 Usage   : my $chrom = $self->__GetRefChrom()
           my $chrom = $self->__GetTargetChrom()

 Function: get THE reference chromosome (should exist only 1 ref/target chrom)
 Returns : THE chromosome name
=cut
sub __GetRefChrom
{
	my $self = shift;

	my @chrom=@{$self->{ref_chroms}};
	if (scalar @chrom >1)
	{
		warning("There is more than one chromosome but only one is asked ");
	}
	return $chrom[0];
}
sub __GetTargetChrom
{
	my $self = shift;

	my @chrom=@{$self->{target_chroms}};
	if (scalar @chrom >1)
	{
		warning("There is more than one chromosome but only one is asked ");
	}
	return $chrom[0];
}

=head2 Function __NbRefChrom,__NbTargetChroms
 Function : Returns the number of chromosomes
=cut
sub __NbRefChroms
{
	my $self = shift;

	return (scalar @{$self->{ref_chroms}});
}
sub __NbTargetChroms
{
	my $self = shift;

	return (scalar @{$self->{target_chroms}});
}

=head2 Function __GetRefChromsAsString,__GetTargetChromsAsString

	Title   : __GetRefChromsAsString,__GetTargetChromsAsString
	Usage   : my $str = $self->__GetRefChromsAsString()
	          my $str = $self->__GetTargetChromsAsString()
	Function: Get a string representation of the reference/target chromosomes
	Returns : The string representation
	Args    : none

=cut
sub __GetRefChromsAsString
{
	my $self = shift;

	if (defined $self->{_ref_chroms_as_string}) 
	{
		return  $self->{_ref_chroms_as_string};
	};
	
	my @chroms = $self->__GetRefChroms();
    my $chroms_str=(scalar @chroms)	? join("-",@chroms) : 0;
	return $chroms_str;
}
sub __GetTargetChromsAsString
{
	my $self = shift;

	if (defined $self->{_target_chroms_as_string}) 
	{
		return  $self->{_target_chroms_as_string};
	};
	
	my @chroms = $self->__GetTargetChroms();
    my $chroms_str=(scalar @chroms)	? join("-",@chroms) : 0;
	return $chroms_str;
}

=head2 Procedure __SetDotplotCutoff

	Title      :	__SetDotplotCutoff
	Usage      :	$string_type = $self->SetDotplotCutoff()
	Prerequiste:	none
	Function   :	set the dutplot cut-off according to the program
	Returns    :	nothing
	Args       :	none
=cut
sub __SetDotplotCutoff
{
	my $self = shift;

    my $o_io   =  $self->_Get( '__o_io' );
	my $program = $self->_Get( '__program' );

	my $cutoff_key="DOTPLOT_CUTOFF_".uc($program);

	$self->{__DotplotCutoff}=($o_io->AreDefined($cutoff_key)) ? $o_io->GetCfg($cutoff_key) : $o_io->GetCfg('DOTPLOT_CUTOFF');

}

=head2 Procedure __SetDotplotType

	Title      :	__SetDotplotType
	Usage      :	$string_type = $self->SetDotplotType()
	Prerequiste:	none
	Function   :	get the type of the dotplot 'partial','one2one' or 'full'
						'one2one' : only 2 chromosomes are involved (1 for reference and 1 for target)
						'partial' : only a subset of ref and target chromosomes are considered
						'full'    : all chromosomes
	Returns    :	the target chromsome name
	Args       :	a reference on the dotplot hash
=cut

sub __SetDotplotType
{
	my $self = shift;

	my @ref_chroms = $self->__GetRefChroms();
	my @target_chroms = $self->__GetTargetChroms();
	my $nb_ref_chrom=scalar @ref_chroms;
	my $nb_target_chroms=scalar @target_chroms;

	if ( $nb_ref_chrom == 0 && $nb_target_chroms == 0 )
	{
		$self->_Set( '__type','unset' );
	}
	elsif ( $nb_ref_chrom==1 && $nb_target_chroms==1 )
	{
		$self->_Set( '__type','one2one' );
	}
	else
	{
		$self->_Set( '__type','partial' );
	}
	
	if ( defined $self->{_query_type} && $self->{_query_type} eq "soft" )
	{
		$self->_Set( '__type','unset' );
	}
	
}

=head2 Procedure __GetDotPlotType

=cut
sub __GetDotPlotType
{
	my $self = shift;

	return $self->_Get( '__type');

}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut


1;
