#
# 	emmanuelle.permal@toulouse.inra.fr
#	Created: 25 06 2013
#	$Id: DsDataAnalyser.pm 
#
# Copyright INRA/CNRS

# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package DsDataAnalyser;

=pod

=head1 NAME

DsDataAnalyser - An object that gives tools for DS data analysis

=head1  SYNOPSIS

my $o_dba = New DsDataAnalyser (species => $species, -o_inst_obs => $o_io, file => $file);

=head1 DESCRIPTION

This class gives tools for the analysis of DS data. Input file can be given by ncs_ComputeDs.pl.

=cut

use strict;
use warnings;
use Data::Dumper;
use base qw( LipmObject );
use Bio::Seq;
use Bio::SeqIO;
use Bio::DB::Fasta;
use FastaReader;
use General;
use FindBin;
use Cwd;
use File::Path qw( rmtree ); 
use Cwd 'abs_path';
use Runner;
use List::Util qw(max);
my $DEBUG = 0;

=head2 method _Init

 Title        : _Init
 Usage        : my $o_dsa = New DSAnalyser(species => $species, -o_inst_obs => $o_io, $ds_file, $gene_file);
 Prerequisite : none
 Fonction     : Init
 Returns      : $self
 Args         : $file (tabulated formatted as : seq1	seq2	S	dN	dS	dN/dS)
  
=cut

sub _Init
{
	my $self = shift;
	
	my %h_args = (  
					species => undef,
					-o_inst_obs => undef,
					ds_file => undef,
					gene_file => undef,
					block_file => undef,
					fam_file => undef,
					ds_data => undef,
					gene_data =>undef,
					block_data => undef,
					fam_data => undef,
					@_);

	$self -> SUPER::_Init( @_ );
	
	my ($species, $o_io, $ds_file, $gene_file, $block_file, $fam_file) = @h_args{ qw( species -o_inst_obs ds_file gene_file block_file fam_file) };

	my $path_to_data = abs_path ($ds_file);
	my $path_to_gene_data = abs_path ($gene_file);
	my $path_to_blk_file = undef;
	my $path_to_fam_file = undef;
	
	if ($block_file ne "")
	{
		$path_to_blk_file = abs_path($block_file);
	}
	if ($fam_file ne "")
	{
		$path_to_fam_file = abs_path($fam_file);
	}

	#Parse file and Create hash
	print "Creating Ds hash from DS file \n";
	my $rh_ds_data = $self -> _ParseDsFile ($path_to_data);
	$self -> _Set (	ds_data => $rh_ds_data);
	
	#Parse gene file and create hash
	print "Creating gene hash from gene file \n";
	my ($rh_gene_data, $ra_chr_list, $rh_geneid_protid) = $self -> _ParseGeneFile ($path_to_gene_data);
	$self -> _Set ( gene_data => $rh_gene_data);
	$self -> _Set (chr_list => $ra_chr_list);
	$self -> _Set ( gene_prot_data => $rh_geneid_protid);
	
	#~print Dumper (%$rh_gene_data);
	#Parse family file and create hash
	if (defined $path_to_fam_file)
	{
		print "Creating fam hash from fam file \n";
		my $rh_fam_data = $self -> _ParseFamilyFile ($path_to_fam_file);
		$self -> _Set ( fam_data => $rh_fam_data);
	}

	#Parsing Blocks file
	if (defined $path_to_blk_file)
	{
		print "Creating blk hash from blk file \n";
		my $rh_blocks = $self -> _ParseBlockFile ($path_to_blk_file);
		$self -> _Set (block_data =>$rh_blocks);
	}
	
	return $self;
} # _Init


=head2 Procedure _GetRefDsHash

 Usage        : $self -> __GetDsHash;
 Procedure    : Gets hash with all dS informations and returns it
 Args         : none
 Returns      : $rh_ds_data

=cut

sub _GetRefDsHash
{
	my $self = shift;
	return $self -> _Get ('ds_data');
}

=head2 Procedure _GetRefGeneHash

 Usage        : $self -> __GetRefGeneHash;
 Procedure    : Gets hash with all gene informations and returns it
 Args         : none
 Returns      : $rh_gene_data

=cut

sub _GetRefGeneHash
{
	my $self = shift;
	return $self -> _Get ('gene_data');
}


=head2 Procedure _GetRefBlockHash

 Usage        : $self -> _GetRefBlockHash;
 Procedure    : Gets hash with all block informations and returns it
 Args         : none
 Returns      : $rh_block_data

=cut

sub _GetRefBlockHash
{
	my $self = shift;
	return $self -> _Get ('block_data');
}

=head2 Procedure _GetRefFamilyHash

 Usage        : $self -> _GetRefFamilyHash
 Procedure    : Gets hash with all block informations and returns it
 Args         : none
 Returns      : $rh_family_data

=cut

sub _GetRefFamilyHash
{
	my $self = shift;
	return $self -> _Get ('fam_data');
}

=head2 Procedure _GetRefParaclusterHash

 Usage        : $self -> _GetRefParaclusterHash
 Procedure    : Gets hash with all paraclusters informations and returns it
 Args         : none
 Returns      : $rh_paracluster

=cut

sub _GetRefParaclusterHash
{
	my $self = shift;
	return $self -> _Get ('rh_paracluster');
}

=head2 Procedure _GetRefChrList

 Usage        : $self -> _GetRefChrList;
 Procedure    : Gets ref to array with all represented chromosomes
 Args         : none
 Returns      : $ra_chr_list

=cut

sub _GetRefChrList
{
	my $self = shift;
	return $self -> _Get ('chr_list');
}


=head2 Procedure _GetRefGeneProtHash

 Usage        : $self -> _GetRefGeneProtHash;
 Procedure    : Gets ref to array with all represented chromosomes
 Args         : none
 Returns      : $ra_chr_list

=cut

sub _GetRefGeneProtHash
{
	my $self = shift;
	return $self -> _Get ('gene_prot_data');
}

=head2 Procedure _ParseBlockFile

 Usage        : $self -> _ParseBlockFile();
 Procedure    : Fills a hash with ds info
 Args         : $file with ds informations
 Returns      : A hash with filled with ds information

=cut

sub _ParseBlockFile
{
	my $self = shift;
	my ($file) = (@_);
	
	my $fi_file = &GetStreamIn($file);
	my %h_block_data = ();
	
	my $headers = <$fi_file>;
	chomp ($headers);
	my @a_headers = split("\t", $headers);
	
	@a_headers = $self -> RemoveSpaceFromArray(@a_headers);
	
	foreach my $line (<$fi_file>)
	{
		
		my %h_line= ();
		chomp ($line);
		my @a_values = split ("\t", $line);
		
		@h_line{@a_headers} = @a_values;
		#print Dumper (%h_line);
		
		my $blockid = $h_line{$a_headers [0]}; #Patch PB hidden character mc vs unix
		
		$h_block_data {$blockid}={		index_start => $h_line {index_start},
										index_stop => $h_line {index_stop},
										chr => $h_line { chr },
										inverted => $h_line {inverted},
										type => $h_line {type}
										};
	}
	
	return \%h_block_data;
}


=head2 Procedure _ParseDsFile

 Usage        : $self -> _ParseDsFile();
 Procedure    : Fills a hash with ds info
 Args         : $file with ds informations
 Returns      : A hash with filled with ds information

=cut

sub _ParseDsFile
{
	my $self = shift;
	my ($file) = (@_);
	my $fi_file = &GetStreamIn($file);
	my %h_ds_data = ();
	
	my $headers = <$fi_file>;
	chomp ($headers);
	my @a_headers = split("\t", $headers);
	@a_headers = $self -> RemoveSpaceFromArray(@a_headers);
	my $index = 0;
	foreach my $line (<$fi_file>)
	{
		$index += 1;
		my %h_line= ();
		chomp ($line);
		my @a_values = split ("\t", $line);
		@h_line{@a_headers} = @a_values;
		my $couple_id = "pair_".$index;
		$h_ds_data {$couple_id} = {	seq1 => $h_line { seq1 },
									seq2 => $h_line { seq2 },
									dS => $h_line { dS}
									};
	}
		
	return \%h_ds_data;
}

=head2 Procedure _ParseGeneFile

 Usage        : $self -> _ParseGeneFile();
 Procedure    : Fills a hash with gene info
 Args         : $file with gene informations
 Returns      : A hash with filled with gene information

=cut

sub _ParseGeneFile
{
	my $self = shift;
	my ($gfile) = (@_);
	my $fi_gfile = &GetStreamIn($gfile);
	
	my $headers = <$fi_gfile>;
	chomp ($headers);
	my @a_headers = split("\t", $headers);
	@a_headers = $self -> RemoveSpaceFromArray(@a_headers);
	
	my %h_gene_data = ();
	my @a_chr_list = ();
	my %h_geneid_protid = ();
	my $last_chr="";
	foreach my $line (<$fi_gfile>)
	{
		my %h_line= ();
		chomp ($line);
		my @a_values = split ("\t", $line);
		@h_line{@a_headers} = @a_values;
		$h_gene_data {@h_line { qw (ProteinID) }} = {	GeneID => $h_line { GeneID},
														index => $h_line {index},
														chr => $h_line { chr },
														start => $h_line {start},
														end => $h_line { end},
														strand => $h_line {strand},
														nbexon => $h_line { nbexon}
													};
		$h_geneid_protid {$h_line { GeneID}} = $h_line { ProteinID};
		my $chr = $h_line { chr };
		if ($chr ne $last_chr)
		{
			push(@a_chr_list, $chr);
		}
		$last_chr = $chr;
	}
	
	return \%h_gene_data, \@a_chr_list, \%h_geneid_protid;
}

=head2 Procedure _ParseFamilyFile

 Usage        : $self -> _ParseFamilyFile();
 Procedure    : Fills a hash with gene info
 Args         : $file with gene informations
 Returns      : A hash with filled with gene information

=cut

sub _ParseFamilyFile
{
	my $self = shift;
	my ($fam_file) = (@_);
	my $rh_gene_prot = $self -> _GetRefGeneProtHash();
	
	my $fi_file = &GetStreamIn($fam_file);
	
	my $headers = <$fi_file>;
	chomp ($headers);
	my @a_headers = split("\t", $headers);
	@a_headers = $self -> RemoveSpaceFromArray(@a_headers);
	
	my %h_family_data = ();
	
	my $count = 0;
	foreach my $line (<$fi_file>)
	{
		my $mb_id = "";
		my %h_line= ();
		chomp ($line);
		my @a_values = split ("\t", $line);
		@h_line{@a_headers} = @a_values;
		my $mb_list = $h_line { qw (Members) };
		$mb_list =~ s/\]//;
		$mb_list =~ s/\[//;
		$mb_list =~ s/\s//g;
		my @a_members = split (',', $mb_list); 
		
		#patch sur ID :
		my @a_new_members =();
		foreach my $mb (@a_members)
		{
			$mb_id = $self -> _CheckFamilyAndGeneConcordance ($rh_gene_prot, $mb); #returns 1 if not in gene list
			next if ($mb_id eq "");
			push (@a_new_members, $mb_id);
		}
		#patch sur ID 
		
		if (scalar @a_new_members <= 1)
		{
			$count += 1;
			next;
		}
		
		$h_family_data {@h_line { qw (GeneTreeID) }} = \@a_new_members;
	}
	print "The process removed $count families \n";	
	return \%h_family_data;
}


=head2 Procedure GetRefDsHash
 
 Usage        : $o_dsa-> GetRefDsHash();
 Procedure    : Get the ref to the hash containing all ds info
 Args         : none
 Returns      : A reference to a hash
 
=cut

sub GetRefDsHash
{
	my $self = shift;
	return $self -> _GetRefDsHash();
}

=head2 Procedure GetRefParaclusterHash
 
 Usage        : $o_dsa-> GetRefParaclusterHash();
 Procedure    : Get the ref to the hash containing all ds info
 Args         : none
 Returns      : A reference to a hash
 
=cut

sub GetRefParaclusterHash
{
	my $self = shift;
	return $self -> _GetRefParaclusterHash();
}

=head2 Procedure GetRefGeneHash
 
 Usage        : $o_dsa-> GetRefGeneHash();
 Procedure    : Get the ref to the hash containing all Gene info
 Args         : none
 Returns      : A reference to a hash
 
=cut

sub GetRefGeneHash
{
	my $self = shift;
	return $self -> _GetRefGeneHash();
}
##################################TO TEST start #####################################
=head2 Procedure GetRefBlockHash
 
 Usage        : $o_dsa-> GetRefBlockHash();
 Procedure    : Get the ref to the hash containing all Gene info
 Args         : none
 Returns      : A reference to a hash
 
=cut
sub GetRefBlockHash
{
	my $self = shift;
	return $self -> _GetRefBlockHash();
}

##################################TO TEST end #####################################

=head2 Procedure RemoveSpaceFromArray

 Usage        : $odsa -> RemoveSpaceFromArray(@a_array);
 Procedure    : remove space from data in array
 Args         : @a_array
 Returns      : @a_array without spaces

=cut

sub RemoveSpaceFromArray
{
	my $self = shift;
	my ( @a_array_to_change)= (@_);
	my @a_changed_array = ();
	foreach my $mb (@a_array_to_change)
	{
		$mb =~ s/\s//;
		push (@a_changed_array, $mb);
	}
	return @a_changed_array;
}

=head2 Procedure GetProteinID

 Usage        : $o_dsa-> GetProteinID($gene_id);
 Procedure    : Get the Protein id
 Args         : $protein_id the protein id
 Returns      : The protein id name

=cut

sub GetProteinID
{
	my $self = shift;
	my ( $gene_id)= (@_);
	my $rh_gene_prot = $self -> _GetRefGeneProtHash();
	my $protein_id = $rh_gene_prot -> {$gene_id};
	
	return $protein_id;
}


=head2 Procedure GetProteinIDFromIndex

 Usage        : $o_dsa-> GetProteinIDFromIndex( $chr, $gene_index);
 Procedure    : Get the Protein id
 Args         : The chr $chr and the gene index $gene_index
 Returns      : The protein id name

=cut

sub GetProteinIDFromIndex
{
	my $self = shift;
	my ( $chr, $gene_index)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash;
	
	foreach my $id (keys %$rh_gene)
	{
		if ($chr eq $rh_gene ->{$id}->{chr})
		{
			if ($gene_index == $rh_gene ->{$id}->{index})
			{
				return $id;
			}
		}
		
	}
}

##################################TO TEST start #####################################
=head2 Procedure GetGeneID

 Usage        : $o_dsa-> GetGeneID($protein_id);
 Procedure    : Get the Gene id
 Args         : $protein_id the protein id
 Returns      : The protein id name

=cut

sub GetGeneID
{
	my $self = shift;
	my ( $protein_id)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash();
	my $gene_id = $rh_gene -> {$protein_id} -> {GeneID};
	
	return $gene_id;
}
##################################TO TEST end #####################################
#OK start#
=head2 Procedure GetGeneIndex

 Usage        : $o_dsa-> GetGeneIndex($protein_id);
 Procedure    : Get the Gene index
 Args         : $protein_id the protein id
 Returns      : The index of the protein id

=cut

sub GetGeneIndex
{
	my $self = shift;
	my ( $protein_id)= (@_);

	my $rh_gene = $self -> _GetRefGeneHash();
	my $gene_index = $rh_gene -> {$protein_id} -> {index};
	
	return $gene_index;
}


=head2 Procedure GetNbExon

 Usage        : $o_dsa-> GetNbExon($protein_id);
 Procedure    : Get the number of exons
 Args         : $protein_id the protein id
 Returns      : The number of exons of the protein id

=cut

sub GetNbExon
{
	my $self = shift;
	my ( $protein_id)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash();
	my $nb_exon = $rh_gene -> {$protein_id} -> {nbexon};
	
	return $nb_exon;
}

=head2 Procedure GetChrom

 Usage        : $o_dsa-> GetChrom($protein_id);
 Procedure    : Get the chromosome name
 Args         : $protein_id the protein id
 Returns      : The chromosome name of the protein id

=cut

sub GetChrom
{
	my $self = shift;
	my ($protein_id)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash();
	my $chr_name = $rh_gene -> {$protein_id} -> {chr};
	
	return $chr_name;
}

=head2 Procedure GetStrand

 Usage        : $o_dsa-> GetStrand($protein_id);
 Procedure    : Get the Strand
 Args         : $protein_id the protein id
 Returns      : The strand of the protein id

=cut

sub GetStrand
{
	my $self = shift;
	my ( $protein_id)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash();
	my $strand = $rh_gene -> {$protein_id} -> {strand};
	return $strand;
}

#OK end#

=head2 Procedure GetCoords

 Usage        : $o_dsa-> GetCoords($protein_id);
 Procedure    : Get the coordinates
 Args         : $protein_id the protein id
 Returns      : a ref to hash of the protein id start and end

=cut

sub GetCoords
{
	my $self = shift;
	my ( $protein_id)= (@_);
	my $rh_gene = $self -> _GetRefGeneHash();
	my %h_coords = ();
	my $start = $rh_gene -> {$protein_id} -> {start};
	my $end = $rh_gene -> {$protein_id} -> {end};
	%h_coords = (start => $start, end => $end);
	
	return \%h_coords;
}
#OK start#
=head2 Procedure DefineOrientation

 Usage        : $o_dsa-> DefineOrientation($ra_couple);
 Procedure    : Define the physical orientation betwen two protein of a couple
 Args         : the $ra_couple
 Returns      : parallel, divergent, convergent e.g. the orientation

=cut

sub DefineOrientation
{
	my $self = shift;
	my ( $couple_id)= (@_);
	my %h_type = ( t1 => "parallel", t2 => "divergent", t3=> "convergent");
	my $rh_ds_data = $self-> _GetRefDsHash();
	my $seq1 = $rh_ds_data -> {$couple_id} -> {seq1};
	my $seq2 = $rh_ds_data -> {$couple_id} -> {seq2};
	
	if ($self -> AreOnSameStrand ($seq1, $seq2))
	{
		return $h_type{t1} ;
	}
	else
	{
		my $index1 = $self-> GetGeneIndex ($seq1);
		my $index2 = $self-> GetGeneIndex ( $seq2);
		my $strand2 = $self->  GetStrand ($seq2);
		
		if ($index2 > $index1)
		{
			if ($strand2 == -1)
			{
				return $h_type{t3};
			}
			else
			{
				return $h_type{t2};
			}
		}
		else
		{
			if ($strand2 == -1)
			{
				return $h_type{t2};
			}
			else
			{
				return $h_type{t3};
			}
		}
	}
}


=head2 Procedure AreOnSameStrand

 Usage        : $o_dsa-> AreOnSameStrand($ra_couple);
 Procedure    :Tell about strand orientation for a couple
 Args         : the $ra_couple
 Returns      : Boolean 0 if false and 1 if true

=cut

sub AreOnSameStrand
{
	my $self = shift;
	my ( $id1, $id2)= (@_);
	my $strand1 = $self -> GetStrand($id1);
	my $strand2 = $self -> GetStrand($id2);
	return 0 if ($strand1 ne $strand2);
	return 1;
}


=head2 Procedure DefinePhysicalLink

 Usage        : $o_dsa-> DefinePhysicalLink($ra_couple);
 Procedure    : Define the physical link betwen two protein of a couple
 Args         : the $ra_couple
 Returns      : inter or intra e.g. the physical link

=cut

sub DefinePhysicalLink
{
	my $self = shift;
	my ( $couple_id)= (@_);
	my $rh_ds_data = $self-> _GetRefDsHash();
	my $id1 = $rh_ds_data -> {$couple_id} -> {seq1};
	my $id2 = $rh_ds_data -> {$couple_id} -> {seq2};
	my %h_link = (t1 => 'inter', t2 => 'intra');

	my $chr1 = $self -> GetChrom ($id1);
	my $chr2 = $self -> GetChrom ($id2);
	
	if ($chr1 eq $chr2)
	{
		return $h_link {t2};
		
	}
	else
	{
		return $h_link {t1};
	}
}

=head2 Procedure GetGeneDistance

 Usage        : $o_dsa-> GetGeneDistance($ra_couple);
 Procedure    : Define the distance between two protein of a couple
 Args         : the $ra_couple
 Returns      : $distance the distance in gene
 
=cut

sub GetGeneDistance
{
	my $self = shift;
	my ( $couple_id)= (@_);
	my $rh_ds = $self -> _GetRefDsHash();
	my $id1 = $rh_ds -> { $couple_id } -> { seq1 };
	my $id2 = $rh_ds -> { $couple_id } -> { seq2 };
	if ($self -> DefinePhysicalLink ($couple_id) eq "intra")
	{
		my $distance = abs( $self -> GetGeneIndex ($id1) - $self -> GetGeneIndex ($id2));
		return $distance;
	}
	else
	{
		return "NA";
	}	
}

=head2 Procedure GetParaclusterFromId

 Usage        : $o_dsa-> GetParaclusterFromId($ra_couple);
 Procedure    : Define the distance between two protein of a couple
 Args         : the $ra_couple
 Returns      : $distance the distance in gene
 
=cut

sub GetParaclusterFromId
{
	my $self = shift;
	my ( $protein_id)= (@_);
	my $rh_paracluster = $self -> _GetRefParaclusterHash();
	
	my $chr = $self -> GetChrom($protein_id);
	my $index = $self -> GetGeneIndex ($protein_id);
	
	foreach my $para_id (keys %$rh_paracluster)
	{
		if ( $chr eq $rh_paracluster -> {$para_id} -> {chr})
		{
			if (grep (/$index/,@{$rh_paracluster -> {$para_id} -> {ra_gene}}))
			{
				#~print "INEDX $index \n";
				#~print "MB ", join(",", @{$rh_paracluster -> {$para_id} -> {ra_gene}}), "\n";
				return $para_id;
			}
		}
	}
	return "NA";
}




=head2 Procedure IsInParaCluster

 Usage        : $o_dsa-> IsInParaCluster($ra_couple, $cutoff);
 Procedure    : Define if the two members are part of a cluster if their indexes are within cutoff value 
 Args         : the $ra_couple, the $cutoff for inex difference
 Returns      : boolean answer

=cut

sub IsInParaCluster
{
	my $self = shift;
	my ( $index1, $index2, $cutoff)= (@_);
	
	my $dist = abs ($index2 - $index1);
	if ($dist <= $cutoff)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}
=head2 Procedure BuildParaId
=cut

sub BuildParaId
{
	my $self = shift;
	my ($fam, $index) = (@_);
	
	return $fam. "_para_".$index;
}

=head2 Procedure DefineDsClass

 Usage        : $o_dsa-> DefineDsClass($ra_couple, $rh_classes);
 Procedure    : Define the couple class according to its Ds 
 Args         : the $ra_couple, the $rh_classes where classes are defined for example C1: 0<= dS <0.01, C2: 0.01<= dS <0.03, and so on...
 Returns      : the ds class

=cut

sub DefineDsClass
{
	my $self = shift;
	my ( $couple_index)= (@_);
	my $rh_ds = $self -> _GetRefDsHash();
	
	my %h_bounds = (b1 => 0.01, b2 => 0.03, b3 => 0.1, b4 => 0.2, b5 => 0.4, b6=> 0.6);
	my $ds = $rh_ds -> {$couple_index} ->{dS};
	my $ds_class = "";

	return "NA" if ($ds =~ /nan/i);
	
	if ($ds < $h_bounds {b1})
	{
		$ds_class = "C1";
	}
	elsif ($ds >= $h_bounds {b1} && $ds < $h_bounds {b2})
	{
		$ds_class = "C2";
	}
	elsif ($ds >= $h_bounds {b2} && $ds < $h_bounds {b3})
	{
		$ds_class = "C3";
	}
	elsif ($ds >= $h_bounds {b3} && $ds < $h_bounds {b4})
	{
		$ds_class = "C4";
	}
	elsif ($ds >= $h_bounds {b4} && $ds < $h_bounds {b5})
	{
		$ds_class = "C5";
	}
	elsif ($ds >= $h_bounds {b5} && $ds < $h_bounds {b6})
	{
		$ds_class = "C6";
	}
	elsif ($ds >= $h_bounds {b6} )
	{
		$ds_class = "C7";
	}
	else
	{
		$ds_class = $ds;
	}
	
	return $ds_class;
}

=head2 Procedure DefineExonClass

 Usage        : $o_dsa-> DefineExonClass($ra_couple);
 Procedure    : Define the couple exon class according to its nb of exon (Single, Multi, or Mixed)
 Args         : the $ra_couple
 Returns      : 

=cut

sub DefineExonClass
{
	my $self = shift;
	my ( $couple_id)= (@_);
	
	my $rh_ds = $self -> _GetRefDsHash();
	
	my %h_exonClass = (c1 => "single", c2 => "multi", c3 => "mixed");
	
	my $nb_exon1 = $self -> GetNbExon ($rh_ds -> { $couple_id } -> { seq1 });
	my $nb_exon2 = $self -> GetNbExon ($rh_ds -> { $couple_id } -> { seq2 });
	
	if ($nb_exon1 <= 1)
	{
		if ($nb_exon2 <= 1)
		{
			return $h_exonClass{c1};
		}
		else
		{
			return $h_exonClass{c3};
		}
	}
	else
	{
		if ($nb_exon2 > 1)
		{
			return $h_exonClass{c2};
		}
		else
		{
			return $h_exonClass{c3};
		}
	}
}

=head2
=cut

sub DefineParaclusterClass
{
	my $self = shift;
	my ($pair_id ) = (@_);
	
	my $rh_ds = $self -> _GetRefDsHash();
	my $id1 = $rh_ds -> {$pair_id} -> {seq1};
	my $id2 = $rh_ds -> {$pair_id} -> {seq2};
	my $para_cluster1 = $self -> GetParaclusterFromId ($id1);
	my $para_cluster2 = $self -> GetParaclusterFromId ($id2);
	my $para_class = "";
	
	my $link = $self -> DefinePhysicalLink($pair_id);
	return "NA" if ($link eq "inter");
	
	if ($para_cluster1 eq $para_cluster2)
	{
		if ($para_cluster1 eq "NA")
		{
			return "NA";
		}
		else
		{
			return "single";
		}
	}
	else
	{
		return "mixed";
	}
	
	
}
#OK end#
##################################TO TEST start #####################################
#~=head2 Procedure DefineBlockClass
#~
 #~Usage        : $o_dsa-> DefineBlockClass($ra_couple);
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub DefineBlockClass
#~{
	#~my $self = shift;
	#~my ($ra_couple)= (@_);
	#~my %h_block_type = (t1 => "S", t2 => "D", t3 => "X"); 
	#~my %h_id_blk =();
	#~my $rh_gene = $self -> _GetRefGeneHash();
	#~my $rh_blocks = $self -> _GetRefBlockHash();
	#~
	#~#Assign
	#~foreach my $id (@$ra_couple)
	#~{
		#~my $chr = $self -> GetChrom ($id, $rh_gene);
		#~my $rh_chr_blk = $self -> GetChrBlocks ($chr, $rh_blocks);
		#~my ($block, $type) = $self -> LocateInBlocks ($id, $rh_chr_blk);
		#~$h_id_blk {$id} = {bloc_id => $block, type => $type};
	#~}
#~
	#~return \%h_id_blk;
#~}
#~
#~=head2 Procedure GetChrBlocks
#~
 #~Usage        : $o_dsa-> GetChrBlocks($chr, $rh_blocks);
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub GetChrBlocks
#~{
	#~my $self = shift;
	#~my ( $chr)= (@_);
	#~my $rh_blocks = _GetRefBlockHash();
	#~my %h_chr_blk =();
	#~foreach my $blk (keys %$rh_blocks)
	#~{
		#~if ($rh_blocks -> {$blk} -> {chr} == $chr)
		#~{
			#~$h_chr_blk {$blk} = {start =>  $rh_blocks -> {$blk} -> {start},
			                     #~end =>  $rh_blocks -> {$blk} -> {end}};
		#~}
	#~}
	#~
	#~return \%h_chr_blk;
#~}
#~
#~
#~=head2 Procedure LocateInBlocks
#~
 #~Usage        : $o_dsa-> LocateInBlocks($protein_index, $rh_chr_blocks);
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub LocateInBlocks
#~{
	#~my $self = shift;
	#~my ( $protein_index, $rh_chr_blocks )= (@_);
	#~
	#~foreach my $block ( keys %$rh_chr_blocks )
	#~{
		#~if ($protein_index >=  $rh_chr_blocks -> {$block} -> {start})
		#~{
			#~if ($protein_index <=  $rh_chr_blocks -> {$block} -> {end})
			#~{
				#~my $type = $rh_chr_blocks -> {$block} -> {type};
				#~return $block, $type;
			#~}
		#~}
	#~}
	#~return "NA","X";
#~}
#~
#~=head2 Procedure AssignFamily
#~
 #~Usage        : $o_dsa-> AssignFamily($protein_index, $rh_blocks);
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub AssignFamily
#~{
	#~my $self = shift;
	#~my ( $protein_id )= (@_);
	#~my $rh_family = _GetRefFamilyHash();
	#~foreach my $fam (keys %$rh_family)
	#~{
		#~if (grep (/$protein_id/, @{$rh_family -> {$fam}}))
		#~{
			#~return $fam;  
		#~}
	#~}
	#~return "NA" ;
#~}
#~
#~=head2 Procedure MapGeneFamilyByChr
#~
 #~Usage        : $o_dsa-> MapGeneFamilyByChr ($protein_index, $rh_blocks);
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub MapGeneFamilyByChr
#~{
	#~my $self = shift;
	#~my ( $chr)= (@_);
	#~my $rh_gene = $self -> _GetRefGeneHash();
	#~my $rh_family = $self -> _GetRefFamilyHash();
	#~my %h_mapped = ();
	#~
	#~foreach my $id (keys %$rh_gene)
	#~{
		#~if ($rh_gene -> {$id} -> {chr} eq $chr)
		#~{
			#~my $gene_index = $self -> GetGeneIndex($id);
			#~my $fam = $self -> AssignFamily ($id); #Attention à la nomenclature
			#~$h_mapped {$gene_index} = {id => $id, fam => $fam};
		#~}
	#~}
	#~return \%h_mapped;
#~}

##################################TO TEST end #####################################

##################################TO FILL start#####################################
=head2 Procedure GetGoAnnotation

 Usage        : $o_dsa-> GetGoAnnotation($protein_index, $rh_blocks);
 Procedure    : 
 Args         : 
 Returns      : 

=cut

sub GetGoAnnotation
{
	my $self = shift;
	my ( $protein_id)= (@_);
	
	#Here Get GO ?!! But where ?? Look in Narcisse ??
	my $go="";
	return $go ;
}

##################################TO FILL end #####################################

=head2 Procedure BuildParaCluster

 Usage        : $o_dsa-> BuildParaCluster($rh_family);
 Procedure    : 
 Args         : 
 Returns      : $rh_paracluster with $h_para {$para_id} = {chr => $current_chr, fam => $fam, ra_gene => \@a_keep_array}
				in @a_keep_array there are the gene index involved in the parcaluster

=cut

sub BuildParaCluster
{
	my $self = shift;
	
	my ($cutoff) = (@_);
	my $rh_family = $self -> _GetRefFamilyHash();
	my $ra_chr = $self -> _GetRefChrList();

	#The has with paracluster results
	my %h_para = ();
	
	#Foreach family
	foreach my $fam (keys %$rh_family)
	{
		#~print "FAM: $fam \n";
		#Get members of this family
		my $ra_members = $rh_family ->{$fam};
		#Creating a hash with all chr as key
		my %h_chrom = ();
		@h_chrom{@$ra_chr} = ();
		
		#For each memebr of the family
		foreach my $geneID (@$ra_members)
		{
			#Get correspondig porteinid
			my $id = $self ->GetProteinID ($geneID);
			#Get chr and index
			my $chr = $self -> GetChrom ($id);
			my $idx = $self -> GetGeneIndex ($id);
			#print "Mb: $geneID  Prot: $id chr: $chr Indx: $idx\n";
			#Un hash avec les chr et un array avec les membres
			
			if (defined $h_chrom {$chr})
			{
				my $ra_g_in_chr = $h_chrom {$chr};
				push (@$ra_g_in_chr, $idx);
				$h_chrom {$chr} = $ra_g_in_chr;
			}
			else
			{
				my @a_g_in_chr = ();
				push (@a_g_in_chr, $idx);
				$h_chrom {$chr} = \@a_g_in_chr;
			}
		}

		#Foreach chr of the family build paracluster
		my @a_sort_chrom = sort @$ra_chr;
		my $prec_chr = $a_sort_chrom[0];
		my @a_para_mb = ();
		my $para_index = 1;
		my $para_id = $self -> BuildParaId($fam ,$para_index);
		
		#Start walk on families 
		foreach my $current_chr (@a_sort_chrom)
		{
			if (scalar @a_para_mb > 0)
			{
				#Paracluster array non empty
				if (!exists $h_para {$para_id})
				{
					my @a_keep_array = @a_para_mb;
					$h_para {$para_id} = {chr => $prec_chr, fam => $fam, ra_gene => \@a_keep_array}; #En index !!!
					$para_index += 1;
					$para_id = $self -> BuildParaId($fam ,$para_index);	
				}
			}
			#~#Empty para cluster array for new one
			@a_para_mb = ();
			
			if (!defined $h_chrom {$current_chr})
			{
				next ; #Nothing in that Chr
			}
			else
			{
				#Ya qqc sur le $current_chr#
				my @a_mb = @{$h_chrom {$current_chr}};
				#No more than two members on the same chr#	
				next if (scalar @a_mb < 2);
				
				my @a_ordered_mb = sort {$a <=> $b} (@a_mb);
				my $prec_index = shift @a_ordered_mb;
					
				foreach my $current_index (@a_ordered_mb)
				{
					my $is_para = $self-> IsInParaCluster ($prec_index, $current_index, $cutoff);
					
					if (!$is_para)
					{
						if (scalar @a_para_mb > 1)
						{
							#At least two mb in a paracluster so we need at leats two mb on chromosme
							my @a_keep_array = @a_para_mb;
							$h_para {$para_id} = {chr => $current_chr, fam => $fam, ra_gene => \@a_keep_array}; #En index !!!
							#Empty para array
							@a_para_mb = ();
							#New paraId
							$para_index += 1;
							$para_id = $self -> BuildParaId($fam ,$para_index);	
						}
					}
					else
					{
						if (scalar @a_para_mb < 1)
						{
							#Pushing the first pair
							push (@a_para_mb, $prec_index);
							push(@a_para_mb, $current_index);
						}
						else 
						{	
							#Arry is not empty, adding an element
							push(@a_para_mb, $current_index);
						}
					}
					$prec_index = $current_index;
				}
			}
			$prec_chr= $current_chr; 
		}
	}
	
	$self -> _Set ('rh_paracluster' => \%h_para);
	
	#~print "DUMP\n\n\n";
	#~print Dumper (%h_para);
	
	return \%h_para ;
}

#TO TEST START########################################################################"""

#~=head2 Procedure ChromosomeWalk
#~
 #~Usage        : $o_dsa-> ChromosomeWalk();
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub ChromosomeWalk
#~{
	#~my $self = shift;
	#~my ( $rh_family, $rh_ds_data, $rh_gene)= (@_);
	#~
	#~my $ra_chr = $self -> BuildChrList( $rh_gene);
	#~
	#~my $current_index = 0;
	#~foreach my $chr (@$ra_chr)
	#~{
		#~$current_index += 1;
		#~my $rh_chr_info = $self -> MapGeneFamilyByChr($chr);
		#~my $family = $rh_chr_info -> {$current_index} -> {fam};
		#~if ($family ne "NA")
		#~{
			#~if (exists $h_seen_family {$family})
			#~{
				#~#Est-ce que ce gene a cet index est dans le paracluster
				#~my $prec_index = $h_seen_family {$family} -> {prec_index};
				#~
				#~if ($self -> IsInParaCluster($prec_index, $current_index, $cutoff))
				#~{
					#~my $ra_para = $h_seen_family {$family} -> {paracluster};
					#~push (@$ra_para, $current_index);
					#~
					#~$h_seen_family {$family} = (prec_index => $current_index,
												#~paracluster => $ra_para);
				#~}
			#~else
			#~{
				#~my @a_para = ();
				#~push (@a_para, $current_index);
				#~$h_seen_family {$family} = (prec_index => $current_index,
											#~paracluster => \@a_para);		
			#~}	
		#~}
	#~}
	#~
	#~return $go ;
#~}
##################################TO TEST start #####################################
#~=head2 Procedure BuildDsGeneIDList
#~
 #~Usage        : $o_dsa-> BuildDsGeneIDList();
 #~Procedure    : 
 #~Args         : 
 #~Returns      : 
#~
#~=cut
#~
#~sub BuildDsGeneIDList
#~{
	#~my $self = shift;
	#~my ($rh_ds_data)= (@_);
	#~
	#~my @a_ds_gene_ids = ();
	#~
	#~foreach my $idx (keys %$rh_ds_data)
	#~{
		#~my @a_ids = $rh_ds_data -> {$idx} -> (qw (seq1 seq2));
		#~
		#~foreach my $id (@a_ids)
		#~{
			#~next if (grep /$id/, @a_ds_gene_ids);
			#~push(@a_ds_gene_ids, $id);
		#~}
	#~}
	#~
	#~return \@a_ds_gene_ids;
#~}
##################################TO TEST end #####################################

##################################TO FILL start #####################################
=head2 Procedure ClassifyDuplication

 Usage        : $o_dsa-> ClassifyDuplication();
 Procedure    : TAG or Retrodup or WGD or Drift
 Args         : 
 Returns      : 

=cut

sub ClassifyDuplication
{
	my $self = shift;
	
	my %h_duplication = (d1 => "WGS", d2 => "TAG", d3 => "Retro", d4 => "Drift");
	#Classify the duplication
	
	#Retroduplication = long range duplication
	#Look for single exon, within the same family there is a multiexon parent, paralle orientation (direct repeat), polyA tail (???), different chr
	
	#TAG
	#Look for same chr duplication with low Ds, with 10 genes proximity
	
	#WGD
	#Look for high dS maybe ??
	
	#Drift
	#Random orientation,  physical link between tandem and retro, with a translocation from another chr between (maybe ??)
	
	return ;
}

##################################TO FILL end #####################################


=head2 Procedure _CreateWorkDir

 Usage        : $self -> _CreateWorkDir($prefix);
 Procedure    : Creates a temporary working directory to avoid index problems with Bio::DB::Fasta
 Args         : $prefix a prefix
 Returns      : the $working_dir and the $currentdirectory
 
=cut

sub _CreateWorkDir
{
	my $self = shift;
	my ($prefix) = (@_);
	
	my $sav_dir = cwd;
	my $working_dir = File::Temp::tempdir( $prefix."_XXXX", CLEANUP => 0, TMPDIR => 1 );
	chdir($working_dir);
	
	return $sav_dir, $working_dir;
}

=head2 Procedure _RemoveWorkDir

 Usage        : $self -> _RemoveWorkDir($prefix);
 Procedure    : Removes temporary working directory 
 Args         : $prefix a prefix
 Returns      : none
 
=cut

sub _RemoveWorkDir
{
	my $self = shift;
	my ($sav_dir, $working_dir) = (@_);
	chdir $sav_dir;
	rmtree($working_dir);
	return ;
}

=head2 Procedure _CreateSymLink

 Usage        : $self -> _CreateSymLink($path_to_old_file);
 Prerequisite : 
 Procedure    : Create a symbolic link to $path_to_old_file
 Args         : $path_to_old_file the path to the file to be linked
 returns      : The name of the symbolic linked file without the path in it

=cut

sub _CreateSymLink
{
	my $self = shift;
	my ($path_to_old_file) = (@_);	
	my $new_file = (split("\\/", $path_to_old_file))[-1];
	my $symlink_exists=eval {symlink("",""); 1};
	if ($symlink_exists) 
	{
		symlink ($path_to_old_file, $new_file);
	}
	else
	{
		print "Linking db file $path_to_old_file failed\n";
		exit(1);
	}
	
	return $new_file;	
}

##################################TO TEST start#####################################
sub _CheckFamilyAndGeneConcordance
{
	my $self = shift;
	my ($rh_gene_prot, $id) = (@_);
	
	my $newid = $id . ".1";
	if (exists $rh_gene_prot -> {$newid})
	{
		return $newid;
	}
	else
	{
		return "";
	}
}
##################################TO TEST end #####################################

1;
