# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet - Main class of Narcisse data objects


=head1 DESCRIPTION

Main class for the components of a Chromosome object.
"Mixing" of Data.pm, Display.pm and Representation_graphique.pm.

This class will automatically load data files informations, to display them you should use Display method.

B<!! Warnings !!>
B< - You don't have to instanciate this class directly. This class must be inherited.>
B< - Data are considered to be well sort in data files.>

B<How to create a new class :>
To create a new class, you need to copy the code of the Synopsis, to change some attributes and to overload some methods :
See Template.pm to have a 'skeleton' of a new ElementsSet object.

I<1) Attributes :>
- file_directory : by default we considere that data are store in a directory called "class_name/"

- file_extension : by default an ElementsSet inherited class will read only files which file extension is ".class_name". If yours files have an other extension you need to change this parameter.
GetElement
- ra_data_headers : this is the data headers array. By default headers are 'symbol', 'start', 'end', 'strand' in Narcisse data.
If there are more or other headers in your data files, you have to set them like Synopsis show you.
Moreover this attribute is used by GetElementFields method to know which column have to be returned.

- primary_key : if the reference column not equals to 'element', you have to change this attribute.

- start_label : if the label of your start column not equals to 'start', you have to change this attribute.

- end_label : if the label of your end column not equals to 'end', you have to change this attribute.


I<2) Overload methods :>
By default, the Display method isn't usable, you have to overload this method to show loading data.


B<Graphic parameters :>
You could use a file to set graphic parameters using UpdateGraphicCfg method.
Then you have to use GetGraphicParameter method to access to those values.


The parameters file must be formatted such as :
I<#tag	value
first_tag	first_value>


B<Object attributes :>

 - feature			=> the feature name of the ElementsSet inherited object
 - organism			=> short name of the organism, according to the data directory label
 - chromosome		=> chromosome number or label
 - start			=> start of the slice to show
 - end				=> end of the slice
 - ra_map			=> array reference for mapping coordinates
 - rh_data			=> hash reference of extracted data
 - ra_sort_elements	=> hash reference of sorted primary keys for data mining
 - ra_data_headers	=> list of sort headers, keys of rh_data elements (default ('element', 'start', 'end', 'strand'))
 - graph_orientation=> vertical orientation of the representation. Positive orientation show data from top to bottom. ('-' or '+')


B<File parameters :>
 - file_extension	=> optional : the extension of data files (default 'feature')
 - file_directory	=> optional : directory of data (default 'feature')
 - primary_key		=> optional : primary key of the file (default 'element')
 - start_label		=> optional : label of the column which store start data ('start' by default)
 - end_label		=> optional : label of the column which store end data ('end' by default)

=head1 SYNOPSIS

package ElementsSet::Class;

use strict;
use warnings;

use base qw( ElementsSet );
our $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
				attribute	=> 'default_value',

				@_
				);

	# New attributes :
	$self -> {attribute} = $h_param{attribute};

	# If you have to change a defaut parameter (file extension, start_label, data headers ...) :
	my @a_data_headers = ('first_header','second_header','n_header');
	my %h_new_param = ( file_directory => 'data_dir_label',
						file_extension => 'class',
						ra_data_headers => \@a_data_headers,

						%h_param
						);

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

    return $self;
}


=head1 SUBROUTINES

_Init
_ReadNarcisseFormatedFile
AddCoordinatesOnMap
AddFieldsOnArea
Display
GetAreaCoordinates
GetAreaFields
GetAttributes
GetData
GetDataOnInterval
GetEnd
GetFileName
GetGraphicParameter
GetLinksAreaCoordinates
GetNucleicMax
GetNucleicMin
GetRange
GetSortElements
GetStart
UpdateGraphicCfg
New
PrintElementsList
ReOrientedStartEnd
ReversePosition
SaveMap
SaveParameters
SetGraphicParameters

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package ElementsSet;

use strict;
use warnings;

use base qw( LipmObject );

use Scalar::Util qw( blessed );
use Data::Dumper;
use Carp;

use Time::HiRes qw( gettimeofday tv_interval);

use Image;
use IO::File;
use NARCISSE_utils;
use Error qw( :try );
use LipmError::IOException;
use LipmError::ObjectException;
use LipmError::ParameterException;
use ParamParser;
use General;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

our $TIME = 0;

# Global parameters
# our $GRAPHIC_FILE;
our $BOOL_FILE_STATUS;	# True(1) if parameters are correctly saved in GRAPHIC_FILE, else False(0).
use constant ID_SEPARATOR => '-';

# Defines the layer name for particular objects. In most of the cases the layer name is the name
# of the object itself (Gene => gene, Rna => rna). This is not the case for ocean for example which
# belongs to the chrom layer.

our %LAYER=( 'ocean' => "chrom"
			);

#Some of the elemnts are paired elements (syntenies, hsps, prbh) while othre are single (gene, rna, etc, ....)
# This will be used for the map information

use constant SINGLE => 'single';
use constant PAIRED => 'paired';

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : $self -> _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param a hash of parameters
 ##PAS A JOUR ### NOT UP TO DATE ##
 					mandatory
                        o_io            => InstanceObserver
  						organism  		=> short name of the organism
  						chromosome		=> chromosome name or number
  						o_param_graph   => ParamParser object, with Graphics.cfg data

  					optional
  						feature			 	=> simple name of this feature of object (default 'class_name')
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 0)
  						file_directory		=> directory of data (default 'feature')
  						file_extension		=> extension of the data files (default 'feature')
						graph_orientation 	=> the orientation (default '+')
						primary_key 		=> in data files, this is the reference column (default 'element')
						start_label 		=> in data files, this is the start column  (default 'start')
						end_label	 		=> in data files, this is the end column  (default 'end')
						ra_data_headers		=> array ref., this is the array of sorted column which have to be store when data are read (default ('element','start','end','strand'))
						graphic_param_file	=> file for the graphic parameters initialization

 Attributes	  : objects attributes :
 					Element attributes :
 						feature, organism, chromosome, start, end
 					File and directory attributes :
 						file_directory, file_extension, ra_data_headers, primary_key, start_label, end_label
 					Representation attributes :
 						ra_map, graph_orientation

_o_species
__root_data  --a virer ? pas utilise on dirait.
_id
__ra_map_keys
__dna_cylinders
__dna_cylinder_height
_o_data
_chromosome
__o_map_area
_o_param_graph
__start
__end
__start_label
__end_label
__file_directory
__file_extension
__primary_key
_graph_orientation
_feature
__ra_headers
_o_chrom_slice
__area_id_prefix


 Globals	  :	none
 Errors		  : LipmError::ObjectException

=cut

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init( @_ );

	# Default parameters
	my @a_default_data_headers = ('symbol','start','end','strand');
	my $feature = lc(ref($self));
	$feature =~ s/.*::(.*)/$1/;

	# Parameters initialization
	my %h_params = (
		# Mandatory parameters
		o_species           => undef,
		chromosome			=> undef,

		o_maparea			=> undef,
        o_param_graph		=> undef,

        # Optional parameters with default value
        o_chrom_slice       => 'bidouille',
        feature				=> $feature,
        start				=> 0,
        end					=> 1,
#         file_directory		=> $feature,
#         file_extension		=> $feature,
        graph_orientation	=> '+',
        primary_key			=> 'element',
        start_label			=> 'start',
        end_label			=> 'end',
#         ra_data_headers		=> \@a_default_data_headers,
        graphic_param_file	=> '',
        area_id_prefix		=> '',
        return_type			=> 'HASH',
        use_cache			=> 0,
        @_ );

	#0- assert mandatories => no undef
	LipmError::ParameterException::Assert( %h_params );

	#0.1- cas o_species
	my $o_species = $h_params{ o_species };
	if( ! &SafeIsa( $o_species, 'SpeciesObserver' ) )
	{
		throw LipmError::ObjectException( $self, '!attribute' );
	}
	$self->_Set( _o_species  => $o_species,
				 __root_data => $o_species->GetDataPath() ); #/!\ c'est species_data pas root

	#1- Set attributes from args
	$self->_SetAttributesFromParams( %h_params );

=pod
	#Giving a name to the object : is this really usefull ???
	#$self->_SetElemenSetId();
	my @a_id =( ref($self) );
	my @a_attr = qw( _chromosome __start __end );
	my( $chrom, $start, $end ) = $self->_Get( @a_attr );
# warn "$chrom, $start, $end";
	push( @a_id, $o_species->GetShortName() );
	push( @a_id, $chrom, $start, $end );
	$self->_Set( __id => join( "_", @a_id ) );
=cut

	#2 Some configurations
	$self->UpdateGraphicCfg( $h_params{graphic_param_file} );

	$self->SetElementSetLayerName( $feature );

	# graphical initialization
	#This is a dirty trick necessary because neither synt nor prbh depends on a chrom_slice
	my $o_slice = $self->_Get( '_o_chrom_slice' );
	if( &TRUE == &SafeIsa( $o_slice, 'ChromosomeSlice' ) )
	{
		my $cylinders = $o_slice->GetDNACylinders();
		my $height    = $o_slice->GetDNACylinderHeight();
		$self->_Set( __dna_cylinders => $cylinders,
					 __dna_cylinder_height => $height );
	}

	# Elements can't be shown if they belong to a too long or too short range/slice
	# throw exception if pb => no object created
	$self->_AssertBetweenThresholds();

	#3 Data initalization
	my $file = $self->GetFileName();

	my %h_storage_opt = ( return_type => $h_params{ return_type },
							use_cache => $h_params{ use_cache } );

	my @a_attr = qw( _chromosome __start __end );
	my( $chrom, $start, $end ) = $self->_Get( @a_attr );

	# If data can't be loaded an exception is thrown
	my $o_data = $self->_ReadNarcisseFormatedFile($file, $start, $end, %h_storage_opt );
	$self->_Set( _o_data => $o_data );

    return;
} # _Init

#sub DESTROY
#{
#	my $self = shift;
#	warn "DESTROYED";
#	$self -> _Set ( '_o_chrom_slice', undef);
#	$self -> _Set ( '_o_map_area', undef);
#}

=head2 _SetAttributesFromParams

 Usage       : $o_x->_SetAttributesFromParams( %h_params )
 Function    : set attributes which values are in %h_params.
 Args        : %h_params
 Error       : none.

=cut

sub _SetAttributesFromParams
{
	my $self   = shift;
	my %h_params = @_;

	# not defined attributes
	$self->_Set( _rh_mapareas => {} );

	my @a_keys = qw( chromosome o_maparea o_param_graph );
	my( $chrom, $o_map, $o_param ) = @h_params{ @a_keys };
	$self->_Set( _chromosome     => $chrom,
				 _o_map_area    => $o_map,
				 _o_param_graph  => $o_param );

	@a_keys = qw( start start_label end end_label );
	my( $start, $start_lb, $end, $end_lb ) = @h_params{ @a_keys };
	$self->_Set( __start   => $start,
				 __end     => $end,
				 __start_label => $start_lb,
				 __end_label   => $end_lb );

	@a_keys = qw( file_directory file_extension primary_key graph_orientation );
	my( $dir, $ext, $key, $orient ) = @h_params{ @a_keys };
	$self->_Set( #__file_directory => $dir,
				 #__file_extension => $ext,
				 __primary_key    => $key,
				 _graph_orientation => $orient );

	@a_keys = qw( ra_data_headers feature );
	my( $ra_headers, $feature ) = @h_params{ @a_keys };
	$self->_Set( _feature => $feature,
				 __ra_headers => $ra_headers );

	@a_keys = qw( o_chrom_slice area_id_prefix );
	my( $o_slice, $area_id ) = @h_params{ @a_keys };
	$self->_Set( _o_chrom_slice  => $o_slice,
				 __area_id_prefix => $area_id );

	return;
} #_SetAttributesFromParams

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

=head2 DrawElement

 Title        : DrawElement
 Usage        : $object -> DrawElement($x0,$y0,$x1,$y1,$dna_cyl,$color,$side);
 Prerequisite : The appropriate layer should have been set
 Procedure    : Draws an element, gene, rna or repeat, on the DNA cylinder and on the correct strand
 Returns      : an array of coordinates of a rectangular area : (X0,Y0,X1,Y1)
 Args         :
 				$o_image : an image object
 				$x0, $y0, $x1, $y1 : the rectangular coordinates of the element
 				$color : a GD image color
 				$dna_strand : either +1 or -1

=cut
sub DrawElement
{
	my $self = shift;
	my ($o_image, $x0, $y0, $x1, $y1,  $color, $dna_strand) = @_;

	#Restrict the element to the drawing area defined by the start and end of this ElsementSet object
	my $draw_y0 = ( $y0 > $self->_GetStart() ) ? $y0 : $self->_GetStart();
	my $draw_y1 = ( $y1 < $self->_GetEnd() ) ? $y1 : $self->_GetEnd();

	my $ellipse_height = $self->GetDNACylinderHeight();

	if ( $dna_strand > 0 )
	{
	 return $o_image->RightVerticalCylinder( $x0, $draw_y0, $x1, $draw_y1, $ellipse_height, $color);
	} else
	{
	 return $o_image->LeftVerticalCylinder ( $x0, $draw_y0, $x1, $draw_y1, $ellipse_height, $color);
	}

}

=head2 PrintElementsList

 Title        : PrintElementsList
 Usage        : $object -> PrintElementsList();
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Print locus_tag : start - end for each element
 				This method could be use to check data (on a debug mode)
 Returns      : string
 Args         : none
 Globals      : none

=cut
sub PrintElementsList
{
	my $self = shift;
	my @a_sort_elements = $self->GetSortElements();
	my $return = "";

	foreach my $element (@a_sort_elements)
	{
		my $ra_headers = $self->_Get( '__ra_data_headers' );
		foreach my $i ( $self->GetFields("$element", $ra_headers ) )
		{
			$return .= "$i\t";
		}
		$return .= "\n";
	}
	$return .= "\n";

	return $return;
}

=head2 Display

 Title        : Display
 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 Procedure    : Display data of the Element inherited object on the Image object.
 				You need to overload this method on your own class.
 				Moreover, you can store drawing elements in 'areas' by using AddAreaOnMap method.
 Returns      : none
 Args         : o_image	-> Image object
                o_box   -> a box object associated to the chromosome slice of the elements under consideration
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			@_
			);

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	return;
}
sub DisplayText{};

=head2 ProceedDisplay

 Title        : ProceedDisplay
 Usage        : $object -> ProceedDisplay(o_image [,side]);
 Procedure    : Check if the provided image parameter is an image object, if the box is a box object etc...
 Returns      : 1 if its OK 0 if its not !
 Args         : same as Display

=cut

sub ProceedDisplay
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			@_
			);

	# Don't draw data if it wasn't loaded
	return 0 if ( $self->IsEmpty() );

	# Testing mandatory parameters
	if( &FALSE == &SafeIsa( $h_param{image}, 'Image' ) )
	{
		throw LipmError::ParameterException('Image object', 'missing');
	}
	if( &FALSE == &SafeIsa( $h_param{o_box}, 'Box' ) )
	{
		throw LipmError::ParameterException('Box object', 'missing');
	}

	return 1;
}

=head2 UpdateGraphicCfg

 Title        : UpdateGraphicCfg
 Usage        : $o_element -> UpdateGraphicCfg($file)
 				ElementsSet-> UpdateGraphicCfg($file)
 				&ElementsSet::UpdateGraphicCfg($file)
 Prerequisite : none
 Procedure    : Initialize graphical parameters, managed by o_param_graph, with data in file $file.
 Returns      : none
 Args         : $file  configuration file name

 Error		  : LipmError::IOException

=cut
sub UpdateGraphicCfg
{
	my $self = shift;
	my $file = shift;

	# Initialized graphic parameters with the file $file.
	try
	{
		my $species_cfg  = $self->GetLocalGraphicsCfgFile();

		my $msg = "read $species_cfg for graphics configuration.\n";
		$self->_Trace( $msg, '.', 'IE' );

		$self->{ o_param_graph }->Update( $species_cfg, 'O' )	if ( -e $species_cfg );
		$self->{ o_param_graph }->Update( $file, 'O' )	if ( -e $file );

	}
	catch LipmError with
	{
		my ($E,$scalar) = (shift,shift);
		my $message = $E->Explain();
		$message .= "$file couldn't be read.\n\n"; #! a quoi sert explain alors ?
# 		warn $message;
	}
	finally
	{
		$BOOL_FILE_STATUS = 0;
	};

	return;
}

=head2 IsMapDefined

 Title        : IsMapDefined
 Usage        : $self -> IsMapDefined( $element_id )
 Prerequisite : none
 Function     : Check if a map element exists for element_id
 Returns      : 0 or 1
 Args         : none
 Attributes   : element_id
 Globals      : none

=cut
sub IsMapDefined
{
	my $self = shift;
	my $element_id =shift;

	return $self->_Get( '_o_map_area' )->IsMapDefined( $element_id );
}

#------------------------------------------------#
#           Managing the data                    #
#------------------------------------------------#

=head2 IsEmpty

 Title        : IsEmpty
 Usage        : $self -> IsEmpty()
 Prerequisite : none
 Function     : Check if the object effectively contains data ( rh_data should be a reference on a non empty hash )
 Returns      : 1 if rh_data is not defined or has no keys
 				or
 				0 if not

=cut
sub IsEmpty
{
	my $self = shift;

	my $o_data = $self->_Get( '_o_data' );
	return (!defined( $o_data ) || $o_data->IsEmpty() );
}


=head2 IsDataAvailable

 Title        : IsDataAvailable
 Usage        : $self -> IsDataAvailable()
 Prerequisite : none
 Function     : Check if the data exists (most of the time this consists in verifying if the file exists)
 Returns      : 1 if it foes exist
 				or
 				0 if not

=cut
sub IsDataAvailable
{
	my $self = shift;
	my $file = shift;

	return (-e $file);
}

=head2 _ReadNarcisseFormatedFile

 Title        : _ReadNarcisseFormatedFile
 Usage        : $self -> _ReadNarcisseFormatedFile($file,$start,$end);
 Prerequisite : none
 Function     : Get rh_data and ra_sort_elements
 Returns      : a NarcisseData object
 Args         : $file 	-> the data file
 				$start 	-> start of analyze
 				$end 	-> end of analyze
 Error        : LipmError::ObjectException.
                LipmError::ParameterException.

=cut
sub _ReadNarcisseFormatedFile
{
	my $self = shift;
	my ( $file, $start, $end, %h_opts ) = @_;

	if( !-e $file )
	{
		my $msg = "$file does not exist.\n";
		#throw LipmError::ObjectException( $self, '!build', -text => $msg );
        throw LipmError::IOException($file, 'f', -value => 0);
	}

	$start = 0	unless defined( $start );
	$end   = 0	unless defined( $end );

	# Get file parameters.
	my @a_attr = qw( __primary_key __start_label __end_label );
	my( $key, $start_lb, $end_lb ) = $self->_Get( @a_attr );
	my %h_param = ( start_label => $start_lb,
					end_label   => $end_lb,
					start       => $start,
					end         => $end );

	# Try to collect data.
	my $o_data;
	my %h_args = ( file => $file, primary_key_name => $key,
					%h_param, %h_opts );

	$o_data = &NARCISSE_utils::ReadChromFeatureData( %h_args );

	# returns the NarcisseData object
	return $o_data;
}


#------------------------------------------------#
#           Managing the tracks                  #
#------------------------------------------------#

=head1 The tracks

	A track defined a rectangular area where elements will be displayed.
	This rectangular area is defined by its coordinates : X0, X1, Y0 and Y1

=cut

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : Defines the track on which the ElementsSet will be displayed
 Returns      : a reference on a hash with the parameters of the track
 				  - essentially X0 and X1, the absolute X coordinates of the trak
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut
sub GetTrack
{
	warn("Must be defined fo each children object");
}

#------------------------------------------------#
#           Managing the layers                  #
#------------------------------------------------#

=head2 SetElementSetLayerName

 Title        : SetElementSetLayerName
 Usage        : $self -> SetElementSetLayerName()
 Function     : sets the layer name for the elemenSet
               (some elements don't have their own layer see %LAYER at the top of the file)

=cut
sub SetElementSetLayerName
{
	my $self = shift;
	my $layer_name = shift;

	my $name=(defined $LAYER{$layer_name}) ? $LAYER{$layer_name} : $layer_name;
	$self->_Set( __layer => $name );

	return;
}

=head2 GetGraphicParameter

 Title        : GetGraphicParameter
 Usage        : $o_x->GetGraphicParameter (parameter_label [, default_value])
 Prerequisite : none
 Function     : Return the value of the graphical parameter named 'parameter_label'.
                If it isn't set and 'default_value' is defined,
                    this parameter value is set to 'default_value'
 Returns      : parameter value
 Args         : parameter_label -> name of the parameter in the config file
                default_value	-> default value for this parameter.
                    It is advisable to fill it.
                    Do not use it anymore, policy seemed to be changed.

=cut
sub GetGraphicParameter
{
	my $self = shift;
	my ( $label, $default_value ) = @_;

	my $value = $default_value;

	my $o_param_graph = $self->_Get( '_o_param_graph' );
	$value = $o_param_graph->Get( $label )	if( $o_param_graph->IsDefined( $label ) );

	if( !defined( $value ) or $value eq '' )
	{
		my ( $pack, $file, $line ) = caller();
		my $msg = "Check your graphics.cfg file, it seems not to have $label.";
		$msg   .= " ( $file, $line )\n";
		$self->_Trace ( $msg, '.', 'IEV', $pack );
	}

	return $value;
}

=head2 GetFileName

 Title        : GetFileName
 Usage        : $self -> GetFileName()
 Prerequisite : object should be initialized
 Function     : get the url of the file which contain data for this ElementsSet object
 Returns      : a string
 Args         : none
 Attributes   : feature, organism, chromosome

=cut
sub GetFileName
{
	my $self = shift;

	my @a_attr = qw( _o_species _chromosome _feature );
	my( $o_species, $chrom, $feature ) = $self->_Get( @a_attr );

	my $filepath = $o_species->GetFeatureFiles( -chrom => $chrom, key => uc($feature) );

	return $filepath;
}

=head2 GetLocalGraphicsCfgFile

 Title        : GetLocalGraphicsCfgFile
 Usage        : $self -> GetLocalGraphicsCfgFile()
 Prerequisite : none
 Function     : get the filepath to the species specific graphics.cfg
 Returns      : a string
 Args         : none

=cut
sub GetLocalGraphicsCfgFile
{
	my $self = shift;

	my $o_species = $self->_Get( '_o_species' );
    my $o_io = $o_species->GetInstanceObserver();
	my $filename = $o_io->GetCfg( 'LOCAL_GRAPHICS_CFG' );

	my @a_dirs = $self->_Get( qw( __root_data ) );
	push( @a_dirs, $o_species->GetShortName() );

	my $file_path = File::Spec->catfile( @a_dirs, $filename );

	return $file_path;
}


=head2 GetAttributes

 Title        : GetAttributes
 Usage        : $object -> GetAttributes()
 Prerequisite : none
 Function     : Get all general attributes of the element $object.
                Used to build another ElementsSet From one.
 Returns      : hash-map containing:
 				- o_species
 				- chromosome
 				- start
 				- end
 				- graph_orientation
 Args         : none

=cut
sub GetAttributes
{
	my $self = shift;

	my $o_species = $self->_Get( '_o_species' );

	my %h_params = ( start => $self->_GetStart(),
	                 end   => $self->_GetEnd(),
	                 o_species => $o_species );

	my @a_attr = qw( _o_chrom_slice _o_map_area _o_param_graph _chromosome _graph_orientation );
	my @a_keys = qw( o_chrom_slice o_maparea o_param_graph chromosome graph_orientation );
	@h_params{ @a_keys } = $self->_Get( @a_attr );

	return %h_params;
}

#------------------------------------------------#
#               Various accessors                #
#------------------------------------------------#

=head2 GetElementField

 Title        : GetElementField
 Usage        : $object -> GetElementField($id,$field)
 Prerequisite : none
 Function     : return the value of the field $field in the hash rh_data for the key $id
 Returns      : $field value
 Args         : id : the element id
 				field : name of the field to return
 Attributes   : rh_data
 Globals      : none

=cut
sub GetElementField
{
	my $self = shift;
	my ( $id, $field ) = @_;

	my ( $value ) = $self->GetFields($id, $field );

	return $value;
}

=head2 GetFields

 Title        : GetFields
 Usage        : $object -> GetFields($id,@a_fields)
 Prerequisite : none
 Function     : return the values of each field of @fields in the hash rh_data for the key $id
 Returns      : @fields values
 Args         : see NarcisseData object

=cut
sub GetFields
{
	my $self = shift;

	# We make the assumption that the fields exist !!
	return  $self->_Get( '_o_data' )->GetFields( @_ );

}

=head2 AddFields

 Title        :
 Usage        : $object -> AddFields($id,%attributes)
 Prerequisite : none
 Function     : add or modify one or more fields to the data structure corresponding to $elt.
 Returns      : none
 Args         : see NarcisseData object

=cut
sub AddFields
{
	my $self = shift;

	return  $self->_Get( '_o_data' )->AddFields( @_ );

}

=head1 Accessors

=head2 _GetEnd, _GetStart, GetElementSetLayer, GetDNACylinderHeight

 Usage        : $object -> _GetEnd()
 Function     : Accessors for some attributes.
 Returns      : .
 Args         : none

=cut
sub _GetEnd
{
	my $self = shift;
	return $self->_Get( '__end' );
}
sub _GetStart
{
	my $self = shift;
	return $self->_Get( '__start' );
}
sub GetElementSetLayer
{
	my $self = shift;
	return $self->_Get( '__layer' );
}
sub GetDNACylinderHeight
{
	my $self = shift;

	return $self->_Get( '__dna_cylinder_height' );
}

=head2 GetNucleicMin

 Title        : GetNucleicMin
 Usage        : $object -> GetNucleicMin()
 Prerequisite : rh_data must be initialized
 Function     : Get the start of the first element, this is the minimum position.
 Returns      : integer
 Args         : none
 Attributes   : start
 Globals      : none

=cut
sub GetNucleicMin
{
	my $self = shift;

	my $first_element_ID = ($self->GetSortElements)[0];

	return $self->GetElementField($first_element_ID,'start');
}

=head2 GetNucleicMax

 Title        : GetNucleicMax
 Usage        : $object -> GetNucleicMax()
 Prerequisite : rh_data must be initialized
 Function     : Get the end of the last element, this is the maximum position.
 Returns      : integer
 Args         : none
 Attributes   : end
 Globals      : none

=cut
sub GetNucleicMax
{
	my $self = shift;

	my $last_element_ID = ($self->GetSortElements)[-1];

	return $self->GetElementField($last_element_ID,'end');
}

=head2 GetRange

 Title        : GetRange
 Usage        : $object -> GetRange()
 Prerequisite : rh_data must be initialized
 Function     : Get the range between the first and the last locus.
 Returns      : integer
 				undef if there is a problem on data set
 Args         : none
 Attributes   : start,end
 Globals      : none

=cut
sub GetRange
{
	my $self = shift;
	my $range = $self->_GetEnd() - $self->_GetStart();

	$range = 0 if ($range <= 0);		# Return 0 if the range is negative (problem on rh_data)
	return $range;
}

=head2 GetSortElements

 Title        : GetSortElements
 Usage        : $object -> GetSortElements()
 Prerequisite : data must be initialized
 Function     : Get ra_sort_elements array
 Returns      : The array ra_sort_elements
 Args         : none
 Attributes   : ra_sort_elements
 Globals      : none

=cut
sub GetSortElements
{
	my $self = shift;

	return $self->_Get( '_o_data' )->GetSortedElements();
}

=head2 GetNumElements

 Title        : GetNumElements
 Usage        : $object -> GetSortElements()
 Prerequisite : data must be initialized
 Function     : returns the number of loaded elements
 Return       : the number of loaded elements
 Args         : none

=cut
sub GetNumElements
{
	my $self = shift;

	return $self->_Get( '_o_data' )->Size();
}

=head2 Function GetDNACylinder

 Title        : GetDNACylinder
 Usage        : $o_xx->GetDNACylinder( $side );
 Procedure    : see ChromosomeSclice::GetDNACylinder
 Args         : the side of interest
 Returns      : a refence on a hash with lft and right coordinate

=cut
sub GetDNACylinder
{
	my $self = shift;
	my $side = shift;

	return $self->_Get( '__dna_cylinders' )->{$side} ;
}

#------------------------------------------------#
#          Managing the maps                     #
#------------------------------------------------#

=head2 NewMapElementSet

 Title        : NewMapElementSet
 Usage        : $o_element -> NewMapElementSet();
 Procedure    : notify the maparea object the existence of map information for the calling object
 Returns      : none
 Args         : none

=cut
sub NewMapElementSet
{
	my $self = shift;

	my( $o_map, $id ) = $self->_Get( qw( _o_map_area _id ) );
	#$o_map->NewElementSet( $id, $self );
	$o_map->NewElementSet( $self );

	return;
}

=head2 NewMapElement

 Title        : NewMapElement
 Usage        : $o_element -> NewMapElement();
 Procedure    : Adding a map element : a box (the first box is called __primary) and the element id
 Returns      : nothing
 Args         : nothing

=cut
sub NewMapElement
{
	my $self = shift;
	my ( $element_id, $ra_box ) = @_;

	my( $rh_map_area, $o_map ) = $self->_Get( qw( _rh_mapareas _o_map_area ) );
	$$rh_map_area{$element_id} = { boxes => { __primary => $ra_box },
									elt => $element_id,
									type => &SINGLE };
	$o_map->RegisterElement($element_id, $$rh_map_area{$element_id} );

	return;
}

=head2 Procedure AddMapElement

 Title        : AddMapElement
 Usage        : &AddMapElement(  $element_id, $ra_box, $tag );
 Prerequisite : $element_id mus be an existing keys
 Procedure    : adding an element area to an existing map element
 Args         :

=cut
sub AddMapElement
{
	my $self = shift;
	my( $element_id, $ra_box, $tag ) = @_;

	my $rh_map = $self->_Get( '_rh_mapareas' )->{$element_id};
	if( !defined( $rh_map ) )
	{
		throw  LipmError::ParameterException($element_id,'unknown element');
	}

	$$rh_map{ boxes }{ $tag } = $ra_box;

	return;
}

=head2 function IsaPairedMapElt

 Title        : IsaPairedMapElt
 Usage        : IsaPairedMapElt();
 Procedure    :
 Args         : none

=cut
sub IsaPairedMapElt
{
	my $self = shift ;
	my $element = shift;

	my $map_element = $self->GetMapElement( $element );
	return ( $map_element->{ type } eq &PAIRED ) ? &TRUE : &FALSE;

}

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map coordinate informations (.coord file)
 Returns      : none
 Args         : The map file (.coord file)

=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;

	#warn("DumpMapInformation :: no map information for $self");

}

=head2 GetMapElement

 Title        : GetMapElement
 Usage        : $o_element -> GetMapElement($element_id);
 Procedure    : Interface to o_maparea GetMapElement (see MapArea)
 Returns      : see MapArea
 Args         : see MapArea
 status

=cut
sub GetMapElement
{
	my $self = shift;
	my $element_id = shift;

	my $o_map = $self->_Get( '_o_map_area' );

	return $o_map->GetMapElement( $element_id ) ;
}


=head2 GetMapElements

 Title        : GetMapElements
 Usage        : $o_element -> GetMapElements();
 Procedure    :
 Returns      :
 Args         :
status        : DEPRECATED (not used any more)
=cut
sub GetMapElements
{
	my $self = shift;

	my ($element_type) = (ref($self) =~ /\:\:(\w+)$/);

	my $o_map = $self->_Get( '_o_map_area' );

	return $o_map->GetTypeArea( $element_type ) ;
}

=head2 function AllocateColors

	Title        : AllocateColors
	Usage        : $rh_indexes = &AllocateColors( $o_image, $rh_colors [,$layer] );
	Prerequisite : none
	Function     : initialize some color values and return their indexes.
	Args         :  $rh_colors, ref on hash.
						keys are color labels and values are default values.
					$layer, layer where colors belong to, optional.
	Returns      :  $rh_indexes, hash ref.
						keys are color labels and values are indexes.

=cut
sub AllocateColors
{
	my $self = shift;
	my ( $o_image, $rh_colors, $layer ) = @_;
	my %h_indexes = ();

	return \%h_indexes	unless (defined $o_image and ref($o_image) =~ /(Image)|(GD::Image)/  );
	return \%h_indexes	unless ( defined $rh_colors );

	# get colors RGB value
	while( my ( $color_label, $default_value ) = each( %$rh_colors ) )
	{
		my $rgb = $self->GetGraphicParameter( $color_label, $default_value );
		my $color_index  = $o_image->colorAllocate( $rgb, $layer );
		$h_indexes{ $color_label } = $color_index;
	}

	return \%h_indexes;
}


=head2 Method _AssertBetweenThresholds

 Usage        : return	if( 0 == $o_xx->_AssertBetweenThresholds() );
 Prerequisite : none
 Procedure    : thow exception if slice is
                    under bottom threshold
                    over top threshold
 Args         : none

=cut

sub _AssertBetweenThresholds
{
	my $self = shift;

	my @a_attr = qw( __start __end _feature );
	my ( $start, $end, $feature ) = $self->_Get( @a_attr );
	my $floor = $self->GetGraphicParameter( 'floor_threshold_' . $feature );
	my $ceil  = $self->GetGraphicParameter( 'ceil_threshold_' . $feature );

	my $delta  = $end - $start;

	if( defined $floor )
	{
		if( $delta < $floor )
		{
			my $msg = "the slice is $delta long,";
			$msg   .= " we do not show $feature for less than $floor\n";
			$self->_Trace( $msg, '.', 'ICV', 'ElementsSet' );
			throw LipmError( -text => $msg, -value => 2 ); # value = non fatal
		}
	}

	if( defined $ceil )
	{
		if( $delta > $ceil )
		{
			my $msg = "the slice is $delta long,";
			$msg   .= " we do not show $feature for more than $ceil\n";
			$self->_Trace( $msg, '.', 'ICV', 'ElementsSet' );
			throw LipmError( -text => $msg, -value => 2 ); # value = non fatal
		}
	}

	return;
} #_AssertBetweenThresholds

=head2 _TargetProgram

 Title        : G_TargetProgram
 Usage        : _TargetProgram(target_label)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get program of target_label (target_list store target data as : 'name__program')
 Returns      : string = the target program
 Args         : none
 Globals	  : target_list

=cut
sub _TargetProgram
{
	my $target_label = shift;

	# Test parameter
	return '' if (!defined $target_label);

	my $target_program;

	if ($target_label =~ /__/)
	{
		$target_program = $';			# We take the expression after '__'. Example : "mouse__prbh" => "prbh"
	}

	return $target_program;
}

=head2 Infos

 Title        : Infos
 Usage        : $self->Infos
 Function     : Returns some information abour the current object
 Returns      : string = the target program
 Args         : none
 Globals	  : target_list

=cut
sub Infos
{
	my $self = shift;

	my $start = $self->_GetStart();
	my $end   = $self->_GetEnd();
	my( $o_species, $chrom ) = $self->_Get( qw( _o_species _chromosome ) );

	my @a_fields = ( $o_species->GetShortName(), $chrom, $start, $end );

	return join("\t", @a_fields );
}


=head2 Infos

 Title        : Infos
 Usage        : ElementsSet::_inflate(X0,Y0,X1,Y1)
 Function     : inflates the size of the box for the purpose of facilitating the click
 Returns      : the inflated X0,Y0,X1,Y1 coordinates
 Args         : the usual four coordinates characterizing a box upper left corner (X0,Y0) and low right corener (X1,Y1)

=cut
sub _inflate
{
	my @a_coord = @_;

# 	carp "empty array"  if( 0 == @a_coord );
	return ( int($a_coord[0]-4) , int($a_coord[1]-2) ,int($a_coord[2]+4) , int($a_coord[3]+2) );
}
1;

__END__



=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

#
# DETECTE PAR Devel::Cycle::find_cycle !
# Mais je n'ai pas reussi à les enlever pour eviter les fuites de memoire
#
Cycle (2):
	$ElementsSet::Synt::N->{'_o_map_area'} => \%MapArea::O                  
	$MapArea::O->{'__displayed_elementsets'} => \%P                           
	$P->{'ElementsSet::Cyto=HASH(0x442da18)'} => \%ElementsSet::Cyto::Q        
	$ElementsSet::Cyto::Q->{'_o_chrom_slice'} => \%ChromosomeSlice::Reference::R
	$ChromosomeSlice::Reference::R->{'h_elements'} => \%S                           
	                  $S->{'Gene'} => \%ElementsSet::Gene::T        
	$ElementsSet::Gene::T->{'_o_chrom_slice'} => \%ChromosomeSlice::Reference::R


