# Copyright INRA/CNRS

=pod

# The Cecill license

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=cut

=head1 NAME

ElementsSet inherited object to display Narcisse cytological data.

=head1 DESCRIPTION

To draw cyto labels, set show_label attribute value to 1.

See L<ElementsSet>

=head1 SUBROUTINES

Display

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Cyto;

use strict;
use warnings;

use Data::Dumper;
use File::Path;
use File::Basename;
use ElementsSet qw( _inflate );

use base qw( ElementsSet );

use General;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end [, $rh_param])
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : rh_cfg		=> ref of the cfg hash map
 				organism	=> organism
 				chr			=> chromosome
 				start		=> start
 				end			=> end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- organism
 					- chromosome
 					- start
 					- end

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
				show_label	=> 0,

				@_
				);

	# New attributes :
	$self -> {show_label} = $h_param{show_label};

	# Change defaut parameter : ra_data_headers, ra_map_keys
	my @a_map_keys     = ('label');
	my @a_data_headers = ('element','start','end','color');
	my %h_new_param = ( ra_data_headers => \@a_data_headers,
						ra_map_keys		=> \@a_map_keys,

						%h_param );

	# Base class initialization
	$self -> SUPER::_Init( %h_new_param );

    return $self;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw name of each data and their rectangle representation.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box => a box object associated to the chromosome slice of the elements under consideration
 					side  => optional, 'right' or 'left'
 						 }
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side	=> 'right',
			@_
			);

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	#Some elements will be displayed, we register the current object in MapArea
	$self->NewMapElementSet();

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box	= $h_param{o_box};
	my $side 	= $h_param{side};

	#======================================
	# image layer
	my $layer = $self->GetElementSetLayer();

	# Colors
	my $rh_cytobandcolor = $self->_GetCytoBandsColors( $o_image, $layer );
	#my $color_clear_grey = $self->GetGraphicParameter('color_cyto_pos','grey_clear');
	#my $color_dark_grey  = $self->GetGraphicParameter('color_cyto_neg','grey_dark');
	#my $gpos  = $o_image->colorAllocate($color_clear_grey,$layer);
	#my $gneg  = $o_image->colorAllocate($color_dark_grey,$layer);
	
	my $black = $o_image->colorAllocate('black', $layer);

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );

	my $X0 = $rh_track->{ X0 };
	my $X1 = $rh_track->{ X1 };

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	# coords of the centromere
	my ( $y0_centro, $y1_centro );

	#======================================
	# Draw and store each elements.
	my @a_fields = ('chrom','element','start','end','color');
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	foreach my $element_id ($self->GetSortElements())
	{
		my ($chrom,$element,$start,$end,$elt_color) = $self->GetFields($element_id, @a_fields);
		my @a_coord = ();
			#warn "$element_id $element $start $end $elt_color";
		# Bounds test : the first and the last element are set in the frame limits:
		next if ( $start > $elt_end || $end < $elt_start );

		#Restrict the element to the drawing area defined by the start and end of this ElsementSet object
		my $draw_y0 = ( $start > $elt_start ) ? $start : $elt_start;
		my $draw_y1 = ( $end < $elt_end ) ? $end : $elt_end;

		if( 'acen' ne $elt_color )
		{
			# dessin de la bande
			#my $color = ($elt_color eq 'gpos') ? $gpos : $gneg;
			#We use the predefined colors and gvar when not available
			my $color = ( defined $rh_cytobandcolor->{ $elt_color } ) ? $rh_cytobandcolor->{ $elt_color } : $rh_cytobandcolor->{ "gvar" };
			@a_coord  = $o_image -> borderedFilledRectangle($X0,$draw_y0,$X1,$draw_y1, $color, $layer);
			my $msg   = "cyto : ($a_coord[0],$a_coord[1]) to ($a_coord[2],$a_coord[3])\n";
			$self->_Trace( $msg, '.', 'IEV' );
			# pas encore de logger !

			# stockage des coordonnees pour les cartes de clic
			$self->NewMapElement( $element_id, \@a_coord  );
		}
		else
		{
			# les donnees arrivent triees
			# on stocke le debut et fin du centromere pour plus tard
			$y0_centro = $start unless defined $y0_centro;
			$y1_centro = $end;
		}
	}# fin lecture ligne a ligne du fichier de donnees.

	$self->__DisplayCentromere( $o_image, $X0,$y0_centro, $X1,$y1_centro, $black );

	return;
}

=head2 GetCytoBandsColors

 Title        : GetCytoBandsColors
 Usage        : $self->_GetCytoBandsColors( $o_image, $layer)
 Procedure    : Construct a hash with the corresponding cytological colors
 Returns      : a reference on the hash
 Args         : the $o_image and the corresponding $layer

=cut
sub _GetCytoBandsColors
{	
	my ($self, $o_image, $layer) = @_;
    
    my %h_cytobandcolor = ();
	my @a_bandcolors = qw( gpos100 gpos75 gpos66 gpos50 gpos33 gpos25 gpos gneg gvar );
	foreach my $color ( @a_bandcolors )
	{
		my $colorname = "grey_".$color;
		$h_cytobandcolor{ $color } = $o_image->colorAllocate( $colorname, $layer);
	}
	
	return \%h_cytobandcolor;
}

=head2 DisplayText

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw name of each data and their rectangle representation.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box => a box object associated to the chromosome slice of the elements under consideration
 					side  => optional, 'right' or 'left'
 						 }
 Globals      : none

=cut
sub DisplayText
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side	=> 'right',
			@_
			);

	# Textual layer
	my $layer = $self->GetElementSetLayer();
	my $o_textual = $h_param{o_textual};
	return unless ($o_textual -> IsLayer ( $layer ) == &TRUE);
	
	my $o_file= $o_textual -> GetLayer( $layer );

	my @a_attr = qw( _o_species _chromosome );
	my ( $o_species, $chr ) = $self->_Get( @a_attr );
	my $species_name = $o_species -> GetLongName();

	#======================================
	# Print each element
	my @a_fields = ('element','start','end','color');
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();
	$o_textual -> PrintHeader ( $layer, 'species', 'chromosome', @a_fields );
	foreach my $element_id ($self->GetSortElements())
	{
		my ($element,$start,$end,$elt_color) = $self->GetFields($element_id, @a_fields);
		next if ( $start > $elt_end || $end < $elt_start );
		print $o_file join ( "\t", $species_name, $chr, $element, $start, $end, $elt_color )."\n";

	}
	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : Defines the track on which the ElemenSet will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1 position of the cyto track
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut

sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_
					);

	#We need the specification of the box to create it
	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side  = $h_param{ side };

	my %h_track_param;

	# x position of the cyto-representation
	my $x_cyto = $self->GetGraphicParameter('x_cyto',100);

	# width of the cyto representation
	my $width  = $self->GetGraphicParameter('width_cyto',10);

	my ( $X0, $X1 );

	if ( $side eq "right" )
	{
		$X0 = $o_box->GetX0() + $x_cyto;
		$X1 = $X0 + $width;
	}
	else
	{
		$X1 = $o_box->GetX1() - $x_cyto;
		$X0 = $X1 - $width;
	}

	%h_track_param = ( X0 => $X0, X1 => $X1, desc => "cyto elements");

	return \%h_track_param;

}

=head2 DumpMapInformation

 Title     : DumpMapInformation
 Usage     : $o_xxx -> DumpMapInformation( $filehandle )
 Procedure : Writes the map information corresponding to the cyto elements
 	The file format is the following
 	X0 Y0 X1 Y1 KARYOTYPE species chromosome
 Returns   : none
 Args      : The filehandle of the .coord file

=cut

sub DumpMapInformation
{
	my $self = shift;
	my( $fh_outfile ) = @_;

	my( $rh_map, $o_species ) = $self->_Get( '_rh_mapareas','_o_species' );
	my @a_fields = ('chrom','element','start','end','color');
	my $species = $o_species->GetLongName();
	foreach my $rh_box ( values %$rh_map )
	{
			my $ra_coords = $$rh_box{ boxes }{ __primary };
			my @a_coords  = ElementsSet::_inflate( @$ra_coords );
            my $ref_elt   = $$rh_box{ elt };
			my ($chrom,$elt,$strt,$end,$col) = $self->GetFields( $ref_elt, @a_fields );

			#my $line = join( "\t", @a_coords, "KARYOTYPE",$species,@a_values );
			my $line = join( "\t", @a_coords, '"KARYOTYPE"',"\"$species\"","\"$chrom\"","\"$elt\"","\"$strt\"","\"$end\"","\"$col\"");
			print $fh_outfile "$line\n";

	}

	return;
}

=head2 __DisplayCentromere
 Usage        : $o_x->__DisplayCentromere( $X0,$y0, $X1,$y1 )
 Function     : Draw two diagonales.
 Args         : $X0,$y0, $X1,$y1

=cut

sub __DisplayCentromere
{
	my $self   = shift;
	my( $o_image, $X0,$y0, $X1,$y1, $color ) = @_;

	return unless( defined $y0 && defined $y1 );

	$o_image->line( $X0,$y0, $X1,$y1, $color );
	$o_image->line( $X1,$y0, $X0,$y1, $color );

	return;
}

=head2 DumpMapInformation

 Title     : DumpMapInformation
 Usage     : $o_xxx -> DumpMapInformation( $filehandle )
 Procedure : Writes the map information corresponding to the cyto elements
 	The file format is the following
 	X0 Y0 X1 Y1 KARYOTYPE species chromosome
 Returns   : none
 Args      : The filehandle of the .coord file

=cut

sub DumpMapInformation_SUPPRESSED
{
	my $self = shift;
	my( $fh_outfile ) = @_;

	my( $rh_map, $o_species ) = $self->_Get( '_rh_mapareas','_o_species' );
	my @a_fields = ('chrom','element','start','end','color');
	my $species = $o_species->GetLongName();
	foreach my $rh_box ( values %$rh_map )
	{
			my $ra_coords = $$rh_box{ boxes }{ __primary };
			my @a_coords  = ElementsSet::_inflate( @$ra_coords );
            my $ref_elt   = $$rh_box{ elt };
			my @a_values  = $self->GetFields( $ref_elt, @a_fields );
			my @a_quoted_values = map ( "\"$_\"", @a_fields);

			my $line = join( "\t", @a_coords, "\"KARYOTYPE\"","\"$species\"",@a_quoted_values );
			print $fh_outfile "$line\n";

	}

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


# ------------------------------------------------------
# 	sebastien.letort@toulouse.inra.fr
#	Created: January 30, 2009
#	$Id$
#

package ElementsSet::Cyto::Karyotype;

=pod

=head1 NAME

 Cyto::Karyotype

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet::Cyto );

=head2 procedure _Init

 Usage        : my $o_x = New ElementsSet::Cyto::Main( %h );
 Prerequisite : none
 Fonction     : initialise an instance of ElementsSet::Cyto::Main class.
 Returns      : a valid ElementsSet::Cyto::Main object
 Args         :  %h

=cut

sub _Init
{
	my $self = shift;
	my %h = ( o_io => undef,
        feature => 'cyto',
        file_directory => 'cyto',
        file_extension => 'cyto',
        @_ );

	$self->SUPER::_Init( %h );
# 	warn Dumper $self;

	return;
}


=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : Defines the track on which the ElemenSet will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1 position of the cyto track
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut

sub GetTrack
{
	my $self   = shift;
	my %h_args = (  o_box => undef,
					@_ );

	# mandatories
	LipmError::ParameterException::Assert( 'o_box' => $h_args{ o_box } );

	my %h_box = $h_args{ o_box }->GetDisplayedArea();
	my $box_width = $h_box{ X1 } - $h_box{ X0 };

	# width of the cyto representation
	my $width  = $self->GetGraphicParameter( 'width_cyto',10 );

	my $X0 = ( $h_box{ X0 } + $h_box{ X1 } - $width ) / 2;
	my $X1 = $X0 + $width;

	return { X0 => $X0, X1 => $X1,
	         Y0 => $h_box{ Y0 }, Y1 => $h_box{ Y1 },
	         desc => 'cyto elements for karotype' };

}


sub Display
{
	my $self = shift;
	$self->SUPER::Display( @_ );

	my %h_param = ( image 	=> undef,
					o_box   => undef,
					side	=> 'right',
					@_ );

	my $o_image = $h_param{ image };

	# add half circle on the ends
	my $black = $o_image->colorAllocate('black', $o_image->GetCurrentLayer() );

	my $rh_track = $self->GetTrack( @_ );
	my ( $X0, $X1 ) = @$rh_track{ qw( X0 X1 ) };

	my $width = $X1 - $X0;

	my $cx = $X0 + $width/2;
	$o_image->arc( $cx,$self->_GetStart(), $width,$width, 180,360, $black );

	$cx = $X1 - $width/2;
	$o_image->arc( $cx,$self->_GetEnd(), $width,$width, 0,180, $black );

	# label du chrom
	my $label = $self->_Get( '_chromosome' );
	my $font  = GD::Font->Large;
	my $label_width = length( $label ) * $font->width();
	my $sx = $cx - $label_width/2;
	# j'y arrive pas avec la methode de Image
# 	$o_image->string( $font, $cx, $font->height(), $label, $black, 'center' );

	my $sy = $$rh_track{ Y0 } - $width - $font->height();
	$o_image->GetLayer( 'karyotype' )-> string($font, $sx,$sy, $label, $black );

	return;
}

1;
