# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet inherited object to display Narcisse exons data.

=head1 DESCRIPTION

See L<ElementsSet>

=head1 SUBROUTINES

GetExonsList

=cut
package ElementsSet::Exon;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end)
 Prerequisite : none
 Procedure    : Initialization called by __LoadExons in the constructor of Gene
 Returns      : reference of the blessed object
 Args         : see the parent constructor

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = ( @_ );

	# Change defaut parameter : file extension
	my @a_data_headers = ('symbol','start','end','strand');
	my %h_new_param = ( #file_directory => 'exon',
						#file_extension => 'exon',
						ra_data_headers => \@a_data_headers,
						%h_param
						);

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

	# New attributes :
	$self->{displayed_start} = $h_param{displayed_start};
	$self->{displayed_end}   = $h_param{displayed_end};

	# Test new attributes :
	$self->{displayed_start} = $self->_GetStart() if (!defined $self->{displayed_start});
	$self->{displayed_end}   = $self->_GetEnd()   if (!defined $self->{displayed_end});

	#Organize data in order to facilitate subsequent extraction of the exons corresponding to a particular gene
	$self->__OrganizeExonData();

    return $self;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw data and store positions.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box => a box object associated to the chromosome slice of the elements under consideration
 					side  => optional, 'right' or 'left'
 					exons_list => list of elements to draw
 						 }
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			exons_list	=> undef,
			strand => 1,
			@_
			);

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	return unless ( defined $h_param{exons_list} );

	#Some elements will be displayed, we register the current object in MapArea
	$self->NewMapElementSet();

	my $strand = $h_param{ strand }; #Exons of the same gene are always on the same strand

	# General parameters
	my $o_image				= $h_param{image};
	my $o_box				= $h_param{o_box};
	my $side 				= $h_param{side};
	my @a_sort_elements		= @{$h_param{exons_list}};
	@a_sort_elements 		= reverse(@a_sort_elements) if ($self->_Get( '_graph_orientation' ) eq '-');

	# Test if we have to draw exons twice (for the reference representation)
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left');
	}

	#======================================
	# Graphic parameters :

	#The image layer
	my $layer = $self->GetElementSetLayer();

	# colors
	my $color_exons 	= $self->GetGraphicParameter('color_exons','green');
	my $color 		= $o_image -> colorAllocate($color_exons,$layer);
	my $link_color 	= $o_image -> colorClearer($color,2,$layer);

	# x positions
	my $rh_track = $self->GetTrack(  o_box => $o_box, side => $side );

	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };
	my $width_exon_link = $rh_track->{ link_width };

	#From now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	#======================================
	# Draw and store each exons.
	my $previous_end = undef;
	my @fields = qw( symbol start end );
	my @drawed_exons = ();

	# c'est quoi la difference entre displayed_start et _GetStart ? <- seb
	my $elt_start = $self->{ displayed_start }; # $self->_GetStart();
	my $elt_end   = $self->{ displayed_end   }; # $self->_GetEnd();

	foreach my $element_id (@a_sort_elements)
	{

		my ( $name, $start, $end ) = $self->GetFields($element_id, @fields );

		next if( $end < $elt_start );
		last if( $start > $elt_end );

		$start = $elt_start if( $start < $elt_start );
		$end   = $elt_end   if( $end   > $elt_end   );

		my @a_box_coord = 	$self->DrawElement( $o_image, $x0, $start, $x1, $end, $color, $strand);
		my @a_coord = $self->DrawSuperImposedElement( $o_image, $x0, $start, $x1, $end, $color, $strand);

		push @drawed_exons, { elt => $element_id, coord => \@a_coord, Y_pos => $a_coord[1]};

		#$self->NewMapElement($self->GetElement($element_id),$element_id, \@a_coord  );
		$self->NewMapElement($element_id, \@a_coord  );

	}

	$o_image -> SetAbsoluteFrame();
	#------------------------------------
	# 2) Draw a link between exons

	my @sorted_drawed_exons = sort { $a->{ Y_pos } <=> $b->{ Y_pos }} @drawed_exons;
	for my $i (0.. ($#sorted_drawed_exons-1) )
	{
		$self->DrawExonLink($o_image, $sorted_drawed_exons[$i],$sorted_drawed_exons[$i+1], $width_exon_link, $link_color, $strand);
	}
	$o_image -> SetRelativeFrame();

	return;
}

=head2 DisplayText

 Title        : DisplayText
 Usage        : $o_element -> DisplayText(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw data and store positions.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box => a box object associated to the chromosome slice of the elements under consideration
 					side  => optional, 'right' or 'left'
 					exons_list => list of elements to draw
 						 }
 Globals      : none

=cut
sub DisplayText
{
	my $self = shift;
	my %h_param = (
			o_textual	=> undef,
			exons_list	=> undef,
			strand      => 1,
			locus_tag   => undef,
			@_
			);

	return unless ( defined $h_param{exons_list} );

	my @a_attr = qw( _o_species _chromosome );
	my ( $o_species, $chr ) = $self->_Get( @a_attr );
	my $strand    = $h_param{ strand }; #Exons of the same gene are always on the same strand
	my $locus_tag = $h_param{ locus_tag };

	# General parameters
	my $o_textual			= $h_param{o_textual};
	my @a_sort_elements		= @{$h_param{exons_list}};
	@a_sort_elements 		= reverse(@a_sort_elements) if ($self->_Get( '_graph_orientation' ) eq '-');

	#======================================
	# Graphic parameters :

	#The image layer
	my $layer = $self->GetElementSetLayer();
	my $o_file= $o_textual -> GetLayer( $layer );


	# Draw and store each exons.
	my $previous_end = undef;
	my @fields = qw( symbol start end );
	my @drawed_exons = ();

	# c'est quoi la difference entre displayed_start et _GetStart ? <- seb
	my $elt_start = $self->{ displayed_start }; # $self->_GetStart();
	my $elt_end   = $self->{ displayed_end   }; # $self->_GetEnd();
	
	my $species_name = $o_species -> GetLongName();
	$o_textual -> PrintHeader ( $layer, 'species', 'chromosome', 'parent','symbol','start','end' );
	foreach my $element_id (@a_sort_elements)
	{

		my ( $name, $start, $end ) = $self->GetFields($element_id, @fields );

		next if( $end < $elt_start );
		last if( $start > $elt_end );

		$start = $elt_start if( $start < $elt_start );
		$end   = $elt_end   if( $end   > $elt_end   );
		
		print $o_file join ("\t",$species_name,$chr,$locus_tag,$name,$start,$end)."\n";
	}
	return;
}

=head2 DrawSuperImposedElement

 Title        : DrawSuperImposedElement
 Usage        : $object -> DrawSuperImposedElement($x0,$y0,$x1,$y1,$dna_cyl,$color,$side);
 Prerequisite : The appropriate layer should have been set
 Procedure    : Print superimposed exon ( some pixels away from the DNA cyinder): start - end for each element
 Returns      : the coordinates
 Args         : bunch

=cut
sub DrawSuperImposedElement
{
	my $self = shift;
	my ($o_image, $x0, $y0, $x1, $y1, $color, $dna_strand) = @_;

	my $width_exon  = $self->GetGraphicParameter('width_exon',3);		# width of exon representation
	my $space_gene_exon = $self->GetGraphicParameter('space_default',5);

	my ( $x0_rect, $x1_rect );

	if ( $dna_strand > 0 )
	{
		$x0_rect = $x1 + $space_gene_exon;
		$x1_rect = $x0_rect + $width_exon;
	}
	else
	{
		$x1_rect = $x0 - $space_gene_exon;
		$x0_rect = $x1_rect - $width_exon;
	}

	#Restrict the element to the drawing area defined by the start and end of this ElsementSet object
	my $draw_y0 = ( $y0 > $self->_GetStart() ) ? $y0 : $self->_GetStart();
	my $draw_y1 = ( $y1 < $self->_GetEnd() ) ? $y1 : $self->_GetEnd();

	return $o_image->filledRectangle ( $x0_rect, $draw_y0, $x1_rect, $draw_y1, $color);


}

=head2 DrawExonLink

 Title        : DrawExonLink
 Usage        : $object -> DrawExonLink($o_image, $rh_exon1, $rh_exon2, $spacer, $link_color, $dna_strand);
 Prerequisite : The superimposed exons must have been drawn
 Procedure    : Draw links between exons (according to absolute coordinates)
 Returns      : none
 Args         :

=cut
sub DrawExonLink
{
	my $self = shift;
	my ($o_image, $rh_exon1, $rh_exon2, $spacer, $link_color, $dna_strand) = @_;

	my @coord_exon1 = @{ $rh_exon1->{coord} };
	my @coord_exon2 = @{ $rh_exon2->{coord} };

	if ( $dna_strand > 0 )
	{
		my ($x0,$y0) = @coord_exon1[2,3];
		my ($x1,$y1) = @coord_exon2[2,1];

		my $mid_point = ( $y0 + $y1 )/2;

		$o_image ->line ($x0,$y0,$x0+$spacer,$mid_point,$link_color);
		$o_image ->line ($x0+$spacer,$mid_point,$x1,$y1,$link_color);
	}
	else
	{
		my ($x0,$y0) = @coord_exon1[0,3];
		my ($x1,$y1) = @coord_exon2[0,1];

		my $mid_point = ( $y0 + $y1 )/2;

		$o_image ->line ($x0,$y0,$x0-$spacer,$mid_point,$link_color);
		$o_image ->line ($x0-$spacer,$mid_point,$x1,$y1,$link_color);
	}


}

=head2 DumpMapInformation

 Title        :
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the exon elements
 	X0 Y0 X1 Y1 EXON  species element start end strand parent symbol
 Returns      : none
 Args         : A filehandle for the map file (.coord file)

=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;

	#Dump exon map
	my @fields = qw( element start end strand parent symbol);

	my @a_attr = qw( _feature _o_species _rh_mapareas );
	my( $feature, $o_species, $rh_map ) = $self->_Get( @a_attr );

	foreach my $element ( keys %$rh_map )
	{
		my $rh_info = $$rh_map{ $element };
		my @coords  = ElementsSet::_inflate(@{$rh_info->{boxes}{__primary}});
		my @values = ( uc($feature), $o_species->GetLongName() );
		push( @values, $self->GetFields($element,@fields) );
		my @quoted_values = map ("\"$_\"",@values);
		#print $file_handle join("\t",@coords)."\t",join("\t",@values)."\n";
		print $file_handle join("\t",@coords)."\t",join("\t",@quoted_values)."\n";
	}

	return;
}

=head2 __OrganizeExonData
 Title        : __OrganizeExonData
 Usage        : $self -> __OrganizeExonData()
 Prerequisite : object should be initialized
 Function     : Constructs a hash structure with the parent field of the exons as the key, and a reference
                on the list of the corresponding exons as value
                Will be usefull for the subsequent extraction of exons for each transcript
 Returns      : none
 Args         : none
 Globals      : none

=cut
sub __OrganizeExonData
{
	my $self = shift;

	my %h_organized_exons;

	foreach my $exon ( $self->GetSortElements() )
	{
		my ($parent) = $self->GetFields( $exon, ('parent') );
		push @{ $h_organized_exons{ $parent } }, $exon;
	}
	$self->{ __organised_exons } = \%h_organized_exons;

	return;
}

=head2 GetExonsList
 Title        : GetExonsList
 Usage        : $self -> GetExonsList(gene_name [,transcript_number])
 Prerequisite : object should be initialized
 Function     : Return the list of exons for the transcript of the gene named gene_name.
 				By default, the method return results for the first transcript.
 Returns      : Array of sort elements.
 Args         : none
 Globals      : none

=cut
sub GetExonsList
{
	my $self = shift;
	my %h_args = (
					gene => undef,
					transcript => undef,
					@_
				);
	(defined $h_args{ gene } )
		or throw LipmError::ParameterException('gene','missing');

	my $transcript_id = (defined $h_args{ transcript }) ? $h_args{ transcript } : $h_args{ gene }."t1";

	return @{ $self->{ __organised_exons }{$transcript_id } };
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     :
 Returns      :
 Args         :

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_
					);

	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };

	my %h_track_param;

	my $width_exon_link = $self->GetGraphicParameter('width_exon_link',3);

	my $rh_dna_cyl = $self->GetDNACylinder( $side ) ;

	my $X0 = $rh_dna_cyl->{X0} ;
	my $X1 = $rh_dna_cyl->{X1};

		%h_track_param = ( X0 => $X0, X1 => $X1,  link_width => $width_exon_link, desc => "exon elements");

	return \%h_track_param;

}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
