# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet inherited object to display Narcisse genes data.
Also use to draw graduations.

=head1 DESCRIPTION

See also L<ElementsSet>

Genes is an ElementsSet inherit class to get genes data and display them.
It loads transcripts and exons data too.

B<Attributes :>
- exons
- transcripts


=head1 SYNOPSIS

See L<ElementsSet/Synopsis>

=head1 SUBROUTINES

__LoadExons
__StoreFirstTranscriptData
Display
DisplayGraduation
GetExon
GetMap
GetTranscripts
AreExonsDisplayed

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Gene;

use strict;
use warnings;

use Data::Dumper;
use Error qw( :try );

use General;

use ElementsSet::Transcript;
use ElementsSet::Exon;

use ElementsSet qw( _inflate );

use base qw( ElementsSet );

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

use constant X0 => 0;
use constant Y0 => 1;
use constant X1 => 2;
use constant Y1 => 3;

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init(o_species, $chromosome, $start, $end [, $rh_param])
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 				chr			=> chromosome
 				start		=> start
 				end			=> end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- chromosome
 					- start
 					- end
 					- transcripts

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
        gene_structure => 0,
#         gene_search    => undef,
        @_
        );

	# New attributes :

	$self -> {gene_structure} = $h_param{gene_structure};
	$self -> {gene_search}	  = $h_param{gene_search};
	# TODO - je crois bien que gene_structure est TOUJOURS a 1
# 	warn "KOUKOU " . $self -> {gene_structure};

	# Change defaut parameter : file extension, ra_data_headers, ra_map_keys
	my @a_map_keys     = ('strand','element_id','name','locus_tag','protein_id','prbh','accession');
	my @a_data_headers = ('symbol','start','end','strand','locus_tag');
	my %h_new_param = (
#         file_extension	=> 'gene',
        ra_data_headers => \@a_data_headers,
        ra_map_keys		=> \@a_map_keys,
        storage_mode	=> 'array',
        %h_param
        );

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

	return $self	if ( $self->IsEmpty() );

	# Load transcipt informations and store transcript data.
	$self -> __StoreTranscriptData();

	# Load exons informations if gene_structure is set to 1.
	# NO - Now loaded when needed, in Display or DisplayText
	# $self -> __LoadExons() if ($self->AreExonsDisplayed() == 1);

    return $self;
}

#------------------------------------------------#
#               Procedures                       #
#------------------------------------------------#

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw name of each genes and a 'cylinder-art' representation.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box   => Box object
 					side  => optional, 'right' or 'left'
 					b_label => optional, boolean to draw genes names
 						 }
=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			side	=> 'right',
			o_box => undef,
			b_label	=> 1,
			@_
			);

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	#Some elements will be displayed, we register the current object in MapArea
	$self->NewMapElementSet();

	# General parameters
	my $o_image				= $h_param{image};
	my $o_box				= $h_param{o_box};
	my $side 				= $h_param{side};
	my $b_label				= $h_param{b_label};

	my @a_attr = qw( _o_species _chromosome _graph_orientation );
	my ( $o_species, $chr, $graph_orientation ) = $self->_Get( @a_attr );

	# Test if we have to draw elements twice (for the reference representation)
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left', b_label => 0);
	}

	#======================================
	# Graphics parameters

	# image layer
	my $layer = $self->GetElementSetLayer();

	# fonts
	my $font 				= $o_image -> gdTinyFont;

	# Colors
	my $color_grey_blue			= $self->GetGraphicParameter('color_genes_neg','grey_blue');
	my $color_grey_blue_dark 	= $self->GetGraphicParameter('color_genes_pos','grey_blue_dark');
	my $color_grey_blue_darker 	= $self->GetGraphicParameter('color_genes_label','grey_blue_darker');
	my $alpha 					= $self->GetGraphicParameter('alpha_genes',30);

	my $neg_color 		= $o_image->colorAllocateAlpha($color_grey_blue,$alpha,$layer);
	my $pos_color 		= $o_image->colorAllocateAlpha($color_grey_blue_dark,$alpha,$layer);
	my $neg_label_color = $o_image->colorAllocate($color_grey_blue,$layer);
	my $pos_label_color = $o_image->colorAllocate($color_grey_blue_darker,$layer);

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );

	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };

	my $x_label = $rh_track->{ X_label };
	my $width_genes_label 	= $rh_track->{ width_label };
	my $max_char 			= $width_genes_label/$font->width();

	my $inversion = ( $graph_orientation eq '-' ) ? -1 : 1;

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	#======================================
	# Draw genes if they are between the object start or end
	my $old_label_position = ($graph_orientation eq '-') ? $rh_track->{Y1} : $rh_track->{Y0};
	my @fields = qw( symbol start end strand locus_tag );
	my @transcript_fields = qw( protein_id prbh accession );
	my $bool_name_displayed;
	foreach my $element_id ( $self->GetSortElements() )
	{
		my ( $symbol, $start, $end, $strand, $locus_tag ) = $self->GetFields( $element_id, @fields );
		#my (protein_id,$prbh,$accession) = à suivre
		my @a_box_coord = ();
		my @a_label_coord = ();

		#Reverse strand is the slice is represented up-side-down (graph_orientation = -1)
		$strand *= $inversion;

		#----------------------#
		# 1) Gene representation
		next if ($end < $self->_GetStart() || $start > $self->_GetEnd() );
		my $gene_color = ($strand eq '-1') ? $neg_color : $pos_color;

		@a_box_coord = 	$self->DrawElement( $o_image, $x0, $start, $x1, $end, $gene_color, $strand);

		#----------------------#
		# 2) Gene Label
		my $name = ($symbol ne '') ? $symbol : $locus_tag;
		$bool_name_displayed = 0;

		if ($b_label != 0)
        {
	        # Calculate the position of the label to set it at the center of the rectangle representation
			my ( $Y0, $Y1 )= @a_box_coord[Y0,Y1];
			my $y_label = ( $Y0 + $Y1 )/2;

	        # Test name width and cut it if it's too long.
	        my $text = $name;
			substr($text,int($max_char/2)-1,length($text)-($max_char-2)) = '..' if (length($name) > $max_char);

	        # Draw gene name if space between this label and the one below is superior to 1px
			my $name_space = ($graph_orientation eq '-') ? ($old_label_position-$y_label) : ($y_label-$old_label_position);
			my $underline = $self->__IsGeneSearched($symbol,$locus_tag);

			if ( $name_space > $font->height() or $underline )
			{
				my $name_color = ($strand eq '-1') ? $neg_label_color : $pos_label_color;
				@a_label_coord = $o_image -> stringVerticalAligned($font, $x_label, $y_label, $text, $name_color, side => $side, underline => $underline, absolute_frame => 1 );

				$old_label_position = $y_label;
				$bool_name_displayed = 1;
			}
        }

		$self->NewMapElement( $element_id, \@a_box_coord  );
		$self->AddMapElement( $element_id, \@a_label_coord, "name"  ) if ( $bool_name_displayed );

	}

	#======================================
	# Draw exons of all genes represented
	#TODO  The method should be called draw transcript
	# GetExonsList is not correctly written
	if ($self->AreExonsDisplayed() == 1)
	{
		$self -> __LoadExons();
		foreach my $ref_elt ($self->GetSortElements())
		{
			my ( $element_id, $strand ) = $self->GetFields( $ref_elt, ( 'element','strand' ) );
			my @a_exons_list = $self -> GetExonsList(gene => $element_id);

			next if (scalar(@a_exons_list) == 0);
			$strand *= $inversion;
			$self->{ o_exon } -> Display(%h_param, exons_list => \@a_exons_list, strand => $strand );
		}
	}
	#======================================


	return;
}

=head2 DisplayText

 Title        : DisplayText
 Usage        : $o_element -> DisplayText( textual => $o_textual );
 Procedure    : Print textual data about the genes
 Returns      : none
 Args         : %h_param {
 					textual => Textual object
 						 }
=cut
sub DisplayText
{
	my $self = shift;
	my %h_param = (
			o_textual 	=> undef,
			@_
			);

	# General parameters
	my $o_textual = $h_param{o_textual};

	my @a_attr = qw( _o_species _chromosome _graph_orientation );
	my ( $o_species, $chr, $graph_orientation ) = $self->_Get( @a_attr );
	my $inversion = ( $graph_orientation eq '-' ) ? -1 : 1;

	#======================================
	# Graphics parameters

	# image layer
	my $layer = $self->GetElementSetLayer();
	my $o_file= $o_textual -> GetLayer( $layer );

	#======================================
	# Print genes if they are between the object start or end
	my @fields = qw( symbol start end strand locus_tag );
	my @transcript_fields = qw( protein_id prbh accession );
	my $species_name = $o_species -> GetLongName();
	$o_textual -> PrintHeader ( $layer, 'species', 'chromosome', @fields );
	foreach my $element_id ( $self->GetSortElements() )
	{
		my ( $symbol, $start, $end, $strand, $locus_tag ) = $self->GetFields( $element_id, @fields );
		next if ($end < $self->_GetStart() || $start > $self->_GetEnd() );
		print $o_file join ( "\t", $species_name, $chr, $symbol, $start, $end, $strand, $locus_tag )."\n";

	}

	#======================================
	# Print exons of all genes represented, whatever the range
	if ($self->AreExonsDisplayed() == 1)
	{
		$self -> __LoadExons(); 
		foreach my $ref_elt ($self->GetSortElements())
		{
			my ( $element_id, $strand, $locus_tag ) = $self->GetFields( $ref_elt, ( 'element','strand','locus_tag' ) );
			my @a_exons_list = $self -> GetExonsList(gene => $element_id);
	
			next if (scalar(@a_exons_list) == 0);
			$strand *= $inversion;
			$self->{ o_exon } -> DisplayText(%h_param, exons_list => \@a_exons_list, strand => $strand, locus_tag => $locus_tag );
		}
	}
	#======================================


	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> GetTrack( )
 Function     : Defines the track on which the ElemenSet will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - the coordinates of the the dna cylinder
                 - X_label : absolute X position for the display of the label
                 - width_label : size of space allocated to the gene names
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_
					);

	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };

	my %h_track_param;

	my $x_genes_label   = $self->GetGraphicParameter('x_genes_label',0.275);	# width of rna representation
	my $width_genes_label 		= $self->GetGraphicParameter('width_genes_label',50);	# max width for gene labels

	my $rh_dna_cyl = $self->GetDNACylinder( $side );

	my $X0  = $rh_dna_cyl->{X0};
	my $X1  = $rh_dna_cyl->{X1};

	my %h_area=$o_box->GetBoxArea();

	my $X_label;

	if ( $side eq "right")
	{
		$X_label = $h_area{X0} + $x_genes_label;
	}
	else
	{
		$X_label = $h_area{X1} - $x_genes_label;
	}

	%h_track_param = ( %{$rh_dna_cyl}, width_label => $width_genes_label, X_label => $X_label, desc => "gene elements and gene label");

	return \%h_track_param;

}

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the gene elements
                The file format is the following
                X0 Y0 X1 Y1 GENE speciesLongName chrom start end strand element symbol locus_tag protein_id element accession X'0 Y'0 X'1 Y'1
                X0 Y0 X1 Y1 are the coordinates of the gene name box
                X'0 Y'0 X'1 Y'1 are the coordinates of the gene on the dna cylinder
 Returns      : none
 Args         : A filehandle

=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;

	my @fields = qw( chrom start end strand element symbol locus_tag );
	my @transcript_fields = qw( protein_id element accession );
	my @prbh_fields = qw( prbh );

	my @a_attr = qw( _rh_mapareas feature _o_species );
	my( $rh_map, $feature, $o_species ) = $self->_Get( @a_attr );
	foreach my $element ( keys %$rh_map )
	{
		my $rh_info = $$rh_map{ $element };
		if (defined $$rh_info{ boxes }{"name"})
		{
			my @coords=ElementsSet::_inflate(@{$rh_info->{boxes}{__primary}});
			my @coords_name=ElementsSet::_inflate(@{$rh_info->{boxes}{"name"}});
            my @values = ( 'GENE', $o_species->GetLongName() );
			my $ref_elt=$rh_info->{elt};
			push @values,  $self->GetFields($ref_elt,@fields);
			my $transcript_elt = $element.'t1';
			push @values, $self->{ o_transcript }->GetFields($transcript_elt, @transcript_fields);
			my @quoted_values = map ("\"$_\"",@values);
			#print $file_handle join("\t",@coords_name)."\t".join("\t",@values)."\t".join("\t",@coords)."\n";
			print $file_handle join("\t",@coords_name)."\t".join("\t",@quoted_values)."\t".join("\t",@coords)."\n";
		}
	}

	return;
}

=head2 __LoadExons

 Title        : __LoadExons
 Usage        : $o_gene -> __LoadExons();
 Prerequisite : none
 Procedure    : initialized Exon object if 'exon' is not initialized.
 Returns      : the blessed object or undef
 Args         : none

=cut
sub __LoadExons
{
	my $self = shift;

	my %h_param = ( $self->GetAttributes(),
					start			=> $self->GetNucleicMin(),
					end				=> $self->GetNucleicMax(),
					displayed_start => $self->_GetStart(),
					displayed_end	=> $self->_GetEnd()
# 					-o_logger		=> $self->GetLogger()
					);

	# protection rajoutee vite fait, a revoir
	# peut etre laisser tomber le AssertBetweenThreshold
	try
	{
		# You may call __LoadExons several times, it will be executed only once
		if (!defined $self->{exon})
		{
			$self->{ o_exon } = ElementsSet::Exon -> New (%h_param);
		}
	}
	catch LipmError with
	{
		my $E = shift;
		my $msg = $E->Explain(). " - element = exon";
		$self->_Trace( $msg, '.', 'IE' );
	};

	return $self;
}

=head2 __StoreFirstTranscriptData

 Title        : __StoreTranscriptData
 Usage        : $o_gene -> __StoreFirstTranscriptData();
 Prerequisite : rh_data, ra_sort_elements and transcripts must be initialized
 Procedure    : store some transcripts informations in genes data (protein_id ,prbh and accession )
 Returns      : none
 Args         : none
 Globals      : rh_data, ra_sort_elements and transcripts

=cut
sub __StoreTranscriptData
{
	my $self = shift;

	# Test if data is correctly stored.
	return if ( $self->IsEmpty());

	$self -> { o_transcript } = ElementsSet::Transcript -> New($self->GetAttributes());

}

=head2 __IsGeneSearched
 Title        : __IsGeneSearched
 Usage        : $self -> __IsGeneSearched(@a_labels)
 Prerequisite : object should be initialized
 Function     : Test if this label is searched
 Returns      : boolean, 0/1
 Args         : a_labels => list of labels of the gene to test
 Attributes   : gene_search

=cut
sub __IsGeneSearched
{
	my $self = shift;
	my @a_labels = @_;

	# Test parameters
	return 0 if (scalar(@a_labels) == 0 or !defined $self->{gene_search});

	# Test the label
	foreach my $label (@a_labels)
	{
		return 1 if ($label =~ /$self->{gene_search}/);
	}

	return 0;
}

=head2 AreExonsDisplayed
 Title        : AreExonsDisplayed
 Usage        : $self -> AreExonsDisplayed([factor])
 Prerequisite : none
 Function     : Check if exons data should be display or not.
 Returns      : a boolean
 Args         : factor (optional) -> (gene_range/image_height). It's the percentage of the image which is occupied by the gene.
 Globals      : exon

=cut
sub AreExonsDisplayed
{
	my $self = shift;

	my $o_species = $self->_Get( '_o_species' );
    my $o_io = $o_species->GetInstanceObserver();
	if ($o_io->GetKingdom() eq "bacteria")
    {
        return &FALSE;
    }
    else
    {
        return $self->{gene_structure};
    }
}

=head2 GetTranscripts
 Title        : GetTranscripts
 Usage        : $self -> GetTranscripts()
 Prerequisite : object should be initialized
 Function     : return the `Transcripts` object
 Returns      : $self -> {transcripts}
 Args         : none
 Attributes   : transcripts

=cut
sub GetTranscripts
{
	my $self = shift;

	return $self -> {transcripts};
}

=head2 GetExon
 Title        : GetExon
 Usage        : $self -> GetExon()
 Prerequisite : object should be initialized
 Function     : return the `Exons` object
 Returns      : $self -> {transcripts}
 Args         : none
 Globals      : exon

=cut
sub GetExon
{
	my $self = shift;

	return $self -> {exon};
}

=head2 GetExonsList
 Title        : GetExonsList
 Usage        : $self -> GetExonsList()
 Prerequisite : object should be initialized
 Procedure    : Call the Exons' method if the object is initialized
 Returns      : $self -> {transcripts}
 Args         : none
 Globals      : exon

=cut
sub GetExonsList
{
	my $self = shift;

	# Test if the object is initialized
	return () if (!defined $self->{ o_exon } or $self->{ o_exon }->IsEmpty() );

	# Else return Exons method :
	return $self->{ o_exon } -> GetExonsList(@_);
}


# =head2 GetFileName
#
#  Usage        : $filepath = $self -> GetFileName()
#  Prerequisite : object should be initialized
#  Function     : get the filepath of data.
#  Returns      : $filepath, string
#  Args         : none
#
# =cut
#
# sub GetFileName
# {
# 	my $self = shift;
#
# 	my @a_attributes = qw( o_io organism chromosome );
# 	my( $o_io, $ref, $chrom ) = $self->_Get( @a_attributes );
#
# 	my $filepath = $o_io->GetGeneFile( species => $ref, chrom => $chrom );
#
# 	return $filepath;
# }
#

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
