# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Narcisse class to display online compared data.

=head1 DESCRIPTION

This class display online glint data on the Narcisse respresentation.

Sequences regions are load in temporary files and then an online compare is launch using glint program.

=head1 SUBROUTINES

Display
LoadSequenceRegion
IsStrandOk
OnlineCompare
GetFileName

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Hsp;

use strict;
use warnings;

use Data::Dumper;
use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf
use File::Basename;
use IO::File;

use base qw( ElementsSet );
use General;
use Runner;
use Error qw( :try );
use LipmError::IOException;
use LipmError::ParameterException;
use NarcisseServiceFactory;
use Configuration::InstanceObserver;
use Sequence_utils;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#              Constructor                       #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : o_io     => Instance observer
 				organism => organism
 				chr      => chromosome
 				start    => start
 				end      => end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- organism
 					- chromosome
 					- start
 					- end

=cut
sub _Init
{
	my $self = shift;

	my %h_param = ( 
        o_synteny       => undef,
        level           => undef,
#         OnlineCompare	=> undef,
        @_
        );

    #TODO set some parameter verification here
	#======================================
	# 1) Set object attributes
	my @a_keys        = keys( %h_param );
	@$self{ @a_keys } = @h_param{ @a_keys };

	# je comprends pas la logique de cette initialisation
	#ici on a des attributs start et end qui devraient etre privee a la mere,
	# accessibles par les methodes protegees _GetStart et GetEnd
	# je ne peux pas les utiliser

	# Change defaut parameter : file extension, data headers, primary_key, start_label, end_label, ra_map_keys
	my @a_map_keys     = ('target_organism','target_chromosom','target_start','target_end','score','identity','strand','file');
	my @a_data_headers = ('q_start','q_end','s_start','s_end','s_chrom','score','identity','strand');
	my $o_io = $h_param{o_species}->GetInstanceObserver();
	my %h_new_param = ( 'root_data'		=> $o_io->GetRootSessDir(),
						'start_label'	=> "q_start",
						'end_label'		=> "q_end",
						'ra_data_headers' => \@a_data_headers,
						'ra_map_keys'     => \@a_map_keys,
						%h_param
						);
	
	
	$self->{ target } = $self->{ o_synteny }->{ target };
	
	# Changed back to root_sessdir (root_data is redefined by SUPER::_Init)
	# previous attempt was made to redefine root_data as ROOT_SESSDIR (seem dangerous to me)
	$self->{ root_sessdir } = $o_io->GetRootSessDir();

	$self->__GetTargetSlicesFromCompareParameters() if ( defined $h_param{ OnlineCompare } );

	# Initialization
	$h_new_param{ OnlineCompare } = 'bidouille';
	$self -> SUPER::_Init( %h_new_param );

	# bidouille
	$self->_Set( OnlineCompare => $h_param{ OnlineCompare } );
    return $self;
}

=head2 _GetTargetSlicesFromCompareParameters

 Title        : _GetTargetSlicesFromCompareParameters
 Usage        : $self -> _GetTargetSlicesFromCompareParameters
 Procedure    : Adapt start and end to OnlineCompare Parameters and identify synt elements corresponding to the region
 Returns      : none
 Globals      : none

=cut

sub  __GetTargetSlicesFromCompareParameters
{
	my $self = shift;
	my $o_synteny = $self->{ o_synteny };

	my ( $ref_start, $ref_end ) = $self->GetOnlineCompareParam();

	my @target_slices = ();		
	my @a_fields = qw( start end );
	my @a_slice_fields = qw( slice_id start end s_chrom s_start s_end sign );
	
	foreach my $sub_synt ( $o_synteny->GetSortElements() )
	{
		my ($slice_start, $slice_end) = $o_synteny->GetFields($sub_synt, @a_fields);
	
		last if ( $slice_start > $ref_end );
		next if ( $slice_end < $ref_start );
		
		my %h_slice;
		@h_slice{ @a_slice_fields } = $o_synteny->GetFields( $sub_synt, @a_slice_fields );

		push @target_slices, \%h_slice;
	}
		
	#Definition of seq_start and seq_end of target according to the s_start and s_end attributes
	foreach  my $target_slice ( @target_slices )
	{
		$target_slice->{ seq_start } = $target_slice->{ s_start };
		$target_slice->{ seq_end }   = $target_slice->{ s_end };
	}
		
	#Correction of the target slices according to $ref_start and $ref_end
	my $first_slice = 	$target_slices[0];
	my $last_slice = 	$target_slices[$#target_slices];

	&_SetTargetSliceStartEnd( $first_slice, "start" , $ref_start);
	&_SetTargetSliceStartEnd( $last_slice, "end" , $ref_end);
	
	$self->{ __selected_slices } = \@target_slices;

	return;	
	
=head2 _SetTargetSliceStartEnd

 Title        : _SetTargetSliceStartEnd
 Usage        : _SetTargetSliceStartEnd( $rh_slice, $label, $ref_start )
 Function     : Determine the target slice start (resp. end) position corresponding to ref_start (resp. ref_end)          
 Returns      : none
 Args         : $rh_slice : a reference on the slice object (hash in fact)
                $label    : label to be adapted (start or end)
                $position : the corresponding position on the reference genome

=cut	
	sub _SetTargetSliceStartEnd
	{
		my $slice = shift;
		my $label =  shift;
		my $position = shift;	

		#Modify them (if necessary) according to the label and to the sign
		if ( $slice->{sign} eq "+" )
		{
			$slice->{ "seq_$label" } = &ProjectedPosition( $slice, $position );			
		}
		else
		{
			my $reverse_label = ( $label eq "start" ) ? "end" : "start";
			$slice->{ "seq_$reverse_label" } = &ProjectedPosition( $slice, $position );			
		}

		return;
	}
}

=head2 ProjectedPosition

 Title        : ProjectedPosition
 Usage        : Synteny::ProjectedPosition( $rh_synt_elt, $pos, [reverse])
 Function     : From a position ($pos) on one genome, calculate the predicted position on the other:
                - default : from ref --> target
                - reverse : from target --> ref
 Returns      : position of undef if not successfull
 Args         : $rh_synt_elt, a refereence on a synt element, $pos : the position of interest

=cut
sub ProjectedPosition
{
	my $rh_elt = shift;
	my $position = shift;
	my $direction = ( scalar @_ ) ? shift : "default";
		
	(defined $rh_elt->{s_start} && defined $rh_elt->{s_end} && defined $rh_elt->{sign})
		#or throw LipmError::ParameterException('synt element','not in valid format');
        or throw NarcisseError( -bd => "Could not find anything in target <br />for a valid comparison" );
	
	my $projected_position = undef;
	my @fields = qw( start end s_start s_end sign);
	my ( $start, $end, $s_start, $s_end, $sign ) = @{$rh_elt}{ @fields };		
			
	if ( $direction eq "default" )
	{
		$projected_position = &Projection( $rh_elt->{sign}, $start, $end, $s_start, $s_end, $position );			
	}	
	else # reverse projection
	{
		$projected_position = &Projection( $rh_elt->{sign}, $s_start, $s_end, $start, $end, $position );					
	}
			 				 	
	return $projected_position;
	
}


=head2 Projection

 Title        : Projection
 Usage        : Synteny::Projection( $sign, $q_start, $q_end, $s_start, $s_end, $position)
 Function     : From a position ($pos) on one genome, calculate the predicted position on the other:

 Returns      : position always defined !!
 Args         : $sign : the sign of the synteny,
 				$q_start, $q_end : interval on the query genome
 				$s_start, $s_end : interval on the subject genome
 				$pos : the position of interest
=cut
sub Projection
{
	my ( $sign, $q_start, $q_end, $s_start, $s_end, $position ) =@_;
	
	my $projected_position;
	
	my $factor = ( $s_end -  $s_start ) / ( $q_end -  $q_start );
	
	if ( $position <  $q_start )
	{
		$position = $q_start;
	}
	if ( $position > $q_end)
	{
		$position = $q_end;
	}

	if ( $sign eq "+" )
	{
		$projected_position = $s_start + ( $position - $q_start ) * $factor;
	}
	else
	{
		$projected_position = $s_end   - ( $position - $q_start ) * $factor;	
	}
	
	return int( $projected_position );
}


#------------------------------------------------#
#                  Display methods               #
#------------------------------------------------#

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Display Prbh data in target's frame coordinates.
 Returns      : none
 Args         : %h_param {
 					image 		 => Image object
 					reference_box=> reference Box object
 					target_box   => target Box object
 						 }
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
		image 			=> undef,
		reference_box	=> undef,
		target_box		=> undef,
		target_slice_boxes => undef,
		@_
		);
	
	#======================================
	# Testing object data and break the method if their is no data to draw.
	return if $self->IsEmpty();
	
	#Some elements will be displayed, we register the current object in MapArea
	$self->NewMapElementSet();
		
	# Testing mandatory parameters
	throw LipmError::ParameterException('image_object', 'missing') if ( !defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );
	throw LipmError::ParameterException('reference box object', 'missing') if ( !defined $h_param{reference_box} || ref($h_param{reference_box}) !~ /Box/  );
	throw LipmError::ParameterException('target box object', 'missing') if ( !defined $h_param{target_box} || ref($h_param{target_box}) !~ /Box/  );

	# General parameters
	my $o_image   = $h_param{image};
	my $o_ref_box = $h_param{reference_box};
	my $o_tgt_box = $h_param{target_box};
	my $synt_factor = $self->{synt_factor};

	my $target_box_justification = $o_tgt_box->GetJustification();

	#======================================
	# Move the frame object
	
	my ($X_ref,$X_tgt,$o_border_box) = $self->SetBorderBox( reference_box => $o_ref_box, target_box => $o_tgt_box);
	$o_image -> SetFrameChange( $o_border_box->GetDisplayedArea(), y0 => $self->{start}, y1 => $self->{end});
	
	my $reference_frame = $o_image->{ frame };

	my $rh_slice_frames = $self->SetTargetFrames(target_slice_boxes => $h_param{ target_slice_boxes } );
	my %slice_frames=%$rh_slice_frames;
	#======================================
	# Graphic parameters
	
	# image layer
	my $layer = $self->GetElementSetLayer();

	#======================================
	# Colors :
	$self->__InitColors( $o_image, $layer );
	my $color = $self->GetGraphicParameter('color_hsp','grey_clear');	
	my $link_color	= $o_image->colorAllocate($color, $layer);
	
	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_border_box );
		
	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };
	
	# from now on objects will be drawn on a single layer : $layer	
	$o_image -> SetCurrentLayer( $layer );

	my $o_layer = $o_image -> GetCurrentLayer( );

	$o_image -> SetAbsoluteFrame();  #Changing to absolute coordinates
	
	#Temporary patch for displaying the alignment
	#TODO remove this awkward path
	my $filename = $self->GetFileName();
	
	#======================================
	# Foreach hsp data stored :
	my @a_fields = qw( q_start q_end s_start s_end s_chrom score identity strand slice_id);
	foreach my $element_id ( $self->GetSortElements())
	{
		my ($ref_start,$ref_end,$tgt_start,$tgt_end,$tgt_chrom,$score,$identity,$strand, $slice_id) = $self->GetFields( $element_id,@a_fields );		

		#-----------------
		# Conditions
		next if ($ref_start < $self->{start} || $ref_end > $self->{end});
		#next if ($tgt_chrom ne $self->{'tgt_chrom'});
					
		# Reference coordinate change
		my $Y_ref_start  = $reference_frame->YAxisCoordinateChange( $ref_start );
		my $Y_ref_end    = $reference_frame->YAxisCoordinateChange( $ref_end );		
		
		# Target coordinate change
		my $Y_tgt_start = $slice_frames{ $slice_id }->YAxisCoordinateChange( $tgt_start ); 
		my $Y_tgt_end   = $slice_frames{ $slice_id }->YAxisCoordinateChange( $tgt_end ); 
				
		#-----------------
		# Score color :
		my $score_color = $self->_GetHspColor( $identity, $strand );
		my @a_coord = $o_image -> filledRectangle($x0,$Y_ref_start,$x1,$Y_ref_end,$score_color);
				
		#Tracing back the coordinates of the box
		my ( $Y_start_box, $Y_end_box ) = @a_coord[1,3];		
		my ( $x_tgt_link, $x_ref_link ) = ( $x0, $x1 );
				
		if ( $target_box_justification eq "right" )
		{
			$x_ref_link = $x0;
			$x_tgt_link = $x1;			
		}
		
		# The correspondance with the reference					
		$o_image -> line( $x_ref_link, $Y_start_box, $X_ref, $Y_ref_start, $link_color );
		$o_image -> line( $x_ref_link, $Y_end_box,   $X_ref, $Y_ref_end,   $link_color );
						
		# The correspondance with the target		
		$o_image -> line( $X_tgt, $Y_tgt_start, $x_tgt_link, $Y_start_box, $link_color );
		$o_image -> line( $X_tgt, $Y_tgt_end,   $x_tgt_link, $Y_end_box,   $link_color );						
				
		$self->AddFields( $element_id, '__rfilename' => \$filename);
		 		
		#-----------------
		# Store map areas
		#$self->NewMapElement($self->GetElement($element_id),$element_id, \@a_coord  );
		$self->NewMapElement( $element_id, \@a_coord  );
				
	}
	
	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : Defines the track on which the ElemenSet exons will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1
 Args         : o_box   -> the box on which the hsp will be displayed
 				
=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = ( 
					o_box => undef,
					@_
					);
		
	( defined $h_param{ o_box } )	
		or throw LipmError::ParameterException('o_box','missing');
		
	my $o_box = $h_param{ o_box };
		
	my %h_track_param;
	
	my $width = 6;
	
	my $x_center = ( $o_box->GetX0() + $o_box->GetX1() )/2;
	my $X0 = $x_center - int( $width / 2);
	my $X1 = $x_center + int( $width / 2);
	
	%h_track_param = ( X0 => $X0, X1 => $X1,  desc => "hsp elements");
		
	return \%h_track_param;
	
}

=head2 SetBorderBox
 Title        : 
 Usage        : $o_xxx -> SetBorderBox( )
 Function     : Defines the track on which the ElemenSet exons will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1
 Args         : o_box   -> the box on which the hsp will be displayed
 				
=cut
sub SetBorderBox
{
	my $self = shift;
	my %h_param = ( reference_box => undef,
					target_box => undef,
					@_
					);
					
	my $o_ref_box = $h_param{reference_box};
	my $o_tgt_box = $h_param{target_box};
	
	my ( $X0_border, $X1_border, $X_ref, $X_tgt);
	my $width_hsp_box	= $self->GetGraphicParameter('x_genes',20);
	my $cyl_pos 		= $self->GetGraphicParameter('x_genes',20);
	
	my $target_box_justification = $o_tgt_box->GetJustification();
	
	if ( $target_box_justification eq 'left' )
	{
		$X0_border = $o_ref_box->GetX0() -  int( $width_hsp_box/2 );
		$X1_border = $o_ref_box->GetX0() +  int( $width_hsp_box/2 );
		$X_ref = $o_ref_box->GetX0() + $cyl_pos - 3 ;
		$X_tgt = $o_tgt_box->GetX1() - $cyl_pos + 3;
	}
	else
	{
		$X0_border = $o_ref_box->GetX1() -  int( $width_hsp_box/2 );
		$X1_border = $o_ref_box->GetX1() +  int( $width_hsp_box/2 );
		$X_ref = $o_ref_box->GetX1() - $cyl_pos + 3 ;
		$X_tgt = $o_tgt_box->GetX0() + $cyl_pos - 3 ;
		
	}
	
	my $o_border_box = Box->New( X0 => $X0_border, Y0 => $o_ref_box->GetY0(), X1 => $X1_border, Y1 => $o_ref_box->GetY1() );
						
	return 	($X_ref, $X_tgt, $o_border_box);			
}

=head2 SetBorderBox
 Title        : 
 Usage        : $o_xxx -> SetBorderBox( )
 Function     : Defines the track on which the ElemenSet exons will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1
 Args         : o_box   -> the box on which the hsp will be displayed
 				
=cut
sub SetTargetFrames
{
	my $self = shift;
	my %h_param = ( 
					target_slice_boxes => undef,
					@_
					);
	
	my $rh_target_slice_boxes=$h_param{ target_slice_boxes };
	
	#Constructing the frame objects for each target slice
	my $o_synteny = $self->{ o_synteny };
	my %slice_frames = ();
	my @a_fields = qw( sign slice_id );
	foreach my $slice ( $self->GetTargetSlices() )
	{		
		my %h_slice_param = $o_synteny->GetTargetSliceParameters( $slice );
		my ( $sign, $slice_id ) = $o_synteny->GetFields( $slice, @a_fields );		
		my $target_slice_box = $rh_target_slice_boxes->{ $slice_id };
        
        # Avoiding an error for very high zoom levels !
        throw LipmError  (-text => "Hsp cannot be drawn (no target here)\n", -value => 2) if ( !defined ($target_slice_box) or $target_slice_box eq '' );
        
		my $target_slice_frame = FrameTransformer::1D->New( $target_slice_box->GetDisplayedArea( $sign ), y0 => $h_slice_param{start}, y1 => $h_slice_param{end});
		$slice_frames{ $slice_id } = $target_slice_frame;		
	}
			
	return \%slice_frames;
}

=head2 __InitColors

 Title        : InitColors
 Usage        : $self -> InitColors()
 Function     : Init display colors for hsps
 Returns      : none
 
=cut
sub __InitColors
{
	my $self = shift;
	my $o_image = shift;
	my $layer = shift;
	
	my %h_colors;
	
	# Strand plus/plus :
	my $plus_color_low	= $o_image->colorAllocate('green_bright', $layer);
	my $plus_color_75	= $o_image->colorAllocate('green_yellow', $layer);
	my $plus_color_85	= $o_image->colorAllocate('green_light', $layer);
	my $plus_color_90	= $o_image->colorAllocate('green', $layer);
	my $plus_color_95	= $o_image->colorAllocate('green_olive', $layer);
	my $plus_color_100	= $o_image->colorAllocate('green_dark', $layer);
	
	@{ $h_colors{ 'plus/plus' } {level} } = ( 100, 95, 90, 85, 80, 75, 0);
	@{ $h_colors{ 'plus/plus' } {color} } = ( $plus_color_100, $plus_color_95, $plus_color_90, $plus_color_85, $plus_color_75, $plus_color_low);
			
	# Strand plus/minus :
	my $minus_color_low	= $o_image->colorAllocate('yellow', $layer);
	my $minus_color_75	= $o_image->colorAllocate('golden_yellow', $layer);
	my $minus_color_85	= $o_image->colorAllocate('orange', $layer);
	my $minus_color_90	= $o_image->colorAllocate('red_light', $layer);
	my $minus_color_95	= $o_image->colorAllocate('red', $layer);
	my $minus_color_100	= $o_image->colorAllocate('red_dark', $layer);
	
	@{ $h_colors{ 'plus/minus' } {level} } = ( 100, 95, 90, 85, 80, 75, 0);
	@{ $h_colors{ 'plus/minus' } {color}} = ( $minus_color_100, $minus_color_95, $minus_color_90, $minus_color_85, $minus_color_75, $minus_color_low);
	
	$self->{ __hsp_colors} = \%h_colors;

}

sub _GetHspColor
{
	my $self = shift;
	my ( $p_identity, $strand ) = ( shift, shift );

	my $rh_colors = $self->{ __hsp_colors}{ $strand };
	my @levels = @{ $rh_colors->{ level }};
	
	my $i=0;
	while ( $levels[$i] > $p_identity)	 
	{
		$i++;
	} 
	
	return 	$rh_colors->{ color }[$i];
}

=head2 IsStrandOk
 Title        : IsStrandOk
 Usage        : $self -> IsStrandOk($strand)
 Prerequisite : graph_orientation should be initialized.
 Function     : This function test if $strand parameter is 'in agreement' with graph_orientation attribute.
 Returns      : A boolean (0 or 1)
 Args         : $strand = the Hsp value of 'strand' column. Set as 'PLus/Plus' or 'PLus/Minus'
 Attributes   : graph_orientation

=cut
sub IsStrandOk
{
	my $self = shift;
	my $strand = shift;

	# Test parameter
	return 0 if (!defined $strand);
	return 0 if ($strand !~ /^(plus|minus)\/(plus|minus)|$/);

	# We test the second plus/minus value and compare with the orientation of the synteny fragment.
	# Return 1 if there are equals, else return 0
	if ($self->{graph_orientation} eq '+')
	{
		return 0 if ($strand eq 'plus/minus');
	}
	elsif ($self->{graph_orientation} eq '-')
	{
		return 0 if ($strand eq 'plus/plus');
	}

	return 1;
}

#------------------------------------------------#
#            Managing the map                    #
#------------------------------------------------#

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the hsp elements
 	X0 Y0 X1 Y1 HSP query q_chrom q_start q_end subject s_chrom s_start s_end score identity strand
 Returns      : none
 Args         : A filehandle on the map file (.coord)
 
=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;	
    
    my $o_io = &InstanceObserver::Singleton();

	#my @fields = qw( query q_chrom q_start q_end subject s_chrom s_start s_end score identity strand );
	
	my @a_attr = qw( _rh_mapareas feature );
	my( $rh_map, $feature ) = $self->_Get( @a_attr );
	foreach my $element ( keys %$rh_map )
	{
		my $rh_info = $$rh_map{ $element };
		my @coords=ElementsSet::_inflate(@{$rh_info->{boxes}{__primary}});	

		my @values = ( 'HSP' );
        my ($query, $subject) = $self->GetFields( $element, 'query', 'subject' );
        ($query,$subject) = $o_io->GetSpeciesLongName($query,$subject);
		push @values,  $query;
        push @values, $self->GetFields($element,qw( q_chrom q_start q_end ));
        push @values, $subject;
        push @values, $self->GetFields($element, qw( s_chrom s_start s_end score identity strand )); 
		#my ($r_filename) = $self->GetFields($element,( '__rfilename' ) );
		#push @values,  $$r_filename;
		my @quoted_values = map ("\"$_\"",@values);
		#print $file_handle join("\t",@coords)."\t",join("\t",@values)."\n";		
		print $file_handle join("\t",@coords)."\t",join("\t",@quoted_values)."\n";		
	}
}

#------------------------------------------------#
#            Comparing the sequences           #
#------------------------------------------------#

=head2 OnlineCompare

 Title        : OnlineCompare
 Usage        : $self -> OnlineCompare()
 Prerequisite : none
 Function     : call glint and transform results to a narcisse readable format.
 Returns      : a string
 Args         : none
 Attributes   : organism, chromosome, target, tgt_chrom
 Globals      : none

=cut
sub OnlineCompare
{
	my $self = shift;

	my $o_species  = $self->_Get( '_o_species' );
	my $o_inst_obs = $o_species->GetInstanceObserver();
    my $img_refresh_button = '<img src="' . $o_inst_obs->GetCfg('HTTP_LOGO').'/move/view-refresh.png' .'" alt="Refresh icon" />';
    my $error_footer       = "click the <a href=\"javascript:O_CONTROL_BLOCK.Reload()\">$img_refresh_button</a> icon to continue";
	# Get start/end of the comparison for each organism
	my ( $ref_start, $ref_end ) = $self->GetOnlineCompareParam();

	my $msg = "OnlineCompare : range=$ref_start:$ref_end\n";
	$self->_Trace( $msg, '.', 'IEV', __PACKAGE__ );

	# temporary files:
	my $tmpl = 'OnLineXXXX';
	my $o_ref_file = new File::Temp( TEMPLATE => $tmpl, CLEANUP => 0);
	my $o_tgt_file = new File::Temp( TEMPLATE => $tmpl, CLEANUP => 0);
	my $o_comp_file = new File::Temp( TEMPLATE => $tmpl, CLEANUP => 0);


	#------------------------------
	# Get sequences :
	#The reference sequence
	my %seq_slices;
	my @a_attr = qw( _chromosome target );
	my( $chrom, $target ) = $self->_Get( @a_attr );
	my $ref    = $o_species->GetShortName();
	my( $start, $end )    = ( $ref_start, $ref_end );

    my $ref_seq = $self->ExtractSequenceRegion( $ref, $chrom, $start, $end );
    print $o_ref_file ">reference\n";
    print $o_ref_file $ref_seq;
	$seq_slices{ "reference" } = {	species => $ref, chrom => $chrom,
									start => $ref_start, end => $ref_end};

    #The target sequences sequence

    #my @a_toto = $self->GetSelectedSlices();
    #warn Dumper \@a_toto;
	foreach my $target_slice ( $self->GetSelectedSlices() )
	{
		my @a_keys = qw( s_chrom seq_start seq_end slice_id );
		my( $chrom, $start, $end, $slice_id ) = @{$target_slice}{ @a_keys };
        
        my $slice_seq = $self->ExtractSequenceRegion( $target, $chrom, $start, $end );
		print $o_tgt_file ">$slice_id\n";
		print $o_tgt_file $slice_seq;
		$seq_slices{ $slice_id } = { species => $target, chrom => $chrom, start => $start, end => $end };
	}

	#------------------------------
	# Glint command :
	my $glint     = $o_inst_obs->GetCfg('ROOT_GLINT');
	my $ref_file  = $o_ref_file->filename();
	my $tgt_file  = $o_tgt_file->filename();
	my $comp_file = $o_comp_file->filename();
	my $cmd       = "$glint -o $comp_file  $tgt_file $ref_file";
    my $timeout   = $o_inst_obs->GetCfg('GLINT_TIMEOUT');
    
    $self->_Trace( "NOW RUNNING $cmd\n", '.', 'IEV' );
    my ($status, $killed) = RunExt(
        -cmd     => $cmd,
        -timeout => $timeout);    
    
    if ($status != 0)
    {
        warn "ERROR: glint returned with the following status: $status";
        throw NarcisseError( -bd => "ERROR: glint returned with the following status: $status" );
    };
	my $out_nb_lines = `wc -l $comp_file`;
	$out_nb_lines =~ s/(\d+) .+$/$1/;
	$out_nb_lines += 0;
	if ( $out_nb_lines == 1 )
	{
		warn "WARNING: glint did not find any match";
		throw NarcisseError( -bd => "WARNING: glint did not find any match", -ft => $error_footer );
	}
    if ($killed == -1)
    {
        warn "ERROR: Timeout exceeded";
        throw NarcisseError( -bd => "ERROR: Time exceeded for the alignment <br />Please try again with a narrower range",
         -ft => $error_footer );   
    }
    
    if ( ! -e $comp_file or -z $comp_file)
    {
        warn "ERROR: No output from glint";
        throw NarcisseError( -bd => "ERROR: No output from glint", -ft => $error_footer );
    }
			
	# Format glint results
	my $result_file = $self->GetFileName( &FALSE );
	$self->_ReformatGlint( $comp_file, $result_file, \%seq_slices);
		
	#-----------------------------
	# Delete temporary files
# 	unlink($ref_file,$tgt_file,$comp_file);
		
	return;

}

=head2 _ReformatGlint

 Usage        : _ReformatGlint(  $file)
 Prerequisite : none
 Function     : reformats the glint file
 Returns      : none
 Args         : none

=cut

sub _ReformatGlint
{
	my $self = shift;
	my ( $file_in, $file_out, $rh_info) = ( shift, shift, shift );

	my $fh_in = new IO::File( $file_in );
	if( !defined $fh_in )
	{
		throw LipmError::IOException( $file_in, 'f', -depth => 1 );
	}

	my $header = <$fh_in>;
	my @fields =qw(element query q_chrom q_start q_end subject s_chrom s_start s_end score identity strand slice_id details );

	my @hsps;
	while ( my $line = <$fh_in> )
	{
		chomp($line);

		my ( $query, $subject, $script ) = split(/\t+/,$line);
		my  ( $reference, $ref_chrom) =  @{$rh_info->{ $query }}{ qw( species chrom )};
		my  ( $target, $tgt_chrom) =  @{$rh_info->{ $subject }}{ qw( species chrom )};

		if ($script =~ /^\[(\d+)-(\d+)\]->\[(\d+)-(\d+)\]\(\d*,(.*),(\d*),\w*,(\d+)%.*\)/)
		{
			my $q_start	= $rh_info->{ $query }{ start } + $1-1;
			my $q_end	= $rh_info->{ $query }{ start } + $2-1;

			my $s_start	= $rh_info->{ $subject }{ start } + $3-1;
			my $s_end	= $rh_info->{ $subject }{ start } + $4-1;

			my $details = $script;
			my $strand	= lc("$5");
			my $score	= $6;
			my $identity= $7;

			my @desc = ($reference,$ref_chrom,$q_start,$q_end,$target,$tgt_chrom,$s_start,$s_end);
			my $elt_id = join("_",@desc);

			my %h_hsp;
			@h_hsp{ @fields }=
				( $elt_id, $reference, $ref_chrom, $q_start, $q_end, $target, $tgt_chrom, $s_start, $s_end, $score, $identity, $strand, $subject, $details);

			push @hsps, \%h_hsp;
		}
		else
		{
			die "Invalid file format ";
		}
	}

	$fh_in->close();

	my( $basename, $dirpath ) = fileparse( $file_out );
	my @a_new_dirs = mkpath( $dirpath );
	if( 0 != @a_new_dirs )
	{
		my $msg = "@a_new_dirs have been created\n";
		$self->_Trace( $msg , '.', 'IE' );
	}

	my $fh_out;
	if( -e $file_out )
	{
		$fh_out = new IO::File( $file_out, 'a' )
					or throw LipmError::IOException( $file_out, 'f' );
	}
	else
	{
		$fh_out = new IO::File( $file_out, 'w' )
						or throw LipmError::IOException( $file_out, 'f' );
		print $fh_out "#".join("\t",@fields)."\n";
	}

	foreach my $rh_hsp (sort { $a->{q_start} <=> $b->{q_start} } @hsps)
	{
		my $string = join("\t",@{$rh_hsp}{@fields});
		print $fh_out $string."\n";
	}
	$fh_out->close();
}


=head2 __MakeComparison
 Title        : __MakeComparison
 Usage        : $self -> __MakeComparison()
 Prerequisite : OnlineCompare must be initialized <-- meme pas vrai ! (seb)
 Function     : Test OnlineCompare attribute to know if the object should load data, or just try to open pre-calculated data
 Returns      : Boolean
 Args         : none
 Attributes   : OnlineCompare

=cut
sub __MakeComparison
{
	my $self = shift;
	my $compare = $self->{'OnlineCompare'};
		
	# If OnlineCompare is undefined, return 0
	return 0 if (!defined $compare);

	# Else if the format is not allowed, return 0
	return 0 if ($compare !~ /^\d+:\d+$/);
	#return 0 if ($compare !~ /^\d+(:\d+){1,3}$/);
		
	# Else test values
	my ($start,$end) = ($1,$2) if ($compare =~ /^(\d+):(\d+)/);

	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

		# Start must be inferior to representation end
	return 0 if ( $elt_end < $start );

		# End must be superior to representation start
	return 0 if ( $end < $elt_start );

		# Start must be inferior to end
	return 0 if ( $end < $start );

	return 1;
}

=head2 GetOnlineCompareParam
 Title        : GetOnlineCompareParam
 Usage        : ($ref_start,$ref_end,$tgt_start,$tgt_end) = $self -> GetOnlineCompareParam()
 Prerequisite : object should be initialized
 Function     : Get the start/end of the online comparison we need to do for each organism.
 				If 'OnlineCompare' is not valid, return (0,0).
 Returns      : ref_start, ref_end
 Args         : none
 Attributes   : OnlineCompare

=cut
sub GetOnlineCompareParam
{
	my $self = shift;

	my ( $ref_start, $ref_end );
	my $onlinecompare = $self->{'OnlineCompare'};
	
	# 1) Get reference start/end for the comparison
	($ref_start,$ref_end) = ($1,$2) if ($self->{'OnlineCompare'} =~ /^(\d+):(\d+)/);

	# start/end should be in the representation
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	# cf note dans la methode _init
	$elt_start = $self->{ start }	if( !defined $elt_start );
	$elt_end   = $self->{ end }		if( !defined $elt_end );

	$ref_start = $elt_start if( $ref_start < $elt_start );
	$ref_end   = $elt_end   if( $ref_end   > $elt_end );
	
	return ( $ref_start, $ref_end );
}

=head2 ExtractSequenceRegion

 Title		  : ExtractSequenceRegion
 Usage        : ExtractSequenceRegion( organism,chrom, start, end )
 Prerequisite : object attributes should be initialized
 Procedure	  : Get the sequence corresponding to the region
 Args		  : organism, chromosome, start end
 Returns   	  : the sequence as a string

=cut
sub ExtractSequenceRegion
{
	my $self = shift;
	my ( $organism, $chrom, $start, $end ) = @_;

	# Test parameters
	(defined $organism && defined $chrom && defined $start && defined $end )
		or throw LipmError::ParameterException('org, chrom, start or end','missing');

	my %h_params = ( species => $organism,
					chrom   => $chrom,
					ra_regions => [ { start => $start, end => $end} ]
					);
	my $o_ns = NarcisseServiceFactory::GetSlicesFromRegions( %h_params );

	my $rh_results = $o_ns->FetchResults();

	#It's a single seq
	my $header = (keys %$rh_results)[0];

	my $fasta_seq = Sequence_utils::FormatSequenceLine( seq => $$rh_results{ $header } );

	return $fasta_seq ;
}

=head2 GetSelectedSlices

 Title		  : GetSelectedSlices
 Usage        : $self->GetSelectedSlices()
 Procedure	  : accessor
 Args		  : none
 Returns      : an array of the selected slices

=cut
sub GetSelectedSlices
{
	my $self = shift;
	
	return @{$self->{ __selected_slices }};
}

=head2 GetTargetSlices

 Title		  : GetTargetSlices
 Usage        : $self->GetTargetSlices()
 Procedure	  : accessor
 Args		  : none
 Returns      : an array of the selceted slices

=cut
sub GetTargetSlices
{
	my $self = shift;
	my $o_synteny = $self->{ o_synteny };
			
	return $o_synteny->GetSortElements() ;
}

#------------------------------------------------#
#       Overloading some ElementSet methods     #
#------------------------------------------------#

=head2 _ReadNarcisseFormatedFile

 Title        : _ReadNarcisseFormatedFile
 Usage        : $self -> _ReadNarcisseFormatedFile($file,$start,$end);
 Prerequisite : none
 Function     : Overwrite default method calling OnlineCompare to generate comparaison
 Returns      : none
 Args         : $file 	-> the data file
 				$start 	-> start of analyze
 				$end 	-> end of analyze
 Attributes   : rh_data, ra_sort_elements
 Globals      : none

=cut
sub _ReadNarcisseFormatedFile
{
	my ($self,$file, $start, $end)= (shift, shift, shift, shift);

	# Load reference and target sequences and make the comparaison
	$self -> OnlineCompare() if ($self->__MakeComparison() == 1);

	# Apply default method and return results
	# we cancel optimisation to force re-read file
	return $self->SUPER::_ReadNarcisseFormatedFile($file, $start, $end,
	                                                -cancel_optimisation => 1);
}

=head2 GetFileName
 Title        : GetFileName
 Usage        : $self -> GetFileName()
 Prerequisite : object should be initialized
 Function     : get the url of the file which contain data for this ElementsSet object
 Returns      : a string
 Args         : none
 Attributes   : organism, chromosome

=cut

sub GetFileName
{
	my $self = shift;
# 	my( $check ) = @_;

	my $check = &FALSE; #&TRUE unless defined( $check );

	my @a_attr = qw( _o_species _chromosome target level );
	my( $o_species, $chrom, $target, $level ) = $self->_Get( @a_attr );

	my $filepath = $o_species->GetHspFiles(
        prog   => 'glint',
        level  => $level,
        target => $target,
        -chrom => $chrom,
        -check => $check );

	return $filepath;
}

=head2 IsDataAvailable

 Title        : IsDataAvailable
 Usage        : $self -> IsDataAvailable()
 Prerequisite : none
 Function     : Check if the data exists
 Returns      : 1 if it foes exist
 				or
 				0 if not
 
=cut
sub IsDataAvailable
{
	my $self = shift;
	my $file = shift;
	
	return (-e $file || defined $self->{ OnlineCompare });			
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
