=head1 NAME

	MultiGraph - an ElementsSet that can draw many graph.

=head1 DESCRIPTION

	This class is a "trompe l'oeil" for ChromosomeSlice.
	It can draw many ElementsSet::SuppChart::X objects like it is one.

	Colors can be defined into graphics.cfg $feature is the feature name
	=> $feature_color_axis  [black]        # axis !
	=> $feature_color_graph [grey_light]   # a dashed-line in the middle
	=> $feature_width       [30]           # the width of the representation

=head1 SUBROUTINES

	__DrawAxis
	__DrawInnerElements

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
Letort Sebastien : sebastien.letort@toulouse.inra.fr

=cut

package ElementsSet::MultiGraph;

use strict;
use warnings;

use Data::Dumper;

use ElementsSet::SuppChart::Histograms;
use ElementsSet::SuppChart::LinePlots;
use Legend;
use LipmError::ParameterException;

use base qw( ElementsSet );
my $VERSION = do {my @r = (q$Rev: 3457 $ =~ /\d+/g); $r[0]};

# those params are used if no value are set in graphics.cfg
# in graphics.cfg, key is the wanted key followed by '_feature'
my %H_GRAPHICS_DEFAULT = ( width => 30, pos_x => 200, side => 'left' );

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : New ElementSet::MultiGraph( %h_param )
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param a hash of parameters
 					mandatory
  						o_species       => SpeciesObserver object
  						chromosome		=> chromosome name or number

  					optional
  						feature			 	=> simple name of this feature of object (default 'class_name')
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 0)
  						file_directory		=> directory of data (default 'feature')
  						file_extension		=> extension of the data files (default 'feature')
						graph_orientation 	=> the orientation (default '+')
						primary_key 		=> in data files, this is the reference column (default 'element')
						start_label 		=> in data files, this is the start column  (default 'start')
						end_label	 		=> in data files, this is the end column  (default 'end')
						ra_data_headers		=> array ref., this is the array of sorted column which have to be store when data are read (default ('element','start','end','strand'))
						graphic_param_file	=> file for the graphic parameters initialization

					specific to this class
						y_axis_side => string, 'left' _default, 'right', 'center'
							where to put the y axis
						graph_class  => string [LinePlots], SuppChart class to draw

 Attributes	  : objects attributes :
 					Element attributes :
 						feature, organism, chromosome, start, end
 					File and directory attributes :
 						file_directory, file_extension, ra_data_headers, primary_key, start_label, end_label
 					Representation attributes :
 						ra_map, graph_orientation
 					Specific to this class :
 						_y_axis_side, string, 'left' _default, 'right', 'center'
							where to put the y axis
						__graph_class, string SuppChart class to draw
						_rao_inner_element_set, $ref on list of
						                          ElementsSet::SuppChart object
						_width, width of the representation
						_pos_x, x position of the feature

 Globals	  :	%CFG
 Errors		  : LipmError::ObjectException

=cut

sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
				y_axis_side    => 'left',
				graph_class    => 'LinePlots',	# class of the ElementSet,
				file_directory => undef,
				with_legend    => 1,
				#o_param        => New ParamParser(), #nawak
				@_ );

	# where the ordinate axis will be drawn
	$self->{ __graph_class } = $h_param{ graph_class };
	$self->{ __has_legend }  = $h_param{ with_legend };
	$self->_Set( _file_dir => $h_param{ file_directory } );

	# ref list of inner element set
	$self->{ __rao_inner_element_set } = [];

	# bidouille pour repasser %h_param aux constructeurs des Features.
	$self->SetParams( \%h_param );

	$self->{ _width } = undef;
	# sa valeur sera donnee dans _ReadNarcisseFormatedFile parce que j'ai
	# besoin de connaitre $self->{ feature }.
	# et que je ne me sers pas de la largeur avant.

	# Initialization
	$self -> SUPER::_Init( %h_param );

	return $self;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

=head2 Display

 Title        : Display
 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 				$self->IsDataLoad() must return 1.
 Procedure    : Display data of the Element inherited object on the Image object.
 Returns      : none
 Args         : o_image	-> Image object
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image => undef,
			side  => 'right',
			o_box => undef,
			@_ );

	# Don't draw data if they weren't loaded
	return if $self->IsEmpty();

	# Testing mandatory parameters
	if( !defined( $h_param{ image } )
		or ( ref( $h_param{ image } ) !~ /(Image)|(GD::Image)/  ) )
	{
		throw LipmError::ParameterException('Image object', 'missing');
	}

	# General parameters
	my( $o_image, $o_box, $side ) = @h_param{ qw( image o_box side ) };

	# If side = center :
	# 	1) This Display is set to a 'right' side mode
	#	2) Call another Display with side equal to 'left'.
	# This is use for example when you want to draw the reference at center and 2 targets to each side of it.
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left');
	}

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );
	my $X0  = $rh_track->{X0};

	# si on pouvait avoir la position de la base 0
	# seulement quand on a des coordonnees negative pour le slice (zoom au debut du chrom.)
	# ça eviterait d'avoir le test dans la boucle de LinePlots.
	my $y0      = $self->_GetStart();

	my $layer = $self->__DrawInnerElements( %h_param, x => $X0, y => $y0 );

	$self->__DrawAxis( $o_image, $layer, $X0, $y0 );

	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : 
 Returns      : 
 Args         : 

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = ( 
					o_box => undef,
					side => "right",
					@_ );
					
	( defined $h_param{ o_box } )	
		or throw LipmError::ParameterException('o_box','missing');
		
	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };
		
	my $width   = $self->{ _width };
	my $pos_x	= $self->{ _pos_x };

	my ($X0,$X1,$X);
	
	if ( $side eq "right" )
	{
		$X0 = $o_box->GetX0() + $pos_x;
		$X1 =  $X0 + $width;
		$X = $X0;
	}
	else
	{
		$X1 = $o_box->GetX1() - $pos_x;
		$X0 =  $X1 - $width;
		$X = $X1;		
	}
	
	$self->{ _pos_x  } = $X0;
		
	# 	left 	center	right
	# 	+----	--+-- 	----+
	# 	|    	  |   	    |
	# 	|    	  |   	    |
	# 	|    	  |   	    |	
		
	my $x0;
	if( $self->{ _y_axis_side } eq 'left' )
	{	$x0 = $X;	}
	elsif( $self->{ _y_axis_side } eq 'center' )
	{	$x0 = $X + $width/2;	}
	elsif( $self->{ _y_axis_side } eq 'right' )
	{	$x0 = $X + $width;	}

	my %h_track_param = ( X0 => $X0, X1 => $X1, X => $X, desc => "supp_data elements" );

	return \%h_track_param;
}


=head2 Procedure __DrawAxis

 Title        : __DrawAxis
 Usage        : $o_xx->__DrawAxis(  $o_image, $layer, $x0, $y0 );
 Prerequisite : none
 Procedure    : Draws axis of the graph.
 Args         : $o_image : Image object
                $layer : string, layer name, where we draw.
                $x0,$y0, int,int : it is the position of the top left corner

=cut
sub __DrawAxis
{
	my $self = shift;
	my ( $o_image, $layer, $x0, $y0 ) = @_;

	my( $feat, $graph_ori ) = $self->_Get( qw( _feature _graph_orientation ) );
	my $width = $self->{ _width  };
	my $pos_x = $self->{ _pos_x  };

	# get colors indexes
	my %h_colors = ( 'color_axis_' . $feat  => 'black',
	             	 'color_graph_' . $feat  => 'grey_light' );
	my $rh_colors = $self->AllocateColors( $o_image, \%h_colors, $layer );
	my $axis_color = $$rh_colors{ 'color_axis_' . $feat };

	# from now on objects will be drawn on a single layer : $layer	
	$o_image -> SetCurrentLayer( $layer );

	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	my $y_axis_pos = ( $graph_ori eq '+' ) ? $elt_start : $elt_end;
	
	#abcisses
	$o_image->line( $x0, $y0, $x0, $elt_end, $axis_color, $layer );

	#ordonnées
	$o_image->line( $pos_x,$y_axis_pos, $pos_x+$width,$y_axis_pos, $axis_color, $layer );

	# one more line is drawn in the middle width.
	if( $self->{ _y_axis_side } ne 'center' )
	{
		# half graph
		$o_image->dashedLine( $pos_x+$width/2,$y0, $pos_x+$width/2,$elt_end,
		                      $$rh_colors{ 'color_graph_' . $feat }, $layer );
	}

	return;
} #__DrawAxis


=head2 Function __DrawInnerElements

 Title        : __DrawInnerElements
 Usage        : $layer = $o_xx->__DrawInnerElements( %h_params );
 Prerequisite : none
 Function     : For each inner ElementsSet object, use its Display method.
                	to create a gd ressource.
                All gd ressource will then be merged into an unique gd ressource.
                This ressource will be the layer for the ElementSet::Multigraph
                object.
 Args         : $rh_params, ref onto the hash needed by Display method
                $x0,$y0, int,int, position of the top left corner of the image.
 Returns      : $layer, string, the new layout name

=cut
sub __DrawInnerElements
{
	my $self     = shift;
	my %h_params = @_;

	my @a_tmp_layers  = ();
	my $o_image       = $h_params{ image };
	my $o_legend      = undef;

	# Draw each elements.
	my $rao_inner = $self->{ __rao_inner_element_set };
	foreach my $o_inner_element_set ( @$rao_inner )
	{
		# MaJ de _pos_x pas propre, mais pas le temps de faire mieux
		$o_inner_element_set->{ _pos_x } = $self->{ _pos_x };

		# will build a ressource (layer) with the extension as layer's name
		# this layer needs to be destroyed as soon a Multi layer is set.
		$o_inner_element_set->Display( %h_params );
		push( @a_tmp_layers, $o_inner_element_set->_Get( '_file_ext' ) );

		if( 1 == $self->{ __has_legend } )
		{
			$o_legend = $self->__BuildLegend( $o_inner_element_set,
			                                  $o_image, $o_legend );
		}
	}

	# merge tmp_layers
	my $layer = $self->_Get( '_feature' );
	$o_image->MergeLayers( $layer, @a_tmp_layers );

	# copy legend into layer
	if( defined( $o_legend ) )
	{
		my $o_gd   = $o_legend->copyRotate90();
		#my ($x,$y) = $o_image->ToAbsoluteCoordinates( @h_params{ qw( x y ) } );
		my ($x,$y) = $o_image->PointCoordinateChange( @h_params{ qw( x y ) } );
		$x += $self->{ _width };

		my $o_dest = $o_image->GetLayer( $layer );
		$o_dest->copy( $o_gd, $x,$y, 0,0, $o_gd->getBounds() );
	}

	return $layer;

=head2 function

 Title        : __BuildLegend
 Usage        : $o_lgd = $o_xx->__BuildLegend($o_element, $o_image, $o_legend );
 Prerequisite : ***Valeurs en dur***
 Function     : build a new Legend object, which is the concatenation
                of $o_legend, and a new one created with $o_element informations
 Returns      : $o_lgd, a new Legend object.
 Args         : $o_element, a ElementsSet object
                $o_image, an Image object
                $o_legend, a Legend object or undef

=cut

	sub __BuildLegend
	{
		my $self = shift;
		my ( $o_element, $o_image, $o_legend ) = @_;

		# build legend
		my $rh_margins = { top_bottom => 2, left_right => 4,
		                   inter_elt_width => 5,
		                   inter_elt_height => 5 };

		my $ra_data = $o_element->GetLegendInfos( $o_image );
		my $o_lgd   = New Legend( $ra_data, $rh_margins, 10, 6 );
		my $o_return= Legend::ConcatenateLegend( $o_legend, 'width', $o_lgd );

		return $o_return;
	}
} # __DrawInnerElements


=head2 Function _ReadNarcisseFormatedFile

 Title        : _ReadNarcisseFormatedFile
 Usage        : $self -> _ReadNarcisseFormatedFile($foo,$start,$end);
 Prerequisite : none
 Function     : this overload intends only not to read data.
                It creates one inner ElementsSet for each extension
                of the current file directory.
                Data will be read by those object.
 Returns      : undef, undef
 Args         : $foo 	-> no use, it's just to be compliant
                		with mother method call.
 				$start 	-> start of analyze.
 				$end 	-> end of analyze.
 Attributes   : rh_data, hash ref, data read.
                ra_sort_elements, list ref, column's names.
 Globals      : %CFG

=cut

sub _ReadNarcisseFormatedFile
{
	my $self = shift;
	my ( $foo, $start, $end ) = @_;
	
	# if __graph_class is undef, it is a call from this class-mother constructor.
	return	unless defined( $self->{ __graph_class } );

	my $feature_name = $self->_Get( '_feature' );
	my $ra_objets    = $self->{ __rao_inner_element_set } = [];

	$self->{ _width } = $self->GetGraphicParameter(
							'width_' . $feature_name, $H_GRAPHICS_DEFAULT{ width } );
	$self->{ _pos_x } = $self->GetGraphicParameter(
							'x_' . $feature_name, $H_GRAPHICS_DEFAULT{ pos_x }  );
	$self->{ _y_axis_side } = $self->GetGraphicParameter(
							'side_' . $feature_name, $H_GRAPHICS_DEFAULT{ side }  );

	# besoin du nom du fichier de donnees
	my $filepath_no_ext = $self->__BuildFilePathNoExtension();

	# pour chacun des sous-dossier, je cherche les extensions
	my $feat_path    = $self->_Get( '_file_dir' );
	my $ra_extension = &NARCISSE_utils::FindAllExtensions( $feat_path );
	foreach my $ext ( @$ra_extension )
	{
		next	if( $ext eq 'cfg' );
		my $class = "ElementsSet::SuppChart::$self->{ __graph_class }";
		my $rh_params = $self->GetParams();
		my $o_element_set = $class->New( %$rh_params,
		                        feature        => $feature_name . '_' . $ext,
# 		                        file_extension => $ext,
		                        filepath    => $filepath_no_ext . $ext,
		                        width       => $self->{ _width },
		                        pos_x       => $self->{ _pos_x },
		                        file_ext    => $ext,
		                        return_type => 'ARRAY'
		                    );
		push( @$ra_objets, $o_element_set )	if( defined $o_element_set );
	}
		
	# to show that I respect signature
	return undef;
}


=head2 Function __BuildFilePathNoExtension

 Usage        : my $path = $o_x->__BuildFilePathNoExtension( none )
 Function     : Build the filepath, without the extension.
 Returns      : $path
 Args         : none
 Error        : none

=cut

sub __BuildFilePathNoExtension
{
	my $self = shift;

	my @a_attr = qw( _file_dir _chromosome _o_species );
	my( $feat_path, $chrom, $o_species ) = $self->_Get( @a_attr );

	# besoin de recuperer le masque de fichier
	my $o_io = $o_species->GetInstanceObserver();
	my $mask = $o_io->GetCfg( 'MASK_FILE_CHROM' );

	my $filename_no_ext = $mask . $chrom . '.';

	my $filepath = File::Spec->catfile( $feat_path, $filename_no_ext );

	return $filepath;
} #__BuildFilePathNoExtension


=head2 function IsDataLoad

	Title        : IsDataLoad
	Usage        : 1 = $o_xx->IsDataLoad();
	Prerequisite : unlike ElementSet, Multigraph does not have rh_data attr.
	Function     : return the number of inner ElementSet.
	               The name is for compatibility reason.
	Returns      : $n, integer, the number of inner ElementSets object.
	Args         : none

=cut
sub IsDataLoad
{
	my $self = shift;

	my $rao_inner = $self->{ __rao_inner_element_set };

	my $n = scalar( @$rao_inner );

	return $n;
}

=head2 function IsEmpty

	Title        : IsEmpty
	Usage        : if ( $o_xx->IsEmpty() )
	Prerequisite : unlike ElementSet, Multigraph does not have rh_data attr.
	Function     : return 1 is the number of inner ElementSet is 0
	Returns      : a boolean

=cut
sub IsEmpty
{
	my $self = shift;
	
	my $rao_inner = $self->{ __rao_inner_element_set };

	return ( 0 == scalar( @$rao_inner ) );

}

=head2 IsDataAvailable

 Title        : IsDataAvailable
 Usage        : $self -> IsDataAvailable()
 Prerequisite : unlike ElementSet, Multigraph does not have a single specific file
 Function     : Only checks __graph_class is defined
 Returns      : 1 if it does exist
 				or
 				0 if not
 
=cut
sub IsDataAvailable
{
	my $self = shift;
				
	return ( defined( $self->{ __graph_class } ) );			
}

=head2 Procedure SetParams

 Title        : SetParams
 Usage        : &SetParams( $rh_params );
 Prerequisite : it a kind of 'bidouille',
               	if someone know how to do it properly, go for it.
 Procedure    : just store $rh_params into the object.
 Args         : $rh_params

=cut
sub SetParams
{
	my $self = shift;
	my ( $rh ) = @_;

	$self->{ bidouille }{ param } = $rh;

	return @_;
}

=head2 function GetParams

	Title        : GetParams
	Usage        : $rh_param = $o_xx->GetParams();
	Prerequisite : it a kind of 'bidouille',
	               	if someone know how to do it properly, go for it.
	Function     : Get what was stored by SetParams.
	Returns      : $rh_param, what is inside $self->{ bidouille }{ param }
	Args         : none

=cut
sub GetParams
{
	my $self = shift;

	return $self->{ bidouille }{ param };
}

=head2 GetFileName

 Title        : GetFileName
 Usage        : $self -> GetFileName()
 Prerequisite : none
 Function     : Do nothing, just present to disabled the mother method.
 Returns      : nothing
 Args         : none

=cut

sub GetFileName
{
	my $self = shift;
	return;
}


1;

__END__

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
