# Copyright INRA/CNRS

# The Cecill license

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

 ElementsSet::Ocean - Class to represent sequence of N composed regions.

=head1 DESCRIPTION

 This class read files that contains position of regions composed uniquely
 with N.

=head1 SUBROUTINES

Add new subroutines here.


=head1 AUTHOR

Sebastien Letort: sebastien.letort@toulouse.inra.fr

=cut
package ElementsSet::Ocean;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );

my $VERSION = do {my @r = (q$Rev: 3189 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : $self -> _Init(%h_param)
 Prerequisite : none
 Procedure    : Everything is done by the mother class

=cut

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 				$self->IsEmpty() must return 0.
 Procedure    : Display data of the Element inherited object on the Image object.
 				see the display method of the Template object for details
 Returns      : none
 Args         : o_image	-> Image object
 				o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box 	=> undef,
			side 	=> undef,
			@_
			);

	#======================================
	# Don't draw data if their weren't loaded
	return if $self->IsEmpty();

	# Testing mandatory parameters
	throw LipmError::ParameterException('Image object', 'missing') if (!defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box   = $h_param{o_box};
	my $side 	= $h_param{side};

	# Test if we have to draw elements twice (for the reference representation)
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left', b_label => 0);
	}

	# Get the box height - If > some threshold, use drawing mode FilledRect
	#                      Else, use drawing mode arc

	use constant FILLED_RECT_MODE => 0;
	use constant ARC_DRAWING_MODE => 1;
	use constant HEIGHT_THRESHOLD => 100000;

	$self->_Trace("Now drawing ocean boxes ",'','IE');
	if ( $o_image->GetMax_y()-$o_image->GetMin_y() > &HEIGHT_THRESHOLD )
	{
		$self->_Set('__drawing_mode',&FILLED_RECT_MODE);
		$self->_Trace("Switched to FILLED_RECT_MODE\n",'','IE');
	}
	else
	{
		$self->_Set('__drawing_mode',&ARC_DRAWING_MODE);
		$self->_Trace("Switched to ARC_DRAWING_MODE\n",'','IE');
	}



	#======================================
	# Graphic parameters :

	# image layer
	my $layer = $self->GetElementSetLayer();

	# graphic constants
	my $color_name = $self->GetGraphicParameter('color_ocean','red');
	my $angle      = $self->GetGraphicParameter('angle_ocean',10);

	# colors
	my $color	= $o_image->colorAllocate( $color_name, $layer );

	# x positions
	my $rh_track = $self->GetTrack( name => $layer, o_box => $o_box, side => $side );

	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	my @a_fields = qw( start end );
	foreach my $element_id ( $self->GetSortElements() )
	{
		my ( $start, $end ) = $self->GetFields($element_id, @a_fields );

		my $elt_start = $self->_GetStart();
		my $elt_end   = $self->_GetEnd();

		next if ( $start > $elt_end || $end <  $elt_start );

		$start = $elt_start if ( $start <  $elt_start );
		$end   = $elt_end   if ( $end   >  $elt_end );

		# We have TWO methods for drawing the oceans:
		#	1/ filledRectangle, very fast but awful at high zoom levels
		#	2/ drawing an arc, rather slow
		# The method is selected above

		if ( $self->_Get('__drawing_mode') == &FILLED_RECT_MODE )
		{
			$o_image->filledRectangle( $x0, $start, $x1, $end, $color );
		}
		else
		{
			my $w  = $x1-$x0;
			my $cx = int( 0.5 + $x0 + $w/2 );
			my $h  = $self->GetDNACylinderHeight();

			#	warn "$end-$start"  if( $end-$start > 200 );
			for( my $i=$start; $i<$end; ++$i )
			{
				$o_image->arc( $cx,$i, $w,$h, 90-$angle,90+$angle, $color );
			}
		}

# 		$self->NewMapElement( $element_id, \@a_coord  );

	}

	return;
}


=head2 GetTrack
 Title        : GetTrack
 Usage        : $self -> GetTrack( )
 Function     : Using the appropriate graphic parameters, defines the track in which the elements will be displayed
 Returns      : a reference on a hash with the position of the track between X0 <--> X1
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_
					);

	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };

	my %h_track_param;

	my $ocean_width= $self->GetGraphicParameter('width_ocean',4);

	my $rh_dna_cyl = $self->GetDNACylinder( $side ) ;

	my $x_center = 	( $rh_dna_cyl->{X0} + $rh_dna_cyl->{X1} )/2;

	my ( $X0,$X1 );
	if ( $self->_Get('__drawing_mode') == &ARC_DRAWING_MODE )
	{
		( $X0,$X1 ) = @$rh_dna_cyl{ qw( X0 X1 ) };
	}
	else
	{
		$X0 = $x_center - $ocean_width/2;
		$X1 = $x_center + $ocean_width/2;
	}

	%h_track_param = ( X0 => $X0, X1 => $X1,  desc => "ocean elements");

	return \%h_track_param;

}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
