# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Narcisse class to display personal data.

=head1 DESCRIPTION

This class display personal data on the Narcisse respresentation.

User must load his data in a tabbed file with this format :
 #chr  element  start  end  [color  label]
 chromosome  element_name  start  end  color  label


=head1 SUBROUTINES

Display

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Personal;

use strict;
use warnings;

use base qw( ElementsSet );
my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

use constant X0 => 0;
use constant Y0 => 1;
use constant X1 => 2;
use constant Y1 => 3;

=head2 _Init

 Title        : _Init
 Usage        : _Init($o_species, $chromosome, $start, $end [, $rh_param])
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 				o_species	=> SpeciesObserver object
 				chr			=> chromosome
 				start		=> start
 				end			=> end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- organism
 					- chromosome
 					- start
 					- end

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
        show_label	=> 0,
        @_
        );

	# New attributes :
	$self -> {show_label} = $h_param{show_label};

	# Change defaut parameter : file extension
	my @a_map_keys	   = ('text');
	my @a_data_headers = ('element','start','end','chr','color','label');
	my %h_new_param = ( ra_data_headers => \@a_data_headers,
						ra_map_keys		=> \@a_map_keys,

						%h_param
						);

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

    return $self;
}

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Procedure    : Draw name of each element and their rectangle representation.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					o_box   => Box object
 					side  => optional, 'right' or 'left'
 						 }
 Globals      : none

=cut

sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			@_ );

# warn "ICI PERSONAL";
	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box 	= $h_param{o_box};
	my $side 	= $h_param{side};

	#======================================
	# Graphic parameters :

	# image layer
	my $layer = $self->GetElementSetLayer();

	# graphic constants
	my $x_perso		= $self->GetGraphicParameter('x_perso',210);
	my $width_perso	= $self->GetGraphicParameter('width_perso',3);

	# fonts
	my $font 		= $o_image -> gdTinyFont;

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );

	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };

	my $x_label = $rh_track->{ X_label };
	my $width_genes_label 	= $rh_track->{ width_label };
	my $max_char 			= $width_genes_label/$font->width();

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	my $strand = 1;

	#======================================
	# Draw and store each personal data.
	my @a_fields  = qw(element start end chrom color label);
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	foreach my $element_id ($self->GetSortElements())
	{
		my ($element,$start,$end,$chrom,$color,$label) = $self->GetFields( $element_id, @a_fields );
# warn "start     -> end       $start $end  \n";
# warn "elt_start -> elt_end   $elt_start $elt_end \n";
		next if ($end < $elt_start || $start > $elt_end );

		# Allocate the color index
		$color = "black" if (!defined $color);
		my $element_color = $o_image->colorAllocate($color,$layer);

		# 1) Line representation
		my @a_box_coord = $o_image->filledRectangle($x0,$start,$x1,$end,$element_color,$layer);

		# Calculate the label vertical position
		my ( $Y0, $Y1 )= @a_box_coord[Y0,Y1];
		my $y_label = ( $Y0 + $Y1 )/2;

        # Test name width and cut it if it's too long.
	    my $text = (defined $label) ? $label : $element;
		substr($text,int($max_char/2)-1,length($text)-($max_char-2)) = '..' if (length($label) > $max_char);

		# 2) Draw the text if $side does not equal 'center'
		$o_image -> stringVerticalAligned($font, $x_label, $y_label, $text, $element_color, side => $side, absolute_frame => 1);
	}

	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        :
 Function     : Defines the track on which the ElemenSet will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1 as usual (X coordinate of the track)
                 - X_label : absolute X position for the display of the label
                 - width_label : size of space allocated to the label names
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_ );

	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };

	# graphic constants
	my $x_perso		= $self->GetGraphicParameter('x_perso',210);
	my $width_perso	= $self->GetGraphicParameter('width_perso',3);
	my $x_label		= $self->GetGraphicParameter('x_perso_label',40);

	my $width_genes_label 	= $self->GetGraphicParameter('width_genes_label',50);	# max width for gene labels

	my ($X0,$X1,$X);
	my $X_label;

	if ( $side eq "right" )
	{
		$X0 = $o_box->GetX0() + $x_perso;
		$X1 =  $X0 + $width_perso;
		$X = $X0;
		$X_label = $X1 + $x_label;
	}
	else
	{
		$X1 = $o_box->GetX1() - $x_perso;
		$X0 =  $X1 - $width_perso;
		$X = $X1;
		$X_label = $X0 - $x_label - $width_genes_label;
	}

	my %h_track_param = ( X0 => $X0, X1 => $X1, X => $X, width_label => $width_genes_label, X_label => $X_label, desc => "personal data");

	return \%h_track_param;

}


=head2 GetFileName

 Title        : GetFileName
 Usage        : $self -> GetFileName()
 Prerequisite : object should be initialized
 Function     : get the url of the file which contain data for this ElementsSet object
 Returns      : a string
 Args         : none
 Attributes   : feature, organism, chromosome

=cut
sub GetFileName
{
	my $self = shift;

	my @a_attr = qw( _o_species _chromosome _feature );
	my( $o_species, $chrom, $feature ) = $self->_Get( @a_attr );

	my $filepath = $o_species->GetPersoFiles( -chrom => $chrom );

	return $filepath;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
