# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet inherited object to display Narcisse PRBH data.

=head1 DESCRIPTION

See L<ElementsSet>

=head1 SUBROUTINES

GetFileName

=cut
package ElementsSet::Prbh;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );
use Box::ChromosomeSliceBox;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : rh_cfg		=> ref of the cfg hash map
 				organism	=> organism
 				chr			=> chromosome
 				start		=> start
 				end			=> end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- organism
 					- chromosome
 					- start
 					- end

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
				target			=> undef,
				@_
				);
								
	# New attributes :
	my @attributes=qw( target );	
	@{$self}{ @attributes } = @h_param{ @attributes };			

	# Change defaut parameter : file extension, data headers, primary_key, start_label, end_label
	my @a_data_headers = ('q_start','q_end','s_start','s_end','s_chr');
	my %h_new_param = ( primary_key => "q_prot",
						start_label => "q_start",
						end_label => "q_end",
						ra_data_headers => \@a_data_headers,

						%h_param
						);	
	# Initialization
	$self -> SUPER::_Init( %h_new_param );

    return $self;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Display Prbh data in target's frame coordinates.
 Returns      : none
 Args         : %h_param {
 					image 		 => Image object
 					reference_box=> reference Box object
 					target_box   => target Box object
 						 }
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
		o_image 		=> undef,
		-o_textual      => undef,
		reference_box	=> undef,
		target_box		=> undef,
		@_
		);

	return if ( defined $h_param{ dont_display_anything } && 1 == $h_param{ dont_display_anything });

	#======================================
	# Testing object data and break the method if their is no data to draw.
	return if $self->IsEmpty();

	# General parameters
	my $o_image   = $h_param{ o_image };
	my $o_ref_box = $h_param{ reference_box };
	my $o_tgt_box = $h_param{ target_box };
	my $o_textual = $h_param{ -o_textual };
	my $layer = 'prbh';
	my $o_file;
	if ( defined $o_textual )
	{
		$o_file= $o_textual -> GetLayer( $layer );
	}

	#======================================
	# Graphics parameters
	my $x 	  = $self->GetGraphicParameter('x_genes',40);
	my $space = $self->GetGraphicParameter('space_default',5);
	my $color = $self->GetGraphicParameter('color_rbh','green');

	my $link_color = $o_image->colorAllocate($color, $layer);

	$o_image -> SetCurrentLayer( $layer );

	#TODO this will have to be changed and managed by the box object !!
	# x position of the box for a negative strand gene
	my $x_genes = $self->_Get( '_o_param_graph' )->Get( 'x_genes' );
			
	#The reference box cylinder
	my %h_reference_area=$o_ref_box->GetBoxArea();
	my %h_target_area=$o_tgt_box->GetBoxArea();
	
	my $reference_justification = $o_ref_box->GetJustification();
	my $target_justification = $o_tgt_box->GetJustification();
		
	my $ref_cyl_pos;	
	my $tgt_cyl_pos;
	
	if ($target_justification eq "right" )  # We are on the right side
	{
		$ref_cyl_pos = $h_reference_area{X1} - $x_genes ;
		$tgt_cyl_pos = $h_target_area{X0} + $x_genes ;
	
	} elsif ( $target_justification eq "left" ) # We are on the left side
	{
		$ref_cyl_pos = $h_reference_area{X0} + $x_genes ;
		$tgt_cyl_pos = $h_target_area{X1} - $x_genes ;		
	}	
			
	$o_image -> SetAbsoluteFrame();  #Changing to absolute coordinates
				
	#======================================
	# Foreach prbh data stored :
	if ( defined $o_textual )
	{
		my $o_species = $self -> _Get ( '_o_species' );
		my $reference = $o_species -> GetLongName();
		my $target    = $self->{target};
		# TODO il doit y avoir des moyens plus malins
		$target    =~ s/_.+//; 
		$target    = $o_species -> GetInstanceObserver() -> GetSpeciesLongName($target);

		$o_textual -> PrintHeader ( $layer, 'reference', 'chromosome', 'protein', 'start','end','target','target chromosome','target protein','target start','target end' );
		my @a_fields = qw( q_chr q_prot q_start q_end s_chr s_prot s_start s_end );
		foreach my $element_id ( $self->GetSortElements() )
		{
			my ($q_chr, $q_prot, $q_start, $q_end, $s_chr, $s_prot, $s_start, $s_end) = $self->GetFields($element_id, @a_fields );
			print $o_file join("\t",$reference,$q_chr, $q_prot, $q_start, $q_end,$target,$s_chr, $s_prot, $s_start, $s_end) . "\n";
		}
	}
	else
	{
		my $o_map = $self->_Get( '_o_map_area' );
		my @a_fields = qw( q_gene q_start q_end s_chr s_gene s_start s_end );
		foreach my $element_id ( $self->GetSortElements() )
		{
			my ($q_transcript,$q_start,$q_end,$s_chr,$s_transcript,$s_start,$s_end) = $self->GetFields($element_id, @a_fields );
			my $q_gene = _ParentGene( $q_transcript );
			my $s_gene = _ParentGene( $s_transcript );
	
			if ( $self->IsMapDefined( $q_gene ) &&  $self->IsMapDefined( $s_gene ) )
			{
				# Draw link
				my ( $x_ref, $x_tgt ) = ( $ref_cyl_pos, $tgt_cyl_pos );
				my $y_ref = $o_map->GetCenterY( $q_gene );
				my $y_tgt = $o_map->GetCenterY( $s_gene );
	
				my ($x0,$y0,$x1,$y1) = &Image::ReorderCoordinates($x_ref, $y_ref, $x_tgt, $y_tgt);
				$o_image -> brokenLine( $x0, $y0, $x1, $y1, $space, $link_color, $layer);
			}
		}
	}
	
	$o_image -> SetRelativeFrame();  #Changing back to relative coordinates	

	return;

	sub _ParentGene
	{
		my $elt = shift;
		$elt=~s/t\d+//;
		return $elt;
	}

}# Display

=head2 GetFileName
 Title        : GetFileName
 Usage        : $self -> GetFileName([$level])
 Prerequisite : object should be initialized
 Function     : overload method
 Returns      : a string
 Args         : level -> optional, the level of synteny file
 Globals      : type, organism, chromosome, reference

=cut
sub GetFileName
{
	my $self = shift;

	my @a_attributes = qw( _o_species _chromosome target );
	my( $o_species, $chrom, $target ) = $self->_Get( @a_attributes );

	my $filepath = $o_species->GetOrthologyFiles(
										-chrom    => $chrom,
										target   => &_TargetOrganism( $target )
									 );

	return $filepath;
}


=head2 _TargetOrganism

 Title        : _TargetOrganism
 Usage        : _TargetOrganism(target_label)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get the right organism of target_label (target_list store target data as : 'name__program')
 Returns      : string = the target short name
 Args         : none
 Globals	  : target_list

=cut
sub _TargetOrganism
{
	my $target_label = shift;

	# Test parameter
	return '' if (!defined $target_label);

	if ($target_label =~ /__/)
	{
		# We take the expression before '__'.
		# Example : "mouse__prbh" => "mouse"
		$target_label = $`;
	}

	return $target_label;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
