# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet inherited object to display Narcisse Rna data.

=head1 DESCRIPTION

See L<ElementsSet>

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Rna;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end [, $rh_param])
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : rh_cfg		=> ref of the cfg hash map
 				organism	=> organism
 				chr			=> chromosome
 				start		=> start
 				end			=> end

 Globals	  : objects attributes
					- type
					- file_extension
 					- rh_cfg
 					- organism
 					- chromosome
 					- start
 					- end

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = ( @_ );

	# New attributes :

	# Change defaut parameter : ra_map_keys
	my @a_map_keys  = ('strand','element_id','name');
	my %h_new_param = ( ra_map_keys		=> \@a_map_keys,

						%h_param
						);

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

    return $self;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Title        : Display
 Usage        : $o_element -> Display(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Draw data and store positions.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					side  => optional, 'right' or 'left'
 						 }
 Globals      : none

=cut
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			@_
			);

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

    #Some elements will be displayed, we register the current object in MapArea
    $self->NewMapElementSet();

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box	= $h_param{o_box};
	my $side 	= $h_param{side};

	# Test if we have to draw exons twice (for the reference representation)
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left');
	}

	#======================================
	# Graphic parameters :

	# image layer
	my $layer = $self->GetElementSetLayer();

	# colors
	my $color_rna = $self->GetGraphicParameter('color_rna','red_bright');	# color of the representation.
	my $color = $o_image->colorAllocate($color_rna, $layer);

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );

	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	#======================================
	# Draw and store each rna.
	my @a_fields= qw( symbol start end strand);
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	foreach my $element_id ($self->GetSortElements())
	{
		my ($name,$start,$end,$strand) = $self->GetFields($element_id, @a_fields);
		next if ( $end < $elt_start || $start > $elt_end );

		# 1) Rna representation
		my @a_coord = $self->DrawElement( $o_image, $x0, $start, $x1, $end, $color, $strand);

       	# 2) Coordinate storage
		$self->NewMapElement($element_id, \@a_coord  );

	}


	return;
}

=head2 DisplayText

 Title        : DisplayText
 Usage        : $o_element -> DisplayText(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : Print data.
 Returns      : none
 Args         : %h_param {
 					image => Image object
 					side  => optional, 'right' or 'left'
 						 }
 Globals      : none

=cut
sub DisplayText
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> 'right',
			@_
			);

 	# image layer
	my $layer = $self->GetElementSetLayer();
	my $o_textual = $h_param{o_textual};
	my $o_file= $o_textual -> GetLayer( $layer );

	my @a_attr = qw( _o_species _chromosome );
	my ( $o_species, $chr ) = $self->_Get( @a_attr );
	my $species_name = $o_species -> GetLongName();

	#======================================
	# Print each repeat.
	my @fields= qw( symbol start end strand );
	$o_textual -> PrintHeader ( $layer, 'species', 'chromosome', @fields );
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();
	foreach my $element_id ($self->GetSortElements())
	{
		my ($symbol,$start,$end,$strand) = $self->GetFields($element_id, @fields);
		next if ( $end < $elt_start || $start > $elt_end );
		print $o_file join ( "\t", $species_name, $chr, $symbol, $start, $end, $strand )."\n";
	}

	return;
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     :
 Returns      :
 Args         :

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_
					);

	#We need the specification of the box to create it
	( defined $h_param{ o_box } )
		or throw LipmError::ParameterException('o_box','missing');

	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };

	my %h_track_param;

	my $rh_dna_cyl = $self->GetDNACylinder( $side ) ;

	my $X0  = $rh_dna_cyl->{X0};
	my $X1  = $rh_dna_cyl->{X1};

	%h_track_param = ( X0 => $X0, X1 => $X1,  desc => "rna elements");

	return \%h_track_param;

}

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the rna elements
                The file format is the following
                X0 Y0 X1 Y1 RNA speciesLongName chrom start end strand symbol locus_tag
                X0 Y0 X1 Y1 are the coordinates of the gene name box
 Returns      : none
 Args         : A filehandle

=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;

	my @fields = qw( chrom start end strand symbol);

	my @a_attr = qw( _rh_mapareas feature _o_species );
	my( $rh_map, $feature, $o_species ) = $self->_Get( @a_attr );
	foreach my $element ( keys %$rh_map )
	{
		my $rh_info = $$rh_map{ $element };
		if (defined $$rh_info{ boxes }{ __primary })
		{
			my @coords = ElementsSet::_inflate(@{$rh_info->{ boxes }{ __primary }});
            my @values = ( 'RNA', $o_species->GetLongName() );
			my $ref_elt=$rh_info->{elt};
			push @values,  $self->GetFields($ref_elt,@fields);
			my @quoted_values = map ("\"$_\"",@values);
			#print $file_handle join("\t",@coords)."\t".join("\t",@values)."\n";
			print $file_handle join("\t",@coords)."\t".join("\t",@quoted_values)."\n";
		}
	}

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
