# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

 ElementsSet::Snp, a class to draw SNPs on dna sticks.

=head1 DESCRIPTION

###
This class is just a snp to construct a new ElementsSet inherited object.

B<Steps to construct a new ElementsSet object :>

2- Changes on _Init method :
	* add new attributes with their default values.
	* change defaults parameters like : file extension, start_label, data headers ...

3- Overload Display method if you want to draw data

4- Overload GetFileName method if data are not store in the same path (/organism/file_directory/chromosome.file_extension )


B<Add a new check box to manage your image :>
See ChromosomeSlice and javascript documentation.

=head1 SUBROUTINES

Add new subroutines here.

=cut


package ElementsSet::Snp;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );
my $VERSION = do {my @r = (q$Rev: 3835 $ =~ /\d+/g); $r[0]};


=head2 _Init

 Title        : _Init
 Usage        : $self -> _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param a hash of parameters
 					mandatory
  						organism  		=> short name of the organism
  						chromosome		=> chromosome name or number

  					optional
  						feature			 	=> simple name of this feature of object (default 'class_name')
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 0)
  						file_directory		=> directory of data (default 'feature')
  						file_extension		=> extension of the data files (default 'feature')
						graph_orientation 	=> the orientation (default '+')
						primary_key 		=> in data files, this is the reference column (default 'element')
						start_label 		=> in data files, this is the start column  (default 'start')
						end_label	 		=> in data files, this is the end column  (default 'end')
						ra_data_headers		=> array ref., this is the array of sorted column which have to be store when data are read (default ('element','start','end','strand'))
						graphic_param_file	=> file for the graphic parameters initialization

 Attributes	  : objects attributes :
 					Element attributes :
 						feature, organism, chromosome, start, end
 					File and directory attributes :
 						file_directory, file_extension, ra_data_headers, primary_key, start_label, end_label
 					Representation attributes :
 						ra_map, graph_orientation

 Globals	  :	none
 Errors		  : LipmError::ObjectException

=cut

sub _Init
{
	my $self = shift;

	my %h_param = ( end_label => 'position', start_label => 'position',
					@_ );

	# If you have to change a default parameter (file extension, start_label, data headers ...) :
	my @a_data_headers = qw();# element position values src_id );
	$self -> SUPER::_Init(	ra_data_headers => \@a_data_headers ,
							%h_param );

	# your new attributes
	$self->_Set( __nb_snps => 0 );

    return $self;
}

# _AssertBetweenThresholds is empty because this makes no sens for this type of object
sub _AssertBetweenThresholds{};

=head2 Display

 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 				$self->IsEmpty() must return 0
 Procedure    : Display data of the Element inherited object on the Image object.
 				Moreover, you can store drawing elements in 'areas' by using the NewMapElement method.
 Returns      : none
 Args         : o_image	-> Image object
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
				b_label -> optional, boolean to draw SNPs names
 Globals      : none

=cut

sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			o_box   => undef,
			side 	=> undef,
			b_label	=> 0,
			@_ );

	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );
	
	# If there are too many snp to display, we give up
	#warn $self->GetSortElements() . " " . $self->GetGraphicParameter( 'ceil_threshold_snp' );
	return if ( $self->GetSortElements() >= $self->GetGraphicParameter( 'ceil_threshold_snp' ) );
	
	# General parameters
	my $o_image	= $h_param{image};
	my $o_box 	= $h_param{o_box};
	my $side 	= $h_param{side};
	my $b_label = $h_param{b_label};

	my @a_attr = qw( _graph_orientation );
	my ( $graph_orientation ) = $self->_Get( @a_attr );

	# !! Optional !!
	# If side = center :
	# 	1) This Display is set to a 'right' side mode
	#	2) Call another Display with side equal to 'left'.
	# This is use for example when you want to draw the reference at center and 2 targets to each side of it.
	# Then data are draw at each side => repeat some informations.
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left');
	}

	#--------------------------
	# 3-1 Initialize parameters
	#--------------------------

	# Name of the Image layer you want to draw in. By default this is "snp" (class name).
	my $layer = $self->GetElementSetLayer();

	# graphic constants
	my $x 		   = $self->GetGraphicParameter('x_snp',40);
	my $color_name = $self->GetGraphicParameter('color_snp','red');
	my $snp_width  = $self->GetGraphicParameter('width_snp',4);
	my $height     = $self->GetGraphicParameter( 'width_genes' )/ 3;
	my $label_color= $self->GetGraphicParameter( 'color_snp_label', 'red_dark');

	# colors
	my $color	= $o_image->colorAllocate( $color_name, $layer );
	my $black	= $o_image->colorAllocate( 'black', $layer );
	my $res_lbl_color = $o_image->colorAllocate( $label_color, $layer );
	# Moreover, you could add an alpha to use transparency.
# 	my $color_alpha = $o_image->colorAllocateAlpha($color_name,$alpha,$layer);

	# x positions
	my $rh_track = $self->GetTrack( name => $layer, o_box => $o_box, side => $side );
	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };

	#&General::ShowDebug( "SNP " . Dumper [ $layer, $x, $color_name, $snp_width, $x0, $x1, $side ] );

# obsolete ?
# 	#--------------------------
# 	# 3-2 Calculate x positions, according to size parameter
# 	#--------------------------
# 	# This allow user to reverse the representation (left<=>right)
# 	my $x0 = $x;
# 	my $x1 = $x0 + $width;
# 	if ($side eq 'left')
# 	{
# 		$x0 = $o_image->ReversePositionInFrame('x',$x0);
# 		$x1 = $o_image->ReversePositionInFrame('x',$x1);
#
# 		($x0,$x1) = &Image::SwapCoordinates($x0,$x1);	# Test both number and return them setting smallest at first.
# 	}
#

	# fonts	 * Use fonts :
	my $font = $o_image -> gdTinyFont;	# See GD::Image doc to know the list of fonts.

	my $x_label         = $rh_track->{ X_label };
	my $width_snp_label = $rh_track->{ width_label };
	my $max_char        = $width_snp_label / $font->width();

	# from now on objects will be drawn on a single layer : $layer
	$o_image -> SetCurrentLayer( $layer );

	# to have information on the element with the mouse
	$self->NewMapElementSet();
	my $nb = $self->_Get( '__nb_snps' );

	my $old_label_position = ($graph_orientation eq '-') ? $rh_track->{Y1} : $rh_track->{Y0};

	#--------------------------
	# 3-3 Draw elements that are between start-end
	# Specify here the fields that are specific to the elements you want to draw
	my @fields= qw( position mutation extern_id);
	my $bool_name_displayed;
	foreach my $element_id ( $self->GetSortElements() )
	{
		my( $pos, $mutation, $snp_id ) = $self->GetFields( $element_id, @fields );
# &General::ShowDebug( $pos, $mutation );
		my @a_label_coord = ();

		my @a_coord = $o_image->FilledCurvedRectangle( $x0,$pos-0.5, $x1,$pos+0.5, $height,$color );

		if( &General::IsNotNull( $mutation ) )
		{
			# est-ce que ca marchera si c'est un cible ?
			# la position est marquée en absolue parce que j'arrive a rien
			# si j'utilise $x0 _ce qui a mon sens, il faudrait faire_
			# pourquoi on est oblige de preciser la couche ici alors qu'on a deja fait un SetCurrentLayer ?
			$o_image->string( $o_image->gdTinyFont, 85, $pos, $mutation, $black, 'left', $layer );
		}

		$bool_name_displayed = 0;
		if( 1 == $b_label )
        {
	        # Calculate the position of the label to set it at the center of the rectangle representation
			my ( $Y0, $Y1 )= @a_coord[1,3];
			my $y_label = ( $Y0 + $Y1 )/2;

	        # Draw gene name if space between this label and the one below is superior to 1px
			my $name_space = ($graph_orientation eq '-') ? ($old_label_position-$y_label) : ($y_label-$old_label_position);

			if ( $name_space > $font->height() )
			{
				@a_label_coord = $o_image->stringVerticalAligned( $font, $x_label, $y_label, $snp_id, $res_lbl_color, side => $side, absolute_frame => 1 );

				$old_label_position = $y_label;
				$bool_name_displayed = 1;
			}
        }

		# 2) formule magique pour ecrire les donnees dans un fichier lu par javascript
		$self->NewMapElement( $element_id, \@a_coord );
		++$nb;
	}

	$self->_Set( __nb_snps => $nb );

	return;
}


=head2 Function GetTrack

 Usage        : $self -> GetTrack( )
 Function     : Using the appropriate graphic parameters,
                it defines the track in which the elements will be displayed.
                In other word it fixes $x0 and $x1.
 Returns      : a reference on a hash with the position of the track between X0 <--> X1
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut

sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_ );

	if( !defined $h_param{ o_box } )
	{
		throw LipmError::ParameterException('o_box','missing');
	}

	my( $o_box, $side ) = @h_param{ qw( o_box side ) };
	my %h_track_param;

# change for your need
	my $snp_width       = $self->GetGraphicParameter( 'width_snp', 4 );
	my $x_snp_label     = $self->GetGraphicParameter( 'x_snp_label', 0.275 );
	my $width_snp_label = $self->GetGraphicParameter( 'width_snp_label', 50 ); # max width for snp labels

	my $rh_dna_cyl = $self->GetDNACylinder( $side );

	my $x_center = ( $rh_dna_cyl->{X0} + $rh_dna_cyl->{X1} )/2;
	my $X0 = $x_center - $snp_width/2;
	my $X1 = $X0 + $snp_width;

	my $X_label;
	my %h_area = $o_box->GetBoxArea();
	if( $side eq "right" )
	{
		$X_label = $h_area{X0} + $x_snp_label;
	}
	else
	{
		$X_label = $h_area{X1} - $x_snp_label;
	}

 
	%h_track_param = ( %{$rh_dna_cyl}, width_label => $width_snp_label, X_label => $X_label, desc => "snp elements");
	# %h_track_param = ( X0 => $X0, X1 => $X1, width_label => $width_snp_label, X_label => $X_label, desc => "snp elements");

	return \%h_track_param;
}


=head2 DumpMapInformation

 Title        :
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the snp
 	X0 Y0 X1 Y1 SNP  species element start end strand parent symbol
 Returns      : none
 Args         : A filehandle for the map file (.coord file)

=cut

sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;

	# fields to show in remote control
	# <=> fields to print in mapX.coord
	my @a_fields = qw( chrom position strand values extern_id mutation );

	my @a_attr = qw( _feature _o_species _rh_mapareas );
	my( $feature, $o_species, $rh_map ) = $self->_Get( @a_attr );

	# doit-on generer le fichier ?
	my $nb = $self->_Get( '__nb_snps' );
	my $snp_map_threshold = $self->GetGraphicParameter( 'ceil_threshold_map_snp' );
	#warn "$snp_map_threshold  $nb";
	return if( $snp_map_threshold < $nb );

	foreach my $element ( keys %$rh_map )
	{

		my $ra_boxe  = $$rh_map{ $element }{ boxes }{ __primary };
		my @a_coords = ElementsSet::_inflate( @$ra_boxe );

		my @a_values = ( uc($feature), $o_species->GetLongName() );
		push( @a_values, $self->GetFields( $element, @a_fields ) );
		my @a_quoted_values = map( "\"$_\"", @a_values );
		#print $file_handle join( "\t", @a_coords, @a_values ), "\n";
		print $file_handle join( "\t", @a_coords, @a_quoted_values ), "\n";
	}

	return;
}


=pod

=head1 AUTHORS

 Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
 Sebastien Letort : sebastien.letort@toulouse.inra.fr

=head1 COPYRIGHT NOTICE

 This software is governed by the CeCILL license - www.cecill.info

=cut

1;
