# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

	SuppChart.pm - an interface to all chart of supplementary data

=head1 DESCRIPTION

	This classe is an interface that provides functions to integrate data
	of supp_X directory as chart.

	all data in sub directory of supp_X will be displayed as X chart.
	X can be line_plots, histograms, scatter_plots, heat_map or tiles.
	Each of them will be a class that derives from SuppChart.


=head1 SUBROUTINES

	__ReadLocalConf
	_NormalizeValue

=cut

package ElementsSet::SuppChart;

use strict;
use warnings;

use Data::Dumper;
use POSIX;

use base qw( ElementsSet );
my $VERSION = do {my @r = (q$Rev: 3728 $ =~ /\d+/g); $r[0]};

# those params are used if no value are set in graphics.cfg
# in graphics.cfg, key is the wanted key followed by '_feature'
my %H_GRAPHICS_DEFAULT = ( color_data => 'red' );

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : $self -> _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param a hash of parameters
 					mandatory
  						o_species    => SpeciesObserver object
  						chromosome => chromosome name or number

  					optional
  						feature			 	=> simple name of this feature of object (default 'class_name')
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 0)
  						file_directory		=> directory of data (default 'feature')
  						file_extension		=> extension of the data files (default 'feature')
						graph_orientation 	=> the orientation (default '+')
						primary_key 		=> in data files, this is the reference column (default 'element')
						start_label 		=> in data files, this is the start column  (default 'start')
						end_label	 		=> in data files, this is the end column  (default 'end')
						ra_data_headers		=> array ref., this is the array of sorted column which have to be store when data are read (default ('element','start','end','strand'))
						graphic_param_file	=> file for the graphic parameters initialization

 Attributes	  : objects attributes :
 					Element attributes :
 						feature, organism, chromosome, start, end
 					File and directory attributes :
 						file_directory, file_extension, ra_data_headers, primary_key, start_label, end_label
 					Representation attributes :
 						ra_map, graph_orientation

 					specific to this class
 					 _width, int, the width of the representation.
 					 _pos_x, int, the x position of the feature
 					 _lgd_label, string, label of this Chart into the legend.

 Globals	  :	none
 Errors		  : LipmError::ObjectException
=cut

sub _Init
{
	my $self = shift;
	my %h_params = ( @_ );

	# pas sur que ce soit le plus efficace
	$self->_Set( _width => $h_params{ width },
				 _pos_x => $h_params{ pos_x },
				 _file_ext => $h_params{ file_ext },
				 _filepath => $h_params{ filepath } );

	my @a_data_headers = qw( start end value );
	$self -> SUPER::_Init( %h_params,
	                    ra_data_headers => \@a_data_headers,
	                    primary_key => 'id'
	               );

	return if( $self->IsEmpty() );

	$self->__ReadLocalConf();
# 	$self->__SetGraphicParameters();

	return;
}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

=head2 Display

 Title        : Display
 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 Procedure    :	must be inherited.
 Returns      : none
 Args         : o_image	-> Image object
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
 Globals      : none

=cut
sub Display
{
	warn "Must be inherited";

	return;
}

=head2 function _NormalizeValue

	Title        : _NormalizeValue
	Usage        : $value = $o_xx->_NormalizeValue( $value, $x0 );
	Prerequisite : none
	Function     : set value to max if value > max
	               and value to min if value < min
	               It also normalize values.
	Returns      : $value, int the value as min < value < max
	Args         : $value, int, the value to modify

=cut
sub _NormalizeValue
{
	my $self = shift;
	my ( $value ) = @_;

	$value = $self->{ _min }	if ( $value < $self->{ _min } );
	$value = $self->{ _max }	if ( $value > $self->{ _max } );

	$value = ( $value - $self->{ _min } ) * $self->{ _ratio };

	return $value;
}

=head2 __SetGraphicParameters

	Title        : 
	Usage        : $o_xx->_NormalizeValue();
	Prerequisite : none
	Function     : set some local grahic parameters for optimization purpose
	               min, max and ratio
	Returns      : none
	Args         : none

=cut
sub __SetGraphicParameters
{
	my $self = shift;

	my $feat = $self->_Get( '_feature' );

	my $min   = $self->GetGraphicParameter( 'min_' . $feat );
	my $max   = $self->GetGraphicParameter( 'max_' . $feat );
	my $ratio = $self->GetGraphicParameter( 'ratio_' . $feat );

	$self->{ _min } = $min;
	$self->{ _max } = $max;
	$self->{ _ratio } = $ratio;

	return ;
}

=head2 function

	Title        : __ReadLocalConf
	Usage        : $o_xx->__ReadLocalConf();
	Prerequisite : none
	Function     : Read conf file (extension.cfg),
	               	set min, max, ratio and ra_color_data.
	               in the file (key= value), keys can be
	               min=X
	               max=X
	               color_data=X
	Args         : none

	Note         : c'est pas propre, mais je dois faire vite.
		=> il n'y a plus de fichier de conf local, on va lire les valeur dans graphics.cfg.

=cut
sub __ReadLocalConf
{
	my $self = shift;

	my $base_key = $self->_Get( '_feature' );	# type supp_xxx_ext
# 	if( $color_data eq '' )
# 	{	$color_data = $self->GetGraphicParameter( 'color_data_' . $base_key, 'red' );	}

	my $min = $self->GetGraphicParameter( 'min_' . $base_key );
	my $max = $self->GetGraphicParameter( 'max_' . $base_key );

	if( !defined $min or !defined $max )
	{
		my( $data_min, $data_max ) = $self->__GetMinMaxValues();
		$min = $data_min  if( !defined $min );
		$max = $data_max  if( !defined $max );
	}

	# we cannot have same values for min and max.
	++$max  if( $min == $max );

	my $ratio = $self->{ _width } / ( $max - $min );
# 	$self->_Get( '_o_param_graph' )->Set( 'ratio_' . $base_key, $ratio );

	$self->_Set( _min   => $min,
				 _max   => $max,
				 _ratio => $ratio );

	return ;
}

=head2 function GetLegendInfos

	Title        : GetLegendInfos
	Usage        : $ra_data = $o_xx->GetLegendInfos( $o_image );
	Prerequisite : none
	Function     : it prepare data that will be used into Legend object
	Returns      : $ra_data, a ref list which
	             	first data is data_color
	             	second data is the label
	Args         : $o_image, an Image object
	               it is used only if $self->{ _rh_cfg }{ ra_color_data } is a label
	               so if extension.cfg uses label instead of r,g,b

=cut
sub GetLegendInfos
{
	my $self = shift;
	my ( $o_image ) = @_;

	my $ra_color = $self->_GetColor( $o_image );

	my @a = ( $ra_color, $self->_Get( '_file_ext' ) );

	return \@a;
}


# no doc too ugly
sub _GetColor
{
	my $self = shift;
	my ( $o_image ) = @_;

	my $base_key = $self->_Get( '_feature' );	# type supp_xxx_ext
	my $color    = $self->GetGraphicParameter( 'color_data_' . $base_key,
						$H_GRAPHICS_DEFAULT{ color_data } );

	if( !defined $color )
	{
		my $msg =  'color_data_' . $base_key . " doesn't seemed to exist\n";
		$self->_Trace( $msg, '.', 'ICV' );

		return undef;
	}

	my @a_color = split( /,/, $color );
	if( 3 != @a_color )
	{	# il faut trouver le rbg a partir du label _un peu complexe
		my $index = $o_image->colorAllocate( $color );
		@a_color  = $o_image->rgb( $index );
	}

	return \@a_color;
}

=head2 GetFileName

 Title        : GetFileName
 Usage        : $self -> GetFileName()
 Prerequisite : object should be initialized
 Function     : get the url of the file which contain data for this ElementsSet object
 Returns      : a string
 Args         : none
 Attributes   : feature, organism, chromosome

=cut

sub GetFileName
{
	my $self = shift;

	my $filepath = $self->_Get( '_filepath' );

	return $filepath;
}


=head2 __GetMinMaxValues

 Usage    : my @a_min_max = $o_x->__GetMinMaxValues()
 Function : Read all the data file and retrieve the minimum and maximum values.
 Returns  : @a_min_max, an array of two integers.
 Args     : none
 Error    : none

=cut

sub __GetMinMaxValues
{
	my $self = shift;

	my $filename = $self->GetFileName();
	my $o_ndr = New DataReader::Whole();
	$o_ndr->Prepare( -from => $filename );
	my $ra_data = $o_ndr->FetchallArrayref(-line_type => 'HASH');

	my( $min, $max ) = ( INT_MAX, INT_MIN );
	foreach my $rh_row ( @$ra_data )
	{
		my $value = $$rh_row{ value };
		$min = $value  if( $value < $min );
		$max = $value  if( $max < $value );
	}

	return ( $min, $max );
} #__GetMinMaxValues


1;

__END__

=pod

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr
Sebastien Letort : sebastien.letort@toulouse.inra.fr

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
