# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ElementsSet inherited object to display Narcisse syntenies data.

=head1 DESCRIPTION

See also L<ElementsSet>

Syntenies is an ElementsSet inherit class.
Store syntenies data between the query organism and a subject.

Calculate the position of each subject syntenies and get bounds of each "sub-elements".


B<New attributes:>
- subject	=> short name of the target organism compare with the reference
- program	=> name of the program use to compare both organisms (default 'prbh')
- level		=> synteny level (default 5)
- view_level=> (default 0)
- gene_structure => boolean to show exons (default 0)

B<Colors management:>
The color management of this object use 3 globals paramaters :
 %H_CHROM_COLORS
 $DEFAULT_COLOR
 $LEVEL_CHAIN_LABEL

Those global parameters allow user to be sure that all segments of a same chromosome will be show with one and only one color.
To set a new colors file, use InitColors method.

The colors file must be formatted such as (rgb colors) :
I<#chrom	color
chrom_number	color_name>


=head1 SUBROUTINES

GetSliceParameters
GetTargetSliceParameters
__CalculateSyntenyZoomFactor
GetFileName

B<Class methods :>
GetChromColor
InitColors

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut
package ElementsSet::Synt;

use strict;
use warnings;

use POSIX;
use IO::File;
use Data::Dumper;
use Error qw(:try);
use Carp;
use Data::Dumper;

use base qw( ElementsSet );
use Box::ChromosomeSliceBox;
my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

# Constant
use constant MIN_PIXEL_HEIGHT => 5;
use constant REFERENCE => 'ref';
use constant TARGET => 'tgt';

# Global parameters
our %H_CHROM_COLORS;
our $DEFAULT_COLOR = 'grey_light';
our $LEVEL_CHAIN_LABEL = 'q_chain';		# Name of the column which link syntenies levels together

use constant ID_SEPARATOR => &ElementsSet::ID_SEPARATOR;

#TODO move this in a cfg file (graphics or defaults)
use constant MAX_SYNT => 1_000;

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 _Init

 Title        : _Init
 Usage        : _Init($rh_cfg, $organism, $chromosome, $start, $end [, $rh_param] [, colors_file])
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : target
 				program
 				level		=> the top-synteny level
 				view_level	=> the sub-synteny level
 				gene_structure => boolean to know if exons should be draw in target slices

 Attributes	  : objects attributes
					- target
 					- program
 					- level		=> the top-synteny level
 					- view_level=> the sub-synteny level
 					- gene_structure => boolean to know if exons should be draw in target slices

 Globals	  :	none

=cut
sub _Init
{
	my $self = shift;

	# Parameters initialization
	my %h_param = (
				__displayed_synt => {},
				target	=> '',
				program	=> 'prbh',
				level	=> 5,
				view_level => 0,
				gene_structure => 0,
				@_
				);
	
	# Change defaut parameters : data headers and file extension
	my @a_map_keys	   = ('reference_order','target_organism','target_chromosom','target_start','target_end','target_order','program');
	my @a_map_keys1	   = ('feature','target_organism','target_chromosom','target_start','target_end','target_order','organism','chromosom','start','end','reference_order','program');
	my @a_data_headers = ('element','start','end','sign','s_chrom','s_start','s_end','s_order');
	my $organism = $h_param{ o_species }->GetShortName();
	my %h_new_param = ( ra_data_headers => \@a_data_headers,
						ra_map_keys		=> \@a_map_keys,
						area_id_prefix	=> $organism.&ID_SEPARATOR.$h_param{target}.&ID_SEPARATOR,
						storage_mode		=> 'array',
						%h_param
						);
												
	# Some new attributes :
	my @attribute_names=qw(target program level view_level gene_structure);
	@{$self}{ @attribute_names }=@h_param{ @attribute_names };
	$self -> {ra_map_keys1}	= \@a_map_keys1;
	
	my $target_level = $self->{ level } - $self->{ view_level };
	$self->{ target_level } = $target_level;

	# Initialization
	$self -> SUPER::_Init( %h_new_param );

	#TODO modifier ici la gestion de ce test
	if ( $self->IsEmpty() && 0)
	{
		$self->_Trace( "no data loaded", '', 'IEV' );
		$self->{ rh_view_data }         = {};
		$self->{ rh_sort_view_element } = {};
		return $self;
	}	
		
	( $self->GetNumElements() < &MAX_SYNT )
			or throw NarcisseError( -bd => "Trying to load ". $self->GetNumElements() ." syntenies is too much work for the lazy Narcisse. <br />Please try changing range or synteny levels." );
	
	# TODO temporary patch waiting for a more thorough definition of the element id in the .synt files
	# WARNING this is absolutly mandatory as a lot of methods are using this slice id
	$self->__SetAllSliceId();
			
	$self->__LoadLowerLevelChains();
	
    return $self;
}

=head2 __SetTargetSlice

 Title        : __SetTargetSlice
 Usage        : $self->__SetTargetSlice($element_id,$target_box,$Y0,$Y1,$y0,$y1)
 Prerequisite : none
 Porcedure    : Store the hash parameter to instanciate the target's ChromosomeSlice $element_id.
 Returns      : none
 Args         : element_id : the ID of top-level synteny to store in a slice.
 				target_box : the upper Box object of target element.
 				Y0,Y1	   : absolute ordinates coordinates of the sub-box (pixel).
 				y0,y1	   : relative ordinates coordinates of the sub-box (nucleic).
 Attributes	  : slices
 Globals      : none

=cut
sub __SetTargetSlice
{
	my $self = shift;
	my ( $element_id, $target_box, $Y0, $Y1, $y0, $y1 ) = @_ or return;

	my ($pad_top, $pad_bot) = ($target_box->{padding_top},$target_box->{padding_bottom});
	my ($tgt_Y0, $tgt_Y1) 	= ($target_box->{Y0},$target_box->{Y1});

	# Calculate new box paddings :
	my $slice_pad_top = ($Y0-$pad_top < 0) ? abs($Y0-$pad_top+$tgt_Y0) : $pad_top;
	my $slice_pad_bot = ($Y1 > $tgt_Y1 - $pad_bot) ? $Y1+$pad_bot-$tgt_Y1 : $pad_bot;

	# Calculate new box bounds :
	my $slice_Y0 = ($Y0 - $pad_top < 0) ? 0 : $Y0-$pad_top;
	my $slice_Y1 = ($Y1+$pad_bot-$tgt_Y1 > 0) ? $Y1 : $Y1+$pad_bot;

	# Set Box attributes :
	my %h_slice_param = $target_box -> GetAttributes();
	%h_slice_param = ( 	%h_slice_param,
						Y0 => $slice_Y0,
					   	Y1 => $slice_Y1,
					    y0 => $y0,
					    y1 => $y1,
					    padding_top    => $pad_top,
					    padding_bottom => $pad_bot
						);

	# Store ChromosomeSliceBox data :
	$self->{slices}{ $element_id } = \%h_slice_param;
		
	return;
}

=head2 __LoadLowerLevelChains

 Title        :
 Usage        : $self -> __LoadLowerLevelChains();
 Prerequisite : none
 Procedure    : Parse each sub-level files to initialize data with view_level.
 Returns      : rh_view_data, ra_sort_view_elements
 Args         : none
 Globals	  : none

=cut
sub __LoadLowerLevelChains
{
	my $self = shift;

	my $root_level = $self->{ level };
	my $view_level = $self->{ view_level };
	my $target_level =  $self->{ target_level };

	my @a_o_syntdata = ();

	#-----------------------------------------------------
	# 1) Store root-level infos
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();

	for (my $k = $root_level; $k >= $target_level; $k--)
	{
		if ( $k == $root_level ) # Nothing has to be read, the data is already available
		{
			$a_o_syntdata[$k] = $self->_Get( '_o_data' );
		}
		else
		{
			my $file = $self->GetFileName( $k );
			$a_o_syntdata[$k] = $self->_ReadNarcisseFormatedFile( $file, $elt_start, $elt_end );
		}		
	}

	# We consider the hierarchical structure of synteny chains (a forest)
	# a chain of level k is composed of n chains of level k-1 (they share the same root level ancestor)
	# Let C.(k) be all the chains of level k and Cj(k-1) be a subchain of Ci(k)
	# Let A() denote the ancestor of a chain 
	# A ( chain_i ) := ancestor( chain_i )
	# We have
	# A( Cj(k-1) ) := ancestor( Cj(k-1) ) = ancestor( Ci(k) )
	# We store for each ancestor the different chains of level k :
	# push @{sons[k-1]{A( Cj(k-1) )}}, Cj(k-1)
	# Result :
	# 	$sons[$target_level] is a hash with the root_level synt id as keys 
	#   and the corresponding (array of) target level synt id as values 
	
	my @ancestor =();
	my @sons =();	
	
	
	if ( $root_level == $target_level ) # sons == fathers 
	{
		my @a_root_level_elements = $a_o_syntdata[$root_level]->GetSortedElements();	
		foreach my $root_elt ( @a_root_level_elements ) # A améliorer
		{
			push @{$sons[$root_level]{$root_elt}}, $root_elt;
		}
	}
	else   # we have to look for the progeny
	{
		my @a_fields=( $LEVEL_CHAIN_LABEL );	
		for (my $k=$root_level; $k>$target_level; $k--)
		{
			foreach my $k_level_chain ( $a_o_syntdata[$k]->GetSortedElements() )
			{	
				my ( $q_chains )= $a_o_syntdata[$k]->GetFields( $k_level_chain, @a_fields);
				foreach my $sub_chain ( split(":",$q_chains) )
				{
					$ancestor[$k-1]{$sub_chain}= ( $k == $root_level ) ? $k_level_chain : $ancestor[$k]{$k_level_chain} ;
					if ( $a_o_syntdata[$k-1]->Exist( $sub_chain ) && defined $ancestor[$k-1]{$sub_chain}) 
					#If the synt element is out of range or if it has no ancestor we omit it
					# (some level k synt might not be included in level k+1)
					{ 
						push @{$sons[$k-1]{$ancestor[$k-1]{$sub_chain}}}, $sub_chain;
					}
				}	 
			}		
		}
	}

	# We store the synteny data corresponding to each level as well as the hierarchical 
	# relationship between syntenies
	$self->{ __ra_synt_data }			=	\@a_o_syntdata;
	$self->{ __ra_synt_hierarchy }		=	\@sons;		
								
}

=head2 __SetAllSliceId

 Title        : __SetAllSliceId
 Usage        : $self ->__SetAllSliceId();
 Procedure    : Add a slice id to each synt element. 
                They will be used for creating links between ref and tgt syntenie segments
 				Synt elts are always loaded by the reference genome, hence the following slice nomenclature
                  ref_$organism_$target_$s_order 
                $q_order and $s_order are unique identifiers : we choose q_order (identifiers in q_chain will be used for subchains) 
 Returns      : none
 Args         : none

=cut
sub __SetAllSliceId
{
	my $self = shift;
		
	my @a_fields=qw( element );
	my $organism = $self->_Get( '_o_species' )->GetShortName();
	foreach my $element_id ( $self->GetSortElements() )
	{
		my ( $q_order ) = $self->GetFields( $element_id, @a_fields);
		my $slice_id  = join(&ID_SEPARATOR, (  "ref", $organism, $self->{ target }, $element_id, $q_order) );
		$self->AddFields($element_id, 'slice_id' => $slice_id );
	}
			
}

#------------------------------------------------#
#                  Predicates                    #
#------------------------------------------------#

=head2 IsDataAvailable

 Title        : IsDataAvailable
 Usage        : $self -> IsDataAvailable()
 Prerequisite : none
 Function     : Check if the data exists (most of the time this consists in verifying if the file exists)
 Returns      : 1 if it foes exist
 				or
 				0 if not

=cut
sub IsDataAvailable
{
	my $self = shift;
	my $file = shift;
	
	unless ( -e $file )
	{
		throw LipmError::ObjectException("Empty or inexistant synteny file $file",'!build',-dbg => "Probably a wrong level or program");
	}
		
	return (-e $file);
}

#------------------------------------------------#
#                  Display methods               #
#------------------------------------------------#

=head2 DisplaySlice

 Title        : DisplaySlice
 Usage        : $o_element -> DisplaySlice(%h_param);
 Prerequisite : rh_data and ra_sort_elements must be initialized
 Procedure    : After calling StoreTargetSlicesInfos, this procedure allow user to draw each targets slices syntenies infos.
 				1) Draw reference syntenies arrows if target box is conjointed to reference one.
 				2) Draw target level_id's slice arrows.
 				3) Draw links between the arrows if target box is conjointed to reference one.
 Returns      : none
 Args         : %h_param {
 					image 		 => Image object
 					reference_box=> reference Box object
 					target_box   => target Box object
 					level_id	 => id of the synteny to draw
 						 }
 Globals      : none

=cut
sub DisplaySlice
{
	my $self = shift;
	my %h_param = (
		o_image 		=> undef,
		-o_textual      => undef,
		reference_box	=> undef,
		target_box		=> undef,
		level_id		=> undef,
		@_
		);
	
	return if ( defined $h_param{ dont_display_anything } && 1 == $h_param{ dont_display_anything });

	#======================================
	# Testing object data and break the method if their is no data to draw.
	return if $self->IsEmpty();

	# Testing mandatory parameters
	my @mandatory = qw( o_image reference_box target_box level_id);
	foreach my $param ( @mandatory )
	{
		(defined $h_param{ $param } )
			or throw LipmError::ParameterException($param, 'missing');
	}
	
	
	# Some elements will be displayed, we register the current object in MapArea
	$self->NewMapElementSet();
	
	# General parameters
	my $o_image		= $h_param{o_image};
	my $o_ref_box 	= $h_param{reference_box};
	my $o_tgt_box 	= $h_param{target_box};
	my $level_id	= $h_param{level_id};
	my $layer       = $self->GetElementSetLayer();
	my $o_textual   = $h_param{ -o_textual };
	my $o_file;
	if ( defined $o_textual )
	{
		$o_file= $o_textual -> GetLayer( $layer );
	}

	my ($start,$end,$synt_sign) = $self->GetFields($level_id,qw( start end sign ) );
	
	#======================================
	# 1) Draw references syntenies view levels if the two boxes are neighbours
	if ($o_tgt_box->IsReferenceNeighbor() == 1)
	{
		$self -> DisplayReferenceViewLevels (image=> $o_image, level_id => $level_id, o_box => $o_ref_box, side => $o_tgt_box->GetJustification());
	}
	
    #======================================
	# 2) Draw target syntenies view levels.
	$self -> DisplayTargetViewLevels( image=>$o_image, level_id => $level_id, o_box => $o_tgt_box);

    #======================================
	# 3) Draw links between reference and target syntenies arrows.
	if ($o_tgt_box->IsReferenceNeighbor() == 1)
    {
		my $rh_displayed_synts=$self->GetDisplayedSynt();
		if ( defined $o_textual )
		{
			my $o_species = $self -> { _o_species };
			my $o_io      = $o_species -> GetInstanceObserver();
			my $reference = $o_species -> GetLongName();
			my @a_fields = qw( start end sign s_chrom s_start s_end );
			$o_textual -> PrintHeader ( $layer, 'reference', 'chromosome', 'start','end','sign','target','target chromosome','target start','target end' );
			my $level = $self -> { level };
			my $rh_synt_data = $self -> { __ra_synt_data } -> [$level];
			# TO DO - ça NE MARCHE QUE SI view_level = 0 !!!
			foreach my $ref_element (keys %$rh_displayed_synts)
			{
				my ($crap,$ref,$target,$id)=split(&ID_SEPARATOR,$ref_element);
				$target = $o_io -> GetSpeciesLongName($target);
				#warn $id . '==> ' . Dumper ($rh_synt_data -> { _rh_data });
				my $ra_data = $rh_synt_data -> { _rh_data } -> { $id };
				if ( defined $ra_data)
				{
					my ($crap1, $chrom, $start,$end,$sign,$s_chrom,$s_start,$s_end) = @$ra_data;
					print $o_file join ("\t",$reference,$chrom,$start,$end,($sign eq '-')? 1:-1,$target,$s_chrom,$s_start,$s_end) . "\n";
				}
			}
		}
		else
		{
			$o_image -> SetAbsoluteFrame();
	
			my $x 	  = $self->GetGraphicParameter('x_synt',20);	# x position of the syntenies representation
			my $space = $self->GetGraphicParameter('space_default',5);
			
			#my $x_mid = $o_joined_box->Width()/2;
			#my $layer = $self->GetElementSetLayer();;
			my $link_color = $o_image->colorAllocate($DEFAULT_COLOR, $layer);
			foreach my $ref_element (keys %$rh_displayed_synts)
			{
				my ($crap,$ref,$target,$id)=split(&ID_SEPARATOR,$ref_element);					
				my $rh_ref_info=$self->GetMapElement($ref_element);
				my $tgt_element=join(&ID_SEPARATOR,(&TARGET, $target,$ref,$id));
				my $rh_tgt_info=$self->GetMapElement($tgt_element);
				
				next unless (defined $rh_ref_info and  $rh_tgt_info);
				
				$rh_ref_info->{type} = $self->SUPER::PAIRED;
				$rh_tgt_info->{type} = $self->SUPER::PAIRED;
	
				my @coords_ref=@{$rh_ref_info->{boxes}{__primary}};
				my @coords_tgt=@{$rh_tgt_info->{boxes}{__primary}};	
				my ( $x_ref, $y_ref ) = ( int ( ($coords_ref[0]+$coords_ref[2])/2 ) , int( ($coords_ref[1]+$coords_ref[3])/2 ) );
				my ( $x_tgt, $y_tgt )=( int ( ($coords_tgt[0]+$coords_tgt[2])/2 ) , int( ($coords_tgt[1]+$coords_tgt[3])/2 ) );
				my ($x0,$y0,$x1,$y1) = &Image::ReorderCoordinates($x_ref, $y_ref, $x_tgt, $y_tgt);
				$o_image -> brokenLine( $x0, $y0, $x1, $y1, 5, $link_color, $layer);							
	    	}
		}
    	
    	#We go back to relative frame
		$o_image -> SetRelativeFrame();
    }

	return;
}

=head2 DisplayReferenceViewLevels

 Title        : DisplayReferenceViewLevels
 Usage        : $o_element -> DisplayReferenceViewLevels(%h_param);
 Prerequisite : Data should be correctly initialized.
 Procedure    : Draw reference's syntenies infos on Image object.
 Returns      : none
 Args         : %h_param {
 					image 	=> Image object
 					box		=> Box object (reference's box)
 					level_id=> id of the synteny to draw
 						 }
 Attributes	  :	none
 Globals      : none

=cut
sub DisplayReferenceViewLevels
{
	my $self = shift;
	my %h_param = (
		image	 => undef,
		level_id => undef,
		o_box	 	 => undef,
		side	 => undef,
		@_
		);

	#======================================
	# Testing mandatory parameters
	return if ( !defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );
	return if ( !defined $h_param{level_id} );
	return if ( !defined $h_param{o_box} );

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box	= $h_param{o_box};
	my $level_id= $h_param{level_id};
	my $side	= $h_param{side};
	
	if (!defined $side)
	{
		$side = ( $o_box->GetJustification() eq 'left') ? 'right' : 'left';
	}

	#======================================
	# Move the frame.
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();
	my $graph_orientation = $self->_Get( '_graph_orientation' );

	$o_image -> SetFrameChange( $o_box->GetDisplayedArea( $graph_orientation ),
								y0 => $elt_start, y1 => $elt_end );

	#======================================
	# Graphics parameters :
	my $x_synt	  = $self->GetGraphicParameter( 'x_synt' , 20 );			# x position of the syntenies representation
	my $width = $self->GetGraphicParameter( 'width_synt' , 2 );		# width of the syntenies representation
	my $layer = 'synt';
	my $color = &GetChromColor($o_image,$self->GetElementField($level_id,'chrom'), $layer);
	
	$o_image -> SetCurrentLayer( $layer );
	
	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );
		
	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };
	my $x = $rh_track->{ X };
	
	my $arrow_width = 6;
		
	# The target level syntenies
	my $ra_synt_level	= $self->{ __ra_synt_data };
	my $target_level = $self->{ target_level };
	my $o_target_level_synt = $ra_synt_level->[$target_level];	
	
	my $always_plus = "+";
		
	#======================================
	# Draw subchains data : arrow representation and links between sub-levels	
	my @drawed_syntenies=();
	my @fields = qw( start end sign );
	my @a_subchains = $self->GetSortedSubElements( $level_id );
	my $organism = $self->_Get( '_o_species' )->GetShortName();
	foreach my $subchain_id ( @a_subchains )
	{
		my ( $start, $end, $sign ) = $o_target_level_synt->GetFields( $subchain_id, @fields );	
		
		next if ($end < $elt_start || $start > $elt_end );		
		
		my @a_coord = $self->DrawSyntElement($o_image, $x0, $start, $x1, $end, $always_plus, $arrow_width, $color, $side );

		#Name of the corresponding element on the target side (see nomenclature defined by __SetAllSliceId)
		my $element_key = join( &ID_SEPARATOR, &REFERENCE, $organism, $self->{target}, $subchain_id );

		$self->NewMapElement( $element_key, \@a_coord  );
		$self->AddDisplayedSynt( $element_key );

		push @drawed_syntenies, { elt => $element_key, coord => \@a_coord, Y_pos => $a_coord[1]};
	}

	#In order to prevent dangling subchains due to a limited range
	my $chain_0 = $a_subchains[0];
	my $chain_n = $a_subchains[$#a_subchains];
	$self->FillReferenceChain( $o_target_level_synt, $level_id, $chain_0, $chain_n, $o_image, $rh_track, $color );
    
	#======================================
	# Draw links.
	$o_image -> SetAbsoluteFrame();
	
	my @sorted_drawed_syntenies = sort { $a->{ Y_pos } <=> $b->{ Y_pos }} @drawed_syntenies;
	my $spacer = $width +1;
	for my $i (0.. ($#sorted_drawed_syntenies-1) )
	{
		$self->DrawSyntChainLinks($o_image,$sorted_drawed_syntenies[$i],$sorted_drawed_syntenies[$i+1],$spacer,$color,$side)
	}
				
	$o_image -> SetRelativeFrame();

	return;
}

=head2 FillReferenceChain

 Title        : FillReferenceChain
 Usage        : $o_element -> FillReferenceChain(%h_param);
 Procedure    : Fill with a straight line remaining top_level synteny segment 
            	When zomming in some subchains might get out of range and are therfore absent from
            	the __ra_synt_hierarchy hierachy. Because this may appear on the image as a synteny 
            	disruption we fill the corresponding space with a straight line.
 Returns      : none 
 Args         : bunch

=cut
sub FillReferenceChain
{
	my $self = 	shift;
	my ($o_target_level_synt, $level_id, $chain_0, $chain_n, $o_image, $rh_track, $color ) = @_;
	
	my $X = $rh_track->{ X_in };
	my @sub_fields = qw( start end );
	my ( $level_id_start, $level_id_end ) = $self->GetFields( $level_id, @sub_fields );
	
	if (!defined $chain_0 && !defined $chain_n)#Subchains might not be available
	{
		$o_image->line( $X, $level_id_start, $X, $level_id_end, $color );
	}
	else
	{
		my ( $start_0, $end_0 ) = $o_target_level_synt->GetFields( $chain_0, @sub_fields );
    	$o_image->line( $X, $level_id_start, $X, $start_0, $color );
    	my ( $start_n, $end_n ) = $o_target_level_synt->GetFields( $chain_n, @sub_fields );
    	$o_image->line( $X, $end_n, $X, $level_id_end, $color );	
	}
		
	return ;	
}

=head2 DisplayTargetViewLevels

 Title        : DisplayTargetViewLevels
 Usage        : $o_element -> DisplayTargetViewLevels(%h_param);
 Prerequisite : Data should be correctly initialized.
 Procedure    : Draw target's slice syntenies infos on Image object.
 				Draw the chromosome number of this synteny.
 Returns      : none
 Args         : %h_param {
 					image 	=> Image object
 					box		=> Box object (target's slice box)
 					level_id=> id of the synteny to draw
 						 }
 Attributes	  :	none
 Globals      : none

=cut
sub DisplayTargetViewLevels
{
	my $self = shift;
	my %h_param = (
		image	 => undef,
		o_box => undef,
		level_id => undef,
		@_
		);

	#======================================
	# Testing mandatory parameters
	return if ( !defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );
	return if ( !defined $h_param{level_id} );
	return if ( !defined $h_param{o_box} );

	# General parameters
	my $o_image	= $h_param{image};
	my $o_box	= $h_param{o_box};
	my $level_id= $h_param{level_id};
	my $side	= ($o_box->GetJustification() eq 'left') ? 'right' : 'left';
	
	#======================================
	# Move the frame.
	my %h_slice_param = $self->GetTargetSliceParameters($level_id);
	$o_image -> SetFrameChange($o_box->GetDisplayedArea( $self->GetSliceOrientation($level_id) ), y0 => $h_slice_param{start}, y1 => $h_slice_param{end});
		
	#======================================
	# Graphics parameters :
	my $x_synt	  = $self->GetGraphicParameter('x_synt',20);			# x position of the syntenies representation
	my $width = $self->GetGraphicParameter('width_synt',2);		# width of the syntenies representation
	my $layer = 'synt';
	my $color = &GetChromColor($o_image,$self->GetElementField($level_id,'s_chrom'), $layer);
	
	$o_image -> SetCurrentLayer( $layer );

	# x positions
	my $rh_track = $self->GetTrack( o_box => $o_box, side => $side );
		
	my $x0 = $rh_track->{ X0 };
	my $x1 = $rh_track->{ X1 };
	my $x = $rh_track->{ X };
	
	my $arrow_width = 6;

	# The target level syntenies
	my $ra_synt_level	= $self->{ __ra_synt_data };
	my $target_level = $self->{ target_level };
	my $o_target_level_synt = $ra_synt_level->[$target_level];			

	#======================================
	# Draw "view level" data : arrow representation
	my @drawed_syntenies=();
	my @fields = qw( s_start s_end sign );
	my @a_subchains = $self->GetSortedSubElements( $level_id );
	my $organism = $self->_Get( '_o_species' )->GetShortName();
	foreach my $subchain_id ( @a_subchains )
	{
		my ( $start, $end, $sign ) = $o_target_level_synt->GetFields( $subchain_id, @fields );				
		#Difficult to know here if we are out of bounds........
		
		my @a_coord = $self->DrawSyntElement($o_image, $x0, $start, $x1, $end, $sign, $arrow_width, $color, $side );
			
		#Name of the corresponding element on the target side (see nomenclature defined by __SetAllSliceId)
		my $element_key = join( &ID_SEPARATOR, &TARGET, $self->{target}, $organism, $subchain_id );
		$self->NewMapElement( $element_key, \@a_coord  );		
		
		push @drawed_syntenies, { elt => $element_key, coord => \@a_coord, Y_pos => $a_coord[1]};
		
	}

	#In order to prevent dangling subchains due to a range that do not span the whole chain
	my $chain_0 = $a_subchains[0];
	my $chain_n = $a_subchains[$#a_subchains];
	$self->FillTargetChain( $o_target_level_synt, $level_id, $chain_0, $chain_n, $o_image, $rh_track, $color );

	#======================================
	# Draw links.
	$o_image -> SetAbsoluteFrame();
	
	my @sorted_drawed_syntenies = sort { $a->{ Y_pos } <=> $b->{ Y_pos }} @drawed_syntenies;
	my $spacer = $width +1;
	for my $i (0.. ($#sorted_drawed_syntenies-1) )
	{
		$self->DrawSyntChainLinks($o_image,$sorted_drawed_syntenies[$i],$sorted_drawed_syntenies[$i+1],$spacer,$color,$side)
	}
		
	$o_image -> SetRelativeFrame();

	#======================================
	# Draw chromosome number.
	my $chrom_number = $self->GetElementField($level_id,"s_chrom");
	my $x_label		 = ($side eq 'left') ? $x-7 : $x+7;
	my $y_middle	 = ( $h_slice_param{start} + $h_slice_param{end} )/2;
	my $font 		 = $o_image->gdSmallFont;
	my $label_color  = &GetChromColor($o_image,$self->GetElementField($level_id,'s_chrom'), $layer);
	my $label_side	 = ($side eq 'left') ? 'right' : 'left';

	if ( $o_box->Height() > $font->height() )
	{
		$o_image -> stringVerticalAligned($font,$x_label,$y_middle,$chrom_number,$label_color, side => $label_side);
	}
	
	return;
}

=head2 FillTargetChain

 Title        : FillTargetChain
 Usage        : $o_element -> FillTargetChain(%h_param);
 Procedure    : exactly the same as FillReferenceChain but for the target
 Returns      : none 
 Args         : bunch

=cut
sub FillTargetChain
{
	my $self = 	shift;
	my ($o_target_level_synt, $level_id, $chain_0, $chain_n, $o_image, $rh_track, $color ) = @_;
	
	my $X = $rh_track->{ X_in };
	my @sub_fields = qw( s_start s_end );
	my ( $level_id_start, $level_id_end ) = $self->GetFields( $level_id, @sub_fields ); 
    
    if (!defined $chain_0 && !defined $chain_n)#Subchains might not be available
	{
		$o_image->line( $X, $level_id_start, $X, $level_id_end, $color );
	}
	else
	{
    	my ( $start_0, $end_0 ) = $o_target_level_synt->GetFields( $chain_0, @sub_fields );
    	my ( $start_n, $end_n ) = $o_target_level_synt->GetFields( $chain_n, @sub_fields );
	
		my $sign = $self->GetFields( $level_id, 'sign' ); 
	
		if ( $sign eq '+' )
		{
			$o_image->line( $X, $level_id_start, $X, $start_0, $color );
    		$o_image->line( $X, $end_n, $X, $level_id_end, $color );
		}
		else
		{
			$o_image->line( $X, $level_id_start, $X, $start_n, $color );
    		$o_image->line( $X, $end_0, $X, $level_id_end, $color );
		}
	}
   		
	return ;	
}

=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : 
 Returns      : 
 Args         : 

=cut
sub GetTrack
{
	my $self = shift;
	my %h_param = ( 
					o_box => undef,
					side => "right",
					@_
					);
		
	( defined $h_param{ o_box } )	
		or throw LipmError::ParameterException('o_box','missing');
		
	my $o_box = $h_param{ o_box };
	my $side = 	$h_param{ side };
		
	my %h_track_param;
	
	my $x_synt	  = $self->GetGraphicParameter('x_synt',20);			# x position of the syntenies representation
	my $width = $self->GetGraphicParameter('width_synt',2);		# width of the syntenies representation
		
	my ($X0,$X1,$X,$X_in,$X_out);
		
	if ( $side eq "left" )
	{
		$X0 = $o_box->GetX0() + $x_synt;
		$X1 = $X0 + $width;	
		$X = $X0;
		$X_in = $X1;
		$X_out = $X0;
	}
	else
	{
		$X1 = $o_box->GetX1() - $x_synt;
		$X0 = $X1 - $width;		
		$X = $X1;
		$X_in = $X0;
		$X_out = $X1;
	}
	%h_track_param = ( X0 => $X0, X1 => $X1, X => $X, X_in => $X_in, X_out => $X_out, desc => "synteny elements");
		
	return \%h_track_param;
	
}

=head2 DrawSyntElement

 Title        : DrawSyntElement
 Usage        : $object -> DrawSyntElement($x0,$y0,$x1,$y1,$orientation, $arrow_thickness,$color,$side);
 Procedure    : Draw the arrow associated with the syntny element
 Returns      : none
 Args         : bunch

=cut
sub DrawSyntElement
{
	my $self = shift;
	my ($o_image, $x0, $y0, $x1, $y1, $orientation, $arrow_thickness, $color, $side ) = @_;
		
	my $frame_orientation = $o_image->{frame}->GetFrameOrientation();

	my $direction = ( $orientation eq "+" ) ? 1 : -1  ;
				
	#Restrict the element to the drawing area defined by the start and end of this ElsementSet object
	my $draw_y0 = ( $y0 > $o_image->GetMin_y() ) ? $y0 : $o_image->GetMin_y();
	my $draw_y1 = ( $y1 < $o_image->GetMax_y() ) ? $y1 : $o_image->GetMin_y();		
	
	if ( $side eq "right")
	{
	 return $o_image->LeftJustifiedHalfArrow( $x0, $y0, $x1, $y1, $direction, $arrow_thickness, $color);
	} else
	{
	 return $o_image->RightJustifiedHalfArrow( $x0, $y0, $x1, $y1, $direction, $arrow_thickness, $color);
	}
			
}

=head2 DrawSyntChainLinks

 Title        : DrawSyntChainLinks
 Usage        : $object -> DrawSyntChainLinks($o_image,$synt_previous,$synt_next,$color,$side);
 Prerequisite : The appropriate layer should have been set
 Procedure    : Draw the element on the DNA cylinder on the correct strand
 Returns      : string
 Args         : bunch

=cut
sub DrawSyntChainLinks
{
	my $self = shift;
	my ($o_image, $rh_synt1, $rh_synt2, $spacer, $link_color, $side) = @_;
	
	my @coord_synt1 = @{ $rh_synt1->{coord} };
	my @coord_synt2 = @{ $rh_synt2->{coord} };
	
	if ( $side eq "left" )
	{
		my ($x0,$y0) = @coord_synt1[2,3];
		my ($x1,$y1) = @coord_synt2[2,1];	
		
		my $mid_point = ( $y0 + $y1 )/2; 	
		
		$o_image ->line ($x0,$y0,$x0-$spacer,$mid_point,$link_color); 					 
		$o_image ->line ($x0-$spacer,$mid_point,$x1,$y1,$link_color);	
	}
	else
	{
		my ($x0,$y0) = @coord_synt1[0,3];
		my ($x1,$y1) = @coord_synt2[0,1];	
		
		my $mid_point = ( $y0 + $y1 )/2; 	
		
		$o_image ->line ($x0,$y0,$x0+$spacer,$mid_point,$link_color); 					 
		$o_image ->line ($x0+$spacer,$mid_point,$x1,$y1,$link_color);	
			
	}
	
}

=head2 InitColors
 Title        : InitColors
 Usage        : $self -> InitColors($file)
 				ElementsSet::Synt -> InitColors($file [,$session_dir [,cfg_dir]])
 Prerequisite : none
 Procedure    : Initialize the global hash with data of the file $file.
 				Catch an exception if the file doesn't exist, but store file name for the future saving.
 Returns      : none
 Args         : $file
 Globals      : %H_CHROM_COLORS, color_file
 Error		  : LipmError::IOException

=cut
sub InitColors
{
	my $self = shift if ($_[0] =~ /Synt/);
	my $file = shift;

	try
	{
		my $o_ndr = New DataReader::Whole();
		$o_ndr->Prepare( -from   => $file );	  	               		
		my $rh_data = $o_ndr->FetchallHashref( "chrom", -line_type => "HASH" );

		foreach my $chrom (keys(%$rh_data))
		{
			my $rh_element_data = ${$rh_data}{$chrom};
			$H_CHROM_COLORS{$chrom} = ${$rh_element_data}{color};
		}
	}
	catch LipmError::IOException with
	{
		my ($E,$scalar) = (shift,shift);
		my $message = $E->Explain();
		warn $message;
	};

	return;
}

=head2 AddDisplayedSynt
 Title        :
 Usage        : 
 Function     : 
 Args         : synt $element_id  the element ID

=cut
sub AddDisplayedSynt
{
	my $self = shift;
	my $element_id = shift;
	
	$self->{ __displayed_synt }{$element_id}++;
}

=head2 GetDisplayedSynt
 Title        :
 Usage        : 
 Function     : 
 Returns      :

=cut
sub GetDisplayedSynt
{
	my $self = shift;
	
	return $self->{ __displayed_synt};
}

=head2 __CalculateSyntenyZoomFactor

 Title        : __CalculateSyntenyZoomFactor
 Usage        : $self->__CalculateSyntenyZoomFactor($element_id)
 Prerequisite : none
 Function     : Calculate the "synteny factor" which is the factor of zoom between query and subject elements
 Returns      : an integer
 Args         : $element_id : the element ID
 Globals	  : organism, chromosome, start, end

=cut
sub __CalculateSyntenyZoomFactor
{
	my ($self,$element_id) = (shift,shift);

	#======================================
	# Init factor and test parameters.
	my $zoom_factor = 1;
	my @a_fields = qw( start end s_start s_end );
	my( $start, $end, $s_start, $s_end ) = $self->GetFields( $element_id, @a_fields );
	
	#======================================
	# Calculate this factor.
	my $ref_range = $end - $start;
	my $tgt_range = $s_end - $s_start;

	$zoom_factor = $tgt_range/$ref_range if ($tgt_range != 0 && $ref_range != 0);

	return $zoom_factor;
}

=head2 __RecalculateTargetStartEnd

 Title        : __RecalculateTargetStartEnd
 Usage        : $self->__RecalculateTargetStartEnd($element)
 Prerequisite : none
 Function     : Recalculate start/end for the target slice, according to the reference segments that could be ignore
 				(resegmentation due to zoom or scroll)
 				If a part of the whole synteny is not represented :
 					- we recalculate bounds of target side using a "zoom" factor, which represent the ratio between target and reference range.
 					- we tried to center target representation on the nearest center reference element.
 Returns      : array: (start, end)
 Args         : $element : the element ID
 Globals	  : rh_data

=cut
sub __RecalculateTargetStartEnd
{
	my ($self,$element_id) = (shift,shift);

	my ($ref_start,$ref_end,$tgt_start,$tgt_end,$tgt_chrom,$sign) = $self->GetFields($element_id, qw( start end s_start s_end s_chrom sign ) );
# &General::ShowDebug( [ $ref_start, $ref_end, $tgt_start, $tgt_end, $tgt_chrom, $sign ] );
	my $synt_factor = $self->__CalculateSyntenyZoomFactor($element_id);

	#======================================
	# Get display start/end of reference and calculate reference middle.
	my $obj_start = $self->_GetStart();
	my $obj_end   = $self->_GetEnd();

	my $view_ref_start = ( $obj_start > $ref_start ) ? $obj_start : $ref_start;
	my $view_ref_end   = ( $obj_end   < $ref_end )   ? $obj_end   : $ref_end;

	my $ref_middle = int(($view_ref_start+$view_ref_end)/2);
	my $tgt_middle = int(($tgt_start+$tgt_end)/2);

# 	my $root_level = $self->{ level };
	my $view_level = $self->{ view_level };
# 	my $target_level =  $self->{ target_level };
# warn "$root_level:$view_level:$target_level";
# Calculate correct target middle
	if( $obj_start > $ref_start && $obj_end < $ref_end )
	{
		my $file    = $self->GetFileName(0);
# 		my $file    = $self->GetFileName( $view_level );

		my $o_ndr = New DataReader::Region();
		$o_ndr->Prepare( -from   => $file,
		                 -where  => {  VALUE => { s_chrom => $tgt_chrom  }, RANGE => { start => $view_ref_start, end => $view_ref_end} }
		                );
		my $ra_data = $o_ndr -> FetchallArrayref();
		my $o_data  = New NarcisseData::ArrayReturn(ra_data => $ra_data, rh_headers => $o_ndr -> GetHeadersColumns());
		my $closest_elt = $o_data->FindNearestElement( $ref_middle );
# &General::ShowDebug( $o_nd, $ref_middle, $rh_return ) if( !defined( $closest_elt ) );
# &General::ShowDebug( $file, $tgt_chrom, $view_ref_start, $view_ref_end );
# &General::ShowDebug( $file, $tgt_chrom, $view_ref_start, $view_ref_end ) if( !defined( $closest_elt ) );

		my @attributes = qw( start end s_start s_end );
		my ( $elt_start, $elt_end, $elt_tgt_start, $elt_tgt_end) = $o_data->GetFields( $closest_elt, @attributes );

		# Test positions
	    $elt_start = $obj_start if( $obj_start > $elt_start );
	    $elt_end   = $obj_end   if( $obj_end   < $ref_end );
	    $elt_tgt_start = $tgt_start if ($elt_tgt_start < $tgt_start);
	    $elt_tgt_end   = $tgt_end   if ($elt_tgt_end   > $tgt_end);

	    # Working with the linear equation y = ax + b
	    my $a = ($sign eq '+') ? $synt_factor : -$synt_factor;
	    my $b = $elt_tgt_start - $a*$elt_start;
	    $tgt_middle = $a*$ref_middle+$b;
	}
	else
	{
		if ($sign eq '+')
		{
			$tgt_middle = $tgt_start + int(($ref_middle-$ref_start)*$synt_factor);
		}
		else
		{
			$tgt_middle = $tgt_end - int(($ref_middle-$ref_start)*$synt_factor);
		}
	}

	#======================================
	# Recalculate target start/end :
	my $tgt_range = int(($view_ref_end-$view_ref_start)*$synt_factor);
	my $view_tgt_start = $tgt_middle - $tgt_range/2;
	my $view_tgt_end   = $tgt_middle + $tgt_range/2;

	#======================================
	# Test start/end values :
	if ($view_tgt_start < $tgt_start)
	{
		$view_tgt_start = $tgt_start;
		$view_tgt_end   = $view_tgt_end + ($tgt_start-$view_tgt_start);
	}
	elsif ($view_tgt_end > $tgt_end)
	{
		$view_tgt_end   = $tgt_end;
		$view_tgt_start = $view_tgt_start - ($view_tgt_end-$tgt_end);
	}
	elsif ($view_tgt_start < $tgt_start && $view_tgt_end   > $tgt_end)
	{
		$view_tgt_start = $tgt_start;
		$view_tgt_end   = $tgt_end;
	}

	return (int($view_tgt_start),int($view_tgt_end));
}

#------------------------------------------------#
#             Managing the maps                  #
#------------------------------------------------#

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_xxx -> DumpMapInformation( $filehandle )
 Function     : Writes the map information corresponding to the synteny elements
	The file format is the following
 	X0 Y0 X1 Y1 SYNT species chrom start end element s_species s_chrom s_start s_end s_order X'0 Y'0 X'1 Y'1
 	where
 		X0 Y0 X1 Y1 are the coordinates of the reference synteny box
 		X'0 Y'0 X'1 Y'1 are the coordinates of the target synteny box
 	For Paired elements the information is written twice :  
 	 X0 Y0 X1 Y1 SYNT Ref_info Target_info  X'0 Y'0 X'1 Y'1
 	 X'0 Y'0 X'1 Y'1 SYNT  Target_info Ref_info X0 Y0 X1 Y1
 Returns      : none
 Args         : A filehandle
 
=cut
sub DumpMapInformation
{
	my $self = shift;
	my $file_handle = shift;
	
	#All the synt subchain are stored in __ra_synt_data at $target_level
	# The target level syntenies
	my $ra_synt_level	= $self->{ __ra_synt_data };
	my $target_level = $self->{ target_level };
	my $o_target_level_synt = $ra_synt_level->[$target_level];
		
	my @fields_ref = qw( chrom start end element );
	my @fields_tgt = qw( s_chrom s_start s_end s_order );

	my @a_attr = qw( _rh_mapareas _o_species _feature );
	my( $rh_map, $o_species, $feature ) = $self->_Get( @a_attr );
	my $organism = $o_species->GetShortName();
	
	#Specific treatment for synteny elements to prevent from huge .coord file
	my $o_io = $o_species->GetInstanceObserver();
	my $num_elements = scalar (keys %$rh_map);
	
	#Useless to print map information when there is more than $max_map_elt	
	if ( $num_elements > $o_io->GetCfg( 'MAX_MAP_ELEMENTS' ) )
	{
		warn "Too many map elements. No synteny elements will be stored in .coord file \n";
		return ;	
	}	
		
	foreach my $element ( keys %$rh_map )
	{
		my ($crap,$ref,$target,$id)=split(&ID_SEPARATOR,$element);
		
        my ($ref_long,$target_long) = $o_io->GetSpeciesLongName($ref,$target);
		my $str_feature   = uc( $feature );
		my $str_genomic   = "genomic";
		
        #Two synteny segments are available, one for the query and one for the target										
		if ( $self->IsaPairedMapElt( $element ) ) 
		{		
			next unless ($ref eq $organism );
											
			my $ref_element = $element;
			#The corresponding element on the target side (see nomenclature defined by __SetAllSliceId)
			my $tgt_element=join(&ID_SEPARATOR,(&TARGET,$target,$ref,$id)); 
		
			my ( $ra_ref_coord, $ra_ref_values) = $self->_GetElementInfos($o_target_level_synt, $id, $ref_element, $ref_long, \@fields_ref);
			my ( $ra_tgt_coord, $ra_tgt_values) = $self->_GetElementInfos($o_target_level_synt, $id, $tgt_element, $target_long, \@fields_tgt);

			my @a_fields    = ( @$ra_ref_coord, "\"$str_feature\"", @$ra_ref_values, @$ra_tgt_values, "\"$str_genomic\"", @$ra_tgt_coord );
			my $line_ref = join( "\t", @a_fields );
			print $file_handle $line_ref."\n";

			@a_fields = ( @$ra_tgt_coord, "\"$str_feature\"", @$ra_tgt_values, @$ra_ref_values, "\"$str_genomic\"", @$ra_ref_coord );
			my $line_tgt = join( "\t", @a_fields );
			print $file_handle $line_tgt."\n";
		}
        
        #Only one synteny segment available and this is a target (case of multiple targets not neighbour to the reference)
		else 
		{
			my $tgt_element = $element;
			my $ref_element = "__none__";
			
			my @a_ref_coord = (0,0,0,0);
			my @a_ref_values  = ( "\"$target_long\"", "\"" . $o_target_level_synt->GetFields( $id, @fields_ref) . "\"" );
			my ( $ra_tgt_coord, $ra_tgt_values ) = $self->_GetElementInfos($o_target_level_synt, $id, $tgt_element, $ref_long, \@fields_tgt);

			#A single line will be displayed, the line corresponding to the target synteny (inverted values because we are on the target side)
			my @a_fields = ( @$ra_tgt_coord, "\"$str_feature\"", @$ra_tgt_values, @a_ref_values, "\"$str_genomic\"", @a_ref_coord );
			my $line_tgt = join( "\t", @a_fields );
			print $file_handle $line_tgt."\n";	
		}
	}

}


=head2 _GetElementInfos

 Title        : 
 Usage        :  $rh_infos = _GetElementInfos( $element_id )
 Function     : Extracts information and coordinate position of and element
 Returns      : a reference on a hash with the correspinding information
 Args         : The element id
 
=cut
sub _GetElementInfos
{
	my $self = shift;
	my $o_target_level_synt = shift;
	my $elt_id = shift;
	my $side_elt_id = shift;
	my $species = shift;
	my $ra_fields = shift;
	my $rh_infos;
	my @a_area_coord=();
	my @a_values=();
	
	my $rh_map = $self->_Get( '_rh_mapareas' );
	if(  !defined( $$rh_map{ $side_elt_id } ) )
	{
		@a_area_coord = (0,0,0,0);
		@a_values =  map("\"$_\"", ("","","","","") );		
	} 
	else
	{
		my $rh_info = $$rh_map{ $side_elt_id };
		@a_area_coord  = ElementsSet::_inflate(@{$rh_info->{boxes}{__primary}});
		#@a_values = ($species,  $o_target_level_synt->GetFields($elt_id,@$ra_fields) );
		@a_values = map("\"$_\"",($species,  $o_target_level_synt->GetFields($elt_id,@$ra_fields) ) );
	}
	
	return ( \@a_area_coord, \@a_values );

}

#------------------------------------------------#
#               The slices       
#           									 #
#------------------------------------------------#

#  The slices denote the synteny higher level elements

=head2 StoreTargetSlicesInfos

 Title        : StoreTargetSlicesInfos
 Usage        : $self -> StoreTargetSlicesInfos()
 Prerequisite : none
 Procedure    : For each reference synteny, store infos to construct sub-box and target slices.
 Returns      : none
 Args         : %h_param {
 					image 			=> Image object,
					reference_box	=> reference Box object,
					target_box		=> target Box object,
 						}
 Globals      :

=cut
sub StoreTargetSlicesInfos
{
	my $self = shift;
	my %h_param = (
		image 			=> undef,
		reference_box	=> undef,
		target_box		=> undef,
		@_
		);

	#======================================
	# Testing object data and break the method if their is no data to draw.
	return if ( $self->IsEmpty() );

	# Testing mandatory parameters
	#TODO modify the following 3 tests
	return if ( !defined $h_param{image} || ref($h_param{image}) !~ /(Image)|(GD::Image)/  );
	return if ( !defined $h_param{reference_box} || ref($h_param{reference_box}) !~ /Box/  );
	return if ( !defined $h_param{target_box} || ref($h_param{target_box}) !~ /Box/  );

	# General parameters
	my $o_image		= $h_param{image};
	my $o_ref_box 	= $h_param{reference_box};
	my $o_tgt_box 	= $h_param{target_box};

	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();
	my $graph_orientation = $self->_Get( '_graph_orientation' );

	$o_image -> SetFrameChange($o_ref_box->GetDisplayedArea( $graph_orientation ), y0 => $elt_start, y1 => $elt_end );
 
 	# The target chromosomes will be stored
 	my %h_target_chroms;
 
	#======================================
	# For each 'top' level syntenies, store target slices infos (ChromosomeSliceBox and ChromosomeSlice parameters)
	my @a_fields = qw( slice_id start end sign s_chrom);
	foreach my $element_id ( $self->GetSortElements() )
	{
		my ($slice_id,$start,$end,$synt_sign,$s_chrom) = $self->GetFields( $element_id, @a_fields );

		# Don't store out-of-bounds syntenies :
		next if ($end < $elt_start || $start > $elt_end );

	    $h_target_chroms{ $s_chrom }++;

		# 1) Calculate correct nucleic target slice bounds :
		my ($y0,$y1) = $self -> __RecalculateTargetStartEnd( $element_id );

		# 2) Calculate correct absolute target slice bounds :
			# Test if synteny reference bounds are include in representation. Else use correct values.
		$start 	= $elt_start if( $start < $elt_start );
		$end	= $elt_end	 if( $end > $elt_end );
			# Transform relative to absolute values
		my ($X0,$Y0) = $o_image->ToAbsoluteCoordinates(0,$start);
		my ($X1,$Y1) = $o_image->ToAbsoluteCoordinates(0,$end);

		# 3) Store the target slice box parameters.
		$self -> __SetTargetSlice ( $element_id, $o_tgt_box, $Y0,$Y1, $y0,$y1 );
	}

	$self->_Set( 'target_chrom_occ' => \%h_target_chroms );

	return;
}

=head2 GetSliceAbsoluteBounds

 Title        : GetSliceAbsoluteBounds
 Usage        : $self->GetSliceAbsoluteBounds($element_id)
 Prerequisite : StoreTargetSlicesInfos should be call before
 Function     : Get the absolute ordinates values of the slice $element_id
 Returns      : (Y0,Y1)
 Args         : $element_id : ID of top-level synteny of this slice.
 Attributes	  : slices
 Globals      : none

=cut
sub GetSliceAbsoluteBounds
{
	my $self = shift;
	my $element_id = shift ;

	return (0,0) if (!defined $element_id || !defined $self->{slices}->{$element_id});

	my $rh_param = $self->{slices}->{$element_id};
	my @a_return = ($rh_param->{Y0}+$rh_param->{padding_top},$rh_param->{Y1}-$rh_param->{padding_bottom});

	return @a_return;
}

=head2 GetSliceRelativeBounds

 Title        : GetSliceRelativeBounds
 Usage        : $self->GetSliceRelativeBounds($element_id)
 Prerequisite : StoreTargetSlicesInfos should be call before
 Function     : Get the relative ordinates values of the slice $element_id
 Returns      : (y0,y1)
 Args         : $element_id : ID of top-level synteny of this slice.
 Attributes	  : slices
 Globals      : none

=cut
sub GetSliceRelativeBounds
{
	my $self = shift;
	my $element_id = shift ;

	return (0,0) if (!defined $element_id || !defined $self->{slices}->{$element_id});

	my $rh_param = $self->{slices}->{$element_id};
	my @a_return = ($rh_param->{y0},$rh_param->{y1});
	
	return @a_return;
}

=head2 GetSliceClosestSyntElt

 Title        : 
 Usage        : $self->GetSliceClosestSyntElt($element_id,$pos)
 Function     : Get the synt elt belonging to the slice identified by element_id closest to $pos
 Returns      : a ref_hash containing the infos of the corresponding view_elt_id
 Args         : $element_id : ID of top-level synteny of this slice, a position on the slice
 Attributes	  : slices
 Globals      : none

=cut
sub GetSliceClosestSyntElt
{
	my $self = shift;
	my $element_id = shift ;
	my $pos = shift;

	return undef if ( $pos < $self->GetSliceStart( $element_id ) || $pos > $self->GetSliceEnd($element_id) );

	my @synt_elt = $self->GetSortedSubElements( $element_id );

	# A little dichotomy game 
	my ( $left, $right )=( 0, $#synt_elt );
	
	while ( $left != $right )
	{
		my $middle =  ($left + $right )/2 ;
		my $sub_chain = $synt_elt[ int($middle) ];
		if ( $pos < $self->GetElementField( $sub_chain , 'start' )  )
		{
			$right = int( $middle ) ;
		}
		else
		{
			$left = ceil( $middle );
		}
	}
		
	return $self->GetViewLevelElement( $synt_elt[$left] );
}

=head2 GetTargetProjectedPos

 Title        : 
 Usage        : $self->GetTargetProjectedPos($element_id,$pos)
 Function     : Get the target position corresponding to reference $pos according to a projection 
                based on the sub-chain closest to $pos (we are within a higher level synt hence on
                a target single chromosome)
 Returns      : a target position
 Args         : $element_id : ID of top-level synteny of this slice,
                $pos        : a position on the slice
 Attributes	  : slices
 Globals      : none

=cut
sub GetTargetProjectedPos
{
	my $self = shift;
	my $element_id = shift ;
	my $pos = shift;

	return undef if ( $pos < $self->GetSliceStart( $element_id ) || $pos > $self->GetSliceEnd( $element_id ) );

	my @synt_elt = $self->GetSortedSubElements( $element_id );

	# The target level syntenies
	my $ra_synt_level	= $self->{ __ra_synt_data };
	my $target_level = $self->{ target_level };
	my $o_target_level_synt = $ra_synt_level->[$target_level];			
	
	# A little dichotomy game 
	my ( $left, $right )=( 0, $#synt_elt );
	
	while ( $left != $right )
	{
		my $middle =  ($left + $right )/2 ;
		my $sub_chain = $synt_elt[ int($middle) ];
		
		my ( $sub_chain_start ) = $o_target_level_synt->GetFields($sub_chain , 'start' );
		
		if ( $pos < $sub_chain_start  )
		{
			$right = int( $middle ) ;
		}
		else
		{
			$left = ceil( $middle );
		}
	}
	my ( $s_start, $s_end ) = $o_target_level_synt->GetFields( $synt_elt[$left], qw( s_start s_end ) );
		
	return  ( $s_start + $s_end )/2 ;
}

=head2 GetSliceRepresentationType

 Title        : GetSliceRepresentationType
 Usage        : GetSliceRepresentationType($element)
 Prerequisite : DisplaySlice should be call before.
 Function     : Determine which eleemnts should be shown in this slice, if it's height is too small we didn't draw all slice elements
 Returns      : A string : basic, full
 Args         : $element : the element ID
 Globals	  : none

=cut
sub GetSliceRepresentationType
{
	my $self = shift;
	my $element_id = shift;

	my ($px_start,$px_end) = $self->GetSliceAbsoluteBounds( $element_id );
	my $min_pixel_height   = $self->GetGraphicParameter('min_pixel_height', &MIN_PIXEL_HEIGHT);

	# Tests if the slice box is big enough, in pixels, in order to be legible
	my $representation_type = ( ($px_end-$px_start) < $min_pixel_height) ? 'basic' : 'full';

	return $representation_type;
}

=head2 GetTargetSliceBox

 Title        : GetTargetSliceBox
 Usage        : $self -> GetTargetSliceBox($element_id)
 Prerequisite : none
 Function     : return the reference of the sub-box Box::ChromosomeSlice object corresponding to the element ID $element_id.
 Returns      : Box::ChromosomeSliceBox ref
 Args         : $element_id : ID of top-level synteny of this slice.

=cut
sub GetTargetSliceBox
{
	my $self = shift;
	my $element_id = shift ;

	return () if (!defined $element_id || !defined $self->{slices}->{$element_id});

	my %h_param = %{$self->{slices}->{$element_id}};
	
	my $slice_box = Box::ChromosomeSliceBox -> New( %h_param );

	return $slice_box;
}

=head2 GetTargetSliceParameters

 Title        : GetTargetSliceParameters
 Usage        : $o_synt -> GetTargetSliceParameters($slice_id)
 Prerequisite : DisplaySlice should be call before.
 Function     : Return the hash which allow user to instanciate a new ChromosomeSlice object for the slice $slice_id.
 Returns      : The ChromosomeSlice::New parameter.
 Args         : $slice_id : the element ID
 Globals	  : none

=cut
sub GetTargetSliceParameters
{
	my $self = shift;
	my $elt_id = shift;

	#======================================
	# Set parameters and store them in an hash-map.
	my @fields = qw( s_chrom sign );
	my ( $s_chrom, $sign ) = $self->GetFields( $elt_id, @fields );

	my ($start,$end)= $self->GetSliceRelativeBounds( $elt_id );
	my $representation_type = $self->GetSliceRepresentationType( $elt_id );

	my( $target, $o_species ) = $self->_Get( qw( target _o_species ) );
	my $o_inst_obs = $o_species->GetInstanceObserver();
	my $o_target = New SpeciesObserver( species => $target,
										-o_inst_obs => $o_inst_obs );
	my %h_return = (  o_species             => $o_target,
					  chromosome 			=> $s_chrom,
					  start 				=> $start,
					  end 					=> $end,
					  reference				=> $o_species->GetShortName(),
					  o_synt				=> $self,
					  synt_id				=> $elt_id,
					  graph_orientation 	=> $sign,
					  representation_type	=> $representation_type
					);
					
	return %h_return;
}

=head2 GetSliceOrientation

 Title        : GetSliceOrientation
 Usage        : $o_synt -> GetSliceOrientation($slice_id)
 Function     : Return the synteny orientation (sign in the Narcisse files)
 Returns      : the synteny orientation
 Args         : $slice_id : the element ID

=cut
sub GetSliceOrientation
{
	my $self = shift;
	my $elt_id = shift;

	my ( $sign ) = $self->GetFields( $elt_id, qw( sign ));
	
	return $sign;
}

=head2 GetViewLevelElement

 Title        : GetViewLevelElement
 Usage        : $object ->GetViewLevelElement($view_level_id)
 Function     : returns a hash with the corresponding information
 Returns      : a reference ona hash
 Args         : an element identifier

=cut
sub GetViewLevelElement
{
	my $self = shift;
	my $id  = shift;

	return $self->{ rh_view_data }->{ $id };
}

=head2 GetViewLevelField

 Title        : GetViewLevelField
 Usage        : $object -> GetViewLevelField($view_level_id,$field)
 Prerequisite : rh_view_data must be initialized
 Function     : Get the value of the field $field.
 Returns      : $field value.
 Args         : $view_level_id = name of the element (primary_key_name on data)
				$field		   = name of the field to return
 Attributes	  :	rh_view_data
 Globals      : none

=cut
sub GetViewLevelField
{
	my $self = shift;
	my ( $id,$field ) = @_;

	# Test parameters
	return undef if (!defined $self->{ rh_view_data }->{ $id });

	return $self->{ rh_view_data }->{ $id }->{ $field };
}

=head2 GetViewLevelFields

 Title        : GetViewLevelFields
 Usage        : $object -> GetViewLevelFields($view_level_id,@fields)
 Prerequisite : rh_view_data must be initialized
 Function     : Get the values of all fields @field.
 Returns      : The array corresponding to @fields values
 Args         : $view_level_id = name of the element (primary_key_name on data)
				@fields		   = list of the fields
 Attributes	  :	rh_view_data
 Globals      : none

=cut
sub GetViewLevelFields
{
	my $self = shift;
	my ( $id, @a_fields ) = @_;

	return @{$self->{ rh_view_data }->{ $id }}{ @a_fields };
	
=pod
	my @a_return = ();

	foreach my $field (@a_fields)
	{
		push (@a_return, $self->GetViewLevelField($id,$field));
	}

	return @a_return;
=cut

}

=head2 GetViewLevelIDList

 Title        : GetViewLevelIDList
 Usage        : $o_element->GetViewLevelIDList($element_id)
 Prerequisite : none
 Function     : Get the list of view_level IDs for the level $element_id
 Returns      : Array
 Args         : rh_sort_view_elements
 Globals      : none

=cut
sub GetViewLevelIDList
{
	my $self           = shift;
	my ( $element_id ) = @_;

	return () if (!defined $self->{rh_sort_view_elements}->{$element_id});

	my @a_return = @{$self->{rh_sort_view_elements}->{$element_id}};

	return @a_return;
}

=head2 GetSortedSubElements

 Title        : GetSortedSubElements
 Usage        : $o_element->GetSortedSubElements( $element_id )
 Prerequisite : none
 Function     : Get the list of target_level subchain IDs corresponding to the synt $element_id (of level level)
 Returns      : Array of element_id
 Args         : $element_id
 Globals      : none

=cut
sub GetSortedSubElements
{
	my $self           = shift;
	my ( $element_id ) = @_;

	my $target_level = $self->{ target_level };
		
	return () if (!defined $self->{ __ra_synt_hierarchy }[$target_level]{$element_id});

	my @a_return = @{ $self->{ __ra_synt_hierarchy }[$target_level]{$element_id} };

	return @a_return;
}

=head2 GetSliceParameters

 Title        : GetSliceParameters
 Usage        : GetSliceParameters($element)
 Prerequisite : StoreTargetSliceInfos should be call before.
 Function     : Calculate the new bounds of the subject element according to the query element position.
 Returns      : An hash-map : (query_organism, chromosome, start, end, subject_organism, synteny_factor)
 Args         : $element : the element ID
 Globals	  : none

=cut
sub GetSliceParameters
{
	my $self = shift;
	my $element = shift;

	# Get target nucleic start/end
	my ($tgt_start,$tgt_end)= $self->GetSliceRelativeBounds($element);

	# Get synteny zoom factor
	my $zoom_factor = $self->__CalculateSyntenyZoomFactor($element);

	# Get some element attributes
	my @fields = qw( start end s_chrom sign );
	my ($start, $end, $tgt_chrom, $sign) = $self->GetElementField($element, @fields );
		
	my $elt_start = $self->_GetStart();
	my $elt_end   = $self->_GetEnd();
	my @a_attr    = qw( _chromosome _o_species _o_map_area );
	my( $chrom, $o_species, $o_map ) = $self->_Get( @a_attr );
	my $organism = $o_species->GetShortName();

	# Get reference nucleic start/end
	my $ref_start = ( $start > $elt_start ) ? $start : $elt_start;
	my $ref_end   = ( $end   < $elt_end   ) ? $end   : $elt_end;

	my %h_return = (  organism 			=> $organism,
					  chromosome 		=> $chrom,
					  target			=> $self->{target},
					  start 			=> $ref_start,
					  end				=> $ref_end,
					  synt_factor		=> $zoom_factor,
					  tgt_start			=> $tgt_start,
					  tgt_end			=> $tgt_end,
					  tgt_chrom			=> $tgt_chrom,
					  graph_orientation => $sign,
					  o_maparea			=> $o_map
					);

	return %h_return;
}

=head2 GetSliceStart

 Title        : GetSliceStart
 Usage        : GetSliceStart($element)
 Function     : Retruns the start of the synt element called a slice here (the top level synt element)
 Returns      : the start of the synt element
 Args         : $element : the element ID
 Globals	  : none

=cut
sub GetSliceStart
{
	my $self = shift;
	my $element = shift;
	
	return $self->GetElementField( $element, 'start' );

}

=head2 GetSliceEnd

 Title        : GetSliceEnd
 Usage        : GetSliceEnd($element)
 Function     : Returns the end of the synt element called a slice here (the top level synt element)
 Returns      : the end of the synt element
 Args         : $element : the element ID
 Globals	  : none

=cut
sub GetSliceEnd
{
	my $self = shift;
	my $element = shift;
	
	return $self->GetElementField( $element, 'end' );

}

#------------------------------------------------#
#                ElementSet overloads            #
#------------------------------------------------#

=head2 GetFileName

 Title        : GetFileName
 Usage        : $self -> GetFileName([$level])
 Prerequisite : object should be initialized
 Function     : overload method
 Returns      : a string
 Args         : level -> optional, the synteny level
 Attributes   : type, organism, chromosome, subject
 Globals	  : none

=cut
sub GetFileName
{
	my $self = shift;
	my( $level ) = @_;

	$level = $self->_Get( qw( level ) ) unless( defined $level );

	my @a_attr = qw( _o_species _chromosome target program );
	my( $o_species, $chrom, $target, $prog ) = $self->_Get( @a_attr );

	my $filepath = $o_species->GetSyntFiles(
        target => $target,
        prog   => $prog,
        level  => $level,
        -chrom  => $chrom
        );

	return $filepath;
}

=head2 GetSortedSliceId

 Title        : GetSortedSliceId
 Usage        : $object -> GetSortedSliceId()
 Prerequisite : data must be initialized
 Function     : Returns an array of reference on each line (in fact an array organized as in the file)
 Returns      : an array of reference
 Args         : none
 Attributes   : ra_sort_elements
 Globals      : none

=cut
sub GetSortedSliceId
{
	my $self = shift;
	
	my @sorted_slice_id;
	
	foreach my $element ( $self->GetSortElements() )
	{
		my ($slice_id) = $self->GetFields($element, 'slice_id');
		push @sorted_slice_id, $slice_id;
	}
	
	return @sorted_slice_id;
}

#------------------------------------------------#
#                Utility functions               #
#------------------------------------------------#

=head2 GetChromColor

 Title        : GetChromColor
 Usage        : Syntenies::GetChromColor(o_image,chromosome_number [,layer])
 Prerequisite : none
 Function     : Get the index of the color of chromosome 'chromosome_number', or set the default color
 Returns      : Color index (integer >=0)
 Args         : o_image 		  -> object image
 				chromosome_number -> chromosome number
 				layer 			  -> optional, choose the layer in which colors will be allocated.
 Globals      : $H_CHROM_COLORS, $DEFAULT_COLOR

=cut
sub GetChromColor
{
	my ($o_image,$chrom_nb) = (shift,shift) or return 0;
	my $layer = shift;
	my $index = 0;

	if (!defined $H_CHROM_COLORS{$chrom_nb})
	{
		$index = $o_image->colorAllocate($DEFAULT_COLOR,$layer);
	}
	else
	{
		$index = $o_image->colorAllocate($H_CHROM_COLORS{$chrom_nb},$layer);
	}

	return $index;
}

=head2 Infos

 Title        : Infos
 Usage        : $self->Infos
 Function     : Returns some information abour the current object
 Returns      : string = the target program
 Args         : none
 Globals	  : target_list

=cut
sub Infos
{
	my $self = shift;
	
	my $prefix = $self->SUPER::Infos();
	
	my @a_fields = split( "_", $prefix);

	my ( $target, $program, $target_level ) = $self->_Get( qw (target program target_level) );
	
	push @a_fields, ( $target, $program, $target_level );
	
	return join("\t", @a_fields );
}


package ElementsSet::Synt::Karyotype;
# NOT YET FINISHED

=pod

=head1 NAME

 Synt::Karyotype

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet::Synt );

=head2 procedure _Init

 Usage        : my $o_x = New ElementsSet::Synt::Karyotype( %h );
 Prerequisite : none
 Fonction     : initialise an instance of ElementsSet::Synt::Karyotype class.
 Returns      : a valid ElementsSet::Synt::Karyotype object
 Args         :  %h

=cut

sub _Init
{
	my $self = shift;
	my %h = ( 
			target	=> 'porcine',
			program	=> 'genomic',
			level	=> 3,
        @_ );

	$self->SUPER::_Init( %h );
# 	warn Dumper $self;

	return;
}


=head2 GetTrack
 Title        : GetTrack
 Usage        : $object -> NewTrack( )
 Function     : Defines the track on which the ElemenSet will be displayed
 Returns      : a reference on a hash with the follownig parameters :
                 - X0 and X1 position of the cyto track
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut

sub GetTrack
{
	my $self   = shift;
	my %h_args = (  o_box => undef,
					@_ );

	# mandatories
	LipmError::ParameterException::Assert( 'o_box' => $h_args{ o_box } );

	my %h_box = $h_args{ o_box }->GetDisplayedArea();
	my $box_width = $h_box{ X1 } - $h_box{ X0 };

	# width of the cyto representation
	my $width  = $self->GetGraphicParameter( 'width_cyto',10 );

	my $X0 = ( $h_box{ X0 } + $h_box{ X1 } - $width ) / 2;
	my $X1 = $X0 + $width;

	return { X0 => $X0, X1 => $X1,
	         Y0 => $h_box{ Y0 }, Y1 => $h_box{ Y1 },
	         desc => 'cyto elements for karotype' };

}



sub Display
{
	my $self = shift;
	$self->SUPER::Display( @_ );

	my %h_param = ( image 	=> undef,
					o_box   => undef,
					side	=> 'right',
					@_ );

	my $o_image = $h_param{ image };

	my $rh_track = $self->GetTrack( @_ );
	my ( $X0, $X1 ) = @$rh_track{ qw( X0 X1 ) };

	my $width = $X1 - $X0;
# warn Dumper [ keys %$self ];
#my $chrom = $self->_Get( '_chromosome' );
#my $o_data = $self->_Get( '_o_data' );
#my $ra_synt = $self->_Get( '__ra_synt_data' );
#warn Dumper [ $ra_synt ];

	# my $cx = $X0 + $width/2;
	# $o_image->arc( $cx,$self->_GetStart(), $width,$width, 180,360, $black );

	# $cx = $X1 - $width/2;
	# $o_image->arc( $cx,$self->_GetEnd(), $width,$width, 0,180, $black );

	return;
}

1;


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

