# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Template to create a new ElementsSet inherited object.

=head1 DESCRIPTION

This class is just a template to construct a new ElementsSet inherited object.

B<Steps to construct a new ElementsSet object :>

1- Copy this code and change the package name

2- Changes on _Init method :
	* add new attributes with their default values.
	* change defaults parameters like : file extension, start_label, data headers ...

3- Overload Display method if you want to draw data

4- Overload GetFileName method if data are not store in the same path (/organism/file_directory/chromosome.file_extension )


B<Add a new check box to manage your image :>
See ChromosomeSlice and javascript documentation.

=head1 SUBROUTINES

Add new subroutines here.


=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

#----------------------------------------
# 1- Change package name
#----------------------------------------
package ElementsSet::Template;

use strict;
use warnings;

use Data::Dumper;

use base qw( ElementsSet );
my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};


=head2 _Init

 Title        : _Init
 Usage        : $self -> _Init(%h_param)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param a hash of parameters
 					mandatory
  						organism  		=> short name of the organism
  						chromosome		=> chromosome name or number

  					optional
  						feature			 	=> simple name of this feature of object (default 'class_name')
  						start				=> start of the representation (default 0)
  						end					=> end of the representation  (default 0)
  						file_directory		=> directory of data (default 'feature')
  						file_extension		=> extension of the data files (default 'feature')
						graph_orientation 	=> the orientation (default '+')
						primary_key 		=> in data files, this is the reference column (default 'element')
						start_label 		=> in data files, this is the start column  (default 'start')
						end_label	 		=> in data files, this is the end column  (default 'end')
						ra_data_headers		=> array ref., this is the array of sorted column which have to be store when data are read (default ('element','start','end','strand'))
						graphic_param_file	=> file for the graphic parameters initialization

 Attributes	  : objects attributes :
 					Element attributes :
 						feature, organism, chromosome, start, end
 					File and directory attributes :
 						file_directory, file_extension, ra_data_headers, primary_key, start_label, end_label
 					Representation attributes :
 						ra_map, graph_orientation

 Globals	  :	none
 Errors		  : LipmError::ObjectException

=cut
sub _Init
{
	my $self = shift;
	my %h_param = @_;

	#----------------------------------------
	# 2-2 Change defaults parameters :
	#----------------------------------------
	# If you have to change a default parameter (file extension, start_label, data headers ...) :
	my @a_data_headers = ('first_header','second_header','n_header');	# => Default column returned by GetFields method
	# @a_data_headers n'est plus utilise
	$self -> SUPER::_Init(	#ra_data_headers => \@a_data_headers ,
							%h_param );

	# your new attributes
	$self->_Set( __attribute => $value );

	return $self;
}


=head2 Display

 Usage        : $object -> Display(o_image [,side]);
 Prerequisite : Object must be initialized correctly.
 				$self->IsEmpty() must return 0
 Procedure    : Display data of the Element inherited object on the Image object.
 				Moreover, you can store drawing elements in 'areas' by using the NewMapElement method.
 Returns      : none
 Args         : o_image	-> Image object
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')
 Globals      : none

=cut
#----------------------------------------
# 3- Write your own display method
#----------------------------------------
sub Display
{
	my $self = shift;
	my %h_param = (
			image 	=> undef,
			side 	=> undef,
			o_box   => undef,
			@_ );

	#======================================
	# Don't draw data if their weren't loaded
	#Checking if the elementset is not empty and if the image and o_box parameters are OK
	return unless $self -> ProceedDisplay( %h_param );

	# General parameters
	my $o_image	= $h_param{image};
	my $side 	= $h_param{side};

	# !! Optional !!
	# If side = center :
	# 	1) This Display is set to a 'right' side mode
	#	2) Call another Display with side equal to 'left'.
	# This is use for example when you want to draw the reference at center and 2 targets to each side of it.
	# Then data are draw at each side => repeat some informations.
	if ($side eq 'center')
	{
		$side = 'right';
		$self -> Display(%h_param, side => 'left');
	}

	#--------------------------
	# 3-1 Initialize parameters
	#--------------------------

	# Name of the Image layer you want to draw in. By default this is "template" (class name).
	my $layer = $self->GetElementSetLayer();

	#	graphic constants * Graphics parameters => using $self->GetGraphicParameter
	my $x 		   = $self->GetGraphicParameter('x_template',40);
	my $width	   = $self->GetGraphicParameter('width_template',100);
	my $color_name = $self->GetGraphicParameter('color_template','blue');

	# colors	 * Allocate colors index to each layers you want to draw in.
	my $color_id	= $o_image->colorAllocate($color_name,$layer);
	my $color_alpha = $o_image->colorAllocateAlpha($color_name,$alpha,$layer); # Moreover, you could add an alpha to use transparency.

	my $color = "whatever";

	# fonts	 * Use fonts :
	my $font = $o_image -> gdTinyFont;	# See GD::Image doc to know the list of fonts.

	# x positions
	my $rh_track = $self->GetTrack( name => $layer, o_box => $o_box, side => $side );
	my( $x0, $x1 ) = @$rh_track{ qw( X0 X1 ) };

	# from now on objects will be drawn on a single layer : $layer	
	$o_image -> SetCurrentLayer( $layer );

	#--------------------------
	# 3-3 Draw elements that are between start-end
	#--------------------------
	my @fields= qw( symbol start end strand);  # Specify here the fields that are specific to the elements you want to draw
	foreach my $element_id ($self->GetSortElements())
	{
		my ($symbol,$start,$end,$strand) = $self->GetFields($element_id, @fields);	

		# 1) template representation
		my @a_coord = 	$self->DrawElement( $o_image, $x0, $start, $x1, $end, $color, $strand);

		# 2) Coordinate storage
		$self->NewMapElement( $element_id, \@a_coord  );
	}

	return;
}


=head2 Function GetTrack

 Usage        : $self -> GetTrack( )
 Function     : Using the appropriate graphic parameters,
                it defines the track in which the elements will be displayed.
                In other word it fixes $x0 and $x1.
 Returns      : a reference on a hash with the position of the track between X0 <--> X1
 Args         : o_box   -> the box corresponding to the chromosome slice
 				side	-> Set the position of the representation : 'left', 'center' or 'right' (default = 'right')

=cut

sub GetTrack
{
	my $self = shift;
	my %h_param = (
					o_box => undef,
					side => "right",
					@_ );

	if( !defined $h_param{ o_box } )
	{
		throw LipmError::ParameterException('o_box','missing');
	}

	my( $o_box, $side ) = @h_param{ qw( o_box side ) };
	my %h_track_param;

# change for your need
# 	my $template_width = $self->GetGraphicParameter( 'width_template', 4 );
# 	my $rh_dna_cyl = $self->GetDNACylinder( $side );
# 
# 	my $x_center = ( $rh_dna_cyl->{X0} + $rh_dna_cyl->{X1} )/2;
# 	my $X0 = $x_center - $template_width/2;
# 	my $X1 = $x_center + $template_width/2;
# 
# 	%h_track_param = ( X0 => $X0, X1 => $X1,  desc => "template elements");

	return \%h_track_param;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
