#
#	sebastien.letort@toulouse.fr
#	Created: April 03, 2008
#	$Id: FastaReader.pm 3702 2011-07-27 12:37:29Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package FastaReader;

=pod

=head1 NAME

 FastaReader - a class to easily read fasta file, especially multi fasta file

=head1  SYNOPSIS

 my $o_fr = New FastaReader( $file );

 # get the 2nd sequence in raw format
 my $ra_headers = $o_fr->GetHeaders();
 my $header     = $$ra_headers[ 1 ];
 my $seq        = $o_fr->GetRawSequence( $header );	# one string, one line

 # your multifasta file contains mamalian sequence
 # species is specified in sequence header with sp=human for example
 # you want to write human fasta sequence into separated files
 my $i = 0;
 my $ra_headers = $o_fr->GetHeaders( 'sp=human' );
 foreach $header ( @$ra_headers )
 {
 	my $fs = new IO::File( "human_$i", 'w' );
 	$seq   = $o_fr->GetFastaSequence( $header );
 	print $fs $seq;
 }

 # you want all the headers sequences related to glucose, aldose, and enzyme.
 # this information is noted into the header of the sequence.
 @a_headers = $o_fr->GetHeaders( 'glucose|aldose|ase\s' );

=head1 DESCRIPTION

 This class provide easy access to fasta file.
 It is especially designed if you access specific sequence of a multi fasta
 file or want to read those sequences in another order than the reading order.

=head1 BUGS

 headers do not have "\n" at the end anymore.

=cut

use strict;
use warnings;

# use Data::Dumper;
use IO::File;

use General;
use LipmError::IOException;
use LipmError::ParameterException;

use base qw( LipmObject );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New FastaReader( $filename );
 Prerequisite : none
 Fonction     : The constructor
 Args         : $filename, filename of a fasta file

=cut

sub _Init
{
	my $self = shift;
	my ( $filename ) = @_;

	# -d "$filename" the quotes are necessary because of a warning when we come from CGI.pm....
	throw LipmError( -text => "$filename is a directory\n" ) if( -d "$filename" );

	# If the input file is compressed (or stdin), we first copy the file to a temporary
	if ( &WhichStreamIn( $filename ) =~ /\|/ )
	{
		my $tmp_file = $filename . '.tmp';
		my $fh_in = &GetStreamIn ( $filename ) or throw LipmError::IOException( $filename, 'f' );
		my $fh_out= &GetStreamOut( $tmp_file ) or throw LipmError::IOException( $tmp_file, 'w' );
		while ( my $line = <$fh_in> )
		{
			print $fh_out $line;
		}
		$self -> _Set('__tmp_file',$tmp_file);
		$filename = $tmp_file;
	}
	else
	{
		$self -> _Set('__tmp_file',"");
	}
			
	my $fh = $self->{ __fh } = new IO::File( $filename, 'r' )
					or throw LipmError::IOException( $filename, 'f' );

	$self->{ __h_seqs } = {};
	my $rh_seqs = $self->{ __h_seqs };

	# Read the fasta file and keep only the headers and the position of the sequence
	my @a_headers = ();
	while( my $line = <$fh> )
	{
		# Skip all lines except headers
		next if( $line !~ /^>/ );

		# The line is a header: keep the header and the position
		chomp $line;
		my $header = $line;
		my $offset = $fh->getpos();

		# Each header should be uniq
		if( defined( $$rh_seqs{ $header } ) )
		{
			my $txt = "you have more than one sequence with the header -$header-\n";
			my $dbg = "As we use a hash to store sequences, we cannot allow two sequences to have the same header. Headers are our keys.\n";
			throw LipmError( -user_msg => $txt, -text => $dbg );
		}

		# We may add the sequence position, as well as the header
		# NB - The array of headers is in the order defined by the fasta file
		push @a_headers, $header;
		$$rh_seqs{ $header } = $offset;
	}

	$self->{ __ra_headers } = \@a_headers;

	return;
}
sub DESTROY
{
	my $self = shift;
	my $tmp_file = $self->_Get('__tmp_file');
	if ( $tmp_file ne '')
	{
		unlink $tmp_file;
	}
}

sub WriteIndicedFasta
{
	my $self = shift;
	my ($file_fasta) = @_;

	my $fh_fasta = GetStreamOut($file_fasta,'w') or throw IOException( $file_fasta, 'f');
	my $ra_headers = $self->{ __ra_headers };
	my $rh_seqs    = $self->{ __h_seqs };
	for (my $i=1; $i <= @$ra_headers; $i++)
	{
		my $header = $ra_headers->[$i-1];
		print $fh_fasta ">" . $i . "\n";
		print $fh_fasta $self->GetSequence( $header ) . "\n";
	}
}

sub GetIndicedHeader
{
	my $self = shift;
	my ($index) =@_;
	my $ra_headers = $self->{ __ra_headers };
	if ($index <= 0 or $index > @$ra_headers)
	{
		throw LipmError::ParameterException( 'index', 'value' );
	}
	$index--;
	return $ra_headers->[$index];
}


=head2 function GetFastaSequence

 Title        : GetFastaSequence
 Usage        : $seq = $o_xx->GetFastaSequence( $header );
 Prerequisite : none
 Function     : none
 Returns      : $seq, a sequence fasta formatted
 Args         : $header, a header that correspond to only one sequence
 TODO - Throw an exception if $header not defined
 REMOVED BY MANU - UTILISER GetSequence A LA PLACE !!!!
=cut

#sub GetFastaSequence
#{
#	my $self = shift;
#	my ( $header ) = @_;
#
#	return undef	if( !defined $header );
#	my $seq = "";
##
#	#locate the seq in the file
#	my $fh = $self->{ __fh };
#	my $offset = $self->{ __h_seqs }{ $header };
## 	$fh->seek( $offset, SEEK_SET );
#	$fh->setpos( $offset );
#	while( my $line = <$fh> )
#	{
#		last if( $line =~ /^>/ );
#
#		#suppression des espaces s'il y en a
#		# pas sur que ca restera ici ca ; cette fonction ne devait au depart que reformater.
#		# faire une fonction GetCleanedFastaSequence ?
#		$line =~ s/ //g;
#
#		$seq .= $line;
#	}
#
#	return $seq;
#}

=head2 function GetSequence

 Title        : GetSequence
 Usage        : $seq = $o_obj->GetSequence( $header );
 Function     : Return a string containing the sequence
                No space, no EOL characters, length ($seq) returns the REAL sequence length
 Returns      : The sequence
 Args         : $header, a header
 Error        : Throw an exception if the header is not found
=cut

sub GetSequence
{
	my $self = shift;
	my ( $header ) = @_;

	my $rh_seqs = $self->{ __h_seqs };
	if (!defined $header || !defined $rh_seqs->{$header})
	{
		throw LipmError::ParameterException( 'index', 'value' );
	}

	# Find the position of the sequence inside the file
	my $fh_fasta = $self->{ __fh };
	my $offset   = $rh_seqs->{$header};
	#$fh_fasta->seek( $offset, SEEK_SET );
	$fh_fasta->setpos( $offset );

	my $seq = "";
	while( my $line = <$fh_fasta> )
	{
		last if( $line =~ /^>/ );

		# removing spaces and eol if any
		chomp $line;
		$line =~ s/ //g;
		$seq .= $line;
	}

	return $seq;
}



#	return undef	if( !defined $header );
#	my $seq = "";
##
#	#locate the seq in the file
#	my $fh = $self->{ __fh };
#	my $offset = $self->{ __h_seqs }{ $header };
## 	$fh->seek( $offset, SEEK_SET );
#	$fh->setpos( $offset );
#	while( my $line = <$fh> )
#	{
#		last if( $line =~ /^>/ );
#
#		#suppression des espaces s'il y en a
#		# pas sur que ca restera ici ca ; cette fonction ne devait au depart que reformater.
#		# faire une fonction GetCleanedFastaSequence ?
#		$line =~ s/ //g;
#
#		$seq .= $line;
#	}
#
#	return $seq;
#}


=head2 function GetRawSequence

 Title        : GetRawSequence
 Usage        : $seq = $o_xx->GetRawSequence( $header );
 Prerequisite : none
 Function     : read and return the sequence corresponding to the header.
 Returns      : $seq, the sequence, into only one string
 Args         : $header, a header that corresponds to only one sequence
                $flg = If false (default) the line is chopped
                       If true, the line is kept AS IS
 TODO - Throw an exception if $header not defined
 TODO - Voire cette histoire louche de flags (cf. SnpReader)

=cut

sub GetRawSequence
{
	my $self = shift;
	my ( $header ) = @_;

	return undef	if( !defined $header );
	my $seq = "";

	#locate the seq in the file
	my $fh = $self->{ __fh };
	my $offset = $self->{ __h_seqs }{ $header };
# 	$fh->seek( $offset, SEEK_SET );	# ne fonctionne pas !
	$fh->setpos( $offset );
	while( my $line = <$fh> )
	{
		last	if( $line =~ /^>/ );
		$seq .= $line;
	}

	return $seq;
}


=head2 function GetHeaders

 Title        : GetHeaders
 Usage        : $ra_headers = $o_xx->GetHeaders( $regexp );
 Prerequisite : none
 Function     : search all headers that match the $regexp.
 Returns      : $ra_headers, a sorted list ref to all headers.
 Args         : $regexp, a regular expression.
                	will be used as ($header =~ /$regexp/);
                if $regexp is undef, all headers are returned.

 Note         : beware if you look for a specific header and if it contains '|'
                be sure to neutralized them with '\' or "\Q$regexp\E"

=cut

sub GetHeaders
{
	my $self = shift;
	my ( $regexp ) = @_;

# 	my @a_headers = keys %{ $self->{ __h_seqs } };
	my @a_headers = @{ $self->{ __ra_headers } };
	return \@a_headers	if( !defined $regexp );

	# $regexp is defined, search throught the list
	my $ra_headers = [];
	foreach my $header ( @a_headers )
	{
		push( @$ra_headers, $header)	if( $header =~ /$regexp/ );
	}

	return $ra_headers;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
