# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

FrameTransformer - a class to define the frame in which an Image object will draw its elements

=head1 SYNOPSIS

FrameTransformer -> New(X0,Y0,X1,Y1);
FrameTransformer -> New(X0,Y0,X1,Y1,x0,y0,x1,y1);

=head1 DESCRIPTION

FrameTransformer is a class for computing absolute positions in a final image from relative coordinates, like nucleic positions in Narcisse.
Moreover this is a virtual representation of the frame in which we will draw particular elements. (like a ChromosomeSlice representation)


=head2 CONSTRUCTOR PARAMETERS

X0,Y0,X1,Y1			=> 	Absolute coordinates of the frame. In pixel
x0,y0,x1,y1			=>	Relative coordinates of the frame. In base pair in Narcisse data.
						If you do not want to convert any coordinate, do not set


=head1 SUBROUTINES

	ToAbsoluteCoordinates
	RelativeCoordinate
	ValideCoordinate
	IsSetAsDefault
	CalculateScaleFactor
	ChangeReferences
	GetAttributes
	GetRelativeBounds
	GetBottomCorner
	GetFactor
	GetTopCorner
	MaskHeight
	MaskWidth

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package FrameTransformer;

use strict;
use warnings;

use Data::Dumper;

use base qw( LipmObject );

use LipmError::ObjectException;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

##################################################
###########			Constructor		 #############
##################################################
=head2  _Init

	Title        : 	_Init
	Usage        : 	_Init(%h_param)
	Prerequisite : 	none
	Procedure    : 	Initialization
	Returns      : 	reference of the blessed object
				   	else
					undef if the object can't be initialized
	Args         :	X0,Y0	=> absolute coordinates of the frame origin (top-left)
					X1,Y1	=> absolute coordinates of the frame end (bottom-right)
					x0,y0	=> relative coordinates of the frame origin (top-left)
					x1,y1	=> relative coordinates of the frame end (bottom-right)
	Globals	  	 : 	object attributes
	Errors		 : 	LipmError::ObjectException

=cut
sub _Init
{
	my $self = shift;

	# 0) Set default attributes to make the object happy
	my %h_param = (
				# Absolute coordinates
				X0	=> 0,
				Y0	=> 0,
				X1	=> undef,
				Y1	=> undef,

				# Relative coordinates
				x0	=> undef,
				y0	=> undef,
				x1	=> undef,
				y1	=> undef,

				@_);

	# 1) Test mandatory parameters (absolute coordinates)
	foreach my $required ( qw( X1 Y1 ) )
	{
		if( !defined( $h_param{ $required } ) )
		{
			my $dbg = "Parameter 'X1' missing";
			throw LipmError::ObjectException( __PACKAGE__, '!build', -text => $dbg );
		}
	}

	# 2) Store object attributes and calculate scale factors
	$self -> SetFrame(%h_param);

	return $self;
}

##################################################
###########			Procedures		 #############
##################################################
=head2  SetFrame
	Title		 :	SetFrame
	Usage		 :	$self -> SetFrame(%h_param)
	Prerequisite :	instanciate the object
	Procedure	 :	Change some or all frame attributes.
	Returns	  	 : 	none
	Args		 :	%h_param (
						X0,Y0	=> absolute coordinates of the frame origin (top-left)
						X1,Y1	=> absolute coordinates of the frame end (bottom-right)
						x0,y0	=> relative coordinates of the frame origin (top-left)
						x1,y1	=> relative coordinates of the frame end (bottom-right)
					)
	Globals	  	 :	none
	Errors		 : LipmError::ObjectException

=cut
sub SetFrame
{
	my $self = shift;
	my %h_param = @_;

	if (scalar (keys(%h_param)) == 0)
	{
		$self->SetAbsoluteFrame();
	}

	#----------------------------
	# Store coordinates.
	my @a_attributes_keys = keys(%h_param);
	@$self{@a_attributes_keys} = @h_param{@a_attributes_keys};

	#----------------------------
	# Test relative coordinates.
	# If one of the relative coordinate axes is not completly defined  =>  we force absolute coordinate for this axes
	if (!defined $h_param{x0} || !defined $h_param{x1})
	{
		($self->{x0}, $self->{x1}) = (0, $self->MaskWidth());

		my $dbg = "Forcing absolute coordinates for the x axis";
		$self->_Trace( $dbg, '.', 'IIV' );
	}
	if (!defined $h_param{y0} || !defined $h_param{y1})
	{
		($self->{y0}, $self->{y1}) = (0, $self->MaskHeight());
		my $dbg = "Forcing absolute coordinates for the y axis";
		$self->_Trace( $dbg, '.', 'IIV' );
	}

	#----------------------------
	# Calculate abscissa and ordinate factors.
	$self->__SetFrameScale();

	return;
}

=head2  TestRelativeCoordinates
	Title		 :	TestRelativeCoordinates
	Usage		 :	$self -> TestRelativeCoordinates()
	Prerequisite :	Absolute coordinates should be defined.
	Procedure	 :	Set relative coordinates which are not defined to absolute coordinates values.
	Returns	  	 : 	none
	Args		 :	none
	Globals	  	 :	none

=cut
sub TestRelativeCoordinates
{
	my $self = shift;

	my @a_relative_keys = ('x0','y0','x1','y1');
	foreach my $key (@a_relative_keys)
	{
		$self->{$key} = $self->{uc($key)} if (!defined $self->{$key});
	}

	return;
}

=head2  __SetFrameScale
	Title		 :	__SetFrameScale
	Usage		 :	$self -> SetFrameScale()
	Prerequisite :	instanciate the object
	Procedure	 :	Check the attributes of the object and calculates the appropriate scale factors  fx et fy
	Returns	  	 : 	none
	Args		 :	none
	Globals	  	 :	none

=cut
sub __SetFrameScale
{
	my $self = shift;

	$self->{fx} = &CalculateScaleFactor($self->{x0}, $self->{x1}, $self->{X0}, $self->{X1});
	$self->{fy} = &CalculateScaleFactor($self->{y0}, $self->{y1}, $self->{Y0}, $self->{Y1});

	$self->{__valid_frame}=1;

	return;
}

##################################################
###########			Functions		 #############
##################################################

=head2 YAxisCoordinateChange
	Title		 :	YAxisCoordinateChange
	Usage		 :	$self -> YAxisCoordinateChange(y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
=cut
sub YAxisCoordinateChange
{
	my $self = shift;
	my $y = shift;

	return int( $self->{Y0} + ($y-$self->{y0})*$self->{fy} );	
}

=head2  PointCoodinateChange
	Title		 :	PointCoodinateChange
	Usage		 :	$self -> PointCoodinateChange(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
=cut
sub PointCoordinateChange
{
	my $self = shift;
	my ($x,$y) = (shift, shift);

	#my $X = int( $self->{X0} + ($x-$self->{x0})*$self->{fx} );
	my $X = $x;
	my $Y = int( $self->{Y0} + ($y-$self->{y0})*$self->{fy} );

	return ($X,$Y);	
}

=head2  RectangleCoordinateChange
	Title		 :	RectangleCoordinateChange
	Usage		 :	$self -> RectangleCoordinateChange($x0,$y0,$x1,$y1)
	Prerequisite :	none
 	Function	 :  calls PointCoodinateChange for each edge of the rectange and ensures in addition the ordering  X0<X1 and Y0<Y1
 	Returns	  	 :	X0,Y0,X1,Y1	(in pixel)
 	Args		 :	x0,y0,x1,y1	(whatever)
=cut
sub RectangleCoordinateChange
{
	my $self = shift;
	my ( $x0, $y0, $x1, $y1 ) = @_;
	
	my ( $X0, $Y0 ) = $self->PointCoordinateChange( $x0, $y0 );
	my ( $X1, $Y1 ) = $self->PointCoordinateChange( $x1, $y1 );

	( $X0, $X1 ) = ( $X1, $X0 ) if ( $X0 > $X1 );
	( $Y0, $Y1 ) = ( $Y1, $Y0 ) if ( $Y0 > $Y1 );

	return ( $X0, $Y0, $X1, $Y1 );	
}

=head2  ToAbsoluteCoordinates
	Title		 :	ToAbsoluteCoordinates
	Usage		 :	$self -> ToAbsoluteCoordinates(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
 	Globals	  	 :	none

 	TODO         : search if it has sens that this function return something
 	               like  $Y < Y0 or Y1 < $Y

 	               because in Image::filledRectangle and Image::verticalCylinderPart
 	               value are corrected so that Y0 <= $Y <= Y1

 	               It seems that a one pixel decalage can appear.

=cut
sub ToAbsoluteCoordinates
{
	my $self = shift;
	my ($x,$y) = (shift, shift);

	my $X = int( $self->{X0} + ($x-$self->{x0})*$self->{fx} );
	my $Y = int( $self->{Y0} + ($y-$self->{y0})*$self->{fy} );

	return ($X,$Y);
}

=head2  ToRelativeCoodinates
	Title		 :	ToRelativeCoodinates
	Usage		 :	$self -> ToRelativeCoodinates(x,y)
	Prerequisite :	none
	Function	 :	calculate the relative coordinate (x,y) of the final point (X,Y) in the actual frame
	Returns	  	 :	x,y	(in base-pair)
	Args		 :	X,Y (in pixel)
	Globals	  	 :	none

=cut
sub ToRelativeCoordinates
{
	my $self = shift;
	my ($X,$Y) = (shift, shift);

	my $x = int ( $self->{x0} + ($X-$self->{X0})/$self->{fx} );
	my $y = int ( $self->{y0} + ($Y-$self->{Y0})/$self->{fy} );

	return ($x,$y);
}

=head2  ToAbsoluteYRange
	Title		 :	ToAbsoluteYRange
	Usage		 :	$self -> ToAbsoluteYRange(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
 	Globals	  	 :	none

=cut
sub ToAbsoluteYRange
{
	my ($self,$range) = (shift, shift) or return 0;

	return $range*$self->{fy};
}

=head2  ToAbsoluteXRange
	Title		 :	ToAbsoluteXRange
	Usage		 :	$self -> ToAbsoluteXRange(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
 	Globals	  	 :	none

=cut
sub ToAbsoluteXRange
{
	my ($self,$range) = (shift, shift) or return 0;

	return $range*$self->{fx};
}

=head2  ToRelativeYRange
	Title		 :	ToRelativeYRange
	Usage		 :	$self -> ToRelativeYRange(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
 	Globals	  	 :	none

=cut
sub ToRelativeYRange
{
	my ($self,$range) = (shift, shift) or return 0;

	if ($self->{fy} == 0)
	{
		return 0;
	}
	else
	{
		return $range/$self->{fy};
	}
}

=head2  ToRelativeXRange
	Title		 :	ToRelativeXRange
	Usage		 :	$self -> ToRelativeXRange(x,y)
	Prerequisite :	none
 	Function	 :  calculate the absolute coordinate (X,Y) of the point (x,y) in the final image
 	Returns	  	 :	X,Y	(in pixel)
 	Args		 :	x,y	(in base-pair)
 	Globals	  	 :	none

=cut
sub ToRelativeXRange
{
	my ($self,$range) = (shift, shift) or return 0;

	if ($self->{fx} == 0)
	{
		return 0;
	}
	else
	{
		return $range/$self->{fx};
	}
}

=head2  CalculateScaleFactor
 	Title        : 	CalculateScaleFactor
 	Usage        : 	&FrameTransformer::CalculateScaleFactor(a,b,A,B)
 	Prerequisite : 	none
 	Function     : 	Calculates the scale factor between an old interval and a new interval
 					Factor unit in Narcisse : px/bp
 	Returns      : 	The scale factor
 	Args         : 	a,b = relative values (in base pair in Narcisse)
	 				A,B = absolute values (in pixel)
 	Globals      : 	none

=cut
sub CalculateScaleFactor
{
	my ($a,$b,$A,$B) = @_;
# 	my ($package, $filename, $line) = caller;

	# Test if parameters are good.
	return undef if (!defined $a or !defined $b or !defined $A or !defined $B);
	#return undef if ( $B-$A < 0 or $b-$a <= 0 );
	
	#this should not happen too often however but we just want to avoid Illegal division by zero errors
	return 0 if ($b == $a);
	
	# Calculate factor between relative and absolute vectors.
	my $factor = ($B-$A) / ($b-$a);

	return $factor;
}

=head2 ReversePositionInFrame
	Title        : 	ReversePositionInFrame
 	Usage        : 	$new_position = $o_frame -> ReversePositionInFrame (axe, position)
 	Prerequisite : 	none
 	Function     : 	Swap a pixel position according to the actual frame bounds.
 	Returns      : 	new_position = the pixel of the revert value
 	Args         : 	axe 	 = axe on which we make this transformation ('x' or 'y')
 					position = the position value to transform.
 	Globals      : 	none
 	Attribute    :	frame

=cut
sub ReversePositionInFrame
{
	my $self = shift;
	my ($axe,$position) = (shift,shift);

	if ($axe eq 'x')
	{
		$position = $self->{x0} + $self->{x1} - $position;
	}
	elsif ($axe eq 'y')
	{
		$position = $self->{y0} + $self->{y1} - $position;
	}

	return $position
}

=head2  GetBounds
	Title        :	GetBounds
	Usage        :	$self->GetBounds()
	Prerequisite :	none
	Function     :	Get the coordinates of the bounds (top-left and bottom-right) of the box
	Returns      : 	Hash-map (X0,Y0,X1,Y1)
	Args         : 	none
	Attributes	 :	X0,Y0,X1,Y1
	Globals      : 	none

=cut
sub GetBounds
{
	my $self = shift;

	my %h_return = ( X0 => $self->{X0},
					 Y0 => $self->{Y0},
					 X1 => $self->{X1},
					 Y1 => $self->{Y1}
					);

	return %h_return;
}

=head2  GetFactor
	Title        : 	GetFactor
 	Usage        : 	GetFactor
 	Prerequisite : 	none
 	Function     : 	Get the factor of the frame
 	Returns      : 	factor (long)

=cut
sub GetFactor
{
	my $self = shift;

	return $self->{fy};
}

=head2  GetFrameOrientation
	Title        : 	GetFrameOrientation
 	Usage        : 	GetFrameOrientation
 	Prerequisite : 	none
 	Function     : 	Get the frame orientation (positive or negative)
 	Returns      : 	factor (long)

=cut
sub GetFrameOrientation
{
	my $self = shift;

	return ( $self->{fy} > 0 ) ? 1 : -1;
}

=head2  GetRelativeBounds
	Title        : 	GetRelativeBounds
 	Usage        : 	GetRelativeBounds
 	Prerequisite : 	none
 	Function     : 	Get the relative minimum of the frame parameters
 	Returns      : 	$self->{y0}
 	Args         : 	none
 	Globals      : 	y0

=cut
sub GetRelativeBounds
{
	my $self = shift;

	return ($self->{x0},$self->{y0},$self->{x1},$self->{y1});
}

=head2  MaskHeight
 	Title		 :	MaskHeight
 	Usage		 :	$self -> MaskHeight()
 	Prerequisite :	instanciate the object
 	Function	 :	Get the height of the frame (pixel)
 	Returns	  	 :	integer
 	Args		 :	none
 	Globals	  	 :	Y0,Y1

=cut
sub MaskHeight
{
	my $self = shift;
	my $Y0 = $self -> {Y0};
	my $Y1 = $self -> {Y1};

	return abs($Y1 - $Y0);
}

=head2  MaskWidth
	Title		 :	MaskWidth
	Usage		 :	$self -> MaskWidth()
	Prerequisite :	instanciate the object
	Function	 : 	Get the width of the frame (pixel)
	Returns	  	 :	integer
	Args		 :	none
	Globals	  	 :	X0,X1

=cut
sub MaskWidth
{
	my $self = shift;
	my $X0 = $self -> {X0};
	my $X1 = $self -> {X1};

	return abs($X1 - $X0);
}

#------------------------------------------------#
#                  Accessors                     #
#------------------------------------------------#

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

