
package GlintFish;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

GlintFish - Selecting primers for region specific probes

=head1  SYNOPSIS

my $o_glintFish = New GlintFish(	    chr          => $chrom, 
										start        => $start, 
										end          => $end,
										species      => $species,
										instanceObs  => $o_instanceObserver,
										-target      => $target_dir,
										-prefix      => $output_prefix,
										-web          => &FALSE,
										-o_logger   => $o_logger,
										param        => $rh_Param
								      );									          
$o_glintFish->Run();

=head1 DESCRIPTION

=cut

use strict;
use warnings;

use Data::Dumper;
use Switch;
use File::Path;
 
use General;
use LipmError::ObjectException;
use Runner;
use ParamParser;

use base qw( LipmObject );

use Cwd;
use File::Temp qw/ tempfile tempdir /;
use Term::ANSIColor;

use Bio::SeqIO;

use Region;
use Probe;

use constant start  => 'start';
use constant end    => 'end';
use constant middle => 'middle';
use constant GLINTFISH_PRIMER3_PRIMER => 'GLINTFISH_PRIMER3_PRIMER';

my $DEBUG = 0;

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New GlintFish(  chr          => $chrom, 
										  start        => $start, 
										  end          => $end,
										  species      => $species,
										  instanceObs  => $o_instanceObserver
										)
 Prerequisite : none
 Fonction     : GlintFish constructor
 Returns      : 
 Args         : 
 Attributes   :  
     
=cut

sub _Init
{
	my $self = shift;
	
	my %h_args = (	chr          => undef,
					start        => undef,
					end          => undef,
					species      => undef,
					instanceObs  => undef,
					-web         => &FALSE,
					-no_clean    => &FALSE,
					-o_logger    => undef,
					-target      => undef,
					-prefix      => 'glintFish',
					-param       => {},
					@_
				 );
	
	$self->SUPER::_Init( @_ );		
	
	#Error hash
	#Errors in arguments settingw will be stored here
	$self -> _Set( '__error_status'  => []);
	
	$self->_InitFromOptions( %h_args );	
	
	$self->_SetGenomeFilePath();
	
	#Setting the parameters from configuration file defaults.cfg
	#replacing the values of those that are specified in $h_args{ -param }
	#primer3 specific parameters are set in __primer3_param
	$self->_SetGlintFishParameters( $h_args{ -param } );	
	
	$self->_SetExecutableLocation( );
	
	#general options
	$self->_Set( '__web', $h_args{ -web } );
	$self->_Set( '__force', $h_args{ -param }{ force } );
	$self->_Set( '__verbose', $h_args{ -param }{ verbose } );
	$self->_Set( '__debug', $h_args{ -param }{ debug } );
	
	$self->_Set( '__query_filename', $self->_SetQueryFileName() );
	$self->_Set( '__output_filename', $self->_SetOutputFileName( $h_args{ -prefix } ) );
	$self->_Set( '__output_prefix',  $h_args{ -prefix } );
	
	#The different arrays
	$self->_Set( '__rao_suitables'  => [],
	             '__rao_probes'     => [],
				 '__rao_validProbes' => [],
				 );
	
	$self->Trace( "Now constructing the GlintFish Object\n" );
					
	return $self;
} # _Init


=head2 Procedure _InitFromOptions

 Usage        : $o_xx->_InitFromOptions( %h_args );
 Prerequisite : none
 Procedure    : Initialize specific object parameters
                arguments. 
 Args         : cf _Init
=cut

sub _InitFromOptions
{
	my $self = shift;
	my %h_args = @_;
	
	LipmError::Function::Assert( %h_args );	
	
	my ( $chr, $start, $end, $species ) = @h_args{ qw( chr start end species ) };
	my ( $o_instanceObserver, $prefix, $rh_param ) = @h_args{ qw( instanceObs prefix mode param ) };
	my ( $no_clean ) =  @h_args{ qw( -no_clean )};
		
	my $o_speciesObserver = $o_instanceObserver->GetSpeciesObserver( $species );
	
	my $target_dir;
	if ( defined $h_args{ -target } )
	{
		$target_dir	= 	$h_args{ -target };
	}
	else
	{
		my $sessdir = $o_instanceObserver->GetSessionDirectory();
		if ( defined $sessdir && $sessdir ne '' ) 
		{	
			$target_dir = $sessdir; 
		}
		else 
		{ 	
			$target_dir = tempdir( 'e-fish-XXXX', CLEANUP => 0 ); 
		}
	}
			
	my $sav_dir = getcwd();	
		
	$self->_Set( '__chr'    => $chr,
				 '__start'  => $start,
				 '__end'    => $end,
				 '__species'=> $species,			
				 '__prefix' => $prefix,
				 '__execution_dir' =>  $sav_dir,
				 '__targetDir' => $target_dir,
				 '__no_cleaning'  => $no_clean,
				 '__o_instanceObserver' => $o_instanceObserver,
				 '__o_speciesObserver'  => $o_speciesObserver
				 );	
				 
	$self -> _CheckRegionParameters();
							
}

=head2 Procedure _SetGenomeFilePath
 Usage        : $o_xx->_GetGenomeFilePath( );
 Function     : Sets the genome index file path and associated sequence name
 Returns      : none
 Args         : none
=cut

sub _SetGenomeFilePath
{
	my $self = shift;
	
	my $o_speciesObserver = $self->_Get( '__o_speciesObserver' );
	
	my $ra_index_file = $o_speciesObserver->GetChromIndexFilepaths( -where => 'INSTANCE' );
	my @a_index_file = @$ra_index_file;
	my $genomeIndexFile  = $a_index_file[0];
	
	$self->_Set( '__genome' => $genomeIndexFile );
}

=head2 Procedure _SetQueryFileName
 Usage        : $o_xx->_SetQueryFileName( );
 Function     : Constructs the name of the query file
 Returns      : $filepath
 Args         : none
=cut
sub _SetQueryFileName
{
	my ( $self ) = @_;
	
	my $sp    = $self->Species();
	my $chr   = $self->GetChr();
	my $start = $self->GetStart();
	my $end   = $self->GetEnd();
	
	$sp=~s/\s/_/g;
	my $query_Filename = "$sp.$chr:$start-$end.fa";
	
	return $query_Filename;
}

=head2 Function _SetOutputFileName
 Usage        : $o_xx->_SetOutputFileName( );
 Function     : Constructs the name of the output filename
 Returns      : the filename
 Args         : none
=cut
sub _SetOutputFileName
{
	my $self = shift;
	my $prefix = shift;
	return $prefix.".probes";
}

=head2 Function _SetGlintFishParameters
 Usage        : $o_xx->_SetGlintFishParameters( $rh_params )
 Function     : Parse and rename the GlintFish specific parameters obtained from the cfg file
 Returns      : none
 Args         : $rh_params 
=cut

sub _SetGlintFishParameters
{
	my $self = shift;
	my ( $rh_param ) = @_;
	
	my $o_instanceObserver = $self->_Get( '__o_instanceObserver' );			
	my $rh_cfg = $o_instanceObserver->GetConfigHash();
	my $errors = "";
	#Getting data with "GLINTFISH" tag from defaults.cfg 	
	my @glintfish_keys = grep $_ =~ /^GLINTFISH/, keys %$rh_cfg;
	
	#Looking at those for PRIMER3 and setting values
	foreach my $glintfish_key ( @glintfish_keys )
	{
		next if ( $glintfish_key =~ /PRIMER3/ );
		my $key_name = $glintfish_key;
		$key_name =~ s/GLINTFISH\_//;
		$key_name = "__\L$key_name";
		my $key_value = $o_instanceObserver->GetCfg( $glintfish_key );
		if ( exists $rh_param->{ $glintfish_key } )
		{
			#Check max GC percent before update
			my $newGC = $rh_param->{ $glintfish_key };	
			my $switch_key = $self -> _CheckGCparameter($newGC);
			if ($switch_key == 1)
			{
				$key_value = $rh_param->{ $glintfish_key };
			}
		}
		$self->_Set( $key_name => $key_value );
	}
	
	my @primer_keys = grep $_ =~ /^GLINTFISH_PRIMER3/, keys %$rh_cfg;

	#Setting a hash with primer3 parameters obtained from %rh_cfg
	
	my %h_primer3_param = ();	
	foreach my $primer3_key ( @primer_keys )
	{
		my $key_name = $primer3_key;
		$key_name =~ s/GLINTFISH_PRIMER3_//;
		
		my $key_value = $o_instanceObserver->GetCfg( $primer3_key );
		if ( exists $rh_param->{ $primer3_key} )
		{
			#Check parameters before update
			my $newkey_value = $rh_param->{ $primer3_key };
			my $param_ok = $self -> _CheckPrimer3Parameters($newkey_value, $primer3_key);
			if ($param_ok == 1)
			{	
				$key_value = $rh_param->{ $primer3_key };
			}
		}
		$h_primer3_param{  $key_name  } = $key_value;		
	}
		
	$self->_Set( '__primer3_param', \%h_primer3_param );		
}

=head2 _CheckPrimer3Parameters

 Title     : _CheckPrimer3Parameters
 Usage     : $self->_CheckPrimer3Parameters();
 Function  : Check parameters consistency for primer 3 execution
			 Must be used before updating default parameters
=cut

sub _CheckPrimer3Parameters
{
	my $self = shift;
	
	my ( $newvalue, $primer3_key) = (@_);
	my $param_OK = 1;
	my $o_instanceObserver = $self->_Get( '__o_instanceObserver' );	

	unless ($newvalue)
	{
		$self -> _PushArgError({field => 'Probe size or Primer size or Tm value', message => "Empty field"});
		$param_OK = 0;
		return $param_OK;
	}
	#~#Check Probe size	
	if ($primer3_key =~ /.*PRODUCT_SIZE_RANGE.*/)
	{	
		my $probe_dft_range = $o_instanceObserver-> GetCfg(&GLINTFISH_PRIMER3_PRIMER.'_PRODUCT_SIZE_RANGE');
		my ($dft_base, $dft_top) = split(/-/, $probe_dft_range);
		my ($base, $top) = split(/-/, $newvalue);
		
		if (($top < $base)||($base < $dft_base) || ($top > $dft_top))
		{
			$self -> _PushArgError({field => 'Probe size', message => "It should be in default range: $probe_dft_range"});
			$param_OK = 0;		
		}	
	}
	
	#~#Check Primer size
	if ($primer3_key =~ /.*PRIMER_OPT_SIZE.*/)
	{
		my $primer_min_size = $o_instanceObserver-> GetCfg (&GLINTFISH_PRIMER3_PRIMER.'_MIN_SIZE');
		my $primer_max_size = $o_instanceObserver-> GetCfg (&GLINTFISH_PRIMER3_PRIMER.'_MAX_SIZE');
		my $primer_dft_range = $primer_min_size."-".$primer_max_size;
		
		if (($newvalue > $primer_max_size) || ($newvalue < $primer_min_size))
		{
			$self -> _PushArgError({field => 'Primer size', message => "It should be in default range: $primer_dft_range"});
			$param_OK = 0;	
		}	
	}
	
	#~#Check Tm value
	if ($primer3_key =~ /.*OPT_TM.*/)
	{
		my $min_Tm_value = $o_instanceObserver-> GetCfg(&GLINTFISH_PRIMER3_PRIMER.'_MIN_TM');
		my $max_Tm_value = $o_instanceObserver-> GetCfg (&GLINTFISH_PRIMER3_PRIMER.'_MAX_TM'); 
		my $Tm_dft_range = $min_Tm_value."-".$max_Tm_value ;
		
		if (($newvalue > $max_Tm_value) || ($newvalue < $min_Tm_value))
		{
			$self -> _PushArgError({field => 'Tm value', message => "It should be in default range: $Tm_dft_range"});
			$param_OK = 0;	
		}
	}
	return $param_OK;
}

=head2 _CheckRegionParameters

 Title     : _CheckRegionParameters
 Usage     : $self->_CheckRegionParameters;
 Function  : Check parameters consistency for primer 3 execution in GlintFish
			 
=cut

sub _CheckRegionParameters
{
	my $self = shift;
	#Here parameters have to be set in the Init
	my ( $chrom, $start, $end ) = $self->_Get( qw( __chr __start __end ) );
	my $o_speciesObserver = $self->_Get( '__o_speciesObserver' );
	my $ra_chr_list = $o_speciesObserver -> GetChromsNames();
	
	unless ( $start && $end )
	{
		$self -> _PushArgError({field => 'Start or End', message => "Empty field"});
	}
	
	#~#Check Start and End
	if ($start > $end)
	{
		$self -> _PushArgError({field => 'Start or End', message => "Incompatible"});
	}
	
	#Check chromosome id
	if ($chrom ne "")
	{
		my @chr_ok = grep $_ =~ /$chrom/, @$ra_chr_list;
		
		my $dft_list = join(",", @$ra_chr_list);
		
		if (scalar @chr_ok == 0)
		{
			$self -> _PushArgError({field => 'Chromosome name', message => "Available chromosomes are: $dft_list."});
		}
	}
	else
	{
		$self -> _PushArgError({field => 'Chromosome name', message => "Empty field"});
	}
}

=head2 _CheckGCparameters

 Title     : _CheckGCparameter
 Usage     : $self->_CheckGCparameter;
 Function  : Check GC percent for probe obtained by primer3
			 
=cut

sub _CheckGCparameter
{
	my $self = shift;
	#Here parameters have to be set in the Init
	my ($newGC) = (@_);
	my $bool_gc = 1;
	unless ($newGC)
	{
		$self -> _PushArgError({field => 'Max Probe GC percent', message => "Empty field"});
		$bool_gc = 0;
	}
	if ($newGC > 1)
	{
		$self -> _PushArgError({field => 'Max Probe GC percent', message => "Value can't be higher than 1"});
		$bool_gc = 0;
	}
	return $bool_gc;
}
	
=head2 Function _PushArgError
 Usage        : $self-> _PushArgError
 Function     : Fill an array reference with an error when parameters are note correctly set
 Returns      : none
 Args         : $rh_error. {field => 'start', $message => "Empty field"}
=cut

sub _PushArgError
{
	my $self = shift;
	my ($rh_error) = (@_);

	my $ra_arg_errors = $self -> _Get('__error_status');
	push (@$ra_arg_errors, $rh_error);
	
}

=head2 Function DisplayErrors
 Usage        : $o_glintFish-> DisplayErrors
 Function     : Displays the content of the glintfish field '__error_status'
 Returns      : displays error messages
 Args         : none
=cut

sub DisplayErrors
{
	my $self = shift;

	my $ra_arg_errors = $self -> _Get('__error_status');
	my $html_errors = "<br>";
	if (scalar @$ra_arg_errors > 0)
	{
		foreach my $rh_error (@$ra_arg_errors)
		{
			$html_errors .= "<b>WARNING:</b> $rh_error->{field} - <b>ERROR MESSAGE:</b> $rh_error->{message}</br>";
		}
		return $html_errors;
	}
	else
	{
		return "000";
	}
}


=head2 Function _SetExecutableLocation
 Usage        : $o_xx->_SetExecutableLocation(  )
 Function     : Get the executable pathsfrom $o_instanceObserver
 Returns      : none
 Args         : none 
=cut

sub _SetExecutableLocation
{
	my $self = shift;
	my $o_instanceObserver = $self->_Get( '__o_instanceObserver' );	
	
	$self->_Set( '__glintFreq_exe', $o_instanceObserver->GetCfg( 'ROOT_GLINTFREQ'  ) );
	$self->_Set( '__glint_exe'    , $o_instanceObserver->GetCfg( 'ROOT_GLINT_PURE' ) );
	$self->_Set( '__primer3_exe'  , $o_instanceObserver->GetCfg( 'ROOT_GLINTFISH_PRIMER3'    ) );
	$self->_Set( '__pairEnd_exe'  , $o_instanceObserver->GetCfg( 'ROOT_GLINTFISH_GLINTREPCR' ) );	
}

=head2 Procedure Run

 Usage        : $o_xx->Run(  )
 Function     : Run the complete process of selecting the probe primers
 Returns      : none
 Args         : none 
=cut

sub Run
{
	my $self = shift;
	
	$self->_SetWorkingDir();
		
	$self->Trace("1. Extracting the sequence\n");
	$self->ExtractRegionSequence();
		
	$self->Trace("2. Detecting suitable subregions\n");
	$self->DetectSuitableSubregion() ;
		
	$self->Trace("3. Detecting probes\n");
	$self->DetectAllProbes() ;

	$self->Trace("4. Probes validation\n");	
	$self->ProbesValidation();
						
	$self->_LeavingWorkingDir();		
	$self->_CleaningWorkingDir();

}

=head2 function ExtractRegionSequence

 Title    : ExtractRegionSequence
 Usage    : Extract a region sequence and create the query file in the target directory
 Access   : Public
 Function : Extract a region sequence and create the query file in the target directory
 Args     : $chromosome_file => the chromosome file '.fa'
            $rh_region       => the hash ( chr, start, end ) used to create the file
            $target_dir      => the target directory used 
 Returns  : the output file name
 Globals  : none
=cut

sub ExtractRegionSequence
{
	my ( $self ) = @_;

	my ( $chrom, $start, $end ) = $self->_Get( qw( __chr __start __end ) );
	
	my $o_speciesObserver = $self->_Get( '__o_speciesObserver' );
	#my $chromosome_file = $o_speciesObserver->GetChromFastaFile( chrom => $chrom ); 
	my $chromosome_file = $o_speciesObserver->GetChromFastaFile( chrom => $chrom, -where => 'INSTANCE' ); 
	my $force = $self->_Get( '__force' );
				
	my $output_file_path = $self->_Get( '__query_filename' );

	if ( !( -e $output_file_path ) || ( defined $force ) )
	{
		my $seq_in        = Bio::SeqIO->new( -file => "<$chromosome_file", -format => 'fasta' );
		my $seq_obj       = $seq_in->next_seq();
		my $seq_region    = $seq_obj-> subseq( $start, $end );
		my $seq_out       = Bio::SeqIO->new( -file => ">$output_file_path", -format => 'fasta' );
		my $seqregion_obj = Bio::Seq->new( -seq => $seq_region, -display_id => "region_$chrom\_$start-$end" );
		$seq_out->write_seq( $seqregion_obj );
	}
}

=head2 function DetectSuitableSubregion

 Title        : DetectSuitableSubregion
 Usage        : $region->DetectSuitableSubregion( $bank, $queryFile );
 Function     : Run glintfreq program and call LoadSuitable
 Returns      : none
 Args         : none
=cut
sub DetectSuitableSubregion
{
	my $self = shift;
		
	$self->LaunchGlintFreq();
	$self->LoadSuitable();
	
	$self->Trace( "\t".$self->NbSuitables()." suitable subregions\n" );
	
	if ( $DEBUG )
	{
		$self->SuitableRegionsLog();
	}
}

=head2 function GetGlintFreqParameters

 Title        : GetGlintFreqParameters
 Usage        : $glintParam = $self->GetGlintFreqParameters;
 Function     : Return a string wth all the glint param from ParamParser
 Access       : Public
 Returns      : $glint param
 Args         : none
=cut
sub GetGlintFreqParameters
{
	my $self = shift;
		
	my @a_parameters = qw( count_cutoff conseq_freq conseq_unfreq condemned_min_size suitable_min_size suitable_max_size suitable_chunk_size );
	
	my $glintParam ="-w ".$self->_Get( '__w' );
	foreach my $param_key ( @a_parameters )
	{
		my $param_value = $self->_Get( '__'.$param_key );
		$glintParam.=" --$param_key $param_value";	
	}		
	return $glintParam;
}

=head2 function LoadSuitable

 Title        : LoadSuitable
 Usage        : $region->LoadSuitable( $glintfreqFile );
 Function     : Parses the glintfreq output      
                contruct an object region for each suitable
                and push in the suitable region array
 Returns      : none
 Args         : none 
=cut
sub LoadSuitable
{
	my $self             = shift;
	my $region_start     = $self->GetStart();

	my $glintFreqOutputFile = &GetTempFileName( 'glintfreq', 'out' ) ;
	
	my $fh = new IO::File( $glintFreqOutputFile )
		or die "Impossible to open file glintFreqOutputFile for reading";
	
	my $id = undef;
	my ( $header, $rel_start, $rel_end, $sequence ) ;
	while ( my $line = <$fh> )
	{
		next if ( $line =~ /^\s*$/ );
		if ( $line =~ /^>/ )
		{
			if ( defined $id )
			{			
				my $o_suitable = $self->__NewSuitable( $id, $rel_start, $rel_end, \$sequence );
				#print STDERR "$id\t".$self->_Get( '__chr' )."\t".join("\t",($rel_start,$rel_end,$abs_start,$abs_end))."\t".substr( $sequence, 0, 20 )."\n";
				$self->PushSuitable( $o_suitable ) ;	
			}
			( $id, $header, $rel_start, $rel_end ) = ( $line =~  /^>(\d+)\|(\S+)\tstart : (\d+)\tend : (\d+)/ );			
			$sequence = "";
		}
		else
		{
			chomp( $line );
			$sequence.= $line;
		}				
	}	
	#Adding the last suitable region
	my $o_suitable = $self->__NewSuitable( $id, $rel_start, $rel_end, \$sequence );
	$self->PushSuitable( $o_suitable ) ;
}


sub __NewSuitable
{	
	my $self = shift;
	my ( $id, $rel_start, $rel_end, $r_sequence ) = @_;
	
	my $abs_start = $self->_Get( '__start' ) + $rel_start - 1;
	my $abs_end   = $self->_Get( '__start' ) + $rel_end - 1;
	my $o_suitable = New Region( id    => $id,
								 chr   => $self->_Get( '__chr' ),
							     start => $abs_start,
							     end   => $abs_end,
							     seq   => $$r_sequence,
							     o_speciesObserver => $self->_Get( '__o_speciesObserver' )
							    );
	return $o_suitable;
}


=head2 function DetectAllProbes
 Title        : DetectAllProbes
 Usage        : $region->DetectAllProbes( $primerInputFile );
 Function     : Run Primer3 program foreach suitable region
                Parses the output file of Primer3
                and push all the probes in the probe array of the region object
 Returns      : none
 Args         : none
=cut
sub DetectAllProbes
{
	my $self = shift;
	
	my $primer3_exe    = $self->_Get( '__primer3_exe'    );
	my $query_Filename = $self->_Get( '__query_filename' );
	
	my $rao_suitables = $self->_Get( '__rao_suitables' );
	
	my $min_size_4primer3 = $self->_Get( '__suitable_primers_min_size' );
	
	#$self->DebugTrace( "$primer3_exe -p3_settings_file primerParamFile <primerInputFile -output primerOutputFile\n" );
		
	foreach my $o_suitable ( @$rao_suitables )
	{
		if ( $o_suitable->GetSize() > $min_size_4primer3  )
		{ 			
			my $primer3_intput = $self->GeneratePrimer3InputFile( $o_suitable );		
			my $primer3_output = $self->LaunchPrimer3( $primer3_intput );
			my $ra_probes = &ParsePrimer3Output( $primer3_output );
			$self->LoadProbes( $o_suitable, $ra_probes );
		}
	}	
	$self->Trace("\t".$self->GetNbProbes()." candidate probes\n");
}

=head2 function ParsePrimer3Output

 Title        : ParsePrimer3Output
 Usage        : my $ra_probes = &ParsePrimer3Output( $primer3_output )
 Function     : parse the primer3 output and return a ref on an array with the associated probes
 Returns      : a reference on an array with the probe infos
 Args         : the primer3 output file
 
=cut
sub ParsePrimer3Output
{
	my $primer3_output_file = shift;

	my $fh = new IO::File( $primer3_output_file )
		or die "Immpossible to open file $primer3_output_file for reading";
	
	my @a_probes = ();
	#Get all the lines of the file
	my @lines = <$fh>;
	foreach my $line ( @lines )
	{
		chomp( $line );
		&ParseLine(  $line, \@a_probes );
	}	
		 	
	return \@a_probes;
} # LoadProbe

sub ParseLine
{
	my ( $line, $ra_probes ) = @_;	
	
	if ( $line =~ /^PRIMER_(\w+)\_(\d+)=([\w\.\,]+)$/ )
	{
		my ( $start, $length ) = split(",",$3);
		$ra_probes->[$2]{ $1."_START" } = $start;
		$ra_probes->[$2]{ $1."_LENGTH" } = $length;	
		return ;
	}
	if ( $line =~ /^PRIMER_(\w+)\_(\d+)\_(\w+)=([\w\.]+)$/ )
	{
		$ra_probes->[$2]{ $1."_".$3 } = $4;
		return ;
	}	
}

=head2 function LoadProbe

 Title        : LoadProbes
 Usage        : $subregion->LoadProbe( $o_suitable, $rao_probes );
 Function     : For each probe construct a probe object and push it in the suitable region probe array
 Returns      : none
 Args         : none
=cut
sub LoadProbes
{
	my $self = shift;
	my ( $o_suitable, $ra_probes ) = @_ ;
	
	foreach my $rh_probe ( @$ra_probes )
	{
		my $o_probe = New Probe( o_region => $o_suitable, rh_primer3 => $rh_probe );
		$o_suitable->PushProbe( $o_probe );
	}		
} # LoadProbe

=head2 function GetNbProbes

 Title        : 
 Usage        : $o_xxx->GetNbProbes(  ); 
 Returns      : The total number of probes
 Args         : none
=cut
sub GetNbProbes
{
	my $self = shift;	
	
	my $rao_probes = $self->_Get( '__rao_suitables' );	
	my $nb_probes = 0;
	foreach my $o_suitable ( @$rao_probes )
	{
		$nb_probes+= $o_suitable->GetNbProbes();
	}
	return $nb_probes;	
}

=head2 function GeneratePrimer3InputFile

 Title        : GeneratePrimer3InputFile
 Usage        : $region->GeneratePrimer3InputFile();
 Function     : Create the Primer3 input file using the region object
 Access       : Public
 Returns      : the primer3 input filename
 Args         : a region object
=cut

sub GeneratePrimer3InputFile
{
	my $self = shift;
	my $o_region = shift;
	
	my $input_Filename = &GetTempFileName( 'Primer3', 'in' );
	
	my $rh_primer3_param = $self->_Get( '__primer3_param' );
	
	my $start = 1;
	my $end = $o_region->GetSize()-1;
	my $seq = $o_region->GetSequence();
			
	open( FILE, ">$input_Filename" ) || die "Erreur E/S:$!\n";
		
	print FILE "SEQUENCE_ID=Region |\tstart : $start\tend : $end\n";
	print FILE "SEQUENCE_TEMPLATE=$seq\n";
	while ( my ( $key, $value ) = each %$rh_primer3_param )
	{
	  print FILE "$key=$rh_primer3_param->{$key}\n";
	}
	print FILE '=';	
	close ( FILE );
	
	return $input_Filename;
}

=head2 Procedure ProbesValidation

 Title        : ProbesValidation
 Usage        : $region->ProbesValidation();
 Function     : Check the proposed probes :
                  - Do they amplify specifically the correct region (RePCRProbes)
                  - Do the amplified product hybridize specifically to the correct region (CheckPcrProduct)
                  - Make sure the GC percent is OK (CheckGCPercent)
                  - Identify the representative probes (a subset of non-overlapping probes)
 Returns      : none
 Args         : none
=cut
sub ProbesValidation
{
	my $self = shift ;
		
	$self->RePCRProbes();	
	$self->DetectRepresentativeProbes();	
	$self->CheckPcrProduct();
	$self->CheckGCPercent();
	
	if ( $DEBUG )
	{
		$self->SelectedProbesLog();	
	}

}

=head2 function RePCRProbes

 Title        : RePCRProbes
 Usage        : $o_xxx->RePCRProbes();
 Function     : Do they amplify specifically the correct region (glint mappe eprc and check)
 Returns      : none
 Args         : none
=cut
sub RePCRProbes
{
	my $self = shift;
	
	$self->Trace("\trePCR\n");
	
	my $rao_allprobes = $self->GetAllValidProbes();
		
	my ( $fwd, $bwd ) = $self->GenerateGlintPairEndInputFile( $rao_allprobes );			
	my $pairend_out   = $self->LaunchGlintPairEnd( $fwd, $bwd );	
				
	my $rh_output = $self->ReadGlintRePCROutput( $pairend_out );
	
	my $num_probes = scalar keys %$rh_output;
	
	my $nb_valid = 0;
	my $num = 0;
	foreach my $o_probe ( @$rao_allprobes )
	{
		my $id = $num + 1;
		my $o_probe = $rao_allprobes->[ $num++ ];
		if (!exists $rh_output->{ $id } )
		{
			$o_probe->UnValid();
			#print "Unvalid  (no hit) --------> ".$o_probe->ChrInfos()."\n";
			next;		
		}
		if ( scalar @{$rh_output->{ $id }} > 1 )
		{
			$o_probe->UnValid();
			#print "Unvalid  (multiple hit)  --------> ".$o_probe->ChrInfos()."\n";
			next;	
		}	
		my $r_pehsp =  $rh_output->{ $id }[0];
		my $o_region = $o_probe->GetRegion();	
		#Checking the location of the amplified fragment
		if ( 
			$r_pehsp->{ chr } ne $o_probe->GetChr() ||
			$r_pehsp->{ start } != $o_probe->GetChrStart() || 
			$r_pehsp->{ end } != $o_probe->GetChrEnd() 
			)
		{
			$o_probe->UnValid();
			#print "Unvalid (uncorrect location)  --------> ".$o_probe->ChrInfos()."\n";
			next;		
		}
		#print "Valid ".$id." : ".$o_probe->ChrInfos()."\n";	
		$nb_valid++;		
	}
	$self->Trace("\t\t".$nb_valid." remaining probes \n");							
}

=head2 function ReadGlintRePCROutput

 Title        : ReadGlintRePCROutput
 Usage        : $o_xxx-ReadGlintRePCROutput( $glint-mappe-epcr-output );
 Function     : Parse the glint mappe epcr output and return the information in a hash
 Returns      : none
 Args         : none
=cut
sub ReadGlintRePCROutput
{
	my $self = shift;
	my $file = shift;
		
	#WARNING, we assume here that each primer size == PRIMER_OPT_SIZE, hence perfect_score = 2 * PRIMER_OPT_SIZE
	my $rh_primer3_param = $self->_Get( '__primer3_param' );
	my $rePCR_perfect_score = 2 * $rh_primer3_param->{ 'PRIMER_OPT_SIZE' };
		
	my $o_speciesObserver = $self->_Get( '__o_speciesObserver' );
	my $species =  $o_speciesObserver->GetShortName();
	
	my $fh = new IO::File ( $file, "r" )
		or die "Impossible to open $file for reading";
	my @a_lines = <$fh>;
	my %h_data = ();
	my @a_fields = qw( fwd_chr fwd_start fwd_end bwd_chr bwd_start bwd_end id score fwd_strand bwd_strand );
	foreach my $line ( @a_lines )
	{
		chomp( $line );
		my %h_line = ();
		@h_line{ @a_fields } = split( /\t/, $line );
		_SetPeHspLocus( \%h_line, $species );	
		#Only perfect matches are kept
		next unless ( $h_line{ score } == $rePCR_perfect_score );	
		my $id = $h_line{ id  };	
		push @{$h_data{ $id }}, \%h_line; 	
	}
		
	return \%h_data;
	
	#Setting the start and end of the pe-hsp (warning bedpe is 0-based for start only : hence +1 sor start)
	sub _SetPeHspLocus
	{
		my ( $rh_line, $species ) = @_;
		
		my ($start, $end );
		if (  $rh_line->{ bwd_end } > $rh_line->{ fwd_start } )
		{
			$start = $rh_line->{ fwd_start } + 1 ;
			$end = $rh_line->{ bwd_end } ;
		}
		else
		{
			$start = $rh_line->{ bwd_start } + 1 ;
			$end = $rh_line->{ fwd_end } ;	
		}
		$rh_line->{ start } = $start;
		$rh_line->{ end }   = $end;
		$rh_line->{ chr } = $rh_line->{ fwd_chr };
		$rh_line->{ chr } =~ s/^$species//;		
	}
}

=head2 function GenerateGlintPairEndInputFile

 Title        : GenerateGlintPairEndInputFile
 Usage        : $o_xxx->GenerateGlintPairEndInputFile( $rao_probes );
 Function     : Generate the glint mappe epcr input files
 Returns      : fwd and bwd primers in a fasta file
 Args         : a reference on an array of probes
=cut
sub GenerateGlintPairEndInputFile
{
	my $self = shift;	
	my $rao_allprobes = shift;
	
	my $fwd_filename = &GetTempFileName( 'fwd_pairEnd', 'in' );
	my $bwd_filename = &GetTempFileName( 'bwd_pairEnd', 'in' );
	
	#write the inputFile with tags and values
	open( fwd_FILE, ">$fwd_filename" ) || die "Erreur E/S:$!\n";
	open( bwd_FILE, ">$bwd_filename" ) || die "Erreur E/S:$!\n";
	
	my $num = 0;
	foreach my $o_probe ( @$rao_allprobes )
	{
		$num++;
		my $leftPrimer  = $o_probe->GetLeftPrimer();
		my $rightPrimer = $o_probe->GetRightPrimer();
		print fwd_FILE ">$num\n$leftPrimer\n";
		print bwd_FILE ">$num\n$rightPrimer\n";	
	}
	return ( $fwd_filename, $bwd_filename );	
}

=head2 Procedure CheckPcrProduct

 Title        : CheckPcrProduct
 Usage        : $o_xxx->CheckPcrProduct();
 Function     : Do the amplified product hybridize specifically to the correct region (glint and check)
 Returns      : none
 Args         : none
=cut
sub CheckPcrProduct
{
	my $self = shift;	
	
	my $rao_allprobes = $self->GetAllSelectedProbes();
	
	$self->Trace("\tChecking PCR products\n");
		
	my $sequence_file = $self->GenerateGlintInputFile( $rao_allprobes );	
	my $glint_output = $self->LaunchGlint( $sequence_file );
	
	my $rh_output = $self->ReadGlintOutput( $glint_output );
		
	my $nb_valid = 0;
	foreach my $id ( keys %$rh_output )
	{
		my $num = $id -1;
		my $o_probe = $rao_allprobes->[ $num ];
		my @a_hits = @{	$rh_output->{$id} };
		my $opt_hit = $a_hits[0];
		$o_probe->SetNumHits( scalar @a_hits );
		#Checking the location of the best hit	
		if ( $o_probe->OutOfPlaceHit( $opt_hit ) )
		{
			$o_probe->DeSelect();
			my $infos = $o_probe->ChrInfos();
			my $chrom = $opt_hit->{ chr };
			my $species = $self->_Get( '__species' );
			$self->DebugTrace("Unvalid (location) $infos $chrom $species \n");
			
			#print STDERR "Unvalid (location)  --------> ".$o_probe->ChrInfos()." vs ".$opt_hit->{ start }."-".$opt_hit->{ end }."\n";
			next ;		
		}		
		#Checking the total number of large hits
		if ( GetNbLargeHits( $rh_output->{$id} ) > 10 )
		{
			$o_probe->DeSelect();
			#$self->Trace("Unvalid (large nb hits) \n");
			#print STDERR "Unvalid (large nb hits)  --------> ".$o_probe->ChrInfos()."\n";
			next ;	
		}
		#Identifying potential high scoring far away suboptimal hits
		if ( scalar @a_hits > 1 )
		{			
			my $sub_opt_hit = $a_hits[1];
			my $o_region = $o_probe->GetRegion();
			if ( $sub_opt_hit->{ size } > 2_000 && $sub_opt_hit->{ score } > 1_500 && $o_region->FarAway( $sub_opt_hit ) )
			{
				$o_probe->DeSelect();
				#print STDERR "Unvalid (High sub optimal hit)  --------> ".$o_probe->ChrInfos()."\n";
				next ;	
			}	
		}
		$nb_valid++;	
	}	
	$self->Trace("\t\t".$nb_valid." remaining probes \n");		
}


=head2 Function GetNbLargeHits

 Title        : GetNbLargeHits
 Usage        : &GetNbLargeHits( $ra_hits );
 Function     : Nb of hsp of size larger than 500
 Returns      : number of such hits
 Args         : an array with all thehsp correcponding to a specific query probe
=cut
sub GetNbLargeHits
{
	my $ra_hits = shift;
	
	my $nb_hits = 0;
	foreach my $r_hit ( @$ra_hits )
	{
		$nb_hits++ if ( $r_hit->{ size } > 500 );	
	}
	
	return $nb_hits;
}

sub GenerateGlintInputFile
{
	my $self = shift; 	
	my $rao_probes = shift;
	
	my $glint_in = &GetTempFileName( 'glint', 'in' );	
	#my $seq_out    = Bio::SeqIO->new( -file => ">$glint_in", -format => 'fasta' );
	my $seq_out_fh = new IO::File( $glint_in, "w" );
	my $num = 0;
	foreach my $o_probe ( @$rao_probes )
	{
		$num++;
		my $seq_region   = $o_probe->GetSequence();
		my $description = $o_probe->GetChrStart()."-".$o_probe->GetChrEnd();
		print $seq_out_fh ">$num $description\n";
		print $seq_out_fh Sequence_utils::FormatSequenceLine( seq => $seq_region );
		#my $seqregion_obj = Bio::Seq->new( -seq => $seq_region, -display_id => "$num", -desc => $description );
		#$seq_out->write_seq( $seqregion_obj );	
	}
	return $glint_in;
}

sub ReadGlintOutput
{
	my $self = shift;
	my $file = shift;
		
	my $species = $self->_Get( '__species' );
	my $fh = new IO::File ( $file, "r" )
		or die "Impossible to open $file for reading";
	my $header= <$fh>;
	chomp( $header );
	$header=~s/^#//;
	my @a_fields = split(/\t/,$header);
	my %h_data = ();
	while ( my $line = <$fh> )
	{
		chomp( $line );
		my %h_line = ();
		@h_line{ @a_fields } = split( /\t/, $line );	
		_SetHspLocus( \%h_line, $species );
		my $id = $h_line{ q_name  };
		push @{$h_data{ $id }}, \%h_line; 	
	}
	
	return \%h_data;
	
	sub _SetHspLocus 
	{
		my ( $rh_line, $species ) = @_;
		
		my ( $chrom, $start, $end ) = @{ $rh_line }{qw( s_name s_start s_end )};  
		$rh_line->{ start } = $start;
		$rh_line->{ end }   = $end;
		$rh_line->{ chr } = $chrom;
		$rh_line->{ chr } =~ s/^$species//;					
	}
}

=head2 CheckGCPercent

 Title        : ChechGCPercent
 Usage        : $o_xxx->ChechGCPercent( )
 Prerequisite : none
 Function     : Check the GC percent of all the valid pcr product
 Returns      : void
 Args         : none
=cut

sub CheckGCPercent
{
	my $self = shift;
	
	$self->Trace("\tChecking GC content\n");
	
	my $rao_allprobes = $self->GetAllSelectedProbes();		
	my $nb_valid = 0;
	my $probe_gc_max = $self ->_Get ('__probe_max_gc_percent');
	
	foreach my $o_probe ( @$rao_allprobes )
	{
		if ( $o_probe->GetAmpliconGC() > $probe_gc_max )
		{
			$o_probe->DeSelect();
			my $probe_id = $o_probe->ProbeId();
			my $gcpercent = sprintf("%2.1f",$o_probe->GetAmpliconGC());
			$self->DebugTrace("$probe_id not selected based on gc% : $gcpercent\n");	
			next;	
		}
		
		$nb_valid++;
	}	
	$self->Trace("\t\t".$nb_valid." remaining probes \n");		
}


=head2 Procedure DetectRepresentativeProbes

 Title        : DetectRepresentativeProbes
 Usage        : $o_xxx->DetectRepresentativeProbes();
 Function     : For each region detect a set of non-overlapping valid probes (see DetectRepresentativeProbes)
 Returns      : none
 Args         : none
=cut
sub DetectRepresentativeProbes
{
	my $self = shift;	
	
	$self->Trace("\tSelecting representative probes\n");
	
	my $rao_suitables = $self->_Get( '__rao_suitables' );
	my $nb=0;
	foreach my $o_suitable ( @$rao_suitables )
	{
		$nb+=$o_suitable->DetectRepresentativeProbes();		
	}	
	$self->Trace("\t\t".$nb." remaining probes\n");
	
}

=head2 Function _GetAllProbes

 Title        : _GetAllProbes
 Usage        : $o_xxx->_GetAllProbes();
 Function     : Identify the set of all probes according to a selection criteria
 Returns      : an array of probes 
 Args         : none
=cut
sub _GetAllProbes
{
	my $self = shift;
	my %h_args = ( filter => 'none', 
	               @_ 
	              );
	my $rao_suitables = $self->_Get( '__rao_suitables' );
	my @a_allprobes = ();
	foreach my $o_suitable ( @$rao_suitables )
	{	
		my $rao_validprobes = $o_suitable->GetAllProbes( filter => $h_args{ filter } );	
         push @a_allprobes, @$rao_validprobes;
	}
	return \@a_allprobes;
}

sub GetAllValidProbes {
	my $self = shift;	
	return $self->_GetAllProbes( filter => 'valid' );
}	

sub GetAllSelectedProbes {
	my $self = shift;
	return $self->_GetAllProbes( filter => 'selected' );
}	

=head2 Function ProbeOutput

 Title        : ProbeOutput
 Usage        : $o_xxx->ProbeOutput();
 Function     : Dumpt the selectd probes in the outputfile
 Returns      : an array of probes 
 Args         : none
=cut
sub ProbeOutput
{
	my $self = shift;	
		
	my $output = $self->_Get( '__output_filename');				
	my $ra_data = $self->GetSelectedProbesArray();
	$ra_data->[0]=~ s/^#//;
	
	my $fh = new IO::File( 	$output, "w" )
		or die "Impossible to open $output for writing";
	print $fh join("", @$ra_data);
	$fh->close();
	my $num = (scalar @$ra_data)-1;
	$self->Trace("results: $num selected probes in $output\n");		
}

=head2 Function FormatProbeOutputForNarcisse

 Title        : FormatProbeOutputForNarcisse
 Usage        : $o_xxx->FormatProbeOutputForNarcisse();
 Function     : Format the output in a Narcisse format 
 Returns      : a ref on a hash of probes with info
 Args         : none
=cut
sub FormatProbeOutputForNarcisse
{
	my $self = shift;	
		
	my $chr = $self->GetChr();
	my $rao_allprobes = $self->GetAllSelectedProbes();	
		
   my %h_probes = ();
   my $id = 1;
   foreach my $o_probe ( sort { $a->GetChrStart() <=> $b->GetChrStart() } @$rao_allprobes )
   {
	   my $probe_start =  $o_probe->GetChrStart();
	   my $probe_end   = $o_probe->GetChrEnd();
	   $h_probes{ $id } ={ chrom => $chr, element => $id, start => $probe_start, end => $probe_end, color => "green", label => "probe$id"};
	   $id++;
   } 
	
	return \%h_probes;
}

=head2 Function ProbeOutputNarcisseFormat

 Title        : ProbeOutputNarcisseFormat
 Usage        : $o_xxx->ProbeOutputNarcisseFormat();
 Function     : Dump the results in the Narcisse format in a file (eg. glintFish.nc)
 Returns      : none
 Args         : none
=cut
sub ProbeOutputNarcisseFormat
{
	my $self = shift;
	my $rh_probes = $self->FormatProbeOutputForNarcisse();	
	my $out_prefix = $self->_Get( '__output_prefix');
	my $output = $out_prefix.".nc";
	
	my @a_fields = qw( chrom element start end label color );
	my $fh = new IO::File( 	$output, "w" )
		or die "Impossible to open $output for writing";
	print $fh "#".join("\t",@a_fields)."\n";
	foreach my $probe (sort { $a <=> $b } keys %$rh_probes)
	{
		print $fh join("\t",@{$rh_probes->{$probe}}{@a_fields})."\n";
	} 
	$fh->close();	
}

=head2 Function GetSelectedProbesArray

 Title        : GetSelectedProbesArray
 Usage        : $o_xxx->GetSelectedProbesArray();
 Function     : Construct an array with the information on the selected probes
 Returns      : an array of probes 
 Args         : none
=cut

sub GetSelectedProbesArray
{
	my $self = shift;	
	my $rao_allprobes = $self->GetAllSelectedProbes();
	
	my @a_sorted_probes = sort { $a->GetChrStart() <=> $b->GetChrStart() } @$rao_allprobes;
	my @a_data = ();
	my @a_generalfields = qw(chrom start end size element);
	my @a_primer3fields = qw( leftPrimer rightPrimer leftTm rightTm leftEndStability rightEndStability leftSelfEnd rightSelfEnd leftSelfAny rightSelfAny PairPenality PairComplAny PercentRepeat);
	push @a_data, "#".join("\t",(@a_generalfields,@a_primer3fields))."\n";	
	my $num = 0;
	foreach my $o_probe ( @a_sorted_probes )
	{
		$num++;
		my @a_probeinfo = ();
		push @a_probeinfo, ($o_probe->GetChr(),$o_probe->GetChrStart(),$o_probe->GetChrEnd(),$o_probe->GetSize(),$o_probe->ProbeId() );
		push @a_probeinfo, ($o_probe->GetLeftPrimer(),$o_probe->GetRightPrimer());
		push @a_probeinfo, ($o_probe->GetLeftTm(),$o_probe->GetRightTm());
		push @a_probeinfo, ($o_probe->GetLeftEndStability(),$o_probe->GetRightEndStability());
		push @a_probeinfo, ($o_probe->GetLeftSelfEnd(),$o_probe->GetRightSelfEnd());
		push @a_probeinfo, ($o_probe->GetLeftSelfAny(),$o_probe->GetRightSelfAny());
		push @a_probeinfo, ($o_probe->GetPairPenality(), $o_probe->GetPairComplAny());
		push @a_probeinfo, (sprintf("%.1f", $o_probe->GetAmpliconRepeat()*100));
		push @a_data, join("\t",@a_probeinfo)."\n";
	}	
	return \@a_data;	
}


=head2 Function SelectedProbesLog

 Title        : GetSelectedProbesArray
 Usage        : $o_xxx->GetSelectedProbesArray();
 Function     : Construct an array with the information on the selected probes
 Returns      : an array of probes 
 Args         : none
=cut
sub SelectedProbesLog
{
	my $self = shift;	
	my $rao_allprobes = $self->GetAllSelectedProbes();
	
	my @a_sorted_probes = sort { $a->GetChrStart() <=> $b->GetChrStart() } @$rao_allprobes;	
	my $num = 0;
	foreach my $o_probe ( @a_sorted_probes )
	{
		$num++;
		my $freq = &CalculateRepeatFreq( $o_probe->GetSequence() );
		print STDERR join("\t",($o_probe->GetChr(),$o_probe->GetChrStart(),$o_probe->GetChrEnd(),$o_probe->GetSize(),sprintf("%.1f",$freq*100),$o_probe->GetNumHits()))."\n";
	}		
}


sub SuitableRegionsLog
{
	my $self = shift;		
	my $rao_suitables = $self->_Get( '__rao_suitables' );
		
	my $num = 0;
	foreach my $o_region ( @$rao_suitables )
	{
		$num++;
		my $freq = &CalculateRepeatFreq( $o_region->GetSequence() );
		print STDERR join("\t",($o_region->GetChr(),$o_region->GetChrStart(),$o_region->GetChrEnd(),$o_region->GetSize(),sprintf("%.1f",$freq*100)))."\n";
	}		
}

sub CalculateRepeatFreq
{
	my $seq = shift;

	my @seqarray = split('',$seq);
	my $count = 0;
	foreach my $base (@seqarray) 
	{
		$count++ if $base =~ /[a|c|g|t|N]/;
	}
	my $len = $#seqarray+1;
	return $count / $len;
}

=head2 CreateNarcisseInput

 Title        : CreateNarcisseInput
 Usage        : $o_xxx->CreateNarcisseInput(  )
 Function     : Create a file corresponding to the csv data file format used as an input for Narcisse 
 Returns      : void
 Args         : none
 Globals      : none
=cut
sub CreateNarcisseInput
{
	my ( $self ) = @_;
	
	my $o_speciesObserver = $self->_Get( '__o_speciesObserver' );		
	my $o_logger= $self->_Get('_o_logger');
	
	my $rh_probes = $self->FormatProbeOutputForNarcisse();
    my $chr = $self->GetChr();
    
	my @a_headers = qw( chrom element start end color label );
	my $file_by_chrom = $o_speciesObserver->GetPersoFiles( -check => &FALSE, -chrom => $chr);
	my $o_ndw         = New NarcisseDataWriter::OnMemory( $file_by_chrom, \@a_headers, { append => &TRUE }, undef, $o_logger );
	$o_ndw -> Update($rh_probes);
   		
}


###### Launching commands ######

=head2 function LaunchGlintFreq

 Title        : LaunchGlintFreq
 Usage        : &LaunchGlintFreq();
 Function     : Run GlintFreq software
 Access       : Public
 Returns      : none
 Args         : none
 
=cut
sub LaunchGlintFreq
{
	my $self = shift ;
	
	my $glintFreq_exe = $self->_Get( '__glintFreq_exe' ) ;
	
	my $param = $self->GetGlintFreqParameters();
	
	my $bank      = $self->_Get( '__genome' ) ;
	my $queryFile = $self->_Get( '__query_filename' ) ;
		
	my $output_file_name = &GetTempFileName( 'glintfreq', 'out' ) ;
	my $log_file_name    = &GetTempFileName( 'glintfreq', 'log' ) ;
	my $err_file_name    = &GetTempFileName( 'glintfreq', 'err' ) ;
	
	my $cmd = $glintFreq_exe." $queryFile -C 0 $param -o $output_file_name";
		
	my ( $status, $killed ) = RunExt( -cmd => $cmd, -stderr => ">$err_file_name" );		
}

=head2 function LaunchPrimer3

 Title        : LaunchPrimer3
 Usage        : &LaunchPrimer3();
 Function     : Run Primer3 software
 Access       : Public
 Returns      : none
 Args         : none
=cut
sub LaunchPrimer3
{
	my $self = shift;
	
	my $primer3_exe = $self->_Get( '__primer3_exe' );
	
	my $primerInputFile  = &GetTempFileName( 'Primer3', 'in' );
	my $primerOutputFile = &GetTempFileName( 'Primer3', 'out');
				
	my $command = $primer3_exe." <$primerInputFile -output $primerOutputFile";
	system( $command );	
		
	return 	$primerOutputFile;
}

=head2 function LaunchGlintPairEnd

 Title        : LaunchGlintPairEnd
 Usage        :my $output = &LaunchGlintPairEnd($fwd, $bwd);
 Function     : Run GlintFreq software
 Access       : Public
 Returns      : none
 Args         : none
 
=cut
sub LaunchGlintPairEnd
{
	my $self = shift;
	my ( $fwd, $bwd ) = @_;
	
	my $genome               = $self->_Get( '__genome' );
	my $glintPairEnd_exe     = $self->_Get( '__pairEnd_exe' );
		
	my $processus = 'pairEnd';
	my $output = &GetTempFileName( $processus, 'out' ) ;
	my $log    = &GetTempFileName( $processus, 'log' ) ;
	my $err    = &GetTempFileName( $processus, 'err' ) ;
	
	my $cmd = $glintPairEnd_exe;
	$cmd.=" $genome $fwd,$bwd";
	$cmd.=" -m bed --mate-maxdist 20000 -o $output";
	my ( $status, $killed ) = RunExt( -cmd => $cmd, -stderr => ">$err" );		
	
	return $output;
}

=head2 function LaunchGlint

 Title        : LaunchGlint
 Usage        : &LaunchGlint();
 Function     : Run Glint software
 Access       : Public
 Returns      : none
 Args         : none
 
=cut
sub LaunchGlint
{
	my ( $self, $input ) = @_ ;
		
	my $genome               = $self->_Get( '__genome' );
	my $glint_exe            = $self->_Get( '__glint_exe' );
	
	my $glint_scoreThreshold = $self->_Get( '__glint_score_threshold' );
	my $param                = "-s $glint_scoreThreshold";
	
	my $processus = 'glint';
	my $output = &GetTempFileName( $processus, 'out' ) ;
	my $log    = &GetTempFileName( $processus, 'log' ) ;
	my $err    = &GetTempFileName( $processus, 'err' ) ;
	
	my $cmd = $glint_exe." $genome $input $param -m 2 -C 0 -w 3 --step 3 -s 100 -c 30 -X 40 -o $output";
	#my $cmd = $glint_exe." $genome $input $param -m 2 -C 0 -o $output";
	my ( $status, $killed ) = RunExt( -cmd => $cmd, -stderr => ">$err" );	
	
	return $output;	
}

###### PUSHING ######

=head2 function PushSuitable

 Title        : PushSuitable
 Usage        : $region->PushProbes($suitable);
 Function     : Push a suitable subregion in the suitable subregion array of the region object
 Access       : Public
 Returns      : none
 Args         : $suitable => the suitable subregion pushed

=cut
sub PushSuitable
{
	my ( $self, $o_suitable ) = @_;
	
	my $rao_suitables = $self->_Get( '__rao_suitables' );
	push( @$rao_suitables, $o_suitable );
}

sub  _SetWorkingDir
{
	my $self = shift;
		
	unless ( $self->_Get( '_web' ) )
	{	
		my $work_dir = $self->_Get( '__targetDir' );
		chdir( $work_dir );
		$self->Trace("working in $work_dir\n");		
	}	
}

sub _LeavingWorkingDir
{
	my $self = shift;
		
	unless ( $self->_Get( '_web' ) )
	{	
		chdir( 	$self->_Get( '__execution_dir' ) );
	}
}

sub _CleaningWorkingDir
{
	my $self = shift;
		
	unless ( $self->_Get( '_web' ) || $self->_Get( '__no_cleaning' ))
	{	
		my $work_dir = $self->_Get( '__targetDir' );
		rmtree( $work_dir );	
		$self->Trace("removing $work_dir\n");		
	}
}


=head2 function GetTempFileName

 Title        : GetTempFileName
 Usage        : $tempFileName = &GetTempFileName($process);
 Function     : Accessor, returns the Amplicon start position
 Access       : Public
 Returns      : $from
 Args         : none

=cut

sub GetTempFileName
{
	my ( $processus, $type, $ext ) = @_;
	
	unless ( defined $ext ) { $ext = '' ; }
	die "\n Error : Wrong argument in GetTempFileName $!\n" if ( $type !~ /(in|out|log|param|err)/ );
	
	my $tempFileName = $processus.'.'.$type.'.GF-tmp'.$ext;
	
	return $tempFileName;
}

=head2 function GetFileName

 Title        : GetFileName
 Usage        : $fileName = &GetFileName($process);
 Function     : Construct the filename according toe the process and type (ex: glintfreq.out.GF-tmp)
 Returns      : $from
 Args         : none

=cut
sub GetFileName
{
	my ( $processus, $type, $ext ) = @_;
	
	unless ( defined $ext ) { $ext = '' ; }
	die "\n Error : Wrong argument in GetTempFileName $!\n" if ( $type !~ /(in|out|log|param|err)/ );
	
	my $fileName = $processus.'.'.$type.'.GF'.$ext;
	
	return $fileName;
}

=head2 function CleanTempFileName

 Title        : CleanTempFileName
 Usage        : &CleanTempFileName();
 Function     : Erase all files created
 Access       : Public
 Returns      : none
 Args         : none
 
=cut
sub CleanTempFileName
{
	#~ my $command = "rm -I *.GF-tmp*";
	my $command = "rm *.GF-tmp*";
	system ( $command );
}

## Some accessors

sub GetChr
{
	my $self = shift;
	return $self->_Get('__chr');
}

sub GetStart
{
	my $self = shift;
	return $self->_Get('__start');
}

sub GetEnd
{
	my $self = shift;
	return $self->_Get('__end');
}

sub Species
{
	my $self = shift;
	return $self->_Get('__species');
}

sub NbSuitables
{
	my $self = shift;
	return 	scalar @{$self->_Get( '__rao_suitables')};	
}


#### Tracing messages ###
sub Trace
{
	my ( $self, $message ) = @_;
	
	if ( $self->_Get( '__web' ) )
	{
		#print STDERR "Hello \n";
		warn $message;
		$self->_Trace( $message, '.', 'GS' );	
	}		
	if ( $self->_Get( '__verbose' ) )
	{
		print STDERR $message;
		#print STDERR colored ( $message, 'white' ) ; 
	}	
}

sub DebugTrace
{
	my ( $self, $message ) = @_;
	
	if ( $self->_Get( '__debug' ) )
	{
		print STDERR colored ( $message, 'white' ) ; 
	}	
}


=head2 Function GetSelectedProbesHash

 Title        : GetSelectedProbesHash
 Usage        : $o_xxx->GetSelectedProbesHash();
 Function     : Construct an array with one hash per probe with the information on the selected probes
 Returns      : an array of hash 
 Args         : none
=cut

sub GetSelectedProbesHash
{
	my $self = shift;	
	my $rao_allprobes = $self->GetAllSelectedProbes();
	
	my @a_sorted_probes = sort { $a->GetChrStart() <=> $b->GetChrStart() } @$rao_allprobes;
	my @a_data = ();
	#~my @a_selectedfields = qw(chrom start end size element leftPrimer rightPrimer leftTm rightTm);
	#~push @a_data, "#".join("\t",(@a_selectedfields))."\n";	
	#~push @a_data, join("\t",(@a_selectedfields))."\n";		
	my $num = 0;
	foreach my $o_probe ( @a_sorted_probes )
	{
		$num++;
		my %h_probeInfo =();
		%h_probeInfo = (
										chrom 				=> $o_probe->GetChr(),
										start 				=> $o_probe->GetChrStart(),
										end 				=> $o_probe->GetChrEnd(),
										size 				=> $o_probe->GetSize(),
										element 			=> $o_probe->ProbeId(), 
										leftPrimer 			=> $o_probe->GetLeftPrimer(),
										rightPrimer 		=> $o_probe->GetRightPrimer(),
										leftTm 				=> $o_probe->GetLeftTm(),
										rightTm 			=> $o_probe->GetRightTm(),
										leftEndStability 	=> $o_probe->GetLeftEndStability(),
										rightEndStability 	=> $o_probe->GetRightEndStability(),
										leftSelfEnd			=> $o_probe->GetLeftSelfEnd(),
										rightSelfEnd		=> $o_probe->GetRightSelfEnd(), 
										leftSelfAny			=> $o_probe->GetLeftSelfAny(),
										rightSelfAny		=> $o_probe->GetRightSelfAny(),
										PairPenality		=> $o_probe->GetPairPenality(), 
										PairComplAny		=> $o_probe->GetPairComplAny()
										);
			
		
		push @a_data, \%h_probeInfo;
	}	
	return \@a_data;	
}
1;
