# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Image - a GD inherited class for Narcisse data representations.

=head1 SYNOPSIS

Image -> New();					# Create a 100x100 image
Image -> New(GD::Image);
Image -> New(%h_param);


B<Example :>
	my $image = Image->New(width=>800, height=>800);

	# Use GD methods naturally
	my $red  = $image -> colorAllocate(255,0,0);
	my $blue = $image -> colorAllocate(0,0,255);

	$image->filledRectangle(100,100,200,300,$blue);
	$image->string($image->{font}, 10,10,"Test text",$blue);
	$image->filledTriangle(360,150,700,275,450,500,$blue);

	# Change the frame and paint same objects in a different color.
	$image->AdaptFrameToView(100,100,500,500,0,0.5);	# Change FrameTransformer parameters
	$image->filledRectangle(100,100,200,300,$red);
	$image->filledTriangle(360,150,700,275,450,500,$red);
	$image->string($image->{font}, 10,10,"Test text",$red);

	# Look the result :
	$image->PrintLayers('png','my_test');

=head1 DESCRIPTION

Image is a GD inherited class for Narcisse data representations.
AUTOLOAD method has been overload to inherited GD methods, so all GD methods could be call normally.

This class use a FrameTransformer object to control the draw area and to calculate all pixel coordinates from other data.
To draw in the picture normally (like in a GD::Image object) just don't touch to this FrameTransformer object.
Otherwise, if you want to adapt nucleic data, or other, on pixel coordinates, you have to use AdaptFrameToView method to change the area and the "factor" of drawing.

Moreover, this class use layers attributes to manage differents levels of drawing.
Each layer is a separate GD::Image object.


B<Overloaded subroutines :>
- arc
- char
- charUp
- colorAllocate
- colorAllocateAlpha
- colorDeallocate
- dashedLine
- fill
- filledRectangle
- line
- rectangle
- setPixel
- string
- stringUp

B<New subroutines :>
- AdaptFrameToView
- AddLayer
- borderedFilledRectangle
- brokenArc
- brokenLine
- colorClearer
- colorDarker
- filledEquilateralTriangle
- filledTriangle
- GetFactor
- GetFrameAttributes
- InitColorsLibrary
- PrintImage
- PrintLayers
- randomColor
- Rotate
- stringBox
- verticalCylinder
- verticalCylinderPart
- verticalHalfArrow

=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package Image;

use strict;
use warnings;

use base qw( LipmObject );

use Scalar::Util qw( blessed );
use Data::Dumper;
use GD;

use Error qw(:try);

use LipmError::IOException;
use DataReader::Whole;

use FrameTransformer::2D;
use FrameTransformer::1D;

our $AUTOLOAD;

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};
my $DEBUG = 0;

# Global parameters

my @A_FOREGROUND_COLOR 	= (0,0,0);
my @A_BACKGROUND_COLOR 	= (255,255,255);
use constant LAYER => 'background';
my @A_DEFAULT_LAYERS	= ( &LAYER );	# pourquoi il n'y a pas de 'default' qui sert de test plus tard ?

# Class parameter
our %H_COLORS_LIBRARY;
our %H_FONT_LIBARY;
our %H_COLOR_INDEX;

=head2 _Init

 Title        : _Init
 Usage        : _Init()
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : %h_param = (
 						width 	=> integer,
 						height 	=> integer,
 						layers	=> array_reference)
 				or
 				GD::Image object

=cut
sub _Init
{
	my $self = shift;
	my @a_param = @_;

	my %h_default_param = (
				width 	=> 100,
				height 	=> 100,
				frame_transf_type => "2D",
				layers	=> \@A_DEFAULT_LAYERS
				);

	#==================================
	# 0-1) Test to copy data from a GD::Image object
	if ( scalar (@a_param) == 1 && ref($a_param[0]) eq 'GD::Image' )
	{
		my $o_gd_image = $a_param[0];

		# Create a new object Image with GD width and height
		$self = $self->_Init( width => $o_gd_image->width(),
							  height=> $o_gd_image->height() );

		# Replace the default layer by GD::Image object.
		$self->_SetLayer( &LAYER, $o_gd_image);

		return $self;
	}
	# 0-2) Test to copy data from an image file
	elsif ( scalar (@a_param) == 1 )
	{
		$self = $self->_Init( new GD::Image($a_param[0]) );

		return $self;
	}

	#==================================
	# 1) Overload default parameters :
	my %h_param = ( %h_default_param, @a_param);

	my $frame_type=$h_param{frame_transf_type};

	#==================================
	# 2) Set object attributes
	$self -> {width}		  = $h_param{width};
	$self -> {height}		  = $h_param{height};
	$self -> {ra_layers}	  = ();
	$self -> {rh_layers}	  = {};
	$self -> {orientation} 	  = 'portrait';
	$self -> {frame} 		  = "FrameTransformer::$frame_type" -> New( X1 => $h_param{width},
														 Y1 => $h_param{height});

	#==================================
	# 3) Initialize layers :
	my @a_layers = @{$h_param{layers}};
	foreach my $o_layer (@a_layers)
	{
		$self -> AddLayer($o_layer);
	}
	
	$self->{ __absolute_frame } = 0;	
	my $default_layer = $a_layers[0];
	$self->{ __current_layer } = $self->GetLayer( $default_layer );

	return $self;
}

=head2 AUTOLOAD

 Title        : AUTOLOAD
 Usage        : automatically call by perl when the calling method is unknown
 				$image -> GD_method();
 				$image -> New();
 Prerequisite : none
 Function     : *use the GD::Image method if it's not overload
 				*apply the GD::Image method which is call on the layer $o_layer if the last parameter is set.
 				*call the New method of UniversalObject
 Returns      : the result of the calling method if it exist
 Args         : none
 Globals	  : none

=cut
sub AUTOLOAD
{
	my $self = shift;
	my ($class,$func_name) = $AUTOLOAD=~/(.+)::([^:]+)$/;
	my @func_parameters = @_;

	my $o_layer = (defined $func_parameters[-1] && defined (${$self->{rh_layers}}{$func_parameters[-1]}) ) ? pop(@func_parameters) : ${$self->{ra_layers}}[0];
	my @result;

	return if $func_name eq 'DESTROY';				# Return if DESTROY method is call

	#----------------------------------------------------------
	# Test the method to call
	# 	- Use the UniversalObject method New
	if ($func_name eq "New")
	{
		$self->UNIVERSAL::New(); # ?? LipmObject
	}
	# 	- Use GD method on gd object of the layer $o_layer if it exist
	elsif (ref $self && exists ${$self->{rh_layers}}{$o_layer})
	{
		my $o_layer = $self->GetLayer($o_layer);
		@result = $o_layer -> $func_name(@func_parameters);
	}
	# 	- Use GD::Image class method if the Image object is not instanciate.
	else
	{
		@result = GD::Image->$func_name(@func_parameters);
	}


	#----------------------------------------------------------
	# Test of @result (to return the right GD method results) :
	#	- copy of a GD::Image object
	if ( ref($result[0]) eq 'GD::Image' )
	{
		$self -> _SetLayer($o_layer, $result[0]);
		return $self;
	}
	#	- test @result to return a scalar
	elsif ( scalar @result == 1 )
	{
		return $result[0];
	}
	#	- test @result to return an array
	else
	{
		return @result;
	}
}

#------------------------------------------------#
#              Managing the layers               #
#------------------------------------------------#

=head2 _SetLayer
 Title		  :	_SetLayer
 Usage		  : $self -> _SetLayer(name, $o_gd)
 Prerequisite :	Image have to be initialized
 Procedure	  : Stores the  GD object $o_gd in the $>self->{ rh_layers } hash with hash key $name
 Returns	  : none
 Args		  : name -> which layer have to be set
 				$o_gd-> the new GD::Image

=cut
sub _SetLayer
{
	my $self = shift;
	my $name = shift if (defined $_[0]) or return;
	my $o_gd = shift if (defined $_[0] && ref($_[0]) eq 'GD::Image') or return;
	${$self->{rh_layers}}{$name} = $o_gd;

	return;
}

=head2 SetCurrentLayer
 Title		  : SetCurrentLayer
 Usage		  : $self -> SetCurrentLayer(name)
 Procedure	  : Set __current_layer to the layer named 'name'. 
                Enables subsequent drawing actions to be performed directly on this current layer
 Returns	  : the previous current layer
 Args		  : the name of the layer as a string

=cut
sub SetCurrentLayer
{
	my $self  = shift;
	my $name  = shift;

	my $previous_current_layer = $self->{ __current_layer };
	my $new_current_layer =  $self->GetLayer($name);
	$self->{ __current_layer } = $new_current_layer;
	
	return $previous_current_layer;
}

=head2 GetCurrentLayer
 Title		  : GetCurrentLayer
 Usage		  : $self -> GetCurrentLayer()
 Procedure	  : returns the current_layer which is a GD object.
 Args		  : none

=cut
sub GetCurrentLayer
{
	my $self  = shift;
	
	return $self->{ __current_layer };
}

=head2 GetLayer
 Title		  : GetLayer
 Usage		  : $self -> GetLayer(name)
 Procedure	  : Get the reference to GD::Image object of the layer named 'name'.
 				If this layer is a new one, we call AddLayer and return this object.
 Returns	  : reference on a GD::Image blessed object.
 Args		  : name
 Attributes	  : layers

=cut
sub GetLayer
{
	my $self     = shift;
	my ( $name ) = @_;

	# Test parameter
	$name = ($self->GetLayersList)[0] if (!defined $name);

	if ( blessed($name) && $name->isa( "GD::Image" ))
	# The name argument is in fact already a GD::Image object, this is for backward compatibility
	{
		return $name;
	}	
	elsif (defined ${$self->{rh_layers}}{"$name"})
	# If this layer is defined, we return the GD object associated with :
	{
		return ${$self->{rh_layers}}{"$name"};
	}

	# Else we add a new GD::Image object on layers, named $name :
	else
	{
		return $self->AddLayer($name);
	}
}

=head2 AddLayer
 Title		  :	AddLayer
 Usage		  : $o_image -> AddLayer(name [,file])
 Prerequisite :	Image have to be initialized
 Procedure	  : Add a layer on level +1 (A layer is a new GD::Image)
 				You can add a new blank layer setting just named param, or add an existing image file.
 Returns	  : The blessed object created.
 Args		  : name : name of the layer
 				file : image file (png,gif or jpg)
 Attributes	  : rh_layers,ra_layers

=cut
sub AddLayer
{
	my $self = shift;
	my ( $name, $file ) = @_;

	# Test parameters.
	$name = 'layer'.scalar(@{$self->{ra_layers}}) if (!defined $name);
	$file = '' if (!defined $file);

	# If this layer isn't already filled, we add the gd object and initialize default parameters :
	if (!defined ${$self->{rh_layers}}{$name})
	{
		#-----------------------
		# 1) Make a blank GD::Image by default
		my $o_gd = GD::Image -> new($self->{width}, $self->{height});

		#-----------------------
		# 2) Set colors indexes and transparancy :
		my $bgcolor = $o_gd->colorAllocate(@A_BACKGROUND_COLOR);
		my $fgcolor = $o_gd->colorAllocate(@A_FOREGROUND_COLOR);
		$o_gd -> transparent($bgcolor);

		#-----------------------
		# 3) Try to load the image from $file
		$o_gd = GD::Image -> new($file) if (-f $file);

		#-----------------------
		# 4) Store the layer name:
		push (@{$self->{ra_layers}}, $name);

		#-----------------------
		# 5) Store the layer object:
		$self->_SetLayer($name,$o_gd);

		return $o_gd;
	}

	return;
}

=head2 DeleteLayers
 Title		  :	DeleteLayers
 Usage		  : $o_image -> DeleteLayer(@a_layers)
 Prerequisite :	Image have to be initialized
 Procedure	  : Delete existing layers.
 Returns	  : None
 Args		  : @a_layers : list of the layers to delete
 Attributes	  : rh_layers,ra_layers

=cut
sub DeleteLayers
{
	my ($self,@a_layers) = (shift,@_);

	# Test parameters.
	return if (scalar(@a_layers) == 0);

	# Delete on rh_layers
	foreach my $layer (@a_layers)
	{
		${$self->{rh_layers}}{$layer} = undef if (exists ${$self->{rh_layers}}{$layer});
	}

	# Delete on ra_layers.
	my @a_new_list = ();
	foreach my $element (@{$self->{ra_layers}})
	{
		push (@a_new_list, $element) if ( scalar( grep({$_ eq $element} @a_layers)) == 0 );
	}
	$self->{ra_layers} = \@a_new_list;

	return;
}

=head2 MergeLayers
 Title		  :	MergeLayers
 Usage		  : $o_image -> MergeLayers(destination_layer, @a_layers)
 Prerequisite :	Image have to be initialized
 Procedure	  : Merge all layers af the list in a destination one.
 				If the destination already exists, it's merged too.
 				The order of drawing is the same as parameters, so destination_layer will be draw first and after the layers are ordered like the list.
 Returns	  : none
 Args		  : destination_layer : name of the destination layer
 				@a_layers : list of existing layers to merge with.
 Attributes   : rh_layers, ra_layers

=cut
sub MergeLayers
{
	my ($self, $dest_layer, @a_layers) = (shift, shift, @_);

	# Test parameters.
	return if (!defined $dest_layer or scalar(@a_layers) == 0);

	# Add a layer if destination not exist.
	my $o_dest = $self->GetLayer($dest_layer);

	# For each layer name of the list, merge it to the destination and delete old layer.
	foreach my $layer (@a_layers)
	{
		next if (!exists ${$self->{rh_layers}}{$layer});
		$o_dest -> copy($self->GetLayer($layer),0,0,0,0,$self->{width},$self->{height});
	}
	$self -> DeleteLayers(@a_layers);

	return;
}

=head2 GetLayersList
 Title		  :	GetLayersList
 Usage		  : $self -> GetLayersList()
 Prerequisite :	Image have to be initialized
 Procedure	  : Return the actual names list of layers composing the image
 Returns	  : Array of sorted layers names
 Args		  : none
 Attributes	  : ra_layers

=cut
sub GetLayersList
{
	my $self = shift;

	return @{$self->{ra_layers}};
}

#------------------------------------------------#
#              Managing the colors               #
#------------------------------------------------#

=head2 AddColorOnLibrary
 Title		  :	AddColorOnLibrary
 Usage		  : &Image::AddColorOnLibrary(color_name,r,g,b)
 Prerequisite :	Image have to be initialized
 Procedure	  : Add 'color_name' in the library
 Returns	  : none
 Args		  : color_name,r,g,b
 Globals	  : colors_library

=cut
sub AddColorOnLibrary
{
	my $self = shift if ($_[0] =~ /Image/);
	my ($color_name,$r,$g,$b) = (shift,shift,shift,shift) or return;

	my @a_rgb = ($r,$g,$b);

	$H_COLORS_LIBRARY{$color_name} = \@a_rgb if ($r=~/^\d+$/ && $g=~/^\d+$/ && $b=~/^\d+$/);

	return;
}

=head2 InitColorsLibrary
 Title        : InitColorsLibrary
 Usage        : $o_image -> InitColorsLibrary($file)
 Prerequisite : none
 Procedure    : Initialize the global hash with data of the file $file.
 				Catch an exception if the file doesn't exist, but store file name for the future saving.
 Returns      : none
 Args         : $file
 Globals      : colors_library
 Error		  : LipmError::IOException

 TODO : utiliste du try ?

=cut
sub InitColorsLibrary
{
	my $self = shift if ($_[0] =~ /Image/);
	my $file = shift;

	try
	{
		my $o_ndr = New DataReader::Whole();
		$o_ndr->Prepare( -from   => $file );	  	               		
		my $rh_data = $o_ndr->FetchallHashref( "color_label", -line_type => 'HASH' );

		foreach my $color_label (keys(%$rh_data))
		{
			my $rh_element_data = ${$rh_data}{$color_label};
			&AddColorOnLibrary($color_label,${$rh_element_data}{r}, ${$rh_element_data}{g}, ${$rh_element_data}{b});
		}

	}
	catch LipmError::IOException with
	{
		return;
	};

	return;
}

=head2 randomColor
 Title        : randomColor
 Usage        :	$image -> randomColor([min_color_difference][layer])
 Prerequisite : none
 Procedure    :	Get the index of a new random color.
 Returns      : the index of the new color
 Args         :
 Globals      : none

=cut
sub randomColor
{
	my $self = shift;
	my $min_color_diff = (defined $_[0] && $_[0]=~/^\d*$/) ? shift : 30;
	my $o_layer = $self->GetLayer(shift);

	my ($r,$g,$b) = (rand(255),rand(255),rand(255));
	my $new_color_index;
	my @rgb_nearest = $o_layer->rgb($o_layer->colorClosest($r,$g,$b));
	my $bool_already_exist = ($o_layer -> colorExact($r,$g,$b) != -1 || abs($rgb_nearest[0]-$r)<$min_color_diff || abs($rgb_nearest[1]-$g)<$min_color_diff || abs($rgb_nearest[2]-$b)<$min_color_diff) ? 1 : 0;


	if ($bool_already_exist == 1)
	{
		$new_color_index = $self->randomColor($o_layer);
	}
	else
	{
		$new_color_index = $o_layer->colorAllocate($r,$g,$b);
	}

	return $new_color_index;
}

=head2 colorDarker
 Title        : colorDarker
 Usage        :	$image -> colorDarker(color_index [,factor])
 Prerequisite : none
 Procedure    :	Obscure the color which index is color_index
 Returns      : the index of the new color
 Args         : color_index -> GD color index
 				factor -> optional, the 'hardness' of the darkening (between 1 to 5)
 Globals      : none

=cut
sub colorDarker
{
	my $self = shift;
	my $color_index = shift;
	my $factor = (defined $_[0] && "$_[0]" =~ /^[1-5]$/) ? shift : 1;
	my $o_layer = shift;

	my $dark_factor = 30*$factor;
	my $new_color_index = $self -> colorAllocate(0,0,0,$o_layer);
	my @rgb = $self -> GetLayer($o_layer) -> rgb($color_index) or return $new_color_index;

	foreach my $color (@rgb)
	{
		$color = ($color - $dark_factor >= 0) ? $color - $dark_factor : 0;
	}

	$new_color_index = $self -> colorAllocate(@rgb,$o_layer);

	return $new_color_index;
}

=head2 colorClearer
 Title        : colorClearer
 Usage        :	$image -> colorClearer(color_index)
 Prerequisite : none
 Procedure    :	Clear up the color which index is color_index
 Returns      : the index of the new color
 Args         : color_index -> GD color index
 				factor -> optional, the 'hardness' of the cleaning (between 1 to 5)
 Globals      : none

=cut
sub colorClearer
{
	my $self = shift;
	my $color_index = shift;
	my $factor = (defined $_[0] && "$_[0]" =~ /^[1-5]$/) ? shift : 1;
	my $o_layer = shift;

	my $clear_factor = 30*$factor;
	my $new_color_index = $self -> colorAllocate(255,255,255,$o_layer);
	my @rgb = $self -> GetLayer($o_layer) -> rgb($color_index) or return $new_color_index;

	foreach my $color (@rgb)
	{
		$color = ($color + $clear_factor <= 255) ? $color + $clear_factor : 255;
	}

	$new_color_index = $self -> colorAllocate(@rgb,$o_layer);

	return $new_color_index;
}

=head2 colorAllocate
 Title        : colorAllocate
 Usage        :	$image -> colorAllocate(r,g,b [,layer])
 				$image -> colorAllocate(color_label [,layer])
                $image -> colorAllocate()
 Prerequisite : none
 Procedure    :	overload the colorAllocate GD method 
 TODO         : PASSER LES PARAMETRES PAR UN HASH
 Returns      : none
 Args         : none
 Globals      : gd

=cut
sub colorAllocate
{
	my $self    = shift;
	my @a_param = @_;

	my (@rgb, $o_layer, $index);

	# Test parameters :
    # if (r,g,b,[layer])
	if (scalar(@a_param) >= 3 && $a_param[0]=~/^\d+$/ && $a_param[1]=~/^\d+$/ && $a_param[2]=~/^\d+$/)
	{
        $o_layer = $self->GetLayer($a_param[3]);
		@rgb = @a_param[0..2];
	}
    
    # if (color_label,[layer])
	elsif (scalar(@a_param) >=1 && scalar(@a_param) <= 2)
	{
        my $layer_name = $a_param[1];
        $o_layer = $self->GetLayer($layer_name);
        
        # if color_label already allocated
        if (defined @{$H_COLORS_LIBRARY{$a_param[0]}})
        {
            @rgb =  @{$H_COLORS_LIBRARY{$a_param[0]}};
        }
        else
        {
            return $self->colorAllocate(@A_FOREGROUND_COLOR, $layer_name);
        }
	}
    
    # if NO PARAMETER = default is foreground color
	else
	{
		return $self->colorAllocate(@A_FOREGROUND_COLOR);
	}


	# Set color index :
	if ($o_layer->colorsTotal() < 256)
	{
		$index = ($o_layer->colorExact(@rgb) == -1) ? $o_layer->colorAllocate(@rgb) : $o_layer->colorExact(@rgb);
	} else
	{
		$index = $o_layer->colorClosest(@rgb);
	}

	return $index;
}

=head2 colorAllocateAlpha
 Title        : colorAllocateAlpha
 Usage        :	$image -> colorAllocateAlpha(r,g,b,alpha)
 Prerequisite : none
 Procedure    :	overload GD method to test if the color index already exist.
 Returns      : none
 Args         : none
 Globals      : gd

=cut
sub colorAllocateAlpha
{
	my $self = shift;
	my @a_param = @_;

	my (@rgba, $o_layer, $index);

	# Test parameters :
	if (scalar(@a_param) >= 4 && $a_param[0]=~/^\d+$/ && $a_param[1]=~/^\d+$/ && $a_param[2]=~/^\d+$/ && $a_param[3]=~/^\d+$/)
	{
		$o_layer = (defined $a_param[4]) ? $self->GetLayer($a_param[4]) : $self->GetLayer();
		@rgba = @a_param[0..3];
	}
	elsif (scalar(@a_param) >=2 && scalar(@a_param) <= 3)
	{
		$o_layer = (defined $a_param[2]) ? $self->GetLayer($a_param[2]) : $self->GetLayer();
		my @rgb = (defined @{$H_COLORS_LIBRARY{$a_param[0]}}) ? @{$H_COLORS_LIBRARY{$a_param[0]}} : return $self->colorAllocate(@A_FOREGROUND_COLOR, $o_layer);
		@rgba = (@rgb, $a_param[1]);
	}
	else
	{
		return $self->colorAllocate(@A_FOREGROUND_COLOR, $o_layer);
	}


	# Set color index :
	if ($o_layer->colorsTotal() < 256) {
		$index = ($o_layer->colorExactAlpha(@rgba) == -1) ? $o_layer->colorAllocateAlpha(@rgba) : $o_layer->colorExactAlpha(@rgba);
	}
	else
	{
		$index = $o_layer->colorClosestAlpha(@rgba);
	}

	return $index;
}

=head2 colorDeallocate
 Title        : colorDeallocate
 Usage        :	$image -> colorDeallocate([index][layer])
 Prerequisite : none
 Procedure    :	deallocate an index for on all layers.
 Returns      : none
 Args         : none
 Globals      : gd

=cut
sub colorDeallocate
{
	my $self = shift;
	my $index = (defined $_[0] && $_[0]=~/^\d+$/) ? shift : undef;
	my @layer_list = @_;

	@layer_list = $self->GetLayersList() if (scalar(@layer_list) == 0);
	foreach my $o_layer (@layer_list)
	{
		my $o_layer = $self->GetLayer($o_layer);
		if (defined $index)
		{
			$o_layer -> colorDeallocate($index);
		}
		else
		{
			for (my $i=1; $i<$o_layer->colorsTotal(); $i++)
			{
				$o_layer -> colorDeallocate($i);
			}
		}
	}

	return;
}

#------------------------------------------------#
#              Managing the fonts                #
#------------------------------------------------#

=head2 InitFontLibrary
 Title        : InitFontLibrary
 Usage        : $o_image ->InitFontLibraryy($file)
 Prerequisite : none
 Procedure    : Initialize the global hash with data of the file $file.
 				Catch an exception if the file doesn't exist, but store file name for the future saving.
 Returns      : none
 Args         : $file containing the font definition
 				$font_directory : the font directory on the server
 Globals      : none
 Error		  : LipmError::IOException

 TODO : utilite du try ?

=cut
sub InitFontLibrary
{
	my $self = shift if ($_[0] =~ /Image/);
	my $file = shift;
	my $font_directory= shift;
	my $rh_fonts;

	try
	{
		my $rh_param=New ParamParser($file);
		foreach my $font ($rh_param->GetKeys()) {
			$H_FONT_LIBARY{$font} = "$font_directory/".$rh_param->Get($font);
		}

	}
	catch LipmError::IOException with
	{
		return;
	};

	return;
}

=head2 font
 Title        : font
 Usage        : $o_image ->font($font_name)
 Prerequisite : InitFontLibrary has been called
 Function     : returns the correctct font as initialized by InitFontLibrary
 Returns      : the correct font
 Args         : $font_name
 Globals      : none
 Error		  : LipmError::IOException

=cut
sub font
{
	my $self = shift;
	my $font_name = shift;

	if( !defined $H_FONT_LIBARY{ $font_name } )
	{
		my $dbg = "Font name $font_name undefined";
		throw LipmError::ObjectException(__PACKAGE__, '!build', -text => $dbg);
	}

	return $H_FONT_LIBARY{$font_name};
}

#------------------------------------------------#
#              Managing the frame                #
#------------------------------------------------#

=head2 SetFrameChange
 Title		  :	SetFrameChange
 Usage		  : $self -> SetFrameChange(%h_param)
 Prerequisite :	FrameTransformer object initialized
 Procedure	  : Call SetFrame to enable a correct frame change
 Returns	  : none
 Args		  : a hash : (X0 => , Y0 => , X1 => , Y1 => , x0 => , y0 => , x1 => , y1 => )
				(X0,Y0,X1,Y1)	=> Coordinates to move the frame box.
 				(x0,y0,x1,y1)	=> Coordinates of the relative box.
 Globals	  : none

=cut
sub SetFrameChange
{
	my $self = shift;
	my @a_param = @_;

	if (scalar (@a_param) == 0)
	{
		$self->{frame} -> SetFrame( X0=>0, Y0=>0, X1=>$self->width(), Y1=>$self->height() );
	}
	elsif ( scalar (@a_param) == 1 && $a_param[0]->isa("FrameTransformer") )
	{
		$self->{frame} = $a_param[0];
	}
	else
	{
		$self->{frame} -> SetFrame( @a_param );
	}

	return;
}

=head2 SetAbsoluteFrame
 Title		  :	SetAbsoluteFrame
 Usage		  : $self -> SetAbsoluteFrame()
 Procedure	  : set to TRUE (1) the __absolute_frame absolute, enables to draw in absolute coordinates
 Returns	  : none
 Args		  : none
 Gl

=cut
sub SetAbsoluteFrame
{
	my $self = shift;
	
	$self -> { __absolute_frame } = 1;	
		
}

=head2 SetRelativeFrame
 Title		  :	SetRelativeFrame
 Usage		  : $self -> SetRelativeFrame()
 Procedure	  : set to FALSE (0) the __absolute_frame absolute, enables to draw in coordinate defined by frame transformer
 Returns	  : none
 Args		  : none
 Gl

=cut
sub SetRelativeFrame
{
	my $self = shift;
	
	$self -> { __absolute_frame } = 0;	
		
}

=head2 GetFrame
 Title		  : GetFrame
 Usage		  : $self -> GetFrame()
 Procedure	  : returns  the frame object
 Args		  : none

=cut
sub GetFrame
{
	my $self  = shift;
	
	return $self->{ frame };
}

=head2 GetMax_y
 Title		  : GetMax_y
 Usage		  : $self -> GetMax_y()
 Procedure	  : returns the highest position in the image box
 Args		  : none

=cut
sub GetMax_y
{
	my $self  = shift;
	
	return $self->{ frame }{ y1 };
}
=head2 GetMin_y
 Title		  : GetMin_y
 Usage		  : $self -> GetMin_y()
 Procedure	  : returns the lowest position in the image box
 Args		  : none

=cut
sub GetMin_y
{
	my $self  = shift;
	
	return $self->{ frame }{ y0 };
}

=head2 ReversePositionInFrame
	Title        : 	ReversePositionInFrame
 	Usage        : 	$new_position = $o_image -> ReversePositionInFrame (axe, position)
 	Prerequisite : 	none
 	Function     : 	Swap a pixel position according to the actual frame bounds.
 	Returns      : 	new_position = the pixel of the revert value
 	Args         : 	axe 	 = axe on which we make this transformation ('x' or 'y')
 					position = the position value to transform.
 	Globals      : 	none
 	Attribute    :	frame

=cut
sub ReversePositionInFrame
{
	my $self = shift;
	my ($axe,$position) = (shift,shift);

	return $self->{frame} -> ReversePositionInFrame($axe,$position);
}

=head2 SwapCoordinates
	Title        : 	SwapCoordinates
 	Usage        : 	&Image::SwapCoordinates (n1,n2)
 	Prerequisite : 	none
 	Function     : 	Test both number and return them setting smallest at first.
 	Returns      : 	array (n1',n2')
 	Args         : 	(n1,n2) = positions to swap.
 	Globals      : 	none

=cut
sub SwapCoordinates
{
	my ($n1,$n2) = (shift,shift);

	return ($n1 < $n2) ? ($n1,$n2) : ($n2,$n1);
	
}

=head2 ReorderCoordinates
	Title        : 	
 	Usage        : 	&Image::ReorderCoordinates(n1,n2)
 	Prerequisite : 	none
 	Function     : 	Test both number and return them setting smallest at first.
 	Returns      : 	array (n1',n2')
 	Args         : 	(n1,n2) = positions to swap.

=cut
sub ReorderCoordinates
{
	my ($x0,$y0,$x1,$y1) = (shift,shift,shift,shift);

	return ($x0 < $x1) ? ($x0,$y0,$x1,$y1) : ($x1,$y1,$x0,$y0) ;
	
}

=head2 RectangleToAbsoluteCoordinates
 Title        : RectangleToAbsoluteCoordinates
 Usage        : RectangleToAbsoluteCoordinates
 Function     : see FrameTransFormer::RectangleCoordinateChange
 Returns      :
 Args         : none
 Globals      : frame

=cut
sub RectangleToAbsoluteCoordinates
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1) = @_ or return;

	return $self->{frame} -> RectangleCoordinateChange($x0,$y0,$x1,$y1)
}

=head2 PointCoordinateChange
 Title        : PointCoordinateChange
 Usage        : ($X,$Y) = $o_image->PointCoordinateChange($x,$y)
 Function     : see FrameTransFormer::PointCoordinateChange
 Returns      : (X,Y) absolute coordiantes on the image
 Args         : (x,y) relative coordinates on the image
=cut
sub PointCoordinateChange
{
	my $self = shift;
	my ($x,$y) = @_ or return;

	return $self->{frame} -> PointCoordinateChange($x,$y)
}


=head2 ToAbsoluteCoordinates
 Title        : AbsoluteCoordinates
 Usage        : AbsoluteCoordinates
 Prerequisite : none
 Function     :
 Returns      :
 Args         : none
 Globals      : frame

=cut
sub ToAbsoluteCoordinates
{
	my $self = shift;
	my ($x,$y) = @_ or return;

	return $self->{frame} -> ToAbsoluteCoordinates($x,$y);
}

=head2 ToAbsoluteRange
 Title        : ToAbsoluteRange
 Usage        : ToAbsoluteRange
 Prerequisite : none
 Function     :
 Returns      :
 Args         : none
 Globals      : frame

=cut
sub ToAbsoluteRange
{
	my ($self,$range,$axe) = @_ or return;

	if ($axe eq 'y')
	{
		return $self->{frame} -> ToAbsoluteYRange ($range);
	}
	elsif ($axe eq 'x')
	{
		return $self->{frame} -> ToAbsoluteXRange ($range);
	}
}

#------------------------------------------------#
#              Managing the images               #
#------------------------------------------------#

=head2 Rotate
 Title		  :	Rotate
 Usage		  : $self -> Rotate()
 Prerequisite :	none
 Procedure	  : Image rotation : portrait <-> landscape
 Returns	  : $self->{orientation}
 Args		  : none
 Globals	  : none

=cut
sub Rotate
{
	my $self = shift;

	$self->{orientation} = ($self->{orientation} eq 'portrait') ? 'landscape' : 'portrait';

	foreach my $o_layer ($self->GetLayersList())
	{
		$self -> copyRotate270($o_layer) if ($self->{orientation} eq 'landscape');
		$self -> copyRotate90($o_layer) if ($self->{orientation} eq 'portrait');
	}

	return $self->{orientation};
}

=head2 PrintImage

 Title        : PrintImage
 Usage        : PrintImage(file [,(ra_layers => \@a_layer, ...)]);
 Prerequisite : GD image initialized
 Procedure    : Print all layers (or a selection) into only one file name $file.
 				N layers = 1 file
 Returns      : file
 Args         : file    => file name.
 				@a_layer=> list of layers to print. By default we print all layers
 Attributes   : ra_layers

=cut
sub PrintImage
{
	my $self = shift;
	my $file = shift;

	# Some optional parameters may be specified here
	my %h_param = (
					ra_layers => [],
					offset_x  => 0,
					offset_y  => 0,   #__
					@_
					);

	my @a_layer = @{$h_param{ra_layers}};

	#------------------------------
	# Test file parameter
	return '' if (!defined $file);

	#------------------------------
	# Get the image format
	my $format = $1 if ($file =~ /\.(\w*)$/);
	return '' if ($format !~ /^(png|jpeg|gif)$/);

	#------------------------------
	# Test the argument of the format
	my $format_arg = '';
	$format_arg = 5  if ($format eq 'png');
	$format_arg = 80 if ($format eq 'jpeg');

	#------------------------------
	# Make a new blank Image
	my $o_image = new GD::Image ($self->{width}, $self->{height});
	#Suppressed to enable transparency again !!
	my $bg_color=$o_image->colorAllocate(@A_BACKGROUND_COLOR);
	if ($file =~/transparent/)
	{
		my $fg_color = $o_image->colorAllocate(@A_FOREGROUND_COLOR); # ? recuperer fg pour utiliser bg ! <---Seb.
		$o_image->transparent($bg_color);
	} else
	{
			$o_image->transparent(-1);
	}

	#------------------------------
	# Always merge the first layer to this blank image
	$o_image -> copy($self->GetLayer(),$h_param{offset_x},$h_param{offset_y},0,0,$self->{width}, $self->{height});

	#------------------------------
	# Merge other layers to this image
	@a_layer = $self->GetLayersList() if ( scalar(@a_layer) == 0 );
	foreach my $layer (@a_layer)
	{
		next if (!exists ${$self->{rh_layers}}{$layer});

		my $o_layer = $self->GetLayer($layer);
		$o_image -> copy($o_layer,$h_param{offset_x},$h_param{offset_y},0,0,$self->{width}, $self->{height});
	}

    #------------------------------
	# Print the merged image
	my $fs = new IO::File( $file, 'w' )
				or throw LipmError::IOException( $file, 'f' );
#     open (FILE, ">$file") or print STDERR "Erreur a la creation du fichier : $file\n";
	print $fs  $o_image->$format($format_arg);
#     print (FILE $o_image->$format($format_arg)) or print STDERR "Erreur a l'impression de l'image\n";
#     close FILE;
	$fs->close();

    return $file;
}

=head2 PrintLayers

 Title        : PrintLayers
 Usage        : PrintLayers(root,suffix,format [,format_argument]);
 Prerequisite : GD image initialized
 Procedure    : Print each layer in files named 'root/layer.suffix.format'
 				1 layer = 1 file
 Returns      : List of generated files.
 Args         : root   => the root were the file will be saved.
 				suffix => a suffix to add after layer name.
 				format => the output file format.
 				format_argument => png -> compression level (0-9)
 								   jpeg-> quality (0-100)
 Globals      : none

=cut
sub PrintLayers
{
	my $self = shift;
	my ($root,$suffix,$format,$format_arg) = (shift,shift,shift,shift);

	#------------------------------
	# Test parameters
	$root   = '' if (!defined $root);
	$suffix = '' if (!defined $suffix);
	$format = 'png' if (!defined $format or (defined $format && $format !~ /^(png|gif|jpeg)$/) );
	$format_arg = '' if (!defined $format_arg);

	#------------------------------
	# Test the argument of the format
	if ($format eq 'png' && $format_arg !~ /^\d$/)
	{
		$format_arg = 5;
	}
	elsif ($format eq 'jpeg' && $format_arg !~ /^\d{1,3}$/)
	{
		$format_arg = 80;
	}
	else
	{
		$format_arg = '';
	}

	#------------------------------
	# Convert all images to PNG or GIF and print it on standard output
	my @a_files = ();
    foreach my $layer ($self->GetLayersList())
    {
    	# Test if the layer is the default one, in this case we don't add 'default' on the file name.
    	$layer = '' if ($layer eq &LAYER ); #'default');

    	# We compute a file name AND a link name, adding a stamp to the file name
        # This trick is important because of the browser's cache: if the file is computed several times, we are sur that the LAST version will be sent to the browser
        my $cache_stamp = time.$$;
        my $file_name   = $layer.$suffix.'_'.$cache_stamp.'.'.$format;
        my $path_name   = $root."/".$file_name;
    	my $link_name   = $root."/".$layer.$suffix.'.'.$format;

    	my $o_layer = $self->GetLayer($layer);

		my $file = new IO::File( $path_name, 'w' )
			or throw LipmError::IOException( $path_name, 'f' );
		print $file $o_layer->$format($format_arg);
    	$file->close();

        unlink $link_name;
        symlink($file_name,$link_name);
    	push( @a_files, $link_name );
    }

    return @a_files;
}

#------------------------------------------------#
#              Drawing methods                   #
#------------------------------------------------#

=head2 filledTriangle
 Title        : filledTriangle
 Usage        :	$image -> filledTriangle(x0,y0,x1,y1,x2,y2,color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x0,y0 	=> first top point coordinates, with units like (px, base-pair)
 				x1,y1 	=> second top point coordinates, with units like (px, base-pair)
 				x2,y2 	=> third top point coordinates, with units like (px, base-pair)
 				color 	=> color of the triangle
 Globals      : gd

=cut
sub filledTriangle
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$x2,$y2,$color) = (shift,shift,shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self -> GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	unless ( $self->{ __absolute_frame} )
	{
		($x0,$y0) = $self->{frame}->PointCoordinateChange($x0,$y0);
		($x1,$y1) = $self->{frame}->PointCoordinateChange($x1,$y1);
		($x2,$y2) = $self->{frame}->PointCoordinateChange($x2,$y2);
	};

	# Set Polygon object
	my $polygon = new GD::Polygon();
	$polygon->addPt($x0,$y0);
	$polygon->addPt($x1,$y1);
	$polygon->addPt($x2,$y2);
	$o_layer->filledPolygon($polygon,$color);

	# Search and returns bounds (method $polygon->bounds() doesn't work)
	my ($x_min,$y_min,$x_max,$y_max) = ($x0,$y0,0,0);
	foreach my $i ($x0,$x1,$x2)
	{
		$x_min = $i if ( $i<$x_min );
		$x_max = $i if ( $i>$x_max );
	}
	foreach my $j ($y0,$y1,$y2)
	{
		$y_min = $j if ( $j<$y_min );
		$y_max = $j if ( $j>$y_max );
	}

	return ($x_min,$y_min,$x_max,$y_max);
}

=head2 filledEquilateralTriangle
 Title        : filledEquilateralTriangle
 Usage        :	$image -> filledEquilateralTriangle(x,y,side,direction,color)
 Prerequisite : none
 Procedure    :	New method to draw an equilateral triangle.
 				By default the triangle reference apex is direct to the left
 Returns      : none
 Args         : x,y 	  => reference apex coordinates, with units like (px, base-pair)
 				side 	  => length of sides, with units like (px, base-pair)
 				direction => direction of the reference apex : left,right,top,bottom (left by default)
 				color 	  => color of the triangle
 Globals      : none

=cut
sub filledEquilateralTriangle
{
	my $self = shift;
	my ($x,$y,$side,$direction,$color) = (shift,shift,shift,shift,shift) or return;
	my $o_layer = shift;

	my ($x1,$y1,$x2,$y2);
	my $o_frame = $self->{frame};
	my $factor  = $o_frame->GetFactor();
	my $h 		= sqrt(3/4 * $side**2);		# Triangle height

	# Set apex for each directions
	if ($direction eq 'top')
	{
		$x1 = $x + $side/2;
		$y1 = $y + $h/$factor;
		$x2 = $x - $side/2;
		$y2 = $y + $h/$factor;
	}
	elsif ($direction eq 'bottom')
	{
		$x1 = $x + $side/2;
		$y1 = $y - $h/$factor;
		$x2 = $x - $side/2;
		$y2 = $y - $h/$factor;
	}
	elsif ($direction eq 'right')
	{
		$x1 = $x - $h;
		$y1 = $y - $side/(2*$factor);
		$x2 = $x - $h;
		$y2 = $y + $side/(2*$factor);
	}
	else
	{
		$x1 = $x + $h;
		$y1 = $y - $side/(2*$factor);
		$x2 = $x + $h;
		$y2 = $y + $side/(2*$factor);
	}

	# Draw triangle and return bounds
	return $self->filledTriangle($x,$y,$x1,$y1,$x2,$y2,$color,$o_layer);
}

=head2 verticalCylinder
 Title        : verticalCylinder
 Usage        :	$image -> verticalCylinder(x0,y0,x1,y1,cheight,color)
 Prerequisite : none
 Procedure    :	draw a cylinder.
 Returns      : none
 Args         : x0,y0 	=> cylinder origin coordinates, in pixel
 				x1,y1 	=> cylinder end coordinates, in pixel
 				cheight	=> "circle height" : height of the elipse will be draw on top of the cylinder
 				color 	=> color of the cylinder
 				layer   => layer onto which draw
 Globals      : gd

=cut
sub verticalCylinder
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$circle_height,$color) = (shift,shift,shift,shift,shift,shift) or return;
	
	# Unless we choosed to, draw in absolute frame me make the appropriate coordinate transformation
	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );

	return $self->__verticalCylinder($x0,$y0,$x1,$y1,$circle_height,$color);
}

=head2 __verticalCylinder
 Title        : __verticalCylinder
 Usage        :	$image -> __verticalCylinder(x0,y0,x1,y1,cheight,color)
 Prerequisite : none
 Procedure    :	draw a cylinder.
 Returns      : none
 Args         : x0,y0 	=> cylinder origin coordinates, in pixel
 				x1,y1 	=> cylinder end coordinates, in pixel
 				cheight	=> "circle height" : height of the elipse will be draw on top of the cylinder
 				color 	=> color of the cylinder
 				layer   => layer onto which draw
 Globals      : gd

=cut
sub __verticalCylinder
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$circle_height,$color) = (shift,shift,shift,shift,shift,shift) or return;

	my $o_layer = $self -> GetCurrentLayer();

	# Set colors
	my $grey  = $self -> colorAllocate(210,210,210,$o_layer);  
	my $white = $self -> colorAllocate(255,255,255,$o_layer);

	# Take care of padding
	my $padding = $circle_height/2 - 1;
	$y0 = $y0 + $circle_height/2 -1;
	$y1 = $y1 - $circle_height/2;

	# Graphic parameters.
	my $width = abs($x1-$x0);
	my $cx = $x0 + $width/2;

	# Draw vertical lines (2px width).
	$o_layer -> setThickness(2);
	$o_layer -> line($x0,$y0,$x0,$y1,$color);
	$o_layer -> line($x1+1,$y0,$x1+1,$y1,$color);
	$o_layer -> setThickness(1);

	# Delete previous data on bottom :	(je ne comprends pas ce choix (Seb) )
	$o_layer -> filledArc($cx,$y1,$width,$circle_height,0,360,$white);
	$o_layer -> filledRectangle($x0+1,$y1-$circle_height,$x1-1,$y1,$white);

	# Draw top elipse and bottom arc.
	$o_layer -> filledArc($cx,$y0,$width,$circle_height,0,360,$grey);	#elipse pleine
	$o_layer -> arc($cx,$y0,$width+2,$circle_height,0,360,$color);		#contour elipse
	$o_layer -> arc($cx,$y0,$width+2,$circle_height,0,180,$color);		#demi elipse superieure
	$o_layer -> arc($cx,$y1,$width+2,$circle_height,0,180,$color);		# demi elipse inferieure

	# Correction of "bad pixels" of the arc :
	$o_layer -> setPixel($cx,$y0-$circle_height/2,$o_layer->getPixel($cx+1,$y0-$circle_height/2) );
	$o_layer -> setPixel($cx,$y0+$circle_height/2,$white ) if (($y1-$y0) > 1);
	$o_layer -> setPixel($cx,$y1+$circle_height/2,$o_layer->getPixel($cx+1,$y1+$circle_height/2) );

	# Fill the cylinder with the background color:
	# commente pour laisser apparaitre ce qu'on met sur la meme couche. (Seb)
# 	$o_layer -> fillToBorder($x0+5,$y0+5,$color,$white) if (($y1-$y0) >2 );

	# Return the drawing coordinates :
	$y0 = $y0+$circle_height/2;
	$y1 = $y1+$circle_height/2;

	return ($x0,$y0,$x1,$y1);
}

=head2 LeftVerticalCylinder
 Title        : LeftVerticalCylinder
 Usage        :	$image -> LeftVerticalCylinder(x0,y0,x1,y1,$ellpise_height,$from_alpha, $to_alpha, $color)
 Prerequisite : none
 Procedure    :	draw a half-part of a cylinder.
 Returns      : none
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				cheight	=> "circle height" : height of the elipse will be draw
 				color 	=> color of the rectangle
 				factor	=> which part of the full cylinder will be draw.
 				side	=> 'right' or 'left' : draw starting to the left or to the right
 Globals      : gd

=cut
sub LeftVerticalCylinder
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$ellipse_height,$color) = (shift,shift,shift,shift,shift,shift) or return;
		
	my ( $X0, $Y0, $X1, $Y1 ) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
		
	my $Cx = int ( ($X0 + $X1)/2 );
	my $ellipse_width = $X1-$X0;
	my $height = $Y1-$Y0;
		
	# We draw one arc by pixel of 'height'
	my $iterations = ($height%10 < 5) ? $height : $height+1;

	for (my $i=0; $i<$iterations; $i++)
	{
		$self->{ __current_layer } -> arc($Cx,$Y0+$i,$ellipse_width,$ellipse_height,90,180,$color);
	}	

	my @area_coord = ( $X0,$Y0,$Cx,$Y1 );
	
	return @area_coord;
}

=head2 RigthVerticalCylinder
 Title        : RigthVerticalCylinder
 Usage        :	$image -> RigthVerticalCylinder(x0,y0,x1,y1,$ellpise_height,$from_alpha, $to_alpha, $color)
 Prerequisite : none
 Procedure    :	draw a half-part of a cylinder.
 Returns      : none
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				cheight	=> "circle height" : height of the elipse will be draw
 				color 	=> color of the rectangle
 				factor	=> which part of the full cylinder will be draw.
 				side	=> 'right' or 'left' : draw starting to the left or to the right
 Globals      : gd

=cut
sub RightVerticalCylinder
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$ellipse_height,$color) = (shift,shift,shift,shift,shift,shift) or return;
		
	my ( $X0, $Y0, $X1, $Y1 ) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
		
	my $Cx = int ( ($X0 + $X1)/2 );	
	my $ellipse_width = $X1-$X0;
	my $height = $Y1-$Y0;

	# We draw one arc by pixel of 'height'
	my $iterations = ($height%10 < 5) ? $height : $height+1;
		
	for (my $i=0; $i<$iterations; $i++)
	{
		$self->{ __current_layer } -> arc($Cx,$Y0+$i,$ellipse_width,$ellipse_height,0,90,$color);
	}	
		
	my @area_coord = ( $Cx,$Y0,$X1,$Y1 );	
		
	return @area_coord;
}

=head2 RigthJustifiedHalfArrow
 Title        : RigthJustifiedHalfArrow
 Usage        :	$image ->RightJustifiedHalfArrow(x0,y0,x1,y1,$arrow_thickness, $color)
 Prerequisite : none
 Procedure    :	draw a half-part of a cylinder.
 Returns      : none
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				cheight	=> "circle height" : height of the elipse will be draw
 				color 	=> color of the rectangle
 				factor	=> which part of the full cylinder will be draw.
 				side	=> 'right' or 'left' : draw starting to the left or to the right
 Globals      : gd

=cut
sub RightJustifiedHalfArrow
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$direction,$arrow_size,$color) = (shift,shift,shift,shift,shift,shift,shift) or return;
			
	my ( $X0, $Y0, $X1, $Y1 ) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
		
	my $o_layer = $self -> GetCurrentLayer();

	if ( ($Y1 - $Y0) <= $arrow_size ) 
	{	 # No triangle is drawn
		$o_layer -> filledRectangle($X0,$Y0,$X1,$Y1,$color);	
	} 
	else
	{
		my ( $A_x,$A_y,$B_x,$B_y, $C_x, $C_y );
	
		if ( $direction > 0 )
		{
			$o_layer -> filledRectangle($X0,$Y0,$X1,$Y1 -$arrow_size ,$color);
			($A_x,$A_y) = ( $X1 - $arrow_size , $Y1 - $arrow_size );
			($B_x,$B_y) = ( $X1 , $Y1 - $arrow_size );
			($C_x,$C_y) = ( $X1, $Y1 );
		}
		else
		{
			$o_layer -> filledRectangle($X0,$Y0 + $arrow_size,$X1,$Y1 ,$color);
			($A_x,$A_y) = ( $X1, $Y0 );
			($B_x,$B_y) = ( $X1, $Y0 + $arrow_size );
			($C_x,$C_y) = ( $X1 - $arrow_size , $Y0 + $arrow_size);		 
		}
		$self->SetAbsoluteFrame();
		$self -> filledTriangle( $A_x,$A_y,$B_x,$B_y, $C_x, $C_y, $color);
		$self->SetRelativeFrame();
	}
		
	return  ($X0,$Y0,$X1,$Y1 );
}

=head2 LeftJustifiedHalfArrow
 Title        : LeftJustifiedHalfArrow
 Usage        :	$image ->LeftJustifiedHalfArrow(x0,y0,x1,y1,$arrow_thickness, $color)
 Prerequisite : none
 Procedure    :	draw a half-part of a cylinder.
 Returns      : none
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				cheight	=> "circle height" : height of the elipse will be draw
 				color 	=> color of the rectangle
 				factor	=> which part of the full cylinder will be draw.
 				side	=> 'right' or 'left' : draw starting to the left or to the right

=cut
sub LeftJustifiedHalfArrow
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$direction,$arrow_size,$color) = (shift,shift,shift,shift,shift,shift,shift) or return;
			
	my ( $X0, $Y0, $X1, $Y1 ) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
		
	my $o_layer = $self -> GetCurrentLayer();
							
	if ( ($Y1 - $Y0) <= $arrow_size ) 
	{	 # No triangle is drawn
		$o_layer -> filledRectangle($X0,$Y0,$X1,$Y1,$color)		
	} 
	else
	{			
		my ( $A_x,$A_y,$B_x,$B_y, $C_x, $C_y );
	
		if ( $direction > 0 )
		{
			$o_layer -> filledRectangle($X0,$Y0,$X1,$Y1 -$arrow_size,$color);		
			($A_x,$A_y) = ( $X0, $Y1 - $arrow_size );
			($B_x,$B_y) = ( $X0 + $arrow_size, $Y1 - $arrow_size );
			($C_x,$C_y) = ( $X0, $Y1 );
		}
		else
		{
			$o_layer -> filledRectangle($X0,$Y0+$arrow_size,$X1,$Y1,$color);
			($A_x,$A_y) = ( $X0, $Y0 );
			($B_x,$B_y) = ( $X0, $Y0 + $arrow_size );
			($C_x,$C_y) = ( $X0 + $arrow_size , $Y0 + $arrow_size);		 
		}

		$self->SetAbsoluteFrame();
		$self -> filledTriangle( $A_x,$A_y,$B_x,$B_y, $C_x, $C_y, $color);
		$self->SetRelativeFrame();		
		
	}
		
	return ( $X0,$Y0,$X1,$Y1 );
}

=head2 borderedFilledRectangle
 Title        : borderedFilledRectangle
 Usage        :	$image -> borderedFilledRectangle(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				color 	=> color of the rectangle
 Globals      : gd

=cut
sub borderedFilledRectangle
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$color) = (shift,shift,shift,shift,shift) or return;
		
	my $o_layer = $self -> GetCurrentLayer();
	my $border_color = $self->colorAllocate(0,0,0,$o_layer);
	
	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation

	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );

	$o_layer -> rectangle($x0,$y0,$x1,$y1,$border_color);
	$o_layer -> filledRectangle($x0+1,$y0+1,$x1-1,$y1-1,$color);

	return ($x0,$y0,$x1,$y1);
}

=head2 ribbon
 Title        : ribbon
 Usage        :	$image -> ribbon(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         :
 Globals      : gd

=cut
sub ribbon
{
	my $self = shift;
	my ($x_topleft,$y_topleft,$x_botleft,$y_botleft,$x_topright,$y_topright,$x_botright,$y_botright) = (shift,shift,shift,shift,shift,shift,shift,shift);
	my $color = shift;
	my $layer = shift;

	my $o_layer = $self -> GetLayer($layer);
	my @a_coord;

	# Get absolute coordinates values :
	($x_topleft,$y_topleft) = $self->{frame}->ToAbsoluteCoordinates($x_topleft,$y_topleft);
	($x_botleft,$y_botleft) = $self->{frame}->ToAbsoluteCoordinates($x_botleft,$y_botleft);
	($x_topright,$y_topright) = $self->{frame}->ToAbsoluteCoordinates($x_topright,$y_topright);
	($x_botright,$y_botright) = $self->{frame}->ToAbsoluteCoordinates($x_botright,$y_botright);
	if ( $y_topleft > $y_botleft )
	{
		my ($x_temp,$y_temp)	= ($x_topleft,$y_topleft);
		($x_topleft,$y_topleft) = ($x_botleft,$y_botleft);
		($x_botleft,$y_botleft) = ($x_temp,$y_temp);
	}
	if ( $y_topright > $y_botright )
	{
		my ($x_temp,$y_temp)	  = ($x_topright,$y_topright);
		($x_topright,$y_topright) = ($x_botright,$y_botright);
		($x_botright,$y_botright) = ($x_temp,$y_temp);
	}


	# Draw a simple line on middle if height of one side is inferior to 2px
	if ( ($y_botleft-$y_topleft) < 2 or ($y_botright-$y_topright) < 2 )
	{
		my $x_midleft = ($x_botleft > $x_topleft) ? $x_topleft + ($x_botleft-$x_topleft)/2 : $x_botleft + ($x_topleft-$x_botleft)/2;
		my $y_midleft = $y_topleft + ($y_botleft-$y_topleft)/2;
		my $x_midright = ($x_botright > $x_topright) ? $x_topright + ($x_botright-$x_topright)/2 : $x_botright + ($x_topright-$x_botright)/2;
		my $y_midright = $y_topright + ($y_botright-$y_topright)/2;

		$o_layer -> line($x_midleft,$y_midleft,$x_midright,$y_midright,$color);
	}

	# Set Polygon object
	my $polygon = new GD::Polygon();
	$polygon->addPt($x_topleft,$y_topleft);
	$polygon->addPt($x_topright,$y_topright);
	$polygon->addPt($x_botright,$y_botright);
	$polygon->addPt($x_botleft,$y_botleft);
	$o_layer->filledPolygon($polygon,$color);

	return ($x_topleft,$y_topleft,$x_botright,$y_botright);
}

=head2 ribbon_OLD
 Title        : ribbon_OLD
 Usage        :	$image -> ribbon_OLD(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         :
 Globals      : gd

=cut
sub ribbon_OLD
{
	my $self = shift;
	my ($x_topleft,$y_topleft,$x_botleft,$y_botleft,$x_topright,$y_topright,$x_botright,$y_botright) = (shift,shift,shift,shift,shift,shift,shift,shift);
	my $color = shift;
	my $layer = shift;

	my $o_layer = $self -> GetLayer($layer);
	my @a_coord;

	# Get absolute coordinates values :
	($x_topleft,$y_topleft) = $self->{frame}->ToAbsoluteCoordinates($x_topleft,$y_topleft);
	($x_botleft,$y_botleft) = $self->{frame}->ToAbsoluteCoordinates($x_botleft,$y_botleft);
	($x_topright,$y_topright) = $self->{frame}->ToAbsoluteCoordinates($x_topright,$y_topright);
	($x_botright,$y_botright) = $self->{frame}->ToAbsoluteCoordinates($x_botright,$y_botright);
	if ( $y_topleft > $y_botleft )
	{
		my ($x_temp,$y_temp)	= ($x_topleft,$y_topleft);
		($x_topleft,$y_topleft) = ($x_botleft,$y_botleft);
		($x_botleft,$y_botleft) = ($x_temp,$y_temp);
	}
	if ( $y_topright > $y_botright )
	{
		my ($x_temp,$y_temp)	  = ($x_topright,$y_topright);
		($x_topright,$y_topright) = ($x_botright,$y_botright);
		($x_botright,$y_botright) = ($x_temp,$y_temp);
	}


	# Draw a simple line on middle if height of one side is inferior to 2px
	if ( ($y_botleft-$y_topleft) < 2 or ($y_botright-$y_topright) < 2 )
	{
		my $x_midleft = ($x_botleft > $x_topleft) ? $x_topleft + ($x_botleft-$x_topleft)/2 : $x_botleft + ($x_topleft-$x_botleft)/2;
		my $y_midleft = $y_topleft + ($y_botleft-$y_topleft)/2;
		my $x_midright = ($x_botright > $x_topright) ? $x_topright + ($x_botright-$x_topright)/2 : $x_botright + ($x_topright-$x_botright)/2;
		my $y_midright = $y_topright + ($y_botright-$y_topright)/2;

		$o_layer -> line($x_midleft,$y_midleft,$x_midright,$y_midright,$color);
	}

	# Set Polygon object
	my $polygon = new GD::Polygon();
	$polygon->addPt($x_topleft,$y_topleft);
	$polygon->addPt($x_topright,$y_topright);
	$polygon->addPt($x_botright,$y_botright);
	$polygon->addPt($x_botleft,$y_botleft);
	$o_layer->filledPolygon($polygon,$color);

	return ($x_topleft,$y_topleft,$x_botright,$y_botright);
}


=head2 verticalHalfArrow
 Title        : verticalHalfArrow
 Usage        :	$image -> verticalHalfArrow(x,width,start,end,color [,direction][,orientation])
 Prerequisite : none
 Procedure    :
 Returns      : the index of the new color
 Args         : x		=> x position
 				start	=> start
 				end		=> end
 				color	=> GD::Image color index
 				direction	=> 'up'/'down' or '+'/'-'. By default 'down'
 				orientation	=> 'right' or 'left'. Position of the arrow. 'left by default'
 Globals      : none

=cut
sub verticalHalfArrow
{
	my $self = shift;
	my ($x,$width,$start,$end,$color) = (shift,shift,shift,shift,shift) or return;
	my ($direction, $orientation) = ('up','left');
	my $o_layer = '';

	while (defined $_[0])
	{
		my $temp_param = shift;
		if ($temp_param =~ s/^(up)|(\+)$/up/ or $temp_param =~ s/^(down)|(\-)$/down/)
		{
			$direction = $temp_param;
		}
		elsif ($temp_param =~ /^(left)|(right)$/)
		{
			$orientation = $temp_param;
		}
		else
		{
			$o_layer = $temp_param;
		}
	}

	# Test start/end
	($start,$end) = &SwapCoordinates($start,$end);


	# Test width value : if the drawing triangle is superior to the range of representation, we decrease it
	my $arrow_dimensions = $self->{frame} -> ToRelativeYRange($width*4);
	my $arrow_bool =  ( ($end-$start) < $arrow_dimensions ) ? 0 : 1;
	my $pixel_shift = $width-$width/2;

	# Calculate arrow positions :
	my $x1_line = ($orientation eq 'left') ? $x-$width : $x+$width;
	my $x1_arrow = ($orientation eq 'left') ? $x-$width*3 : $x+$width*3;
	my $y0_arrow = ($direction eq 'down') ? $start : $end;
	my $y1_arrow = ($direction eq 'down') ? $y0_arrow+$arrow_dimensions : $y0_arrow-$arrow_dimensions;
	my $y0_line = ($direction eq 'down') ? $y1_arrow : $start;
	my $y1_line = ($direction eq 'down') ? $end : $y1_arrow;
	($y0_line,$y1_line) = ($start,$end) if ($arrow_bool == 0);

	#Drawn line
	$self -> filledRectangle($x,$y0_line,$x1_line,$y1_line,$color,$o_layer);
	# Draw the 'half' arrow.
	$self -> filledTriangle($x,$y0_arrow,$x1_arrow,$y1_arrow,$x,$y1_arrow,$color,$o_layer) if ($arrow_bool == 1);


	# Return coordinates :
	my ($x0, $x1) = ($orientation eq 'left') ? ($x1_arrow,$x) : ($x,$x1_arrow);

	return ($self->{frame}->ToAbsoluteCoordinates($x0,$start), $self->{frame}->ToAbsoluteCoordinates($x1,$end));
}

=head2 verticalHalfArrow
 Title        : verticalHalfArrow
 Usage        :	$image -> verticalHalfArrow(x,width,start,end,color [,direction][,orientation])
 Prerequisite : none
 Procedure    :
 Returns      : the index of the new color
 Args         : x		=> x position
 				start	=> start
 				end		=> end
 				color	=> GD::Image color index
 				direction	=> 'up'/'down' or '+'/'-'. By default 'down'
 				orientation	=> 'right' or 'left'. Position of the arrow. 'left by default'
 Globals      : none

=cut
sub verticalHalfArrow_OLD
{
	my $self = shift;
	my ($x,$width,$start,$end,$color) = (shift,shift,shift,shift,shift) or return;
	my ($direction, $orientation) = ('up','left');
	my $o_layer = '';

	while (defined $_[0])
	{
		my $temp_param = shift;
		if ($temp_param =~ s/^(up)|(\+)$/up/ or $temp_param =~ s/^(down)|(\-)$/down/)
		{
			$direction = $temp_param;
		}
		elsif ($temp_param =~ /^(left)|(right)$/)
		{
			$orientation = $temp_param;
		}
		else
		{
			$o_layer = $temp_param;
		}
	}

	# Test start/end
	($start,$end) = &SwapCoordinates($start,$end);


	# Test width value : if the drawing triangle is superior to the range of representation, we decrease it
	my $arrow_dimensions = $self->{frame} -> ToRelativeYRange($width*4);
	my $arrow_bool =  ( ($end-$start) < $arrow_dimensions ) ? 0 : 1;
	my $pixel_shift = $width-$width/2;

	# Calculate arrow positions :
	my $x1_line = ($orientation eq 'left') ? $x-$width : $x+$width;
	my $x1_arrow = ($orientation eq 'left') ? $x-$width*3 : $x+$width*3;
	my $y0_arrow = ($direction eq 'down') ? $start : $end;
	my $y1_arrow = ($direction eq 'down') ? $y0_arrow+$arrow_dimensions : $y0_arrow-$arrow_dimensions;
	my $y0_line = ($direction eq 'down') ? $y1_arrow : $start;
	my $y1_line = ($direction eq 'down') ? $end : $y1_arrow;
	($y0_line,$y1_line) = ($start,$end) if ($arrow_bool == 0);

	#Drawn line
	$self -> filledRectangle($x,$y0_line,$x1_line,$y1_line,$color,$o_layer);
	# Draw the 'half' arrow.
	$self -> filledTriangle($x,$y0_arrow,$x1_arrow,$y1_arrow,$x,$y1_arrow,$color,$o_layer) if ($arrow_bool == 1);


	# Return coordinates :
	my ($x0, $x1) = ($orientation eq 'left') ? ($x1_arrow,$x) : ($x,$x1_arrow);

	return ($self->{frame}->ToAbsoluteCoordinates($x0,$start), $self->{frame}->ToAbsoluteCoordinates($x1,$end));
}

=head2 brokenArc
 Title        : brokenArc
 Usage        :	$image -> brokenArc(x0,y0,x1,y1,width,side,color)
 Prerequisite : none
 Procedure    :	new method
 Returns      : none
 Args         : x0,y0	-> first point
 				x1,y1	-> second point
 				width	-> width of the broken-arc
 				side	-> right or left
 				color	-> GD reference color
 Globals      : gd

=cut
sub brokenArc
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$width,$side,$color) = (shift,shift,shift,shift,shift,shift,shift);
	my $o_layer = shift;

	# Calculate the third point coordinates :
	my $x2 = ($side eq 'left') ? $x0-$width : $x0+$width;
	my $y2 = $y0 + ($y1-$y0)/2;

	# Draw lines :
	$self -> GetLayer($o_layer) -> line ($self->{frame}->ToAbsoluteCoordinates($x0,$y0),$self->{frame}->ToAbsoluteCoordinates($x2,$y2),$color);
	$self -> GetLayer($o_layer) -> line ($self->{frame}->ToAbsoluteCoordinates($x2,$y2),$self->{frame}->ToAbsoluteCoordinates($x1,$y1),$color);

	return;
}

=head2 brokenLine
 Title        : brokenLine
 Usage        :	$image -> brokenLine(x0,y0,x1,y1,width,$color)
 Prerequisite : none
 Procedure    :	new method
 Returns      : none
 Args         : x0,y0 	=> origin coordinates, with units like (px, base-pair)
				x1,y1 	=> end coordinates, with units like (px, base-pair)
				width	=> space between oblic line and horizontal line.
 				color 	=> color
 Globals      : gd

=cut
sub brokenLine
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$width,$color) = (shift,shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self->{ __current_layer } ;

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );
		
	$o_layer -> line($x0,$y0,$x0+$width,$y0,$color);
	$o_layer -> line($x0+$width,$y0,$x1-$width,$y1,$color);
	$o_layer -> line($x1,$y1,$x1-$width,$y1,$color);

	return ($x0,$y0,$x1,$y1);
}

#------------------------------------------------#
#              Overloaded GD  methods            #
#------------------------------------------------#

=head2 arc
 Title        : arc
 Usage        :	$image -> arc(cx,cy,width,height,start,end,$color)
 Prerequisite : none
 Procedure    :	overload the colorAllocate GD method
 Returns      : none
 Args         : cx,cy 			=> center arc coordinates, with units like (px, base-pair)
 				width,height 	=> width and height of the arc, with units like (px, base-pair) !!! For now, only in pixels
 				start,end		=> The portion of the ellipse covered by the arc, start/end are given in degree (0-360)
 				color 			=> color
 Globals      : gd

=cut
sub arc
{
	my $self    = shift;
	my( $cx,$cy, $width,$height, $start,$end, $color ) = @_;
	
	my $o_layer = $self -> GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	if( ! $self->{ __absolute_frame } )
	{
		( $cx,$cy ) = $self->{ frame }->PointCoordinateChange( $cx,$cy );

		# je ne fais pas la transformation parce que je n'en ai pas besoin
		# et que ca ne marche pas bien
# 		( $width,$height ) = $self->{ frame }->PointCoordinateChange( $width,$height );
	}

	$o_layer->arc( $cx,$cy, $width,$height, $start,$end, $color);

	return ( $cx,$cy, $width,$height );
}

=head2 fill
 Title        : fill
 Usage        :	$image -> fill(x,y,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x,y 	=> origin coordinates, with units like (px, base-pair)
 				color 	=> color
 Globals      : gd

=cut
sub fill
{
	my $self = shift;
	my ($x,$y,$color) = (shift,shift,shift) or return;
	my $o_layer = shift;

	return $self -> GetLayer($o_layer) -> fill($self->{frame}->ToAbsoluteCoordinates($x,$y),$color);
}

=head2 filledRectangle
 Title        : filledRectangle
 Usage        :	$image -> filledRectangle(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using RectangleToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				color 	=> color of the rectangle

=cut
sub filledRectangle
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$color) = (shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self -> GetCurrentLayer();
	
	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );

# warn Dumper { image => { A => [$x0,$y0], B => [$x1,$y1] } };
	$o_layer-> filledRectangle($x0,$y0,$x1,$y1,$color);

	return ($x0,$y0,$x1,$y1);
}


=head2 FilledCurvedRectangle

 Usage        :	( $x0,$y0, x1,y1 ) = $image->FilledCurvedRectangle( x0,y0, x1,y1, $h, $color )
 Prerequisite : none
 Procedure    : draw a filledRectangle like but it is curved.
 				It can be used to draw filledRectangle onto the cylinder of genes.
 				It uses RectangleToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				$h		=> height of the curved (the DNA cylinder uses width_gene/3)
 				color 	=> color of the rectangle

 Note : New implementation force to have $x0 < $x1 and $y0 < $y1
        $color cannot be a gdBrushed nor a gdTiled

=cut

sub FilledCurvedRectangle
{
	my $self = shift;
	my( $x0,$y0, $x1,$y1, $h, $color ) = @_;

	my $o_layer = $self->GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	if( ! $self->{ __absolute_frame} )
	{
		($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
	}

	# for arc
	my $w  = ($x1-$x0);
	my $cx = int( 0.5 + $x0 + $w/2 ); # pas normal que $x0 puisse avoir des demi pixels.

	# I thought that almost nobody will zoom enough to show big SNP.
	# There's no use to have a view where a base is 50 pixels high.
	# even though if one do that, there will be not that much SNP to display.
	# that why i use this loop
	for( my $y = $y0; $y <= $y1; ++$y )
	{
		$o_layer->arc( $cx,$y, $w,$h, 0,180, $color ); # top arc
	}

# 	# bottom-top
# 	$o_layer->arc( $cx,$y0, $w,$h, 0,180, $color ); # top arc
# 	$o_layer->arc( $cx,$y1, $w,$h, 0,180, $color ); # top arc
# 
# 	# left-right
# 	$o_layer->line( $x0,$y0, $x0,$y1, $color );
# 	$o_layer->line( $x1,$y0, $x1,$y1, $color );
# 
# 	# fill
# 	if( ( $y0+1 < $y1 ) && ( $x0+1 < $x1 ) )
# 	{
# 		my $y = ($y1-$y0)/2 + $y0;
# 		my $x = ($x1-$x0)/2 + $x0;
# 		$o_layer->fill( $x,$y, $color );
# 	}

	

	return ($x0,$y0,$x1,$y1);
}


=head2 CurvedRectangle

 Usage        :	( $x0,$y0, x1,y1 ) = $image->CurvedRectangle( x0,y0, x1,y1, $h, $color )
 Prerequisite : none
 Procedure    : draw a Rectangle but it is curved to wrap the cylinder.
 				It uses RectangleToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : absolute coordinates (x0,y0,x1,y1)
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				$h		=> height of the curved (the DNA cylinder uses width_gene/3)
 				color 	=> color of the rectangle

=cut

sub CurvedRectangle
{
	my $self = shift;
	my( $x0,$y0, $x1,$y1, $h, $color ) = @_;

	my $o_layer = $self->GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	if( ! $self->{ __absolute_frame} )
	{
		($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1);
	}

	$x0 += 1;
	$x1 -= 1;

	# for arc
	my $w  = ($x1-$x0);
	my $cx = int( 0.5 + $x0 + $w/2 ); # pas normal que $x0 puisse avoir des demi pixels.

	# arcs top and bottom
	$o_layer->arc( $cx,$y0, $w,$h, 0,180, $color ); # top
	$o_layer->arc( $cx,$y1, $w,$h, 0,180, $color ); # bottom

	# vertical lines
	$o_layer->line( $x0,$y0, $x0,$y1, $color ); # left
	$o_layer->line( $x1,$y0, $x1,$y1, $color ); # right

	return ($x0,$y0,$x1,$y1);
}


=head2 line
 Title        : line
 Usage        :	$image -> line(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x0,y0 	=> origin coordinates, with units like (px, base-pair)
				x1,y1 	=> end coordinates, with units like (px, base-pair)
 				color 	=> color
 Globals      : gd

=cut
sub line
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$color) = (shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self -> GetCurrentLayer();
# warn Dumper [ [$x0,$x1],[$y0,$y1] ];
	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	unless ( $self->{ __absolute_frame} )
	{
		($x0,$y0)  = $self->{frame} ->PointCoordinateChange($x0,$y0);
		($x1,$y1)  = $self->{frame} ->PointCoordinateChange($x1,$y1);		
	}
# warn Dumper [ [$x0,$x1],[$y0,$y1] ];

	$o_layer -> line($x0,$y0,$x1,$y1,$color);

	return ($x0,$y0,$x1,$y1);
}

=head2 dashedLine
 Title        : dashedLine
 Usage        :	$image -> dashedLine(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x0,y0 	=> origin coordinates, with units like (px, base-pair)
				x1,y1 	=> end coordinates, with units like (px, base-pair)
 				color 	=> color
 Globals      : gd

=cut
sub dashedLine
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$color) = (shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self -> GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );

	return $o_layer -> dashedLine($x0,$y0,$x1,$y1,$color);
}

=head2 rectangle
 Title        : rectangle
 Usage        :	$image -> rectangle(x0,y0,x1,y1,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x0,y0 	=> rectangle origin coordinates, with units like (px, base-pair)
 				x1,y1 	=> rectangle end coordinates, with units like (px, base-pair)
 				color 	=> color of the rectangle
 Globals      : gd

=cut
sub rectangle
{
	my $self = shift;
	my ($x0,$y0,$x1,$y1,$color) = (shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self -> GetCurrentLayer();

	# Unless we choosed to draw in absolute frame me make the appropriate coordinate transformation
	($x0,$y0,$x1,$y1) = $self->RectangleToAbsoluteCoordinates($x0,$y0,$x1,$y1) unless ( $self->{ __absolute_frame} );

	$o_layer -> rectangle($x0,$y0,$x1,$y1,$color);

	return ($x0,$y0,$x1,$y1);
}

=head2 setPixel
 Title        : setPixel
 Usage        :	$image -> setPixel(x,y,$color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : x,y 	=> point coordinates, with units like (px, base-pair)
 				color 	=> color
 Globals      : gd

=cut
sub setPixel
{
	my $self = shift;
	my ($x,$y,$color) = (shift,shift,shift) or return;
	my $o_layer = shift;

	

	return $self -> GetLayer($o_layer) -> setPixel($self->{frame}->ToAbsoluteCoordinates($x,$y),$color);
}

#------------------------------------------------#
#              Drawing strings                   #
#------------------------------------------------#


=head2 stringVerticalAligned
 Title        : stringVerticalAligned
 Usage        :	$image -> stringVerticalAligned(font,x,y,text,color,layer, [justification => 0|1],[underline => ])
 Prerequisite : none
 Procedure    :	New method, draw the text 'text' aligned on 'y' position as the vertical center
 Returns      : bounds
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				text 	=> text to write
 				color 	=> color of the text
 				justification 	=> 'left', 'right' or 'center'. 'left' by default. Text is justified according to the x coordinate.
 				underline	=> boolean (0 or 1), to underline this text.
 Globals      : gd

=cut
sub stringVerticalAligned
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my %h_param= ( side => 'left',
	               underline      => 0,
	               absolute_frame => $self->{ __absolute_frame},
	               @_
	               );
	my $justification = $h_param{ side };
	my $underline = $h_param{ underline };
	
	# Unless we choosed to draw in absolute frame, me make the appropriate coordinate transformation
	($x,$y) = $self->{ frame }->PointCoordinateChange( $x, $y)  unless ( $h_param{ absolute_frame } );
		
	return $self->__stringVerticalAligned($font,$x,$y,$text,$color,$justification,$underline);
}

=head2 __stringVerticalAligned
 Title        : __stringVerticalAligned
 Usage        :	$image -> __stringVerticalAligned(font,x,y,text,color [,justification][,underline])
 Prerequisite : none
 Procedure    :	New method, draw the text 'text' aligned on 'y' position as the vertical center
 Returns      : bounds
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				text 	=> text to write
 				color 	=> color of the text
 				justification 	=> 'left', 'right' or 'center'. 'left' by default. Text is justified according to the x coordinate.
 				underline	=> boolean (0 or 1), to underline this text.
 Globals      : gd

=cut
sub __stringVerticalAligned
{
	my $self = shift;
	my ($font,$x,$y,$text,$color,$justification,$underline) = (shift,shift,shift,shift,shift,shift,shift) or return;
	
	my $o_layer = $self ->{ __current_layer };

	# Justification :
	if ( $justification eq 'right' )
	{
		$x = $x - length($text)*$font->width();
	}
	elsif ( $justification eq 'center' )
	{
		$x = $x - length($text)*$font->width() /2;
	}

	# Aligned text with y as middle
	$y -= $font->height()/2;

	# Calculate bounds
	my $x1 = $x + length($text)*$font->width();
	my $y1 = $y + $font->height();

	# Underline text :
	if ($underline)
	{
		my $underline_color = $self -> colorAllocateAlpha('yellow',40,$o_layer);
		$o_layer -> filledRectangle($x,$y,$x1,$y1,$underline_color);
	}

	# Write text :
	$o_layer -> string($font,$x,$y,$text,$color);

	return ($x,$y,$x1,$y1);
}

=head2 string
 Title        : string
 Usage        :	$image -> string(font,x,y,text,color [,justification])
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : bounds
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				text 	=> text to write
 				color 	=> color of the text
 				justification 	=> 'left', 'right' or 'center'. 'left' by default. Text is justified according to the x coordinate.
 Globals      : gd

=cut
sub string
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my $justification = (defined $_[0] && $_[0] =~ /(right)|(left)|(center)/) ? shift : 'left';
	my $o_layer = shift;

	return if ($self->{frame}->MaskHeight() < $font->height());

	# Calculate absolute coordinates :
	($x,$y) = $self->{frame}->ToAbsoluteCoordinates($x,$y);

	# Justification :
	if ( $justification eq 'right' )
	{
		$x = $x - length($text)*$font->width();
	}
	elsif ( $justification eq 'center' )
	{
		$x = $x - length($text)*$font->width() /2;
	}

	# Calculate bounds
	my $x1 = $x + length($text)*$font->width();
	my $y1 = $y + $font->height();

	# Write text :
	$self -> GetLayer($o_layer) -> string($font,$x,$y,$text,$color);

	# Return bounds :
	return ($x,$y,$x1,$y1);
}

=head2 stringUp
 Title        : stringUp
 Usage        :	$image -> stringUp(font,x,y,text,color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : bounds
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				text 	=> text to write
 				color 	=> color of the text
 Globals      : gd

=cut
sub stringUp
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my $o_layer = shift;

	# Calculate absolute coordinate and draw text :
	($x,$y) = $self->{frame}->ToAbsoluteCoordinates($x,$y);
	$self -> GetLayer($o_layer) -> stringUp($font,$x,$y,$text,$color);

	# Calculate bounds and return
	my ($x0,$y0,$x1,$y1) = ($x,$y-length($text)*$font->width(),$x+$font->height(),$y);
	return ($x0,$y0,$x1,$y1);
}

=head2 char
 Title        : char
 Usage        :	$image -> char(font,x,y,char,color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				char 	=> char to write
 				color 	=> color of the char
 Globals      : gd

=cut
sub char
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my $o_layer = shift;

	return $self -> GetLayer($o_layer) -> char($font,$self->{frame}->ToAbsoluteCoordinates($x,$y),$text,$color);
}

=head2 charUp
 Title        : charUp
 Usage        :	$image -> charUp(font,x,y,text,color)
 Prerequisite : none
 Procedure    :	overload GD::Image method, using FrameTransformer method ToAbsoluteCoordinates to convert nucleic 'Y' position
 Returns      : none
 Args         : font	=> the gdfont to use
 				x,y 	=> coordinates, with units like (px, base-pair)
 				text 	=> text to write
 				color 	=> color of the text
 Globals      : gd

=cut
sub charUp
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my $o_layer = shift;

	return $self -> GetLayer($o_layer) -> charUp($font,$self->{frame}->ToAbsoluteCoordinates($x,$y),$text,$color);
}

=head2 absoluteString
 Title        : absoluteString
 Usage        :	$image -> abs-string(font,x,y,text,color [,justification])
 Prerequisite : none
 Procedure    :	calls  GD::Image string on the correct layer
 Returns      : bounds
 Args         : font	=> the gdfont to use
 				x,y 	=> absolute coordinates
 				text 	=> text to write
 				color 	=> color of the text
 Globals      : none

=cut
sub absoluteString
{
	my $self = shift;
	my ($font,$x,$y,$text,$color) = (shift,shift,shift,shift,shift) or return;
	my $o_layer = shift;

	# Write text :
	$self -> GetLayer($o_layer) -> string($font,$x,$y,$text,$color);

	# Return bounds :
	return ;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
