#
# 	sebastien.letort@toulouse.inra.fr
#	Created: February 16, 2009
#	$Id$
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Karyotype;

=pod

=head1 NAME

 Karyotype - a Karyotype object for Narcisse

=head1  SYNOPSIS

	# minimal
	my $o_karyo = New Karyotype( o_species => $o_species );

	#maximal
	my %h_param = (
        o_io => $o_io,
		org => $org,
		-o_logger => $o_log,
		-feature_list => [?non teste?],
		-height       => 300,
		-boxes_width  => 100,
		-supp_data    => '^gd$', # only gd dir data
		-rh_padding => { top => $top, bottom => $bottom, # need to be >=30 for chrom label and extremity decorations
		                 left => $left, right => $right,
		                 inner_width => $width, #space between columns
		                 inner_height=> $height #space between rows
		               }
	my $o_karyo = New Karyotype( %h_param );

=head1 DESCRIPTION

 This class manages Karyotype in Narcisse.
 It associates with each chrom its ElementsSet::Cyto::Karyotype,
 it can also associate supplementary data.

 It provides a function to display the Karyotype into an Image object.
 Another function allows to plot as rectangle any element given its chrom's label
 and its start-end position.

=cut

use strict;
use warnings;

use Data::Dumper;
use Error qw( :try );

use base qw( LipmObject );

use NARCISSE_utils;
use BoxesAdministrator::Karyotype;
use SuppDataLoader;
use MapArea;
use ParamParser;
use ElementsSet::Cyto; # Cyto::Karyo est dans le meme fichier que cyto.
# use ElementsSet::Gene;
use General; #SafeIsa, &FALSE
# use LipmError;

=head2 method _Init

 Usage        : my $o_x = New Karyotype( o_species => $o_species,
                                         -feature_list => $ra_features,
                                         -o_logger    => $o_logger,
                                         -rh_padding  => $rh_padding,
                                         -height       => $height,
                                         -boxes_width  => $width,
                                         -supp_data    => $re_filter );
 Prerequisite : none
 Fonction     : initialise an instance of Karyotype class.
 Returns      : a valid Karyotype object
 Args         :
     $o_species, a SpeciesObserver object.
     $ra_features, list ref of classname of features (ElementsSet).
     $o_logger, Logger object
     $height, int [150], height in pixels of the longuest chrom
     $width,  int [100], width in pixel of the box for one chrom
     $re_filter, string ['.'], regular expression to select which supplementary data will be loaded, default means all
     $rh_padding, padding around boxes, unit is pixel,
                  keys are top,bottom,top,left,inner_height,inner_width
 Error        : LipmError::ParameterException
 
 Attributes   :
    __o_logger, #inherited from LipmObject

    __org, org shrot name
    __o_map_area, MapArea object. Needed for ElementsSet objects.
    __rao_features, ref list of every ElementsSet objects.
    __rh_chroms, hash ref, chroms returned by GetSpeciesChromInfo
    __o_param_graph_cfg, ParamParser object containing ROOT_GRAPHICS_CFG info
    __o_boxes_adm, BoxesAdministrator object.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( #o_io          => undef,
	               o_species     => undef,
                   -o_logger     => undef,
# 	               org           => undef,
	               -feature_list => [ qw( Cyto::Karyotype ) ],
	               -o_logger     => undef,
	               -rh_padding   => undef,
	               -height       => 200,
	               -boxes_width  => 100,
	               -supp_data    => '.',
	               -col          => 10,
	               @_ );
	$self->SUPER::_Init( %h_args );

	my $o_species = $h_args{ o_species };
	my $o_io = $o_species -> GetInstanceObserver();

	# assert that o_species
	LipmError::ParameterException::Assert( o_species => $o_species );
	if( &FALSE == &SafeIsa( $o_species, 'SpeciesObserver' ) )
	{
		my $msg = "You give a non SpeciesObserver object for the o_species key.\n";
		throw LipmError::Object( $self, '!build', -text => $msg );
	}

	my $short = $o_species->GetShortName();

	$self->_Set( __org          => $short,
	             __o_map_area   => New MapArea( reference => $short ),
	             __rao_features => [],
	             __o_species    => $o_species,
				 __supp_data_filter => $h_args{ -supp_data } );

	# obtenir la liste des chroms, avec leur taille, position du centro
	# pas sur qu'il soit nécessaire de le conserver comme attributs
	# les objet Cyto semble avoir toutes les infos necessaires
	my $rh_chroms;
	if ( $o_species -> IsPseudo() )
	{
		$rh_chroms = $o_species->GetChromsInfo();
	}
	else
	{
		$rh_chroms = $o_species->GetChromsInfo($o_io -> GetCfg( 'MAX_DISPLAYED_SCAFFOLDS' ));
	}

	my $o_param_cfg = New ParamParser( $o_io->GetCfg( 'ROOT_GRAPHICS_CFG' ) );

	$self->_Set( __rh_chroms    => $rh_chroms,
	             __o_param_graph_cfg => $o_param_cfg );

	# update __rao_features
	$self->__LoadFeatures( $h_args{ -feature_list } );
	# update __rao_features
	$self->__LoadSupplementaryData( $h_args{ -supp_data } );

	# les parametres du constructeur de BoxesAdministrator::Karyotype
	my @a_keys = qw( -o_logger -rh_padding );
	my %h_params = ( ra_chroms => $self->GetSortedChroms() );
	@h_params{ @a_keys } = @h_args{ @a_keys };
# 	$h_params{ ra_chroms }	= $self->GetSortedChroms();
	@h_params{ qw( height boxes_width col ) }
							= @h_args{ qw( -height -boxes_width -col ) };

	my $o_box_adm = New BoxesAdministrator::Karyotype( %h_params );
	$self->_Set( __o_boxes_adm => $o_box_adm );

	return;
} # _Init


=head2 __LoadFeatures

 Usage        : $o_x->__LoadFeatures( $ra_features )
 Function     : Load all features of $ra_features for each chroms.
 Args         : $ra_features, ref list of every ElementsSet objects.
 Error        : none.

 Note         : Do we have to allow something else than Cyto::Karyo ?

=cut

sub __LoadFeatures
{
	my $self   = shift;
	my( $ra_features ) = @_;

	my @ao_elements  = ();

	# rh_params contient o_io, o_param_graph, chromosome, start, end
	my $r_func = sub {
		my ( $rh_params ) = @_;

		foreach my $feature ( @$ra_features )
		{
			try
			{
				my $class = "ElementsSet::$feature";
				my $o_element = New $class( %$rh_params );
				push( @ao_elements, $o_element );
			}
			catch LipmError with
			{
				my $o_err = shift;

				$self->_Trace( $o_err->Explain(), '.', 'IK' );

				$o_err->throw() if( &TRUE == $o_err->IsFatal() );
			};
		}
	};

	$self->__LoadFeatureForAllChromsWithCode( code => $r_func );

	my $rao_feat = $self->_Get( '__rao_features' );
	push( @$rao_feat, @ao_elements );

	return;
}


=head2 __LoadFeatureForAllChromsWithCode

 Usage        : $o_x->__LoadFeatureForAllChromsWithCode( code => $rf_code, rh_param => $rh_param )
 Function     : run the function $rf_code for each chrom of the object.
 Args         :
      $rf_code, code to execute
      $rh_param, params of the code

=cut

sub __LoadFeatureForAllChromsWithCode
{
	my $self   = shift;
	my %h_args = ( code => sub{},
				   @_ );

	my $r_func = $h_args{ code };
	
	my @a_attr = qw( __o_species __rh_chroms  );
	my( $o_species, $rh_chroms ) = $self->_Get( @a_attr );

	@a_attr    = qw( __o_map_area __o_param_graph_cfg );
	my( $o_map_area, $o_param_graph ) = $self->_Get( @a_attr );

    my $o_io   = $o_species->GetInstanceObserver();

	# pour chaque chrom, charger toutes les structures ( ElementsSet )
	foreach my $rh_chrom ( values %$rh_chroms )
	{
		my $rh_params = { o_io          => $o_io,
						  o_species     => $o_species, # tjs besoin de o_io ?
                          o_param_graph => $o_param_graph,
						  o_maparea     => $o_map_area,
						  chromosome    => $$rh_chrom{ chr },
						  start         => 1,
						  end           => $$rh_chrom{ length }
						};
		&$r_func( $rh_params );
	}

	return;
}


=head2 __LoadSupplementaryData

 Usage        : $o_x->__LoadSupplementaryData( $filter )
 Function     : Load supp_ features through SuppDataLoader for each chroms.
 Args         :
     $filter, regular expression to select which supplementary data will be loaded
 Error        : none.

=cut

sub __LoadSupplementaryData
{
	my $self   = shift;
	my( $filter ) = @_;

	my @a_attr = qw( __o_species __supp_data_filter );
	my( $o_species, $rx_filter ) = $self->_Get( @a_attr );

	my $supp_path = $o_species->GetSuppDataPath( 'INSTANCE' );
	my $o_loader  = New SuppDataLoader( o_species      => $o_species,
										ra_directories => [ $supp_path ],
										-feat_prefix   => 'karyo',
										-filter   => $rx_filter,
										-o_logger => $self->GetLogger()
									);

	# pour chaque chrom, charger toutes les structures ( ElementsSet )
	my @ao_elts_set = ();
	my $r_func = sub
	{
		my( $rh_param )  = @_;

		try
		{
			my $rao_elts_set = $o_loader->BuildFeatures( %$rh_param );
			push( @ao_elts_set, @$rao_elts_set );
		}
		catch LipmError with
		{
			my $o_err = shift;

			$self->_Trace( $o_err->Explain(), '.', 'IK' );

			$o_err->throw() if( &TRUE == $o_err->IsFatal() );
		};
	};
	$self->__LoadFeatureForAllChromsWithCode( code => $r_func );

	my $rao_feat = $self->_Get( '__rao_features' );
	push( @$rao_feat, @ao_elts_set );

	return;
}


=head2 GetSortedChroms

 Usage        : my $ra_chroms = $o_x->GetSortedChroms()
 Function     : sort chroms like Dotplot and return them.
 Returns      : $ra_chroms, a ref list on ref hashes like :
                { chr => 'chr_name', length => $nb_bases, centro => $centro_position }
 Args         : none

=cut

sub GetSortedChroms
{
	my $self   = shift;

	my @a_rh_chroms = ();
	my $o_species = $self -> _Get ( '__o_species' );
	my $o_io = $o_species -> GetInstanceObserver();
	my $rh_chroms;
	my $ra_names;
	if ($o_species -> IsPseudo())
	{
		$ra_names  = $o_species -> GetChromsNames();
		$rh_chroms = $o_species -> GetChromsInfo();
	}
	else
	{
		my $max_displayed_scaffolds = $o_io -> GetCfg ( 'MAX_DISPLAYED_SCAFFOLDS' );
		$ra_names  = $o_species -> GetChromsNames('SIZE',$max_displayed_scaffolds);
		$rh_chroms = $o_species -> GetChromsInfo ($max_displayed_scaffolds);
	}
	foreach my $chr ( @$ra_names )
	{
		push @a_rh_chroms,$rh_chroms -> { $chr };
	}
	return \@a_rh_chroms;
}


=head2 Display

 Usage        : $o_x->Display( image => $o_image )
 Function     : Display the object into an Image object.
 Args         : $o_image, Image object

=cut

sub Display
{
	my $self   = shift;
	my %h_args = (  image => undef,
		        @_ );

	&LipmError::ParameterException::Assert( 'image' => $h_args{ image } );

	my $o_boxes = $self->_Get( '__o_boxes_adm' );
	my $rh_frame_transfo = $o_boxes->GetFrameTransfo(); # methode ad hoc = bidouille
	my $o_img   = $h_args{ image };

	my $rh_chroms   = $self->_Get( '__rh_chroms' );

	my $rao_features = $self->_Get( '__rao_features' );
	foreach my $o_feature ( @$rao_features )
	{
		my $chr = $o_feature->_Get( '_chromosome' );
		# HACK ! _Get ne peut pas etre appele ici par $o_feature !
		# je l'ai fait parce que je suis oblige de passer un objet Box

		my $o_box = $o_boxes->GetBox( $chr );
		
		# HACK from manu - Si $o_box n'est pas definie ca veut dire qu'on ne doit rien afficher pour ce chromosome
		#                  Je ne pige absolument pas pourquoi on essaie de le dessiner alors qu'il a ete elimine au debut
		#                  (cf. GetSortedChroms quand on n'est pas en PSEUDOMOLECULES)
		next unless defined $o_box;
		my %h_absolute_coords = $o_box->GetDisplayedArea();

		my $o_frame = $$rh_frame_transfo{ $chr };
		$o_img->SetFrameChange( $o_frame );

		# dessine sur couche 'Karyotype'
		my $feature = $o_feature->_Get( '_feature' );
		$self->_Trace( "display $feature ($chr)\n", '.', 'IK' );
		$o_feature->Display( image => $h_args{ image },
							 o_box => $o_box );
	}

	return;
}


=head2 GetImageParameters

 Usage        : $o_x->GetImageParameters()
 Function     : call GetImageParameters from __o_boxes_adm.
 Args         : none
 Return       : hash, with keys width and height

=cut

sub GetImageParameters
{
	my $self = shift;

	return $self->_Get( '__o_boxes_adm' )->GetImageParameters();
}


=head2 Plot

 Usage      : $o_x->Plot( $o_img, $chrom, $start, $end, $label, $color )
 Function   : Plot a rectangle under the chrom given its start-end position.
 Args       :
     $o_img, Image object
     $chrom, string, chrom label
     $start, int, start position of the element
     $end,   int, end position of the element
     $label, string, label to print []
     $color, string, color label used by Image object [red]
 Error      : none.
 Note       : I think I used this order to fit Highlights format of Circos.
              Label have been add by me, like color.

=cut

sub Plot
{
	my $self   = shift;
	my( $o_img, $chr, $start, $end, $label, $str_color ) = @_;
# warn Dumper { plot => [ $chr, $start, $end ] };

	my $rh_frame_transfo = $self->_Get( '__o_boxes_adm' )->GetFrameTransfo();
	my $o_frame = $$rh_frame_transfo{ $chr };

	if( !defined $o_frame )
	{
		my $msg = "The chromosome $chr is not known,";
		$msg   .= " associated data will be skipped.\n";
		$self->_Trace( $msg, '.', 'IK' );

		return;
	}
	$o_img->SetFrameChange( $o_frame );


	my $layer = $o_img->GetLayer( 'under_karyo' );
	my $old_layer = $o_img->SetCurrentLayer( $layer );

	# trouve la boite
	my $o_box = $self->_Get( '__o_boxes_adm' )->GetBox( $chr );
	if( !defined $o_box )
	{
		my $msg = "Your chrom $chr doesn't seem to exist in this karyotype.";
		throw LipmError::ParameterException( 'chr', 'value', -text => $msg );
	}

	my %h_coords = $o_box->GetDisplayedArea();
	my( $X0,$X1 )  = @h_coords{ qw( X0 X1 ) };

	# largeur de la representation = un peu plus que la largeur du chrom
	# faudrait relire le fichier graphics expres pour ca,
	# je suis contre et le marque donc en dur. on verra plus tard.
	my $width = 18;
	my $A0 = ( $X0 + $X1 - $width ) / 2; # $X0 + ($X1-$X0)/2 - $width/2
	my $A1 = $A0 + $width;


	$str_color = 'red' unless defined( $str_color );
    # tout ce qui est sous la couleur disparait !
    # 	my $color = $o_img->colorAllocateAlpha('red',64, $layer );
	my $color = $o_img->colorAllocate( $str_color, $layer );
	my $black = $o_img->colorAllocate( 'black', $layer );

	my ($B0,$B1);
	( $A0,$B0, $A1,$B1 ) = $o_img->filledRectangle( $A0,$start, $A1,$end, $color );

	# ecriture du label
	return unless ( defined $label );

# 	my $h = $B1 - $B0;
# 	if( $h < 5 ) # au moins 5 pixels verticalement
# 	{
# 		my $msg = "Not enough";
# 	}

	my $pos_y = ($start+$end) /2;
	$o_img->stringVerticalAligned( $o_img->gdTinyFont, $A1+5,$pos_y, $label, $black );

	#je remets dans l'etat precedent
	$o_img->SetCurrentLayer( $old_layer );

	return;
}


=head2 LoadGraphicsCfg

 Usage        : $o_x->LoadGraphicsCfg( file => $file )
 Function     : Set __o_param_graph_cfg attribute with a ParamParser object build on $file.
 Args         : none

=cut

sub LoadGraphicsCfg
{
	my $self   = shift;

    my $o_io = $self->_Get('__o_species')->GetInstanceObserver();
	my %h_args = ( file => $o_io->GetCfg('ROOT_GRAPHICS_CFG' ), @_ );

	$self->_Set( __o_param_graph_cfg => New ParamParser( $h_args{ file } ) );

	return;
}


=head2 Procedure DumpMapInformation

 Title      : DumpMapInformation
 Usage      : $o_x->DumpMapInformation( $outfile )
 Procedure  : Writes the map coordinate informations (.coord file) into $outfile
 Returns    : none
 Args       : $outfile, the output filename

=cut

sub DumpMapInformation
{
	my $self = shift;
    my( $outfile ) = @_;

	#Dumping the Map information for all the displayed objects
	$self->_Get( '__o_map_area' )->DumpMapInformation( $outfile );

	return;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;


