#
# 	sebastien.letort@toulouse.inra.fr
#	Created: February 12, 2009
#	$Id: KaryotypeMaker.pm 3664 2011-03-10 12:53:47Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package KaryotypeMaker;

=pod

=head1 NAME

 KaryotypeMaker - A builder for the Karyotype class that managed the image.

=head1  SYNOPSIS

	my $rh_padding   = { inner_height => 30, top => 10, bottom => 10 };

	my %h_karyo_params = ( filer => 1 ); # no supp data with 1 so no displayed
	my $o_karyo = New KaryotypeMaker( -o_logger => $o_log,
	                                    org => $org,
	                                    -rh_padding => $rh_padding,
	                                    %h_karyo_params
	                                  );

	my $ra_elts = $$rh_params{ elements };
	foreach my $rh_elt ( @$ra_elts )
	{
		$o_karyo->Plot( @$rh_elt{ qw( chrom start end label ) } );
	}

	$o_karyo->Print( %$rh_params );

=head1 DESCRIPTION

 This class aims to print Karyotype as png image.
 With it you can also print elements over the chrom through Plot function.

 All actions related to the Image object (layers,colors) are hidden.
 Colors are those found in cyto files and defined in $CFG{ ROOT_COLORS_CFG }.
 !! %CFG need to be initialized for ROOT_COLORS_CFG.

 See Karyotype.pm for more informations.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( LipmObject );

use NARCISSE_utils;
use Karyotype;
use ElementsSet::Cyto; # pour classe ElementsSet::Cyto::Karyotype
use Image;
use Carp;

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New KaryotypeMaker( o_species => $o_species,
                                              -rh_padding => $rh_pad,
 											  -height => $height,
 											  -o_logger => $o_logger );
 Prerequisite : none
 Fonction     : initialise an instance of KaryotypeMaker class.
 Returns      : a valid KaryotypeMaker object
 Args         : same as Karyotype _Init class args
 Attributes   :
    __o_logger, #inherited from LipmObject

    __o_image, Image object
    __o_karyo, Karyotype object
=cut

sub _Init
{
	my $self = shift;
	my %h_args = (	o_species => undef,
                    -o_logger => undef,

					#keys for direct transmission to Karyotype
					# -supp_data
					#-rh_padding
					#boxes_width
					#height
					#org
	               	@_ );
	$self->SUPER::_Init( %h_args );

	my $o_karyo = New Karyotype( %h_args );

	# Image
	my $o_io    = $h_args{ o_species }->GetInstanceObserver();
	Image->InitColorsLibrary( $o_io->GetCfg( 'ROOT_COLORS_CFG' ) );
	my $o_image = Image->New( $o_karyo->GetImageParameters(),
                              frame_transf_type => '1D' );

	# display the karyotype into the image.
	$o_karyo->Display( image => $o_image ); #, o_boxes => $o_box_adm );

	$self->_Set( __o_image => $o_image,
	             __o_karyo => $o_karyo );

	return;
} # _Init


=head2 Print

 Usage        : $o_x->Print( output => $out )
 Function     : call the PrintImage method of the Image object.
 Args         : $out, string [karyotype.png], ouput filename
 Error        : none.

=cut

sub Print
{
	my $self   = shift;
	my %h_args = ( 
			output => 'karyotype.png',
			coord  => 'karyotype.coord',
			@_ 
		);

	my $msg = "Will print all Image's layers in $h_args{ output }\n";
	$self->_Trace( $msg, '.', 'IKV' );

	my $o_img    = $self->_Get( '__o_image' );
	my @a_layers = $o_img->GetLayersList();

	$o_img->PrintImage( $h_args{ output }, ra_layers => \@a_layers );
	# rename the files ?

	# en test
	$self->_Get( '__o_karyo' )->DumpMapInformation( $h_args{ coord } );

	return;
}


=head2 Plot

 Usage        : $o_x->Plot( $chrom, $start, $end, $label, $color )
 Function     : Plot a rectangle under the chrom given its start-end position.
 Args         :
     $chrom, string, chrom label
     $start, int, start position of the element
     $end,   int, end position of the element
     $label, string, label to print []
     $color, string, color label used by Image object [red]
 Error        : none.

=cut

sub Plot
{
	my $self   = shift;
	my( $chrom, $start, $end, $label, $color ) = @_;

	my( $o_karyo, $o_img ) = $self->_Get( qw( __o_karyo __o_image ) );

	$o_karyo->Plot( $o_img, $chrom, $start, $end, $label, $color );

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;


