#
#	sebastien.letort@toulouse.inra.fr
#	Created: October 2, 2008
#	$Id: KnowledgeRemoteReader.pm 3728 2011-09-28 13:24:44Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

 KnowledgeRemoteReader

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

package KnowledgeRemoteReader;

use strict;
use warnings;

use base qw( LipmObject );

use Data::Dumper;
use File::Temp;
use File::Basename;

use General;
use DataReader::Whole;
use NARCISSE_utils;
use Enum;
use AnotherTemplate;
use iANT::DB::Entry;

use constant DATA_FILE => 'knowledge.list';

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New KnowledgeRemoteReader( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of KnowledgeRemoteReader class.
 Returns      : a valid KnowledgeRemoteReader object
 Args         : %h_args with keys :
                o_io          => obj, instance observer
                species       => string, knowledge short species name,
                ra_locus_tags => array list of locus tags.

                optional:
                -o_logger => a Logger object, to debug
                -cache    => string, filename of a cache,
                             where data are stored in Narcisse Format.
 Attributes   :
	__species  : string, species knowledge short name
	__ra_tags  : ref list of locus_tags for this species
	__o_logger : Logger object, for debug purpose
	__cache    : string, a filename.
	             The file contains informations read from iANT XML,
	             and stored in Narcisse Format (for fastest access).
	__rh_data  : the knowledge data.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( o_io           => undef,
                   species        => undef,
	               -ra_locus_tags => [],
	               -cache         => undef,
	               -o_logger      => undef,
	               @_ );

	$self->SetLogger( $h_args{ -o_logger } );

	if( !defined( $h_args{ species } ) )
	{
		# ObjectException ?
		my $msg = "species is a mandatory argument of ";
		$msg   .= __PACKAGE__ . " constructor.";
		throw LipmError::ParameterException('species','missing',-text => $msg);
	}

	@$self{ qw( __species ) } = @h_args{ qw( species ) };
	$self->_Set( __o_inst_obs => $h_args{ o_io } );

	# optional TODO si rien d'autre pas de tranche !
	@$self{ qw( __cache __ra_tags ) } = @h_args{ qw( -cache -ra_locus_tags ) };

	$self->{ __rh_data } = {};

	my $ra_tags = $self->{ __ra_tags };
# 	return  if( 0 == @$ra_tags ); # pourquoi ca a ete rajoute ca ?

	# syntaxe perl 5.10
	my $filename = $self->{ __cache };
	$self->Debug( "try to read >$filename<\n" );
	if( -e -r $filename )
	{
		$self->Log( "will read cached file $filename\n" );

		my $o_ndr = New DataReader::Whole();
		$o_ndr->Prepare( -from => $filename );
		$self->{ __rh_data } = $o_ndr->FetchallHashref( 'line', -line_type => 'HASH' );
	}
	else
	{
		my $o_db_entry = $self->__LaunchWebService();
		$self->__ParseDbEntry( $o_db_entry );	# MaJ __rh_data
	}

# warn Dumper { init => $self->{ __rh_data } };
	# creation d'un lien symbolique pour faciliter le traitement
	#	des scripts suivants nc_display_map.pl
	#	Attention, avec ca, seul le dernier appel est pris en compte
	if( $self->{ __cache } ne &DATA_FILE )
	{
		my @a_ = unlink( &DATA_FILE );
		@a_    = symlink( $self->{ __cache }, &DATA_FILE );
	}

	return;
}

=head2 function __LaunchWebService

 Title        : __LaunchWebService
 Usage        : $o_db_entry = $o_xx->__LaunchWebService( none );
 Prerequisite : none
 Function     : none
 Returns      : $o_db_entry
 Args         : none

 TODO    : Faire utiliser
=cut

sub __LaunchWebService
{
	my $self = shift;

	my ( $species, $ra_locus ) = @$self{ qw( __species __ra_tags ) };

    my $o_io = $self->_Get('__o_inst_obs');

	#pour AnotherTemplate, je dois mettre les locus dans un fichier accessible par url
	my $file = $self->__PrintTmpFile();
	my $url  = $o_io->GetCfg('HTTP_SESSDIR') . '/' . $file;

	# AnotherTemplate string
	my $str = "#BIOMOBY|LegooGetKnowledgeBaseRelationships\@www.legoo.org|";
	$str   .= "inputs => [ { name => 'infile', type => 'List_Text',";
	$str   .= " id => 'file', url_content => '$url' } ],";
# 	$str   .= " id => 'file', content => '$locus_tag' } ],";
	$str   .= "output => 'raw',";
	$str   .= "parameters => { genotype => '$species', check_ID => 'false'";
	$str   .= ", object_type => 'gene' }#";

	$self->Log( "AnotherTemplate($str)\n" );

	# lancement du service a travers AnotherTemplate
	my $o_tmpl = New AnotherTemplate();
	$o_tmpl->Update($str);
	$o_tmpl->Parse();
	my $xml = $o_tmpl->Get( 'parsed_template' );
	#warn "KOUKOU str => $str";
	#warn "KOUKOU xml => $xml";
	my $o_db_entry = New iANT::DB::Entry( $xml );

	return $o_db_entry;
}


=head2 function

 Title        : __PrintTmpFile
 Usage        : $url = $self->__PrintTmpFile();
 Prerequisite : none
 Function     : print all locus tags into a file accessible via url
 Returns      : $url, string, the url to access the file
 Args         : none

=cut

sub __PrintTmpFile
{
	my $self = shift;
    my $o_io = $self->_Get( '__o_inst_obs' );
	my $sessdir = $o_io->GetCfg( 'ROOT_SESSDIR' );
	my $fh_tmp  = new File::Temp( DIR => $sessdir, UNLINK => 0 );

	my $ra_tags = $self->{ __ra_tags };
	print $fh_tmp join( "\n", @$ra_tags );

	my $file = $fh_tmp->filename();
	my $nb = @$ra_tags;
	$self->Debug( "$file have been written with $nb tags\n" );
`chmod 777 $file`;

	return &File::Basename::basename( $file );
}


=head2 Procedure __ParseDbEntry

 Title        : __ParseDbEntry
 Usage        : $o_xx->__ParseDbEntry( $o_db_entry );
 Prerequisite : none
 Procedure    : fill rh_data with entries and print then as a NarcisseDataFile
 Args         : $o_db_entry, a iANT::DB::Entry

=cut

sub __ParseDbEntry
{
	my $self = shift;
	my ( $o_db_entry ) = @_;

	$self->Debug( "parse db entry object\n" );

	my $rh_data = $self->{ __rh_data };
	my $num_line = 1;
	my @a_headers;

	my $ra_entries = $o_db_entry->GetEntries();
	foreach my $o_entry ( @$ra_entries )
	{
		my $o_db_features = $o_entry->GetEntryFeatures();
		my $ra_patterns = $o_db_features->GetPatterns();
		foreach my $o_pattern ( @$ra_patterns )
		{
			my $pattern = $o_pattern->GetDefinition();
			my %h_data = ( pattern => $pattern, id => $o_entry->Get('id'),
			           line => $num_line );
			my $rao_properties = $o_pattern->GetProperties();
			foreach my $o_property ( @$rao_properties )
			{
				my $id  = $o_property->GetId();
				my $val = $o_property->GetValue();
				$h_data{ $id } = $val;
			}
			$$rh_data{ $num_line } = \%h_data;
			++$num_line;

			# Note : !defined @a est obsolete
			@a_headers = keys %h_data	if( !@a_headers );
		}
	}

	if( 0 != scalar( keys %$rh_data ) )
	{
# 		my @a_headers = ( @a_properties, 'pattern', 'id', 'line' );
		&NARCISSE_utils::PrintNarcisseData( $self->{ __cache }, \@a_headers,
	                                        $rh_data, { sort => 'num' } );
	}

	return;
}


=head2 function

 Title        : GetPatterns
 Usage        : $ra_patterns = $o_xx->GetPatterns();
 Prerequisite : none
 Function     : return all the patterns found into __rh_data
 Returns      : $ra_patterns
 Args         : none

=cut

sub GetPatterns
{
	my $self = shift;

	return $self->Get( 'pattern', @_ );
}


=head2 function

 Title        : GetIds
 Usage        : $ra_locus_tags = $o_xx->GetIds();
 Prerequisite : none
 Function     : return all the ids found into __rh_data
 Returns      : $ra_ids, array ref of all ids found
 Args         : none

=cut

sub GetIds
{
	my $self = shift;

	return $self->Get( 'id', @_ );
}


=head2 function

 Title        : GetValues
 Usage        : $ra_vals = $o_xx->GetValues( @a_lines );
 Prerequisite : none
 Function     : return the record values found into __rh_data
                that correspond to lines into @a_lines
 Returns      : $ra_vals, array ref of values found
 Args         : @a_lines, array of sorted lines number, optional
                if none, return all values

 Note         : I do not use array because if i do (which i should),
                i must write a function to print narcisse data
                with array of hash ref. I don't have time nor envie !
=cut

sub GetValues
{
	my $self = shift;

	return $self->Get( 'factor', @_ );
}


=head2 function

 Title        : Get
 Usage        : $ra_key = $o_xx->Get( $key, @a_lines );
 Prerequisite : none
 Function     : return all the $key-value found into __rh_data
 Returns      : $ra_key, arrayref, values of __rh_data{ $key }
 Args         : $key, a key of __rh_data
                id, pattern, ... whatever inside xml file
                @a_lines, array of lines number. optional
                          if none, return all.

=cut

sub Get
{
	my $self = shift;
	my ( $key, @a_lines ) = @_;

	return $self->Get( $key )	if( 0 == scalar( @a_lines ) );

	my $rh_data = $self->{ __rh_data };

	my @a_val   = ();
	foreach my $line_number ( @a_lines )
	{
		my $rh_row = $$rh_data{ $line_number };

		if( !defined $rh_row )
		{
			my $err_msg = "The parameter doesn't seem to be a line number\n";
			$err_msg   .= Dumper $rh_data;

			throw LipmError::ParameterException( '$line_number', 'value',
			                                     -text => $err_msg );
		}
		push( @a_val, $$rh_row{ $key } );
	}

	return \@a_val;
}


=head2 function

 Title        : GetRow
 Usage        : $rh_ = $o_xx->GetRow( %h_args );
 Prerequisite : none
 Function     : read __rh_data and return one row, regarding %h_args
 Returns      : $rh_ one row/record of __rh_data
 Args         : %h_args, keys are column filtration
                 for now only 'line'

=cut

sub GetRow
{
	my $self = shift;
	my %h_args = ( line => undef,
	               @_ );

	my $rh_data = $self->{ __rh_data };
	foreach my $rh_ ( values %$rh_data )
	{
		return $rh_	if( $$rh_{ line } eq $h_args{ line } );
	}

	return undef;
}


=head2 function

 Title        : GetData
 Usage        : $rh_data = $o_xx->GetData();
 Prerequisite : none
 Function     : accessor to __rh_data
 Returns      : $rh_data, __rh_data value
 Args         : none

=cut

sub GetData
{
	my $self = shift;

	return $self->{ __rh_data };
}


=head2 Procedure Log

 Title        : Log
 Usage        : $o_xx->Log( $msg );
 Prerequisite : none
 Procedure    : none
 Args         : $msg

=cut

sub Log
{
	my $self   = shift;
	my( $msg ) = @_;

	$self->_Trace( $msg, '.', 'Ik', __PACKAGE__ );

	return;
}


=head2 Procedure Debug

 Title        : Debug
 Usage        : $o_xx->Debug( $msg );
 Prerequisite : none
 Procedure    : none
 Args         : $msg

=cut

sub Debug
{
	my $self   = shift;
	my( $msg ) = @_;

	$self->_Trace( $msg, '.', 'IkV', __PACKAGE__ );

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;
