#
# 	Sebastien.Letort@toulouse.inra.fr
#	Created: May 21, 2007
#	$Id$
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package Legend;

use warnings;
use strict;

use base qw( LipmObject );

=pod

=head1 NAME

Legend

=head1 SYNOPSIS


=head1 DESCRIPTION

	With this class you generate ressource image for a legend.

	attributes :
	- _side
	- _font
	- _img	=	la ressource gd
	- _map	=	la carte des glyphes
	- _max_len
	- _ra_data
	- _nb_elt	=	le nb d'elements dans la legende
	- _h_margin
	- _h_box
	- height	=	hauteur de la legende
	- width	= largeur de la legende

=cut

use GD::Image;
use IO::File;
use Data::Dumper;


# for TRUE and FALSE
use General qw(:DEFAULT ShowDebug);

# color manipulation
eval "use Colors::Color";
my $USE_COLORS = ( $@ eq "" ) ? &TRUE : &FALSE;

# pour la fonction blessed
use Scalar::Util;

# error management
use LipmError::ParameterException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        :	_Init
 Function     :	Constructor, initialize the object, do not use directly, use New method instead
 Usage        :	my $o_legend = New Legend( $ra_data, $rh_margin, $box_width, $box_height, $side );
 Prerequisite : less than 253 elements will be represented
 Returns      :	an new Legend object
 Args         : 2 ref on hash tables :
               	$ra_data, the data to represent like [ $o_color, "text label" ],
               		- $o_color is Color object or a ref to an array like [$r, $g, $b].
               	$rh_margin, spacing between element like
               		{	top_bottom => $h, left_right => $w,
               			inter_elt_height => $eh, inter_elt_width => $ew	}
               		- top_bottom represent the margin at the top and at the bottom of the picture.
               		- left_right represent the margin at the left and at the right of the picture.
               		- inter_elt_width is the width spacing between the box and its label.
               		- inter_elt_height is the height spacing between 2 elements boxes.
               	$box_width, the color box element width
               	$box_height, the color box element height
               	$side, 'left' by default or 'right', where to place the color box.
 Error        :
 Globals      :	none

=cut

sub _Init
{
	my $self = shift;
	my ($ra_data, $rh_margin, $box_width, $box_height, $side, $gd_font) = @_;

	$self->_CheckData($ra_data);
	$self->_CheckMargin($rh_margin);
	$self->_CheckBoxDim($box_width, $box_height);

	$side = 'left'	unless defined( $side );
	$self->{_side} = $side;

	$gd_font = GD::gdSmallFont	unless defined( $gd_font );
	$self->{_font} = $gd_font;

	# set the final dimension of the legend
	$self->_ComputeLegendDimensions();

	# the image
	$self->{_img} = new GD::Image( $self->{width}, $self->{height} );
	$self->{_img}->interlaced('true');
	# background color
	my $white = $self->{_img}->colorAllocate(255,255,255);

	# build the ressource image with colors
	$self->_DrawImage();

	return;
}


sub ToPng
{
	my $self = shift;
	my ($filename) = @_;

	# to make it smart, I add a border line
	my $black = $self->{_img}->colorExact(0,0,0);
	$self->{_img}->setThickness(3);
	$self->{_img}->rectangle(0,0, $self->{width}-1,$self->{height}-1, $black);

	my $file = new IO::File($filename, "w");
	$file->binmode;

	print $file $self->{_img}->png;
}


=head2 Procedure GetGdRessource

 Title        : GetGdRessource
 Usage        : $o_xx->GetGdRessource();
 Prerequisite : none
 Procedure    : Accessor to the GD ressource _img
 Args         : none

=cut

sub GetGdRessource
{
	my $self = shift;
	return $self->{ _img };
}


#prerequis : on a un objet couleur par label
sub _CheckData
{
	my $self = shift;
	my ($ra_data) = @_;

	my ($max_len) = (0);

	my $size = @$ra_data;
	if( ($size % 2) != 0 )
	{
		throw LipmError::ParameterException($ra_data, 'value',
								-should_be => "one color for one label !");
	}
	for( my $i=0; $i<@$ra_data; $i+=2)
	{
		my $color = $$ra_data[$i];
		my $bool = &FALSE;
		if ( &Scalar::Util::blessed($color) )
		{	$bool |= $color->isa('Colors::Color');	}
		$bool |= ( ( ref($color) eq "ARRAY" ) and ( scalar(@$color) == 3 ) );
		if( !$bool )
		{
			my $msg = 'a Colors::Color object or a ref to an array like [$r,$g,$b]';
			throw LipmError::ParameterException( $color, 'type',
									-should_be => $msg );
		}

		# get the max length of labels
		my $len = length( $$ra_data[$i+1] );
		$max_len = $len	if( $len > $max_len );
	}

	# set the attributes
	$self->{_max_len} = $max_len;
	$self->{_ra_data} = $ra_data;
	$self->{_nb_elt} = $size/2;

	return;
}


sub _CheckMargin
{
	my $self = shift;
	my ($rh_margin) = @_;
	my @a_keys = ('top_bottom', 'left_right', 'inter_elt_width', 'inter_elt_height');

	foreach my $key (@a_keys)
	{
		if( !exists( $$rh_margin{$key} ) )
		{	throw LipmError::ParameterException($key, 'missing');	}

		#set attributes
		$self->{_h_margin}{$key} = $$rh_margin{$key};
	}

	return;
}


sub _CheckBoxDim
{
	my $self = shift;
	my ( $width, $height ) = @_;

	# i use another array not to duplicate the code
	my @vals = ( $width, $height );
	foreach my $val (@vals)
	{
		if( !defined( $val ) )
		{	throw LipmError::ParameterException('one color box dim.', 'missing');	}
	}

	# set the attribute
	$self->{_h_box}{width} = $width;
	$self->{_h_box}{height} = $height;

	return;
}


sub _ComputeLegendDimensions
{
	my $self = shift;
# 	my () = @_;
	my $font_width = $self->{_font}->width;
	my $font_height = $self->{_font}->height;

	# width
	my $width = 2*$self->{_h_margin}{left_right};
	$width += $self->{_h_box}{width} + $self->{_h_margin}{inter_elt_width};
	$width += $self->{_max_len} * $font_width;
	$self->{width} = $width;

	# height
	my $h = $self->{_h_box}{height};
# 	my $elt_height = ( $h > $font_height) ? $h : $font_height;
	my $elt_height = $h;
	my $height = 2*$self->{_h_margin}{top_bottom};
	$height += $elt_height;	# le premier elt

	#les autres elt
	$height += ( $elt_height + $self->{_h_margin}{inter_elt_height} ) * ( $self->{_nb_elt} - 1 );
	$self->{height} = $height;

	return;
}


sub _DrawImage
{
	my $self = shift;
	my ($x0, $xi, $string_x0) = $self->_GetXCoordinates();

	my $y0;
	# minus operation is to facilitate the add operation into the loop
	# this is a bidouille !
	my $yi = $self->{_h_margin}{top_bottom} - $self->{_h_margin}{inter_elt_height};

	#les variable de la police
	my $black = $self->{_img}->colorAllocate(0,0,0);

	my $ra_data = $self->{_ra_data};
	for( my $i=0; $i<@$ra_data; $i+=2 )
	{
		my ($color_data, $label) = @$ra_data[$i, $i+1];

		# color allocation
		my $color;
		if( ref($color_data) eq 'ARRAY' )
		{	$color = $self->{_img}->colorAllocate(@$color_data);	}
		elsif($USE_COLORS)	# it's a Color object
		{	$color = $self->{_img}->colorAllocate($color_data->GetRGB());	}
		else	# pour eviter un crash = pas de couleur !
		{	$color = $self->{_img}->colorAllocate(255,255,255);	}

		# the color box
			# coordinates
		$y0 = $yi + $self->{_h_margin}{inter_elt_height};
		$yi = $y0 + $self->{_h_box}{height};
		$self->{_img}->filledRectangle($x0,$y0, $xi,$yi, $color);
		$self->{_img}->rectangle($x0,$y0, $xi,$yi, $black);

		#the label
		my $string_y0 = $yi - $self->{_font}->height;
		$self->{_img}->string($self->{_font}, $string_x0, $string_y0, $label, $black);
	}

	return;
}

# $x0, $xi = are the x coordinates of the glyph box $x0 < $xi
# $string_x0 is the x coordinate of the label
sub _GetXCoordinates
{
	my $self = shift;
	my ( $x0, $xi, $string_x0 );

	if( $self->{_side} eq 'left' )
	{
		$x0 = $self->{_h_margin}{left_right};
		$xi = $x0 + $self->{_h_box}{width};
		$string_x0 = $xi + $self->{_h_margin}{inter_elt_width};
	}
	else
	{
		$xi = $self->{width} - $self->{_h_margin}{left_right};	# -1 ?
		$x0 = $xi - $self->{_h_box}{width};
		$string_x0 = $self->{_h_margin}{left_right};
	}

	return ( $x0, $xi, $string_x0 );
}

# renvoie x0, y0, xi, yi
sub GetCoordinates
{
	my $self = shift;

	return $self->{_img}->clip;
}


# prerequis, $img_dest est undef ou bien une ressource image valide !
sub ConcatenateLegend
{
	my ($img, $direction, @ao_legends ) = @_;

	#1- on calcule les dimensions de la nouvelle image
	my ($width, $height) = (!defined $img) ? (0,0) : $img->getBounds();
	foreach my $o_legend (@ao_legends)
	{
		if( $direction eq 'width' )
		{
			$width += $o_legend->{_img}->width;
			my $leg_height = $o_legend->{_img}->height;
			$height = $leg_height	if($height < $leg_height);
		}
		else
		{
			my $leg_width = $o_legend->{_img}->width;
			$width = $leg_width		if($width < $leg_width);
			$height += $o_legend->{_img}->height;
		}
	}

	#2- on cree le receptacle
	my $img_dest = new GD::Image($width, $height);
	my $white = $img_dest->colorAllocate(255,255,255);

	#3- on cree une liste de ressource gd
	my @a_imgs = (defined $img) ? ($img) : ();
	foreach my $o_legend (@ao_legends)
	{	push(@a_imgs, $o_legend->{_img});	}

	#4- on concatene
	my ($x0, $y0) = (0,0);
	$width = $height = 0;
	foreach my $img_src (@a_imgs)
	{
		if( $direction eq 'width' )
		{	$x0 += $width;	}
		else
		{	$y0 += $height;	}

		($width, $height) = $img_src->getBounds();
		$img_dest->copy( $img_src, $x0,$y0, 0,0, $width, $height );
	}

	# je ne libere pas la ressource de $img, perl s'en charge ?
	return $img_dest;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
