#
# $Id: Locus.pm 3570 2010-12-03 14:38:49Z manu $
#

package Locus;

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Locus - describes any locus on a sequence

=head1 SYNOPSIS

 my $o_locus = New Locus('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_locus);

 

=head1 DESCRIPTION

 In addition to the mandatory locus features, any pair of key/values can be stored in the objet
 A locus is identified by a start, end positions on the genomic sequence AND by a locus_tag
 The locus_tag should be  - but is not always uniq, so we manage the possibility to get the same locus_tag for several loci
 The parent is another locus inside which the current locus lives, ex:
 cds --> mRNA --> gene --> chromosome
 The last parent in this chain must be an object with following methods:
     - GetChromosomeLabel()
     - GetSpecies()
     - GetKingdom()
     - GetCodonTable()
 Typically an object deriving from AnnotatedSeqFile is a OK

=cut

use Carp;
use LipmError::ParameterException;
use LipmObject;
use General;
use base qw( LipmObject );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 3570 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_locus = New Locus('start'=>10,'end'=>100,'strand'=>-1,'locus_tag'=>'GNI','type'=>'gene','product'=>'enzyme');
 Prerequisite : start < end, strand =~ /1|\-1|\+|\-/
 Returns      : none
 Args         : Mandatory: start      position of the start (pb) of the object on the chromosome
                Mandatory: end        position of the end (bp) of the object on the chromosome
				Mandatory: strand     identify the strand on which is located the object; values=1/-1|+/-
				Mandatory: locus_tag  accession on the molecule (expected to by unique)
				Optional : type       type of the object (type of rna, type of marker, ...)
				optional : o_parent   The parent locus
				Optional : rh_contig_info A ref to a hash containing information about the contig in which this locus lives. 
				           This is used to compute the position offset
				Optional : any other pair of key/value
 Globals      : none

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	# mandatory parameters
	foreach my $required ('start','end','strand','locus_tag')
	{
		if ( ! defined($h_param{$required}) )
		{
			#Carp::croak("Fatal error in constructor: $required must be provided to the constructor\n");
			throw LipmError::ParameterException( $required, 'missing');
		}
	}

	# Some sanity checks
	if ( $h_param { start } > $h_param { end } )
	{
		throw LipmError::ParameterException( 'start and end','value',-text => "start should be < end (start=$h_param{start} end=$h_param{end})" );
	}
	if ($h_param { strand } !~ /[-+]?1/)
	{
		throw LipmError::ParameterException( 'strand','value',-text => 'should be one of -1,1 (not ' . $h_param { strand } . ')' );
	}
	
	# the internal key is __externalparam (see GetAll)
	foreach my $param (keys(%h_param))
	{
		$$self{'__'.$param}      = $h_param{$param};
	}
	
	$$self{ __contig_offset } = 0;
	$$self{ __contig_strand } = '.';
	
	# __start_contig, __end_contig = The positions, relative to the contig
	$$self{ __start_contig } = $$self{ __start };
	$$self{ __end_contig }   = $$self{ __end };
    if ( defined $$self{ __rh_contig_info } )
    {
        my $rh_contig_info = $$self{ __rh_contig_info };
        
        # Strand + : the offset is the contig start
        $self -> { __contig_strand } = $rh_contig_info -> { orientation };
        if ( $self -> { __contig_strand } eq '+' )
        {
	        $self -> { __contig_offset } = $rh_contig_info -> {chr_start} - 1;
		}
		
		# Strand - : the offset is the contig end, we exchange (start,end) positions, we change the strand
		else
		{
			$self -> { __contig_offset } = $rh_contig_info -> {chr_stop} + 1;
			$$self{ __start } = $$self{ __end_contig };
			$$self{ __end }   = $$self{ __start_contig };
			$$self{ __strand }= -$$self{ __strand };
		}
		
		# Correct start, end positions
        $$self{ __start }  = $self -> __CorrectOffset( $$self{ __start } );
        $$self{ __end   }  = $self -> __CorrectOffset( $$self{ __end } );
    }
   
    $$self{ __o_parent } = undef unless exists $$self{ __o_parent };
	$$self{ __type }     = "" unless exists $$self{ __type };
}

=head2 function __CorrectOffset

 Title   : __CorrectOffset
 Usage   : $position = $self -> __CorrectOffset( $position)
 Function: Correct the position passed by parameter against the contig offset
 
 Args    : A position
 Return  : The corrected position
 Globals : none

=cut
sub __CorrectOffset
{
    my $self = shift;
    my ($position) = @_;
    my $offset = $self -> { __contig_offset };
	my $contig_strand = $self -> { __contig_strand };

	# strand = '.' ==> default value: no correction
	if ( $contig_strand eq '.' )
	{
		return $position;
	}
	elsif ( $contig_strand eq '+' )
    {
        $position += $offset;
	}
	else
	{
		$position = $offset - $position;
	}
 
    ###warn "KIKI $position" . " offset " . $self -> { __contig_offset } . " version " . $self -> { __rh_contig_info } -> { feature_name } . " " . $self -> GetLocusTag() . " " . $self -> GetType() . "\n";
    return $position;
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_locus->setParent( $o_locus );
 Procedure    : Set the locus parent
 Prerequisite : The parent should not already set, (it is possible to set the parent only once)
                The method could be overloaded in the derived class, to control the parent type
                BUT YOU SHOULD CALL $self->SUPER::SetParent from the overloaded function
                In the Locus class, the parent could be set to any object, the only restrictions are:
                - The parent should 'can' GetParent (thus should be a Locus object)
                - OR the parent shoud 'can' GetChromosomeLabel, GetSpecies and GetKingdom 
                However this is not checked here
 
 Args         : $o_locus : The parent locus
 Globals      : none

=cut
sub SetParent
{
	my $self      = shift;
	my ($o_locus) = @_;
	if ( defined $self -> { __o_parent } )
	{
		throw LipmError::ObjectException($self,'!attribute', -text => 'Parent is already set');
	}
	else
	{
		$self -> { __o_parent } = $o_locus;
	}
	
	if ( ! defined $self -> GetAncestor() )
	{
		throw LipmError::ObjectException($self,'!attribute', -text => 'There is a missing link in the parents chain');
	}
}

=head2 function GetLocusTag

 Title        : GetLocusTag
 Function     : Return the locus tag
 Usage        : my $loc_id = $o_locus->GetLocusTag();
 Prerequisite : none
 Returns      : The locus tag
 Args         : none
 Globals      : none

=cut
sub GetLocus { return GetLocusTag( @_ ); };
sub GetLocusTag
{
	my ($self) = @_;
	return $$self{'__locus_tag'};
}

=head2 function GetGeneName

 Title        : GetGeneName
 Function     : Return the gene name, if not found ""
 Usage        : my $gene_name = $o_mcds->GetGeneName();
 Returns      : The gene name (if possible)

=cut

sub GetGeneName
{
	my ($self) = @_;

	if ( defined $self -> { __gene_name } )
	{
		return $self -> { __gene_name };
	}
	else
	{
		return "";
	}
}

=head2 function GetProduct

 Title        : GetProduct
 Function     : Return the product (= annotation) if found
                If not found and if possible, return the parent's product
 Usage        : my $gene_name = $o_mcds->GetProduct();
 Returns      : The product, or the parent's product (if possible)

=cut
sub GetProduct
{
	my ($self) = @_;
	

	if ( defined $self -> { __product } )
	{
		return $self -> { __product };
	}
	else
	{
		return ( $self -> GetParent() -> isa ( 'Locus' ) ) ? $self -> GetParent() -> GetProduct() : '';
	}
}

=head2 function GetEnds

 Title        : GetEnds
 Function     : Return the locus ends (start,end)
 Usage        : my ($start,$end) = $o_locus->GetEnds();
 Returns      : The ends of the locus
 Args         : none
 Globals      : none

=cut
sub GetEnds
{
	my ($self) = @_;
	return ($self->{'__start'},$self->{'__end'});
}

=head2 function GetEndsContig

 Title        : GetEndsContig
 Function     : Return the locus ends (start,end) relative to the contig
 Usage        : my ($start,$end) = $o_locus->GetEnds();
 Returns      : The ends of the locus
 Args         : none
 Globals      : none

=cut
sub GetEndsContig
{
	my ($self) = @_;
	return ($self->{'__start_contig'},$self->{'__end_contig'});
}	

=head2 function GetAccession

 Title        : GetAccession
 Function     : Return the locus accession if exists else it returns the locus tag
 Usage        : my $loc_acc = $o_locus->GetAccession();
 Prerequisite : none
 Returns      : a locus accession | locus id
 Args         : none
 Globals      : none

=cut

sub GetAccession
{
	my ($self) = @_;

	return $$self{'__accession'} if (defined($$self{'__accession'}));
	return $$self{'__locus_tag'};
}

=head2 function GetNarcisseId

 Title        : GetNarcisseId
 Function     : Return the narcisse Id, computed by some child class
 Usage        : my $id = $o_locus -> GetNarcisseId()
 Returns      : the Narcisse Id
 Args         : none
 Globals      : none
 TODO - __narcisse_id devrai s'appeler _narcisse_id (variable protégéee)

=cut

sub GetNarcisseId
{
	my ($self) = @_;
	return $self -> { __narcisse_id };
}

=head2 function GetRegions

 Title        : GetRegions
 Function     : Returns the regions associated to a composite locus like mRNA or CDS
                Return a string like 1..10,20..30" or an array ('1..10','20..30']
 Usage        : my $region_string = $o_locus->GetRegions();
                my @a_regions = $o_locus->GetRegions();
 Returns      : The regions, string or array formatted, '' or empty array if no region specified
 Args         : none

=cut

sub GetRegions
{
	my ($self) = @_;
	
	if ( defined $self -> {__regions} )
	{
		my $regions = $self -> {__regions};
		return wantarray ? split (',',$regions) : $regions;
	}
	else
	{
		return wantarray ? () : '';
	}
}

=head2 function GetRegionNbExonsAndLength

 Title        : GetRegionNbExonsAndLength
 Function     : Returns the number of exons and the len of a composite locus
 Usage        : my ($nb_exons,$length) = $locus->GetRegionNbExonsAndLength($region)
 Returns      : (nb of exons, total length)
 Args         : none

=cut

sub GetRegionNbExonsAndLength
{
	my $self = shift;

	my @a_exon   = $self -> GetRegions();
	my $nb_exons = @a_exon;
	my $len=0;
	foreach my $exon (@a_exon)
	{
		my ($start,$end) = split(/\.\./,$exon);
		$len += abs($end-$start+1);
	}
	return($nb_exons,$len);
}


=head2 function GetStrand

 Title        : GetStrand
 Function     : Returns -1/+1
 Usage        : my $strand = $o_locus->GetStrand( &FALSE );
 Returns      : The strand (-1/+1 or -/+)
 Args         : If &FALSE (default) return -1/+1
                If &TRUE return -/+ (useful for GFF3)

=cut

sub GetStrand
{
	my $self   = shift;
	my ($gff3) = @_;
	$gff3 = &FALSE unless defined $gff3;
	
	my $strand = $self -> {__strand};
	
	if ($gff3 == &TRUE)
	{
		$strand = '-' if $strand == -1;
		$strand = '+' if $strand ==  1;
		return $strand;
	}
	else
	{
		$strand = -1 if $strand eq '-';
		$strand = +1 if $strand eq '+';
		return $strand+0;
	}
}

=head2 function GetType

 Title        : GetType
 Function     : Returns the type of locus 
 Usage        : my $type = $o_locus->GetType();
 Returns      : The type
 Args         : none

=cut

sub GetType
{
	my ($self) = @_;
	return $self -> { __type };
}

=head2 procedure AddRegion

 Title        : AddRegion
 Procedure    : Add a new region to a composite locus like mRNA or CDS
 Usage        : $o_locus->AddRegion($start,$end);
 Prerequisite : the regions must be added in increasing order    
                MANU 7/10/2009 This prerequisite is removed
 Returns      : none
 Args         : $start: start position on the pseudomolecule
                $end:   end position   on the pseudomolecule
 Globals      : none

=cut

# POUR COMPATIBILITE AVEC LE CODE EXISTANT
sub Update { &AddRegion( @_ ); };
sub AddRegion
{
    my ($self,$start,$end) = @_;
	
	my $contig_strand = $self -> { __contig_strand };

	# If the strand of the contig is +, juste correct the offset
	if ( $contig_strand eq '+' )
	{
	    $start = $self -> __CorrectOffset( $start );
	    $end   = $self -> __CorrectOffset( $end );
	}
	elsif ( $contig_strand eq '-' )
	{
		my $tmp = $start;
	    $start = $self -> __CorrectOffset( $end );
	    $end   = $self -> __CorrectOffset( $tmp );
	}		
    
	my $sep = "";
	if ( defined($$self{'__regions'}) )
	{
		$sep = ",";
		my ($max) = $$self{'__regions'} =~ /(\d+)$/;
		if ( $start < $max )
		{
            # MANU 7/10/2009 Old behaviour: if not well ordered, we giveup
            #                New behaviour: if not well ordered, we reorder
			#Carp::confess("Update: locations must be added in increasing order\nDebug: prev=$$self{'__regions'}, curr=$start,$end\n");
            $self->__SortAndUpdate($start,$end);
            return;
		}
	}
	$$self{'__regions'} .= "$sep$start..$end";
	$$self{'__end'} = $end if ( $end > $$self{'__end'} ); # do not change the mRNA boundaries when it is already initialised (like in gff3)
}

=head2 procedure __SortAndUpdate

 Title        : __SortAndUpdate
 Procedure    : Same as Update, but inserts the new region so that $self->{'__regions'} is kept sorted
 Usage        : Called by Update if necessary
 Args         : $start: start position on the pseudomolecule
                $end: end position on the pseudomolecule
 Globals      : none

=cut
sub __SortAndUpdate
{
    my ($self,$start,$end) = @_;

    my @a_regions = $self -> GetRegions();
    push (@a_regions,"$start..$end");
	my @a_sorted_regions = sort __CmpRegions @a_regions;
    $self->{'__regions'} = join(',',@a_sorted_regions);
    $$self{'__end'} = $end if ( $end > $$self{'__end'} ); # do not change the mRNA boundaries when it is already initialised (like in gff3)
    #warn "KOUKOU " . $self->{'__regions'} . "\n";
    
    sub __CmpRegions
    {
        my ($e_a) = $a =~ /(\d+)$/;
        my ($s_b) = $b =~ /^(\d+)/;
        return -1 if ($e_a<$s_b);
        
        my ($s_a) = $a =~ /^(\d+)/;
        my ($e_b) = $b =~ /(\d+)$/;
        return 1 if ($e_b<$s_a);
        
        # NOTE - If 2 regions are identical, this IS an overlap - So , we NEVER return 0
		# return 0 if ($s_a==$s_b and $e_a==$e_b);

		throw Error::Simple ( "Locus::AddRegion: overlap between those regions $a $b" );
    }
}

=head2 Function GetSequence

 Title        : GetSequence
 Function     : Returns the sequence of a locus, initializes the __length private field
 Usage        : my ($len,$seq) = $o_locus->GetSequence($r_seq);
 Returns      : ($len, $seq)
 Args         : $r_seq: reference on a fasta sequence
 Globals      : none
 TODO - STUPIDE DE RENVOYER DEUX TRUCS, GetLength permet de connaitre la longueur de la sequence !!!
=cut

sub GetSequence
{
	my ($self,$r_seq) = @_;
	my $seq='';
	my $len=0;
	my @a_regions = $self -> GetRegions();
	my $strand    = $self -> GetStrand();

	my $seqlen = length($$r_seq);

	foreach my $region (@a_regions)
	{
		my($start,$end)=split(/\.\./,$region);
		my $len = $end-$start+1;
		if ( $seqlen < $end )
		{
			throw Error::Simple("ASSEMBLY ERROR: coordinates out of range: sequence length=$seqlen, region [$start..$end], check the assembly");
		}
		$seq .= substr ($$r_seq,$start-1,$len);
	}
	if ( $strand < 0 )
	{
		my $tmpseq = reverse($seq);
		$tmpseq =~ tr/ATGCatgc/TACGtacg/;
		$seq = $tmpseq;
	}
	$len = length($seq);
	$self -> { __length } = $len;
	return($len,$seq);
}

=head2 Function GetSequenceLength

 Title        : GetSequenceLength
 Function     : Returns the sequence length, calling GetSequence if necessary
 Usage        : my $seq = $o_locus->GetSequenceLength($r_seq);
 Returns      : The sequence length
 Args         : $r_seq: reference on a fasta sequence - This is NOT required if GetSequence was already called once
 Globals      : none
 TODO - HOULALA Pas super ce truc il vaudrait mieux passer $r_seq dans le constructeur !!!
=cut

sub GetSequenceLength
{
	my ($self,$r_seq) = @_;
	
	return $self -> { __length } if (defined $self -> { __length });
	if (! defined $r_seq)
	{
		croak ("Fatal error - GetSequence was never called, please give r_seq in parameter");
	}
	else
	{
		$self->GetSequence( $r_seq);
		return $self -> { __length };
	}
}

=head2 procedure GetParent

 Title    : GetParent
 Function : Return the parent locus
 Usage    : my $o_parent = $o_locus -> GetParent();
 Args     : none
 Return   : an object, generally Locus, but may be some other class (at top of string)
 Globals  : none

=cut
sub GetParent
{
	my $self = shift;
	return $self -> { __o_parent };
}

=head2 procedure GetAncestor

 Title    : GetAncestor
 Function : Return the ancestor locus, ie the farest parent (should be an AnnotatedSeqFile object)
 Usage    : my $o_ancestor = $o_locus -> GetAncestor();
 Args     : none
 Return   : an AnnotatedSeqFile object
 Globals  : none

=cut
sub GetAncestor
{
	my $self = shift;
	return undef unless defined $self -> GetParent();
	return (defined ($self -> GetParent() -> can ("GetAncestor")) ? $self -> GetParent() -> GetAncestor() : $self -> GetParent());
}

=head2 Functions GetKingdom, GetSpecies,GetChromosomeLabel,GetCodonTable, GetContigInfo

 Title    : GetKingdom, GetSpecies,GetChromosomeLabel
 Function : Call GetKingdom, Getspecies, GetChromosomeLabel from the locus ancestor (see GetAncestor)
 Usage    : $chr = $o_locus -> GetChromosomeLabel();
 Args     : none
 Globals  : none

=cut
sub GetKingdom
{
	my $self = shift;
	return $self -> GetAncestor() -> GetKingdom();
}
sub GetSpecies
{
	my $self = shift;
	return $self -> GetAncestor() -> GetSpecies();
}
sub GetChromosomeLabel
{
	my $self = shift;
	return $self -> GetAncestor() -> GetChromosomeLabel();
}
sub GetCodonTable
{
	my $self = shift;
	return $self -> GetAncestor() -> GetCodonTable();
}
sub GetContigInfo
{
	my $self      = shift;
	my ($contig ) = @_;
	return $self -> GetAncestor() -> GetContigInfo( $contig );
}

=head2 function Get

 Title        : Get
 Function     : Returns the value of a given attribute
 Usage        : my $att = $o_locus->Get('att');
 Prerequisite : none
 Returns      : none
 Args         : $tag: key of the attribute
 Globals      : none
 TODO - A VIRER ASAP - MANU

=cut

sub Get
{
my ($self,$tag) = @_;

	if ( defined($$self{'__'.$tag}))
	{
		return $$self{'__'.$tag};
	}
	return '';
}

=head2 procedure Print

 Title        : Print
 Function     : Print a selection of features attached to the locus (All mandatory and some common optional)
 Usage        : $o_locus->Print();
 Returns      : none
 Args         : none

=cut

sub Print
{
	my($self) = @_;

	foreach my $field ('__locus_tag','__start','__end','__strand','__type','__accession','__definition','__product')
	{
		print $self->{$field}."\t" if ( defined($self->{$field}));
	}
	if ( defined($self->{'__regions'}))
	{
		print $self->{'__regions'};
	}
	print "\n";
}

=head2 procedure DumpFastaNarcisse

 Title        : DumpFastaNarcisse
 Function     : Print on the current filehandle the fasta sequences corresponding to the CDS, the protein, the mrna and the gene
                Warning: only protein coding genes are considered
				Warning: when mode is set, the procedure mimics the procedure that dump gene information (in Locus/Gene.pm)
 Usage        : print $o_gene->DumpFastaNarcisse($chr,$sp,$r_seq,$mode,$kingdom)
 Prerequisite : none
 Returns      : none
 Args         : $chr: chromosome id
                $sp: species short code
                $r_seq: reference on the fasta sequence
                $mode: prot|mrna_old|cds|gene
                $kingdom: bacteria|eukaryota
				$table: condon usage table
 Globals      : none

=cut

sub DumpFastaNarcisse
{
	my($self,$chr,$sp,$r_seq,$mode,$kingdom,$table)=@_;

	my $code = $mode;
	if ( $mode =~ /mrna|gene|mrna_old|cds|prot/ )
	{
		my $region = $self->GetRegions();
		my %h_param = $self->GetAll('locus_tag','strand','product','gene_name','accession');
		my ($len,$sseq) = $self->GetSequence($r_seq,$h_param{strand},$region);
		return if ( $len == 0);
		$sseq  =~ s/(\w{60})/$1\n/g;
		my $gn = ( defined($h_param{gene_name}) && $h_param{gene_name} ne '' )  ? " gn=$h_param{gene_name}" : "";
		print ">$code:$h_param{accession} locus_tag=$h_param{locus_tag}$gn len=$len chr=$chr strand=$h_param{strand} sp=$sp $h_param{product}\n$sseq\n";
	}
}

=head2 procedure GetForNarcisse

 Title        : GetForNarcisse
 Function     : Return a hash with the locus properties useful for Narcisse
 Usage        : my $rh_row = $o_locus -> GetForNarcisse()
 Args         : None

=cut

sub GetForNarcisse
{
	my $self = shift;

	my $chr          = $self -> GetChromosomeLabel();
	my $uid          = $self -> GetNarcisseId();
	my $locus_tag    = $self -> GetLocusTag();
	my ($start,$end) = $self -> GetEnds();
	my $strand       = $self -> GetStrand();
	my $type         = $self -> { __type };

	my @a_h_locus = qw( chrom element symbol start end strand type );
	my %h_rvl;
	@h_rvl{ @a_h_locus } = ( $chr, $uid, $locus_tag, $start, $end, $strand, $type );
	return ( \%h_rvl );
}


=head2 procedure DumpGFF3

 Title        : DumpGFF3
 Function     :
 Usage        :
 Prerequisite : none
 Returns      : none
 Args         : $chr: chromosome ID
                $uid: locus unique ID
 Globals      : none
 TODO - VERIFIER QUE CA MARCHE ENCORE (MANU)

=cut

sub DumpGFF3
{
my($self,$chr,$uid,$mode)=@_;

	print "$chr\tsource\t$mode\t$self->{__start}\t$self->{__end}\t.\t$self->{__strand}\t.\tID=$self->{__locus_tag} ou $uid\n";
	# voir si c'est locus_tag ou $uid
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut


1;
