#
# $Id: Gene.pm 2955 2009-10-23 08:48:47Z manu $
#

package Locus::Cds;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Cds - describes a Cds

=head1 SYNOPSIS

 my $o_gene = New Locus::Cds('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_gene);

=head1 DESCRIPTION

 A Cds extends the Locus class
 In the constructor, the parent MUST be a Locus::Mrna object

=cut


use strict;
use warnings;
use General;
use Carp;
use Bio::Seq;
use Data::Dumper;

use base qw( Locus );


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 2955 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, see Locus::_Init for the details

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$self->SUPER::_Init(%h_param);

	# The default is 'consistent'
	$$self{ __consistency } = 'C';


}

=head2 function GetConsistency

 Title        : GetConsistency
 Function     : Return the consistency flag
 Usage        : my $consis = $o_locus->GetConsistency();
 Returns      : 'C', 'I', 'F'
 Args         : none
 Globals      : none

=cut

sub GetConsistency
{
	my ($self) = @_;
	return $self -> { __consistency };
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_cds->setParent( $o_mrna );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set, $o_mrna should be a Locus::Mrna object
 
 Args         : $o_rmna : The parent mRNA
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_rmna) = @_;
	
	if ( ref($o_rmna) ne 'Locus::Mrna' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be a Locus::Mrna object' );
	}
	$self->SUPER::SetParent( $o_rmna );

	my $nc_id = $o_rmna -> GetNarcisseId();
	
	# caeel_sIg1t1 ==> caeel_sIg1c1
	$nc_id    =~ s/t(\d+)$/c$1/;
	$self -> { __narcisse_id } = $nc_id;
}

=head2 Function GetProteinSequenceLength

 Title        : GetProteinSequenceLength
 Function     : If consistent or fragment: returns the sequence length, calling Locus::GetLength, thus maybe Locus::GetSequence
				If inconsistent: returns 0
 Usage        : my $seq = $o_locus->GetProteinSequenceLength($r_seq);
 Returns      : The proteinsequence length
 Args         : $r_seq: reference on a fasta sequence - This is NOT required if GetSequence was already called once
 Globals      : none
 TODO - HOULALA Pas super ce truc il vaudrait mieux passer $r_seq dans le constructeur !!!
=cut

sub GetProteinSequenceLength
{
	my ($self,$r_seq) = @_;
	if ($self -> GetConsistency eq 'I')
	{
		return 0;
	}
	else
	{
		# Return the nucleic sequence length / 3 and take into account the start codon
		return int ( $self -> GetSequenceLength($r_seq) / 3 ) - 1;
	}
}

=head2 procedure GetTranslation

 Title     : GetTranslation
 Usage     : my $prot_seq = $o_cds->GetTranslation();
 Function  : Compute and return the protein translation of the CDS, using bioperl
             MAY SET THE CONSISTENCY FLAG OF THE OBJECT (thus the object is MODIFIED by this function)

 Args      : $r_seq: Ref to a string containing the genomic sequence
 Return    : The translated protein, "" if nothing to translate or if inconsistency found

=cut
sub GetTranslation
{
	my $self = shift;
	
	my ($r_seq) = @_;
	
	# TODO - A REVOIR - Ensembl NOT SUPPORTED YET
	my $data_source_ensembl = &FALSE;					
	# JG changement suite au merge API my ($len,$sseq) = $self->GetSequence($r_seq,$h_param{strand},$exoncds);
	my ($len,$sseq);
	if ( $data_source_ensembl )
	{
		($len,$sseq) = $self->GetParent()->GetSequence($r_seq);
	}
	else
	{
		#($len,$sseq) = $self->GetSequence($r_seq,$h_param{strand},$exoncds);
		($len,$sseq) = $self->GetSequence($r_seq);
	}
	return "" if ( $len == 0 );

	if ( $data_source_ensembl )
	{
		# On a deja fait un GetSequence pourquoi on le corrige ? (MANU)
		my $o_mrna = $self -> GetParent();
		$sseq = substr($sseq,$o_mrna->Get('cds_start')-1,$o_mrna->Get('cds_end')-$o_mrna->Get('cds_start')+1);
		$len = length $sseq;
		#printf STDERR "$sseq#%d\n",$r_mrna->Get('cds_end')-$r_mrna->Get('cds_start')+1;
	}
				
	# Useful for warning or error messages
	my $gene_locus_tag = $self -> GetLocusTag();

	# Detecting fragment or inconsistent proteins:
	#
	# 1/ If START or STOP codon is missing, the protein is a FRAGMENT
	# 2/ If there is a STOP codon inside the CDS, the protein is a FRAGMENT and we try another phase
	# 3/ If there is still a STOP codon inside the CDS, AND we are in phase 2, we give up and the protein is INCONSISTENT

	$self -> SetFragment() if ($len % 3 != 0);
	my $display_id = "prot:" . $self -> GetAccession();
	#my $gene_obj     = Bio::Seq->new( -display_id => "$display_id", -seq => $sseq);
	my $gene_obj     = Bio::Seq->new( -display_id => "$display_id" );
	$gene_obj -> seq ( $sseq, 'dna' );
	my $seq_prot;
	foreach my $phase (0,1,2)
	{
		#my $prot_obj = $gene_obj->translate(
		#	-codontable_id => $self -> GetCodonTable(),
		#	-frame         => $phase,
		#	-complete      => 1,
		#	);
		my $prot_obj = $gene_obj->translate(undef,undef,$phase,$self -> GetCodonTable(),1);

		$seq_prot = $prot_obj->seq;
		
		# If it does not start with M, it is a fragment
		$self -> SetFragment() if ( $seq_prot !~ /^M/ );

		# If we have stop codons in the middle, it is at least a fragment, but we check another phase
		if ( $seq_prot =~ /\*/ )
		{
			$self -> SetFragment();
			#... and if in phase 2, we mark the protein as inconsistent before giving up
			$self -> SetInconsistent() if ($phase == 2);
			warn "WARNING gene $gene_locus_tag - Gene::DumpFastaNarcisse: in-phase stop codon for $display_id, phase $phase\n>$seq_prot\n";
			next;
		}
	
		# OK - No terminator codon inside CDS
		# If we have a stop codon at end it's OK
		# NOTE - We have to check the end codon at the nucleic level because the terminator is chopped by the translate function
		my ($lastcodon) = ($sseq =~ /(\w{3})$/);
		#print "\nLASTCODON $lastcodon\n";
		if  ( $lastcodon =~ /(TAA|TGA|TAG)$/i)
		{
			last;
		}
		
		# if there is no stop codon, or if stop codon contains 'N', as in EuGene, it is a fragment
		else
		{
		  $self -> SetFragment();
		  last;
		}
	}
	
	# display the consistency status if necessary
	if ($self -> GetConsistency() eq 'I')
	{
		warn "WARNING gene $gene_locus_tag - Gene::DumpFastaNarcisse: protein $display_id is inconsistent (removed)";
		$seq_prot = "";
	}
	elsif ( $self -> GetConsistency() eq 'F' )
	{
		warn "WARNING gene $gene_locus_tag - Gene::DumpFastaNarcisse: protein $display_id is a fragment";
	}
	
	return $seq_prot;
} 

=head2 procedure SetFragment,SetInconsistent

 Title        : SetFragment,SetInconsistent
 Procedure    : Set the consistency flag to 'F' (fragment) or 'I' (inconsistent)
                when marked as 'I' it is not possible to return to the 'F' state
                when marked as 'F' or 'I', it is not possible to return to the 'C' state
 Usage        : $o_locus->SetFragment()
                $o_locus->SetInconsistent()
 Globals      : none

=cut

sub SetFragment
{
	my ($self) = @_;
	$self -> { __consistency } = 'F' unless $self -> { __consistency } eq 'I';
}
sub SetInconsistent
{
	my ($self) = @_;
	$self -> { __consistency } = 'I';
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


