#
# $Id: Gene.pm 3410 2010-07-06 12:12:25Z manu $
#

package Locus::Gene;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Gene - describes a protein coding gene

=head1 SYNOPSIS

 my $o_gene = New Locus::Gene('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_locus);

=head1 DESCRIPTION

 A Gene extends the Locus class, mainly including an array to store the mRNAs (Locus::Mrna)

=cut

=head1 NAME

Locus::Gene - describes a protein coding gene

=head1 SYNOPSIS

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Data::Dumper;
use General;

use Locus;
use base qw( Locus );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 3410 $ =~ /\d+/g); $r[0]};
}

# PRE2_1 = 1 ==> retrouver le comportement de narcisse 2.0
use constant PRE2_1 => 0;

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_gene = New Gene('start'=>10,'end'=>100,'strand'=>-1,'locus_tag'=>'GNI','type'=>'gene','parent' => $o_annot,'product'=>'enzyme');
 Prerequisite : start < end, strand =~ /1|\-1|\+|\-/
 Returns      : none
 Args         : Mandatory: start      position of the start (pb) of the object on the chromosome
                Mandatory: end        position of the end (bp) of the object on the chromosome
                Mandatory: strand     identify the strand on which is located the object; values=1/-1|+/-
                Mandatory: locus_tag  accession on the molecule (expected to by unique)
                Mandatory: type       type of the object
                Optional: any other pair of key/value
 Globals      : none
 TODO type est INUTILE MAINTENANT IL SERA SUPPRIME ASAP

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$self->SUPER::_Init(%h_param);

	# An array of Locus::Mrnas objects
	$$self{'__a_mrnas'} = [];
	
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_gene->SetParent( $o_annot );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set
 
 Args         : $o_annot : The parent, an AnnotatedSeqfile.pm object (also a chromosome !)
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_annot) = @_;
	
	$self->SUPER::SetParent( $o_annot );
	
	my $nc_id = $o_annot -> GetNarcisseId();
	$nc_id .= 'g' . $o_annot->GetNbGenes();
	$self -> { __narcisse_id } = $nc_id;
}

=head2 GetNbMrnas

 Title    : GetNbMrnas
 Function : Accessors to the number of mRNAS or the number of CDS
 Usage    : my $nb = $o_gene -> GetNbMrnas();
 Return   : The nb of mrnas
 
=cut
sub GetNbMrnas
{
	my $self = shift;
	my $ra_mrna = $self -> { '__a_mrnas' };
	return scalar @$ra_mrna;
}

=head2 GetMrnas

 Title    : GetMrnas
 Function : Accessor to the array of mRNA
 Usage    : my $ra_mrnas = $o_gene -> GetMrnas();
 Arg      : None
 Return   : A ref to an array of mRNAs objects
 
=cut
sub GetMrnas
{
	my $self = shift;
	return $self -> { __a_mrnas };
}

=head2 GetMrna, GetCDS

 Title    : GetMrna, GetCDS
 Function : Accessors to a single mRNA or CDS
 Usage    : my $o_mrna = $o_gene -> GetMrna($index);
            my $o_cds  = $o_gene -> GetCDS($index);
 Arg      : $index, the index of the mRNA or CDS to retrieve
 Return   : Locus::Mrna or Locus::CDS object found at index $index
 TODO - A SUPPRIMER ??? PAS ENCORE cf LES SANITY CHECKS et AnnotatedSeqfile !!! (MANU)
 
=cut
sub GetMrna
{
	my $self = shift;
	my ($index) = @_;
	
	my $ra_mrna = $self -> { '__a_mrnas' };
	return $ra_mrna -> [ $index ];
}

=head2 procedure AddmRNA

 Title        : AddmRNA
 Function     : Add a new mRNA, maybe modifying the start/end of the gene
 Usage        : $ind = $o_gene->AddmRNA($o_mrna)
 Returns      : The added mRNA
 Args         : $o_mrna: A Locus::Mrna object to add
 Globals      : none

=cut

sub AddmRNA
{
	my ($self,$o_mrna) = @_;
	throw LipmError::ParameterException( 'o_mrna', 'type', -text => 'should by a Locus::Mrna object' ) unless ref $o_mrna eq 'Locus::Mrna';
	
	my $ra_mrna = $self -> { __a_mrnas };
	push @$ra_mrna, $o_mrna;
	my ($mrna_start,$mrna_end) = $o_mrna -> GetEnds();
	
	### MANU - Si necessaire j'agrandis un peu le gene - Important dans certains cas (gff3):
	#          - Pas de description de gene
	#          - Second transcrit plus grand que le premier = le gene est cree a partir du premier, le second depasse
	#          - Commenter les deux lignes ci-dessous pour retrouver le comportement de la 2.0 !
	# TODO - MERDOUILLE 
	if ( &PRE2_1 == 0 )
	{
		$self -> { __start } = $mrna_start if ( $mrna_start < $self -> { __start } );
		$self -> { __end }   = $mrna_end   if ( $mrna_end   > $self -> { __end } );
	}
	$o_mrna -> SetParent ( $self );

	return $o_mrna;
}


=head2 procedure Print

 Title        : Print
 Function     : Print the content of the gene
 Usage        : my $o_gene->Print();
 Prerequisite : none
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Print
{
my ($self) = @_;

	if ( ! $self->{'__nb_cds'} && ! $self->{'__nb_mrna'} )
	{
		print "Warning: ".$self->{'__locus_tag'}." removed, does not contain neither mRNA nor CDS (see ncRNA)\n";
		return;
	}
	$self->SUPER::Print();
	print STDOUT "\tMRNA ....................\n";
	for(my $i=0; $i<$$self{'__nb_mrna'}; $i++)
    {
		printf "\t%d\t",$i+1;
		$$self{'__a_mrnas'}[$i]->Print();
    }
	print STDOUT "\tCDS ....................\n";
	for(my $i=0; $i<$$self{'__nb_cds'}; $i++)
    {
		printf "\t%d\t",$i+1;
		$$self{'__a_cds'}[$i]->Print();
    }
}

=head2 function IsProteinCoding

 Title        : IsProteinCoding
 Function     : Return true if the gene contains at least 1 CDS
 Usage        : my $isproteincoding = $o_gene->IsProteinCoding();
 Prerequisite : none
 Returns      : Return true/false

=cut

sub IsProteinCoding
{
	my($self) = @_;
	my $ra_mrna = $self -> { __a_mrnas };
	
	foreach my $o_mrna (@$ra_mrna)
	{
		return &TRUE if defined $o_mrna -> GetCDS();
	}
	return &FALSE;
}

=head2 procedure SyncCDS2Mrna

 Title        : SyncCDS2Mrna
 Function     : If necessary, synchronize the mRNAs from the CDS
                Synchonization is necessary for an mRNA if the number of exons in the mRNA is lower than the number of exons in the corresponding CDS
 Usage        : my $o_gene->UpdCDS($r_locus,$index_cds)
 Returns      : none
 Args         : $r_obj: a basic locus object 
                $index_cds: index of the CDS to be updated
 Globals      : none

=cut

sub SyncCDS2Mrna
{
	my $self      = shift;
	my ($verbose) = @_;

	my $ra_mrnas = $self -> { __a_mrnas };

	foreach my $o_mrna (@$ra_mrnas)
	{
		my $o_cds = $o_mrna -> GetCDS();
		next unless defined $o_cds;

		my @a_exons_cds = $o_cds -> GetRegions();
		my @a_exons_mrna= $o_mrna-> GetRegions();
		
		if ( $#a_exons_mrna < $#a_exons_cds )
		{
			printf STDERR "WARNING - Gene::SyncCDS2Mrna: mRNA %s updated to match the CDS %s\n", $o_mrna->GetLocusTag(),$o_cds->GetLocusTag() if ( $verbose );
			# MANU - Je ne synchronise pas les start/end, parce que dans GFF3 j'ai une description des limites du transcrit
			#        et si je synchronsie start/end je perds de l'information !
			#        Il faut decommenter ci-dessous pour retrouver le comportement de narcisse 2.0
			if ( &PRE2_1 == 1 )
			{
				$o_mrna->{'__start'}  = $o_cds->{'__start'};
				$o_mrna->{'__end'}    = $o_cds->{'__end'};
			}
			$o_mrna->{'__strand'} = $o_cds->{'__strand'};
			$o_mrna->{'__regions'}= $o_cds->{'__regions'};
		}
		else
		{
			warn "Nothing to sync " . $o_mrna->GetLocus() if ( $verbose );
		}
    }
}




=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
