#
# $Id: Gene.pm 2955 2009-10-23 08:48:47Z manu $
#

package Locus::Mrna;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Mrna - describes a mRNA

=head1 SYNOPSIS

 my $o_gene = New Locus::Mrna('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_gene);

=head1 DESCRIPTION

 A Mrna extends the Locus class, mainly including a Locus::CDS object
 In the constructor, the parent MUST be a Locus::Gene object

=cut

use strict;
use warnings;
use Carp;
use Bio::Seq;
use Data::Dumper;

use Locus;

use base qw( Locus );


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 2955 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, see Locus::_Init for the details
                Copy constructor (constructing an Mrna from a Cds object)
 Usage        : 1/ my $o_rmna = New Locus::Mrna ( %h_param );
                2/ my $o_rmna = New Locus::Mrna ( $o_cds, %h_param );
                   where $o_cds is a Locus::CDS object
                NOTE - We ignore the UTR when building a mRNA from a CDS

=cut

sub _Init
{
	my $self = shift;
	
	# 1 parameter = it SHOULD BE a $o_cds object
	# We copy only some parameters from $o_cds to the new object
	my %h_param;
	if (ref $_[0] eq 'Locus::Cds')
	{
		my $o_cds =shift;
		(%h_param) = @_;
		
		$h_param { start }     = $o_cds -> { __start_contig }     unless defined $h_param{ start };
		$h_param { end }       = $o_cds -> { __end_contig }       unless defined $h_param{ end };
		$h_param { strand }    = $o_cds -> { __strand }    unless defined $h_param{ strand };
		
		$h_param { rh_contig_info } = $o_cds -> { __rh_contig_info }    unless defined $h_param{ rh_contig_info };
		$h_param { locus_tag } = $o_cds -> { __locus_tag } unless defined $h_param{ locus_tag };
		#$h_param { type }      = $o_cds -> { __type }     unless defined $h_param{ type };
		$h_param { gene_name } = $o_cds -> { __gene_name } unless defined $h_param{ gene_name };
		$h_param { accession } = $o_cds -> { __accession } unless defined $h_param{ accession };
	}
	elsif ( ref $_[0] eq '' )
	{
		(%h_param) = @_;
	}
	else
	{
		my $bad_type = ref $_[0];
		throw LipmError::ParameterException( 'o_cds', 'type', -text => "should by a Locus::Cds object OR a scalar, NOT a $bad_type" );
	}
	
	$self->SUPER::_Init(%h_param);

	# A Locus::CDS object living inside this Mrna
	$$self{'__o_cds'} = undef;
	
	# See SetParent and GetLocusNarcisse Id 
	$$self{'__narcisse_index' } = 0;
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_locus->setParent( $o_gene );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set, $o_gene should be a Locus::Gene object
 
 Args         : $o_gene : The parent gene
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_gene) = @_;
	
	if ( ref($o_gene) ne 'Locus::Gene' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be a Locus::Gene object' );
	}
	$self->SUPER::SetParent( $o_gene );
	
	my $nc_id = $o_gene -> GetNarcisseId();
	$nc_id .= 't' . $o_gene->GetNbMrnas();
	$self -> { __narcisse_id } = $nc_id;
	$self -> { __narcisse_index } = $o_gene->GetNbMrnas();
	
}

=head2 function GetGeneName

 Title        : GetGeneName
 Function     : Return the gene name, if not found the cds gene name, or ""
 Usage        : my $gene_name = $o_mrna->GetGeneName();
 Returns      : The gene name (if possible)

=cut

sub GetGeneName
{
	my ($self) = @_;

	if ( defined $self -> { __gene_name } )
	{
		return $self -> { __gene_name };
	}
	elsif ( defined $self -> { __o_cds } )
	{
		return $self -> { __o_cds } -> GetGeneName();
	}
}

=head2 procedure GetNarcisseLocusId

 Title     : GetNarcisseLocusId
 Usage     : my $id = $o_locus->GetNarcisseLocusId()
 Function  : If the gene locus tag is different from the Mrna locus tag, return the Mrna locus tag
             If they are equal and there is only ONE mrna, return the locus tag
             If they are equal BUT there are several mrnas return a Narcisse locus tag: gene_locus_tag#Nb
 
 Args      : none
 Return    : A built locus id
 Globals   : none

=cut
sub GetNarcisseLocusId
{
	my $self = shift;
	
	my $locus_tag        = $self -> GetLocusTag();
	my $o_parent         = $self -> GetParent();
	my $parent_locus_tag = $o_parent -> GetLocusTag();
	
	if ( $locus_tag eq $parent_locus_tag && $o_parent -> GetNbMrnas() > 1 )
	{
		return $locus_tag . '#' . $self -> { __narcisse_index };
	}
	else
	{
		return $locus_tag;
	}
}

=head2 procedure AddCDS

 Title        : AddCDS
 Usage        : $o_locus->AddCDS( $o_cds );
                $o_locus->AddCDS( $o_locus );
 Procedure    : If the CDS is not defined: Set the CDS, calling SetParent on the CDS
                In this case, the parameter MUST BE a Locus::Cds object
                If the CDS is already defined: Add the region to the CDS
                In this case, the parameter MAY BE any Locus object
 Args         : $o_locus: The CDS to add or the region to add to the CDS
                NOTE - $o_locus MAY be modified by this procedure
 Globals      : none

=cut
sub AddCDS
{
	my $self = shift;
	my ($o_locus) = @_;
	
	if ( defined $self -> { __o_cds } )
	{
		my $o_cds = $self -> { __o_cds };
		$o_cds -> AddRegion ( $o_locus -> GetEndsContig() );
	}
	else
	{
		if ( ref($o_locus) ne 'Locus::Cds' )
		{
			throw LipmError::ParameterException( 'o_cds', 'type', -text => 'should be a Locus::Cds object' );
		}
		$o_locus -> SetParent ( $self );
		$self    -> { __o_cds } = $o_locus;
		
		# A CDS MUST have at least 1 region defined: add the region if necessary
		if ( $o_locus->GetRegions() eq '' )
		{		
			$o_locus -> AddRegion ( $o_locus -> GetEndsContig() );
		}
	}
}

=head2 procedure GetCDS

 Title        : GetCDS
 Usage        : my $o_cds = $o_locus->GetCDS();
 Function     : Accessor to the CDS
 Args         : none
 Globals      : none

=cut
sub GetCDS{
	my $self = shift;
	if ( defined $self -> { __o_cds } )
	{
		return $self -> { __o_cds };
	}
	else
	{
		warn "WARNING !!! mrna = " . $self -> GetLocusTag() . " CDS NOT DEFINED ";
		return undef;
	}
}

=head2 function GetTranslationStartStop

 Title        : GetTranslationStartStop
 Function     : Identify the translational start on a mRNA
 Usage        : my ($min_trans,$max_trans,$nb_exons) = &GetTranslationStartStop($strand,$mrna_region,$cds_region)
 Prerequisite : none
 Returns      : $min_trans: lower boundary participating to the translation   (ie the cds start, relative to the Mrna start)
                $max_trans: higher boundary participating to the translation  (ie the cds stop, relative to the Mrna end)
                $nb_exons: number of mrna exons
                (-1,-1,-1) is returned when there is an inconsistency (whether already set by Gene::DumpFastaNarcisse OR set by this function)
                WARNING - The Inconsistency status of this mrna object MAY BE MODIFIED !
 Args         : $strand: the strand of the locus
                $mrna_region: a comma-separated list of mrna exons
                $cds_regions: a comma-separated list of CDS exons
 Globals      : none

=cut

sub GetTranslationStartStop
{
	my $self = shift;
	
	my $strand = $self -> GetStrand();
	my @a_exon = $self -> GetRegions();
	my ($nb_exons,$len_mrna) = $self -> GetRegionNbExonsAndLength();
	my $o_cds  = $self -> GetCDS();
	return (-1,-1,-1) if ($o_cds -> GetConsistency() eq 'I');
	
	my $cds_region = $o_cds -> GetRegions();
	
	# 1..10,50..100  ==> mincds=1,maxcds=100
	my ($mincds) = $cds_region =~ /^(\d+)/;
	my ($maxcds) = $cds_region =~ /(\d+)$/;

	my $mintrans=0;
	foreach my $exon (@a_exon)
	{
		my($start,$end) = split(/\.\./,$exon);
		if ( $mincds > $end )
		{
			$mintrans += ($end-$start+1);
		}
		else
		{
			$mintrans +=  ($mincds-$start+1);
			last;
		}
	}
	
	my $maxtrans=0;
	foreach my $exon (reverse @a_exon)
	{
		my($start,$end) = split(/\.\./,$exon);
		if ( $maxcds < $start )
		{
			$maxtrans += ($end-$start+1);
		}
		else
		{
			$maxtrans +=  ($end-$maxcds+1);
			last;
		}
	}
	if ( $mintrans < 1 || $maxtrans < 1 )
	{
		print STDERR "Warning: strong inconsistency between CDS and mRNA detected (" . $self -> GetAccession() .")\n";
		$o_cds -> SetInconsistent();
		return(-1,-1,-1);
	}
	
	$maxtrans = $len_mrna-$maxtrans+1;
	my $mod = ((abs($maxtrans-$mintrans)+1 ) % 3);
	if ( $mod != 0 )
	{
		$maxtrans -= $mod;
		#print	STDERR "Warning: not modulo 3 subsequence " . $self -> GetAccession() . "($len_mrna) \n" . "regions mrna = " .$self->GetRegions() . "\nregions  cds = $cds_region \n";
		print	STDERR "Warning: not modulo 3 subsequence " . $self -> GetAccession() . "($len_mrna) \n";
		$o_cds -> SetFragment();
	}
	
	if ( $strand < 0 )
	{
		my ($min,$max) = ($mintrans,$maxtrans);
		$mintrans = $len_mrna - $max + 1;
		$maxtrans = $len_mrna - $min + 1;
	}

	return($mintrans,$maxtrans,$nb_exons);

}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


