#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: October 24, 2008
#	$Id:  $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package MapArea;

use strict;
use warnings;
use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf 

use base qw( LipmObject );

=head1 NAME

MapArea - An object to store and manage boxes for a subsequent use the maps of html pages
          The principle of this object is to delegate to the ElementsSet objects the 
          management of all the map information

=head1 SYNOPSIS

1 The simple case

my $o_maparea   = MapArea->New();

my $o_anyobject = New AnyObject();   #Any object you like. You must however have a method DumpMapInformation which does the job

#Storing $o_anyobject as an object that deserves a map
$o_maparea->NewElementSet( "my_any_obj_id", $o_object );

#This method calls the DumpMapInformation method of all the registered object
$o_maparea->DumpMapInformation( $map_file);

2 You have the possibility to register a map area for subsequent use

my $o_maparea   = MapArea->New();

my $map_area = { any information you would like to keep }

$o_maparea->RegisterElement( "my_map_area_id", $map_area );

if ( $o_maparea->IsDefined( "my_map_area_id" ))
{
	my $map_area = $o_maparea->Get( "my_map_area_id" );	
}

This mechanism enables to link objects that have been drawn (and registered)

=head1 DESCRIPTION

The MapArea object enables to manage the elements displayed box in order to generate the map information 
in the html pages.

=head1 SUBROUTINES

	NewElementSet
	RegisterElement
	IsMapDefined
	GetMapElement
	DumpMapInformation
	
=cut

use Data::Dumper;
use Carp;
use General;
use LipmError;
use LipmError::ObjectException;

use ElementsSet;

#------------------------------------------------#
#               Constructor                      #
#------------------------------------------------#

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New MapArea(?);
 Prerequisite : none
 Fonction     : initialise an instance of Pairwise class.
 Returns      : a valid Pairwise object
 Args         : ?

=cut
sub _Init
{
	my $self = shift;
	my %h_param= (  'reference' => undef,
					'verbose' => 0,
							@_
					);
			
	my @mandatory_param = qw( reference );
	
	foreach my $param ( @mandatory_param )
	{
		(defined $h_param{$param})
			or throw LipmError::ParameterException($param,'missing');	
	}
	@{$self}{@mandatory_param} = @h_param{@mandatory_param};
		
	$self->{ __displayed_elementsets }={};
																														
	return $self;
}

# DEVRAIT EVITER LES FUITES DE MEMOIRE EN CASSANT LES CYCLES - NE MARCHE PAS BIEN
#sub DESTROY
#{
#	my $self = shift;
#	my $rh_displayed = $self->{ __displayed_elementsets };
#	foreach my $elt ( keys %$rh_displayed )
#	{
#		warn 'BOUM';
#		delete $rh_displayed -> { $elt };
#	}
#}

#------------------------------------------------#
#                  Methods                       #
#------------------------------------------------#

=head2 Procedure NewElementSet

 Title        : NewElementSet
 Usage        : &NewElementSet( $ref_element );
 Prerequisite : The stored object should have a DumpMapInformation method
 Procedure    : Adding an ElementSet object to the list of displayed Elements (objects wanting a map area to be written)
 Args         : $msg, a string

=cut
sub NewElementSet
{
	my $self = shift;
	my $ref_element = shift;
	
	#warn "New map elementset for $ref_element : ".$ref_element->Infos();
	
	$self->{ __displayed_elementsets }{ $ref_element } = $ref_element;
    
	#push @{$self->{ __displayed_elementsets }}, $ref_element;
		
}

=head2 Procedure RegisterElement

 Title        : RegisterElement
 Usage        : $o_xxx->RegisterElement( $element_id, $rh_area );
 Prerequisite : none
 Procedure    : registering an area element (enables a subsequent use to draw links between objects)
 Args         : $element_id and $rh_area

=cut
sub RegisterElement
{
	my $self = shift;
	my $element_id = shift;
	my $rh_area = shift;
	
	$self->{__index}{$element_id}=$rh_area;

}

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_areamap -> DumpMapInformation( $map_file )
 Function     : Calls the DumpMapInformation methods of all the registered obejsct
                into $map_file
 Returns      : none
 Args         : The map file (.coord file),
                => the output filename, erase if already exists.
 
=cut
sub DumpMapInformation
{
	my $self = shift;
	my $map_file = shift;
				
	#----------------------------------------
	# Open file
	my $file_handle = new IO::File(">$map_file") or throw LipmError::IOException($map_file,'f');

	# TODO - Actuellement on trie les keys en ordre alphabetique, c'est bien car snp vient APRES rna et il est utile
	#        que les lignes snp soient APRES rna, sinon la carte RNA se met par-dessus la carte SNP
	#        C'est juste un coup de bol, il faudra definir un ordre de tri plus subtil a l'avenir
	foreach my $elt_id ( sort keys  %{ $self->{ __displayed_elementsets } })
	{
    	my $ref_elt = 	$self->{ __displayed_elementsets }{ $elt_id };
		$ref_elt->DumpMapInformation( $file_handle );		
	}
	
}

=head2 GetMapElement

 Title        : GetElement
 Usage        : $o_element -> GetMapElement();
 Prerequisite :  IsMapDefined should have been called (?!)
 Procedure    : returns the stored information
 Returns      : the stored information
 Args         : 
 				
=cut
sub GetMapElement
{
	my $self = shift;
	my $element_id  = shift ;
		
	return $self->{__index}{$element_id};	
}

=head2 _GetElementInfos

 Title        : 
 Usage        :  $rh_infos = _GetElementInfos( $element_id )
 Function     : Extracts information and coordinate position of and element
 Returns      : a reference on a hash with the correspinding information
 Args         : The element id
 
TODO  Remove this method, same argument stands as below : (the MapArea should be agnostic to how boxes are stored !!)
 
=cut
sub _GetElementInfos
{
	my $self = shift;
	my $elt_id = shift;
	my $species = shift;
	my $ra_fields = shift;
	my $rh_infos;
	
	if (  ! $self->IsMapDefined( $elt_id ) )
	{
		$rh_infos = { coord => [0,0,0,0], values => ["","","",""] };
	} 
	else
	{
		my $rh_info = $self->{__index}{ $elt_id };
		my @coords  = _inflate(@{$rh_info->{boxes}{__primary}});
		my @values = map("\"$_\"",($species, @{ $rh_info->{ infos } }{ @$ra_fields }));
		$rh_infos = { coord => \@coords, values => \@values };
	}
	
	return $rh_infos;

	sub _inflate
	{
		my @a_coord = @_;
		
		return ( int($a_coord[0]-4) , int($a_coord[1]-2) ,int($a_coord[2]+4) , int($a_coord[3]+2) );
	}

}

sub GetTypeArea
{
	my $self = shift;
	my $type = shift;
	
	( defined $self->{__areas}{$type} )
		or throw LipmError::ParameterException("type ".$type,'undefined');
	
	return $self->{__areas}{$type};	
}

sub IsMapDefined
{
	my $self = shift;
	my $element_id =shift;
		
	return (defined $self->{__index}{$element_id});
}

=head2 Procedure NewElement

 Title        : NewElement
 Usage        : &NewElement( $type, $element_id, $rh_info, $ra_box );
 Prerequisite : none
 Procedure    : adding an element area of type $type
 Args         : $msg, a string

#=cut
sub NewElement
{
	my $self = shift;
	my $type = shift;
	my $organism = shift;
	my $element_id = shift;
	my $rh_info = shift;
	my $ra_box = shift;
			
	$self->{__areas}{$type}{$element_id}={ organism => $organism, type => $type, infos => $rh_info, boxes => { __primary => $ra_box}, elt => $element_id};
	$self->{__index}{$element_id}=$self->{__areas}{$type}{$element_id};
	push @{$self->{__organisms}{ $organism }}, $self->{__areas}{$type}{$element_id};
	
}
=cut

#------------------------------------------------#
#                  Accessors                     #
#------------------------------------------------#

#TODO move these methods to the ElemenSets object (the MapArea should be agnostic to how boxes are stored !!)

sub GetCenterY
{
	my $self = shift;
	my $element_id =shift;
	
	my $r_coord=$self->{__index}{$element_id}{boxes}{__primary};
	
	return ($r_coord->[1]+$r_coord->[3])/2;
		
}

sub GetCenterX
{
	my $self = shift;	
	my $element_id =shift;
	
	my $r_coord=$self->{__index}{$element_id}{boxes}{__primary};
	
	return ($r_coord->[0]+$r_coord->[2])/2;	
}

sub GetCoord
{
	my $self = shift;	
	my $element_id =shift;
		
	return @{$self->{__index}{$element_id}{boxes}{__primary}};	
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;

