#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 17, 2008
#	$Id: MultiService.pm 3835 2013-06-03 15:03:33Z  $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package MultiService;

=pod

=head1 NAME

 MultiService

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( LipmObject );

use Error qw( :try );
use Carp;

use General;
#use Appli;
use ParamParser;
use Logger;
# use Configuration::NarcisseObserver;
use ConfigurationFactory;
use LipmError;
use LipmError::ObjectException;
use Sequence_utils;
use Args_utils;

# le nom des options
use constant HELP       => 'help';
use constant MOBYLE     => 'mobyle_dir';

use constant INSTANCES  => 'instance';

use constant OUT_FILE   => 'output_file';

use constant LOG_LEVEL => 'IOVISVIDV';
use constant LOG_FLAG  => 'log';


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New MultiService( rh_descr => $rh_descr,
                                            rh_help  => $rh_help,
                                            log_file => $log_filename,
                                            ra_options => $ra_options,
                                            prog_name => $prog );
 Prerequisite : none
 Fonction     : initialise an instance of MultiService class.
 Returns      : a valid MultiService object
 Args         :
     $rh_descr, hash ref, description for Appli constructor.
     $rh_help, hash ref, contains strings to add to help message.
         base    => string, the base help, should not be used.
         params  => string, parameters to provide.
         options => string, options of the script.
     $log_filename, string, the name of the file where we will log via Logger.
     $ra_options, list ref of strings, elements are GetOptLong strings.
     $prog, string, the program's name
 Attributes   :
    __rao_instance_observer, ref list of InstanceObserver object
#     __ra_species, list ref of species short/long name (in fact what was given by --species )
    __rh_current, hash ref, values change while processing
        o_inst_obs, the instance observer object
        species, the species short name
    __rh_out, hash ref,
        ra_results, list ref of results, one result per species
             a result is a hash with 2 keys
                 hash => with NarcisseService::__rh_results value
                 list => with NarcisseService::__ra_results value
        rh_filename, hash ref
            results, string, name of the output file for results.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( rh_descr => undef,
	               rh_help  => {},
	               log_file => '',
	               ra_options => [],
	               prog_name  => 'prog',
	               @_ );

	# assert that nothing is not undef
	LipmError::ParameterException::Assert( %h_args );

	my $sub = $self->__GenerateHelp( @h_args{ qw( rh_help prog_name ) } );
	ParamParser::SetDefaultUsage( $sub );

	# collecte et lecture des options attendues
	my $ra_options = $self->_GetOptLongList( $h_args{ ra_options } );
# warn Dumper $ra_options;
	my $o_param = New ParamParser( 'GETOPTLONG', @$ra_options );
# my $rh_ = {};
# $o_param->Dump( 'HASH', $rh_ );
# warn Dumper $rh_;

	# on veut une trace de ce qui se passe
	if( $o_param->IsDefined( &LOG_FLAG ) && $h_args{ log_file } ne '' )
	{
		my $o_logger = New Logger( '>>' . $h_args{ log_file }, &LOG_LEVEL );
		$self->SetLogger( $o_logger );
	}

	# on veut la description Mobyle
	my $mobyle_dir = $o_param->Get( &MOBYLE );
	if( $mobyle_dir ne '' )
	{
        # Appli is integrated to playmoby, this switch should not be used if playbomy is not installed
        eval("use Appli");
        if ($@ ne "")
        {    # Something wrong in eval
            die "ERROR - COULD NOT FIND Appli ! Maybe playmoby is not installed ?";
        }
		$self->_Set( __flag_mobyle => &TRUE );
		$self->__GenerateMobyleFile( $mobyle_dir, %h_args );
		return $self;
	}
	$self->_Set( __flag_mobyle => &FALSE,
	             __rh_current  => {} );
	# c'est surement pas top de ne pas intialiser les attributs dans _InitAttributes
	# mais je suis presse.

	# on initialise les attributs
	$self->_InitAttributes( $o_param );

	# set attributes for mandatories args
	$self->_FetchAndSetMandatories( $o_param );

	return $self;
} # _Init

# -------Help functions---------------------------------

=head2 function

 Title        : GenerateHelp
 Usage        : $sub = $o_xx->GenerateHelp( $rh_help, $prog_name );
                ParamParser::SetDefaultUsage( $sub );
 Prerequisite : none
 Function     : while Appli cannot do it itself.
                help will look like "$base\n$params\n\n$options\n"
                $params = $self->_GetHelpParams() . $params_str
                $options = $self->_GetHelpOptions() . $options_str
                $base = $self->_GetHelpBase( $prog_name ) . $base_str
 Returns      : $sub
 Args         : $rh_help, like { params => $params_str, options => $options_str, base => $base_str }
     $params_str, string, description of parameters
     $options_str, string, description of options,
     $base_str, string, general description of the program.

     $prog_name is the name of the program.

=cut

sub __GenerateHelp
{
	my $self   = shift;
	my ( $rh_help, $prog ) = @_;

	my $params  = $self->_GetHelpParams();
	$params    .= "\n" . $$rh_help{ params } if( defined $$rh_help{ params } );

	my $options = $self->_GetHelpOptions();
	$options   .= "\n" . $$rh_help{ options } if( defined $$rh_help{ options } );

	my $help_line = $self->_GetHelpBase( $prog );
	$help_line   .= "\n" . $$rh_help{ base } if( defined $$rh_help{ base } );

	return sub
	{
		print "$help_line\n$params\n\n$options\n";
	}
}


=head2 function _GetHelpBase

  Usage        : $string = $self->_GetHelpBase( $prog );
  Prerequisite : none
  Function     : Build a base help line including &HELP and &MOBYLE
  Returns      : $string
  Args         : $prog, program name

=cut

sub _GetHelpBase
{
	my $self = shift;
	my ( $prog ) = @_;

	my $string = "Usage :";
	$string   .= "\n\t$prog --" . &HELP . "\n\t\t=> print this message.\n";
	$string   .= "\n\t$prog --" . &MOBYLE . " <directory>\n\t\t=> save the mobyle file into <directory>.\n";
	$string   .= "\n\t$prog PARAMS [OPTIONS]\n";

	return $string;
}


=head2 function _GetHelpOptions

 Usage        : $string = $o_xx->_GetHelpOptions();
 Prerequisite : none
 Function     : none
 Returns      : $string, description of options &INSTANCES and &SPECIES
 Args         : none

=cut

sub _GetHelpOptions
{
	my $self      = shift;

	my $format = '%-20s';

	my $p_instance = sprintf( $format, &INSTANCES );
	my $p_log      = sprintf( $format, &LOG_FLAG );
	my $string = 'OPTIONS :
    --' . $p_instance . ' : string, instance name or coma separated list of instance name.
    --' . $p_log      . ' : flag, use it if you want a log file. (log level : ' . &LOG_LEVEL . ').';

	return $string;
}


=head2 function _GetHelpParams

 Usage        : $string = $o_xx->_GetHelpParams();
 Prerequisite : none
 Function     : build a standard string.
 Returns      : $string, description of parameters
 Args         : none

=cut

sub _GetHelpParams
{
	my $self   = shift;

	my $format = '%-20s';

	my $p_output    = sprintf( $format, &OUT_FILE );
	my $string = 'PARAMS :
    --' . $p_output . ' : string, name of the output file.';

	return $string;
}

# -------GetOptLong functions---------------------------------

=head2 function _GetOptLongList

 Usage        : $ra_options = $o_xx->_GetOptLongList( $ra_options );
 Prerequisite : none
 Function     : builf the option string list for ParamParser
 Returns      : $ra_options
 Args         : none

=cut

sub _GetOptLongList
{
	my $self = shift;
	my( $ra_options ) = @_;

	my @a_options = ( &INSTANCES. '=s',
	                  &MOBYLE   . '=s',
	                  &HELP,
	                  &LOG_FLAG,
	                  &OUT_FILE . '=s',
	                  @$ra_options
                    );

	return \@a_options;
}


# -------Mobyle related methods-------------------------

=head2 Procedure

 Title        : __GenerateMobyleFile
 Usage        : $o_xx->__GenerateMobyleFile( $mobyle_dir, %h_args );
 Prerequisite : none
 Procedure    : none
 Args         : $mobyle_dir, %h_args

=cut

sub __GenerateMobyleFile
{
	my $self = shift;
	my ( $mobyle_dir, %h_args ) = @_;

	my $o_appli = $self->_BuildAppli( $h_args{ rh_descr } );

	my $xml_page = $o_appli->GetMobyleXml( $mobyle_dir );
	if( defined $xml_page )
	{
		my $msg = "File can't be written into '$mobyle_dir'\n";
		warn $msg;
		print STDOUT $xml_page;
	}

	return;
}


=head2 Function _BuildAppli

 Usage        : my $o_appli = $o_x->_BuildAppli( $rh_descr )
 Function     : Build an Appli object given the description.
                This function add the descriptions of &INSTANCES and the &SPECIES.
 Returns      : $o_appli, Appli object.
 Args         :
     $rh_descr, hash ref, description need by Appli constructor.
 Error        : none

=cut

sub _BuildAppli
{
	my $self = shift;
	my( $rh_descr ) = @_;

	# besoin ajout de l'url dans description
	# et de doclink
	my $o_appli  = New Appli( %$rh_descr );

	# entree commune a une classe
	my @a_inputs = $self->_GetInputsDescription();
	foreach my $rh_input ( @a_inputs )
	{
		$o_appli->AddInput( %$rh_input );
	}

	# instance
	my @a_params = $self->_GetCommonParams();
	foreach my $rh_param ( @a_params )
	{
		$o_appli->AddParam( %$rh_param );
	}

	return $o_appli;
}


=head2 function

 Title        : _GetInputsDescription
 Usage        : @a_descr = $o_xx->_GetInputsDescription();
 Prerequisite : none
 Function     : return a list of hash ref containing mobyle description of inputs.
 Returns      : @a_descr
 Args         : $from

=cut

sub _GetInputsDescription
{
	my $self   = shift;

	my @a_descr = ();

	return @a_descr;
}


=head2 function

 Title      : _GetCommonParams
 Usage      : @a_list = $o_x->_GetCommonParams();
 Prerequiste: none
 Function   : return a list of hash ref of 2 parameters &INSTANCES and &SPECIES
              to all NarcisseService _instance and species.
 Args       : none.
 Return     : a hash containing the description of the instance parameter
              and the species param.
 Error      : none.

=cut

sub _GetCommonParams
{
	my $self         = shift;

	# pour obtenir la liste des instances dispo
	# j'ai besoin d'un objet NarcisseObserver
	my $o_nar_obs   = New NarcisseObserver( -o_logger => $self->_GetLogger() );
	my %h_instances = $o_nar_obs->GetInstancesDescr();

	# plus all
	$h_instances{ all } = 'toutes les instances.';

	my %h_instance = (
				name        =>	&INSTANCES,
				datatype    =>	'Choice',
				type_biomoby=>	'String',
				enum        =>	\%h_instances,
				default 	=>	'all',
				descr       =>	"name of the instance."
			);

	return ( \%h_instance );
}


# -------Action functions---------------------------------

=head2 Procedure

 Title        : _InitAttributes
 Usage        : $o_xx->_InitAttributes( $o_param );
 Prerequisite : none
 Procedure    : retrieve all InstanceObserver object regarding &INSTANCE
     argument of the program.
     Note that in web-service context, o_conf will be a NarcisseObserver,
     but in web context (keyword search, blast), o_conf will be a InstanceObserver.
 Args         : $o_param

=cut

sub _InitAttributes
{
	my $self = shift;
	my ( $o_param ) = @_;

	my $o_conf = &ConfigurationFactory::GetConfiguration();

	my $rao_inst_obs = [];
	if( $o_conf->isa( 'NarcisseObserver' ) )
	{
		# definir la liste des objets InstanceObserver
		my $str_instance = $o_param->Get( &INSTANCES );
		my $ra_names  = $self->_FetchArgs( $str_instance );
		$rao_inst_obs = $o_conf->FindInstanceObservers( ra_instances => $ra_names );
	}
	else
	{
		push( @$rao_inst_obs, $o_conf );
	}

	$self->_SetInstanceObserversListAttribute( $rao_inst_obs );

	return;
} # _InitAttributes


=head2 _SetInstanceObserversListAttribute

 Usage        : $o_x->_SetInstanceObserversListAttribute( $rao_inst_obs )
 Prerequiste  : do not override _setter
 Function     : $self->_Set( __rao_instance_observer => $rao_inst_obs );
 Args         :
     $rao_inst_obs, list ref of InstanceObserver object
 Error        : none.

 Note : c'est peut etre une liste d'objet Configuration

=cut

sub _SetInstanceObserversListAttribute
{
	my $self   = shift;
	my( $rao_inst_obs ) = @_;
	$self->_Set( __rao_instance_observer => $rao_inst_obs );

	return;
}


=head2 Function _GetCurrentData

 Usage        : my $rh_current = $o_x->_GetCurrentData()
 Function     : .
 Returns      : $rh_current
 Args         :
     none
 Error        : none

=cut

sub _GetCurrentData
{
	my $self = shift;

	my $rh_current = $self->_Get( '__rh_current' );

	return $rh_current;
}


=head2 _SetOutHashRef

 Usage        : $o_x->_SetOutHashRef( $rh_out )
 Prerequiste  : do not override _setter
 Function     : add $rh_out to hash ref __rh_out
 Args         :
     $rh_out, hash ref
 Error        : none.

 Note         : Set or Append ?

=cut

sub _SetOutHashRef
{
	my $self   = shift;
	my( $rh_out ) = @_;

	my $rh_out_ori = $self->_Get( '__rh_out' ) || {};

	my %h_out = ( %$rh_out_ori, %$rh_out );

	$self->_Set( __rh_out => \%h_out );

	return;
}


=head2 function

 Title        : _FetchArgs
 Usage        : $ra_list = $o_xx->_FetchArgs( $str, $filename );
 Prerequisite : none
 Function     : when args can be a coma separated list or when values can be
                into a file, this function read all those sources and return
                one ref list.
 Returns      : $ra_list, a ref list of all values read
 Args         : $str, string, a value or a coma separated list of values
                $filename, string, name of the file where to read data
                           if '-', will read STDIN.

=cut

sub _FetchArgs
{
	my $self = shift;
	my ( $str, $filename ) = @_;

	my $ra_list = &Args_utils::GetFromStringOrFile( str  => $str,
													file => $filename );

	return $ra_list;

# 	my @a_list = ();
# 
# 	if( &General::IsNotNull( $str ) )
# 	{
# 		push( @a_list, split( /,\s*/, $str ) );
# 	}
# 
# 	if( defined $filename )
# 	{
# 		# if the string is stdin
# 		if( $filename eq '-' )
# 		{
# 			__ReadStream( *STDIN, \@a_list );
# 		}
# 		# if the string is a file
# 		elsif( -e $filename )
# 		{
# 			my $fe = new IO::File( $filename, 'r' )
# 				or throw LipmError::IOException( $filename, 'f' );
# 			__ReadStream( $fe, \@a_list );
# 		}
# 	}
# 
# 	# mise en majuscule pour le service qui ne passe pas par l'API
# 	# => GetAnnotationFromLocus_tags
# # 	return map( uc, @a_words );
# 
# 	return \@a_list;
# 
# =head2 Procedure __ReadStream
# 
#  Title        : __ReadStream
#  Usage        : &__ReadStream( $flux, $ra_words );
#  Prerequisite : none
#  Procedure    : none
#  Args         : $flux, $ra_words
# 
# =cut
# 
# 	sub __ReadStream
# 	{
# 		my ( $flux, $ra_words ) = @_;
# 
# 		while( my $line = <$flux> )
# 		{
# 			next	if( $line =~ /^#/ );
# 			chomp( $line );
# 			my( $tag, @a_scrap ) = split( '#', $line );
# 			if( $tag )
# 			{
# 				$tag =~ s/^\s+|\s+$//g ;	#trim function
# 				push( @$ra_words, $tag );
# 			}
# 		}
# 
# 		return;
# 	}

} # _FetchArgs


=head2 _FetchAndSetMandatories

 Usage        : $o_x->_FetchAndSetMandatories( $o_param )
 Function     : Read mandatories attributes and throw Error if they are missing.
                Then set attributes with values.
 Args         :
     $o_param
 Error        : none.

=cut

sub _FetchAndSetMandatories
{
	my $self   = shift;
	my( $o_param ) = @_;

	my $outfile = $o_param->Get( &OUT_FILE );
	my $file_attr;

	# le service GetLocusTagsFromKeyword utilise dans l'interface doit ecrire sur STDOUT
	if( $outfile eq '' )
	{
# 		my %h_msg = ( -user_msg => "You need to specify an output file." );
# 		throw LipmError::ParameterException( &OUT_FILE, 'missing', %h_msg );
		use IO::Handle;
		my $o_handle = new IO::Handle();
		$file_attr = $o_handle->fdopen( fileno( STDOUT ), 'w' ); # il y a pe + simple
	}
	else
	{
		# file_attr is a filehandle here
		# but as the method is protected, it can be overloaded, it can be any scalar
		$file_attr = $self->_BuildFileAttribute( $outfile );
	}

	my $rh_out = { file => $file_attr,
	               ra_results  => [] };

	$self->_SetOutHashRef( $rh_out );

	return;
}


=head2 Function _BuildFileAttribute

 Usage        : my $scalar = $o_x->_BuildFileAttribute( $outfile )
 Function     : This file open the $outfile file and return its filehandle.
 Returns      : $scalar, the file handle on the open file.
 Args         :
     $outfile, string, the output filename.
 Error        : IOException

=cut

sub _BuildFileAttribute
{
	my $self = shift;
	my( $outfile ) = @_;
	
	my $fs_out = new IO::File( $outfile, 'w' )
	                or throw LipmError::IOException( $outfile, 'f' );

	return $fs_out;
}


# should not be overloaded
sub _GetFileAttribute
{
	my $self = shift;

	my $rh_out = $self->_Get( '__rh_out' );
	
	return $$rh_out{ file };
}


=head2 Procedure

 Title        : Launch
 Usage        : $o_xx->Launch();
 Prerequisite : none
 Procedure    : none
 Args         : none

=cut

sub Launch
{
	my $self = shift;

	# si on a genere le fichier mobyle
	return  if( &TRUE == $self->_Get( '__flag_mobyle' ) );

	my $rh_current   = $self->_GetCurrentData();
	my $rao_inst_obs = $self->_Get( '__rao_instance_observer' );
	foreach my $o_inst_obs ( @$rao_inst_obs )
	{
		$$rh_current{ o_inst_obs } = $o_inst_obs;

		my $msg = 'working on ' . $o_inst_obs->GetName() . "\n";
		$self->_Trace( $msg, '.', 'IS' );

		my $stop_flag;
		try
		{
			$stop_flag = $self->_LaunchServiceOnInstance( $o_inst_obs );
		}
		catch LipmError::ObjectException with
		{
			my( $o_err, $r ) = @_;

			# a voir si c'est une bonne facon de faire
			if( $o_err->value eq &SpeciesObserver::ERR_SPECIES_NOT_IN_INSTANCE )
			{
				# on logue et on passe a l'instance suivante
				$self->_Trace( $o_err->Explain(), '*', 'IS' );
				$r = 1;
			}
		}
		catch LipmError with
		{
			my( $o_err, $r ) = @_;

			if( !$o_err->IsFatal() )
			{
				$self->_Trace( $o_err->text(), '.', 'IS' );
				$r = 1;
			}
			else
			{
				$o_err->throw();
			}
		}; # ne pas enlever ;

		last if( defined $stop_flag && $stop_flag==1 );
		# apres conversion de toute les procedures _LaunchServiceOnSpecies en fonction
		# virer le defined
	}

	$self->_PrintOutResults();

	my $msg = "End of MultiService execution.\n";
	$self->_Trace( $msg, '.', 'IS' );

	return;
}


=head2 Procedure

 Title        : _LaunchServiceOnInstance
 Usage        : $o_xx->_LaunchServiceOnInstance( $ra_species );
 Prerequisite : none
 Procedure    : none
 Args         : $ra_species

=cut

sub _LaunchServiceOnInstance
{
	my $self = shift;
	my ( $o_inst_obs ) = @_;

	my %h_params = ( o_instance_obs => $o_inst_obs );

	my $o_ns = $self->_BuildNarcisseService( %h_params );

	my ( $ra_results, $rh_results ) = $o_ns->FetchResults();
# warn Dumper { $o_inst_obs->GetName() => [ $ra_results, $rh_results ] };
	$self->_StoreResults( { hash => $rh_results, list => $ra_results } );

	my $nb = @$ra_results;
	my $msg   = "found $nb results\n"; # ! faux si data => []
	$self->_Trace( $msg, '.', 'ISV' );

	return;
}


=head2 _BuildNarcisseService

 Usage        : my $o_ns = $o_x->_BuildNarcisseService( %h_args ).
 Prerequiste  : Must be overloaded.
 Function     : This function must build and return a NarcisseService object.
 Returns      : $o_ns, NarcisseService object
 Args         : %h_args,
     o_instance_obs => $o_inst_obs, ? getcurrent ?
 Error        : LipmError::ObjectException.

 Note : Je pense que cette methode correspond au patron 'monteur'.
        Il aurait donc peut-etre fallut appeler cette methode NarcisseServiceBuilder.

=cut

sub _BuildNarcisseService
{
	my $self   = shift;
	my %h_args = @_;

	my $dbg = "MultiService need to be derived to be used.\n";
	my $msg = "Somebody don't do his job well, tell him to check error file.\n";

	my %h_params = ( -text => $dbg, -user_msg => $msg );
	throw LipmError::ObjectException( $self, 'abstract method', %h_params );

	my $o_ns = undef;

	return $o_ns;
}


=head2 Procedure

 Title        : _StoreResults
 Usage        : $o_xx->_StoreResults( $data );
 Prerequisite : none
 Procedure    : none
 Args         : $ra_data

=cut

sub _StoreResults
{
	my $self     = shift;
	my ( $data ) = @_;

	my $rh_out = $self->_Get( '__rh_out' );

	my $ra_ = $$rh_out{ ra_results };
	push( @$ra_, $data );

	return;
}


=head2 Procedure _PrintOutResults

 Usage        : $o_xx->_PrintOutResults( none );
 Prerequisite : none
 Procedure    : call _PrintArray or _PrintHash on every results
 Args         : none

=cut

sub _PrintOutResults
{
	my $self    = shift;

	my $rh_out = $self->_Get( qw( __rh_out ) );

	# un enregistrement par instance
	my $ra_  = $$rh_out{ ra_results };
	foreach my $rh_data ( @$ra_ )
	{
		my ( $rh_results, $ra_results ) = @$rh_data{ qw( hash list ) };
# warn Dumper [$rh_results, $ra_results];
		# TODO il me semble qu'avec data, on ne passe plus jamais par ici, a verifier
		if( 0 == scalar( keys %$rh_results ) )
		{
			$self->_PrintArray( $ra_results );
		}
		else
		{
			$self->_PrintHash( $rh_results, $ra_results );
		}
	}

	return;
}


=head2 _PrintArray

 Usage        : $o_x->_PrintArray( $ra_data )
 Function     : .
 Args         :
     $fh_out, $ra_data
 Error        : none.

=cut

sub _PrintArray
{
	my $self   = shift;
	my( $ra_data ) = @_;

	my $fh_out = $self->_GetFileAttribute();

	foreach my $data ( @$ra_data )
	{
		print $fh_out $data . "\n";
	}
	

	return;
}


=head2 _PrintHash

 Usage        : $o_x->_PrintHash( $rh_results, $ra_results )
 Function     : .
 Args         :
     $fh_out, $rh_results
     $ra_results
 Error        : none.

=cut

sub _PrintHash
{
	my $self   = shift;
	my( $rh_results, $ra_results ) = @_;

	my $fh_out = $self->_GetFileAttribute();

# 	$ra_result contient les clef triees de $rh_results
#	A voir s'il faut conserver ceci !

	foreach my $header ( @$ra_results )
	{
		my $data = $$rh_results{ $header };
		if( 'ARRAY' eq ref( $data ) )
		{
			print $fh_out join( "\n", @$data ), "\n";
		}
		else
		{
			my $fasta_seq = Sequence_utils::FormatSequenceLine( seq => $data );
# warn Dumper $$rh_results{ $header };

			print $fh_out $header . "\n", $fasta_seq . "\n";
		}
	}

	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;
