#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 10 March, 2009
#	$Id: Accessor.pm 3540 2010-11-05 17:35:18Z sletort $
#

package MultiService::Accessor;

=pod

=head1 NAME

 MultiService::Accessor - 

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 What's different between MS::Accessor and MS ?
 -> instance observer is optional.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( MultiService::ThroughSpecies );

#use General;
use LipmError::ParameterException;

use NarcisseService::Accessor;
use NarcisseService::Accessor::WhatDecorator::GenesDecorator;
use NarcisseService::Accessor::WhatDecorator::CDS_TransDecorator;
use NarcisseService::Accessor::WhatDecorator::ProteinsDecorator;
use NarcisseService::Accessor::WhatDecorator::SlicesDecorator;
use NarcisseService::Accessor::WhatDecorator::SlicesDecorator::UpstreamsDecorator;
use NarcisseService::Accessor::WhatDecorator::SlicesDecorator::DownstreamsDecorator;
use NarcisseService::Accessor::FromDecorator::RegionsDecorator;
use NarcisseService::Accessor::FromDecorator::IdsDecorator;

use NarcisseServiceKeywords qw( &NS_GENES &NS_PROTEINS &NS_SLICES NS_TRANSCRIPTS 
                                &NS_UPSTREAM &NS_DOWNSTREAM &NS_CDS );

# options relatives a la sortie
use constant OUT_TYPE   => 'output_type'; # valeur possible : locus_tag fasta infos
use constant OUT_NOT_FOUND => 'not_found_file';

# options relatives aux entrees
use constant REGIONS    => 'regions';
use constant TAGS       => 'locus_tags';
use constant CHROM      => 'chromosomes';
use constant LENGTH     => 'length';
use constant STREAM     => 'stream'; # valeur attendues up/down

use constant REGIONS_FILE => &REGIONS . '_file';
use constant TAGS_FILE    => &TAGS    . '_file';
use constant CHROM_FILE   => &CHROM   . '_file';

use constant SAFE       => 'stop_on_found';


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New MultiService( rh_descr => $rh_descr,
                                            rh_help  => $rh_help,
                                            log_file => $log_filename,
                                            ra_options => $ra_options,
                                            what  => $what,
                                            from  => $from );
 Prerequisite : none
 Fonction     : initialise an instance of MultiService class.
 Returns      : a valid MultiService object
 Args         :
     $rh_descr, hash ref, description for Appli constructor.
     $rh_help, hash ref, contains strings to add to help message.
         base    => string, the base help, should not be used.
         params  => string, parameters to provide.
         options => string, options of the script.
     $log_filename, string, the name of the file where we will log via Logger.
     $ra_options, list ref of strings, elements are GetOptLong strings.
    $from, string, 'Regions', 'LocusTags'
            indicates the decoration to use as a FromDecoration
    $what, string, &NS_GENES, ...
            indicates the decoration to use as a WhatDecoration
 Attributes   :
    __ra_chrom,   ref list of chromosome label
    __ra_regions, ref list of hashes regions { start => x, end => y }
    __ra_tags, ref list of locus tags
    __from, string, 'Regions', 'LocusTags'
            indicates the decoration to use as a FromDecoration
    __what, string, &NS_GENES, ...
            indicates the decoration to use as a WhatDecoration
    __rh_current, hash ref, values change while processing
#?        o_inst_obs, the instance observer object
    __rh_out, hash ref,
        ra_results, list ref of results, one result per species
             a result is a hash with 2 keys
                 hash => with NarcisseService::__rh_results value
                 list => with NarcisseService::__ra_results value
        o_not_found, Enum object of entries not found
        type, string cf OUT_TYPE constant
        rh_filename, hash ref
            results, string, name of the output file for results.
            not_found, string, name of the output file for not_found results.
    __safe_search, boolean, true to stop at the first species with results.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( what  => undef,
	               from  => undef,
	               @_ );

	# initialisation de __what et __from ici car on en a besoin pour generer l'aide
	$self->_Set( __what => $h_args{ what },
	             __from => $h_args{ from },
	             __safe_search => 0 );

	$self->SUPER::_Init( %h_args );

	return;
} # _Init


# -------Help functions---------------------------------

=head2 function _GetHelpOptions

 Usage        : $string = $o_xx->_GetHelpOptions();
 Prerequisite : none
 Function     : none
 Returns      : $string, description of options &INSTANCES and &SPECIES
 Args         : none

=cut

sub _GetHelpOptions
{
	my $self      = shift;

	my $string = $self->SUPER::_GetHelpOptions();

	my $format = '%-20s';
	my $what   = $self->_Get( '__what' );
	if( ( $what eq &NS_GENES )	# en test
	   || ( $what eq &NS_CDS )
	   || ( $what eq &NS_PROTEINS )
	   || ( $what eq &NS_TRANSCRIPTS ) )
	{
		my $p_type = sprintf( $format, &OUT_TYPE );
		$string .= '
    --' . $p_type . ' : string, output type
            <fasta> fetch multi-fasta sequence, *Default*
            <locus> fetch only locus tags,
            <infos> fetch the locus tags and other infos.';
	}
	elsif( ( $what eq &NS_UPSTREAM )
	     || ( $what eq &NS_DOWNSTREAM ) )
	{
		my $p_length  = sprintf( $format, &LENGTH );
		my $def_value = &WhatDecorator::SlicesDecorator::DEF_LENGTH;
		$string .= '
    --' . $p_length . " : [$def_value] integer, number of bases to extract.";
	}

	my $from = $self->_Get( qw( __from ) );
	if( $from eq 'LocusTags' )
	{
		my $p_safe = sprintf( $format, &SAFE );
		$string   .= '
    --' . $p_safe . " : [0] boolean, 1 to stop at the first species with results.";
	}

	return $string;
}


=head2 function _GetHelpParams

 Usage        : $string = $o_xx->_GetHelpParams();
 Prerequisite : none
 Function     : build a standard string.
 Returns      : $string, description of parameters
 Args         : none

=cut

sub _GetHelpParams
{
	my $self   = shift;

	my $string = $self->SUPER::_GetHelpParams();

	my $format = '%-20s';
	my $p_not_found = sprintf( $format, &OUT_NOT_FOUND );
	$string .= '
    --' . $p_not_found . ' : string, name of the file where not found entries will be written.';

	my $from = $self->_Get( '__from' );
	if( $from eq 'Regions' )
	{
		my $p_chrom  = sprintf( $format, &CHROM . '|' . &CHROM_FILE );
		my $p_region = sprintf( $format, &REGIONS . '|' . &REGIONS_FILE );
		$string .= '
    --' . $p_chrom . ' : string, chromosome name or name of the file containing one chromosome name per line.
    --' . $p_region . ' : start..end, region start/end. (ex: 256, 2kb, 2Mb) or name of the file containing one region per line.';
    }

	if( $from eq 'LocusTags' )
	{
		my $p_tags   = sprintf( $format, &TAGS . '|' . &TAGS_FILE );
		$string .= '
    --' . $p_tags . ' : string, locus tags or name of the file containing one locus tags per line.';
    }

	return $string;
}


# -------GetOptLong functions---------------------------------

=head2 function _GetOptLongList

 Usage        : $ra_options = $o_xx->_GetOptLongList( $ra_options );
 Prerequisite : none
 Function     : builf the option string list for ParamParser
 Returns      : $ra_options
 Args         : none

=cut

sub _GetOptLongList
{
	my $self = shift;
	my( $ra_new_options ) = @_;

	my $ra_options = $self->SUPER::_GetOptLongList( $ra_new_options );

	my @a_opt_out = ( &OUT_TYPE      . '=s',
	                  &OUT_NOT_FOUND . '=s'
	                );
	push( @$ra_options, @a_opt_out );

	my( $from, $what ) = $self->_Get( qw( __from __what ) );

	if( $from eq 'Regions' )
	{
		my @a_opt_regions = ( &REGIONS . '=s',
		                      &CHROM   . '=s',
		                      &REGIONS_FILE . '=s',
		                      &CHROM_FILE   . '=s'
		                    );
		push( @$ra_options, @a_opt_regions );
	}
	elsif( $from eq 'LocusTags' )
	{
		my @a_opt_regions = ( &TAGS . '=s',
		                      &TAGS_FILE . '=s',
		                      &SAFE
		                    );
		push( @$ra_options, @a_opt_regions );
	}

	if( ( $what eq &NS_DOWNSTREAM )
	 || ( $what eq &NS_UPSTREAM ) )
	{
		my @a_opt_regions = ( &LENGTH . '=i' );
		push( @$ra_options, @a_opt_regions );
	}

	return $ra_options;
}


# -------Mobyle functions---------------------------------

=head2 function _GetInputsDescription

 Usage        : @a_descr = $o_xx->_GetInputsDescription();
 Prerequisite : none
 Function     : return a list of hash ref containing mobyle description of inputs.
 Returns      : @a_descr
 Args         : $none

=cut

sub _GetInputsDescription
{
	my $self   = shift;
	my @a_descr = $self->SUPER::_GetInputsDescription();

	my $from = $self->_Get( '__from' );

	if( $from eq 'Regions' )
	{
		push( @a_descr, __GetChromDescr() );
		push( @a_descr, __GetRegionsDescr() );
	}
	elsif( $from eq 'LocusTags' )
	{
		push( @a_descr, __GetLocusTagsDescr() );
	}

	return @a_descr;
}


=head2 function

 Title      : _GetCommonParams
 Usage      : @a_list = $o_x->_GetCommonParams();
 Prerequiste: this is used only by NS_DOWNSTREAM and NS_UPSTREAM,
              other only call mother.
 Function   : call mother and if __what is NS_DOWNSTREAM and NS_UPSTREAM,
              describe LENGTH parameter.
 Args       : none.
 Return     : @a_list, a list of hash ref of 3 parameters &INSTANCES, &SPECIES,
              and LENGTH.
 Error      : none.

=cut

sub _GetCommonParams
{
	my $self         = shift;

	# species and instances
	my @a_params = $self->SUPER::_GetCommonParams();

	my $what = $self->_Get( qw( __what ) );

	if( ( $what eq &NS_UPSTREAM )
		|| ( $what eq &NS_DOWNSTREAM ) )
	{
		my $def_value = &WhatDecorator::SlicesDecorator::DEF_LENGTH;
		my %h_length = (
					name        =>	&LENGTH,
					datatype    =>	'int',
					type_biomoby=>	'Integer',
					descr       =>	"[$def_value] integer, number of bases to extract."
				);

		push( @a_params, \%h_length );
	}

	return @a_params;
}


=head2 function

  Title        : __GetChromDescr
  Usage        : $rh_descr = &__GetChromDescr();
  Prerequisite : none
  Function     : return the chromosome parameter description.
  Returns      : $rh_descr
  Args         : none
  Globals      : none

=cut

sub __GetChromDescr
{
	my %h_descr = ( name         => 'ChromosomeName',
	                descr        => "the chromosome's name.",
	                namespace    => 'Narcisse',
	                type_biomoby => 'List_Text',
# 	                datatype     => 'string',
	                cmd          => '--' . &CHROM_FILE . '=$value'
				);
	return \%h_descr;
}


=head2 function __GetRegionsDescr

  Usage        : $rh_descr = &__GetRegionsDescr();
  Prerequisite : none
  Function     : return the region parameter description.
  Returns      : $rh_descr
  Args         : none
  Globals      : none

=cut

sub __GetRegionsDescr
{
	my %h_descr = ( name         => 'RegionsFile',
	                descr        => "the region limits.",
	                namespace    => 'Narcisse',
	                type_biomoby => 'text-formatted',
# 	                datatype     => 'string',
	                cmd          => '--' . &REGIONS_FILE . '=$value'
	              );
	return \%h_descr;
}


=head2 function __GetLocusTagsDescr

  Title        : __GetLocusTagsDescr
  Usage        : $rh_descr = &__GetLocusTagsDescr();
  Prerequisite : none
  Function     : return the locus_tags_file parameter description.
  Returns      : $rh_descr, cf Appli.pm
  Args         : none

=cut

sub __GetLocusTagsDescr
{
	my %h_descr = ( name         => 'LocusTagsFile',
	                descr        => "The locus tags to search.",
	                namespace    => 'Narcisse',
	                type_biomoby => 'List_Text',
# 	                datatype     => 'string',
	                cmd          => '--' . &TAGS_FILE . '=$value'
	              );
	return \%h_descr;
}


# -------Action functions---------------------------------

=head2 Procedure

 Title        : _InitAttributes
 Usage        : $o_xx->_InitAttributes( $o_param );
 Prerequisite : none
 Procedure    : none
 Args         : $o_param

=cut

sub _InitAttributes
{
	my $self = shift;
	my ( $o_param ) = @_;

	$self->SUPER::_InitAttributes( $o_param );

	my $out_type = $o_param->Get( &OUT_TYPE );
	$self->_Set( #__o_not_found => undef,
	             __out_type => $out_type );

	my $what = $self->_Get( qw( __what ) );

	my $ra_length = $self->_FetchArgs( $o_param->Get( &LENGTH ) );
	if( ( $what eq &NS_UPSTREAM )
	 || ( $what eq &NS_DOWNSTREAM ) )
	{
		$self->_Set( __length => $$ra_length[0] );
	}

	$self->_Set( __safe_search => 1 )  if( $o_param->IsDefined( &SAFE ) );

	return;
} # _InitAttributes


=head2 Procedure

 Title        : __SetRegionsAttributes
 Usage        : $o_xx->__SetRegionsAttributes( $ra_regions );
 Prerequisite : none
 Procedure    : none
 Args         : $ra_regions, a list of string in the format start..end

=cut

sub __SetRegionsAttributes
{
	my $self = shift;
	my ( $ra_regions ) = @_;

	my @a_list = ();

	foreach my $region ( @$ra_regions )
	{
# 		$region =~ /^(\d+)(-|(\.\.))(\d+)$/;
		my ( $start, $end ) = split( /\.\.|-/, $region );
		$start = &Sequence_utils::Unit2Base( $start );
		$end   = &Sequence_utils::Unit2Base( $end );

		if( !defined( $start ) or !defined( $end )
		or ( $start !~ /^\d+$/ ) or ( $end !~ /^\d+$/ )
		)
		{
			no warnings; # pour pallier les cas undef
			my $msg = "the region <$start>..<$end> is invalid.";
			use warnings;
			$self->_Trace( $msg, '.', 'IS' );
		}
		else
		{
			push( @a_list, { start => $start, end => $end } );
		}
	}

	$self->_Set( __ra_regions => \@a_list );

	return;
}


=head2 _FetchAndSetMandatories

 Usage        : $o_x->_FetchAndSetMandatories( $o_param )
 Function     : Read mandatories attributes and throw Error if they are missing.
                Then set attributes with values.
 Args         :
     $o_param
 Error        : none.

=cut

sub _FetchAndSetMandatories
{
	my $self   = shift;
	my( $o_param ) = @_;

	$self->SUPER::_FetchAndSetMandatories( $o_param );

	my $not_found = $o_param->Get( &OUT_NOT_FOUND );
	if( $not_found eq '' )
	{
		my $msg = "You need to specify an output file for not_found entries.";
		my %h_msg = ( -user_msg => $msg );
		throw LipmError::ParameterException( &OUT_NOT_FOUND,'missing', %h_msg );
	}
	my $fs_out = new IO::File( $not_found, 'w' )
	                or throw LipmError::IOException( $not_found, 'f' );
	my $rh_out = { filehandle => { not_found => $fs_out } };
	$self->_Set( __rh_out => $rh_out );

	# verifie la coherence des param.
	my $from = $self->_Get( qw( __from ) );

	if( $from eq 'Regions' )
	{
		# regions args need to be reformatted into hashes
		my $ra_regions = $self->_FetchArgs( $o_param->Get( &REGIONS, &REGIONS_FILE ) );
		$self->__SetRegionsAttributes( $ra_regions );

		my $ra_chrom  = $self->_FetchArgs( $o_param->Get( &CHROM, &CHROM_FILE ) );

		if( 1 != scalar( @$ra_chrom ) )
		{
			my $msg  = "You have to set ONE chromosome.";
			my $join = join (',', @$ra_chrom );
			my $dbg  = "\noption chromosome was set with <$join>\n";
			throw LipmError( -text => $msg . $dbg, -user_msg => $msg );
		}
		$self->_Set( __ra_chrom => $ra_chrom );
	}
	else #locus
	{
		my $ra_tags  = $self->_FetchArgs( $o_param->Get( &TAGS, &TAGS_FILE ) );

		if( 0 == @$ra_tags )
		{
			my $msg = "You have to specify a list of locus tags --" . &TAGS;
			$msg   .= " or --" . &TAGS_FILE . "\n";
			my $dbg = "No tags were found.\n$msg";
			throw LipmError( -text => $dbg, -user_msg => $msg );
		}
		$self->_Set( __ra_tags => $ra_tags );
	}

	return;
}


=head2 function _LaunchServiceOnSpecies

 Usage        : my $bool = $o_xx->_LaunchServiceOnSpecies( none );
 Prerequisite : none
 function     : Build the service, gets the results, store them.
 Args         : none
 return       : $bool, boolean, true if we want to stop searching.

=cut

sub _LaunchServiceOnSpecies
{
	my $self = shift;

	# si trop lent, (j'en doute), sortir cet appel de la boucle espece et passe IO et species en parametres.
# 	my $rh_current = $self->_GetCurrentData();

	my %h_params = ( ); # ?
	my $o_ns = $self->_BuildNarcisseService( %h_params );

	my ( $ra_results, $rh_results ) = $o_ns->FetchResults();
	$self->_StoreResults( { hash => $rh_results, list => $ra_results } );

	my $nb = @$ra_results;
	my $msg   = "found $nb results\n"; # ! faut si data => []
	$self->_Trace( $msg, '.', 'ISV' );

	my $ra_not_found = $o_ns->FetchNotFound();

	# on ne stocke pas les regions sans trucs, sauf s'il y a de l'info
	my $from = $self->_Get( '__from' );
	return  if( ( $from eq 'Regions' ) && ( ref( $$ra_not_found[0] ) ne 'HASH' ) );

	$self->__StoreNotFound( $ra_not_found );

	return 1  if( ( 0 < $nb ) && ( 1 == $self->__PerformSafeSearch() ) );

	return 0;
}


=head2 Function __PerformSafeSearch

 Usage        : my $bool = $o_x->__PerformSafeSearch()
 Function     : accessor on the boolean attribute __safe_search.
 Returns      : $bool, true for a safe search (stop at first results ).
 Args         : none
 Error        : none

=cut

sub __PerformSafeSearch
{
	my $self = shift;

	return $self->_Get( '__safe_search' );
} #__PerformSafeSearch



=head2 _BuildNarcisseService

 Usage        : my $o_ns = $o_x->_BuildNarcisseService( %h_args ).
 Prerequiste  : Must be overloaded.
 Function     : This function must build and return a NarcisseService object.
 Returns      : $o_ns, NarcisseService object
 Args         : %h_params,
     -o_instance_obs => $o_inst_obs,
     -o_logger       => $o_logger
 Error        : LipmError::ObjectException.

=cut

sub _BuildNarcisseService
{
	my $self   = shift;
# 	my %h_args = ( o_instance_obs => undef,
# 	               @_ );

	# toutes les clef doivent être renseignées
# 	LipmError::ParameterException::Assert( %h_args );

	my $out_type = $self->_Get( '__out_type' );
	my $rh_current = $self->_GetCurrentData();

	my %h_params = ( species    => $$rh_current{ species },
					 -out_type  => $out_type,
					 -o_logger  => $self->_GetLogger(),
					 -o_instance_obs => $$rh_current{ o_inst_obs } );
	my $o_ns = New NarcisseService::Accessor( %h_params );
	$o_ns    = $self->AddFromDecoration( $o_ns );
	$o_ns    = $self->AddWhatDecoration( $o_ns );

	return $o_ns;
}


=head2 function AddFromDecoration

 Title        : AddFromDecoration
 Usage        : $o_ns_out = $o_xx->AddFromDecoration( $o_ns_in );
 Prerequisite : none
 Function     : none
 Returns      : $o_ns_out
 Args         : $o_ns_in

=cut

sub AddFromDecoration
{
	my $self        = shift;
	my ( $o_ns_in ) = @_;

	my $class    = '';
	my %h_args   = ( o_accessor => $o_ns_in, -o_logger => $self->_GetLogger() );

	my @a_attr = qw( __from __ra_chrom __ra_regions __ra_tags __what );
	my( $from, $ra_chrom, $ra_regions, $ra_tags, $what ) = $self->_Get( @a_attr );

	if( $from eq 'Regions' )
	{
		$class = 'RegionsDecorator';
		my $chrom = $$ra_chrom[0];
		if( 0 == @$ra_regions )
		{
			my $chr_length = $self->__GetChromSize( $chrom );
			$ra_regions = [ { start => 1, end => $chr_length } ];
		}

		$h_args{ chrom }      = $chrom;
		$h_args{ ra_regions } = $ra_regions;
	}
	elsif( $from eq 'LocusTags' )
	{
		$class = 'LocusTagsDecorator';

		$h_args{ ra_tags } = $ra_tags;
		$h_args{ feature } = $what;
	}
	throw LipmError( -text => "erreur : $from" )	if( $class eq '' );

	$class = 'FromDecorator::' . $class;
	my $o_ns_out = New $class( %h_args );

	return $o_ns_out;
}


=head2 function AddWhatDecoration

 Title        : AddWhatDecoration
 Usage        : $o_ns_out = $o_xx->AddWhatDecoration( $o_ns_in );
 Prerequisite : none
 Function     : Add a decoration to the $o_ns_in object
                regarding the __what attribute.
 Returns      : $o_ns_out
 Args         : $o_ns_in

=cut

sub AddWhatDecoration
{
	my $self        = shift;
	my ( $o_ns_in ) = @_;

	my $class    = '';
	my %h_args   = ( o_accessor => $o_ns_in, -o_logger => $self->_GetLogger()  );

	my( $what, $length ) = $self->_Get( qw( __what __length ) );

	if( $what eq &NS_GENES )
	{	$class = 'GenesDecorator';	}
	elsif( $what eq &NS_TRANSCRIPTS ) #! le mot clef &PROTEIN est actuellement le meme aue TRANSCRIPT
	{	$class = 'TranscriptsDecorator';	}
	elsif( $what eq &NS_CDS )
	{	$class = 'CDSsDecorator';	}
	elsif( $what eq &NS_PROTEINS )
	{	$class = 'ProteinsDecorator';	}
	elsif( $what eq &NS_SLICES )
	{	$class = 'SlicesDecorator';	}
	elsif( $what eq &NS_UPSTREAM )
	{
		$class = 'SlicesDecorator::UpstreamsDecorator';
		$h_args{ -length } = $length if( defined $length );
	}
	elsif( $what eq &NS_DOWNSTREAM )
	{
		$class = 'SlicesDecorator::DownstreamsDecorator';
		$h_args{ -length } = $length if( defined $length );
	}
	throw LipmError( -txt => "erreur : $what")	if( $class eq '' );

	$class = 'WhatDecorator::' . $class;
	my $o_ns_out = New $class( %h_args );

	return $o_ns_out;
}


=head2 function

 Title        : __GetChromSize
 Usage        : $chr_length = $o_xx->__GetChromSize( $chrom );
 Prerequisite : none
 Function     : Get the size of $chrom of current species of current instance
 Returns      : $chr_length
 Args         : $chrom

=cut

sub __GetChromSize
{
	my $self     = shift;
	my( $chrom ) = @_;

	my $rh_current = $self->_GetCurrentData();
	my $o_inst_obs = $$rh_current{ o_inst_obs };
	my $species    = $$rh_current{ species };
	my $o_species  = $o_inst_obs->GetSpeciesObserver( $species );
	my $o_dba      = $o_species->GetDBAdaptor();
	my $rh_karyo   = $o_dba->get_SliceAdaptor()->karyotype();

	my $chr_length = $$rh_karyo{ $chrom }{ length }
	                 or die( "chrom doesn't exist __TODO__" );

	return $chr_length;
}


=head2 Procedure __StoreNotFound

 Usage        : $o_xx->__StoreNotFound( $ra_data );
 Prerequisite : none
 Procedure    : .
 Args         : $ra_data

=cut

sub __StoreNotFound
{
	my $self = shift;
	my ( $ra_data ) = @_;

	my $rh_out = $self->_Get( '__rh_out' );

	# $o_not_found est un objet Enum::NotFound, classe incluse ici.
	my $o_not_found = $$rh_out{ o_not_found };
	my $o_enum;
	if( defined $o_not_found )
	{
		$o_enum = $o_not_found->Intersec( @$ra_data );
	}
	else
	{	$o_enum = New Enum::NotFound( @$ra_data );	}

	$$rh_out{ o_not_found } = $o_enum;

	return;
}


=head2 Procedure

 Title        : __PrintOutNotFound
 Usage        : $o_xx->__PrintOutNotFound();
 Prerequisite : none
 Procedure    : none
 Args         : none

=cut

sub __PrintOutNotFound
{
	my $self     = shift;

	my $rh_out = $self->_Get( '__rh_out' );
	my $o_fs = $$rh_out{ filehandle }{ not_found };

	my $o_enum = $$rh_out{ o_not_found };
	my @a_list = defined( $o_enum ) ? $o_enum->ListItems() : ();

	print $o_fs join( "\n", @a_list );

	return;
}


=head2 Procedure _PrintOutResults

 Usage        : $o_xx->_PrintOutResults( none );
 Prerequisite : none
 Procedure    : call _PrintArray or _PrintHash on every results
 Args         : none

=cut

sub _PrintOutResults
{
	my $self    = shift;

	#found
	$self->SUPER::_PrintOutResults();

	#not_found
	$self->__PrintOutNotFound();

	return;
}

1;

# Classe ad hoc pour repondre a nos besoins sans revoir tout le stockage des not found
# En fait remplace quasi completement Enum
# l'unique raison de sa presence est de permettre d'ecrire qqc a la suite
# du locus tag dans le fichier de sortie de not_found
# fait pour le cas des transcripts incoherents
package Enum::NotFound;

use Data::Dumper;
use base qw( Enum );

=head2 function

 Title        : New
 Usage        : $o_enum = $o_xx->New( @a_list );
 Prerequisite : Enum does not inherite from LipmObject
 Function     : none
 Returns      : $o_enum
 Args         : @a_list

=cut

sub New
{
	my $class = shift;
	my $self = {};
	bless( $self, $class );

	my( @a_list ) = @_;

	$self->{ __hash } = {};
	my @a_values = ();

	foreach my $item ( @a_list )
	{
		my $val = $item;
		if( ref( $item ) eq 'HASH' )
		{
			my( $key, $value ) = each( %$item );
			$val   = $key;
			$self->{ __hash }{ $key } = $value;
		}

		push( @a_values, $val );
	}

	$self->AddValues( @a_values );

	return $self;
}


=head2 function

 Title        : ListItems
 Usage        : @a_list = $o_xx->ListItems( none );
 Prerequisite : none
 Function     : none
 Returns      : @a_list
 Args         : none

=cut

sub ListItems
{
	my $self = shift;

	my @a_list = ();

	my @a_items = $self->SUPER::ListItems();
	foreach my $item ( @a_items )
	{
		next  if( $item eq '__hash' );
		my $val = $item;

		if( defined $self->{ __hash }{ $item } )
		{
			$val = $item . "\t" . $self->{ __hash }{ $item };
		}

		push( @a_list, $val );
	}

	return @a_list;
}


=head2 function

 Title        : Intersec
 Usage        : $o_enum  = $o_xx->Intersec( @a_list );
 Prerequisite : none
 Function     : lists items present both in
                   $o_xx and @a_list
 Returns      : $o_enum, Enum::NotFound object
 Args         : @a_list, list, can contain hash ref

=cut

sub Intersec
{
	my $self   = shift;
	my @a_list = @_;

	my $o_enum = New Enum::NotFound( @a_list );

	my @a_both = ();

	foreach my $data ( keys %$o_enum )
	{
		next  if( $data eq '__hash' );

		if( $self->Includes( $data ) )
		{
			#/!\ on ecrase la precedente
			# ==> ultra rare, les locus sont deja presque uniques
			if( defined( $o_enum->{ __hash }{ $data } ) )
			{
				my $value = $o_enum->{ __hash }{ $data };
				push( @a_both, { $data => $value } );
			}
			elsif( defined( $self->{ __hash }{ $data } ) )
			{
				my $value = $o_enum->{ __hash }{ $data };
				push( @a_both, { $data => $value } );
			}
			else
			{
				push( @a_both, $data );
			}
		}
	}

	return New Enum::NotFound( @a_both );
}


1;

__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
