#
# $Id: NARCISSE_inlineC.pm 2474 2009-04-14 08:28:30Z tfaraut $
#

package NARCISSE_inlineC;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

NARCISSE_inlineC - C implementation, using Inline::C, of a few routines to access data files
                 

=head1  SYNOPSIS

 use NARCISSE_inlineC;

 my $ra_data = &NARCISSE_inlineC::ReadFile( $file );
 ...
 my %h_args = ( 
			   file			=> $file,	
			   start 		=> $start,
               end   		=> $end,
               start_label 	=> 'start',
               end_label  	=> 'end',
               mode			=> 0,
               keyname 		=> 'element',
               VALUE 		=> { 's_chrom' => 4}
              );

 my $rh_data = &NARCISSE_inlineC::ReadAndFormatData( \%h_args );

=head1 DESCRIPTION

 This module consists in C implementation, using the perl module Inline::C, of essentially two methods to read
 Narcisse formated data files.
    
=head1 The two methods

	ReadFile 			:	reads a Narcisse tabular data file in one shot.
	            			Every line is stored in an array and the methods returns a reference
	            			on an array of references on those arrays. 
	
	ReadAndFormatData 	: 	reads a Narcisse tabular data file an stores data according to specified
	                    	conditions :
	                    		i) range conditions specified by
	                    			start, end, start_label and end_label
	                    		ii) value conditions specified by
	                    			a hash of the type  { col1_name => val1, col2_name => val2, ... },
	                    				
=cut

use strict;
use warnings;
use Carp;

use Data::Dumper;
use IO::File;

use General;

our $INLINE_C_CONFIG;

BEGIN
{
	use File::Basename;
	use Cwd;
	use FindBin;

#	my $mask_cgilib_dir = "Core/lib/InlineClib";
	my $mask_cgilib_dir = "Core/lib/";
	my $dirprg = $FindBin::Bin;

# 
# 	if ( $dirprg eq '.' )
# 	{
# 		$dirprg = $current_dir;
# 	}

	my $cgilib_dir = "$dirprg/$mask_cgilib_dir";

	while ( ! -d $cgilib_dir )
    {
			last if ( $dirprg eq '/' );
			$dirprg = dirname($dirprg);
			$cgilib_dir = "$dirprg/$mask_cgilib_dir";
	}

	if ( ! -d $cgilib_dir )
	{
		my $current_dir = &Cwd::cwd();
		$cgilib_dir = $current_dir;
		print STDERR "Not a regular Narcisse tree, compiling in Inline_C in $current_dir \n";
	}

    $INLINE_C_CONFIG = $cgilib_dir .'/InlineClib';

    if ( ! -d $INLINE_C_CONFIG )
    {
        mkdir $INLINE_C_CONFIG or die "ERROR - THE DIRECTORY $INLINE_C_CONFIG DOES NOT EXIST !";
        warn "$INLINE_C_CONFIG just created";
    }
}

my $VERSION = do {my @r = (q$Rev: 2474 $ =~ /\d+/g); $r[0]};

use Inline ( Config => DIRECTORY => $INLINE_C_CONFIG );

use Inline C => <<'END_OF_C_CODE'; 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINESIZE 260000				// maximum size (in char) of a line, for fgets 

/* Constants for the return type of ReadAndFormatData */
#define ARRAY  0						
#define HASH 1

/* Constants for the header */
#define MAX_FIELDS 50					// number of fields and 
#define MAX_STRING_SIZE 20000				// size (in char) of each field name

#define MAX_COND 10						// maximum authorized value conditions

/*For the indexing scheme */
#define MAX_POSITIONS 1000				// maximum number of position in the index line
#define SLICE_UNIT 100000				// the size of each slice is given in 100kb unit 

#define HEADER_MARKUP '#' 				// first character of the header line
#define COMMENT_MARKUP ';' 				// first character of comment line

#define EXIT_ON_FAILURE 1				// guess what

#define EOL '\n'						// guess what

char* strtok_alt(char *str, char *token,const char delim);

/*
Name	: ReadFile
usage	: $ra_data = ReadFile( filename )
Function: reads a tabulated file (Narcisse format) and returns a reference on the array of arrays
Args	: the filename as a string
returns	: an array of references on array, each line is stored in an array (each tab-separated string in a cell) 
*/
SV* ReadFile( char* file_name ) {
  FILE *fp;
  char line[MAXLINESIZE];
  char token[MAXLINESIZE];
  char *str, *suffix;
  AV* array;
  AV* line_array;
   	
  if( (fp = fopen(file_name, "r") ) == NULL ) {
    fprintf(stderr, "Cannot open file %s.\n",file_name);
    exit(1);
  }  
  array = newAV();
   
  while( !feof(fp) ) {
    if( fgets(line, MAXLINESIZE, fp) )
	{
	if ( 1==strlen(line) || line[0] == COMMENT_MARKUP ) //If the line is empty or commented out
		continue;
	line_array = newAV();
        str = line;
		while ( (suffix = strtok_alt(str, token, '\t')) !=NULL )
		{
			av_push(line_array, newSVpvf("%s", token));
			str = suffix;
		}
		av_push(array,newRV_noinc( (SV*)line_array ));
    }   
  }

  fclose(fp);

  return newRV_noinc((SV*)array);
}

/*
	A very elementary structure to store strings from a tabulated formated line
*/
typedef struct
{
	char t[MAX_FIELDS][MAX_STRING_SIZE];
	unsigned int size;
} tabuline;


/*
	A very elementary structure to store the value conditions:
	  colum_name1 => value1 : key = colum_name1, val = value, key_col is the column number corresponding to colum_name1
*/
typedef struct
{
	char key[MAX_STRING_SIZE];
	char val[MAX_STRING_SIZE];
	unsigned int key_col;
} condition;

/*
	A structure to store the different ReadAndFormatData parameters
	  essentially a recopy of the perl hash given as the only argument
*/
typedef struct
{
	char file[MAX_STRING_SIZE];
	long start;
	long end;
	char start_label[MAX_STRING_SIZE]; 
	char end_label[MAX_STRING_SIZE]; 
	int return_type;
	char keyname[MAX_STRING_SIZE];
	int num_condition;
	condition value_cond[MAX_COND];
	
} parameters;


void GetArguments(SV* hash_ref,parameters* args);             // Stores in a parameters structures the arguments given by has_ref
char* _GetHashValue(HV* h, const char* key_name);			  // wrapper for retrieving from a perl hash, given the key, a string value 

/* Small structure to manage the header : associating column numbers to column names */
tabuline* new_header(char* line);
int __GetHeaderIndex(tabuline* h, char *label);

/* Taking advantage of the Narcisse indexing scheme */
long __GetFilePosFromIndexline(char *line,long start, long end);

/* Managing the value conditions */
unsigned int _StoreValueCond(HV* h_cond, condition* value_cond );
void __StoreConditionHeaderIndex(tabuline* h, parameters* args);
unsigned int __ValidateValueConditions(tabuline* tb, parameters* args);

/*
Name	: ReadAndFormatData
usage	: $rh_data = ReadAndFormatData( \%h_args )
Function: reads a tabulated file (Narcisse format) and stores, in ad-hoc structures, the lines verifying the conditions
Args	: a reference on a hash with the following keys :		
			( 
			   file			: The filename as a string
			   start 		: start position of the wanted chromosome slice
               end   		: end position of the wanted chromosome slice
               start_label 	: the name of the colum on which the start value is tested
               end_label  	: the name of the colum on which the end value is tested
               mode			: return type 0|1 ARRAY or HASH
               keyname 		: the primary key if the return type is HASH
               VALUE 		: a reference on a perl hash describing the value conditions
              );
returns	: a reference on a perl hash with the following keys:
 				  _ra_return 		: an array of the array refs
 				  _rh_return 		: a hash with the array refs
 				  _ra_headers		: an array with the field names
 				  _ra_sorted_keys	: an array with the keys ordered as in the file
*/
SV* ReadAndFormatData(SV* hash_args)
{
  FILE *fp;
  char line[MAXLINESIZE];
  char token[MAXLINESIZE];
  char *str, *suffix;
  int start_label_col, end_label_col, keyname_label_col;
  long data_offset, relevant_offset;
  char* keyname_val;
  tabuline tab_line; 
  tabuline* p_header;
  int i;  
  
  parameters args;
  GetArguments( hash_args, &args);

  long start = args.start;
  long end = args.end;
   
  if( (fp = fopen( args.file, "r") )==NULL) {
    printf("Cannot open file.\n");
    exit(EXIT_ON_FAILURE);
  }  
  fgets(line, MAXLINESIZE, fp);
  relevant_offset = 0;
  data_offset = ftell( fp ); //where the actual data start
  if ( line[0] == '@' )
  {
  	relevant_offset = __GetFilePosFromIndexline(line, start, end);	
	fgets(line, MAXLINESIZE, fp);
	data_offset = ftell( fp ); //where the actual data start
  }
  // Skip comments or empty lines
  while ( 1==strlen(line) || line[0] == COMMENT_MARKUP )
  {
	  fgets(line, MAXLINESIZE, fp);
	  data_offset = ftell( fp ); //where the actual data start
  }
  p_header=new_header(line);
  
  start_label_col   = __GetHeaderIndex(p_header,args.start_label);
  end_label_col     = __GetHeaderIndex(p_header,args.end_label);
   
  keyname_label_col = -1; //Making sure this will never be used in the cae of an ARRAY
  if ( args.return_type == HASH )
  {
  	keyname_label_col = __GetHeaderIndex(p_header,args.keyname);  
  }
  
  if ( args.num_condition > 0 )
  {
  	__StoreConditionHeaderIndex(p_header,&args);
  }
     
  AV* array 	= newAV();	
  HV* hash 		= newHV();
  AV* sorted_keys = newAV();
  AV* line_array;
     
  if ( relevant_offset >= 0 ) // $relevant_offset == -1 means we are outside the chromosome range for this file
  {
  	fseek ( fp , relevant_offset+data_offset , SEEK_SET );
  	while(!feof(fp)) 
	{
  		if( fgets(line, MAXLINESIZE, fp) )
		{			
			i=0;  
			str = line;  
			while ( (suffix = strtok_alt(str, token,'\t')) !=NULL )
			{
				strcpy(tab_line.t[ i++ ],token);
				str = suffix;	
			};
			tab_line.size=i;
			
			// We check if the current line (element) is within the slice start-end
			if ( atoi( tab_line.t[ start_label_col ] ) > end )
				break;		
			if ( atoi( tab_line.t[ end_label_col ] ) < start )
				continue;
			
			// We check if the current line (element) verifies at least one of the value conditions			
			if ( args.num_condition > 0 )
			{
				if ( 0 == __ValidateValueConditions(  &tab_line, &args ) )
					continue;
			}			
			
			// if we get here the previous conditions have been fullfilled so we store the line in a perl array 				
			line_array = newAV();
			for ( i=0; i < tab_line.size ; i++ )
			{
				av_push(line_array, newSVpvf("%s", tab_line.t[i]));
			}
			if ( args.return_type == ARRAY )
			{
				av_push(array,newRV_noinc((SV*)line_array));
			}
			else
			{	
				keyname_val = tab_line.t[ keyname_label_col ];
				hv_store(hash, keyname_val, strlen(keyname_val), newRV_noinc((SV*)line_array), 0);
				av_push(sorted_keys, newSVpvf("%s", keyname_val));
			}		
		}   
  	}
  }
  
  fclose(fp);
  
  /* The header        */
  AV* header_array = newAV();	
  for ( i=0; i < p_header->size ; i++ )
  {
     av_push(header_array, newSVpvf("%s", p_header->t[i]));
  }
  
  HV* hash_return = newHV();
  hv_store(hash_return,"_ra_return",strlen("_ra_return"),newRV_noinc((SV*)array), 0);
  hv_store(hash_return,"_rh_return",strlen("_rh_return"),newRV_noinc((SV*)hash), 0);
  hv_store(hash_return,"_ra_headers",strlen("_ra_headers"),newRV_noinc((SV*)header_array), 0);
  hv_store(hash_return,"_ra_sorted_keys",strlen("_ra_sorted_keys"),newRV_noinc((SV*)sorted_keys), 0);
  
  return newRV_noinc((SV*)hash_return);
}


/*
Name	: __ValidateValueConditions
usage	: __ValidateValueConditions(&tab_line, &args)
Function: check for the given tab_line, the value conditions
Args	: a reference on the tab_line a reference on the ReadAndFormat parameters
returns	: 1 if at least one condition is verified otherwise 0 
*/
unsigned int __ValidateValueConditions(tabuline* tb, parameters* args)
{
	unsigned int flag =0;
	int i;
	
	for ( i=0 ; i< args->num_condition ; i++ )
	{ 
		if ( 0 == strcmp(tb->t[ args->value_cond[i].key_col ], args->value_cond[i].val)  )
		{
			flag++;
			break;
		} 
	}
	return flag;
} 

/*
Name	: GetArguments
usage	: GetArguments(hash_ref,&args)
Function: copy the perl hash into the parameters args structure
Args	: a reference on the perl hash given as argument to ReadAndFormatData and a reference on a parameters structure
returns	: none
*/
void GetArguments(SV* hash_ref,parameters* args)
{
	HV* hash;
	HV* h_cond;
			
	if (! SvROK(hash_ref) )
	{
		fprintf(stderr,"hash_ref is not a reference\n");
		exit(EXIT_ON_FAILURE);
	}
	
	hash = (HV*) SvRV( hash_ref );
	
	strcpy( args->file, _GetHashValue(hash, "file") );	
	args->start = atol(	_GetHashValue(hash, "start") );
	args->end = atol(	_GetHashValue(hash, "end") );
	strcpy( args->start_label, _GetHashValue(hash, "start_label") );
	strcpy( args->end_label,   _GetHashValue(hash, "end_label") );
	args->return_type = atoi( _GetHashValue(hash, "mode") );
	strcpy( args->keyname,  _GetHashValue(hash, "keyname") );		
	
	if ( hv_exists( hash, "VALUE", strlen("VALUE") ) )
	{
		SV** r= hv_fetch( hash, "VALUE", strlen("VALUE"), 0);
		if (! SvROK(*r) )
		{
			fprintf(stderr,"VALUE is not a reference\n");
			exit(EXIT_ON_FAILURE);	
		}
		h_cond = (HV*) SvRV( *r );	
		args->num_condition = _StoreValueCond( h_cond, args->value_cond);
	}
	else
	{
		args->num_condition = 0;
	}		
}

/*
Name	: _GetHashValue
usage	: _GetHashValue(h, key_name)
Function: returns the value associated with the key key_name of the perl hash h
Args	: a reference on the perl hash and a string
returns	: the value as a string
*/
char* _GetHashValue(HV* h, const char* key_name)
{
	
	SV** r= hv_fetch(h, key_name, strlen(key_name), 0);
	if ( r!=NULL )
	{
		return SvPV(*r, PL_na);
	}
	else
	{
		fprintf(stderr,"%s is not a key of your perl hash\n",key_name);
		exit(EXIT_ON_FAILURE);
	}	
}

/*
Name	: _StoreValueCond
usage	: _StoreValueCond( h_cond, value_cond )
Function: stores in the value_cond conditions-array the values of the perl hash h_cond (the VALU hash)
Args	: a perl hash and a reference on a reference on a condition array
returns	: the number of conditions
*/
unsigned int _StoreValueCond( HV* h_cond, condition* value_cond )
{
	HE* h_entry;
    int num_keys, i;
    SV* sv_key;
    SV* sv_val;
    
    num_keys = hv_iterinit( h_cond );
    for (i = 0; i < num_keys; i++) {
      h_entry = hv_iternext(h_cond);
      sv_key = hv_iterkeysv(h_entry);
      sv_val = hv_iterval(h_cond, h_entry);
      strcpy( value_cond[i].key, SvPV(sv_key, PL_na));
      strcpy( value_cond[i].val, SvPV(sv_val, PL_na));  
      value_cond[i].key_col = 0;              
    }
    
	return i;
}

/* Extracting and storing the header information */
/*
Name	: new_header
usage	: new_header(line)
Function: stores the header line in a tabuline structure
Args	: a string
returns	: a reference on a new tabuline structure
*/
tabuline* new_header(char* line)
{	
	char *str1;
	tabuline* p_tabuline = (tabuline*) malloc( sizeof(tabuline) );
	int l = 0;
	
	str1 = strtok(line, "\t");
	/* The HEADER_MARKUP is now optional
	if ( str1[0] != HEADER_MARKUP )
	{
		fprintf(stderr,"Bad file header format %s\n",line);
		exit(2);
	}
	*/
	if ( str1[0] == HEADER_MARKUP )
	{
		str1++; // Skipping the (obsolete) HEADER_MARKUP
	}
	
	strcpy(p_tabuline->t[l++],str1);
	while ( (str1 = strtok(NULL, "\t\n")) !=NULL )
	{
		strcpy(p_tabuline->t[l++],str1);	
	};
		
	p_tabuline->size=l;	
		
	return 	p_tabuline;
}

/*
Name	: __GetHeaderIndex
usage	: __GetHeaderIndex(p_header, label)
Function: retrieves the column number associated with the column name label
Args	: a reference on the tabuline object storing the header line and a string
returns	: an integer (the column number)
*/
int __GetHeaderIndex(tabuline* h, char *label)
{
	int i=0;
	int index = -1;
	
	while ( i < h->size && strcmp(h->t[i],label)!= 0 ) i++;
	
	if ( i >= h->size )
	{
		fprintf(stderr,"Label %s absent from header\n",label);
		exit(EXIT_ON_FAILURE);
	}
	
	return i;
}

/*
Name	: __StoreConditionHeaderIndex
usage	: __StoreConditionHeaderIndex(p_header, args)
Function: stores foreach value condition the column number corresponding to the column_name for that condition
Args	: a reference on the tabuline object storing the header line and a reference on the parameters
returns	: none
*/
void __StoreConditionHeaderIndex(tabuline* h, parameters* args)
{
	int i;

	for ( i=0 ; i< args->num_condition ; i++ )
	{
		args->value_cond[i].key_col = __GetHeaderIndex( h, args->value_cond[i].key );
	}
}

/* Using the index to retrieve the first relevant position */
/*
Name	: __GetFilePosFromIndexline
usage	: __GetFilePosFromIndexline(char *line,long start, long end)
Function: using the index line (the first one) as the index of the file, returns the first relevant line acccording to start and end
Args	: the index line as a string and the start and end range condition
returns	: the first relevant byte of the file
*/
long __GetFilePosFromIndexline(char *line,long start, long end)
{
	long chrom_start, chrom_end, start_offset;
	long slice_size_base, relevant_offset;
	int nb_pos, index;
	long indexes[MAX_POSITIONS];
	char *str1;
	int i;
		
	str1 = strtok(line, ";");
	str1++; /* Skipping the first character */
	chrom_start = atol(str1);	
	chrom_end = atol( strtok(NULL, ";") );
	slice_size_base = atol( strtok(NULL, ";") );
	nb_pos = atol( strtok(NULL, ";") );
	str1 = strtok(NULL, ";\n");

	/* Now reading the index positions */
	indexes[ 0 ] = atol( strtok(str1, ":\n") );
	for ( i=1 ; i < nb_pos ; i++ )
	{
		indexes[ i ] = atol( strtok(NULL, ":\n") );
	}

	start_offset = start - chrom_start;		
	if ( start > chrom_end || end < chrom_start ) // nothing to read
	{
			return -1;
	}
	else if ( start_offset > (slice_size_base*SLICE_UNIT) ) // If not we return 0
	{	
		index = ( start_offset / (slice_size_base*SLICE_UNIT) ) - 1; //Warning integer division
		if ( index >= nb_pos ) index =  nb_pos - 1 ;  // in case of (we have limited indeed the number of indexing cells to NUM_INDEX_SLICE = 50, change this ?)
		relevant_offset=indexes[ index ];		
	}
	else
	{
		relevant_offset = 0;
	}
	
	return relevant_offset;  
}


/* An alternative to strtok that avoids to merge adjacent delimiters.*/
/*
Name	: strtok_alt
usage	: suffix = strtok_alt(string,token, '\t')
Function: very close to the strtok function but does not merge adjacent delimiter 
Warning : The token must point to an array with enough space (!!)
Args	: the string to process, the token, and the delimiter
returns	: a pointer on the first char just after the first delim
*/
char* strtok_alt(char *str, char *token, const char delim) 
{
    /* Search for the token. */
    int i = 0; 
 	token[i]='\0';
	
	/*If there is no string or we are at the end of a string we return NULL */
	if ( str == NULL || *str == '\0' )
		return NULL;
	
	while( *str )
    {
            if(*str == delim)
            {
            	str++;
				break;
            }
			token[i++] = *str++;
    }
	
	
	/* We just don't want to consider end-of-line character */
	if ( token[i-1] == EOL)
	{
		token[i-1] = '\0';
	}
	else
	{
		token[i] = '\0';
	}
  
	return str;
}


END_OF_C_CODE

=pod

=head1 AUTHORS

	Thomas.Faraut@toulouse.inra.fr

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
