#
# $Id: NARCISSE_utils.pm 3811 2013-01-31 15:23:39Z  $
#

package NARCISSE_utils;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

NARCISSE_utils - A lot of very useful functions, may used by any script in Narcisse
                 NOTE - This module does not export anything, you must call the functions
                        using their complete name ex. NARCISSE_utils::GetSpeciesRoot("human")
                        SOME SYMBOLS (NONE) are exported now

=head1  SYNOPSIS

 use NARCISSE_utils;


=head1 DESCRIPTION

    We put in this module the functions that will be used by several Narcisse scripts

=cut

use strict;
use warnings;

use Carp;
use Data::Dumper;
use IO::File;
use Cwd;
use File::Spec::Functions qw(rel2abs);
use File::Basename;
use File::Path;
use CGI;

use IO::Uncompress::AnyUncompress qw(anyuncompress $AnyUncompressError) ;
use IO::Compress::Gzip qw(gzip $GzipError) ;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;

use LipmError::ParameterException;
use LipmError::IOException;
use Runner;
use ParamParser;
use General;
use Enum;
use DataReader;
use DataReader::Whole;
use DataReader::Region;
use NarcisseData::HashReturn;
use NarcisseData::ArrayReturn;
use SpeciesObserver;

# NONE is a special species name
use constant NONE => 'none';

BEGIN
{
	use vars       qw(@EXPORT);
	use base qw(Exporter);
	@EXPORT      = qw( &NONE );


	my $VERSION     = do {my @r = (q$Rev: 3811 $ =~ /\d+/g); $r[0]};
}

my $TIME = 0;

#definition of some errors returned by NARCISSE_utils::validParam
use constant
CH_ERR => {
    NO_ERROR	=> 0,
    SAME_SPECIES=> 1,
    FORBID_CHAR	=> 2,
    UNKNOWN_ORG => 3,
    NO_ORGANISM	=> ""
};

=head1 SUBROUTINES


=head2 function GetDate

 Title        : PrintDate
 Usage        : &NARCISSE_utils::PrintDate();
 Function     : Return the date, mm-dd-yy formatted
 Returns      : The date
 Args         : none

=cut
sub GetDate
{
    my @a_time = localtime();
    my $rvl = sprintf("%02d-%02d-%02d",$a_time[4]+1,$a_time[3],$a_time[5] % 100);
    return $rvl;
}

=head2 function GetTime

 Title        : GetTime
 Usage        : &NARCISSE_utils::GetTime();
 Function     : Return the date, mm-dd-yy-hh-mn-ss formatted
 Returns      : The date
 Args         : none

=cut
sub GetTime
{
    my @a_time = localtime();
    my $rvl = sprintf("%02d-%02d-%02d at %02d:%02d:%02d",$a_time[4]+1,$a_time[3],$a_time[5] % 100,$a_time[2],$a_time[1],$a_time[0]);
    return $rvl;
}

=pod

=head2 function PrettyDisplayPosition

 Title: PrettyDisplayPosition
 Usage:     pretty_pos = Prettydisplay(pos);
 Function:  display a genomic position in b, kb or mb
 Args:      pos The position in b
 return:    The position, pretty displayed (string)

=cut

sub PrettyDisplayPosition
{
	my ($pos) = @_;
	$pos += 0;
	if ($pos < 1000)
	{
		return $pos;
	};
	if ($pos < 1000000)
	{
		my $p = sprintf ("%ikb",$pos/1000);
		return $p;
	}
	my $p = sprintf("%imb",$pos/1000000);
	return $p;
}

=head2 function ReadNarcisseTabFile

 Title        : ReadNarcisseTabFile
 Usage        : ReadNarcisseTabFile(file => $file,[, $rh_param])
                ReadNarcisseTabFile is a wrapper for DataReader::Whole, it returns a hash of hashes
 Function     : Creates an hash %data= (
 					element	=> a hash of Narcisse data = ( for example
 									chr => ''
 									locus_tag => ''
 									symbol => ''
 									start => ''
 									end => ''
 									strand => ''
 								)
 					)
 Returns      : ref on %data
 Args         : Mandatory
 				$file	: the file to open
 				Optional
 				$rh_param	: a hash with parameters :
 					- primary_key_name ('element')
 					- start_label      ('start')
 					- end_label        ('end')
					- start
					- end
					-cancel_optimisation do not use the cache, force a new file read

 Globals      : none
 Errors		  : LipmError::ParameterException, LipmError::IOException

=cut
sub ReadNarcisseTabFile
{
	my %h_param = (
			file  				=> undef,
			primary_key_name	=> 'element',
			start				=> 0,
			end					=> 0,
			-cancel_optimisation => 0,
			@_
		);

	#Controlling the range conditions
	my $rh_where_condition;
	if ( $h_param{ start } == 0 && $h_param{ end } == 0 )
	{
		$rh_where_condition = { };
	}
	else
	{
		$rh_where_condition = { RANGE => \%h_param };
	}
	my $o_ndr = New DataReader::Whole();
	$o_ndr->Prepare( -from   => $h_param{ file },
	                 -where  => $rh_where_condition,
	                 -force_reading => $h_param{ -cancel_optimisation }
	               );

	my $rh_data = $o_ndr->FetchallHashref( $h_param{ primary_key_name }, -line_type => 'HASH' );
	return $rh_data;
}

=head2 function ReadChromFeaturesData

 Title        : ReadChromFeaturesData
 Usage        : ReadChromFeaturesData(file => $file,[, $rh_param])
                ReadChromFeaturesData is a wrapper for DataReader::Region
 Prerequisite : none
 Function     :
 Returns      : a NarcisseData object contaning the read data
 Args         :
 				Mandatory
 				  file				: the file to read
 				Optional
				  primary_key_name (element) : when return type is a hash, the field name to use as a key
				  start                      : staring position of the desired slice
				  end                        : ending position of the desired slice
				  start_label (start)        : label of the column used for start (q_start for prbh)
				  end_label	  (end)          : when the file format does not use 'end' as an end label (prbh has q_end)
				  return_type (HASH)         : return type

 Globals      : none
 Errors		  : LipmError::ParameterException, LipmError::IOException

=cut
sub ReadChromFeatureData
{
	my %h_param = (
			file  				=> undef,
			primary_key_name	=> 'element',
			start				=> 0,
			end					=> 0,
			start_label			=> 'start',
			end_label			=> 'end',
			return_type			=> 'HASH',
			@_
		);

	my $o_ndr = New DataReader::Region();

	my %h_range_param;
	my @range_attributes = qw( start end start_label end_label);
	@h_range_param{ @range_attributes } = @h_param{ @range_attributes };

	(defined $h_param{primary_key_name} && $h_param{primary_key_name}=~/\w+/)
		or throw LipmError::ParameterException( 'primary_key_name', 'value' );

	$o_ndr->Prepare( -from   => $h_param{ file },
	                 -where  => { RANGE => \%h_range_param }
	               );

	if ( $h_param{ return_type } eq 'HASH' )
	{
		#warn $h_param{file};
		my $rh_data        = $o_ndr -> FetchallHashref ( $h_param{primary_key_name}, -keep_order => &TRUE );
		my $ra_sorted_keys = $o_ndr -> GetSortedKeys();
		my $rh_headers     = $o_ndr -> GetHeadersColumns();
		return New NarcisseData::HashReturn ( rh_data => $rh_data, ra_sorted_keys => $ra_sorted_keys, rh_headers => $rh_headers);
	}
	else
	{
		my $ra_data     = $o_ndr -> FetchallArrayref();
		my $rh_headers  = $o_ndr -> GetHeadersColumns();
		return New NarcisseData::ArrayReturn ( ra_data => $ra_data, rh_headers => $rh_headers);
	}
}

=head2 function TabArray2Html
 Title        : TabArray2Html
 Usage        : my $html = &TabArray2Html(\@a_data);
 Prerequisite : The array must be initialized with a Narcisse-formatted file
                $a_data[0] contains the file header
 Function     : build an html table equivalent to the data in the array
 Returns      : an html table
 Args         : $ra_array   A ref to the array
 Globals      : The table belongs to the class TableData, each line belongs to the classes odd/even
 Errors		  : LipmError

=cut

sub TabArray2Html
{
    my ($ra_data) = (@_);

    if ($ra_data->[0] !~ /^#/)
    {
        throw LipmError( -text => "Headers line missing from data");
    }

    my $row0 = shift(@$ra_data);
    chomp($row0);
    my @a_headers = split('\t',$row0);
    $a_headers[0] =~ s/^#//;
    my $h_nb = @a_headers;

    # The headers
    my $html = "<table class=\"TableData\">\n<tr><th>";
    $html .= join('</th><th>',@a_headers);
    $html .= "</th></tr>\n";

    # The data
    my $p=0;
    my @a_parity=('odd','even');
    foreach my $row (@$ra_data)
    {
        chomp $row;
        $html .= "<tr class=\"$a_parity[$p]\"><td>";
        $html .= join('</td><td>',split('\t',$row,$h_nb));
        $html .= "</td></tr>\n";
        $p = 1 - $p;
    }

    # The trailer
    $html .= "</table>\n";

    return $html;
}

=head2 function TabFile2Html

 Title        : TabFile2Html
 Usage        : my $html = &TabFile2Html("file");
 Prerequisite : The file must be a narcisse-formatted data file, the first line contains the file header
 Function     : build an html table equivalent to the file
 Returns      : an html table
 Args         : $file   A file name
 Globals      : The table belongs to the class TableData, each line belongs to the classes odd/even
 Errors		  : LipmError, LipmError::IOException

=cut

sub TabFile2Html
{
    my ($file) = (@_);

	# Test parameters :
	if ( !defined $file ) {
		throw LipmError::ParameterException("\$file", 'missing');
    }

	# Open file
	my $fh_file = new IO::File($file) or throw LipmError::IOException($file,'f');

    my $line = <$fh_file>;
    chomp($line);
    if ($line !~ /^#/)
    {
        throw LipmError( -text => "Headers line missing from data" );
    }

    my @a_headers = split('\t',$line);
    $a_headers[0] =~ s/^#//;
    my $h_nb = @a_headers;

    # The headers
    my $html = "<table class=\"TableData\">\n<tr><th>";
    $html .= join('</th><th>',@a_headers);
    $html .= "</th></tr>\n";

    # The data
    my $p=0;
    my @a_parity=('odd','even');
    while (my $line = <$fh_file>)
    {
        chomp($line);
        $html .= "<tr class=\".$a_parity[$p].\"><td>";
        $html .= join('</td><td>',split('\t',$line,$h_nb));
        $html .= "</td></tr>\n";
        $p = 1 - $p;
    }

    # The trailer
    $html .= "</table>\n";

    return $html;
}

=head2 TaxIDtoSwissProtID

 Title        : TaxIDtoSwissProtID
 Usage        : TaxIDtoSwissProtID ($tax_id, $speclist_file)
 Prerequisite : none
 Fonction     : Converts the TaxID in SwissProt ID (5 or 6 letters), using the file speclist.txt
                Croaks if speclist_file not found
 Returns      : The Swissprot ID, or "" if not found
 Args         : $tax_id The tax ID to convert
                $speclist_file The path to the file speclist
 Globals      : none

=cut

sub TaxIDtoSwissProtID
{
	my ( $tax_id, $speclist_file ) = (shift,shift);

	my $tax = new IO::File($speclist_file) or Carp("Cannot open $speclist_file");
	while ( my $line = <$tax> )
	{
		next if ( $line =~ /^\s+/ );
		my @a_st = split(/\s+/, $line);
		if ( defined($a_st[2]) && ($a_st[2] =~ /^(\d+):$/) )
		{
			if ( $tax_id == $1 )
			{
				return $a_st[0];
#				return "$a_st[0].$a_st[1]";
			}
		}
	}
	return "";
}

=head2 procedure SortFile

 Title        : SortFile
 Usage        : SortFile(@array)
 Prerequisite : none
 Fonction     : rewrite files sorted on the start field
 Returns      : none
 Args         : @array : liste des fichiers a trier
 Globals      : none

FONCTION A SUPPRIMER A TERME !!!
Utilisee seulement par Pairwise::Genomic et Pairwise::Genetic

=cut

sub SortFile
{
	my @a_list_file = @_;

	foreach my $file (@a_list_file)
	{
		#print STDERR "Sorting $file ..., ";
		my %h = ();

		my $fh = new IO::File("$file") or throw LipmError::IOException($file,'f');
		my @a_lines  = <$fh>;
		$fh->close;

		my ($indice_start,$nb_keys)=&NARCISSE_utils::AssertGetColumn('start',$a_lines[0]);
		my ($indice_end,$crap0)=&NARCISSE_utils::AssertGetColumn('end',$a_lines[0]);

		#Not very clean this selection on element|gene
		my ($indice_element,$crap1)=&NARCISSE_utils::AssertGetColumn('gene|element',$a_lines[0]);

		for ( my $i=1 ; $i<@a_lines ; $i++ )
		{
			my @st = split('\t',$a_lines[$i]);
			if ( scalar(@st) != $nb_keys )
			{
				Carp::croak("SortFile/Error: file $file line $i+1: inconsistency between header ($nb_keys keys) and content ($#st+1 keys)\n");
			}
			my $idx = $st[$indice_element];
			my $second_sort_key;
			if ($idx =~/(\d+)$/)
			{
				$second_sort_key=$1; # sorting using internal id number
			} else
			{
				$second_sort_key=$st[$indice_end];
			}
			my ($nc_internal_id_number)=($idx =~/(\d+)$/);
			$h{$idx}={line => $a_lines[$i], start => $st[$indice_start], second_sort_key => $second_sort_key};
		}

		my @a_key = sort { $h{$a}{start} <=> $h{$b}{start} || $h{$a}{second_sort_key} <=> $h{$b}{second_sort_key}} ( keys(%h) );

		$fh = new IO::File(">$file") or throw LipmError::IOException($file,'f');

		# writing header
		print $fh $a_lines[0];
		foreach my $id ( @a_key )
		{
			print $fh $h{$id}{line};
		}
		$fh->close;
	}
}

=head2 function GetStream

 Usage        : my $stream = GetStream($filename);
 Fonction     : analyse a filename, if gzipped, gunzip it on the fly else return the filename
 Returns      : a stream, either a filename or a zcat on a gzipped file
 Args         : $o_io,$filename

=cut

sub GetStream
{
    my($o_io,$filename) = @_;

	if ( ! -e $filename )
	{
		Carp::croak("GetStream Error: a filename is expected\n");
	}
	if ( $filename =~ /\.gz/ )
	{
        my $zcat = $o_io->GetUnixPath('UNIX_ZCAT');
		$filename = "$zcat $filename|";
	}
	return $filename;
}

=head2 procedure RunCmd

 Title        : RunCmd
 Usage        : my $sts = &RunCmd($cde,$stdout,$stderr)
 Fonction     : Run the command $cde using Runner.pm's RunExt, optionally redirecting stdout and stderr
                The return value and kill status is printed on stdout
                Die if the return value != 0
 Returns      : nothing (but may die)
 Args         : $cde The command
                $stdout (optional): the stdout file. If not specified, the file is opened new (>)
                $stderr (optional): The stderr file.
 Globals      :

=cut

sub RunCmd
{
    my ($cde,$stdout,$stderr) = @_;
    my ($status, $killed);
    if (defined $stdout and $stdout !~ /^>/)
    {
        $stdout = '>'.$stdout;
    };
    if (defined $stderr and $stderr !~ /^>/)
    {
        $stderr = '>'.$stderr;
    };

    if (defined $stdout and defined $stderr)
    {
        ($status, $killed) = RunExt(-cmd=>$cde,
                                    -stdout=>$stdout,
                                    -stderr=>$stderr);
    }
    elsif (defined $stdout)
    {
        ($status, $killed) = RunExt(-cmd=>$cde,
                                    -stdout=>$stdout);
    }
    else
    {
       ($status, $killed) = RunExt(-cmd=>$cde);
    };
    if ($status==0 and $killed==0)
    {
        my @cde = split(' ',$cde);
        foreach my $token (@cde)
        {
            $token = basename($token);
        };
        my $short_cde = substr(join(' ',@cde),0,100);
        print $short_cde . " ... OK\n";
    }
    else
    {
        print "ERROR !!! - $cde - EXIT CODE $status - KILL CODE $killed\n";
        print "LEAVING NOW\n" and die;
    }
}


=head2 procedure RunCmdFromFile

 Title        : RunCmdFromFile
 Usage        : &RunCmdFromFile($file,$stdoe)
 Procedure    : Run the commands found in $file (one command per line), redirecting stdout and stderr to $stdoe
                We use RunCmd, and die with a message if a value != 0 is returned
 Returns      : nothing
 Args         : $file : a file containing the comands to run (1 cmd per line)
                $stdoe: stdout + stderr
 Globals      :

=cut

sub RunCmdFromFile
{
	my ($file, $stdoe) = @_;
	open(CMD,$file) or die "Cannot open $file";
	while(my $cde=<CMD>)
	{
		chomp($cde);
		&RunCmd($cde,">$stdoe",">>$stdoe");
	}
}


=head2 procedure Welcome

 Title        : Welcome
 Usage        : &Welcome($program,$version)
 Procedure    : Report information on the program itself
 Returns      : nothing
 Args         : $program: program name
 				$version: program version
				$cfg: narcisse configuration file
				$genome_repository: genome repository
				$snp repository: snp repository
 Globals      : none

=cut

sub Welcome
{
    my ($program,$version,$cfg,$genome_repository,$snp_repository) = @_;

	print '-' x 30 . "\n" . basename($program) . " rel $version: \n" . '.' x 5 . "\n";
	print "CONFIGURATION FILE         = $cfg\n";
	print "NARCISSE_GENOME_REPOSITORY = $genome_repository\n" if (defined $genome_repository);
	print "NARCISSE_SNP_REPOSITORY    = $snp_repository\n" if (defined $snp_repository);

}

=head2 procedure RedirectOE

 Title        : RedirectOE
 Usage        : &RedirectOE($o_io,$output,$error)
 Procedure    : Redirect std output and std error to the files, without disturbing the default stream
 Returns      : nothing
 Args         : $output: the output file name
 				$error: the error file name

=cut

sub RedirectOE
{
    my ($o_io,$stdout_file,$stderr_file) = @_;
    my $tee = $o_io->GetUnixPath('UNIX_TEE');
    open (STDOUT, "| $tee -a $stdout_file") or croak "Cannot redirect stdout to $stdout_file";
    open (STDERR, "| $tee -a $stderr_file 2>&1") or croak "Cannot redirect stderr to $stderr_file";
}

=head2 function AssertGetColumn

 Title        : AssertGetColumn
 Usage        : my ($ind, $nb_col) = &AssertGetColumn($string_to_seek,$header_line)
 Fonction     : returns the indice of the string in the header, die if not found
 Args         : $tag: A column name to seek
                WARNING - We are seeking the col name through a regex match. If you want to check with an EQUALITY,
                please use anchors: ^foo$ rather than only foo
                $header_line: The header line of the file to check - must start by #
 Returns      : The column number (0 = 1st col)
                The total number of columns
 Globals      : none

=cut

sub AssertGetColumn
{
    my($tag,$line) = @_;

	my $ind = -1;
	my @a = split('\t', $line);
	for ( my $i=0 ; $i<=$#a ; $i++ )
	{
		if ( $a[$i] =~ /$tag/ )
		{
			$ind = $i;
			last; # I want the first start which can be: start, q_start, s_start, start_translation
		}
	}
	&Carp::croak("Error: $tag not found in header >$line<\n") if ( $ind == -1 );
	return ($ind,$#a+1);
}

=head2 procedure AssertCfgParam

 Title        : AssertCfgParam
 Usage        : &AssertCfgParam($o_io,$o_param,$cfg)
 Procedure    : Canonize the 'cfg' parameter value, and check its value

 Args         : $o_io   : an Instance Observer
                $o_param: a ParamParser object
                $cfg    : the parameter to check ('cfg', 'cfgQ', etc)
                $where  : may be INSTANCE or SESSION (default = INSTANCE)
 Returns      : nothing - Croaks if anything wrong

=cut

sub AssertCfgParam
{
    my ($o_io,$o_param,$cfg,$where) = @_;
    $where = 'INSTANCE' unless(defined $where);

    if ($o_param->IsDefined($cfg))
    {
        # If not a file path, it is a species short name: compute the path
        if ($o_param->Get($cfg) !~ /[\/.]/)
        {
            my $short = $o_param->Get($cfg);

            # Check this name
            if ($o_io->GetSpeciesShortName($short) eq 'none')
            {
                &Carp::croak ("ERROR - The genome $short does not exist in this instance\n");
            }

            # compute the cfg file
            my $species_dir = $o_io->GetSpeciesCfgDirpath($where);
            $o_param->Set($cfg,File::Spec->catfile($species_dir,$short.'.cfg'));
        }

        # Make an absolute path
        $o_param->Set($cfg,rel2abs($o_param->Get($cfg)));
	    $o_param->AssertFileExists($cfg);
	}
}

=head2 procedure ShowErrorPage

	Title       :	ShowErrorPage
	Usage       :	&ShowErrorPage($error_title, $message [, $url])
	Prerequiste :	none
	Procedure   :	print a section of html file dedicated to display errors
	Returns     :	none
	Args        :	$error_title,	a title for this error
	            	$message,	    the message that caracterise/explain the error
                	$url,	        optional
	Globals     :	none

=cut

sub ShowErrorPage
{
	my ($error_title, $error_msg, $home_url) = @_;

    print "<div class=\"error\"><h2>$error_title</h2><p>$error_msg</p>";
    print "<p>Go to <a href=\"$home_url\">$home_url</a> to continue</p>" if (defined($home_url));
    print "</div>\n";
    return;
}

=head2 function ChromIDComparison

 Title        : ChromIDComparison
 Usage        : ChromIDComparison($chra,$chrb)
Prerequisite  : this comparison test doesn't seem to be exported
 Function     : sort using chromosome ID which can be: an integer, a string with integer, a string
 Returns      : just like <=> or cmp
 Args         : the two chromosom names to compare
 Globals      : none
=cut

sub ChromIDComparison
{
 	my ($a,$b)=(shift,shift);

	my $aIsNumber = $a =~ /^\d+$/;
	my $bIsNumber = $b =~ /^\d+$/;

	if ($aIsNumber && $bIsNumber)           # both are numbers
	{
		return $a <=> $b;
	}
	elsif ($aIsNumber)                      # Only a is a number
	{
		return -1;
	}
	elsif ($bIsNumber)                      # Only b is a number
	{
		return 1;
	}
	else
	{
		# cas particulier reference circos : le nom du chrs commence par un prefix
		use Circos::Narcisse;
		my $prefix = &Circos::Narcisse::GetPrefixRef();

		my $alphaa = ( $a =~ /^$prefix(.*)$/) ? $1 : $a;
		my $alphab = ( $b =~ /^$prefix(.*)$/) ? $1 : $b;

		# cas particluier des chromosomes Unknown et MT
		$alphaa = ($alphaa =~ /^(MT|Un)(.*)$/) ? lc($1).$2 : $alphaa;
		$alphab = ($alphab =~ /^(MT|Un)(.*)$/) ? lc($1).$2 : $alphab;

		my ($numa, $numb);
		($alphaa,$numa) = ($alphaa =~ /^([^0-9]+)(\d+)$/) ? ($1,$2) : ($alphaa,1);
		($alphab,$numb) = ($alphab =~ /^([^0-9]+)(\d+)$/) ? ($1,$2) : ($alphab,1);

		return ($alphaa cmp $alphab || $numa <=> $numb);                    # alphanumeric sort
	}

}

=head2 function SortByChromIDComparison

 Title        : SortByChromIDComparison
 Usage        : sort SortByChromIDComparison @array
 Prerequisite : this comparison test doesn't seem to be exported
 Function     : sort using chromosome ID which can be: an integer, a string with integer, a string
 Returns      : none
 Args         : none
 Globals      : $a and $b

=cut

sub SortByChromIDComparison
{
	return ChromIDComparison($a,$b);
}

=head2 function SortByChromID

 Title        : SortByChromID
 Usage        : @sorted_array = SortByChromID (@array)
 Prerequisite : the comparison test doesn't seem to be exported
 Function     : sort using chromosome ID which can be: an integer, a string with integer, a string
 Returns      : none
 Args         : @array, an array containning the name of chromosomes
 Globals      : none

=cut

sub SortByChromID
{
	return sort SortByChromIDComparison @_;
};

=head2 function GetAdmAnswer

 Title        : GetAdmAnswer
 Usage        : $answer = GetAdmAnswer($defaults,"1|2|3|4","2")
 Prerequisite : none
 Fonction     : return the answer.
                If $defaults != "", the answer is taken from $defaults, and there is no dialog
                Else, <STDIN> is read for the answer, if "" is entered the 3rd parameter is taken as the answer
                Else a match is done agains 2nd parameter, and <STDIN> is read until the match succeeds
 Returns      : none
 Args         : $def_param If different from "", this value is returned without any dialog
                $allowed   A regex, the answer should match the regex
                $default   Default value, if the user answers Return", this value is returned
 Globals      : none

=cut

sub GetAdmAnswer
{
    my ($def_param,$allowed,$default) = @_;
    my $answer;

	if ( $def_param )
	{
		$answer = "$def_param";
	}
	else
	{
		do
		{
			$answer = <STDIN>;
			chomp($answer);
			if ( $answer eq "" )
			{
				$answer = $default;
			}
			$answer = "\U$answer";
		}
		while ($answer !~ /$allowed/ && $allowed ne "" && (print "Try again, should be in $allowed :") );
	}
	return "$answer";
}

=head2 AddIds

 Usage    : $o_x->AddIds( $ra_data, $key )
 Procedure: Add an identifier to each hash ref of $ra_data.
            the key name is $key, the value is the position of the row in the array ref.
 Args     :
     $ra_data, array ref of hash ref. Each hash ref is a row of data.
	 $key, string, key name of the identifier field.
 Error    : none.

=cut

sub AddIds_DONOTUSE
{
	my( $ra_data, $key ) = @_;

	for( my $i=0; $i < @$ra_data; ++$i )
	{
		my $rh_row = $$ra_data[$i];
		$$rh_row{ $key } = $i+1;
	}

	return;
} #AddIds



=head2 Procedure GetDataAsHtmlTable

 Usage        : $html_table = GetDataAsHtmlTable( $ra_headers, $ra_data, $rh_options )
 Function     : .
 Args         :
     $ra_headers, $ra_data, $rh_options
 Error        : none.

 Note : elements of $ra_data shouldn't be undef.
 TODO - Utilise SEULEMENT par evaluate_certification.pl, peut-etre a supprimer

=cut

sub GetDataAsHtmlTable
{
	my( $ra_headers, $ra_data, $rh_options ) = @_;

	my %h_options = ( fn_sort  => \&SortNarcisseDataOnStartEnd,
					  o_logger => undef,
					  erase    => &TRUE,
					  %$rh_options );
	my @a_keys = qw( fn_sort o_logger erase );
	my ( $fn_sort, $o_logger, $erase ) = @h_options{ @a_keys };

	# headers
	my $headers   = "\t\t<tr>\n";
	foreach my $header ( @$ra_headers )
	{
		$headers .= "\t\t\t<th>$header</th>\n";
	}
	$headers     .= "\t\t</tr>";

	#data
	my $ra_rows = &$fn_sort( $ra_data );

	my $data = "";
	foreach my $rh_row ( @$ra_rows )
	{
		my @a_row = @$rh_row{ @$ra_headers };
		$data .= "\t\t<tr>\n";
		foreach my $elt ( @a_row )
		{
			$data .= "\t\t\t<td>$elt</td>\n";
		}
		$data .= "\t\t</tr>\n";
	}

	my $html_table = "<table>\n\t<thead>$headers\n\t</thead>\n";
	$html_table   .= "\t<tboby>$data\t</tboby>\n</table>\n";

	return $html_table;
} #GetDataAsHtmlTable


=head2 SortNarcisseDataOnStartEnd

 Usage    : my $ra_data = $o_x->SortNarcisseDataOnStartEnd( $ra_data )
 Function : Sort the array $ra_data, using $a->{start} and $a->{end} as sort keys
 Args     : $ra_data  An array of rows (ie an array of ref to hashes)
 Return   : A ref to the sorted array

 Error    : none

=cut

sub SortNarcisseDataOnStartEnd
{
	my( $ra_data ) = @_;

	my @a_sorted = sort
		{
			$$a{ start } <=> $$b{ start }
			|| $$a{ end } <=> $$b{ end } # deuxieme clef de tri
		} @$ra_data;

	return \@a_sorted;
}

=head2 function FindAllExtensions

  Title        : __FindAllExtensions
  Usage        : $ra_ext = &__FindAllExtensions( $path );
  Prerequisite : Enum is used to avoid doublon
  Function     : List and return all file extensions of $path
                 It is non recursive.
  Returns      : $ra_ext, a ref list of string, the extensions.
  Args         : $path, string, the path of the dir to examine.
                 	$path will be open by IO::Dir.
  Globals      : none

  Error        : LipmError::IOException, if $path can't be opened.

=cut

sub FindAllExtensions
{
	my ( $path ) = @_;

	my $o_enum = New Enum();
	my $dh = new IO::Dir( $path )
	             or throw LipmError::IOException( "-$path-", 'd', -depth => 1 );

	while( my $file = $dh->read() )
	{
		next	if( $file =~ /^\./ );
		$file =~ /([^.]+)$/;
		$o_enum->AddValues( $1 );
	}
	$dh->close();

	my @a_ext = $o_enum->ListItems();

	return \@a_ext;
}

=head2 function GetFileHandle

 Title        : GetFileHandle
 Usage        : $fh = $o_xx->GetFileHandle( $filename, $mode );
 Prerequisite : none
 Function     : this is essentially an overload of the IO::FILE constructor enabling in addition to deal with zipped files

 Returns      : a file handle such as those provided by IO::File of IO::Uncompress::Gunzip
 Args         : $key, string, a key of %CFG

#TODO

=cut
sub GetFileHandle
{
	my $file = shift;
	my $mode = (scalar @_ ) ? shift @_ : 'r';

	( defined $file )
		or throw LipmError::ParameterException('file','missing');

	( $mode =~ /[rw]/ )
		or throw LipmError::ParameterException('mode',"unauthorized mode $mode for GetFileHandle");

	my $fh;
	if ( $mode eq 'r' )
	{
		if (!-e $file)
		{
			if (-e "$file.gz")
			{
				$file.=".gz";
			}
			elsif ( -e "$file.zip" )
			{
				$file.=".zip";
			}
			elsif( -e "$file.bz2" )
			{
				$file.=".bz2";
			}
		}
		$fh = new IO::Uncompress::AnyUncompress $file
        	or die "anyuncompress failed: $AnyUncompressError\n";

	}
	elsif ( $mode eq 'w' &&  $file =~/\.gz$/ )
	{
		$fh = new IO::Compress::Gzip $file
        	or die "gzip failed: $GzipError\n";
	}
	else
	{
		$fh = new IO::File $file, $mode
			or throw LipmError::IOException( $file, 'f' );
	}

	return $fh ;

}

=head2 Procedure Unzip

 Title        : Unzip
 Usage        : Unzip( $filename )
 Prerequisite : none
 Procedure    : Unzip the file if needed, wathever the compression algorithm is, and change the filename accordingly
 Args         : a filename

=cut

sub Unzip
{
	my $file = shift;
	my $file_gz;

	if( !defined $file )
	{
		# le rapport donne la ligne d'appel de la fonction
		throw LipmError::ParameterException( '$file', 'missing', -depth => 0.5 );
	}

	if ( -s $file && $file !~ /\.(gz|bz2|zip)$/ )
	{
		#print STDERR "Already uncompressed file $file\n";
		return ;
	}

	my $suffix=undef;
	if (!-s $file)
	{
		my @suffixes = qw( gz zip bz2 );
		foreach my $suff ( @suffixes )
		{
			if (-e "$file.$suff")
			{
				$suffix = $suff;
				last;
			}
		}
		unless (defined $suffix)
			{
				warn("File $file.(gz|bz2|zip) is missing we therefore skip the uncompressing step");
				return ;
			}
		$file_gz = "$file.$suffix";
	}
	else
	{
		$file_gz = $file;
		$file =~ s/\.(gz|bz2|zip)$//;
	}

	my $status = anyuncompress $file_gz => $file
        	or die "uncompressed failed: $AnyUncompressError\n";
    unlink( $file_gz );

	return;
}

=head2 Procedure Zip

 Title        : Zip
 Usage        : Zip( $filename )
 Prerequisite : none
 Procedure    : Zip the file if not already zipped
 Args         : a filename

=cut

sub Zip
{
	my $filename = shift;

	unless (-e $filename)
	{
		warn("File  $filename is missing we therefore skip the compressing step");
		return ;
	}


	unless ( $filename =~ /\.(gz|bz2|zip)$/ )
    {
		my $input = $filename;
		my $output = "$filename.gz";
		#print STDERR "Zipping file $input \n";
		my $status = gzip $input => $output
        or die "gzip failed: $GzipError\n";
        unlink( $input );
	}

	return;
}

=head2 function DownloadData

 Title        : DownloadData
 Usage        : (1) if ( &NARCISSE_utils::DownloadData($o_io,$file) ) ...
                (2) my $path = &NARCISSE_utils::DownloadData($o_io,$file,$download_dir[,$o_looger])
 Fonction     : Download the file if needed (ftp) or test that it really exists (file://, or /)
                (1) NARCISSE_utils::DownloadData($o_io,$file);
                    Returns 1 if the file exists, 0 if it does not
                (2) NARCISSE_utils::DownloadData($o_io,$file,$download_dir,$o_looger);
                    Download the file and return the path of the downloaded file
 Returns      : (1) 0 or 1
                (2) The path of the download file
 Args         : $o_io               = An instance observer
                $uri                = uri of the file to download
                $download_dir (opt) = the download directory
                $o_logger (opt)     = A logger object
 Globals      : none

=cut

sub DownloadData
{
	my ($o_io,$uri,$download_dir,$o_logger) = @_;
	my $only_chk = (defined $download_dir) ? 0 : 1;
	return unless defined $uri;
	return unless $uri ne '';

	my $unix_wget= $o_io->GetUnixPath('UNIX_WGET');
	my $unix_curl= $o_io->GetUnixPath('UNIX_CURL');
	my $unix_cp  = $o_io->GetUnixPath('UNIX_CP');

	# Download through ftp
	if ( $uri =~ /^ftp:\/\// )
	{
        if ($only_chk)
        {
			# If the file exists, it should return 63, because it is longer than 10 bytes.
			# If the file is in fact a directory, it should return 0, because --max-filesize does not apply
			# If the file does not exist, it should return something else (probably 78 for a file, 9 for a directory, etc.)
			my $cmd = "$unix_curl --max-filesize 10 $uri";
			my ($status, $killed) = RunExt(-cmd=>$cmd, -stdout=>">/dev/null", -stderr=>">/dev/null");
			return ($status==63 || $status==0) ? 1 : 0;
        }
        else
        {
			my $sav_dir = getcwd();
			chdir $download_dir;
			my $local_file = $uri;
			$local_file =~ s/^ftp:\/+/\//;
			$local_file = basename $local_file;
			unlink $local_file;
			$o_logger->Trace("Starting downloading $uri\n",'','A') if ( defined $o_logger);
			$o_logger->StartTimer('Download');
			#&NARCISSE_utils::RunCmd("$unix_wget $uri");
			&NARCISSE_utils::RunCmd("$unix_curl -o $local_file $uri");
			$o_logger->Trace("End download $local_file\n",'Download','A');
			chdir $sav_dir;
			return File::Spec->catfile($download_dir,$local_file);
		}
	}

	# only local file, we just check
	elsif ( $uri =~ /^file:\/\// or $uri =~ /^\// )
	{
		my $src_file = $uri;
		$src_file =~ s/^file:\/+/\//;

		if ($only_chk)
		{
			return (-e $src_file) ? 1 : 0;
		}
		else
		{
			my $local_file = basename $src_file;
			$local_file = File::Spec->catfile($download_dir,$local_file);
			`$unix_cp $src_file $local_file`;
			return $local_file;
		}
	}

	# download through something else: ERROR
    else
    {
        &Carp::croak("ERROR - $uri should start with 'ftp://' or 'file:///' or '/'\n");
	}
}

=head2 ParseNarcisseRegion

 Usage    : my $rh_regions = ParseNarcisseRegion( $o_io, $ra_nc_regions, rh_options )
 Function : a Narcisse region is a string, we need to extract each part of it.
 Prerequiste : All species of $ra_nc_regions should belong to $o_io, else none will be their short name.
 Args     :
     $o_io, InstanceObserver object, needed to convert long name to short name.
     $ra_nc_regions, list ref of string, the Narcisse region "nom_long:chrom:start:end:strand".
	 $rh_options, hash ref for options, not yet implemented.
 Returns  : $rh_regions, hash ref of regions like this :
    { short_species_name => { chrom_label => [ { start => 1, end => 5000, strand => + } ] } }

=cut

sub ParseNarcisseRegion
{
	my( $o_io, $ra_nc_regions, $rh_options ) = @_;

	my %h_regions = ();

	foreach my $nc_region ( @$ra_nc_regions )
	{
		my( $species, $chrom, $start, $end, $strand ) = split( /:/, $nc_region );

		my $short = $o_io->GetSpeciesShortName( $species );

		my $ra_chrom = $h_regions{ $short }{ $chrom } || [];
		push( @$ra_chrom, { start => $start, end => $end, strand => $strand } );

		$h_regions{ $short }{ $chrom } = $ra_chrom;
	}

	return \%h_regions;
}



=head2 Procedure PrintNarcisseData

 Title        : PrintNarcisseData
 Usage        : &PrintNarcisseData( $filepath, $ra_headers, $rh_data, $rh_options );
 Prerequisite : none
 Procedure    : print data of $rh_data into $filepath.
                This function will be rewrite/copy in near-future,
                into the module to deal with Narcisse data.
 Args         : $filepath, string the filepath
                $ra_headers,
                $rh_data,
                $rh_options,
                        sort, string, to specify how to sort value of rh_data keys,
                        verbose, bool, true if you want some infos. -> Not use for now

=cut

sub PrintNarcisseData
{
        my ( $filepath, $ra_headers, $rh_data, $rh_options ) = @_;

my @a_ = caller();
warn "deprecated, try the New NarcisseDataWriter\n@a_";
        #do nothing if no data
        my @a_ids = keys %$rh_data;     # one key per output line
        if( 0 == scalar( @a_ids ) )
        {
                my $msg = "no data, $filepath was NOT written.";
                &General::Verbose( $rh_options, $msg ); # TODO => logger

                return;
        }

        my $SEP = "\t";

        #creation de l'arborescence si elle n'existe pas
        my $dirname = dirname( $filepath );
        if( ! -e $dirname )
        {
                mkpath( $dirname );
        }

        my $fh = new IO::File( $filepath, 'w' );
        # donner un code d'erreur pour "fichier de donnee non ecrit nombre pair car non fatal."
        throw LipmError::IOException( $filepath, 'f', -value => 0 )     unless( defined $fh );

        # print headers
        print $fh '#' . join( $SEP, @$ra_headers ) . "\n";

        if( $$rh_options{ sort } and $$rh_options{ sort } eq 'num' )
        {
                @a_ids = sort { $a <=> $b } @a_ids;
        }
        else
        {
                @a_ids = sort @a_ids;
        }

        # print data
        foreach my $id ( @a_ids )
        {
                my $rh_values = $$rh_data{ $id };
                my @a_values = @$rh_values{ @$ra_headers };
                print $fh join( $SEP, @a_values ) . "\n";
        }

        $fh->close();

        my $msg = "$filepath was successfully written.";
        &General::Verbose( $rh_options, $msg ); # TODO => logger
#       print $msg . "\n";

        return;
}

=head2 Procedure PrintNarcisseData2

 Title        : PrintNarcisseData2
 Usage        : &PrintNarcisseData2( $filepath, $ra_headers, $rh_data, $rh_options );
 Prerequisite : none
 Procedure    : print data of $rh_data into $filepath.
                This function will be rewrite/copy in near-future,
                into the module to deal with Narcisse data.
 Args         : $filepath, string the filepath
                $ra_headers,
                $rh_data,
                $rh_options,
                        sort, string, to specify how to sort value of rh_data keys,
                        verbose, bool, true if you want some infos. -> Not use for now

TODO : rh_data devient ra_data
        trie : si ref func => on l'utilise
                        si list ref de string, on utilise les clef dans l'ordre
                        ajouter une option sorted order -> asc/des ?
                        ==> ne pas inclure l'option de trie ?
                                => ecriture = trie plus appel a PrintNarcisseData2
=cut

sub PrintNarcisseData2
{
        my ( $filepath, $ra_headers, $rx_data, $rh_options ) = @_;

        my %h_options = ( fn_sort  => \&SortNarcisseDataOnStartEnd,
                                          o_logger => undef,
                                          erase    => &TRUE,
                                          comments => undef,
                                          %$rh_options );
        my @a_keys = qw( fn_sort o_logger erase );
        my ( $fn_sort, $o_logger, $erase ) = @h_options{ @a_keys };

my @a_ = caller();
warn "deprecated, try the New NarcisseDataWriter\n@a_";
        #do nothing if no data
        my $is_array_ref = ( 'ARRAY' eq ref( $rx_data ) ); # si c'est un array ref, a-t-on besoin de trier ?
        my @a_rows = $is_array_ref ? @$rx_data : values %$rx_data;
        if( 0 == scalar( @a_rows ) )
        {
                my $msg = "no data, $filepath was NOT written.\n";
                $o_logger->Trace( $msg, '.', 'W' )  if( defined $o_logger );

                return;
        }

        my $SEP = "\t";

        #creation de l'arborescence si elle n'existe pas
        my $dirname = dirname( $filepath );
        if( ! -e $dirname )
        {
                my @a_created = mkpath( $dirname );
                my $msg = join( ', ', @a_created ) . "created\n";
                $o_logger->Trace( $msg, '.', 'WV' )  if( $o_logger );
        }

        my $fh = new IO::File( $filepath, 'w' )
                                or throw LipmError::IOException( $filepath, 'f' );

        # print comments if some
        if( defined $h_options{ comments } )
        {
                print $fh "; $h_options{ comments }\n";
        }

        # print headers
        print $fh '#' . join( $SEP, @$ra_headers ) . "\n";

        my $ra_rows = &$fn_sort( \@a_rows );

        # print data
        foreach my $rh_values ( @$ra_rows )
        {
                next if( 0 == keys %$rh_values );

                my @a_values = @$rh_values{ @$ra_headers };
                print $fh join( $SEP, @a_values ) . "\n";
        }
        $fh->close();

        my $msg = "$filepath was successfully written.\n";
        $o_logger->Trace( $msg, '.', 'W' ) if( defined( $o_logger ) );

        return;
} #PrintNarcisseData2


=pod

=head1 AUTHORS

	Yoann.Beausse@toulouse.inra.fr
	sebastien.letort@toulouse.inra.fr
	Jerome.Gouzy@toulouse.inra.fr
	Thomas.Faraut@toulouse.inra.fr
	Emmanuel.courcelle@toulouse.inra.fr

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
