#
# 	sebastien.letort@toulouse.inra.fr
#	Created: August 22, 2008
#	$Id: NDetector.pm 3677 2011-04-17 10:34:18Z sletort $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package NDetector;

use strict;
use warnings;

use base qw( ThroughSpeciesComputer );


=pod

=head1 NAME

	NDetector - class to detect N base into

=head1  SYNOPSIS

	# code

=head1 DESCRIPTION

 This classe compute Ocean of N ; it is a N detector !

	'-file_location', string, can be 'INSTANCE' or 'SESSION'.
			'SESSION', default value, will print data into session directory.
			'INSTANCE', will print data into instance data directory.
 Attributes  :
	_file_location, string, where data will be written, into session or instance

=cut

use Data::Dumper;
use English;
use Error qw( :try );

use General;
use NARCISSE_utils;


=head2 method _Init

	Title        : _Init
	Usage        : my $o_x = New NDetector( %h_args );
	Prerequisite : none
	Fonction     : initialise an instance of NDetector class.
	Returns      : a valid NDetector object
	Args         : %h_args
	Attributes   :

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -size      => 5_000,
	               -force     => &FALSE,
	               -file_location => 'SESSION',

	              # mother
# 	               -species   => [],
# 	               -chroms     => [],
# 	               -o_logger  => undef,
	              @_ );

	# set class attributes
	$self->_Set( _force  => $h_args{ -force },
	             _regexp => qr/N{$h_args{ -size },}/,
	             _file_location => $h_args{ -file_location } );

	# Check
	$self->SUPER::_Init( %h_args );

	my $msg = ref( $self ) . " launch for ";
	$msg   .= "\tregexp = " . $self->{ _regexp };
	$self->_Trace( $msg . "\n", '.', 'IC', __PACKAGE__ );

	return $self;
}


=head2 function _ComputeData

	Title        : _ComputeData
	Usage        : $rh_data = $o_xx->_ComputeData();
	Prerequisite : none
	Function     : none
	Returns      : $rh_data
	Args         : none

=cut

sub _ComputeData
{
	my $self = shift;

	my $filepath   = $self->_ComputeFilepath();
	if( -e $filepath and ( &FALSE == $self->{ _force } ) )
	{
		my $msg = "Skip detection\n";
		$msg   .= "file already exists : $filepath\n\n";
		$self->_Trace( $msg, '.', 'IC', __PACKAGE__ );

		return;
	}

	my $rh_data = $self->_SearchInChrom();

	return $rh_data;
}


=head2 function _SearchInChrom

	Title        : _SearchInChrom
	Usage        : $rh_data = $o_xx->_SearchInChrom( none );
	Prerequisite : none
	Function     : none
	Returns      : $rh_data
	Args         : none

=cut

sub _SearchInChrom
{
	my $self = shift;

	my %h_data = ();
	my $i = 1;

	my $rh_current = $self->_Get( '_rh_current' );
	my $chrom      = $$rh_current{ chrom };
	my $o_sla      = $$rh_current{ o_slice_adaptor };

	# chrom name circos-compliant
	my $chr_name = $o_sla->db()->species() . '-' . $chrom;

	# obtenir la sequence
	my $o_slice = $o_sla->fetch_by_region( 'chromosome', $chrom );
	my $seq = $o_slice->seq();

	my $regexp = $self->_Get( '_regexp' );
	while( $seq =~ /$regexp/g )
	{
		my $end   = pos($seq);        # last base included in 1-based
		my $len   = length( $MATCH ); # $MATCH is internal variable, containing the matched regexp
		my $start = $end - $len + 1;

		$self->_Trace( "$start-$end => $len\n", '.', 'ICV', __PACKAGE__ );

		$h_data{ $i } = { element => $i, chrom => $chr_name,
						start => $start, end => $end };

		++$i;
	}

	return \%h_data;
}


=head2 function __ComputeFilepath

	Title        : _ComputeFilepath
	Usage        : $filepath = $o_xx->_ComputeFilepath();
	Prerequisite : none
	Function     : Compute the filepath and make the path if necessary.
	Returns      : $filepath, the formatted absolute filepath
	Args         : none

=cut

sub _ComputeFilepath
{
	my $self = shift;

	my @a_attr = qw( _rh_current _file_location );
	my( $rh_current, $where ) = $self->_Get( @a_attr );

	my $chrom      = $$rh_current{ chrom };
	my $o_sla      = $$rh_current{ o_slice_adaptor };
	my $species    = $o_sla->db()->species();

	my $o_species = New SpeciesObserver( species => $species );
	my $data_path = $o_species->GetDataPath( $where );

	my $o_inst_obs = $self->_GetInstanceObserver();
	my @a_keys = qw( MASK_FILE_CHROM DIR_N );
	my ( $mask, $dir_n ) = $o_inst_obs->GetCfg( @a_keys );

	my $path = File::Spec->catdir( $data_path, $dir_n );
	File::Path::mkpath( $path, { verbose => $self->{ _verbose } } );

	my $filename = join( '', $mask, $chrom, '.', $dir_n );
	my $filepath = File::Spec->catfile( $path, $filename );

	return $filepath;
}


=head2 function _GetHeaderFile

	Title        : _GetHeaderFile
	Usage        : $ra_headers = $o_xx->_GetHeaderFile();
	Prerequisite : none
	Function     : this function returns the headers of the datafile.
	             	by default it is
	             	[ 'chrom', 'start', 'end', 'value', 'opt_circos', 'id' ]
	Returns      : $ra_headers, list ref of headers,
	              	their order is their order in the datafile
	Args         : none.
	Note         : It should be redefine in daughter class.

=cut

sub _GetHeaderFile
{
	my @a_headers = qw( chrom element start end );

	return \@a_headers;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

