#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: March 22, 2009
#	$Id: NarcisseData.pm 2115 2009-03-22 07:24:38Z manu $
#

=pod

=head1 NAME

	NarcisseData - Interface class between DataReader and Objects using the data
	               NarcisseData defines seeters and accessors, you may use without knowing the encapsulated data structures
	               The data structures can be hash of arrays or array of arrays

=head1  SYNOPSIS


	my $o_ndr = 	New DataReader::Region();
	$o_ndr->Prepare( 
	                 -from   => $file,
	                 -where  => { RANGE => \%h_range_param },
	               );	  
	my $rh_data        = $o_ndr -> FetchallHashref ( key_name => 'element' )
	my $ra_sorted_keys = $o_ndr -> GetSortedKeys();
	my $ra_headers     = $o_ndr -> GetHeaders();
	foreach my $element ( $o_nd -> GetSortedElements() )
	{
		# Do something with each element	
	}
	

=head1 DESCRIPTION

	This class is used to enable an interface between the data reading object and the objects actually using of data
	This class is created in order to enable an easy reimplementation of NarcisseDataReader class with Inline::C++

=cut

package NarcisseData;

use strict;
use warnings;

use base qw( LipmObject );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;

use General;

#=head2 method _Init

 #Usage        : my $o_x = New NarcisseData();
 #Prerequisite : none
 #Fonction     : Constructor
 #Args         : none
 #Attibuts     :
  #__rh_return		: a reference on the hash containing the data
  
  #__rh_headers 		:  keys are columns name, values are columns position
  #__ra_sorted_keys 	: when data are retrived in a hash,
                      #we need, for backward compatibility to give a way to retrieve the order of row.  
  #__storage_mode 	:  how each line of the data file is stored : as a HASH or as an ARRAY (default : HASH)
  #__return_type  	: indicates the return type (default : HASH)

#=cut
#sub _Init
#{
	#my $self = shift;
	#my %h_args = (  
				  #rh_data        => {},
				  #ra_data	     => [],         # a reference on an array of arrays (the Narcisse data one array per line)
				  #rh_headers     => {},         # a reference on a hash specifying the meaning of the columns : keys column names, values column numbers
				  #@_	
				#);
			
	#$self->_Set( 
				 #_ra_data   		=>  $h_args{ ra_data },
	             #_rh_data   		=>  $h_args{ rh_data },
	             #_rh_headers 		=>  $h_args{ rh_headers },
	             #_ra_sorted_elts    => []                      # will be initialized in the sons
	             #);
							
	#return $self;

#}

#----------------------------------
#    The methods
#---------------------------------

=head2 IsEmpty

 Title        : IsEmpty
 Usage        : $self -> IsEmpty()
 Prerequisite : none
 Function     : Check if the object effectively contains data ( ra_data should be a reference on a non empty array )
 Returns      : 1 if ra_data is a reference on an empty array
 				or
 				0 if not
 
=cut
sub IsEmpty
{
	my $self = shift;
	my $ra_sorted_elts = $self->_Get( '_ra_sorted_elts' );
	
	my $IsEmpty= 1;
	$IsEmpty=0 if ( scalar @$ra_sorted_elts  != 0  );
		
	return $IsEmpty;
			
}

=head2 Size

 Title        : Size
 Usage        : $self -> Size()
 Prerequisite : none
 Function     : calculates and returns the number of elements
 Returns      : number of elements
 
=cut
sub Size
{
	my $self = shift;
	my $ra_sorted_elts = $self->_Get( '_ra_sorted_elts' );
	
	return scalar @$ra_sorted_elts;
}


=head2 AddFields

 Title        : AddFields
 Usage        : $object -> AddFields( $elt, field_key => field_value [, field_key2 => field_value2 , ... ] )
 Prerequisite : none
 Function     : add or modify one or more fields to the data structure corresponding to $elt.
                The field label is created if necessary
 Returns      : none
 Args         : elt :  the element id or ref (hashreturn or ArrayReturn)
                a hash 

=cut
sub AddFields
{
	my $self = shift;
	my $elt  = shift;
	my %h_attr = @_;
	
	my $rh_headers = $self->_Get( '_rh_headers' );
	
	foreach my $key ( keys %h_attr )
	{
		if ( !defined $rh_headers->{ $key } )
		{
			$self->_NewLabel( $key );
		}
		my $value = $h_attr{ $key };	
		my $ref_data=$self->GetElementData( $elt );
		$ref_data->[ $rh_headers->{ $key }] = $value;
	}

	return;
	
}

=head2 _NewLabel

 Title        : _NewLabel
 Usage        : $object -> _NewLabel( field_key )
 Prerequisite : none
 Function     : add a label to rh_headers, the corresponding column label is set
 Returns      : none
 Args         : a string corresponding to the label

=cut
sub _NewLabel
{
	my $self = shift;
	my $label = shift;
	
	my $rh_headers = $self->_Get( '_rh_headers' );
	my @a_labels = keys %$rh_headers;
	
	$rh_headers->{ $label } = $#a_labels +1;
}

=head2 GetFields

 Title        : GetFields
 Usage        : $object -> GetFields($elt, @a_fields )
 Function     : return the values of each field of @fields corresponding to $elt (only implemented in HashReturn of Arrayreturn)
 Returns      : @a_values The corresponding values
 Args         : $elt : For ArrayReturn, a ref to a line
                       For HashReturn, a key
 				@a_fields : list of the fields to return
 Warning      : not implemented here		
=cut
sub GetFields
{
	my $self = shift;
	throw LipmError::ObjectException( $self, 'abstract method' );
}

=head2 GetElementData

 Title        : GetElementData
 Usage        : $object -> GetElementData( $elt )
 Prerequisite : none
 Function     : returns the line referred to by $elt
 Returns      : a reference on the data (thus a ref to an array)
 Args         : $elt : the element id
 Warning      : IMPLEMENTED ONLY IN HashReturn !
=cut
sub GetElementData
{
	my $self = shift;
	throw LipmError::ObjectException( $self, 'abstract method' );
}

=head2 GetSortedElements

 Title        : GetSortElements
 Usage        : $object -> GetSortElements()
 Prerequisite : data must be initialized
 Function     : If ArrayReturn: Returns an array of reference on each line (in fact an array organized as in the file)
                If HashReturn : Returns an array of keys in the same order as in the file
 Returns      : an array of reference
 Args         : none

=cut
sub GetSortedElements
{
	my $self = shift;
	return @{ $self->_Get( '_ra_sorted_elts' ) };
}

=head2 FindNearestElement

 Title        : FindNearestElement
 Prerequisite : The data are loaded, the file was sorted using the 'start' field so that GetSortedElements returns a correct stuff
 Usage        : my $element = $self->FindNearestElement($position)
 Fonction     : We return the data element whose 'start' field is closest from $position (dichotomy algorithm)                
                This method is use to know bounds of the representation when user use a 'zoom_level' (for example if zoom_level 'gene' show 5 genes, we search the index of the third one)
 Args         : $position : a position next to which we want to find the corresponding element
                            It is an index in the array returned by GetSortedElements
 Returns      : the corresponding element (the key if HashReturn, or the row if ArrayReturn)

=cut

sub FindNearestElement
{
    my $self = shift;
    my $position = shift;

 	my @a_sorted_elements = $self->GetSortedElements();

    # Test index
    my $start = 0;
    my $end   = $#a_sorted_elements;
    my $middle = int($end/2);
	
    # Calculate the index nearest to $position.
    while ( $end - $start > 1)
    {
        my $middle_elt = $a_sorted_elements[ $middle ];
        my ( $middle_elt_start ) = $self->GetFields( $middle_elt, 'start' );
        if ($position <= $middle_elt_start)
        {
            $end = $middle;
        }
        else
        {
            $start = $middle;
        }
        $middle = int( ($end+$start)/2);
    };

	my $nearest_elt = $a_sorted_elements[ $start ];

    return $nearest_elt;
}

1;
