#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: March 22, 2009
#	$Id: NarcisseData.pm 2115 2009-03-22 07:24:38Z manu $
#

=pod

=head1 NAME

	NarcisseData - Interface class between NarcisseDataReader and Objects using the data

=head1  SYNOPSIS

	# code

=head1 DESCRIPTION

	This class is used to enable an interface between data reading and actual use of data
	This class is created in order to enable an easy reimplementation of NarcisseDataReader class with Inline::C++

=cut

package NarcisseData::HashReturn;

use strict;
use warnings;

use base qw( NarcisseData );

use Data::Dumper;
use Carp;

use LipmError::ObjectException;
use General;

=head2 method _Init

 Usage        : my $o_x = New NarcisseData::HashReturn ( rh_data => \%h_data, ra_sorted_keys => \@a_sorted_keys, rh_header => \%h_header );
 Fonction     : Constructor
 Args         : rh_data         The data, should be a hash of arrays
                rh_header       The header, a hash: key = labels, values = column numbers
                ra_sorted_keys  The keys in the same order as in the file
=cut
sub _Init
{
  my $self = shift;
  my (%h_args) = @_;
  throw LipmError::ParameterException("rh_data", 'missing')        unless defined $h_args{ rh_data };
  throw LipmError::ParameterException("rh_header", 'missing')      unless defined $h_args{ rh_headers };
  throw LipmError::ParameterException("ra_sorted_keys", 'missing') unless defined $h_args{ ra_sorted_keys };

  $self -> _Set ( _rh_data        => $h_args { rh_data } );
  $self -> _Set ( _ra_sorted_elts => $h_args { ra_sorted_keys } );
  $self -> _Set ( _rh_headers      => $h_args { rh_headers } );
		
  return $self;
}


#----------------------------------
#    The methods
#---------------------------------


=head2 GetElementData

 Title        : GetElementData
 Usage        : $object -> GetElementData( $elt )
 Prerequisite : none
 Function     : returns the structure containing the element data
 Returns      : a reference on the data structure
 Args         : elt : the element internal id
 				
=cut
sub GetElementData
{
	my $self = shift;
	my $elt = shift;
	
	my $rh_data = $self->_Get( '_rh_data' );
	
	return $rh_data->{ $elt };
		
}

=head2 GetFields

 Title        : GetFields
 Usage        : $object -> GetFields($ref_a, @a_fields )
 Prerequisite : none
 Function     : return the values of each field of @fields in the hash ra_data corresponding to the
                columns of the keys $id
 Returns      : @fields values
 Args         : elt : the element internal id
 				fields : list of the fields to return

=cut
sub GetFields
{
	my $self = shift;
	my ( $elt, @a_fields ) = @_;

	#We are not using $self->_Get( 'rh_data' ) for efficiency reason
	my $rh_data = $self->{ _rh_data };
	
	my @a_return = ();

	# If the element doesn't exist an empty array of the correct size is returned
	if (!defined $rh_data->{ $elt })
	{
		my @a_return = ();
		for (my $i=0; $i< scalar @a_fields; $i++)
		{
			push @a_return, "";
		}
			
		return @a_return;
	}
	
	my $ref_elt = $rh_data->{ $elt }; 

	# The different labels
	my %h_headers = %{ $self->{ _rh_headers } };

	# We make the assumption that the fields exist in the header of the file!!
	return @{$ref_elt}[ @h_headers{@a_fields} ];
}

=head2 Exist

 Title        : Exist
 Usage        : $object -> Exist( $elt_id )
 Prerequisite : none
 Function     : returns true is $elt_id is an existing element_id
 Returns      : 0|1
 Args         : elt : a candidate element internal id
 				
=cut
sub Exist
{
	my $self = shift;
	my $elt  = shift;
	
	my $rh_data = $self->_Get( '_rh_data' );
	
	return ( defined $rh_data->{ $elt } );
}

1;
