#
#	Created: Feb 26, 2010
#	$Id: FastaReader.pm 3201 2010-02-05 13:32:36Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package NarcisseDataWriter;

=pod

=head1 NAME

 NarcisseDataWriter - a class to write data, Narcisse-formatted

=head1  SYNOPSIS

 Create an object, specifying the file path, the headers, some options
	my $o_wrt = New NarcisseDataWriter($filepath, $ra_headers, $rh_options, $o_logger)
	
 Add a chunk to the file: data are in $rx_data, a ref to a hash or to an array
 $comment is an optional comment which will be inserted at top of file
 
    $o_wrt->Update($rx_data,$comment);
    
 Update the file: the file is read to a hash using the 'element' column as the key,
 then data from $rh_data are used to complete the hash
    $o_wrt->UpdateFile($rh_data,$rh_options,$comment);
    
    
=head1 DESCRIPTION

 This class is useful to write Narcisse-formatted files from data in memory, whether in a hash or in an array
 You may write a lot of small chunks, sorting the file at the end: this is convenient for very big files
 In the opposite, you may also update the file, ie read the existing file, add data and sort in memory, the dump everything back
 
 NarcisseDataWriter is an abstract class, DO NOT INSTANTIATE
 Only derived classes should be instantiated, they MUST redefine the following methods:
      _OpenFile
      Update

  Log is obtained with 'WV'.

=head1 BUGS


=cut

use base qw( LipmObject );

use strict;
use warnings;

use Data::Dumper;
use IO::File;

use General;
use LipmError::IOException;
use LipmError::ParameterException;

use File::Basename;
use File::Path;

use constant SEPARATOR => "\t";

=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New NarcisseDataWriter( $filename, $ra_headers, $rh_options, $ra_sort_keys, $o_logger );
                my $o_x = New NarcisseDataWriter( $filename, undef, { append => &TRUE } ); $o_x = undef;
 Prerequisite : none
 Function     : (1) The file is opened, if necessary it is created and the headers are written
                (2) The file, which must exist, is just sorted
 Args         : $filename (required)                    : filename of a NArcisse formatted file
                $ra_headers (required unless append)    : The headers, created if the file does not exist or is empty, 
                                                          checked if the file already exists (append mode)
                                                          In append mode, this parameter is not required
                $rh_options: The options
                    sort: boolean [true]                : Sort rows in memory before writing them to the file
                    sort_numerical: boolean [true]      : Sort all fields numerically
                    OR
                    sort_numerical: { 'start'=>1,'end'=>1 }   : Sort numerically ONLY the fields start and end, the other fields, if sorted,
                                                                are sorted alphabetically
                    sort_ascending: boolean [true]      : Sort ascending
                    sort_uniq     : boolean [false]     : Keep only one value of each key (see sort -u)
                    append: boolean [false]             : If true, open the file in append mode
                    comment: string (undef)             : Comment to write before the headers
                $ra_sort_keys: ref_to_array ['start','end'] : The list of keys used to sort            
                                                          To sort using the columns start,end:
                                                              New NarcisseDataWriter('file',$ra_headers);
                                                          To sort using the column "position":
                                                              New NarcisseDataWriter('file',$ra_headers,$rh_options,[ 'position' ]);
                                                          To sort alphabetically on the column "element":
                                                              New NarcisseDataWriter('file',$ra_headers,
                                                                                     { sort_numerical => &FALSE }, [ 'element' ]);
                                                          To sort alphabetically on the column species, then numerically on the columns start,end:
                                                              New NarcisseDataWriter('file',
                                                                                     [ species, ..., start, end ],
                                                                                     { sort_numerical => { 'start'=>1, 'end'=>1 ] },
                                                                                     [ 'species','start','end' ] );
                $o_logger [undef], a Logger object.
                
 Error        : IOException, ParameterEXception

=cut

sub _Init
{
	my $self = shift;
	my ( $filename, $ra_headers, $rh_options, $ra_sort_keys, $o_logger ) = @_;
	throw LipmError::ParameterException('filename','missing') unless defined $filename;
	$rh_options = {} if (! defined $rh_options);

	if( ( defined $ra_sort_keys ) && ( 'ARRAY' ne ref( $ra_sort_keys ) ) )
	{
		throw LipmError::ParameterException( 'ra_sort_keys', 'type',
												-should_be => "array ref" );
	}
	$ra_sort_keys = [ 'start','end' ] if (! defined $ra_sort_keys || @$ra_sort_keys==0 );

	$self->_Set('_sort',           (defined $rh_options->{sort}) ? $rh_options->{sort} : &TRUE );
	
	if ( defined $rh_options->{sort_numerical} && ( 'HASH' eq ref( $rh_options->{sort_numerical} ) ) )
	{
		$self->_Set('_sort_numerical',&FALSE);
		$self->_Set('_sort_numerical_byfield',$rh_options->{sort_numerical});
	}
	else
	{
		$self->_Set('_sort_numerical', (defined $rh_options->{sort_numerical}) ? $rh_options->{sort_numerical} : &TRUE );
		$self->_Set('_sort_numerical_byfield',{});
	}
	$self->_Set('_sort_ascending', (defined $rh_options->{sort_ascending}) ? $rh_options->{sort_ascending} : &TRUE );
	$self->_Set('_sort_uniq',      (defined $rh_options->{sort_uniq}) ? $rh_options->{sort_uniq} : &FALSE );

	$self->_Set('__ra_headers',$ra_headers);
	$self->_Set('_ra_sort_keys',$ra_sort_keys);
	$self->_Set('_filename', $filename);
	$self->_Set('_fh_file', undef);

	$self->SUPER::_Init( $o_logger );

	# If not in append mode:
	#    - $ra_headers is required and must be an array
	#    - We remove the existing file
	if (!defined($rh_options->{append}) || $rh_options->{append} == &FALSE)
	{
		throw LipmError::ParameterException('ra_headers','missing') unless defined $ra_headers;
		throw LipmError::ParameterException('ra_headers','type') unless (ref($ra_headers) eq 'ARRAY');
		unlink $filename;
	}

	$self->__CreatePath();
	$self->__CheckHeaders();

	my $wrt_head_flg = $self->_Get('__wrt_head_flg');
	$self->_OpenFile($wrt_head_flg,$rh_options->{comment});

	return $self;
}

sub GetHeaders
{
	my $self = shift;
	return $self->_Get('__ra_headers');
}

=head2 __CreatePath

 Usage    : $self->__CreatePath()
 Procedure: If necessary, Create the path towards the wanted file
 Args     : none

 Error    : See perldoc File::Path

=cut

sub __CreatePath
{
	my $self = shift;

	my $filename = $self->_Get('_filename');	
	my $dirname  = dirname( $filename );
	if( ! -e $dirname )
	{
		my @a_created = mkpath( $dirname );
		my $msg = join( ', ', @a_created ) . "created\n";
		$self->_Trace( $msg, '*', 'WV' );
	}

	return;
}


=head2 __CheckHeaders

 Usage    : $self->__CheckHeaders();
 Procedure: If the file does not exist or is Zero length, does nothing
            If the headers is still undef (append mode), read and set the headers
            Else, read the first lines, search for the first line, may be skipping comments
            ; some comment
            ; some other comment
            col1	col2 the header
            If found, the header is compared with the __ra_headers private attribute
            If not found or different from private attribute, an exception is thrown

 Error    : IOException or ParameterException

=cut

sub __CheckHeaders
{
	my $self = shift;
	
	my $filename = $self->_Get('_filename');
	if (-e $filename && ! -z $filename)
	{
		my $line = '';
		my $fh_filename = new IO::File($filename) or throw LipmError::IOException($filename,'f');
		while( $line = <$fh_filename> )
		{
			next if $line =~ /^;/;
			last;

			# on n'arrive jamais ici si on a un en-tete
			throw LipmError::IOException($filename,'f',-text => 'ERROR - File format not recognized');
		}
		
		chomp $line;
		$line =~ s/^#//;
		
		my $ra_headers = $self->_Get('__ra_headers');
		if ( ! defined $ra_headers )
		{
			my @a_headers = split ( /\t/, $line );
			$self -> _Set( '__ra_headers',\@a_headers );
		}
		else
		{
			my $headers    = join( "\t", @$ra_headers );
			if( $headers ne $line )
			{
				throw LipmError::ParameterException('labels','value',-text => "Requested labels = $headers - Labels in file = $line");
			}
		}
		$self->_Set('__wrt_head_flg',&FALSE);
	}
	else
	{
		my $ra_headers = $self->_Get('__ra_headers');
		if ( ! defined $ra_headers )
		{
			throw LipmError::ParameterException('ra_headers','missing',-text => "Headers are required here, even in append mode, because the file is empty");
		}
		$self->_Set('__wrt_head_flg',&TRUE);
	}

	return;
}


=head2 _Openfile

 Usage    : $self->_OpenFile($wrt_head_flg);
 Procedure: If the file does not exist or is Zero length, open the file, write the comment and the headers
            Else, just open in append mode
            Set the __fh_file attribute

 Error    : IOException

=cut
sub _OpenFile
{
	throw LipmError::ObjectException( 'NarcisseDataWriter', 'abstract method' )
}

sub Update
{
	throw LipmError::ObjectException( 'NarcisseDataWriter', 'abstract method' )
}		

=head2 Procedure _PrintComments

 Title        : _PrintComments
 Prerequisite : The file must be opened (we do not check)
 Usage        : $self->_PrintComments($comment)
 Procedure    : Print a comment, being sure we print only 1 line
 Args         : @a_comments The comments to print
 
           
=cut
sub _PrintComments 
{
	my $self = shift;
	my (@a_comments) = @_;
	
	my $fh_file = $self->_Get('_fh_file');
	foreach my $c (@a_comments)
	{
		my @a_c = split( /\n/, $c );	# Be sure there is only 1 line of comments 
		$a_c[0] =~ s/^;+//;
		print $fh_file ";$a_c[0]\n";
	}
}
	
=head2 Procedure _PrintHeaders

 Title        : _PrintHeaders
 Prerequisite : The file must be opened (we do not check)
 Usage        : $self->_PrintHeaders()
 Procedure    : Print the header, being sure we print only 1 line
 Args         : none
           
=cut
sub _PrintHeaders
{
	my $self = shift;
	
	my $ra_header  = $self->_Get('__ra_headers');
	foreach my $h (@$ra_header)
	{
		$h =~ s/\n//gm;
	}
	my $fh_file = $self->_Get('_fh_file');
	#print $fh_file '#' . join( &SEPARATOR, @$ra_header ) . "\n";
	print $fh_file join( &SEPARATOR, @$ra_header ) . "\n";
}

=head2 _SortOnStartEnd

 Usage    : &_SortOnStartEnd( $ra_data )
 Function : Sort the array $ra_data, using $a->{start} and $a->{end} as sort keys
            NOTE - This is NOT a method, this is an ordinary function
 Args     : $ra_data  An array of rows (ie an array of ref to hashes)
 Return   : A ref to the sorted array

 Error    : none

=cut

sub _SortOnStartEnd
{
	my( $ra_data ) = @_;

	my @a_sorted = sort
		{
			$$a{ start } <=> $$b{ start }
			|| $$a{ end } <=> $$b{ end } # deuxieme clef de tri
		} @$ra_data;

	return \@a_sorted;
}



=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
