#
#	Created: Feb 26, 2010
#	$Id: FastaReader.pm 3201 2010-02-05 13:32:36Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package NarcisseDataWriter::OnMemory;

=pod

=head1 NAME

 NarcisseDataWriter - a class to write data, Narcisse-formatted, using feww IO but keeping the fole file in memory

=head1  SYNOPSIS

 See NarcisseDataWriter... BUT:
     1/ Function Update: the data MUST be in a hash, NOT in an array
     2/ Function __Init: not all sort options are recognized
    
=head1 DESCRIPTION

 At each call to Update, the data are stored to a hash in memory
 When the destructor is called the file is written
 
=head1 BUGS


=cut

use base qw( NarcisseDataWriter );

use strict;
use warnings;

use Data::Dumper;
use IO::File;

use General;
use LipmError::IOException;
use LipmError::ParameterException;

use File::Basename;
use File::Path;

=head2 method _Init

 Fonction     : The constructor. Just die if some switch is not implemented, and call the superclass

=cut

sub _Init
{
	my $self = shift;
	
	$self->SUPER::_Init(@_);
	
	if ( $self->_Get('_sort_uniq') == &TRUE )
	{
		die "ERROR " . ref ($self) . ": Option sort_uniq NOT IMPLEMENTED";
	}

	return;
}

=head2 _OpenFile

 Usage    : $self->_OpenFile($wrt_head_flg,$comment);
 Procedure: Open the file and read it using NarcisseDataReader
            We keep the comments inside an array
 Args     : $wrt_head_flg (true/false)
            $comment [undef]: A comment to insert at top of file
 Error    : IOException

=cut

sub _OpenFile
{
	my $self = shift;
	my ($wrt_head_flg,$comment) = @_;

	my $filename     = $self->_Get('_filename');
	
	# read the file if not empty
	if ($wrt_head_flg == &TRUE)
	{
		$self->_Set('__ra_data',[]);
		$self->_Set('__ra_comments',[]);
	}
	
	# ... read the existing file if it does not exist
	else
	{
		# Read the file to an hash, using NarcisseDataReader
		my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
		$o_ndr->Prepare( -from => $filename );
		my $ra_data = $o_ndr->FetchallArrayref();
		$self->_Set('__ra_data',$ra_data);
	
		# Extract the comments
		my $o_conf   = &ConfigurationFactory::GetConfiguration();
		my $unix_grep= $o_conf -> GetUnixPath('UNIX_GREP');
		my $cmd = "$unix_grep -E '^;' $filename";
		my @a_comments = `$cmd`;
		$self->_Set('__ra_comments',\@a_comments);
	}
	
	# Write the comment
	if (defined $comment)
	{
		my $ra_comments = $self->_Get('__ra_comments');
		push @$ra_comments,$comment;
	}

	return;
}

=head2 Procedure Update

 Title        : Update
 Usage        : $o_obj->Update( $rx_data, $comment );
 Procedure    : Write a chunk of data to the underlying file
                See also __SortFile

 Args         : $rx_data: A ref to a hash OR to an array
                          Each element is a hash, the key names are the names of the header array
                          We do not check if those names are the same as declared in $ra_header
                $comment: A comment to insert with the data (optional)
                	
=cut

sub Update
{
	my $self = shift;
	my ( $rx_data, $comment ) = @_;
	
	# Do nothing if no data
	my $is_array_ref = ( 'ARRAY' eq ref( $rx_data ) );
	my @a_rows = $is_array_ref ? @$rx_data : values %$rx_data;
	if( 0 == scalar( @a_rows ) )
	{
		my $filename = $self->_Get('_filename');
		my $msg = "no data, $filename was NOT written.\n";
		$self->_Trace( $msg, '*', 'W' );
		return;
	}
	
	# Append data to the data already in memory
	my $ra_data = $self->_Get('__ra_data');
	push (@$ra_data,@a_rows);

	# Keep comments
	if( defined $comment )
	{
		my $ra_comments = $self->_Get('__ra_comments');
		push(@$ra_comments,$comment);
	}

	return;
}


=head2 Procedure __SortData

 Usage     : $self->__SortData ();
 Procedure : Sort the data
 Args      : none
=cut 

sub __SortData
{
	my $self = shift;
	
	my $ra_keys  = $self->_Get('_ra_sort_keys');
	my $ascending= $self->_Get('_sort_ascending');
	my $numerical= $self->_Get('_sort_numerical');
	
	# TODO - NOT IMPLEMENTED !!!
	#my $uniq     = $self->_Get('_sort_uniq');
	
	# TODO - NOT IMPLEMENTED !!
	# sort_numerical by field (see mother class, and the OnDisk implementation)
	my $ra_data  = $self->_Get('__ra_data');
	
	my $s;
	my @a_sorted_data;

	if ($numerical == &TRUE)
	{
		@a_sorted_data = sort
		{
			foreach my $k (@$ra_keys) { $s=($$a{$k} <=> $$b{$k});last if($s!=0); }; $s
		} @$ra_data;
	}
	else
	{
		@a_sorted_data = sort
		{
			foreach my $k (@$ra_keys) { $s=($$a{$k} cmp $$b{$k});last if($s!=0); }; $s
		} @$ra_data;
	}
	
	if ($ascending == &FALSE)
	{
		my @a_reversed = reverse @a_sorted_data;
		$self->_Set('__ra_data',\@a_reversed);
	}
	else
	{
		$self->_Set('__ra_data',\@a_sorted_data);
	}
}	

=head2 Procedure __WriteData

 Usage     : $self->__WriteData ();
 Procedure : Write the data to the disk
 Args      : none
=cut 
sub __WriteData
{
	my $self = shift;
	my $filename = $self->_Get('_filename');
	my $fh_file  = new IO::File($filename, 'w') or throw LipmError::IOException($filename,'f');
	$self->_Set('_fh_file',$fh_file);
	my $ra_comments = $self->_Get('__ra_comments');
	my $ra_headers  = $self->GetHeaders();
	my $ra_data     = $self->_Get('__ra_data');
	
	# Write the comments and the header
	$self->_PrintComments(@$ra_comments);
	$self->_PrintHeaders();

	# Write the data
	foreach my $rh_values ( @$ra_data )
	{
		next if( 0 == keys %$rh_values );
		my @a_values = @$rh_values{ @$ra_headers };
		print $fh_file join( &NarcisseDataWriter::SEPARATOR, @a_values ) . "\n";
	}

	$fh_file -> close();
	$self->_Set('_fh_file',undef);
}
	
=head2 Procedure DESTROY

 Usage     : THE DESTRUCTOR
 Procedure : Close the file, then call __SortFile if the sort was asked by the constructor
 Args      : none
=cut 

sub DESTROY
{
	my $self = shift;

	$self->__SortData() if ($self->_Get('_sort') == &TRUE);
	$self->__WriteData();
	
	return;
}




=head2 Procedure __Close

 Title        : __Close
 Usage        : $self->__Close()
 Procedure    : Close the file and set __fh_file to undef
 Args         : none
           
=cut
sub __Close { 
	my $self = shift;
	my $fh_file = $self->_Get('__fh_file');
	$fh_file->close();
	$fh_file->_Set('__fh_file',undef);
}

=head2 Procedure __Reopen

 Title        : __Reopen
 Usage        : $self->__Reopen()
 Procedure    : If already opened, does nothing
                If closed, opens the file
 Args         : none
 
 Error        : IOException
           
=cut
sub __Reopen {
	my $self = shift;
	my $fh_file = $self->_Get('__fh_file');
	unless (defined $fh_file)
	{
		my $filename = $self->__Get('__filename');
		$fh_file = new IO::File($filename, 'a') or throw LipmError::IOException($filename,'f');
		$self->_Set('__fh_file',$fh_file);
	}
}
	

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
