#
# 	sebastien.letort@toulouse.inra.fr
#	Created: december 09, 2008
#	$Id: NarcisseService.pm 3407 2010-07-06 09:05:58Z sletort $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package NarcisseService;

=pod

=head1 NAME

 NarcisseService - abstract class to manage "services" for One instance.

=head1  SYNOPSIS

 # code
 # this class have to be inherited
 # to redefine _Run, _TreatOptions methods

=head1 DESCRIPTION

 This new version of NarcisseService aims to simplify the writing of
 services for one instance of Narcisse.
 This class will be derived in classes that are specifics to a service type.
 This is more an interface than a real Class as it only provided user functions
 FetchResults, FetchNotFound and the initialisation of the attribute _o_instance_obs.

=cut

use strict;
use warnings;

use Data::Dumper;
use LipmError::ObjectException;

use base qw( LipmObject );

use Configuration::InstanceObserver;


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New NarcisseService( -o_logger => $o_logger,
                                               -o_instance_obs => $o_inst_obs
                                             );
 Prerequisite : none
 Fonction     : initialise an instance of NarcisseService class.
                It relies on an InstanceObserver object to access instance data.
                This method set a logger, the _o_instance_obs attribute
                then call _TreatOptions methods.
 Returns      : a valid NarcisseService object
 Args         : both optional
                $o_logger => a logger object, if none, no log.
                $o_inst_obs => an InstanceObserver object, if none, an new one
                              is build with default args ($ENV{ NARCISSE_CFG }).
 Attributes   :
     __o_logger      : Logger object [undef]

     _o_instance_obs : InstanceObserver object [undef]
     __rh_results    : hash ref of whatever {empty}
                           keys are ordered in __ra_results
#                        If stored data are not hash, they are stored as 'data' =>
# undef
     __ra_results    : list ref of whatever [empty]
     __ra_not_found  : list ref of whatever is not found [empty].

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger       => undef,
	               -o_instance_obs => undef,
	               @_ );
	$self->SUPER::_Init( @_ );

	$self->_TreatOptions( %h_args );

	$self->_Set( __rh_results    => {},
	             __ra_results    => [],
	             __ra_not_found  => [] );

	return $self;
}

=head2 Function FetchResults

Title        : FetchResults
Usage        : my ($ra_headers,$rh_data) = $o_xx->FetchResults();
Function     : Launch the search and return the found data
               the _Run method, responsable for populating the _ra_results array,
               is called
Args         : none
return       : $ra_results: an array ref of sorted headers
               $rh_results: a  hash ref:
                   key = header
                   val = sequence (in a string, unformatted)
               You may want to use the function Sequence_utils::FormatSequenceLine to print a fasta file
               NOTE - Please note we return the whole data set, NOT an iterator

=cut

sub FetchResults
{
	my $self = shift;

	$self->_Run(); # name due to retro compatibility

	my $ra_results = $self->_GetArrayResults();
	my $rh_results = $self->_GetHashResults();

	return ( $ra_results, $rh_results );
}


=head2 Function FetchNotFound

 Title        : FetchNotFound
 Usage        : my $ra_not_found = $o_xx->FetchNotFound();
 Prerequisite : FetchResults must have been previously called
 Function     : Return the list of input data which did not produce any result
                NOTE - The objects decorated by FromDecorator::RegionsDecorator do NOT return any FetchNotFound identifier 
 Args         : none
 return       : $ra_results, a ref list of identifiers.

=cut

sub FetchNotFound
{
	my $self = shift;

	return $self->_GetNotFound();
}


=head2 Procedure _StoreResults

 Title        : _StoreResults
 Usage        : $o_xx->_StoreResults( $anything );
 Prerequisite : none
 Procedure    : push $anything into __rh_results
     if $anything is a hash ref, only one key is considered.
 Args         : $anything, whatever can be stored

=cut

sub _StoreResults
{
	my $self       = shift;
	my( $anything )= @_;

	my $ra_results = $self->_GetArrayResults();
	my $rh_results = $self->_GetHashResults();

	if( ref( $anything ) eq 'HASH' )
	{
		my ( $key, $value )  = each( %$anything );
		$$rh_results{ $key } = $value;
		push( @$ra_results, $key );
	}
	else
	{
# 		$$rh_results{ $anything } = undef;
# 		push( @$ra_results, $anything );

		# changement a verifier
		if( defined $$rh_results{ data } )
		{
			my $ra_ = $$rh_results{ data };
			push( @$ra_, $anything );
		}
		else
		{
			my $ra_ = $$rh_results{ data } = [];
			push( @$ra_, $anything );
			push( @$ra_results, 'data' );
		}
	}

	return;
}


=head2 Procedure _StoreNotFound

 Title        : _StoreNotFound
 Usage        : $o_xx->_StoreNotFound( $anything );
 Prerequisite : none
 Procedure    : push $anything into __ra_not_found
 Args         : $anything, whatever can be stored

=cut

sub _StoreNotFound
{
	my $self       = shift;
	my @a_anything = @_;

	my $ra_not_found = $self->_GetNotFound();
	push( @$ra_not_found, @a_anything );

	return;
}


=head2 Procedure _TreatOptions

 Usage        : $o_xx->_TreatOptions( %h_args );
 Prerequisite : none
 Procedure    : This method will be call to treat specific constructor's
                arguments. That's why it have to be redefined if needed.
                This method sets _o_instance_obs attribute.
 Args         : cf _Init

=cut

sub _TreatOptions
{
	my $self = shift;
	my %h_args = @_;

	LipmError::Function::Assert( %h_args );

	# si pas d'observer, on le cree
	my $o_inst_obs;
	if( defined $h_args{ -o_instance_obs } )
	{
		$o_inst_obs = $h_args{ -o_instance_obs };
		my $name = $o_inst_obs->GetName();
		my $msg = "Will use InstanceObserver passed as argument ($name).\n";
		$self->_Trace( $msg, '.', 'ISV' );
	}
	else
	{
		my $msg = "No InstanceObserver passed as argument, ";
		$msg   .= "will build a new one.\n";
		$self->_Trace( $msg, '.', 'ISV' );
		$o_inst_obs = InstanceObserver::Singleton( -o_logger => $self->_GetLogger() );
	}

	$self->_Set( _o_instance_obs => $o_inst_obs );

	return;
}


=head2 Procedure _Run

 Usage        : $o_xx->_Run();
 Prerequisite : Overide to make something
 Procedure    : This method should do the job.
 Args         : none

=cut

sub _Run
{
	my $self = shift;

	my $msg = "Somebody doesn't do his job well,";
	$msg   .= " tell your administrator to check error file.\n";

	my %h_args = ( -user_msg => $msg, -depth => 0.5 );
	throw LipmError::ObjectException( $self, 'abstract method', %h_args );

	return;
}


#----- Some accessors, required by the decoration pattern -----

=head2 Function Accessors

 Title        : _GetArrayResults, _GetHashResults
 Usage        : my $ra_res = $o_x->_GetArrayResults()
 Function     : accessor on __ra_results, __rh_results.
 Returns      :
     _GetArrayResults => list ref
     _GetHashResults  => hash ref
     _GetNotFound     => list ref
 Args         : none
 Error        : none

=cut

sub _GetArrayResults
{
	my $self = shift;

	return $self->_Get( qw( __ra_results ) );
}

sub _GetHashResults
{
	my $self = shift;

	return $self->_Get( qw( __rh_results ) );
}

sub _GetNotFound
{
	my $self = shift;

	return $self->_Get( qw( __ra_not_found ) );
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
1;
