#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 09, 2008
#	$Id: Accessor.pm 3688 2011-05-24 08:33:59Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package NarcisseService::Accessor;

=pod

=head1 NAME

 Accessor - a NarcisseService Class to access any feature of Narcisse instance.

=head1  SYNOPSIS

 # code this class need to be decorated to be usefull.
 my $o_ns = New NarcisseService::Accessor( species => 'mtr' );
 $o_ns    = New FromDecorator::RegionsDecorator( o_accessor => $o_ns,
                                  chrom => 5,
                                  ra_regions =>[{ start => 10, end => 100}]);
 $o_ns    = New WhatDecorator::GenesDecorator( o_accessor => $o_ns );
 my $ra_results = $o_ns->FetchResults();

=head1 DESCRIPTION

 This class derives from NarcisseService. It is specific to access features
 through the adapted Ensembl API ( Gene, Protein, Exon ... ).
 This class do nothing by itself.
 You have to use decoration, like describe in design pattern Decorator.
 Two decorations are needed, a FromDecorator which will redefine the
 _GetAPILinks method, and a WhatDecorator which will redefine the _Fetch method.

 The _Run method calls _GetAPILinks to have an access to the API, then _Fetch
 to call _StoreResults with whatever feature obtain by the API.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( NarcisseService );
use Carp;

use LipmError::ParameterException;
use NarcisseServiceKeywords qw( &NS_INFOS &NS_FASTA &NS_LOCUS );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New NarcisseService::Accessor( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of NarcisseService::Accessor class.
 Returns      : a valid NarcisseService::Accessor object
 Args         : -o_logger => Logger object
                species   => species short name
                -out_type => $type
 Attributes   :
     _o_instance_obs : InstanceObserver object

     _species        : species short name
     _o_db_adaptor   : DBAdaptor object (API)
     _out_type       : string, one of those constant &NS_INFOS &NS_FASTA &NS_LOCUS

=cut

sub _Init
{
	my $self = shift;

	# define _o_instance_obs and call TreatOptions
	$self->SUPER::_Init( @_ );

	my @a_attr = qw( _o_instance _o_species _species );
	my( $o_inst_obs, $o_species, $species ) = $self->_Get( @a_attr );

	# question a moi-meme : pourquoi ca n'est pas dans treatOptions ?
	# une piste peut-etre : avant ca utilisait instanceObserver.
	my $o_dba = $o_species->GetDBAdaptor();
	if( !$o_dba )
	{
		my $dbg = "Your species ($species) probably does not belong to ";
		$dbg   .= "the instance " . $o_inst_obs->GetName()."\n";

		# value doit etre pair pour permettre de ne pas mourir
		throw LipmError::ParameterException( 'species', 'value', -text => $dbg, -value => 02 );
	}
	$self->_Set( _o_db_adaptor => $o_dba );

	return $self;
}


=head2 Procedure

 Title        : _TreatOptions
 Usage        : $o_xx->_TreatOptions( species  => $species
                                      -out_type => $type );
 Prerequisite : called by NarcisseService::_Init
 Procedure    : set $species as attribute _species
 Args         : $species, string, short species name

=cut

sub _TreatOptions
{
	my $self   = shift;
	my %h_args = ( species  => undef,
	               -out_type => &NS_FASTA,
	               @_ );

	#to set Instance_observer
	$self->SUPER::_TreatOptions( %h_args );

	my $species = $h_args{ species };
	LipmError::ParameterException::Assert( species => $species );

	my $out_type;
	my $o_enum = New Enum( &NS_INFOS, &NS_FASTA, &NS_LOCUS );
	if( $o_enum->Includes( $h_args{ -out_type } ) )
	{
		$out_type = $h_args{ -out_type };
	}
	else
	{
		my ( $input, @a_list ) = ( $h_args{ -out_type }, $o_enum->ListItems() );
		my $msg = "$input does not belong to the list '@a_list',";
		$msg   .= " we'll use default value : " . &NS_FASTA . "\n";
		$self->_Trace( $msg, '.', 'IS' );

		$out_type = &NS_FASTA;
	}

	my $o_inst_obs = $self->_Get( '_o_instance_obs' );
	my $o_species = New SpeciesObserver( species => $species,
										 -o_inst_obs => $o_inst_obs );

	$self->_Set( _species => $species,
				 _o_species => $o_species,
	             _out_type => $out_type );

	return;
}


=head2 function

 Title        : GetSpeciesName
 Usage        : $ra_species = $o_xx->GetSpeciesName();
 Prerequisite : none
 Function     : accessor to _species attribute
 Returns      : $ra_species
 Args         : none

=cut

sub GetSpeciesName
{
	my $self      = shift;

	return $self->_Get( '_species' );
}


=head2 Procedure

 Title        : _Run
 Usage        : $o_xx->_Run();
 Prerequisite : Called by NarcisseService::FetchResults
 Procedure    : just call _GetAPILinks to get a link to the API
                then _Fetch, to get what is wanted.
 Args         : none

=cut

sub _Run
{
	my $self = shift;

	my $rao_links = $self->_GetAPILinks();
	$self->_Fetch( links => $rao_links );

	return;
}


=head2 function

 Title        : _GetAPILinks
 Usage        : $rao_links = $o_xx->_GetAPILinks();
 Prerequisite : should be overridden in a decoration FromDecorator
 Function     : this methods return a ref list of object that are entries point
                of the API.
 Returns      : $rao_links, a ref list of anything.
 Args         : none

=cut

sub _GetAPILinks
{
	my $self = shift;

	my $dbg = "_GetAPILinks should be overridden by a Decorator.\n";
	my $msg = "Somebody doesn't do his job well,";
	$msg   .= " tell your administrator to check error file.\n";

	my %h_args = ( -text => $dbg, -user_msg => $msg, -depth => 0.5 );
	throw LipmError::ObjectException( $self, 'abstract method', %h_args );

	my $rao_links = undef;

	return $rao_links;
}


=head2 procedure

 Title        : _Fetch
 Usage        : $whatever = $o_xx->_Fetch( links => $rao_links );
 Prerequisite : should be overridden in a decoration WhatDecorator
 Procedure    : this method should use _StoreResults.
 Args         : none

=cut

sub _Fetch
{
	my $self = shift;
	my %h_args = ( links => undef,
	               @_ );

	my $dbg = "You try to use a service without enough decorations.\n";
	my $msg = "Somebody don't do his job well, tell him to check error file.\n";

	my %h_params = ( -text => $dbg, -user_msg => $msg );
	throw LipmError::ObjectException( $self, 'abstract method', %h_params );

	return;
}


=head2 function

 Title        : _GetHeader
 Usage        : $header = $o_xx->_GetHeader( $o_x, $locus_tag );
 Prerequisite : none
 Function     : build and return a header like defined in Narcisse Project
                http://lipm-grpware.toulouse.inra.fr/cgi-bin/trac/narcisse/trac.cgi/wiki/FileFormats
 Returns      : $header, a string
 Args         : $o_x, a Bio::EnsEMBL::x object
                $locus_tag (optional)
                MANU 24 05 11 - Added $locus_tag to the calls of _GetFeatureId - 
                                Will be used by some methods only (GetUpstream, GetDownstream)

=cut

sub _GetHeader
{
	my $self   = shift;
	my( $o_feature, $locus_tag ) = @_;

	my $o_dba = $o_feature->adaptor()->db();

	my ( $start, $end, $length, $strand, $species, $chrom );
	$length   = $o_feature->length();
	$species  = $o_dba->conf()->GetLongNameWithoutSpace();

	if( !$o_feature->isa('Bio::EnsEMBL::Translation') )
	{
		$start     = $o_feature->start();
		$end       = $o_feature->end();
		$strand    = $o_feature->strand();

		$chrom     = $o_feature->seq_region_name();
	}

	my ( $type, $parent, $id );
	if( $o_feature->isa( 'Bio::EnsEMBL::Slice' ) )
	{
		$type   = 'dna';
		$id     = $self->_GetFeatureId( $o_feature, $locus_tag );
		$parent = $chrom;
# warn Dumper [ keys %{ $self->_Get( '_o_accessor' )->_Get( '_o_accessor' ) } ];
	}
	elsif( $o_feature->isa('Bio::EnsEMBL::Gene') )
	{
		$type   = 'gene';
		$parent = $chrom;
		$id     = $self->_GetFeatureId( $o_feature, $locus_tag ); # accession ou locus tag ?
	}
	elsif( $o_feature->isa( 'Bio::EnsEMBL::Transcript' ) )
	{
		$type   = 'transcript';
		$id     = $self->_GetFeatureId( $o_feature, $locus_tag ); # accession ou locus tag ?
		$parent = $id;	# ou accession/locus_tag du gene
# $gene_adaptor->fetch_by_transcript_stable_id();
	}
	elsif( $o_feature->isa('Bio::EnsEMBL::Translation') )
	{
		$type    = 'protein';

		$id     = $self->_GetFeatureId( $o_feature, $locus_tag ); # accession ou locus tag ?
# *		$length  = $o_feature->length();
# *		$species = $o_dba->conf()->GetLongNameWithoutSpace();

		$start   = $o_feature->genomic_start();
		$end     = $o_feature->genomic_end();

		$chrom   = $o_feature->start_Exon()->slice()->seq_region_name();

		# l'API oblige a passer par le transcript pour avoir les infos du gene
		# on aurait pu passer par le gene, mais nos fichiers ne lient pas le gene a la proteine
		my $o_tr_a      = $o_dba->get_TranscriptAdaptor();
		my $o_transcript= $o_tr_a->fetch_by_translation_stable_id( $id );

		$parent = $o_transcript->stable_id();
		$strand    = $o_transcript->strand();
	}

	my $header = '>nc|';
	my $more   = $self->_GetMoreHeaderData(); # en prevision

	$header .= "$type|$id parent=$parent chrom=$chrom";
	$header .= " range=$start-$end length=$length strand=$strand  sp=$species";
	$header .= " $more" if( defined $more );

	return $header;
}


=head2 function _GetMoreHeaderData

 Usage        : $more = $o_xx->_GetMoreHeaderData();
 Prerequisite : none
 Function     : return a string that will be added at the end of header sequence.
 Returns      : $more, string, anything or undef
 Args         : none

=cut

sub _GetMoreHeaderData
{
	my $self = shift;

	return undef;
}


=head2 function _GetFeatureId

 Usage        : $id = $o_xx->_GetFeatureId( $o_feature );
 Prerequisite : none
 Function     : return an Id that describes the feature.
                I use a function to allow overloading by SliceDecorator.
 Returns      : $id, string, id of $o_feature
 Args         : $o_feature, a Bio::EnsEMBL::x object

=cut

sub _GetFeatureId
{
	my $self = shift;
	my( $o_feature ) = @_;

	if( $o_feature->can( 'stable_id' ) )
	{
		my $id = $o_feature->stable_id();
		return $id;
	}

	# lancer une exception ?
	my $msg = "o_feature doesn't have *stable_id* method.\n";
	&Carp::cluck( $msg . "o_feature is a " . ref( $o_feature ) );
	
# MANU - Quand on demande une proteine a partir de l'interface graphique
#        $o_feature est un Bio::EnsEMBL::translation
#        => j'ai essaye de tester isa( 'Bio::EnsEMBL' ) mais ca ne marche pas, ce qui est zarbi
#           je vire le test, c'est comme ca dans la version 2.0, ca arrange mon pb.... mais peut-etre que ca en cree d'autres ?
# Seb - Bio::EnsEMBL::translation n'est pas un objet 'Bio::EnsEMBL::Feature',
#	il derive de 'Bio::EnsEMBL::Storable'.
#	il n'existe pas d'objet 'Bio::EnsEMBL'.
#	J'ai donc changé pour un test avec 'can'.
#	-> je tiens au test car j'ai le souvenir (vague il est vrai) d'avoir perdu du temps
#		a chercher une erreur ici.
	return;
}

#----- Quelques accesseurs necessaires pour la decoration -----

=head2 function _GetDBAdaptor

 Usage        : $o_dba = $o_xx->_GetDBAdaptor();
 Prerequisite : none
 Function     : accessor of _o_db_adaptor attribute
 Returns      : $o_dba, an API DB_Adaptor object.
 Args         : none

=cut

sub _GetDBAdaptor
{
	my $self      = shift;

	return $self->_Get( '_o_db_adaptor' );
}


=head2 function _GetOutType

 Usage        : $o_dba = $o_xx->_GetOutType();
 Prerequisite : none
 Function     : accessor of attribute _out_type
 Returns      : value of out_type
 Args         : none

=cut

sub _GetOutType
{
	my $self = shift;

	return $self->_Get( '_out_type' );
}


=head2 function GetOutType

 Usage        : $o_dba = $o_xx->GetOutType();
 Prerequisite : A valider par les responsables.
 Function     : accessor of attribute _out_type
 Returns      : value of out_type
 Args         : none

=cut

sub GetOutType
{
	my $self = shift;

	return $self->_Get( '_out_type' );
}


#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 15, 2008
#	$Id: Accessor.pm 3688 2011-05-24 08:33:59Z manu $
#

package AccessorDecorator;

=pod

=head1 NAME

 AccessorDecorator - mother class of decorator for Accessor

=head1  SYNOPSIS

 # code see WhatDecorator or FromDecorator
 # this class should only be inherited

=head1 DESCRIPTION

 Redefine _Init method so that all attributes of Accessor are accesible.
 This is an implementatin of the design pattern Decorator
 http://en.wikipedia.org/wiki/Decorator_pattern

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( NarcisseService::Accessor );


=head2 method _Init

 Title        : _Init
 Usage        : $self->SUPER::_Init( o_accessor => $o_accessor );
 Prerequisite : do not use directly, call it from a derived class.
 Fonction     : initialise an instance of AccessorDecorator class.
 Returns      : a valid AccessorDecorator object
 Args         : $o_accessor, an Accessor object
 Attributes   :
     _o_instance_obs : InstanceObserver object

     _species        : species short name
     _o_db_adaptor   : DBAdaptor object (API)

     _o_accessor     : NarcisseService::Accessor object.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( o_accessor => undef,
	               @_ );

	my $o_accessor = $h_args{ o_accessor };
	LipmError::ParameterException::Assert( 'o_accessor', $o_accessor );

	if( !$o_accessor->isa( 'NarcisseService::Accessor' ) )
	{
		throw LipmError::ObjectException( $self, '!build' );
	}

# 	# copie de tous les attributs de la mere
# 	# pour etre propre il faudrait redefinir seulement les accesseurs
# 	my @a_attributes = keys %$o_accessor;
# 	@$self{ @a_attributes } = @$o_accessor{ @a_attributes };

	# on ecrase au besoin la valeur de _o_accessor
	$self->_Set( _o_accessor => $o_accessor );
	$self->SetLogger( $h_args{ -o_logger } );

	my $msg = "A decoration have been created.\n";
	$self->_Trace( $msg, '.', 'ISV' );

	return $self;
}

#----- Redefinition des accesseurs -----

=head2 function Accessors

 Title        : _GetDBAdaptor, _GetOutType,
                _GetArrayResults, _GetHashResults, _GetNotFound
 Usage        : $o_dba = $o_xx->_GetDBAdaptor();
 Prerequisite : needed for Decoration
 Function     : accessor, just return attribute value
 Returns      :
     GetSpeciesName=> string

     _GetDBAdaptor => an API DB_Adaptor object.
     _GetOutType   => string

     _GetArrayResults => list ref
     _GetHashResults  => hash ref
     _GetNotFound     => list ref
 Args         : none

=cut

sub GetSpeciesName
{
	my $self      = shift;

	return $self->_Get( '_o_accessor' )->GetSpeciesName();
}


sub _GetDBAdaptor
{
	my $self      = shift;

	return $self->_Get( '_o_accessor' )->_GetDBAdaptor();
}

sub _GetOutType
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->_GetOutType();
}

sub GetOutType
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->GetOutType();
}

sub _GetArrayResults
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->_GetArrayResults();
}

sub _GetHashResults
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->_GetHashResults();
}

sub _GetNotFound
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->_GetNotFound();
}


1;
