#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 17, 2008
#	$Id: IdsDecorator.pm 3695 2011-07-22 14:30:15Z sletort $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package FromDecorator::IdsDecorator;

=pod

=head1 NAME

 FromDecorator::IdsDecorator - a FromDecorator to access API through ids like locus tags or protein ids.

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given ids.
 This class is abstract as the _FetchGene is not implemented.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New FromDecorator::X(
                                       o_accessor => $o_accessor,
                                       ra_ids     => $ra_ids,
                                       feature    => $feature
                                      );
                X can be LocusTags, TranscriptLocusTags, ProteinIds.
 Prerequisite : none
 Fonction     : Assert that args are ok.
                Then call grand-ma _Init (AccessorDecorator).
                and last set specific attributes.
 Returns      : a valid FromDecorator::IdsDecorator object
 Args         : $o_accessor, a NarcisseService::Accessor object
                $ra_ids, a ref list of ids.
                $feature, string, tell what kind of Adaptor you need
                (Gene, Transcript)
 Attributes   :
     __o_logger       : Logger object

#      _o_instance_obs : InstanceObserver object
# 
#      _species        : species short name
#      _o_db_adaptor   : DBAdaptor object (API)
#      _ra_results     : array ref that contains any results.
# 
#      _ra_tags        : ref list of locus tag ids (string) OR of protein ids (cf GetGenesFromProteinIds)
#      _feature        : string, the Feature you have locus tag ids.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( ra_ids => undef,
	               feature => undef,
	               @_ );
	# warn Dumper \%h_args;

	my $ra_ids = $h_args{ ra_ids };
	$ra_ids = $h_args{ ra_tags } if( !defined $ra_ids );# pour retro-compatibilité

	LipmError::ParameterException::Assert( 'ra_ids', $ra_ids );
	throw LipmError( -text => "Array of ids" )	if( ref( $ra_ids ) ne 'ARRAY' );

	my $feature = $h_args{ feature };
	my $o_enum = New Enum( qw( Gene Transcript ) );
	LipmError::ParameterException::Assert( 'feature', $feature,
	                                       -o_enum => $o_enum );

	$self->SUPER::_Init( %h_args );

	$self->_Set( _ra_ids => $ra_ids,
	             _feature => $feature,
				 
	# ajoute pour permettre de faire un lien entre l'id en entree et la sortie
	# avant on avait une liste en entree, une liste en sortie,
	# mais sans possibilite de relier les 2
	             _ra_found_ids => [] );

	return $self;
}


=head2 function _GetAPILinks

 Title        : _GetAPILinks
 Usage        : $o_bfa = $o_xx->_GetAPILinks();
 Prerequisite : none
 Function     : get the API DB_Adaptor, then return the BaseFeatureAdaptor.
 Returns      : $o_bfa, a BaseFeatureAdaptor.
 Args         : none

=cut

sub _GetAPILinks
{
	my $self = shift;

	my @ao_genes = ();

	my @a_attrs = qw( _ra_ids _ra_found_ids );
	my( $ra_ids, $ra_founds ) = $self->_Get( @a_attrs );

	foreach my $id ( @$ra_ids )
	{
		# _FetchGene is not a good name. list ref is updated.
		my $bool = $self->_FetchGene( $id, \@ao_genes );
		if( 1 == $bool )
		{
			push( @$ra_founds, $id );
			next; # on s'arrete au premier trouve.
		}

        # Written for medicago locus_tags : if AC3234.4 not found, we search AC3234
		if( $id =~ /.+\.\d+$/ )
		{
			my $altered_lt = $id;
			$altered_lt =~ s/\.\d+$//;
			$bool = $self->_FetchGene( $altered_lt, \@ao_genes );
			next  if( 1 == $bool );
		}

		$self->_StoreNotFound( $id );
	}

	return \@ao_genes;
}


=head2 function _FetchGene

 Title        : _FetchGene
 Usage        : $bool = $o_xx->_FetchGene( $locus_tag, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $locus_tags, $ra_return

=cut

# sub _FetchGene
# {
# 	my $self = shift;
# 	my( $locus_tag, $ra_return ) = @_;
# 
# 
# 	return 0;
# }

1;

package FromDecorator::LocusTagsDecorator;

=pod

=head1 NAME

 FromDecorator::LocusTagsDecorator - a FromDecorator to access API through Locus tags as ids.

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given LocusTag ids.
 You build this decoration with a list of locus tags.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

 Note : The correct name of the package should be FromDecorator::IdsDecorator::LocusTagsDecorator
        but for convenient use, IdsDecorator is omitted.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator::IdsDecorator );


=head2 function _FetchGene

 Title        : _FetchGene
 Usage        : $bool = $o_xx->_FetchGene( $locus_tag, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $locus_tags, $ra_return

=cut

sub _FetchGene
{
	my $self = shift;
	my( $locus_tag, $ra_return ) = @_;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_gene_adpt = $o_dba->get_GeneAdaptor();

	my $o_gene = $o_gene_adpt->fetch_by_stable_id( $locus_tag );
	if( defined $o_gene )
	{
		push( @$ra_return, $o_gene );
		return 1;
	}

	return 0;
}

1;

#------------------------------------

package FromDecorator::TranscriptLocusTagsDecorator;

=pod

=head1 NAME

 FromDecorator::TranscriptLocusTagsDecorator - a FromDecorator to access API through transcripts Locus tags as ids.

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given transcripts LocusTag ids.
 You build this decoration with a list of locus tags.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator::IdsDecorator );


=head2 function _FetchGene

 Title        : _FetchGene
 Usage        : $bool = $o_xx->_FetchGene( $locus_tag, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $locus_tags, $ra_return

=cut

sub _FetchGene
{
	my $self = shift;
	my( $locus_tag, $ra_return ) = @_;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_gene_adpt = $o_dba->get_GeneAdaptor();

    my $o_gene = $o_gene_adpt->fetch_by_translation_stable_id( $locus_tag );
	if( defined $o_gene )
	{
		push( @$ra_return, $o_gene );
		return 1;
	}

	return 0;
}

1;

# -----------------------------------

package FromDecorator::ProteinIdsDecorator;

=pod

=head1 NAME

 FromDecorator::ProteinIdsDecorator - a FromDecorator to access API through protein ids (or translation_stable_id for the API)

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given protein ids
 which are also transcript stable ids.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator::IdsDecorator );


=head2 function _FetchGene

 Usage        : $bool = $o_xx->_FetchGene( $protein_id, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $protein_id, $ra_return

=cut

sub _FetchGene
{
	my $self = shift;
	my( $protein_id, $ra_return ) = @_;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_gene_adpt = $o_dba->get_GeneAdaptor();


    my $o_gene = $o_gene_adpt->fetch_by_translation_stable_id( $protein_id );
	if( defined $o_gene )
	{
		push( @$ra_return, $o_gene );
		return 1;
	}

	return 0;
}

1;
# -----------------------------------

package FromDecorator::SymbolsDecorator;

=pod

=head1 NAME

 FromDecorator::SymbolsDecorator - a FromDecorator to access API through symbol.

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given symbols.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator::IdsDecorator );


=head2 function _FetchGene

 Usage        : $bool = $o_xx->_FetchGene( $symbol, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $symbol, $ra_return

=cut

sub _FetchGene
{
	my $self = shift;
	my( $symbol, $ra_return ) = @_;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_gene_adpt = $o_dba->get_GeneAdaptor();


    my $o_gene = $o_gene_adpt->fetch_by_symbol( $symbol );
	if( defined $o_gene )
	{
		push( @$ra_return, $o_gene );
		return 1;
	}

	return 0;
}

1;
# -----------------------------------

package FromDecorator::AccessionNumbersDecorator;

=pod

=head1 NAME

 FromDecorator::AccessionNumbersDecorator - a FromDecorator to access API through accession number (or xxx for the API)

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API given accession number.
 The links returned by _GetAPILinks will be a BaseFeatureAdaptor.

=cut

use strict;
use warnings;

use Data::Dumper;

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use Enum;


use base qw( FromDecorator::IdsDecorator );


=head2 function _FetchGene

 Usage        : $bool = $o_xx->_FetchGene( $accession, $ra_return );
 Prerequisite : none
 Function     : none
 Returns      : $bool
 Args         : $accession, $ra_return

=cut

sub _FetchGene
{
	my $self = shift;
	my( $accession, $ra_return ) = @_;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_gene_adpt = $o_dba->get_GeneAdaptor();


    my $o_gene = $o_gene_adpt->fetch_by_transcript_accession( $accession );
	if( defined $o_gene )
	{
		push( @$ra_return, $o_gene );
		return 1;
	}

	return 0;
}

1;
