#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 12, 2008
#	$Id: RegionsDecorator.pm 3185 2010-01-20 12:55:30Z sletort $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package FromDecorator::RegionsDecorator;

=pod

=head1 NAME

 FromDecorator::RegionsDecorator - a FromDecorator to access API through slices.

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a FromDecorator, specific to access API through Slice object.
 You build this decoration with a list of start-end region and a chromosome,
 The links returned by _GetAPILinks will be a list of slice object that
 represents a slice of the chromosome sequence.

=cut

use strict;
use warnings;

use Data::Dumper;
use Error qw( :try );

# chemin d'access a FromDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclur dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::FromDecorator;
use LipmError;
use base qw( FromDecorator );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New FromDecorator::RegionsDecorator(
                                       o_accessor => $o_accessor,
                                       chrom      => $chrom,
                                       ra_regions => $ra_regions
                                      );
 Prerequisite : none
 Fonction     : Assert that args are ok.
                Then call grand-ma _Init (AccessorDecorator).
                and last set specific attributes.
 Returns      : a valid FromDecorator::RegionsDecorator object
 Args         : $o_accessor, a NarcisseService::Accessor object
                $chrom, string, chrom label
                $ra_region, a ref list of ref hashes with start and end keys.
 Attributes   :
     __o_logger       : Logger object

     _o_instance_obs : InstanceObserver object

     _species        : species short name
     _o_db_adaptor   : DBAdaptor object (API)
     _ra_results     : array ref that contains any results.

     _chrom          : chromosome label
     _ra_regions     : ref list of ref hashes of start-end values

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( chrom      => undef,
	               ra_regions => undef,
	               @_ );

	my $chrom = $h_args{ chrom };
	LipmError::ParameterException::Assert( 'chrom', $chrom );

	my $ra_regions = $h_args{ ra_regions };
	LipmError::ParameterException::Assert( 'ra_regions', $ra_regions );
	throw LipmError( -text => "Array for regions" )	if( ref( $ra_regions ) ne 'ARRAY' );

	$self->SUPER::_Init( %h_args );

	$self->{ _chrom } = $chrom;
	$self->{ _ra_regions } = $ra_regions;

	return $self;
}


=head2 function

 Title        : _GetAPILinks
 Usage        : $rao_links = $o_xx->_GetAPILinks();
 Prerequisite : none
 Function     : get the API DB_Adaptor, then for each region of _ra_regions
                build a slice object and return them as a ref list.
 Returns      : $rao_links, a ref list of API Slice objects.
 Args         : none

=cut

sub _GetAPILinks
{
	my $self      = shift;

	my $o_dba = $self->_GetDBAdaptor();
	my $o_sla = $o_dba->get_SliceAdaptor();
	if( !$o_sla )
	{
		my $species = $self->GetSpeciesName();
		my $msg = "no slice adaptor has been generated for $species\n";
		throw LipmError( -text => $msg );
	}

	my $chr = $self->_Get( qw( _chrom ) );

	my @ao_links = ();

	# $ra_regions est une ref de liste contenant
	# des ref de hash style { start => x, end => y }
	my $ra_regions = $self->_Get( qw( _ra_regions ) );
	foreach my $rh_region ( @$ra_regions )
	{
		# j'ajoute un try-catch car si on precise un mauvais chrom, on peut passer au suivant.
		# ou bien dans le cas du multi espece, cael a une notatino en chiffre romain.
		try
		{
			my( $start, $end ) = @$rh_region{ qw( start end ) };
			my @a_options = ( $chr, $start, $end );

			my $o_slice   = $o_sla->fetch_by_region( 'chromosome', @a_options );

			if( !$o_slice )
			{
				warn Dumper \@a_options;
			}

			push( @ao_links, $o_slice );

#		on pourrait nommer la region, pour le moment utilite seulement dans legoo
#		donc je ne le fait pas, je suis presse !
# 		my $name = $$rh_region{ name } || "$start..$end";
#		push( @ao_links, { $name => $o_slice } );
		}
		catch LipmError with
		{
			my( $o_err, $r ) = @_;
			$self->_Trace( $o_err->Explain(), '.', 'IS' );

			$r = 1; # pour poursuivre
		};
	}

	return \@ao_links;
}


1;
