#
# 	sebastien.letort@toulouse.inra.fr
#	Created: December 12, 2008
#	$Id: WhatDecorator.pm 3695 2011-07-22 14:30:15Z sletort $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package WhatDecorator;

=pod

=head1 NAME

 WhatDecorator - one type of AccessorDecorator

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 This class should be inherited.
 It redefines _Fetch, the method that give a way to enter the API.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( AccessorDecorator );

use NarcisseServiceKeywords qw( &NS_INFOS &NS_FASTA &NS_LOCUS );


=head2 function

 Title        : _GetAPILinks
 Usage        : $rao_links = $o_xx->_GetAPILinks();
 Prerequisite : Design pattern Decorator
 Function     : just call _GetAPILinks of the inner object.
 Returns      : $rao_links
 Args         : none

=cut

sub _GetAPILinks
{
	my $self = shift;

	return $self->_Get( '_o_accessor' )->_GetAPILinks( @_ );
}


=head2 procedure

 Title        : _Fetch
 Usage        : $o_xx->_Fetch();
 Prerequisite : none
 Procedure    : By defining this methods I assure that it is not the mother one
                that is called.
 Args         : none

=cut

sub _Fetch
{
	my $self = shift;
	my %h_args = ( links => undef,
	               @_ );

	my $rao_links = $h_args{ links };

	# I add a parameter to the _FetchX methods
	# so that a link can be made between input and output.
	for( my $i=0; $i<@$rao_links; ++$i )
	{
		my $o_link = $$rao_links[$i];

		if( $o_link->isa( 'Bio::EnsEMBL::Slice' ) )
		{
			$self->_FetchFeaturesFromSlice( $o_link, $i+1 );
		}
		elsif( $o_link->isa( 'Bio::EnsEMBL::Gene' ) )
		{
			$self->_FetchFeaturesFromGene( $o_link, $i+1 );
		}
		else
		{
			my $msg = "Your query can't be achieved due to a lack of programmation\n";
			my $dbg = "Tried to fetch genes with -" . ref( $o_link );
			$dbg   .= "- no function defined for that\n";
			throw LipmError( -text => $dbg, -user_msg => $msg );
		}
	}

	return;
}


=head2 _StoreFeature

 Usage        : $o_x->_StoreFeature( $o_feature )
                for up/downstream $o_x->_StoreFeature( $o_feature, $locus_tags )
 Function     : .
 Args         :
     $class,
 Error        : none.
 Note       : $locus_tag was only added to put into the header of the sequence.
              This is required for Up/Downstream _class redefine _GetHeader.
              For other class is will be undef.

=cut

sub _StoreFeature
{
	my $self   = shift;
	my( $o_feature, $locus_tag ) = @_;

	LipmError::ParameterException::Assert( 'o_feature' => $o_feature );

	my $out_type = $self->_GetOutType();
	if( $out_type eq &NS_LOCUS )
	{
		if( !$o_feature->isa( 'Bio::EnsEMBL::Slice') )
		{
			$self->_StoreResults( $o_feature->stable_id() );
		}
		else
		{
		# faire un fetch all gene, puis sur chaque un stable_id
		# pas fait pour le moment, on a deja GetGenesFromRegions
			1;
		}
	}
	else
	{
		my $header = $self->_GetHeader( $o_feature, $locus_tag );

		if( $out_type eq &NS_INFOS )
		{
			$self->_StoreResults( $header );
		}
		else # &NS_FASTA => on veut la sequence
		{
			my $seq = $self->_GetFeatureSeq( $o_feature );
			$self->_StoreResults( { $header => $seq } );
		}
	}


	return;
}


=head2 Function _GetFeatureSeq

 Usage        : my $seq = $o_x->_GetFeatureSeq( $o_feature )
 Function     : Return the sequence corresponding to the feature.
 Returns      : $seq, string
 Args         :
     $o_feature, EnsEMBL object (Slice, Gene, Protein, or Transcript)
 Error        : none

=cut

sub _GetFeatureSeq
{
	my $self = shift;
	my( $o_feature ) = @_;

	return $o_feature->seq();
}


sub _FetchFeaturesFromSlice
{
	my $self = shift;
	throw LipmError::ObjectException( $self, 'abstract method' );
}


sub _FetchFeaturesFromGene
{
	my $self = shift;
	throw LipmError::ObjectException( $self, 'abstract method' );
}

1;
