#
# 	sebastien.letort@inserm.fr
#	Created: April 20, 2010
#	$Id: SynteniesDecorator.pm 3743 2011-11-20 13:48:22Z sletort $
#

package WhatDecorator::SynteniesDecorator;

=pod

=head1 NAME

 WhatDecorator::SynteniesDecorator - a WhatDecorator to fetch synteny

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 NarcisseService::Accessor needs a FromDecorator and a WhatDecorator to be
 usable. This is a WhatDecorator, specific to access sequences (Slice) through API.
 What is found by _Fetch is stored with _StoreResults.
 Results are NarcisseData::HashReturn objects.

=cut

use strict;
use warnings;

use Data::Dumper;
use File::Basename;

use Error qw( :try );

use Enum;
use NARCISSE_utils;

# chemin d'access a WhatDecorator depuis lib/
# 'use' obligatoire pour avoir plusieurs paquets dans le meme fichier.
# si on ne le met pas ici,il faut le mettre dans l'executable
# je trouve plus propre de n'inclure dans l'executable que les paquets
# que l'on instancie.
use NarcisseService::Accessor::WhatDecorator;
use base qw( WhatDecorator );

use constant DEF_LVL => 0;


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New WhatDecorator::SynteniesDecorator(
                                       o_accessor => $o_accessor,
                                       -level     => $level,
									   prog       => $prog
									   target     => $target
                                      );
 Prerequisite : none
 Fonction     : call grand-ma _Init (AccessorDecorator).
                set the specific attribute _level and target.
 Returns      : a valid NarcisseService::Accessor object
 Args         : $o_accessor, a NarcisseService::Accessor object
                $level, the synteny level.
				$target, the target organism.
				$prog, the program used to compute the syntenies.
 Attributes   :
     _o_instance_obs : InstanceObserver object
     _species        : species short name
     _o_db_adaptor   : DBAdaptor object (API)
     _ra_results     : array ref that contains any results.

     __level         : the level of the synteny
	 __target        : the organism short name to use as target.
	 __prog          : the program used to compute the syntenies.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = (  -level => &DEF_LVL,
					target => undef,
					prog   => undef,
	               @_ );

	$self->SUPER::_Init( %h_args );

	$self->_Set( __level  => $h_args{ -level } );
	$self->_Set( __target => $h_args{ target } );
	$self->_Set( __prog   => $h_args{ prog } );

	return;
}


=head2 procedure

 Title        : _Fetch
 Usage        : $o_xx->_Fetch( links => $rao_links );
 Prerequisite : should populate _ra_results attribut.
 Procedure    : Foreach link, fetch synteny through a method that is dependant of
               the object link. *only FromSlices implemented*
 Args         : $rao_links, list ref of link objects.

=cut

sub _Fetch
{
	my $self = shift;
	my %h_args = ( links => undef,
	               @_ );

	my $rao_links = $h_args{ links };
	foreach my $o_link ( @$rao_links )
	{
		if( !defined $o_link )
		{
			$self->_StoreResults( undef );
		}
		elsif( $o_link->isa( 'Bio::EnsEMBL::Slice' ) )
		{
			$self->__FetchSyntenyFromSlices( $o_link );
		}
		else
		{
			my $msg = "Your query can't be achieved due to a lack of programmation\n";
			my $dbg = "Tried to fetch synteny with -" . ref( $o_link );
			$dbg   .= "- no function defined for that\n";
			throw LipmError( -text => $dbg, -user_msg => $msg );
		}
	}

	return;
}


=head2 Procedure

 Title        : __FetchSyntenyFromSlices
 Usage        : $o_xx->__FetchSyntenyFromSlices( $o_slice );
 Prerequisite : none
 Procedure    : none
 Args         : $o_slice, a EnsEMBL::Slice object.

=cut

sub __FetchSyntenyFromSlices
{
	my $self = shift;
	my ( $o_slice ) = @_;

	# range
	my $chrom = $o_slice->seq_region_name();
	my $start = $o_slice->start();
	my $end   = $o_slice->end();
	my @a_ref_region = ( $chrom, $start, $end );

	my $o_syntenies;
	try
	{
		my $filepath = $self->__GetSynteniesFilepath( $chrom );
		# warn $filepath;

		my %h_args = (	file  => $filepath,
						start => $start,
						end   => $end,
	#					return_type => 'ARRAY',
						return_type => 'HASH',
						use_cache   => 0 );
		# $o_syntenies is a Bio::Narcisse::Utils::NarcisseData object.
		$o_syntenies = &NARCISSE_utils::ReadChromFeatureData( %h_args );
	}
	catch LipmError::IOException with
	{
		my( $o_err, $r ) = @_;

		my $file = $o_err->file();
		if( $file =~ /SpeciesObserver.pm$/ )
		{
			my $pb_file = fileparse( $o_err->{ -handler } );
			my $keyword = $o_err->{ -keyword };
			my $expl    = $o_err->Explain();
			if( $pb_file =~ /.synt$/ && $keyword eq "!exist" )
			{
				warn "There's no syntenie for the region $chrom:$start..$end.\n";
			}
			else
			{	$r = 1; }
		}
		else
		{	$r = 1;	} # to continue
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		my $msg = "Error in SynteniesDecorator while fetching syntenies from slices.\n";
		$msg   .= "No synteny will be retrieved for your region $chrom:$start..$end.\n";
		$msg   .= $o_err->Explain() . "\n";
		warn $msg;
	};

	# finally we store the object or undef;
	$self->_StoreResults( $o_syntenies );
	
	return;
}


sub __GetSynteniesFilepath
{
	my $self = shift;
	my( $chrom ) = @_;

	my @a_attr = qw( __level __prog __target );
	my( $lvl, $prog, $tgt ) = $self->_Get( @a_attr );
	
	my $o_species = $self->_GetDBAdaptor()->conf();

	# assert Method
	my @a_progs = $o_species->GetMethods( target => $tgt );
	if( $prog eq 'auto' )
	{
		$prog = $a_progs[0];
	}
	else
	{
		my $o_methods = New Enum( @a_progs );
		LipmError::ParameterException::Assert(
				'prog' => $prog, -enum => $o_methods );
	}

	# get synt file
	my %h_args    = ( prog   => $prog,
					  level  => $lvl,
					  target => $tgt,
					  -chrom => $chrom );
	my $filepath  = $o_species->GetSyntFiles( %h_args );
	
	return $filepath;
}

1;
