#
# 	sebastien.letort@toulouse.inra.fr
#	Created: March 03, 2009
#	$Id$
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package NarcisseService::AccessorWithPrimerSearchInput;

=pod

=head1 NAME

 AccessorWithPrimerSearchInput - a NarcisseService Class to access any feature found into region identify by primerSearch.

=head1  SYNOPSIS

 my $o_ns = New NarcisseService::AccessorWithPrimerSearchInput( file => $file,
                                                                -out_type => $type );
 my $ra_results = $o_ns->FetchResults();

=head1 DESCRIPTION

 This class derives from NarcisseService. It is specific to access features
 found into regions described by PrimerSearch file.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( NarcisseService );

use LipmError::ParameterException;
use NarcisseServiceFactory;
use NarcisseServiceKeywords qw( &NS_INFOS &NS_FASTA &NS_LOCUS );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New NarcisseService::AccessorWithPrimerSearchInput( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of NarcisseService::AccessorWithPrimerSearchInput class.
 Returns      : a valid NarcisseService::AccessorWithPrimerSearchInput object
 Args         : -o_logger => Logger object
                -o_instance_obs => InstanceObserver object,
                file      => string, PrimerSearch file
                -what      => string 'locus' or 'sequence'
                     locus will retrieve all genes locus tags
                     sequence will retrieve regions sequence.
#                 -out_type => NarcisseServiceKeywords ( NS_LOCUS )
 Attributes   :
     _o_instance_obs : InstanceObserver object

     __fh_file          : PrimerSearch file handler
     __what             : string, 'locus' or 'sequence' depending on what we want to retrieve
#      _o_db_adaptor   : DBAdaptor object (API)
#      _out_type       : string, one of those constant &NS_FASTA &NS_LOCUS

=cut

sub _Init
{
	my $self = shift;

	# define _o_instance_obs and call TreatOptions
	$self->SUPER::_Init( @_ );

	$self->__ExtractRegionsFromPrimerSearchFile();

	return;
}


=head2 Procedure

 Title        : _TreatOptions
 Usage        : $o_xx->_TreatOptions( file     => $file
                                      -what    => $what );
 Prerequisite : called by NarcisseService::_Init
 Procedure    : set $species as attribute _species
 Args         : $file => string, PrimerSearch file
                $what => string, locus or sequence

=cut

sub _TreatOptions
{
	my $self   = shift;
	my %h_args = ( file  => undef,
# 	               -out_type => &NS_LOCUS,
	               -what  => 'locus',
# 	               -o_instance_obs => undef,
	               @_ );

	# to set _o_instance_observer
	$self->SUPER::_TreatOptions( %h_args );

	my $file = $h_args{ file };
	LipmError::ParameterException::Assert( file => $file );

	my $fh_file = new IO::File( $file, 'r' )
		or throw LipmError::IOException( $file, 'f' );

	my @a_vals = qw( locus sequence );
	my $what = $self->AssertOptions( \@a_vals, $h_args{ -what } );

# 	@a_vals = ( &NS_LOCUS, &NS_FASTA );
# 	my $type = $self->AssertOptions( \@a_vals, $h_args{ -out_type } );

	$self->_Set( __fh_file => $fh_file,
# 	             _out_type => $out_type,
	             __what    => $what );

	return;
}


=head2 __ExtractRegionsFromPrimerSearchFile

 Usage        : $o_x->__ExtractRegionsFromPrimerSearchFile()
 Function     : read PrimerSearch file and reformat regions informations.
 Args         : none
 Error        : none.

 Note         : Si on passe aux regions nommees, il faudra revoir la structure.

=cut

sub __ExtractRegionsFromPrimerSearchFile
{
	my $self   = shift;

	my %h_regions = ();

	my $fh_in = $self->_Get( qw( __fh_file ) );
	while( my $line = <$fh_in> )
	{
		next if( $line =~ /^#/ );
		chomp( $line );

		my( $primer, $str_regions ) = split( /\t/, $line );
		next if( !defined $primer );

		my @a_nc_regions = split( /,/, $str_regions );

		foreach my $nc_region ( @a_nc_regions )
		{
			my( $species, $chrom, $start, $end ) = split( /:/, $nc_region );

# 		my( $primer ) = $line =~ /^(.+?)\t/;
# 
# 		# .+? pour avoir le plus petit modele possible
# 		my $regexp = '\t(.+?)_(.+?)-(\d+)-(\d+)-(\d+)-\d+-\d+';
# 		while(  $line =~ /$regexp/gi )
# 		{
# 			# $len est pour le DBG
# 			my( $species, $chrom, $start, $end, $len ) = ( $1, $2, $3, $4, $5 );
# # print Dumper [ $species, $chrs, $deb, $fin, $len ];
# 
			my $rh_chrom   = $h_regions{ $species }{ $primer } || {};
			my $ra_regions = $$rh_chrom{   $chrom   } || [];

			push( @$ra_regions, { start => $start, end => $end } );

			$h_regions{ $species }{ $primer }{ $chrom } = $ra_regions;
		}
	}

	$self->_Set( __rh_regions => \%h_regions );

	return;
}


=head2 Procedure

 Title        : _Run
 Usage        : $o_xx->_Run();
 Prerequisite : Called by NarcisseService::FetchResults
 Procedure    : just call _GetAPILinks to get a link to the API
                then _Fetch, to get what is wanted.
 Args         : none

=cut

sub _Run
{
	my $self = shift;

	my @a_attr = qw( _o_instance_obs __rh_regions );
	my( $o_inst_obs, $rh_regions ) = $self->_Get( @a_attr );

	# la verification que l'espece appartient à l'instance se fera dans l'accesseur
	foreach my $species ( keys %$rh_regions )
	{
# 		commentee parce qu'il faut trouver un moyen de loguer ca, meme en mode web-service
# 		next if $o_inst_obs->HasSpecies( $species )

		my %h_params = ();
		$h_params{ species } = $species;

		my $rh_primers = $$rh_regions{ $species };
		foreach my $primer ( keys %$rh_primers )
		{
			my $rh_chroms = $$rh_primers{ $primer };
			foreach my $chrom ( keys %$rh_chroms )
			{
				$h_params{ chrom } = $chrom;

				my $ra_regions = $$rh_chroms{ $chrom };
				$h_params{ ra_regions } = $ra_regions;

				my $nb = @$ra_regions;
				my $msg = "Primer $primer : will search on $nb region(s) of chrom $chrom of $species\n";
				$self->_Trace( $msg, '.', 'IS' );

				$self->__RunAccessor( $primer, \%h_params );
			}
		} # primer
	} # species
}


=head2 __RunAccessor

 Usage        : $o_x->__RunAccessor( $primer, $rh_args )
 Function     : seek for locus tags/sequence for the primer $primer
                with 'classic' accessors.
 Args         :
     $primer, string, primer's name
     $rh_args, hash ref, argument of accessor
 Error        : none.

=cut

sub __RunAccessor
{
	my $self   = shift;
	my( $primer, $rh_args ) = @_;

	my $what = $self->_Get( qw( __what ) );

	$$rh_args{ -o_logger } = $self->_GetLogger();
	$$rh_args{ -o_instance_obs } = $self->_Get( qw( _o_instance_obs ) );

	my $o_ns;
	if( $what eq 'locus' )
	{
		$o_ns = NarcisseServiceFactory::GetGenesFromRegions( %$rh_args,
		                                              -out_type => &NS_LOCUS );
	}
	else
	{
		$o_ns = NarcisseServiceFactory::GetSlicesFromRegions( %$rh_args,
		                                               -out_type => &NS_FASTA );
	}

	my $rh_results = $o_ns->FetchResults();

	my $ra_data = $$rh_results{ data };
	my $nb = ( defined $ra_data ) ? scalar @$ra_data : 0;
	my $msg = "found $nb results for $primer\n";
	$self->_Trace( $msg, '.', 'IS' );
# warn Dumper $rh_results;
	$self->_StoreResults( $primer, $rh_results );

	my $ra_not_found = $o_ns->FetchNotFound();
	 $nb = scalar( @$ra_not_found );
	$msg = "not found $nb values for $primer\n";
	$self->_Trace( $msg, '.', 'IS' );
# warn Dumper $ra_not_found;
	$self->_StoreNotFound( { $primer => $ra_not_found } )  if( $nb != 0 );

	return;
}


=head2 Procedure

 Title        : _StoreResults
 Usage        : $o_xx->_StoreResults( $primer, $rh_results );
 Prerequisite : none
 Procedure    : 
 Args         :
     $primer, string, primer's name
     $rh_results, hash ref, results of the accessor

=cut

sub _StoreResults
{
	my $self = shift;
	my( $primer, $rh_results ) = @_;

# 	warn Dumper { $primer => $rh_results };

	my $ra_final_results = $self->_GetArrayResults();
	my $rh_final_results = $self->_GetHashResults();

	my $ra_locus = $$rh_results{ data };
	if( defined( $ra_locus ) )
	{ # au moins un locus correspond a l'amorce (GetGenesLocusTagsFromPSRegions)
		foreach my $tag( @$ra_locus )
		{
			my $ra_list = $$rh_final_results{ $tag } || [];
			push( @$ra_list, $primer );
			$$rh_final_results{ $tag } = $ra_list;
		}
	}
	else
	{ # dans quel cas ? a preciser
		while( my( $head, $seq ) = each %$rh_results )
		{
			$head .= "primer=$primer";

			$$rh_final_results{ $head } = $seq;
			push( @$ra_final_results, $head );
		}
	}

# 	warn Dumper { $primer => [ $ra_final_results, $rh_final_results ] };

	return;
}

=head2 Function AssertOptions

 Usage        : my $value = $o_x->AssertOptions( $ra_values, $val )
 Function     : Check that $val belong to $ra_values.
                If true, return $val, else return the first element of the list
                and make a _Trace.
 Returns      : $val, one of the value of $ra_values
 Args         :
     $ra_values, a list ref of allowed values
     $val, scalar, the value to test.
 Error        : none

=cut

sub AssertOptions
{
	my $self = shift;
	my( $ra_values, $val ) = @_;

# warn Dumper( $ra_values, $val );
	my $return;

	my $o_enum = New Enum( @$ra_values );
	if( $o_enum->Includes( $val ) )
	{
		$return = $val;
	}
	else
	{
		my $default = $$ra_values[0];
		my $msg = "$val does not belong to the list '@$ra_values',";
		$msg   .= " we'll use default value : $default.\n";
		$self->_Trace( $msg, '.', 'IS' );

		$return = $default;
	}

	return $val;
}



1;