#
# 	sebastien.letort@toulouse.inra.fr
#	Created: November 09, 2009
#	$Id$
#

package NarcisseService::AccessorWithNarcisseRegionInput;

=pod

=head1 NAME

 AccessorWithNarcisseRegionInput - a NarcisseService Class to access any feature found into region identify by a NarcisseRegion.

=head1  SYNOPSIS

 my $o_ns = New NarcisseService::AccessorWithNarcisseRegionInput( ra_nc_regions => $ra_nc_regions,
                                                                -out_type => $type );
 my $ra_results = $o_ns->FetchResults();

=head1 DESCRIPTION

 This class derives from NarcisseService. It is specific to access features
 found into regions described by PrimerSearch file.

 Maybe one day all regions will be defined as NarcisseRegion.

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( NarcisseService );

use LipmError::ParameterException;
use NarcisseServiceFactory;
use NarcisseServiceKeywords qw( &NS_INFOS &NS_FASTA &NS_LOCUS );


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New NarcisseService::AccessorWithNarcisseRegionInput( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of NarcisseService::AccessorWithNarcisseRegionInput class.
 Returns      : a valid NarcisseService::AccessorWithNarcisseRegionInput object
 Args         :
     -o_logger       => Logger object
     -o_instance_obs => InstanceObserver object,
	 ra_nc_regions   => list ref of NarcisseRegion
     -what           => string 'locus' or 'sequence'
                        locus will retrieve all genes locus tags
                        sequence will retrieve regions sequence.
 Attributes   :
     _o_instance_obs : InstanceObserver object

     __fh_file          : PrimerSearch file handler
     __what             : string, 'locus' or 'sequence' depending on what we want to retrieve

=cut


=head2 Procedure

 Title        : _TreatOptions
 Usage        : $o_xx->_TreatOptions( ra_nc_regions => $ra_nc_regions,
                                      -what    => $what );
 Prerequisite : called by NarcisseService::_Init
 Procedure    : 
 Args         : cf _Init.

=cut

sub _TreatOptions
{
	my $self   = shift;
	my %h_args = ( ra_nc_regions => undef,
# 	               -out_type => &NS_LOCUS,
	               -what  => 'locus',
# 	               -o_instance_obs => undef,
	               @_ );

	# to set _o_instance_observer
	$self->SUPER::_TreatOptions( %h_args );

	my $fh_file = new IO::File( $file, 'r' )
		or throw LipmError::IOException( $file, 'f' );

	my @a_vals = qw( locus sequence );
	my $what = $self->AssertOptions( \@a_vals, $h_args{ -what } );

	# pourquoi ça ne fonctionne pas ?
# 	@a_vals = ( &NS_LOCUS, &NS_FASTA );
# 	my $type = $self->AssertOptions( \@a_vals, $h_args{ -out_type } );

	my $rh_regions = &NARCISSE_utils::ParseNarcisseRegions( $h_args{ ra_nc_regions } );

	$self->_Set( __rh_regions => $rh_regions,
# 	             _out_type => $out_type,
	             __what    => $what );

	return;
}


=head2 Procedure

 Title        : _Run
 Usage        : $o_xx->_Run();
 Prerequisite : Called by NarcisseService::FetchResults
 Procedure    : just call _GetAPILinks to get a link to the API
                then _Fetch, to get what is wanted.
 Args         : none

=cut

sub _Run
{
	my $self = shift;

	my @a_attr = qw( _o_instance_obs __rh_regions );
	my( $o_inst_obs, $rh_regions ) = $self->_Get( @a_attr );

	my $inst_name = $o_inst_obs->GetName();

	# la verification que l'espece appartient à l'instance se fera dans l'accesseur
	foreach my $species ( keys %$rh_regions )
	{
		if( &FALSE == $o_inst_obs->HaveSpecies( $species ) )
		{
			my $long = $o_inst_obs->GetSpeciesLongName( $species );
			my $msg = "$long does not belong to $inst_name.\n";
			$self->_Trace( $msg, '.', 'IS' );
		}

		my %h_params = ();
		$h_params{ species } = $species;

		my $rh_chroms = $$rh_regions{ $species };
		foreach my $chrom ( keys %$rh_chroms )
		{
			$h_params{ chrom } = $chrom;

			my $ra_regions = $$rh_chroms{ $chrom };
			$h_params{ ra_regions } = $ra_regions;

			my $nb  = @$ra_regions;
			my $msg = "Primer $primer : will search on $nb region(s) of chrom $chrom of $species\n";
			$msg   .= "For now strand is ignored, only + strand is returned.\n";
			$self->_Trace( $msg, '.', 'IS' );

			$self->__RunAccessor( \%h_params );
		}
	} # species
}


=head2 __RunAccessor

 Usage        : $o_x->__RunAccessor( $rh_args )
 Function     : call 'classic' accessors.
 Args         :
     $rh_args, hash ref, argument of accessor
 Error        : none.

=cut

sub __RunAccessor
{
	my $self   = shift;
	my( $rh_args ) = @_;

	my $what = $self->_Get( qw( __what ) );

	$$rh_args{ -o_logger }       = $self->_GetLogger();
	$$rh_args{ -o_instance_obs } = $self->_Get( qw( _o_instance_obs ) );

	my $o_ns;
	if( $what eq 'locus' )
	{
		$o_ns = NarcisseServiceFactory::GetGenesFromRegions( %$rh_args,
		                                              -out_type => &NS_LOCUS );
	}
	else
	{
		$o_ns = NarcisseServiceFactory::GetSlicesFromRegions( %$rh_args,
		                                               -out_type => &NS_FASTA );
	}

	my $rh_results = $o_ns->FetchResults();
	my $msg = "found " .  scalar( keys %$rh_results ) . " results for $primer\n";
	$self->_Trace( $msg, '.', 'IS' );
# warn Dumper $rh_results;
	$self->_StoreResults( $primer, $rh_results );

	my $ra_not_found = $o_ns->FetchNotFound();
	my $nb = scalar( @$ra_not_found );
	$msg = "not found $nb values for $primer\n";
	$self->_Trace( $msg, '.', 'IS' );
# warn Dumper $ra_not_found;
	$self->_StoreNotFound( { $primer => $ra_not_found } )  if( $nb != 0 );

	return;
}


=head2 Procedure

 Title        : _StoreResults
 Usage        : $o_xx->_StoreResults( $primer, $rh_results );
 Prerequisite : none
 Procedure    : 
 Args         :
     $primer, string, primer's name
     $rh_results, hash ref, results of the accessor

=cut

sub _StoreResults
{
	my $self = shift;
	my( $primer, $rh_results ) = @_;

# 	warn Dumper { $primer => $rh_results };

	my $ra_final_results = $self->_GetArrayResults();
	my $rh_final_results = $self->_GetHashResults();

	my $ra_locus = $$rh_results{ data };
	if( defined( $ra_locus ) )
	{ # au moins un locus correspond a l'amorce (GetGenesLocusTagsFromPSRegions)
		foreach my $tag( @$ra_locus )
		{
			my $ra_list = $$rh_final_results{ $tag } || [];
			push( @$ra_list, $primer );
			$$rh_final_results{ $tag } = $ra_list;
		}
	}
	else
	{ # dans quel cas ? a preciser
		while( my( $head, $seq ) = each %$rh_results )
		{
			$head .= "primer=$primer";

			$$rh_final_results{ $head } = $seq;
			push( @$ra_final_results, $head );
		}
	}

# 	warn Dumper { $primer => [ $ra_final_results, $rh_final_results ] };

	return;
}

=head2 Function AssertOptions

 Usage        : my $value = $o_x->AssertOptions( $ra_values, $val )
 Function     : Check that $val belong to $ra_values.
                If true, return $val, else return the first element of the list
                and make a _Trace.
 Returns      : $val, one of the value of $ra_values
 Args         :
     $ra_values, a list ref of allowed values
     $val, scalar, the value to test.
 Error        : none

=cut

sub AssertOptions
{
	my $self = shift;
	my( $ra_values, $val ) = @_;

# warn Dumper( $ra_values, $val );
	my $return;

	my $o_enum = New Enum( @$ra_values );
	if( $o_enum->Includes( $val ) )
	{
		$return = $val;
	}
	else
	{
		my $default = $$ra_values[0];
		my $msg = "$val does not belong to the list '@$ra_values',";
		$msg   .= " we'll use default value : $default.\n";
		$self->_Trace( $msg, '.', 'IS' );

		$return = $default;
	}

	return $val;
}



1;