#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 25 03, 2009
#	$Id: GetProteinsFromRegions.pm 3142 2009-12-22 12:38:52Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


package GetProteinsFromRegions;

=pod

=head1 NAME


=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use base qw( NarcisseService::GrepBased );

use Error qw( :try );

=head2 Procedure

 Title        : _TreatOptions
 Usage        : $o_xx->_TreatOptions( -species  => $species
                                      -db => $db,
                                      ra_tags => $ra_tags );
 Prerequisite : called by NarcisseService::_Init
 Procedure    : set $species as attribute _species
 Args         : $species, string, short species name

=cut

sub _TreatOptions
{
	my $self   = shift;
	my %h_args = ( ra_regions => [],
					chrom => undef,
					-species => undef,
	               @_ );

	$self->SUPER::_TreatOptions( %h_args );

	# set __ra_regions
	$self->__SetRegionsAttributes( $h_args{ ra_regions } );

	if( !defined $h_args{ chrom } )
	{
		my $msg = "You must specified one chromosome.";
		throw LipmError::ObjectException( $self, '!build', -text => $msg );
	}

	$self->_Set( __chrom      => $h_args{ chrom },
				 __species    => $h_args{ -species } );

	return;
}


=head2 Procedure _Run

 Usage        : $o_xx->_Run();
 Prerequisite : Overide to make something
 Procedure    : This method should do the job.
 Args         : none

=cut

sub _Run
{
	my $self = shift;

	# obtenir les slices des regions
	my $rao_slices = $self->__GetSlices();

	#pour chaque slice, obtenir les id des proteines
	my $ra_prot_ids = $self->__GetProteinIdsFromSlices( $rao_slices );

	my $fasta_file = $self->_GetDbFile();

	my $regexp_tags = join( '|', @$ra_prot_ids );

	$self->_Trace( "Parsing $fasta_file with $regexp_tags\n", '.', 'ISV' );

	my $fe = new IO::File( $fasta_file, 'r' )
					or throw LipmError::IOException( $fasta_file, 'f' );
	while( my $line = <$fe> )
	{
		if( $line =~ /^>prot:(\S+)/ )
		{
				my $prot_id = $1;
				next	unless( $prot_id =~ $regexp_tags );

				chomp( $line );
				my $head = $line;

				my $seq_fasta = '';
				my $pos = $fe->getpos();
				while( ( defined( $fe ) )
						&& ( $line = <$fe> )
						&& ( $line !~ /^>/ ) )
				{
					$pos=$fe->getpos();
					chomp( $line );
					$seq_fasta .= $line;
				}

				$self->_StoreResults( { $head => $seq_fasta } );
				$fe->setpos( $pos ); # on revient sur la ligne d'avant pour la relecture
		} # fasta header
	} # end <file>

	return;
}



sub __GetProteinIdsFromSlices
{
	my $self = shift;
	my( $rao_slices ) = @_;

	my @a_prot_ids = ();

	foreach my $o_slice ( @$rao_slices )
	{
		my $rao_transcripts = $o_slice->get_all_Transcripts();
		foreach my $o_transcript ( @$rao_transcripts )
		{
			push( @a_prot_ids, $o_transcript->display_xref() );
		}
	}

	return \@a_prot_ids;
}


sub __GetSlices
{
	my $self = shift;

	my $species = $self->_Get( '__species' );
	my $o_so  = $self->_Get( '_o_instance_obs' )->GetSpeciesObserver( $species );
	my $o_dba = $o_so->GetDBAdaptor();
 	
	my $o_sla = $o_dba->get_SliceAdaptor();
	if( !$o_sla )
	{
		my $msg = "no slice adaptor have been generated for $species\n";
		throw LipmError( -text => $msg );
	}

	my( $ra_regions, $chr ) = $self->_Get( qw( __ra_regions __chrom ) );
	my @ao_links = ();

	# $ra_regions est une ref de liste contenant
	# des ref de hash style { start => x, end => y }
	foreach my $rh_region ( @$ra_regions )
	{
		# j'ajoute un try-catch car si on precise un mauvais chrom, on peut passer au suivant.
		# ou bien dans le cas du multi espece, cael a une notatino en chiffre romain.
		try
		{
			my( $start, $end ) = @$rh_region{ qw( start end ) };
			my @a_options = ( $chr, $start, $end );

			my $o_slice   = $o_sla->fetch_by_region( 'chromosome', @a_options );

			if( !$o_slice )
			{
				warn Dumper \@a_options;
			}

			push( @ao_links, $o_slice );

#		on pourrait nommer la region, pour le moment utilite seulement dans legoo
#		donc je ne le fait pas, je suis presse !
# 		my $name = $$rh_region{ name } || "$start..$end";
#		push( @ao_links, { $name => $o_slice } );
		}
		catch LipmError with
		{
			my( $o_err, $r ) = @_;
			$self->_Trace( $o_err->Explain(), '.', 'IS' );

			$r = 1; # pour poursuivre
		};
	}

	return \@ao_links;
}



=head2 Procedure

 Title        : __SetRegionsAttributes
 Usage        : $o_xx->__SetRegionsAttributes( $ra_regions );
 Prerequisite : none
 Procedure    : none
 Args         : $ra_regions, a list of string in the format start..end

=cut

sub __SetRegionsAttributes
{
	my $self = shift;
	my ( $ra_regions ) = @_;

	my @a_list = ();

	foreach my $region ( @$ra_regions )
	{
# 		$region =~ /^(\d+)(-|(\.\.))(\d+)$/;
		my ( $start, $end ) = split( /\.\.|-/, $region );
		$start = &Sequence_utils::Unit2Base( $start );
		$end   = &Sequence_utils::Unit2Base( $end );

		if( !defined( $start ) or !defined( $end )
		or ( $start !~ /^\d+$/ ) or ( $end !~ /^\d+$/ )
		)
		{
			no warnings; # pour pallier les cas undef
			my $msg = "the region <$start>..<$end> is invalid.";
			use warnings;
			$self->_Trace( $msg, '.', 'IS' );
		}
		else
		{
			push( @a_list, { start => $start, end => $end } );
		}
	}

	$self->_Set( __ra_regions => \@a_list );

	return;
}



1;

__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
