#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: October 24, 2008
#	$Id$
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Pairwise;

use strict;
use warnings;

use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf
use File::Spec;
use Cwd;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;

use base qw( LipmObject );

=pod

=head1 NAME

 PairWise - An object to manage pairwise genome comparison

=head1 DESCRIPTION
  - a virtual object : only PairWise::Genomic or PairWise::Genetic are really instanciated
  - maintains information for both organisms under consideration
  - methods for initializing the result directories (raw as well as processed)
  - methods for reformating of data prior to an installation in Narcisse tree

=cut

use Data::Dumper;
use Carp;
use File::Spec;

use General;
use Error;
use LipmError;
use LipmError::ObjectException;
use ParamParser;
use Bio::Narcisse::Utils::NarcisseDataReader::LineByLine;
use NarcisseDataWriter;
use NarcisseDataWriter::OnDisk;
use Runner;

use base qw( LipmObject );

use constant CFGQ => "cfgQ";
use constant CFGS => "cfgS";

our %H_DEFBEHAVIOUR;

$H_DEFBEHAVIOUR{'verbose'} = 1;

=head2 method _Init

 Usage        : my $o_x = New Pairwise( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of Pairwise class.
 Returns      : a valid Pairwise object
 Args         : %h_args, a hash with the following keys
 -- mandatory --
 	query   : the SpeciesObserver of the query
 	subject : the SpeciesObserver of the subject
 	o_param : A Paramparser object
 	o_logger : A Logger object
 	-- optional --
 	-verbose : flag for debugging purpose

 Attributes  :
 __method
 __results_path
 _selfcomparison
 __raw_results_dir
 __X_filename
 _o_io,    InstanceObserver object
 _o_query, SpeciesObserver object onto query species
 _o_target, SpeciesObserver object onto target/subject species
 _o_param

=cut

sub _Init
{
	my $self = shift;
	my %h_param = (	'query'    => undef,
					'subject'  => undef,
					'o_param'  => undef,
					'o_io'     => undef,
					'o_logger' => undef,
					'method'   => undef,
                    '-where'   => undef,
					'-verbose' => 0,
					@_ );

	#Set Logger
	$self->SUPER::_Init( %h_param );

	$H_DEFBEHAVIOUR{'verbose'} = $h_param{'verbose'};

	my @mandatory_param = qw( o_io query subject o_param method o_logger );
	foreach my $param ( @mandatory_param )
	{
		&LipmError::ParameterException::Assert( $param => $h_param{ $param } );
	}

	# set attributes corresponding to mandatories params
	my @a_attr = qw( _o_io _o_query _o_target _o_param __method );
	my %h_attr;
	@h_attr{ @a_attr } = @h_param{ @mandatory_param };
	$self->_Set( %h_attr );

	# undef values will be set farther
	$self->_Set( __results_path    => undef,
				 __raw_results_dir => undef,
				 _cutoff_name      => undef,
                 __where           => $h_param{-where} );

	$self->__SetResultsPath();

	$self->__SetChainingParameters();

	#Self comparison is at first always false
	$self->_Set( _selfcomparison => 0 );

	#Do we force the comparison
	my $o_param = $self->_Get( '_o_param' );
	my $force   = ( $o_param->IsDefined('force') ) ? &TRUE : &FALSE ;
	$self->_Set( '__force' => $force );

	if( $force )
	{
		my $msg = "--force has been invoked. The comparison will be forced.\n";
		$self->_Trace( $msg . "\n", '.', 'IP' );
		print STDERR $msg;
	}

    # This attribute is set by the _OpenChrFileHandle method
    my @a_list_file = ();
    $self->_Set('_ra_list_file',\@a_list_file);
	return $self;
}


=head2 function __SetResultsPath

 Usage        : $self->__SetResultsPath()
 Prerequisite : none
 Function     : setting the directory tree for the output filenames
 					comparison/method/species_q/species_s 
 						for blastp2genome, blastp and glint outputs (genomic)
 						for primersearch outputs (genetic)
 Returns      : none
 Args         : none

=cut

sub __SetResultsPath
{
	my $self = shift;

	my @a_attr = qw( _o_query _o_target __method );
	my( $o_query, $o_tgt, $method ) = $self->_Get( @a_attr );
	my $results_path = $o_query->GetComparisonPath( method => $method,
													target => $o_tgt->GetShortName(),
                                                    -where => $self->_Get('__where'),
													-mkdir => &TRUE );

	$self->_Set( __results_path => $results_path );

	return;
}


=head2 procedure __SetChainingParameters

 Usage        : $self->__SetChainingParameters();
 Prerequisite : none
 Function     : Set the correct chaining cutoff parameters for the corresponding method, kingdom and instance name (when possible)
 Args         : none
 Return       : none

=cut

sub __SetChainingParameters
{
	my $self = shift;

	my $o_param =  $self->_Get( qw( _o_param ) );

	my $chaining_parameters;

	if( $o_param->IsDefined( 'chaining_param' ) )
	{
		$chaining_parameters = $o_param->Get( 'chaining_param' ); 
		$self->_Set( '_cutoff_name' => "User defined" );
	}
	else
	{
		$chaining_parameters = $self->__GetChainingParamFromCfgFile();
	}

	$self->_Set( '__chaining_parameters' => $chaining_parameters);

	my @a_level_parameters = split( ":", $chaining_parameters );

    # 0 and 1 are required levels (hence +1), the other levels are defined by the chaining parameters
	my $level_max = scalar @a_level_parameters + 1;
	$self->_Set( '__level_max' => $level_max );

	return;
}

=head2 Procedure __GetChainingParamFromCfgFile

 Usage        : $o_xxx->__GetChainingParamFromCfgFile();
 Prerequisite : none
 Procedure    : Set the correct chaining cutoff parameters for the corresponding method, kingdom and instance name (when possible)
                using the information from the config file

=cut

sub __GetChainingParamFromCfgFile
{
	my $self = shift;

	my $method = $self->GetMethod();

	my $o_io   = $self->_Get( '_o_io' );
	my $rh_cfg = $o_io->GetConfigHash();
	my $instance_name = $o_io->GetName();
	my $kingdom_name  = $o_io->GetKingdom();

	#Get available instance default parameters
	my $chaining_cutoff_key    = uc( $method ) . '_CHAINING_CUTOFFS';
	my @a_chaining_cutoff_keys = grep( /$chaining_cutoff_key/, keys %$rh_cfg );

	#Select the cutoff parameters corresponding to the kingdom or the instance name
	my %h_matching_keys = ();
	foreach my $cutoff_key ( @a_chaining_cutoff_keys )
	{
		my ($cutoff_name) = ($cutoff_key =~ /^(\w*?)\_/);
		if ( $instance_name =~ /$cutoff_name/i )
		{
			$h_matching_keys{ $cutoff_key } ++;
		}
		if ( $kingdom_name =~ /$cutoff_name/i )
		{
			$h_matching_keys{ $cutoff_key } ++;
		}
	}

	my $default_chaining_cutoff_key =  'DEFAULT_'.uc($method).'_CHAINING_CUTOFFS';
	my @a_matching_keys = keys %h_matching_keys;
	
	my $cutoff_key_name;
	if ( scalar @a_matching_keys == 1 ) #If a single $cutoff_key_name has been found we take it
	{
		$cutoff_key_name = $a_matching_keys[0];
	}
	else  #No corresponding cutoffs default parameters or ambiguous instance-kingdom name
	{
		$cutoff_key_name = $default_chaining_cutoff_key;
	}

	$self->_Set( '_cutoff_name' => $cutoff_key_name );
	my $chaining_parameters = $o_io->GetCfg( $cutoff_key_name );

	return $chaining_parameters;
} #__GetChainingParamFromCfgFile

=head2 Procedures DoComparison, DoOrthologDetection

 Usage     : $o_xxx->DoComparison()
             $o_xxx->DoOrthologDetection()
 Procedure : Those procedures DO NOT DO ANYTHING
             If the derived class needs to implement DoComparison (it should) or DoOrthologDetetion (it could),
             the REAL code is there
 
=cut
sub DoComparison { return; };
sub DoOrthologDetection { return; };

=head2 Procedure DoBuildChain

 Usage        : $o_xxx->DoBuildChain();
 Prerequisite : The glint and blastp2genome outputfiles must have been computed first
 Procedure    : Calls DoChainning an performs some postprocessing steps
 				- reformating the data into Narcisse format for syntenies
 				- moving the files into the Narcisse tree
=cut

sub DoBuildChain
{
	my $self = shift;

	print "STEP: building chains\n\n";	

	my $sav_dir = cwd;
	my $working_dir = File::Temp::tempdir( 'build_chain_XXXX', CLEANUP => 0, TMPDIR => 1 );

	# build and launch the chain command
	$self->DoChaining( $working_dir );

	# data reformat for narcisse (creating the different directories)
	$self->_ReformatData( $working_dir );

	# move the corresponding files to the Narcisse tree
	$self->_MoveInNarcisse( $working_dir );

	chdir $sav_dir;
	rmtree($working_dir);

	return;
}

=head2 procedure _ReformatData

 Title        : _ReformatData
 Usage        : $self->_ReformatData( ? );
 Prerequisite : The working dir must contain files produced by the chaining algorithm and therefore prefixed by chain and followed
                by the level value
 Function     : Performs a reformating of the data according to the narcisse file format
 Args         : the working dir

=cut

sub _ReformatData
{
	my $self = shift;
	my $work_dir  = shift;

	my $o_io      = $self->_Get( '_o_io' );
	my $unix_sort = $o_io -> GetUnixPath('UNIX_SORT');
	my $unix_grep = $o_io -> GetUnixPath('UNIX_GREP');

	chdir($work_dir);

	my $o_param = $self->_Get( '_o_param' );
    my $self_comparison = $o_param->IsDefined( 'self' );

    my ( $query, $subjt ) = $self->GetBothShort();

	mkdir ( $query );
    mkdir ( $subjt ) unless ( $self_comparison == &TRUE );
    
	my $dh_work_dir = new IO::Dir( $work_dir )
				or throw LipmError::IOException( $work_dir, 'd' );

	my @a_files = $dh_work_dir->read();
	@a_files = File::Spec->no_upwards( @a_files );	# removing . ..
	foreach my $file ( @a_files )
	{
		next	if ($file !~ /^chain(\d+)/ );
		my ($level) = $file =~ /^chain(\d+)/;
		next if ( $level > $self->GetMaxLevel() );
	
 		# Calling NarcisseDataWriter in append mode, just for sorting
		warn "sorting " . cwd . " $file";
		my $ndw = New NarcisseDataWriter::OnDisk( $file, 
		                                          undef, 
		                                          { append => &TRUE, sort => &TRUE, sort_numerical => &FALSE }, 
		                                          [ 'q_chrom','s_chrom' ] );
		$ndw    = undef;
		
		# passage very quick and dirty, suite a un bug de glintchain les fichiers sont parfois
		# binaires et dans ce cas sort se plante completement
		my $file_type = `file -b $file`;
		chomp $file_type;
		if ( $file_type =~ /^ASCII/)
		{
			$self->__BuildDataFilesWithLevelFile( $file, $query, $subjt, $level, $self_comparison );
		}
		else
		{
			warn "FILE = $file BINAIRE - IGNORE";
		}
	}

	return;
}

=head2 Procedure __BuildDataFilesWithLevelFile

 Title        : __BuildDataFilesWithLevelFile
 Usage        : $self->__BuildDataFilesWithLevelFile( $file, $query, $subjt, $level, \@a_query_files, \@a_subject_files  );
 Prerequisite : The source file must be sorted using the columns q_chrom and s_chrom
 Procedure    : From the file whose path is passed by parameter, build Narcisse data files
 Args         : $file : The name of a file created by glintchain, and sorted
                $query: The query org short name
                $subjt: The subject org short name
                $level: The synteny level to consider
                $self_comparison &TRUE if self
 Globals      : @A_QUERY_FIELDS, @A_SUBJECT_FIELDS, @A_SYNT_FIELDS (build the arrays only once)

=cut

my @A_QUERY_FIELDS   = qw ( element q_chrom q_start q_end strand s_chrom s_start s_end s_order score q_chain mol );
my @A_SUBJECT_FIELDS = qw ( s_order s_chrom s_start s_end strand q_chrom q_start q_end q_order score s_chain mol );
my @A_SYNT_FIELDS    = qw ( element chrom start end sign s_chrom s_start s_end s_order matching q_chain mol );

sub __BuildDataFilesWithLevelFile
{
    my $self = shift;
	my ( $file, $query, $subjt, $level, $self_comparison ) = @_;

	my %h_ndw_objects;

	# Using the LineByLine version of NarcisseDataReader, to avoid reading the whole (may be huge) file in memory
	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader::LineByLine();
	$o_ndr->Prepare( -from  => $file );

	# For each row from the source file
	my ($q_chrom,$s_chrom) = ('','');
	my ($o_ndw_q,$o_ndw_s);
	
	while ( my $rh_line = $o_ndr->FetchNextLine() )
	{
		my %h_line = %$rh_line;
		
		# on enleve le nom court des noms de chromosome
		$self->__PurgeBothChrName( \$h_line{'q_chrom'}, \$h_line{'s_chrom'}, $query, $subjt );

		if ( $level == 0)
		{	
			$h_line{'q_chain'} = ":";
			$h_line{'s_chain'} = ":";
		}

		# when the chromosome is changed, we create a new file
		if  ( $h_line{q_chrom} ne $q_chrom )
		{
			$o_ndw_q = undef;
			$q_chrom = $h_line{q_chrom};
			$o_ndw_q = $self -> __GetNdw ( $query, $level, $q_chrom );
#			warn "Q $query, $level, $q_chrom";
#			warn Dumper( \%h_line );
		}
		if ( $self_comparison == &FALSE )
		{
			if ( $h_line{s_chrom} ne $s_chrom )
			{
				$o_ndw_s = undef;
				$s_chrom = $h_line{s_chrom};
				$o_ndw_s = $self -> __GetNdw ( $subjt, $level, $s_chrom, \%h_ndw_objects );
#				warn "S $subjt, $level, $s_chrom";
#				warn Dumper( \%h_line );
			}
		}

		my %h_q;
		@h_q { @A_SYNT_FIELDS } = @h_line{ @A_QUERY_FIELDS };
		my @a_q = ( \%h_q );
		$o_ndw_q -> Update ( \@a_q );
		
		if ( $self_comparison == &FALSE )
		{
			my %h_s;
			@h_s { @A_SYNT_FIELDS } = @h_line{ @A_SUBJECT_FIELDS };
			my @a_s = ( \%h_s );
			$o_ndw_s -> Update ( \@a_s );
		}
	}
	
	return;

} # __BuildDataFilesWithLevelFile

=head2 Procedure __PurgeBothChrName

 Title        : __PurgeBothChrName
 Usage        : $self->__PurgeBothChrName( $r_chr_name1, $r_chr_name2, $query, $subjt );
 Prerequisite : none
 Procedure    : Removing org name from chromosome name
 Args         : $r_chr_name1, $r_chr_name2, $query, $subjt

=cut

sub __PurgeBothChrName
{
	my $self = shift;
	my ( $r_chr1, $r_chr2, $query, $subjt ) = @_;

	# on enleve $query du nom du chr
	$$r_chr1 =~ s/$query//g;
	$$r_chr2 =~ s/$query//g;

	# on enleve $subjt du nom du chr s'il existe
	if( defined( $subjt ) )
	{
		$$r_chr1 =~ s/$subjt//g;
		$$r_chr2 =~ s/$subjt//g;
	}

	return;
}

=head2 Function __GetNdw

 Title        : __GetNdw
 Usage        : $self->__GetHandle( $short, $level, $chr, $rh_ndw_objects );
 Function     : Create and return a NarcisseDataWriter object, or return an already opened one
 Return       : A NarcisseDataWriter object
 Args         : $short, the short species name
                $level, the synteny level
                $chr,   the chromosome label
                $rh_cache (default = undef) If specified, the file is kept inside this cache, it is not necessarily reopened

=cut

sub __GetNdw 
{
	my $self = shift;
	my ( $short, $level, $chr, $rh_cache_ndw) = @_;
	
	my @a_attrs = qw( MASK_FILE_CHROM SYNT );
	my( $mask, $synt_dir ) = $self->_Get( '_o_io' )->GetCfg( @a_attrs );

	my $filename = 	File::Spec->catfile( $short, $level, $mask );

	# on ajoute le chr et l'extension
	$filename .= $chr . '.' . $synt_dir;
	
	# If $cache_flg is false, we create a new object
	my $o_ndw;
	if ( ! defined ( $rh_cache_ndw ) )
	{
		$o_ndw = New NarcisseDataWriter::OnDisk ( $filename, 
												  \@A_SYNT_FIELDS, 
												  { append => &TRUE, sort => &TRUE, sort_numerical => { 'start' => 1, 'end' => 1, 's_start' => 1 , 's_end' => 1 } },
												  [ 'start','end','s_chrom','s_start','s_end' ] );
	}
	
	# Else, we keep a ref to the object, it will not be created again
	else
	{
		if ( exists $rh_cache_ndw -> { $filename } )
		{
			$o_ndw = $rh_cache_ndw -> { $filename };
		}
		else
		{
			$o_ndw = New NarcisseDataWriter::OnDisk ( $filename, 
													  \@A_SYNT_FIELDS, 
													  { append => &TRUE, sort => &TRUE, sort_numerical => { 'start' => 1, 'end' => 1, 's_start' => 1 , 's_end' => 1 } },
												      [ 'start','end','s_chrom','s_start','s_end' ] );
			$rh_cache_ndw -> { $filename } = $o_ndw;
		}
	}
		
	return $o_ndw;
}

=head2 procedure _OpenChrFileHandle

 Usage        : _OpenChrFileHandle($o_ref,$target,$chr)
 Function     : Create the orthology chr file if necessary and print the header line the first time the file is accessed
 Returns      : a reference on the filehandle
 Args         : $o_ref: The reference species observer
                $target: the target name
                $chr: chromosome id
 Attribute written: _ra_list_file

=cut

sub _OpenChrFileHandle
{
	my $self = shift;
	my( $o_ref, $target, $chrom ) = @_;

	my $file_chr = $o_ref->GetOrthologyFiles(
        target => $target,
        -chrom => $chrom,
        -check => &FALSE );

	#my $fh = new IO::File ">>$file_chr" or throw LipmError::IOException( $file_chr, 'f' );
    my $fh = new IO::File ">>$file_chr" or confess "cannot open $file_chr";
    if (-z $file_chr)
    {
		print $fh "#q_chr\tq_prot\tq_gene\tq_start\tq_end\ts_chr\ts_prot\ts_gene\ts_start\ts_end\n";
        my $ra_list_file = $self->_Get('_ra_list_file');
		push (@$ra_list_file,$file_chr);
	}
        
	return $fh;
}

=head2 procedure _MoveInNarcisse

 Title        : _MoveInNarcisse
 Usage        : _MoveInNarcisse();
 Prerequisite : ReformatData should have been called first
 Function     : Moves the files to the corresponding Narcisse tree
 Returns      : none
 Args         : $work_dir : the working (temporary) directory where the files have been generated

=cut

sub _MoveInNarcisse
{
	my $self = shift;
	my $work_dir  = shift;

	my( $o_param, $o_io ) = $self->_Get( qw( _o_param _o_io ) );
    my $self_computation  = $o_param->IsDefined ('self' );
    
    my ( $orgaQ, $orgaS )= $self->GetBothShort();
    
    # hum, pas fameux...
	my $dstQ = $o_param->Get( &CFGQ.'DST');
    my $dstS = ( $self_computation == &TRUE ) ? $dstQ : $o_param->Get( &CFGS.'DST');

	my $comp = $o_io->GetCfg('COMPARATIVE');
	my $synt = $o_io->GetCfg('SYNT');
	my $prog = $o_param->Get('method');

	# query
	my @a_elts = ( $dstQ, $comp, $orgaQ, $orgaS, $synt, $prog );
	$self->__MoveIn( $work_dir, $orgaQ, @a_elts );

	#subject
	if ( $self_computation == &FALSE )
	{
		@a_elts = ( $dstS, $comp, $orgaS, $orgaQ, $synt, $prog );
		$self->__MoveIn( $work_dir, $orgaS, @a_elts );
	}

	return;

} # MoveInNarcisse


=head2 Procedure __MoveIn

 Title        : __MoveIn
 Usage        : $self->__MoveIn( $work_dir, @a_elts );
 Prerequisite : none
 Procedure    : none
 Args         : $work_dir, @a_elts

=cut

sub __MoveIn
{
	my $self = shift;
	my ( $work_dir, $orgaQ, @a_elts ) = @_;

	my $path = File::Spec->catfile( @a_elts );
	rmtree( $path );
	mkpath( $path );
	my $src  = File::Spec->catfile( $work_dir, $orgaQ, '*' );
	my $unix_cp = $self->_Get( '_o_io' )->GetUnixPath('UNIX_CP');
	`$unix_cp $src $path`;

	return;
}


=head2 procedure DoChaining

 Usage        : $o_pariwise->DoChaining();
 Function     : This is a virtual function and should not be called by a Pairwise object (which does not exist anyway)

=cut
sub DoChaining
{
	my $self = shift;

	die "This is a virtual function my lord !";

	return;
}


=head2 function MakeRawResultDir

 Usage        : $self->MakeRawResultDir()
 Prerequisite : none
 Function     : setting the directory dir for raw output files 
 					comparison/method/species_q/species_s/raw
 						blast and glint raw results 
 Returns      : the path of the corresponding dir
 Args         : none

=cut

sub MakeRawResultDir
{
	my $self = shift;

	my $results_root_path = $self->_Get( '__results_path' );

	#Raw comparison files
	my $raw_result_dir =  "$results_root_path/raw";
	mkpath( $raw_result_dir ) unless (-d $raw_result_dir);

	$self->_Set( __raw_results_dir => $raw_result_dir );

	return $raw_result_dir;
}


=head2 function RemoveRawResultDir

 Usage        : $self->RemoveRawResultDir()
 Prerequisite : none
 Function     : removes the raw result dir
 Returns      : none
 Args         : none

=cut

sub RemoveRawResultDir
{
	my $self = shift;

	my $raw_result_dir = $self->_Get( '__raw_results_dir' );

	rmtree( $raw_result_dir );	
}


=head2 function GetRawResultDir

 Usage        : $o_p->GetRawResultDir()
 Prerequisite : SetRawResultDir must have been called first
 Function     : accessor
 Returns      : the path to the raw result dir 

=cut

sub GetRawResultDir
{
	my $self = shift;

	my $attr_value = $self->_Get( '__raw_results_dir' );
	( defined $attr_value )
		or throw LipmError::ObjectException( $self, '!attribute', -dbg => " __raw_results_dir is undefined at this point" );
	
	return $attr_value;
}

####################################################
#   Accessors for genome-specific attributes
#      returns info for Both genomes in the order : query subject
####################################################

=head2 function GetBothShort

 Usage        : ( $short_q, $short_s )  = $o_obj->GetBothShort();
 Function     : accessor of the short name of both organisms
 Returns      : ( $short_q, $short_s ), 2 string, the short names
 
=cut

sub GetBothShort
{
	my $self = shift;

	my( $o_query, $o_target ) = $self->_Get( qw( _o_query _o_target ) );

	my $short_q = $o_query->GetShortName();
	my $short_s = $o_target->GetShortName();

	return ( $short_q, $short_s );
}

=head2 function GetBothDataDir

 Usage        : ( $dir_q, $dir_s )  = $self->GetBothDataDir(   );
 Function     : accessor of the data dir of both organisms
 Returns      : ( $dir_q, $dir_s ), 2 string, the dir names

=cut

sub GetBothDataDir
{
	my $self = shift;
    my( $o_query, $o_target ) = $self->_Get( qw( _o_query _o_target ) );
    my $dir_q = $o_query->GetDataPath();
    my $dir_s = $o_target->GetDataPath();
	return ( $dir_q, $dir_s );
}


=head2 function GetBothOrthologDir

 Usage        : ( $dir_q, $dir_s )  = $self->GetBothOrthologDir(   );
 Function     : accessor of the data dir of both organisms
 Returns      : ( $dir_q, $dir_s ), 2 string, the dir names

=cut

sub GetBothOrthologDir
{
	my $self = shift;

	my( $o_query, $o_target ) = $self->_Get( qw( _o_query _o_target ) );

	my %h_args = ( -check => &FALSE, -where => $self->_Get( '__where' ) );

	my $orth_dir_q = $o_query->GetOrthologueDirpath ( target => $o_target->GetShortName(), %h_args );
	my $orth_dir_s = $o_target->GetOrthologueDirpath( target => $o_query->GetShortName(), %h_args );

	my @a_created = mkpath( $orth_dir_q );
	$self->_Trace( "created directories @a_created\n", '.', 'IP' );

	 @a_created = mkpath( $orth_dir_s );
	$self->_Trace( "created directories @a_created\n", '.', 'IP' );

	return ( $orth_dir_q, $orth_dir_s );
}

####################################################
#   Managing filenames according to the method and
#      to the short names of both organisms
####################################################

=head2 Procedure SetFilename

 Usage        : &SetFilename( $blastp_filename )
 Function     : setting the path for the output filename
 Args         : the filename

 Note : la methode semble etre d'un usage privee.

=cut

sub SetFilename
{
	my $self = shift;
	my %h_param = (  type => undef,
# 					 basename => undef, # pas vu d'utilisation de ca
					 @_ );

	# pas de valeur undef
	LipmError::ParameterException::Assert( %h_param );

	my $type = $h_param{ type };

	my $attr_name = "__" . $type . "_filename";

	if (defined $h_param{ basename } )
	{
		$self->_Set( $attr_name => $h_param{basename} );
	}
	else
	{
		my $target_name = join( '_', $type, $self->GetBothShort() ) . ".out";
		$self->_Set( $attr_name => $target_name );
	}

}

=head2 function GetFilename

 Usage        : &GetFilename( $type )
 Prerequisite : none
 Function     : getting the path for the filename corresponding to type
 Returns      : none
 Args         : $type, string
 Globals      : none

=cut

sub GetFilename
{
	my $self = shift;
	my $type = shift;

	my $path = $self->_Get( '__results_path' );
	my $filename = $self->GetBasename( $type );

	return  File::Spec->catfile( $path, $filename );
}


=head2 function GetBaseName

 Usage        : &GetBaseName()
 Prerequisite : none
 Function     : getting the basename of the file corresponding to type
 Returns      : the basename as a string
 Args         : the type
 Globals      : none

 Note : usage prive.

=cut

sub GetBasename
{
	my $self = shift;
	my $type = shift;

	my $o_type = New Enum( qw( blastp glint blastp2genome genetic geneticrbh ) );
	LipmError::ParameterException::Assert( type => $type, -enum => $o_type );

	my $attr = $self->_Get( "__" . $type . "_filename" );
	$self->SetFilename( type => $type ) unless( defined $attr );

	return $self->_Get( "__" . $type . "_filename" );
}


####################################################
#                 Various accessors
####################################################

=head2 function Some Accessors

 Title        : SelfComparison GetResultRoot GetResultsDirpath GetResultRoot GetChainingParameters
                GetMaxLevel GetMethod ForceComparaison
 Usage        : $o_xx->GetMethod()
 Function     : return the value of the corresponding attribute.
 Returns      : string, except for SelfComparison and ForceComparaison which return boolean

=cut

sub SelfComparison
{
	my $self = shift;

	return $self->_Get( '_selfcomparison' );
}

sub GetResultRoot
{
	my $self = shift;	
	return $self->_Get( '__results_path' );
}

sub GetResultsDirpath
{ # =GetResultRoot mais + clair je trouve
	my $self = shift;	
	return $self->_Get( '__results_path' );
}

sub GetChainingParameters
{
	my $self = shift;
	return $self->_Get( '__chaining_parameters' );
}

sub GetMaxLevel
{
	my $self = shift;
	return $self->_Get( '__level_max' );
}

sub GetMethod
{
	my $self = shift;

	return $self->_Get( '__method' );
}

sub ForceComparaison
{
	my $self = shift;
	
	return $self->_Get( '__force' );
}


=head2 Procedure Verbose

 Title        : Verbose
 Usage        : $self->Verbose( $msg );
 Prerequisite : none
 Procedure    : will print $msg onto STDERR if VERBOSE is set
 Args         : $msg, a string

=cut
sub Verbose
{
	my $self = shift;
	my ( $msg ) = @_;

	print STDERR $msg."\n"	if( $H_DEFBEHAVIOUR{'verbose'} );

	return 1;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
