#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: October 24, 2008
#	$Id$
#

package Pairwise::Genetic;

use strict;
use warnings;

use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
use File::Copy;
use File::Basename;

use Bio::Narcisse::Utils::NarcisseDataReader;

use base qw( Pairwise );

=pod

=head1 NAME

 PairWise - An object to manage pairwise genetic comparison

=head1 DESCRIPTION

descr - maintains information for both organisms under consideration
   - defines the absolute path of the files where the proteic (blastp) and nucleic (glint)
     output will be stored

   Some important private attributes

	° related to data
	  __rh_protein_info a reference on a hash containing all the protein gene informations (gene structure)
	  __rh_blast_output a reference on a hash containing all the blastp output information

	° related to files

      __results_root : the root directory where the processed glint, blastp and the constructed blastp2genome files will be stored
      __blastoutput_filename : basename for the blastp output (complete path is constructed by appending __results_root )


=cut

use Data::Dumper;
use Carp;
use General;
use LipmError;
use LipmError::ObjectException;
use LipmError::IOException;
use NARCISSE_utils;

use constant AMPLIMERLIMITSIZE => 2000;
use constant INTERVALMAXBP     => 100000;

use constant GENE => "GENE";
use constant MARKER => "MARKER";

# une explication de ces termes serait la bienvenue <- Seb
# des arg du blast pour la plupart
our %h_default = (
		MISMATCH_PERCENT => 5,
);


=head2 method _Init

 Usage        : my $o_x = New Pairwise::Genetic( $o_org_q, $o_org_s );
 Prerequisite : none
 Fonction     : initialise an instance of Pairwise class.
 Returns      : a valid Pairwise object
 Args         : a hash with the following keys
 				-- mandatory --
 				query   : the SpeciesObserver of the query
 				subject : the SpeciesObserver of the subject
 				o_param : A Paramparser object
 				o_logger : A Logger object
 				-- optional --
 				verbose : flag for debugging purpose

=cut

sub _Init
{
	my $self = shift;
	my %h_params = (
				# those params are needed by mother class.
# 					'query' => undef,
# 					'subject' => undef,
# 					'o_param' => undef,
# 					'o_logger' => undef,
# 					'o_io' => undef,					
# 					'-verbose' => 0,
					@_ );

	$self->SUPER::_Init( %h_params, method => 'genetic' );

	$self->_Set( __primersearch_done => 0,
				 __rh_marker_info    => undef );


	# Load marker information from files
	$self->__LoadMarkerInfoFromFiles();

	return $self;
}

=head2 Procedure DoComparison

 Usage     : $o_xxx->DoComparison()
 Procedure : Call DoGeneticAnchoring, who really does the job
 
=cut
sub DoComparison {
    my $self = shift;
    $self->DoGeneticAnchoring();
}

=head2 function __LoadMarkerInfoFromFiles

 Usage        : $rh_info_prot = $self->__LoadMarkerInfoFromFiles(  )
 Fonction     : Initialize a hash with info related to the markers (reference and target)

=cut

sub __LoadMarkerInfoFromFiles
{
	my $self = shift;

	my $rh_info_marker = {};

	# Recupere toutes les informations des marqueurs
	&_LoadMarkerInfo( $self->_Get( '_o_query' ), $rh_info_marker );
	&_LoadMarkerInfo( $self->_Get( '_o_target' ), $rh_info_marker );
		
	$self->_Set( __rh_marker_info => $rh_info_marker );

	return;

=head2 Procedure _LoadMarkerInfo

 Usage        : &_LoadMarkerInfo( $o_species, $rh_info );
 Prerequisite : none
 Procedure    : fill the $rh_info with data found into $data_dir
 Args         : $o_species
                $rh_info, a hash to receive information

=cut

	sub _LoadMarkerInfo
	{
		my ( $o_species, $rh_info ) = @_;

		my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

		my @a_gene_files = $o_species->GetGeneFiles();
		foreach my $file ( @a_gene_files )
		{
			$o_ndr->Prepare( -from  => $file );
			my $rh_data = $o_ndr->FetchallHashref( 'locus_tag' );

			#All markers info from $rh_local_info are concatenated at once into $rh_info
			my @a_keys = keys %$rh_data;
			@$rh_info{ @a_keys } = @$rh_data{ @a_keys };
		}

		return;
	}

} # __LoadMarkerInfoFromFiles


=head2 Procedure DoChaining

 Usage        : $o_pairwise->DoChaining( $working_dir );
 Prerequisite : none
 Procedure    : build the chain command line and executes it
 Args         : the working_dir

=cut

sub DoChaining
{
	my $self = shift;
	my $working_dir = shift;

	print STDERR "Chaining parameters : ".$self->GetChainingParameters()." ( ".$self->_Get( '_cutoff_name')." ) \n";
	print STDERR "Computed levels : ".$self->GetMaxLevel()." \n";

	# DetectOrthologs is a prerequisite since it constructs the rbh file necessary for chaining
    # Thus it is done here and the function DoOrthologDetection id NOT redefined
    # TODO - oooooh que c'est moche (manu)
	$self->DetectOrthologs();

	my $anchors_filename = $self->GetFilename( 'geneticrbh' );

	#Possibly zipped anchor file
	&NARCISSE_utils::Unzip( $anchors_filename );

	print STDERR "$anchors_filename \n";

	my $chaining_param =  $self->GetChainingParameters();

	my $prog = $self->_Get( '_o_io' )->GetCfg( 'ROOT_GLINT_CHAIN' );
	my $cmd_build_chain = $prog;

	$cmd_build_chain .= " --anchors $anchors_filename";
	$cmd_build_chain .= " --targetdir $working_dir";	
	$cmd_build_chain .= " --cutoffs $chaining_param";
	$cmd_build_chain .= " --noRBH " if ( $self->SelfComparison() );

	print STDERR "$cmd_build_chain\n";
	&NARCISSE_utils::RunCmd($cmd_build_chain);
	
	&NARCISSE_utils::Zip( $anchors_filename );

	return;
}


=head2 Procedure DoGeneticAnchoring

 Usage        : $o_pairwise->DoGeneticAnchoring( );
 Prerequisite : none
 Procedure    : Sets the mapping result ouput file and calls PrimersearchRbh
 Args         : none

=cut

sub DoGeneticAnchoring
{
	my $self = shift;

	my $genetic_anchors_output = $self->GetFilename( 'genetic' );

	# If the blastp2genome output exists we do not compute the proteic comparison again unless the --force option is set to true
	if ( -e "$genetic_anchors_output.gz" && !$self->ForceComparaison())
	{
		my $basename = &basename( "$genetic_anchors_output.gz" );
		print STDERR "Primersearch output already available ( $basename ). Using this file. \n";
		print STDERR "Use --force to force the primersearch comparison\n";

		return ;
	}

	# PrimerSearch
	printf "STEP: Computing PrimerSearch [%s]\n",$self->_Get( '_o_param' )->Get('path_primersearch');

	#First create a temporary working dir for raw output files	
	my $raw_result_dir = $self->MakeRawResultDir();

	$self->DoPrimersearch( );

	#now storing the output file ion the correct place
	my $primersearch_output = $self->GetMappingResultFilename( $raw_result_dir );

	copy($primersearch_output, $genetic_anchors_output );

	#zipping the genetic anchors output file
	&NARCISSE_utils::Zip( $genetic_anchors_output );

	#Removes the temporary working dir
	$self->RemoveRawResultDir();

	return;
}

=head2 procedure Primersearch

 Usage        : &Primersearch()
 Prerequisite : none
 Fonction     : - run the primersearch
                - dump the relationships
 Returns      : none
 Args         : none

=cut

sub DoPrimersearch
{
	my $self = shift;

	my @a_attrs = qw( __primersearch_done _o_param _o_io );
	my( $primersearch_done, $o_param, $o_io ) = $self->_Get( @a_attrs );

	( 1 == $primersearch_done ) and return;

	my $rawresult_dir =  $self->GetRawResultDir();

	my $primersearchoutputfile = $self->GetMappingResultFilename( $rawresult_dir );

	my $path_primersearch =  $o_param->Get('path_primersearch');

	my ( $org_q, $org_s ) = $self->GetBothShort();
	my ( $dir_q, $dir_s ) = $self->GetBothDataDir( );

	my $cde_input = 'cde_cmd';
	my $cde_output= 'cde_oe';
	my @a_outputfiles = ();

	$self->_Trace("Execute Primersearch :\n\n",'Timer','A');

	my $commands = new IO::File(">$cde_input") or die ("ERROR - Cannot open $cde_input for writing");
	my @a_listoffiles = ();
	my $rh_chroms = $self->_Get( '_o_query' )->GetChromsInfo();
	my @a_chr =  keys( %$rh_chroms );

	#Setting some parameters
	my $mismatchpercent = '';
	my $mismatchnb      = '';
	if ( $o_param->Get( &Pairwise::CFGQ.'ORG_TAXID' ) ==  $o_param->Get(&Pairwise::CFGS.'ORG_TAXID') )
	{
		$mismatchpercent = $o_param->Get('mismatchpercentintra');	# primersearch
		$mismatchnb      = $o_param->Get('mismatchnbintra'); # e-PCR
	}
	else
	{
		$mismatchpercent = $o_param->Get('mismatchpercentintra');
		$mismatchnb      = $o_param->Get('mismatchnbintra');
	}
	$self->_Trace( "\t" . gmtime(time) . "\n\tmismatchpercent = $mismatchpercent\tmismatchnb=$mismatchnb\n\n", 'Timer','A' );

	my @fields = qw( locus_tag forward_primer reverse_primer );
	my @a_keys = qw( MASK_FILE_CHROM MARKER DB CHR );
	my( $mask_file_chrom, $marker, $db, $chr ) = $o_io->GetCfg( @a_keys );
	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

	foreach my $chr_name ( @a_chr )
	{
		# je ne sais pas si on peut passer par SO->GetChromFile ou autre

		#my $prefix = $mask_file_chrom.$chr_name;
		#my $marker_file = "$dir_q/$CFG{MARKER}/$CFG{MASK_FILE_CHROM}$chr_name.$CFG{MARKER}";
		my $marker_file = File::Spec->catfile($dir_q,$marker,$mask_file_chrom.$chr_name.".$marker");
		$o_ndr->Prepare( -from  => $marker_file );
		my $rh_primers  = $o_ndr->FetchallHashref( 'locus_tag' );

		my $f_primerpairs = new IO::File("$primersearchoutputfile$chr_name.primers","w");
        my $fasta_file = File::Spec->catfile($dir_s,$db,$chr,"index/${org_s}V1.seq");
		foreach my $marker (keys %$rh_primers)
		{
			if ( $path_primersearch =~ /primersearch/ )
			{
				print $f_primerpairs join("\t",@{$rh_primers->{$marker}}{@fields})."\n";
			}
			else	# e-PCR
			{
				print $f_primerpairs join("\t",@{$rh_primers->{$marker}}{@fields})."\t0\n";
			}
		}
		$f_primerpairs->close;

        #my $fasta_file = File::Spec->catfile($dir_s,$db,$chr,$org_s.$chr_name.'.fasta');
		#my $cde = "$path_primersearch -seqall $fasta_file -mismatchpercent $mismatchpercent -infile $primersearchoutputfile$chr_name.primers -outfile $primersearchoutputfile$chr_name";

		my $cde = '';
		if ( $path_primersearch =~ /primersearch/ )
		{
			$cde = "$path_primersearch -seqall $fasta_file -mismatchpercent $mismatchpercent -infile $primersearchoutputfile$chr_name.primers -outfile $primersearchoutputfile$chr_name";
		}
		else
		{
			$cde = "$path_primersearch $primersearchoutputfile$chr_name.primers $fasta_file N=$mismatchnb W=3 > $primersearchoutputfile$chr_name";
		}
		print $commands "$cde\n";
		push(@a_listoffiles,"$primersearchoutputfile$chr_name");
		$self->_Trace("\t".gmtime(time)."\n\t$cde\n\n",'Timer','A');
	}
	$commands->close;

	&NARCISSE_utils::RunCmd("/bin/sh $cde_input",$cde_output);

	unlink ($primersearchoutputfile) if ( -e $primersearchoutputfile );
    my $unix_cat = $o_io->GetUnixPath('UNIX_CAT');
	foreach my $file (@a_listoffiles)
	{
		my $cde = "$unix_cat $file >> $primersearchoutputfile";
		$self->_Trace("\t".gmtime(time)."\n\t$cde\n\n",'Timer','A');
		system($cde);
	}
	
	return $primersearchoutputfile;
}


=head2 Procedure DetectOrthologs

 Usage        : $o_xxx->DetectOrthologs( );
 Prerequisite : The blast comparison should have been performed
 function     : From the raw blast ouput detects protein orthologs
 Args         : none
 Return       : none

=cut

sub DetectOrthologs
{
	my $self = shift;

	my $genetic_anchors_output =  $self->GetFilename( 'genetic' );

	#Possibly zipped anchor file
	&NARCISSE_utils::Unzip( $genetic_anchors_output );

	$self->EditPrimerSearchRbhResults( $genetic_anchors_output );

	#Now reformating orthologs for glint_chain
	$self->ReformatingRbhFile();

    # Sort the files whose list was built by _OpenChrFileHandle
    my $ra_list_file = $self->_Get('_ra_list_file');
	&NARCISSE_utils::SortFile(@$ra_list_file);

	#zipping back again
	&NARCISSE_utils::Zip( $genetic_anchors_output );

	return;
}

=head2 Procedure ReformatingRbhFile

 Title        : ReformatingRbhFile
 Usage        : $o_xxx->ReformatingRbhFile( );
 Prerequisite : EditPrimerSearchRbhResults must have been called first
 function     : Reformating the rbh file for subsequent use by glintchain
 Args         : none
 Return       : none

=cut

sub ReformatingRbhFile
{
	my $self = shift;

	my @a_attrs = qw( _o_query _o_target );
	my( $o_query, $o_target ) = $self->_Get( @a_attrs );
	my( $query, $target ) = $self->GetBothShort();

	# deleting previous files
	my @a_ortho_files_q = $o_query->GetOrthologyFiles( target => $target, -existing_files => &TRUE );
	map ( unlink, @a_ortho_files_q );
	my @a_ortho_files_s = $o_target->GetOrthologyFiles( target => $query, -existing_files => &TRUE );
	map ( unlink, @a_ortho_files_s );

	#now reformating file
	my $primer_rbh_file = $self->_Get( '__primer_rbh_file' );
	my $fh_primers      = &NARCISSE_utils::GetFileHandle( $primer_rbh_file, 'r' );

	my $genetic_rbh_output = $self->GetFilename( 'geneticrbh' );
	my $fh_genetic_output  = &NARCISSE_utils::GetFileHandle( $genetic_rbh_output, 'w' );

	my @fields = qw ( q_name q_start q_end s_name s_start s_end strand score size ); # __
	print $fh_genetic_output "#".join("\t",@fields)."\n";
	my $cpt = 1;
	while ( my $line = <$fh_primers> )
	{
		chomp($line);
		my @values=split("\t",$line);
		print "$line\n";
		print $fh_genetic_output join("\t",@values)."\n";
		my $A_fh = $self->_OpenChrFileHandle( $o_query, $target, $values[0] );
		print $A_fh "$values[0]\tnclocusmap$cpt\tnclocusmap$cpt\t$values[1]\t$values[2]\t$values[3]\tnclocus$cpt\tnclocus$cpt\t$values[4]\t$values[5]\n";
		my $B_fh =  $self->_OpenChrFileHandle( $o_target, $query, $values[3]);
		print $B_fh "$values[3]\tnclocus$cpt\tnclocus$cpt\t$values[4]\t$values[5]\t$values[0]\tnclocusmap$cpt\tnclocusmap$cpt\t$values[1]\t$values[2]\n";
		$B_fh->close();
		$A_fh->close();
		$cpt++;
	}

	$fh_genetic_output->close();
	$fh_primers->close();

	return;
}

=head2 Procedure EditPrimerSearchRbhResults

 Usage        : $self->EditPrimerSearchRbhResults(  );
 Prerequisite : none
 Procedure    : Edit the PrimerSearch reciprocal best hits
 Args         : none

=cut

sub EditPrimerSearchRbhResults
{
	my $self = shift;
	my $primersearchoutputfile = shift;

	my ( $org_q, $org_s ) = $self->GetBothShort();
	my ( $orthologue_q, $orthologue_s ) = $self->GetBothOrthologDir();
	my $rh_marker_info = $self->GetMarkerInfo();

	$self->_Trace( "Build syntheny like files :\n\n",'Timer','A' );

	my $o_param = $self->_Get( '_o_param' );
	my $method  = $o_param->Get('method');

	my $filename = join( '_', 'rbh', $method, $org_q, $org_s );
	my $FILE_RBH_PRIMER_AB = File::Spec->catfile( $orthologue_q, $filename );

	$filename = join( '_', 'rbh', $method, $org_s, $org_q );
	my $FILE_RBH_PRIMER_BA = File::Spec->catfile( $orthologue_s, $filename );

	my $psfile = new IO::File( $primersearchoutputfile ) or Carp::croak("ERROR, unable to read >$primersearchoutputfile<\n");
#Primer name h2_84g4a
#Amplimer 1
#	Sequence: mtr1
#
#	GGCTGATTTGAAACATGCCA hits forward strand at 1067937 with 0 mismatches
#	GGTGGTTGTGGGACACTTTT hits reverse strand at [25770537] with 0 mismatches
#	Amplimer length: 164 bp

	my $symbol='';
	my $chrS='';
	my $amplimerid='';
	my $amplimerlength=0;
	my $primerstart=0;
	my $primerend=0;
	my %h_min=();
	my %h_max=();
	my $pos='';
	my $ok = 0;
    my $path_primersearch =  $o_param->Get('path_primersearch');
	while(my $line=<$psfile>)
	{
		chomp($line);
		if ( $path_primersearch  =~ /primersearch/ )
		{
			if ( $line =~ /^Primer name\s+(\S+)/ )
			{
				$symbol = $1;
			}
			elsif ( $line =~ /^\s+Sequence:\s+(\S+)/ )
			{
				$chrS = $1;
				$chrS =~ s/$org_s//;
			}
			elsif ( $line =~ /^Amplimer\s+(\S+)/ )
			{
				$amplimerid = "$symbol.$1";
			}
			elsif ( $line =~ /hits forward strand at\s+(\d+)/ )
			{
				$primerstart = $1;
			}
			elsif ( $line =~ /Amplimer length:\s+(\S+)/ )
			{
				$amplimerlength = $1;
				$primerend = $primerstart+$amplimerlength-1;
				$ok = 1;
			}
		}
		else	# mtr0       51226..52367     mtr-5-23381948-23383221-CR932961_39.4_1#3
		{
			($chrS,$pos,$symbol)= split(' ',$line);
			$chrS =~ s/$org_s//;
			($primerstart,$primerend) = split(/\.\./,$pos);
			$amplimerlength = $primerend-$primerstart;
			$ok = 1;
		}
		if ( $ok && ($amplimerlength < &AMPLIMERLIMITSIZE ) )
		{
			my $key_cm = "$rh_marker_info->{$symbol}->{chrom}-$rh_marker_info->{$symbol}->{start}-$rh_marker_info->{$symbol}->{end}";
			$h_min{$key_cm} = $primerstart if ( ! defined($h_min{$key_cm}) ||   $primerstart < $h_min{$key_cm} );
			$h_max{$key_cm} = $primerend   if ( ! defined($h_max{$key_cm}) ||   $primerend > $h_max{$key_cm} );
			$ok = 0;
		}
	}
	$psfile->close;
	my $primerqs = new IO::File($FILE_RBH_PRIMER_AB,'w') or Carp::croak("ERROR, unable to create >$FILE_RBH_PRIMER_AB<\n");
	foreach my $key_cm ( sort {
								my @A=split('-',$a);
								my @B=split('-',$b);
								return ( $A[1] <=> $B[1] ) if ( $A[0] eq $B[0] );
								return ( $A[0] cmp $B[0] );
							  } keys(%h_min) )
	{
		my $str = $key_cm;
		$str =~ s/-/\t/g;
		my @a_ = split(' ',$str);
		my $w = ($h_max{$key_cm}-$h_min{$key_cm});
		if ( $w > &INTERVALMAXBP )
		{
			$h_min{$key_cm}=$h_max{$key_cm}-&INTERVALMAXBP;
			$w = &INTERVALMAXBP;
		}
		printf $primerqs "$str\t$a_[0]\t$h_min{$key_cm}\t$h_max{$key_cm}\t+\t$w\t$w\n"; #,$h_max{$key_cm}-$h_min{$key_cm};
	}
	$primerqs->close;
	
	$self->_Set( '__primer_rbh_file' => $FILE_RBH_PRIMER_AB );
	
	return ;
}


=head2 function GetMarkerInfo

 Usage        : $rh_marker_info = $o_genetic->GetMarkerInfo()
 Function     : accessor to the marker information

=cut

sub GetMarkerInfo
{
	my $self = shift;

	return $self->_Get( '__rh_marker_info' );
}


=head3 function GetMappingResultFilename

 Usage        : $self->GetBlastOutputFilename( $temp_working_dir );
 Prerequisite : none
 Function     : constructs the name of the primersearch output and returns it
 				If the mapping result is provided by the mappingresults argument set __primersearch_done to 1
 Returns      : the filename
 Args         : $temp_working_dir

=cut

sub GetMappingResultFilename
{
	my $self = shift ;
	my $temp_working_dir = shift;

	my $o_param = $self->_Get( '_o_param' );
	my $mapping_results_file;

	if( $o_param->IsDefined('mappingresults') )
	{
		$o_param->AssertFileExists('mappingresults');
		$mapping_results_file = $o_param->Get('mappingresults');

		$self->_Set( __primersearch_done => 1 );
	}
	else
	{
		my $basename = join("_",($self->GetMethod(),$self->GetBothShort()));
		$mapping_results_file = "$temp_working_dir/$basename.out";
	}

	return $mapping_results_file ;
}

1;
