#
# 	Thomas.Faraut@toulouse.inra.fr
#	Created: October 24, 2008
#	$Id:  $
#

package Pairwise::Genomic;

use strict;
use warnings;

use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf 
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
use File::Copy;
use File::Basename;
use LipmError::IOException;

use Bio::Narcisse::Utils::NarcisseDataReader;

use base qw( Pairwise );

=pod

=head1 NAME

 PairWise - An object to manage pairwise genome comparison

=head1 DESCRIPTION

 - maintains information for both organisms under consideration
 - defines the absolute path of the files where the proteic (blastp) and nucleic (glint)
     output will be stored

   Some important private attributes  

    ° related to data
     __rh_protein_info a reference on a hash containing all the protein gene informations (gene structure)
     __rh_blast_output a reference on a hash containing all the blastp output information

    ° related to files

     __results_root : the root directory where the processed glint, blastp and the constructed blastp2genome files will be stored
     __blastoutput_filename : basename for the blastp output (complete path is constructed by appending __results_root )


=cut

use Data::Dumper;
use Carp;
use File::Spec;

use General;
use LipmError;
use LipmError::ObjectException;
use NARCISSE_utils;

our %H_DEFBEHAVIOUR;

use constant RAW_BLAST => 'raw_blast';
use constant SELF => 'self';
use constant MIN_SPAN => 70;
use constant MAX_EVALUE => 5e-20;
use constant TRANSCRIPT => "TRANSCRIPT";

use constant BLASTP_NC_HEADER => qw(q_name s_name pident size mismatch gaps q_start q_end s_start s_end evalue score q_len s_len q_cov s_cov); # __

use constant DEFAULT_COMP_TYPE => 'genomic';

$H_DEFBEHAVIOUR{'verbose'} = 0;

# une explication de ces termes serait la bienvenue <- Seb
# des arg du blast pour la plupart
our %h_default = (
		EVALUE	=>	"1e-3",
		O_GAP	=>	9,
		E_GAP	=>	2,
		L_DB	=>	15*10e6,
		MIN_SPAN=>	50,
		MIN_GSCORE => 30,
		CPU 	=>	1
);


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New Pairwise::Genomic( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of Pairwise class.
 Returns      : a valid Pairwise object
 Args         : %h_args, a hash with the following keys
 	-- mandatory --
 	query   : the SpeciesObserver of the query
 	subject : the SpeciesObserver of the subject
 	o_param : A Paramparser object
 	o_logger: A Logger object
 	o_io    : An InstanceObserver object
 	-- optional --
 	-verbose : flag for debugging purpose

 Attributes
 __X_virtual_genomes
 __cutoff_name

=cut

sub _Init
{
	my $self = shift;
	my %h_param= (
				param => undef, #non renseigne/commente, non utilise
				# those params are needed by mother class.
# 					'query'    => undef,
# 					'subject'  => undef,
# 					'o_param'  => undef,
# 					'o_logger' => undef,
# 					'o_io'     => undef,
#                   '-where'   => undef,
# 					'-verbose' => 0,
					@_ );

	$self->SUPER::_Init( %h_param, 'method' => 'genomic' );

	# undef value defined farther
	$self->_Set( __comparedmolecules => undef,
				 __prbh_results_dir  => undef,
				 _selfcomparison    => undef,
				 __rbh_file_QS       => undef,
				 __rbh_file_SQ       => undef,
				 __rh_protein_info   => undef,
				 __rh_blast_output   => undef,
				 __blastoutput_filename => undef );

	#Hsp Filter parameters
	my $min_span   = $h_param{ param }{ min_span };
	$min_span      = &MIN_SPAN unless defined( $min_span );

	my $max_evalue = $h_param{ param }{ max_evalue };
	$max_evalue    = &MAX_EVALUE unless defined( $max_evalue );

	$self->_Set( filter => { min_span => $min_span,
							 max_evalue => $max_evalue } );

	$self->__SetComparedMol();
	
	$self->__SetPrbhResultDir();

	$self->__SetRbhFilename();

	#Is it a self comparison ?
	my $bool = $h_param{ o_param }->IsDefined('self') ;
	$self->_Set( _selfcomparison => $bool );

	return $self;
}

=head2 Procedure DoComparison

 Usage     : $o_xxx->DoComparison()
 Procedure : Call CompareGenomes, who really does the job
 
=cut
sub DoComparison {
    my $self = shift;
    $self->CompareGenomes();
}

=head2 Procedure DoOrthologDetection

 Usage     : $o_xxx->DoOrthologDetection()
 Procedure : Call DetectOrthologs, who really does the job
 
=cut
sub DoOrthologDetection {
    my $self = shift;
    $self->DetectOrthologs();
}

=head2 __SetRbhFilename

 Usage        : $o_x->__SetRbhFilename( none )
 Function     : .
 Args         :

=cut

sub __SetRbhFilename
{
	my $self   = shift;

	my( $short_q, $short_s ) = $self->GetBothShort();
	my $filename  = join( '_', 'rbh_prot',  $short_q, $short_s );

	my( $o_query, $o_tgt ) = $self->_Get( qw( _o_query _o_target ) );
	my $dirpath_q = $o_query->GetOrthologueDirpath( target => $short_s, -check => &FALSE );
	my $dirpath_t = $o_tgt->GetOrthologueDirpath(   target => $short_q, -check => &FALSE );

    my $rbh_file_QS = File::Spec->catfile($dirpath_q,$filename);
    my $rbh_file_SQ = File::Spec->catfile($dirpath_t,$filename);

	$self->_Set( __rbh_file_QS => $rbh_file_QS,
				 __rbh_file_SQ => $rbh_file_SQ);

	return;
} #__SetRbhFilename


=head2 Procedure __SetPrbhResultDir

 Usage        : $self->__SetPrbhResultDir()
 Prerequisite : none
 Function     : setting the directory dir for prbh raw output files (initially in misc/prbh )
 					comparison/genomic/species_q/species_s/prbh
 						blast and glint raw results as well as old misc/prbh files
 Args         : none

=cut

sub __SetPrbhResultDir
{
	my $self = shift;
	
	my $results_root_path = $self->GetResultsDirpath();
    
	#Prbh files
	my $prbh_result_dir =  "$results_root_path/prbh";
# 	mkpath( $prbh_result_dir ) unless (-d $prbh_result_dir);
	my @a_created = mkpath( $prbh_result_dir );
	$self->_Trace( "created directories : @a_created\n", '.', 'IP' );

	$self->_Set( __prbh_results_dir => $prbh_result_dir );

	return;
}

=head2 Procedure __SetComparedMol

 Usage        : __SetComparedMol( $o_param );
 Procedure    : Set the comparison type according to o_param
 Returns 	  : none

=cut

sub __SetComparedMol
{
	my $self = shift;

	my $o_param = $self->_Get( '_o_param' );
	my $comparison_type = &DEFAULT_COMP_TYPE;
	
	if ( 0 == $self->AreProteinsAvailable || $o_param->IsDefined('only_nucleic') || $o_param->IsDefined('no_proteic'))
	{
		$comparison_type = "nucleic";
	}
	if ( $o_param->IsDefined('only_proteic') || $o_param->IsDefined('no_nucleic'))
	{
		$comparison_type = "proteic";
	}
	
	$self->_Set( __comparedmolecules => $comparison_type );
		
	return ;
}


=head2 Procedure AreProteinsAvailable

 Usage        : $self->AreProteinsAvailable( );
 Procedure    : Check if both genomes have proteins
 Returns 	  : 0|1

=cut

sub AreProteinsAvailable
{
	my $self = shift;

	my( $o_query, $o_tgt ) = $self->_Get( qw( _o_query _o_target ) );

	my $query_has_protein = ( $o_query->GetNbProteins() != 0 );
	my $tgt_has_protein   = ( $o_tgt->GetNbProteins()   != 0 );

	return ( $query_has_protein && $tgt_has_protein );
}


=head2 Procedure CompareGenomes

 Usage        : $o_pairwise->CompareGenomes();
 Prerequisite : none
 Procedure    : Compare both genomes at the nucleic and|or proteic level

=cut

sub CompareGenomes
{
	my $self = shift;

	print STDERR "STEP: Comparing genomes \n";

	#First create a temporary working dir for raw output files
	# utiliser File::Temp tempdir ?
	$self->MakeRawResultDir();

	if ( $self->WithNucleic() )
	{
		$self->DoNucleicComparison(); 
	}
	if ( $self->WithProteic() )
	{
		$self->DoProteicComparison();
	}

	#Removes the temporary working dir 
    $self->RemoveRawResultDir();

	return;
}

=head2 Procedure DoChaining

 Usage        : $o_pairwise = &DoChaining( $working_dir );
 Prerequisite : none
 Procedure    : build the chain command line and executes it
 Args         : the working_dir

=cut

sub DoChaining
{
	my $self = shift;
	my $working_dir = shift;

	my $o_param = $self->_Get( '_o_param' );

	#If the chaining files are povided in the chains_dir we just copy them in $working_dir
	if ( $o_param->IsDefined('chains_dir') )
	{
		my $chains_dir = $o_param->Get('chains_dir');
		my @a_files=glob( "$chains_dir/chain*" );
		
		foreach my $file (  @a_files )
		{
			my $basename = basename($file );
			print STDERR "copying $file\n";	
			copy ($file, "$working_dir/$basename" );
		}

		return ;
	}

	my $chaining_param = $self->GetChainingParameters();

	print STDERR "Chaining parameters : $chaining_param ( ".$self->_Get( '_cutoff_name')." ) \n";
	print STDERR "Computed levels : ".$self->GetMaxLevel()." \n";

	my $glint_output = $self->GetFilename( 'glint' );
	my $blastpgenome_output = $self->GetFilename( 'blastp2genome' );

	&NARCISSE_utils::Unzip( $glint_output );
	&NARCISSE_utils::Unzip( $blastpgenome_output );
	
	my $prog = $self->_Get( '_o_io' )->GetCfg('ROOT_GLINT_CHAIN');
	my $cmd_build_chain = $prog;

	$cmd_build_chain .= " --blastpout $blastpgenome_output " if ( $self->WithProteic() && -e $blastpgenome_output);
	$cmd_build_chain .= " --glintout  $glint_output" if ( $self->WithNucleic() && -e $glint_output);
	$cmd_build_chain .= " --targetdir $working_dir";	
	$cmd_build_chain .= " --cutoffs $chaining_param";
	$cmd_build_chain .= " --self " if ( $self->SelfComparison() );
	$cmd_build_chain .= " --noRBH" if ( $o_param->Get('no_rbh') );
			
	print STDERR "$cmd_build_chain\n";
	&NARCISSE_utils::RunCmd($cmd_build_chain);
	
	&NARCISSE_utils::Zip( $glint_output );
	&NARCISSE_utils::Zip( $blastpgenome_output );		

	return;
}

#-----------    The Nucleic comparison  (methods dedicated to)  -----------------#

=head2 Procedure DoNucleicComparison

 Usage        : &DoNucleicComparison();
 Prerequisite : none
 function     : Perfoms the nucleic comparison using glint
 Args         : none

=cut

sub DoNucleicComparison
{
	my $self  = shift;

	my $glintoutput = $self->GetFilename( 'glint' );

	# In self mode, we do not compute nucleic comparisons
	return if ( &TRUE == $self->SelfComparison() );

	# If the glint output exist we do not compute the nucleic comparison unless the --force option is set to true
	if ( -e "$glintoutput.gz" && !$self->ForceComparaison())
	{
		my $basename = &basename( "$glintoutput.gz" );
		print STDERR "Glint output already available ( $basename ). Using this file. \n";
		print STDERR "Use --force to force the glint comparison\n";

		return ;
	}

	#Checking first the presence of the glint index
	$self->_CheckGlintIndex();

	my $ra_output_files;
	my $o_param = $self->_Get( '_o_param' );

	if ( $o_param->IsDefined('glintresults') )
	{
		$o_param->AssertFileExists('glintresults');
		push( @$ra_output_files, $o_param->Get('glintresults') );
	}
	else
	{
		$ra_output_files = $self->DoRunGlint();
	}

	my $fh_glintoutput = &NARCISSE_utils::GetFileHandle( $glintoutput, 'w' );

	my $start = 1;
	my @fields;
	my ( $orgQ, $orgS ) = $self->GetBothShort();

	print STDERR "Now reformatting files \n";
	foreach my $output_filename (@$ra_output_files)
	{
		my $input = new IO::File( $output_filename, 'r' )
			or throw LipmError::IOException( $output_filename, 'f' );

		my $header = <$input>;
		if ( $header=~/^\#/ )
		{			
			print $fh_glintoutput $header if ( $start );
			$header =~ s/\#//;
			chomp($header);
			@fields=split(/\t/,$header);		
		} else  #No header provided, we act as if it is in the previous output format
		{
			@fields=qw ( q_name q_start q_end s_name s_start s_end strand score size );  #_ 
			print $fh_glintoutput "#".join("\t",@fields)  if ( $start );
			 #already a data line in fact
			my $fmt_line = &_GetFormattedLine( $header, \@fields, $orgQ, $orgS);
			print STDERR $fmt_line."\n";
			print $fh_glintoutput $fmt_line."\n";
		}
		$start = 0;
		while ( my $line = <$input> )
		{
			#removing species prefix from chromosome number
			my $fmt_line=&_GetFormattedLine( $line, \@fields, $orgQ, $orgS);
			print $fh_glintoutput $fmt_line."\n"; 
		}
		$input->close();
	}
	$fh_glintoutput->close();

	#Always zip the files	
	&NARCISSE_utils::Zip( $glintoutput );

	return;

=head2 Function _GetFormattedLine

 Usage        : _GetFormattedLine
 Procedure    : Returns a tabulated formated line without species prefix      
 Args         : $o_both, the BothOrgs object

=cut

	sub _GetFormattedLine
	{
		my ( $line, $ra_fields, $orgQ, $orgS)  = @_;
		
		my %match;
		chomp($line);
		@match{@{$ra_fields}} = split(/\t/,$line);
		
		#Purging chromosome names	
		$match{ 'q_name' } =~ s/$orgQ//;
		$match{ 's_name' } =~ s/$orgS//; 
			
		return join("\t", @match{ @{$ra_fields} } );
	}	
}


=head2 Procedure DoRunGlint

 Usage        : DoRunGlint
 Procedure    : Runs the glint program on the virtual genomes
 Args         : $o_both, the BothOrgs object

=cut

sub DoRunGlint
{
	my $self = shift;

	print "STEP: Glint the genome chromosomes\n";

 	my ( $orgQ, $orgS ) = $self->GetBothShort();

    my $ra_vg_query   = $self->GetVirtualGenomes( $orgQ );
    my $ra_vg_subject = $self->GetVirtualGenomes( $orgS );

	# First check for the presence of the index files of the subject
	# If absent, we die with a message

	my $no_index_flg = &FALSE;
	foreach my $virtual_genome ( @$ra_vg_query )
	{
		if ( ! -e "$virtual_genome.fa" )
		{
			$self -> _PrintVirtualGenomeMessage( $orgQ );
			$no_index_flg = &TRUE;
		};
		last;
	}
	foreach my $virtual_genome ( @$ra_vg_subject )
	{
		if ( ! -e "$virtual_genome.fa" )
		{
			$self -> _PrintVirtualGenomeMessage( $orgS );
			$no_index_flg = &TRUE;
		}
	};
	if ( $no_index_flg )
	{
		print "*** Please reindex and try again\n";
		exit 0;
	}
	
	# Building the glint commands for all virtual genomes
    my $glint_input = 'glint_cmd';
    my $glint_output= 'glint_oe';

	# preparation
	my %h_params = $self->_PrepareGlint();

	# Glint additional parameters
	my $glint_parameters = " -c $h_params{min_gscore}";


	my( $o_io, $o_param ) = $self->_Get( qw( _o_io _o_param ) );
	my @a_output_files;

	my $commands = new IO::File( $glint_input, 'w' )
				or throw LipmError::IOException( $glint_input, 'f' );

	#my $target_dir_path = $self->{ o_param }->Get('rawresult_dir');
	my $target_dir_path = $self->GetRawResultDir();
	my $filename = 'glint_' . $orgQ;
	my $filepath = File::Spec->catfile( $target_dir_path, $filename );

	if( $o_param->IsDefined( 'self' ) )
	{
		if( 1 < scalar( @$ra_vg_query ) )
		{
			&Carp::croak("ERROR - Genomes with multiple virtual genome files cannot be handled yet with the self option\n");
		}

		my $query = shift @$ra_vg_query;
		my $vgQ   = $query;

		$vgQ =~ s/.*(V[0-9]+).*/$1/;

		my $output = $filepath . $vgQ . '_' . $orgQ . $vgQ . '.full.out';
		push( @a_output_files, $output );

		my $cmd = $o_io->GetCfg('ROOT_GLINT_BUILD') ." -o $output -S $query $glint_parameters\n";
		print $commands $cmd;
	}
	else
	{
		foreach my $query ( @$ra_vg_query )
		{
			my $vgQ = $query;

			$vgQ =~ s/.*(V[0-9]+).*/$1/;

			foreach my $subject  ( @$ra_vg_subject )
			{
				my $vgS = $subject;
				$vgS =~ s/.*(V[0-9]+).*/$1/;

				my $output = $filepath . $vgQ . '_' . $orgS.$vgS . '.full.out';
				push( @a_output_files, $output );

				my $cmd = $o_io->GetCfg('ROOT_GLINT_BUILD') ." -o $output $subject $query $glint_parameters\n";
				print $commands $cmd;
			}
		}
	}
	$commands->close();

	#Performing the actual glint comparisons
	&NARCISSE_utils::RunCmdFromFile( $glint_input, $glint_output );
 
# warn Dumper \@a_output_files;
	return \@a_output_files;
}


=head2 function _PrepareGlint

 Usage        : %h_params = &_PrepareGlint( );
 Prerequisite : none
 Function     : Read the glint params from o_param and return them into a hash
 Returns      : %h_params
 Args         : none
 Globals      : none

=cut

sub _PrepareGlint
{
	my $self = shift;

	my %h_p = (
			min_gscore => $self->_Get( '_o_param' )->Get('min_gscore')
		);

	my( $o_query, $o_target ) = $self->_Get( qw( _o_query _o_target ) );

	my $orgQ = $o_query->GetShortName();
	my $orgS = $o_target->GetShortName();

	my $data_q = $o_query->GetDataPath();
	my $data_s = $o_target->GetDataPath();

	my $ortho_q = $o_query->GetOrthologueDirpath( target => $orgS, -check => &FALSE );
	my $ortho_s = $o_target->GetOrthologueDirpath( target => $orgQ, -check => &FALSE );
	# cree les dossiers si inexistant ?

	$self->_Trace("Parameters :\n\n",'Timer','A');
	$self->_Trace("\tOrga A                  : $orgQ\n",'Timer','A');
	$self->_Trace("\tOrga B                  : $orgS\n",'Timer','A');
	$self->_Trace("\tDirectory data A        : $data_q\n",'Timer','A');
	$self->_Trace("\tDirectory data B        : $data_s\n",'Timer','A');
	$self->_Trace("\tMin glint score         : $h_p{ min_gscore } %\n",'Timer','A');

	return %h_p;
} # _PrepareGlint

=head2 Procedure _PrintVirtualGenomeMessage

Print a message explaining how to index the organism whose short name is passed by parameter

=cut
sub _PrintVirtualGenomeMessage
{
	my $self = shift;
	my ($org) = @_;
	
	print "\n";
	print "*** ERROR ! The indexes are absent for the organism $org\n";
	print "***         You may create them with the command:\n\n";
	print "***         nc_get_data.pl --cfg $org --only_index --only_formatdb\n\n";
	print "***         Please have a look to the help file (nc_get_data.pl --help)\n";
	print "***         Some switches may be interesting for you\n\n";
}


#-----------   END OF The Nucleic comparison  (methods dedicated to)  -----------------#

#-----------    The Proteic comparison  (methods dedicated to)  -----------------#

=head2 Procedure DoProteicComparison

 Usage        : &DoProteicComparison();
 Prerequisite : none
 function     : Perfoms the proteic comparison using blast and the following postprocessing steps
                  - edit it to add protein length information
                  - projects the hsps on the genome
                  - detects the reciprocal best hits for orthologue identification
                  - finally zipps the blast outputs
 Args         : none
 Return       : none

=cut

sub DoProteicComparison
{
	my $self = shift ;

	#No protein comparison is performed if one of the organisms lacks proteins
	if ( 0 == $self->AreProteinsAvailable() )
	{
		print STDERR "Skipping protein comparisons \n";
		return;
	}

	my $blastpgenome_output = $self->GetFilename( 'blastp2genome' );

	# If the blastp2genome output exists we do not compute the proteic comparison again unless the --force option is set to true
	if ( -e "$blastpgenome_output.gz" && !$self->ForceComparaison() )
	{
		my $basename = &basename( "$blastpgenome_output.gz" );
		print STDERR "Blast output already available ( $basename ). Using this file. \n";
		print STDERR "Use --force to force the blast comparison\n";
		return;
	}

	$self->__SetBlastOutputFilename( );

	# preparation
	$self->_LoadProteinInfo();
	my %h_params = $self->_PrepareBlast( );

	print "STEP: Blast the genome proteins\n";

	$self->Blast( \%h_params );	
	$self->AddLengthAndSpanToBlastResults(  );

	#The blastp file into the correct narcisse format
	$self->FromBlastp2NarcisseFormat( );

	#The protein hits are transformed into genomic coordinates
	$self->FromProteinToGenome( \%h_params );	
			
	#zipping the blastp2genome output file
	&NARCISSE_utils::Zip( $blastpgenome_output );

	return;
}# end of func DoProteicComparison


=head2 Procedure Blast

 Usage        :&Blast( $blast_out, $o_both, $h_params )
 Prerequisite : none
 Procedure    : Runs the blast or move the raw blast output file specified by the command line arguments in the correct place
 Returns      : none
 Args         : $blast_out : the blastoutput filename
                $o_both: an BothOrgs

=cut

sub Blast
{
	my $self = shift;
	my $rh_params = shift;

	my $o_param = $self->_Get( '_o_param' );

	my $runblast = ( $o_param->IsDefined('blastresults') ) ? 0 : 1;
	
	# bricolage _injection de fichier de resultats de blast brut
	$runblast = 0	if( $o_param->IsDefined( &RAW_BLAST ) );

	# perform the blast if needed
	if( 1 == $runblast )
	{
		$self->RunBlast( %$rh_params );
	}
	else
	{
		print STDERR "Not performing the blast comparison, using instead blastresults file :". $self->GetBlastOutputFilename( )."\n";
	}

	my $blast_out = $self->GetBlastOutputFilename( );
	$self->Verbose( $blast_out . "\n" );
	
	#fn de File::Copy
	if ( $o_param->IsDefined( &RAW_BLAST ) )
	{
		my $raw_blast = $o_param->Get( &RAW_BLAST );
		$self->Verbose( "start copying $raw_blast\n" );
		&copy( $raw_blast, $blast_out );
		$self->Verbose( "end copy\n" );
	}

	return;
}


=head2 Procedure RunBlast

 Usage        : &RunBlast( $blastoutput, $o_both_orgs, %h_params );
 Prerequisite : none
 Procedure    : Runs the blast program with the arguments specified in %h_param and writing it in $blastoutput 
 Args         : $blastoutput : a filename
                $o_both      : a BothOrgs object
                %h_params    : a hash with the blast parameters

=cut

sub RunBlast
{
	my $self = shift;
	my %h_params = @_;

	my $blast_out = $self->GetBlastOutputFilename( );

	my @a_attrs = qw( _o_io _o_query _o_target _o_param );
	my( $o_io, $o_query, $o_target, $o_param ) = $self->_Get( @a_attrs );

	my $bank_q = $o_query->GetProtFastaFile();
	my $bank_s = $o_target->GetProtFastaFile();

	my $blast_cde = "export BLASTMAT=" . $o_io->GetCfg('ROOT_BLASTMAT') . " ;";
	$blast_cde .= $h_params{ path_blast } . " -p blastp";
	$blast_cde .= " -d $bank_s -i $bank_q -m 8 -F F";
	$blast_cde .= " -G $h_params{ o_gap }";
	$blast_cde .= " -E $h_params{ e_gap }";
	$blast_cde .= " -e $h_params{ evalue }";
	$blast_cde .= " -z $h_params{ l_db }";
	$blast_cde .= " -a $h_params{ cpu }";
	$blast_cde .= " -o $blast_out";

	my $blast_input = 'blast_cmd';
	my $blast_output= 'blast_oe';

	$self->_Trace("Execute Blast :\n\n",'Timer','A');
	$self->_Trace("Bank A : $bank_q\n\n",'Timer','A');
	$self->_Trace("Bank B : $bank_s\n\n",'Timer','A');
	$self->_Trace("\t".gmtime(time)."\n\t$blast_cde\n\n",'Timer','A');

	my $fh_command = &NARCISSE_utils::GetFileHandle($blast_input, 'w');
	print $fh_command "$blast_cde\n";
	$fh_command->close();

	&NARCISSE_utils::RunCmd( "/bin/sh $blast_input", $blast_output );

	return;
}


=head2 Procedure DetectOrthologs

 Usage        : &DetectOrthologs( );
 Prerequisite : The blast comparison should have been performed : the blastp output file should be available
 function     : From the raw blast output detects protein orthologs
 Args         : none
 Return       : none

=cut

sub DetectOrthologs
{
	my $self = shift;

	###warn "DectectOrthologs TEMPORARILY REMOVED !";
	###return;
	
	if ( 0 == $self->AreProteinsAvailable() )
	{
		print STDERR "No protein available therefore no orthologs\n";
		return;
	}

	my $blast_out = $self->GetFilename( 'blastp' );

	#Possibly zipped blast file
	&NARCISSE_utils::Unzip( $blast_out );

	(-e $blast_out)
		or throw LipmError( -text => "The blast_output  $blast_out missing where needed", -user_msg => "$blast_out missing where needed" );

	#Preparing info and parameters for orthologs detection
	$self->_LoadProteinInfo();

	my %h_params = $self->_PrepareBlast( );

	#Now editing blast rbh results for orthologous detection
	$self->EditBlastRbhResults( \%h_params );

    # Sort the files whose list was built by _OpenChrFileHandle
    my $ra_list_file = $self->_Get('_ra_list_file');
	&NARCISSE_utils::SortFile(@$ra_list_file);

	#zipping the blastoutput file
	&NARCISSE_utils::Zip( $blast_out );

	return;
}


=head2 Procedure EditBlastRbhResults

 Usage        : $o_xxx->EditBlastResults( $blast_out );
 Prerequisite : none
 Procedure    : Edit the blast result file
              	In self mode _searching for intra chrq link,
              	all duplicated informations will be deleted.
 Args         : $blast_out

=cut

sub EditBlastRbhResults
{
	my $self = shift;
	my $rh_params = shift;

	#creation du repertoire qui va contenir les fichiers	
	my $repository_rbh = $self->GetPrbhResultDir();

	my ( $orgQ, $orgS ) = $self->GetBothShort();
	my $tmp_filename = "$repository_rbh/prbh.$orgQ.$orgS.";
	$tmp_filename .= $$rh_params{ evalue } . '.';
	$self->BlastToCluster(  $rh_params, $tmp_filename );

	# faudrait lui trouver un nom potable a cette fonction
	# j'ai l'impression que des fichiers prbh il y en a plein !
	# et la on en edite un seul.
	$self->ParsePrbhFile( $tmp_filename, $rh_params );

	return;
}

=head2 Procedure BlastToCluster

 Usage        : &BlastToCluster( $blast_out, $o_both_orgs, $rh_params, $tmp_filename );
 Prerequisite : none
 Procedure    : compute the command line for blast2cluster program for each threshold
 Args         : $blast_out, the blast output file
                $o_both, BothOrgs object
                $rh_params, ref to hash that contains blast parameters of the prog
                $tmp_filename, a string will be used as part of the filename

=cut

sub BlastToCluster
{
	my $self = shift ;
	my ( $rh_params, $tmp_filename ) = @_;

	my $blast_out = $self->GetFilename( 'blastp' );
	
	$self->Verbose( "debut BlastToCluster\n" );

	# une tranche de hash
	my( $min_span, $evalue ) = @$rh_params{ qw( min_span evalue ) };

	my $bank_q = $self->_Get( '_o_query' )->GetProtFastaFile();
	my $bank_s = $self->_Get( '_o_target' )->GetProtFastaFile();

	# les seuils
	my @a_cutoff = ( $min_span );
	for(my $cutoff=50;$cutoff<=100;$cutoff+=10)
	{
		push(@a_cutoff,$cutoff) if ( $min_span != $cutoff) ;
	}

	my @a_keys = qw( ROOT_BLAST2CLUSTER UNIX_GZIP );
	my( $prog, $gzip ) = $self->_Get( '_o_io' )->GetCfg( @a_keys );
	foreach my $cutoff (@a_cutoff)
	{
		my $prbh_file = $tmp_filename . $cutoff;
		my $cde = "$prog --sequence $bank_q,$bank_s";
		$cde .= " --collect $blast_out --dirout /tmp";
		$cde .= " --evalue $evalue --pcs $cutoff --pcq $cutoff";
		$cde .= " --famfile $prbh_file --result_format minimal --rbh";
		print STDERR "$cde\n\n";

		&NARCISSE_utils::RunCmd($cde);
		# ecrire directement dans un zip ?
		# http://search.cpan.org/~nwclark/PerlIO-gzip-0.17/gzip.pm
		&NARCISSE_utils::RunCmd( "$gzip $prbh_file");
	}

	$self->Verbose( "fin BlastToCluster\n" );

	return;
}


=head2 Procedure ParsePrbhFile

 Usage        : $rh_info_prot = &ParsePrbhFile( $tmp_filename, $o_both, $rh_params );
 Prerequisite : none
 Function     : parse a gzip file containing on each line the name of 2 proteins
              	it uses those names into the EditFileBestHit function.
 Returns      : a ref hash which keys are proteins names and containing information onto them.
 Args         :
 Globals      : none

=cut

sub ParsePrbhFile
{
	my $self = shift;
	my ( $tmp_filename, $rh_params ) = @_;

    my $o_io = $self->_Get('_o_io');
	my ( $org_q, $org_s )    = $self->GetBothShort();

    my $unix_mkdir = $o_io->GetUnixPath('UNIX_MKDIR');
    my $unix_rm    = $o_io->GetUnixPath('UNIX_RM');
	my $prbh_file = $tmp_filename . $$rh_params{ min_span };
	print STDERR "Parsing: $prbh_file\n";
    
	my @a_attrs = qw( _o_query _o_target );
	my( $o_query, $o_target ) = $self->_Get( @a_attrs );
	my( $query, $target ) = $self->GetBothShort();
	
    # creating the hierarchy if necessary, and removing any file found here
    my @a_paths = $o_io->GetComparativeDirpaths( 'SESSION,INSTANCE' );
    my $created_dir = "";
    foreach my $dir ( @a_paths )
    {
        `$unix_mkdir  $dir`;
        if ( -e $dir )
        {
            $created_dir = $dir;
            last;
        }
    }
    if ($created_dir eq '')
    {
        carp "ERROR - Cannot create directory - I tried " . join(' ',@a_paths);
    }
    
    # TODO - VERRUE - Il manque une fonction dans SpeciesObserver
    # Peut-etre GetcomparativeDirpath mais elle est privee 
    # manu
    
    my @a_directories = ();
    push (@a_directories, File::Spec->catfile( $created_dir, $query, $target, $o_io->GetCfg('ORTHOLOGUE') ) );
    push (@a_directories, File::Spec->catfile( $created_dir, $target, $query, $o_io->GetCfg('ORTHOLOGUE') ) );
    foreach my $d (@a_directories)
    {
        `$unix_mkdir $d`;
        `$unix_rm $d/*`;
    }
    
	my $fh =&NARCISSE_utils::GetFileHandle($prbh_file, 'r' );

	my $nb_prot_link = 0;
	while( my $line=<$fh> )
	{
		#ecoa:prot:NP_752639 ecob:prot:NP_415152
		#ecoa:prot:NP_752639 ecoa:prot:NP_415152	si self

		chomp($line);
		my ( $p1, $p2 ) =  grep( /\w+\:prot\:[\w\.]+/, split(/\s+/, $line) );

		if( !defined $p2 )
		{
			my $msg = "line $line is not in the expected format.\nformat should be :";
			$msg   .= "' $org_q:prot:prot_name $org_s:prot:prot_name'\n";
			print STDERR $msg;
			next;
		}

		$p1 =~ s/prot://g;
		$p2 =~ s/prot://g;

		my ($prot_q,$prot_s);
		if ( $p1 =~ /$org_q:/  && $p2 =~ /$org_s:/ )
		{
			$p1 =~ s/$org_q://;
			$p2 =~ s/$org_s://;
			($prot_q,$prot_s) = ( $p1,$p2 );
		}
		elsif ( $p1 =~ /$org_s:/ && $p2 =~ /$org_q:/ )
		{
			$p2 =~ s/$org_q://;
			$p1 =~ s/$org_s://;
			($prot_q,$prot_s) = ( $p2,$p1 );
		}
		else
		{
			print STDERR "unexpected organisms in >$line< $org_q,$org_s  : $p1 and $p2\n";
			next;
		}

		# inscrit ce meilleur appariement dans les fichiers des chrs concernes
		# ainsi que dans le fichier contenant toutes les lignes
		$self->EditFileBestHit( $prot_q, $prot_s );
		$nb_prot_link++;
	}

	$fh->close();
	$self->_Trace( "\tNumber of reciprocal best blast hit : $nb_prot_link\n",'Timer','A' );

	return;
}


=head2 procedure EditFileBestHit

 Usage        : EditFileBestHit($protA,$protB )
 Prerequisite : none
 Fonction     : Dump the "orthology" relationships (RBH based)
              : les noms des proteines sont lu dans le resultat de blast
              	les infos sur les proteines viennent du dossier de transcript
 Returns      : none
 Args         : $protA = symbol of the gene on the reference
 			  : $protB = symbol of the gene on the target
 Globals      : none

=cut

sub EditFileBestHit
{
	my $self = shift;
	my ( $protA, $protB ) = @_;

	my @a_attrs = qw( _o_query _o_target );
	my( $o_query, $o_target ) = $self->_Get( @a_attrs );
	my( $query, $target ) = $self->GetBothShort();

	my $rh_prot_info = $self->GetProtInfo();
	
	# Il peut arriver ( c'est rare ) que la proteine lu dans le fichier blast
	# ne soit pas presente dans le fichier des transcripts, car pas dans Genbank
	# Si tel est le cas, $$rh_prot_info{ $prot } n'aura que la clef len
	# car sa valeur est lu dans le fichier de sequences venant de ncbi.
	# On ecrira alors une alerte dans le fichier d'erreur
	# et on passera a la proteine suivante
	if( !defined $$rh_prot_info{$protA}->{chrom} )
	{
		warn "no data in transcripts for $protA\n"	;
		return;
	}
	if( !defined $$rh_prot_info{$protB}->{chrom} )
	{
		warn "no data in transcripts for $protB\n"	;
		return;
	}

	my $chrA   = $$rh_prot_info{$protA}->{chrom};
	my $startA = $$rh_prot_info{$protA}->{start};
	my $endA   = $$rh_prot_info{$protA}->{end};
	my $geneA  = $$rh_prot_info{$protA}->{element};
	my $strandA = $$rh_prot_info{$protA}->{strand};

	my $chrB   = $$rh_prot_info{$protB}->{chrom};
	my $startB = $$rh_prot_info{$protB}->{start};
	my $endB   = $$rh_prot_info{$protB}->{end};
	my $geneB  = $$rh_prot_info{$protB}->{element};
	my $strandB = $$rh_prot_info{$protB}->{strand};

	my $file_rbh_prot_ab = $self->GetQueryRBH();

    my $A_fh = $self->_OpenChrFileHandle( $o_query, $target, $chrA);
	my $prot_AB_fh = new IO::File ">>$file_rbh_prot_ab" or throw LipmError::IOException ($file_rbh_prot_ab,'f');

	my $sign = ($strandA eq $strandB) ? "+" : "-" ;

	print $A_fh "$chrA\t$protA\t$geneA\t$startA\t$endA\t$chrB\t$protB\t$geneB\t$startB\t$endB\n";
	print $prot_AB_fh "$chrA\t$startA\t$endA\t$chrB\t$startB\t$endB\t$sign\t1\t1\n";

	unless ( $self->SelfComparison() )
	{
		my $file_rbh_prot_ba = $self->GetSubjectRBH( );
		my $B_fh =  $self->_OpenChrFileHandle( $o_target, $query, $chrB );
		my $prot_BA_fh = new IO::File ">>$file_rbh_prot_ba";

		print $B_fh "$chrB\t$protB\t$geneB\t$startB\t$endB\t$chrA\t$protA\t$geneA\t$startA\t$endA\n";
		print $prot_BA_fh "$chrB\t$startB\t$endB\t$chrA\t$startA\t$endA\t$sign\t1\t1\n";

		$B_fh->close();
		$prot_BA_fh->close();
	}

	$A_fh->close();
	$prot_AB_fh->close();

	return;
}

=head2 procedure AddLengthAndSpanToBlastResults

 Usage        : &AddLengthAndSpanToBlastResults($blastoutputfile)
 Prerequisite : none
 Function     : Edit the blastoutput file in order to add the protein length and the spanning percentage
 Returns      : none
 Args         : $blastoutputfile: the blastoutput filename
 Globals      : %H_INFO_PROT

=cut

sub AddLengthAndSpanToBlastResults
{
	my $self = shift;

	my $blast_out = $self->GetBlastOutputFilename( );	
	my $rh_prot_info =  $self->GetProtInfo();

	# utiliser File::Temp ?
	my $tmp_file = $blast_out.'.'.$$;
    #`cp ${blast_out}.TMPSRC $blast_out`;
    #warn "KOUKOUKOU $tmp_file";
    
	&move( $blast_out, $tmp_file );
	#rename( $blast_out, $tmp_file );
# 	`mv $blastoutputfile $blastoutputfile.$$`;

	my	$fh_blast_src = &NARCISSE_utils::GetFileHandle( $tmp_file, 'r');
	my $fh_blast = &NARCISSE_utils::GetFileHandle( $blast_out, 'w' );
			
	while( my $line=<$fh_blast_src> )
	{
		# a refaire avec la classe d'acces aux fichiers ad hoc
		chomp($line);
        
		# $self->Verbose( $line );
		my ( $prot_q,$prot_s, $pci,$alilen,$mismatch,$gap, $qstart,$qend, $sstart,$send, $evalue,$bitscore ) = split( ' ', $line );
		my @a_blast_cols = ( $prot_q,$prot_s, $pci,$alilen,$mismatch,$gap, $qstart,$qend, $sstart,$send, $evalue,$bitscore );

		my $len_q = &_GetProtLen( $rh_prot_info, $prot_q );
		my $len_s = &_GetProtLen( $rh_prot_info, $prot_s );
		if ( $len_s > 0 && $len_q > 0 )
		{
			my $pcq = sprintf("%.0f",( abs($qend-$qstart+1)*100 ) / $len_q);
			my $pcs = sprintf("%.0f",( abs($send-$sstart+1)*100 ) / $len_s );
			my $line = join("\t",( @a_blast_cols, $len_q, $len_s, $pcq, $pcs));
			print $fh_blast $line . "\n";
		}
		else
		{
			print STDERR "Warning, failure getting the length of one protein: length $prot_q=$len_q, length $prot_s=$len_s, skipped\n";
		}
	}
	$fh_blast->close();
	$fh_blast_src->close();
	unlink( $tmp_file );

	return;

=head2 function _GetProtLen

 Usage        : $length = &_GetProtLen( $rh_prot_info, $prot_name );
 Prerequisite : none
 Function     : accessor to the length of the protein length
 Returns      : $length, the length of the protein whose name is $prot_name
 Args         : $rh_prot_info, a hash ref which keys are protein names
               	$prot_name, the name of the protein we're interested in
 Globals      : none

 Note : I could pass $$rh_prot_info{ $prot_name }
        but i would rather use a class object.

=cut

	sub _GetProtLen
	{
		my ( $rh_prot_info, $prot_name ) = @_;
			
		# suppression du "parasite"
		$prot_name =~ s/prot://;

		my $protein_len = ( defined $$rh_prot_info{ $prot_name }{len} ) ? $$rh_prot_info{ $prot_name }{len} : 0;
		
		return $protein_len;
	}
}

=head2 procedure FromBlastp2NarcisseFormat

 Usage        : ( $blast_out, $o_both)
 Prerequisite : the blast_out must be a valid blastp output
 Fonction     : copy the blastout file into the correct repository and purge the prot: prefix
 Returns      : a filename with the genome hsps
 Args         : none

=cut

sub FromBlastp2NarcisseFormat
{
	my $self = shift;

	my $blastout =  $self->GetBlastOutputFilename( );	
	my $nc_blastpout = $self->GetFilename( 'blastp' );	

	$self->Verbose("Starting FromBlastp2NarcisseFormat");

	my $fh_blastout =  &NARCISSE_utils::GetFileHandle( $blastout,  'r' );
	my $fh_nc_blastpout = &NARCISSE_utils::GetFileHandle( $nc_blastpout , 'w');	

	print $fh_nc_blastpout "#".join("\t",(BLASTP_NC_HEADER))."\n";

	while( my $line = <$fh_blastout> )
	{
		$line =~ s/prot\://g; 
		print $fh_nc_blastpout $line;
	}

	$fh_nc_blastpout->close();
	$fh_blastout->close();	

	#Always zip the files	
	&NARCISSE_utils::Zip( $nc_blastpout );

	return;
}


=head2 Function  HspProjection

 Usage        : HspProjection( blast_output, map)
 Prerequisite : The pairwise object should have been correctly initialize, 
                hence with a valid blastp output file and a map file fot exon structure
 Fonction     : Transforms the list of proteic hsps into a list of genomeic hsps (kind of reverse transcription of hsp) 
 Returns      : a reference on a list with the corrsponding genomic hsps
 Args         : none

=cut

sub HspProjection
{
	my $self = shift;

	my $rh_blast_output = $self->GetBlastOutput();
	my $rh_maps = $self->GetProtInfo();

	my $min_span   = $self->_Get( 'filter')->{ min_span };
	my $max_evalue = $self->_Get( 'filter')->{ max_evalue };

	my @hsps = ();
	my $num_hsp=0;

	foreach my $ra_blast_output ( values %$rh_blast_output )
	{
		foreach my $rh_hsp ( @$ra_blast_output )
		{
			my $q_name = $rh_hsp->{q_name};  #Same as $query
			my $s_name = $rh_hsp->{s_name};

			my $rh_map_q = $$rh_maps{ $q_name };
			my $rh_map_s = $$rh_maps{ $s_name };

			# Proteins with mapping (hence no structure) information are useless
			next unless ( defined $$rh_map_q{ start } && defined $$rh_map_s{ start } );

			# We are not going to deal with inconsistent proteins
			my $query_is_consistent = $$rh_map_q{consistency} eq 'C';
			my $subjt_is_consistent = $$rh_map_s{consistency} eq 'C';
			next unless( $query_is_consistent && $subjt_is_consistent );

			#Checking the coverage of the hsp according to the protein
			next unless ( $min_span < $rh_hsp->{q_cov} && $min_span < $rh_hsp->{s_cov} );

			#Filtering on the evalue
			next unless ( $rh_hsp->{evalue} < $max_evalue );

			# When comparing a genome against himself we are only interested in one half of the diagonal matrix
			next if ( $self->SelfComparison() &&
						( &NARCISSE_utils::ChromIDComparison( $$rh_map_q{chrom},$$rh_map_s{chrom})>0 ||
						( $$rh_map_q{chrom} eq $$rh_map_s{chrom}  &&
						$$rh_map_q{start} >= $$rh_map_s{start} ) )
						);

			$rh_hsp->{cluster} = ++$num_hsp;
			my @decomposed_hsp = &ProtHsp2GenomeHsp( $rh_hsp, $rh_map_q, $rh_map_s );

			push @hsps, @decomposed_hsp;
		}
	}

	return \@hsps;
}

=head2 function  ProtHsp2GenomeHsp

 Usage        : @a_hsps = ProtHsp2GenomeHsp( $hsp, $rh_map_q, $rh_map_s )
 Fonction     : Transforme a single hsp into its genomeic part according to the transcripts structure on both genomes
 Returns      : a list of genome hsps
 Args         : a reference on a hsp and references on the query transcript structure and subject transctipy structure

=cut

sub ProtHsp2GenomeHsp
{
	my( $rh_hsp, $rh_query_prot, $rh_subjt_prot ) = @_;

	my @a_rh_prots = ( $rh_query_prot, $rh_subjt_prot );

	#First transforming the hsp into a transcript hsp	
	Pairwise->Verbose("From protein to transcript");	
	my $transcript_hsp = &ProtHsp2TranscriptHsp( $rh_hsp, @a_rh_prots );

	#Decomposition of the transcript hsp into virtual exons
	my $ra_query_involved_exons = &TranscriptPartition( $transcript_hsp, @a_rh_prots, "query" );
	
	Pairwise->Verbose("Split query thsp");	
	my $ra_query_hsps = &SplitHsp( $ra_query_involved_exons, $transcript_hsp, "query");

	my @a_transcript_hsps = ();
	my @a_genome_hsps     = ();

	Pairwise->Verbose("Split subject hsp");
	foreach my $hsp ( @$ra_query_hsps )
	{
		my $ra_subject_involved_exons = &TranscriptPartition( $hsp, @a_rh_prots, "subject" );
		my $ra_subject_hsps=SplitHsp( $ra_subject_involved_exons, $hsp, "subject" );

		push @a_transcript_hsps, @$ra_subject_hsps;
	}
	Pairwise->Verbose(scalar @a_transcript_hsps." transcript hsp");

	foreach my $hsp ( @a_transcript_hsps )
	{
		push @a_genome_hsps, &TranscriptHsp2Genome( $hsp, @a_rh_prots );
	}

	return @a_genome_hsps;
}


=head2 Function  ProtHsp2TranscriptHsp

 Usage        : $r_hsp= ProtHsp2TranscriptHsp( $rh_hsp, $rh_query_prot, $rh_subject_prot)
 Fonction     : Transforms a protein hsp into its corresponding transcript hsp (coordinates are now on the transcript) 
 Returns      : a reference on this transcript hsp
 Args         : a reference on a hsp and references on the query transcript structure and subject transcritp structure

=cut

sub ProtHsp2TranscriptHsp
{
	my( $rh_hsp, $rh_query_prot, $rh_subjt_prot ) = @_;
	
	my $query_start = &Protein2Transcript( $rh_hsp->{q_start}, $rh_query_prot );
	my $query_end   = &Protein2Transcript( $rh_hsp->{q_end},   $rh_query_prot );

	my $subjt_start = &Protein2Transcript( $rh_hsp->{s_start}, $rh_subjt_prot );
	my $subjt_end   = &Protein2Transcript( $rh_hsp->{s_end},   $rh_subjt_prot );
	
	my %h_hsp = (	q_start => $query_start, q_end => $query_end,
					s_start => $subjt_start, s_end => $subjt_end );

	my @a_immuable_fields = qw( pident size mismatch gaps evalue score q_name s_name cluster );
	@h_hsp{ @a_immuable_fields } = @$rh_hsp{ @a_immuable_fields };

	return \%h_hsp;
}


=head2 Function : TranscriptPartition

 Usage          : $rh_virtual_exons = _TranscriptPartition( hsp, $rh_query_prot, $rh_subjt_prot, $org
 Fonction       : Transform a given interval on the transcript on a list of transcript intervals with exon borders as borders
                  a kind of Reverse-Transcriptase behavior
 Returns        : a list of potentially truncated exons containing both transcript and genomes coordinates
 Args           : an interval ?<-- a expliquer

=cut

sub TranscriptPartition
{
	my( $rh_hsp, $rh_query_prot, $rh_subjt_prot, $side ) = @_;
	
	my( $StartField, $EndField, $rh_prot );

	if ( $side eq "query" )
	{
		$StartField = "q_start";
		$EndField   = "q_end";
		$rh_prot    = $rh_query_prot;
	}
	else
	{
		$StartField = "s_start";
		$EndField   = "s_end";
		$rh_prot    = $rh_subjt_prot;
	}

	my $rh_transcript_pos = { start => $rh_hsp->{ $StartField },
							 end   => $rh_hsp->{ $EndField } };

	return  &_TranscriptPartition( $rh_transcript_pos, $rh_prot );
}


=head2 Function  TranscriptHsp2Genome

 Usage        :  $r_hsp= TranscriptHsp2Genome( $rh_hsp, $rh_query_prot, $rh_subject_prot)
 Prerequisite : The hsp givne in argument must be restricted to a single exon on both sides
 Fonction     :  Transforms a transcript hsp into its corresponding genomic hsp (coordinates are now on the genome) 
 Returns      : a reference on the genome hsp
 Args         : a reference on a hsp and references on the query transcript structure and subject transcritp structure

=cut

sub TranscriptHsp2Genome
{
	my $rh_hsp = shift;
	my $rh_query_prot = shift;
	my $rh_subject_prot = shift;
		
	my $query_start = &Transcript2Genome( $rh_hsp->{q_start}, $rh_query_prot );
	my $query_end   = &Transcript2Genome( $rh_hsp->{q_end}, $rh_query_prot );
	my $subjt_start = &Transcript2Genome( $rh_hsp->{s_start}, $rh_subject_prot );
	my $subjt_end   = &Transcript2Genome( $rh_hsp->{s_end}, $rh_subject_prot );

	if ( $rh_query_prot->{strand} eq "-1" )
	{
		( $query_start, $query_end ) = ( $query_end, $query_start );
# 		my $temp     = $query_start;
# 		$query_start = $query_end;
# 		$query_end   = $temp;

		if ( $rh_subject_prot->{strand} eq "-1")
		{
			( $subjt_start, $subjt_end ) = ( $subjt_end, $subjt_start );
# 			my $temp     = $subject_start;
# 			$subjt_start = $subject_end;
# 			$subjt_end   = $temp;
		}
	}

	#Suppressed in order to keep the information of the hit orientation
	#if ( $rh_subject_prot->{strand} eq "-1")
	#{
	#	my $temp = $subject_start;
	#	$subject_start = $subject_end;
	#	$subject_end = $temp;
	#}

	my $q_size = abs( $query_end - $query_start );
	my $s_size = abs( $subjt_end - $subjt_start );
	my $size = ( $q_size > $s_size ) ? $q_size : $s_size;
	my %hsp = ( q_start => $query_start, q_end => $query_end,
	            s_start => $subjt_start, s_end => $subjt_end,
	            size => $size );
	my @immuable_fields= qw(pident mismatch gaps evalue score q_name s_name cluster);

	@hsp{ @immuable_fields } = @{ $rh_hsp }{ @immuable_fields };

   return \%hsp;
}


=head2 Function SplitHsp

 Usage        : $ra_hsps = SplitHsp( $ra_involved_exons, $hsp, $side );
 Fonction     : Given a partition of the transcript on one side (according to the transcript structure)
                splits the hsps into parts belonging each to a single exon
 Returns      : a reference on a list of hsps (restricted on one side to a single exon each)
 Args         : a reference on a list of exons involved in the hsps, a reference on the hsp, a side ("query" or "subject")

=cut

sub SplitHsp
{
	my $ra_exons = shift ;
	my $rh_hsp = shift;
	my $side = shift;

	my @a_new_hsps=();

	my ($Q_Start, $Q_End, $S_Start, $S_End );
	if ( $side eq "query" )
	{
		$Q_Start = "q_start";
		$Q_End = "q_end";
		$S_Start = "s_start";
		$S_End = "s_end";
	} else
	{
		$Q_Start = "s_start";
		$Q_End = "s_end";
		$S_Start = "q_start";
		$S_End = "q_end";
	}

	#We are dealing with virtual exons corresponding exactly to the partition of the hsp on the ("query"|"subject") side
	foreach my $exon ( @$ra_exons )	
	{
		my $q_start = $exon->{texon}{start};
		my $q_end = $exon->{texon}{end};
				
		my $s_start= $rh_hsp->{ $S_Start } + ( $q_start - $rh_hsp->{ $Q_Start } );
		my $s_end  = $rh_hsp->{ $S_Start } + ( $q_end - $rh_hsp->{ $Q_Start } );
				
		my $denominator = ($rh_hsp->{ $Q_End } -$rh_hsp->{ $Q_Start });
				
		my $ratio = ($denominator > 0 ) ? (($q_end - $q_start)/($denominator)) : 1;
		
		#We have to stay within the bounds (this is completely arbitrary !!)
		$s_start = ( $s_start > $rh_hsp->{ $S_End } ) ? $rh_hsp->{ $S_End } : $s_start;
		$s_end   = ( $s_end > $rh_hsp->{ $S_End } ) ? $rh_hsp->{ $S_End } : $s_end;
		
		my %new_hsp = (  );
		@new_hsp{($Q_Start, $Q_End, $S_Start, $S_End )} = ( $q_start, $q_end, $s_start, $s_end );
		
		$new_hsp{score} = $rh_hsp->{score}*$ratio;
		$new_hsp{size}  = $rh_hsp->{size}*$ratio;
		my @immuable_fields= qw(pident mismatch gaps evalue q_name s_name cluster);
		@new_hsp{ @immuable_fields } = @{ $rh_hsp }{ @immuable_fields };
				
		push @a_new_hsps, \%new_hsp;
	}
	
	return \@a_new_hsps;
}


=head2 Function : _TranscriptPartition

 Usage          : $rh_virtual_exons = _TranscriptPartition( $rh_transcript_pos, $rh_prot )
 Fonction       : Transform a given interval on the transcript on a list of transcript intervals with exon borders as borders
                  a kind of Reverse-Transcriptase behavior
 Returns        : a list of potentially truncated exons containing both transcript and genomes coordinates
 Args           : an interval ?<--

=cut

sub _TranscriptPartition
{
	my( $rh_interval, $rh_transcript_info ) = @_;

	my( $start, $end ) = @$rh_interval{ qw( start end) };

	#Getting the first and last involved exons
	my $start_exon = &GetCorrespondingExon( $rh_transcript_info, $start );
	my $end_exon   = &GetCorrespondingExon( $rh_transcript_info, $end   );

	my $ra_transcript_exons = $rh_transcript_info->{transcript_exons};
	my @a_involved_exons = @$ra_transcript_exons[ $start_exon..$end_exon ];

	#Copying all the exons in a new array in order not to overwrite the data
	my @involved_parts = ();
	foreach my $exon ( @a_involved_exons )
	{
		my $gexon={};
		@{$gexon}{("start", "end")}=@{$exon->{gexon}}{("start", "end")};
		my $texon={};
		@{$texon}{("start", "end")}=@{$exon->{texon}}{("start", "end")};
		push @involved_parts, { gexon => $gexon, texon => $texon};
	}

	# Clearing the nucleotides of the first and last exon not involved in the interval	
	my $last_exon = (scalar @a_involved_exons) -1;
	my $excluded_5prime = $start-$a_involved_exons[0]{texon}{start} ;
	my $excluded_3prime = $a_involved_exons[$last_exon]{texon}{end}-$end;
		
	$involved_parts[0]{texon}{start} += $excluded_5prime ;
	$involved_parts[$last_exon]{texon}{end} -= $excluded_3prime ;
	if ( $rh_transcript_info->{strand} eq "1")
	{
		
		$involved_parts[0]{gexon}{start} += $excluded_5prime ;
		$involved_parts[$last_exon]{gexon}{end} -= $excluded_3prime ;
		
	} else
	{
		$involved_parts[$last_exon]{gexon}{start} += $excluded_3prime ;
		$involved_parts[0]{gexon}{end} -= $excluded_5prime ;
	}	
	
	return \@involved_parts;
}


=head2 Function GetCorrespondingExon

 Usage        : &GetCorrespondingExon( $rh_transcript_info, $pos )
 Fonction     : Detect the exon corresponding to a transcript position and returns its number
 Returns      : the order number of and exon
 Args         : a reference on a hash constructed from the transcript Narcisse file (see documentaion) and a position on the transcript

=cut

sub GetCorrespondingExon
{
	my $rh_transcript_info = shift; 
	my $pos = shift;

	my $match_info={};

	( 0 < $pos )
		|| die "$pos outside of transcript !! in GetCorrespondingExon";

	my $selected_exon;
	EXON : foreach my $exon (@{$rh_transcript_info->{transcript_exons}})
	{
		$selected_exon=$exon;
		last EXON if ($pos <= $selected_exon->{texon}{end});
	}

# 	unless ( $pos <= $selected_exon->{texon}{end} )
	if( $selected_exon->{texon}{end} < $pos )
	{
		#print STDERR join("\t",@{$HSP}{(keys %$HSP)})."\n";
		Pairwise->Verbose("Warning: $pos outside of transcript $selected_exon->{texon}{end}!! in GetCorrespondingExon");
	}

	return $selected_exon->{num};
}


sub Protein2Transcript
{
	my $pos = shift;
	my $rh_prot = shift;
	
	#Start translation is the first nucleotide of the first amino acid
	return 	$rh_prot->{start_translation} + ($pos-1)*3;
}


sub Transcript2Protein
{
	my $pos = shift;
	my $rh_prot = shift;
	
	#Checking the pposition according to the coding part
	if ( $pos < $rh_prot->{start_translation} || $pos > $rh_prot->{stop_translation} )
	{
		die "$pos outside the coding part of the transcript is simply not possible (Transcript2Protein)";
	}
	
	return ( $pos - $rh_prot->{start_translation} + 1 ) / 3	;
}

sub Protein2Protein
{
	my $q_pos =shift;
	my $rh_hsp= shift;
	
	return $rh_hsp->{s_start} + ( $q_pos - $rh_hsp->{q_start} );
}

=head2 Function Transcript2Genome

 Title        : Transcript2Genome
 Usage        : Transcript2Genome
 Fonction     : Given a position on the transcript returns the corresponding position on the genome
 
 Returns      : the corresponding positin on the genome
 Args         : a position on the transcript and a reference on a hash with the transcript information

=cut
sub Transcript2Genome
{
	my $pos = shift;
	my $rh_transcript = shift;
	
	my $genome_pos = undef;
	
	#Getting the first and last involved exons
	my $exon_num = &GetCorrespondingExon( $rh_transcript, $pos );

	my $exon = $rh_transcript->{transcript_exons}[$exon_num];

	if ( $rh_transcript->{strand} eq "1")
	{		
		$genome_pos = $exon->{gexon}{start} +  ( $pos - $exon->{texon}{start} );				
	} else
	{
		$genome_pos = $exon->{gexon}{start} +  ( $exon->{texon}{end} - $pos );		
	}	
	
	return $genome_pos;
}

#-----------   END OF The Proteic comparison  (methods dedicated to)  -----------------#

=head2 function WithNucleic

 Usage        : if ( $self->WithNucleic() )
 Function     : returns true if the method includes the computation and use of nucleic alignments
 Returns      : 0|1
 Args         : none

=cut

sub WithNucleic 
{
	my $self = shift;

    my $o_param   = $self->_Get('_o_param');
	my $molecules = $self->_Get( '__comparedmolecules' );
    
    return 0 if ( $o_param->IsDefined(SELF) );
	return ( $molecules eq 'genomic' || $molecules eq 'nucleic' );
}

=head2 function WithProteic

 Usage        : if ( $self->WithProteic() )
 Function     : returns true if the method includes the computation and use of proteic alignments
 Returns      : 0|1
 Args         : none

=cut
sub WithProteic 
{
	my $self = shift;

	my $molecules = $self->_Get( '__comparedmolecules' );

	return ( $molecules eq 'genomic' || $molecules eq 'proteic');
}


=head2 function GetQueryRBH

 Usage        : $data_filename = &GetQueryRBH( $o_pp );
 Prerequisite : none
 Function     : compute the rbh filename of the reference (query)
 Returns      : $data_filename
 Args         : $o_pp, a ParamParser object
 Globals      : none

=cut

sub GetQueryRBH
{
	my $self = shift;

	my( $filename, $o_param ) = $self->_Get( qw( __rbh_file_QS _o_param ) );

	my $data_filename = $filename;
	$data_filename   .= '.' . $o_param->Get( 'min_span' );
	$data_filename   .= '.' . $o_param->Get( 'e' );

	return $data_filename;
}


=head2 function GetSubjectRBH

 Usage        : $data_filename = &GetSubjectRBH( $o_pp );
 Prerequisite : none
 Function     : compute the rbh filename of the target (subject)
 Returns      : $data_filename
 Args         : $o_pp, a ParamParser object
 Globals      : none

=cut

sub GetSubjectRBH
{
	my $self = shift;

	my( $filename, $o_param ) = $self->_Get( qw( __rbh_file_SQ _o_param ) );

	my $data_filename = $filename;
	$data_filename   .= '.' . $o_param->Get( 'min_span' );
	$data_filename   .= '.' . $o_param->Get( 'e' );

	return $data_filename;
}

=head2 function GetRBHOutputfilename

 Usage        : &GetRBHOutputfilename();
 Function     : accessor

=cut

sub GetRBHOutputfilename
{
	my $self = shift;

	my $filename = $self->_Get( '__rbh_output_filename' );

	(defined $filename )
		or throw LipmError::ObjectException( $self, '!attribute', -text => "__rbh_output_filename is undefined at this point" );

	return $filename;
}

=head2 function GetBlastOutput

 Usage        : &GetBlastOuput()
 Function     : accessor

=cut

sub GetBlastOutput
{
	my $self = shift;

	my $output = $self->_Get( '__rh_blast_output' );

	(defined $output )
		or throw LipmError::ObjectException( $self, '!attribute', -text => "__rh_blast_output is undefined at this point" );
	
	return $output;
	
}

=head2 function GetBlastOutputFilename

 Usage        : &GetBlastOuputFilename()
 Function     : accessor 

=cut

sub GetBlastOutputFilename
{
	my $self = shift;

	my $blast_output = $self->_Get( '__blastoutput_filename' );

	(defined $blast_output )
		or throw LipmError::ObjectException( $self, '!attribute', -text => "__blastoutput_filename is undefined at this point" );

	return $blast_output;
}


=head2 function GetProtInfo

 Usage        : $rh_protinfo=&GetProtInfo();
 Function     : accessor

=cut

sub GetProtInfo
{
	my $self = shift;

	my $rh_proteins = $self->_Get( '__rh_protein_info' );

	( defined $rh_proteins )
		or throw LipmError::ObjectException( $self, '!attribute', -text => "__rh_protein_info is undefined at this point" );

	return $rh_proteins;
}


=head2 function _CheckGlintIndex

 Usage        : &_CheckGlintIndex()
 Prerequisite : none
 Function     : setting the list of the glint indexes
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub _CheckGlintIndex
{
	my $self = shift;	

	my( $o_query, $o_target ) = $self->_Get( qw( _o_query _o_target ) );

    my $o_io = $self->_Get( '_o_io' );
	
	my $data_dir=$o_io->GetCfg('ROOT_DATA');
	
	foreach my $o_species ( $o_query, $o_target )
	{
		my $ra_virtual_genomes = $o_species->GetChromIndexFilepaths( -ext => 'seq' );
		my $short = $o_species->GetShortName();

		if( 0 == @$ra_virtual_genomes )
		{
			print "ERROR, no virtual genome found for $short\n" and die;
		}

		# OK, keep track of the virtual genomes
		my $attr_name = '__' . $short . '_virtual_genomes';
		$self->_Set( $attr_name => $ra_virtual_genomes );
	}

	return;
}

=head2 function GetVirtualGenomes

 Usage    : &GetVirtualGenomes()
 Function : accessor

=cut

sub GetVirtualGenomes
{
	my $self = shift;
	my $org = shift;

	LipmError::ParameterException::Assert( org => $org );

	my $genome_key = "__".$org."_virtual_genomes";

	return $self->_Get( $genome_key );
}


=head2 procedure FromProteinToGenome

 Usage        : $rh_info_prot = &FromProteinToGenome( $blast_out, $o_both)
 Prerequisite : the blast_out must be a valid (uncompressed) blastp output
 Fonction     : transforms protein hsps into genome hsps
 Returns      : a filename with the genome hsps
 Args         : none

=cut

sub FromProteinToGenome
{
	my $self = shift ;

	$self->Verbose("Starting FromProteinToGenome");

	$self->_LoadBlastpOutput( );

	my $ra_hsps=$self->HspProjection();

	$self->WriteGenomeOutput( $ra_hsps );

	return;
}

=head3 function __SetBlastOutputFilename

 Usage        : $self->_SetBlastOutputFilename();
 Prerequisite : none
 Function     : read the blastresults param if it exists
                	otherwise compute it
 Returns      : defines the the blast output filename
 Args         : none

=cut

sub __SetBlastOutputFilename
{
	my $self = shift ;

	my $rawresult_dir = $self->GetRawResultDir();
	my $o_param       = $self->_Get( '_o_param' );

	my $br = 'blastresults';
   
	my $blast_out;
	if( $o_param->IsDefined( $br ) )
	{
        $o_param->AssertFileExists( $br ); 
		$blast_out = $o_param->Get( $br );
	}
	else
	{
		my ( $orgQ, $orgS ) = $self->GetBothShort();
			$blast_out = File::Spec->catfile( $rawresult_dir, "raw_blastp.out" );
			#$blast_out .= $orgQ . '_' . $orgS . '_';
			#$blast_out .= $$ . '.' . $self->{ o_param }->Get('e');
			#$blast_out .= '.G' . &__GetDefault( 'O_GAP' );
			#$blast_out .= '.E' . &__GetDefault( 'E_GAP' );
		}

	$self->_Set( __blastoutput_filename => $blast_out );

	return $blast_out;
}

=head2 function _PrepareBlast

 Usage        : %h_params = &_PrepareBlast( $o_both_orgs );
 Prerequisite : none
 Function     : Read blast params from o_param and return them into a hash
                It also print a trace into the log file
                and create a new empty directory for orthologue data
                	eventually erase previous one.
 Returns      : %h_params
 Args         : a lot

=cut

sub _PrepareBlast
{
	my $self = shift ;

	my @a_attr = qw( _o_param _o_query _o_target );
	my( $o_param, $o_query, $o_target ) = $self->_Get( @a_attr );

	my %h_p = (
			evalue     => $o_param->Get('e'),
			o_gap      => $o_param->Get('G'),
			e_gap      => $o_param->Get('E'),
			l_db       => $o_param->Get('z'),
			min_span   => $o_param->Get('min_span'),
			cpu        => $o_param->Get('a'),
			path_blast => $o_param->Get('path_blastall')
		);

	my $orgQ = $o_query->GetShortName();
	my $orgS = $o_target->GetShortName();

	my $data_q = $o_query->GetDataPath();
	my $data_s = $o_target->GetDataPath();

	$self->_Trace("Parameters :\n\n",'Timer','A');
	$self->_Trace("\tOrga A                  : $orgQ\n",'Timer','A');
	$self->_Trace("\tOrga B                  : $orgS\n",'Timer','A');
	$self->_Trace("\tDirectory data A        : $data_q\n",'Timer','A');
	$self->_Trace("\tDirectory data B        : $data_s\n",'Timer','A');
	$self->_Trace("\tE-Value (-e)            : $h_p{ evalue }\n",'Timer','A');
	$self->_Trace("\tOpen Gap (-G)           : $h_p{ o_gap }\n",'Timer','A');
	$self->_Trace("\tExtend Gap (-E)         : $h_p{ e_gap }\n",'Timer','A');
	$self->_Trace("\tLength of DataBase (-z) : $h_p{ l_db }\n",'Timer','A');
	$self->_Trace("\tCPU (-a)                : $h_p{ cpu }\n",'Timer','A');
	$self->_Trace("\tCut Of                  : $h_p{ min_span } %\n",'Timer','A');

	#Nothing to do here !!
	#&_NewEmptyDir( $orth_q );
	#&_NewEmptyDir( $orth_s );

	return %h_p;


# 	sub _NewEmptyDir
# 	{
# 		my ( $path_dir ) = @_;
# 
# 		# like rm -rf
# 		rmtree( $path_dir );
# 
# 		# like mkdir -p
# 		mkpath( $path_dir );
# 	}

} # _PrepareBlast


=head2 function _GetProtInfoFromFiles

 Usage        : $rh_info_prot = &GetProtInfoFromFiles()
 Prerequisite : none
 Fonction     : Initialize a hash with info related to the proteins (reference and target)
 Returns      : a ref hash which keys are proteins name and containing information about them.
 Args         : none

=cut

sub _GetProtInfoFromFiles
{
	my $self = shift;

	my $rh_info_prot = {};

	# Recupere tous les ID des proteines des deux organismes.
	&_GetProteinID( $self->_Get( '_o_query' ), $rh_info_prot );

	# on ne le fait pas 2 fois en cas de self
	if( $self->SelfComparison() == &FALSE )
	{
		&_GetProteinID( $self->_Get( '_o_target' ), $rh_info_prot );
	}

	return $rh_info_prot;


=head2 Procedure _GetProteinID

 Usage        : &_GetProteinID( $data_dir, $rh_info );
 Prerequisite : none
 Procedure    : fill the $rh_info with data found into $data_dir
 Args         : $data_dir, string, the data directory of a given org
                $rh_info, a have that will received information

=cut

	sub _GetProteinID
	{
		my ( $o_species, $rh_info ) = @_;

		#my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

		my @a_files = $o_species->GetTranscriptFiles();
		foreach my $file ( @a_files )
		{
			#warn "LECTURE FICHIER $file";
			my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
			$o_ndr->Prepare( -from  => $file );
			my $rh_data = $o_ndr->FetchallHashref( 'protein_id' );

			#All protein info from $rh_local_info are concatenated at once into $rh_info
			#rh_info will therefore contain information about proteins from all the chromosomes in both species
			#@{%$rh_info}{keys %$rh_local_info}=@{%$rh_local_info}{keys %$rh_local_info};
			#warning the previous syntax is not recognized by perl 5.10.0

			my @a_keys = keys %$rh_data;
			@$rh_info{ @a_keys } = @$rh_data{ @a_keys };
			
			$o_ndr -> Finish();
		}

# TODO find a solution instead of reading the fasta file : This stinks
#      Utiliser plutot les fichiers prot pour les deux trucs !
=pod		
		foreach my $prot (keys %$rh_info)
		{
			my $protein_size=($rh_info->{ $prot }{ stop_translation }-$rh_info->{ $prot }{ start_translation }+1)/3 - 1;
			$$rh_info{ $prot }{ size } = $protein_size;
		}
=cut		

		# analyse l'en-tete des sequences fasta pour recuperer la longueur.
		my $fasta_file = $o_species->GetProtFastaFile();
		my $fh_fasta = new IO::File( $fasta_file, 'r' );
		while ( my $line = <$fh_fasta> )
		{
			next	if( $line !~ /^>prot:(\S+)/ );
			my $prot = $1;
			my ( $len ) = ( $line =~ /\slen=(\d+)\s/ );
			$$rh_info{ $prot }{ len } = $len;
		}
		close($fh_fasta);

		return;
	} # _GetProteinID

}	#GetInfoProt


=head2 procedure _SetTranscriptInfo

 Usage        : &_SetTranscriptInfo( $rh_mapping )
 Fonction     : Modify transcript information, establishing correspondances bewteen genome exons and transcript exons
 
 Returns      : none
 Args         : a reference on a hash constructed from the transcript Narcisse file

=cut

sub _SetTranscriptInfo
{
	my $self = shift;

	my $rh_maps = $self->GetProtInfo();

	foreach my $prot ( keys %$rh_maps )
	{
		my $rh_info = $rh_maps->{ $prot };
		
		(defined $rh_info->{regions})
			or ( $self->Verbose("$prot absent from transcript file !! ") and next );

		my @raw_exons= split( ",", $rh_info->{regions} );
		my @exons;

		if ( $rh_info->{strand} == 1 )
		{
			@exons = @raw_exons; 
		}
		else
		{
			@exons = reverse @raw_exons;
		}
		my @transcriptexons=();
		my( $start, $num ) = ( 1, 0 );

		foreach my $exon ( @exons )
		{
			my( $exon_start, $exon_end ) = split( /\.{2}/, $exon );

			my $end = $start + $exon_end - $exon_start;

			push @transcriptexons, { texon => {start => $start, end => $end},
									 gexon => {start => $exon_start, end => $exon_end},
									 num => $num++ };
			$start = $end + 1;
		}

		$$rh_maps{ $prot }{ transcript_exons } = \@transcriptexons;
	}

	return;
} # _SetTranscriptInfo


=head2 function WriteGenomeOutput

 Usage        : &WriteGenomeOutput( $rh_blastout, $rh_mapping)
 Fonction     : Transforms blast hits position information (q_start, q_end, ...) into genome position information

 Returns      : none
 Args         : two reference on hashes provided by 1) load_blast_output and 2) by load_tab_file

=cut

sub WriteGenomeOutput
{
	my $self = shift;
	my $ra_hsps = shift;

	my $rh_maps = $self->GetProtInfo();
	my $outputfile = $self->GetFilename( 'blastp2genome' );

	my $fh=&NARCISSE_utils::GetFileHandle( $outputfile, 'w' );

	my @fields = qw (q_name q_prot q_start q_end s_name s_prot s_start s_end score identity evalue size cluster); #_ _
	print $fh "#".join("\t",@fields)."\n";

	foreach my $hsp (@{$ra_hsps})
	{
		my $q_name = $hsp->{q_name};
		my $s_name = $hsp->{s_name};
		my $line="$rh_maps->{$q_name}{chrom}\t";
		$line.=join("\t",@{$hsp}{qw(q_name q_start q_end)});
		$line.="\t$rh_maps->{$s_name}{chrom}\t";
		$line.=join("\t",@{$hsp}{qw(s_name s_start s_end )});
		$line.=sprintf("\t%d\t%3.2f\t%3.2f\t%d\t%d",@{$hsp}{qw(score pident evalue size cluster)});
		print $fh $line."\n";
	}
	$fh->close();

	#Always zip the files
	&NARCISSE_utils::Zip( $outputfile );
}


=head2 function _LoadBlastpOutput

 Usage        : &load_blast_output($file)
 Fonction     : load a tab formated -m 8 blast output into a hash
 Returns      : the content of the tab formated table in a hash returned by reference
 Args         : $file the complete path to the blast output

=cut

sub _LoadBlastpOutput 
{
	my $self =shift;

	my $blastp_output = $self->GetFilename( 'blastp' );

	my $fh_blastp_output = &NARCISSE_utils::GetFileHandle( $blastp_output, 'r' );

	my $header = <$fh_blastp_output>;
	chomp($header);
	$header =~ s/\#//;
	my @fields = split( /\t/, $header );

	my %h_blastoutput;
	# pourquoi il y a un label ? <- seb
	LINE : while ( my $line = <$fh_blastp_output> )
	{
		chomp($line);
		my @champs = split( /\t/, $line );
		my $accession=$champs[0];
		my %rh=();
		@rh{@fields}=@champs;
		push @{$h_blastoutput{$accession}}, \%rh;
		#print STDERR "$line \n;"
	}

	$self->_Set( __rh_blast_output => \%h_blastoutput );

	return;
}


=head2 Procedure _LoadProteinInfo

 Usage        : _LoadProteinInfo
 Fonction     : loads the protein infos from the transcript files of the Narcisse tree
 Returns      : none
 Args         : none
 TODO Si deja execute, pas la peine de recommencer !

=cut

sub _LoadProteinInfo 
{
	my $self = shift;

	# protein information
	$self->_Set( '__rh_protein_info', $self->_GetProtInfoFromFiles( ));
	$self->_SetTranscriptInfo();

	return;
}

=head2 function GetPrbhResultDir

 Usage        : $o_p->GetPrbhResultDir()
 Prerequisite : none
 Function     : accessor
 Returns      : the path to the prbh result dir ( see __SetResultTree )
 Args         : none

=cut

sub GetPrbhResultDir
{
	my $self = shift;

	return $self->_Get( '__prbh_results_dir' );
}


sub hsp2string
{
	my $hsp=shift;
	return join("\t",values %$hsp);
}

sub __GetDefault
{
	my ( $key ) = @_;

	return $h_default{ $key };
}

sub __GetDefaults
{
	my ( @a_keys ) = @_;
	my @a_values = ();

	foreach my $key ( @a_keys )
	{	push( @a_values, $h_default{ $key } );	}

		return @a_values;
}

1;
