# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

ChromosomeMap

=head1 DESCRIPTION

ChromosomeMap is used with NcdisplayMap
This class is use to calculate the new range after a 'zoom','scroll' or 'search' click.


=head2 ATTRIBUTES

reference		=> reference organism short name
chromosome		=> reference chromosome number (or label)
targets_list	=> sorted list of organisms compared with the reference
level			=> synteny level of the comparison
view_level		=> synteny level show (must be inferior or equal to 'level')
start			=> start of the slice shown
end				=> end of the slice shown
gene_structure	=> boolean, checkbox value to display exons data
chr_lenght		=> lenght of the reference chromosome (store to read corresponding file only once)
boxes			=> hash of the boxes that will structured the image, each box corresponding to an organism representation (reference or compared)
ref_position	=> position of the reference box, could be use to manage boxes organisation
gene			=> optional, if searching option is initialize to find a reference gene
position		=> optional, if searching option is initialize to find a reference position



=head1 SUBROUTINES

__InitBoxes
__InitRange
__SetReferenceCenterPos
__Zoom
_Init
AreSynteniesDefined
CenterSynteny
GetBox
GetBoxCoordinates
GetBoxDisplayAttributes
GetBoxesNumber
GetBoxJustification
GetImageParameters
GetRange
GetRangeSize
GetRangeToString
GetSynteniesInfos
GetTargetOrganism
GetTargetProgram
GetTargetsList
MoveView
SearchGene
SetBox
SetRange
ZoomLevel


=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package Planner::ChromosomeMap;

use strict;
use warnings;
use Data::Dumper;

use POSIX;

use base qw( Planner );

use File::Basename;
use General;
use NARCISSE_utils;
use ParamParser;
use Box::ChromosomeSliceBox;
use Image;
use ElementsSet::Synt;
use Error qw( :try );
use LipmError;
use LipmError::ObjectException;
use LipmError::ParameterException;
use NarcisseError;
use NarcisseServiceFactory;
use NarcisseServiceKeywords qw( &NS_INFOS );

my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

use constant DEFAULT_PROGRAM_GENOMIC 	=> 'genomic';
use constant DEFAULT_REF_POSITION		=> 'center';

# This is a COPY of the same constant found in chromosomeSlice.pm
use constant A_CORE_FEATURES 			=> qw( Gene Cyto Rna Repeat Personal Ocean Snp Contig Super Clone );

use constant ALL  => "All";
use constant SUPP => "Supp";
use constant CORE => "Core";

use DataReader::Whole;

=head2 _Init

 Title        : _Init
 Usage        : _Init(o_io => $o_io,o_param => $o_param, ...)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : $o_io: Instance observer object
                $o_param : ParamParser object
                $o_logger: Logger object
                $rh_organism: cf. NcScript.pm for the details about this parameter
 Globals	  : object attributes :
 					- reference		: short name of the main organism
 					- chromosome	: reference chromosome
 					- targets_list	: list of all organisms compared with reference.
 					- level			: synteny level of the comparison
 					- view_level	: synteny level show (must be inferior or equal to 'level')
 					- chr_length 	: lenght of the reference chromosome
 					- gene 			: search option (to search by name)
 					- position		: search option	(to search by position)
 					- gene_structure: boolean, display exons or not

 Errors		  : LipmError::ObjectException, LipmError::ParameterException

=cut
sub _Init
{
	my $self = shift;
	my %h_args = (
        o_io        => undef,
        o_param     => undef,
        o_logger    => undef,
        rh_organisms=> undef,
        @_
        );

    $self->SUPER::_Init( %h_args );

	my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get('_o_io');

    #Some default parameters
    $self->__SetDefaultAttributes( );

    #Set the attributes associated to the reference genome and targets
	$self->__SetGenomeAttributes( );

	# Getting display options from $o_param
	#     selected features (Core Supp)
	#	  level view_level gene position gene_structure ref_position ck_gs OnlineCompare
	$self->__SetDisplayAttributes( $o_param );

	# Getting the selected features (Core Supp)
	$self->__SetSelectedFeatures( );

	#Graphic initialization (Image and MapArea)
	$self->__InitGraphicAttributes();

	# Boxes management :
	$self->__InitBoxes();

	#Now set the current view according to the arguments (click, zoom, center_centeny, ...)
	$self->__SetCurrentView( $o_param );

	#Sets the graduation scale
	$self->__SetGraduationScale() if ( !defined $self->{ graduation_scale });

	# Test the graduation scale to know if we add exon representation
 	$self->{ gene_structure } = &TRUE if ( $self->GetGraduationScale() <= 2*10**4 );
    #$self->{ gene_structure } = &TRUE;
	return $self;
}

#----------------------------------------#
#      Initialization methods            #
#----------------------------------------#

=head2 procedure __SetDefaultAttributes

 Title        : __SetDefaultAttributes
 Usage        : $self->__SetDefaultAttributes()
 Function     : Set default arributes

=cut

sub __SetDefaultAttributes
{
	my $self = shift;

	my $o_io    = $self->_Get('_o_io');

	my $default_box_width  = $o_io->GetCfg('SYNT_MAP_WIDTH');
	my $default_box_height = $o_io->GetCfg('SYNT_MAP_HEIGHT');
	my $default_level      = $o_io->GetGuidef('level');
	my $default_view_level = $o_io->GetGuidef('view_level');
	my $param_graph    = $o_io->GetCfg( 'ROOT_GRAPHICS_CFG' );
    my $o_param_graph  = New ParamParser( $param_graph );

	my %h_initial_attributes = (

        program			 	=> &DEFAULT_PROGRAM_GENOMIC,
        level				=> $default_level,          # The synteny level used for the comparative mapping representation
        view_level			=> $default_view_level,     # The conserved chains making up the corresponding synteny segment ( view_level <= level )
        gene				=> &NONE,					# Value of the search field when searching a gene name.
        position			=> &NONE,					# Value of the search field when scrolling to a nucleic position.
        gene_structure		=> &FALSE,					# Draw exons or not
        OnlineCompare		=> undef,					# Start and en position of the reference region used for comparison, when not set no comparison is performed
        ref_position		=> &DEFAULT_REF_POSITION,  	# Position of the reference : could be 'center', 'left' or 'right'
        selected_features	=> [],						# features to display (if empty the default features defined in chromosome slice are used)
        targets_list		=> [],						# the target organism in the format target1_program1, target2_program2, ...

        _boxes				=> {},						# Maintains a hash on the boxes that will be used later
        __box_width			=> $default_box_width,		# the default box width
        __box_height		=> $default_box_height,

        o_param_graph		=> $o_param_graph
        );

	# Set the initial default attributes
	my @initial_attributes = keys %h_initial_attributes;
	@{$self}{ @initial_attributes } = @h_initial_attributes{ @initial_attributes };

	return ;
}


=head2 procedure __SetGenomeAttributes

 Usage        : $self ->__SetGenomeAttributes()
 Function     : Set attributes associated to the reference genome and targets
                The info is taken from the protected field _rh_organisms if possible, OR from o_param

=cut

sub __SetGenomeAttributes
{
	my $self = shift;

	my $o_param = $self->_Get('_o_param');

    my ($ra_species,$ra_chroms,$ra_programs) = ('','','');
    my $rh_organisms = $self->_Get('_rh_organisms');
    if (defined $rh_organisms)
    {
        $ra_species   = $rh_organisms->{ SPECIES };
        $ra_chroms    = $rh_organisms->{ CHROMS };
        $ra_programs  = $rh_organisms->{ PROGRAMS };

        $self->_Set(
            'reference' => $ra_species->[0],
            'chromosome'=> $ra_chroms->[0]
            );

		my @a_targets_list = ();
        for (my $i=1; $i <= $#$ra_species; $i++)
        {
        		push (@a_targets_list, $ra_species->[$i] . '__' . $ra_programs->[$i]);
        }
        $self->{ targets_list } = \@a_targets_list;
	}
    else
    {
        #Mandatory parameters
        my @mandatory_param =qw( reference chroms_reference );
        foreach my $param ( @mandatory_param )
        {
            if ( !$o_param->IsDefined( $param ) )
            {
                throw LipmError::ParameterException("ParamParser, parameter $param", 'missing');
            }
        }

        my ( $reference, $chromosome ) = $o_param->Get( @mandatory_param );

        $self->_Set( 'reference' => $reference,
            'chromosome'=> $chromosome
            );
    }

	#Load and set chromosome information: chromosome length (chr_length) and chromosome centromere position ()chr_centro)
	my ( $chrom_length, $chrom_centro  ) = $self->__LoadChromosomeInformation( );

	$self->_Set( 'chr_length' => $chrom_length,
				 'chr_centro' => $chrom_centro );

	#The different targets
	$self->__SetTargetList( $o_param );

	return;
}


=head2 procedure __LoadChromosomeInformation

 Title        : __LoadChromosomeInformation
 Usage        : $o_planner ->__LoadChromosomeInformation( )
 Function     : Load the chromosome information for the reference chrom (length and centromere position)
 Returns      : none

=cut
sub __LoadChromosomeInformation
{
	my $self = shift;

	my $o_io = $self->_Get('_o_io');

	# mettre comme attribut ?
	#my $o_ref = New SpeciesObserver( species => $self->GetRefOrg(), -o_inst_obs => $o_io );
    my $o_ref = $o_io->GetSpeciesObserver( $self->GetRefOrg() );


	#----------------------------------
	# Add Chromosome information
	my $rh_chroms         = $o_ref->GetChromsInfo();

	my $chrom_length 	= $rh_chroms->{ $self->{ chromosome } }{ length };
	my $chrom_centro 	= $rh_chroms->{ $self->{ chromosome } }{ centro };

	return ( $chrom_length, $chrom_centro  );
}

=head2 procedure __SetTargetList

 Title        : __SetTargetList
 Usage        : $o_planner ->__SetTargetList( $o_param )
 Function     : Constructs from the arguments in the paramparser object, the list of targets
 					The list of target organism are provided in the following form :
 						target1, target2, target3, .... program1, program2, program3, ...
					The resulting list of targets has the following format
						target1_program1,  	target2_program2, target3_program3
				If programi is missing we use the 'program' value instead
 Returns      : none
 Args         : $o_param

 TODO - Probablement plus tres utile, cf. __SetGenomeAttributes
=cut
sub __SetTargetList
{
	my $self = shift;
	my $o_param = shift;
	my $o_io = $self->_Get('_o_io');

	my @a_targets_list = ();
	my $n = 1;
    my $display_max_org = $o_io->GetCfg('DISPLAY_MAX_ORG')+0;

	while ($n <= $display_max_org)
	{
		last unless ($o_param->IsDefined("target$n"));

		my $target_name = $o_io->GetSpeciesShortName($o_param->Get("target$n"));
		my $target_program = ($o_param->IsDefined("program_target$n")) ? $o_param->Get("program_target$n") : $self->{ program };

		push (@a_targets_list, $target_name.'__'.$target_program);
		$n++;
	}

	$self->{ targets_list } = \@a_targets_list;

	return ;
}

=head2 procedure __SetDisplayAttributes

 Title        : __SetDisplayAttributes
 Usage        : $o_planner ->__SetDisplayAttributes( $o_param )
 Function     : Checks and replace the attribute modified by $o_param
 Returns      : none
 Args         : $o_param
 Comment	  : Again a patch to encapsulate previous initialization that blured the constructor

=cut
sub __SetDisplayAttributes
{
	my $self = shift;
	my $o_param = shift;

	#Eventually modifying some previously defined attributes
	my @parameters = qw( level view_level program gene ref_position OnlineCompare );
	foreach my $param ( @parameters )
	{
		if ( $o_param->IsDefined( $param ) )
		{
			$self->_Set( $param => $o_param->Get( $param ) );
		}
	}

	# view_level must always be smaller or equal to level
	( $self->_Get( 'view_level' ) <=  $self->_Get( 'level' ) )
		or throw LipmError(-depth => 1, -text=>"view_level should always be lower than level.");

	if ( $o_param->IsDefined( 'features' ) )
	{
		my $features = $o_param->Get( 'features' );
		my @features = split(/\s+/, $features);
		$self->_Set( 'selected_features' => \@features);
	} else
	{
		$self->_Set( 'selected_features' => [ "All" ] );
	}

	#We have to handle here parameters that change their names (what a shame)
	#if ( $o_param->IsDefined( 'pos' ) )
	#{
	#	$self->_Set( 'position' => $o_param->Get( 'pos' ) );
	#}
	#if ( $o_param->IsDefined('ck_gs')  )
	#{
	#	$self->_Set( 'gene_structure' => $o_param->Get('ck_gs') );
	#}

	return ;
}

=head2 Function __SetSelectedFeatures

 Title        : __SetSelectedFeatures
 Usage        : $o_xx->__SetSelectedFeatures();
 Procedure    : Sets the feature list according to the selected features given as arguments
 Args         : none

=cut
sub __SetSelectedFeatures
{
	my $self = shift;

	my @a_feature_list = ();
	my %h_features_batch = () ;

	foreach my $feat ( @{ $self->{ selected_features } } )
	{
		if ( $feat eq &ALL || $feat eq &SUPP || $feat eq &CORE )
		{
			$h_features_batch{ $feat }++;
		} else
		{
			push @a_feature_list, $feat;
		}
	}
	# We overwrite the feature list if we have either All features or Core features selected
	if (defined $h_features_batch{ &ALL } || defined $h_features_batch{ &CORE } )
	{
			@a_feature_list = &A_CORE_FEATURES;
	}
	$self->{ _features_batch } = \%h_features_batch;
	$self->{ feature_list } = \@a_feature_list;

}

=head2 procedure __InitGraphicAttributes

 Title        : __InitGraphicAttributes
 Usage        : $self ->__InitGraphicAttributes()
 Function     : Set attributes associated to the reference genome and targets
=cut
sub __InitGraphicAttributes
{
	my $self = shift;

	 my $o_io    = $self->_Get('_o_io');

	#======================================
	# Initialization of graphics parameters
	Image 		-> InitColorsLibrary( $o_io->GetCfg('ROOT_COLORS_CFG') );
	ElementsSet::Synt -> InitColors( $o_io->GetCfg('ROOT_SYNTENY_COLORS_FILE'));

	#The map area object for html maps
	my $o_maparea = MapArea->New( reference => $self->GetRefOrg() );
	$self->_Set( 'o_maparea' => $o_maparea );

}

=head2 __InitBoxes

 Title        : __InitBoxes
 Usage        : __InitBoxes()
 Prerequisite : none
 Procedure    : Initialization of boxes
 Returns      : none
 Args         : none
 Error		  : LipmError::ObjectException

=cut
sub __InitBoxes
{
	my $self = shift;

	my @a_targets = $self->GetTargetsList();
	my $ref_position = $self->{ ref_position };

	#----------------------------------
	# Set default boxes padding values
	&Box::SetDefaultPaddings(7,0,3,0);

	#TODO, think of a mechanism enabling to change these parameters somewhere in configuration or parameters
	$self->{ __ref_box_width }		= $self->{ __box_width };
	$self->{ __ref_target_width }	=   $self->{ __box_width };

	#----------------------------------
	# Re-order organisms list :
	my (@a_final_list,@a_left_targets,@a_right_targets);
	my $ref_index;

	# 1) If reference is on the left side
	if ($ref_position eq 'left')
	{
		@a_final_list = ($self->GetRefOrg(), @a_targets);
		$ref_index = 0;
	}
	# 2) If reference is on the right side
	elsif ($ref_position eq 'right')
	{
		@a_final_list = (reverse(@a_targets), $self->GetRefOrg());
		$ref_index = scalar(@a_targets);
	}
	# 3) If reference is centered
	else
	{
		my (@a_left_targets,@a_right_targets);
		for (my $i=0; $i<scalar(@a_targets); $i++)
		{
			push (@a_left_targets, $a_targets[$i]) if ($i%2 == 0);
			push (@a_right_targets, $a_targets[$i]) if ($i%2 != 0);
		}

		# Add an empty box if their is only reference.
		if (scalar(@a_left_targets) == 0)
		{
			push (@a_left_targets,' ');
			$ref_index++;
		}

		@a_final_list = (reverse(@a_left_targets), $self->GetRefOrg(), @a_right_targets);
		$ref_index = scalar(@a_left_targets);
	}

	#----------------------------------
	# Initialize boxes :
	my ($X0,$Y0) = (0,0);
	my ($X1,$Y1)= (undef,$self->{ __box_height });
	my $box_width = undef;
	for (my $j=0; $j<scalar(@a_final_list); $j++)
	{
		# Boolean to mark if a target box is near reference's one
		my $ref_neighbor = ( abs($j-$ref_index) == 1 ) ? 1 : 0;

		# By default the justification is set to center when reference is on middle
		my $justification = 'center';
		# If reference is on the left side, or a target is on the left of reference
		$justification = 'left' if ($j-$ref_index < 0 or $j==0);
		# If reference is on the right side, or a target is on the right of reference
		$justification = 'right' if ($j-$ref_index > 0 or $j == scalar(@a_final_list-1));

		$box_width = ( $ref_index == $j ) ? $self->{ __ref_box_width }	: $self->{ __ref_target_width };
		$X1 = $X0 + $box_width ;

		$self->SetBox ($a_final_list[$j], $X0, $Y0, $X1, $Y1, $justification, $ref_neighbor);

		$X0 = $X1 + 1 ;
	}

	# Test boxes initialization, return an ObjectException if a problem occured.
	if ($self->GetBoxesNumber() < scalar(@a_targets))
	{
		my $dbg = "Bad boxes initialization";
		throw LipmError::ObjectException( __PACKAGE__, -dbg => $dbg );
	}

	# Setting the final image height and width parameters
	$self->{ __image_width } = $X1;
	$self->{ __image_height } = $Y1;


	return;
}

=head2 __SetGraduationScale
 Title        : __SetGraduationScale
 Usage        : $self -> __SetGraduationScale()
 Prerequisite : The range must have been set first
 Procedure    : Set the attribute graduation_scale value. It's a 10 factor
 Returns      : none
 Args         : graduation_scale -> optional
 Globals      : none

=cut
sub __SetGraduationScale
{
	my $self = shift;

	# ulate the scale.
	my $range = $self -> GetRangeSize();

	# en perl la fonction 'log' represente en maths la fonction 'ln'
	my $n = $range/200;
	my $factor1 = ceil(log(5*$n)/log(10)); # log( $range/1000 )
	my $factor2 = ceil(log($n)/log(10));   # log( $range/200 )

	$self->{ graduation_scale } = ($factor1 < $factor2) ? 5*10**$factor1 : 10**$factor2;

	return;
}

#-----------------------------------------#
#  Selecting the appropriate view         #
#-----------------------------------------#

=head2 procedure __SetCurrentView

 Title        : __SetCurrentView
 Usage        : $o_planner ->__SetCurrentView( $o_param )
 Function     : Set the range according to $o_param parameters
 Returns      : none
 Args         : $o_param
 Comment	  : Again a patch to encapsulate previous initialization that blured the constructor

=cut
sub __SetCurrentView
{
	my $self = shift;
	my $o_param = shift;
    $self->_Trace( "__SetCurrentView entered\n", '.', 'MVV' );

	# Test 'range' parameter to set start and end values.
	# Moreover, test maximum bounds setting by 'max_range' field.
	$self->__InitRange();

	# center on a synteny, may be zooming also
    my @keys_center_synt = $o_param->GetKeys('center_synt\d*');
	if ( @keys_center_synt > 0 )
	{
        my $key = $keys_center_synt[0]; # There should exist only ONE key, if several the others are ignored
        $self->_Trace( "__SetCurrentView: center_synt ($key)\n", '.', 'MVV' );
        my $flg = ($o_param->Get('click') eq 'zoom_level_synteny');
		$self->CenterSynteny($key,$flg);
	}

	# Test 'gene' or 'pos' to search a gene or go to some position
    # Note: if 'gene' or 'pos' is specified together with a zoom_level ('click=zoom_level_xxx'), both actions may be performed
    elsif (($o_param->IsDefined('gene') and $o_param->Get('search') eq '1'))
    {
        $self->_Trace( "__SetCurrentView: goto gene\n", '.', 'MVV' );
        $self->__GotoGene($o_param->Get('gene'));
    }

    elsif ($o_param->IsDefined('pos') and $o_param->Get('search') eq '1')
    {
        $self->_Trace( "__SetCurrentView: goto pos\n", '.', 'MVV' );
        $self->__GotoPosition($o_param->Get('pos'));
        if ($o_param->IsDefined('click'))
        {
	        $self->_Trace( "__SetCurrentView: calling MoveView\n", '.', 'MVV' );
	   		$self->MoveView();
	   	}
    }

	# Test 'click' parameter to apply zoom/scroll. Zoom could be called as 'up'/'down' or by a special level like 'chromosome' or 'arm'.
   	elsif ($o_param->IsDefined('click'))
   	{
        $self->_Trace( "__SetCurrentView: calling MoveView\n", '.', 'MVV' );
   		$self->MoveView();
   	}
    $self->_Trace( "__SetCurrentView exited\n", '.', 'MVV' );
	return;
}

=head2 procedure __InitRange

 Title        : __InitRange
 Usage        : $o_planner -> __InitRange($o_param)
 Function     : Initialize range attributes (start,end, max_start, max_end) using ParamParser parameters.
 				# Test 'range' parameter to set start and end values.
				# Moreover, test maximum bounds setting by 'max_range' field.
 Returns      : none
 Args         : $o_param -> ParamParser object
 Attributes	  : max_start, max_end, start, end

=cut
sub __InitRange
{
	my $self = shift;
	my $o_param = $self->_Get('_o_param');

	#----------------------------------
	# Test parameter.
	return if (!defined $o_param);
	return if (ref($o_param) !~ /ParamParser/);

	#----------------------------------
	# Initialize max range attributes.
	my $max_range = $o_param->Get('max_range');
	if (defined $max_range && $max_range =~ /^(\d+):(\d+)/ && $max_range ne '0:0')
	{
		$self->{max_start}	= $1;
		$self->{max_end}	= $2;
	}
	else
	{
		$self->{max_start}	= 1;
		$self->{max_end}	= $self->_Get( 'chr_length' );
	}

	#----------------------------------
	# Initialize range attributes.
	my $range = $o_param->Get('range');
	if (defined $range && $range =~ /^(-?\d+):(\d+)/)
	{
		$self->SetRange($1,$2);
	}
	else
	{
		$self->SetRange();
	}

	return;
}

=head2 procedure __SetReferenceCenterPos

 Title        : __SetReferenceCenterPos
 Usage        : $o_planner -> __SetReferenceCenterPos($pos)
 Function     : Change a range in order to center the view on the new position
 Returns      : none
 Args         : a valid position on the chromosome

=cut
sub __SetReferenceCenterPos
{
	my $self = shift;
	my $pos = shift;
	my %h_args = ( inflate_factor => 1,
					@_
					);

	my $range_size = $self->GetRangeSize();
	#Inflate the range size
	$range_size *= $h_args{ inflate_factor };
	my $new_start  = int($pos - $range_size/2);
	my $new_end    = int($pos + $range_size/2-1);

	$self->SetRange($new_start,$new_end);

	return;
}

=head2 procedure GetReferenceCenterPos

 Title        : GetReferenceCenterPos
 Usage        : $o_planner -> GetReferenceCenterPos()
 Function     : Returns the center position of the map (reference genome)
 Returns      : a position in base pair
 Args         : none
=cut
sub GetReferenceCenterPos
{
	my $self = shift;

	return  int ( $self->{start} + $self->{end} )/2;
}

=head2 __Zoom

 Title        : __Zoom
 Usage        : __Zoom($factor)
 Prerequisite : none
 Procedure    : Changing the range (start and end) according to current middle position, range and zoom factor
 Returns      : none
 Args         : factor = the zoom factor
 Attributes	  : o_param
 Globals	  : none
 Error		  : none

=cut
sub __Zoom
{
	my $self = shift;
	my $factor = (defined $_[0] && $_[0] =~ /^\d*\.?\d+$/ && $_[0] != 0) ? shift : throw LipmError::ParameterException('factor', 'null');
	my ($start,$end) = $self->GetRange();
	my $range  = $end - $start;

	my $middle 	= $start + $range/2;
	my $step	= $range/(2*$factor);

	my $new_start	= int($middle - $step);
	my $new_end		= int($middle + $step);

	$self->SetRange($new_start, $new_end);

	return;
}

=head2 MoveView

 Title        : MoveView
 Usage        : $self -> MoveView();
 Prerequisite : ChromosomeMap object must be initialized.
 Procedure    : Test the 'click' parameter, and apply zoom or scroll
 Returns      : none
 Args         : none
 Attributes	  : o_param,start, end

=cut
sub MoveView
{
	my $self = shift;
	my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get('_o_io');

	my $click_param = lc($o_param->Get('click'));
	my ($start,$end) = $self->GetRange();

	#==============================================
	# Zoom :
	if ($click_param =~ /zoom(_level_)?(\w+)$/)
	{
		my $zoom = $2;
        my $zoom_step = $o_io->GetCfg('ZOOM_STEP') / 100.0;
        $zoom_step = .1  if ($zoom_step <= .1);
        $zoom_step = .9  if ($zoom_step >= .9);
        $self -> __Zoom(1 + $zoom_step) if ($zoom eq "pos");
        $self -> __Zoom(1 - $zoom_step) if ($zoom eq "neg");

		$self -> ZoomLevel($zoom) if ($click_param =~ /level/);
	}

	#==============================================
	# Scrolls :
	if ($click_param =~ /^scroll(\w+)$/ )
	{
		my $scroll = $1;
        my $move_pas = $o_io->GetCfg('MOVE_PAS');
		my $pas = ($move_pas =~ /^(\d+)$/ && $move_pas > 1) ? $move_pas : 4;
		my $range = $end-$start;
		my $step = int($range/$pas);

		if ($scroll eq 'up')
		{
			$end	= ($start-$step > $self->{max_start}) ? $end  -$step : $self->{max_start}+$range;
			$start 	= ($start-$step > $self->{max_start}) ? $start-$step : $self->{max_start};
		}
		else
		{
			$start  = ($end + $step <= $self->{max_end}) ? $start+$step : $self->{max_end}-$range;
			$end	= ($end + $step <= $self->{max_end}) ? $end  +$step : $self->{max_end};
		}

        $self -> SetRange($start,$end);
	}

	return;
}

=head2 procedure __GotoPosition

 Title        : __GotoPosition
 Usage        : $self -> __GotoPosition('50mb');
 Procedure    : Try to go to the searched position
                Recognized strings are (don't care case): 50, 50m, 50mb, 50k, 50kb, 50 mb etc., 900kb:1mb, .9:1mb
                If the parameter does not look like a position, we call __GotoGene
 Args         : The position we want to reach

=cut
sub __GotoPosition
{
	my $self = shift;

    my ($position_string) = @_;
    if ($position_string eq 'none')
    {
        return;
    }

    # If 'none', ignore the pos paramter
    elsif ($position_string =~ /^[.\d]+[\d.kmb:\- ]*$/i)
    {
        my $global_unit=undef;
        if ($position_string =~ /([bmk])[bp]?$/i)
        {
            $global_unit = $1;
        }

        my @a_positions = split(/[:\-]|\.\./,$position_string,2);
        foreach my $position (@a_positions)
        {
            my ($position_nb, $position_unit) = ($position =~ /^([.\d]+)([bmk])?/i);
            $position_unit = $global_unit if (!defined $position_unit && defined $global_unit);
            $position = int($position_nb)      if ( defined $position_unit && $position_unit =~ /b/i );
            $position = $position_nb * 1000	   if ( defined $position_unit && $position_unit =~ /k/i );
            $position = $position_nb * 1000000 if ( defined $position_unit && $position_unit =~ /m/i );

        }
        if (@a_positions==2 && $a_positions[1] > $a_positions[0])
        {
            $self->SetRange(@a_positions);
        }
        else
        {
            $self->__SetReferenceCenterPos($a_positions[0]);
        }
    }

    # We don't recognize this as a position, let's try as a gene
    else
    {
        $self->__GotoGene($position_string);
    }

    return;
}

=head2 __GotoGene

 Title        : __GotoGene
 Usage        : $self -> __GotoGene('FixJ1');
 Procedure    : Try to go to the searched gene
 Returns      : none
 Args         : The gene we want to reach
 Error        : may throw a NarcisseError if the gene is not found

=cut
sub __GotoGene
{
	my $self = shift;

    my $o_param = $self->_Get('_o_param');
    my ($gene_id) = @_;
    $self->_Set('gene',$gene_id);
    my $rh_gene_infos = $self->SearchGene();

    my $start = $$rh_gene_infos{start};
    my $end   = $$rh_gene_infos{end};
    $self -> SetRange($start,$end);
    $self -> ZoomLevel('gene_name',&TRUE);

    return;
}


=head2 CenterSynteny

 Title        : CenterSynteny
 Usage        : $self -> CenterSynt($o_param);
 Prerequisite : Planner object need to be initialized.
 Procedure    : Set the range attribute according to the requested synteny
 Returns      : none
 Args         : $key The key to consider   : center_synt, center_synt1, center_synt2 etc.
                $zoom_flg (opt, def=&FALSE): if &TRUE, call ZoomLevel to zoom to this synteny
 Attributes	  : o_param

=cut
sub CenterSynteny
{
	my $self = shift;

    my ($key,$zoom_flg) = @_;
    $zoom_flg = &FALSE unless defined $zoom_flg;

	my $o_param    = $self->_Get('_o_param');
	my $o_io       = $self->_Get('_o_io');

	my $synteny_level = $self->{level} - $self->{view_level};
    my $rank;
    if ($key eq 'center_synt')
    {
        $rank = 1;
    }
    else
    {
        ($rank) = ($key =~ /center_synt(\d)+/);
    }
	my $synt = $o_param->Get($key);


	#TODO modify this to use object attributes (?!)
	#my @a_keys = qw( reference chroms_reference target1 program_target1 );
	#my( $ref, $chrom, $target, $prog ) = $o_param->Get( @a_keys );

    my $ref   = $self->GetRefOrg();
    my $chrom = $self->GetRefChrom();
    my $target= $self->_Get('_rh_organisms')->{ SPECIES } -> [$rank];
    my $prog  = $self->_Get('_rh_organisms')->{ PROGRAMS } -> [$rank];

	my $o_ref = New SpeciesObserver( species => $ref );
	my $filepath = $o_ref->GetSyntFiles( -chrom => $chrom,
										 target => $target,
										 prog   => $prog,
										 level  => $synteny_level );

	my $o_ndr = New DataReader::Whole();

	$o_ndr->Prepare( -from   => $filepath,
			 -where  => { VALUE => { element => $synt } }
            );

	my $ra_data   = $o_ndr->FetchallArrayref( -line_type => "HASH" );

    #warn "KOUKOU $synt#$filepath#" . @$ra_data;
    if (  0 == @$ra_data )
    {
        throw NarcisseError(-bd => "Sorry, the synteny $synt does not exist at level $synteny_level");
    }
	my $synt_data = $ra_data->[0];

    # Zoom to this synteny
    if ( $zoom_flg == &TRUE )
    {
        $self->SetRange($synt_data->{'start'},$synt_data->{'end'});
        $self->ZoomLevel('synteny',&TRUE);
    }

	# Change range to center the image on synteny $synt.
    else
    {
        $self->__SetReferenceCenterPos(($synt_data->{'start'}+$synt_data->{'end'})/2, inflate_factor => 1.2);
    }

    $o_param->Delete($key);
    return;
}

=head2 SetRange

 Title        : SetRange
 Usage        : $self -> SetRange(25,100);
 Prerequisite : ChromosomeMap object need to be initialized.
 Procedure    : Test values of start and end, set start/end attributes and return them.
 Returns      : (start, end)
 Args         : $start, $end = two integers to test
 Attributes	  : start, end, max_start, max_end

=cut
sub SetRange
{
	my ($self,$start,$end) = (shift,shift,shift);

	#---------------------------
	# Test parameters.
	$start = $self->{max_start} if (!defined $start);
	$end   = $self->{max_end}	if (!defined $end);

	# End must be higher than start
	$end = $start + 1 if ($start==$end);
	if ($start > $end)
	{
		my $tmp = $start;
		$start = $end;
		$end = $tmp;
	}

	#---------------------------
	# Test bounds => 3 cases :
	# 1) Both start and end are invalid :
	my $range = $end-$start;
=pod
	if ( $start < $self->{max_start} && $end > $self->{max_end} )
	{
		$start = $self->{max_start};
		$end   = $self->{max_end};
	}
	# 2) Start is inferior to max_start
	elsif ( $start < $self->{max_start} )
	{
		$start = $self->{max_start};
		$end   = ( $start+$range <= $self->{max_end} ) ? $start+$range : $self->{max_end};
	}
	# 3) End is superior to max_end
	elsif ( $end > $self->{max_end} )
	{
		$end   = $self->{max_end};
		$start = ( $end-$range >= $self->{max_start} ) ? $end-$range : $self->{max_start};
	}
=cut
	#---------------------------
	# Set object attributes : We always want integer values for range
	$self->{start} 	= int($start);
	$self->{end}	= int($end);

	#---------------------------
	# Set the range value and the overview value in ParamParser.
    my $o_param = $self->_Get('_o_param');
	$o_param -> Set('range',$self->GetRangeToString());
    my $length = $self->_Get( 'chr_length');
    my $overview_length = int(100*($end-$start)/$length);

    $overview_length=1 unless($overview_length);  # Avoiding 0...
    my $overview_top   = int(100*$start/$length);
    $o_param -> Set('overview',"$overview_top:$overview_length");

	return ($start,$end);
}

=head2 ZoomLevel

 Title        : ZoomLevel
 Usage        : $self -> ZoomLevel(zoom_level)
 Procedure    : Convert a key word, corresponding to a biological level, to the nucleic range of the representation and call SetRange
                We start from the current range. We look for the nearest object (arm,gene,synteny) and we change the
                range to get the objet surrounded by the display.
                Chromosome level: display the whole chromosome
                Chromosome arm:   display the whole arm, selecting the current arm (given by the range)
                Synt level, gene level:
                     - If parameter $flg_object is NOT set find the nearest gene or synteny
                     - Center the display on this object
                     - Select a range to 5x the gene length or 2x the synteny length, in order to display also the context
                gene structure: same thing as gene level, except that:
                     - The range is just around the gene
                     - The exons are displayed
                TODO - Le facteur 5 ou 2 est en dur, ca devrait etre dans un parametre (fichier zoom_level.list, pas utilise actuellement)
                TODO - Pour le calcul de $middle et $span, il y a des methodes qui existent. C'est plus propre
 Args         : zoom_level : chromosome, arm, synteny, gene_name, gene_struct
                flg_object (optional, used only with gene_name/synteny): it set, the range is already surrounding the object
 Globals      : none

=cut
sub ZoomLevel
{
    my $self = shift;

    my ($zoom_level,$flg_object) = @_;
    $flg_object = &FALSE unless defined $flg_object;

	my @a_attr = qw( reference chromosome gene _o_io );
	my( $ref, $chrom, $gene, $o_io ) = $self->_Get( @a_attr );
	
    my ($start,$end) = $self->GetRange();

	# Chromosome level :
	if ($zoom_level eq 'chromosome')
	{
		$start = 1;
		$end   = $self->_Get( 'chr_length' );
	}

	# Arm level :
	elsif ($zoom_level eq 'arm')
	{
    	my $length =  $self->_Get( 'chr_length' );
    	my $centro =  $self->_Get( 'chr_centro' );

        my $middle = int(($start+$end)/2);
    	$start 	= ( $middle < $centro) ? 1 : $centro;
		$end	= ( $middle < $centro) ? $centro : $length;
	}

    #Gene level/gene structure, try to find the gene from the current range
    elsif ( $zoom_level eq 'gene_name' || $zoom_level eq 'gene_struct' )
    {
        if ( $flg_object==&FALSE )
        {
            my $o_ref = New SpeciesObserver( species => $ref );
            my $filepath = $o_ref->GetGeneFiles( -chrom => $chrom );

            # Read the whole data file
            my $o_ndr = New DataReader::Whole();
            $o_ndr->Prepare( -from   => $filepath );
            my $ra_data    = $o_ndr->FetchallArrayref();
            my $rh_headers = $o_ndr->GetHeadersColumns();
            my $o_data  = New NarcisseData::ArrayReturn( ra_data => $ra_data, rh_headers => $rh_headers );

            # Find the nearest gene
            my $middle     = int(($start+$end)/2);
            my $ra_nearest = $o_data -> FindNearestElement ( $middle );

            # Find the start,end positions of the gene
            $start = $ra_nearest -> [ $rh_headers->{'start'}];
			$end   = $ra_nearest -> [ $rh_headers->{'end'}];
        }

        # gene_name: the range is N x the gene length, N is read in defaults.cfg
        if ( $zoom_level eq 'gene_name')
        {
            my $middle = int( ($start + $end) / 2 );
            my $span  = $end - $start;
	        my $factor = $o_io -> GetCfg ( 'ZOOM_LEVEL_FACTOR_GENE' );
            $start    = $middle - int( $factor * $span/2 );
            $end      = $middle + int( $factor * $span/2 );
        }

        # gene_structure: the range is OK, but we must show the exons
        else
        {
            my $o_param = $self->_Get('_o_param');
            $o_param->Set('ck_exon',1);
            $self->{gene_structure} = 1;
        }
    }

    # Synteny level, try to find the current synteny from the current range
    elsif ( $zoom_level eq 'synteny' )
    {
        if ( $flg_object==&FALSE )
        {
            # Get file name to calculate the new range according to the number of elements which must be displayed
            my $synteny_level = $self->{level} - $self->{view_level};
            my $ra_targets = $self->_Get( qw( targets_list ) );
            my $target = $self->GetTargetOrganism( $$ra_targets[0] );
            my $prog   = $self->GetTargetProgram(  $$ra_targets[0] );

            # Throw a Narcisse Error if the user clicks on synteny level with no target
            if ($target eq '')
            {
                throw NarcisseError(-bd => "Please select a target before zooming at synteny level");
            }

            my $o_ref = New SpeciesObserver( species => $ref );
            my $filepath = $o_ref->GetSyntFiles( -chrom => $chrom,
											target => $target, # peut etre undef => lancement erreur ok
											prog   => $prog,
											level  => $synteny_level );

            # Read the whole data file
            my $o_ndr = New DataReader::Whole();
            $o_ndr->Prepare( -from   => $filepath );
            my $ra_data    = $o_ndr->FetchallArrayref();
            my $rh_headers = $o_ndr->GetHeadersColumns();
            my $o_data  = New NarcisseData::ArrayReturn( ra_data => $ra_data, rh_headers => $rh_headers );

            # Find the nearest synteny
            my $middle     = int(($start+$end)/2);
            my $ra_nearest = $o_data -> FindNearestElement ( $middle );

            # Find the start,end positions of the synt
            $start = $ra_nearest -> [ $rh_headers->{'start'}];
			$end   = $ra_nearest -> [ $rh_headers->{'end'}];
        }

        # the range is N x the synt length (defaults to 2, cf. defaults.cfg)
        my $middle = int( ($start + $end) / 2 );
        my $span  = $end - $start;
        my $factor = $o_io -> GetCfg ( 'ZOOM_LEVEL_FACTOR_SYNTENY' );
        $start    = $middle - int( $factor * $span/2 );
        $end      = $middle + int( $factor * $span/2 );
    }

	# Set the new range and return
	$self->SetRange($start, $end);
    return;
}

=head2 function SearchGene

 Title        : SearchGene
 Usage        : &SearchGene($orga,$chrom,$gene)
 Prerequisite : none
 Fonction     : Looking for the gene in the 'gene' file
 Returns      : The row of the gene, if found, or undef if not found
 Args         : A lot
 Globals      : none

=cut

sub SearchGene
{
	my $self = shift;

	my @a_attr = qw( reference chromosome gene _o_io );
	my( $ref, $chrom, $gene_id, $o_io ) = $self->_Get( @a_attr );
	
	# Strip leading of tailing blanks
	$gene_id =~ s/^\s*(\S+)\s*$/$1/;

	my %h_args = (	species => $ref,
					chrom   => $chrom,
					ra_ids  => [ $gene_id ],
					-o_instance_obs => $o_io,
					-out_type  => &NS_INFOS,
					-o_logger  => $self->GetLogger()
                    );

	my $o_ns = &NarcisseServiceFactory::GetGenesFromLocusTags( %h_args );
	my ($ra_headers,$rh_data) = $o_ns->FetchResults();

	if( 0 == @$ra_headers )
	{ # It's not a locus tags, let's try a protein id
		$self->_Trace( "$gene_id is not a locus tag. will try protein id.\n", '.', 'MVV' );

		$o_ns = &NarcisseServiceFactory::GetGenesFromProteinIds( %h_args );
		($ra_headers,$rh_data) = $o_ns->FetchResults();
	}

	if( 0 == @$ra_headers )
	{ # It's not a locus tags nor protein ids, let's try symbols
		$self->_Trace( "$gene_id is not a protein id. will try symbol.\n", '.', 'MVV' );

		$o_ns = &NarcisseServiceFactory::GetGenesFromSymbols( %h_args );
		($ra_headers,$rh_data) = $o_ns->FetchResults();
	}

	if( 0 == @$ra_headers )
	{ # It's not a locus tags nor protein ids nor symbol, try accession numbers
		$self->_Trace( "$gene_id is not a symbol. will try accession.\n", '.', 'MVV' );

		$o_ns = &NarcisseServiceFactory::GetGenesFromAccessionNumbers( %h_args );
		($ra_headers,$rh_data) = $o_ns->FetchResults();
	}

	if( 0 == @$ra_headers )
	{
		$self->_Trace( "$gene_id is not an accession number. will die.\n", '.', 'MVV' );

		my $msg = "The gene related to the id '$gene_id' was not found";
		$msg   .= " in $ref species, nor any proteins.";
		throw NarcisseError( -bd => $msg );
	}

	my $ra_gene = $$rh_data{ data };

	my $str_gene= $$ra_gene[0]; # on obtient un en-tete fasta
	my( $start, $end ) = $str_gene =~ /range=(\d+)-(\d+)/;

	return { start => $start, end => $end };
}

=head2 GetAdditionalFeatures

 Title        : GetAdditionalFeatures
 Usage        : $o_xxx->GetAdditionalFeatures()
 Fonction     : Checking for the additional features to display: Exons and the possible supplementary features (supp_*)
 Returns      : an array corresponding to the available additional features
 Args         : none
 Globals      : none

=cut
sub GetAdditionalFeatures
{
	my $self = shift;

	my @a_add_features = ();

	push @a_add_features, "Exon" if ($self->{ gene_structure} == &TRUE);

	return  @a_add_features;
}

##########################################
#          Predicates
##########################################

=head2 Function IncludeSuppData

 Title        : IncludeSuppData
 Usage        : if ( $o_xx->IncludeSuppData() )
 Procedure    : Returns true (1) is supplementary data have to be displayed
 Args         : none

=cut
sub IncludeSuppData
{
	my $self = shift;

	my $bool_include_supp = 0;
	my $rh_features_batch = $self->_Get( '_features_batch' );

	if ( defined $rh_features_batch->{ &ALL } || defined $rh_features_batch->{ &SUPP })
	{
		$bool_include_supp = 1;
	}
	return 	$bool_include_supp;
}

=head2 Function IncludePrbhHsp

 Title        : IncludePrbhHsp
 Usage        : $o_xx->IncludePrbhHsp();
 Procedure    : Returns true (1) is hsp and prbh have to be displayed
 Args         : none

=cut
sub IncludePrbhHsp
{
	my $self = shift;

	my $bool_include_supp = 0;
	my $rh_features_batch = $self->_Get( '_features_batch' );

	if ( defined $rh_features_batch->{ &ALL } || defined $rh_features_batch->{ &CORE })
	{
		$bool_include_supp = 1;
	}
	return 	$bool_include_supp;
}

=head2 AreSynteniesDefined

 Title        : AreSynteniesDefined
 Usage        : $boolean = $o_planner -> AreSynteniesDefined();
 Prerequisite : ChromosomeMap object need to be initialized.
 Function     : Test if syntenies are defined
 Returns      : boolean
 Args         : none
 Globals	  : targets_list

=cut
sub AreSynteniesDefined
{
	my $self = shift;
	my $return = (defined $self->{targets_list} && scalar($self->{targets_list}) > 0) ? 1 : 0;

	return $return;
}

#------------------------------------------------#
#              Managing boxes                    #
#------------------------------------------------#

=head2 GetBoxesNumber

 Title        : GetBoxesNumber
 Usage        : $o_planner -> GetBoxesNumber()
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get the number of boxes
 Returns      : integer
 Args         : none
 Globals	  : boxes

=cut
sub GetBoxesNumber
{
	my $self = shift;

	my $boxes_number = scalar(keys(%{$self->{_boxes}}));

	if ($boxes_number == 1 && $self->{ref_position} eq 'center')
	{
		return 2;
	}
	else
	{
		return $boxes_number;
	}
}

=head2 GetBoxCoordinates

 Title        : GetBoxCoordinates
 Usage        : $o_planner -> GetBoxCoordinates(box_name)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get image coordinates of the box 'box_name'.
 Returns      : Array (X0,Y0,X1,Y1)
 Args         : box_name = the box label

=cut
sub GetBoxCoordinates
{
	my $self = shift;
	my $box_name = shift;
	$box_name = $self->{reference} if (!defined $box_name);

	my $o_box = $self->GetBox($box_name);

	return $o_box -> GetBounds();
}

=head2 GetBoxJustification

 Title        : GetBoxJustification
 Usage        : $o_planner -> GetBoxJustification(box_name)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get image position of the box 'box_name'.
 Returns      : position : left, right, center
 Args         : box_name = the box label
 Globals	  : boxes

=cut
sub GetBoxJustification
{
	my $self = shift;
	my $box_name = shift;
	$box_name = $self->{reference} if (!defined $box_name);

	my $o_box = $self->GetBox($box_name);

	return $o_box->GetJustification();
}

=head2 GetBoxDisplayAttributes

 Title        : GetBoxDisplayAttributes
 Usage        : $o_planner -> GetBoxDisplayAttributes(box_name)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get all parameters to call Display method :
 					- jutification : image position of the box 'box_name'.
 					- paddings
 Returns      : %param
 Args         : box_name = the box label
 Globals	  : boxes

=cut
sub GetBoxDisplayAttributes
{
	my $self = shift;
	my $box_name = shift;
	$box_name = $self->{reference} if (!defined $box_name);

	my $o_box = $self->GetBox($box_name);

	my %h_param = (	side => $o_box->GetJustification(),$o_box->GetPaddings() );

	return %h_param;
}

=head2 GetBox

 Title        : GetBox
 Usage        : $o_planner -> GetBox(box_name)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get reference of the hash corresponding to this box.
 Returns      : Hash ref
 Args         : box_name = the box label
 Globals	  : boxes

=cut
sub GetBox
{
	my $self = shift;
	my $box_name = shift;

	$box_name = $self->{ reference } if (!defined $box_name or !defined $self->{_boxes}{$box_name});

	return $self->{ _boxes }{$box_name};
}

=head2 SetBox

 Title        : SetBox
 Usage        : $o_planner -> SetBox(box_name, X0, Y0, position [,ref_neighbor])
 Prerequisite : ChromosomeMap need to be initialized
 Procedure    : Add a new box on the 'virtual' image.
 Returns      : none
 Args         : box_name = label of the box
 				(X0,Y0)	 = origin coordinates
 				position = position of the box (center, right, left)
 				ref_neighbor = boolean to manage multiple representation of synteny links (0 or 1)

=cut
sub SetBox
{
	my $self = shift;
	my ($box_name, $X0, $Y0, $X1, $Y1, $position, $ref_neighbor) = (shift, shift, shift, shift, shift, shift, shift);
	return if (!defined $box_name or $box_name eq "");
	my $rh_boxes = $self->{_boxes};

	# Test parameters :
	return if ($X0 < 0 || $Y0 < 0);
	return if ($position !~ /^right|left|center$/);
	$ref_neighbor = 1 if (!defined $ref_neighbor or $ref_neighbor!~/^0|1$/);

	# Set box values :
	my %h_param = (	X0 				=> $X0,
					Y0 				=> $Y0,
					X1 				=> $X1,
					Y1 				=> $Y1,
					justification 	=> $position,
					ref_neighbor	=> $ref_neighbor
					);

	$$rh_boxes{$box_name} = Box::ChromosomeSliceBox->New (%h_param, o_param_cfg => $self->{ o_param_cfg });

	return;
}

#------------------------------------------------#
#  Managing  and  Printing the images            #
#------------------------------------------------#

=head2 GetImageParameters

 Title        : GetImageParameters
 Usage        : $o_planner -> GetImageParameters()
 Function     : Minimum information required for describing the image : width, height, frame_trans_type
 Returns      : a reference on a hash

=cut
sub GetImageParameters
{
	my $self = shift;

	my $width =  $self->{ __image_width };
	my $height = $self->{ __image_height };

	my %h_param = (
			frame_transf_type => "1D",
			width	=> $width,
			height	=> $height
			);

	return \%h_param;
}

#------------------------------------------------#
#               Some accessors                   #
#------------------------------------------------#

=head2 GetGraduationScale
 Title        : GetGraduationScale
 Usage        : $self -> GetGraduationScale()
 Prerequisite : none
 Function     : Return graduation_scale value.
 Returns      : graduation_scale
 Args         : none
 Globals      : none

=cut
sub GetGraduationScale
{
	my $self = shift;

	return $self->{ graduation_scale };
}

=head2 GetRange

 Title        : GetRange
 Usage        : ($start,$end) = $o_planner -> GetRange()
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Return start and end values
 Returns      : An array (start,end)
 Args         : none
 Globals	  : start, end

=cut
sub GetRange
{
	my $self = shift;

	return ($self->{start}, $self->{end});
}

=head2 function GetRangeSize

 Title        : GetRangeSize
 Usage        : $size = $o_planner -> GetRangeSize()
 Function     : Return the size of the range : end-start+1
 Returns      : the size in the correct unit
 Args         : none

=cut
sub GetRangeSize
{
	my $self = shift;

	return ( $self->{end}-$self->{start}+1 );
}

=head2 GetRangeToString

 Title        : GetRangeToString
 Usage        : $range = $o_planner -> GetRangeToString()
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Return start and end values in string format
 Returns      : A string  'start:end'
 Args         : none
 Globals	  : start, end

=cut
sub GetRangeToString
{
	my $self = shift;

	return "$self->{start}:$self->{end}";
}

=head2 function GetRefOrg

 Title        :
 Usage        : $o_planner ->GetRefOrg ($target_name)
 Function     : accessor to the short name of the reference organism
 Returns      : the short name of the reference genome

=cut
sub GetRefOrg
{
	my $self = shift;
    return $self->_Get('reference');
}

=head2 function GetRefChrom

 Title        :
 Usage        : $o_planner ->GetRefChrom ($target_name)
 Function     : accessor to the reference organism chromosome
 Returns      : the short name of the reference chromosome

=cut
sub GetRefChrom
{
	my $self = shift;
		return $self->_Get('chromosome');
}

=head2 GetSynteniesInfos

 Title        : GetSynteniesInfos
 Usage        : $o_planner -> GetSynteniesInfos($target_name)
 Prerequisite : ChromosomeMap object need to be initialized.
 Function     : Get an hash-map which have all data for the target Syntenies object initialization
 Returns      : An hash ref : (organism, chromosome, start, end, subject, program, level, view_level, gene_structure)
 Args         : $target_name
 Globals	  : organism, chromosome, start, end, level, view_level, gene_structure

=cut
sub GetSynteniesInfos
{
	my $self = shift;
	my $target_label = shift;

use Carp;
carp( "Mais si, GetSynteniesInfos est toujours utilise !" );

	# Test parameter :
	return () if (!defined $target_label);

	my $target 	= $self->GetTargetOrganism( $target_label );
	my $program = $self->GetTargetProgram( $target_label );

	my %h_param = (  organism 	=> $self->{ reference },
					 chromosome => $self->{ chromosome },
					 start		=> $self->{ start },
					 end		=> $self->{ end },
					 target 	=> $target,
					 program 	=> $program,
					 level 		=> $self->{ level },
					 view_level => $self->{ view_level },
					 o_maparea  => $self->{ o_maparea },
					 gene_structure	=> $self->{ gene_structure }
					);

	return %h_param;
}

=head2 GetTargetsList

 Title        : GetTargetsList
 Usage        : $o_planner -> GetTargetsList();
 Prerequisite : ChromosomeMap object need to be initialized.
 Function     : return the list of compared elements.
 Returns      : An array
 Args         : none
 Globals	  : targets_list

=cut
sub GetTargetsList
{
	my $self = shift;

	return @{$self->{targets_list}};
}

=head2 GetTargetOrganism

 Title        : GetTargetOrganism
 Usage        : $o_planner -> GetTargetOrganism(target_label)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get the right organism of target_label (target_list store target data as : 'name__program')
 Returns      : string = the target short name
 Args         : none
 Globals	  : target_list

=cut
sub GetTargetOrganism
{
	my $self = shift;
	my $target_label = shift;

	# Test parameter
	return '' if (!defined $target_label);

	if ($target_label =~ /__/)
	{
		$target_label = $`;			# We take the expression before '__'. Example : "mouse__prbh" => "mouse"
	}

	return $target_label;
}

=head2 GetTargetProgram

 Title        : GetTargetProgram
 Usage        : $self -> GetTargetProgram(target_label)
 Prerequisite : ChromosomeMap need to be initialized
 Function     : Get program of target_label (target_list store target data as : 'name__program')
 Returns      : string = the target program
 Args         : none
 Globals	  : target_list

=cut
sub GetTargetProgram
{
	my $self = shift;
	my $target_label = shift;

	# Test parameter
	return '' if (!defined $target_label);

	my $target_program = $self->{ program };

	if ($target_label =~ /__/)
	{
		$target_program = $';			# We take the expression after '__'. Example : "mouse__prbh" => "prbh"
	}

	return $target_program;
}

#------------------------------------------------#
#               Map information                  #
#------------------------------------------------#

=head2 DumpMapInformation

 Title        : DumpMapInformation
 Usage        : $o_planner -> DumpMapInformation( $map_file )
 Function     : Writes the map coordinate informations (.coord file)
 Returns      : none
 Args         : The map file name (.coord file)

=cut
sub DumpMapInformation
{
	my $self = shift;

    my ($map_file) = @_;
	my $o_param 		= $self->_Get( '_o_param' );
	my $o_param_graph 	= $self->_Get( 'o_param_graph' );
    my $o_io            = $self->_Get( '_o_io' );

	#my $map_file = $o_io->CalcCoordNames($o_param,1,'map');

	#Dumping the Map information for all the displayed objects
	$self->{ o_maparea }->DumpMapInformation( $map_file );

    # Keep the coordinates for cursors along the reference chromosome graduation
    my ($x1,$y1,$x2,$y2) = $self->GetBoxCoordinates();
	my $x_grad = $o_param_graph->Get('x_grad');# The position of the axis
	my $x_box = $x1;

	$x1 = $x_box + $x_grad - 10;
	$x2 = $x_box + $x_grad + 10;

	my $ra_chrom_coord = $self->__GetChromosomeCoord($x1,$y1,$x2,$y2);
	open FILE, ">> $map_file";
	foreach my $ra_line (@$ra_chrom_coord)
	{
		foreach my $elt (@$ra_line)
		{
			print FILE "$elt\t";
		}
		print FILE "\n";
	}
	close FILE;

	return;
}

# ==================== The coord management =================
#
# When computing the images, we compute ALSO the coordinates of the graphics objects
# Those are stored, with most needed information, in the .coord files
#

=head2 function $self->__GetChromosomeCoord

 Title        : __GetChromosomeCoord
 Usage        : my $ra_map = $self->__GetChromosomeCoord($x1,$y1,$x2,$y2);
 Function     : return the coordinates of the map corresponding to the cursor on the reference chromosome
                (100 areas along the chromosome scale)
 Args         : $x1,$y1,$x2,$y2 = The box surrounding the reference chromosome
 Return       : An array of arrays
 Globals      : none

=cut
sub __GetChromosomeCoord
{
    my $self = shift;
    my ($x1,$y1,$x2,$y2) = @_;

    my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get( '_o_io' );

    #my $reference = $o_param->Get('reference');
    my $reference        = $self->GetRefOrg();
    #my $chroms_reference  = $o_param->Get('chroms_reference');
    my $chroms_reference = $self->GetRefChrom();
    my $range = ( $o_param->IsDefined('range') ) ? $o_param->Get('range') : "0:0" ;
    my $l_flg = ( $o_param->IsDefined('ck_orien') ) ? $o_param->Get('ck_orien') : undef;
    my ($start,$end) = split(':',$range);
    $start = int($start);
    $end = int($end);

	# mettre en attributs ?
    my $o_ref = $o_io->GetSpeciesObserver( $reference );

	# ce ne serait pas l'attribut chr_length ?
	my $length = $o_ref->GetChromLength( $chroms_reference );

	if ( $start == 0 && $end == 0 )
	{
		warn("Something wrong with the range");
	}

    my @rvl;

    # We create 100 areas along the scale
    # We do not create them at very high zoom level (less than 100 pb displayed, is it only possible ?)
    if (($end - $start) >= 100)
    {
        my $step_px    = ($y2 - $y1) / 100;
        my $step_bases = ($end - $start) / 100;
        $y2 = $y1+10; # Starting at top of image, 10 pixels height
        my $pos= $start;    # The starting position, in base
        my $ref_long = $o_io->GetSpeciesLongName($reference);
        for (my $i=0; $i < 100; $i++)
        {
            my @a_coord = &ValidCoord($x1,$y1,$x2,$y2,$l_flg);
            push(@a_coord,"\"CHROMOSOME_CURSOR\"",int($pos),"\"$ref_long\"","\"$chroms_reference\"",$length,$start,$end);
            #push(@a_coord,"CHROMOSOME_CURSOR",int($pos),$ref_long,$chroms_reference,$length,$start,$end);
            push(@rvl,\@a_coord);

            $y1 += $step_px;
            $y2 += $step_px;
            $pos+= $step_bases;
        }
    }
    return \@rvl;
}

=head2 ValidCoord

 Title        : ValidCoord
 Usage        : ValidCoord($x1,$y1,$x2,$y2,$flag)
 Prerequisite : none
 Fonction     : Rounds the coordinates and check we do not have y1==y2 or x1==x2
                The area is at least 10 pixels in width or in height
                If y1==y2 We add 10 pixels to y2 to get a correct area
                If $flag != 0, we are in landscape mode: $x,$y are swapped
 Args         : $x1,$y1,$x2,$y2 : The coord to be validated
                $flag           : If 1, we are in landscape mode
 Returns      : $x1,$y1,$x2,$y2 : Coordonnees verifiees
 Globals      : none

=cut
sub ValidCoord
{
	my $x1 = shift;
    my $y1 = shift;
    my $x2 = shift;
    my $y2 = shift;
    my $flag=shift;

    if (!defined $flag)
    {
        $flag = 0;
    }

    # round the coord
    $x1 = int($x1);
    $x2 = int($x2);
    $y1 = int($y1);
    $y2 = int($y2);

    # reorder the coord (should not happen)
    #($x1,$x2) = ($x2,$x1) if ($x1>$x2);
    #($y1,$y2) = ($y2,$y1) if ($y1>$y2);

    # Mini 10x10 area
	if ($y2-$y1<10)
    {
        $y2 = $y1 + 10;;
    };
    if ($x2-$x1<10)
    {
        $x2 = $x1 + 10;
    }

    # swap x,y if landscape mode
    if ($flag!=0)
    {
        ($x1,$y1) = ($y1,$x1);
        ($x2,$y2) = ($y2,$x2);
    }

	return ($x1,$y1,$x2,$y2);
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
