# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

GenomeMap

=head1 DESCRIPTION

ChromosomeMap is used with NcdisplayMap


=head2 ATTRIBUTES

root_data		=> data directory
reference		=> reference organism short name
chromosome		=> reference chromosome number (or label)
targets_list	=> sorted list of organisms compared with the reference
level			=> synteny level of the comparison
view_level		=> synteny level show (must be inferior or equal to 'level')
start			=> start of the slice shown
end				=> end of the slice shown
gene_structure	=> boolean, checkbox value to display exons data
chr_lenght		=> lenght of the reference chromosome (store to read corresponding file only once)
boxes			=> hash of the boxes that will structured the image, each box corresponding to an organism representation (reference or compared)
ref_position	=> position of the reference box, could be use to manage boxes organisation
gene			=> optional, if searching option is initialize to find a reference gene
position		=> optional, if searching option is initialize to find a reference position


=head1 AUTHOR

Stahl Olivier : Olivier.Stahl@toulouse.inra.fr

=cut

package Planner::GenomeMap;

use strict;
use warnings;

use base qw( Planner );

use NARCISSE_utils;
use ParamParser;
use LipmError::ObjectException;
use LipmError::ParameterException;

#use base qw( Planner );
use base qw( LipmObject );
my $VERSION = do {my @r = (q$Rev: 1 $ =~ /\d+/g); $r[0]};

#Some constant parameters prior to writing them in a config file

use constant DOPLOT_TOP_MARGIN	=> 30;
use constant DOPLOT_BOTTOM_MARGIN => 0;
use constant DOPLOT_LEFT_MARGIN  => 30;
use constant DOPLOT_RIGHT_MARGIN => 0;

##################################################
###########			Constructor		 #############
##################################################
=head2 _Init

 Title        : _Init
 Usage        : _Init(o_io => $o_io,o_param => $o_param, ...)
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : $o_param -> ParamParser object
 Globals	  : objects attributes :
 					- root_data		: data directory
 					- reference		: short name of the main organism
 					- chromosome	: reference chromosome
 					- targets_list	: list of all organisms compared with reference.
 					- level			: synteny level of the comparison
 					- view_level	: synteny level show (must be inferior or equal to 'level')
 					- chr_length 	: length of the reference chromosome
 					- gene 			: search option (to search by name)
 					- position		: search option	(to search by position)
 					- gene_structure : boolean, checkbox value to display exons data

 Errors		  : LipmError::ObjectException, LipmError::ParameterException

=cut
sub _Init {
	my $self = shift;
	my %h_args = ( 
        o_io     => undef,
        o_param  => undef,
        o_logger => undef,
        @_
        );

    $self->SUPER::_Init( %h_args );
    
	my %h_param= ( reference => undef,
					target	  => undef,
					level	  => $SUPER::DEFAULT_LEVEL,
					program   => $SUPER::DEFAULT_PROG,
					@_
					);
					
	my @a_parent_parameters = qw( reference );
	
	my %h_parent_params;
	@h_parent_params{@a_parent_parameters}=@h_param{@a_parent_parameters};
		
	
	#TODO get rid of this call to the Planner.pm constructor as it will probably be removed
	#$self-> SUPER::_Init( %h_parent_params );
	
	my @specific_parameters=qw( target level program );
	foreach my $param ( @specific_parameters )
	{
		defined $h_param{$param} or throw LipmError::ParameterException("GenomeMap, parameter $param", 'missing',-text=>'The $param parameter is missing');
	}
	@{$self}{@specific_parameters}=@h_param{@specific_parameters};
	
    my $o_io = $self->_Get('_o_io');
	$self->{root_data} = $o_io->GetCfg('ROOT_DATA');
	
	$self->{_boxes} = {};

	return $self;
}

##################################################
###########			Procedures		 #############
##################################################

=head2

 Title        :
 Usage        :
 Prerequisite : Planner need to be initialized
 Function     :
 Returns      :
 Args         : box_name, the box label, X0,X1,Y0,Y1 the box coordinates
 Globals	  : none

=cut
sub SetBox
{
	my $self = shift;
	my $box_name = shift;
	my $bool_valid_box=0;

	my ($X0,$X1,$Y0,$Y1)=(shift,shift,shift,shift);

	if( 0 == _CheckValidIntegerCoordinates($X0,$X1,$Y0,$Y1) )
	{
		my $dbg = 'Box coordinates are not integers';
		throw LipmError::ObjectException( __PACKAGE__, 'x', -text => $dbg ) ;
	}

	$self->{_boxes}{$box_name}={X0 => $X0, X1 => $X1,Y0 => $Y0, Y1 => $Y1};

	return {X0 => $X0, X1 => $X1,Y0 => $Y0, Y1 => $Y1};
}

=head2

 Title        :
 Usage        :
 Prerequisite : Planner need to be initialized
 Function     :
 Returns      :
 Args         : box_name, the box label, X0,X1,Y0,Y1 the box coordinates
 Globals	  : none

=cut
sub SetDotplotBox
{
	my $self = shift;
	my $box_name = shift;

	my ($X0,$X1,$Y0,$Y1)=(shift,shift,shift,shift);

	return $self->SetBox($box_name,$X0,$X1,$Y0,$Y1);
}

=head2 _CheckValidIntegerCoordinates

 Title        : _CheckValidIntegerCoordinates
 Usage        : _CheckValidIntegerCoordinates(@coordinates)
 Prerequisite : none
 Function     : check if the array @coordinates contains only integer coordinates
 Returns      : 1 if this is true
 Args         : an array of numbers (to be tested as integers)
 Globals	  : none

=cut
sub _CheckValidIntegerCoordinates
{
	my @numbers=@_;
	my $bool_are_integers=1;

	foreach my $number (@numbers)
	{
		$bool_are_integers=$bool_are_integers and ($number =~/^\d+$/);
	}

	return $bool_are_integers;
}

=head2 GetBoxCoordinates

 Title        : GetBoxCoordinates
 Usage        : $o_planner -> GetBoxCoordinates(box_name)
 Prerequisite : Planner need to be initialized
 Function     : Get image coordinates of the box 'box_name'.
 Returns      : Array (X0,X1,Y0,Y1)
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub GetBoxCoordinates
{
	my $self = shift;
	my $box_name = shift;

	my @coordinates=@{$self->{_boxes}{$box_name}}{('X0','X1','Y0','Y1')};

	return @coordinates;
}

=head2 GetBoxHCoordinates

 Title        : GetBoxHCoordinates
 Usage        : $o_planner -> GetBoxHCoordinates(box_name)
 Prerequisite : Planner need to be initialized
 Function     : Get image coordinates of the box 'box_name'.
 Returns      : a ref on a hash {X0 => $X0, X1 => $X1,Y0 => $Y0, Y1 => $Y1};
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub GetBoxHCoordinates
{
	my $self = shift;
	my $box_name = shift;
	my %box_coordinates=();

	@box_coordinates{('X0','X1','Y0','Y1')}=@{$self->{_boxes}{$box_name}}{('X0','X1','Y0','Y1')};

	return \%box_coordinates;
}

=head2 BoxWidth

 Title        : BoxWidth
 Usage        : $o_planner -> BoxWidth(box_name)
 Prerequisite : Planner need to be initialized and the box to be created
 Function     :
 Returns      : the width of the box : X1-X0
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub BoxWidth
{
	my $self = shift;
	my $box_name = shift;

	return 	$self->{_boxes}{$box_name}{X1}-$self->{_boxes}{$box_name}{X0};
}

=head2 BoxHeight

 Title        : BoxHeight
 Usage        : $o_planner -> BoxHeight(box_name)
 Prerequisite : Planner need to be initialized and the box to be created
 Function     :
 Returns      : the width of the box : X1-X0
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub BoxHeight
{
	my $self = shift;
	my $box_name = shift;

	return 	$self->{_boxes}{$box_name}{Y1}-$self->{_boxes}{$box_name}{Y0};
}

=head2 BoxXMiddle

 Title        : BoxHeight
 Usage        : $o_planner -> BoxHeight(box_name)
 Prerequisite : Planner need to be initialized and the box to be created
 Function     :
 Returns      : the width of the box : X1-X0
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub BoxMiddlex
{
	my $self = shift;
	my $box_name = shift;

	return 	($self->{_boxes}{$box_name}{X0}+$self->{_boxes}{$box_name}{X1})/2;
}
=head2 BoxXMiddle

 Title        : BoxHeight
 Usage        : $o_planner -> BoxHeight(box_name)
 Prerequisite : Planner need to be initialized and the box to be created
 Function     :
 Returns      : the width of the box : X1-X0
 Args         : box_name = the box label
 Globals	  : boxes, BOX_WIDTH, BOX_HEIGHT

=cut
sub BoxMiddley
{
	my $self = shift;
	my $box_name = shift;

	return 	($self->{_boxes}{$box_name}{Y0}+$self->{_boxes}{$box_name}{Y1})/2;
}


=head2 GetBoxPosition

 Title        : GetBoxPosition
 Usage        : $o_planner -> GetBoxPosition(box_name)
 Prerequisite : Planner need to be initialized
 Function     : Get image position of the box 'box_name'.
 Returns      : position : left, right, center
 Args         : box_name = the box label
 Globals	  : boxes

=cut
sub GetBoxPosition
{
	my $self = shift;
	my $box_name = shift;
	$box_name = $self->{reference} if (!defined $box_name);

	my $rh_box = $self->GetBox($box_name);
	return $$rh_box{position};
}


=head2 GetBox

 Title        : GetBox
 Usage        : $o_planner -> GetBox(box_name)
 Prerequisite : Planner need to be initialized
 Function     : Get reference of the hash corresponding to this box.
 Returns      : Hash ref
 Args         : box_name = the box label
 Globals	  : boxes

=cut
sub GetBox
{
	my $self = shift;
	my $box_name = shift;
	$box_name = $self->{reference} if (!defined $box_name or !defined $self->{_boxes}{$box_name});

	return $self->{_boxes}{$box_name};
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;