#!/usr/bin/perl -I perl

# $Id$

=pod
=head1 Name

Probe.pm

=head1  SYNOPSIS

my $o_probe = New Probe( o_region => $o_suitable, rh_primer3 => $rh_probe );

=head1 DESCRIPTION

A simple object to store probe information.

=cut

=head1 SUBROUTINES

=cut

package Probe;

use strict;
use warnings;

use Data::Dumper;
use base qw( LipmObject );

use General;

=head2 function _Init

 Title    : _Init
 Usage    : The constructor
 Access   : Public
 Function : construct the probe object
            
 Args     : 
 Returns  : The object
 Globals  : none

=cut
sub _Init
{
	my $self = shift;
	my %h_args = ( o_region => undef,
				   rh_primer3 => undef,
				  @_
				  ); 
	$self->_Set( __o_region => $h_args{ o_region } );	
	
	$self->_Set( '__ampliconGC' => undef );
	$self->_Set( '__sequence' => undef );
	
	$self->_InitFromParams( $h_args{ rh_primer3 } );
	
	$self->_RetrieveProbeSequence();
					
	return $self;		
}

sub _InitFromParams
{
	my $self = shift;	
	my $rh_params = shift;
		
	foreach my $key ( keys %$rh_params )
	{
		my $key_name = "__\L$key";
		#print STDERR "$key_name $rh_params->{$key}\n";
		$self->_Set( $key_name => 	$rh_params->{ $key } );
	}			
	
	#primer3 output is 0-based hence +1
	my $start = $self->_Get( '__left_start' ) + 1;
	my $end   = $self->_Get( '__right_start' ) + 1;
					
	$self->_Set( '__start'    => $start, 
				 '__end'      => $end,
				 '__valid'    => 1,
				 '__selected' => 0
			   );	
}

sub _RetrieveProbeSequence
{
	my $self = shift;	
	
	my $o_region = 	$self->_Get( '__o_region') ;
	my $region_seq = $o_region->GetSequence();
	
	my ( $start, $end ) = $self->_Get( '__start', '__end' );
	my ( $offset, $length ) = ( $start - 1, $end - $start + 1 );
	
	my $probe_seq = substr($region_seq, $offset, $length);
	
	$self->_Set( '__sequence' => $probe_seq );
}



sub Overlap
{	
	my ( $o_a, $o_b ) = @_;	
	my ( $o_first, $o_second ) = ( $o_a->GetStart() <= $o_b->GetStart() ) ? ($o_a,$o_b) : ($o_b,$o_a);
	
	my $overlap = 0;
	if ( $o_second->GetStart() <= $o_first->GetEnd() )
	{
		$overlap = $o_first->GetEnd() - $o_second->GetStart() + 1;	
	}
		
	return $overlap;	
}

=head2 function 

 Title        : 
 Usage        : $o_xxx->_ExtractAmpliconProductPcrSequence( );
 Function     : Extract the sequence corresponding to the product amplified by the probes
 Args         : none
 Returns      : the DNA sequence
=cut
sub _ExtractAmpliconProductPcrSequence
{
	my $self = shift;
	my $o_region = 	$self->_Get( '__o_region') ;
	my $o_speciesObserver = $o_region->GetSpeciesObserver();
	my $chrom = $self->GetChr();
		
	my $chromosome_file = $o_speciesObserver->GetChromFastaFile( chrom => $chrom, -where => 'INSTANCE' );
		
	my $seq_in        = Bio::SeqIO->new( -file => "<$chromosome_file", -format => 'fasta' );
	my $seq_obj       = $seq_in->next_seq();
	
	my ( $start, $end ) = ( $self->GetChrStart(), $self->GetChrEnd() );
	my $seq_region    = $seq_obj->subseq( $start, $end );
	
	return $seq_region;
}

sub IsValidforFilter {
	my $self = shift;
	my %h_args = ( filter => 'none', 
	               @_ 
	              );	
	if ( $h_args{ filter } eq "valid" )
	{
		return $self->_Get( '__valid' );
	}
	if ( $h_args{ filter } eq "selected" )
	{
		return $self->_Get( '__selected' );
	}
	return 1;
}

# Some setters

sub Select {
	my $self = shift;	
	$self->_Set( '__selected' => 1) ;	
}

sub DeSelect {
	my $self = shift;	
	$self->_Set( '__selected' => 0) ;	
}

sub IsSelected {
	my $self = shift;	
	return $self->_Get( '__selected' );	
}

sub UnValid {
	my $self = shift;	
	$self->_Set( '__valid' => 0) ;		
}

sub SetNumHits {
	my $self = shift;	
	my $num = shift;
	$self->_Set( '__numHits' => $num) ;		
}

sub ProbeId{
	my $self = shift;		
	my ( $species, $chrom, $start, $end ) = ( $self->GetSpecies(), $self->GetChr(), $self->GetChrStart(), $self->GetChrEnd() );
	my $probe_id = join("_",($species, $chrom, $start, $end));	
	return $probe_id;
}

# Some accessors

sub GetSpecies{
	my $self = shift;
	my $o_region = 	$self->_Get( '__o_region') ;	
	return $o_region->GetSpecies();		
}

sub GetChr {
	my $self = shift;
	my $o_region = 	$self->_Get( '__o_region') ;	
	return $o_region->GetChr();	
}

sub GetChrStart {
	my $self = shift;
	my $o_region = 	$self->_Get( '__o_region') ;	
	return ($o_region->GetChrStart() + $self->_Get( '__start') - 1);	
}

sub GetChrEnd {
	my $self = shift;	
	my $o_region = 	$self->_Get( '__o_region') ;	
	return ($o_region->GetChrStart() + $self->_Get( '__end') - 1) ;	
}

sub GetRegion {
	my $self = shift;
	return $self->_Get( '__o_region') ;	
}

sub GetStart {
	my $self = shift;	
	return $self->_Get( '__start') ;	
}

sub GetEnd {
	my $self = shift;	
	return $self->_Get( '__end') ;	
}

sub GetSize {
	my $self = shift;	
	return $self->_Get( '__end') - $self->_Get( '__start') +1;	
}

sub GetLeftPrimer {
	my $self = shift;	
	return $self->_Get( '__left_sequence') ;	
}

sub GetRightPrimer {
	my $self = shift;	
	return $self->_Get( '__right_sequence') ;	
}

sub GetLeftTm {
	my $self = shift;	
	return $self->_Get( '__left_tm') ;	
}

sub GetRightTm {
	my $self = shift;	
	return $self->_Get( '__right_tm') ;	
}

sub GetLeftEndStability {
	my $self = shift;	
	return $self->_Get( '__left_end_stability') ;	
}

sub GetRightEndStability {
	my $self = shift;	
	return $self->_Get( '__right_end_stability') ;	
}

sub GetLeftSelfEnd {
	my $self = shift;	
	return $self->_Get( '__left_self_end') ;	
}

sub GetRightSelfEnd {
	my $self = shift;	
	return $self->_Get( '__right_self_end') ;	
}

sub GetLeftSelfAny {
	my $self = shift;	
	return $self->_Get( '__left_self_any') ;	
}

sub GetRightSelfAny {
	my $self = shift;	
	return $self->_Get( '__right_self_any') ;	
}

sub GetPairPenality {
	my $self = shift;	
	return $self->_Get( '__pair_penalty') ;	
}

sub GetPairComplAny {
	my $self = shift;	
	return $self->_Get( '__pair_compl_any') ;	
}

sub GetSequence{
	my $self = shift;	
	my $sequence = $self->_Get( '__sequence' );
	unless ( defined $sequence )
	{
		 my $ampliconseq = $self->_ExtractAmpliconProductPcrSequence();
		 $self->_Set( '__sequence' => $ampliconseq );	
	}
	return $self->_Get( '__sequence') ;	
}

sub GetAmpliconGC {
	my $self = shift;
	my $amplicon_gc = $self->_Get('__ampliconGC');
	unless ( defined $amplicon_gc )
	{
		$amplicon_gc = CalculateGc( $self->GetSequence() );
		$self->_Set( '__ampliconGC'	=> $amplicon_gc );
	}
	return $self->_Get('__ampliconGC');
}

sub GetAmpliconRepeat {
	my $self = shift;
	my $amplicon_repeat = $self->_Get('__ampliconRepeat');
	unless ( defined $amplicon_repeat )
	{
		$amplicon_repeat = CalculateRepeatFreq( $self->GetSequence() );
		$self->_Set( '__ampliconRepeat'	=> $amplicon_repeat );
	}
	return $self->_Get('__ampliconRepeat');
}

sub GetPrimerLeftSize
{
	my $self = shift;
	return length($self->GetLeftPrimer());	
}

sub GetPrimerRightSize
{
	my $self = shift;
	return length($self->GetRightPrimer());	
}

sub GetNumHits {
	my $self = shift;	
	return $self->_Get( '__numHits') ;		
}

sub sprint {
	my $self = shift;		
	return $self->_Get( '__start')."-".$self->_Get( '__end');
}

sub ChrInfos {
	my $self = shift;		
	my $o_region = 	$self->_Get( '__o_region') ;
	return $self->GetChrStart()."-".$self->GetChrEnd()." ".$self->GetSize();
}

sub OutOfPlaceHit {
	my $self = shift;
	my $opt_hit = shift;
	
	if ( $opt_hit->{ chr } ne $self->GetChr() ||
	     $opt_hit->{ end }  <  $self->GetChrStart() ||
	     $opt_hit->{ start } >  $self->GetChrEnd()
	)
	{
		return &TRUE;	
	}	
	return &FALSE;	
}

sub CalculateGc 
{
	my $seq = shift;

	my @seqarray = split('',$seq);
	my $count = 0;
	foreach my $base (@seqarray) 
	{
		$count++ if $base =~ /[G|g|c|C]/i;
	}
	my $len = $#seqarray+1;
	return $count / $len;
}

sub CalculateRepeatFreq
{
	my $seq = shift;

	my @seqarray = split('',$seq);
	my $count = 0;
	foreach my $base (@seqarray) 
	{
		$count++ if $base =~ /[a|c|g|t|N]/;
	}
	my $len = $#seqarray+1;
	return $count / $len;
}


=pod   OldStuff
        
=cut


1;
