
package Region;

=pod

=head1 Name

Region.pm

=head1  SYNOPSIS

my $o_suitable = New Region( id    => $id,
						     chr   => $chr,
							 start => $start,
						     end   => $end,
						     seq   => $sequence,
						     o_speciesObserver => $self->_Get( '__o_speciesObserver' )
						    );

=head1 DESCRIPTION

A simple Region object (chr, start, end) to manage a set of candidate probes

=cut

use strict;
use warnings;

use Data::Dumper;
use base qw( LipmObject);

use General;
use Probe;

use constant MAX_OVERLAP => 3_000;
use constant MAX_NEIGHBOUR_DIST => 50_000;

############################################################
#####                CONSTRUCTEUR                      #####
############################################################

=head2 function _Init

 Title    : _Init
 Usage    : The constructor
 Access   : Private
 Function : Set all arguments and create the probe array
 Args     : $chr     => the chromosome number
            $start   => the start position
            $end     => the end position
            $ra_suitables => the suitable subregion reference array
 Returns  : The object
 Globals  : none

=cut
sub _Init
{
	my $self = shift;	
	my %h_args = ( 
				   id          => undef,
				   chr         => undef,
				   start       => undef,
				   end         => undef,
				   seq         => undef,
				   o_speciesObserver => undef,					
					@_
				  );
	$self->SUPER::_Init( @_ );		  
	
	$self->_InitFromOptions( %h_args );	

	$self->_Set( '__rao_probes' => [],
				 '__rao_connected_probes' => []
				 );		
	
	return $self;
} # _Init

sub _InitFromOptions
{
	my $self = shift;
	my %h_args = @_;
	
	LipmError::Function::Assert( %h_args );	
	
	$self->_Set( '__o_speciesObserver' => $h_args{ o_speciesObserver },
				 '__id'          => $h_args{ id },
				 '__chr'         => $h_args{ chr },
				 '__start'       => $h_args{ start },
	             '__end'         => $h_args{ end },
	             '__seq'         => $h_args{ seq } 
	            );	   
}


sub FarAway
{	
	my $self = shift;
	my $hit = shift;
		
	if ( $hit->{ chr } ne $self->GetChr() )
	{
		return &TRUE;
	} 
	# The hit is declared as far away if either :		
	#     - s_end is smaller and far away from region start
	#     - s_start is greater and far away than region end
	# (far away means a distance larger than region size)
	if ( $self->GetChrStart() - $hit->{ s_end } > $self->GetSize ||   $hit->{ s_start } - $self->GetChrEnd() > $self->GetSize )
	#Changing from region size to a larger spectrum
	#if ( $self->GetChrStart() - $hit->{ s_end } > &MAX_NEIGHBOUR_DIST ||   $hit->{ s_start } - $self->GetChrEnd() > &MAX_NEIGHBOUR_DIST )
	{
		return &TRUE;	
	}	
	return &FALSE;
}

sub GetSpeciesObserver{
	my $self = shift;	
	return $self->_Get( '__o_speciesObserver' );
}

sub GetSpecies{
	my $self = shift;	
	my $o_speciesobserver = $self->_Get( '__o_speciesObserver' );
	return $o_speciesobserver->GetShortName();
}

sub GetChr {
	my $self = shift;	
	return $self->_Get( '__chr') ;	
}

sub GetChrStart {
	my $self = shift;	
	return $self->_Get( '__start') ;	
}

sub GetChrEnd {
	my $self = shift;	
	return $self->_Get( '__end') ;	
}

sub GetSize {
	my $self = shift;	
	return $self->_Get( '__end') - $self->_Get( '__start') +1;	
}

sub GetSequence {
	my $self = shift;	
	return $self->_Get( '__seq');	
}

sub GetProbes {
	my $self = shift;	
	return 	$self->_Get( '__rao_probes' );	
}

sub GetValidProbes {
	my $self = shift;
	return $self->GetAllProbes( filter => 'valid' );
}

sub sprint {
	my $self = shift;
	return 	$self->_Get( '__chr').":".$self->_Get( '__start')."-".$self->_Get( '__end');
}

sub GetAllProbes 
{
	my $self = shift;
	my %h_args = ( filter => 'none', 
	               @_ 
	              );
	my $rao_probes =$self->_Get( '__rao_probes' ); 
	my @a_validprobes = ();
	foreach my $o_probe ( @$rao_probes )
	{
		if ( $o_probe->IsValidforFilter( filter => $h_args{ filter } ) )
		{
			push @a_validprobes, $o_probe;
		}
	}
	return \@a_validprobes;	
}


=head2 function PushProbes
 Title        : PushProbes
 Usage        : $region->PushProbes( $o_Probe );
 Function     : Push a new o_probe in the probe array of the region object
 Access       : Public
 Args         : $o_probe :  a reference on a probe object
=cut
sub PushProbe
{
	my ( $self, $o_probe ) = @_;
	
	my $rao_probes = $self->_Get( '__rao_probes' );
	push( @$rao_probes, $o_probe );
}

sub GetNbProbes
{
	my $self = shift;	
	my $rao_probes = $self->_Get( '__rao_probes' );
	return scalar @$rao_probes;
}

sub CheckProbesPairEnd
{
	my $self = shift;	
	my $rao_repprobes = $self->_Get( '__rao_repprobes' );	
		
	$self->GenerateGlintPairEndInputFile();		
}

=head2 Procedure DetectRepresentativeProbes 
 Title        : DetectRepresentativeProbes
 Usage        : $o_xxx->DetectRepresentativeProbes(  );
 Function     : SDelect a subset of non-overlapping probes spanning the region
					Given a set of probes, construct an overlap graph (only strong overlap are considered : more than &MAX_OVERLAP)
					For each connected component select a single probe (the largest one)
 Args         : none
 Returns      : the number of representative probes
=cut
sub DetectRepresentativeProbes
{
	my $self = shift;		
	$self->ConnectedProbes();
	
	my $rao_repprobes = $self->_Get( '__rao_repprobes' );
	my $rao_connected_probes = $self->_Get( '__rao_connected_probes' );
	
	my $nb = 0;
	foreach my $ra_coco ( @$rao_connected_probes )
	{
		my @a_probes = sort { $b->GetSize() <=> $a-> GetSize() } @{$ra_coco};
		my $o_selprobe = $a_probes[0];
		$o_selprobe->Select();	
		#print $o_selprobe->ChrInfos()."\n";
		$nb++;
	}	
	return $nb;		
}

=head2 Procedure  ConnectedProbes
 Title        : ConnectedProbes
 Usage        : $o_xxx->ConnectedProbes(  );
 Function     : Detect connected components
 Args         : none
=cut
sub ConnectedProbes
{
	my $self = shift;
	
	my $rao_probes = $self->GetValidProbes();
	my @ao_probes = @$rao_probes;
	
	my @a_sortedprobdes = sort { $a->GetStart() <=> $b->GetStart() } @$rao_probes;
	my @a_nodes = @a_sortedprobdes;	
	my $rh_graph = _SetOverlapGraph( \@a_nodes );

	my $size = scalar @a_nodes;
	my %h_rep = ();		
	my %h_coco = ();
	for (my $i=0; $i < $size ; $i++ )
	{
		$h_rep{ $i } = $i;
		$h_coco{ $i }{ $i }++;	
	}
	my @a_edges = @{$rh_graph->{edges}};
	foreach my $edge ( @a_edges )
	{
		if ( $h_rep{ $edge->{u} } != $h_rep{ $edge->{v} } )
		{
			my $n = &Union( $edge, \%h_rep, \%h_coco );
		}	
	}		
	
	my $rao_connected_probes = $self->_Get( '__rao_connected_probes' );
	foreach my $coco ( keys %h_coco )
	{	
		my @a_coco = ();
		foreach my $n ( keys %{$h_coco{ $coco }} )
		{
			push @a_coco, $a_nodes[$n];
		}	
		push @{$rao_connected_probes}, \@a_coco;
	}
}

sub Union
{
	my ( $edge, $rh_rep, $rh_coco ) = @_;	
	
	my $a_n =  $edge->{u};
	my $b_n =  $edge->{v};	
	my $rep_a = $rh_rep->{ $a_n };
	foreach my $node ( keys %{$rh_coco->{$b_n}} )
	{
		$rh_coco->{$rep_a}{$node}++;
		$rh_rep->{$node} = $rep_a;
	}
	delete $rh_coco->{$b_n};
	return $a_n;	
}

sub _SetOverlapGraph
{
	my $rao_nodes = shift;
	
	my @a_nodes = @$rao_nodes;
	my @a_edges = ();
	
	my $size = scalar @a_nodes;
	for (my $i=0; $i < $size ; $i++ )
	{
		for (my $j=$i+1; $j < $size ; $j++ )
		{
			my $overlap = &Probe::Overlap( $a_nodes[$i], $a_nodes[$j] );
			if ( $overlap >= &MAX_OVERLAP )
			{
				my %h_edge = ( u => $i, v => $j, w => $overlap );
				push @a_edges, \%h_edge;
			}
			
		}
	}
	my %h_graph = ( nodes => \@a_nodes, edges => \@a_edges );
	return 	\%h_graph;
}


1;
