# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Sequence_utils;

=pod

=head1 NAME

 Sequence_utils - a module that provide function to deal with sequences.

=head1 DESCRIPTION

 As there is no Sequence object, at least this module provide some functions to deal with biological sequences.

=cut

use strict;
use warnings;

use Data::Dumper;

my $VERSION = 0.1;

# our Modules
# use Logger;

=head2 Function FormatSequenceLine

 Usage      : my $seq = FormatSequenceLine( seq => $seq, -len => $len);
 Prerequiste: none
 Function   : will format any string to a multi line string with
                  with no more than $len characters.
 Returns    : a string, generally multi lines with no more than $len char per line.
 Args       : $seq, the string to format
              $len, integer [70] the max. nb of character per lines
 Globals    : none

=cut

sub FormatSequenceLine
{
	my %h_args = ( seq => undef, -len => 70, @_ );
	my ( $sequence, $length_line ) = @h_args{ qw( seq -len ) };

	if( !defined $sequence )
	{	&General::ShowDebug( { deep=>2 }, 'sequence not defined !' )	}

	my $size_sequence = length($sequence);
	my $nb_lines = $size_sequence/$length_line;

	# un arrondi superieur
	$nb_lines = int($nb_lines)+1  if( $nb_lines =~ /\./ );

	my $formatted_seq = '';
	for (my $j=0; $j<$nb_lines; ++$j)
	{
		$formatted_seq .= substr($sequence,($j)*$length_line,$length_line)."\n";
	}

	return $formatted_seq;
}


=head2 Unit2Base

 Usage        : my $int = Unit2Base( $str )
 Function     : convert a position with unit like 2kb, 2Mb, 2Gb to integer.
                     1    => 1
                     1 kb => 1_000
                     1 Mb => 1_000_000
                     1 Gb => 1_000_000
                space between position and unit is tolerated.
 Returns      : $int, return 1 if string mismatchs.
 Args         : $str like \d+\s*[kb|Mb|Gb]

=cut

sub Unit2Base
{
	my ( $string ) = @_;

	return undef  if( !defined $string ); #throw error ?

	return $string  if( $string =~ /^\d+$/ );

	return $1 * 1_000          if( $string =~ /^([\d\.]+)\s*kb$/i );
	return $1 * 1_000_000      if( $string =~ /^([\d\.]+)\s*Mb$/i );
	return $1 * 1_000_000_000  if( $string =~ /^([\d\.]+)\s*Gb$/i );

	return 1; #throw error ?
}


=head2 Base2Unit

 Usage    : my $str = Base2Unit( $int );
            my $str = Base2Unit( $int, { -format => '%.2f',
                                         -max => 100,
                                         -ra_units => [ qw( b, kb, Mb ) ] } );
 Function : convert a position without unit to a position with unit
                     1 => 1
                 1_000 => 1kb
             1_000_000 => 1Mb
         1_000_000_000 => 1Gb
 Returns  : string
 Args     :
     $int, integer, the number of base.
     $rh_options, ref hash containing options :
         -format, string, [%d], used by sprintf to format the string.
         -max, integer, [1000], the returned value will be lower than it.
         -ra_units, array ref of units to use [ '', 'kb', 'Mb', 'Gb' ].

=cut

sub Base2Unit
{
	my ( $value, $rh_options ) = @_;

	my %h_opt = ( -format   => '%d',
	              -max      => 1000, # will return a number <1000 what the unit is.
	              -ra_units => [ '', 'kb', 'Mb', 'Gb' ],
	              %$rh_options );

	my $i;
	for( $i = 0; $h_opt{ -max } < $value; ++$i )
	{	$value /= 1_000;	}

	my $form_value = sprintf( $h_opt{ -format }, $value );
	my $ra_units   = $h_opt{ -ra_units };

	return "$form_value $$ra_units[$i]";
}


1;

__END__

=pod

=head1 AUTHOR

sebastien.letort@toulouse.inra.fr

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

