
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
#	sebastien.letort@toulouse.inra.fr
#	Created: september 03, 2009
#	$Id: SequencesLocator.pm 3728 2011-09-28 13:24:44Z manu $
#

package SequencesLocator;

=pod

=head1 NAME

 SequencesLocator - localise sequence into a genome and integrate to Narcisse.

=head1  SYNOPSIS

 # abstract class
 my $seq_int = New SequencesLocator::X( file      => $file,
                                          o_species => $o_human );
 # compute data
 $seq_int->LaunchGlint();
 $seq_int->CureData();
 # or
 $seq_int->GenerateData();

 #Get them
 $seq_int->GetNarcissePrintableData();
 # or
 $seq_int->GetData();

=head1 DESCRIPTION

 This class allows to localise sequences into a genome which belongs to Narcisse.
 Then the position of these sequence can be retrieve in a format that allow to
 generate Narcisse formatted file.

 It was first used with SNPs, then with ESTs.

=cut

use strict;
use warnings;

use Data::Dumper;
use IO::File;
use File::Temp;

use LipmError::Function;
use Runner;
use FastaReader;
use General;

use base qw( LipmObject );

#use General;

=head2 method _Init

 Usage        : my $o_x = New SequencesLocator( file      => $file,
                                                o_species => $o_species,
                                                -raw_output => $output,
												-huge_file  => $huge_file,
                                                -o_logger => $o_logger );
 Abstract class.
 Fonction     : initialise an instance of SequencesLocator class.
 Returns      : a valid SequencesLocator object
 Args         :
	$file, string, filepath to a fasta file containning sequence(s) to align.
	$o_species, SpeciesObserver object,
	$output, filepath, output file of the alignment program.
			If none is passed, the file will be removed after work.
	$huge_file, booleen [FALSE], if true, an optimisation for huge file (lot of sequences)
			will be used. _the file will be rewritten indiced.
 Attributes   :
    _file, string, filepath, file to localise.
    _raw_output, string, filepath, output file of the alignment program.
    __delete_output, boolean, &TRUE if _raw_output should be removed.
    _o_species, SpeciesObserver object.
    __rh_data, hash ref, data generated, values are Feature object (SNP or EST, coming soon)

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger   => undef,
	               file        => undef,
	               o_species   => undef,
	               -raw_output => undef,
	               -huge_file  => &FALSE,
	               @_ );
	$self->SUPER::_Init( %h_args );

	# assert mandatories
	LipmError::Function::Assert( %h_args );

	$self->_Set( _file      => $h_args{ file },
				 _o_species => $h_args{ o_species },
				 __rh_data  => {}	);
	$self->__SetRawOutput( $h_args{ -raw_output } );

	$self->__IndexFile() if( &TRUE == $h_args{ -huge_file } );

	return;

	# write a copy of _file, labels are replaced with indices
	sub __IndexFile
	{
		my $self = shift;

		my $indiced_file = $self->_Get( qw( _file ) );
		my $o_reader     = New FastaReader::SnpReader();

		$indiced_file    =~ s/\.(.+)$/.indiced.$1/;
		$o_reader->WriteIndicedFasta( $indiced_file );

		return;
	}

	sub __SetRawOutput
	{
		my $self = shift;
		my( $outfile ) = @_;

		my $bool = &FALSE;
		my $o_species = $self->_Get( '_o_species' );
		if( !defined $outfile )
		{
			my $tmp_dir  = $o_species->GetTmpPath( 'INSTANCE', &FALSE );
			my $ofh_tmp  = new File::Temp( TEMPLATE => 'glint_out_XXXX',
										  UNLINK   => 0,
										  DIR      => $tmp_dir );
			$outfile = $ofh_tmp->filename();
			$bool    = &TRUE;
		}

		# j'ai un doute sur l'utilité de __delete_output : si on utilise UNLINK => 1,
		# l'objet File::Temp ne sera détruit qu'avec l'objet (si j'ai bien compris).
		$self->_Set( _raw_output     => $outfile,
					 __delete_output => $bool );

		return;
	}

} # _Init


=head2 DESTROY

 Usage    : $o_x->DESTROY()
 Procedure: remove _raw_output file.
 Args     : none
 Error    : none.

=cut

sub DESTROY
{
	my $self = shift;

	my( $outfile, $delete_it ) = $self->_Get( qw( _raw_output __delete_output ) );
	unlink( $outfile )  if( &TRUE == $delete_it );

	return;
} #DESTROY


=head2 GenerateData

 Usage    : $o_x->GenerateData()
 Procedure: This is the method that final user should use.
            Developpers should implement this method in daughter class.
 Args     : none.
 Error    : none.

=cut

sub GenerateData
{
	my $self = shift;

	$self->LaunchGlint();
	$self->CureData();

	return;
} #GenerateData


=head2 LaunchGlint

 Usage        : $o_x->LaunchGlint( -glint_opt => $glint_opt )
 Function     : Launch glint _o_species as a reference.
 Args         :
    $glint_opt, string optionnal, any option to add to glint
 Error        : none.

=cut

sub LaunchGlint
{
	my $self   = shift;
	my %h_args = (	-glint_opt => '',
					@_ );

	my @a_attrs = qw( _raw_output _file _o_species );
	my( $outfile, $tgt_db, $o_species ) = $self->_Get( @a_attrs );
	if (-s $outfile )
	{
		my $msg = "WARNING - $outfile already exists - Not executing glint";
		warn $msg;
		$self->_Trace("$msg\n",'*','IsV');
	}
	else
	{
		my $o_io      = $o_species->GetInstanceObserver();
		my( $glint, $glint_opt ) = $o_io->GetCfg( qw( ROOT_GLINT_PURE GLINT_SEQ_LOCATOR ) );
		my $prog      = "$glint $glint_opt $h_args{ '-glint_opt' }";

		my $ra_index  = $o_species->GetChromIndexFilepaths();
		my $ref_db    = $$ra_index[ 0 ]; # il n'y a qu'un seul fichier indexe apparemment

	# warn "j'utilise un fichier de test perso comme ref.";
	# # $ref_db = '/home/sletort/test_glint/human21.fasta';
	# $ref_db = '/home/sletort/test_glint/humanMT.fasta';
		my $cmd       = "$prog -o $outfile $ref_db $tgt_db";

		$self->_Trace( "running $cmd\n", '*', 'Is' );
		my ( $status, $killed ) = RunExt( -cmd => $cmd );
		$self->_Trace( "end of $cmd status($status)\n", '*', 'Is' );
	}
	return;
} #LaunchGlint


=head2 CureData

 Usage        : $o_x->CureData()
 Function     : parse _raw_output file and store valuable data in __rh_data.
 Args         : $o_species, SpeciesObserver.
 Error        : none.

=cut

sub CureData
{
	my $self   = shift;

	throw LipmError::ObjectException( $self, 'abstract method' );

	# code given as example
# 	my @a_attrs = qw( _raw_output _file );
# 	my( $raw_output, $file ) = $self->_Get( @a_attrs );

# 	my $o_reader= New FastaReader( $file );

# 	my $o_ndr   = New DataReader();
# 	$o_ndr->Prepare( -from => $raw_output );
# 	my $ra_data = $o_ndr->FetchallArrayref(-line_type => 'HASH');

# 	foreach my $rh_raw_data ( @$ra_data )
# 	{
# 		my $o_feat  = New Feature( ... );
# 		my $feat_id = $o_feat->GetIdentifier();
# 		my $chrom   = $o_feat->GetChrom();
# 		$h_features{ $chrom }{ $feat_id } = $o_feat;
# 	}
# 	$self->_Set( __rh_data => \%h_features );

	return;
} #CureData


=head2 GetData

Usage        : my $rh_snp = $o_x->GetData()
Function     : accessor for __rh_data.
Returns      : $rh_snp, hash ref, value is a Feature object (coming next SNP or EST )
Args         : none
Error        : none

=cut

sub GetData
{
	my $self = shift;

	my $rh_data = $self->_Get( '__rh_data' );

	return $rh_data;
} #GetData


=head2 GetNarcissePrintableData

 Usage        : my $rh_data = $o_x->GetNarcissePrintableData()
 Function     : accessor for snp data.
 Returns      : $rh_data, hash ref, keys are chrom label, values hash ref snp.
 Args         : none
 Error        : none

 Note : une methode PrintData capable d'appeler une methode format ou stringify de o_snp serait plus jolie.

=cut

sub GetNarcissePrintableData
{
	my $self = shift;

	my $rh_data = $self->GetData();
	my $rh_out  = {};

	foreach my $o_feature ( values %$rh_data )
	{
		my $rh_feat = $o_feature->ToRefHash();
		my $chrom   = $o_feature->GetChrom();
		my $feat_id = $o_feature->GetIdentifier();

		$$rh_out{ $chrom }{ $feat_id } = $rh_feat;
	}

	return $rh_out;
} #GetNarcissePrintableData


=head2 Accessor GetSpecies, IsHugeFile

 Usage    : my $o_species = $o_x->GetSpecies()
 Returns  : $o_species
 Error    : none

=cut

sub GetSpecies
{
	my $self = shift;

	return $self->_Get( '_o_species' );
} #GetSpecies

sub IsHugeFile
{
	my $self = shift;

	return $self->_Get( '__huge_file' );
} #GetSpecies


1;

__END__
