
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
#	sebastien.letort@toulouse.inra.fr
#	Created: september 03, 2009
#	$Id: Generic.pm 3728 2011-09-28 13:24:44Z manu $
#

package SequencesLocator::Generic;

=pod

=head1 NAME

 SequencesLocator::Generic - localise sequence into a genome and integrate to Narcisse.

=head1  SYNOPSIS

 my $seq_int = New SequencesLocator::Generic( file      => $file,
                                          o_species => $o_human );
 # compute data
 $seq_int->LaunchGlint();
 $seq_int->CureData();
 # or
 $seq_int->GenerateData();

 #Get them
 $seq_int->GetNarcissePrintableData();
 # or
 $seq_int->GetData();

=head1 DESCRIPTION

 This class allows to localise sequences into a genome which belongs to Narcisse.
 Then the position of these sequence can be retrieve in a format that allow to
 generate Narcisse formatted file.

 It was first used with SNPs, then with ESTs.

=cut

use strict;
use warnings;

use Data::Dumper;
use IO::File;
use File::Temp;

use LipmError::Function;
use Runner;
use FastaReader;
use General;

use base qw( LipmObject );

#use General;

=head2 method _Init

 Usage        : my $o_x = New SequencesLocator::Generic( file      => $file,
                                                o_species => $o_species,
                                                -raw_output => $output,
												-huge_file  => $huge_file,
                                                -o_logger => $o_logger );
 Abstract class.
 Fonction     : initialise an instance of SequencesLocator class.
 Returns      : a valid SequencesLocator object
 Args         :
	$file, string, filepath to a fasta file containning sequence(s) to align.
	$o_species, SpeciesObserver object,
	$output, filepath, output file of the alignment program.
			If none is passed, the file will be removed after work.
	$huge_file, booleen [FALSE], if true, an optimisation for huge file (lot of sequences)
			will be used. _the file will be rewritten indiced.
 Attributes   :
    _file, string, filepath, file to localise.
    _raw_output, string, filepath, output file of the alignment program.
    _o_species, SpeciesObserver object.

=cut



=head2 CureData

 Usage        : $o_x->CureData()
 Function     : parse _raw_output file and store valuable data in __rh_data.
 Args         : $o_species, SpeciesObserver.
 Error        : none.

=cut

sub CureData
{
	my $self   = shift;

	# code given as example
	my @a_attrs = qw( _raw_output _file );
	my( $raw_output, $file ) = $self->_Get( @a_attrs );

	my $o_reader= New FastaReader( $file );

	my $o_ndr   = New DataReader::Whole();
	$o_ndr->Prepare( -from => $raw_output );
	my $ra_data = $o_ndr->FetchallArrayref(-line_type => 'HASH');

	die "NOT YET IMPLEMENTED";
	# foreach my $rh_raw_data ( @$ra_data )
	# {
		# my $o_feat  = New Feature( ... );
		# my $feat_id = $o_feat->GetIdentifier();
		# my $chrom   = $o_feat->GetChrom();
		# $h_features{ $chrom }{ $feat_id } = $o_feat;
	# }
	# $self->_Set( __rh_data => \%h_features );

	return;
} #CureData


=head2 GetData

Usage        : my $rh_snp = $o_x->GetData()
Function     : accessor for __rh_data.
Returns      : $rh_snp, hash ref, value is a Feature object (coming next SNP or EST )
Args         : none
Error        : none

=cut

sub GetData
{
	my $self = shift;

	my $rh_data = $self->_Get( '__rh_data' );

	return $rh_data;
} #GetData


=head2 GetNarcissePrintableData

 Usage        : my $rh_data = $o_x->GetNarcissePrintableData()
 Function     : accessor for snp data.
 Returns      : $rh_data, hash ref, keys are chrom label, values hash ref snp.
 Args         : none
 Error        : none

 Note : une methode PrintData capable d'appeler une methode format ou stringify de o_snp serait plus jolie.

=cut

sub GetNarcissePrintableData
{
	my $self = shift;

	my $rh_data = $self->GetData();
	my $rh_out  = {};

	foreach my $o_feature ( values %$rh_data )
	{
		my $rh_feat = $o_feature->ToRefHash();
		my $chrom   = $o_feature->GetChrom();
		my $feat_id = $o_feature->GetIdentifier();

		$$rh_out{ $chrom }{ $feat_id } = $rh_feat;
	}

	return $rh_out;
} #GetNarcissePrintableData


=head2 Accessor GetSpecies, IsHugeFile

 Usage    : my $o_species = $o_x->GetSpecies()
 Returns  : $o_species
 Error    : none

=cut

sub GetSpecies
{
	my $self = shift;

	return $self->_Get( '_o_species' );
} #GetSpecies

sub IsHugeFile
{
	my $self = shift;

	return $self->_Get( '__huge_file' );
} #GetSpecies


1;

__END__
