#
# 	sebastien.letort@toulouse.inra.fr
#	Created: August 05, 2008
#	$Id: SlidingWindowCalculator.pm 3696 2011-07-22 14:32:16Z sletort $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package SlidingWindowCalculator;

use strict;
use warnings;

use base qw( ThroughSpeciesComputer );

=pod

=head1 NAME

 SlidingWindowCalculator - a root class that manage a sliding window over species chromosomes

=head1  SYNOPSIS

	# do not instanciate : inherit it and overwrite this method.
	use base ( SlidingWindowCalculator )

	sub _ComputeDataOnWindow
	{
		my $self = shift;
		my ( $o_slice ) = @_;
		...
		return { value => $val };
	}

	(and eventually this one)
	sub _GetHeaderFile
	{
		...
		return $ra_headers;
	}

=head1 DESCRIPTION

	this class compute a sliding window over chromosomes and species.
	Daughter class must overwrite _ComputeDataOnWindow methods to produce data.

	you can specify none or more species to study, none or more chromosome too.
	If no species is given, all species of the instance will be used.
	If no chromosome is given, all chrom of the species will be used.
	You can specify window size and step size of course.
	You can also specify where to start and where to end the calculation.

=cut

use Data::Dumper;
use Error qw( :try );
use File::Path;
use File::Spec;


use General;
use NARCISSE_utils;
use Configuration::InstanceObserver;
use LipmError::ParameterException;
use LipmError::IOException;

use constant DEBUG => 1;


=head2 method _Init

 Usage        : $self->SUPER::_Init( %h_args );
 Prerequisite : none
 Fonction     : initialise a SlidingCalculator class.
 Returns      : a valid SlidingCalculator object
 Args         : %h_args as follow :
	'-species', [empty] list ref of species short name
	'-chrom', [empty] list ref of chrom label
	'-window', [100_000] positive non null integer, the window width
	'-step', [0] positive integer, the step in bases used when sliding the window
			     if 0, same value as window width (-window)
	'-start', [1] integer, if negative, the window will start that far from the end of the chromosome
	'-end', [-1] integer, where the sliding window will stop.
                 If negative, it will use all chromosome length.
	'-graph', string, a keyword representing the way data
                 will be represented. Allowed values are :
                      'line_plots', 'scatter_plots', 'histograms','tiles' or 'heat_map'.
    '-feature', string, the name of the subdir that will contain
			all data for ONE representation (image).
	'-extension', [dat] string, the extension of the file generated
	'-file_location', string, can be 'INSTANCE' or 'SESSION'.
			'SESSION', default value, will print data into session directory.
			'INSTANCE', will print data into instance data directory.
 Attributes  :
 		#~ _ra_species, ref to a list of short species names
 		             #~ if the list is empty, all species of the instance
 		             	#~ will be used.
 		             #~ if one species does not belong to the instance
			#~ _ra_chroms, ref to a list of chromosome labels
 		_window, positive non null integer, the size of the window
 					_the length of the subsequence to examine
 		_step, positive integer, the number of base that you have to move the window
 		       before computing another treatement
 		       if null, its value will be _window / 2
 		_start, integer, if _start is negative, the window
              		will start that far from the end of the chromosome
 		_end, integer, where the sliding window will stop.
              		If negative, it will use all chromosome length.
         _extension, string, the extension of the file generated
         _file_location, string, where data will be written, into session or instance

 		=> you do one treatement onto one window every step base from start to end.

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -window    => 100_000,
	               -step      => 0,
	               -start     => 1,
	               -end       => -1,
	               -feature   => '1',
	               -graph     => 'scatter_plots',
	               -extension => 'dat',
	               -file_location => 'SESSION',

	               # mother
# 					-species   => [],
# 					-chroms    => [],
# 					-o_logger  => undef,
	              @_ );

	# set class attributes
	my %h_attr = ();
	@h_attr{ qw( _window _step ) }   = @h_args{ qw( -window -step ) };
	@h_attr{ qw( _start _end ) }     = @h_args{ qw( -start -end )};
	@h_attr{ qw( _graph _feature ) } = @h_args{ qw( -graph -feature )};
	$h_attr{ _extension }            = $h_args{ -extension };
	$h_attr{ _file_location }        = $h_args{ -file_location };
	$self->_Set( %h_attr );

	# mother
	$self->SUPER::_Init( %h_args );

	my( $win, $step ) = $self->_Get( qw( _window _step ) );
	my $msg = ref( $self ) . " launch for ";
	$msg   .= "window = $win\tstep = $step\n";
	$self->_Trace( $msg, '.', 'IC' );

	return;
} # _Init


=head2 procedure _CheckAttributesAndSetDefaults

	Title        : _CheckAttributesAndSetDefaults
	Usage        : $bool = $o_xx->_CheckAttributesAndSetDefaults();
	Prerequisite : none
	Procedure    : check some attributes. cf _Init doc.
	Args         : none
	Error        : LipmError::ParameterException

=cut

sub _CheckAttributesAndSetDefaults
{
	my $self = shift;

	# for species
	$self->SUPER::_CheckAttributesAndSetDefaults();

	my( $win, $step ) = $self->_Get( qw( _window _step ) );

	# check step - window param
	if( $win <= 0 )
	{
		my $msg = "window parameter can't be negative nor null.";
		throw LipmError::ParameterException( 'window', 'value',
		                      -text => $msg, -should_be => ' > 0 ' );
	}

	if( $step < 0 )
	{
		my $msg = "step parameter can't be negative.";
		throw LipmError::ParameterException( 'parameter', 'value',
		                      -text => $msg, -should_be => ' >= 0 ' );
	}

	# cas particulier step = 0;
	if( $step == 0 )
	{	$self->_Set( _step => $self->{ _window } );	}

	return;
}


=head2 Procedure _ComputeData

 Title        : _ComputeData
 Usage        : $o_xx->_ComputeData();
 Prerequisite : none
 Procedure    : compute bounds between which window will move.
                move the window, get results, get headers, and write data.
 Args         : $o_sla, a SliceAdaptor object (API)
                $rh_chrom, a hash ref with at least keys
                	len     => integer, chrom length
                	chrName => string, chrom name
 Error        : LipmError::IOException,
                	thrown by NARCISSE_utils::PrintNarcisseData

 TODO  : A revoir : public avec args ou protege avec current ?
=cut

sub _ComputeData
{
	my $self = shift;

	my $rh_current = $self->_Get( '_rh_current' );
	my $o_sla   = $$rh_current{ o_slice_adaptor };
	my $chrom   = $$rh_current{ chrom };
	my $len     = $$rh_current{ rh_chroms }{ $chrom }{ length };
	my $chrName = $$rh_current{ rh_chroms }{ $chrom }{ chr };

	# Synchronize window start and window end regarding chromosome length
	my $rh_bounds = $self->__SynchronizeBounds( $o_sla, $len );

	my $msg = "From $$rh_bounds{ start } to $$rh_bounds{ end }\n";
	$self->_Trace( $msg, '.', 'IC' );

	my $rh_data = $self->__MoveWindowOverChromosome( $o_sla,
	                                       $chrName, $rh_bounds );

	$self->_Trace( "Data generated\n", '.', 'IC' );

	return $rh_data;
}


=head2 function __MoveWindowOverChromosome

	Title        : __MoveWindowOverChromosome
	Usage        : $rh_data = $o_xx->__MoveWindowOverChromosome( $o_slice_adaptor, $chrom, $rh_bounds );
	Prerequisite : none
	Function     : this function will read chromosome sequence
	               from $$rh_bounds{ start } to $$rh_bounds{ end }
	Returns      : $rh_data, a hash ref of data.
	               	keys are index.
	               	values are hash ref, whose keys are colum's name.
	Args         : $o_slice_adaptor, a SliceAdaptor (API)
	               $chrom, chrom name
	               $rh_bounds, bounds of the chrom slice to analyse.

=cut

sub __MoveWindowOverChromosome
{
	my $self = shift;
	my ( $o_slice_adaptor, $chrom, $rh_bounds ) = @_;

	my( $win, $step ) = $self->_Get( qw( _window _step ) );

	my $win_start = $$rh_bounds{ start };
	my $win_end   = $win_start + $win - 1;

	# chrom name circos-compliant
	my $chr_name = $o_slice_adaptor->db()->species() . '-' . $chrom;

	# count through each region
	my %h_data = ();
	my $i = 1;
	while( $win_end <= $$rh_bounds{ end } )
	{
# warn "$win_start, $win_end\n";
		#  the slice ie the window
		my $o_slice = $o_slice_adaptor->fetch_by_region( 'chromosome',
		                                        $chrom, $win_start, $win_end );

		# ComputeDataOnWindow return a hash ref to allow multi values on the same slice.
		my $rh_values = $self->_ComputeDataOnWindow( $o_slice );
		$h_data{ $i } = { id => $i, chrom => $chr_name,
#		                  start => $win_start, end => $win_end,
						  start => $win_start, end => $win_start + $step - 1, # JE NE SUIS PAS SUR - MANU
		                  opt_circos => '=',
		                  %$rh_values };

		# MaJ des variables
		$win_start += $step;
		$win_end    = $win_start + $win - 1;
		++$i;
	}

	return \%h_data;
}


=head2 function __SynchronizeBounds

	Title        : __SynchronizeBounds
	Usage        : $rh_bounds = $o_xx->__SynchronizeBounds( $o_slice_adaptor, $chrom );
	Prerequisite : none
	Function     : Sets start and end position for the current chromosome.
	Returns      : $rh_bounds, hash ref keys are :
	              		start => integer where reading start
	              		end   => integer where reading stop
	Args         : $o_slice_adaptor, a SliceAdaptor object (API)
	               $chrom, string, the chromosome label.

=cut

sub __SynchronizeBounds
{
	my $self = shift;
	my ( $o_slice_adaptor, $chrom_len ) = @_;
	
	my @a_attr = qw( _start _end );
	my( $start, $end ) = $self->_Get( @a_attr );

	# if _start is negative,
	# 	the window will start that far from the end of the chromosome
	my $start_value = ( $start < 0 ) ? $chrom_len + $start : $start;

	# if _end is on the chromosome, use it else use chromosome length
	my $end_value;
	if( ( 1 < $end ) and ( $end < $chrom_len ) )
	{	$end_value = $end }
	else
	{	$end_value = $chrom_len;	}

	return { start => $start_value, end => $end_value };
}


=head2 function _ComputeDataOnWindow

 Usage        : $rh_values = $o_xx->_ComputeDataOnWindow( $o_slice );
 Prerequisite : none
 Function     : this function computes whatever onto the slice.
 Returns      : $rh_values, hash ref keys are value (colum's) name
                'gc' => $gc, 'value' => $val, ....
 Args         : $o_slice, Slice object (API), it is the reading window,
                where data are compute.
 Note         : It must be redefine in daughter class.
 Error        : LipmError::ObjectException

=cut

sub _ComputeDataOnWindow
{
	my $self = shift;
	my ( $package, $filename, $line) = caller;
	my $dbg = "_ComputeDataOnWindow method must be redefine in $package class.";

	throw LipmError::ObjectException( $self, 'abstract method',
	                                  -text => $dbg );

	return undef;
}


=head2 function _ComputeFilepath

	Usage        : $filepath = $o_xx->_ComputeFilepath( $species, $chrom );
	Prerequisite : none
	Function     : Compute the filepath and make the path if necessary.
	Returns      : $filepath, the formatted absolute filepath
	Args         : $species, string species short name
	               $chrom, string chromosome label

=cut

sub _ComputeFilepath
{
	my $self = shift;

	my @a_attr = qw( _rh_current _graph _feature _extension _file_location );
	my( $rh_current, $graph, $feat, $ext, $where ) = $self->_Get( @a_attr );

	my $o_sla   = $$rh_current{ o_slice_adaptor };
	my $species = $o_sla->db()->species();

	my $o_species = New SpeciesObserver( species => $species );
	my $suppdata_path = $o_species->GetSuppDataPath( $where, &FALSE );

	my $path = File::Spec->catdir( $suppdata_path, $graph, $feat );
	File::Path::mkpath( $path, { verbose => &DEBUG } );

	my @a_keys = qw( MASK_FILE_CHROM );
	my ( $mask ) = $o_species->GetInstanceObserver()->GetCfg( @a_keys );

	my $chrom   = $$rh_current{ chrom };
	my $filename = $mask . $chrom . '.' . $ext;
	my $filepath = File::Spec->catfile( $path, $filename );

	return $filepath;
}


=head2 function _GetHeaderFile

	Usage        : $ra_headers = $o_xx->_GetHeaderFile();
	Prerequisite : none
	Function     : this function returns the headers of the datafile.
	             	by default it is
	             	[ 'chrom', 'start', 'end', 'value', 'opt_circos', 'id' ]
	Returns      : $ra_headers, list ref of headers,
	              	their order is their order in the datafile
	Args         : none.
	Note         : It should be redefine in daughter class.

=cut

sub _GetHeaderFile
{
	my @a_headers = qw( chrom start end value opt_circos id );

	return \@a_headers;
}


=head2 _CleanData

 Title        : _CleanData
 Usage        : $o_xx->_CleanData( $rh_data, $col, $r_value );
 Prerequisite : $rh_data MUST have a column named 'value'
 Procedure    : This method empty $rh_data if data should not be printed.
 Args         : $rh_data, a ref hash of hash ref,
                like the one used by NARCISSE_utils::PrintNarcisseData.

=cut

sub _CleanData
{
	my $self = shift;
	my ( $rh_data ) = @_;

	my $value;
	my $flat_curve = $self->__CheckDataForFlatness($rh_data, 'value', \$value);

	if ( &TRUE == $flat_curve )
	{
		my $msg = "Auto-clean : ";
		if( defined $value )
		{
			$msg .= "only one value found ($value)";
			$msg .= "the data will not be saved.\n";
		}
		else
		{	$msg .= "no data found.\n";	}
		$self->_Trace( $msg, '.', 'IC' );

		%$rh_data = ();
	}

	return;
}


=head2 __CheckDataForFlatness

 Title        : __CheckDataForFlatness
 Usage        : $o_xx->__CheckDataForFlatness( $rh_data, $col, $r_value );
 Prerequisite : none
 Function     : This method return true if the data in the column
                    $col have same values.
 Args         : $rh_data, a ref hash of hash ref,
                like the one used by NARCISSE_utils::PrintNarcisseData.
                $col, the name of the tested column
                $r_value Will receive the one value found if we return TRUE
 Return       : &TRUE is the data in the column $col have only one value
                (ie 0 for all data, for example
                     - In this case, $$r_value will be filled with 0 )
                &FALSE is they have different values

=cut

sub __CheckDataForFlatness
{
	my $self = shift;
	my ( $rh_data, $col, $r_val ) = @_;

	# in a hurry : no time to do it really nice
	$col = 'value'	unless (defined $col );

	my ($val_ref, $val ) ;
	foreach my $rh_row ( values %$rh_data )
	{
		$val_ref = $$rh_row{ $col }	unless( defined $val_ref );
		$val     = $$rh_row{ $col };

		return &FALSE if( $val ne $val_ref );
	}

	# all data have the same value
    $$r_val = $val_ref;
	return &TRUE;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
