#
# 	sebastien.letort@toulouse.inra.fr
#	Created: August 06, 2008
#	$Id: GeneSurfaceCalculator.pm 3142 2009-12-22 12:38:52Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package GeneSurfaceCalculator;

=pod

=head1 NAME

	GeneSurfaceCalculator - To compute gene surface into a sliding window.

=head1  SYNOPSIS

	# to compute for all species, all chrom, whole length.
	my $o_gdc = New GeneSurfaceCalculator(
	            		-window => $window,
	            		-step => $step
	            	);
	$o_gdc->RollOverSpecies();

	# to compute for Medicago truncatula and Vitis vinifera, chrom 1 and 3
	# between position 156_000 and 2_350_000
	# in a 10_000 bases window, 2_000 bases step
	# with debug information
	my $o_gdc = New GeneSurfaceCalculator(
	            		-species => [ 'mtr', 'vvin' ],
	            		-chrom => [ '1', '3' ]
	            		-window => $window,
	            		-step => $step
	            		-start => 156_000,
	            		-end => 2_350_000,
	            		-verbose => 1
	            	);
	$o_gdc->RollOverSpecies();


=head1 DESCRIPTION

	This class can compute gene surface.
	It means that every step position, it counts how many bases are within a gene
	into a 'window' length sequence.
	For each value a line is written in a file ( NarcisseData formatted ) like
	#id	start	end	value

=cut

use strict;
use warnings;

use Data::Dumper;

use base( 'SlidingWindowCalculator' );

use General;


=head2 function _ComputeDataOnWindow

	Title        : _ComputeDataOnWindow
	Usage        : $rh_values = $o_xx->_ComputeDataOnWindow( $o_slice );
	Prerequisite : none
	Function     : this function compute the percentage of the slice
	               	that is part of genes.
	Returns      : $rh_values, hash ref { value => $value }
	Args         : $o_slice, Slice object (API), it is the reading window,
	               	where data are compute.

=cut

sub _ComputeDataOnWindow
{
	my $self = shift;
	my ( $o_slice ) = @_;

	my $length = $o_slice->length();
	my $gene_surface = 0;

	my $ra_genes = $o_slice->get_all_Genes();
	foreach my $o_gene ( @$ra_genes )
	{
		$gene_surface += $o_gene->length();
	}

	# to have a percentage.
	# value is the percentage of the slice surface that compose genes.
	my $value = $gene_surface / $length;

	# return the data
	return { value => $value };
}


1;
