#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 30-04, 2009
#	$Id: SpeciesObserver.pm 3782 2013-01-09 10:34:31Z  $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package SpeciesObserver;

=pod

=head1 NAME

 SpeciesObserver - Access to configuration information about a species

=head1  SYNOPSIS

 my $o_so = NewSpeciesObserver( -o_inst_obs => $o_io, species => 'species_name' )

 Create a new species observer, related to an instance observer - If you don't specify any instance observer,
 an instance observer is created using the InstanceObserver Singleton class method

 my $o_so = $o_io->GetSpeciesObserver( 'some_scpeis' );

 where $o_io is an InstanceObserver

=head1 DESCRIPTION

 Be carefull : SpeciesObserver does not inherite from Configuration.
 To be usefull, this class needs that the InstanceObserver object linked to it
 includes the 'species'.

=cut

use strict;
use warnings;

use Data::Dumper;
use File::Spec;
use File::Path;
use Carp;
use Error qw( :try );

use base qw( LipmObject );

use LipmError::ParameterException;
use LipmError::Function;
use Configuration::InstanceObserver;
use General;
use Bio::Narcisse::DBAdaptor;
use FastaReader;

use constant ERR_SPECIES_NOT_IN_INSTANCE => 0003;

=head2 method _Init

 Usage        : my $o_x = New SpeciesObserver( -o_inst_obs => $o_inst_obs,
                                               species => $species );
 Fonction     : initialise an instance of SpeciesObserver
 Returns      : a valid SpeciesObserver object
 Args         : $o_inst_obs, InstanceObserver object.
                $species, string, species name
 Attributes   :
    _cfg_path, inherited, string, path to cfg file.
    __short_name,
    __long_name,
    __o_inst_obs,
    __rh_chroms_info,
=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger => undef,
	               -o_inst_obs => undef,
	               species   => undef,
	               @_ );

	# for mandatories
	LipmError::Function::Assert( %h_args );

	my( $species, $o_inst_obs ) = @h_args{ qw( species -o_inst_obs ) };


	# obtenir un objet InstanceObserver - si MULTI defini => NO
# 	my $o_nar_obs = New NarcisseObserver()
	$o_inst_obs = InstanceObserver::Singleton() unless defined( $o_inst_obs );

	my $short   = $o_inst_obs->GetSpeciesShortName( $species );
	if( $short eq &InstanceObserver::NONE )
	{
		my $name = $o_inst_obs->GetName();
		my $msg  = "Your species ($species) does not belong to the instance $name\n";
		throw LipmError::ObjectException( __PACKAGE__, '!build', -text => $msg,
										 -value => &ERR_SPECIES_NOT_IN_INSTANCE );
	}

	my $rh_infos = $o_inst_obs->GetSpeciesInfo( $short );

	$self->_Set( __rh_infos => $rh_infos,
				 __short    => $short, # raccourci existe deja dans rh_infos
				 __o_inst_obs => $o_inst_obs,
				 __rh_chroms_info => undef,
				 __nb_proteins    => undef,
				 __genome_size    => 0 );

	return;
} # _Init

=head1 Some Accessors

=head2 Functions GetShortName, GetLongName, GetAssembly, GetTaxId, GetWhere

 Usage        : my $short = $o_x->GetShortName();
                my $long  = $o_x->GetLongName();
                my $ass   = $o_x->GetAssembly();
                my $taxid = $o_x->GetTaxId();
                my $where = $o_x->GetWhere();
 Function     : accessors.
 Returns      : string, whatever is asked.
 Args         :
     $species, string, short or long species name.
 Error        : none

=cut

sub GetShortName
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return $$rh_{ short }
}

sub GetLongName
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return $$rh_{ long }
}

sub GetLongNameWithoutSpace
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	my $long = $$rh_{ long };
	$long    =~ s/ /_/g;

	return $long;
}

sub GetAssembly
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return $$rh_{ assembly }
}

sub GetAssemblyPhase
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return (defined $$rh_{ assembly_phase }) ? $$rh_{ assembly_phase } : 'PSEUDOMOLECULES';
}

sub GetTaxId
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return $$rh_{ taxid }
}

sub GetWhere
{
	my $self = shift;
	my $rh_  = $self->_Get( '__rh_infos' );
	return $$rh_{ where }
}

sub GetTimestamp
{
    my $self = shift;
    my $file_time_stamp = $self->GetTimestampFile();
    my $fh_time_stamp = new IO::File($file_time_stamp) or throw LipmError::IOException($file_time_stamp,'f');
    my $time_stamp = <$fh_time_stamp>;
    chomp $time_stamp;
    return $time_stamp;
}

=head2 Function IsPseudo

 Usage        : if ($o_x->IsPseudo())...
 Function     : Accessor: returns true if assembly phase is pseudomolecules
 Returns      : a boolean
 Args         : none

=cut
sub IsPseudo
{
	my $self = shift;
	return ($self -> GetAssemblyPhase() eq 'PSEUDOMOLECULES');
}

=head2 Function GetInstanceObserver

 Usage        : my $o_inst_obs = $o_x->GetInstanceObserver()
 Function     : Accessor to InstanceObserver attribute.
 Returns      : $o_inst_obs
 Args         : none
 Error        : none

=cut

sub GetInstanceObserver
{
	my $self = shift;

	return $self->_Get( '__o_inst_obs' );
} #GetInstanceObserver


=head1 Some Public methods to access files

=head2 Method __GetXPath

 Usage    : my $path = $o_x->__GetXPath( %h_args )
 Function : generic method to acces path through InstanceObserver
            regarding the 'where' keywords.
 Return   : $path, string, the path where to read data.
                It can be into instance data dir or into session data dir.
                Return the first that exists.
 Args     : %h_args
     io_method => an InstanceObserver method name to get paths
     where => string, cf GetWhere, it is passed to InstanceObserver::method,
     -errors => boolean [&TRUE], if true, throw exception if nothing found.
     -check  => boolean [&TRUE] if true, check if the file exists.
                Warning : if false, only the first 'where' keyword is consider.
     -ra_dirs => list ref of any directory cfg keyword to add to each path.
 Error    :

=cut

sub __GetXPath
{
	my $self = shift;
	my %h_args = (	io_method => undef,
					where     => undef,
# 					-error    => &TRUE,
# 					-check    => &TRUE,
					-ra_dirs  => [],
					@_ );

	my @a_attr = qw( __o_inst_obs __short );
	my( $o_inst_obs, $species ) = $self->_Get( @a_attr );

	my( $where, $ra_dirs ) = @h_args{ qw( where -ra_dirs ) };

	my @a_dirs  = $o_inst_obs->GetCfg( @$ra_dirs );

	my $method  = $h_args{ io_method };
	my @a_paths = $o_inst_obs->$method( $where );
	foreach my $path ( @a_paths )
	{
		my $data_path = File::Spec->catdir( $path, $species, @a_dirs );
        return $data_path;
	}

	return;
}

=head2 Function GetDataPathes

 Usage    : my @a_pathes = $o_so->GetDataPathes( $where )
 Function : uses its internal InstanceObserver to access pathes
            corresponding the 'where' keywords.
 Return   : @a_pathes, an array of pathes
 Args     : $where, string, optional 'INSTANCE', 'SESSION', or 'INSTANCE,SESSION'
 Error    : ParameterException

=cut

sub GetDataPathes
{
	my $self = shift;
	my ( $where ) = @_;

	$where = $self->GetWhere() if( !defined $where );
	my $species = $self -> GetShortName();

	my @a_pathes = $self -> GetInstanceObserver() -> GetDataPaths( $where );
	my @a_data_pathes = map ( File::Spec->catdir( $_, $species), @a_pathes );
	return @a_data_pathes;
}

=head2 Function GetDataPath

 Usage    : my $path = $o_so->GetDataPath( $where, $check )
 Function : uses its internal InstanceObserver to access path
            corresponding the 'where' keywords.
 Return   : $path, string, the path where to read data.
                It can be into instance data dir or into session data dir.
                Return the first existing (if mkdir==&FALSE), or the first that could be created (mkdir=&TRUE)
 Args     : $where, string, optional [INSTANCE] or SESSION.
                    if specified, passed to InstanceObserver::GetDataPaths.
            $mkdir, boolean [false], if true, create the directory
 Error    : ParameterException

 Note pas sur que ca fonctionne avec les donnees perso

=cut

sub GetDataPath
{
	my $self = shift;
	my( $where, $mkdir  ) = @_;
	$mkdir = &FALSE unless defined $mkdir;

	$where = $self->GetWhere() if( !defined $where );
	my @a_paths = $self -> GetInstanceObserver() -> GetDataPaths( $where );
	my $species = $self -> GetShortName();
	foreach my $path (@a_paths)
	{
		my $datapath = File::Spec->catdir( $path, $species );
		mkpath $datapath if ( $mkdir == &TRUE && !-d $datapath );
		return $datapath if ( -d $datapath );
	}
	
	# TODO - si on arrive la c'est que ca merde il faudrait lancer une exception ?
	return '';
}

sub GetDataPath_OLD
{
	my $self = shift;
	my( $where, $check  ) = @_;

	$where = $self->GetWhere() if( !defined $where );
	

# 	$check = &TRUE unless defined $check;

# 	my $path = $self->__GetXPath( io_method => 'GetDataPaths',
# 								  where     => $where );
# 								  , -check    => $check );

	# ca devrait etre dans InstanceObserver je pense
	my $key = ( $where eq 'SESSION' ) ? 'ROOT_LOCAL_DATA' : 'ROOT_DATA';

	my @a_attr = qw( __o_inst_obs __short );
	my( $o_inst_obs, $species ) = $self->_Get( @a_attr );

	my $path = File::Spec->catdir( $o_inst_obs->GetCfg( $key ), $species );

	return $path;
}


=head2 Function GetTmpPath, GetDownloadPath
 Usage       : my $path = $o_so->GetDataPath( $where, $check )
 Function    : Call GetDataPath, create and return the tmp / tmp_data subdirectories
 Args, Error : see GetDataPath

=cut
sub __GetSubdirPath
{
	my $self   = shift;
	my $subdir = shift;
	my $o_io = $self->GetInstanceObserver();
	my $path = File::Spec->catdir($self->GetDataPath(@_), $subdir);
	mkpath $path;
	return $path;
}
sub GetTmpPath
{
	my $self = shift;
	my $o_io = $self->GetInstanceObserver();

	# Pass tmp as the FIRST parameter
	unshift (@_,$o_io->GetCfg( 'TMP'));
	return $self->__GetSubdirPath(@_);
}
sub GetTmpDataPath
{
	my $self = shift;
	my $o_io = $self->GetInstanceObserver();

	# Pass TmpData as the FIRST parameter
	unshift (@_,$o_io->GetCfg( 'TMP_DATA'));
	return $self->__GetSubdirPath(@_);
}
sub GetDownloadPath
{
	my $self = shift;
	my $o_io = $self->GetInstanceObserver();

	# Pass tmp as the FIRST parameter
	unshift (@_,$o_io->GetCfg( 'TMP_DOWNLOAD'));
	return $self->__GetSubdirPath(@_);
}

=head2 Function GetSuppDataPath

 Usage    : my $path = $o_so->GetSuppDataPath( $where, $check )
 Function : uses its internal InstanceObserver to access path
                corresponding the 'where' keywords.
 Return   : $path, string, the path where to read data.
                It can be into instance data dir or into session data dir.
                Return the first that exists.
 Args     : $where, string, optional
                   if specified, it is passed to InstanceObserver::GetDataPaths.
            $check, boolean [true], if true, assert dir exists.
 Error    : ParameterException, LipmError (erreur grave)

 TODO : utiliser un hash pour les param comme pour les autres methodes.

=cut

sub GetSuppDataPath
{
	my $self = shift;
	my( $where, $check ) = @_;

	if( !defined $where )
	{
		$where = 'INSTANCE';
        #&Carp::carp("the where parameter should be defined ");
	}

# 	$check = &TRUE unless defined $check;

# 	my $path = $self->__GetXPath( io_method => 'GetDataPaths',
# 								  where     => $where,
# # 								  -check    => $check,
# 								  -ra_dirs  => [ 'SUPP_DATA' ] );
	my $key = ( $where eq 'SESSION' ) ? 'ROOT_LOCAL_DATA' : 'ROOT_DATA';

	my @a_attr = qw( __o_inst_obs __short );
	my( $o_inst_obs, $species ) = $self->_Get( @a_attr );

	my( $datapath, $supp_dir ) = $o_inst_obs->GetCfg( $key, 'SUPP_DATA' );

	if( $datapath =~ /D_SESSDIR/ )
	{
		my $msg = "You try to access session directory without specifying the session directory";
		throw LipmError( -text => $msg, -usr_msg => "BUG ! call the dev A-team" );
	}

	my $path = File::Spec->catdir( $datapath, $species, $supp_dir );

	return $path;
}


=head2 Function __GetComparativeDirpath

 Usage    : my $dirpath = $o_so->__GetComparativeDirpath( %h_args )
 Function : uses its internal InstanceObserver to access path.
            If -where is specified, return the corresponding path,
            else use GetWhere and return the first existing path.
 Return   : $dirpath, string, the path where to read comparative data (ex. data/ref/
            It can be into instance or session dir.
 Args     : hash with following keys
            target, string, short species name to use as target
            key,    string, 'ORTHOLOGUE','SYNT' or ''
            -subdir, string, any subdir to add at the end of the path (ad hoc for synt)
                             ONLY IF key is NOT ''
            -where, string [INSTANCE] or SESSION
           -check, boolean [&TRUE], if &TRUE, return only existing dirpath.
 Error    : ParameterException, IOException

=cut

sub __GetComparativeDirpath
{
	my $self = shift;
	my %h_args = (	target => undef,
					key    => undef,
					-where => undef,
					-check => &TRUE,
					-subdir => undef, # bidouille pour synt - pas sur que ca ait sa place ici.
					@_ );
	LipmError::Function::Assert( %h_args );

	my @a_keys = qw( target -subdir key -where -check );
	my( $target, $subdir, $key, $where, $check ) = @h_args{ @a_keys };

	my @a_attr = qw( __o_inst_obs __short );
	my( $o_inst_obs, $species ) = $self->_Get( @a_attr );

	$where = $self->GetWhere() unless defined $where;

	my @a_dirpaths = $o_inst_obs->GetComparativeDirpaths( $where );
    my $dir = '';
    if ($key eq '')
    {
        if ( defined $subdir )
        {
            throw LipmError::ParameterException( '-subdir', 'value',-text => "should NOT be specified because key=''\n");
        }
    }
    else
    {
        $dir       = $o_inst_obs->GetCfg( $key );
    }
	my @a_subdirs  = ( $species, $target, $dir, $subdir );
	pop(@a_subdirs) unless defined $subdir; # avoiding a warning from catdir

	foreach my $dirpath ( @a_dirpaths )
	{
		my $data_path = File::Spec->catdir( $dirpath, @a_subdirs );
# a voir si on peut factoriser avec __GetXPath -ra_dir pose pb.
# 	my $path = $self->__GetXPath( io_method => '__GetComparativeDirpaths',
# 								  where    => $where );
# 								  , -error => &FALSE );


		if( ( &FALSE == $check ) or ( -e $data_path ) )
		{
			return $data_path;
		}
	}

	my $subdirs = File::Spec->catdir( @a_subdirs );
	my $msg = "datapath ( $subdirs ) found into neither of this paths @a_dirpaths\n";
	throw LipmError::IOException( '','',-text => $msg );

	return;
}


=head2 Function GetComparisonPath

 Usage    : my $path = $o_so->GetComparisonPath( %h_args )
 Function : uses its internal InstanceObserver to access comparison path.
            Then build the complete path to the final directory.
            Directories can be built.
 Return   : $path, string, the path to the comparison directory.
                It can be into instance data dir or into session data dir,
                depending on the __where attribute.
 Args     : hash with following keys :
     method, string, the computationnal method
     target, string, short name of the target species
     -mkdir, boolean [&FALSE], if &TRUE, the missing directories will be created
 Error    : none

=cut

sub GetComparisonPath
{
	my $self = shift;
	my %h_args = (	method => undef,
					target => undef,
					-mkdir => &FALSE,
                    -where => undef,
					@_ );

	# check mandatories
	&LipmError::Function::Assert( %h_args );

	my @a_attr = qw( __o_inst_obs __short );
	my( $o_inst_obs, $species ) = $self->_Get( @a_attr );

	# only take the first one
    my $where = defined $h_args{-where} ? $h_args{-where} : $self->GetWhere();
	my ( $path ) = $o_inst_obs->GetComparisonPaths( $where );
	my ( $method, $target ) = @h_args{ qw( method target ) };

	$path = File::Spec->catfile( $path, $method, $species, $target );

	if( &TRUE == $h_args{ -mkdir } )
	{
		my @a_created = mkpath( $path );
		$self->_Trace( "Created directories @a_created\n", '.', 'IO' );
	}

	return $path;
}


=head2 Function GetKaryotypeFile,GetTimestampFile,GetSummaryFile,__GetKeyFile

 Usage        : my $karyo_path = $o_x->GetKaryotypeFile( ...)
 Function     : Retrieve the karyotype/timestamp file, __GetKeyFile is called by the other two
 Returns      : $karyo_path or timestamp path, string
 Args         : -where, -check (optional)
 Error        : none

=cut

sub GetKaryotypeFile
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'MASK_FILE_KARYOTYPE' );
}
sub GetTimestampFile
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'TIMESTAMP' );
}
sub GetSummaryFile
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'SUMMARY' );
}
sub GetTimestampMirror
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'TIMESTAMP_MIRROR' ) ;
}
sub GetReleaseFile
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'MASK_FILE_RELEASE' );
}
sub GetReadmeFile
{
	my $self = shift;
	my %h_args = ( @_ );
    return $self->__GetKeyFile( %h_args,key => 'README' );
}
sub __GetKeyFile
{
	my $self = shift;
	my %h_args = ( -where => undef,
                   -check => &TRUE,
                   -tmp   => &FALSE,
                   -mkdir => &FALSE,
                   key => undef,
                   @_ );

	my $o_inst_obs = $self->GetInstanceObserver();
    my $key        = $h_args{ key };
	my $filename   = $o_inst_obs->GetCfg( $key );
	my ( $where )  = @h_args{ qw( -where ) };
	$where = $self->GetWhere() unless( defined( $where ) );

	my @a_data_pathes = $self->GetDataPathes( $where );
	my $tmp_data= ($h_args{'-tmp'} == &TRUE) ? $o_inst_obs -> GetCfg( 'TMP_DATA' ) : "";
	foreach my $datapath (@a_data_pathes)
	{
		#my $datapath = $self->GetDataPath( $keyword );
		my $dirpath;

		# If tmp, path is data/human/tmp_data/gene
		if ( $h_args{ -tmp } == &TRUE )
		{
			$dirpath  = File::Spec->catdir( $datapath, $tmp_data );
		}
		else
		{
			$dirpath = $datapath;
		}
		
		# If mkdir is true, create the directories if possible
		if( &TRUE ==  $h_args{ -mkdir })
		{
			my @a_created = mkpath( $dirpath );
			$self->_Trace( "Tried to create directories @a_created\n", '.', 'IO' );
		}
		my $filepath = File::Spec->catdir( $dirpath, $filename );
		return $filepath if( ( &FALSE == $h_args{ -check } ) || ( -e $filepath ) );
	}


	#my $usr = "Your $feature data files look to be missing\n";
	my $msg = "Species: " . $self->GetShortName() . " file: $filename";
    throw LipmError::IOException( '','',-text => $msg );
} #__GetKeyFile

=head2 GetCfgFilepath

 Usage        : my $filepath = $o_x->GetCfgFilepath( $where )
 Function     : return the filepath to cfg/species/x.cfg.
 Returns      : $filepath, string, a filepath.
 Args         : $where, string, optionnal. 'SESSION' or 'INSTANCE'
                If undef, GetWhere will be used.
 Error        : none

 Note    : l'argument $where se justifie pour le moment *uniquement* dans un but d'ecriture.
           --> obtenir le chemin d'acces au fichier.

=cut

sub GetCfgFilepath
{
	my $self = shift;
	my( $where ) = @_;

	$where = $self->GetWhere()  unless defined( $where );

	my $o_io    = $self->GetInstanceObserver();
	my $dirpath = $o_io->GetSpeciesCfgDirpath( $where );

	my $file    = $self->GetShortName() . '.cfg';

	my $filepath= File::Spec->catfile( $dirpath, $file );

	return $filepath;
} #GetCfgFilepath

=head1 Methods to access feature files

=head2 Function GetFeatureFiles

 Usage        : my $file_path = $o_x->GetFeatureFiles( %h_params )
                my @a_filepaths = $o_x->GetFeatureFiles( %h_params )
 Function     : Given the chrom and the feature, retrieve the filepath
 Returns      : $file_path, string, absolute path to feature file.
                @a_file_paths, list of string if no chrom were specified.
 Args         :
   -chrom  => $chrom, string, chromosome label.
   key     => $feature, string, key of feature dir and extension in conf file.
              (GENE SYNT RNA ...)
   -where  => string, ['INSTANCE'] coma separated list of keywords ex. 'INSTANCE,SESSION'
   -tmp    => boolean, [&FALSE] if true, we return the tmp_data file
                                (ex: Site/data/human/tmp_data/gene)
   -check  => boolean, [&TRUE], if true, will check that the file exists.
   -mkdir  => boolean, [&FALSE], if true create the directory
 Error        : ParameterException, LipmError

=cut

sub GetFeatureFiles
{
	my $self   = shift;
	my %h_args = (	-chrom => undef,
					key    => undef,
					-check => &TRUE,
					-existing_files => &FALSE,
					-where => undef,
					-tmp   => &FALSE,
					-mkdir => &FALSE,
					@_ );

	my @a_keys = qw( -chrom -check -existing_files key -tmp -mkdir );
	my( $chrom, $check, $only_if_exists, $key, $tmp, $mkdir ) = @h_args{ @a_keys };

	# If chrom not specified, we return an array of files
	# However, GetFeatureFiles may always be called in void context (may be useful for catching the exceptions)
	if( defined(wantarray) && wantarray==0 && !defined $chrom )
	{
		my $msg = "In scalar context, you have to specify a chrom value";
		throw LipmError::ParameterException( '-chrom', 'missing',
											-text => $msg );
	}

	my $o_io = $self->GetInstanceObserver();
	@a_keys = ( qw( MASK_FILE_CHROM ), $key );
	my( $mask, $feature ) = $o_io->GetCfg( @a_keys );

	my ( $where ) = @h_args{ qw( -where ) };
	$where        = $self->GetWhere() unless( defined( $where ) );
	my $tmp_data  = ($tmp == &TRUE) ? $o_io->GetCfg( 'TMP_DATA' ) : "";

	my @a_data_pathes = $self -> GetDataPathes( $where );
	foreach my $datapath (@a_data_pathes)
	{
		#my $datapath = $self->GetDataPath( $keyword );
		my $dirpath;

		# If tmp, path is data/human/tmp_data/gene
		if ( $tmp == &TRUE )
		{
			$dirpath  = File::Spec->catdir( $datapath, $tmp_data, $feature );

			# If mkdir is true, create the directories if possible
			if( &TRUE == $mkdir )
			{
				my @a_created = mkpath( $dirpath );
				$self->_Trace( "Created directories @a_created\n", '.', 'IO' );
			}
		}
		else
		{
			$dirpath  = File::Spec->catdir( $datapath, $feature );
		}

# 		next  if( ( &TRUE == $check ) && ( !-d $dirpath ) );

		# No chrom specified, we call GetFeatureFiles for each chromosome
		if( !defined( $chrom ) )
		{
			if( &TRUE == $only_if_exists )
			{
				if( -d $dirpath )
				{
						my @a_files = glob( "$dirpath/*.$feature" );
						return @a_files;
				}
				next;
			}
			else
			{
				my @a_files;
				my $rh_chroms = $self->GetChromsInfo();
				foreach my $chr ( keys %$rh_chroms )
				{
					#my %h_args_1 = ( -chrom => $chr, %h_args );
					my %h_args_1 = %h_args;
					$h_args_1 { -chrom } = $chr;

					push @a_files, $self->GetFeatureFiles ( %h_args_1 );
				}
				return @a_files;
			}
		}

		# on a un chrom, on veut le bon fichier
		my $filename = $mask . $chrom . '.' . $feature;
		my $filepath = File::Spec->catfile( $dirpath, $filename );

		return $filepath if( ( &FALSE == $check ) || ( -e $filepath ) )
	}

	return () if( $only_if_exists );

	#my $usr = "Your $feature data files look to be missing\n";
	my $msg = "Species: " . $self->GetShortName() . " feature: $feature\n\t**no file found.**\n";
    #$msg    .= "Check that your $feature data are generated.\n";
    # throw LipmError::IOException( '','',-text => $msg );
    throw LipmError( -text => $msg, -depth => 1 );

	return;
} #GetFeatureFiles


=head2 GetGeneFiles, GetExonFiles, GetRepeatFiles, GetMarkerFiles, GetTranscriptFiles
       GetCytoFiles, GetPersoFiles, GetSnpFiles, GetOceanFiles

 Usage        : my $filepath = $o_x->GetXyzFiles( %h_params );
                my @a_filepaths  = $o_x->GetXyzFiles( %h_params );
 Function     : call GetFeatureFiles with correct key.
 Returns      : $filepath, string, the path to the wanted file.
                @a_filepaths, list of string, if no chrom were specified.
 Args         :
   -chrom   => string, chrom label
   -check  => boolean, [&FALSE] set to True if you want to test that the file exists.
 Error        : none.

=cut

sub GetGeneFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'GENE' );
} #GetGeneFiles

sub GetExonFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'EXON' );
} #GetExonFiles

sub GetCytoFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'CYTO' );
} #GetCytoFiles

sub GetRepeatFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'REPEAT' );
} #GetRepeatFiles

sub GetRnaFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'RNA' );
}

sub GetProtFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'PROT' );
}

sub GetMarkerFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'MARKER' );
}

sub GetTranscriptFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'TRANSCRIPT' );
} #GetTranscriptFiles

sub GetPersoFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, -where => 'SESSION', key => 'PERSONAL' );
} #GetPersoFiles

sub GetSnpFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'SNP' );
} #GetPersoFiles


sub GetOceanFiles
{
	my $self   = shift;
	return $self->GetFeatureFiles( @_, key => 'OCEAN' );
} #GetPersoFiles


=head1 Methods to access comparative files

=head2 Function GetSyntFiles, GetOrthologyFiles

 Usage    : $filepath = $o_x->GetSyntFiles( %h_params );
                @a_filepaths = $o_x->GetSyntFiles( %h_params );
 Function : given a target, a prog, a level, a chromosome
                return the synteny filepath,
                if no chrom is given, return a list of filepaths.
 Returns  : $filepath, a string,
            @a_filepaths, a list of string
 Args     : prog   => $prog,
            level  => $level,
            target => $tgt, string, target short name
            -chrom => $chrom, string ref chrom's label
            -check => boolean, [&TRUE]
            -existing_files => boolean, [&FALSE], if &TRUE, -check is ignored
            -where => string
 Error    : ParameterException, IOException

 Note     : -existing_files is a "bequille", if one refactors Narcisse, consider this.
  we need 3 behaviours :
  	- get all existing files ( for Dotplot for exemple )
  	- get all filenames
  	- get all filenames and throw an exception if one is missing.

 	One possible way to do better is to do a GetFile method, returning or throwing error.

=cut

sub GetComparativeFiles
{
	my $self   = shift;
	my %h_args = (	-chrom => undef,
		key    => undef,
		target => undef,
		# 					-subdir => undef, # bidouille pour synt
		-check => &TRUE,
		-existing_files => &FALSE,
		-where => undef,
		@_ );

	my @a_keys = qw( -chrom -check -existing_files key );
	my( $chrom, $check, $only_if_exists, $key ) = @h_args{ @a_keys };

	# If chrom not specified, we return an array of files
	if( 0 == wantarray() && !defined( $chrom ) )
	{
		my $msg = "In scalar context, you have to specify a chrom value.";
		throw LipmError::ParameterException( '-chrom', 'missing',
            -text => $msg, -depth => 1.5 );
	}

	my $dirpath = $self->__GetComparativeDirpath( %h_args );

	my $o_inst_obs = $self->GetInstanceObserver();

	@a_keys = ( qw( MASK_FILE_CHROM ), $key );
	my( $mask, $feature ) = $o_inst_obs->GetCfg( @a_keys );

	# pas besoin pour le moment <-- where est utilise par __GetComparativeDirpath.
	# 	my ( $where ) = @h_args{ qw( -where ) };
	# 	$where = $self->GetWhere() unless( defined( $where ) );
	#
	# 	foreach my $keyword ( split( /,/, $where ) )
	# 	{
	# 		my $datapath = $self->GetComparativePath( $keyword, $h_args{ target } );
	# 		my $datapath = $self->GetComparativePath( $h_args{ target } );
	# 		my $dirpath  = File::Spec->catdir( $datapath, $feature, $dir );
	# 		next  if( !-d $dirpath );

	# No chrom specified, we call GetComparativeFiles for each chromosome
	if( !defined( $chrom ) )
	{
		if( &TRUE == $only_if_exists )
		{
			if( -d $dirpath )
			{
					my @a_files = glob( "$dirpath/*.$feature" );
					return @a_files;
			}
		}
		else
		{
			my @a_files = ();
			my $rh_chroms = $self->GetChromsInfo();
			foreach my $chr ( keys %$rh_chroms )
			{
				my %h_args_1 = ( %h_args, -chrom => $chr );
				try
				{
					push( @a_files, $self->GetComparativeFiles ( %h_args_1 ) );
				}
				catch LipmError with
				{
					my( $err, $r ) = @_;

					$err->throw()  if( &FALSE == $only_if_exists );
				};
			}
			return @a_files;
		}
	}
	else
	{
		my $filename = $mask . $chrom . '.' . $feature;
		my $filepath = File::Spec->catfile( $dirpath, $filename );

		return $filepath if( ( &FALSE == $check ) || ( -e $filepath ) );

		#my $usr = "Your $feature data files seem to be missing\n";

		my $ref = $self->GetShortName();
		my $tgt = $h_args{ target };
		#my $msg = "Check that your $feature data are generated (ref=$ref, tgt=$tgt).\n";
		# $msg   .= "reference: $ref, target: $tgt, feature: $feature";
		throw LipmError::IOException( $filepath, 'f', -depth => 1 );#,-text => $msg );
	}

	return;
}

sub GetSyntFiles
{
	my $self   = shift;
	my %h_args = (	prog   => undef,
					level  => undef,
					# pour infos
# 					target => undef,
# 					-chrom => undef,
# 					-check => &FALSE,
# 					-where =>
					@_ );

	my @a_keys = qw( prog level );
	my( $prog, $level ) = @h_args{ @a_keys };

	my $dirs = File::Spec->catdir( $prog, $level );

	return $self->GetComparativeFiles( %h_args, -subdir => $dirs, key => 'SYNT' );
} #GetSyntFiles

sub GetOrthologyFiles
{
	my $self   = shift;
	my %h_args = (  #target => undef,
# 					-chrom => undef,
# 					-check => &FALSE,
# 					-where =>
					@_ );

	return $self->GetComparativeFiles( %h_args, key => 'ORTHOLOGUE' );
} #GetOrthologyFiles

sub GetHspFiles
{
	my $self   = shift;
	my %h_args = (  prog   => undef,
					level  => undef,
#                   target => undef,
# 					-chrom  => undef,
# 					-check  => &FALSE,
				# where is always SESSION whatever the args are.
					@_ );

	my @a_keys = qw( prog level );
	my( $prog, $level ) = @h_args{ @a_keys };

	my $dirs = File::Spec->catdir( $prog, $level );

	return $self->GetComparativeFiles( %h_args, -subdir => $dirs, -where => 'SESSION', key => 'HSP' );
} #GetHspFiles


# Return ..../comparative/bana/banb/ortho
sub GetOrthologueDirpath
{
	my $self   = shift;
	my %h_args = (  #target => undef,
# 					-check => &FALSE,
# 					-where =>
					@_ );

	return $self->__GetComparativeDirpath( %h_args, key => 'ORTHOLOGUE' );
} #GetOrthologueDirpath

# Return ..../comparative/bana/banb/synt
sub GetSyntDirpath
{
	my $self   = shift;
	my %h_args = (  #target => undef,
# 					-check => &FALSE,
# 					-where =>
					@_ );

	return $self->__GetComparativeDirpath( %h_args, key => 'SYNT' );
} #GetSyntDirpath

# Return ..../comparative/bana/banb
# NOTE - With target => '', return ..../comparative/bana
sub GetComparativeDirpath
{
	my $self   = shift;
	my %h_args = (  #target => undef,
# 					-check => &FALSE,
# 					-where =>
					@_ );

	return $self->__GetComparativeDirpath( %h_args, key => '' );
} #GetComparativeDirpath

=head2 Function GetMethods

 Usage      : my @a_methods = $o_xx->GetMethods( target => $target );
 Function   : Look for the comparative directory path, synt subdirectory
              When found, return the directory list found here, as there is 1 directory / method
 Return     : @a_methods, the list of found methods
 Args       :
     target, string, required, the target organism
     -where, string, optional, xwhere to look for. If not specified, searchs in the whole directory path
 Error      : ParameterException, LipmError (if data are not found)
 TODO       : regarder partout dans where, ne pas renvoyer le premier repertoire venu

=cut

sub GetMethods
{
    my $self = shift;
    my %h_args = ( target => undef,
                   -where => undef,
                   @_);

	# check mandatories
	&LipmError::Function::Assert( %h_args );

	my @a_attr = qw( __o_inst_obs __short );
	my ( $o_inst_obs, $species ) = $self->_Get( @a_attr );
    my $target = $h_args{target};

	# only take the first one
    my $where = defined $h_args{-where} ? $h_args{-where} : $self->GetWhere();

    my @a_dirpaths = $o_inst_obs->GetComparativeDirpaths( $where );

    # Looking for the first data/comparative/ref/target/synt found in path
    foreach my $d (@a_dirpaths)
    {
        my $prog_path  = File::Spec->catdir( $d,$species,$target,$o_inst_obs->GetCfg('SYNT') );
        # Found, returning the list of files found here (may be none)
        if (-d $prog_path)
        {
            my @a_dirs = glob( "$prog_path/*" );
            my @a_methods;
            foreach my $p ( glob( "$prog_path/*" ) )
            {
                if (-d $p) # Considering only subdirectories
                {
                    my ($vol,$d,$f) = File::Spec->splitpath( $p );
                    push(@a_methods,$f);
                }
            }
            return @a_methods;
        }
    }

    # No directory found in path: we throw an exception
    my $usr = "Your synt data files seem to be missing\n";
    my $ref = $self->GetShortName();
    my $tgt = $h_args{ target };
    my $msg = "Check that your synt data are generated (ref=$species, tgt=$target).\n";
    throw LipmError( -text => $msg, -user_msg => $usr );
}


=head2 Function GetFastaFile

 Usage      : my $filepath = $o_xx->GetFastaFile( feature => $feature,
                                                  chrom   => $chrom,
                                                  -rep     => &FALSE,
                                                  -tmp    => &FALSE,
                                                  -mkdir => &FALSE,
                                                  -check => &TRUE );
 Function   : retrieve the path to the fasta file with the $feature.
 Return     : $filepath, the path to the fasta file.
 Args       : $feature, string, one of qw( PROT CDS CHR GENE TRANSCRIPT )
              $chrom, string, chrom label, used only with 'chr' feature
              rep, boolean, used only with the 'PROT' feature: selects the protein representative of the genes
 Error      : ParameterException, IOException
 DO NOT USE THIS FUNCTION, The GetChromFastaFile etc. are MORE CONVENIENT

=cut

sub GetFastaFile
{
	my $self   = shift;
	my %h_args = ( feature => undef,
	               chrom   => undef,	# pour chr feature
	               -rep    => &FALSE,    # pour prot feature
   	               -tmp   => &FALSE,
	               -mkdir => &FALSE,
	               -check => &TRUE,	               
# 	               keyword => 'INSTANCE', # keyword : a remplacer par 'where'
	               @_ );

	my $feat_dir = $self -> GetFastaDir ( @_ );

	my @a_keys = qw( feature chrom -rep -check );
	my( $feature, $chrom, $rep, $check ) = @h_args{ @a_keys };
	
	my $feat_key = (($rep == &FALSE) ? 'FASTA_' : 'FASTA_REP_' ) . $feature;
	my $feat_ext = $self->GetInstanceObserver()->GetCfg( $feat_key ); 
	my $species  = $self->GetShortName();

	my $filename;
	if( defined $chrom )
	{
		$filename = $species . $chrom . $feat_ext;
	}
	else
	{	
		$filename = $species . '_' . $feat_ext;	
	}

	my $filepath = File::Spec->catfile( $feat_dir, $filename );

	if ($check == &TRUE)
	{
		return $filepath  if( -e $filepath );
		throw LipmError::IOException( $filepath,'f');
	}
	else
	{
		return $filepath;
	}

	return;
}

=head2 Function GetFastaDir

 Usage      : my $dirpath = $o_xx->GetFastaDir( feature => $feature )
 Function   : retrieve the directory to the fasta files with the $feature.
 Return     : $dirpath, the path to the directory.
 Args       :
     $feature, string, one of qw( PROT CDS CHR GENE TRANSCRIPT )
 Error      : ParameterException, IOException

=cut
sub GetFastaDir
{
	my $self = shift;
	my %h_args = ( feature => undef,
   	               -tmp   => &FALSE,
	               -mkdir => &FALSE,
	               -check => &TRUE,
	               -where => undef,
# 	               keyword => 'INSTANCE', # keyword : a remplacer par 'where'
	               @_ );
	# SUPPRIME PAR manu LE 12/1/12               
	#my @a_keys = qw( feature -tmp -mkdir -check keyword );
	#my( $feature, $tmp, $mkdir, $check, $where ) = @h_args{ @a_keys };

	# pb avec cette fonction si pas de species, on ne reste pas sur root_data
	#my $datapath = $self->GetDataPath( $where );
	#my $datapath = $self->GetDataPath();
	
	# REMPLACE PAR CE QUI SUIT
	# si -where non specifie $where vaut undef et on retrouve la meme chose qu'auparavant
	# si -where est specifie il vaut INSTANCE ou SESSION et dans ce cas on le passe a GetDataPath
	my @a_keys = qw( feature -tmp -mkdir -check -where );
	my( $feature, $tmp, $mkdir, $check, $where ) = @h_args{ @a_keys };
	my $datapath = $self->GetDataPath( $where );

	@a_keys = ( 'DB', $feature );
	my( $db, $feat_dir ) = $self->GetInstanceObserver()->GetCfg( @a_keys );

	my $species = $self->GetShortName();
	my $tmp_data= ($tmp == &TRUE) ? $self->GetInstanceObserver()->GetCfg( 'TMP_DATA' ) : "";

	my $dirpath = File::Spec->catfile( $datapath, $tmp_data, $db, $feat_dir );
	
	if ( -d $dirpath )
	{
		return $dirpath;
	}
	else
	{
		# If mkdir is true, create the directories if possible
		if( &TRUE == $mkdir )
		{
			my @a_created = mkpath( $dirpath );
			$self->_Trace( "Created directories @a_created\n", '.', 'IO' );
		}
		if ( &TRUE == $check )
		{
			return $dirpath  if( -d $dirpath );
			throw LipmError::IOException( $dirpath,'d');
		}
		return $dirpath;
	}
}

=head2 Function GetChromFastaFile, GetCDSFastaFile, GetGeneFastaFile, GetProtFastaFile, GetTranscriptFastaFile

 Usage    : my $filepath = $o_io->GetChromFastaFile( chrom   => $chrom )
            my $filepath = $o_io->GetCDSFastaFile()
            my $filepath = $o_io->GetGeneFastaFile()
            my $filepath = $o_io->GetProtFastaFile()
            my $filepath = $o_io->GetTranscriptFastaFile()
 Function : return the path of the fasta file for the wanted chrom.
 Return   : $filepath, string.
 Args     :
     $chrom, string, chrom label
 Error    :

=cut

sub GetChromFastaFile
{
	my $self = shift;
	return $self->GetFastaFile( @_, feature => 'CHR' );
}
sub GetCDSFastaFile
{
	my $self = shift;
	return $self->GetFastaFile( @_, feature => 'CDS' );
}
sub GetGeneFastaFile
{
	my $self = shift;
	return $self->GetFastaFile( @_, feature => 'GENE' );
}
sub GetProtFastaFile
{
	my $self = shift;
	return $self->GetFastaFile( @_, feature => 'PROT' );
}
sub GetTranscriptFastaFile
{
	my $self = shift;
	return $self->GetFastaFile( @_, feature => 'TRANSCRIPT' );
}



=head2 GetChromIndexFilepaths

 Usage        : my $ra_filepath = $o_x->GetChromIndexFilepaths( %h_args )
 Function     : .
 Returns      : $filepath
 Args         : hash with following keys
     -ext, string, file extension, (anything after the FIRST dot) _seq,seq.bi,seq.bl,seq.fas,seq.he
     -num, integer, the integer following the 'V'
     -where : string, optional [INSTANCE] or SESSION.
     		if you want many files, use *
 Error        : none

=cut

sub GetChromIndexFilepaths
{
	my $self = shift;
	my %h_args = (	-ext => 'seq',
					-num => 1,
					-where => undef,
					@_ );

	my $dirpath = $self->GetDataPath( $h_args{ -where } );
	my @a_dirs  = $self->GetInstanceObserver()->GetCfg( qw( DB CHR INDEX ) );

	my $filename= $self->GetShortName() . 'V';
	$filename  .= $h_args{ -num } . '.' . $h_args{ -ext };

	my $filepath = File::Spec->catfile( $dirpath, @a_dirs, $filename );

	my @a_filepaths = glob( $filepath ); # in case of *

	return \@a_filepaths;
} #GetChromIndexFilepaths



=head2 Function GetChromsInfo

 Usage    :(1) my $rh_chroms = $o_so->GetChromsInfo()
           (2) my $rh_chroms = $o_so->GetChromsInfo(10)
 Function : return the chromosomes information, as a hash
            (1) return ALL information
            (2) limit the information returned to the 10 longest chromosomes
 Return   : a reference to a hash of hash:
              key = chr_name
              val = { chr => $chrName, length => $length, centro => $centro }
 Args     : none
 Error    : ParameterException

=cut

sub GetChromsInfo
{
	my $self = shift;
	my ($limit) = @_;

	my @a_attr = qw( __o_inst_obs __short __rh_chroms_info );
	my( $o_inst_obs, $species, $rh_chroms ) = $self->_Get( @a_attr );
	
	if ( !defined($rh_chroms) )
	{
		# appeler GetSpeciesChromInfo de IO ?
		# => pour eviter de relire plusieurs fois le fichier karyotype
		my $o_dba = $self->GetDBAdaptor();
		my $slice_adaptor = $o_dba->get_SliceAdaptor();
		$rh_chroms        = $slice_adaptor->karyotype();
	}
	
	if (!defined $limit)
	{
		return $rh_chroms;
	}
	else
	{
		my %h_chroms_limited;
		my $ra_names = $self -> GetChromsNames( 'SIZE', $limit );
		foreach my $chr_name ( @$ra_names )
		{
			$h_chroms_limited {$chr_name} = $rh_chroms -> {$chr_name };
		}
		return \%h_chroms_limited;
	}
}

=head2 Function GetChromsNames

 Usage    : (1) my $ra_names = $o_so->GetChromsNames();
            (2) my $ra_names = $o_so->GetChromsNames( 'SIZE' );
			(3) my $ra_names = $o_so->GetChromsNames( 'SIZE', 10 );
			(4) my $ra_names = $o_so->GetChromsNames( 'SIZE', 3, [ 'I','II','III','IV' ]);
			(1,2) returns the names of all chromosomes, cleverly sorted
			(3)   returns the names of the 10 longest chromosomes, sorted from longest to shortest
			(4)   returns the names of the 3 longest chromosomes among the 4 specified, sorted from longest to shortest

 Function : return the sorted list of chromosome names.
 Return   : a reference to an array (sorted list of names)
 Args     : $sort_type (string): ID (default) sort by ID, using NARCISSE_utils::SortByChromID
                                 SIZE sort by chromosome size, longest first
            $limit (int)       : If specified, returns only this number of names
            $ra_restrict       : If specified, restrict to the specified list

=cut

sub GetChromsNames
{
	my $self = shift;
	
	my ($sort_type, $limit, $ra_restrict) = @_;
	$sort_type = 'ID' unless defined $sort_type;
	if ( $sort_type !~ /(^ID$)|(^SIZE$)/ )
	{
		throw LipmError::ParameterException( 'sort_type', 'value',-text => "should be ID or SIZE");
	}

	my $rh_chroms = $self -> GetChromsInfo();
	my @a_unsorted_names = (defined $ra_restrict) ? @$ra_restrict : keys %$rh_chroms;
	my @a_names;
	if ( $sort_type eq 'ID' )
	{
		@a_names = &NARCISSE_utils::SortByChromID(@a_unsorted_names);
	}
	else
	{
		@a_names = sort { $rh_chroms->{$b}->{length} <=> $rh_chroms->{$a}->{length} } @a_unsorted_names;
	}
	if (! defined $limit)
	{
		return \@a_names;
	}
	else
	{
		$limit -= 1;
		my @a_limited_names = @a_names[0..$limit];
		return \@a_limited_names;
	}
}

=head2 Function GetGenomeSize

 Usage    : my $size = $o_sp -> GetGenomeSize();

 Function : Compute if necessary and return the genome length;
 Return   : a reference to an array (sorted list of names)
 Args     : $sort_type (string): ID (default) sort by ID, using NARCISSE_utils::SortByChromID
                                 SIZE sort by chromosome size, longest first
            $limit (int)       : If specified, returns only this number of names
            $ra_restrict       : If specified, restrict to the specified list

=cut
sub GetGenomeSize
{
	my $self = shift;
	
	if ( $self -> _Get('__genome_size') == 0 )
	{
		my $size = 0;
		my $rh_chroms = $self -> GetChromsInfo();
		foreach my $chr (keys %$rh_chroms)
		{
			$size += $rh_chroms ->  { $chr } -> { length };
		}
		$self -> _Set ( '__genome_size', $size );
	}
	return $self -> _Get('__genome_size');
}

=head2 Function GetChromLength

 Usage    : my $len = $o_x->GetChromLength( $chrom )
 Function : return the chromosome length
 Return   : The chromosome length
 Args     : $chrom  : A chromosome label
 Error    : ParameterException

=cut

sub GetChromLength
{
	my $self = shift;
	my( $chrom ) = @_;

	my $rh_chrom_info = $self->GetChromsInfo();

	return $$rh_chrom_info{ $chrom }{ length };
}


=head2 Function ExtractChromNameFromFilename

 Usage        : my $chrom = $o_x->ExtractChromNameFromFilename( $filepath )
 Function     : raccourci vers InstanceObserver::ExtractChromNameFromFilename.
 Returns      : $chrom
 Args         : $filepath
 Error        : none

=cut

sub ExtractChromNameFromFilename
{
	my $self = shift;

	my $o_inst_obs = $self->GetInstanceObserver();
	my $chrom = $o_inst_obs->ExtractChromNameFromFilename( @_ );

	return $chrom;
} #ExtractChromNameFromFilename


=head2 function GetDBAdaptor

 Title      : GetDBAdaptor
 Usage      : my $o_dba = $o_x->GetDBAdaptor();
 Prerequiste: Use API.
 Function   : Build a DBadaptor object for the species
 Returns    : the dba object for the species
 Args       : none.
 Error      :

=cut

sub GetDBAdaptor
{
	my $self = shift;

	my $o_dba = new Bio::Narcisse::DBAdaptor( -o_species => $self );

	return $o_dba;
}


=head2 GetNbProteins

 Usage        : my $nb = $o_x->GetNbProteins()
 Function     : return the number of proteins from rep_prot <=> nb genes.
 Returns      : $nb
 Args         : none
 Error        : none

=cut

sub GetNbProteins
{
	my $self = shift;

	my $nb = $self->_Get( '__nb_proteins' );

	if( !defined $nb )
	{
		# get db
		my $fasta = $self->GetProtFastaFile();
		my @a_keys = qw( FASTA_PROT FASTA_REP_PROT );
		my( $bad, $good ) = $self->GetInstanceObserver()->GetCfg( @a_keys );
		$fasta =~ s/$bad$/$good/; # le bon fichier

		# 10 fois rapide qu'avec Bio::SeqIO
		my $o_fasta = New FastaReader( $fasta );
		my $ra_headers = $o_fasta->GetHeaders();
		$nb = scalar( @$ra_headers );

		$self->_Set( __nb_proteins => $nb );
	}

	return $nb;
} #GetNbProteins


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
