# $Id: Sufine.pm 1927 2008-07-25 11:41:54Z manu $
#

package Sufine;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

 Sufine - 

=head1 PUBLIC METHODS

=cut

use strict;
use warnings;
use Carp;
use File::Basename;
use Cwd;
use IO::File;

@Sufine::ISA = qw( IO::File );

# Some global variables - Please modify those variables if the compress programs are not in the path

$Sufine::COMPRESS = 'compress';
$Sufine::GZIP     = 'gzip';
$Sufine::BZIP2    = 'bzip2';
$Sufine::UNCOMPRESS= 'uncompress';
$Sufine::GUNZIP    = 'gunzip';
$Sufine::BUNZIP2   = 'bunzip2';

# Some module-related variables

our $date_regex    = '[0-2][0-9][0-1][0-9][0-3][0-9]';  # yymmdd, valid from 2000 to 2029
our $ver_regex     = '[0-9]{1,2}';                      # 0,10,11,12,...
our $text_regex    = '[a-zA-Z0-9_]+';                   # Allowed characters for method, ext
our $name_regex    = '[a-zA-Z0-9_\\/]+';                # Allowed characters for name: $text_regex + '/'
our $name_regex_wc = '[a-zA-Z0-9_\[\]*?.]+';            # Same as $name_regex, but wildcard added and / removed
our $dir_regex     = '[a-zA-Z0-9_\\/\\-]+';             # Allowed characters for dir: $name_regex + '-', supporting the Sufine directories in -dir args
our $compress_regex= '\\.(Z)|(gz)|(bz2)';               # Compression indicator


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1927 $ =~ /\d+/g); $r[0]};
}

sub DESTROY
{
    my $self = shift;
    $self->close();
    my $absolute_path_name = ${*$self}{'__absolute_path_name'};
    return unless (defined $absolute_path_name);        # If called by an exception generated in the constructor
    &__Compress($absolute_path_name);   # Compress the file if necessary
}

=pod

=head2    New 

 --- CREATING A SUFINE ---
 
 Usage:       my $o_sufine = New Sufine (-dir         => 'some_directory',
                                         -name        => '>some_name',
                                         -method      => 'some_method_name',
                                         -ext         => 'some_extension',
                                         -compression => '','Z','gz','bz2',
                                         -mode        => 'R','W','A');
 
 or
 
              my $o_sufine = New Sufine ('>some_name'); (pass -name, the other parameters are set to their default values)
 
 Constructor: Create and open a file named: some_directory/some_name-050925-5-some_method_name.some_extension
              where 050925 is the date, 5 is the version number
 Parameters:  -dir         = create the file in this directory - Defaults to . (the current directory)
              -name        = MANDATORY - The name of the file to create, > indicates file creation, >> file append
              -method      = The name of the "method" this file is an output of. Defaults to basename($0)
              -ext         = The extension. Defaults to txt
              -compression = Creates the file through compress ('Z') or gzip ('gz') or bzip2 ('bz2') - Defaults to No compression
              -mode        = Read/Write/Append mode. You may also specify <,>,>>  in front of the name field
              
 --- OPENING A SUFINE FOR READ or FOR APPEND ---
  
 Usage:       my $o_sufine = New Sufine (-dir    => 'some_directory',
                                         -name   => '<some_name', or '>>some_name'
                                         -version=> 'some_number',
                                         -method => 'some_method_name',
                                         -ext    => 'some_extension');
 
 or
 
              my $o_sufine = New Sufine ('some_name'); (pass -name, other parameters are set to their default values)
              my $o_sufine = New Sufine ('>>some_name');
              
 Constructor: Open for reading the file named: some_directory/some_name-YYMMDD-some_number-some_method_name.some_extension
 Parameters:  -dir    = open the file in this directory - Defaults to . (the current directory)
              -name   = MANDATORY - The name of the file to open, may begin with < for read, must begin with >> for append
              -version= A version number, defaults to the highest possible version taking into account the other parameters
              -method = A method name, defaults to * (any method name)
              -ext    = An extension name, defaults to * (any extension)
              
 WARNING - The file MUST BE ALREADY CREATED, even for opening in append mode.
           This is different from the append mode in perl or in shell

=cut

sub New
{
    my $type = shift;
    my %h_args;
    my $name;
    if (@_==1) 
    {                   # Only 1 argument
        $name = shift;
        $h_args{'-name'} = $name;
    }
    else
    {
        %h_args = @_;
        $name = $h_args{'-name'};
    };

    my ($prefix) = ($name =~ /\A(>{0,2})/);
    my $mode = 'R';
    $mode = 'W' if ($prefix eq '>');
    $mode = 'A' if ($prefix eq '>>');
    
    my $path_name = &GetPath(%h_args);
    return undef unless (defined $path_name);
    
    my $absolute_path_name = &__CalcAbsolutePathName($path_name);

    my $path_to_create = &__UnCompress($absolute_path_name,$mode);                     # uncompress the file, if necessary
    my $class = ref($type) || $type || "IO::File::Sufine";
    my $fh = $class->SUPER::new($prefix.$path_to_create);
    return undef unless (defined $fh);

    ${*$fh}{'__path_name'}          = $path_name;
    ${*$fh}{'__absolute_path_name'} = $absolute_path_name;
    ${*$fh}{'__mode'}               = $mode;

    return $fh;
};

#
# We must forbid the use of new, because the Sufine will be partially constructed (new is the constructor of its superclass)
#

sub new
{
    die 'Please use New instead of new to build a new Sufine object';
};

=pod

=head2 GetMode

 Usage:    my $mode = $o_sufine->GetMode();
 Function: Return the mode of the sufine
 Access:   Public
 Return:   The mode of the sufine: R,W,A
 
=cut

sub GetMode 
{
    my $self = shift;
    return ${*$self}{'__mode'};
};
 
=head2 GetFileName  

 Usage:    my $file_name = $o_sufine->GetFileName();
 
 or
 
           my $file_name = Sufine::GetFileName(...);  Same parameters as for the constructors
 Function: Member function (1st) or class function (2nd)
           When used as a class function and file creation (>some_name), a 0-sized file is created
 Access:   Public
 Return:   The file name corresponding to the sufine
           NB - The directory part of the file is NOT returned. See GetPath
 
=cut

sub GetFileName 
{
    my $file_name = &GetPath(@_);
    return basename($file_name);
};
 
=pod

=head2 GetAbsolutePath  

 Usage:    my $file_path = $o_sufine->GetAbsolutePath();
  or
           my $file_name = Sufine::GetAbsolutePath(...);  Same parameters as for the constructors

 Function: Member function (1st) or class function (2nd)
           When used as a class function and file creation (>some_name), a 0-sized file is created
 Access:   Public
 Return:   The absolute path corresponding to the sufine
 
=cut

sub GetAbsolutePath
{
    my $self = $_[0];
    if (ref($self))                 # We were called as $obj->GetAbsolutePath()
    {
        return ${*$self}{'__absolute_path_name'};
    }
    else                            # We were called as Sufine::GetAbsolutePath(...)
    {
        my $path = &GetPath(@_);
        return &__CalcAbsolutePathName($path);
    };
};

=pod

=head2 GetPath  

 Usage:    my $file_path = $o_sufine->GetPath();
 or
           my $file_name = Sufine::GetPath(...);  Same parameters as the constructor

 Function: Member function (1st) or class function (2nd)
           When used as a class function and file creation (>some_name), a 0-sized file is created
 Access:   Public
 Return:   The file path corresponding to the sufine, or undef
           The <, >, >> signs are suppressed from the returned path
 
=cut

sub GetPath 
{
    if (@_==0) {
        croak ("SUFINE ERROR - You must specify at least the name field");
    };
    my $self = $_[0];
    if (ref($self))                          # We were called as $obj->GetPath() or $obj->GetFileName()
    {
        return ${*$self}{'__path_name'};
    }
    else                                     # We were called as Sufine::GetPath(...) or Sufine::GetFileName(...)
    {
        my %h_args;
        if (@_==1)      # Only 1 argument: it is the name
        {
            $h_args{'-name'} = shift;
        }
        else
        {
            %h_args = @_;
        };
        
        if ($h_args{'-name'} =~ /\A\// and exists($h_args{'-dir'}))
        {
            croak ("SUFINE ERROR - -name is an absolute name: ".$h_args{'-name'}." and -dir is specified: ".$h_args{'-dir'});
        };
        
        if (exists($h_args{'-mode'}))
        {
            croak ("SUFINE ERROR - illegal mode specified: ".$h_args{'-mode'}) unless ($h_args{'-mode'} =~ /\A[ARW]\Z/);
        }
        
        my $path = '';        
        if ($h_args{'-name'} =~ /\A>>/ or (exists($h_args{'-mode'}) and $h_args{'-mode'} eq 'A'))         # Append mode
        {
            $h_args{'-name'} =~ s/\A>>//;
            $path = &__CalcPathFileExists(%h_args);
        }
        elsif ($h_args{'-name'} =~ /\A>/ or (exists($h_args{'-mode'}) and $h_args{'-mode'} eq 'W'))       # Write mode
        {
            $h_args{'-name'} =~ s/\A>//;
            $path = &__CalcPathFileNew(%h_args);
        }
        else                                    # Read mode
        {
            $h_args{'-name'} =~ s/\A<//;
            $path = &__CalcPathFileExists(%h_args);
        }
        return $path;        
    }
};

=head1 PUBLIC CLASS FUNCTIONS

=head2 SymLink  

 Usage:    Sufine::SymLink(-dir  => 'directory',
                           -name => 'name',...);
 Procedure: Create in the current directory a symlink towards a Sufine in the directory dir
            Same parameters as specified to the constiructor, when opening a Sufine for reading
 Access:   Public
 Return:   NOTHING - WARNING This function is NOT a method, and DOEST NOT create a new object.
           If necessary, you have to create a new Sufine, using the locally created file as parameter.
 
=cut

sub SymLink 
{
    my %h_args = @_;
    if (ref($_[0]))
    {
        croak ("SUFINE ERROR - Sufine::SymLink is NOT a method");
    };
    if (!exists($h_args{'-dir'}) or $h_args{'-dir'} eq '.')
    {
        croak ("SUFINE ERROR - Cannot make a link against myself");
    };
    if (!exists($h_args{'-name'}))
    {
        croak ("SUFINE ERROR - name is not specified");
    };
    if ($h_args{'-name'} =~ /\A>/)
    {
        croak ("SUFINE ERROR - To create a symlink, you cannot specify > or >>");
    };

    my $path_name = &__CalcPathFileExists(%h_args);
    my $target_version = __GetVer($path_name);
    
    delete $h_args{'-dir'};
    my $local_path_name = &__CalcPathFileExists(%h_args);
    my $local_version = (defined $local_path_name) ? &__GetVer($local_path_name) : 0;
    if ($local_version >= $target_version)
    {
        croak ("SUFINE ERROR - Version conflict");
    };
    
    my $file_name = basename($path_name);
    
    
    if (defined($path_name))
    {
        symlink($path_name,$file_name) or carp("SUFINE ERROR - Cannot create the symlink to $path_name -  $!");
    }
    else
    {
        croak("SUFINE ERROR - I did not find the file to link to");
    };
};

=head2 MkDir

 Usage:    my $path_name = Sufine::MkDir(-dir  => 'some_directory',
                                         -name => 'some_name');
 Access:   Public
 Function: Create a directory named: some_directory/some_name-050925-5
           where 050925 is the date, 5 is the version number
 Parameters: -dir         = make the dir in this directory - Defaults to . (the current directory)
             -name        = MANDATORY - The name of the directory to make
 Return:   The path name, or undef if the directory cannot be created

=cut

sub MkDir {
    my %h_args;
    my $dir;
    my $name;
    
    if (@_==1) 
    {                   # Only 1 argument
        $name = shift;
        $dir = '.';
        $h_args{'-name'} = $name;
    }
    else
    {
        %h_args = (-name=>'',
                   -dir=>'.',@_);
        croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');
        
        $dir    = $h_args{'-dir'};
        $name   = $h_args{'-name'};
    };
    
    #
    # We call __CalcPathDirExists
    #

    my $existing_file_name = &__CalcPathFileExists(%h_args);
    if (defined($existing_file_name))
    {
        croak "SUFINE ERROR - There is a Sufine file called $existing_file_name";
    };
    my $existing_dir_name  = &__CalcPathDirExists(%h_args);
    my $date = &__CalcDateField();
    my $ver = '';
    if (!defined($existing_dir_name)) 
    {
        $ver = 1;
    }
    else
    {
        $ver = &__GetVer($existing_dir_name) + 1;
        if ($ver >= 100)
        {
            croak ("SUFINE ERROR - The version number CANNOT BE Higher than 99");
        };
    };
    
    my $path = $name.'-'.$date.'-'.$ver;
    if ($dir ne '.')
    {
        $path = $dir.'/'.$path;
    }; 
    if (mkdir($path))
    {
        return $path;
    }
    else
    {
        carp ("SUFINE ERROR - Cannot create $path - $!");
        return undef;
    };
};

=head2 ChDir

 Usage:    my $path_name = Sufine::ChDir(-dir  => 'some_directory',
                                         -name => 'some_name',
                                         -ver  => 'some_version');
 Access:   Public
 Function: Change to a directory named some_directory/some_name-050925-5
           where 050925 is the date, 5 is the highest version number for a directory called some_name in some_directory
 Parameters: -dir         = changes to a dir in this directory - Defaults to . (the current directory)
             -name        = MANDATORY - The name of the directory to change to
             -ver         = The version number of the directory - Defaults to the highest version number
 Return:   The path name, or undef if it was impossible to change to the directory

=cut

sub ChDir {
    my %h_args;
    my $dir;
    my $name;
    
    if (@_==1) 
    {                   # Only 1 argument
        $name = shift;
        $dir = '.';
        $h_args{'-name'} = $name;
    }
    else
    {
        %h_args = (-name=>'',
                   -dir=>'.',@_);
        croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');
        
        $dir    = $h_args{'-dir'};
        $name   = $h_args{'-name'};
    };
   
    #
    # We call __CalcPathDirExists
    #
    
    my $existing_dir_name = &__CalcPathDirExists(%h_args);
 
    if (!defined($existing_dir_name)) {
        carp ("SUFINE ERROR - directory not found");
        return undef;
    };
    
    if (!chdir($existing_dir_name))
    {
        carp ("SUFINE ERROR - Cannot change to $existing_dir_name - $!");
        return undef;
    };
    return ($existing_dir_name);
};    

=head2 SymDirLink  

 Usage:    Sufine::SymDirLink(-dir  => 'directory',
                              -name => 'name',
                              -VER  => 'VER');
 Procedure: Create in the current directory a symlink towards a directory Sufine in the directory dir
            Same parameters as specified to chdir
 Access:   Public
 Return:   nothing
 
=cut

sub SymDirLink 
{
    my %h_args = @_;
    if (ref($_[0]))
    {
        croak ("SUFINE ERROR - Sufine::SymDirLink is NOT a method");
    };
    if (!exists($h_args{'-dir'}) or $h_args{'-dir'} eq '.')
    {
        croak ("SUFINE ERROR - Cannot make a link against myself");
    };
    if (!exists($h_args{'-name'}))
    {
        croak ("SUFINE ERROR - name is not specified");
    };

    my $path_name = &__CalcPathDirExists(@_);
    my $file_name = basename($path_name);
    
    if (defined($path_name))
    {
        symlink($path_name,$file_name) or carp("SUFINE ERROR - Cannot create the symlink to $path_name -  $!");
    }
    else
    {
        croak("SUFINE ERROR - I did not find the file to link to");
    };
};

=pod

=head2 GetDirName, GetDirPath, GetDirAbsolutePath  

 Usage:    my $file_name = Sufine::GetDirName(...);  Same parameters as for the ChDir
 Function: Returns the directory name, the path name, or the absolute path name of
           an EXISTING directory.
 Access:   Public
 Return:   The file or path name corresponding to the sufine directory
 
=cut

sub GetDirName 
{
    return basename(&GetDirPath(@_));
};
 
sub GetDirAbsolutePath
{
    return &__CalcAbsolutePathName(&GetDirPath(@_));
};

sub GetDirPath
{
    return (&__CalcPathDirExists(@_));
}

=head2 Glob  

 Usage:    my @sufines = Sufine::Glob(-select => 'type', -dir => 'dir', -name => 'name', ...)
           or
           my @sufines = Sufine::Glob('name')
 Function: Returns the list of Sufines matching the parameters
           Glob works for the files AND for the directories (parameter -filetype)
           GlobDir calls Glob with the parameter -filetype set to 'DIR'
 Access:   Public
 Arguments: -select: 'ALL'  Retrieve all files matching the other criteria (default)
                     'LAST' For each matching name, retrieve only the highest version file
                     'OLD'  For each matching name, retrieve all versions, except the highest.
            -filetype = 'FILE' or 'DIR' (default 'FILE')
            -dir,-name,-ver,-method,-ext cf. New for the meaning of those parameters
 Return:   an array, with the list of retrieved files in array context
           a string, with the list of retrieved files as space-separated file names in scalar context.

=cut

sub GlobDir
{
    my %h_args = (-filetype=>'DIR',@_);
    return Glob(%h_args);
};

sub Glob
{
    if (@_==0) {
        croak ("SUFINE ERROR - You must specify at least the name field");
    };
    my %h_args;
    if (@_==1)
    {
        %h_args = (-select=>'ALL',
                   -filetype=>'FILE',
                   -name=>shift,
                   -dir=>'.',
                   -ver=>'*',
                   -method=>'*',
                   -ext=>'*');
    }
    else
    {
        %h_args = (-select=>'ALL',
                   -filetype=>'FILE',
                   -name=>'',
                   -dir=>'.',
                   -ver=>'*',
                   -method=>'*',
                   -ext=>'*',@_);
    };
    croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');

    #
    # Check and save the parameters
    #

    my $select = $h_args{'-select'};
    my $filetype=$h_args{'-filetype'};
    my $dir    = $h_args{'-dir'};
    my $name   = $h_args{'-name'};
    my $nf     = basename($name);
    my $nd     = dirname($name);
    my $ver    = $h_args{'-ver'};
    my $method = $h_args{'-method'};
    my $ext    = $h_args{'-ext'};

    if ($select !~ /\A[AOL]/)
    {
        croak ("SUFINE ERROR - Bad value for the select argument: -select=>$select");
    };
    if ($filetype !~ /\A[FD]/)
    {
        croak ("SUFINE ERROR - Bad value for the filetype argument: -filetype=>$filetype");
    };
    
    if (($dir ne '.' and $dir !~ /\A$dir_regex\Z/) or
        (($nd ne '.' and $nd !~ /\A$dir_regex\Z/)) or
        ($nf !~ /\A$name_regex_wc\Z/) or
        ($ver ne '*' and $ver !~ /\A$ver_regex\Z/) or
        ($method ne '*' and $method !~ /\A$text_regex\Z/) or
        ($ext ne '*' and $ext !~ /\A$text_regex\Z/))
    {
        croak ("SUFINE ERROR - Bad value for one argument: -dir=>$dir, -name=>$name, -ver=>$ver, -method=>$method, -ext=>$ext");
    };
    
    # Save current directory, then change to specified directory, if any
    # Croak if a directory was specified, but impossible to chdir
    
    my $dirsav = '.';
    if ($dir ne '.') {
        $dirsav = cwd;
        unless (chdir $dir)
        {
            croak ("SUFINE ERROR - Cannot chdir to $dir - $!");
        };
    };
    
    # will be passed to glob: a relax filer
    # will be filtered after glob: a more strict filter
    # croak if some argument is not correct
    
        
    my $glob_pattern       = $name.'-'.'*'.'-'.$ver;
    my $after_glob_pattern = '\A'.$name_regex.'-'.$date_regex.'-'.$ver_regex;
    my $after_glob_comp_pattern;
    
    if ($filetype =~ /\AF/)
    {
        $glob_pattern       .= '-'.$method.'.'.$ext.'*';
        $after_glob_pattern .= '-'.$text_regex.'.'.$text_regex;
        $after_glob_comp_pattern = $after_glob_pattern . $compress_regex;
    }
    else            # For a directory, the compression indicator is a nonsense
    {
        $after_glob_comp_pattern = $after_glob_pattern;
    }
    
    $after_glob_pattern        .= '\Z';
    $after_glob_comp_pattern   .= '\Z';
    
    #
    # Keep in a hash of arrays the files matching the patterns
    #      The files sharing the same name field are kept in the same array
    #
    
    my %names = ();
    foreach my $f (glob($glob_pattern)) {
        chomp($f);
        if (basename($f) =~ /$after_glob_pattern/ or basename($f) =~ /$after_glob_comp_pattern/) 
        {
            my $name = &__GetName($f);
            if (exists $names{$name})
            {
                push(@{$names{$name}},$f);
            }
            else
            {
                my @n = ($f);
                $names{$name} = \@n;
            };
        };
    };

    #
    # No file found: return undef
    #

    if (keys(%names)==0) {
        chdir ($dirsav);
        return undef;
    };

    #
    # Some specific version requested: only 1 file / name should be returned
    #
    
    if ($h_args{'-ver'} ne '*')
    {
        foreach my $k (keys %names)
        {
            if (@{$names{$k}}!=1)
            {                           # some file name is ambiguous, ERROR
                croak ("SUFINE ERROR - Several files have the same name: ".join(' - ',@{$names{$k}}));
            };
        };
    };

    #
    # No specific version requested: the files must be sorted
    #
    
    foreach my $k (keys %names)
    {
        @{$names{$k}} = sort {&__GetVer($a)<=>&__GetVer($b)} @{$names{$k}};
    };

    # 
    # Return the result
    #

    my @rvl;

    if ($select =~ /\AA/i)           # ALL mode selected
    {
        if ($dir ne '.')
        {
            foreach my $k (sort keys %names)
            {
                push(@rvl, map ($dir.'/'.$_, @{$names{$k}}));
            }
        }
        else
        {
            foreach my $k (sort keys %names)
            {
                push(@rvl, @{$names{$k}});
            }
        };
    };

    if ($select =~ /\AO/i)            # OLD mode selected (all but the last one)
    {
        if ($dir ne '.')
        {
            foreach my $k (sort keys %names)
            {
                pop(@{$names{$k}});     # Removing the last file for each name
                push(@rvl, map ($dir.'/'.$_, @{$names{$k}}));
            }
        }
        else
        {
            foreach my $k (sort keys %names)
            {
                pop(@{$names{$k}});     # Removing the last file for each name
                push(@rvl, @{$names{$k}});
            }
        };
    };

    if ($select =~ /\AL/i)            # LAST mode selected
    {
        if ($dir ne '.')
        {            
            foreach my $k (sort keys %names)
            {
                push(@rvl, $dir.'/'.pop(@{$names{$k}}));  # Keeping only the last file of each family
            }
        }
        else
        {
            foreach my $k (sort keys %names)
            {
                push(@rvl, pop(@{$names{$k}}));           # Keeping only the last file of each family
            }
        };
    };
    
    chdir($dirsav);
    if (wantarray)
    {
        return @rvl;
    }
    else
    {
        return join(' ',@rvl);
    };
};

=head1 PRIVATE METHODS and FUNCTIONS

=head2    __CalcPathFileExists 

 Usage:     my $path_name = Sufine::__CalcPathFileExists(-dir   => 'directory',
                                                         -name  => 'name',
                                                         -ver   => N,
                                                         -method=>'method',
                                                         -ext   => 'ext')
 
 Access:    private
 Function:  Calc the path name, from the arguments
            We are looking for an existing file.
 Return:    The path name, or undef if the file does not exist
 Args:      -dir  The directory in which the file lives OPTIONAL, DEFAULT='.'
            -name The name of the file
            -ver  The version number, DEFAULT=highest version in the directory
            -method The creation method OPTIONAL, DEFAULT = '*'
            -ext    The extension OPTIONAL, default = '*'

=cut

sub __CalcPathFileExists {
    my %h_args = (-name=>'',
                -dir=>'.',
                -ver=>'*',
                -method=>'*',
                -ext=>'*',@_);

    croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');

    #
    # Check for the validity of the arguments
    #
    
    if (($h_args{'-name'}!~/\A$name_regex\Z/) or
    ($h_args{'-ver'}ne'*' and $h_args{'-ver'}!~$ver_regex) or
    ($h_args{'-method'}ne'*' and $h_args{'-method'}!~/\A$text_regex\Z/) or
    ($h_args{'-ext'}ne'*' and $h_args{'-ext'}!~/\A$text_regex\Z/))
    {
        croak ("SUFINE ERROR - Bad value for one argument: -name=>".$h_args{'-name'}.", -ver=>".$h_args{'-ver'}.", -method=>".$h_args{'-method'}.", -ext=>".$h_args{'-ext'});
    };
    
    #
    # Use -select = LAST, returning 0 or 1 file
    #
    
    $h_args{'-select'} = 'LAST';
    return &Glob(%h_args);
}
        
=head2    __CalcPathFileNew 

 Usage:     my $path_name = Sufine::__CalcPathFileNew(-dir   => 'directory',
                                                      -name  => 'name',
                                                      -method=>'method',
                                                      -ext   => 'ext', 
                                                      -compression=>'Z/gz/bz2')
 
 Access:    private
 Precondition: -name arg starts with ONLY 1 >
 Function:  Calc the complete path name, from the arguments
            We are creating a file: the file is created, zero-sized
            We look for files withe the same name, and increment the version number.
 Return:    The complete path name, or undef if file cannot be created
 Args:      -dir  The directory in which the file lives OPTIONAL, DEFAULT='.'
            -name The name of the file
            -method The creation method OPTIONAL, DEFAULT = $0 (name of the calling program)
            -ext    The extension OPTIONAL, DEFAULT = 'txt'
            -compression The compression flag, DEFAULT = ''

=cut

sub __CalcPathFileNew {
    my $method = basename($0);
    $method =~ s/\A($text_regex).*/$1/;         # Removing the extensions or other forbidden characters
    my %h_args = (-name=>'',
                  -dir=>'.',
                  -method=>$method,
                  -ext=>'txt',
                  -compression=>'',
                  @_);
    croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');
    croak ("SUFINE ERROR - bad compression indicator") if ($h_args{'-compression'} ne '' 
                                                           and $h_args{'-compression'} ne 'Z' 
                                                           and $h_args{'-compression'} ne 'gz' 
                                                           and $h_args{'-compression'} ne 'bz2');

    my $dir    = $h_args{'-dir'};
    my $name   = $h_args{'-name'};
    $method = $h_args{'-method'};
    my $ext    = $h_args{'-ext'};
    my $compression = $h_args{'-compression'};
    
    #
    # We call __CalcPathFileExists, keeping only the arguments -dir and -name
    #
    
    delete ($h_args{'-method'});
    delete ($h_args{'-ext'});
    delete ($h_args{'-compression'});

    my $existing_dir_name = &__CalcPathDirExists(%h_args);
    if (defined($existing_dir_name))
    {
        croak "SUFINE ERROR - There is a Sufine directory called $existing_dir_name";
    };

    my $existing_file_name = &__CalcPathFileExists(%h_args);
    my $date = &__CalcDateField();
    my $ver = '';
    if (!defined($existing_file_name)) 
    {
        $ver = 1;
    }
    else
    {
        $ver = &__GetVer($existing_file_name) + 1;
        if ($ver >= 100)
        {
            croak ("SUFINE ERROR - The version number CANNOT BE Higher than 99");
        };
    };
    
    my $path = $name.'-'.$date.'-'.$ver.'-'.$method.'.'.$ext;
    if ($dir ne '.')
    {
        $path = $dir.'/'.$path;
    }; 
    if ($compression ne '')
    {
        $path .= '.'.$compression;
    };
    
    open(TEST,">$path") or croak ("SUFINE ERROR - Cannot create $path - $!");
    close(TEST);
    return $path;
};
    
=head2    __CalcPathDirExists 

 Usage:     my $dir_path_name = Sufine::__CalcPathDirExists(-dir   => 'directory',
                                                            -name  => 'name',
                                                            -ver   => N)
 Access:    private
 Function:  Calc the path name, from the arguments
            We are looking for an existing directory.
 Return:    The path name, or undef if the dir does not exist
 Args:      -dir  The directory in which the directory lives OPTIONAL, DEFAULT='.'
            -name The name of the directory
            -ver  The version number, DEFAULT=highest version in -dir

=cut

sub __CalcPathDirExists {
    my %h_args = (-name=>'',
                -dir=>'.',
                -ver=>'*',@_);

    croak ("SUFINE ERROR - name is not specified") if ($h_args{'-name'} eq '');

    #
    # Check for the validity of the arguments
    #
    
    if (($h_args{'-name'}!~/\A$name_regex\Z/) or
    ($h_args{'-ver'}ne'*' and $h_args{'-ver'}!~$ver_regex))
    {
        croak ("SUFINE ERROR - Bad value for one argument: -name=>".$h_args{'-name'}.", -ver=>".$h_args{'-ver'});
    };

    #
    # Use -select = LAST, returning 0 or 1 file
    #
    
    $h_args{'-select'} = 'LAST';
    return &GlobDir(%h_args);
}
 
=head2    __CalcDateField 

 Usage:     my $date_filed = &__CalcDateField(); 
 Access:    private
 Function:  Calc the date field, from the today's date
 Return:    The date field: YYMMDD
 Args:      none
 
=cut

sub __CalcDateField 
{
    my @today = localtime();
    my $rvl   = sprintf("%02i%02i%02i",$today[5]-100,$today[4]+1,$today[3]);
    return $rvl;
};

=head2    __CalcAbsolutePathName

 Usage:     my $absolute_path_name = &__CalcAbsolutePathName($path_name)
 Access:    private
 Function:  Compute the absolute path from the path name and from the current directory.
            Take into account the > and >> characters that may be specified in frnt of $path_name
 Return:    The absolute path
 Args:      The relative path
 
=cut

sub __CalcAbsolutePathName {
    my $path_name = shift;
    my $cwd = cwd;
    my $rvl = '';
    if ($path_name =~ /\A>{0,2}\//) {  # The path is already absolute
        $rvl = $path_name;
    } 
    elsif ($path_name =~ /\A(>{0,2})(.+)/)
    {
        $rvl = $1.$cwd.'/'.$2;
    };
    return $rvl;
};

=pod

=head2 __GetName  

 Usage:    my $name = $o_sufine->GetName()
 or
           my $name = Sufine::GetName(some_name);

 Function: Member function (1st) or class function (2nd)
           returns the name field
 Access:   Private
 Return:   The version (an integer)
 Error:    Die if the parameter is not a Sufine-type name
 
=cut

sub __GetName 
{   
    if (@_==0) {
        croak ("SUFINE ERROR - No parameter specified");
    };
    my $self = $_[0];
    if (ref($self))                          # We were called as $obj->GetVer()
    {
        die "Not yet implemented";
        #return ${*$self}{'__path_name'};
    }
    else                                     # We were called as Sufine::GetVer(...)
    {
        my $file_name=basename($_[0]);
        if ($file_name =~ /\A(${text_regex})-${date_regex}-${ver_regex}-${text_regex}\.${text_regex}/)
        {
            return $1;
        }
        elsif ($file_name =~ /\A(${text_regex})-${date_regex}-(${ver_regex})\Z/)  # A directory
        {
            return $1;
        }
        else
        {
            croak "SUFINE ERROR - Bad file name $file_name";
        };
    };
};

=head2 __GetVer  

 Usage:    my $ver = $o_sufine->GetVer()
 or
           my $ver = Sufine::GetVer(some_name);  Same parameters as for the constructors

 Function: Member function (1st) or class function (2nd)
           returns the Version field, AS AN INTEGER
 Access:   Private
 Return:   The version (an integer)
 Error:    Die if the parameter is not a Sufine-type name
 
=cut

sub __GetVer 
{   
    if (@_==0) {
        croak ("SUFINE ERROR - No parameter specified");
    };
    my $self = $_[0];
    if (ref($self))                          # We were called as $obj->GetVer()
    {
        die "Not yet implemented";
        #return ${*$self}{'__path_name'};
    }
    else                                     # We were called as Sufine::GetVer(...)
    {
        my $file_name=basename($_[0]);
        if ($file_name =~ /\A${text_regex}-${date_regex}-(${ver_regex})-${text_regex}\.${text_regex}/)
        {
            my $ver = $1 + 0;
            return $ver;
        }
        elsif ($file_name =~ /\A${text_regex}-${date_regex}-(${ver_regex})\Z/)  # A directory
        {
            my $ver = $1 + 0;
            return $ver;
        }
        else
        {
            croak "SUFINE ERROR - Bad file name $file_name";
        };
    };
};

=head2    __Compress

 Usage:     my $compressed_path = &__Compress($path_to_compress);
 Access:    private, called by the destructor
 Function:  If the path ends with .Z, .gz, .bz2, remove this extension and call the correct compression function.
 Return:    The path of the compressed file
 Args:      none
 Globals:   $COMPRESS_EXE, $GZIP_EXE, $BZIP2_EXE
 
=cut

sub __Compress
{
    my $absolute_path = shift;
    my $compress_exe = '';

    # Example: 
    #    if   $absolute_path      = >data1-050928-20-method2.txt.Z
    #    then 
    #         $uncompressed_path  = data1-050928-20-method2.txt
    #         $compress_extension = Z
    #         $canonical_path     = data1-050928-20-method2.txt.Z
    #
    #    We are in compression mode only if $compress_extension = Z, gz, or bz2
    #
 
    my ($uncompressed_path,$compress_extension) = ($absolute_path =~ /(.+)\.(.+)\Z/);
    my $canonical_path = $uncompressed_path.'.'.$compress_extension;

    if (!defined($compress_extension))
    {
        return $absolute_path;       # Nothing to compress (but should not happen)
    }
    elsif ($compress_extension eq 'gz')
    {
        $compress_exe = $Sufine::GZIP;
    }
    elsif ($compress_extension eq 'bz2')
    {
        $compress_exe = $Sufine::BZIP2;
    }
    elsif ($compress_extension eq 'Z')
    {
        $compress_exe = $Sufine::COMPRESS;
    }
    else
    {
        return $absolute_path;        # Nothing to compress
    };
    
    my $cmd = $compress_exe.' '.$uncompressed_path;
    system($cmd);
    return $absolute_path; 
};

=head2    __UnCompress

 Usage:     my $uncompressed_path = &__UnCompress($path_to_uncompress,$mode);
 Access:    private
 Function:  If the path ends with .Z, .gz, .bz2, remove this extension and call the correct compression function.
            Called by the constructor
 Return:    The path of the uncompressed file
 Args:      $path, $mode (RWA)
 Globals:   $COMPRESS_EXE, $GZIP_EXE, $BZIP2_EXE
 
=cut

sub __UnCompress
{
    my $absolute_path = shift;
    my $mode = shift;
    my $uncompress_exe = '';

    # Example: 
    #    if   $absolute_path      = data1-050928-20-method2.txt.Z
    #    then 
    #         $uncompressed_path  = data1-050928-20-method2.txt
    #         $compress_extension = Z
    #         $canonical_path     = data1-050928-20-method2.txt.Z
    #
    #    We are in compression mode only if $compress_extension = Z, gz, or bz2
    #
 
    my ($uncompressed_path,$compress_extension) = ($absolute_path =~ /(.+)\.(.+)\Z/);
    my $canonical_path = $uncompressed_path.'.'.$compress_extension;

    if (!defined($compress_extension))
    {
        return $absolute_path;       # Nothing to compress (but should not happen)
    }
    elsif ($compress_extension eq 'gz')
    {
        $uncompress_exe = $Sufine::GUNZIP;
    }
    elsif ($compress_extension eq 'bz2')
    {
        $uncompress_exe = $Sufine::BUNZIP2;
    }
    elsif ($compress_extension eq 'Z')
    {
        $uncompress_exe = $Sufine::UNCOMPRESS;
    }
    else
    {
        return $absolute_path;        # Nothing to compress
    };
    
    if ($mode eq 'W')         # File creation: the file may exist, it must be removed
    {
        unlink($canonical_path);
    }
    else
    {
        my $cmd = $uncompress_exe.' '.$canonical_path;
        system($cmd);
    };
    return $uncompressed_path; 
};

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

