#
# 	sebastien.letort@toulouse.inra.fr
#	Created: February 9, 2009
#	$Id$
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package SuppDataLoader;

=pod

=head1 NAME

 SuppDataLoader - 

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

# perl
use Data::Dumper;
use File::Basename;
use Error qw( :try );
use LipmError;
use base qw( LipmObject );

# lipm
use ElementsSet::MultiGraph;
use LipmError::Function;
use General;


=head2 method _Init

 Title        : _Init
 Usage        : my $o_x = New SuppDataLoader( ? );
 Prerequisite : none
 Fonction     : initialise an instance of SuppDataLoader class.
 Returns      : a valid SuppDataLoader object
 Args         : ?
 Attributes   :
    __o_logger, #inherited from LipmObject

    __org, string short species name
    __rao_elt_set, list ref on ElementsSet created
    __ra_data, list ref of hash ref, keys are
			graph_class
			feature
			file_directory
    __filter, regexp, select which feature will be displayed

=cut

sub _Init
{
	my $self = shift;
	my %h_args = (
			o_species => undef,
			ra_directories => [],

			-o_logger => undef,
			-filter   => '.', # anything,
			-feat_prefix => undef,
			@_ );

	&LipmError::Function::Assert( %h_args );

	$self->SUPER::_Init( %h_args );

	$self->_Set( __o_species => $h_args{ o_species },
	             __filter    => $h_args{ -filter },
	             __ra_data   => [],
				 __feature_prefix => $h_args{ -feat_prefix }
				);

	my $ra_supp_dirs = $h_args{ ra_directories };
	foreach my $dir_path ( @$ra_supp_dirs )
	{
		# explore and set __ra_data with a list of valid data
		$self->__ExploreSuppDirectory( $dir_path );
	}

	return;
} # _Init


=head2 Procedure __ExploreSuppDirectory

 Usage        : $o_xx->__ExploreSuppDirectory( $supp_path );
 Prerequisite : none
 Procedure     : browse a supp dir to look for supplementary data.
                When some are found a hash ref is build and added to the returned data.
 Args         : $supp_path, path of a supp_data directory.
	$supp_path should be the output of o_species->GetSuppDataPath( $x ),
	but any directory can be analysed.

=cut

sub __ExploreSuppDirectory
{
	my $self = shift;
	my( $supp_path ) = @_;

	# return unless( -d $dir_path );
	my $dh_supp = new IO::Dir( $supp_path )
					or return;

	my $msg = "$supp_path will be explored to seek supp_data files.\n";
	$self->_Trace( $msg, '.', 'IL' );

	my @a_attr = qw( __ra_data __filter __feature_prefix );
	my( $ra_data, $rx_filter, $prefix ) = $self->_Get( @a_attr );

	my $feature_prefix = 'supp_';
	$feature_prefix   .= $prefix . '_' if( defined $prefix );

	#graph_dir can be lineplots or histogram (for now).
	# it represents the ES::MultiGraph class that will be used for their data.
	while( my $graph_dir = $dh_supp->read() )
	{
		next if( $graph_dir =~ /^\./ );

		my $graph_path  = File::Spec->catdir( $supp_path, $graph_dir );

		my $graph_class = &General::Camelize( $graph_dir );

		my $dh_graph = new IO::Dir( $graph_path )
						or next;
		while( my $feat_dir = $dh_graph->read() )
		{
			next if( $feat_dir =~ /^\./ );
			next unless( $feat_dir =~ $rx_filter );

			my $feat_path = File::Spec->catdir( $graph_path, $feat_dir );
			$msg = "$feat_path contains supp_data.\n";
			$self->_Trace( $msg, '.', 'IL' );
			
			my $feat_name = $feature_prefix . $feat_dir;

			my $rh_ = { feature        => $feat_name,
						file_directory => $feat_path,
						graph_class    => $graph_class };
			push( @$ra_data, $rh_ );
		}
	}

	return;
}


=head2 Function BuildFeatures
 Usage   : my $rh_features = $o_x->BuildFeatures( ... )
 Function:  Build the elementsSet corresponding to the supplementary features.
 Returns : $rao_elements
 Args    : A lot of parameters

=cut
sub BuildFeatures
{
	my $self   = shift;
	my %h_args = (  chromosome => undef,
					start => 1,
					end   => 1,
					# o_param  => undef,
					@_ );

	my @ao_elts = ();

	my( $ra_data ) = $self->_Get( '__ra_data' );
	my $n = @$ra_data;
	my $msg = __PACKAGE__ . " will try to build $n object ES::MultiGraph.\n";
	$self->_Trace( $msg, '.', 'ILV' );

	# rh_feature contient
	# { feature => $feat_dir,
	# file_directory => $feat_path,
	# graph_class => $graph_class }
	foreach my $rh_feature ( @$ra_data )
	{
		my %h_param = (
				%$rh_feature,
				-o_logger => $self->_GetLogger(),
				%h_args
			);
		my $error_flg = &FALSE;
		try
		{
			my $o_element = New ElementsSet::MultiGraph( %h_param );
			push( @ao_elts, $o_element );
		}
		catch LipmError with
		{
			my ($o_error) = @_;
			$error_flg = &TRUE;
		};
		
		# next in the catch block does not work well
		next if ($error_flg);
		my $msg = "A new ElementsSet::MultiGraph has been built for";
		$msg   .= " $h_args{ chromosome } [ $h_args{ start } - $h_args{ end } ]";
		$msg   .= " feature: $$rh_feature{ feature }.\n";
		$self->_Trace( $msg, '.', 'ILV' );

	}

	return \@ao_elts;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

