# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Textual - a class for Narcisse textual representations of objects (genes etc.)

=head1 SYNOPSIS

 my $o_textual = New -> Textual();
 
 # Add a "layer" (ie a csv file)
 $o_textual -> AddLayer ( $name, $file_name );
 
 # Get a layer from its name, may be creating a new layer
 my $o_file = $o_textual -> GetLayer( $name );

 # GetLayersList()
 my $ra_list = $o_textual -> GetLayersList();
 

=head1 DESCRIPTION

 Textual plays the same role as Image, but is used for textual representations of objects and of links between objects
 A 'layer' is a csv file, opened by the method AddLayer
 PrintHeader prints the header... unless already done: you may call PrintHeader several times, only the FIRST header will be printed
 GetLayer returns the opened file, you just have to write data to this file
 TODO - Utiliser NarcisseDataWriter ?

=head1 AUTHOR

Emmanuel Courcelle : emmanuel.courcelle@toulouse.inra.fr

=cut

package Textual;

use strict;
use warnings;

use base qw( LipmObject );

use General;
use Data::Dumper;
use Error qw(:try);
use LipmError::IOException;

my $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};

=head2 _Init

 Title        : _Init
 Usage        : _Init()
 Prerequisite : none
 Procedure    : Initialization
 Returns      : reference of the blessed object
 Args         : none

=cut
sub _Init
{
	my $self = shift;
	
	# key = layer name, value = object IO::File
	$self -> _Set ( '__rh_layers', {} );
	
	# key = layer name, value = 0/1
	$self -> _Set ( '__rh_header_printed', {} );

	return $self;
}

=head2 AddLayer
 Title		  :	AddLayer
 Usage		  : $o_image -> AddLayer(name,file)
 Procedure	  : Add a layer 
 Args		  : name : name of the layer
 				file : name of the corresponding file

=cut
sub AddLayer
{
	my $self = shift;
	my ( $name, $file ) = @_;
	throw LipmError::ParameterException('name', 'missing') unless( defined $name );
	throw LipmError::ParameterException('file', 'missing') unless( defined $file );

	# If the layer already exists, nothing to do
	my $rh_layers = $self -> _Get ('__rh_layers');
	if ( !defined $rh_layers -> {$name} )
	{
		my $rh_files = $self -> _Get ('__rh_layers');
		my $rh_file = new IO::File($file,'w') or throw LipmError::IOException($file,'f');
		$rh_files -> { $name } = $rh_file;
	}
}

=head2 DeleteLayer
 Title		  :	DeleteLayer
 Usage		  : $o_image -> DeleteLayer(name)
 Procedure	  : Delete a layer 
 Args		  : name : name of the layer

=cut
sub DeleteLayer
{
	my $self = shift;
	my ($name) = @_;
	my $rh_layers = $self -> _Get( '__rh_layers');
	my $rh_header_printed = $self -> _Get('__rh_header_printed');
	my $layer = $self -> GetLayer ($name);
	$layer -> close();
	$layer = '';
	delete $rh_layers -> { $name };
	delete $rh_header_printed -> { $name };
}

=head2 GetLayer
 Title		  : GetLayer
 Usage		  : my $rh_file = $self -> GetLayer(name)
 Procedure	  : Get the reference to the IO::File object of the layer named 'name'.
 				Throw an exception if this layer does not exist
 Returns	  : $rh_file
 Args		  : name

=cut
sub GetLayer
{
	my $self     = shift;
	my ( $name ) = @_;
	throw LipmError::ParameterException("name", 'missing') unless ( defined $name );
	throw LipmError::ParameterException("layer $name", 'missing') unless ( $self -> IsLayer( $name ) );
	return $self -> _Get( '__rh_layers') -> { $name };
}

=head2 IsLayer
 Title		  : IsLayer
 Usage		  : if ( $self -> IsLayer(name) ) ...
 Function     : Return &TRUE if the layer exist, &FALSE if it does not
 Returns	  : &TRUE/&FALSE
 Args		  : name

=cut
sub IsLayer
{
	my $self     = shift;
	my ( $name ) = @_;
	throw LipmError::ParameterException("name", 'missing') unless ( defined $name );
	my $rh_layers = $self -> _Get( '__rh_layers');
	return ( defined $rh_layers->{$name} ) ? &TRUE : &FALSE;
}

=head2 GetLayersList
 Title		  :	GetLayersList
 Usage		  : $self -> GetLayersList()
 Procedure	  : Return the actual list of layer names
 Returns	  : Array layers names
 Args		  : none
 Attributes	  : ra_layers

=cut
sub GetLayersList
{
	my $self = shift;

	my $rh_layers = $self -> _Get ( '__rh_layers' );
	return keys %$rh_layers;
}

=head2 PrintHeader
 Title		  :	PrintHeader
 Usage		  : $self -> PrintHeader ( $name, @a_labels)
 Procedure	  : Prints the header to the file/layer $name, unless already printed
 Args		  : $name: The name of the layer to consider
                An array of labels, will be printed as a CSV line (separated by \t)

=cut
sub PrintHeader
{
	my $self = shift;
	my ($name, @a_labels) = @_;
	my $rh_file = $self -> GetLayer( $name );
	my $rh_header_printed = $self -> _Get('__rh_header_printed');
	unless (exists $rh_header_printed -> { $name })
	{
		print $rh_file join("\t",@a_labels) . "\n";
		$rh_header_printed -> { $name } = 1;
	}
}

=head2 DeleteAll
 Title		  :	DeleteAll
 Usage		  : $o_textual -> DeleteAll();
 Procedure	  : Delete all layers
 Args		  : None
 
=cut
	
sub DeleteAll
{
	my $self = shift;
	my @a_layers = $self -> GetLayersList();
	foreach my $layer (@a_layers)
	{
		$self -> DeleteLayer( $layer );
	}
}

=head2 DESTROY
 Title		  :	DESTROY
 Usage		  : $o_textual = undef;
 Procedure	  : Destructor - flush and close all files
 Args		  : None
 
=cut
sub DESTROY
{
	my $self = shift;
	$self -> DeleteAll();
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
