#
# 	sebastien.letort@toulouse.inra.fr
#	Created: August 05, 2008
#	$Id: ThroughSpeciesComputer.pm 3773 2012-07-06 09:52:43Z sletort $

# Copyright INRA/CNRS
#
# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr
#
# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
package ThroughSpeciesComputer;

use strict;
use warnings;

use base qw( LipmObject );

=pod

=head1 NAME

 ThroughSpeciesComputer - an abtract class for class that compute whatever on Narcisse data.

=head1  SYNOPSIS

	# do not instantiate : inherit from it and overwrite the following methods:
	
	
	use base ( ThroughSpeciesComputer );
	sub _ComputeData
	sub _GetHeaderFile
	sub _CleanData
	sub _GetPrintOptions
	sub _ComputeFilepath

=head1 DESCRIPTION

 This class contains methods needed by classes
 performing computations onto Narcisse data.

 It defines 2 list refs corresponding to species and to chromosomes, using InstanceObserver

 You can specify one or more species, one or more chromosomes too.
 If no species is given, all species of the instance will be used.
 If no chromosome is given, all chrom of the species will be used.

=cut

use Data::Dumper;
use Error qw( :try );
use File::Path;
use File::Spec;

use General;
use NARCISSE_utils;
use Configuration::InstanceObserver;
use NarcisseDataWriter::OnDisk;
use LipmError::ParameterException;
use LipmError::IOException;


=head2 method _Init

 Usage        : $self->SUPER::_Init();
 Prerequisite : none
 Fonction     : Initialize common attributes.
 Returns      : a valid object
 Args         : %h_args as follow :
              	'-species', [empty] list ref of species short name
              	'-chroms', [empty] list ref of chrom label

 Attributes   :
 	_ra_species, ref to a list of short species names
 	             if the list is empty, all species of the instance
 	             	will be used.
 	             if one species does not belong to the instance
 	_ra_chroms, ref to a list of chromosome labels

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -species  => [],
	               -chroms   => [],
	               -o_logger => undef,
	               -o_io     => undef,
	              @_ );
	# Set logger
	$self->SUPER::_Init( %h_args );

	# set attributes
	my $ra_species = $h_args{ -species };
	my $ra_chroms  = $h_args{ -chroms };
	my $rh_current = { o_slice_adaptor => undef,
	 				   chrom           => undef,
					   rh_chroms       => undef };

	my $o_inst_obs;
	if( !defined( $h_args{ -o_io } ) )
	{
		$o_inst_obs = &InstanceObserver::Singleton( -o_logger => $h_args{ -o_logger } );
	}
	else
	{
		$o_inst_obs = $h_args{ -o_io };
	}
	
	
	$self->_Set( __ra_species => $ra_species,
				 __ra_chroms  => $ra_chroms,
				 _rh_current => $rh_current,
				 __o_inst_obs => $o_inst_obs );

	# no abnormal values - can throw error to stop constructor
	$self->_CheckAttributesAndSetDefaults();

	return $self;
} # _Init


=head2 procedure _CheckAttributesAndSetDefaults

	Title        : _CheckAttributesAndSetDefaults
	Usage        : $bool = $o_xx->_CheckAttributesAndSetDefaults();
	Prerequisite : none
	Procedure    : check species attributes. If none was specified, will be set to all instances species.
	Args         : none
	Error        : LipmError::ParameterException

	Note         : Daugther class should call this one.

=cut

sub _CheckAttributesAndSetDefaults
{
	my $self = shift;
	
	my @a_attr = qw( __ra_species __o_inst_obs );
	my( $ra_species, $o_inst_obs ) = $self->_Get( @a_attr );

	# species
	#	=> if empty, all species of the instance
	#	=> if species not in the instance, warn and return false;
	if( 0 == @$ra_species )
	{
		$ra_species = $o_inst_obs->ExtractMySpecies( [] );
		if( 0 == @$ra_species )
		{
			my $msg = "Your instance is probably misconfigured. ";
			$msg   .= "No species found into it.";
			throw LipmError( -user_msg => $msg );
		}
	}
	else
	{
		foreach my $species ( @$ra_species )
		{
			if( !$o_inst_obs->HaveSpecies( $species ) )
			{
				my $msg = "$species does not belong to your instance.";
				my $dbg = "Check into narcisse.cfg.db for GENOME hash";
				$dbg   .= ", and check that $species belongs to it.\n";

				throw LipmError::ParameterException( 'species', 'value',
				                      -user_msg => $msg, -text => $dbg );
			}
		}
	}

	return;
}


=head2 Function RollOverSpecies

 Usage        : $o_xx->RollOverSpecies( %h_args );
 Prerequisite : none
 Function     : for each 'species' involved and for each 'chrom' involved,
                this method will call DoTheJobForOneChrom method,
                If no chrom is specified, all chrom of the species will be used.
 Args         : %h_args with keys
	               '-auto_clean', [0] boolean,
		                              if true and if computed data are equal on each point, (or equals 0 ?)
		                              no file will be written.

=cut

sub RollOverSpecies
{
	my $self = shift;

	#valeurs par defaut pour infos
	my %h_args = ( #-auto_clean=> &FALSE,
	               @_ );

	# foreach species build a SliceAdaptor
	my @a_attr = qw( __ra_species _rh_current );
	my( $ra_species, $rh_current ) = $self->_Get( @a_attr );
	foreach my $species ( @$ra_species )
	{
		$self->_Trace( "current species : $species\n", '.', 'IC' );
		my $o_sla = $self->__BuildSliceAdaptor( $species );

		$$rh_current{ o_slice_adaptor } = $o_sla; # effacement en cours ( 13-10-2009 ) TODO
		$$rh_current{ o_dba }           = $self->_GetDBAdaptor( $species );
		$$rh_current{ rh_chroms }       = $o_sla->karyotype();

		$self->_RollOverChroms( %h_args );
	}
	
	return;
}


=head2 Function _RollOverChroms

 Usage        : $o_xx->_RollOverChroms( %h_args );
 Prerequisite : none
 Function     : for each chrom involved in the current species,
                this method will call __DoTheJobForOneChromosome method,
                If no chrom is specified, all chrom of the species will be used.
 Args         : %h_args with keys
	               '-auto_clean', [0] boolean,
	                    if true and if computed data are equals on each point, (or equals 0 ?)
	                    no file will be written.

=cut

sub _RollOverChroms
{
	my $self = shift;

	#valeurs par defaut pour infos
	my %h_args = ( #-auto_clean=> &FALSE,
	               @_ );

	# get chrom infos
	my @a_attr = qw( __ra_chroms _rh_current );
	my( $ra_chroms, $rh_current ) = $self->_Get( @a_attr );

	if( 0 == @$ra_chroms )
	{
		my $rh_chroms = $$rh_current{ rh_chroms };
		my @a_chroms  = keys %$rh_chroms;
		$ra_chroms    = \@a_chroms;

		my $msg = "no chrom specified, will use all (";
		$msg   .= scalar( @$ra_chroms ) . ") chroms of the org\n";
		$self->_Trace( $msg, '.', 'IC' );
	}

	# foreach chromosome - move the window onto it
	foreach my $chrom ( @$ra_chroms )
	{
		$$rh_current{ chrom } = $chrom;
		$self->__DoTheJobForOneChromosome( %h_args );
	} # end chrom loop

	return;
} # _RollOverChroms


=head2 Function __DoTheJobForOneChromosome

 Usage        : $o_xx->__DoTheJobForOneChromosome( %h_args );
 Prerequisite : none
 Function     : Calls ComputeData in a safe mode.
                If an LipmError is thrown, only this calculation will be broken.
				When data are computed they are printed in a file with NARCISSE_utils::PrintNarcisseData
 Args         : %h_args with keys
	               '-auto_clean', [0] boolean,
	                    if true and if computed data are equals on each point,
	                    no file will be written.
 Error        : LipmError are caught here.

=cut

sub __DoTheJobForOneChromosome
{
	my $self = shift;
	my %h_args = ( -auto_clean => &FALSE,
	               @_ );

	# juste pour la trace.
	my $rh_current = $self->_Get( '_rh_current' );
	my $chrom    = $$rh_current{ chrom };
	my $rh_chrom = $$rh_current{ rh_chroms }{ $chrom };
	my ( $chrName, $len ) = @$rh_chrom{ qw( chr length ) };
	my $msg = "current chrom : $chrName - $len\n";
	$self->_Trace( $msg, '.', 'IC' );

	try
	{
		my $rh_data = $self->_ComputeData();

		# _CleanData peut vider $rh_data si besoin
		$self->_CleanData( $rh_data )  if ( &TRUE == $h_args{ -auto_clean } );

		if( 0 == keys %$rh_data )
		{
			my $msg = "No data computed, or the data were dropped.\n";
			$self->_Trace( $msg, '.', 'IC' ); # or throw LipmError ?
			return;
		}

		$self->__PrintData( $rh_data );
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		warn $o_err->Explain();
	};
	# else we don't catch it

	return;
}


=head2 __PrintData

 Usage    : $o_x->__PrintData( $rh_data )
 Procedure: Print the data to the file returned by _ComputeFilepath()
 Args     : $rh_data/$ra_data

=cut

sub __PrintData
{
	my $self = shift;
	my( $rh_data ) = @_;

	# args for the constructor
	my $filepath   = $self->_ComputeFilepath();
	my $ra_headers = $self->_GetHeaderFile();
	my ($rh_options,$ra_sort_keys) = $self->_GetPrintOptions();

	my $msg = "Will write --@$ra_headers-- in $filepath\n";
	$self->_Trace( $msg, '.', 'ICV', __PACKAGE__ );

	# Note : si c'est possible utiliser un ra_data au lieu de rh_data
	# ça m'arrangerait pour certify.pl
	my $o_logger = $self->_GetLogger();
	my $o_ndw    = New NarcisseDataWriter::OnDisk( $filepath, $ra_headers, $rh_options, $ra_sort_keys, $o_logger );
	$o_ndw       = $o_ndw->Update($rh_data);
	$o_ndw       = undef;

	return;
} #__PrintData


=head2 function __BuildSliceAdaptor

 Usage        : $o_slice_adaptor = $o_xx->__BuildSliceAdaptor( $species );
 Prerequisite : none
 Function     : this method builds a dbAdaptor from the
 				InstanceObserver object, to get a slice adaptor.
 Returns      : $o_slice_adaptor, a SliceAdaptor object for $species.
 Args         : $species, string, the short name of the species
 Error        : LipmError if API pb.

 Note : I think that as get_SliceAdaptor almost never fail, this method should
       be removed. Please deal with the new method _GetDBAdaptor instead.

=cut

sub __BuildSliceAdaptor
{
	my $self = shift;
	my ( $species ) = @_;

	# get DBAdaptor for this species
	my $o_dba = $self->_GetDBAdaptor( $species );

	my $o_slice_adaptor = $o_dba->get_SliceAdaptor();

	my $msg = "no slice adaptor could have been generated for $species\n";
	throw LipmError( -text => $msg )	if( !defined($o_slice_adaptor ) );

	return $o_slice_adaptor;
}


=head2 _GetDBAdaptor

 Usage    : my $o_dba = $o_x->_GetDBAdaptor( $species )
 Function : Build a DBAdaptor for the given species.
 Returns  : $o_dba, API DBAdaptor object.
 Args     : $species, string, species short/long name
 Error    : LipmError if API pb.

=cut

sub _GetDBAdaptor
{
	my $self = shift;
	my( $species ) = @_;

	my $o_io      = $self->_GetInstanceObserver();
	my $o_species = $o_io->GetSpeciesObserver( $species );
	my $o_dba     = $o_species->GetDBAdaptor();

	my $msg = "no DBAdaptor could have been generated for $species\n";
	throw LipmError( -text => $msg )	if( !defined($o_dba ) );

	return $o_dba;
} #_GetDBAdaptor


=head2 accessors

  _GetInstanceObserver

=cut

sub _GetInstanceObserver
{
	my $self = shift;
	
	return $self->_Get( '__o_inst_obs' );
}


=head2 function _GetHeaderFile

	Title        : _GetHeaderFile
	Usage        : $ra_headers = $o_xx->_GetHeaderFile();
	Prerequisite : none
	Function     : this function returns the headers of the datafile.
	Returns      : $ra_headers, list ref of headers,
	             	their order is their order in the datafile
	             	something like
	             	[ 'chrom', 'start', 'end', 'value', 'opt_circos', 'id' ]
	Args         : none.

	Note         : It *MUST* be redefined in daughter class.

=cut

sub _GetHeaderFile
{	return [];	}


=head2 function _ComputeData

 Usage        : $rh_data = $o_xx->_ComputeData();
 Prerequisite : none
 Function     : this function computes and return the data.
                $rh_data will be passed to NarcisseDataWriter
 Returns      : $rh_data, like { id => { key1 => val1, ..., keyn => valn } }
 Args         : none.

 Note         : It *SHOULD* be redefined in daughter class.

=cut

sub _ComputeData
{	return {};	}




=head2 function _ComputeFilepath

 Title        : _ComputeFilepath
 Usage        : $filepath = $o_xx->_ComputeFilepath();
 Prerequisite : none
 Function     : Computes the filepath and make the path if necessary.
 Returns      : $filepath, the formatted absolute filepath
 Args         : none

 Note         : It *MUST* be redefined in daughter class.

=cut

sub _ComputeFilepath
{
	my $self = shift;

	throw LipmError::ObjectException( $self, 'abstract method' );

	return '';
}


=head2 function _GetPrintOptions

 Title        : _GetPrintOptions
 Usage        : $rh_options = $o_xx->_GetPrintOptions();
                This function is protected, it may be overriden by the derived classes
                The original function uses the default values from NarcisseDataWriter
 Function     : return options to pass NarcisseDataWriter
 Returns      : $rh_options (see NarcisseDataWriter for the details)
                $ra_sort_keys (the keys used to sort data)
 Args         : none
 
=cut

sub _GetPrintOptions
{
	my $self = shift;
	return (undef, undef);
}

=head2 function _GetPrintOptions_AJETER

 Title        : _GetPrintOptions_AJETER
 Usage        : $rh_options = $o_xx->_GetPrintOptions();
 Prerequisite : none
 Function     : return options to pass to NARCISSE_utils::PrintNarcisseData
 Returns      : $rh_options, with at least 2 keys
 				sort, string, to specify how to sort value of rh_data keys,
 				verbose, bool, true if you want some infos.
 Args         : none
 Note         : We now use PrintNarcisseData2, so check what is its options.

=cut

sub _GetPrintOptions_AJETER
{
	my $self = shift;

	return {
		fn_sort  => sub
			{
				my( $ra_list ) = @_;
				return sort { $a <=> $b } @$ra_list;
			},
		verbose => 1,
		o_logger => $self->GetLogger()
	};
}


=head2 _CleanData

 Title        : _CleanData
 Usage        : $o_xx->_CleanData( $rh_data, $col, $r_value );
                Should be overwritten
 Procedure    : This method empty $rh_data if data should not be printed.
 Args         : $rh_data, a ref hash of hash ref, used by NarcisseDataWriter

=cut

sub _CleanData
{
	my $self = shift;
	my ( $rh_data ) = @_;

	# do nothing, should be overwritten.
	return;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

