#
# 	emmanuel.courcelle@toulouse.inra.fr
#	Created: december 05, 2008
#	$Id: WebBuilderFactory.pm 2317 2009-01-22 14:23:43Z sletort $
#

package NcScript;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 NAME

 NcScript.pm

=head1 SYNOPSIS


=head1 DESCRIPTION

 NcScript is a BASE CLASS for the Narcisse script objects
 For each Narcisse request:
    - one object deriving from NcScript is created
    - This object connects himself to a Webbuilder object (using the WebBuilder's method AddObject)
    -
    
 NcScript maintains one private variable with the information about the reference, the targets, the chromosomes, described in a hash of perl arrays
    _rh_organisms
         SPECIES  => ra_species  = The reference ([0]) and targetN ([N]) parameters (SHORT NAMES ONLY)
         PROGRAMS => ra_programs = The 'program_targetN' parameters ([0] = 'target' parameter)
         CHROMS   => ra_chroms   = The 'chroms_reference ([0]) or the 'chroms_targetN' ([1]) parameters
 
=cut
use strict;
use warnings;

use Data::Dumper;
use File::Basename;
use IO::File;
use File::Spec;

use Error qw( :try );
use LipmError;
use LipmError::ObjectException;
use NarcisseError;

use General;
use WebBuilder;
use LipmError::IOException;
use NarcisseServiceFactory;
use NarcisseServiceKeywords qw( &NS_FASTA );
use Sequence_utils;
use AnotherTemplate;
use ElementsSet;

use base qw( LipmObject );

# Used among other things by the _YUIAddXXX protected functions
use constant DEBUG => 1;
use constant FILENAME_SEPARATOR => '_';
#TODO : changer le separateur ou bien interdire le '_' dans le nom court d'espece.

use constant IMG_EXT   => ".png";
use constant COORD_EXT => ".coord";
use constant TXT_EXT   => ".txt";
use constant JS_EXT    => ".js";
use constant CSV_EXT   => ".csv";
use constant TSV_EXT   => ".tsv";
use constant CST_EXT   => ".cst";   # A .txt acting as a flag for the .csv files

# The parameters in this array will NOT be saved to local.cfg etc. (see DESTROY)
my @A_DO_NOT_SAVE = qw( __wba_ajax __wba_start __wba_index history_stamp  gene center_synt features 
    click center_synt search pos UploadSession SaveSession GenomicSeq GeneSeq ProteinSeq CdsSeq UploadData sequence UploadSeq
    SaveData ViewAlign GlintFish LaunchGlintFish PrintImage OnlineCompare knowledge_list WithMaps Textual );

=head2 function _Init

 Title        : _Init
 Fonction     : the constructor
 Args         : $o_io         : Instance Observer (the configuration object)
                $o_web_builder: A valid WebBuilder object
                $o_logger     : A valid logger
                $flg_local    : flag - If set, a timestamp is computed and kept in $o_param and in the object
                                It will be used at destruction time (see DESTROY)
                                Defaults to &FALSE

=cut

sub _Init
{
	my $self = shift;

	my ($o_io,$o_wb,$o_logger,$flg_local) = @_;
	$flg_local = &FALSE unless (defined $flg_local);

	$self->SUPER::_Init( $o_logger );

	$self->_Set('_o_io',$o_io);
	$self->_Set('_rh_cfg',$o_io->GetConfigHash());
	$self->_Set('__error_message','');
    $self->_Set('_http_status',200);
    $self->_Set('_http_message','');
    
	my $o_param = $o_wb->GetObjectParam();
	$self->_Set('_o_wb',$o_wb);
	$self->_Set('_o_param',$o_param);

	# Log something for every Verbose log level
	$self->_Trace("Generic constructor of NcScript - now building ".ref($self)."\n","*","V");

    # Clean the parameters and create the _rh_organisms protected member
	$self->__CleanParameters();

	# Set the WebBuilder function
	$o_wb->SetObject($self);

	# If the parameter __json_A_GENSEL_DATA is set, keep the gensel configuration to a special file
	if ($o_param->IsDefined('__json_A_GENSEL_DATA'))
	{
		my $f_gensel  = $o_io->GetCfg('ROOT_LOCAL_GENSEL');
		my $fh_gensel = new IO::File(">$f_gensel") or throw LipmError::IOException($f_gensel,'f' );
		print $fh_gensel $o_param->Get('__json_A_GENSEL_DATA');
		$fh_gensel->close();
		$o_param->Delete('__json_A_GENSEL_DATA');
	}

	if ($flg_local == &TRUE)
	{
		# Compute a time stamp and keep it in $o_param and in $self
		my $history_stamp = $o_logger->CalcStamp();
		$history_stamp    =~ s/#//g;          # remove the # as they cannot be used in the urls
		$o_param->Set('history_stamp',$history_stamp);
		$self->_Set('__history_stamp',$history_stamp);
	}
	else
	{
		# No time stamp, $o_param will not be saved
		$self->_Set('__history_stamp','');
	}
	
    #warn "koukou " . ref($self) . ' ' . $o_param->Get('level') . ' ' . $o_param->Get('view_level') . "\n"; 
	return;
}


=head2 PrepareData

 Usage    : $o_x->PrepareData()
 Function : Prepare the data, BEFORE calling Webbuilder's Realize
            Call the protected method _PrepareData
            The protected attributes _http_status and _http_message may be modified by _PrepareData
            If _http_status is different from 200, the WebBuilder's method SetHttpStatus is called
            (see the cgi)
 Args     : none

=cut

sub PrepareData
{
	my $self = shift;
    $self->_PrepareData();
    
    if ( $self->_Get('_http_status') ne '200' )
    {
        my $o_wb = $self->_Get('_o_wb');
        $o_wb->SetHttpStatus( $self->_Get('_http_status','_http_message') );
    }

} #PrepareData

sub _PrepareData{};

=head2 GetHttpStatus

 Usage    : my $sts = $o_x->GetHttpStatus()
 Function : return the http status which has been "prepared" by PrepareData
            Generally unchanged from default value (200), but could be changed in some situations
            (see NcAddBookmark.pm)
 Args     : none
 Return   : the http status and the http message

=cut

sub GetHttpStatus
{
    my $self = shift;
    return $self->_Get( '_http_status','_http_message' );
}

=head2 Procedure __CleanParameters

 Usage     : $self->&__CleanParameters();
 Procedure : delete unwanted parameters, initialize _rh_organisms, __tgt_nb
 Args      : none

=cut

sub __CleanParameters
{
    my $self = shift;
    my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get('_o_io');

	my $flg_rebuild = &FALSE;
    my $flg_auto    = &FALSE;
    
	# Parameters validation (for security)
	my $err = $self->__ValidParams();
	if( $err != &NARCISSE_utils::CH_ERR->{NO_ERROR} )
	{
		throw NarcisseError(  -hd => "NARCISSE ERROR", -bd => "some parameters are not validated (error $err)" );
	}

    my (@a_species,@a_chroms,@a_programs);
    
    # The reference and chroms_reference
    if ($o_param->IsDefined('reference'))
    {
        $a_species[0] = &__LongToShort($o_io,$o_param->Get('reference'),\$flg_rebuild);
    }
    else
    {
        $a_species[0] = 'none';
    };
    
    $a_chroms[0] = $o_param->Get('chroms_reference') if ($o_param->IsDefined('chroms_reference'));
    
    # The targetN, program_targetN, chroms_targetN (N should be <=N !)
    $a_programs[0] = ($o_param->Get('program') ne '') ? $o_param->Get('program') : 'auto';
    $flg_auto = &TRUE if ( $a_programs[0] eq 'auto' );
    
    # If 'target' defined, it is renamed target1 and all other targetN are ignored
    if ($o_param->IsDefined('target'))
    {
        $a_species[1] = &__LongToShort($o_io,$o_param->Get('target'),\$flg_rebuild);
        $a_programs[1]= $a_programs[0];
        $self->_Set('__tgt_nb',1);
        $flg_rebuild = &TRUE;
    }
    
    # No targetN at all, or at least target1.
    # We should have a perfect sequence, else we ignore the highest targets
    # A targetN called 'none' means: no targetN
    else
    {    
        my $j = 0;
        foreach my $k (sort $o_param->GetKeys('^target[1-9]'))
        {
            $j += 1;
            my ($ki) = ($k =~ /^target(.*)/);
            if ($j == $ki and $o_param->Get($k) ne '' and $o_param->Get($k) ne 'none')
            {
                my $p = 'program_target'.$j;
                my $c = 'chroms_target' .$j;
                $a_species[$j] = &__LongToShort($o_io,$o_param->Get($k),\$flg_rebuild);
                $a_chroms[$j] = $o_param->Get($c) if ($o_param->IsDefined($c)); 
                if ($o_param->IsDefined($p))
                {
                    $a_programs[$j] = $o_param->Get($p);
                }
                else
                {
                    $a_programs[$j] = $a_programs[0];
                }
                $flg_auto = &TRUE if ( $a_programs[$j] eq 'auto' );
            }
            else
            {
                $flg_rebuild = &TRUE;
                $j -= 1;
                last;
            }
        }
        $self->_Set('__tgt_nb',$j);
    }
    
    # If some 'auto' program was specified, replace it with a correct method name
    if ( $flg_auto == &TRUE )
    {
        my $ref = $a_species[0];
        if ($ref ne 'none')
        {
            my $o_so = $o_io->GetSpeciesObserver( $ref );
            for (my $j = 1; $j <= $self->_Get('__tgt_nb'); $j++)
            {
                if ( $a_programs[$j] eq 'auto' )
                {
                    my $target = $a_species[$j];
                    my @a_methods;
                    
                    # Call GetMethods, trapping every exception - at the end we have the possible methods in the array
                    try
                    {
                        @a_methods = $o_so->GetMethods( target => $target );
                    }
                    catch Error with {};
                    
                    # NOTE - This works well when there is only ONE method for this couple of species
                    #        But it could be useful to get a more cler algorithm for the general case !
                    $a_programs[$j] = $a_methods[0];
                    $flg_rebuild = &TRUE;
                }
            }
        }
    }

    # Re-init $o_param only if necessary
    if ( $flg_rebuild == &TRUE )
    {
        # delete all targetN, program_targetN, chroms_targetN
        foreach my $k ($o_param->GetKeys('^target'))
        {
            $o_param->Delete($k);
        }
        foreach my $k ($o_param->GetKeys('^program_target'))
        {
            $o_param->Delete($k);
        }
        foreach my $k ($o_param->GetKeys('^chroms_target'))
        {
            $o_param->Delete($k);
        }
                
        # Rebuild those parameters from the protected arrays
        $o_param->Set('reference',$o_io->GetSpeciesLongName($a_species[0]));
        for (my $i =1; $i <= $#a_species; $i++)
        {
            $o_param->Set('target'.$i,$o_io->GetSpeciesLongName($a_species[$i]));
            $o_param->Set('program_target'.$i,$a_programs[$i]);
            $o_param->Set('chroms_target'.$i,$a_chroms[$i]);
        }
    }

    # Keep the @a_species, @a_chroms, @a_programs arrays as protected data
    my %h_organisms = ( 
        SPECIES  => \@a_species,
        PROGRAMS => \@a_programs,
        CHROMS   => \@a_chroms
    );
    $self->_Set('_rh_organisms',\%h_organisms);
    
    # TODO - INTEGRER CELA A WebBuilder
    # If __wb_rest=1, __wba_ajax=j
    if ($o_param->GetInteger('__wb_rest') != &FALSE)
    {
        $o_param->Set('__wba_ajax','j');
        $self->_Set('_rest',&TRUE);
    }
    else
    {
        $self->_Set('_rest',&FALSE);
    };
    
	return;
    
    
    
    # Converting long to short names
    # $flg is set if the name is already a short species name
    # Usage: $new_name = &__LongToShort($o_io,$name,\$flg);
    sub __LongToShort
    {
        my ($o_io,$name,$r_flg) = @_;
        
        my $rvl;
        
        # If already a short name, return it and set the flag
        if ($o_io->IsSpeciesShortName($name)==&TRUE)
        {
            $rvl   = $name;
            $$r_flg |= &TRUE;
        }
        
        else
        {
            # If NOT a long name set the flag
            $$r_flg |= &TRUE if ($o_io->IsSpeciesLongName($name) == &FALSE);
            
            # Try to get the short name anyway
            $rvl  = $o_io->GetSpeciesShortName($name);

            #Throw an exception if not found
            if ($rvl eq 'none')
            {
                throw NarcisseError( -hd => "NARCISSE ERROR", -bd => "Error in parameters: $name is NOT a known species");
                #throw LipmError::ParameterException("species $name", 'missing')
            }
        }
        return $rvl;
    }
}

=head2 function __ValidParams

 Usage    : if ($self->__ValidParams())..
 Fonction : Parameters validation - We check each parameter for a set of characters
 Returns  : 0 ==> all is OK
            2 ==> Bad character in sequence
            Die if there is any other serious security concern
 Args     : none
 Access   : private
 Globals  : none

=cut

sub __ValidParams
{
    my $self = shift;
    my $o_param = $self->_Get('_o_param');

    # === SECURITY ===
	#\s has been added to allow copy-paste in the text area of the blast form
    my $authorized_characters = '[>/0-9a-zA-Z:_.@+\#\-\s\047]*';
    my $authorized_characters_seq = '[>/0-9a-zA-Z:_.@+\#\-\s=,\047]*';

    # If defined, the sequence parameter must match the authorized characters, but please do not die !!!
    if ($o_param->IsDefined('sequence'))
    {
		my $seq = $o_param->Get('sequence');
        if ($o_param->Get('sequence') !~ /^$authorized_characters_seq$/)
        {
            return &NARCISSE_utils::CH_ERR->{FORBID_CHAR};
        };
        
        # Even from linux, a copy/paste inserts ^M characters (at least with firefox)
        $seq =~ s/\r//g;
        $o_param -> Set('sequence',$seq);
    }

    $o_param->UnsetBehaviour('assert_strict');

    # We check every parameter (except sequence), only some characters are allowed
    # Calling AsserAllowedValue, we die if there is an error
    # If some parameter is "none", it is deleted
    my @a_keys=$o_param->GetKeys();
    foreach my $k (@a_keys)
    {
        next if ($k eq 'sequence');
        # __wb_url est l'ensemble des parametres de WebBuilder codes
        # donc avec des caracteres etranges
        next if( $k eq '__wb_url');
        #$o_param->AssertAllowedValue($k,$authorized_characters);
    }

    return &NARCISSE_utils::CH_ERR->{NO_ERROR};
};

=head2 function _IncludeFromTmpl

	Title       : _IncludeFromTmpl
	Usage       : my $html = $self->_IncludeFromTmpl($filepath, [%h_param])
	Prerequiste : The included files must NOT contain any variable named 'warning'
                  Substituted variables are written as '#PERLEVAL|var# in the template file
                  (see AnotherTemplate.pm for the details)
	function    : Open and read the file passed as param and substitute variables.
                  SUBSTITUTED VARIABLES:
                     IOBS      ==> $o_io (object)
                     PAMPAR    ==> $o_param (object)
                     #PERLEVAL|some_var#  ==> $h_param{some_var}, using the parameter %h_param
                     
                  Example: $html .= $self->_IncludeFromTmpl( $file.inc, warning => $warning_msg );
                  You get the html file $file.inc, with a lot of possible substitutions:
                      #PERLEVAL|$warning_msg#                ==> $warning_msg
                      #OBJECT|PAMPAR|Get('__wb_session')#    ==> The Narcisse session number
                      #OBJECT|IOBS|GetCfg('HTTP_HELP_ICON')# ==> The help icon
                      Every stuff accessible from ParamParser and from Instance Observer can be retrieved
  	Returns     : A string: the read file after variable substitution
                  If the file cannot be read and warning => 0 is specified, return ""
	Args        : $filepath	the file to read
                  %h_param,	optional, list of variables to substitute
                  if ($h_param{warning}==0) no error will be generated if the file can't be read
                  default is 1 (warnings generated)
    Access      : protected
	Globals     : none

=cut

sub _IncludeFromTmpl
{
    my $self = shift;
	my $filename = shift;

	#other parameters, default is true for warning
	my %h_param = (warning => 1, @_);

	my $page = "";
	if (!-r $filename or -z $filename)
	{
		if ($h_param{warning}==1)
		{
            throw LipmError::IOException($filename,'f');
		}
        return '';
	}
	else
	{
        my $o_tmpl = New AnotherTemplate($filename);
        my $rh_cfg = $self->_Get('_rh_cfg');
        my $o_io   = $self->_Get('_o_io');
        my $o_param= $self->_Get('_o_param');

		# PLEASE DO NOT USE THE CFG HASH - 
        $o_tmpl->AddVariable('CFG',$rh_cfg);
        $o_tmpl->AddObject('IOBS',$o_io);
		$o_tmpl->AddObject('PAMPAR',$o_param);

        foreach my $k (keys %h_param)
        {
            $o_tmpl->AddVariable($k,$h_param{$k});
        }
        $o_tmpl->Parse();
        return $o_tmpl->Get('parsed_template');
    }
}

=head2 procedure _SaveParamParser

	Title       : _SaveParamParser
	Usage       : $self->_SaveParamParser();
	Procedure   : Save to local.cfg most parameters stored in ParamParser
	              This function may be called from a derived class, this is a convenient way changing some parameter
	              to the main Narcisse page (see NcSendSequences::__MyComputation for an example)
	Args        : none
    Access      : protected
	Globals     : none

=cut
sub _SaveParamParser
{
	my $self    = shift;
	my $o_param = $self->_Get('_o_param');
	
	# Work on a copy and delete the paramters we do not want to save
	my $o_param_alt = New ParamParser('PARAMPARSER',$o_param);
	foreach my $k (@A_DO_NOT_SAVE)
	{
		$o_param_alt->Delete($k);
	};
	
	# Dump to the 'local.cfg' file, or rather to its target (it should be a symlink)...
	my $local_cfg    = 'local.cfg';
	if ( -l $local_cfg )
	{
		$local_cfg = readlink $local_cfg;
	}
	# ... but start with reloading data, in case some parameters are missing (ex: target1 => none vs target1 undefined)
	$o_param_alt -> Update($local_cfg, 'A');
	$o_param_alt -> Dump($local_cfg);
}
	

=head2 procedure DESTROY

 Title     : DESTROY
 Procedure : the destructor
             If the attribute __history_stamp was set by the constructor, call _SaveParamParser
=cut
sub DESTROY
{
	my $self = shift;

	if (IsNotNull $self->_Get('__history_stamp'))
	{
		my $stamp        = $self->_Get('__history_stamp');
		my $o_param      = $self->_Get('_o_param');
		my $local_cfg    = 'local.cfg';
		
        # Return if $o_param undefined (may happen if interrupted by an exception)
        return unless defined $o_param;
        
		# If __wba_index is set; nothing to save
		return if ($o_param->IsDefined('__wba_index'));

		# Remove local.cfg (it is a symlink)
		unlink ($local_cfg);
					
		# Recreate the symlink
		my $history_cfg  = $stamp.'.cfg';
		symlink($history_cfg,$local_cfg);
		
		# Dump ParamParser
		$self -> _SaveParamParser();
		
	}
}

=head2 Procedure PrintHtml()

 Title    : WbFunction
 Usage    : THROUGH THE WebBuilder's Realize
 Procedure: This procedure is called through The realize function of WebBuilder
            PrintHtml is the default value for __wb_function 
 Returns  : The generated html
 Access   : protected
 Args     : none

=cut
sub PrintHtml
{
    my $self   = shift;
    if ($self->_Get('__error_message') ne '')
    {
        print $self->_Get('__error_message');
    }
    else
    {
        my $html = $self->GetHtml();
        if (defined $html and $html ne '')
        {
            print $html;
        }
    }
}

=head2 function GetHtml()

 Title    : GetHtml
 Usage    : $html = $o_ncs->GetHtml()
 Function : Generate the html code, generally called by the WebBuilder's Realize function
            NOTE - This method can be used AS A FUNCTION: the html is RETURNED and printed alltogether
                                        or AS A PROCEDURE: the html is printed from the function body, and the function returns ""
 Returns  : The generated html
 Access   : protected
 Args     : none

=cut
sub GetHtml
{
    my $self = shift;
    throw LipmError::ObjectException( $self, 'abstract method', -text => "ERROR - You should overload NcScript::GetHtml in your derived class");
}

=head2 function _ToJson
    Title  : _ToJson
    Usage  : $self->_ToJson( $scalar );
             $self->_ToJson( \%h_hash );
             $self->_ToJson( \@a_array );
    Function : Return the json representation of a hash with 3 keys:
                      __wb_session the session
                      params       the parameters
                      data         ref to the input variable
    Args   : The data to convert in json
    Return : the json representation of the parameter
    
=cut
sub _ToJson
{
    my $self = shift;
    my ($data) = @_;

	my $o_param = $self->_Get('_o_param');
    my %h_params;
    $o_param->Dump('HASH',\%h_params);
    my $json = new JSON( );
     
    my %h_rvl;
    my $session = '';
    $session = $o_param->Get('__wb_session') if ($o_param->IsDefined('__wb_session'));
    $h_rvl{__wb_session} = $session;
    
    $h_rvl{params}       = \%h_params;
    $h_rvl{data}         = $data;

    return $json->encode( \%h_rvl );
}

=head2 function _YUIAddXXX

	Title      : _YUIAddXXX
	Usage      : $self->_YUIAddXXX()
	Procedure  : Add to the encapsultad WebBuilder object some javascript specifications useful for some YUI functionality
	Args       : none
    Access     : Protected
	globals    : none

=cut
sub _YUIAddDom
{
	warn "_YUIAddDom is OBSOLETE";
 }

sub _YUIAddAutocomplete
{
	warn "_YUIAddAutocomplete is OBSOLETE";
}

sub _YUIAddPanel
{
	warn "_YUIAddPanel is OBSOLETE";
}

sub _YUIAddTabView
{
	warn "_YUIAddTabView is OBSOLETE";
}

sub _YUIAddTreeView
{
	warn "_YUIAddTreeView is OBSOLETE";
}

sub _YUIAddDataTable
{
	warn "_YUIAddAutocomplete is OBSOLETE";
}

=head2 _GetSequenceLocusTag

Title     : _GetSequenceLocusTag
Usage     : _GetSequenceLocusTag($object,$species,$locus_tag,$file,$o_logger)
            _GetSequenceLocusTag($object,$species,\@a_locus_tag,$file,$o_logger)
Procedure : Get the sequence corresponding to the locus tag and write it to $file
Args      : $object:    Gene, Protein, cds
            $species:   The species to consider
            $locus_tag: The locus tag or
            $ra_locus_tag: A ref to an array of locus_tags
            $file:      the output file
Access    : Protected

=cut
sub _GetSequenceLocusTag
{
	my $self = shift;

	my ($object,$species,$locus_tag,$file)=(@_);
	
	my $ra_loci = ( ref $locus_tag eq 'SCALAR' ) ? [ $locus_tag ] : $locus_tag;
	my $o_io = $self->_Get('_o_io');
	my $o_request;
	if ( $object eq 'Gene' )
	{
		$o_request = &NarcisseServiceFactory::GetGenesFromLocusTags(
			species         => $species,
			ra_tags         => $ra_loci,
			-o_logger       => $self->_GetLogger(),
			-o_instance_obs => $o_io,
			out_type        => &NS_FASTA
			);
	}
	elsif ( $object eq 'Protein' )
	{
		$o_request = &NarcisseServiceFactory::GetProteinsFromLocusTags(
			species         => $species,
			ra_tags         => $ra_loci,
			-o_logger       => $self->_GetLogger(),
			-o_instance_obs => $o_io,
			out_type        => &NS_FASTA
			);
	}
	elsif ( $object eq 'Cds' )
	{
		$o_request = &NarcisseServiceFactory::GetCDSsFromLocusTags(
			species         => $species,
			ra_tags         => $ra_loci,
			-o_logger       => $self->_GetLogger(),
			-o_instance_obs => $o_io,
			out_type        => &NS_FASTA
			);
	}
	elsif ( $object eq 'Genomic' )
	{
		throw LipmError::ParameterException( 'GenomicSeq', 'value', -text=> 'Please Ask a region, not a locus tag with GenomicSeq' );
	}
	else
	{
		throw LipmError::ParameterException( 'Protein sequence','value', -text=> 'Request unknown: '.$object )
	}

	my $err_file = $file;
	if ($file =~ /\.fasta$/)
	{
		$err_file =~ s/\.fasta$/\.err/;
	}
	else
	{
		$err_file .= '.err';
	};

    $self->_Trace( "Calling " . ref($o_request) . '->FetchResults' . "\n", '.', 'ISV' );
    $self->GetLogger()->StartTimer('F0');
	my ($ra_header,$rh_data) = $o_request->FetchResults();
    $self->_Trace( "Returned from " . ref($o_request) . '->FetchResults' . "\n", 'F0', 'ISV' );

#	my $header  = $ra_header->[0];
#	my $data    = $rh_data->{$header};

	my $fh_file = new IO::File(">$file") or throw LipmError::IOException( $file, 'f' );
#	print $fh_file "$header\n";
#	print $fh_file &Sequence_utils::FormatSequenceLine(seq => $data);
	foreach my $h (@$ra_header)
	{
		print $fh_file "$h\n";
		my $data = $rh_data -> {$h};
		print $fh_file &Sequence_utils::FormatSequenceLine(seq => $data);
	}
	$fh_file->close();

	my $ra_not_found = $o_request->FetchNotFound();
	my $fh_err_file = new IO::File(">$err_file") or  throw LipmError::IOException( $err_file, 'f' );
	print $fh_err_file @$ra_not_found;
	$fh_err_file->close();

	return;
}

=head2 _GetSequenceRegion

 Title     : _GetSequenceRegion
 Usage     : _GetSequenceRegion($object,$ra_regions, $file)
 Procedure : Get the sequences corresponding to the regions and write them to $file
 Args      :
     $object:    Gene,Protein, cds,genomic
     $ra_regions:An array ref of (species,chrom,start,end)
                 Several regions can be sent,
                 thus the number of elements should be a multiple of 4
     $file:      the output file
     $o_logger   A logger object
 Access    : Protected

=cut

sub _GetSequenceRegion
{
	my $self = shift;
	my( $object, $ra_global_regions, $file, $o_logger ) = @_;

    my $o_io = $self->_Get('_o_io');
    
	# Reformat data: loop through $ra_regions
	# We'll create an $o_request object / species_chrom
	my %h_calls;
	for (my $i=0;$i<$#$ra_global_regions;$i+=4)
	{
		my $species = $ra_global_regions->[$i];
        $species    = $o_io->GetSpeciesShortName($species);
		my $chrom   = $ra_global_regions->[$i+1];
		my $key     = "$species $chrom";

		# Regions specific to this chromosome
		my $ra_specific_regions;
		if (exists $h_calls{$key})
		{
			$ra_specific_regions = $h_calls{$key};
		}
		else
		{
			$ra_specific_regions    = [];
			$h_calls{$key} = $ra_specific_regions;
		}
		
		# Compute the default value for $end if not speficied (0)
		my $end;
		if ( $ra_global_regions->[$i+3] != 0 )
		{
			$end = $ra_global_regions->[$i+3];
		}
		else
		{
			my $chr_len = $o_io->GetSpeciesObserver($species)->GetChromLength( $chrom );
			$end = $chr_len;
		}
		my $rh_region  = { start => $ra_global_regions->[$i+2], end => $end };
		push(@$ra_specific_regions,$rh_region);
	}

    # Open the output file
    my $fh_file = new IO::File(">$file") or throw LipmError::IOException( $file, 'f' );
    
    # Loop through the ($species,$chrom)
    foreach my $key (sort keys %h_calls)
    {
        my $o_request;
        my ($species,$chrom) = split(' ',$key,2);
        my $ra_regions       = $h_calls{$key};

        if ( $object eq 'Gene' )
        {
            $o_request = &NarcisseServiceFactory::GetGenesFromRegions(
                species    => $species,
                chrom      => $chrom,
                ra_regions => $ra_regions,
                -o_logger => $self->_GetLogger(),
                out_type  => &NS_FASTA
                );
        }
        elsif ( $object eq 'Protein' )
        {
            $o_request = &NarcisseServiceFactory::GetProteinsFromRegions(
                species    => $species,
                chrom      => $chrom,
                ra_regions => $ra_regions,
                -o_logger => $self->_GetLogger(),
                out_type  => &NS_FASTA
                );
        }
        elsif ( $object eq 'Cds' )
        {
            $o_request = &NarcisseServiceFactory::GetCDSsFromRegions(
                species    => $species,
                chrom      => $chrom,
                ra_regions => $ra_regions,
                -o_logger => $self->_GetLogger(),
                out_type  => &NS_FASTA
                );
        }
        elsif ( $object eq 'Genomic' )
        {
			$o_request = &NarcisseServiceFactory::GetSlicesFromRegions(
                species    => $species,
                chrom      => $chrom,
                ra_regions => $ra_regions,
                -o_logger => $self->_GetLogger(),
                out_type  => &NS_FASTA
                );
        };
        
        # Launch the request and print the results to the output file
        if ( ! defined $o_request )
        {
			my $regions = Dumper $ra_regions;
			my $msg = "object  = $object\nspecies = $species\nchrom    = $chrom\nregions=$regions\n \$o_request NOT DEFINED";
		    throw LipmError::ObjectException( $o_request, '!build', -text => "$msg" );
		}
		else
		{
	        my ($ra_header,$rh_data) = $o_request->FetchResults();
	        foreach my $header (@$ra_header)
	        {
	            my $data    = $rh_data->{$header};
	            print $fh_file "$header\n";
	            print $fh_file &Sequence_utils::FormatSequenceLine(seq => $data);
	        }
		}
    }
    
    # close the output file
    $fh_file->close();
}

=head2 CalcTxtNames,CalcImageNames,CalcCoordNames

 Title        : CalcTxtNames
 Usage        : my ($txt1,$txt2,...) = $o_io -> CalcTxtNames($path_flg,'prefix1','prefix2',...)
                my $txt1 = $o_io -> CalcTxtNames($o_param,$path_flg,'prefix1')
 Fonction     : Compute the file txt/image/coord names from the parameters
                For each prefix we return 1 file name
 Args         : $path_flg: return the path if 1, the url if 0 
                'prefix1','prefix22',... A list of prefixes for which we want to know the file names
                CalcTxtNames is used to compute the file names in which you store the list of existing features for each request
                The prefix may be Core or Supp
                CalcImageNames is used to compute the image names, CalcCoordNames for the coordinates names
                The prefix is some feature (Synt, Exon, etc.)
                
 Return       : 1/ ('txt1','txt2',...) A corresponding list of names
                2/ txt1 (in scalar context)

=cut
sub CalcTxtNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&TXT_EXT,@_);
}
sub CalcImageNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&IMG_EXT,@_);
}
sub CalcCoordNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&COORD_EXT,@_);
}
sub CalcJsNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&JS_EXT,@_);
}
sub CalcCsvNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&CSV_EXT,@_);
}
sub CalcTsvNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&TSV_EXT,@_);
}
sub CalcCstNames
{
    my $self=shift;
    return $self->__CalcImageCoordNames(&CST_EXT,@_);
}

=head2 __CalcImageCoordNames

 Title        : __CalcImageCoordName
 Usage        : my ($file1,$file2,...) = $self->__CalcCoordImageName($type,$path_flg,'prefix1','prefix2',...)
 Fonction     : Only called by CalcImageNames, CalcCoordNames, CalcTxtNames
 Args         : $type: the extension (.png, .txt, .coord...)
                $path: return the path if 1, the url if 0
                'prefix1','prefix2',... A list of features for which we want to know the file names
 Return       : ('file1','file2',...) A corresponding list of names
                $file1 a string, only one name (scalar context only)

=cut
sub __CalcImageCoordNames
{
	my $self = shift;

	my ($ext,$path_flg,@a_prefix) = @_;
	my $o_param = $self->_Get('_o_param');
	my $o_io    = $self->_Get('_o_io');
	my @a_rvl;

	if( !wantarray and @a_prefix != 1 )
	{
		my $msg = "Only 1 prefix must be specified in scalar context ";
		throw LipmError::ParameterException('a_prefix', 'type', -text => $msg );
	}
	if( $path_flg != 0 and $path_flg != 1 )
	{
		throw LipmError::ParameterException('type', 'value', -should_be => "0,1" );
	}

	return @a_rvl if (@a_prefix==0);

	my $dirname = ($path_flg==1) ? $o_io->GetCfg('ROOT_SESSDIR') : $o_io->GetCfg('HTTP_SESSDIR');
	my $suffix = $self->_GetImagesSuffix();

	foreach my $t (@a_prefix)
	{
		push (@a_rvl,$dirname.'/'.$t.$suffix.$ext);
	}
	return wantarray ? @a_rvl:$a_rvl[0];
}



=head2 _GetImagesSuffix

 Title        : _GetImagesSuffix
 Usage        : my $suffix = $self->__GetImagesSuffix()
 Fonction     : Get the suffix used to save images (depends on paramters)
				This function is used by __CalcImageCoordNames AND by NcDisplayMap,
				thus it is protected
 Args         : none
 Return       : $suffix => a string
 Globals      : none

=cut 
sub _GetImagesSuffix
{
    my $self = shift;

	my $o_param = $self ->_Get('_o_param');

	# Set the suffix :
    my $rh_organisms = $self->_Get('_rh_organisms');
    my $ra_species   = $rh_organisms->{ SPECIES };
    my $ra_chroms    = $rh_organisms->{ CHROMS };
    my $ra_programs  = $rh_organisms->{ PROGRAMS };

	my $suffix = '_';
    $suffix .= $ra_species->[0] .'_' . (defined ($ra_chroms->[0])?$ra_chroms->[0]:'') ;

    # Add targets info
    my $tgt_nb = $self->GetTargetsNumber();
    for ( my $i=1; $i<=$tgt_nb; $i++ )
    {
    	$suffix .= '_'.$ra_species->[$i] . '_' . $ra_programs->[$i];
    }

    # Add other info
	$suffix .= '_'.$o_param->Get('level').'_'.$o_param->Get('view_level');
    $suffix .= '_'.$o_param->Get('range')	if ($o_param->IsDefined('range'));
    #$suffix .= '_'.$o_param->Get('gene') 	if ($o_param->IsDefined('gene'));

	# Remove spaces from url or file names, if any
	$suffix =~ s/ /+/g;

    return $suffix;
}


=head2 GetTargetsNumber

 Title        : GetTargetsNumber
 Usage        : $self->GetTargetsNumber()
 Function     : return the number of targets
 Returns      : the number of targets
 Args         : none

=cut
sub GetTargetsNumber 
{
	my $self = shift;
    return $self->_Get('__tgt_nb');
}

=head2 procedures PrintRelNotes, PrintHelp, PrintCredits

	Title       :	PrintRelnotes
	Usage       :	&PrintRelnotes()
	Prerequiste :
	Procedure   :	1/ Print the list of genomes
	                2/ Print the release notes from 2 history files:
                          -1 file for the instance
                          -1 file for the program
	Returns     :	none
	Args        :	none

=cut

sub PrintRelnotes
{
    my $self = shift;
    
    my $o_io = $self->_Get('_o_io');
    my ($http_logo,$root_cfg,$root_core) = $o_io->GetCfg('HTTP_LOGO','ROOT_CFG','ROOT_CORE');
	print "<div id=\"ID_STANDARD_PAGE\" style=\"background-image: url($http_logo/logo/background.jpg)\" >";

    print "<h2>The genomes contained in this release</h2>";
    $self->__PrintGenomes($o_io);
    $self->__PrintHistory($root_cfg.'/'.'history.list',"<h2>The Data history</h2>");
    $self->__PrintHistory($root_core.'/'.'history.list',"<h2>The Narcisse history</h2>");

    print "</div>\n";
    
}

sub PrintMoreInfo
{
    my $self = shift;
    
    my $o_io = $self->_Get('_o_io');
    my $o_param = $self->_Get('_o_param');
    if ($o_param->IsDefined( 'more' ))
    {
		my $species= $o_param -> Get( 'more' );
		my $o_sp = $o_io -> GetSpeciesObserver( $species );
		my $long = $o_sp -> GetLongName();
		my $summary_file = $o_sp->GetSummaryFile(-check => &FALSE,-where => "INSTANCE");
		print "<h1>$long</h1>";
		if ( -e $summary_file )
		{
			my $o_sp_param = New ParamParser($summary_file);
			if ($o_sp -> IsPseudo())
			{
				my $ra_chroms  = $o_sp -> GetChromsNames();
				
				print "<table class=\"TableRelease\"><tr><th>Chromosome</th><th>Length</th><th>Genes</th><th>Transcripts</th><th>Exons</th><th>Repeats</th><th>RNA</th></tr>";
				my $parity;
				my ($length,$genes,$transcripts,$exons,$repeats,$rnas) = (0,0,0,0,0,0);
				foreach my $chr (@$ra_chroms)
				{
					$o_sp_param -> SelectNameSpace( "CHROMOSOME${chr}_" );
					$length      = &NARCISSE_utils::PrettyDisplayPosition($o_sp -> GetChromLength( $chr));
					$genes       = $o_sp_param -> Get ( "GENE" );
					$transcripts = $o_sp_param -> Get ( "TRANSCRIPT" );
					$exons       = $o_sp_param -> Get ( "EXON" );
					$repeats     = $o_sp_param -> Get ( "REPEAT" );
					$rnas        = $o_sp_param -> Get ( "RNA" );
			        $parity = ($parity eq "odd") ? "even" : "odd";
					print "<tr class=\"$parity\"><td>$chr</td><td>$length</td><td>$genes</td><td>$transcripts</td><td>$exons</td><td>$repeats</td><td>$rnas</td></tr>";
				}
				$parity = ($parity eq "odd") ? "even" : "odd";
				$o_sp_param -> SelectNameSpace( "TOTAL_" );
				$length = &NARCISSE_utils::PrettyDisplayPosition($o_sp_param -> Get ( "LENGTH" ));
				$genes       = $o_sp_param -> Get ( "GENE" );
				$transcripts = $o_sp_param -> Get ( "TRANSCRIPT" );
				$exons       = $o_sp_param -> Get ( "EXON" );
				$repeats     = $o_sp_param -> Get ( "REPEAT" );
				$rnas        = $o_sp_param -> Get ( "RNA" );
				print "<tr class=\"$parity\"><td><strong>TOTAL</strong></td><td>$length</td><td>$genes</td><td>$transcripts</td><td>$exons</td><td>$repeats</td><td>$rnas</td></tr>";
				print "</table>\n";
			}
			else
			{
				$o_sp_param -> SelectNameSpace( "TOTAL_" );
				print "<p>Only " . $o_sp -> GetAssemblyPhase() . " are available for this species</p>";
				print "<table class=\"TableRelease\"><tr><th>Length</th><th>Genes</th><th>Transcripts</th><th>Exons</th><th>Repeats</th><th>RNA</th></tr>";
				print "<tr class=\"odd\"><td>" . &NARCISSE_utils::PrettyDisplayPosition($o_sp_param -> Get ( "LENGTH" )) . "</td>";
				print "<td>" . $o_sp_param -> Get ( "GENE" ) . "</td>";
				print "<td>" . $o_sp_param -> Get ( "TRANSCRIPT" ) . "</td>";
				print "<td>" . $o_sp_param -> Get ( "EXON" ) . "</td>";
				print "<td>" . $o_sp_param -> Get ( "REPEAT" ) . "</td>";
				print "<td>" . $o_sp_param -> Get ( "RNA" ) . "</td>";
				print "</tr></table>";
			}
		}
		else
		{
			print "<p>No summary about this species</p>";
		}
		
		my $readme_file = $o_sp->GetReadmeFile(-check => &FALSE,-where => "INSTANCE");
		if (-r $readme_file)
		{
			print "<h2>$long README:</h2>\n";
			my $fh_readme = new IO::File( $readme_file );
			my @a_html = <$fh_readme>;
			my $html = "";
			foreach my $h (@a_html)
			{
				$h =~ s/[^"'](http:[^ ]+)/<a href="$1">$1<\/a>/g;
				$h =~ s/[^"'](ftp:[^ ]+)/<a href="$1">$1<\/a>/g;
				$html .= $h;
			}
			print $html;
		}
		else
		{
			print "<p>Sorry, there is no README for this species</p>";
		}
	}
}
	
sub PrintCredits
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    $self->__PrintMenuItem(File::Spec->catfile($o_io->GetCfg('ROOT_DOCUMENTATION'),'credits.inc'),$o_io);
}
sub PrintInstallDoc
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    $self->__PrintMenuItem(File::Spec->catfile($o_io->GetCfg('ROOT_DOCUMENTATION'),'install.inc'),$o_io);
}
sub PrintHelp
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    $self->__PrintMenuItem(File::Spec->catfile($o_io->GetCfg('ROOT_DOCUMENTATION'),'help.inc'),$o_io);
}
sub PrintDataFormat
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    $self->__PrintMenuItem(File::Spec->catfile($o_io->GetCfg('ROOT_DOCUMENTATION'),'upload_data_format.inc'),$o_io);
}
sub BackToHome
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    my $o_wb = $self->_Get('_o_wb');

    my $redirect_url = $o_io->GetCfg('HTTP_NARCISSE_CGI') . '?script=HOME&amp;advanced=1&amp;__wb_session=' . $o_wb->GetSession();
    $o_wb->RedirectJs($redirect_url);
}

# Internal functions
=head2 procedure __PrintHistory

	Title     :	__PrintHistory
	Usage     :	$self->__PrintHistory($o_conf)
	Procedure :	Print the program or data history, using the file history.list
                If the file is empty (only the headers), nothing is printed
                If the file is not found, a wild red screen is displayed, but this should NOT happen
	Args      :	$history_file: path to the history.list file
                $title: a title to print just befor the table

=cut
sub __PrintHistory
{
    my $self = shift;
    
    my ($history_file,$title) = @_;
    my $ra_history;
    if (-r $history_file and ! -z $history_file )
    {
        my $o_ndr = New DataReader::Whole();
        $o_ndr->Prepare(-from   => $history_file,-force_reading => &TRUE );
        $ra_history = $o_ndr->FetchallArrayref(-line_type => 'HASH');
        return if (0==@$ra_history);
    }
    else
    {
        &NARCISSE_utils::ShowErrorPage("NOT FOUND !","The release note file is empty or not found");
    };
    
    my $rvl = $title;
    my $last_rel="";
    my $parity="even";
    $rvl .= "<table class=\"TableRelease\"><tr><th>Release</th><th>Date</th><th>Comment</th></tr>";
    for (my $i=$#$ra_history;$i>=0;$i--)
    {
        my $rh_data = $ra_history->[$i];
        my $release = "&nbsp;";
        if ($rh_data->{release} ne $last_rel)
        {
            $last_rel = $rh_data->{release};
            $release  = $last_rel;
            $parity = ($parity eq "odd") ? "even" : "odd";
        }
        
        $rvl .= "<tr class=\"$parity\" ><td class=\"rel\">".$release."</td><td class=\"date\">".$rh_data->{date}."</td><td class=\"com\">".$rh_data->{comment}."</td></tr>";
    }
    $rvl .= "</table>\n";
    print $rvl;
}

=head2 procedures __PrintGenomes, __PrintGenomesSingleInstance, __PrintGenomesMultiInstance

	Title     :	__PrintGenomes, __PrintGenomesSingleInstance, __PrintGenomesMultiInstance
	Usage     :	(1) $self->__PrintGenomes($o_conf)
                    (2) $self->__PrintGenomesSingleInstance($o_io)
                    (3) $self->__PrintGenomesMultiInstance($o_conf)
	Procedures:	(1) call (2) or (3)
                (2) Print the list of genomes for the instance corresponding to $o_io
                (3) For each known instance, call (2)
	Args      :	$o_conf: A narcisse or instance observer

=cut
sub __PrintGenomes
{
    my $self = shift;
    
    my ($o_conf) = @_;
    if ($o_conf->isa( "NarcisseObserver" ))
    {
        $self->__PrintGenomesMultiInstance($o_conf);
    }
    else
    {
        $self->__PrintGenomesSingleInstance($o_conf);
    }
}
 
sub __PrintGenomesMultiInstance
{
    my $self = shift;
    
    my ($o_no) = (@_);
    my $rh_instances = $o_no->GetInstanceObservers();
    foreach my $key (sort keys %$rh_instances)
    {
        my $o_io = $rh_instances->{ $key };
        print "<h3>" . $o_io->GetName() . "</h3>";
        $self->__PrintGenomesSingleInstance($o_io);
    }
}
        
sub __PrintGenomesSingleInstance
{
    my $self = shift;
    
    my ($o_io) = (@_);
    my $o_param = $self -> _Get('_o_param');
    
    my $url_taxonomy = $o_io->GetCfg( 'URL_TAXONOMY' );
    my @a_infos = $o_io->GetSpeciesInfo( $o_io->GetSpeciesShortName( sort $o_io->GetSpeciesLongName() ) );
    
    # Split the species in two groups: the species pseudomolecule-based, the species scaffolds or contigs-based
    my (@a_pseudomolecules, @a_scaffolds);
    foreach my $sp_info (@a_infos)
    {
		my $o_sp = $o_io->GetSpeciesObserver($sp_info->{short});
		if ($o_sp -> IsPseudo())
		{
			push @a_pseudomolecules, $sp_info;
		}
		else
		{
			push @a_scaffolds, $sp_info;
		}
	}
    
    my $nb_pseudo = @a_pseudomolecules;
    my $nb_scaff  = @a_scaffolds;
    
    print "<h4>$nb_pseudo species for which the assembly has reached the pseudomolecules phase</h4>";
    print "<p><em>A pseudomolecule corresponds to a Chromosome, but sometimes a \"special\" chromosome is added, labelled <strong>0</strong> or <strong>Un</strong>";
    print " it contains the remaining sequences which could not be correctly assembled. They are generally separated with ";
    print "a lot (5000 or so) of 'N' letters, which appear in red in the Narcisse display</em></p>";
    my $parity="even";
    my $rvl = "<table class=\"TableRelease\"><tr><th>Genome name</th><th>Taxonomy</th><th>Assembly</th><th>Release</th><th>Release date</th><th>Update date</th><th>Pseudomol.</th><th>Length</th><th>Genes</th><th>RNA</th><th>More...</th></tr>";
    foreach my $sp_info (@a_pseudomolecules)
    {
        my $o_sp = $o_io->GetSpeciesObserver($sp_info->{short});
        #warn Dumper $sp_info;
        #if ($o_param -> Get("UPGRADING") ne "1")
        #{
	    #    next if ($sp_info -> { upgrade } > 0);
		#}
        $parity = ($parity eq "odd") ? "even" : "odd";
        my $taxid=$sp_info->{taxid};
        my $url = $url_taxonomy;
        $url =~ s/~TAXID~/$taxid/;
        my $long = $sp_info->{long};
        my $time_stamp = $o_sp->GetTimestamp();
        my ($sec,$min,$hour,$mday,$mon,$year) = localtime $o_sp->GetTimestamp();
        $mon  += 1;
        $year += 1900;
        my ($chroms,$length,$genes,$rnas,$more)=$self -> __GetSummary($o_sp);
        $rvl  .= "<tr class=\"$parity\" ><td class=\"com\">".$long."</td><td class=\"com\"><a href=\"$url\" class=\"ExtLinks\" title=\"Get taxonomy information about $long\">$taxid</a></td><td class=\"com\">".$sp_info->{assembly}."</td><td>".$sp_info->{release}."</td><td>".$sp_info->{reldate}."</td><td>${year}-${mon}-${mday}</td>";
        $rvl  .= "<td>$chroms</td><td>$length</td><td>$genes</td><td>$rnas</td><td>$more</td></tr>\n";
    }
    $rvl .= "</table>\n";
    print $rvl;
    
    $rvl = "<h4>$nb_scaff species for which the assembly has reached a phase earlier than pseudomolecules</h4>";
    $rvl .= "<p><em>The assembly is less advanced for those species. They are generally compared to completely assembled species only, and ";
    $rvl .= " are never self compared. They are included in Narcisse mainly for annotation purpose</em></p>";
    $parity="even";
    $rvl .= "<table class=\"TableRelease\"><tr><th>Genome name</th><th>Taxonomy</th><th>Assembly phase</th><th>Assembly</th><th>Release</th><th>Release date</th><th>Update date</th><th>Assembly units</th><th>Length</th><th>Genes</th><th>RNA</th><th>More...</th></tr>";
    my $flg_found=&FALSE;
    foreach my $sp_info (@a_scaffolds)
    {
        my $o_sp = $o_io->GetSpeciesObserver($sp_info->{short});
		$flg_found=&TRUE;
        $parity = ($parity eq "odd") ? "even" : "odd";
        my $taxid=$sp_info->{taxid};
        my $url = $url_taxonomy;
        $url =~ s/~TAXID~/$taxid/;
        my $long = $sp_info->{long};
        my $time_stamp = $o_sp->GetTimestamp();
        my ($sec,$min,$hour,$mday,$mon,$year) = localtime $o_sp->GetTimestamp();
        $mon  += 1;
        $year += 1900;
        my ($chroms,$length,$genes,$rnas,$more)=$self -> __GetSummary($o_sp);
		my $assembly_phase = $o_sp -> GetAssemblyPhase();
        $rvl  .= "<tr class=\"$parity\" ><td class=\"com\">".$long."</td><td class=\"com\"><a href=\"$url\" class=\"ExtLinks\" title=\"Get taxonomy information about $long\">$taxid</a></td><td>$assembly_phase</td><td class=\"com\">".$sp_info->{assembly}."</td><td>".$sp_info->{release}."</td><td>".$sp_info->{reldate}."</td><td>${year}-${mon}-${mday}</td>";
        $rvl  .= "<td>$chroms</td><td>$length</td><td>$genes</td><td>$rnas</td><td>$more</td></tr>\n";
    }
    $rvl .= "</table>\n";
    print $rvl if $flg_found;
		
   
}

=head2 __GetSummary

 Title    : __GetSummary
 Usage    : my ($chroms,$genes,$transcripts,$more) = $self->__GetSummary($o_sp);
 Function : Get the number of chroms, of genes, of transcripts read from the summary.cfg file
            No error (but no info) if the summary.cfg file does not exist
 Args     : $o_sp (a species observer)
 Return   : Data read from the summary.cfg file, and a link to get more info, if possible
 
=cut
sub __GetSummary
{
	my $self = shift;
	
	my ($o_sp) =@_;
	my $o_io = $o_sp -> GetInstanceObserver();
	my ($chroms,$length,$genes,$rnas,$more)=('nc','nc','nc','nc','&nbsp;');
	my $summary_file = $o_sp->GetSummaryFile(-check => &FALSE,-where => "INSTANCE");
	if ( -e $summary_file)
	{
		my $ra_names = $o_sp -> GetChromsNames();
		$chroms   = @$ra_names;

		my $o_sp_param = New ParamParser($summary_file);
		$o_sp_param -> SelectNameSpace( 'TOTAL_' );
		
		$length      = &NARCISSE_utils::PrettyDisplayPosition($o_sp_param -> Get ( 'LENGTH' ));
		$genes       = $o_sp_param -> Get ( 'GENE' );
		$rnas        = $o_sp_param -> Get ( 'RNA' );
		
		my $short = $o_sp -> GetShortName();
		my $url = $o_io->GetCfg( 'HTTP_NARCISSE_CGI' ) . '?__wba_popup=1&amp;__wb_function=PrintMoreInfo&amp;more='.$short;
		$more = "<a class=\"MoreInfo\" href=\"$url\">more...</a>";
	}
	return ($chroms,$length,$genes,$rnas,$more);
}	

=head2 PrintMenuItem

 Title    : PrintMenuItem
 Usage    : $self->__PrintMenuItem(tmpl_file)
 Procedure: Print a template
 Args     : The template file to print, $o_io
 
=cut
sub __PrintMenuItem
{
    my $self = shift;
    my ($tmpl,$o_io) = @_;
    
    my $version             = $o_io->GetNarcisseVersion();
    my $short_version       = $version;
    $short_version          =~ s/^(\w+)\.(\w+).*$/$1_$2/;
    my $nottooshort_version = $version;
    $nottooshort_version    =~ s/^(\w+)\.(\w+)\.(\w+).*$/$1.$2.$3/;
    my $page_html = $self->_IncludeFromTmpl($tmpl,version => $version,short_version => $short_version, nottooshort_version => $nottooshort_version);
    print $page_html;
}
1;
