#
# 	emmanuel.courcelle@toulouse.inra.fr
#	Created: december 05, 2008
#	$Id$
#

package NcDisplayMap;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 NAME

 NcDisplayMap.pm

=head1 SYNOPSIS


=head1 DESCRIPTION

 NcDisplayMap derives from NcView, this object implements the 'DisplayMap' representation

=cut

use strict;
use warnings;

use LipmError;
use LipmError::ObjectException;

use Data::Dumper;
use File::Basename;
use Error qw( :try );
use Carp qw(cluck);

use General;
use WebBuilder;
use Runner;
use cgilib::NcScript;
use cgilib::NcScript::NcView;

use NARCISSE_utils;

use Image;
use Textual;
use Planner::ChromosomeMap;
use ChromosomeSliceBigMac;
use ChromosomeSlice::Reference;
use ChromosomeSlice::Target;
use ElementsSet;
use ElementsSet::Synt;
use ElementsSet::Prbh;
use ElementsSet::Hsp;
use MapArea;

use DataReader::Whole;
use NarcisseDataWriter::OnDisk;

eval ("use KnowledgeRemoteReader");

use constant IMG_EXT   => ".png";
use constant COORD_EXT => ".coord";
use constant JS_EXT    => ".js";

use constant SPECIES_NAME_BAND_SIZE => 80;

use base qw( NcView );

use constant CORE_FEATURES => "Core";
use constant SUPP_FEATURES => "Supp";

use constant A_DEFAULT_FEATURES => qw( chrom map synt );   #The features that are always defined (and removed when MODE_SUPP is set)
use constant MODE_CORE     => 0;
use constant MODE_SUPP     => 1;
use constant MODE_NONE     => 2;

=head2 function _Init

 Title        : _Init
 Fonction     : the constructor
                Build the object, setting some default values (only if necessary) to a minimal set of parameters
 Args         : $o_io         : the instance observer
                $o_web_builder: A valid WebBuilder object
                $o_logger     : a Logger

=cut

sub _Init
{
	my $self = shift;

	my ($o_io,$o_web_builder,$o_logger) = (@_);

	my $o_param = $o_web_builder->GetObjectParam();

	# Minimal default values for the parameters, they may be changed in guidef.cfg
    # EST-CE BIEN UTILE ICI ?
	my $rh_guidef = $o_io->GetGuiDefHash();
	$o_param->SetUnlessDefined('level',$rh_guidef->{level}) if (defined $rh_guidef->{level});
	$o_param->SetUnlessDefined('view_level',$rh_guidef->{view_level}) if (defined $rh_guidef->{view_level});
	$o_param->SetUnlessDefined('program',$rh_guidef->{program}) if (defined $rh_guidef->{program});

    # The chroms_reference parameter should be specified for displayMap
    if (! $o_param->IsDefined('chroms_reference'))
    {
        throw NarcisseError( -hd => "NARCISSE ERROR", -bd => "Error in parameters: No chroms_reference specified");
    }

  	#warn Dumper($o_param->{__h_opt});

	$self->SUPER::_Init($o_io,$o_web_builder,$o_logger,&TRUE);

	return;
}

=head2 Procedure _ComputeImages

 Title        : _ComputeImages
 Usage        : $self->_ComputeImage()
 Procedure    : Implement an abstract function of NcView: compute the images
 Args         : none
 Access       : Protected
 Globals      : none

=cut
sub _ComputeImages
{
	my $self = shift;
	my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get('_o_io');
	my $flg_textual = $o_param->GetInteger('Textual');

    # Set __fmode and the features parameter
    if ($o_param->IsDefined('__wb_rest') and $o_param->Get('__wb_rest')==&TRUE)
    {
        $o_param->SetUnlessDefined('features',&CORE_FEATURES);
        if ($o_param->Get('features') eq &CORE_FEATURES)
        {
            $self->_Set('__fmode',&MODE_CORE);
        }
        else
        {
            $self->_Set('__fmode',&MODE_SUPP);
        }
    }
    elsif ( $flg_textual )
    {
		$self->_Set('__fmode',&MODE_CORE);
	}
    elsif ($o_param->IsDefined('__wba_ajax') and $o_param->IsDefined('__wba_index'))
    {
        $self->_Set('__fmode',&MODE_SUPP);
        $o_param->Set('features',&SUPP_FEATURES);
    }
    elsif ($o_param->IsDefined('__wba_ajax'))
    {
        $self->_Set('__fmode',&MODE_CORE);
        $o_param->Set('features',&CORE_FEATURES);
    }
    else
    {
        $self->_Set('__fmode',&MODE_NONE);
        $o_param->Set('features',&CORE_FEATURES);
        return;
    }

    # Initialization of Planner::ChromosomeMap
    # The constructor may modify o_param (typically the range is modified when zooming and scrolling)
    # So we init Chromosome Map NOW, to get the correct parameters while computing the file names
    # THUS, the file names must be computed AFTER this line

    my $o_planner = New Planner::ChromosomeMap(
    					o_io => $o_io,
    					o_param => $o_param,
    					rh_organisms => $self->_Get('_rh_organisms'),
    					-o_logger => $self->GetLogger() );

    # Compute the file names
    my ($file_core,$file_supp) = $self->CalcTxtNames(1,'core','supp');

    # If knowledge_list is defined: Compute the knowledge information and remove the images already computed for the supp data
    if ( $o_param->IsDefined('knowledge_list') and $self->_Get('__fmode') == &MODE_CORE)
    {
        $self->DrawKnowledgeInfos( $o_param );
 		unlink( glob("supp_*") );
        $self->_Trace("Removing supp_*.txt because of knowledge_list\n","*","MV");
   }

	# If OnlineCompare is defined: remove the images already computed for the supp data, forcing a new computation
	if( $o_param->IsDefined('OnlineCompare') and  $self->_Get('__fmode') == &MODE_CORE)
	{
		unlink( glob("core_*") );
		$self->_Trace("Removing core_*.txt because of OnlineCompare\n","*","MV");
	}

	# If Reload pressed: remove the images already computed for the supp data, forcing a new computation
    #                    generally, we want to reload data if something wrong happened, so we compute everything from scratch
	if( $o_param->IsDefined('Reloaded') and  $self->_Get('__fmode') == &MODE_CORE)
	{
		unlink( glob("core_*") );
		$self->_Trace("Removing core_*.txt because of Reloaded\n","*","MV");
	}

	my $file_features = '';
	my $fast_mode = 0;

	if ($self->_Get('__fmode') == &MODE_SUPP)
	{
		$file_features = $file_supp;
	}
	else
	{
		$file_features = $file_core;
	}

	# If the file exists, we assume that the png and coord files exist too and we do not compute them again
	# In Textual mode, we NEVER $fast_mode: people should use this mode only from time to time

	$fast_mode=1 if ($flg_textual==&FALSE && -e $file_features);

	# When you debug the code, it is more convenient to completely avoir $fast_mode
	# You just have to comment out the following line
	# $fast_mode = 0;

	# Fast mode - We just read the features list
	if ($fast_mode==1)
	{
		$self->_Trace("FOUND $file_features - Entering fast mode\n","*","MV");
		my $fh_file_features = new IO::File($file_features)
					or throw LipmError::IOException( $file_features, 'f' );

		# If the file is empty, using an empty string
		my $features = <$fh_file_features> || '';
		$fh_file_features->close();
		chomp $features;
		my @a_features = split(' ',$features);

		$self->_Set('__ra_features',\@a_features);
	}

	# OK - we have to compute all layers
	else
	{
		$self->_Trace("NOT FOUND $file_features - Need to compute everything\n","*","MV");

		#---------------------------------------
		# Initialization of Image :
		#	- The frame transformation acts only in the vertical dimension (the chromosomes)
		#   - chrom and map are two obligate layers
		my $rh_ImageParameters = $o_planner->GetImageParameters();
		my @mandatory_layers = qw( chrom map );

		my $o_image = Image->New( frame_transf_type => "1D",
									width			=> $rh_ImageParameters->{ width },
									height 			=> $rh_ImageParameters->{ height },
									layers			=> \@mandatory_layers
								);
								
		my $o_textual = undef;
		if ($flg_textual==&TRUE)
		{
			$o_textual = New Textual ();
			#
			# Create a layer for each feature
			#
			my $ra_features = $self -> _GetFeatures();
			my @a_file_names = $self -> CalcCsvNames (1,@$ra_features);
			for (my $i=0; $i < @$ra_features; $i++)
			{
				my $name      = $ra_features -> [$i];
				my $file_name = $a_file_names[$i];
				$o_textual -> AddLayer( $name, $file_name );
			}
		}

		#---------------------------------------
		# The reference organism

		my @a_org_attributes = qw( reference chromosome start end );
		my ( $organism, $chromosome, $start, $end ) = $o_planner->_Get( @a_org_attributes );

		my $o_ref_box 		= $o_planner->GetBox();  						# The box (rectangular area) of the reference genome

		my $o_ref = New SpeciesObserver( species => $organism, -o_inst_obs => $o_io );
        my $o_reference = ChromosomeSlice::Reference -> New(
            		#Reference chromosome slice specific attributes
#             		organism 			=> $organism,
            		chromosome			=> $chromosome,
            		start				=> $start,
            		end					=> $end,
            		textual				=> $o_param->GetInteger('Textual'),
            		#Additional technical attributes
#             		o_io                => $o_io,
					o_species           => $o_ref,
            		o_planner			=> $o_planner,
            		o_box 				=> $o_ref_box,
            		o_image             => $o_image,
            		-o_textual          => $o_textual,
            		-o_logger 			=> $self->_GetLogger()
            );

		$o_reference -> Display();

   		#---------------------------------------
        # Foreach target, draw the corresponding synteny slices : Le BigMac
        foreach my $target ( $o_planner->GetTargetsList() )
        {
            $self->_Trace("Loading $target data...\n","*","MV");
			my @attributes = qw( reference chromosome start end level view_level );
			my ( $organism, $chromosome, $start, $end, $level, $view_level ) = $o_planner->_Get( @attributes );
			try
			{
				my $o_bigmac = ChromosomeSliceBigMac -> New(
					#Target chromosome slice BigMac specific attributes
					organism 			=> $organism,
					chromosome			=> $chromosome,
					start				=> $start,
					end					=> $end,
					level				=> $level,
					view_level			=> $view_level,
					target  			=> $target,
					program 			=> $o_planner->GetTargetProgram( $target ),
					#Additional technical attributes
					o_io                => $o_io,
					o_planner 			=> $o_planner,
					o_image  			=> $o_image,
					-o_textual			=> $o_textual,
					-o_logger 			=> $self->_GetLogger()
					);

				$o_bigmac -> Display();
				my $o_target_box = $o_planner->GetBox( $target );			# The box (rectangular area) of the target genome

				if ( $o_target_box->IsReferenceNeighbor() == 1 && $o_planner->IncludePrbhHsp())
				{
					# if some files are missing, an exception is thrown
					# but to allow the code to continue with other files
					# i add a new try catch bloc in a loop.
					foreach my $method ( qw( DisplayPrbh DisplayHsp ) )
					{
						try
						{
							$o_bigmac->$method( reference_box => $o_ref_box, target_box => $o_target_box );
						}
						catch LipmError::IOException with
						{
							my $o_err = shift;

							my $msg = $o_err->Explain() . "\n";
							$self->_Trace( $msg, '.', 'MV' );
						};
					}
				}
			}
			catch LipmError::IOException with
			{
				my $o_err = shift;

				my $msg = $o_err->Explain() . "\n";
				$self->_Trace( $msg, '.', 'MV' );
			};


        }

		my $ra_features;
		if ($flg_textual == &FALSE)
		{
	        # Save map information for all elements (.coord files)
	        my $feature_mode = $self->_Get('__fmode' );
	        if ( $feature_mode eq &MODE_CORE )
	        {
	        	$o_planner->DumpMapInformation( $self->CalcCoordNames(1,'map'));
	        }
	
	        # Print each feature in separate images and keep track of the features printed
	        $ra_features=$self->PrintImages( $o_image );
	        $self->_Set('__ra_features',$ra_features);

			# this is useful for the cache mechanism
	        my $fh_file_features = new IO::File(">$file_features") or throw LipmError::IOException( $file_features, 'f' );
	        print $fh_file_features join(' ',@$ra_features);
	        $fh_file_features->close();

		}
		else
		{
			$ra_features = $self -> _GetFeatures();    # TODO: bizarre non ?
			$o_textual -> DeleteAll();
			$self -> __MergeFeatureFiles();
		}
	}

	return;
}

=head2 Procedure __MergeFeatureFiles
 Title        : __MergeFeatureFiles
 Usage        : $self->__MergeFeatureFiles()
 Procedure    : Merge the textual files produced by _ComputeImages and corresponding to the objects related to the chromosomes.
                                       
 Args         : None

=cut
sub __MergeFeatureFiles
{
	my $self = shift;
	my $kingdom = $self -> { _o_io } -> GetCfg( 'KINGDOM' );
	
	my @a_features = ( $kingdom eq 'bacteria' ) ? qw( cyto gene rna repeat ) : qw( cyto gene exon rna repeat );
	my @a_files = $self -> CalcCsvNames(1,@a_features);
	
	my @a_headers = qw( species chromosome type locus_tag symbol element parent color start end strand );

	my ($obj_file) = $self -> CalcCsvNames(1,'objects');
	#my $o_ndw = New NarcisseDataWriter::OnDisk($obj_file, \@a_headers);
	
	my $o_ndw = New NarcisseDataWriter::OnDisk($obj_file, 
	                                           \@a_headers, 
	                                           {sort_numerical => { species => 0, chromosome => 1, start => 1, end => 1 } },
	                                           [ 'species','chromosome','start','end' ]);
	
	for (my $i=0; $i < @a_files; $i++)
	{
		my $file = $a_files    [$i];
		my $type = $a_features [$i];
		my $o_ndr = New DataReader::Whole();
		next if ( ! -r $file );
		$o_ndr -> Prepare ( -from => $file );
		my $ra_rows = $o_ndr -> FetchallArrayref(-line_type => 'HASH');
		foreach my $row ( @$ra_rows )
		{
			foreach my $hdr (@a_headers)
			{
				$row -> { $hdr } = '' unless (defined $row -> { $hdr });
			}
			$row -> { type } = $type;
		}
		$o_ndw -> Update ( $ra_rows );
	}
	$o_ndw = undef;
}

=head2 Procedure PrintImages

 Title        : PrintImages
 Usage        : $self->PrintImages($o_image)
 Procedure    : 
 Args         : 

=cut
sub PrintImages
{
	my $self      = shift;
	my ($o_image) = @_;
	my $o_io      = $self->_Get('_o_io');
	my $o_param   = $self->_Get( 'o_param' );

	my $root   = $o_io->GetCfg('ROOT_SESSDIR').'/';
	my $suffix = $self->_GetImagesSuffix();

	my $feature_mode = $self->_Get('__fmode' );

	# When the MODE_SUPP (or MODE_NONE) is requested none of the A_DEFAULT_FEATURES should be included
	if ( $feature_mode eq &MODE_SUPP || $feature_mode eq &MODE_NONE)
	{
		$o_image->DeleteLayers( (A_DEFAULT_FEATURES) );
	}

    my @a_files = $o_image->PrintLayers($root,$suffix);
    my @a_features = $o_image->GetLayersList();

    my @a_displayed_features = @a_features;

    #$self->{ _ra_features } = \@a_displayed_features;

    return \@a_features;
}

=head2 function _GetMainImages

 Title        : _GetMainImages
 Usage        : $html .= $self->_GetMainImages()
 Fonction     : generate the html code to read the main images
                Implement an abstract function from NcView
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetMainImages
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    my $o_param =$self->_Get('_o_param');

    my $html = '';

    my $ra_features = $self->_Get('__ra_features');
    if (defined $ra_features)
    {
        my $msg = "Features " . join(' ',@$ra_features) . "\n";
        $self->_Trace( $msg, '.', 'MVV' );
    }
    else
    {
        $self->_Trace( "Features NONE\n", '.', 'MVV' );
    }

    # This mode is used when we display the whole page (ie it is NOT an ajax request)
    # We don't display anything, as the images will be displayed by the next Ajax requests
    if ($self->_Get('__fmode') eq &MODE_NONE)
    {
        $o_param->Set('__wba_start',1);
        $o_param->Set('__wba_ajax',2);
        $html .= "<div class=\"PartialDisplay\"></div><div class=\"PartialDisplay\"></div><div class=\"PartialDisplay\"></div>\n";

        my $max_key = $o_io->GetCfg('DISPLAY_MAX_ORG');
        for (my $i=1; $i<=$max_key; $i++)
        {
            $o_param->SetUnlessDefined("target$i",'none');
        }
    }

    # the core features: the html looks like:
    # <div class="PartialDisplay">
    #    <img ...> (the images of the chrom layer and the core features)
    # </div>
    # <div class="PartialDisplay">
    #    &nbsp; (reserved for the supplementary features)
    # </div>
    # <div>
    #    The maps at top of layers
    # </div>
    elsif ($self->_Get('__fmode') eq &MODE_CORE)
    {
        my $html_map = "";
        $html .= "<div class=\"PartialDisplay\">";
        foreach my $ftr (@$ra_features)
        {
            if ($ftr eq 'map')
            {
                $html_map .= $self->__GetFeatureImage($ftr);
            }
            else
            {
                $html .= $self->__GetFeatureImage($ftr)
            }
        }
        $html .= "</div><div class=\"PartialDisplay\">&nbsp;</div>";
        $html .= "<div class=\"PartialDisplay\">$html_map</div>";

        # Program the partial request
        $o_param->Set('__wba_start',1);
        $o_param->Set('__wba_ajax',2);
        #$o_param->Set('__wba_index',2);
        $o_param->Delete('features');

        my $max_key = $o_io->GetCfg('DISPLAY_MAX_ORG');
        for (my $i=1; $i<=$max_key; $i++)
        {
            $o_param->SetUnlessDefined("target$i",'none');
        }
    }

    # the supp features: the html looks like:
    #    <img ...> (the images of the supplementary features)
    else
    {
        foreach my $ftr (@$ra_features)
        {
            $html .= $self->__GetFeatureImage($ftr)
        }
    }
    return $html;
}

=head2 function _GetMainText

 Title        : _GetMainText
 Usage        : $html .= $self->_GetMainText()
 Fonction     : generate the html code to read the textual representation of the images
                Implement an abstract function from NcView
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetMainText
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    my $o_param =$self->_Get('_o_param');

    my $html = '';
	$html .= "<h2>Download data corresponding to this Narcisse view</h2>";

	# The following is useful at debug time	
	if (0)
	{
		my $ra_features = $self->_GetFeatures();
		my @a_url   = $self -> CalcCsvNames(0,@$ra_features);
		my @a_files = $self -> CalcCsvNames(1,@$ra_features);
	    $html .= '<ul>';
		for (my $i=0; $i< @$ra_features; $i++)
		{
			my $url = $a_url[$i];
			my $ftr = $ra_features -> [$i];
			my $file= $a_files[$i];
			if ( ! -z $file )
			{
				$html .= "<li><a href=\"$url\">$ftr</a></li>";
			}
		}
		$html .= '</ul>';
	}
	else
	{
		my ($objects, $synt, $prbh)       = $self -> CalcCsvNames(0,'objects','synt','prbh');
		my ($f_objects, $f_synt, $f_prbh) = $self -> CalcCsvNames(1,'objects','synt','prbh');
		$html .= "<ul>";
		$html .= "<li> <a href=\"$objects\">Biological objects</a> corresponding to the displayed species</li>\n" unless (-z $f_objects);
		$html .= "<li> <a href=\"$synt\">Syntenies</a></li>\n" unless (-z $f_synt);
		$html .= "<li> <a href=\"$prbh\">Gene homologies</a></li>" unless (-z $f_prbh);
		$html .= "</ul>\n";
	}
	
	return $html;
}

=head2 function _GetMainImagesJson

 Title        : _GetMainImagesJson
 Usage        : $html .= $self->_GetMainImagesFson()
 Fonction     : generate the json code with urls to the main images
                Implement an abstract function from NcView
 Returns      : Json code
 Args         : none
 Globals      : none

=cut
sub _GetMainImagesJson
{
    my $self = shift;
    my $o_io = $self->_Get('_o_io');
    my $o_param =$self->_Get('_o_param');

    my %h_urls;

    my $ra_features = $self->_Get('__ra_features');

    my $max_key = $o_io->GetCfg('DISPLAY_MAX_ORG');
    for (my $i=1; $i<=$max_key; $i++)
    {
        $o_param->SetUnlessDefined("target$i",'none');
    }

    foreach my $ftr (@$ra_features)
    {
        # TODO - Gerer les maps en JSON
        next if $ftr eq 'map';
        my $url = $self->__GetFeatureImage($ftr);
        if ($url ne '')
        {
            $h_urls{$ftr} = $url;
        }
    }
    return $self->_ToJson(\%h_urls);
}

=head2 function __GetFeatureImage

 Title        : __GetFeatureImage
 Usage        : $html .= $self->__GetFeatureImage()
 Fonction     : generate the html code useful to read the image for one feature
                Called by _GetMainImages
                In rest mode, only return the url of the image
 Returns      : html code
 Args         : $ftr The feature name
 Globals      : none

=cut
sub __GetFeatureImage
{
    my $self = shift;
    my ($ftr) = (@_);
    my $o_param = $self->_Get('_o_param');
    my $o_io    = $self->_Get('_o_io');

    my ($img)   = $self->CalcImageNames(0,$ftr);
    my ($img_p) = $self->CalcImageNames(1,$ftr);

    if ($self->_Get('_rest') == &TRUE)
    {
        return (-e $img_p) ? $img : '';
    }

    my $id = 'ID_'.uc($ftr);

    my $html = '';
    if (! -e $img_p)
    {
        $html = "&nbsp;";
    }
    else
    {

        # Feature 'map': we download the markers, the img, the areas
        # NOTE - There are several images stacked one upon the other, BUT there is  ONLY ONE map, connected to
        #        the (top empty) image. If we manage one map / image, ONLY the top-level maps will be available

        # Retrieving the "marker" blocks - the markers are AFTER the images in the html stream,
        # so that they will be displayed ON the images, BUT they are BEFORE the 'maps' image area tags,
        # to be sure the events are captured
        if ($ftr eq 'map')
        {
            $html .= "<div id=\"ID_DYNAMIC_MARKER\" class=\"MapsImg\" >&nbsp;</div>";
            $html .= "<div id=\"ID_DYNAMIC_MARKER2\" class=\"MapsImg\" >&nbsp;</div>";


            my ($map_coord) = $self->CalcCoordNames(1,$ftr);
            $html .= "<img alt=\"map\" class=\"MapsImg\" id=\"$id\" src=\"$img\" usemap=\"#ID_MAP_MAP\" />\n";

            # Write the areas: we have one area/line in the .coord file
            $html .= "<map name=\"ID_MAP_MAP\" id=\"ID_MAP_MAP\" >\n";
            my $fh_map_coord = new IO::File($map_coord) or throw LipmError::IOException($map_coord,'f');

            #my %h_maps;
            my @a_maps;
            while (my $line=<$fh_map_coord>)
            {
                chomp($line);
                my @f = split(/\t/,$line);

                my $coord = join(',',@f[0..3]);
                my $obj   = $f[4];
                $obj =~s/"//g;
                $obj = 'O_'.$obj;
                my $params= join(',',@f[0..3]).','.join(',',@f[5..$#f]);
                $html .= "<area alt=\"click here\" shape=\"rect\" href=\"#\" coords=\"$coord\" onmouseover='javascript:$obj.MouseOver($params)' onmousedown='javascript:$obj.MouseDown($params)' />\n";
				#$h_maps{$id++} = \@f if ($f[4] eq 'GENE');
                #$h_maps{$id++} = \@f;
                push(@a_maps,\@f);
            }

            # Finished for the areas
            $html .= "</map>\n";

            # Print the coords in JSON format
##            my $json     = new JSON( );
##            my $map_json = $json->encode(\@a_maps);
##            my ($map_js) = $self->CalcJsNames(1,$ftr);
##            my $fh_map_js = new IO::File ($map_js,'w') or throw LipmError::IOException($map_js,'f');
##            print $fh_map_js $map_json;
##            $fh_map_js->close();
       }

        # Other feature: just download the image
        else
        {
            # If $img_p is a symlink, we use the --REAL-- file instead
            if (-l $img_p)
            {
                my $dst = readlink $img_p;
                my $img_dir = dirname $img;
                my $img_file= basename $dst;
                $img = $img_dir.'/'.$img_file;
            }
            $html =  "<img alt=\"$ftr feature\" class=\"MapsImg\" id=\"$id\" src=\"$img\" />\n";
        }
    }
    return $html;
}

=head2 function _GetPrintableImage

 Title        : _GetPrintableImage
 Usage        : $rvl .= $self->_GetPrintableImage()
 Fonction     : generate the html code useful to read the single image
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetPrintableImage
{
    my $self = shift;

	my $o_param   = $self->_Get('_o_param');
 	my $http_printable = $self->__ConstructPrintableImage();
 	my $html ="<img src='".$http_printable."' alt='The printable image' />";

 	return $html;
}

=head2 __ConstructPrintableImage

 Title     : __ConstructPrintableImage
 Usage     : &ConstructPrintableImage()
 Procedure : This function constructs from existing image files in the dir session an image for printing purpose
 Args      : none
 Errors	   : IOException
 Globals   : CFG

=cut
sub __ConstructPrintableImage
{
    my $self = shift;

    my $o_io    = $self->_Get('_o_io');
    my $o_param = $self->_Get('_o_param');

    my $file_core = $self->CalcTxtNames(1,'core');
	my $image_name = "";

    my $rh_organisms = $self->_Get('_rh_organisms');
    my $ra_species   = $rh_organisms->{ SPECIES };
    my $ra_chroms    = $rh_organisms->{ CHROMS };
    my $ra_programs  = $rh_organisms->{ PROGRAMS };
	my $reference                            = $ra_species->[0];
    my $chroms_reference                     = $ra_chroms->[0];

	#This is necessary to deal with the boxes
	my $o_planner = Planner::ChromosomeMap -> New(
			o_io         => $o_io,
			o_param      => $o_param,
			rh_organisms => $self->_Get('_rh_organisms'),
			-o_logger    => $self->GetLogger()
		);

	if (-f $file_core)
	{
		my $root   = $o_io->GetCfg('ROOT_SESSDIR');
		my $suffix = $self->_GetImagesSuffix();
		my $chrom_layer = 'chrom';
		my $chrom_image_file = File::Spec->catfile($root,$chrom_layer.$suffix.'.png');

		#Some image attributes definition
		# # Initialization of graphics parameters :
    	Image->InitColorsLibrary($o_io->GetCfg('ROOT_COLORS_CFG'));
    	Image->InitFontLibrary($o_io->GetCfg('ROOT_FONTS_CFG'),$o_io->GetCfg('ROOT_FONTS'));

		# 1/ Make a new Image object using the first layer.
		# This is just a trick to grab the image bounds

		my $o_temp_image = Image->New($chrom_image_file);
		my ($width,$height)= $o_temp_image->getBounds();

		#Adding an horizontal band for the species names
		$height+=&SPECIES_NAME_BAND_SIZE;
		my $o_image =  Image->New(width=> $width, height=> $height);

		$o_image->AddLayer('chrom',	$chrom_image_file);

		# 2/ Add each checked feature image
		my $ra_features = $self->_GetCheckedFeatures();
		foreach my $feature (@$ra_features)
		{
			$o_image->AddLayer($feature,$root.'/'.$feature.$suffix.'.png');
		}

		# 3/ Print the image.
		$image_name=$o_io->GetCfg('HTTP_SESSDIR').'/'.$o_io->GetCfg('SAVED_NARCISSE_IMG_FILE').".png";
		my $image_file_name = File::Spec->catfile($o_io->GetCfg('ROOT_SESSDIR'),$o_io->GetCfg('SAVED_NARCISSE_IMG_FILE').".png");
		$o_image->PrintImage( $image_file_name,offset_y => SPECIES_NAME_BAND_SIZE);

		#Now upload again the image and print the species name
		my $o_printable_image = Image->New($image_file_name);
		my $dark_blue = $o_printable_image->colorAllocate(0,0,120);
		my $dark_green = $o_printable_image->colorAllocate(0,120,0);

		#Display Organism names
		&_DisplayName($o_io,$o_planner,$o_printable_image,$reference,$dark_blue,14);
		&_DisplayRefChrom($o_planner,$o_printable_image,$reference,$chroms_reference,$dark_blue);
		foreach my $target ($o_planner->GetTargetsList())
		{
			&_DisplayName($o_io,$o_planner,$o_printable_image,$target,$dark_blue,12);
		}

		$o_printable_image->PrintImage( $image_file_name );

	}

	return $image_name;

=head2 Procedure _DisplayName

 Title        : _DisplayName
 Usage        : _DisplayName($o_io,$o_planner,$o_image,$label,$color,$font_size)
 Procedure    : Displays the species Name
 Args         :

=cut

	sub _DisplayName
	{
        my ($o_io,$o_planner,$o_image,$target_label,$color,$font_size ) = (@_);

		my $o_box = $o_planner->GetBox($target_label);
		my $target_name = $o_planner->GetTargetOrganism($target_label);
		my $name=	$o_io->GetSpeciesLongName($target_name);
		$name =~s/\_/ /;
		my $x_text=$o_box->CenterX()-length($name)*5;
		my $y_text=25;

		$o_image->stringFT($color,$o_image -> font("georgiai"),$font_size,0,$x_text,$y_text, $name);

	}

=head2 Procedure _DisplayRefChrom

 Title        : _DisplayName
 Usage        : _DisplayName($o_planner, $o_image,$org )
 Procedure    : Displays the species Name
 Args         :

=cut

	sub _DisplayRefChrom
	{
		my $o_planner = shift;
		my $o_image = shift;
		my $org = shift;
		my $chrom_name = shift;
		my $color =shift;

		my $o_box = $o_planner->GetBox($org);
		my $text = "chr $chrom_name";
		my $x_text=$o_box->CenterX()-length($text)*5;
		my $y_text=43;
		$o_image->stringFT($color,$o_image -> font("georgia"),12,0,$x_text,$y_text,$text);

	}
}

=head2 function _GetMainTitle

 Title        : _GetMainTitle
 Usage        : $rvl .= $self->__GetImagetitle()
 Fonction     : generate a little table, to be displayed at top of images, containing the title
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetMainTitle
{
    my $self = shift;
    return "<div id=\"ID_TITLE\"></div>";
}


=head2 Procedure __PrintFile

 Usage        : $o_x->__PrintFile( $dir_path, $chrom,
                             $rh_data, $extension );
 Prerequisite : none
 Procedure    : erase previous file then print a new one
 Args         : $dir_path, $chrom, $rh_data, $extension

=cut

sub __PrintFile
{
	my $self = shift;
	my ( $dir_path, $chrom, $rh_data, $extension ) = @_;

	# erase previous data
	my $glob_files = File::Spec->catfile( $dir_path, '*.' . $extension );
	my $n = unlink glob( $glob_files );
	my $msg = "$n files ($extension) have been erased into $dir_path\n";
	$self->_Trace( $msg, '.', 'MVV' );


    my $o_io      = $self->_Get('_o_io');
	my $filename  = $o_io->GetCfg('MASK_FILE_CHROM'). "$chrom.$extension";
	my $file_path = File::Spec->catfile( $dir_path, $filename );

	my @a_cols    = qw( chrom start end value opt_circos id );
	my $rh_options= { o_logger => $self->GetLogger() };
	&NARCISSE_utils::PrintNarcisseData2( $file_path, \@a_cols, $rh_data, $rh_options );

	return;
}


=head2 function __SplitDataOnActivation

 Usage        : ($rh_activ, $rh_repress) = $o_x->__SplitDataOnActivation( $rh_data );
 Prerequisite : none
 Function     : puts all data with 'activ*' into $rh_activ
                 and all with 'repress*' into $rh_repress
 Returns      : $rh_activ, $rh_repress, hash ref, usual narcisse hash ref
 Args         : $rh_data, hash ref, usual narcisse hash ref

=cut

sub __SplitDataOnActivation
{
	my $self = shift;
	my ( $rh_data ) = @_;

	my ( %h_activ, %h_repress );
	my ( $i_act, $i_rep ) = ( 1, 1 );

	foreach my $key ( sort keys %$rh_data )
	{
		my $rh_row = $$rh_data{ $key };
		my $action = $$rh_row{ action };
		$h_activ{ $i_act++ }   = $rh_row	if( $action =~ /activ/ );
		$h_repress{ $i_rep++ } = $rh_row	if( $action =~ /repress/ );
	}

	return \%h_activ, \%h_repress;
}


=head2 function __GetKnowledgeData

 Usage        : $rh_data = $o_x->__GetKnowledgeData( $o_know, $ra_records,
                                                $species, $chrom );
 Prerequisite : YUI returned indices of a list staring at 0.
                 Our record starts at line 1.
 Function     : gets all wanted information from $o_know object
                 and builds a hash ref ready to use or to be printed out.
 Returns      : $rh_data, usual narcisse hash ref
                    keys are line number.
 Args         : $o_know, a KnowledgeRemoteReader object
                 $ra_records, array ref on records' line number.
                 $o_so, A species observer
                 $chrom, string, chrom label
 Globals      : none

=cut

sub __GetKnowledgeData
{
	my $self = shift;
	my ( $o_know, $ra_records, $o_so, $chrom ) = @_;

	for( my $i=0; $i < @$ra_records; ++$i )
	{	++$$ra_records[$i];	}

	# obtenir les valeurs et les locus des enregistrements concernes
	my $ra_val    = $o_know->GetValues( @$ra_records );
	my $ra_tags   = $o_know->GetIds( @$ra_records );
	my $ra_action = $o_know->Get( 'relationship', @$ra_records );

	# pour chaque locus obtenir ses bornes
	my $rh_range = &GetRange( $o_so, $ra_tags );

	# construction du hash de donnees
	my $chr      = $o_so->GetShortName() . '-' . $chrom;

	my %h_data = ();
	for( my $i=0; $i<@$ra_records; ++$i )
	{	# $i+1 pour que la premiere ligne ne soit pas 0
		my %h_row = ( chrom => $chr, opt_circos => '=', id => $i+1 );

		my $tag          = $$ra_tags[ $i ];
		$h_row{ start }  = $$rh_range{ $tag }{ start };
		$h_row{ end   }  = $$rh_range{ $tag }{ end   };
		$h_row{ value }  = $$ra_val[ $i ];
		$h_row{ action } = $$ra_action[ $i ];

		$h_data{ $i+1 }  = \%h_row;
	}

	return \%h_data;
}

=head2 function GetRange

  Title        : GetRange
  Usage        : $rh_range = &GetRange( $species, $ra_tags );
  Prerequisite : use API
  Function     : given a list of locus tags, will return a hashref with foreach
                 the start and the end position.
  Returns      : $rh_range, hashref of hash ref, keys are tags
                     { tag => { start => x, end => x }, ... }
  Args         : $o_so A built species observer
                 $ra_tags, array ref of locus tags
  Globals      : none

=cut

sub GetRange
{
	my ( $o_so, $ra_tags ) = @_;

    my $o_dba = $o_so->GetDBAdaptor();
	my $o_gene_a = $o_dba->get_GeneAdaptor();

	my %h_return = ();

	foreach my $tag ( @$ra_tags )
	{
		my $o_gene = $o_gene_a->fetch_by_stable_id( $tag );
		$h_return{ $tag } = { start => $o_gene->start(),
		                      end   => $o_gene->end() };
	}

	return \%h_return;
}

=head2 Procedure DrawKnowledgeInfos

 Usage        : $self->__DrawKnowledgeInfos( $o_param, $o_logger );
 Prerequisite : It does not draw data !
 Procedure    : Read the Knowledge info into knowledge.list file,
                get activation/repression data.
                print files into the session directory so that
                these data will be drawn with "supplementary" objects.
 Args         : $o_param, ParamParser object.
                $o_logger, Logger object.
 Globals      : none

=cut

sub DrawKnowledgeInfos
{
	my $self = shift;
	my ( $o_param ) = @_;

	my ( $species, $list, $chrom ) =
	                       $o_param->Get( qw( species knowledge_list chrom ) );

	my $msg = "DrawKnowledgeInfos ( species : $species, list : $list )\n";
	$self->_Trace( $msg, '.', 'MVV' );

	return	if( '' eq $list );

	# lire le fichier de donnees
    my $o_io      = $self->_Get( '_o_io' );
	my %h_args    = ( o_io => $o_io, species => $species,
						-cache => 'knowledge.list', -o_logger => $self->GetLogger() );

	my $o_know    = New KnowledgeRemoteReader( %h_args );
	my @a_records = split( '_', $list );
    my $o_species = $o_io->GetSpeciesObserver($species);
	my $rh_data   = $self->__GetKnowledgeData( $o_know, \@a_records, $o_species, $chrom );

	my( $rh_activ, $rh_repress ) = $self->__SplitDataOnActivation( $rh_data );

	# print data into a convenient session directory
	my $supp_path = $o_species->GetSuppDataPath( 'SESSION' );
	my $data_dir  = $o_io->GetCfg( 'KNOWLEDGE_DIR' );
	my $dir_path  = File::Spec->catdir( $supp_path, 'histograms', $data_dir );

	# attention : histogram est ecrit en dur
	# ce nom signifie que c'est dans un objet ElementsSet::Histogram qu'il sera represente
	# si cette classe change de nom un jour, les fichiers generes ne seront plus exploites.


# 	my @a_dirs    = ( $species, qw( supp_histograms knowledge ) );
# 	my $dir_path  = File::Spec->catdir( @a_dirs );
	my @a_created = File::Path->mkpath( $dir_path );
	foreach my $dir ( @a_created )
	{
		my $msg = "directory $dir successfully created\n";
		$self->_Trace( $msg, '.', 'MVV', 'dismap' );
	}

	$self->__PrintFile( $dir_path, $chrom, $rh_activ, 'activ' );
	$self->__PrintFile( $dir_path, $chrom, $rh_repress, 'repress' );

	return;
}


=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
