#
# 	emmanuel.courcelle@toulouse.inra.fr
#	Created: december 05, 2008
#	$Id: WebBuilderFactory.pm 2317 2009-01-22 14:23:43Z sletort $
#

package NcDotplot;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 NAME

 NcDotplot.pm

=head1 SYNOPSIS


=head1 DESCRIPTION

 Ncdotplot derives from NCView, , this object implements the 'Dotplot' representation
 
=cut

use strict;
use warnings;

use LipmError;
use LipmError::ObjectException;

use Data::Dumper;
use File::Basename;
use Error qw( :try );
use Carp;
use File::Copy;
 
use General;
use WebBuilder;
use Runner;
use cgilib::NcScript;
use cgilib::NcScript::NcView;

use NARCISSE_utils;

use Image;
use Planner;
use Dotplot;
use Planner::GenomeMap;

use constant IMG_EXT   => ".png";
use constant COORD_EXT => ".coord";
use constant JS_EXT    => ".js";

use constant WITH_PERSONAL_DATA  => 'perso';
use constant PRINTABLE           => 'printable';
use constant DEFAULT 			 => 'default';

use base qw( NcView );


=head2 function _Init

 Title        : _Init
 Fonction     : the constructor
                The Dotplot object is built, and if the drawing is not in the cache, it is computed
 Args         : $o_io         : the instance observer
                $o_web_builder: A valid WebBuilder object
                $o_logger     : a Logger

=cut
sub _Init
{
	my $self = shift;
	my( $o_io,$o_web_builder,$o_logger ) = @_;

	my $o_param = $o_web_builder->GetObjectParam();

	# Minimal default values for the parameters, they may be changed in guidef.cfg
	$o_param->SetUnlessDefined('level',$o_io->GetCfg('GUI_DEFAULTS')->{level}) if (defined $o_io->GetCfg('GUI_DEFAULTS')->{level});
	$o_param->SetUnlessDefined('program',$o_io->GetCfg('GUI_DEFAULTS')->{program}) if (defined $o_io->GetCfg('GUI_DEFAULTS')->{program});

	$self->SUPER::_Init($o_io,$o_web_builder,$o_logger,&TRUE);

    # The target/target1 parameter should be specified for the dotplot
    my $ra_species = $self->_Get('_rh_organisms')->{SPECIES};
    if ($self->GetTargetsNumber()==0)
    {
        throw NarcisseError( -hd => "NARCISSE ERROR", -bd => "Error in parameters: No target1 specified");
    };
    my $o_species = New SpeciesObserver( species => $ra_species->[0], -o_inst_obs => $o_io );
	$self->_Set( __o_species => $o_species );

	return;
}


=head2 Procedure _ComputeImages

 Title        : _ComputeImages
 Usage        : $self->_ComputeImage() - Called by the _Init from NcView
 Fonction     : Compute the images
 Args         : none
 Access       : Protected
 Globals      : none

=cut
sub _ComputeImages
{
	my $self = shift;

	my $o_io    = $self->_Get('_o_io');
	my $o_param = $self->_Get('_o_param');

	# Draw the dotplot, or retrieve the drawing from the cache
	my $rh_organisms = $self->_Get('_rh_organisms');
	my $ra_species   = $rh_organisms->{ SPECIES };
	my $ra_chroms    = $rh_organisms->{ CHROMS };
	my $ra_programs  = $rh_organisms->{ PROGRAMS };

	my ($ra_reference_chroms,$ra_target_chroms)=$self->__GetDotPlotChroms();
	
	my $mode = ( $self->__ExistingPersonnalData() ) ? &WITH_PERSONAL_DATA : &DEFAULT;
	
	my $o_dotplot=Dotplot -> New(
		o_io        => $o_io,
		reference	=> $ra_species->[0],
		target		=> $ra_species->[1],
		program		=> $ra_programs->[1],
		level		=> $o_param->Get('level'),
		ref_chroms 	=> $ra_reference_chroms,
		target_chroms => $ra_target_chroms,
		-mode        => $mode
		);

	$self->_Set('__o_dotplot',$o_dotplot);

	if ( $o_dotplot->DotPlotsAvailable() == 1 ) # we just take them from cache 
	{
		$self->_Trace("Reading dotplots from cache\n","*","D");
	}
	else   # we have to compute them, they will be kept in cache dir
	{
		$self->_Trace("Creating a new dotplot\n","*","D","dotplot ");
		$o_dotplot->ComputeDotPlots();
	}

	my $file_name = $self->CalcImageNames(1,'dotplot');
	my $url_name  = $self->CalcImageNames(0,'dotplot');
	my $coord_name= $self->CalcCoordNames(1,'dotplot');

	return;
}

sub __ExistingPersonnalData
{
	my $self = shift;
	
	my $o_species = $self->_Get( '__o_species' );
	my @a_perso_files = $o_species->GetPersoFiles( -existing_files => &TRUE );
		
	my $bool_exist_perso_file = (scalar @a_perso_files) ? 1 : 0;
	
	return $bool_exist_perso_file;
}


=head2 procedure __GetDotPlotChroms

 Title        : __GetDotPlotChroms
 Usage        : $self->__GetDotPlotChroms()
 Function     : Creates the dotplot object as a hash
 Returns      : a reference on the hash
 Args         : a ParamParser object
 Globals      : none

=cut
sub __GetDotPlotChroms
{
    my $self = shift;
    
	my $o_param   = $self->_Get('_o_param');
	my @a_ref_chroms=();
	my @a_target_chroms=();
	
    my $rh_organisms = $self->_Get('_rh_organisms');
    my $ra_species   = $rh_organisms->{ SPECIES };
    my $ra_chroms    = $rh_organisms->{ CHROMS };
    my $ra_programs  = $rh_organisms->{ PROGRAMS };

    @a_ref_chroms    = split(/\s/,$ra_chroms->[0]) if (defined $ra_chroms->[0]);
    @a_target_chroms = split(/\s/,$ra_chroms->[1]) if (defined $ra_chroms->[1]);

    if ( @a_ref_chroms == 0 || @a_target_chroms == 0)
    {
        @a_ref_chroms    = ( 'all' );
        @a_target_chroms = ( 'all' );
    }
    
	return (\@a_ref_chroms,\@a_target_chroms);
}

=head2 function _GetMainImages

 Title        : _GetMainImages
 Usage        : $rvl .= $self->_GetMainImages()
 Fonction     : generate the html code useful to read the main images
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetMainImages
{
    my $self = shift;
    
	my $o_param   = $self->_Get('_o_param');
	my $o_dotplot = $self->_Get('__o_dotplot');

    my $reference	= $o_param->Get('reference');
	my $target	= $o_param->Get('target1');
	my $program	= $o_param->Get('program');
	my $level = $o_param->Get('level');
	my $zoom = $o_param->Get('zoom');
    my $ajax = $o_param->GetInteger('__wba_ajax');
    my $html;

    my ($path_dotplot,$path_coord_map,$http_dotplot)=$o_dotplot->GetDotPlotPaths();

    $html .= "<div id=\"ID_DYNAMIC_MARKER\" ></div>";
    $html .= "<img alt=\"Dotplot\" class=\"MapsImg\" id=\"ID_DOTPLOT\" usemap=\"#ID_MAP_MAP\" src=\"$http_dotplot\" />\n";

	# Write the areas: we have one area/line in the .coord file
    $html .= "<map name=\"ID_MAP_MAP\" id=\"ID_MAP_MAP\" >\n";
    my $fh_map_coord = new IO::File($path_coord_map) or die "ERROR - Cannot open $path_coord_map file";

	my @a_maps;	
    while (my $line=<$fh_map_coord>)
    {
        chomp($line);
        my @f = split(/\t/,$line);
        my $coord = join(',',@f[0..3]);
        my $obj   = $f[4];
        $obj =~s/"//g;
        $obj = 'O_'.$obj;
        my $params= join(',',@f[0..3]).','.join(',',@f[5..$#f]);
        $html .= "<area href=\"#\" alt=\"click here\" shape=\"rect\" coords=\"$coord\" onmouseover='javascript:$obj.MouseOver($params)' onmousedown='javascript:$obj.MouseDown($params)' />\n";
		push(@a_maps,\@f);

    }

    $html .= "</map>\n";

	# Print the coords in JSON format
#	my ($map_js) = $self->CalcJsNames(1,'map');
#	$map_js =~ s/\.coord$/.js/;
#	my $json     = new JSON( );
#	my $map_json = $json->encode(\@a_maps);
#	my $fh_map_js = new IO::File ($map_js,'w') or throw LipmError::IOException($map_js,'f');
#	print $fh_map_js $map_json;
#	$fh_map_js->close();

    return $html;
}

=head2 function _GetMainImagesJson

 Title        : _GetMainImagesJson
 Usage        : $html .= $self->_GetMainImagesFson()
 Fonction     : generate the json code with urls to the main images
                Implement an abstract function from NcView
 Returns      : Json code
 Args         : none
 Globals      : none

=cut

sub _GetMainImagesJson
{
    my $self = shift;
	my $o_io     = $self->_Get('_o_io');
	my $o_dotplot = $self->_Get('__o_dotplot');

    if ($o_dotplot eq '')
	{
		throw NarcisseError(-bd => 'Sorry, the dotplot representation could not be drawn');
	}
	else
	{
        my ($path_dotplot,$path_coord_map,$http_dotplot)=$o_dotplot->GetDotPlotPaths();
        my %h_urls;
        $h_urls{dotplot} = $http_dotplot;
        return $self->_ToJson(\%h_urls);
    }
}

=head2 function _GetPrintableImage

 Title        : _GetPrintableImage
 Usage        : $rvl .= $self->_GetPrintableImage()
 Fonction     : generate the html code useful to read the single image
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetPrintableImage
{
    my $self = shift;
    
	my $o_param   = $self->_Get('_o_param');
	my $o_dotplot = $self->_Get('__o_dotplot');
    
    my ($path_dotplot,$path_coord_map,$http_dotplot)=$o_dotplot->GetDotPlotPaths();
    
   #We simplify the name of the image
   my ($new_path_dotplot, $new_http_dotplot) = $self->_GetPrintableName();
   copy($path_dotplot, $new_path_dotplot)
   	or die "Copy failed: $!";
    
    #my $http_printable_dotplot = $o_dotplot->ConstructPrintableImage();
 	my $html ="<img src='".$new_http_dotplot."' alt='The printable image' />";	
 	
 	return $html;
}

=head2 function _GetPrintableName

 Title        : _GetPrintableImage
 Usage        : $rvl .= $self->_GetPrintableImage()
 Fonction     : generate the html code useful to read the single image
 Returns      : two string : a simplified dotplot name and corresponding paths (absolute and http)
 Args         : none
 Globals      : none

=cut
sub _GetPrintableName
{
	my $self = shift;
	
	my $o_io    = $self->_Get('_o_io');
	my $o_dotplot = $self->_Get('__o_dotplot');
	
	my $dotplot_name = $o_dotplot->ComputeDotPlotName( -mode => &PRINTABLE );
	
	my $root = $o_io->GetCfg('ROOT_SESSDIR');
	my $http = $o_io->GetCfg('HTTP_SESSDIR');
	
	my $path_dotplot = "$root/$dotplot_name.png";
	my $http_dotplot = "$http/$dotplot_name.png";
	
	return ( $path_dotplot, $http_dotplot );
}

=head2 function _GetMainTitle

 Title        : _GetMainTitle
 Usage        : $rvl .= $self->_GetMainTitle()
 Fonction     : generate a little table, to be displayed at top of images, containing the title
 Returns      : html code
 Args         : none
 Globals      : none

=cut
sub _GetMainTitle
{
    my $self = shift;
    
	my $o_param   = $self->_Get('_o_param');
    my $o_io      = $self->_Get('_o_io');
	my $html;

	my $reference = $o_param->Get('reference');
    my $target = $o_param->Get('target1');
    my $program = $o_param->Get('program');
    my $zoom = $o_param->Get('zoom');
    my $level = $o_param->Get('level');

	$html = '<div id="ID_TITLE">';
    $html .= "<p style=\"text-align: center\"><b>" . $o_io->GetSpeciesLongName($reference) . "</b> vs <b>" . $o_io->GetSpeciesLongName($target) . "</b> using the method <b>$program</b></p>";

    $html .= "</div>";
}



#################################################################
#                 For development purpose only  (a reecrire en cas de besoin)
#################################################################
# 
# sub AddConstantsToCFG
# {
	# my %h_param=@_;
# 
	# @CFG{keys %h_param}=@h_param{keys %h_param};
# 
# }
# 
# sub __DummySelection
# {
	# my %selection=();
# 
	# my @query_chroms=qw( 3 6 9 18 );
	# my @subject_chroms=qw( 1 2 6 12 13);
# 
	# foreach my $chrom (@query_chroms)
	# {
		# $selection{query}{$chrom}=1;
	# }
	# foreach my $chrom (@subject_chroms)
	# {
		# $selection{subject}{$chrom}=1;
	# }
# 
	# return \%selection;
# }
# 
# sub SerializeArray
# {
    # my $ra = shift;
    # my $name = shift;
    # my $file_name = shift;
# 
    # $file_name = CFG{ROOT_SESSDIR}.'/'.$file_name;
# 
    # use Data::Dumper;
    # local $Data::Dumper::Purity = 1;
# 
    # my $msg = Data::Dumper->Dump([$ra], [$name]);
    # open (SERIAL, ">$file_name") or return;
    # print (SERIAL "$msg\n");
    # close(SERIAL);
# }

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
