#
# 	emmanuel.courcelle@toulouse.inra.fr
#	Created: december 05, 2008
#	$Id: WebBuilderFactory.pm 2317 2009-01-22 14:23:43Z sletort $
#

package NcKaryotype;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod
=head1 NAME

 NcKaryotype.pm

=head1 SYNOPSIS


=head1 DESCRIPTION

 NcKaryotype derives from , this object implements the 'Karyotype' representation

=cut

use strict;
use warnings;

use LipmError;
use LipmError::ObjectException;

use Data::Dumper;
use File::Basename;
use Error qw( :try );
use Carp;

use General;
use WebBuilder;
use KaryotypeMaker;
use cgilib::NcScript;
use cgilib::NcScript::NcView;

use NARCISSE_utils;

use Image;
use Planner;
use Dotplot;
use Planner::GenomeMap;
use Planner::ChromosomeMap;

use constant IMG_EXT   => ".png";
use constant COORD_EXT => ".coord";
use constant JS_EXT    => ".js";
use constant SPECIES_NAME_BAND_SIZE => 80;
use constant DEBUG     => 1;

use base qw( NcView );


=head2 function _Init

 Title        : _Init
 Function     : The constructor, call mother and add a SpeciesObserver object as attribute
 Args         : $o_io         : the configuration hash
                $o_web_builder: A valid WebBuilder object
                $o_logger     : a Logger

=cut
sub _Init
{
	my $self = shift;

	$self->SUPER::_Init( @_ );

	my $o_param = $self->_Get('_o_param');
	my $o_io    = $self->_Get('_o_io');
    my $rh_organisms = $self->_Get('_rh_organisms');
    my $ra_species   = $rh_organisms->{ SPECIES };

	my $o_species = New SpeciesObserver( species => $ra_species->[0], -o_inst_obs => $o_io );
	$self->_Set( __o_species => $o_species );
}


=head2 Procedure _ComputeImages

 Title        : _ComputeImages
 Usage        : $self->_ComputeImage() - Called by the _Init from NcView
 Fonction     : Compute the images
 Args         : none
 Access       : Protected
 Globals      : none

=cut

sub _ComputeImages
{
	my $self = shift;

	my $rh_padding   = { top => 40, bottom => 10, left => 10, right => 10, inner_height => 40 };
	my $supp_data = $self->{ _o_io }->GetCfg( 'SUPP_KARYOTYPE' );
# warn $supp_data;
	my %h_karyo_params = (
		o_species    => $self->_Get( '__o_species' ),
		-o_logger    => $self->_GetLogger(),
		-height      => 400,
		-boxes_width => 100,
		# -supp_data   => '^$', # use o_param to fix it.
		-supp_data   => $supp_data,
		-rh_padding  => $rh_padding,

		# pas encore utilisé, à venir
		-feature_list => [ qw( Cyto::Karyotype ) ] #Synt::Karyotype ) ]
		);

	my $o_karyo = New KaryotypeMaker( %h_karyo_params );
	try
	{
		$self->__PlotPersonalData($o_karyo);
	}
	catch LipmError with
	{
		my( $o_err, $r ) = @_;

		$self->_Trace( $o_err->Explain(), '.', 'IKV' );
	};

	my $file_name = $self->CalcImageNames(1,'karyo');
	my $url_name  = $self->CalcImageNames(0,'karyo');
	my $coord_name= $self->CalcCoordNames(1,'karyo');
	$self->_Set('__file_name',$file_name);
	$self->_Set('__url_name',$url_name);
	$self->_Set('__coord_name',$coord_name);
	$o_karyo->Print( output => $file_name, coord => $coord_name );

	return;
}


=head2 Procedure  __PlotPersonalData

 Usage        : $self->__PlotPersonalData( $o_karyo );
 Function     : Get all personal file, read them and
                plot data to the karyotype file.
 Args         : $o_karyo: A karyotypeBuilder object, already initialized
 Globals      : none

=cut

sub __PlotPersonalData
{
	my $self = shift;
	my( $o_karyo ) = @_;

	my $o_species = $self->_Get( '__o_species' );
	my @a_keys = qw( chrom start end color label );
	my @a_perso_files = $o_species->GetPersoFiles( -existing_files => &TRUE );

	foreach my $perso_file ( @a_perso_files )
	{
		my ($rh_data, $ra_keys) =
				&NARCISSE_utils::ReadNarcisseTabFile( file => $perso_file );
		foreach my $rh_row ( values %$rh_data )
		{
			my( $chrom, $start, $end, $color, $label ) = @$rh_row{ @a_keys };
			$o_karyo->Plot( $chrom, $start, $end, $label, $color );
		}
	}

	return;
}


=head2 function _GetMainImages

 Title        : _GetMainImages
 Usage        : $rvl .= $self->_GetMainImages() - Called by NcView's _GetMain
 Fonction     : generate the html code useful to read the main images - Called by NcView's _GetMain
 Returns      : html code
 Args         : none
 Globals      : none

=cut

sub _GetMainImages
{
	my $self = shift;

	$self->_Trace("html generated\n","*","IK");

	my $html = "";
	# $html .= '<img id="ID_KARYO" alt="karyotype" src="' . $self->_Get('__url_name') . '" />';
	$html   .= "<div id=\"ID_DYNAMIC_MARKER\" class=\"MapsImg\" >&nbsp;</div>";
	my $img       = $self->_Get('__url_name');
	my $map_coord = $self->_Get('__coord_name');
	$html   .= "<img alt=\"karyotype\" id=\"ID_KARYO\" src=\"$img\" usemap=\"#ID_KARYO_MAP\" />\n";

	# Write the areas: we have one area/line in the .coord file
	$html .= "<map name=\"ID_KARYO_MAP\" id=\"ID_MAP_MAP\" >\n";
	my $fh_map_coord = new IO::File($map_coord) or thow LipmError::IOException($map_coord,'f');

	my @a_maps;
	my $id=1;
	while( my $line = <$fh_map_coord> )
	{
		chomp($line);
		my @f = split(/\t/,$line);
		my $coord = join(',',@f[0..3]);
		my $obj   = $f[4];
		$obj =~s/"//g;
		$obj = 'O_'.$obj;
		my $params= join(',',@f[0..3]).','.join(',',@f[5..$#f]);
		# warn $params;
		$html .= "<area alt=\"click here\" shape=\"rect\" href=\"#\" coords=\"$coord\" onmouseover='javascript:$obj.MouseOver($params)' onmousedown='javascript:$obj.MouseDown($params)' />\n";
	#	push(@a_maps,\@f);
	}
	$html .= "</map>\n";
	# Print the coords in JSON format
	#my $json     = new JSON( );
	#my $map_json = $json->encode(\@a_maps);

	#my ($map_js) = $self->CalcJsNames(1,'map');
	#warn "KOUKOU $map_js";
	#my $fh_map_js = new IO::File ($map_js,'w') or throw LipmError::IOException($map_js,'f');
	#print $fh_map_js $map_json;
	#$fh_map_js->close();

	return $html;
}

=head2 function _GetMainImagesJson

 Title        : _GetMainImagesJson
 Usage        : $html .= $self->_GetMainImagesFson()
 Fonction     : generate the json code with urls to the main images
                Implement an abstract function from NcView
 Returns      : Json code
 Args         : none
 Globals      : none

=cut
sub _GetMainImagesJson
{
    my $self = shift;

    my %h_urls;
    $h_urls{karyotype} = $self->_Get('__url_name');
    return $self->_ToJson(\%h_urls);
}

=head2 function _GetMainTitle

 Title        : _GetMainTitle
 Usage        : $rvl .= $self->_GetMainTitle()
 Fonction     : generate a little table, to be displayed at top of images, containing the title

 Returns      : html code
 Args         : none
 Globals      : none

=cut

sub _GetMainTitle
{
	my $self = shift;

	my $o_param   = $self->_Get('_o_param');
	my $o_dotplot = $self->_Get('__o_dotplot');
	my $o_io      = $self->_Get('_o_io');
	my $html;

	my $reference = $o_param->Get('reference');
	my $program = $o_param->Get('program');
	my $zoom = $o_param->Get('zoom');
	my $level = $o_param->Get('level');

	$html = '<div id="ID_TITLE">';
	$html .= "<h2>Karyotype</h2><p><b>" . $reference . "</b> </p>";
	$html .= "</div>";

	return $html;
}

=head2 function _GetPrintableImage

 Title        : _GetPrintableImage
 Usage        : $rvl .= $self->_GetPrintableImage()
 Fonction     : generate the html code useful to read the single image
 Returns      : html code
 Args         : none
 Globals      : none

=cut

sub _GetPrintableImage
{
    my $self = shift;

    my $img       = $self->_Get('__url_name');
 	my $html ="<img src='".$img."' alt='The printable image' />";

 	return $html;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
