
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 24-07, 2009
#	$Id: SnpCurator.pm 3696 2011-07-22 14:32:16Z sletort $
#

package SnpCurator;

=pod

=head1 NAME

 SnpCurator - Single Nucleotide Polymorphism

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 An instance of this class is build with information from glint and fasta seq.
 It treats it and can return it in a convenient way.

=cut

use strict;
use warnings;

use Data::Dumper;

use LipmError::Function;
use General;

use base qw( LipmObject );

=head2 method _Init

 Usage        : my $o_x = New SnpCurator( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of SnpCurator class.
 Returns      : a valid SnpCurator object
 Args         : hash with as following
 str_glint => $str_glint, string, an output line of glint result in "m 3" format
 seq       => $seq, string, sequence used to localized the snp, the one which make it unique
 header    => fasta header associated with the snp
 Attributes   :
    __?

=cut

sub _Init
{
	my $self = shift;
	my %h_args = (  -o_logger      => undef,
					rh_glint_data  => undef,
					o_snp_reader   => undef,
					o_species      => undef,
	               @_ );
	$self->SUPER::_Init( %h_args );

	# assert mandatories
	LipmError::Function::Assert( %h_args );

	$self->_Set( __o_species => $h_args{ o_species },
				 __snp_pos   => undef, # sur la sequence ref- 1e base ; position 1
				 __sequence  => undef,
				 __header    => undef,
				 __rh_header => undef, # when we have key=val, it can be stored
				 __o_snp     => undef,
				 __is_valid  => &TRUE,
				 __iupac     => undef );

	# set __chrom attributes
	$self->__ExtractGlintInformation( $h_args{ rh_glint_data }, $h_args{ o_snp_reader } );

	# set __header, __sequence, can change __is_valid
	$self->__SetSequenceInformation( $h_args{ o_snp_reader } );

	return;
} # _Init


=head2 Filter

 Usage    : $o_x->Filter( $rh_data )
 Procedure: .
 Args     :
     $rh_data
 Error    : none.

=cut

sub Filter
{
	my $self   = shift;

	my $rh_data = $self->__GetGlintData();

	# exclude those that don't align
	$self->__AlignTheWholeSequence( $$rh_data{ rh_query } );
	return  if( !$self->IsValid() );

	# exclude those that align with more than one error
	# attention changement entre deletion et insertion a verifier
	my @a_keys  = qw( ra_mismatches ra_insertions ra_deletions );
	my $type    = $self->__AlignWithOneError( @$rh_data{ @a_keys } );
	return  if( !$self->IsValid() );

	# Assert that the position found is the position of the snp_file
	$self->__ComparePositions();
	return if( !$self->IsValid() );

	my @a_attrs = qw( __sequence __o_species __header );
	my( $seq, $o_species, $header )  = $self->_Get( @a_attrs );

	my $snp_pos = $self->_Get( '__snp_pos' );
	my $chrom_position = $$rh_data{ rh_target }{ start } + $snp_pos - 1;
	my $iupac   = substr( $seq, $snp_pos-1, 1 );

	my $o_snp = New SNP( -o_logger  => $self->GetLogger(),
						 o_species  => $o_species,
						 chrom      => $$rh_data{ chrom },
						 position   => $chrom_position,
						 strand     => $$rh_data{ strand },
						 iupac      => $iupac,
						 -type      => $type,
						 -label     => $header,
						 -src       => $$rh_data{ db },
						 -extern_id => $$rh_data{ db_id },
						 -score     => $$rh_data{ score }
					);

	if( &FALSE == $o_snp->CheckValue() )
	{
		my $msg = "The iupac code used in the snp ($iupac) does not include ";
		$msg   .= "the base of our sequence (" . $o_snp->GetRefBase() . ")\n";
		$msg   .= "snp : $$rh_data{ chrom }, $chrom_position, $header\n";
		$self->_Trace( $msg . "\n", '*', 'Is' );
		$self->_Set( __is_valid => &FALSE );
	}
	else
	{
		$self->_Set( __o_snp => $o_snp );
	}

	return;
} #__Filter


=head2 ____ComparePositions

 Usage    : $o_x->__ComparePositions( none )
 Procedure: Set the snp invalid if the position found differs from
            the position given in the header.
 Args     : none
 Error    : none.

=cut

sub __ComparePositions
{
	my $self = shift;

	my $snp_pos         = $self->_Get( '__snp_pos' );
	my( $jsnp, $dbsnp ) = $self->GetHeaderInfo( qw( allelePos pos ) );

	my $pos_ref;
	$pos_ref = $jsnp  if( defined( $jsnp ) );
	$pos_ref = $dbsnp if( defined( $dbsnp ) );

	if( defined( $pos_ref ) && ( $pos_ref != $snp_pos ) )
	{
		my $header = $self->_Get( '__header' );
		my $msg = "*$header* : the position of the snp we found ($snp_pos) is";
		$msg   .= " different than the one refered in its header ($pos_ref).\n";
		$self->_Trace( $msg, '*', 'Is' );

		$self->_Set( __is_valid => &FALSE );
		return;
	}

	return;
} #__ComparePositions


=head2 __AlignTheWholeSequence

 Usage    : $o_x->__AlignTheWholeSequence( $rh_query )
 Procedure: .
 Args     :
     $rh_query
 Error    : none.

=cut

sub __AlignTheWholeSequence
{
	my $self   = shift;
	my( $rh_query ) = @_;

	my( $seq, $header ) = $self->_Get( qw( __sequence __header ) );

	my $len = length( $seq );

	if( ( 1 != $$rh_query{ start } ) # peut etre 0 !
		or ( $len != $$rh_query{ end } ) ) # peut etre len +1 !
	{
		my $msg = "*$header* doesn't align its whole sequence.\n";
		$self->_Trace( $msg, '*', 'IsV' ); # V car un vrai snp peut s'aligner partiellement ailleurs.

		$self->_Set( __is_valid => &FALSE );
	}

	return;
} #__AlignTheWholeSequence


=head2 __AlignWithOneError

 Usage    : my $type = $o_x->__AlignWithOneError( $ra_mis, $ra_ins, $ra_del )
 Function : also set __snp_pos.
 Args     :
     $ra_mis,
     $ra_ins,
     $ra_del
 Return   :
     $type, char, type of SNP :
     	'M' for mutation, 'I' for insertion, 'D' for deletion

     those values can be undef if the snp is not valid.
 Error    : none.

=cut

sub __AlignWithOneError
{
	my $self   = shift;
	my( $ra_mis, $ra_ins, $ra_del ) = @_;

	my( $snp_pos, $type );

	if( 1 == @$ra_mis )
	{
		$snp_pos = $$ra_mis[0] + 1;
		$type    = 'M';
	}
	elsif( 1 == @$ra_ins )
	{
		$snp_pos = $$ra_ins[0] + 1;
		$type    = 'I';
	}
	elsif( 1 == @$ra_del )
	{
		$snp_pos = $$ra_del[0] + 1;
		$type    = 'D';
	}
	else
	{
		my $header = $self->_Get( '__header' );
		my $msg = "*$header* is not a real SNP, it has more than ONE error.\n";
		$self->_Trace( $msg, '*', 'Is' );

		$self->_Set( __is_valid => &FALSE );
	}

	$self->_Set( __snp_pos => $snp_pos );

	return $type;
} #__AlignWithOneError


=head2 __ExtractGlintInformation

 Usage    : $o_x->__ExtractGlintInformation( $rh_glint )
 Function : parse each of glint colum, especially edit script.
 Args     : $rh_glint hash ref, a row returned by NarcisseDataReader (glint output)
            $o_reader           a FastaReader (the snp fasta file)
 Return   : $rh_data:
                      label     => string,
                      rh_query  => { start => int, end => int },
                      rh_target => { start => int, end => int },
                      ra_mismatches => array ref,
                      ra_insertions => array ref,
                      ra_deletions  => array_ref
 Error    : none.

=cut

sub __ExtractGlintInformation
{
	my $self   = shift;
	my( $rh_glint, $o_snp_reader ) = @_;

	#$self->_Trace("Entering __ExtractGlintInformation\n",'*','IsV');

	# glint header
	my @a_header = ( qw( query subject ), "escript([q_start-q_end]->[s_start_s_end](size,strand,score,evalue,pident)" );
	my( $label, $str_chrom, $data ) = @$rh_glint{ @a_header };

	# extraire la BdD et l'id relatif
	$label += 0;
	my $seq_header = $o_snp_reader->GetIndicedHeader( $label );
	#my( $dumm1, $db, $db_id,  ) = split( /\|/, $header );

	# >gnl|dbSNP|rs8896 rs=8896|... ==> keep only dbSNP,"rs8896 rs=8896"
	my ($dummy,$db,$db_id) = split( /\|/, $seq_header );

	# "rs8896 rs=8896" ==> keep only rs8896 (1st word)
	$db_id =~ s/^(\w+).*$/$1/;	#

	# extraire le nom du chrom.
	my $short    = $self->_Get( '__o_species' )->GetShortName();
	my( $chrom ) = $str_chrom =~ /^$short(.+)$/;

	# les infos du script d'edition
	my( @a_mismatches, @a_insertions, @a_deletions ) = ();

	my( $str_query, $str_target, $str_mismatch ) = $data =~ /\[(.*?)\]/g;
	my( $str_del, $str_insert ) = $data =~ /\{(.*?)\}/g;
	my( $str_score            ) = $data =~ /\((.+)\)/;

	my( $q_start, $q_end ) = split( /-/, $str_query );
	my( $t_start, $t_end ) = split( /-/, $str_target );

	my @a_score            = split( /,/, $str_score );
	my $score              = $a_score[2];
	my $str_strand         = $a_score[1];
	my $strand             = ( $str_strand eq 'PLus/Plus' ) ? '1' : '-1';

	push( @a_mismatches,     split( /,/, $str_mismatch ) ) if( defined( $str_mismatch ) );
	push( @a_deletions,      split( /,/, $str_del ) )    if( defined( $str_del ) );
	push( @a_insertions,     split( /,/, $str_insert ) ) if( defined( $str_insert ) );

	my $rh_data = {	chrom     => $chrom,
					strand    => $strand,
					label     => $label,
					score     => $score,
					db        => $db,
					db_id     => $db_id,
					rh_query  => { start => $q_start, end => $q_end },
					rh_target => { start => $t_start, end => $t_end },
					ra_mismatches => \@a_mismatches,
					ra_insertions => \@a_insertions,
					ra_deletions  => \@a_deletions };

	$self->_Set( __rh_glint_data => $rh_data );

	#$self->_Trace("Returning from __ExtractGlintInformation\n",'*','IsV');

	return;
} #__ExtractGlintInformation


=head2 __SetSequenceInformation

 Usage    : $o_x->__SetSequenceInformation( $o_reader, $label )
 Function : extract raw sequence and header from $o_reader with $label as part header.
 Args     :
     $o_reader, FastaReader object
     $label, string, label found in glint output
 Error    : none.

=cut

sub __SetSequenceInformation
{
	my $self  = shift;
	my( $o_reader ) = @_;

	my $rh_glint_data = $self->__GetGlintData();
	my $label = $$rh_glint_data{ label };

	# \Q pour neutraliser les caracteres speciaux comme '|' du label
	###my $ra_headers = $o_reader->GetHeaders( "^\Q>$label\E" . '\s' );

###	if( 1 != @$ra_headers )
###	{
###		my $msg = "*$label* cannot be identified in fasta file\n";
###		$self->_Trace( $msg, '.', 'Is' );

####		J'hesite a lancer une exception, la je mets un attribut booleen
###		$self->_Set( __is_valid => &FALSE );
##### 		throw LipmError( -text => );
###		return;
###	}

###	my $header    = $$ra_headers[0];

	my $header = $o_reader->GetIndicedHeader( $label );
	my $rh_header = {};
	while( $header =~ /(\w+)=(\w+|"(\S+?)")/g )
	{
		my $key   = $1;
		my $value = $3 || $2;
		$$rh_header{ $key } = $value;
	}

	my $snp_seq = $o_reader->GetSequence( $header );

	$self->_Set( __header    => $header,
				 __rh_header => $rh_header,
				 __sequence  => $snp_seq );

	return;
} #__SetSequenceInformation


=head2 GetSnp

 Usage    : my $rh_data = $o_x->GetSnp( none )
 Function : get snp data in a hash ref with the keys to write.
 Returns  : $rh_data (ca evoluera avec le temps)
 Args     : none
 Error    : none

=cut

sub GetSnp
{
	my $self = shift;

	return undef if( &FALSE == $self->IsValid() );

	my $o_snp = $self->_Get( '__o_snp' );

	return $o_snp if defined( $o_snp );

	# sinon on cree l'objet
	my $rh_data = $self->__GetGlintData();
	my @a_keys  = qw( chrom strand rh_target db_id );
	my( $chrom, $strand, $rh_target, $db_id ) = @$rh_data{ @a_keys };

	my @a_attrs = qw( __sequence __o_species __header __snp_pos __type __iupac );
	my( $seq, $o_species, $header, $snp_pos, $type, $iupac )  = $self->_Get( @a_attrs );

	my $chrom_position;
	if( $strand eq '1' )
	{
		$chrom_position = $$rh_target{ start } + $snp_pos - 1;
	}
	else
	{
		$chrom_position = $$rh_target{ end } - $snp_pos + 1;
	}

	my $alleles = $self->GetHeaderInfo( 'alleles' );
	my $other;

	if( $iupac eq 'N' && $alleles =~ /-/ )
	{
		$iupac = '-';
		if( $type ne 'M' )
		{
			# regexp peut etre trop large...
			( $other ) = $alleles =~ /-\/(.+)/;
		}
		else
		{
			$type = 'D';
		}
	}

# &General::ShowDebug( { iupac => $iupac, type => $type, pos => $chrom_position } );
	$o_snp = New SNP(	-o_logger  => $self->GetLogger(),
						 o_species  => $o_species,
						 chrom      => $chrom,
						 position   => $chrom_position,
						 strand     => $strand,
						 iupac      => $iupac,
						 -type      => $type,
						 -label     => $header,
						 -extern_id => $db_id
						 -score     => $$rh_data{ score }
					);

	if( &TRUE == $o_snp->CheckValue( $other ) )
	{
		$self->_Set( __o_snp => $o_snp );
		return $o_snp;
	}

	my $msg = "The iupac code used in the snp ($iupac) does not include ";
	$msg   .= "the base of our sequence (" . $o_snp->GetRefBase() . ")\n";
	$msg   .= "snp : $$rh_data{ chrom }, $chrom_position, $header\n";
	$self->_Trace( $msg . "\n", '*', 'Is' );
	$self->_Set( __is_valid => &FALSE );

	return undef;
} #GetSnp


=head2 IsValid

 Usage    : my $bool = $o_x->IsValid( none )
 Function : accessor to the __is_valid attributes.
            Snp can be invalid if
                - it cannot belocalized in the genome
                - it cannot be retrieved in the fasta file input of glint with glint label
 Returns  : $bool, &TRUE if valid.
 Args     : none
 Error    : none

=cut

sub IsValid
{
	my $self = shift;

	return $self->_Get( '__is_valid' );
} #IsValid


=head2 GetHeaderInfo

 Usage    : my $val = $o_x->GetHeaderInfo( $key )
            @a_vals = $o_x->GetHeaderInfo( @a_keys )
 Function : accessor to the __rh_header attributes.
            With this functin you can acces to value present in the header
            as $key=$val.
 Returns  : $val, string/int
 Args     : $key, string
            @a_keys, list of keys
 Error    : none

=cut

sub GetHeaderInfo
{
	my $self = shift;
	my @a_keys = @_;

	my $rh_header = $self->_Get( '__rh_header' );

	return wantarray ? @$rh_header{ @a_keys } : $$rh_header{ $a_keys[0] };
} #GetHeaderInfo


=head2 ____ComparePositions2

 Usage    : $o_x->__ComparePositions2()
 Procedure: Set the snp invalid if the position found differs from
            the position given in the header.
 Args     : none
 Error    : none.

 Note  : Code faux, il ne tient pas compte des Insertions

=cut

sub __ComparePositions2
{
	my $self = shift;

	my $rh_data = $self->__GetGlintData();
	my @a_keys  = qw( rh_query ra_mismatches ra_insertions ra_deletions );
	my( $rh_query, $ra_mis, $ra_ins, $ra_del ) = @$rh_data{ @a_keys };

	my $pos_ref = $self->GetHeaderInfo( qw( pos ) );

	my( $start, $end ) = @$rh_query{ qw( start end ) };
	my $snp_pos = $pos_ref - $start;

	return $self->__RejectCauseWrongPosition( $pos_ref )  if( $snp_pos < 0 );

	my $type;

	my $o_mismatch  = New Enum( @$ra_mis );
	if( $o_mismatch->Includes( $snp_pos ) )
	{	$type = 'M';	}
	else
	{
		my $o_insertion = New Enum( @$ra_ins );
		if( $o_insertion->Includes( $snp_pos ) )
		{	$type = 'I';	}
		else
		{
			my $o_deletion = New Enum( @$ra_del );
			if( $o_deletion->Includes( $snp_pos ) )
			{	$type = 'D';	}
			else
			{
				&General::ShowDebug( $snp_pos, $pos_ref, $o_mismatch, $o_insertion, $o_deletion );
			}
		}
	}

	$self->_Set( __snp_pos => $snp_pos + 1,
	             __type    => $type );

# 	my $snp_pos         = $self->_Get( '__snp_pos' );
# 	if( defined( $pos_ref ) && ( $pos_ref != $snp_pos ) )
# 	{
# 		my $header = $self->_Get( '__header' );
# 		my $msg = "*$header* : the position of the snp we found ($snp_pos) is";
# 		$msg   .= " different than the one refered in its header ($pos_ref).\n";
# 		$self->_Trace( $msg, '.', 'Is' );
#
# 		$self->_Set( __is_valid => &FALSE );
# 		return;
# 	}

	return;
} #__ComparePositions2


sub Filter3
{
	my $self   = shift;

	$self->__FilterOnScore();
	return  if( !$self->IsValid() );

	$self->__FilterOnSnpPosition();
	return  if( !$self->IsValid() );

	my $pos_ref = $self->GetHeaderInfo( qw( pos ) );
	my $seq     = $self->_Get( '__sequence' );
	my $iupac   = substr( $seq, $pos_ref-1, 1 );
	$self->_Set( __iupac => $iupac );

	return;
}


=head2 __FilterOnSnpPosition

 Usage    : $o_x->__FilterOnSnpPosition()
 Procedure: Compare the SNP position given by the DB and the mismatch/indel
            found in the alignment.
 Args     : none
 Error    : none.

=cut

sub __FilterOnSnpPosition
{
	my $self = shift;

	my $rh_data = $self->__GetGlintData();

	# exclude those that align with more than one error
	my @a_keys  = qw( rh_query ra_mismatches ra_insertions ra_deletions );
	my( $rh_query, $ra_mismatches, $ra_inserts, $ra_deletions ) = @$rh_data{ @a_keys };

	my( $start, $end ) = @$rh_query{ qw( start end ) };
	my $pos_ref = $self->GetHeaderInfo( qw( pos ) );
	my $snp_pos = $pos_ref - $start;

	return $self->__RejectCauseWrongPosition( $pos_ref )  if( $snp_pos < 0 );

	# en cas de suppression <=> insertion de '-' dans la sequence du SNP
	# on doit mettre a jour la position attendu du SNP
	foreach my $deletion ( @$ra_deletions )
	{
		last if( $snp_pos < $deletion );

		if( $deletion == $snp_pos )
		{
			# je pense que ce cas n'arrive jamais
warn 'ben si, ca arrive !';
			$self->_Set( __snp_pos => $snp_pos +1, __type => 'D' );
			return;
		}
		++$snp_pos;
	}

	#Est-ce une base depariee ?
	my $o_mismatches = New Enum( @$ra_mismatches );
	if( $o_mismatches->Includes( $snp_pos ) )
	{
		$self->_Set( __snp_pos => $snp_pos +1, __type => 'M' );

		# cas particulier : la suppression de + d'une base
		# on va avoir /N-+/ dans la seq du SNP
		# et /XX+/ dans la sequence de la ref.
		# exemple : (N-) vs (CA)
		# on regarde donc si on a une deletion a la position suivant le SNP
		my $o_deletions = New Enum( @$ra_deletions );
		if( $o_deletions->Includes( $snp_pos + 1 ) )
		{
			# correction du type
			$self->_Set( __type => 'D' );
		}

		return;
	}

	# est-ce une insertion ?
	my $o_insertions = New Enum( @$ra_inserts );
	if( $o_insertions->Includes( $snp_pos ) )
	{
		$self->_Set( __snp_pos => $snp_pos +1, __type => 'I' );
		return;
	}

	return;
} #__FilterOnSnpPosition


=head2 __RejectCauseWrongPosition

 Usage    : $o_x->__RejectCauseWrongPosition( $pos_ref )
 Procedure: .
 Args     :
     $pos_ref
 Error    : none.

=cut

sub __RejectCauseWrongPosition
{
	my $self = shift;
	my( $pos_ref ) = @_;

	my $header = $self->_Get( '__header' );
	my $msg = "*$header* : No SNP found at the position refered in its header ($pos_ref).\n";
	$self->_Trace( $msg, '*', 'Is' );

	$self->_Set( __is_valid => &FALSE );

	return;
} #__RejectCauseWrongPosition


=head2 __FilterOnScore

 Usage    : $o_x->__FilterOnScore( none )
 Procedure: Set the SNP invalid if its score if less than 95 % of maximum score.
            Maximum is obtained with 100% of identity.
 Args     : none
 Error    : none.

=cut

sub __FilterOnScore
{
	my $self = shift;

	my $rh_data = $self->__GetGlintData();
	my $score   = $$rh_data{ score };

	my( $seq, $header ) = $self->_Get( qw( __sequence __header ) );
	my $score_max       = length( $seq );

	if( $score < 0.95*$score_max )
	{
		my $msg = "*$header* : Score too low ($score/$score_max).\n";
		$self->_Trace( $msg, '*', 'Is' );

		$self->_Set( __is_valid => &FALSE );
	}

	return;
} #__FilterOnScore


=head2 __GetGlintData

 Usage    : my $rh_glint_data = $o_x->__GetGlintData()
 Function : .
 Returns  : $rh_glint_data
 Args     :
     none
 Error    : none

=cut

sub __GetGlintData
{
	my $self = shift;

	return $self->_Get( '__rh_glint_data' );
} #__GetGlintData


1;

__END__
