
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 30 07, 2009
#	$Id: SnpManager.pm 3728 2011-09-28 13:24:44Z manu $
#

package SnpManager;

=pod

=head1 NAME

 SnpManager - This class deals with SNPs

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 This class can search for impact of SNPs onto proteins structure.
 This class have to be used for a batch of SNPs from the SAME CHROMOSOME.

= head1 NOTE

 Je ne suis pas sur que l'ordre des operations de recherche soit optimal.

=cut

use strict;
use warnings;

use Data::Dumper;

use LipmError::Function;
use General;
use SNP;

use base qw( LipmObject );

#use General;

=head2 method _Init

 Usage        : my $o_x = New SnpManager( rh_snps => $rh_snps );
 Prerequisite : none
 Fonction     : initialise an instance of SnpManager class.
 Returns      : a valid SnpManager object
 Args         : $rh_snps, hash ref
                	key is the snp id,
                	values are SNP objects.
 Attributes   :
    __?

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger => undef,
	               rao_snps  => undef,
	               o_species => undef,
	               @_ );
	$self->SUPER::_Init( %h_args );

	# assert mandatories
	LipmError::Function::Assert( %h_args );

	my @a_keys = qw( rao_snps o_species );
	my( $rao_snps, $o_species ) = @h_args{ @a_keys };

	$self->_Set( __rao_snps       => $rao_snps,
				 __o_species      => $o_species,
				 __rao_snp_effect => undef );

	# $self->__SetSnps( $rh_snps );

	return;
} # _Init


=head2 CheckForImpact

 Usage    : $o_x->CheckForImpact( none )
 Procedure: .
 Args     :
     none
 Error    : none.

=cut

sub CheckForImpact
{
	my $self = shift;

	# 1- localiser les SNPs hors-gene, intron ou exon
	$self->LocateSnps();

	# 2- cherche les consequences sur le transcript, la proteine
	$self->__SeekForConsequences();

	return;
} #CheckForImpact


=head2 LocateSnps

 Usage    : $o_x->LocateSnps( none )
 Procedure: .
 Args     :
     none
 Error    : none.

=cut

sub LocateSnps
{
	my $self = shift;

	my @a_attrs = qw( __rao_snps __o_species );
	my( $rao_snps, $o_species ) = $self->_Get( @a_attrs );

	my $o_sla     = $o_species->GetDBAdaptor()->get_SliceAdaptor();
	my @ao_effect = ();

	foreach my $o_snp ( @$rao_snps )
	{
		my $num       = 1; # append to element if more than one gene
		my $in_cds    = 0;
		my $rao_genes = $o_snp->GetAllGenes( $o_sla );
		foreach my $o_gene ( @$rao_genes )
		{
			my $ra_transcripts = $o_gene->get_all_Transcripts();
			foreach my $o_transcript ( @$ra_transcripts )
			{
				my $o_tr_snp = New SnpEffect( o_snp        => $o_snp,
											  o_transcript => $o_transcript );
				my $location = $o_tr_snp->LocateOnMrna();
				push( @ao_effect, $o_tr_snp );

				if( ( defined $location ) && ( $location eq 'CDS' ) )
				{
					++$in_cds;
					last;
				}
			}
			last if( 0 != $in_cds );
		}

		$o_snp->SetInCDS( $in_cds );
	}

	$self->_Set( __rao_snp_effect => \@ao_effect );

	return;
} #LocateSnps


=head2 __SeekForConsequences

 Usage    : $o_x->__SeekForConsequences( none )
 Procedure: .
 Args     : none
 Error    : none.

=cut

sub __SeekForConsequences
{
	my $self = shift;

	my @a_attrs = qw( __rao_snp_effect __o_species );
	my( $rao_snp_effect, $o_species ) = $self->_Get( @a_attrs );

	my $o_sla     = $o_species->GetDBAdaptor()->get_SliceAdaptor();
	foreach my $o_snp ( @$rao_snp_effect )
	{
		$o_snp->DefineMutationType( $o_sla );
# 		$o_snp->SeekForConsequencesOnProtein( $o_sla );
	}

	return;
} #__SeekForConsequences


# =head2 __DuplicateSnpData
# 
#  Usage    : my $rh_snp_clone = &__DuplicateSnpData( $rh_snp, $r_num )
#  Function : .
#  Returns  : $rh_snp_clone
#  Args     :
#      $rh_snp, $r_num
#  Error    : none
# 
# =cut
# 
# sub __DuplicateSnpData
# {
# 	my( $rh_snp, $r_num ) = @_;
# 
# 	my $rh_snp_clone = {};
# 	my @a_keys = qw( chr start end label value type );
# 	@$rh_snp_clone{ @a_keys } = @$rh_snp{ @a_keys };
# 
# 	my $elt = join( '_', $$rh_snp{ element }, $$r_num );
# 	++$$r_num;
# 
# 	return ( $elt, $rh_snp_clone );
# } #__DuplicateSnpData
# 

=head2 GetSnpNarcissePrintableData

 Usage        : my $rh_snps = $o_x->GetSnpNarcissePrintableData()
 Function     : accessor for the SNP information, with newly computed infos	.
                It can be used to rewrite snp files.
 Returns      : $rh_snps, hash ref, first key is SNP's id, value is another hash
                ref with keys 'element' and @a_headers with corresponding values.
 Args         : none
 Error        : none

 Note : Je remets la clef chromosome car je ne comprends pas quel était le probleme.

=cut

sub GetSnpNarcissePrintableData
{
	my $self = shift;

	my $rh_out   = {};

	my $rao_snps = $self->_Get( '__rao_snps' );
	foreach my $o_snp ( @$rao_snps )
	{
		my $id      = $o_snp->GetIdentifier();
		my $chrom   = $o_snp->GetChrom();

		my %h_infos = ( element => $id );
		$$rh_out{ $chrom }{ $id } = $o_snp->ToRefHash();
	}

	return $rh_out;
} #GetSnpNarcissePrintableData


=head2 GetSnpEffectNarcissePrintableData

 Usage        : my $rh_snps = $o_x->GetSnpEffectNarcissePrintableData()
 Function     : accessor for the newly computed information.
                It can be used to print SnpEffect files.
 Returns      : $rh_snps, hash ref, first key is SNP's id, value is another hash
                ref with printable data.
 Args         : none
 Error        : none

=cut

sub GetSnpEffectNarcissePrintableData
{
	my $self = shift;

	my $rh_out = {};

	# not yet implemented

	return $rh_out;
} #GetSnpEffectNarcissePrintableData
1;

__END__
