
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
# 	sebastien.letort@toulouse.inra.fr
#	Created: 04-08, 2009
#	$Id: SNP.pm 3167 2010-01-12 14:51:58Z sletort $
#

package SNP;

=pod

=head1 NAME

 SNP - a class for SNP manipulation

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use LipmError::Function;
use General;
use Bio::EnsEMBL::Utils::Sequence  qw(reverse_comp);

use Bio::Tools::CodonTable; # bioperl

use base qw( LipmObject );

my $SEP = ',';

{
# hash is a copy of Bio::Tools::IUPAC

	my %IUB = ( A => [qw(A)],
				C => [qw(C)],
				G => [qw(G)],
				T => [qw(T)],
				U => [qw(U)],
				M => [qw(A C)],
				R => [qw(A G)],
				W => [qw(A T)],
				S => [qw(C G)],
				Y => [qw(C T)],
				K => [qw(G T)],
				V => [qw(A C G)],
				H => [qw(A C T)],
				D => [qw(A G T)],
				B => [qw(C G T)],
				X => [qw(G A T C)],
				N => [qw(G A T C)]
		     );


=head2 GetBasesFromIupac

 Usage    : my $ra_bases = $o_x->GetBasesFromIupac( $iupac )
 Function : .
 Returns  : $ra_bases
 Args     :
     $iupac
 Error    : none

=cut

	sub GetBasesFromIupac
	{
		my( $iupac ) = @_;

		my $ra_bases = $IUB{ uc( $iupac ) };

		if( !defined( $ra_bases ) )
		{
			throw LipmError( -text => "$iupac is not an IUPAC code\n" );
		}

		return $ra_bases;
	} #GetBasesFromIupac



=head2 Function BaseBelongsToIUPAC

 Usage        : $others = BaseBelongsToIUPAC( $base, $iupac )
 Function     : ad hoc function.
                Checks if $base belongs to $iupac.
 Args         :
     $base, char.
     $iupac char.
 Return       : undef if $base does not belongs to $iupac
	$others, a string, the other bases that are allowed for this $iupac.
 Error        : none.

=cut

	sub BaseBelongsToIUPAC
	{
		my( $base, $iupac ) = @_;

		return $iupac if( '-' eq $iupac );

		my $ra_iupac = GetBasesFromIupac( $iupac );

		$base = uc( $base );
		my $o_enum = New Enum( @$ra_iupac );
		return undef unless( $o_enum->Includes( $base ) );

		$o_enum->RemoveValues( $base );

		return join( $SEP, $o_enum->ListItems() );
	} #BaseBelongsToIUPAC

} # end of closure


=head2 method _Init

 Usage        : my $o_x = New SNP( %h_snp );
 Prerequisite : none
 Fonction     : initialise an instance of SNP class.
 Returns      : a valid SNP object
 Args         : %h_snp, hash with those keys
     chrom,
     position, int, the base number of the SNP
     strand, char [1/-1]
     iupac, char, iupac code allowed for this position, '-' also allowed
     -type, -label, -id, -src, -extern_id, -alignment_score
     cf attribute for meanings.
 Attributes   :
    __o_species, SpeciesObserver object for the species the snp belongs to.
    __chrom, string, chromosome id
    __position, integer, position of the mutation on the chromosome
    __strand, string, (1/-1), strand where the mutation have the value stored.
    __type, string, 'M' for mutation, 'D/I' for deletion/insertion
    __value_ref, char, the base in the sequence, '-' in case of insert
    __value_other, string, all base that are allowed for the snp
    __iupac, string iupac or '-/$iupac', all possible values
    __label, string, anything to show in the interface
    __id, string, unique element id
    __src, string, where the snp comes from.
    __extern_id, string, the id of the snp in the __src
    __alignment_score, integer, score of the alignement of the SNP sequence
        (given by __src) for this chrom-position-strand
        if one __extern_id give us many alignment, we will be able to
        discriminate them on this score.
    __in_cds, boolean, &TRUE if SNP is located in at least one CDS.
    __mutation, char, S,M,N for Silent, Missense, Nonsense

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger  => undef,
	               o_species  => undef,
	               chrom      => undef,
	               position   => undef,
	               strand     => undef,
	               iupac      => undef,
	               -type      => 'M',
	               -label     => '',
	               -src       => '',
	               -extern_id => '',
	               -alignment_score => 0,
# 	               -id        => '',
	               @_ );
	$self->SUPER::_Init( %h_args );

	# assert mandatories
	LipmError::Function::Assert( %h_args );

	$self->_Set( __o_species  => $h_args{ o_species },
	             __chrom      => $h_args{ chrom },
	             __position   => $h_args{ position },
	             __strand     => $h_args{ strand },
	             __iupac      => $h_args{ iupac },
	             __type       => $h_args{ -type },
	             __label      => $h_args{ -label },
# 	             __src        => $h_args{ -src },
	             __extern_id  => $h_args{ -extern_id },
	             __alignment_score => $h_args{ -alignment_score },
# 	             __id         => $h_args{}
	             __value_ref  => undef,
	             __value_other=> undef,

	             # /!\ valeur de l'attribut si plusieurs transcripts
	             __in_cds     => undef,
	             __mutation   => undef
	          );

	return;
} # _Init


=head2 GetLabel,GetRefBase,GetChrom,GetStrand,GetScore

 Usage    : my $string = $o_x->GetX()
 Function : Accessor.
 Returns  : $string
 Args     : none
 Error    : none

=cut

sub GetLabel
{
	my $self = shift;
	return $self->_Get( '__label' );
} #GetLabel

sub GetRefBase
{
	my $self = shift;
	return $self->_Get( '__value_ref' );
} #GetLabel

sub GetChrom
{
	my $self = shift;
	return $self->_Get( '__chrom' );
} #GetLabel

sub GetStrand
{
	my $self = shift;
	return $self->_Get( '__strand' );
} #GetStrand

sub GetScore
{
	my $self = shift;
	return $self->_Get( '__alignment_score' );
} #GetScore

sub GetIupac
{
	my $self = shift;
	return $self->_Get( '__iupac' );
} #GetStrand

sub GetExternId
{
	my $self = shift;
	return $self->_Get( '__extern_id' );
} #GetExternId



=head2 CheckValue

 Usage    : my $bool = $o_x->CheckValue( none )
 Function : .
 Returns  : $bool
 Args     :
     none
 Error    : none

=cut

sub CheckValue
{
	my $self = shift;
	my( $other_bases ) = @_;

	my @a_attrs         = qw( __type __iupac );
	my( $type, $iupac ) = $self->_Get( @a_attrs );
	my( $base_ori, $other );

	if( 'I' eq $type )
	{
		$base_ori = '-';

		if( '-' eq $iupac )
		{	$other = $other_bases;	}
		else
		{
			my $ra_bases = &GetBasesFromIupac( $iupac );
			$other    = join( $SEP, @$ra_bases );
		}
	}
	elsif( 'D' eq $type && defined( $other_bases ) )
	{
		# verifier que le code iupac lu correspond a notre sequence de ref
		$base_ori = $self->__GetBaseRefSequence( length( $other_bases ) );
		if( $base_ori eq $other_bases )
		{	$other = '-';	}
		else
		{	( $base_ori, $other ) = ( '-', $other_bases );	}
	}
	else
	{
		# verifier que le code iupac lu correspond a notre sequence de ref
		$base_ori = $self->__GetBaseRefSequence();

		if( 'D' eq $type && defined( $other_bases ) )
		{	$other = $other_bases	}
		else
		{	$other = &BaseBelongsToIUPAC( $base_ori, $iupac );	}
	}

	$self->_Set( __value_ref   => $base_ori,
				 __value_other => $other );

	return &FALSE  if( !defined $other );

	return &TRUE;
} #CheckValue


=head2 __GetBaseRefSequence

 Usage    : my $seq = $o_x->__GetBaseRefSequence( $nb_bases )
 Function : Return the base corresponding to the SNP in the reference sequence.
 Returns  : $seq, the base or - in case of insertion
 Args     : $nb_bases, int [1], nb of base of the SNP as it is not a single base anymore.
 Error    : none

=cut

sub __GetBaseRefSequence
{
	my $self = shift;
	my( $nb_bases ) = @_;

	$nb_bases = 1 unless defined( $nb_bases );
	my $len = $nb_bases - 1;

	my @a_attrs = qw( __chrom __position __strand __o_species );
	my( $chrom, $pos, $strand, $o_species ) = $self->_Get( @a_attrs );

	my $sla = $o_species->GetDBAdaptor()->get_SliceAdaptor();
	my $o_slice = $sla->fetch_by_region( 'chromosome', $chrom, $pos, $pos+$len, $strand );

	# on dirait que je ne peux obtenir que le brin + d'un slice.
	# la faute a l'appel de SequenceAdaptor::fetch_by_start_end_strand
	# a qui EnsEMBL passe 1 pour le brin.
	my $seq = $o_slice->seq();
	reverse_comp( \$seq ) if( $strand == -1 );
# &General::ShowDebug( $seq, $o_slice->start, $o_slice->end );

	return $seq;
} #__GetBaseRefSequence


=head2 GetAllGenes

 Usage    : my $rao_genes = $o_x->GetAllGenes( $o_sla [, $length ] )
 Function : return the ref list of gene involved with this SNP.
            If no gene found, return empty ref list.
 Returns  : $rao_genes, a ref list of Bio::EnsEMBL::Gene object (API)
 Args     :
     $o_sla, SliceAdaptor object
     $length, integer [1], number of base around the SNP to consider.
         -> example SNP before a gene can modify the begining of this gene.
 Error    : none

=cut

sub GetAllGenes
{
	my $self = shift;
	my( $o_sla, $length ) = @_;

	$length = 0 unless defined( $length );

	my( $pos, $chrom ) = $self->GetPosition();

	my $o_slice   = $o_sla->fetch_by_region( 'chromosome', $chrom, $pos-$length, $pos+$length );
	my $rao_genes = $o_slice->get_all_Genes();

	# set a list of genes' id ?

	return $rao_genes;
} #GetAllGenes


=head2 GetPosition

 Usage    : my $position       = $o_x->GetPosition();
            my( $pos, $chrom, $strand ) = $o_x->GetPosition();
 Function : accessor to position in chromosome coordinate.
 Returns  : $position, integer, first base is number 1.
            $chrom, string, chrom id
            $strand, integer, 1/-1
 Args : none 
 Error : none

=cut

sub GetPosition
{
	my $self = shift;

	my @a_attrs = qw( __chrom __position __strand );
	my( $chrom, $position, $strand ) = $self->_Get( @a_attrs );

	return wantarray ? ( $position, $chrom, $strand ) : $position;
} #GetPosition


=head2 GetPolymorphism

 Usage    : my $mutation = $o_x->GetPolymorphism()
 Function : Accessor for __value attribute.
 Returns  : $mutation, string, __value
 Args     : none
 Error    : none

=cut

sub GetPolymorphism
{
	my $self = shift;

	my $other_base = $self->_Get( '__value_other' );
	my $values = join( $SEP, $self->GetRefBase(), $other_base );

	return $values;
} #GetPolymorphism


=head2 GetIdentifier

 Usage    : my $id = $o_x->GetIdentifier( none )
 Function : return a unique identifier composed of the chrom and the position.
 Returns  : $id, string
 Args     : none
 Error    : none

=cut

sub GetIdentifier
{
	my $self = shift;

	my @a_attrs = qw( __chrom __position __iupac );
	my( $chrom, $position, $iupac ) = $self->_Get( @a_attrs );

	my $id = join( '_', $chrom, $position, $iupac );

	return $id;
} #GetIdentifier


=head2 ToRefHash

 Usage    : my $rh_snp = $o_x->ToRefHash()
 Function : Return as hash ref all attribute to store in snp file.
 Returns  : $rh_snp, keys are columns described in the snp format cf trac.
 Args     : none
 Error    : none

=cut

sub ToRefHash
{
	my $self = shift;

	my @a_attrs = qw( __extern_id __o_species __in_cds __mutation );
	my( $src_id, $o_species, $in_cds, $mutation ) = $self->_Get( @a_attrs );
	my $pos     = $self->GetPosition();

	$in_cds   = '' unless defined( $in_cds );
	$mutation = '' unless defined( $mutation );

	my $rh_snp  = {
					chrom     => $self->GetChrom(),
					position  => $pos,
					strand    => $self->GetStrand(),
					element   => $self->GetIdentifier(),
					values    => $self->GetPolymorphism,
					extern_id => $src_id,
					in_cds    => $in_cds,
					mutation  => $mutation,
					score     => $self->GetScore()
				 };

	return $rh_snp;
} #ToRefHash


=head2 SetInCDS

 Usage    : $o_x->SetInCDS( $bool )
 Procedure: .
 Args     : none
 Error    : none.

=cut

sub SetInCDS
{
	my $self = shift;
	my( $bool ) = @_;

	$self->_Set( __in_cds => $bool );

	return;
} #SetInCDS


=head2 SetMutation

 Usage    : $o_x->SetMutation( $codon, $pos_snp )
 Procedure: use a Bio::Tools::CodonTable to define the type of the mutation. (bioperl)
            if the CodonTable could be defined for the organism, code might be faster.
 Args     :
     $codon, string, the 3 bases of the codon.
     $pos_snp, 0/1/2, position of the SNP in the codon.
               /!\ first base is 0.
               You can consider it as the number of base *before* the SNP.
 Error    : none.

=cut

sub SetMutation
{
	my $self = shift;
	my( $codon, $nb_base_before_snp ) = @_;

	#use iupac code in the codon
	my $iupac = $self->GetIupac();
	substr( $codon, $nb_base_before_snp, 1, $iupac );

	# translation
	my $o_codon_table = new Bio::Tools::CodonTable();
	my $aa            = $o_codon_table->translate( $codon );

	my $mutation;
	if(    $aa eq '*' )
	{
		$mutation = 'N';
	}
	elsif( $aa eq 'X' )
	{
		$mutation = 'M';
	}
	else
	{
		$mutation = 'S';
	}

	$self->_Set( __mutation => $mutation );

	return;
} #SetMutation

1;



package SnpEffect;

=pod

=head1 NAME

 SnpEffect - a class for SNP effect considering a specific transcript

=head1  SYNOPSIS

 # code

=head1 DESCRIPTION

 #descr

=cut

use strict;
use warnings;

use Data::Dumper;

use LipmError::Function;
use LipmError::ObjectException;

use base qw( LipmObject );

#use General;

=head2 method _Init

 Usage        : my $o_x = New SnpEffect( %h_args );
 Prerequisite : none
 Fonction     : initialise an instance of SnpEffect class.
 Returns      : a valid SnpEffect object
 Args         : %h_args, a hash with keys :
     o_snp        => $o_snp, a SNP object
     o_transcript => $o_transcript, a Bio::EnsEMBL::Transcript object
 Attributes   :

=cut

sub _Init
{
	my $self = shift;
	my %h_args = ( -o_logger    => undef,
	               o_snp        => undef,
	               o_transcript => undef,
	               @_ );
	$self->SUPER::_Init( %h_args );

	# assert mandatories
	LipmError::Function::Assert( %h_args );

	my $o_transcript = $h_args{ o_transcript };

	if( !$o_transcript->IsConsistent() )
	{
		my $msg = $o_transcript->dbID() . " is inconsistent.\n";
		throw LipmError::ObjectException( $self, '!build', -text => $msg );
	}

	$self->_Set( __o_snp         => $h_args{ o_snp },
	             __o_transcript  => $o_transcript,
	             __transcript_id => $o_transcript->dbID(),
	             __strand        => $o_transcript->strand(),
	             __gene_location => undef,
	             __o_exon        => undef,
	             __mrna_location => undef );

	return;
} # _Init


=head2 LocateOnMrna

 Usage    : $location = $o_x->LocateOnMrna()
 Function : Seek for the place of the SNP in the transcript and in mRNA.
            $location can be 'Intron' or 'Exon' for the transcript.
            It can be '5_UTR', 'UTR_3' or 'CDS' for the mRNA.
            mRNA location is returned.
 Args     : none
 Return   : $location, string, '5_UTR', 'UTR_3' or 'CDS'
 Error    : none.

=cut

sub LocateOnMrna
{
	my $self = shift;

	my @a_attrs = qw( __o_snp __o_transcript );
	my( $o_snp, $o_transcript ) = $self->_Get( @a_attrs );

	# Intron/Exon
	my ($tr_location, $mrna_location );
# 	my $ra_exons = $o_transcript->get_all_Exons();
# 	if( 0 == @$ra_exons ) # je ne suis pas sur de moi, je ne comprends pas ce que j'ai voulu faire !
# 	{
# 		$tr_location = 'Intron';
# 	}
# 	else
# 	{
# 		$tr_location = 'Exon';

	# will set __o_exon_with_snp and __gene_location
	my $gene_location = $self->__FindAndSetExonWithSnp();
	return $gene_location if( $gene_location eq 'Intron' );

	# else SNP is in a exon, lets locate it on the mRNA
	my $strand   = $o_transcript->strand();

	# 5_UTR/CDS/UTR_3
	# attention $first_base n'est pas la position de l'ATG
	# mais la position de l'extremite sur le brin + :
	# donc la position de l'ATG sur le brin + ou du STOP sur le brin -
	my $first_base = $o_transcript->coding_region_start();
	my $last_base  = $o_transcript->coding_region_end();
	my $snp_pos    = $o_snp->GetPosition();

	if( $snp_pos < $first_base )
	{
		$mrna_location = ( 1 == $strand ) ? '5_UTR' : 'UTR_3';
	}
	elsif( $last_base < $snp_pos )
	{
		$mrna_location = ( 1 == $strand ) ? 'UTR_3' : '5_UTR';
	}
	else
	{
		$mrna_location = 'CDS';
	}

	$self->_Set( __mrna_location => $mrna_location );
warn $o_snp->GetLabel() if( $mrna_location eq 'CDS' );

	return $mrna_location;
} #LocateOnMrna


=head2 __FindAndSetExonWithSnp

 Usage    : $o_x->__FindAndSetExonWithSnp()
 Function : .
 Args     : none
 Error    : none.
 return   : $tr_location, string, where the snp is located in the gene

=cut

sub __FindAndSetExonWithSnp
{
	my $self = shift;

	my @a_attrs = qw( __o_snp __o_transcript );
	my( $o_snp, $o_transcript ) = $self->_Get( @a_attrs );

	my( $snp_pos, $chrom ) = $o_snp->GetPosition();

	# find the exon containing the SNP
	my $gene_location = '';
	my $o_exon_snp;
	my $rao_exons = $o_transcript->get_all_Exons();

	# in - strand, exons are in reverse order, but start < end still.
	@$rao_exons   = reverse @$rao_exons if( -1 == $o_transcript->strand );
	foreach my $o_exon ( @$rao_exons )
	{
		next  if( $o_exon->end < $snp_pos );
		if( $snp_pos < $o_exon->start )
		{
			$gene_location = 'Intron';
		}
		else
		{
			$gene_location = 'Exon';
			$o_exon_snp    = $o_exon;
		}
	}
# if( -1 == $o_transcript->strand )
# {
# 	warn $o_snp->GetLabel();
# 	warn $o_transcript->dbID  . '//' . $snp_pos . '//' . $gene_location;
# }

	$self->_Set( __gene_location => $gene_location,
				 __o_exon      => $o_exon_snp );

	return $gene_location;
} #__FindAndSetExonWithSnp


=head2 DefineMutationType

 Usage    : $o_x->DefineMutationType( $o_sla )
 Procedure: If the SNP is located in the CDS, this function extracts
            the codon containing the mutation, then determine what kind of
            mutation it is (Silent, Missense, Nonsense).
 Args     : $o_sla, a slice adaptor.
 Error    : none.

=cut

sub DefineMutationType
{
	my $self = shift;
	my( $o_sla ) = @_;

	my @a_attrs = qw( __mrna_location __o_transcript __o_snp __o_exon );
	my( $location, $o_transcrit, $o_snp, $o_exon ) = $self->_Get( @a_attrs );
	return unless( defined( $location ) && $location eq 'CDS' );

	my( $snp_pos, $chrom ) = $o_snp->GetPosition();
	my $value              = $o_snp->GetPolymorphism();
	my $tr_strand          = $o_transcrit->strand();

	my $seq_exon = $o_exon->seq()->seq();

	my $o_cds = $o_transcrit->translation();
warn $snp_pos;
warn $o_transcrit->dbID();
warn $o_transcrit->strand();
&General::ShowDebug( { cds => { start => $o_cds->start(),
								g_start => $o_cds->genomic_start(),
								c_start => $o_cds->cdna_start() } } );
&General::ShowDebug( { exon => { start => $o_exon->start(),
								ccs_start => $o_exon->cdna_coding_start($o_transcrit),
								cr_start => $o_exon->coding_region_start($o_transcrit),
								c_start => $o_exon->cdna_start($o_transcrit)
						} } );
&General::ShowDebug( { trans => {
							id => $o_transcrit->dbID(),
							strand => $o_transcrit->strand(),
							cc_start => $o_transcrit->cdna_coding_start(),
							cr_start => $o_transcrit->coding_region_start()
						} } );


# # warn Dumper { exon => [ $o_exon->start(), $o_exon->end() , $o_exon->strand() ] };
# # warn Dumper { cdna => [ $o_exon->cdna_coding_start($o_transcrit), $o_exon->cdna_coding_end($o_transcrit) ] };
# # warn Dumper { position => $snp_pos, value => $value, strand => $o_snp->GetStrand() };
# # warn $seq_exon;
# 
# 		my( $pos_in_cds_exon, $snp );
# 		my $start_cds = $o_exon->cdna_coding_start( $o_transcrit ) - 1;
# if( !defined $o_exon->cdna_coding_start( $o_transcrit ) )
# {
# 	warn $o_transcrit->dbID() . '-' . $o_exon->dbID() . ' => ' . $o_snp->GetLabel();
# 	warn "snp pos:$snp_pos";
# }
# 
# 		if( $o_exon->strand == 1 )
# 		{
# 			$pos_in_cds_exon = $snp_pos - ($o_exon->start()-1);# - $start_cds;
# 		}
# 		else
# 		{
# 			$pos_in_cds_exon = $o_exon->end() - ( $snp_pos -1);# - $start_cds;
# 		}
# 
# 		# le codon contenant le snp commence a la base $pos_in_cds_exon-$nb_avant
# 		my $nb_avant = ( $pos_in_cds_exon - 1 ) % 3;
# 		my $snp_codon_start = $pos_in_cds_exon - $nb_avant;
# # warn Dumper [ length( $seq_exon ), $snp_codon_start-1, $pos_in_cds_exon, $o_snp->GetIdentifier() ];
# # if( $pos_in_cds_exon < 0 )
# # {
# # 	warn Dumper { snp_pos => $snp_pos,
# # 				exon_start => $o_exon->start,
# # 				exon_end => $o_exon->end,
# # 				tr_id    => $o_transcrit->dbID(),
# # 				exon_strand => $o_exon->strand,
# # 				cdna_start => $o_exon->cdna_start( $o_transcrit ),
# # 				cds_start => $start_cds };
# # }
# 		my $snp_codon = substr( $seq_exon, $snp_codon_start-1, 3 );
# # $snp = $snp_codon;
# # if( $o_snp->GetStrand() != $o_exon->strand() )
# # {
# # warn 'inverse';
# # 	use Bio::EnsEMBL::Utils::Sequence  qw(reverse_comp);
# # 	
# # 	reverse_comp( \$snp );
# # }
# 
# 		$o_snp->SetMutation( $snp_codon, $nb_avant );
# 
# # warn Dumper {	pos_in_exon => $pos_in_cds_exon,
# # 				snp_exon =>  $snp_codon,
# # 				alter    => $alter,
# # 				nb_base_avant => $nb_avant };
# 
# # my $o_seq = $o_exon->peptide( $o_transcrit );
# # my $pep = $o_seq->seq();
# # warn $pep;
# # my $n_aa = int( ($pos_in_cds_exon-1)/3 )+1;
# # my $aa = substr( $pep, $n_aa-1, 1 );
# # warn Dumper { $n_aa => $aa };
# 	}


	return;
} #DefineMutationType


=head2 SeekForConsequencesOnProtein

 Usage    : $o_x->SeekForConsequencesOnProtein()
 Procedure: .
 Args     : none
 Error    : none.

=cut

sub SeekForConsequencesOnProtein
{
	my $self = shift;
	my( $o_sla ) = @_;

	return;
} #SeekForConsequencesOnProtein


1;

__END__
