# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#
# 	sebastien.letort@toulouse.inra.fr
#	Created: October 07, 2009
#	$Id: SnpReader.pm 3167 2010-01-12 14:51:58Z sletort $
#

package FastaReader::SnpReader;

=pod

=head1 NAME

 FastaReader::SnpReader - an ad hoc class to parse snp fasta file.

=head1  SYNOPSIS

 my $o_reader = New FastaReader::SnpReader( $filepath );

 # access one of the key=value found in the header
 my $value = $o_reader->GetKey ... ?

 # access the IUPAC letter of a given snp
 my $iupac = $o_reader->GetIupac( $snp_id );

=head1 DESCRIPTION

 Snp fasta files can contain more than 100 000 sequences.
 This class aims to access fastly to information relative to each sequence.

 It is important to not that for now, only sequence from NCBI are allowed.
 Snp can be retrieved through their id : rsxxxxx.

=cut

use strict;
use warnings;

use Data::Dumper;

use LipmError::Function;

use base qw( FastaReader );

#use General;

=head2 method _Init

 Usage        : my $o_x = New FastaReader::SnpReader( filename => $filepath );
 Prerequisite : none
 Fonction     : initialise an instance of FastaReader::SnpReader class.
                call mother and defines attributes.
 Returns      : a valid FastaReader::SnpReader object
 Args         : $filepath
 Attributes   :
    __rh_snps, hash ref of all referenced snps, key is snp_id
    	second keys are :
    		keyval, value is a hash ref of key => value,
    		iupac, value is the iupac code,
    		seq_len, value is the sequence length
    		header, value is the complete header

=cut

sub _Init
{
	my $self = shift;
# 	my %h_args = ( -o_logger => undef,
# 	               filename  => undef,
# 	               @_ );
	$self->SUPER::_Init( @_ );

die;
}
	$self->_Set( __rh_snps => {} );

	# search for snp info
	$self->__ExtractSnpInfos();

	return;
} # _Init


=head2 __ExtractSnpInfos

 Usage    : $o_x->__ExtractSnpInfos( none )
 Procedure: Parse sequence header to extract key=value,
 			then extract the iupac character.
 Args     : none.
 Error    : none.

=cut

sub __ExtractSnpInfos
{
	my $self = shift;

	my $rh_snps = $self->_Get( '__rh_snps' );
die;

	my @a_headers = $self->GetHeaders();
&General::ShowDebug( \@a_headers );
	foreach my $header ( @a_headers )
	{
		# get snp_id from header
&General::ShowDebug( { $header => $snp_id } );
		my $snp_id = &ExtractSnpIdFromHeader( $header );
&General::ShowDebug( { $header => $snp_id } );

		# get key-val info from header
		my $rh_header = {};
		while( $header =~ /(\w+)=(\w+|"(\S+?)")/g )
		{
			my $key   = $1;
			my $value = $3 || $2;
			$$rh_header{ $key } = $value;
		}
		$$rh_snps{ $snp_id }{ keyval } = $rh_header;

		# get iupac from header
		my $raw_seq   = $self->GetRawSequence( $header );
		my $iupac_pos = $$rh_header{ pos };
		my $iupac     = substr( $raw_seq, $iupac_pos-1, 1 );

		my $msg = "Get the iupac code ($iupac) on position $iupac_pos for snp ($snp_id)\n";
		$self->_Trace( $msg, '.', 'IsV' );

		$$rh_snps{ $snp_id }{ iupac }   = $iupac;
		$$rh_snps{ $snp_id }{ seq_len } = length( $raw_seq );
		$$rh_snps{ $snp_id }{ header }  = $header; # pour la compatibilite
	}

die;
	return;
} #__ExtractSnpInfos


=head2 Procedure ExtractSnpIdFromHeader

 Usage        : $snp_id = ExtractSnpIdFromHeader( $header )
 Function     : I assume that header is like >gnl|dbSNP|rsxxx. (ncbi format)
                The funciton will search for the pattern "^>gnl|dbSNP|(rs\d+)".
 Args         : $header
 Error        : none.

=cut

sub ExtractSnpIdFromHeader
{
	my( $header ) = @_;

	my( $snp_id ) = $header =~ /^>gnl|dbSNP|(rs\d+)/;

	return $snp_id;
} #ExtractSnpIdFromHeader


=head2 GetValueFromKey

 Usage    : my $value = $o_x->GetValueFromKey( $snp_id, $key )
 Function : .
 Returns  : $value
 Args     :
     $snp_id, $key
 Error    : none

=cut

sub GetValueFromKey
{
	my $self = shift;
	my( $snp_id, $key ) = @_;

	my $rh_snps = $self->_Get( '__rh_snps' );

	return $$rh_snps{ $snp_id }{ keyval }{ $key };
} #GetValueFromKey


=head2 GetIupacCode

 Usage    : my $header = $o_x->GetIupacCode( $snp_id )
 Function : .
 Returns  : $header
 Args     :
     $snp_id
 Error    : none

=cut

sub GetIupacCode
{
	my $self = shift;
	my( $snp_id ) = @_;

	my $rh_snps = $self->_Get( '__rh_snps' );

	return $$rh_snps{ $snp_id }{ iupac };
} #GetIupacCode


=head2 GetHeader

 Usage    : my $header = $o_x->GetHeader( $snp_id )
 Function : .
 Returns  : $header
 Args     :
     $snp_id
 Error    : none

=cut

sub GetHeader
{
	my $self = shift;
	my( $snp_id ) = @_;

	my $rh_snps = $self->_Get( '__rh_snps' );

	return $$rh_snps{ $snp_id }{ header };
} #GetHeader


1;

__END__
