#! /usr/bin/perl

#
# $Id$
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

ipr2fasta.pl

=head1 SYNOPSIS

ipr2fasta --species <short_name> --ipr <annotation file>

=head1 DESCRIPTION

This program retrieves the ipr annotations from an embl-like file adn integrates them to the files:
     - db/prot/<species>_prot.fasta
     - db/prot/<species>_rep_prot.fasta
     - prot/chromosome*.prot

To retrieve the annotation files from an xml file, here is the complete workflow:
     1/ Run iprscan, using the db/prot/<species>_prot.fasta as input
        iprscan produces an xml file
     2/ Download the software iprscan2annotation from http://lipm-bioinfo.toulouse.inra.fr/download/iprscan2annotation/iprscan2annotation.tgz
        Install the software and follow the instructions in the REDAME file to convert the xml file to an embl-like file
        Let's call the result annotation.txt
     3/ Execute this script using the command:
        ipr2fasta.pl --species <species> --ipr annotation.txt
     4/ Execute the commands:
        nc_get_data.pl --cfg <species> --only_formatdb
        nc_install.pl --only_merge_db

=cut

use strict;
use warnings;

use lib $ENV{NARCISSE_ROOT} . "/Core/lib";
use lib $ENV{NARCISSE_ROOT} . "/Core/API";

use ParamParser;
use FastaReader;
use Data::Dumper;
use File::Basename;
use IO::File;
use Error qw( :try );
use LipmError;
use Configuration::InstanceObserver;
use ConfigurationFactory;
use NarcisseDataWriter::OnDisk;
use General;
use Bio::DB::EMBL;


MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line() . "\n";
		$str   .= $o_err->stacktrace() . "\n";
		$str .= $o_err->Explain() . "\n";
		print STDERR $str;
		exit 1;
	};

	exit( 0 );
}

sub Main
{
	ParamParser::SetDefaultUsage(\&Usage);
	my $o_param = New ParamParser('GETOPTLONG','species:s','ipr:s','help');
	$o_param->AssertDefined( 'species','ipr');
	my ($species,$ipr) = $o_param -> Get ('species','ipr' );
	my $o_io = ConfigurationFactory::GetConfiguration();
	my $o_sp = $o_io -> GetSpeciesObserver( $species );
	
	my $fasta     = $o_sp -> GetProtFastaFile();
	my $fasta_rep = $o_sp -> GetProtFastaFile( -rep => &TRUE );

	print "LECTURE DU FICHIER fasta $fasta\n";
	my $o_fr = New FastaReader( $fasta );
	my $ra_headers = $o_fr -> GetHeaders();
	
	print "LECTURE DU FICHIER fasta_rep $fasta_rep\n";
	my $o_fr_rep = New FastaReader( $fasta_rep );
	my $ra_rep_headers = $o_fr_rep -> GetHeaders();

	# Reading the prot/* files, protein_id is used as a primary key
	# The SEMAE prot_id should be found in the fasta file (>prot:$protein_id)
    my $ra_chromosomes = $o_sp -> GetChromsNames();	
    my %h_prot_data;
	my $o_ndr     = New Bio::Narcisse::Utils::NarcisseDataReader();
	foreach my $chr (@$ra_chromosomes)
	{
		my $prot_file = $o_sp -> GetProtFiles( -chrom => $chr );
		print "READING THE FILE $prot_file\n";
		$o_ndr -> Prepare ( -from => $prot_file );
		my $rh_data = $o_ndr->FetchallHashref( 'protein_id' );
		$h_prot_data{$chr} = $rh_data;
	}
	
	print "READING THE FILE $ipr\n";
	# he non, bioperl ne lit pas le swissprot toulousain
	my %h_annot;
	{
		my $fh_ipr = new IO::File ( $ipr ) || throw LipmError::IOException($ipr,'f');
		local $/ = '//';
		while (my $line = <$fh_ipr>)
		{
			my @a_l = split ("\n",$line);
			my ($id,$de)=('','');
			foreach my $l (@a_l)
			{
				$id=$1 if ($l =~ /^ID   (.+)$/);
				$de=$1 if ($l =~ /^DE   (.+)$/);
			}
			next if ($id eq '');
			$h_annot{$id} = $de;
		}
	}	
	my $fasta_out = $fasta . '.new';
	print "WRITING THE FILE $fasta_out\n";
	my $fh_output = new IO::File ( ">$fasta_out" );
	foreach my $h (@$ra_headers)
	{
		my @a_fields = split ( ' ', $h);
		my ($protein_id,$chr);
		foreach my $f (@a_fields)
		{
			if ($f =~ /^>prot:(.+)$/)
			{
				$protein_id = $1;
				next;
			}
			if ($f =~ /chr=(.+)/)
			{
				$chr = $1;
				next;
			}
		}

		my $desc = defined ($h_annot{$protein_id}) ? $h_annot{$protein_id} : "";
		$desc =~ s/ *AutoIPR: *//;
		if ( exists ($h_prot_data{$chr}->{$protein_id}) )
		{
			$h_prot_data{$chr}->{$protein_id}->{ 'iprscan' } = $desc;
		};
		if ($desc ne "")
		{
			$desc =~ s/IPR\d{6}://g;
			$desc = " AutoIPR: $desc";
		}
		
		# Virer l'ancienne annotation interpro si il y a lieu
		my $new_h = $h;
		$new_h =~ s/ *AutoIPR:.+$//;
		
		print $fh_output "$new_h$desc\n";
		print $fh_output $o_fr->GetRawSequence ( $h, 1 );
	}

	# je sais c'est du copier-coller
	my $fasta_rep_out = $fasta_rep . '.new';
	print "WRITING THE FILE fasta $fasta_rep_out\n";
	my $fh_output_rep = new IO::File ( ">$fasta_rep_out" );
	foreach my $h (@$ra_rep_headers)
	{
		my @a_fields = split ( ' ', $h);
		my ($protein_id,$chr);
		foreach my $f (@a_fields)
		{
			if ($f =~ /^>prot:(.+)$/)
			{
				$protein_id = $1;
				last;
			}
		}
		my $desc = defined ($h_annot{$protein_id}) ? $h_annot{$protein_id} : "";
		$desc =~ s/AutoIPR: //;
		if ($desc ne "")
		{
			$desc =~ s/IPR\d{6}://g;
			$desc = " AutoIPR: " . $desc;
		}
		
		# Virer l'ancienne annotation interpro si il y a lieu
		my $new_h = $h;
		$new_h =~ s/ *AutoIPR:.+$//;
				
		print $fh_output_rep "$new_h$desc\n";
		print $fh_output_rep $o_fr_rep->GetRawSequence( $h, 1 );
	}

	$o_fr = undef;
	$o_fr_rep = undef;
	`mv $fasta ${fasta}.old`;
	`mv $fasta_rep ${fasta_rep}.old`;
	`mv $fasta_out $fasta`;
	`mv $fasta_rep_out $fasta_rep`;
	
	my @a_headers = qw( chrom element locus_tag gene_name protein_id len strand consistency product iprscan );
	foreach my $chr (@$ra_chromosomes)
	{
		my $prot_file = $o_sp -> GetProtFiles( -chrom => $chr );
		print "WRITING THE FILE $prot_file\n";
		my $o_ndw = New NarcisseDataWriter::OnDisk($prot_file, \@a_headers, { sort_numerical => &FALSE }, [ 'element' ] );
		$o_ndw -> Update ( $h_prot_data{ $chr } );
	}

	print "PLEASE EXECUTE NOW nc_get_data.pl --cfg $species --only_formatdb\n";
	print "AND WHEN EVERY SPECIES WILL BE COMPUTED EXECUTE nc_install.pl --only_merge_db\n";
	
	exit 0;
}

sub Usage{
	my $o_param = shift;
	print basename($0) ." --species <nom court> --ipr <annotations ipr> \n";
	print "Integrates the ipr annotations to the files: db/prot/*.fasta and prot/* \n";
	exit 1;
}
