#!/usr/bin/perl

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=pod

=head1 NAME

 nc_analyse_snp

=head1 SYNOPSIS

	obsolete : SnpIntegrator debenu SequencesLocator::Snp
	je pense que get_snp_data le remplace.

 perl nc_analyse_snp @args [ @opts ]

 @args:
  --species_ref : species name (short or long) to use as a reference.
  --snp_file    : fasta file of SNPs to seek in the reference.
  --output      : string, dirpath, where to store produced data. [temporaire, futur stockee dans data/species/]

 @opts
  --glint_options : string, any options you want to add to glint program.

=head1 DESCRIPTION

 This script launch glint with snp_file as query and the bank relative to the
 reference.
 Results are stored into output in NarcisseData format.

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 22-07, 2009
 $Id: nc_analyse_snp.pl 3168 2010-01-12 16:47:04Z sletort $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;
use FindBin;
use File::Spec;

use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../lib/snp";
use lib "$FindBin::Bin/../API";

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;
use SnpIntegrator;
use SnpManager;
use ConfigurationFactory;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3168 $ =~ /\d+/g); $r[0]};

use constant SPECIES     => 'species_ref';
use constant SNP_FILE    => 'snp_file';
use constant OUTPUT_DIR  => 'output';
use constant INPUT_FILE  => 'input';
use constant PROG_OPTIONS=> 'glint_options';

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;
$Error::Debug = &DEBUG;

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{
			warn( $o_err->Explain(), "\n" );
			warn $o_err->stacktrace(); 
		}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $o_log = New Logger( 'nc_analyse_snp.log', 'Is' );
	my $o_io  = &ConfigurationFactory::GetConfiguration( -o_logger => $o_log );

	my $rh_params = &GetParameters();

	my @a_opts = ( &SNP_FILE, &SPECIES, &INPUT_FILE );
	my( $snp_file, $species, $input ) = @$rh_params{ @a_opts };

	LipmError::ParameterException::Assert( &SNP_FILE . ' option' => $snp_file,
	                                       &SPECIES  . ' option' => $species );

	my $o_species = $o_io->GetSpeciesObserver( $species );
	my $rh_snps = {};

	if( !defined $input )
	{
		$rh_snps = &SearchForSnps( $rh_params, $o_log, $o_species );

		# the loop on chrom is just an accomodation with data
		# SnpManager could deal with all snps from all chroms.
		foreach my $rh_snp_for_one_chrom ( values %$rh_snps )
		{
			my $o_snp_man = New SnpManager( -o_logger => $o_log,
											o_species => $o_species,
											snp       => $rh_snp_for_one_chrom );
			$o_snp_man->CheckForImpact();
		}
	}
	else
	{
		my $o_snp_man = New SnpManager( snp       => $input,
										o_species => $o_species,
										-o_logger => $o_log );
		$o_snp_man->CheckForImpact();
	}


	return;
}


sub SearchForSnps
{
	my( $rh_params, $o_log, $o_species ) = @_;

	my @a_opts = ( &SNP_FILE, &PROG_OPTIONS, &OUTPUT_DIR );
	my( $snp_file, $options, $output ) = @$rh_params{ @a_opts };

	# identification des *VRAIS* SNPs (au sens Narcisse)
	my $o_snp_int = New SnpIntegrator( -o_logger => $o_log,
									file      => $snp_file,
									o_species => $o_species );

	$o_snp_int->LaunchGlint( glint_opt => $options );

	my $rh_snps = $o_snp_int->GetSnpData();

	# store data in Narcisse SNP format
	my $o_io = $o_species->GetInstanceObserver();
	my( $mask, $ext ) = $o_io->GetCfg( qw( MASK_FILE_CHROM PERSONAL ) );
	my @a_header = qw( chr element start end value label );
	foreach my $chrom ( keys %$rh_snps )
	{
		my $filepath = $o_species->GetSnpFiles(	-chrom => $chrom,
												-check => &FALSE,
												-where => 'INSTANCE' );
warn "seront enregistrees dans $filepath.";
# 		&NARCISSE_utils::AppendNarcisseData( $filepath, \@a_header, $$rh_snps{ $chrom } );
		my $filename = "$mask$chrom.$ext";
		$filepath = File::Spec->catfile( $output, $filename );
# 		&NARCISSE_utils::AppendNarcisseData( $filepath, \@a_header, $$rh_snps{ $chrom } );
		&NARCISSE_utils::UpdateNarcisseData( $filepath, \@a_header, $$rh_snps{ $chrom }, &FALSE );
	}

	# suppression des fichiers crees par SnpManager
	unlink( qw( snp_glint.out snp_input.fasta ) );

	return $rh_snps;
}


=head2 GetParameters

 Usage        : my $rh_params = $o_x->GetParameters()
 Function     : Read parameters from command line and return them formatted as a hash ref.
 Returns      : $rh_params, hash ref with keys ...
 Args         : none

=cut

sub GetParameters
{
	ParamParser::SetDefaultUsage( sub { pod2usage(1) } );

	my @a_keys   = qw(  help );
	push( @a_keys,
				&SPECIES      . '=s',
				&SNP_FILE     . '=s',
				&OUTPUT_DIR   . '=s',
				&INPUT_FILE   . '=s',
				&PROG_OPTIONS . '=s' );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );

	# valeur par defaut
	$o_param->SetUnlessDefined( &PROG_OPTIONS, '' );
	$o_param->SetUnlessDefined( &OUTPUT_DIR, '.' );

	my $rh_params = {};
	$o_param->Dump( 'HASH', $rh_params );

	return $rh_params;
}



__END__
