#!/usr/bin/perl
#
# $Id: nc_check_dependencies.pl 3464 2010-08-24 13:05:37Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

BEGIN
{
		use File::Basename;
		my $dirprg=dirname($0);
		unshift @INC, "$dirprg/../lib";
		unshift @INC, "$dirprg/../API";
}

use strict;
use warnings;
use Cwd;
use Error qw(:try);
use IO::File;
use Term::ANSIColor;

#use constant UNIX_SVN => '/usr/bin/svn';
#use constant URL_SVN  => 'http://lipm-svn.toulouse.inra.fr/svn/narcisse';
use constant UNIX_RM  => '/bin/rm -rf';
use constant UNIX_TAR => '/bin/tar';
use constant UNIX_LN => '/bin/ln';
use constant UNIX_UNAME => '/bin/uname';
use constant UNIX_MAKE  => '/usr/bin/make';

use constant BLAST_RELEASE => '2.2.21';


# Main program
MAIN:
{

    &CheckRequiredUnixUtilities();
    
    my @a_generic_modules  = qw( DBD::mysql CGI IO::Uncompress::Gunzip IO::Uncompress::AnyUncompress Data::Dumper Error Class::XML XML::Simple XML::Twig GD );
    my %h_generic_methods  = (
        DBD_mysql => "You could install this module with \"cpan Bundle::DBD::mysql\" (but you must have installed first the mysql development files)",
        CGI => "You could install this module with \"cpan CGI\"",
        IO_Uncompress_Gunzip => "You could install this module with \"cpan IO::Uncompress::Gunzip\"",
        IO_Uncompress_AnyUncompress => "You could install this module with \"cpan IO::Uncompress::AnyUncompress\"",
        Data_Dumper => "You could install this module with \"cpan Data::Dumper\"",
        Error => "You could install this module with \"cpan Error\"",
        Class_XML => "You could install this module with \"cpan Class::XML\" (but you must have installed first the expat development files) ",
        XML_Simple => "You could install this module with \"cpan XML::Simple\"",
        XML_Twig => "You could install this module with \"cpan XML::Twig\"",
        GD => "You could install this module with \"cpan GD\" (but you must have installed first the libgd 2.0.28 or higher, development files) "
    );
    
    my @a_circos_modules = qw( Clone Config::General Math::Bezier Math::Round Math::VecStat Params::Validate Set::IntSpan Statistics::Descriptive );
    my %h_circos_methods = (
        Clone => "You could install this module with \"cpan Clone\"",
        Config_General => "You could install this module with \"cpan Config::General\"",
        Math_Bezier => "You could install this module with \"cpan Math::Bezier\"",       
        Math_Round => "You could install this module with \"cpan Math::Round\"",
        Math_VecStat => "You could install this module with \"cpan Math::VecStat\"",
        Params_Validate => "You could install this module with \"cpan Params::Validate\"",
        Set_IntSpan => "You could install this module with \"cpan Set::IntSpan\"",
        Statistics_Descriptive => "You could install this module with \"cpan Statistics::Descriptive\"");
    
    my @a_optional_modules  = qw(  FCGI CGI::Fast Inline );

    my @a_bioperl_modules = qw( Bio::Annotation::DBLink Bio::LocatableSeq Bio::PrimarySeqI Bio::Root::Root
        Bio::Seq Bio::SeqFeature::FeaturePair Bio::SeqFeature::Generic
        Bio::SeqFeatureI Bio::SeqIO Bio::SimpleAlign Bio::Species Bio::Tools::CodonTable );
    
    my $ra_missing_modules;

    print "\nNOW Checking for the presence of Perl general use required modules\n";
    $ra_missing_modules = &CheckRequiredPerlModules(\@a_generic_modules);
    if (@$ra_missing_modules > 0)
    {
        print color 'bold red';
        print "\nERROR - Please install the following perl modules:\n";
        foreach my $m (@$ra_missing_modules)
        {
            my $k = $m;
            $k =~s/::/_/g;
            print $m . "\t" . $h_generic_methods{$k} . "\n";
        }
        print "\nPlease correct those problems and try again\n";
        print color 'reset';
        exit 1;
    }

    print "\nNOW Checking for the presence of modules required by circos\n";
    $ra_missing_modules = &CheckRequiredPerlModules(\@a_circos_modules);
    if (@$ra_missing_modules > 0)
    {
        print color 'bold red';
        print "\nERROR - Please install the following perl modules:\n";
        foreach my $m (@$ra_missing_modules)
        {
            my $k = $m;
            $k =~s/::/_/g;
            print $m . "\t" . $h_circos_methods{$k} . "\n";
        }
        print "\nPlease correct those problems and try again\n";
        print color 'reset';
        exit 1;
    }
    
    print "\nNOW Checking for the presence of BioPerl required modules\n";
    $ra_missing_modules = &CheckRequiredPerlModules(\@a_bioperl_modules);
    if (@$ra_missing_modules > 0)
    {        
        print color 'bold red';
        print "\nERROR - Please install the following bioperl modules:\n";
        print join("\n",@$ra_missing_modules) . "\n";
        print "To install bioperl, you could try:\n";
        print "cpan\n";
        print "cpan[1]> force install CJFIELDS/BioPerl-X.Y.Z.tar.gz\n"; 
        print "However, you are not required to install ALL bioperl. \n";
        print "You may restrict the installation to the groups: utilities, seq, Bio-SeqFeature-Store\n";
        print "Please correct those problems and try again\n";
        print color 'reset';
    }
    
    print "\nNOW checking for the presence of optional perl modules\n";
    $ra_missing_modules = &CheckRequiredPerlModules(\@a_optional_modules);
    if (@$ra_missing_modules > 0)
    {
        print color 'yellow';
        print "\nWARNING - some modules are NOT required, but if installed the performances will be enhanced:\n";
        print join("\n",@$ra_missing_modules) . "\n";
        print "You may want to install those modules anyway\n";
        print color 'reset';
    }
}


sub CheckRequiredPerlModules
{
    my ($ra_tocheck) = @_;
    
    my @a_bad_modules;
    foreach my $m ( @$ra_tocheck )
    {
        my $to_eval = "use $m";
        eval $to_eval;
        if ($@ ne "")
        {
            print color "yellow";
            print "$m => $@\n";
            push(@a_bad_modules, $m);
        }
        else
        {
            print color "green";
            print "$m => Success\n";
        }
    }
    print color "reset";
    return \@a_bad_modules;
}


sub CheckRequiredUnixUtilities
{
    my $cfg_dir = $0;
    $cfg_dir =~s|(.+)/[^/]+$|$1|;       # /path/to/narcisse/Core/sbin/nc_check_dependencies ==> /path/to/narcisse/Core/sbin/
    $cfg_dir =~s|(.+)/[^/]+$|$1|;       # /path/to/narcisse/Core/sbin/                      ==> /path/to/narcisse/Core/
    $cfg_dir =~s|(.+)/[^/]+$|$1|;       # /path/to/narcisse/Core                            ==> /path/to/narcisse/
    $cfg_dir .= '/Site.distrib/cfg';
    
    my $sav_dir = cwd;
    chdir $cfg_dir or die "ERROR - Cannot find the Site.distrib/cfg directory.\n";
    
    # Read the unix_cmd.cfg file
    my %h_unix_cmds;
    my $file_unix_cfg = 'unix_cmd.cfg';
    my $fh_unix_cfg = new IO::File($file_unix_cfg) or die "ERROR - Cannot open $file_unix_cfg";
    while (my $line =<$fh_unix_cfg>)
    {
        chomp $line;
        next if ($line =~ /^#/);
        my ($key,$value) = split('=',$line);
        if (defined($key) and defined($value))
        {
            $key   =~ s/^\s+|\s+$//g;     # remove leading and trailing blanks
            $value =~ s/^\s+//g;          # remove leading blanks
            $value =~ s/^([^ ]+) .*$/$1/; # /bin/rm -r ==> /bin/rm
            $h_unix_cmds{$key}=$value;
        }
    }
    
    # Search every program found here
    # NOTE - We do not try to execute them, let's hope their behaviour is correct !
    my $pb_flg=0;
    foreach my $key (keys %h_unix_cmds)
    {
        my $cmd = $h_unix_cmds{$key};
        if ( -x $cmd )
        {
            print color 'green';
            print "OK - found $cmd\n";
        }
        else
        {
            $pb_flg=1;
            print color 'bold red';
            print "ERROR - $cmd (see $key) is NOT available on this system\n"
        }
    }
    chdir $sav_dir;
    if ($pb_flg)
    {
        print color 'bold red';
        print "Please correct those problems and try again\n";
    }
    print color "reset";
}
    
    
sub Usage
{
    print "\nnc_install_narcisse.pl\n";
    print "Install the Narcisse-related applications\n";
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

