#!/usr/bin/perl
#
# $Id: nc_configure.pl 3865 2014-05-28 15:18:04Z  $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

#BEGIN
#{
#		use File::Basename;
#		my $dirprg=dirname($0);
#		unshift @INC, "$dirprg/../lib";
#		unshift @INC, "$dirprg/../API";
#}

use strict;
use warnings;
use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";
#use lib "$FindBin::Bin/../API/Bio/Narcisse/Utils";

use Data::Dumper;
use File::Basename;
use File::Spec::Functions qw(rel2abs);
use IO::File;
use JSON;
use Error qw( :try );
use Carp;
use Cwd;

use ParamParser;
use Cwd qw( abs_path getcwd );
use URI;
use NARCISSE_version;
use NarcisseError;
use ConfigurationFactory;
use Configuration::InstanceObserver;
use Configuration::NarcisseObserver;
use cgilib::WebBuilderFactory;
use WebBuilder;
use General;
use NARCISSE_utils;
use Logger;
use Sufine;
use Runner;

my $VERSION = do {my @r = (q$Rev: 3865 $ =~ /\d+/g); $r[0]};

use constant UNIX_CP => "/bin/cp -a";
use constant UNIX_MV => "/bin/mv";
use constant UNIX_RM => "/bin/rm -r";
use constant UNIX_BASH => "/bin/bash";
use constant UNIX_GREP => "/bin/grep";

#use constant

# Main program
MAIN:

try
{
    $Error::Debug = 1;
    Main();
}
catch LipmError with
{
     my ( $o_error ) = @_;
     print "ERROR !!! - " . $o_error->Explain() . "\n";
     if ($Error::Debug == 1)
     {
         print $o_error->stacktrace();
     }
}
catch NarcisseError with
{
	my ( $o_error ) = @_;
	print $o_error->bd . "\n";
};

sub Main
{

    # Parsing the command file
    &ParamParser::SetDefaultBehaviour('use_substitution_table');
    
    #&ParamParser::SetDefaultUsage(\&Usage);
    &ParamParser::SetDefaultBehaviour('use_exceptions');
    my $o_param = New ParamParser('ENV',\&Usage);


    # Update from ENV, for NARCISSE_CFG
    $o_param->Update('ENV','O');
    $o_param->Update('GETOPTLONG','O','new_instance:s','edit_instance:s','show_instance:s','remove_instance:s','upgrade_instance:s','rebuild_web','clear_cache','motd','work_in_progress','quiet','help');
    if (($o_param->IsDefined('new_instance') or $o_param->IsDefined('edit_instance') or $o_param->IsDefined('show_instance') or $o_param->IsDefined('remove_instance') or $o_param->IsDefined('upgrade_instance')))
    {
        if ($o_param->IsDefined('ROOT_SBIN') == &FALSE)
        {
            die "ERROR - Please cd to the Narcisse root directory, then source .setenv and try again\n";
        }
        if ($o_param->IsDefined('NARCISSE_CFG') == &TRUE)
        {
            die "ERROR - Please type \"chins\" and try again\n";
        }
        if ($o_param->IsDefined('new_instance'))
        {
            &ConfigureNewInstance($o_param->Get('new_instance'));
        }
        elsif ($o_param->IsDefined('edit_instance'))
        {
            &ReconfigureInstance($o_param->Get('edit_instance'));
        }
        elsif ($o_param->IsDefined('show_instance'))
        {
            &ShowInstance($o_param->Get('show_instance'));
        }
        
        elsif ($o_param->IsDefined('remove_instance'))
        {
            &RemoveInstance($o_param->Get('remove_instance'));
        }
        elsif ($o_param->IsDefined('upgrade_instance'))
        {
            &UpgradeInstance($o_param->Get('upgrade_instance'));
        }
   }
    else
    {
        $o_param->AssertDefined('NARCISSE_CFG');
        $o_param->AssertFileExists('NARCISSE_CFG');
        my $o_conf = ConfigurationFactory::GetConfigurationNoSingleton();
        if ($o_param->Get('rebuild_web'))
        {
            &ConfigureRebuildAndClear($o_conf,&TRUE);
        }
        elsif ($o_param->Get('clear_cache'))
        {
            &ConfigureRebuildAndClear($o_conf,&FALSE);
        }
        elsif ($o_param->Get('motd'))
        {
            &ConfigureMotd($o_conf,$o_param);
        }
        elsif ($o_param->Get('work_in_progress'))
        {
            &ConfigureWorkingMessage($o_conf,$o_param);
        }
        else
        {
            &ConfigureInstance($o_param,$o_conf);
        }
    }
}

=head2 Procedure ConfigureNewInstance

 Title    : ConfigureNewInstance
 Procedure: Configure a new instance, creating its directory if necessary
 Arg      : instance name

=cut

sub ConfigureNewInstance
{
    my ($instance_name) = @_;

    my $current_directory = abs_path(getcwd());

    # We have some things to check - Die if check is not good
	print "using Narcisse v" . $NARCISSE_version::NARCISSE_VERSION . "\n";
	print "checking your distribution directory...\n";
	&__CheckCurrentDirectory($current_directory);
	print "...looks good\n";

    # Ask the instance name if necessary
    if ($instance_name eq "")
    {
        $instance_name = &Ask ("What is the instance name ? ","",\&Blank);
    };

    # Create the directory if not already created (it may be a directory or a link)
    print "creating the instance directory\n";
    &__CreateInstanceDir($current_directory,$instance_name);

    # Common part of the job
    &InstanceBaseConfiguration($current_directory,$instance_name);
}

=head2 Procedure ReconfigureInstance

 Title    : ReconfigureInstance
 Procedure: Configure an already configured instance
 Arg      : instance name

=cut

sub ReconfigureInstance
{
    my ($instance_name) = @_;

    my $current_directory = abs_path(getcwd());

    # We have some things to check - Die if check is not good
    print "checking your distribution directory...\n";
    &__CheckCurrentDirectory($current_directory);
    print "...looks good\n";

    # Ask the instance name if necessary
    if ($instance_name eq "")
    {
        $instance_name = &Ask ("What is the instance name ? ","",\&Blank);
    };

    # the directory should exist already
    my $instance_dir = File::Spec->catfile($current_directory,$instance_name);
    unless (-d $instance_dir or -l $instance_dir)
    {
        die "ERROR - The instance directory does not exists\n";
    }

    # Common part of the job
    try
    {
        &InstanceBaseConfiguration($current_directory,$instance_name);
    }
    catch NarcisseError with
    {
        my ( $o_error ) = @_;
        print "WARNING !!! - PROGRAM VERSION = " . $o_error->pv . ' - INSTANCE VERSION = '.$o_error->cv."\n";
        print "Trying to upgrade\n";
        &UpgradeToVersion($o_error->cv,$o_error->pv,$current_directory,$instance_name);
        print "Upgrade done\n";
        exit;
    };
}

=head2 Procedure UpgradeInstance

 Title    : UpgradeInstance
 Procedure: Upgrade an instance to the current Narcisse version, without reconfiguring anything
 Arg      : instance name

=cut
sub UpgradeInstance
{
    my ($instance_name) = @_;
    
    my $current_directory = abs_path(getcwd());
    
    # We have some things to check - Die if check is not good
    print "checking your distribution directory...\n";
    &__CheckCurrentDirectory($current_directory);
    print "...looks good\n";
    
    # Ask the instance name if necessary
    if ($instance_name eq "")
    {
        $instance_name = &Ask ("What is the instance name ? ","",\&Blank);
    };
    
    # the directory should exist already 
    my $instance_dir = File::Spec->catfile($current_directory,$instance_name);
    unless (-d $instance_dir or -l $instance_dir)
    {
        die "ERROR - The instance directory does not exists\n";
    }

    try
    {
        # Create an o_conf object and retrieve the parameters - should throw an error !
        $ENV{NARCISSE_CFG} = File::Spec->catfile($current_directory,$instance_name,'Site','cfg','narcisse.cfg');
        my $o_conf = ConfigurationFactory::GetConfigurationNoSingleton();
        
        # no error - Nothing to upgrade
        print "Nothing to upgrade, so nothing done\n";
    }
    catch NarcisseError with
    {
        my ( $o_error ) = @_;
        print "WARNING !!! - PROGRAM VERSION = " . $o_error->pv . ' - INSTANCE VERSION = '.$o_error->cv."\n";
        print "Trying to upgrade\n";
        &UpgradeToVersion($o_error->cv,$o_error->pv,$current_directory,$instance_name);
        print "Upgrade done\n";
    };
    exit;
}

=head2 Procedure UpgradeToVersion

 Title    : UpgradeToVersion
 Procedure: Upgrade things to the version passed by parameters
 Arg      : Config version, Program version, current directory, instance name

=cut
sub UpgradeToVersion{
    my ($cv,$pv,$current_directory,$instance_name) = @_;

    #RELEASE 2.0.1 = the UPGRADE IS ONLY POSSIBLE FROM A 2.0.X VERSION
    if ($cv lt '2.0.0')
    {
        die "ERROR - CANNOT UPGRADE FROM $cv\n";
    }
    elsif ($pv =~ /^2\.0/)
    {
		&CopyFromDistrib($instance_name,'defaults.cfg');
		&CopyFromDistrib($instance_name,'unix_cmd.cfg');
    }
    elsif ($pv =~ /^2\.[123]/ )
    {
		&CopyFromDistrib($instance_name,'colors.list');
		&CopyFromDistrib($instance_name,'graphics.cfg');
		&CopyFromDistrib($instance_name,'defaults.cfg');
		&CopyFromDistrib($instance_name,'unix_cmd.cfg');
        &CopyFromCgibin($instance_name,'narcisse.cgi');
		&Upgrade1NarcisseCfg($instance_name);
	}
}

=head2 Procedure CopyFromDistrib

 Title    : CopyFromDistrib
 Procedure: Copy the file passed by parameter from Site.distrib/cfg to Site/cfg
            This is often done when upgrading Narcisse
 Arg      : $instance_name: The instance name
			$file_name    : The file name

=cut
sub CopyFromDistrib
{
    my ($instance_name,$file_name) = @_;
    my $unix_cp = &UNIX_CP;
    my $unix_mv = &UNIX_MV;
    `$unix_mv $instance_name/Site/cfg/$file_name $instance_name/Site/cfg/${file_name}.bkp`;
    `$unix_cp Site.distrib/cfg/$file_name $instance_name/Site/cfg`;
}

=head2 Procedure CopyFromCgibin

 Title    : CopyFromCgibin
 Procedure: Copy the file from cgi-bin to the local cgi-bin directory, 
            if this directory exists (it could be a symlink)
            This is often done when upgrading Narcisse if the instance is used through fastcgi
 Arg      : $instance_name: The instance name
			$file_name    : The file name

=cut
sub CopyFromCgibin
{
    my ($instance_name,$file_name) = @_;
    my $unix_cp = &UNIX_CP;
    my $unix_mv = &UNIX_MV;
    if ( !-l "$instance_name/cgi-bin" )
    {
        `$unix_cp cgi-bin/$file_name $instance_name/cgi-bin`;
    }
}

=head2 Procedure Upgrade1NarcisseCfg

 Title    : Upgrade1NarcisseCfg
 Procedure: Insert the line NARCISSE_SNP_REPOSITORY inside the narcisse.cfg file, unless already here
            This symbol was introduced with the 2.1 release, narcisse.cfg had to be upgraded for the first time
 Arg      : $instance_name: The instance name

=cut
sub Upgrade1NarcisseCfg
{
	my ($instance_name) = @_;
	
	my $nar_cfg = "$instance_name/Site/cfg/narcisse.cfg";
    my $unix_mv = &UNIX_MV;
	my $unix_grep = &UNIX_GREP;
	my $lines   = `$unix_grep -c ^NARCISSE_SNP_REPOSITORY $nar_cfg`;
	chomp $lines;
	if ($lines eq '0')
	{
		my $nar_cfg_bkp = $nar_cfg . '.bkp';
		`$unix_mv $nar_cfg $nar_cfg_bkp`;
		my $fh_nar_cfg_bkp = new IO::File( $nar_cfg_bkp ) || throw LipmError::IOException( $nar_cfg_bkp, 'f' );
		my $fh_nar_cfg     = new IO::File( $nar_cfg,'w' ) || throw LipmError::IOException( $nar_cfg, 'f' );
		while (my $line=<$fh_nar_cfg_bkp>)
		{
			print $fh_nar_cfg $line;
			if ($line =~ /^NARCISSE_GENOME_REPOSITORY/)
			{
				print $fh_nar_cfg "NARCISSE_SNP_REPOSITORY=\n";
			}
		}
	}
}
		
=head2 Procedure InstanceBaseConfiguration

 Title    : InstanceBaseConfiguration
 Procedure: Configure, reconfigure or just show config of an instance
 Arg      : current directory, instance name

=cut

sub InstanceBaseConfiguration
{
    my ($current_directory,$instance_name) = @_;

    # Create an o_conf object and retrieve the parameters
    $ENV{NARCISSE_CFG} = File::Spec->catfile($current_directory,$instance_name,'Site','cfg','narcisse.cfg');
    my $o_conf = ConfigurationFactory::GetConfigurationNoSingleton(-light => &TRUE);

    # Update the narcisse.cfg file
    print "updating narcisse.cfg\n";
    &UpdateNarcisseCFG($o_conf,$instance_name);

    # Dump the configuration file (only useful for dev and debug)
    &DumpConfiguration($o_conf);

    # Update $o_conf
    $o_conf = '';
    $o_conf = ConfigurationFactory::GetConfigurationNoSingleton();

    # Create the data directory
    print "creating the data directory, only if necessary\n";
    &__CreateDataDir($o_conf);

    # Create the Web directory and check the hierarchy
    print "creating the Web directory, only if necessary\n";
    &__CreateWebDir($o_conf);

    # Propose fast cgi through fcgid, but not in multi (this is useless)
    my $fcgi_flg = &FALSE;
    if ($o_conf->GetName() ne 'multi')
    {
        print "Do you want to run this instance through the fcgid module (fast-cgi) ? (Y/[N])";
        $fcgi_flg = (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y') ? &TRUE : &FALSE;
    }

    # Create the Apache configuration file
    print "creating the Apache configuration\n";
    &__CreateApacheConf($o_conf,$fcgi_flg);

    print "Please restart the Apache server\n";

    # If in fcgid, remove the cgi-bin symlink and create a directory instead
    if ($fcgi_flg == &TRUE)
    {
        # Create the cgi-bin directory, unless already here
        # We DO NOT do any symlink for the cgi-bin directory, because this bothers the fcgid module
        print "Removing the cgi_bin symlink and replacing it with a directory, because we are in fast cgi mode\n";
        my $cgi_bin = File::Spec->catfile($current_directory,$instance_name,'cgi-bin');
        if ( -l $cgi_bin )
        {
            unlink $cgi_bin;
            my $unix_cp = &UNIX_CP;
            my $cmd = "$unix_cp " . File::Spec->catfile($current_directory,'cgi-bin') . " $instance_name ";
            `$cmd`;
        }
    }

    print "Base configuration terminated\n";

    # finished !
    return;

    # check the current directory, it must be a Narcisse distribution directory
    sub __CheckCurrentDirectory
    {
        my ($cur_dir) = @_;
        my @a_required_directories = qw( Core cgi-bin Site.distrib ApacheConf.distrib );
        my $msg="";
        foreach my $d (@a_required_directories)
        {
            $msg .= "ERROR - $d in the current directory should NOT be a symbolic link\n" if (-l $d);
            $msg .= "ERROR - the directory $d is NOT found in the current directory\n" if  (!-d $d);
        }
        die $msg if ($msg ne "");
    }

    # create a link if necessary
    sub __CreateLink
    {
        my ($current_dir,$instance_name,$link) = @_;
        my $src = File::Spec->catfile($current_dir,$link);
        my $dst = File::Spec->catfile($instance_name,$link);
        if (! -e $dst)
        {
            symlink ($src, $dst) or throw LipmError::IOException($dst,'d');
        }
    }

    # create a directory for this instance, unless already done
    sub __CreateInstanceDir
    {
        my ($current_dir,$instance_name) = @_;

        # This is required by Sufine (see $dir_regex in Sufine.pm)
        my $instance_dir = File::Spec->catfile( $current_dir, $instance_name );
        if ( $instance_dir !~ /^[a-zA-Z0-9_\/\-]+$/ )
        {
            die "ERROR - The Instance directory should have ONLY those characters: a-zA-Z0-9 _ and -\n";
        }

        # Create the directory
        if (! -e $instance_name)
        {
            mkdir $instance_name or throw LipmError::IOException($instance_name,'d');
        }
        else
        {
            die "ERROR - The directory $instance_dir already exists - Please use the switch --edit if you want to modify the configuration\n";
        }

        # Create the site directory
        my $site_dir = File::Spec->catfile($instance_name,'Site');
        my $unix_cp = &UNIX_CP;
        my $unix_rm = &UNIX_RM;
        if (! -e $site_dir)
        {
            `$unix_cp -r Site.distrib $site_dir`
        }
	
        # Create the Core, Doc, cgi-bin symlinks
        # We use a relative link if possible, an absolute link if necessary
        if ( -l $instance_name )
        {
            &__CreateLink($current_dir,$instance_name,'Core');
            &__CreateLink($current_dir,$instance_name,'Doc');
            &__CreateLink($current_dir,$instance_name,'cgi-bin');
            &__CreateLink($current_dir,$instance_name,'cgi-bina');
        }
        else
        {
            &__CreateLink('..',$instance_name,'Core');
            &__CreateLink('..',$instance_name,'Doc');
            &__CreateLink('..',$instance_name,'cgi-bin');
            &__CreateLink('..',$instance_name,'cgi-bina');
        }
    }

    # Create the data directory
    sub __CreateDataDir
    {
        my ($o_conf) = @_;
        my $unix_cp = $o_conf->GetUnixPath('UNIX_CP');

        my $data_dir = $o_conf->GetCfg('ROOT_DATA');
        mkdir $data_dir unless -e $data_dir;
        mkdir $o_conf->GetCfg('ROOT_DATA_DB');

        # If url.txt does not exist, copy it
        my $url_txt = $o_conf->GetCfg('ROOT_URL_TXT');
        if (! -r $url_txt)
        {
            my $src_url_txt = File::Spec->catfile('data.distrib','url.txt');
            `$unix_cp $src_url_txt $url_txt`;
        }
    }

    # Create the Web hierarchy
    sub __CreateWebDir
    {
        my ($o_conf) = @_;

        my $unix_cp = $o_conf->GetUnixPath('UNIX_CP');
        my $root_dir = $o_conf->GetCfg('ROOT');
        my $web_dir = $o_conf->GetCfg('ROOT_WEB');
        mkdir $web_dir;

        my $web_distrib = File::Spec->catfile(abs_path(getcwd()),'Web');

        # A lot of subdirectories are, in fact, symbolic links
        # TODO This does not work if some symbols (see RH_CSS, RH_DOCUMENTATION etc) are changed in defaults.cfg
        my @a_subdirs_ln = qw( css documentation fonts inc js logo narcisse_multi.xml narcisse.xml ajax.xml );
        if (-l $root_dir)
        {
            foreach my $subdir (@a_subdirs_ln)
            {
                &__CreateLink(File::Spec->catfile($web_distrib),$web_dir,$subdir);
            }
        }
        else
        {
            foreach my $subdir (@a_subdirs_ln)
            {
                &__CreateLink(File::Spec->catfile('..','..','Web'),$web_dir,$subdir);
            }
        }

        # The others must be copied
        my @a_subdirs_cp = qw( pictures );
        foreach my $subdir (@a_subdirs_cp)
        {
            my $src = File::Spec->catfile($web_distrib,$subdir);
            my $dst = $web_dir;
            `$unix_cp $src $dst` unless (-d File::Spec->catfile($web_dir,$subdir));
        };

        # Make the config.js file
        print "Making the Javascript and css Configuration\n";
        &MakeNarcisseAllJsCss($o_conf);

        # Clear the cache
        print "Clearing the cache\n";
        &ClearCache($o_conf);

        # Everything should be correct now
        print "Checking directories\n";
        &CheckDir($o_conf);
    }

    # Create the Apache configuration file for this instance
    sub __CreateApacheConf
    {
        my ($o_conf,$fcgi_flg) = @_;
        my $unix_cp = &UNIX_CP;

        my $d_apache = 'ApacheConf';
        if (! -e $d_apache)
        {
            mkdir $d_apache;
            my $cmd = $unix_cp . ' ' . File::Spec->catfile('ApacheConf.distrib','zzz-apache.conf') . ' ' . File::Spec->catfile($d_apache);
            `$cmd`;
            $cmd    = $unix_cp . ' ' . File::Spec->catfile('ApacheConf.distrib','aaa-apache.conf') . ' ' . File::Spec->catfile($d_apache);
            `$cmd`;
        }
        my $f_apache= File::Spec->catfile('ApacheConf',$o_conf->GetName().'.conf');
        my $http = $o_conf->GetCfg('HTTP');
        my $http_uri = new URI($http);

        #Propose web_service Apache configuration, ONLY in multi
        my $ws = &FALSE;
        if ($o_conf->GetName() eq 'multi')
        {
            print "Do you want to configure Apache for web_service utilisation ? (Y/[N])";
            $ws = (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y') ? &TRUE : &FALSE;
        }

        my $cgi = $o_conf->GetCfg('HTTP_CGI');
        my $cgi_uri = new URI($cgi);

        my $fh_apache= new IO::File (">$f_apache") or throw LipmError::IOException($f_apache);
        print $fh_apache "# DO NOT EDIT !!!\n# this is an automatically generated file\n";

        if ($ws == &TRUE)
        {
            my $playmoby        = $o_conf->GetCfg('HTTP_PLAYMOBY');
            my $playmoby_uri = new URI($playmoby);
            print $fh_apache "ScriptAlias " . $playmoby_uri->path() . '/cgi' . " \"" . $o_conf->GetCfg('ROOT_PLAYMOBY') . '/cgi' . "\"\n";
            print $fh_apache "Alias " . $playmoby_uri->path() . " \"" . $o_conf->GetCfg('ROOT_PLAYMOBY') . "\"\n";
        }

        print $fh_apache "ScriptAlias " . $cgi_uri->path() ." \"" . $o_conf->GetCfg('ROOT_CGI') . "\"\n";

        # The cgi-bin alt directory
        print $fh_apache "# The cgi-bin alternate directory is useful for some special incompatibilities with fcgi_mod\n";
        print $fh_apache "ScriptAlias " . $cgi_uri->path() ."a \"" . $o_conf->GetCfg('ROOT_CGI') . "a\"\n";

        print $fh_apache "Alias " . $http_uri->path() . " \"" . $o_conf->GetCfg('ROOT_WEB') . "\"\n";
        
        print $fh_apache "# Some security mechanism to forbid direct access to the program and data\n";
        print $fh_apache "<Location " . "\"" . $http_uri->path() . "\">\n";
        print $fh_apache "    Options -Indexes\n";
        print $fh_apache "</Location>\n";

        # If in fast cgi, add a couple of lines
        if ($fcgi_flg == &TRUE)
        {
            print $fh_apache "<IfModule fcgid_module>\n";
            print $fh_apache "   <Location " . $cgi_uri->path() . ">\n";
            print $fh_apache "      SetHandler fcgid-script\n";
            print $fh_apache "      Options ExecCGI\n";
            print $fh_apache "   </Location>\n";
            print $fh_apache "</IfModule>\n";
        }

        # If we are in multi, we add a redirection, this is more convenient than typing the whole URL
        if ($o_conf->GetName() eq 'multi')
        {
            print $fh_apache "RedirectMatch " . '^' . dirname($http_uri->path()) . '/?$' . " \"" . $o_conf->GetCfg('HTTP_NARCISSE_CGI') . "\"\n";
        }
        $fh_apache->close();
    }
}

=head2 Procedure RemoveInstance

 Title    : RemoveInstance
 Procedure: Remove an existing instance, removing the apache conf file and renaming the directory
 Arg      : instance name

=cut
sub RemoveInstance
{
    my ($instance_name) = @_;

    my $current_directory = abs_path(getcwd());

    # We have some things to check - Die if check is not good
    print "checking your distribution directory...\n";
    &__CheckCurrentDirectory($current_directory);
    print "...looks good\n";

    # Ask the instance name if necessary
    if ($instance_name eq "")
    {
        $instance_name = &Ask ("What is the instance name ? ","",\&Blank);
    };

    # Nothing to do if the directory does not exist
    unless (-d $instance_name or -l $instance_name)
    {
        die "ERROR - The instance $instance_name does not exists\n";
    };

    # Refuse to do anything if the instance is known by multi
    if (-e 'multi')
    {
        my $chins        = File::Spec->catfile($current_directory,'Core','sbin','chins.sh');
        my $nc_instances = File::Spec->catfile($current_directory,'Core','sbin','nc_instances.pl');

        my $cmd = &UNIX_BASH . " -c '. .setenv.sh; source $chins multi; $nc_instances'";
        my $registered_instances = `$cmd`;
        chomp $registered_instances;
        my @a_registered_instances = split(/ /,$registered_instances);
        if ( grep (/^$instance_name$/, @a_registered_instances) != 0)
        {
            print "ERROR - $instance_name is known by the multi instance\n";
            print "Please execute:\n    chins multi\n    nc_configure.pl\nand try again\n";
            exit 1;
        }
    }

    print "Are you SURE you want to remove the instance $instance_name ? (Y/[N])";
    if (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y')
    {
        my $unix_mv = &UNIX_MV;
        my $dst = $instance_name.'.REMOVED';

        # Create an o_conf object and retrieve the parameters
        $ENV{NARCISSE_CFG} = File::Spec->catfile($current_directory,$instance_name,'Site','cfg','narcisse.cfg');
        my $o_conf = ConfigurationFactory::GetConfigurationNoSingleton();

        # my $data_dir = $o_conf->GetCfg('ROOT_DATA');
        # my $data_dir_dst = $data_dir . '.REMOVED';

        if (-e $dst)
        {
            die "ERROR - The directory $dst already exists - CANCELLED\n";
        }
		
		# We DO NOT move the data directory, this too dangerous as datadirectory may be shared among several instances
		# (at least at the developers' site)
        # `$unix_mv $data_dir $data_dir_dst`;
        `$unix_mv $instance_name $dst`;
        my $f_apache= File::Spec->catfile('ApacheConf',$instance_name.'.conf');
        unlink($f_apache);

        print "The directory $instance_name should have been renamed\n";
        print "The configuration file $f_apache should have been removed\n";
        print "Please check and restart the Apache web server\n";
        print "\n";
        print "May be you should configure the multi instance again ?\n";
        exit(0);
    }
}

=head2 Procedure ConfigureMotd

 Title     : ConfigureMotd
 Usage     : &ConfigureMotd($o_conf)
 Procedure : Ask the user some lines for the motd, and write them to one of the motd files
 Arg       : $o_conf If NarcisseObserver, configure global motd, else configure instance motd

=cut

sub ConfigureMotd
{
    my ($o_conf,$o_param) = @_;

    if ($o_conf->isa('NarcisseObserver')==&TRUE)
    {
        print "YOU ARE NOW ENTERING THE Message Of The Day for the global Narcisse installation\n";
    }
    else
    {
        print "YOU ARE NOW ENTERING THE Message Of The Day for the Narcisse instance " . $o_conf->GetName() . "\n";
    }

    print "Enter 1 or more lines. Html tags allowed - Finish with an empty line\n";
    my @a_buffer;
    while (&TRUE)
    {
        my $line = <STDIN>;
        chomp($line);
        last if ($line eq '');
        push @a_buffer, $line;
    }
    my $buffer = join "\n",@a_buffer;
    
    my $o_wb = WebBuilderFactory::GetWebBuilder( o_io => $o_conf, o_param => $o_param );
    my @a_files_motd = $o_wb -> GetMotdFiles();
    my $file_motd    = $a_files_motd[ -1 ];		# Always updating the LAST file from the xml file
    my $fh_file_motd  = new IO::File(">$file_motd") or throw LipmError::IOException($file_motd,'f');
    print $fh_file_motd $buffer;
    $fh_file_motd->close();
}
=head2 Procedure ConfigureWorkingMessage

 Title     : ConfigureWorkingMessage
 Usage     : &ConfigureWorkingMessage($o_conf)
 Procedure : Ask the user some lines for explaning why the admin is working and this instance not available
 Arg       : $o_conf If NarcisseObserver, disable/enable global installation, else disable/enable only the current instance

=cut

sub ConfigureWorkingMessage
{
    my ($o_conf, $o_param) = @_;

    if ($o_conf->isa('NarcisseObserver')==&TRUE)
    {
        print "\nYOU ARE NOW DISABLING OR ENABLING the global Narcisse installation\n";
    }
    else
    {
        print "\nYOU ARE NOW DISABLING OR ENABLING the Narcisse instance " . $o_conf->GetName() . "\n";
    }

    print "TO DISABLE: Enter 1 or more lines explaining why the instance is disabled. Html tags allowed - Finish with an empty line\n";
    print "TO ENABLE: type enter only, removing any previously entered message\n";
    print "TO CANCEL OPERATION: Type CTRL-C\n";
    my @a_buffer;
    while (&TRUE)
    {
        my $line = <STDIN>;
        chomp($line);
        last if ($line eq '');
        push @a_buffer,$line;
    }
    
    my $buffer = join "\n",@a_buffer;

    my $o_wb = WebBuilderFactory::GetWebBuilder( o_io => $o_conf, o_param => $o_param );
    my @a_files_work = $o_wb -> GetDisabledFiles();
    my $file_work    = $a_files_work[ -1 ];		# Always updating the LAST file from the xml file
    my $fh_file_work  = new IO::File(">$file_work") or throw LipmError::IOException($file_work,'f');
    print $fh_file_work $buffer;
    $fh_file_work->close();
}

=head2 Procedure ConfigureRebuildAndClear

 Title    : ConfigureRebuildAndClear
 Usage    : ConfiguraRebuildAndClear($o_io,$flg)
 Procedure: If $flg is true: Rebuild Web directory (config.js, narcisse_all.js, narcisse_all.css), then clear the cache
            Else: only clear the cache
 Arg      : $o_io,$flg

=cut


sub ConfigureRebuildAndClear
{
    my ($o_io,$flg) = @_;

    if ($o_io->isa( 'InstanceObserver' ) == &FALSE)
    {
        die "The switches --rebuild and --clear work only in a Single Instance context";
    }

    if ($flg == &TRUE)
    {
        print "Making the css and Javascript Configuration\n";
        &MakeNarcisseAllJsCss($o_io);
       
    };

    print "Clearing the cache\n";
    &ClearCache($o_io);
}


=head2 Procedure ConfigureInstance

 Title    : ConfigureInstance
 Procedure: Configure an instance (may be multi)
 Arg      : $o_param The parameters (used only for debugging)
            $o_io, an Instance Observer

=cut

sub ConfigureInstance
{
    my ($o_param,$o_io) = @_;

    my $quiet_flg = ($o_param->IsDefined ('quiet')) ? 1 : 0;

    #
    # create the target directory and change to it
    #

    my $mkdir = $o_io->GetUnixPath('UNIX_MKDIR');
    my $sav_dir = getcwd();
    my $target = 'configure';
    my $target_root = File::Spec->catfile($o_io->GetCfg('ROOT_LOGS'),$target);
    `$mkdir -p $target_root`;

    my $target_path = Sufine::MkDir(-dir => $target_root,-name => $target);

    die unless(defined $target_path);
    die unless(defined Sufine::ChDir(-dir => $target_root,-name => $target));

    # Start logging
    my $o_logger = New Logger (">>narcisse.log",$o_io->GetCfg('LOG_LEVEL'));
    $o_logger->Trace("Starting nc_configure.pl\n","*","A");
    $o_io->SetLogger( $o_logger );

    &NARCISSE_utils::RedirectOE($o_io,"narcisse.out","narcisse.err");
    $o_param->Dump("narcisse.param");

    #
    # Print a welcome message
    #
    &NARCISSE_utils::Welcome($0,$VERSION,$o_io->GetCfgPath());

    # Print and log main parameters
    &PrintLogInfo($o_io);

    # Check the directories
    print "Checking directories\n";
    &CheckDir($o_io);

    # SingleInstance mode
    if ($o_io->isa('InstanceObserver') == &TRUE)
    {
        # Updating species.list
        print "Updating species.list\n";

        print "Do you want to create the species.cfg from the repository ? Y/[N] ";
        if (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y')
        {
            my $msg = "Enter the name of the file containing the species \n";
            $msg .= "(this file must be in the 'species' directory) ";
            my $f_list = &Ask($msg,'',\&Blank);
            $f_list = File::Spec->catfile($o_io->GetSpeciesCfgDirpath('INSTANCE'),$f_list);
            &UpdateSpeciesFromRepository($o_io,$f_list);
        }

        &UpdateSpeciesList($o_io,$quiet_flg);

        # Re-init the instance observer, because species.list may have changed
        # Recreate the serialized rh_cfg (for debug only)
        $o_io = New InstanceObserver( -o_logger => $o_logger );
        $o_io->InitURL();

        # Dump Narcisse.cfg
        &DumpConfiguration($o_io);

        print "Now creating config.js (useful for javascript)\n";
        &MakeNarcisseAllJsCss($o_io);

        # Regenerate update.bash
        print "Do you want to create a new version of update.bash [Y]/N ";
        if (&NARCISSE_utils::GetAdmAnswer("","Y|N","Y") eq 'Y') {
            print "Creating update.bash\n";
            &CreateUpdateScript($o_io);
        };
    }

    # MultiInstances mode
    else
    {
        # Updating species.list
        print "Updating the instance list\n";

        &UpdateInstanceList($o_io);

        # Re-init the narcisse observer, because some parameters may have changed
        $o_io = New NarcisseObserver( -o_logger => $o_logger );

        # Dump Narcisse.cfg
        &DumpConfiguration($o_io);
    }

    $o_logger->Trace("OK\n","*","A");
}

=head2 Procedure ShowInstance

 Title    : ShowInstance
 Procedure: Show parameters of an instance (may be multi)
 Arg      : $o_param The parameters (used only for debugging)
            $o_io, an Instance Observer

=cut

sub ShowInstance
{
    my ($instance_name) = @_;

    my $current_directory = abs_path(getcwd());

    # We have some things to check - Die if check is not good
    print "checking your distribution directory...\n";
    &__CheckCurrentDirectory($current_directory);
    print "...looks good\n";

    # Ask the instance name if necessary
    if ($instance_name eq "")
    {
        $instance_name = &Ask ("What is the instance name ? ","",\&Blank);
    };

    # the directory should exist already
    my $instance_dir = File::Spec->catfile($current_directory,$instance_name);
    unless (-d $instance_dir or -l $instance_dir)
    {
        die "ERROR - The instance directory does not exists\n";
    }

    # Common part of the job
    try
    {
		# Create an o_conf object and retrieve the parameters
		$ENV{NARCISSE_CFG} = File::Spec->catfile($current_directory,$instance_name,'Site','cfg','narcisse.cfg');
		my $o_conf = ConfigurationFactory::GetConfigurationNoSingleton(-light => &TRUE);

		&ShowNarcisseCFG($o_conf);
	}
	catch NarcisseError with
	{
		my ( $o_error ) = @_;
		print "WARNING !!! - PROGRAM VERSION = " . $o_error->pv . ' - INSTANCE VERSION = '.$o_error->cv."\n";
		print "Trying to upgrade\n";
		&UpgradeToVersion($o_error->cv,$o_error->pv,$current_directory,$instance_name);
		print "Upgrade done\n";
		exit;
	};
}


=head2 procedure UpdateInstanceList

 Title        : UpdateInstanceList
 Usage        : UpdateInstanceList($o_no)
 Fonction     : Main procedure for updating the NARCISSE_INSTANCES parameter
 Args         : $o_no, a NarcsseObserver

=cut
sub UpdateInstanceList
{
    my ($o_no) = @_;
    my $o_logger = $o_no->GetLogger();

    print "Do you want to update the instance list ?         Y/[N] ";
    if ( &NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y' )
    {
        my %h_prms;
        $h_prms{NARCISSE_INSTANCE_ROOTS} = $o_no->GetCfg('NARCISSE_INSTANCE_ROOTS');
        &UpdateInstanceListAutomatically($o_no, \%h_prms);
        &WriteNarcisseCfg($o_no->GetCfgPath(),\%h_prms);
        &CreatePalFiles($o_no);
        &PrintLogInfo($o_no);
    }
    else
    {
        $o_logger->Trace("NARCISSE_INSTANCES parameter not updated\n","*","A");
    }
}

=head2 procedure UpdateInstanceListAutomatically

 Title        : UpdateInstanceListAutomatically
 Usage        : UpdateInstanceListAutomatically($o_no,\%h_prms)
 Procedure    : Automatic detection of the instances, from the directory hierarchy
                We look for every directory starting from parent directory to discover instances
                The user is asked for confirmation and the key NARCISSE_INSTANCE_ROOTS from %h_prms is updated
 Returns      : none
 Args         : $o_no A NarcisseObserver
                $rh_prms A hash, its key NARCISSE_INSTANCE_ROOTS will be created/updated

=cut
sub UpdateInstanceListAutomatically
{
    my ($o_no,$rh_prms) = @_;
    my $o_logger = $o_no->GetLogger();

    # NOTE - The following works only because we are sure that multi is a real directory, NOT a symbolic link
    my $multi_dir = $o_no->GetCfg('ROOT');
    my $root_dir  = dirname($multi_dir);

    my @a_instances = ();
    my $choice="Y";   # default answer
    my $dh_root_dir = new IO::Dir($root_dir) or throw LipmError::IOException( $root_dir, 'd' );
    while (my $dir=$dh_root_dir->read())
    {
        next if $dir eq '.';
        next if $dir eq '..';
        next if $dir eq $multi_dir;
        next if $dir =~ /\.REMOVED/;
        #warn "KOUKOU $dir";
        if (-d File::Spec->catfile($root_dir,$dir))
        {
            my $narcisse_cfg = File::Spec->catfile($root_dir,$dir,'Site','cfg','narcisse.cfg');
            if (-r $narcisse_cfg)
            {
                #warn "KOUKOU2 $narcisse_cfg";
                next if ($dir eq 'multi');
                print "Select $dir [Y]/N: ";
                $choice= &NARCISSE_utils::GetAdmAnswer("","Y|N","Y");
                push ( @a_instances,$dir) if ($choice eq 'Y');
            }
        }
    }

    $rh_prms->{NARCISSE_INSTANCE_ROOTS} = join(',',@a_instances);
    $o_logger->Trace("Finished\n","*","A");
	return;
}

=head2 procedure CreatePalFiles

 Title        : CreatePalFiles
 Usage        : CreatePalfiles($o_no)
 Procedure    : Create the .pal files useful for a global blast
 Args         : $o_no A NarcisseObserver

=cut
sub CreatePalFiles
{
    my ($o_no) = @_;

    print "Creating the .pal files, for blasting through all the instances...\n";
    my ($db_cds_pal,$db_prot_pal) = $o_no->GetCfg('DB_CDS_MERGE','DB_PROT_MERGE');
    $db_cds_pal  =~ s/\.fasta$/.pal/;
    $db_prot_pal =~ s/\.fasta$/.pal/;
    $db_cds_pal  =  File::Spec->catfile($o_no->GetCfg('ROOT_DATA_DB'),$db_cds_pal);
    $db_prot_pal =  File::Spec->catfile($o_no->GetCfg('ROOT_DATA_DB'),$db_prot_pal);

    my $fh_db_cds_pal  = new IO::File(">$db_cds_pal") or throw LipmError::IOEXception($db_cds_pal,'f');
    my $fh_db_prot_pal = new IO::File(">$db_prot_pal") or throw LipmError::IOEXception($db_prot_pal,'f');

    print $fh_db_cds_pal  "#\n# DO NOT EDIT\n#\n# alias file created by nc_configure.pl for blast on " . &NARCISSE_utils::GetDate() . "\n#\n";
    print $fh_db_prot_pal "#\n# DO NOT EDIT\n#\n# alias file created by nc_configure.pl for blast on " . &NARCISSE_utils::GetDate() . "\n#\n";

    print $fh_db_cds_pal  "DBLIST ";
    print $fh_db_prot_pal "DBLIST ";

    my $rh_instances = $o_no->GetInstanceObservers();
    foreach my $key (keys %$rh_instances)
    {
        my $o_io = $rh_instances->{$key};
        my ($db_cds_merge,$db_prot_merge) = $o_io->GetCfg('DB_CDS_MERGE','DB_PROT_MERGE');
        $db_cds_merge  =  File::Spec->catfile($o_io->GetCfg('ROOT_DATA_DB'),$db_cds_merge);
        $db_prot_merge =  File::Spec->catfile($o_io->GetCfg('ROOT_DATA_DB'),$db_prot_merge);
        print $fh_db_cds_pal "$db_cds_merge "   if (-r $db_cds_merge);
        print $fh_db_prot_pal "$db_prot_merge " if (-r $db_prot_merge);
    }

    print $fh_db_cds_pal "\n";
    print $fh_db_prot_pal "\n";

    $fh_db_cds_pal->close();
    $fh_db_prot_pal->close();
}

=head2 procedure ClearCache

 Title        : ClearCache
 Usage        : &ClearCache()
 Prerequisite : none
 procedure    : Clear the cache
 Returns      : none
 Args         : $o_io: the Instance Observer

=cut

sub ClearCache
{
    my ($o_io) = @_;
    my $o_logger = $o_io->GetLogger();
    my $logger_flg = defined $o_logger;

    my $unix_find  = $o_io->GetUnixPath('UNIX_FIND');
    my $unix_wc    = $o_io->GetUnixPath('UNIX_WC');
    my $unix_xargs = $o_io->GetUnixPath('UNIX_XARGS');
    my $unix_rm    = $o_io->GetUnixPath('UNIX_RM');
    my $root_cache = $o_io->GetCfg('ROOT_CACHE');
    return unless (-d $root_cache);

    $o_logger->StartTimer("T1") if $logger_flg;
    my $cmd = "$unix_find $root_cache -type f | $unix_wc -l";
    $o_logger->Trace("$cmd\n","","ACV") if $logger_flg;

    my $count = `$cmd`;
    chomp($count);
    if ($count != 0)
    {
        $o_logger->Trace("Clearing the cache\n","T1","A") if $logger_flg;
        $o_logger->StartTimer("T1") if $logger_flg;
        $cmd = "$unix_find $root_cache -type f | $unix_xargs $unix_rm";
        $o_logger->Trace("$cmd\n","","ACV") if $logger_flg;
        `$cmd`;
        $o_logger->Trace("done\n","T1","A") if $logger_flg;
    }
    else
    {
        $o_logger->Trace("The cache is empty\n","T1","A") if $logger_flg;
    }
}

=head2 procedure MakeNarcisseAllJsCss

 Title        : MakeNarcisseAllJsCss
 Usage        : MakeNarcisseAllJsCss($o_io)
 Fonction     : Create the file Web/narcisse_all.js (concatenating all .js files)
 Returns      : none
 Args         : none

=cut
sub MakeNarcisseAllJsCss
{
	my ($o_io) = @_;
	$o_io->InitURL() if ($o_io->isa('InstanceObserver') == &TRUE);

	my $rh_cfg = $o_io->GetConfigHash();

    # Copy the configuration hash, because it is modified
    my %h_cfg = %$rh_cfg;

    # We do not need the GUI_DEFAULTS hash for the javascript
    delete $h_cfg{GUI_DEFAULTS};

    # Create a new GENOME hash, the keys are the long names
    # We throw an error if we get two times the same long name
    my %h_genome_new;
    my $h_genome_old = $h_cfg{GENOME};
    foreach my $k (keys %$h_genome_old)
    {
        my $rh_g_o = $h_genome_old->{$k}; # the old hash
        my %h_g_n  = ();                  # the new hash
        my $l      = $rh_g_o->{long};     # the new key is the long name

        # We copy only some keys, if possible
        foreach my $m ('URL','SYNT_URL','assembly','taxid','assembly_phase')
        {
            $h_g_n{$m}  = $rh_g_o->{$m} if (exists $rh_g_o->{$m});
        }
        $h_genome_new{$l} = \%h_g_n;
    }

    # Drop the old genome hash and replace it with the new
    $h_cfg{GENOME} = \%h_genome_new;

	# Create the config.js file from the modified configuration hash
	my $json_cfg = &Cfg2Json(\%h_cfg);
	my $config_js= $o_io->GetCfg('ROOT_CONFIG_JS');
	my $fd_config = new IO::File(">$config_js") or throw LipmError::IOException( $config_js,'f');
	print $fd_config $json_cfg . "\n";
	$fd_config->close();

	# Create a WebBuilder object
	my $o_param = New ParamParser();
    my $o_web_builder = WebBuilderFactory::GetWebBuilder( o_io => $o_io, o_param => $o_param, -session_flg => &FALSE );
	    
    # Try to compile narcisse_all.js
    if ( 0 )
    {
		# Create the narcisse_all.js file, using WebBuilder.
		$o_web_builder -> CreateDigestJsFiles();

	    my $js_compiler_jar = $o_io->GetCfg('ROOT_JAVASCRIPT_COMPILER');
	    my $js_narcisse_all = File::Spec->catfile($o_io->GetCfg('ROOT_WEB'),'narcisse_all.js');
	    my $js_narcisse_all_compiled = $js_narcisse_all . '.compiled';
	    
	    my $compiler_cmd = "java -jar $js_compiler_jar --warning_level QUIET --compilation_level WHITESPACE_ONLY --js $js_narcisse_all";
	    my ($status) = RunExt( -cmd => $compiler_cmd, -stdout => ">$js_narcisse_all_compiled" );
	    if ($status != 0)
	    {
			warn "WARNING ! The js_compiler returned an error code - narcisse_all.js will NOT be compiled";
		}
		else
		{
			my $unix_mv = $o_io -> GetUnixPath('UNIX_MV');
			my $src = $js_narcisse_all;
			my $dst = $js_narcisse_all . '.source';
			`$unix_mv $src $dst`;
			`$unix_mv $js_narcisse_all_compiled $js_narcisse_all`;
		}
	}

	# Create the narcisse_all.css file, using WebBuilder.
 	$o_web_builder -> CreateDigestCssFiles();
}

=head2 function Cfg2Json

	Title      :	Cfg2Json
	Usage      :	my $js = &Cfg2Json($rh_cfg);
	function   :	Rewrite %O_CFG into a json representation, and generate a tiny javascript
	Returns    :	the produced javascript
	Args       :	$rh_cfg => some hash
	globals    :	none

=cut
sub Cfg2Json
{
    my ($rh_cfg) = (@_);

    # We work on a copy, because we are keeping only some keys
    my %local_cfg;

    # The list of regex, we keep the keys which match with one of theses regexes
    my @a_allowed_keys = split(',',$rh_cfg->{CONFIG_JS_KEYS});
    foreach my $k (keys %$rh_cfg)
    {
        foreach my $ak (@a_allowed_keys)
        {
            if ($k =~ $ak)
            {
                $local_cfg{$k} = $rh_cfg->{$k};
                last;
            }
        }
    }

    my $json = new JSON;
    my $js = $json->encode(\%local_cfg);

	# pour corriger le probleme lie aux ' dans l'adresse ( javascript )
	$js =~ s/'/\\'/g;

    $js = "var O_CFG = '$js'.evalJSON(true);\n";

    return $js;
}

=head2 procedure CheckDir

 Title        : CheckDir
 Usage        : CheckDir($o_io,$root,$root_data)
 Fonction     : Check main directories, change the mode of the RW directories
                Remove cache files (?)
 Returns      : none, dies if something wrong is detected (?)
 Args         : $o_io The Instance Observer
                $root (optional) Root directory to check - NOT NECESSARILY $o_io->GetCfg('ROOT') IN CLONE MODE !
                OUAIS MAIS CA MARCHE PAS, CA MARCHE SEULEMENT EN $o_io !!!
                A REVOIR
                $root_data (optional) Data directory to check - NOT NECESSARILY $o_io->GetCfg('ROOT_DATA') IN CLONE MODE !
 Globals      : %CFG

=cut
sub CheckDir()
{
    my ($o_io,$root,$root_data) = @_;
    my $o_logger = $o_io->GetLogger();
    my $logger_flg = defined $o_logger;
    my $sav_dir = getcwd();

    $root      = $o_io->GetCfg('ROOT')      unless( defined $root );
    $root_data = $o_io->GetCfg('ROOT_DATA') unless( defined $root_data );

    my @a_key_rw = qw( ROOT_TMP ROOT_DATA_CACHE ROOT_MAP_CACHE ROOT_DOTPLOT_CACHE ROOT_DISPLAYFAM_CACHE_50 ROOT_DISPLAYFAM_CACHE_80 );
    my @a_key    = qw( ROOT_SITE ROOT_WEB ROOT_CORE ROOT_BIN ROOT_SBIN ROOT_CFG ROOT_CSS ROOT_LOGO ROOT_JAVASCRIPT );
    my @a_data_dir_key = qw( ROOT_DATA_DB ROOT_URL_TXT );

    # Build @a_dir_rw from @a_key_rw and @a_dir from @a_key and @a_key_rw
    my @a_dir_rw;
    my @a_dir;
    my @a_data_dir_file;
    foreach my $k (@a_key_rw)
    {
        push(@a_dir_rw,$o_io->GetCfg($k));
        push(@a_dir,$o_io->GetCfg($k));
    }
    foreach my $k (@a_key)
    {
        push(@a_dir,$o_io->GetCfg($k));
    }
    foreach my $k (@a_data_dir_key)
    {
        push(@a_data_dir_file,$o_io->GetCfg($k));
    }

	chdir($root);

    $o_logger->Trace("Checking the narcisse directories\n","*","A") if ($logger_flg);

    my $unix_mkdir = $o_io->GetUnixPath('UNIX_MKDIR');
    my $unix_chmod = $o_io->GetUnixPath('UNIX_CHMOD');
	foreach my $dir (@a_dir_rw)
	{
		system("$unix_mkdir $dir");
        `$unix_chmod a+rwX $dir`;
	}
    my $error = 0;
	foreach my $dir (@a_dir)
	{
		if ( !-e($dir) || !-d($dir))
        {
            warn "$dir not found\n";
            $error++;
        }
	}

    # If DATA IN DATABASE is set, look for the database file
    #if ($o_io->GetCfg('DATA_IN_DATABASE')!=0 and ! -r $o_io->GetCfg('ROOT_DATABASE'))
    #{
    #    die ("ERROR - ".$o_io->GetCfg('ROOT_DATABASE')." not found");
    #}

    foreach my $dir_file (@a_data_dir_file)
    {
        if ( !-e($dir_file)) {
            warn "$dir_file not found\n";
            $error++;
        }
    }
    if ($error != 0)
    {
        $o_logger->Trace("$error errors found, terminating\n","*","A") if ($logger_flg);
        die ("$error errors found, terminating\n");
    }
    else
    {
        $o_logger->Trace("OK\n","*","A") if ($logger_flg);
    }
    chdir $sav_dir;
	return;
}

=head2 procedure ShowNarcisseCFG

 Title        : ShowNarcisseCFG
 Usage        : &ShowNarcisseCFG($o_conf)
 Procedure    : Showing the main instance parameters from narcisse.cfg
 Args         : $o_conf

=cut
sub ShowNarcisseCFG
{
    my ($o_conf) = @_;

    # Ask the user for the fondamental parameters
    my %h_prms =(
        ROOT                       => $o_conf->GetCfg('ROOT'),
        ROOT_DATA                  => $o_conf->GetCfg('ROOT_DATA'),
        HTTP                       => $o_conf->GetCfg('HTTP'),
        NARCISSE_ADM_EMAIL         => $o_conf->GetCfg('NARCISSE_ADM_EMAIL'),
        INSTANCE_NAME              => $o_conf->GetCfg('INSTANCE_NAME'),
        LOG_LEVEL                  => $o_conf->GetCfg('LOG_LEVEL')
        );

	if ($o_conf->GetName() ne 'multi')
	{
		$h_prms{KINGDOM}                    = $o_conf->GetCfg('KINGDOM');
		$h_prms{NARCISSE_GENOME_REPOSITORY} = $o_conf->GetCfg('NARCISSE_GENOME_REPOSITORY');
		$h_prms{NARCISSE_SNP_REPOSITORY}    = '';
		$h_prms{INSTANCE_VERSION}           = $o_conf->GetCfg('INSTANCE_VERSION');
		
		# protecting agains errors because we cannot be sure NARCISSE_SNP_REPOSITORY really exists
		# (introduced with release 2.1.0)
		try
		{
			$h_prms{NARCISSE_SNP_REPOSITORY} = $o_conf->GetCfg('NARCISSE_SNP_REPOSITORY');
		}
		catch Error with
		{
		};
	}
	
	my %h_labels=(
		ROOT                       => 'Narcisse instance directory',
		ROOT_DATA                  => 'Narcisse data directory',
		HTTP                       => 'Narcisse instance URL',
		NARCISSE_ADM_EMAIL         => 'Narcisse administrator email',
		INSTANCE_NAME              => 'Narcisse instance name',
		KINGDOM                    => 'kingdom',
		NARCISSE_GENOME_REPOSITORY => 'Narcisse genome repository',
		NARCISSE_SNP_REPOSITORY    => 'Narcisse snp repository',
		INSTANCE_VERSION           => 'Narcisse Instance Version',
	);

	my @a_keys=qw( INSTANCE_NAME INSTANCE_VERSION ROOT ROOT_DATA HTTP NARCISSE_ADM_EMAIL KINGDOM NARCISSE_GENOME_REPOSITORY NARCISSE_SNP_REPOSITORY );

	foreach my $k (@a_keys)
	{
		if ( defined $h_prms{$k} )
		{
			printf "%30s\t%-50s\n",$h_labels{$k},$h_prms{$k};
		}
	}
	
	# Show the species or the instances
	my $name = $o_conf->GetCfg('INSTANCE_NAME');
	my $cmd  = &UNIX_BASH;
	$cmd .= " -c '. .setenv.sh; . \$NARCISSE_ROOT/Core/sbin/chins.sh $name; ";
	if ( $name eq 'multi' )
	{
		$cmd .= "nc_instances.pl";
	}
	else
	{
		$cmd .= "nc_species.pl --short --long";
	}
	$cmd .= "'";
	
	print "\nNow executing $cmd\n";
	system($cmd);
	print "\n";
}

=head2 procedure UpdateNarcisseCFG

 Title        : UpdateNarcisseCFG
 Usage        : &UpdateNarcisseCFG($o_conf)
 Procedure    : Updating the main instance parameters in narcisse.cfg
 Args         : $o_conf
                $instance_name (do not use the name returned by $o_conf->GetName())

=cut
sub UpdateNarcisseCFG
{
    my ($o_conf,$instance_name) = @_;

    # Ask the user for the fondamental parameters
    my %h_prms =(
        ROOT                       => $o_conf->GetCfg('ROOT'),
        ROOT_DATA                  => $o_conf->GetCfg('ROOT_DATA'),
        HTTP                       => $o_conf->GetCfg('HTTP'),
        NARCISSE_ADM_EMAIL         => $o_conf->GetCfg('NARCISSE_ADM_EMAIL'),
        INSTANCE_NAME              => $instance_name
        );

    if ($o_conf->GetName() ne 'multi')
    {
        $h_prms{KINGDOM}                    = $o_conf->GetCfg('KINGDOM');
        $h_prms{NARCISSE_GENOME_REPOSITORY} = $o_conf->GetCfg('NARCISSE_GENOME_REPOSITORY');
        $h_prms{NARCISSE_SNP_REPOSITORY}    = '';
        $h_prms{INSTANCE_VERSION}           = $o_conf->GetCfg('INSTANCE_VERSION');
        
        # protecting agains errors because we cannot be sure NARCISSE_SNP_REPOSITORY really exists
        # (introduced with release 2.1.0)
        try
        {
			$h_prms{NARCISSE_SNP_REPOSITORY} = $o_conf->GetCfg('NARCISSE_SNP_REPOSITORY');
		}
		catch Error with
		{
		};
    }
    
    # give a reasonable value to some of those parameters, if possible
    my $current_dir           = abs_path(getcwd());
    
    # manu, Dec 2010 - Even when reconfigured, compute ROOT - This makes possible to change root directory, then to call
    #                  nc_copnfigure.pl --edit to finish the change
    #$h_prms{ROOT}             = File::Spec->catfile($current_dir,$instance_name) if ($h_prms{ROOT} eq '');
    $h_prms{ROOT}             = File::Spec->catfile($current_dir,$instance_name);
    $h_prms{ROOT_DATA}        = File::Spec->catfile($current_dir,$instance_name,'Site','data') if ($h_prms{ROOT_DATA} eq '');
    $h_prms{HTTP}             = 'http://your.server.name/url/to/narcisse/'.$instance_name if ($h_prms{HTTP} eq '');
    if ($instance_name ne 'multi')
    {
        $h_prms{KINGDOM}          = ($instance_name =~ /bacteria/i) ? 'bacteria' : 'eukaryote';
        $h_prms{INSTANCE_VERSION} = &__DefaultInstanceVersion() if ($h_prms{INSTANCE_VERSION} eq '');
    }

    # Ask the user for all those parameters
    # MANU - This is removed because the root directory MUST BE THE SAME as the Instance directory, which is already chosen
    # $h_prms{ROOT}                       = &Ask("Enter the Narcisse root directory for this instance ",$h_prms{ROOT},\&BlankCreateDir);

    $h_prms{ROOT_DATA}                  = &Ask("Enter the Narcisse data directory for this instance ",$h_prms{ROOT_DATA},\&Blank);

    $h_prms{HTTP}                       = &Ask("Enter the Narcisse root URL for this instance       ",$h_prms{HTTP},\&Blank);
    $h_prms{HTTP} =~ s/\/$//;  # remove the trailing slash, if any
    $h_prms{NARCISSE_ADM_EMAIL}         = &Ask("Enter the Narcisse administrator email              ",$h_prms{NARCISSE_ADM_EMAIL},\&BlankEmail);
    if ($instance_name ne 'multi')
    {
        $h_prms{KINGDOM}                    = &Ask("Enter the kingdom (bacteria/eukaryote)          ",$h_prms{KINGDOM},\&BlankKingdom);
        $h_prms{NARCISSE_GENOME_REPOSITORY} = &Ask("Enter the genome repository                     ",$h_prms{NARCISSE_GENOME_REPOSITORY});
        $h_prms{NARCISSE_SNP_REPOSITORY}    = &Ask("Enter the snp repository                        ",$h_prms{NARCISSE_SNP_REPOSITORY});
        $h_prms{INSTANCE_VERSION}           = &Ask("Enter the instance version                      ",$h_prms{INSTANCE_VERSION},\&Blank);

        # Default answer = Yes if we already have a Beta version, N if we don't
        my $def_answer = ($h_prms{INSTANCE_VERSION} =~ /Beta$/) ? 'Y' : 'N';
        print "Is it a Beta version ? " . ($def_answer eq 'N' ? '(Y/[N])' : '([Y]/N)');
        if (&NARCISSE_utils::GetAdmAnswer("","Y|N",$def_answer) eq 'Y')
        {
            if ($h_prms{INSTANCE_VERSION} !~ /Beta$/)
            {
                $h_prms{INSTANCE_VERSION}.='Beta';
            }
        }
        else
        {
            $h_prms{INSTANCE_VERSION} =~ s/Beta$//;
        }
    }

    # Update the narcisse.cfg file
    print "updating narcisse.cfg\n";
    &WriteNarcisseCfg($o_conf->GetCfgPath(),\%h_prms);

    # Finished !
    return;

    # Generate a default instance version from the date
    sub __DefaultInstanceVersion
    {
        my @now=localtime();
        my $y=$now[5]+1900;
        my $m=$now[4]+1;
        $m='0'.$m if ($m<10);
        my $d=$now[3];
        return $y.$m.$d;
    };
}


=head2 procedure WriteNarcisseCfg
 Title        : WriteNarcisseCfg
 Usage        : &WriteNarcisseCfg($file_cfg,\%h_prms)
 Procedure    : Read the narcisse.cfg file, replacing parameter values by the values found in \%h_prms
 Args         : $file_cfg Path to the narcisse.cfg file
                $rh_prms  Parameters to be overloaded
=cut
sub WriteNarcisseCfg
{
    my ($file_cfg, $rh_prms) = @_;
    throw LipmError::IOException( $file_cfg,'f' ) if ( !-e $file_cfg );

    my $file_cfg_old = $file_cfg . '_old';
    link($file_cfg,$file_cfg_old);
    unlink($file_cfg);

    my $fh_file_cfg_old = new IO::File ($file_cfg_old) or throw LipmError::IOException( $file_cfg_old, 'f' );
    my $fh_file_cfg     = new IO::File (">$file_cfg") or throw LipmError::IOException( $file_cfg, 'f' );
    while (my $line = <$fh_file_cfg_old>)
    {
        if ($line =~ /^([\w]+)=/ and exists $rh_prms->{$1})
        {
            print $fh_file_cfg $1.'='.$rh_prms->{$1}."\n";
        }
        else
        {
            print $fh_file_cfg $line;
        }
    }
    $fh_file_cfg->close();
    $fh_file_cfg_old->close();
    unlink($file_cfg_old);
}

=head2 procedure DumpConfiguration
 Title        : DumpConfiguration
 Usage        : &DumpConfiguration($o_conf)
 Procedure    : Dump $o_conf to a .dump file, this is interesting for debugging
 Args         : $o_conf

=cut
sub DumpConfiguration
{
    my ($o_conf) = @_;
    my $cfg_dump = $o_conf->GetCfgPath().'.dump';
    print "Now creating " . basename $cfg_dump . " (this is very useful for debugging )\n";
    $o_conf->DumpCfg( $cfg_dump );
}

=head2 procedure CreateUpdateScript

 Title        : CreateUpdateScript
 Usage        : CreateUpdateScript($root)
 Prerequisite : none
 Procedure    : Creates the main update.bash script
 Returns      : none
 Args         : none
 TODO         : reecrire tout ca, c'est bien crados !
=cut
sub CreateUpdateScript()
{
	my $o_io   = shift;
    my $o_logger = $o_io->GetLogger();

    my $update_root = $o_io->GetCfg('ROOT_UPDATE');
    my $unix_mkdir  = $o_io->GetUnixPath('UNIX_MKDIR');
    my $unix_rm     = $o_io->GetUnixPath('UNIX_RM');
    `$unix_mkdir -p $update_root`;

    my $update_path = Sufine::MkDir(-dir => $update_root, -name => 'update');
    throw LipmError::IOException( $update_path,'d') unless (defined $update_path);

    my $main_script_name    = $update_path . '/update.bash';

    my $fh_script_name = new IO::File(">$main_script_name") or throw LipmError::IOException( $main_script_name,'f');
    print $fh_script_name "#! /bin/bash\n";
    print $fh_script_name "#PBS -l ncpus=X\n#PBS -l cput=HHH:00:00\n#PBS -M some.one\@some.where.fr -m abe\n";
    print $fh_script_name "# this script should be started from the directory $update_path\n";
    print $fh_script_name "# Change to that directory when submitted through PBS\n";
    print $fh_script_name 'if [ ! -z "$PBS_O_WORKDIR" ] ; then cd $PBS_O_WORKDIR; fi'."\n\n";
    print $fh_script_name 'export ROOT_SBIN="' . $o_io->GetCfg('ROOT_SBIN') . '"'."\n";
    print $fh_script_name 'export NARCISSE_CFG="'.$ENV{NARCISSE_CFG}.'"'."\n";
    if (exists $ENV{NARCISSE_GENOME_REPOSITORY})
    {
        print $fh_script_name 'export NARCISSE_GENOME_REPOSITORY="'.$ENV{NARCISSE_GENOME_REPOSITORY}.'"'."\n";
    };
     if (exists $ENV{NARCISSE_SNP_REPOSITORY})
    {
        print $fh_script_name 'export NARCISSE_SNP_REPOSITORY="'.$ENV{NARCISSE_SNP_REPOSITORY}.'"'."\n";
    };

	print "Do you want to compute the S.N.P. ?    Y/[N] ";
	my $with_snp = (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y') ? &TRUE: &FALSE;

    print "Do you want to force all comparisons ? Y/[N] ";
    my $force_all = (&NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y') ? &TRUE: &FALSE;

    print "How many processors do you want to use for blast ? [1] ";
	my $blast_cpus = &NARCISSE_utils::GetAdmAnswer("","[1-8]","1");
    if ($blast_cpus != 1)
    {
        $blast_cpus = " -a $blast_cpus";
    }
    else
    {
        $blast_cpus = '';
    }

    print "Do you want self calculations (protein only) ? [N] ";
    my $self = &NARCISSE_utils::GetAdmAnswer("","Y|N","N");

    print "Do you want to use paraloop ? [Y] ";
    my $using_paraloop = &NARCISSE_utils::GetAdmAnswer("","Y|N","Y");
    $using_paraloop = ($using_paraloop eq 'Y') ? 1:0;

    my $paraloop_cpus= 0;
    my $paraloop_dir = "";
    if ($using_paraloop)
    {
        print "How many processors do you want to use for paraloop ? [2] ";
        $paraloop_cpus = &NARCISSE_utils::GetAdmAnswer("","[1-9][0-9]*","2");
        if (exists $ENV{PARALOOP})
        {
            $paraloop_dir = $ENV{PARALOOP};
        }
        else
        {
            $paraloop_dir = $o_io->GetCfg('ROOT_PARALOOP_DIR');
        }
        print $fh_script_name 'export PARALOOP="'.$paraloop_dir.'"'."\n";
        print $fh_script_name "\n";

        # build the cfg files and the paraloop command line
        foreach my $name ( 'getting_data','integrating_data' )
        {
            my $paraloop_cfg = $update_path . '/'.$name.'.cfg';
            my $fh_paraloop_cfg = new IO::File(">$paraloop_cfg") or throw LipmError::IOException( $paraloop_cfg,'f');
            print $fh_paraloop_cfg "# Configuration file for paraloop - created by nc_configure.pl\n";
            print $fh_paraloop_cfg "# This is a template - run paraloop.pl --parameters to know the list of available parameters\n\n";
            print $fh_paraloop_cfg "# error,lock,output directories = each time you launch paraloop, a 3 new directories will be created\n";
            print $fh_paraloop_cfg "PARALOOP_error_directory = error-$name-%D%M%Y-%h-%m-%s\n";
            print $fh_paraloop_cfg "PARALOOP_lock_directory  = lock-$name-%D%M%Y-%h-%m-%s\n";
            print $fh_paraloop_cfg "PARALOOP_output          = output-$name-%D%M%Y-%h-%m-%s/out\n";
            print $fh_paraloop_cfg "\n";
            print $fh_paraloop_cfg "# The plugin\n";
            print $fh_paraloop_cfg "PARALOOP_program = Shell\n";
            print $fh_paraloop_cfg "\n";
            print $fh_paraloop_cfg "# For debug 012R, for less debug 01R, for very few log, 0R\n";
            print $fh_paraloop_cfg "# The letter 'R' makes the scripts producing an error to be copied back, so that you can easily replay when necessary\n";
            print $fh_paraloop_cfg "PARALOOP_log_level = 012R\n";
            print $fh_paraloop_cfg "\n";
            print $fh_paraloop_cfg "# This parameter may be useful if you have some switch to specify to the queue\n";
            print $fh_paraloop_cfg "#PARALOOP_qsub_params = -l ncpus=4,mem=4Gb,cput=50:00:00 -m abe -M mail\@somewhere.fr\n";
            print $fh_paraloop_cfg "\n";
            print $fh_paraloop_cfg "# The job is interrupted and resubmitted after this time (in s)\n";
            print $fh_paraloop_cfg "#PARALOOP_fair_time_limit = 7200\n";
            print $fh_paraloop_cfg "\n";
            print $fh_paraloop_cfg "# The input file\n";
            print $fh_paraloop_cfg "PARALOOP_input = $name.shell\n";
            undef $fh_paraloop_cfg;
        }

        # closing main update script
        $fh_script_name->close();
    }

    # Collect the data already present on the system - We check the date of the following directories:
    #   -$species/db (created by nc_get_data)                               ==> stored in %h_get_files (key = a short name, val = a timestamp)
    #   -$species/comparative/$species/synt/glint (created by nc_nucleic)   ==> stored in %h_nucleic_files (key = 2 short names, val = a timestamp)
    #   -$species/comparative/$species/synt/prbh    (created by nc_protein) ==> stored in %h_protein_files (key = 2 short names, val = a timestamp)

    my %h_get_files;
    &BuildTimeStamps($o_io,\%h_get_files);

    # Track the "genetic" species
    my $rep_species = $o_io->GetSpeciesCfgDirpath('INSTANCE');
    my %h_genetic_maps;
    foreach my $spec (sort $o_io->GetSpeciesShortName())
    {
        my $cfg_file =  File::Spec->catfile($rep_species,$spec.'.cfg');
        my $o_param = New ParamParser($cfg_file);
        if ($o_param->Get('FORMAT') eq 'GENETIC')
        {
            my $compare_with = $o_param->Get('COMPARE_WITH');
            $h_genetic_maps{$spec} = ($o_param->IsDefined('COMPARE_WITH')) ? $o_param->Get('COMPARE_WITH') : 'none';
        }
    }


    # build the commands and write them, directly to update.bash or to the shell files
	my $update_flg = &FALSE;
	my $root_sbin = $o_io->GetCfg('ROOT_SBIN');
	foreach my $name ('get_data','get_snp_data','genomic_comparison','genetic_rbh')
	{
		my $file_shell='';
		my $method='';
		my $nc_script = '';
		if ($name eq 'get_data')
		{
			$file_shell= $update_path . '/getting_data.shell';
			$nc_script = "\$ROOT_SBIN/nc_get_data.pl ";
		}
		if ($name eq 'get_snp_data')
		{
			next unless ($with_snp==&TRUE);
			$file_shell= $update_path . '/integrating_data.shell';
			$nc_script = "\$ROOT_SBIN/nc_get_snp_data.pl ";
		}
        if ($name eq 'genetic_rbh')
        {
            $method = 'genetic';
            $file_shell= $update_path . '/integrating_data.shell';
            $nc_script = "\$ROOT_SBIN/nc_genomic_comparison.pl --method=$method --force ";
        }
        if ($name eq 'genomic_comparison')
        {
            $method = 'genomic';
            $file_shell = $update_path . '/integrating_data.shell';
            $nc_script = "\$ROOT_SBIN/nc_genomic_comparison.pl --method=$method --force ";
        }
        #if ($nc_script eq '') {
        #    $nc_script = "\$ROOT_SBIN/nc_" . $name . '.pl';
        #    $file_shell= $update_path . '/'. $name . '.shell';
        #}

        if ($using_paraloop)
        {
            # reopen $fh_script_name, but not to the main script file
            $fh_script_name = new IO::File(">>$file_shell") or die "Cannot open $file_shell for writing";
        }

        # get_data ==> We call nc_get_data only if one of the index files was not found
        if ($name eq 'get_data')
        {
            foreach my $spec (sort $o_io->GetSpeciesShortNamesForUpgrade())
            {
                unless (exists $h_get_files{$spec})
                {
                    print $fh_script_name $nc_script.' --cfg '.$spec. " 2>&1\n";
					$update_flg = &TRUE;
                }
            }
        }

		# get_snp_data ==> We call nc_get_snp_data.pl only if there is snp to compute
		if ($name eq 'get_snp_data')
		{
			foreach my $spec (sort $o_io->GetSpeciesShortNamesForUpgrade())
            {
				my $o_so   = $o_io->GetSpeciesObserver( $spec) ;
				my $o_cfg  = $o_so->GetCfgFilepath( 'INSTANCE' );
				my $o_prms = New ParamParser( $o_cfg );
				my @a_snp_keys = $o_prms->GetKeys( '^FILE_SNP_' );
				if ( @a_snp_keys != 0 )
				{
					print $fh_script_name $nc_script . " --cfg $spec" . " 2>&1 \n";
					$update_flg = &TRUE;
				}
			}
		}

		# genetic_rbh
		if ($name eq 'genetic_rbh')
		{
			foreach my $spec_a (sort keys %h_genetic_maps)
			{
				my $spec_b = $h_genetic_maps{$spec_a};

				# none means = NO SPECIES - skip any processing if none specified !
				next if ($spec_b eq 'none');
				my @a_lst = split(/[,\s]+/,$spec_b);
				foreach my $spec_b (@a_lst)
				{
					if (&IsProcessingNecessary($o_io,$spec_a,$spec_b,\%h_get_files,$method,$force_all))
					{
						print $fh_script_name $nc_script.' --cfgQ '.$spec_a.' --cfgS '.$spec_b." 2>&1 \n";
						$update_flg = &TRUE;
					}
				}
			}
		}

		# genomic_comparison
		if ($name eq 'genomic_comparison')
		{

			# nucleic_rbh, protein_rbh => We call them only if necessary, based on the existing directories's timestamps
			foreach my $spec_a (sort $o_io->GetSpeciesShortNamesForUpgrade())
			{
				# If a genetic map, do not compute nucleic/protein stuff
				next if (exists $h_genetic_maps{$spec_a});
				foreach my $spec_b (sort $o_io->GetSpeciesShortNamesForUpgrade())
				{
					next if ($spec_b lt $spec_a);
					next if ( ($spec_b eq $spec_a) and (($self ne "Y") or ($method ne "prbh" and $method ne "genomic")) );
					next if (exists $h_genetic_maps{$spec_b});
					
					my $o_spa = $o_io -> GetSpeciesObserver( $spec_a);
					my $o_spb = $o_io -> GetSpeciesObserver( $spec_b);
					
					# If none of those species is pseudomolecules based, we skip the comparison
					next if ( $o_spa->IsPseudo()==&FALSE && $o_spb->IsPseudo()==&FALSE);
					
					# If one of those species is scaffolds_based, it is the cfgQ parameter (better glint perfs)
					# Exchange both species ;
					if ( $o_spb->IsPseudo==&FALSE )
					{
						my $o_spc = $o_spa;
						$o_spa = $o_spb;
						$o_spb = $o_spc;
					}

					if ( &IsProcessingNecessary($o_io,$o_spa -> GetShortName(), $o_spb -> GetShortName(),\%h_get_files,$method,$force_all) )
					{
						my $bc = ($name eq 'protein_rbh' or $name eq 'genomic_comparison') ? $blast_cpus:'';
						if ($o_spa -> GetShortName() eq $o_spb -> GetShortName())
						{
							print $fh_script_name $nc_script.' --cfgQ '.$o_spa -> GetShortName().' --self '.$bc." 2>&1 \n";
						}
						else
						{
							print $fh_script_name $nc_script.' --cfgQ '.$o_spa -> GetShortName().' --cfgS '.$o_spb -> GetShortName().' '.$bc." 2>&1 \n";
						}
						$update_flg = &TRUE;
					}
				}
			}
		}
	}

	if ($using_paraloop == &TRUE)
	{
		foreach my $name ('getting_data','integrating_data')
		{
			my $shell = $update_path . '/' . $name . '.shell';
			my $cfg   = $update_path . '/' . $name . '.cfg';
			if ( -s $shell )
			{
				my $fh_script_name = new IO::File(">>$main_script_name") or throw Lipm::IOException($main_script_name,'f');
				print $fh_script_name $paraloop_dir.'/bin/paraloop.pl' . " --cfg $name.cfg --load_balancing --wait --ncpus $paraloop_cpus\n";
			}
			else
			{
				`$unix_rm $shell $cfg`;
			}
		}
    }

    if ($update_flg == &FALSE)
    {
        unlink $main_script_name;
        print "\nNothing to update\n";
        $o_logger->Trace("Nothing to update\n","*","A");
    }
    else
    {
		if ($using_paraloop == &TRUE)
		{
			print "\nYou will use paraloop to generate narcisse files.\n";
			print "You're right, it is a nice program.\n";
			print "But please have a look to the update.bash and *.cfg files, and add the necessary switchs/parameters to paraloop:\n";
			print "May be the switches --local, --ncpus will be useful for you.\n";
		}

        print "\nFiles created for the update in directory $update_path\n";
        $o_logger->Trace("Update.bash created\n","*","A");
        print "Do you want to execute update.bash now ? Y/[N] ";
        if ( &NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y' )
        {
			print "WARNING - WE ARE NOW USING THE UNIX COMMAND at - IF THIS DOES NOT WORK, PLEASE ASK YOUR ADMINISTRATOR\n";
            my $unix_at = $o_io->GetUnixPath('UNIX_AT');
            chdir $update_path;
            `echo "sh update.bash" | $unix_at now`;
        }
    }
}

=head2 procedure BuildTimeStamps

 Title        : BuildTimeStamps
 Usage        : &BuildTimeStamps(\%h_get_files);
 Procedure    : Look for some files produced by nc_get_data.pl inside the data hierarchy
                Keep in the hash the short species name, together with a time stamp.
 Args         : $o_io,$rh_get_files

=cut
sub BuildTimeStamps
{
    my ($o_io,$rh_get_files) = @_;

    my $root_data = $o_io->GetCfg('ROOT_DATA');

    foreach my $spec (sort $o_io->GetSpeciesShortName())
    {
        # We check one of the glint indexation files and one of the blast indexation files,
        # OR the marker, first chromosome file (useful for the genetic maps)
        # We take the mtime of the youngest of those files
        # MANU - JUIN 2010 - On ne regarde QUE l'index glint, PAS l'index prot
        # En effet: 1/ L'index prot peut etre refait ulterieurement sans probleme (ajout d'annotations interpro)
        #           2/ L'index prot n'existe pas pour les markers, ils etaient donc toujours refaits

        my $o_sp = $o_io -> GetSpeciesObserver ( $spec );
		my $file1;
        try
        {
	        $file1 = $o_sp -> GetFastaDir ( feature => 'CHR' );
	        $file1 = File::Spec->catfile ( $file1, 'index', $spec.'V1.seq.fa' );
	        if ( ! -e $file1 )
	        {
				my $ra_chroms = $o_sp -> GetChromsNames();
				$file1 = $o_sp -> GetMarkerFiles ( -chrom => $ra_chroms -> [0], -check => &FALSE );
			}

	        my @stat1 = stat($file1);
	        if ($#stat1!=-1)
	        {
	            $rh_get_files->{$spec} = $stat1[9];
	        }
		}
		catch LipmError::IOException with
		{};

        #my ($file1,$file2);
        #try
        #{
	        #$file1 = $o_sp -> GetFastaDir ( feature => 'CHR' );
	        #$file1 = File::Spec->catfile ( $file1, 'index', $spec.'V1.seq.fa' );
	        #if ( ! -e $file1 )
	        #{
				#my $ra_chroms = $o_sp -> GetChromsNames();
				#$file1 = $o_sp -> GetMarkerFiles ( -chrom => $ra_chroms -> [0], -check => &FALSE );
			#}
		
			#$file2 = $o_sp -> GetFastaDir ( feature => 'PROT' );
			#$file2 = File::Spec->catfile ( $file2, "${spec}_prot.fasta.phr" );

	        #my @stat1 = stat($file1);
	        #my @stat2 = stat($file2);
	        #if ($#stat1!=-1 and $#stat2!=-1)
	        #{
	            #$rh_get_files->{$spec} = ($stat1[9] >= $stat2[9]) ? $stat1[9] : $stat2[9];
	        #}
	        #elsif ($#stat1!=-1)
	        #{
	            #$rh_get_files->{$spec} = $stat1[9];
	        #}
	        #elsif ($#stat2!=-1)
	        #{
	            #$rh_get_files->{$spec} = $stat2[9];
	        #}
		#}
		#catch LipmError::IOException with
		#{};
    }
}

=head2 function IsProcessingNecessary

 Title        : IsProcessingNecessary
 Usage        : if (&IsProcessingNecessary($spec_a,$spec_b,$rh_get_files,$rh_comparison)) ...
 Prerequisite : none
 Function     : Check, for 2 species, if it is necessary to compute the comparison
 Returns      : 0/1
 Args         : $o_io            The instance observer
                $spec_a, $spec_b The two species to check
                $rh_get_files    The hash constaining mtime for the data files
                $method          The method name (glint, prbh, genetic, genomic)
                $force_all       Optional, def = &FALSE - If &TRUE, do not look at the dates

=cut

sub IsProcessingNecessary
{
    my ($o_io,$spec_a,$spec_b,$rh_get_files,$method,$force_all) = @_;
    $force_all = &FALSE unless defined $force_all;

    # do not compare a species against the same species with different releases
    return &FALSE if (&AreTheySameSpeciesOtherRelease($spec_a,$spec_b));

    # If $force_all is set, processing is always necessary
    return &TRUE if $force_all;

    my $time_stamp = &_ComputeTimeStamp($o_io,$spec_a,$spec_b,$method);

    # If one of 3 directories does not exist, return 1
    return &TRUE if (!exists $rh_get_files->{$spec_a});
    return &TRUE if (!exists $rh_get_files->{$spec_b});
    return &TRUE if ($time_stamp == -1);

    # If the comparison directories are too old, return 1
    return &TRUE if ( $time_stamp < $rh_get_files->{$spec_a} );
    return &TRUE if ( $time_stamp < $rh_get_files->{$spec_b} );

    return &FALSE;

    # We check the modification times of the glint or blast results
    # We return the lowest mtime (thus the oldest file)
    # If none found, we check the COMPARAtive directory (synt, level 0)
    # If still not found, we return -1
    sub _ComputeTimeStamp
    {
        my ($o_io,$spec_a,$spec_b,$method) = @_;

        my $root_data = $o_io->GetCfg('ROOT_DATA');
        my ($max_time,$min_time) = (-1, time());
        
        # checking comparison data
        foreach my $file_name ("glint_${spec_a}_${spec_b}.out.gz","glint_${spec_a}_${spec_b}.out","blastp_${spec_a}_${spec_b}.out.gz","blastp_${spec_a}_${spec_b}.out","genetic_${spec_a}_${spec_b}.out.gz")
        {
            my $comparison_path = File::Spec->catfile($root_data,'comparison',$method,$spec_a,$spec_b,$file_name);
            #warn "KOUKOU $comparison_path";
            if (-e $comparison_path)
            {
                my @a_stat = stat($comparison_path);
                $min_time  = $a_stat[9] if ($min_time > $a_stat[9]);
                $max_time  = $a_stat[9] if ($max_time < $a_stat[9]);
            }
		}
		
		# checking comparative data, only if comparison data were removed
		if ( $max_time == -1)
		{
			my $comparative_path = File::Spec->catfile($root_data,'comparative',$spec_a,$spec_b,'synt',$method,'0');
			if (-e $comparative_path)
			{
				my @a_stat = stat($comparative_path);
				$max_time  = $a_stat[9];
				$min_time  = $a_stat[9];
			}
		}

        return ($max_time==-1) ? $max_time : $min_time;
    }

}

=head2 function AreTheySameSpeciesOtherRelease

 Title        : AreTheySameSpeciesOtherRelease
 Usage        : if (&AreTheySameSpecies($spec_a,$spec_b))
 Function     : Check, if $spec_a and $spec_b are the same species with different releases return TRUE
                if $spec_a eq $spec_b OR different species return FALSE
 Returns      : 0/1
 Args         : $spec_a, $spec_b The two species to check

=cut
sub AreTheySameSpeciesOtherRelease
{
    my ($spec_a,$spec_b) = @_;

    return &FALSE if ($spec_a eq $spec_b);

    if ( $spec_a =~ /R/ || $spec_b =~ /R/ )
    {
        my $s_a = $spec_a;
        my $s_b = $spec_b;
        $s_a =~ s/R\d+$//;
        $s_b =~ s/R\d+$//;
        return &TRUE if ( $s_a eq $s_b );
    }
    else
    {
        return &FALSE;
    }
}

=head2 procedure UpdateSpeciesList

 Title        : UpdateSpeciesList
 Usage        : UpdateSpeciesList($o_io)
 Prerequisite : none
 Fonction     : Main procedure for updating the species.list file
 Returns      : none
 Args         : $o_io, $quiet_flg;

=cut
sub UpdateSpeciesList()
{
    my ($o_io,$quiet_flg) = @_;
    my $o_logger = $o_io->GetLogger();

    print "Do you want to (re) create a new species.list file ?        Y/[N] ";
    if ( &NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y' )
    {
        &UpdateSpeciesListAutomatically($o_io,$quiet_flg);
    }
    else
    {
        $o_logger->Trace("species.list not updated\n","*","A");
    }
}

=head2 procedure UpdateSpeciesListAutomatically

 Title        : UpdateSpeciesListAutomatically
 Usage        : UpdateSpeciesListAutomatically()
 Prerequisite : none
 Procedure    : Automatic detection of the species, from the configuration file of each species
 Returns      : none
 Args         : $o_io, $quiet_flg

=cut
sub UpdateSpeciesListAutomatically
{
    my ($o_io,$quiet_flg) = @_;
    my $o_logger = $o_io->GetLogger();

    $o_logger->Trace("Automatically search species for species.list file\n","*","A");
    my $unix_ls= $o_io->GetUnixPath('UNIX_LS');
	my $rep_species = $o_io->GetSpeciesCfgDirpath('INSTANCE');
	my @a_list_cfg = split /\n+/, `$unix_ls -c1 $rep_species`;

    my $rho_param = &ReadListCfg($o_io,$quiet_flg,\@a_list_cfg);
    my $file_species_list = $o_io->GetSpeciesListFilepath('INSTANCE');
	&CreateSpeciesListFile($file_species_list,$rho_param);

    print "Found " . keys (%$rho_param) . " species\n";
    $o_logger->Trace("Finished\n","*","A");
}


=head2 function ReadListCfg

 Title     : ReadListCfg
 Usage     : my $rho_param = &ReadListCfg($o_io,$quiest_flg,$ra_list_cfg)
 Function  : read the .cfg files from the species directory
 Args      : $o_io, $quiest_flg
             $ra_list_cfg A list of files to read
 Return    : A hash, key = short name, value = corresponding $o_param, a $o_param/file
 Error     : Die is something wrong with the cfg files: a strong analysis is performed, to be sure
             that the parameters are correct, the user is told of each problem
=cut
sub ReadListCfg
{
    my ($o_io,$quiet_flg,$ra_list_cfg) = @_;

    my @ao_param_new;
	my $rep_species = $o_io->GetSpeciesCfgDirpath('INSTANCE');
    my $root_data = $o_io->GetCfg('ROOT_DATA');

    my $template = $o_io->GetCfg('TEMPLATE');
    my $error = 0;

	my %h_list_by_short;
    my %h_list_by_long;
    my %ho_param;

    # First loop, skipping the data with ORG_NEW specified
    foreach my $cfg (@$ra_list_cfg)
    {

        # Skip other files
		next if ( $cfg !~ /\.cfg$/ );
		next if ( $cfg =~ /^$template/ );

        # Read files
        my $o_param = New ParamParser("$rep_species/$cfg",'A');
        $o_param->UnsetBehaviour('assert_strict');
        $o_param->SetSubstitution('%d',$root_data);
        $o_param->AssertDefined('DST','ORG_SHORT','ORG_LONG');
        $o_param->SetUnlessDefined('ASSEMBLY_PHASE','PSEUDOMOLECULES');

        my $cfg_name = basename ($cfg, '.cfg');
		my ($short, $long) = $o_param->Get('ORG_SHORT','ORG_LONG');

        # Some reserved words or patterns for short and long names
        if ( $short eq &NONE || $short eq 'all' || $long eq &NONE || $long eq 'all' )
        {
            warn "\tERROR - SKIPPED $long ($short) - \"none\" and \"all\" are RESERVED names\n";
            $error++;
            next;
        };
        if ( $short !~ /^[a-zR_0-9]+$/ )
        {
            warn "\tERROR - SKIPPED $long ($short) - Shortname may have only [a-zR0-1_] characters\n";
            warn "\t      - '_' is reserved for the Genetic maps\n";
            warn "\t      - 'R' is a Revision separator\n";
            $error++;
            next;
        }
        if ( $short =~ /^[_R]/ or $short =~ /[_R]$/ )
        {
            warn "\tERROR - SKIPPED $long ($short) - Short name is illegal, _ and R should NOT be specified at begininning or at end\n";
            $error++;
            next;
        }
        if ( $long !~ /^[a-zA-Z0-9\- ]+$/ )
        {
            warn "\tERROR - SKIPPED $long ($short) - Long name is illegal, Use only space, -, A-Za-z0-9 characters\n";
            $error++;
            next;
        }
        if ( $long =~ /^[ -.]/ or $long =~ /[ -.]$/ or $long =~ /  / )
        {
            warn "\tERROR - SKIPPED $long ($short) - Long name is illegal, space, -, . should NOT be at beginning or at end, twin spaces are forbidden\n";
            $error++;
            next;
        }

        # The filename must be the equal to species short name
        if ($cfg_name ne $short)
        {
            warn "\tERROR - This file should be called $short.cfg, NOT $cfg_name.cfg\n";
            $error++;
        }

        # If this is a new release of an existing organism, keep it for later treatment
        if ( ($o_param->GetInteger( 'ORG_NEW' ) != 0) )
        {
            push ( @ao_param_new,$o_param );
            next;
        };

        # There should be only ONE short name and ONE long name
        # ONE short name means ONLY ONE of banaRxyz AND banaRuvw
        my $lcl_error = 0;
        if ( exists $h_list_by_short{$short} )
        {
            warn "\tERROR - SKIPPED $short - The short name is already used for " . $h_list_by_short{$short} . "\n";
            $lcl_error++;
        }
        if ( exists $h_list_by_long{ uc $long} )
        {
            warn "\tERROR - SKIPPED $long - The long name is already used by " . $h_list_by_long{ uc $long} . " (ignoring the case)\n";
            $lcl_error++;
        }

        if ( $lcl_error > 0 )
        {
            $error += $lcl_error;
            next;
        }

        # Keep the short and the long name, and keep $o_param
        else
        {
            $h_list_by_long{uc $long}   = $short;
            $h_list_by_short{$short} = $long;
            $ho_param{$short}        = $o_param;
            if ($quiet_flg == &FALSE)
            {
				print "\tOK    - FOUND $long - short name $short";
				if ($o_param -> Get( 'ASSEMBLY_PHASE' ) ne 'PSEUDOMOLECULES')
				{
					print " - " . $o_param -> Get( 'ASSEMBLY_PHASE' );
				}
				print "\n";
			}
        }
    }

    # Second loop, considering only the data with ORG_NEW specified
    foreach my $o_param (@ao_param_new)
    {
        my ($short, $old_long) = $o_param->Get('ORG_SHORT','ORG_LONG');
        my $long = $old_long . " NEW UPGRADE";
        if (exists $h_list_by_long{uc $long})
        {
            warn "\tERROR - SKIPPED $long - this species long name was already specified for " . $h_list_by_long{ uc$long} . "\n";
            $error++;
            next;
        }
        else
        {
            $o_param->Set('ORG_LONG',$long);
            $h_list_by_long{uc $long}   = $short;
            $h_list_by_short{$short} = $long;
            $ho_param{$short}        = $o_param;
            print "\tOK    - UPGRADING $old_long - short name $short\n";
        }
    }

    # Third loop, we should NOT have more than 2 short names like mtr, mtrR1, mtrR2
    # Only ONE or TWO allowed, the first for production the second for upgrade
    # The long names should be equal too (except for the string NEW UPGRADE)
    {
        my $old_prefix="";
        my @a_short_same_prefix;
        foreach my $short (sort keys %h_list_by_short)
        {
            my $prefix = ($short =~ /(.+)R(.+)/)? $1 : $short;
            if ($old_prefix eq $prefix)
            {
                push(@a_short_same_prefix,$short);
            }
            else
            {
                if (@a_short_same_prefix>2)
                {
                    warn "\tERROR - Several versions for the same species: " . join(' ',@a_short_same_prefix) . "\n";
                    $error++;
                };
                if (@a_short_same_prefix==2)
                {
                    my $short0 = $a_short_same_prefix[0];
                    my $short1 = $a_short_same_prefix[1];
                    my $long0 = $h_list_by_short{ $short0 };
                    my $long1 = $h_list_by_short{ $short1 };
                    $long0 =~ s/ NEW UPGRADE$//;
                    $long1 =~ s/ NEW UPGRADE$//;
                    if ($long0 ne $long1 )
                    {
                        warn "\tERROR - When upgrading a species, the long name MUST BE UNCHANGED: $long0 vs $long1\n";
                        $error++;
                    }
                    if ( $ho_param{$short0}->GetInteger('ORG_NEW')==1)
                    {
                        $ho_param{$short1}->Set('ORG_NEW',-1);
                    }
                    else
                    {
                        $ho_param{$short0}->Set('ORG_NEW',-1);
                    }
                }
                $old_prefix = $prefix;
                @a_short_same_prefix = ();
                push(@a_short_same_prefix,$short);
            }
        }
    }

    if ($error > 0)
    {
        die "\tERRORS found ($error) - Please correct the found problems and try again\n";
    }

    return \%ho_param;
}

=head2 procedure UpdateSpeciesFromRepository

 Title        : UpdateSpeciesFromRepository
 Usage        : UpdateSpeciesFromRepository($o_io,$f_list)
 Procedure    : For each species found in $f_list, call BuildSpeciesCfg to create the xxx.cfg file
 Args         : $o_io, $f_list

=cut
sub UpdateSpeciesFromRepository
{
    my ($o_io,$species_file) = @_;
    my $o_logger = $o_io->GetLogger();

    my $unix_wget = $o_io->GetUnixPath('UNIX_WGET');
    my $unix_ls   = $o_io->GetUnixPath('UNIX_LS');

    $o_logger->Trace("Generating xxx.cfg files from the repository\n","*","A");
    my $root_species = $o_io->GetSpeciesCfgDirpath('INSTANCE');
    my $repository   = $o_io->GetCfg('NARCISSE_GENOME_REPOSITORY');
    if ($repository eq '')
    {
        if (!exists $ENV{NARCISSE_GENOME_REPOSITORY})
        {
            die "ERROR - The parameter NARCISSE_GENOME_REPOSITORY OR the environnement variable NARCISSE_GENOME_REPOSITORY should be defined if you specify --species_file";
        }
        else
        {
            $repository = $ENV{NARCISSE_GENOME_REPOSITORY};
        }
    }
    if (($repository !~ /^file:/) && ($repository !~ /^\//) )
    {
        die "ERROR - The  NARCISSE_GENOME_REPOSITORY parameter must start with file:/// or /\n";
    }
    else
    {
        $repository =~ s/^file:\/\///;
        die "ERROR - Could not find the repository $repository\n" unless (-d $repository);
    }

    # If speclist.txt does not exist, try to retrieve it with wget
    my $root_speclist     = $o_io->GetCfg('ROOT_SPECLIST');
    my $url_speclist = $o_io->GetCfg('URL_SPECLIST');
    if (! -r $root_speclist)
    {
        $o_logger->Trace("Trying to retrieve the speclist file with wget\n","*","A");
        my $cmd = "$unix_wget -O $root_speclist $url_speclist";
        $o_logger->Trace("$cmd\n",".","ACV");
        `$cmd`;
        if ( ! -r $root_speclist)
        {
            $o_logger->Trace("Failed\n","*","A");
            my $msg = "ERROR - The file speclist cannot be retrieved using wget $url_speclist\n";
            $msg .= "Please retrieve this file by hand and copy it to $root_speclist\n";
            if (&NARCISSE_utils::GetAdmAnswer("OK to continue ?","Y|N","N") ne 'Y')
            {
                die "Interrupted by the user";
            }
        }
        else
        {
            $o_logger->Trace("OK\n","*","A");
        }
    }

    print "Trying to build the species/xxx.cfg files from a genome repository\n";

    # Keeping in a hash the list of long_names/short_names, to avoid creating several times the same file
    my @a_spec_cfg = glob "$root_species/*.cfg";
    my %h_long2short;
    chomp(@a_spec_cfg);
    foreach my $f (@a_spec_cfg)
    {
        next if ($f =~ /template.cfg$/);
        my $o_p = New ParamParser($f);
        my $short = $o_p->Get('ORG_SHORT');
        my $long  = $o_p->Get('ORG_LONG');
        if (($long eq '') or ($short eq ''))
        {
            warn "WARNING - file $f corrupted - ignored\n";
            next;
        };
        $long =~ s/ /_/g;
        $h_long2short{$long} = $short;
    };

    # Calling BuildSpeciesCfg one or several times for each line of species_file
    my $species = new IO::File ($species_file) or die "ERROR: unable to read >$species_file<\n";
    while (my $lign = <$species>)
    {
		chomp($lign);
		next if ( $lign =~ /^#/ || $lign eq "" );        # Empty line or comment are ignored
		my $lign_ori = $lign;
		$lign =~ s/\s+/_/;        # Try
		my($first_char,$others) = $lign =~ /(\w)(.+)/;
		$lign = "\U$first_char\L$others";    # Try anaplasma marginale st maries ==> Anaplasma_marginale_st_maries
		my @a_ls = `$unix_ls -d $repository/$lign_ori 2> /dev/null`;
		if ( $#a_ls == -1 )
		{
			@a_ls = `$unix_ls -d $repository/$lign* 2> /dev/null`;
		}
		if ( $#a_ls == -1 )
		{
			warn "WARNING - $lign not found, skipped !!\n";
            next;
		}
        chomp(@a_ls);
        foreach my $org_dir (sort @a_ls)
        {
            &BuildSpeciesCfg($o_io,$org_dir,$lign,\%h_long2short);
        };
    };
    undef $species;
    $o_logger->Trace("Finished\n","*","A");
}


=head2 procedure BuildSpeciesCfg

 Title        : BuildSpeciesCfg
 Usage        : BuildSpeciesCfg($orgdir,$species,\%h_long2short)
 Prerequisite : none
 Fonction     : Creates the species/xxx.cfg files corresponding to $spec
 Returns      : 2 if already existing, 1 if created, 0 if failed
 Args         : $o_io        = The Instance Observer
                $orgdir      = The directory where are the files of the spec to consider
                $species     = A line of species_files
                %h_long2short= The species already recorded in a .cfg file

=cut
sub BuildSpeciesCfg
{
    my ($o_io,$orgdir,$species,$rh_long2short) = @_;

    my $cfgdir   = $o_io->GetSpeciesCfgDirpath('INSTANCE');
    my $speclist = $o_io->GetCfg('ROOT_SPECLIST');
    my $unix_ls  = $o_io->GetUnixPath('UNIX_LS');

    my $ext = "\LABCDEFGHIJKLMNOPQRSTUVWXYZ";
    my @a_ext = split('',$ext);
    # perl autorise @a_ext = (a..z); <- seb

	my ($str) = $orgdir =~ /${species}_*(.+)/;
	my ($dir,$longname) = $orgdir =~ /(.+)\/(.+)/;

    # Remove heading or trailling _ or spaces
    $longname =~ s/^[ _]//;
    $longname =~ s/[ _]$//;

    # space ==> _ and remove double spaces
    $longname =~ s//_/g;
    $longname =~ s/__/_/g;

    # Remove illegal characters
    $longname =~ s/[^a-zA-Z0-9\-_]//g;

    if (exists $$rh_long2short{$longname})
    {
        my $short = $$rh_long2short{$longname};
        warn "WARNING - $longname is already recorded (short name = $short) - skipped\n";
        return 2;
    };

    my $prefix="";
	my ($init1,$init2) = $longname =~ /^(\w)[^_]+_(\w\w)/;
    if (defined($init1) and defined($init2))
    {
        $prefix = "\L$init1$init2";
    }
    else
    {
        $prefix = "\L$longname";
    }

	# Use default sorting method for ls, ie alphanumeric from the file name: this is most often the correct order
	# for affecting chromosome numbers
	my @a_ls_gbk = `$unix_ls  $orgdir/*.gbk`;
    if (@a_ls_gbk==0)
    {
        warn "WARNING - There is no .gbk file in directory $longname - skipped";
        return 0;
    }

	#print STDERR @a_ls_gbk;
	chomp(@a_ls_gbk);
	my $taxid = &_GetTaxIDFromGbk($o_io,$a_ls_gbk[0]);
	my $shortfromspecies = ($taxid eq "") ? "" : &NARCISSE_utils::TaxIDtoSwissProtID($taxid,$speclist);
	$shortfromspecies = "\L$shortfromspecies";
	my $shortname = ( $shortfromspecies eq "" ) ? $prefix : $shortfromspecies;
	my $begin = ( $shortfromspecies ne "" ) ? 1 : 0;
	if ( $shortfromspecies ne "" && ! -e "$cfgdir/$shortname.cfg" && $shortfromspecies ne 'none' )
	{
		$shortname = $shortfromspecies;
	}
	else
	{
		for(my $i=$begin;$i<=$#a_ext;$i++)
		{
            my $tmp_name = $shortname . $a_ext[$i];
            next if ($tmp_name eq 'none');
			my $cfg_tmp = "$cfgdir/${tmp_name}.cfg";
			if ( ! -e $cfg_tmp )
			{
				#print "cfg=$cfg_tmp\n";
				$shortname = "$tmp_name";
				last;
			}
		}
	}
	$shortname = "\L$shortname";
	my $file_cfg = "$cfgdir/$shortname.cfg";

	my $cfg = new IO::File(">$file_cfg") or die "ERROR $0/BuildNarcisseCfgFile >$file_cfg<\n";
	print $cfg "DST=%d\n";
	print $cfg "ORG_SHORT=$shortname\n";

    # remove the _ in the long name, as this will be used ONLY for display purpose
    my $longname_for_display = $longname;
    $longname_for_display =~ s/_/ /g;

    my $f = basename $a_ls_gbk[0];
    my $d = basename(dirname $a_ls_gbk[0]);

	print $cfg "ORG_LONG=$longname_for_display\n";
	print $cfg "ORG_TAXID=$taxid\n";
	print $cfg "RELEASE_DATE=".&ExtractReleaseDateFromGbk($o_io,$orgdir)."\n";
	#print $cfg "FILE_ASSEMBLY=\n";
	#print $cfg "FILE_MASKING=\n";
	#print $cfg "FILE_IDEOGRAM=\n";

    # It is better using a directory name for TIMESTAMP
	print $cfg "TIMESTAMP_FILE=%g/$d/\n";
	for (my $i=1;$i<=@a_ls_gbk;$i++)
	{
        my $f = basename $a_ls_gbk[$i-1];
		print $cfg "FILE_GBK_${i}=%g/$d/$f\n";
	}
	undef $cfg;
    print "File $shortname.cfg created for $longname\n";
    return 1;

    sub _GetTaxIDFromGbk
    {
        my ($o_io,$file) = @_;
        my $taxid = "";
        return '' if ( $file eq "");

        my $unix_grep = $o_io->GetCfg('UNIX_GREP');
        my @a_grep_taxon = `$unix_grep \"db_xref=.taxon:\" $file`;

        if (defined $a_grep_taxon[0])
        {
            ($taxid) = $a_grep_taxon[0] =~ /taxon:(\d+)/;
        }
        return (defined $taxid) ? $taxid : "";
    }
}

=head2 function ExtractReleaseDateFromGbk

 Title        : ExtractReleaseDateFromGbk
 Usage        : my $release_date = ExtractReleaseDateFromGbk($o_io, $directory)
 Fonction     : Extract the date from the LOCUS line of each .gbk file found in the diretory
                Sort the dates and return the most recent one
 Returns      : The most recent date found (=release date)
 Args         : $o_io The instance observer
                $directory The directory to list
 Globals      : none

=cut
sub ExtractReleaseDateFromGbk
{
	my ($o_io,$directory) = @_;

	my %h_m = ('JAN','01','FEB','02','MAR','03','APR','04','MAY','05','JUN','06','JUL','07','AUG','08','SEP','09','OCT','10','NOV','11','DEC','12');
	my @a_m = ('','Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');
	my $unix_grep = $o_io->GetCfg('UNIX_GREP');
	my @a_dates;
	my @a_sorted_dates;
	
	my @a_locus_lines = `$unix_grep ^LOCUS $directory/*.gbk`;
	chomp @a_locus_lines;
	foreach my $l (@a_locus_lines)
	{
		$l =~ s/.* (..)-(...)-(....)$/$3-$h_m{$2}-$1/;
        push @a_dates, $l;
    };
    @a_sorted_dates = sort @a_dates;
    my $release_date = $a_sorted_dates[-1];
    $release_date =~ s/(\d{4})-(\d{2})-(\d{2})/$3 $a_m[$2] $1/;
    return $release_date;
}
       
=head2 procedure CreateSpeciesListFile

 Title        : CreateSpeciesListFile
 Usage        : CreateSpeciesListFile($file_list,$rh_list)
 Fonction     : Cretaes the file species_list from $rh_list
 Returns      : none
 Args         : $file_list : fichier species_list
              : $rh_list : a ref to an array of ParamParser objects
                  TODO tax_id should NOT be optional
 Globals      : none
 ERROR        : dies if some file has 'none' as a short name

=cut
sub CreateSpeciesListFile
{
	my $file_list = shift;
	my $rho_param = shift;

	my $file_list_old = $file_list."_old";
    link ($file_list,$file_list_old);

    my $out = new IO::File(">$file_list") or die "Cannot open $file_list";
	print $out "#short\tlong\twhere\ttaxid\tassembly\tassembly_phase\trelease\treldate\ttype\tupgrade\n";
	foreach my $key ( keys %$rho_param )
	{
        my $o_param = $rho_param->{$key};
        $o_param->SetUnlessDefined('ORG_NEW',0);
        my $short = $o_param->Get('ORG_SHORT');
        my $long  = $o_param->Get('ORG_LONG');
        my $upgrade = $o_param->Get('ORG_NEW');
        my $where = "SESSION,INSTANCE";
        my $tax_id = $o_param->Get('ORG_TAXID');
        my $assembly = $o_param->Get('ASSEMBLY');
        my $assembly_phase = $o_param->Get('ASSEMBLY_PHASE');
        my $release = $o_param->Get('RELEASE');
        my $reldate = $o_param->Get('RELEASE_DATE');
        #TODO Dans l'avenir dissocier type de donnees et format ! (GENBANK et GENETIC ne devraient pas etre mis sur le meme plan)
        $o_param->SetUnlessDefined('FORMAT','GENBANK');
        my $type = ($o_param->Get('FORMAT') eq 'GENETIC') ? 'genetic' : 'genomic';
		print $out "$short\t$long\t$where\t$tax_id\t$assembly\t$assembly_phase\t$release\t$reldate\t$type\t$upgrade\n";
	}
    undef ($out);
    unlink ($file_list_old);
}

=head2 function CleanPath

 Title        : CleanPath
 Usage        : $cleaned_path = CleanPath($path)
 Prerequisite : none
 Fonction     : Replace // by / in the path
 Returns      : The cloned path
 Args         : $path - The path to clean
 Globals      : none

=cut

sub CleanPath
{
    my $path = shift;
    $path =~ s|//|/|g;
    $path =~ s|/ | |g;
    return $path;
};

=head2 function FindOrgToUpdate

 Title        : FindOrgToUpdate
 Usage        : my $ra_list = &FindOrgToUpdate();
 Prerequisite : none
 Fonction     : Build an array with the short names of the organisms which must be updated ,according to nc_get_data.pl
 Returns      : The list of orgs to update
 Args         : $o_io

=cut
sub FindOrgToUpdate
{
    my ($o_io) = @_;

    my $root_species     = $o_io->GetSpeciesCfgDirpath('INSTANCE');
    my $root_nc_get_data = $o_io->GetCfg('ROOT_NC_GET_DATA');

    my @a_species_to_update;
    foreach my $species ($o_io->GetSpeciesShortName())
    {
        my $result = `$root_nc_get_data --cfg $species --only_check`;
        push (@a_species_to_update,$species) if ($result =~ /-1-/);
    };
    return \@a_species_to_update;
};

=head2 procedure MkDir

 Title        : MkDir
 Usage        : &MkDir($dst,$src)
 Prerequisite : none
 Procedur     : Make a directory, the new directory will have the same permissions as the old one
 Returns      : none
 Args         : $dst = The new directory to create
                $src = The old directory we take the permissions from
 Globals      : none

=cut

sub MkDir
{
    my $dst   = shift;
    my $src   = shift;
    my $perms = (stat($src))[2] & 07777;

    unless (-d $dst)
    {
        mkdir ($dst) or croak "ERROR - Cannot create directory $dst";
    }
    chmod $perms,$dst or croak "ERROR - Cannot change the mode of $dst";
};

=head2 procedure Link

 Title        : Link
 Usage        : &Link($dst,$src)
 Prerequisite : none
 Procedure    : Create a new inode, hard linking it to the old one
                change the permissions of the new inode from the old one
 Returns      : none
 Args         : $dst = The new inode to create
                $src = The old inode
 Globals      : none

=cut
sub Link
{
    my $dst = shift;
    my $src = shift;
    my $perms = (stat($src))[2] & 07777;

    link $src,$dst or croak "ERROR - Cannot create file $dst";
    chmod $perms,$dst or croak "ERROR - Cannot change the mode of $dst";
}


=head2 function Ask

 Title        : Ask
 Usage        : my $variable = &Ask("Enter something ",'DEFAULT',\&Check);
 Fonction     : Print a message to stdout and read the answer from stdin
                If '' is read, the default value is taken
                If specified, the function \&Check is called, and if this function returns FALSE
                the message is printed again, and stdin is read again
 Returns      : The value entered by the user, validated
 Args         : $msg : The message to print
                $default(optional): The default value
                \&Check (optional): The function called for validation

=cut
sub Ask
{
    my ($msg,$default,$test) = @_;
    $default = "" unless defined $default;

    my $answer = "";
    do
    {
        print $msg;
        if ($default ne '')
        {
            print " ($default) ";
        }
        $answer = <STDIN>;
        chomp($answer);
        if ( $answer eq "" )
        {
            $answer = $default;
        }
    }
    while (defined $test and &$test($answer) == &FALSE);
    return $answer;
}

=head2 functions Blank,,BlankExists

 Title    : blank,
 Usage    : through the Ask function
 Fonction : Input validation - Return &FALSE if not validated
 Args     : $value  The value to validate

=cut

# Blank = must be ne ""
sub Blank
{
    return ($_[0] !~ /^\s*$/) ? &TRUE : &FALSE;
}

# BlankExist = must be ne "" and must represent an existing file
sub BlankExist
{
    return ($_[0] ne "" and -e $_[0]) ? &TRUE: &FALSE;
}

# BlankCreateDir = must be ne "" and if directory does not exist, create it
#                  (thus it is not quite a check....)
sub BlankCreateDir
{
    my ($dir) = @_;

    return &FALSE if ( $dir eq "" );
    return &FALSE if ( -e $dir && ! -d $dir );
    return &TRUE  if ( -d $dir );
    return mkdir $dir;
}

# BlankKindom = must be eq 'eukaryote' or 'bacteria'
sub BlankKingdom
{
    return ($_[0] eq 'bacteria' or $_[0] eq 'eukaryote') ? &TRUE : &FALSE;
}

# BlankEmail = must contain an @ character
sub BlankEmail
{
    return ($_[0] =~ /^.+@.+/) ? &TRUE : &FALSE;
}

# A very simple sub to print and log some information
sub PrintLogInfo
{
    my ($o_conf) = @_;
    my $o_logger = $o_conf->GetLogger();
    my $flg_multi= $o_conf->isa('NarcisseObserver');

    print "INSTANCE NAME      = ".$o_conf->GetName()."\n";
    print "INSTANCE VERSION   = ".$o_conf->GetCfg('INSTANCE_VERSION')."\n" unless($flg_multi);
    print "ROOT DIRECTORY     = ".$o_conf->GetCfg('ROOT')."\n";
    print "SERVER URL         = ".$o_conf->GetCfg('HTTP')."\n";
    print "DATA DIRECTORY     = ".$o_conf->GetCfg('ROOT_DATA')."\n";
    print "KINGDOM            = ".$o_conf->GetCfg('KINGDOM')."\n" unless($flg_multi);
    print "GENOME REPOSITORY  = ".$o_conf->GetCfg('NARCISSE_GENOME_REPOSITORY')."\n" unless($flg_multi);
	print "SNP REPOSITORY     = ".$o_conf->GetCfg('NARCISSE_SNP_REPOSITORY')."\n" unless($flg_multi);

    print "ADMIN MAIL         = ".$o_conf->GetCfg('NARCISSE_ADM_EMAIL')."\n";
    print "NARCISSE INSTANCES = ".$o_conf->GetCfg('NARCISSE_INSTANCE_ROOTS')."\n" if($flg_multi);
    print "\n";
    if ($flg_multi)
    {
        $o_logger->Trace("\nROOT DIRECTORY=".$o_conf->GetCfg('ROOT')."\nURL           =".$o_conf->GetCfg('HTTP')."\nDATA DIRECTORY=".$o_conf->GetCfg('ROOT_DATA')."\nADMIN MAIL    =".$o_conf->GetCfg('NARCISSE_ADM_EMAIL')."\nINSTANCES     =".$o_conf->GetCfg('NARCISSE_INSTANCE_ROOTS')."\n","","");
    }
    else
    {
        $o_logger->Trace("\nROOT DIRECTORY=".$o_conf->GetCfg('ROOT')."\nURL           =".$o_conf->GetCfg('HTTP')."\nDATA DIRECTORY=".$o_conf->GetCfg('ROOT_DATA')."\nKINGDOM       =".$o_conf->GetCfg('KINGDOM')."\nADMIN MAIL    =".$o_conf->GetCfg('NARCISSE_ADM_EMAIL')."\n\n","","");
    }
}


=head2 procedure Usage

 Title        : USAGE
 Usage        : USAGE($code)
 Prerequisite : none
 Fonction     : fonction usage
 Returns      : none
 Args         : none
 Globals      : none

=cut
sub Usage
{
    my $prgm = basename $0;
    print <<END

Usage : $prgm

THE FOLLOWING SHOULD BE PRECEDED WITH THE COMMAND:

    chins

$prgm --new_instance some_name     Create a new instance
$prgm --show_instance some_name    Show parameters relative to some instance
$prgm --edit_instance some_name    Reconfigure the main parameters of an existing instance
$prgm --remove_instance some_name  Remove an existing instance
$prgm --upgrade_instance some_name Upgrade (if necessary) an existing instance

THE FOLLOWING SHOULD BE PRECEDED WITH THE COMMAND:

    chins some_instance_name

$prgm    [--quiet]          Configure the instance, soem messages may be removed (--quiet)
$prgm    --rebuild_web      Rebuild some files related to the web, clear the cache, and leave
$prgm    --clear_cache      Clear the server-side cache and leave
$prgm    --motd             Change the Message Of The Day for the instance or for all the instances if working in the instance called "multi"
$prgm    --work_in_progress Temporarily disable/enable the instance of all the instances (if multi), recording an explanation message

THE FOLLOWING SHOULD BE PRECEDED WITH THE COMMAND:

    chins multi

$prgm                     Declare the Narcisse installed instances
$prgm    --motd           Change the Message Of The Day for the whole Narcisse installation

END
;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

