#!/usr/bin/perl
#
# $Id: nc_create_gff3.pl 3701 2011-07-27 12:36:33Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

 nc_create_gff3.pl

=head1 SYNOPSIS

 nc_create_gff3.pl --species <short_name> 

=head1 DESCRIPTION

 Creates a gff3 file for each chromosome for the species specified. Those files are put
 inside the data/short_name/gff3 directory, they are NOT used by Narcisse, but may be useful for exportation to other projects
 
=cut

use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";
use strict;
use warnings;
use ParamParser;
use File::Basename;
use Bio::Narcisse::Utils::NarcisseDataReader;
use Configuration::InstanceObserver;
use File::Path;
use Data::Dumper;
use URI::Escape;

my $VERSION = do {my @r = (q$Rev: 3701 $ =~ /\d+/g); $r[0]};

# Main program
MAIN:
{
	
	
    if  ( ! exists $ENV { 'NARCISSE_CFG' } )
    {
        die "ERROR - Please execute chins <some_instance> before running this program\n";
	}

	ParamParser::SetDefaultUsage( \&Usage );
	ParamParser::SetDefaultBehaviour( 'exit_on_getopt_error' );
	my $o_param = New ParamParser('GETOPTLONG','species=s','help');
	$o_param -> AssertDefined ( 'species' );
	my $species = $o_param -> Get ( 'species' );
    my $o_io = New InstanceObserver();
	my $o_sp = $o_io -> GetSpeciesObserver ( $species );
	print "SPECIES : " . $o_sp -> GetLongName() . " (short name " . $o_sp -> GetShortName() . ")\n";
	my $ra_chr_names = $o_sp -> GetChromsNames();
	
	
	foreach my $chr ( @$ra_chr_names )
	{
		&CreateGff3 ( $o_sp, $chr);
		print "$chr\n";
	}
	exit;	
}

=head2 procedure CreateGff3

 Usage        : &CreateGff3 ( $o_sp, $chr );
 Procedure    : Lit les fichiers Narcisse (gene, transcript, exon, repeat, rna) et genere un fichier gff3
 Args         : $o_sp Un species Observer
                $chr  Le chromosome pour lequel on genere le fichier gff3
                $compute_cds Si true, on recalcule les cds en utilisant l'API EnsEMBL (manu - 25 mai 2011 pour Erika/caeel)

=cut
sub CreateGff3
{
	my ($o_sp, $chr) = @_;
	
	my $chr_len = $o_sp -> GetChromLength ( $chr );
	#my $phase = '0';		# ONLY for gff3 validation !
	my $phase = '.';        # for production (gff3 NOT validated !)
	my $o_ndr;
	my $exon_file = $o_sp -> GetExonFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $exon_file );
	my $ra_exons = $o_ndr -> FetchallArrayref ( );

	my $rh_exons = &Array2Hash($ra_exons);
	$ra_exons = undef;
	$o_ndr = undef;
	
	my $transcript_file = $o_sp -> GetTranscriptFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $transcript_file );
	my $ra_transcripts = $o_ndr -> FetchallArrayref ( );
	
	# Keep the row inside a hash, whose key is 'parents', not elements
	my $rh_transcripts = &Array2Hash($ra_transcripts);

	$ra_transcripts = undef;
	$o_ndr = undef;
	
	my $gene_file = $o_sp -> GetGeneFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $gene_file );
	my $ra_genes = $o_ndr -> FetchallArrayref ();
	
	# Read the prot file, for the annotations
	my $prot_file = $o_sp -> GetProtFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $prot_file );
	my $rh_prot = $o_ndr -> FetchallHashref ( 'element' );
	
	# read the rna file
	my $rna_file = $o_sp -> GetRnaFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $rna_file );
	my $ra_rna = $o_ndr -> FetchallArrayref ();

	# read the repeats file
	my $repeats_file = $o_sp -> GetRepeatFiles ( -chrom => $chr );
	$o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();
	$o_ndr -> Prepare ( -from => $repeats_file );
	my $ra_repeats = $o_ndr -> FetchallArrayref ();

	# The gff3 directory is not supported by SpeciesObserver
	my $data_dir = $o_sp -> GetDataPath();
	mkdir File::Spec->catfile( $data_dir, 'gff3') ;
	##########$data_dir = File
	my $file_gff3 = File::Spec->catfile( $data_dir, 'gff3', 'chromosome' . $chr . '.gff3' );
	my $fh_gff3   = new IO::File ( $file_gff3, 'w' ) || die "ERROR - Cannot open $file_gff3";

	# Create an EnsEMBL API DBAdaptor
	print $fh_gff3 "##gff-version 3\n##sequence-region $chr 1 $chr_len\n";
	foreach my $rh_row ( @$ra_genes )
	{
		my $start = $rh_row -> { start };
		my $end   = $rh_row -> { end };
		my $strand= &CvtStrand($rh_row -> { strand });
		my $gene_id   = $rh_row -> { element };
		my $locus_tag = &uri_escape($rh_row -> { locus_tag },",=;");
		my $symbol    = &uri_escape($rh_row -> { symbol },",=;");
		
		print $fh_gff3 join( "\t", $chr, 'narcisse', 'gene', $start, $end, '.', $strand, $phase, "ID=$gene_id" );
		print $fh_gff3 ";Name=$locus_tag";
		print $fh_gff3 ";Alias=$symbol" if ($symbol ne '');
		print $fh_gff3 "\n";
		
		my $ra_transcripts = $rh_transcripts -> { $gene_id };
		foreach my $rh_transcript ( @$ra_transcripts )
		{
			my $transcript_id = $rh_transcript -> { element };
			my $start = $rh_transcript -> { start };
			my $end   = $rh_transcript -> { end };
			my $strand= &CvtStrand($rh_transcript -> { strand });
			my $locus_tag = &uri_escape($rh_transcript -> { locus_tag }, ",=;");
			my $symbol    = &uri_escape($rh_transcript -> { symbol }, ",=;");
			my $accession = &uri_escape($rh_transcript -> { accession }, ",=;");
			my $protein_id= &uri_escape($rh_transcript -> { protein_id }, ",=;");
			my $consistency = $rh_transcript -> { consistency };
			my @a_regions = split ',', $rh_transcript -> { regions };

			my $prot_id = $transcript_id;
			$prot_id =~ s/t(\d+)$/p$1/;
			my $product = (exists $rh_prot -> { $prot_id }) ? &uri_escape($rh_prot -> { $prot_id } -> { 'product' }, ",=;"):'';
			print $fh_gff3 join( "\t", $chr, 'narcisse', 'mRNA', $start, $end, '.', $strand, $phase, "ID=$transcript_id;Parent=$gene_id" );
			print $fh_gff3 ";Name=$locus_tag";
			print $fh_gff3 &List2Alias($locus_tag,$symbol,$accession);
			print $fh_gff3 ";protein_id=$protein_id" if ($protein_id ne '');
			print $fh_gff3 ";product=$product" if ($product ne '');
			print $fh_gff3 ";Note=inconsistent" if ($consistency eq 'I');
			print $fh_gff3 "\n";
			
			# The exons -
			my $ra_exons = $rh_exons -> { $transcript_id };
			foreach my $rh_exon ( @$ra_exons )
			{
				my $exon_id = $rh_exon -> { element };
				my $start = $rh_exon -> { start };
				my $end   = $rh_exon -> { end };
				my $strand= &CvtStrand($rh_exon -> { strand });
				my $symbol    = &uri_escape($rh_transcript -> { symbol }, ",=;");
				
				print $fh_gff3 join( "\t", $chr, 'narcisse', 'exon', $start, $end, '.', $strand, $phase, "ID=$exon_id;Parent=$transcript_id" );
				print $fh_gff3 ";Name=$symbol" if ($symbol ne '');
				print $fh_gff3 "\n";
			}
			delete $rh_exons -> { $transcript_id };
			
			# The CDS - This section is skipped if the transcript is NOT consistent
			#if ( $rh_transcript -> { consistency } eq 'C' )
			#{
				my @a_regions_start = @a_regions;
				map $_=~s/(\d+)\.\.(\d+)/$1/,@a_regions_start;
				my @a_regions_end   = @a_regions;
				map $_=~s/(\d+)\.\.(\d+)/$2/,@a_regions_end;
				
				my @a_coding_start = @a_regions_start;
				my @a_coding_end   = @a_regions_end;
				my @a_shift;
				
				# Now shifting the regions to get a start position = 1 and to exclude the introns
				for ( my $i=0;$i<@a_regions;$i++)
				{
					$a_shift[$i]         = ($i==0) ? $a_coding_start[0] - 1 : $a_coding_start[$i] - $a_coding_end[$i-1] - 1;
					$a_coding_start[$i] -= $a_shift[$i];
					$a_coding_end[$i]   -= $a_shift[$i];
				}
				
				# Now correcting the start and end positions
				my $start_translation = $rh_transcript -> { start_translation };
				my $stop_translation  = $rh_transcript -> { stop_translation };
				
				my $ttl_length = $a_coding_end[$#a_coding_end];
				
				if ( $strand eq '-' )
				{
					my $tmp = $ttl_length - $start_translation + 1;
					$start_translation = $ttl_length - $stop_translation + 1;
					$stop_translation  = $tmp;
				}
				
				# first clearing the exons outside coding regions
				for ( my $i=0;$i<@a_regions;$i++)
				{
					if ( $a_coding_end[$i] < $start_translation )
					{
						$a_coding_start[$i] = 0;
						$a_coding_end[$i]   = 0;
					}
					if ( $a_coding_start[$i] > $stop_translation )
					{
						$a_coding_start[$i] = 0;
						$a_coding_end[$i]   = 0;
					}
				}
				
				# Then correcting the start position
				for ( my $i=0;$i<@a_regions;$i++)
				{
					if ( $a_coding_start[$i] == 0 )
					{
						next;
					}
					else
					{
						$a_coding_start[$i] = $start_translation;
						last;
					}
				}
				
				# Last correcting the end position
				for ( my $i=$#a_regions;$i>=0;$i--)
				{
					if ( $a_coding_start[$i] == 0 )
					{
						next;
					}
					else
					{
						$a_coding_end[$i] = $stop_translation;
						last;
					}
				}
				
				for (my $i=0; $i<@a_regions; $i++)
				{
					if ( $a_coding_start[$i] == 0 )
					{
						next;
					}
					else
					{
						#my $region       = $a_regions[$i];
						#my ($start,$end) =  split '\.\.',$region;
						my $cds_nb = $i + 1;
						my $cds_id = $transcript_id.'c'.$cds_nb;
						#print $fh_gff3 join( "\t", $chr, 'narcisse', 'CDS', $start, $end, '.', $strand, $phase, "ID=$cds_id;Parent=$transcript_id\n" );
						print $fh_gff3 join( "\t", $chr, 'narcisse', 'CDS', $a_coding_start[$i]+$a_shift[$i], $a_coding_end[$i]+$a_shift[$i], '.', $strand, $phase, "ID=$cds_id;Parent=$transcript_id\n" );
					}
				}
			#}
		}
		delete $rh_transcripts -> { $gene_id };
	}
	foreach my $rh_row ( @$ra_rna,@$ra_repeats )
	{
		my $start = $rh_row -> { start };
		my $end   = $rh_row -> { end };
		my $strand= &CvtStrand($rh_row -> { strand });
		my $id    = $rh_row -> { element };
		my $symbol    = &uri_escape($rh_row -> { symbol }, ",=;");
		my $type      = $rh_row -> { type };
		$type = 'ncRNA'         if ($type eq 'misc_RNA');
		$type = 'repeat_region' if ($type eq 'repeat');
		print $fh_gff3 join( "\t", $chr, 'narcisse', $type, $start, $end, '.', $strand, $phase, "ID=$id" );
		print $fh_gff3 ";Name=$symbol" if ($symbol ne '');
		print $fh_gff3 "\n";
	}
	
}

# RevcompPos et Exchange finalement PAS UTILISES - MANU
sub RevCompPos
{
	my ( $ra_positions, $length) = (@_);
	foreach my $pos ( @$ra_positions )
	{
		if ( $pos == 0 )
		{
			next;
		}
		else
		{
			$pos = $length - $pos + 1;
		}
	}
}

sub Exchange
{
	my ( $ra_1, $ra_2 ) = @_;
	my @a_3 = @$ra_1;
	@$ra_1 = @$ra_2;
	@$ra_2 = @a_3;
}
								
					

=head2 fonction List2Alias

 Usage    : my $alias = &List2Alias($locus_tag,@a_liste_de_noms)
 Fonction : Les parametres sont une liste de noms, qui seront mis sous la forme alias=nom1,nom2...
            a moins qu'ils soient tous nuls (''), dans ce cas la fonction retourne ''
            S'il y a des noms dupliques on ne renvoie qu'un seul alias
            Un nom egal a $locus_tag sera elimine avant traitement
 Args     : Une liste de noms, eventuellement une liste de ''
 Return   : Soit '', soit "Alias=nom1,nom2..."

=cut
sub List2Alias
{
	my ($locus_tag,@a_list	) = @_;
	my %h_corrected_list;
	foreach my $item ( @a_list)
	{
		$h_corrected_list{$item} = 1 if ($item ne '' && $item ne $locus_tag);
	}
	my $alias = '';
	my @a_corrected_list = keys %h_corrected_list;
	if (@a_corrected_list != 0)
	{
		$alias = ';Alias=' . join(',',@a_corrected_list);
	}
	return $alias;
}

=head2 fonction Array2Hash

 Usage    : my $rh_exons = &Array2Hash($ra_exons);
 Fonction : A partir d'une $ra_ retourne par FetchAllArrayref, cree un hash:
              -> La cle est un identifiant de parent
              -> La valeur est une liste de lignes qui partagent le meme parent
            a moins qu'ils soient tous nuls (''), dans ce cas la fonction retourne ''
 Args     : Ref d'array
 Return   : Ref de hash

=cut
sub Array2Hash
{
	my ($ra_data) = @_;
	
	my %h_data;
	my $last_key = '';
	foreach my $rh_row ( @$ra_data)
	{
		my $key = $rh_row -> { parent };
		my $ra_parent;
		if ( $key ne $last_key )
		{
			$last_key = $key;
			if (! exists $h_data { $key } )
			{
				$ra_parent = [];
				$h_data { $key } = $ra_parent;
			}
			else
			{
				$ra_parent = $h_data { $key };
			}
		}
		else
		{
			$ra_parent = $h_data { $key };
		}
		push @$ra_parent, $rh_row;
	}
	
	return \%h_data;
}

=head2 fonction CvtStrand

 Usage    : $strand = CvtStrand ( $strand)
 Fonction : Convertit une indication de Strand de (-1,+1) (conventions Narcisse) vers (-,+) (conventions gff3)

=cut
sub CvtStrand
{
	my ($strand) = @_;
	return ($strand==1)?'+':'-';
}
			
sub Usage
{
	my $self = shift;
	my $exe = basename $0;
	print <<END
Usage : $exe [--species] <species short or long name>

	Generate a gff3 file, from the Narcisse files containend in the species directory
END
;
	exit 1;
}

	
	

	


