#!/usr/bin/perl

=pod
=head1 NAME

	nc_genomic_comparison.pl

=head1 SYNOPSIS

	perl nc_genomic_comparison.pl @args

	@args :
	--cfgQ  :	string, # absolute path to the query org cfg file or simply the query short name
	--cfgS  :	string, # absolute path to the subject org cfg or simply the subject short name (useless if --self is set)
	
	--working_dir :	string, the working directory
	--min_span :	int,
	--a :	int,
	--e :	string,
	--G :	string,
	--E :	string,
	--z :	string,
	--path_blastall :	string, path to the blastall prog
	--blastresults  :	string, filename of the blastp output if already available. In this case no blastp comparison will be performed
	--glintresults  :	string, filename of the glint output if already available. In this case no glint comparison will be performed
	--chains_dir	:   string, read precomputed chin files from dir The files must be in glintchain format and with the following names : chain0, chain1, chain2 (no glint, nor blastp is computed)
	--chaining_param : providing user-defined chaining parameters (see defaults.cfg in the section "Param for building chains" for examples)
	--no_rbh        : no reciprocal best hit when building chains
	
	--(only|no)_proteic    :	flag, # (only|no) proteic comparison will be performed
	--(only_no)_nucleic    :	flag, # (only|no) nucleic comparison will be performed
	--(only|no)_build_chain :	flag, # (only|no) chain building will be performed
	--raw_blast     :	string, #tmp, give a path to a raw blast output, it will be used as input and no blast will be done, but it's different from no_blast
	--self :	flag, # if set, will compute cfgQ against itself
	--help :	flag
	--verbose :	flag, # to have exta information _debugging purpose

=head1 DESCRIPTION

	Compute genomic comparison, nucleic and proteic, and perfosm the chaining of the resulting alignments

=head1 Author

	Thomas.Faraut@toulouse.inra.fr

=head1 SUBROUTINEs

=cut

BEGIN
{
    use File::Basename;
    my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
}

use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf 
use File::Temp;
use File::Copy;
use File::Spec;
use IO::Dir;
use IO::Compress::Gzip qw(gzip $GzipError) ;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
use Cwd;

BEGIN
{		# pathway to our modules
		use File::Basename;
		my $dirprg=dirname($0);
		unshift @INC, "$dirprg/../lib";
		unshift @INC, "$dirprg/../API";
}


# nos modules, nos classes
use LipmError;
use General;
use Configuration::InstanceObserver;
use NARCISSE_utils;
use ParamParser;
use OnlyOrNo;
use Sufine;
use Logger;
use SpeciesObserver;
use Pairwise::Genomic;
use Pairwise::Genetic;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3859 $ =~ /\d+/g); $r[0]};
	
#arguments of the prog
use constant VERBOSE   => 'verbose';
use constant RAW_BLAST => 'raw_blast';
use constant SELF      => 'self';
use constant FORCE     => 'force';
use constant CFGQ      => "cfgQ";
use constant CFGS      => "cfgS";
use constant NO_RBH    => "no_rbh";
#use constant ROOT_LOG  => 'root_log';

use constant BLASTP_NC_HEADER => qw(q_name s_name pident size mismatch gaps q_start q_end s_start s_end evalue score q_len s_len q_cov s_cov); # __

# une classe pour gerer ce mecanisme serait bienvenue ...
# default values _in a closure to prevent overwritting
{
	# une explication de ces termes serait la bienvenue <- Seb
	# des arg du blast pour la plupart
	my %h_default = (			
			METHOD => 'genomic',
			#Specific to the blast comparison
			EVALUE	=>	"1e-3",
			O_GAP	=>	9,
			E_GAP	=>	2,
			L_DB	=>	15*10e6,
			MIN_SPAN=>	50,
			MIN_GSCORE => 30,
			#Specific to the genetic anchoring procedure
			MISMATCH_PERCENT_INTRA => 10,	# same taxid
			MISMATCH_PERCENT_INTER => 20,	# different taxid
			MISMATCH_NB_INTRA => 5,		
			MISMATCH_NB_INTER => 10,
			CPU 	=>	1
		);

	sub GetDefault
	{
		my ( $key ) = @_;

		return $h_default{ $key };
	}

	sub GetDefaults
	{
		my ( @a_keys ) = @_;
		my @a_values = ();

		foreach my $key ( @a_keys )
		{	push( @a_values, $h_default{ $key } );	}

		return @a_values;
	}
} # fin de la fermeture

#----------------------------------------------------------------------------

my $O_LOGGER;
my $O_PARAM;

MAIN:
{
    # to get the stacktrace in case of error
    $Error::Debug = 1;
   #&Main();
   # exit;
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ($err) = @_;
		print STDERR $err->Explain( 1 );
        print STDERR $err->stacktrace(), "\n";
        exit 1;
	}
	catch Error with
	{
		my ($err) = @_;

		print STDERR $err->text(), "\n";
        print STDERR $err->stacktrace(), "\n";
        exit 1;
		#&Verbose( Dumper $err );
	};
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut
sub Main
{
	# create an InstanceObserver and a ParamParser objects
	my ($o_io,$o_param) = &SetParamParser();

	# Read parameters, update paramparser object and return SpeciesObserver objects
	my ( $o_org_q, $o_org_s ) = &GetParametersFromCommandLine( $o_io, $o_param );

	# maintenant on peut ecrire dans un repertoire de session
	my $where = $o_param->IsDefined('sessdir') ? 'SESSION' : 'INSTANCE';
	my $root_data = $o_io->GetDataPaths( $where );
	$o_param->SetSubstitution( '%d', \$root_data );

	&NARCISSE_utils::Welcome(basename($0), &VERSION,
				$o_param->Get('NARCISSE_CFG'),$o_param->Get('NARCISSE_GENOME_REPOSITORY'));

	# compute the target path, create the directory and move to it.
	my $sav_dir = getcwd();
	my( $org_q, $org_s ) = ( $o_org_q->GetShortName(), $o_org_s->GetShortName() );
	my $target_log_path = &SetTargetLogPath( $o_io, $o_param, $org_q, $org_s );

	my $logger_file = ">>narcisse.log";
	$O_LOGGER = New Logger( $logger_file, $o_io->GetCfg('LOG_LEVEL') );
	$O_LOGGER->StartTimer( 'Timer' );

	my $o_pairwise = undef;

	my $o_only = New OnlyOrNo( $o_param );

	my $method = $o_param->Get('method');

    # Two methods supported yet: genomic or genetic
    # classes are called Pairwise::Genomic or Pairwise::Genetic
    my $class_name = 'Pairwise::' . &General::Capitalize( $method );
    $o_pairwise = New $class_name ('query'		=> $o_org_q,
        'subject'	=> $o_org_s, 
        'o_param'	=> $o_param,
        'o_io'		=> $o_io,
        'o_logger'	=> $O_LOGGER,
        '-where'    => $where,
        '-verbose'	=> $o_param->Get( &VERBOSE )
        );
    
    $o_pairwise->DoComparison()        if ( $o_only->Get( 'comparison' ) );
    $o_pairwise->DoOrthologDetection() if ( $o_only->Get( 'ortho' ) );
    $o_pairwise->DoBuildChain()        if ( $o_only->Get( 'build_chain' ) );

	# Back to the original directory
	chdir ($sav_dir);
	print "Log files created in directory: $target_log_path\n";

	return;
}


=head2 function SetTargetLogPath

 Title        : SetTargetLogPath
 Usage        : $target_path = &SetTargetLogPath( $o_io, $o_param, $orgQ, $orgS );
 Prerequisite : none
 Function     : computes the target log path and creates the directory
                It also redirect STDERR and SDTOUT.
 Returns      : $target_path, a string
 Args         : $o_io, $o_param, $org_q, $org_s

=cut

sub SetTargetLogPath
{
	my ( $o_io, $o_param, $org_q, $org_s ) = @_;

	# Compute and print $target, and lock the target
	my $target = $o_param->Get('method').'_'.$org_q.'_'.$org_s;
	print "target $target\n" . '-' x 30 . "\n";

	#my $logs_dir = $o_param->Get( &ROOT_LOG );
	#$logs_dir    = $o_io->GetLogsDirpath( 'INSTANCE' ) if( $logs_dir eq '' );
    my $logs_dir = $o_param->IsDefined( 'sessdir') ? $o_io->GetLogsDirpath( 'SESSION') : $o_io->GetLogsDirpath( 'INSTANCE' );
	# change current dir to target dir
	my @a_dirs = ( $logs_dir, $o_param->Get('method'), $org_q, $org_s );
	my $target_log_root = File::Spec->catfile( @a_dirs );

	# @a_created_dirs pour loguer plus tard.
	my @a_created_dirs = mkpath( $target_log_root );

	my $target_log_path = Sufine::MkDir(-dir => $target_log_root,-name => $target);
	die unless(defined $target_log_path);
	die unless(defined Sufine::ChDir(-dir => $target_log_root,-name => $target));

	&NARCISSE_utils::RedirectOE($o_io,"narcisse.out","narcisse.err");

	$o_param->Dump("narcisse.param");

	return $target_log_path;
}


=head2 Function SetParamParser

 Title        : SetParamParser
 Usage        : my ($o_io,$o_param) = &SetParamParser();
 Prerequisite : none
 Function     : Create an instance observer AND a ParamParser object
                	Update it from ENV
                	Do something relative to $NARCISSE_CFG
 Args         : none
 Return       : Both objects
 Globals      : $O_PARAM;

=cut

sub SetParamParser
{
	&ParamParser::SetDefaultBehaviour('use_substitution_table');

	my $o_param = New ParamParser;
	$o_param->SetUsage(my $usage=sub { &Usage(); } );

	# Update from ENV, for $NARCISSE_CFG
	$o_param->Update('ENV','O');   # for $NARCISSE_CFG
	$o_param->AssertDefined('NARCISSE_CFG');
	$o_param->AssertFileExists('NARCISSE_CFG');
	
    # Create an Instance Observer
    my $o_io = New InstanceObserver();

	# Substitute some parameters - 
	# ne sert rien ici il me semble. %d aura une autre valeur de remplacement
	my $root_data = $o_io->GetCfg('ROOT_DATA');
	$o_param->SetSubstitution('%d',\$root_data);
    my $genome_repository;
    if (exists $ENV{NARCISSE_GENOME_REPOSITORY})
    {
        $genome_repository = $ENV{NARCISSE_GENOME_REPOSITORY};
    }
    else
    {
        $genome_repository = $o_io->GetCfg('NARCISSE_GENOME_REPOSITORY');
    }
    if ($genome_repository ne '')
    {
		$o_param->SetSubstitution('%g',\$genome_repository);
        $o_param->Set('NARCISSE_GENOME_REPOSITORY',$genome_repository);
	};
	return ($o_io, $o_param);
}

=head2 Function GetParametersFromCommandLine

 Title        : GetParametersFromCommandLine
 Usage        : $o_both = &GetParametersFromCommandLine();
 Prerequisite : none
 Function     : Read parameters from the command line
                	it also check mandatories params
                	and set default values
 Returns      : Two SpeciesObserver objects
 Args         : $o_io,$o_param (modified)

=cut
sub GetParametersFromCommandLine
{
	my ($o_io,$o_param) = @_;

	# need more comment or an help function ?
	my @a_args = ( 'cfgQ=s', 'cfgS=s', # the org cfg (query and subject), either path or simply org short names
					'working_dir=s',
					'min_span=i', 'a=i', 'e=s', 'G=s', 'E=s', 'z=s',
					'path_blastall=s',
					'blastresults=s',
					'glintresults=s',
					'mappingresults=s',
					'method=s',
					'min_gscore=s',
					'chaining_param=s',
					&NO_RBH . '=s', #if set no reciprocal best hit will be computed
					'chains_dir=s',
					'only_comparison','no_comparison',
					'only_proteic','no_proteic',
					'only_nucleic','no_nucleic',
					'only_anchor','no_anchor',
					'only_build_chain','no_build_chain',	# do not launch build_chain ?
					'only_ortho','no_ortho',
					'only_clean', 'no_clean',
					'sessdir=s',
					&RAW_BLAST . '=s',	#give a path to a brut blast output
					&SELF,	# if set will compute cfgQ to itself
					&FORCE,	# if set will compute the nucleic comparison and protein comparison even if the corresponding gzip files exists
					'help',
					#&ROOT_LOG . '=s',
					&VERBOSE # to have exta information _debugging purpose
				);

	# Read from command line
	$o_param->Update( 'GETOPTLONG','O', @a_args );
	$o_param->Update('ENV',"O");   # May be overloading some variables

	my $where;
    if ( $o_param->IsDefined( 'sessdir' ))
    {
        my $sessdir = $o_param->Get( 'sessdir' );
		$o_io->SetSessionDirectory( $sessdir );

		# maintenant on peut ecrire dans un repertoire de session
		$where = 'SESSION';
		# MaJ de la liste des organismes
		#$o_io->UpdateGenomesListWithSessionData();
	}
	else
	{	$where = 'INSTANCE';	}
	
	my( $root_data ) = $o_io->GetDataPaths( $where );
	$o_param->SetSubstitution( '%d', \$root_data );
	$o_param->AssertDefined( 'cfgQ' );
	
	my $org_q = &_CheckCfgParams($o_io,$o_param,'cfgQ' );

	my $org_s;
	if( !$o_param->IsDefined( &SELF ) )
	{
		$o_param->AssertDefined( 'cfgS' );
		$org_s = &_CheckCfgParams($o_io,$o_param,'cfgS' );
		if ($org_q eq $org_s)
		{
			my $msg = "You cannot use the same org as cfgQ and cfgS.\n";
			$msg .= " Maybe you should consider using --self option\n";
			throw LipmError::ParameterException( 'cfgS', 'value', -text=>$msg );
		}
	}
	
	$o_param->SetUnlessDefined( 'method','genomic');
	$o_param->AssertAllowedValue( 'method', 'genomic', 'genetic');

	&_SetDefaultValues($o_io,$o_param);
	$o_param->AssertInteger( 'a', 'min_span', 'G', 'E', 'z' );

	my $o_org_q = $o_io->GetSpeciesObserver( $org_q );

	my $o_org_s;

	if( $o_param->IsDefined( 'self' ) )
	{	
        $o_org_s = $o_org_q;
    }
	else
	{
		$o_org_s = $o_io->GetSpeciesObserver( $org_s );
	}
	
	return ( $o_org_q, $o_org_s ) ;


=head3 Procedure _SetDefaultValues

 Title        : _SetDefaultValues
 Usage        : &_SetDefaultValues();
 Procedure    : Give a default value for some optionals arguments
 Args         : $o_io, $o_param

=cut

	sub _SetDefaultValues
	{
        my ($o_io,$o_param) = @_;
        
		$o_param->SetUnlessDefined('a', &GetDefault( 'CPU' ) );
	# je fais comme cela pour compatibilite <-- compatibilite avec quoi ? Seb
		$o_param->SetUnlessDefined('cut_off', &GetDefault( 'MIN_SPAN' ) );
		$o_param->SetUnlessDefined('min_span', $o_param->Get('cut_off'));
		$o_param->SetUnlessDefined('e', &GetDefault( 'EVALUE' ) );
		$o_param->SetUnlessDefined('G', &GetDefault( 'O_GAP' ) );
		$o_param->SetUnlessDefined('E', &GetDefault( 'E_GAP' ) );
		$o_param->SetUnlessDefined('z', &GetDefault( 'L_DB' ) );
		
	# parametres spécifiques à glint
	    $o_param->SetUnlessDefined('min_gscore', &GetDefault( 'MIN_GSCORE' ) );		
		$o_param->SetUnlessDefined('path_blastall', $o_io->GetCfg('ROOT_BLASTALL'));
		$o_param->SetUnlessDefined('build_chain', 0);
		$o_param->SetUnlessDefined('build_link', 0);
		$o_param->SetUnlessDefined( &NO_RBH, 0);
		
		
    # parametres spécifiques à genetic
		$o_param->SetUnlessDefined('mismatchpercentintra', &GetDefault('MISMATCH_PERCENT_INTRA'));	# same TaxID
		$o_param->SetUnlessDefined('mismatchpercentinter', &GetDefault('MISMATCH_PERCENT_INTER'));	# different TaxID
		$o_param->SetUnlessDefined('mismatchnbintra', &GetDefault('MISMATCH_NB_INTRA'));	# same TaxID
		$o_param->SetUnlessDefined('mismatchnbinter', &GetDefault('MISMATCH_NB_NTER'));	# different TaxID
		$o_param->SetUnlessDefined('path_primersearch',  $o_io->GetCfg('ROOT_PRIMERSEARCH'));		
		
	# parametres d'installation
		$o_param->SetUnlessDefined( &VERBOSE, 0);

		return;
	}


=head3 Function _CheckCfgParams

 Title        : _CheckCfgParams
 Usage        : $short = &_CheckCfgParams( $cfg );
 Prerequisite : none
 Function     : Assert that the file $cfg (or the cfg file for the $cfg org) exists.
                Upload the content of this file into $o_param
 Returns      : $short, the short name of the org that have this cfg file
 Args         : $o_io,$o_param,$cfg

=cut

	sub _CheckCfgParams
	{
		my ( $o_io,$o_param,$cfg ) = @_;

        my $where = $o_param->IsDefined('sessdir') ? 'SESSION' : 'INSTANCE';
        &NARCISSE_utils::AssertCfgParam( $o_io,$o_param,$cfg,$where );

		# read the content of the file, updating $o_param with $cfg as namespace
		my $file = $o_param->Get( $cfg );
		$o_param->SelectNameSpace( $cfg );
		$o_param->Update($file,'A');
		$o_param->AssertDefined('DST','ORG_SHORT');
		my $short_name = $o_param->Get( 'ORG_SHORT' );
		$o_param->SelectNameSpace('');

		return $short_name;
	}

} # GetParametersFromCommandLine


#----------------------------------------------------------------------------

=head2 procedure Usage

 Title        : Usage
 Usage        : Usage( $o_param )
 Prerequisite : none
 Fonction     : Usage
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Usage
{
    my ($o_param) = @_;
    
	#$o_param->SelectNameSpace();
    # print STDERR Dumper $o_param;
    my $narcisse_cfg = $ENV{NARCISSE_CFG};
	my @a_wanted = qw( EVALUE O_GAP E_GAP L_DB CPU MIN_SPAN MIN_GSCORE );
	my ( $evalue, $o_gap, $e_gap, $l_db, $cpu, $min_span, $min_gscore )
												= GetDefaults( @a_wanted );

print <<END

NARCISSE_CFG = $narcisse_cfg

Usage : $0

mandatory:
    --cfgQ        configuration file of the query genome, or genome short name
    --cfgS        configuration file of the subject/target genome, or genome short name
    --method	  either genomic or genetic (when genetic is invoked the query organism must be the genetic one)
                  nucleic (resp proteic) can also be specified for the method if only nucleic (resp proteic) comparison are wanted 
optional:
    --working_dir               [.../tmp/blast]
    --chains_dir	<dir> read precomputed chin files from dir The files must be in glintchain format and with the following names : chain0, chain1, chain2
    --sessdir     session directory, where to look for another data directory.

blast specific parameters
    --e             [$evalue]      Expectation value
    --G             [$o_gap]       Cost to open a gap
    --E             [$e_gap]       Cost to extend a gap
    --z             [$l_db]        Effective length of the database
    --a             [$cpu]         Number of processors to use
    --min_span      [$min_span]    Spanning threshold

glint specific parameters
    --min_gscore    [$min_gscore]  minimum score for the glint output
    --blastresults  <file> read precomputed blast results
    --glintresults  <file> read precomputed glint results
    --path_blastall [parameter BLASTALL]  <string>
    --chaining_param providing user-defined chaining parameters (see defaults.cfg in the section "Param for building chains" for examples)
    --no_rbh         no reciprocal best hit when building chains    
	
	
genetic specific parameters    
    --mappingresults <file> read precomputed primersearch results

SELECTING SOME ACTIONS:
    --only_xxx         If some only_xxx switches are specified, ONLY those specific actions will be made
    --no_xxx           If some no_xxx actions are specified, ONLY those specific actions will NOT be made

    --only_comparison  Only the comparison will be performed
    --no_comparison	   no comparison will be performed 
    --only_proteic     Run proteic comparison (blastp on proteins from annoted genes or read precomputed blast results)
    --no_proteic       Skip the proteic comparison
    
    --only_nucleic     Run nucleic comparison (run glint or read precomputed glint results)
    --no_nucleic       Skip the nucleic comparison

    --only_build_chain Run/don't run nc_build_chain.pl
    --no_build_chain

    --self             flag, if set, will compute cfgQ against itself, no reciprocal best hit will be computed
    --force            All comparisons will be made even if the final gzipped files (glint and blastp) are already in place

END
;

}


=head2 Procedure Verbose

 Title        : Verbose
 Usage        : &Verbose( $msg );
 Prerequisite : none
 Procedure    : will print $msg onto STDERR if VERBOSE is set
 Args         : $msg, a string

=cut

#sub Verbose
#{
#	my ( $msg ) = @_;
#
#	print STDERR $msg."\n"	if( $O_PARAM->IsDefined( &VERBOSE ) && ( 1 == $O_PARAM->Get( &VERBOSE ) ) ) ;
#
#	return;
#}


__END__
