#!/usr/bin/perl -w
#
# $Id: nc_get_data.pl 3767 2012-03-18 15:57:45Z sletort $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

nc_get_data.pl

=head1 SYNOPSIS

Core/perl/adm/nc_get_data.pl --cfg Site/species/cele.cfg

=head1 DESCRIPTION

This program manages the integration of one species into narcisse.
It download the data, if needed assemble the pseudo molecule, creates all files required
for narcisse, index these files and extract corresponding proteins, cds, transcripts and
gene sequences.
For annotation, input formats are Genbank and GFF3, for sequence data the input formats
can be either Genbank or Fasta.

The species data are described in a configuration file

	DST=%d
	ORG_SHORT=human
	ORG_LONG=Homo_Sapiens
	FILE_ASSEMBLY=file:///www/narcisse/Site/data/human_data/mapview/seq_contig.md.gz
	ASSEMBLY=reference
	FILE_IDEOGRAM=file:///www/narcisse/Site/data/human_data/mapview/ideogram.gz
	TIMESTAMP_FILE=file:///www/narcisse/Site/data/human_data/README
	FILE_GBK_1=%g/H_sapiens/CHR_01/hs_ref_chr1.gbk.gz
	FILE_GBK_2=%g/H_sapiens/CHR_02/hs_ref_chr2.gbk.gz
	FILE_GBK_3=%g/H_sapiens/CHR_03/hs_ref_chr3.gbk.gz
	FILE_GBK_4=%g/H_sapiens/CHR_04/hs_ref_chr4.gbk.gz
	FILE_GBK_9=file:///www/narcisse/JG4/Site/data/human_data/CHR_09/hs_ref_chr9.gbk.gz
	FILE_SEQ_9=file:///www/narcisse/JG4/Site/data/human_data/CHR_09/hs_ref_chr9.fa.gz


The files must be accessible via an URI which can be: ftp://, http://, file://. The string %g represents
the root of the URI which can be defined using an environment variable NARCISSE_GENOME_REPOSITORY
(e.g: setenv NARCISSE_GENOME_REPOSITORY ftp://ftp.ncbi.nlm.nih.gov/genomes)

For ANNOTATION files:
- the key must match the pattern FILE_(GBK|ANN)_
- the expected value is a Genbank or GFF3 file
- for genbank and ensembl data sources, random chromosomes can be added with the parameter INCL_RAMDOM=1

The format is specified using the key FORMAT. The default format is GENBANK, other values are GFF3, GENETIC, TIGRXML, ENSEMBL

For SEQUENCE files:
- the key must match the pattern FILE_(MFA|FAS|SEQ)_
- the expected value is a fasta or multifasta file
- supersede the sequences embedded in Genbank files.

An alternate translation table can be specified for any sequence using the key prefix (an integer value is expected)
TABLE_
An alternative consistency limit can be specified for any sequence using the key prefix (0 < integer <= 100)
CONSISTENCY_LIMIT_
for both keys, the chromosome id must match an existing chromosome file id


- The sequence is assembled using the file pointed out by FILE_ASSEMBLY when more than one sequence are extracted from
the sequence source (either ANN=Genbank or ANN=Genbank+SEQ=Fasta)
- The program does not manage incremental updating process.
- compressed files are not allowed for TIGRXML format

=cut

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";
use lib "$FindBin::Bin/../API/Bio/Narcisse/Utils";

use File::Path qw( mkpath rmtree ); # mkdir -p et rm -rf
use Bio::SeqIO;
use Bio::Seq;
use Getopt::Long;
use ParamParser;
use Logger;
use Runner;
use Sufine;
use Carp;
use OnlyOrNo;
use Cwd;
use Data::Dumper;
use File::Basename;
use File::Spec;
use IO::File;
use NARCISSE_utils;
use NarcisseDataWriter;
use NarcisseDataWriter::OnMemory;
use Error qw( :try );
use LipmError::ParameterException;
use Configuration::InstanceObserver;
use ConfigurationFactory;
use AnnotatedSeqFile::Genbank;
use AnnotatedSeqFile::Embl;
use AnnotatedSeqFile::Fasta;
use AnnotatedSeqFile::GFF3;
use AnnotatedSeqFile::Ensembl;
use AnnotatedSeqFile::TIGRXML;
use AnnotatedSeqFile::Genetic;
use SlidingWindowCalculator::GCPercentCalculator;
use SlidingWindowCalculator::GeneDensityCalculator;
use SlidingWindowCalculator::GeneSurfaceCalculator;
use SlidingWindowCalculator::LowComplexityDensityCalculator;
use SlidingWindowCalculator::GCSkewCalculator;
use NDetector;
use General;

use Bio::EnsEMBL::Registry;
use LipmError qw(:try);

use constant DEFAULT_CONSISTENCY_LIMIT => 75;

my $VERSION = do {my @r = (q$Rev: 3767 $ =~ /\d+/g); $r[0]};


MAIN:
{
	try
	{
		&Main();
	}
	catch Error with
	{
		my ( $o_err) = @_;
		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line() . "\n";
		$str   .= $o_err->stacktrace() . "\n";
		if ( $o_err -> isa ('LipmError') )
		{
			$str .= $o_err->Explain() . "\n";
		};
		print STDERR $str;
		exit 1;
	};

	exit( 0 );
}


sub Main
{
	&ParamParser::SetDefaultBehaviour('use_substitution_table');

	my $o_param =New ParamParser;
	$o_param->SetUsage( \&Usage );

	# Update from ENV, for $NARCISSE_CFG
	$o_param->Update('ENV','O');   # for $NARCISSE_CFG
	$o_param->AssertDefined('NARCISSE_CFG');
	$o_param->AssertFileExists('NARCISSE_CFG');

	# Init Instance Observer
	my $o_io   = &ConfigurationFactory::GetConfigurationNoSingleton();

	# Substitute some parameters
	my $root_data = $o_io->GetCfg('ROOT_DATA');
	$o_param->SetSubstitution('%d',\$root_data);
    my $genome_repository;
    if (exists $ENV{NARCISSE_GENOME_REPOSITORY})
    {
       $genome_repository = $ENV{NARCISSE_GENOME_REPOSITORY};
    }
    else
    {
       $genome_repository = $o_io->GetCfg('NARCISSE_GENOME_REPOSITORY');
    }
    if ($genome_repository ne '')
    {
		$o_param->SetSubstitution('%g',\$genome_repository);
	};

	# Read and validate parameters from command line
	$o_param->Update('GETOPTLONG',"O","debug=i",
	("cfg=s","restart=s","assembly=s","help",
	"glintindex_mask=s","glintindex_pass=s","virtual_genome_size=i",
	"only_check","no_check","only_download","no_download","only_suppdata","no_suppdata","only_formatdb","no_formatdb","only_index","no_index", "only_clean","no_clean","with_random","check_files", "no_mail", 'sessdir=s' ));

	$o_param->Update('ENV',"O");   # May be overloading some variables

    # If sessdir defined, we are working in a session context
    if ( $o_param->IsDefined( 'sessdir' ))
    {
        my $sessdir = $o_param->Get( 'sessdir' );
		$o_io->SetSessionDirectory( $sessdir );
	}

	$o_param->SetUnlessDefined('assembly',0);

    # mandatory parameters must be org filepath to org.cfg, or just organism short name
    my $where = $o_param->IsDefined('sessdir') ? 'SESSION' : 'INSTANCE';
    $o_param->AssertDefined('cfg');
    &NARCISSE_utils::AssertCfgParam( $o_io,$o_param,'cfg',$where );

	my $o_only = New OnlyOrNo($o_param);

	my $file_arg = $o_param->Get('cfg');

	# Complete $o_param with data from the species' file
	# TODO - Some other ParamParser for these parameters ? - manu
	$o_param->Update($file_arg,'A');
	$o_param->AssertDefined('DST','ORG_SHORT');
	
	# The species observer corresponding to the cfg parameter
	my $o_sp = $o_io -> GetSpeciesObserver( $o_param->Get('ORG_SHORT') );

	# Create the directory for this species, if necessary
	$o_sp -> GetDataPath( undef, &TRUE );
	
	$o_param->SetUnlessDefined('FORMAT','GENBANK');
	$o_param->SetUnlessDefined('ASSEMBLY_PHASE','PSEUDOMOLECULES');

	$o_param->SetUnlessDefined('INCL_RANDOM',0);

	$o_param->Set('INCL_RANDOM',1) if ($o_param->Get('with_random'));

	#
	# if only_check is the only switch specified, just check, print a message and leave
	#
	# WARNING The pattern -1- and -0- may be checked by another script, do not modify it
	#

	if ($o_only->Get('check') and $o_only->HowManyOnly()==1)
	{
		my $org = $o_sp -> GetShortName();
		print &TimeStampMirror($o_sp, $o_param)==1 ? "$org -1- needs to be updated\n" : "$org -0- does not need to be updated\n";
        if ($o_param->IsDefined('check_files'))
        {
            &CheckFiles($o_sp, $o_param);
        }
		exit 0;
	}
    else
    {
        if ($o_param->IsDefined('check_files'))
        {
            die ("ERROR - check_files must be specified ONLY with --only_check AS ONLY SWITCH");
        }
    }

	&NARCISSE_utils::Welcome(basename($0),$VERSION,$o_param->Get('NARCISSE_CFG'),$genome_repository);

	# We check to know if we need an update - If no update is needed, we exit without logging nor creating target directories
	if ($o_only->Get('check'))
	{
		&ExitIfUpToDateData( $o_sp, &TimeStampMirror($o_sp, $o_param), $o_param );
	}

	# Compute and print $target
	my $sav_dir  = getcwd();
    my $target_path = &SetLogPath( $o_sp, $o_param );

	$o_param->Dump("narcisse.param");

	my $logger_file = "narcisse.log";
	my $o_logger = New Logger ($logger_file,$o_io->GetCfg('LOG_LEVEL'));
	$o_logger->StartTimer('Timer');

	$o_logger->Trace("START: $file_arg\n",'Timer','A');

	my $unix_cat= $o_io->GetUnixPath('UNIX_CAT');

	# Telechargement des fichiers ou verif de leur existence
	if ($o_only->Get('download'))
	{
		my $unix_cp = $o_io->GetUnixPath('UNIX_CP');
		my $unix_rm = $o_io->GetUnixPath('UNIX_RM');

		&DownloadFiles($o_sp,$o_param,$o_logger);

		$o_logger->Trace("STEP: Extracting and reformating data\n",'Timer','A');
		my $ra_chromosome_files = &ExtractAndReformatData($o_sp,$o_logger,$o_param);

		$o_logger->Trace("Data extracted\n",'Timer','A');

		$o_logger->Trace("STEP: Moving files to their final destination\n",'Timer','A');
		my $src = join(' ',glob($o_sp -> GetTmpDataPath() . '/*'));
		my $dst = $o_sp -> GetDataPath();
		my $cmd = "$unix_cp $src $dst";
		$o_logger->Trace("$cmd\n",'','A');
		&NARCISSE_utils::RunCmd($cmd);

		if ($o_param->Get('FORMAT') ne 'GENETIC' )
		{
			$o_logger->Trace("STEP: Creating virtual genomes\n",'Timer','A');
			&CreateVirtualGenomes( $o_sp, $o_param, $ra_chromosome_files);
			$o_logger->Trace("Virtual genomes created\n",'Timer','A');
	
			$cmd = "$unix_rm -r " . $o_sp -> GetTmpDataPath();
			$o_logger->Trace("$cmd\n",'','A');
			&NARCISSE_utils::RunCmd($cmd);
			$o_logger->Trace("\nGenome downloaded\n",'Timer','A');
		}
	};

    # Supplementary data
    if ( $o_only->Get('suppdata') && $o_param->Get('FORMAT') ne 'GENETIC' )
    {
		warn "Computing supp data \n";
        &ComputeSuppdata($o_sp,$where,$o_logger);
    };

	# Indexing for blast
	if ($o_only->Get('formatdb') && $o_param -> Get('FORMAT') ne 'GENETIC')
	{
		my (@a_fasta_files,@a_desc,@a_prot);
		push @a_fasta_files, $o_sp->GetFastaFile( feature => 'PROT');
		push @a_desc, 'proteins';
		push @a_prot, &TRUE;
		push @a_fasta_files, $o_sp->GetFastaFile( feature => 'PROT', -rep => &TRUE);
		push @a_desc, 'reference proteins';
		push @a_prot, &TRUE;
		push @a_fasta_files, $o_sp->GetFastaFile( feature => 'CDS');
		push @a_desc, 'cds';
		push @a_prot, &FALSE;

		my $formatdb = $o_io -> GetCfg('ROOT_FORMATDB');
		my $short = $o_sp -> GetShortName();
		for (my $i=0; $i <= $#a_fasta_files; $i++)
		{
			my $fasta_file = $a_fasta_files[$i];
			my $description= $a_desc[$i];
			my $is_prot    = $a_prot[$i];
			my $cmd = "$formatdb -i $fasta_file -t $short";
			$cmd .= ' -p F' if ($is_prot == &FALSE);
			$o_logger->Trace("Indexing $description for blast ... $cmd\n",'Timer','A');
			print "Indexing $description for blast\n";
			&NARCISSE_utils::RunCmd($cmd);
		}
	}
	
	# Indexing for glint
	if ($o_only->Get('index') && $o_param->Get('FORMAT') ne 'GENETIC' )
	{
		warn "Indexing virtual chromosomes \n";
		$o_logger->Trace("STEP: Indexing data for the glint program\n",'Timer','A');
		&IndexSequenceFile( $o_sp, $o_param );
		$o_logger->Trace("Indexing (glintindex) completed\n",'Timer','A');
	};

	# Cleaning
	if ($o_only->Get('clean'))
	{
		my $unix_rm = $o_io->GetUnixPath('UNIX_RM');
		$o_logger->Trace("STEP: Cleaning  temporary download directories\n",'Timer','A');
		my $cmd = "$unix_rm -r " . $o_sp -> GetDownloadPath() . ' ' . $o_sp -> GetTmpPath();
		&NARCISSE_utils::RunCmd($cmd);
	};
	$o_logger->Trace("END: $file_arg\n",'Timer','A');
	
	my @a_log = `$unix_cat $logger_file`;
	my $join=join('',@a_log);
	&SendEmail($o_sp, "Update completed\n$join\n",$o_io->GetCfg('NARCISSE_ADM_EMAIL'),"[NARCISSE]_Update_" . $o_sp->GetShortName(), $o_param);

	# Back to the original directory
	chdir($sav_dir);

	print "Log files created in directory: $target_path\n";

	# WARNING - If the switch --only_check was the ONLY --only_XXX switch specified, we do not go there, because we have already exited
	# We go there if:
	#    --no_check was NOT specified
	#    or if --only_check was specified TOGETHER WITH  some other --only_XXX switch
	#    or if NO --only_xxx/--no_xxx switches were specified
    # TODO --> pas terrible, si je fais --only_download le timestamp est correct alors que les donnees ne sont pas indexees....

	if ($o_only->Get('check') and $o_only->Get('download'))
	{
		
		if ($o_param -> IsDefined ('TIMESTAMP_FILE') and $o_param -> Get('TIMESTAMP_FILE') ne '')
		{
			my $unix_mv = $o_io->GetUnixPath('UNIX_MV');
			my ($timestamp_path,$local_file,$local_file_new) = &GetTimestampPath($o_param);
			my $dir_timestamp = $o_sp -> GetTimestampMirror( -check => &FALSE );
			$local_file_new = File::Spec->catfile ($dir_timestamp, $local_file_new);
			$local_file     = File::Spec->catfile ($dir_timestamp, $local_file);
			if (-e $local_file_new)
			{
				`$unix_mv $local_file_new $local_file`;
				print "now recording the new timestamp file\n";
			}
		}
	}
}

=head2 function SetTargetLogPath

 Title        : SetTargetLogPath
 Usage        : $log_path = &SetTargetLogPath( $o_io, $o_param, $org );
 Function     : computes the target log path and creates the directory
                It also redirect STDERR and SDTOUT.
 Returns      : $log_path, a string
 Args         : $o_sp, $o_param

=cut

sub SetLogPath
{
	my ( $o_sp, $o_param ) = @_;
	my $o_io = $o_sp -> GetInstanceObserver();
	my $org  = $o_sp -> GetShortName();

	# Compute and print $target, and lock the target
	my $target = "get_$org";
	print "target $target\n" . '-' x 30 . "\n";

    my $logs_dir = $o_param->IsDefined( 'sessdir') ? $o_io->GetLogsDirpath( 'SESSION') : $o_io->GetLogsDirpath( 'INSTANCE' );
	# change current dir to target dir
	my @a_dirs = ( $logs_dir,'get',$org );
	my $target_log_root = File::Spec->catfile( @a_dirs );

	# @a_created_dirs pour loguer plus tard.
	my @a_created_dirs = mkpath( $target_log_root );

	my $target_log_path = Sufine::MkDir(-dir => $target_log_root,-name => $target);
	die unless(defined $target_log_path);
	die unless(defined Sufine::ChDir(-dir => $target_log_root,-name => $target));

	&NARCISSE_utils::RedirectOE($o_io,"narcisse.out","narcisse.err");

	$o_param->Dump("narcisse.param");

	return $target_log_path;
}

=head2 procedure ExtractLongestPeptide

 Title        : ExtractLongestPeptide
 Usage        : &ExtractLongestPeptide( $o_sp, $o_param, @a_chromosome_files )
 Procedure    : read existing protein fasta file and extract the longest peptide for each gene : the representative protein of the gene
 				All the representative proteins are stored in the rep_prot file of the db/prot dir (used for orthologue detection)
 Args         : $o_sp, $o_param, \@a_chromosome_files (the list of chrom files)
 Globals      : none

=cut
sub ExtractLongestPeptide
{
	my ($o_sp, $o_param, $ra_chrom_files) = @_;

	my $o_io = $o_sp -> GetInstanceObserver();
	
	# Reading transcript files to identify correspondance between protein_id and gene names (absent from FILE_DB_PROT with Ensembl)
	# %h_transcript:
	#    key   = element
	#    value = The corresponding data row
	my %h_transcript = ();
	#my @a_chrom = split('#',$o_param->Get('AVAILABLE_CHR'));
	foreach my $chrom ( @$ra_chrom_files )
	{
		my $transcript_file = $o_sp->GetTranscriptFiles( -chrom  => $chrom, -tmp => &TRUE );
		my ( $rh_temp_transcript ) = &NARCISSE_utils::ReadNarcisseTabFile( file => $transcript_file );
		next unless defined $rh_temp_transcript; #Check if there is any transcript
		%h_transcript = ( %h_transcript, %$rh_temp_transcript);
	}

	# %h_gene:
	#    key   = protein_id
	#    value = The parent of the corresponding transcript (ie the gene this protein belongs to)
	my %h_gene=();
	foreach my $transcript (keys %h_transcript )
	{
		$h_gene{$h_transcript{$transcript}{protein_id}}=$h_transcript{$transcript}{parent};
	}

	# %h_protein:
	#    key   = The gene the protein found in the fasta file belongs to
	#    value = A ref to a hash with key = $protein_id, value a hash with len,header,seq
	#    ==> For each gene, we keep all the corresnpofing proteins
	my %h_protein=();
	my $prot_fasta = $o_sp -> GetFastaFile( feature => 'PROT', -tmp => &TRUE );

	# preferred to avoid grep dependency
	my $fh_prot_fasta = new IO::File( $prot_fasta ) or  throw LipmError::IOException( $prot_fasta, 'f' );
	my ($protein_id,$locus_tag,$gene_dbID,$length);
	while(my $protline=<$fh_prot_fasta>)
	{
		if ( $protline =~ /^>/ )
		{
			if ($protline =~ /^>prot:(\S+) .*locus_tag=(\S+) .*len=(\d+) /)
			{
				($protein_id,$locus_tag,$length)=($1,$2,$3);
			}
			else
			{
				Carp::croak("ERROR in header format of <$prot_fasta>: $protline\n");
			}
			if (defined $h_gene{$protein_id})
			{
			    $gene_dbID=$h_gene{$protein_id};
			    $h_protein{$gene_dbID}{$protein_id}{len}    = $length;
			    $h_protein{$gene_dbID}{$protein_id}{header} = $protline;
			    $h_protein{$gene_dbID}{$protein_id}{seq}    = "";
			}
			else
			{
			    warn("WARNING - Undefined gene name for $protein_id ");
			}
		}
		else
		{
			$h_protein{$gene_dbID}{$protein_id}{seq}.=$protline;
		}
	}
	$fh_prot_fasta->close;

	my $rep_prot_fasta    = $o_sp -> GetFastaFile( feature => 'PROT', -rep => &TRUE, -tmp => &TRUE, -check => &FALSE );
	my $fh_rep_prot_fasta = new IO::File($rep_prot_fasta,"w") or throw LipmError::IOException( $rep_prot_fasta, 'f' );
	foreach my $gene (keys %h_protein)
	{
		# Sort the proteins related to this gene, longest first
		my @sorted_protid=sort {$h_protein{$gene}{$b}{len} <=> $h_protein{$gene}{$a}{len}} keys %{$h_protein{$gene}};
		
		# Write only the longest protein to the fasta file
		my $protein_id=$sorted_protid[0];
		print $fh_rep_prot_fasta $h_protein{$gene}{$protein_id}{header};
		print $fh_rep_prot_fasta $h_protein{$gene}{$protein_id}{seq};
	}
}

=head2 Function ExtractCytogeneticData

 Title    : ExtractCytogeneticData
 Usage    : ExtractCytogeneticData ($file)
 Fonction : Parse and extract the ideogram data
 Returns  : a hash containing the data:
              Key   = chromosom
              Value = ref-to-hash containing data
 Args     : $o_sp
            $file = the ideogram file
 Globals  : 

=cut
sub ExtractCytogeneticData
{
	my ($o_sp, $file) = @_;
	my $o_io      = $o_sp -> GetInstanceObserver();

	my %h_cyto;
	my %h_list_chr;
	my %h_list_chr_no_num;

	$file = &NARCISSE_utils::GetStream($o_io,$file);
	my $fh_cyto = new IO::File("$file") or throw LipmError::IOException($file,'f');
	my ($chr,$arm,$band,$start,$end,$color) = ("","","","","","");
	my ($chr_mem,$arm_mem,$band_mem,$start_mem,$end_mem,$color_mem) = ("","","","","","");
	my $first_bool = 1;

	# Nouvelle version pour le nouvel assemblage de l'homme 17-03-2006
	
	# Key = the chromosomes, Value = All the lines corresponding to this chromosome
	my %h_cyto_chr;
	my ($chr_pos,$arm_pos,$band_pos,$start_pos,$end_pos,$color_pos) = (0,1,2,5,6,7);
	while ( my $line = <$fh_cyto> )
	{
		next if ( $line =~ /^\#/ );
		my @st = split (/\s+/, $line);
		($chr,$arm,$band,$start,$end,$color) = ($st[$chr_pos],$st[$arm_pos],$st[$band_pos],$st[$start_pos],$st[$end_pos],$st[$color_pos]);

		if ( !defined($h_cyto_chr{$chr}) )
		{
			my @a_tmp = ();
			$h_cyto_chr{$chr} = \@a_tmp;
		}
		push (@{$h_cyto_chr{$chr}},$line);
	}
	$fh_cyto->close();

	# key = $chr, value = $rh_
	#               key = start positions, value = the file line
	my %h_cyto_by_start;
	foreach my $key ( keys %h_cyto_chr )
	{
		$arm_mem = "p";
		foreach my $line ( @{$h_cyto_chr{$key}} )
		{
 			next if ( $line =~ /^\#/ );
			my @st = split (/\s+/, $line);
			($chr,$arm,$band,$start,$end,$color) = ($st[$chr_pos],$st[$arm_pos],$st[$band_pos],$st[$start_pos],$st[$end_pos],$st[$color_pos]);

			# $arm, $chr required fields
			next if ($arm !~ /p|q|cen/i);
			next if ( !$chr );

			# color = default value gneg
			$color = "gneg" if ( !defined($color) || $color !~ /[a-z]+/i);

			# Centro detection
			if ( $arm eq "cen" || ($arm_mem ne "q" && $arm eq "q" ) )
			{
				my $key_centro = "centro_$chr";
 				$h_cyto{$key_centro} = $start if ( !exists($h_cyto{$key_centro}) );
			}

			
			$arm_mem = $arm;
 			if ( $chr =~ /\d+/ )
 			{
				$h_list_chr{$chr} = 1;
 			}
 			else
 			{
 				$h_list_chr_no_num{$chr} = 1;
 			}

			my $key = $arm.$band;
			my %h_band;

			$h_band{chr}   = $chr;
			$h_band{elt}   = $key;
			$h_band{start} = $start;
			$h_band{end}   = $end;
			$h_band{color} = $color;

			if ( !exists($h_cyto_by_start{$chr}{$start}) )
			{
				my @a_tmp = ();
				$h_cyto_by_start{$chr}{$start} = \@a_tmp;
			}

			push( @{$h_cyto_by_start{$chr}{$start}}, \%h_band );
		}
	}

	foreach my $chr ( keys %h_cyto_by_start )
	{
		foreach my $start ( sort {$a<=>$b} keys %{$h_cyto_by_start{$chr}} )
		{
			my $ra_list_cyto = $h_cyto_by_start{$chr}{$start};

			foreach my $rh_band ( sort {$a->{end}<=>$b->{end}} @{$ra_list_cyto} )
			{
				next if ($rh_band->{start} == $rh_band->{end});
				$h_cyto{$chr}->{$rh_band->{elt}}->{start} = $rh_band->{start};
				$h_cyto{$chr}->{$rh_band->{elt}}->{end} = $rh_band->{end};
				$h_cyto{$chr}->{$rh_band->{elt}}->{color} = $rh_band->{color};
				last;
			}
		}
	}
	return \%h_cyto;
}

=head2 procedure DumpCytogeneticData

 Title        : DumpCytogeneticData
 Usage        : DumpCytogeneticData
 Procedure    : When the karyotype does not exists, it is created based on the chromosome length
                Create the cyto/*.cyto files from the cytogenetic data if possible, or from the chr length
 Args         : $rh_cyto      : data extracted by ExtractCytogeneticData, may be empty
                $rh_chr_length: Ref to hash - key = chr, val = chr length
                $o_sp         : species observer
                $o_param, $o_logger
 Globals      : None

=cut

sub DumpCytogeneticData
{
	my ($rh_cyto, $rh_chr_length, $o_sp, $o_param, $o_logger) = @_;

	$o_logger->Trace("Starting CreateKaryotype\n",'Timer','A');

	#my @a_sorted_chr = sort { $H_CHR_LENGTH{$a }<=> $H_CHR_LENGTH{$b} } ( keys ( %H_CHR_LENGTH ));
	my @a_sorted_chr = sort { $rh_chr_length->{$a }<=> $rh_chr_length->{$b} } ( keys ( %$rh_chr_length ) );

    # Create the karyotype file, if necessary
    # Le fichier peut avoir deja ete cree par  ExtractCytogeneticData
    # manu - JE NE CROIS VRAIMENT PAS !!!
    
    my @a_chr_data;
    my @a_chr_fields = qw( chr length centro codon_table );
    foreach my $chr ( @a_sorted_chr )
    {
		my $key_centro = "centro_$chr";
		my $centro = (defined($rh_cyto->{$key_centro})) ? $rh_cyto->{$key_centro} : 1;
		my $table_usage = $o_param->Get("TABLE_$chr");
		my $codon_usage = ( defined($table_usage) && $table_usage =~ /^\d+$/ ) ? $table_usage : "";
		my %h_row;
		@h_row{ @a_chr_fields } = ($chr, $rh_chr_length->{$chr},$centro,$codon_usage);
		push @a_chr_data, \%h_row;
	}
	
    # Create the karyotype file, if necessary
    # Le fichier peut avoir deja ete cree par  ExtractCytogeneticData
    # manu - JE NE CROIS VRAIMENT PAS !!!
	my $file_karyo = $o_sp -> GetKaryotypeFile ( -check => &FALSE, -tmp => &TRUE );
	if ( !-e $file_karyo )
    {
		my $o_ndw = New NarcisseDataWriter::OnMemory( $file_karyo, \@a_chr_fields, { sort => &FALSE } );
		# PAS DE COMMENTAIRE POUR L'INSTANT CA FAIT PLANTER NarcisseDataReader.pl
		#$o_ndw -> Update( \@a_chr_data, "Created by nc_get_data.pl - " . &NARCISSE_utils::GetDate() );
		$o_ndw -> Update( \@a_chr_data );
	}
    else
	{
		$o_logger->Trace("Warning/CreateKaryotype: $file_karyo already exists, skipped\n",'','A');
	}

    # Create the .cyto files
	my @a_cyto_fields = qw( chrom element start end color );
	foreach my $chr ( @a_sorted_chr )
	{
		my @a_cyto_data;

		#chr read from the ideogram file (see ExtractDataCyto)
		if ( defined $rh_cyto->{$chr})
		{
			foreach my $band_cyto ( keys %{$rh_cyto->{$chr}})
			{
				my $start   = $rh_cyto->{$chr}{$band_cyto}->{start};
				my $end     = $rh_cyto->{$chr}{$band_cyto}->{end};
				my $color   = $rh_cyto->{$chr}{$band_cyto}->{color};
				my %h_row;
				@h_row{ @a_cyto_fields } = ($chr, $band_cyto, $start, $end, $color);
				push @a_cyto_data, \%h_row;
			}
		}
		
		# just write the chromosome length
		else
		{
			my %h_row;
			@h_row{ @a_cyto_fields } = ($chr, 1, 1, $rh_chr_length->{$chr}, 'gneg' );
			push @a_cyto_data, \%h_row;
		}
	
		my $file_cyto = $o_sp -> GetCytoFiles( -chrom => $chr, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		my $o_ndw     = New NarcisseDataWriter::OnMemory( $file_cyto, \@a_cyto_fields );
		# PAS DE COMMENTAIRE POUR L'INSTANT CA FAIT PLANTER NarcisseDataReader.pl
		#$o_ndw -> Update( \@a_cyto_data, "Created by nc_get_data.pl - " . &NARCISSE_utils::GetDate() );
		$o_ndw -> Update( \@a_cyto_data );
	}
	
	return;
}
	
=head2 procedure ComputeSuppdata

 Title     : ComputeSuppdata
 Usage     : ComputeSuppdata($species,$where)
 Procedure : Compute genomic statistics:
             1/ gc content
             2/ gene density (nb of genes / genome length)
             3/ gene surface (% of bases belonging to a gene)
 Returns   : none
 Args      : $o_sp   : the species observer
             $where  : 'INSTANCE' or 'SESSION'
             $o_logger
 Globals   : 

=cut

sub ComputeSuppdata
{
	my ( $o_sp, $where, $o_logger ) = @_;
	my $species      = $o_sp -> GetShortName();
	my $o_io         = $o_sp -> GetInstanceObserver();

	my $window_size = $o_io -> GetCfg('SUPP_WINDOW_SIZE');
	my $step  = $window_size / 10;

	# les parametres pour lancer RollOverSpecies
	my %h_params = (-species   => [ $species ],
					-o_logger  => $o_logger,
					-file_location => $where,
					-window    => $window_size,
					-step      => $step,
					-o_io      => $o_io );

	my $o_sk = New GCSkewCalculator( %h_params,
									-graph     => 'line_plots',
									-feature   => 'gcsk',
									-extension => 'sk' );
	$o_sk->RollOverSpecies( -auto_clean => &TRUE );

	my $o_nd = New NDetector( %h_params );
	$o_nd->RollOverSpecies( -auto_clean => &TRUE );

	my $o_gc = New GCPercentCalculator( %h_params,
										-graph     => 'line_plots',
										-feature   => 'gcsk',
										-extension => 'gc' );
	$o_gc->RollOverSpecies( -auto_clean => &TRUE );

	my $o_gs = New GeneSurfaceCalculator( %h_params,
										-graph     => 'line_plots',
										-feature   => 'gs',
										-extension => 'gs' );
	$o_gs->RollOverSpecies( -auto_clean => &TRUE );

	my $o_gd = New GeneDensityCalculator( %h_params,
										-graph     => 'histograms',
										-feature   => 'gd',
										-extension => 'gd' );
	$o_gd->RollOverSpecies( -auto_clean => &TRUE );

	my $o_lc = New LowComplexityDensityCalculator( %h_params,
													-graph    => 'line_plots',
													-feature  => 'lc',
													-extension=> 'lc' );
	$o_lc->RollOverSpecies( -auto_clean => &TRUE );

	return;
}

=head2 Function GetTimestampPath

 Title        : GetTimestampPath
 Usage        : my ($timestamp_path,$local_file,$local_file_new) = &GetTimestampPath($o_param);
 Fonction     : Retrieve the timestamp mirror file path name, file name, and a 'new' file name
 Returns      : 3 names
 Args         : $o_param

=cut
sub GetTimestampPath
{
	my ($o_param) = @_;
	if ( ! $o_param->IsDefined( 'TIMESTAMP_FILE' ) )
	{
		return ('','','');
	}
    my $timestamp_path = $o_param->Get('TIMESTAMP_FILE');
	if ($timestamp_path !~ /^ftp|^http/ and $timestamp_path !~ /^file/ and $timestamp_path !~ /^\//)
	{
		die "ERROR - bad timestamp file - Did not you forget NARCISSE_GENOME_REPOSITORY ?\n";
	};
	    
	# $local_file: you can specify a directory only if using ftp or http
	# In this case, the file retrieved is 'index.html'
	# TODO - Permettre de parametrer cela, pas sur que ce soit TOUJOURS index.html !
	my $local_file;
	if ($timestamp_path !~ /^ftp|^http/ and $timestamp_path =~ /\/$/)
	{
		die "ERROR - bad timestamp file - should NOT be a directory except through ftp or http\n";
	}
	if ($timestamp_path =~ /\/$/)
	{
		$local_file = 'index.html';
	}
	else
	{
		$local_file = basename($timestamp_path);
	}
	my $local_file_new = $local_file.'.new';
	
	return ($timestamp_path,$local_file,$local_file_new);
}

=head2 Function TimeStampMirror

 Title        : TimeStampMirror
 Usage        : TimeStampMirror()
 Prerequisite : none
 Fonction     : check the timestamp of the update
 Returns      : 1 when the update is required, 0 elsewhere
 Args         : $o_sp, $o_param
 Globals      : $PWD, $UNIXCP, $WGET_TIMESTAMP, $UNIXMV, $UNIXDIFF

=cut

sub TimeStampMirror()
{
	my ($o_sp, $o_param) = @_;
	
	my $tobeupdated = 1;
	
	# If the timestamp file is not defined, we do not check and we consider it must be updated
	if ($o_param -> IsDefined( 'TIMESTAMP_FILE' ) == &FALSE)
	{
		$tobeupdated = 1;
	}
	else
	{
		my $o_io      = $o_sp -> GetInstanceObserver();
		
		my $unix_cp   = $o_io->GetUnixPath('UNIX_CP');
		my $unix_wget = $o_io->GetUnixPath('UNIX_WGET').' -K -nd';
		my $unix_mv   = $o_io->GetUnixPath('UNIX_MV');
		my $unix_diff = $o_io->GetUnixPath('UNIX_DIFF');
		
		my $pwd = cwd;
		my $rvl = 1;
	
		my $dir_timestamp = $o_sp -> GetTimestampMirror( -check => &FALSE );
		#$o_logger->Trace("$dir_timestamp\n",".","D");
		if ( ! -d $dir_timestamp)
		{
			mkdir ($dir_timestamp);
		}
		chdir ($dir_timestamp);
	
		my $log_file = "log_wget.$$";
	
	    #my $timestamp_file = $o_param->Get('TIMESTAMP_FILE');
	    #if ($timestamp_file !~ /^ftp|^http/ and $timestamp_file !~ /^file/ and $timestamp_file !~ /^\//)
	    #{
	        #die "ERROR - bad timestamp file - Did not you forget NARCISSE_GENOME_REPOSITORY ?\n";
	    #};
	    
	    ## $local_file: you can specify a directory only if using ftp or http
	    ## In this case, the file retrieved is 'index.html'
	    ## TODO - Permettre de parametrer cela, pas sur que ce soit TOUJOURS index.html !
	    #my $local_file;
	    #if ($timestamp_file !~ /^ftp|^http/ and $timestamp_file =~ /\/$/)
	    #{
			#die "ERROR - bad timestamp file - should NOT be a directory except through ftp or http\n";
		#}
		#if ($timestamp_file =~ /\/$/)
		#{
			#$local_file = 'index.html';
		#}
		#else
		#{
			#$local_file = basename($timestamp_file);
		#}
	    #my $local_file_new = $local_file.'.new';
	    my ($timestamp_path,$local_file,$local_file_new) = &GetTimestampPath($o_param);
	    unlink $local_file_new;
		if ( $timestamp_path =~ /file:\/\// or $timestamp_path =~ /^\//) # We just use a local copy
		{
	        $timestamp_path =~ s/file:\/\///;
	        `$unix_cp $timestamp_path $local_file_new`;
	    }
	    else
	    {
	        my $cde = "$unix_wget -K -nd -o $log_file $timestamp_path";
	        if (-e $local_file)
	        {
	            `$cde`;
	            if (-e "${local_file}.1")
	            {
	                `$unix_mv ${local_file}.1 $local_file_new`;
	            }
	        }
	        else
	        {
	            `$cde`;
	            my $downloaded = $local_file;
	            if (-e $downloaded)
	            {
	                `$unix_mv $downloaded $local_file_new`;
	            }
	        };
	    };
	    if (-e $local_file and -e $local_file_new)
	    {
	        my @a_diff = `$unix_diff $local_file $local_file_new`;
	        $tobeupdated  = ( $#a_diff == -1 ) ? 0 : 1;
	    };
	
		chdir($pwd);
	}
	
	return  $tobeupdated ;
}

=head2 procedure UpdateTimestamp

 Title        : UpdateTimestamp
 Usage        : UpdateTimestamp()
 Procedure    : Update the timestamp
 Returns      : none
 Args         : $o_sp
 Globals      : none

=cut

sub UpdateTimestamp()
{
	my ($o_sp) = @_;
	my $file = $o_sp -> GetTimestampFile( -check => &FALSE );
	
    my $f_time = new IO::File($file,"w") or Carp::croak("Cannot open $file");
	print $f_time time;
	undef $f_time;
	return;
}

=head2 procedure ExitIfUpToDateData

 Title        : ExitIfUpToDateData
 Usage        : ExitIfUpToDateData($bool)
 Prerequisite : none
 Fonction     : Hub that either activates the update or quit when the update is not required
 Returns      : none
 Args         : $o_sp
                $bool : 1 when update is required
                $o_param

=cut

sub ExitIfUpToDateData
{
	my ($o_sp,$tobeupdated,$o_param) = @_;

	my $arg = join(' ',@ARGV);
	my $dir = $o_sp -> GetDataPath();
	my $msg = "The command below was executed to update the directory $dir\n$0 $arg\n";

	if ( $tobeupdated == 0 )
	{
		my $tsf = $o_param -> Get('TIMESTAMP_FILE');
		$msg .= "Based on the TIMESTAMP file >$tsf< the update is NOT NECESSARY\n";
	}
	else
	{
		$msg .= "Update REQUIRED .... in progress\n";
	}
	my $narcisse_adm_email = $o_sp->GetInstanceObserver()->GetCfg('NARCISSE_ADM_EMAIL');
	my $org                = $o_sp->GetShortName();
	&SendEmail($o_sp, $msg,$narcisse_adm_email,"[NARCISSE]_Update_$org",$o_param);
	return if ( $tobeupdated == 1 );
	exit(0);
}

=head2 procedure SendEmail

 Title        : SendEmail
 Usage        : &SendEmail($msg,$email,$subject)
 Prerequisite : none
 Fonction     : Send an email containing the log file to Narcisse admins
 Returns      : none
 Args         : $o_sp   : the species observer
                $msg    : the message
 				$email  : the email address
				$subject: the subject of the message
				$o_param
 Globals      : none

=cut
sub SendEmail
{
	my ($o_sp, $msg,$email,$subject, $o_param) = @_;
	return if( $o_param->IsDefined( "no_mail" ) );
	my $o_io = $o_sp -> GetInstanceObserver();
	
	my $unix_mail = $o_io->GetUnixPath('UNIX_MAIL');
	my $narcisse_install_msg = "/tmp/narcisse_install_email.$$";
    my $f = new IO::File($narcisse_install_msg,"w") or Carp::croak("Cannot open $narcisse_install_msg");
	print $f "$msg\n";
	$f->close;
	my $cde_mail = "$unix_mail -s \"$subject\" $email < $narcisse_install_msg";
	`$cde_mail`;
	unlink("$narcisse_install_msg");
}
			
=head2 procedure DownloadFiles

 Title        : DownloadFiles
 Usage        : &DownloadFiles($o_logger)
 procedure    : Download files from repository
 Returns      : return the list of downloaded annotation and sequence files and keep track of the files location
 Args         : $o_sp
                $o_param ($o_param is MODIFIED by this function)
                $o_logger

=cut
sub DownloadFiles
{
	my ($o_sp, $o_param, $o_logger) = @_;
	
	$o_logger->Trace("STEP: downloading files \n",'Timer','A');
	my $o_io = $o_sp -> GetInstanceObserver();

	# Prepare some keys and regex		
	my $key_ann_regex = '^' . $o_io -> GetCfg ( 'PATTERN_ANNOTATION_FILE' );
	my $key_seq_regex = '^' . $o_io -> GetCfg ( 'PATTERN_SEQUENCE_FILE' );
	my @a_keys = $o_param -> GetKeys( $key_ann_regex );
	push (@a_keys, $o_param -> GetKeys( $key_seq_regex ) );
	#push (@a_keys, 'TIMESTAMP_FILE', 'FILE_ASSEMBLY', 'FILE_IDEOGRAM' );
	push (@a_keys, 'FILE_ASSEMBLY', 'FILE_IDEOGRAM' );

	if (defined $o_param->Get('FORMAT') && $o_param->Get('FORMAT') eq 'ENSEMBL')
	{
		&DownloadEnsemblData($o_sp, $o_param);
	}
	else
	{
		# Download annotation and sequence files
	    foreach my $key ( @a_keys )
	    {
			#warn $key . ' ' . $o_param->Get($key);
			my $downloaded = &NARCISSE_utils::DownloadData($o_io,$o_param->Get($key),$o_sp -> GetDownloadPath(),$o_logger);
			$o_param->Set($key,$downloaded);
	    }
	}

	# Update the time stamp
	&UpdateTimestamp( $o_sp );

	$o_logger->Trace("Files downloaded\n",'Timer','A');
	
    return;
}

=head2 procedure DownloadEnsemblData

 Title        : DownloadEnsemblData
 Usage        : DownloadEnsemblData($o_sp,$o_param)
 Fonction     : Download masked fasta sequences associated to $org et creates the release file.
 Args         : $o_sp (the species observer)
                $o_param IS MODIFIED (the parameters FILE_SEQ_* are added, where * stands for the chromosome names)

=cut

sub DownloadEnsemblData
{
	my ($o_sp, $o_param) = @_;

	my $org = ( $o_param -> IsDefined ( 'ORG_ENSEMBL' ) ) ? $o_param -> Get( 'ORG_ENSEMBL' ) : $o_sp -> GetLongName();
	my $registry = 'Bio::EnsEMBL::Registry';

	# Connecting to Ensembl
	my $o_io = $o_sp -> GetInstanceObserver();
	my $ensembl_host = $o_io -> GetCfg ( 'ENSEMBL_HOST');
	my $ensembl_user = $o_io -> GetCfg ( 'ENSEMBL_USER');
	
	$registry->load_registry_from_db(
    	-host => $ensembl_host,
    	-user => $ensembl_user
	);

	# Release
	my $db_adaptor = $registry->get_DBAdaptor($org, 'core');
	my $dbc = $db_adaptor->dbc;
	
	# Recuperation de la date de creation
	my $created = &__get_dbCreationDate($dbc, $dbc->dbname);
	
	# Recuperation de la version de l'assemblage
	my $assembly = $db_adaptor->get_CoordSystemAdaptor->fetch_by_name('chromosome')->version;

	$assembly = "" if (! defined $assembly);

	my $file_ens = $o_sp -> GetReleaseFile ( -check => &FALSE );
	my $f_ens = new IO::File($file_ens,"w") or Carp::croak("Error: unable to create >$file_ens<\n");
	print $f_ens "#release\tdate\tsite\tassembly\n";
	print $f_ens $db_adaptor->dbc->dbname."\t$created\tensembl\t$assembly";
	$f_ens->close;

	# Sequences masquees des chromosomes
	my $slice_adaptor = $registry->get_adaptor($org,'core','Slice');
	my $ra_chr = $slice_adaptor->fetch_all('chromosome');
	
	# If the parameter CHROMOSOMES is specified, it is used to restrict the choice
	my %h_restrict_chr = ();
	if ( $o_param -> IsDefined ( 'CHROMOSOMES' ) )
	{
		foreach my $chr ( split ( ',', $o_param -> Get ('CHROMOSOMES')) )
		{
			$h_restrict_chr { $chr } = 1;
		}
	}
	else
	{
		$h_restrict_chr { 'all_chromosomes' } = 1;
	}
	
	my $seq_file_key;
	my %h_chr_name=();
	foreach my $chr (@$ra_chr)
	{
		my $chr_name = $chr->seq_region_name;
		if ( exists $h_restrict_chr { 'all_chromosomes' } || exists $h_restrict_chr { $chr_name } )
		{
			$h_chr_name{$chr_name}=1;
		}
	}
	foreach my $chr_name (sort keys(%h_chr_name))
	{
		my $slice_adaptor_chr = $registry->get_adaptor($org,'core','Slice');
		my $chr = $slice_adaptor_chr->fetch_by_region('chromosome',$chr_name);
		#next if ( $chr_name !~ /MT/ ); warn "############### JG Attention: mode debug ##############\n";
		next if ( ! $o_param->Get('INCL_RANDOM')  && $chr_name =~ /_random$/ );
		if ( $chr_name =~ /_random$/ )
		{
			$chr_name =~ s/_random$/_r/;
		}

#		if ($chr_name eq '11' || $chr_name eq '17' || $chr_name eq '21' ||
#		$chr_name eq '7' || $chr_name eq 'Y' || $chr_name eq '13' || $chr_name eq '16' ||
#		$chr_name eq '18' || $chr_name eq '1' || $chr_name eq '22' || $chr_name eq '2'){ next; }

		print "Starting downloading chromosome sequence $chr_name\n";

		$seq_file_key = "FILE_SEQ_$chr_name";
		my $file = File::Spec->catfile( $o_sp -> GetDownloadPath(), $seq_file_key );
		#my $file = $rep_download."/".$org.$chr_name.".fasta";

		$o_param -> Set( $seq_file_key, $file);

		# Download the file only if not already done
		# (can be dangerous if the file is not complete...)
		if ( -e $file && ! -z $file )
		{
			print "Skipping $file: the file already exists!\n";
			next;
		}

		my $f_chr = new IO::File("$file","w") or Carp::croak("Error: unable to create >$file<\n");
		my $seq = undef;
		my $ctries=1;
		while ( !defined($seq) && $ctries <= 3 )
		{
			if ( $ctries > 1 )
			{
				printf STDERR "Try number #%d\n",$ctries;
			}
			eval
			{
				#print Dumper($chr);
				$seq = $chr->seq_region_Slice->get_repeatmasked_seq(undef, 1)->seq;
				if ( length($seq) != $chr->{seq_region_length} )
				{
					warn "Sequence length does not match, I try again\n";
					$seq = undef;
				}
			};
			$ctries++;
		}
		print $f_chr ">$org#$chr_name\n$seq\n";
		$f_chr->close;

		delete $chr->{'seq'};
		undef $slice_adaptor_chr;
		undef $chr;
	}
}

=head2 procedure __get_dbCreationDate

 Title        : __get_dbCreationDate
 Usage        : &__get_dbCreationDate()
 Prerequisite : None
 Fonction     : Retrieve the database creaztion date, when Ensembl data are used
 Returns      : Database creation date
 Args         :
 Globals      : None

=cut

sub __get_dbCreationDate{

	#	RF: Pas d'adaptor Ensembl pour acceder aux donnees de cette table?
	#	RF: Recupere-t-on la bonne date?
	my $db_connection = shift;
	my $db_name = shift;

	my $sth = $db_connection->prepare( "select created from mapping_session where new_db_name='".$db_name."'" );

	$sth->execute();
	my $ref = $sth->fetchrow_hashref();

	if (! defined $ref->{'created'}){
		warn ("Warning: Unable to find creation date for database $db_name\n");
		return "";
	}

	my @a_day_hour = split (" ", $ref->{'created'});
	return $a_day_hour[0];

}

=head2 procedure CheckFiles

 Title        : CheckFiles
 Usage        : &CheckFiles($o_io)
 procedure    : For each file, check if the file could be downloaded and print a message it if can't
 Returns      : nothing
 Args         : $o_sp, the Species observer

=cut
sub CheckFiles
{
	my ($o_sp,$o_param) = @_;
	my $o_io = $o_sp -> GetInstanceObserver();

	my @a_keys        = qw( FILE_ASSEMBLY FILE_IDEOGRAM TIMESTAMP_FILE );
	my $key_ann_regex = '^' . $o_io -> GetCfg ( 'PATTERN_ANNOTATION_FILE' );
	my $key_seq_regex = '^' . $o_io -> GetCfg ( 'PATTERN_SEQUENCE_FILE' );
	push (@a_keys, $o_param->GetKeys( $key_ann_regex ) );
	push (@a_keys, $o_param->GetKeys( $key_seq_regex ) );
    my @a_files = $o_param->Get(@a_keys);
    
    foreach my $file (@a_files)
    {
        next if ($file eq "");
        my $rvl = &NARCISSE_utils::DownloadData($o_io,$file);
        warn("WARNING - COULD NOT CHECK THIS FILE ($file)") if ($rvl==-1);
        warn("ERROR   - THIS FILE DOES NOT EXIST ($file)")  if ($rvl==0);
        print "OK      - CHECKED $file\n"                   if ($rvl==1);
    };
    return;
}



=head2 procedure CreateVirtualGenomes

 Title        : CreateVirtualGenomes
 Usage        : &CreateVirtualGenomes($o_sp, $o_param, \@a_chromosome_files)
 Prerequisite : The data files are already downloaded and masked
 Procedure    : Concatenate the .mask files to create "virtual genome" files of less than 1 Gb long
 Args         : $o_sp, $o_param, \@a_chromosome_files

=cut

sub CreateVirtualGenomes
{
	my ($o_sp, $o_param, $ra_chromosome_files) = @_;
	my $o_io       = $o_sp -> GetInstanceObserver();

	my $unix_rm    = $o_io->GetUnixPath('UNIX_RM');
	my $unix_cat   = $o_io->GetUnixPath('UNIX_CAT');
	my $unix_gzip  = $o_io->GetUnixPath('UNIX_GZIP');
	my $unix_gunzip= $o_io->GetUnixPath('UNIX_GUNZIP');

	# JG Avant integration Ensembl my @a_lst_chr = $o_param->GetKeys($KEY_ANNOTFILES);
	#my @a_lst_chr = split('#',$o_param->Get('AVAILABLE_CHR'));
    my $org       =  $o_sp -> GetShortName();

    my $dst_dir    = $o_sp -> GetFastaDir( feature => 'CHR' );
    my $index_dir  =  File::Spec->catfile( $dst_dir, $o_io -> GetCfg( 'INDEX' ) );
    mkdir $index_dir;

    my $vg_max_size= $o_param->IsDefined('virtual_genome_size') ? $o_param->Get('virtual_genome_size')  : 10000000000;

	# JG Avant integration Ensembl $chr =~ s/^$KEY_ANNOTFILES//;
	#TF only the list of chromosome names are needed
	
	my $key_annotfiles =  $o_io->GetCfg('PATTERN_ANNOTATION_FILE');
	my $key_seqfiles   = $o_io->GetCfg('PATTERN_SEQUENCE_FILE');
	#s/^($key_annotfiles|$key_seqfiles)// foreach @a_lst_chr;
	s/^($key_annotfiles|$key_seqfiles)// foreach @$ra_chromosome_files;

    my $virtual_pattern   = $index_dir.'/'.$org.'V[0-9].seq';
    `$unix_rm $virtual_pattern`;

    my $vg_nb;
    #If only one virtual genome is created the chromosomes are sorted in order to enable an efficient self comparison
  	if ($vg_max_size > $o_io -> GetCfg ('MAX_GENOME_SIZE' ) )
  	{
  		# Create the virtual genomes
    	my $local_cat="";
  		$vg_nb=1;
  		#foreach my $chr (&NARCISSE_utils::SortByChromID(@a_lst_chr))
  		foreach my $chr (&NARCISSE_utils::SortByChromID(@$ra_chromosome_files))
    	{
        	my $seqmask = &Chr2File($dst_dir,$org,$chr);
        	if ($seqmask eq '')
        	{
            	print "ERROR $0 : the chromosome >$chr< does not exist for the organism $org\n" and die;
       	 	};
       	 	my $virtual_name = $org.'V1';
        	my $virtual_genome = $index_dir.'/'.$virtual_name.'.seq';
       	 	if ($seqmask =~ /(.+)\.gz$/)
        	{
            	$local_cat = "$unix_gzip -cd $seqmask ";
        	}
        	else
        	{
            	$local_cat = "$unix_cat $seqmask";
        	}
        	&NARCISSE_utils::RunCmd($local_cat, ">>$virtual_genome");
    	}
  	}
    else  # In the other case we need to sort them carefully to balance the sizes of the virtual genomes
    {
    	# foreach chromosome, we measure the size of the chromosome (uncompressed)
   		my %h_chr_sizes=();
    	#foreach my $chr (@a_lst_chr)
     	foreach my $chr (@$ra_chromosome_files)
	   	{
        	#my $seqmask = "$dstdir/$org$chr.mask";
        	my $seqmask = &Chr2File($dst_dir,$org,$chr);
        	if ($seqmask eq '')
        	{
            	print "ERROR $0 : the chromosome >$chr< does not exist for the organism $org\n" and die;
       	 	};
        	if ($seqmask =~ /(.+)\.gz$/)
        	{
            	my $seqmask_nocomp = $1;
            	system("$unix_gunzip < $seqmask > $seqmask_nocomp");
            	$h_chr_sizes{$chr} = -s $seqmask_nocomp;
            	unlink $seqmask_nocomp;
        	}
        	else
        	{
            	$h_chr_sizes{$chr} = -s $seqmask;
        	};
    	};

    	# Number of virtual genomes
    	my $ttl_size=0;
    	foreach my $k (keys %h_chr_sizes) {$ttl_size += $h_chr_sizes{$k};};

    	$vg_nb = 1 + int($ttl_size / $vg_max_size);

    	warn "$vg_nb because $ttl_size and $vg_max_size \n";

    	# Create the virtual genomes
    	my $local_cat="";
    	my $vg = 1;
    	foreach my $k (sort {$h_chr_sizes{$a}<=>$h_chr_sizes{$b}} keys %h_chr_sizes)
    	{
        	my $virtual_name = $org.'V'.$vg;
        	my $virtual_genome = $index_dir.'/'.$virtual_name.'.seq';
        	my $seqmask = &Chr2File($dst_dir,$org,$k);
        	if ($seqmask eq '')
        	{
            	print "ERROR $0 : the chromosome >$k< does not exist for the organism $org\n" and die;
        	}
        	if ($seqmask =~ /(.+)\.gz$/)
        	{
            	$local_cat = "$unix_gzip -cd $seqmask ";
        	}
        	else
        	{
            	$local_cat = "$unix_cat $seqmask";
        	}

        	&NARCISSE_utils::RunCmd($local_cat, ">>$virtual_genome");
        	$vg = ($vg == $vg_nb) ? 1 : $vg+1;
    	}
    }
}

=head2 procedure IndexSequenceFile

 Title        : IndexSequenceFile
 Usage        : &IndexSequenceFile()
 Prerequisite : The data files are already downloaded and masked
 Fonction     : Concatenate the .mask files to create "virtual genome" files of less than 1 Gb long
                Call glintindex on those files through paraloop (because it is a time consuming process)
                human1.fasta human2.fasta ==> humanv1.fasta
 Returns      : nothing
 Args         : $o_sp,$o_param
 Globals      : $CFG, $TARGET

=cut

sub IndexSequenceFile
{
	my ($o_sp, $o_param) = @_;
	my $o_io = $o_sp -> GetInstanceObserver();
	
	# JG Avant integration Ensembl my @a_lst_chr = $O_PARAM->GetKeys($KEY_ANNOTFILES);
    my $org       = $o_sp -> GetShortName();

    my $dst_dir    = $o_sp -> GetFastaDir( feature => 'CHR' );
    my $index_dir  =  File::Spec->catfile( $dst_dir, $o_io -> GetCfg( 'INDEX' ) );

    my $index_mask= $o_param->IsDefined('glintindex_mask')       ? " -m ".$o_param->Get('glintindex_mask') : "";

    #
    # index_pass = if the user wants a specific value, give him this value
    #              Else, we select 5 for an eukaryote and 1 for the other kingdoms
    #

    my $index_pass="";
    if ($o_param->IsDefined('glintindex_pass'))
    {
        $index_pass = $o_param->Get('glintindex_pass');
    }
    else
    {
        $index_pass= ($o_io -> GetCfg('KINGDOM') eq "eukaryote") ? "-p 5" : "-p 1";
    }

    my $indexation_input = 'indexation_cmd';
    my $indexation_output= 'indexation_oe';

  	my $dh_work_dir = new IO::Dir( $index_dir ) or throw LipmError::IOException( $index_dir, 'd' );
  	my @a_files = $dh_work_dir->read();
	@a_files = File::Spec->no_upwards( @a_files );	#enleve . & ..
	my @a_virtual_genomes=();
	my $virtual_genome_pattern = $org.'V[0-9].seq';
	foreach my $file ( @a_files )
	{
		next	if ($file !~ /$virtual_genome_pattern$/ );
		push @a_virtual_genomes, $index_dir.'/'.$file;
	}

    if (@a_virtual_genomes==0)
    {
            print "ERROR, no virtual genome found for $org\n" and die;
    };

    my $commands = new IO::File(">$indexation_input") or die ("ERROR - Cannot open $indexation_input for writing");
    my $virtual_genome_name_pattern = $org.'V[0-9]';
    foreach my $virtual_genome (@a_virtual_genomes)
    {
        my ($virtual_name) = ( $virtual_genome =~ /($virtual_genome_name_pattern)/);
        my $root_indexer   = $o_io -> GetCfg( 'ROOT_INDEXER' );
        print $commands "$root_indexer $index_mask $index_pass -N $virtual_name $virtual_genome\n";
    };
    undef $commands;

	&NARCISSE_utils::RunCmdFromFile($indexation_input,$indexation_output);

}

=head2 function Chr2File

 Title        : Chr2File
 Usage        : &Chr2File($directory,$organism,chromosome);
 Fonction     : Compute and return a chromosome file name
 Parameters   : The directory, the organism, the chromosome number
 Returns      : The file name: .mask if possible, else .fasta
                               Compressed file if possible
                               "" if no file found
 Args         : directory name, organisme name, chromosome identification
 Globals      : none

=cut

sub Chr2File
{
    my ($dir,$org,$chr) = @_;
    my $prefix = $dir.'/'.$org.$chr;
    my $f;
    $f = $prefix.".mask.gz";  return $f if (-r $f);
    $f = $prefix.".mask";     return $f if (-r $f);
    $f = $prefix.".fasta.gz"; return $f if (-r $f);
    $f = $prefix.".fasta";    return $f if (-r $f);
    return "";
};


=head2 function ExtractAndReformatData

 Title        : ExtractAndReformatData
 Usage        : &ExtractAndReformatData()
 Prerequisite : none
 Fonction     : ExtractAndReformatData
 Returns      : $ra_chromosome_files (the list of chromosome files)
 Args         : $o_sp  : The species observer
                $o_logger, $o_param
 Globals      :

=cut

sub ExtractAndReformatData
{
    my ($o_sp, $o_logger, $o_param) = @_;
    
	my $o_io   = $o_sp -> GetInstanceObserver();
	my $unix_rm = $o_io->GetUnixPath('UNIX_RM');

	my %h_annotation_files=();
	my %h_sequences_files=();
	my %h_chromosome_files=();
		
	# find all annotation/sequence files
	&__DiscoverChromosomesFromAnnotation($o_sp, $o_param);
	&__DiscoverChromosomesFromSequence($o_sp, $o_param);

	# Discover chromosomes from sequence files
	my $key_seq_regex   = '^' . $o_io -> GetCfg ( 'PATTERN_SEQUENCE_FILE' );
	my @a_sequence_keys = $o_param -> GetKeys( $key_seq_regex );
	foreach my $key ( @a_sequence_keys ) 
	{
		my (@a_id)               = ($key =~ /$key_seq_regex(.+)/);
		my $id                   = $a_id[-1];
		$h_chromosome_files{$id} = 1;
		$h_sequences_files{$id}  = $o_param -> Get( $key );
		if ($o_param->Get('INCL_RANDOM')  && $o_param->Get('FORMAT') =~ /GENBANK/i )
		{
			my $chrom=$id."_r";
			$h_chromosome_files{$chrom} = 1;
			$h_sequences_files{$chrom}  = $o_param -> Get( $key );
		}
	}
	
	# Validate annotation keys: the key is deleted if there is no corresponding sequence file
	my $key_ann_regex     = '^' . $o_io -> GetCfg ( 'PATTERN_ANNOTATION_FILE' );
	my @a_annotation_keys = $o_param -> GetKeys( $key_ann_regex );
	foreach my $key ( @a_annotation_keys )
	{
		my (@a_id)               = ($key =~ /$key_ann_regex(.+)/);
		my $id                   = $a_id[-1];
		$h_chromosome_files{$id}        = 1;
		$h_annotation_files{$id}        = $o_param -> Get( $key );
		if ($o_param->Get('INCL_RANDOM')  && $o_param->Get('FORMAT') =~ /GENBANK/i )
		{
			my $chrom=$id."_r";
			$h_chromosome_files{$chrom} = 1;
			$h_annotation_files{$chrom} = $o_param -> Get( $key );
		}
	}
	my @a_chromosome_files = keys %h_chromosome_files;
				
	# Clean directories for most features, because the fasta files will be appended
	# Useless cleaning for CHR feature, because there is 1 file / chromosome
	foreach my $feature ( 'PROT', 'CDS', 'TRANSCRIPT', 'GENE' )
	{
		my $dir = $o_sp -> GetFastaDir( feature => $feature, -tmp => &TRUE, -check => &FALSE );
		`$unix_rm -r $dir/*`;
	}

	my $kingdom = $o_io -> GetCfg( 'KINGDOM' );
	my $tmp_dir = $o_sp -> GetTmpPath();
	my $species = $o_sp -> GetShortName();
	
	my %h_chr_length;
	
	my $o_summary = New ParamParser();
	my %h_total;
	
	# For each chromosome, create an object AnnotatedSeqFile, then call Load and Print methods on it
	# JG Avant integration Ensembl foreach my $chrom ( keys(%h_chromosome_files) )
	my $assembly_phase     = $o_sp -> GetAssemblyPhase();
	$assembly_phase =~ s/S$//; # Fucking the mouches
	#my @a_available_chroms = &NARCISSE_utils::SortByChromID ( split('#',$o_param->Get('AVAILABLE_CHR')) );
	my @a_available_chroms = &NARCISSE_utils::SortByChromID ( @a_chromosome_files );
	my $nb_chroms          = @a_available_chroms;
	my $cnt_chroms         = 0;
	foreach my $chrom ( @a_available_chroms )
	{
		$cnt_chroms++;
		print "========= $assembly_phase $cnt_chroms / $nb_chroms =========\n\n";
		my $annfile  = ( defined($h_annotation_files{$chrom}) ) ? $h_annotation_files{$chrom} : undef;
		printf "$assembly_phase=$chrom - Annotation file: %s\n",(defined($annfile)) ? $annfile : 'NONE';
		
		my $seqfile = ( defined($h_sequences_files{$chrom}) ) ? $h_sequences_files{$chrom} : undef;
		printf "$assembly_phase=$chrom - Sequence file: %s\n", (defined($seqfile)) ? $seqfile : 'NONE';
		
		if ( $o_param -> IsDefined ( 'FILE_ASSEMBLY' ) )
		{
			print "ASSEMBLY NEEDED - Assembly file: ", $o_param -> Get( 'FILE_ASSEMBLY' ) . "\n";
		}
		
		# The data source is the annotation file if possible, else it is the sequence file
		my $src_data = ( defined($annfile) ) ? $annfile : $seqfile;

		my $informat = $o_param->Get('FORMAT');
		if ( !defined($annfile) && $o_param->Get('FORMAT') !~ /FASTA|ENSEMBL/ )
		{
			if ( !defined($seqfile) )
			{
				print "Warning: annotation and/or sequence files not defined for chromosome $chrom\n";
				next;
			}
			else
			{
				print "Only the sequence file is available: input format forced to FASTA for this chromosome\n";
                # manu 8/10/09 - For the format to FASTA, but ONLY for this chrom
				# $O_PARAM->Set('FORMAT','FASTA');
                $informat = 'FASTA';
			}
		}

		my $table_usage = $o_param->Get("TABLE_$chrom");
		my $consistency_limit = $o_param->Get("CONSISTENCY_LIMIT_$chrom");
		$consistency_limit = &DEFAULT_CONSISTENCY_LIMIT if ( !defined($consistency_limit) || $consistency_limit eq '' );
		$o_logger->Trace("Extracting data for Chr $chrom from $src_data\n",'Timer','A');
		$o_logger->Trace("Consistency limit = $consistency_limit%\n",'Timer','A');
		
		# See AnnotatedseqFile::Add...CDS or Add...Exon - This is a to allow for buggy assembled data !
		$o_param->SetUnlessDefined( 'CDS_OVERLAP_ALLOWED',0);
		my $cds_overlap_allowed = $o_param->Get( 'CDS_OVERLAP_ALLOWED' );
		
		# build the parameters to pass to the AnnotatedSeqFile object
		my %h_param = ('infile'              => $annfile,
					   'chr'                 => $chrom,
					   'species'             => $species,
					   'informat'            => $informat,
					   'consistency_limit'   => $consistency_limit, # dans H_sapiens/CHR_Y/hs_ref_chrY.gbk.gz il y a 315 mRNA et seulement 153 CDS !!!!!!
					   'cds_overlap_allowed' => $cds_overlap_allowed,
					   'outformat'           => 'NARCISSE',
					   'tmpdir'              => $tmp_dir,
					   'kingdom'             => $kingdom,
					   'seq_file'            => $seqfile);

        # Parameters related to the assembly
		if( $o_param->IsDefined( 'FILE_ASSEMBLY' ) )
		{
			if( -e $o_param->Get('FILE_ASSEMBLY') )
			{
				$h_param{file_assembly}  = $o_param->Get('FILE_ASSEMBLY');
			}
			else
			{
				my $file_assembly = $o_param->Get('FILE_ASSEMBLY');
				my $msg = "Your FILE_ASSEMBLY ($file_assembly) does not exist.\n";
				throw LipmError( -text => $msg );
			}
		}
		$h_param{assembly_label} = $o_param->Get('ASSEMBLY')      if ($o_param->IsDefined('ASSEMBLY'));

        # Parameters related to the codon usage	
		if ( defined($table_usage) && $table_usage =~ /^\d+$/ )
		{
			# for codon usage table, we continue to automatically manage different tables for bacteria and eukaryote
			$h_param{table}  = $table_usage;
			$o_logger->Trace("Codon usage table = $table_usage\n",'Timer','A');
		}

        # build the $o_annotfile object, using a little factory
        if ( $informat eq 'ENSEMBL' )
        {
			$h_param{ org_ensembl }  = ( $o_param -> IsDefined ( 'ORG_ENSEMBL' ) ) ? $o_param -> Get( 'ORG_ENSEMBL' ) : $o_sp -> GetLongName();
			$h_param{ ensembl_host } = $o_io -> GetCfg ( 'ENSEMBL_HOST');
			$h_param{ ensembl_user } = $o_io -> GetCfg ( 'ENSEMBL_USER');
		}
		
		my $o_annotfile;
		my $err_flg = &FALSE;
		try
		{
			$o_annotfile= AnnotatedSeqFileFactory( %h_param );
		}
		catch LipmError::ParameterException with
		{
			my ( $o_err) = @_;
			$err_flg = &TRUE;
			my $assembly_phase = $o_param -> Get('ASSEMBLY_PHASE');
			print "WARNING - " . $o_err->Explain() . "\n";
			print "Skipping this $assembly_phase\n";
		};
		next if ($err_flg);

		#$o_annotfile= AnnotatedSeqFileFactory( %h_param );
		
		# Goto to debug mode if necessary
		$o_annotfile -> SetDebug() if ( $o_param -> Get('debug') eq '1' );
		
		#print "KOUKOU $file_repeat $file_rna  $file_exon $file_gene $file_transcript $file_seq  \n";

		$o_logger->Trace("Start loading chr $chrom\n",'Timer','A');
		
		# Read the files and load everything in memory
		$o_annotfile->Load();
		
		# Print summary about the loaded data, and keep summary to $o_summary
		$o_annotfile->Print('');

		# only for debug...
		if ( $o_param -> Get('debug') eq '1' )
		{
			$o_annotfile->Print('full');
			$o_annotfile->Validation() ;
		}

		$o_logger->Trace("Start dumping chr $chrom\n",'Timer','A');
				
		# Create the sequence file from the assembly
		my $file_seq_db = $o_sp -> GetFastaFile( feature => 'CHR',  chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		$o_annotfile->BuildChromosomeFile( $file_seq_db );

		$o_param->SetUnlessDefined( 'PROTEIN_ID', 'CDS' );
		my $protein_id = $o_param->Get( 'PROTEIN_ID' );
		
		# Create the fasta files, and set the consistency flag for the objects cds and mrna
		# If PSEUDOMOLECULES, give up if there is some exception
		# Else, just skip the contig (annotation is not so good)
		try
		{
			$o_annotfile->DumpFastaDb(
									  'prot'       => $o_sp -> GetFastaFile( feature => 'PROT', -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE ),
									  'cds'        => $o_sp -> GetFastaFile( feature => 'CDS',  -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE ),
									  'mrna'       => $o_sp -> GetFastaFile( feature => 'TRANSCRIPT', -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE ),
									  'gene'       => $o_sp -> GetFastaFile( feature => 'GENE',       -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE ),
									  'sequence'   => $file_seq_db,
			  					      'protein_id' => $protein_id
									  );
		}
		catch Error::Simple with
		{
			my ( $o_err) = @_;
			$err_flg = &TRUE;
			my $assembly_phase = $o_param -> Get('ASSEMBLY_PHASE');
			print "WARNING - " . $o_err->text() . "\n";
			if ( $assembly_phase eq 'PSEUDOMOLECULES' )
			{
				print "ERROR - WORKING WITH PSEUDOMOLECULES: SHOULD NOT HAPPEN, Giving up\n";
				throw $o_err;
			}
			else
			{
				print "Skipping this $assembly_phase\n";
			}
		};
		next if ($err_flg);
		
		# Verify the consistency limit, do not return if too many inconsistencies
		# NOTE - THIS WAS REMOVED if NOT PSEUDOMOLECULES because sometimes there are juste 2,3 genes in the contig
		$o_annotfile -> AssertConsistencyLimit() if ($o_sp -> GetAssemblyPhase() eq 'PSEUDOMOLECULES' );
		
		# Create the Narcisse-formatted files
		my $file_repeat = $o_sp -> GetRepeatFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		my $file_rna    = $o_sp -> GetRnaFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		my $file_exon   = $o_sp -> GetExonFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		my $file_gene   = $o_sp -> GetGeneFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
		my $file_transcript = $o_sp -> GetTranscriptFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );

		my %h_param1    = ('gene'       => $file_gene, 'transcript' => $file_transcript,
					       'exon'       => $file_exon, 'repeat'     => $file_repeat,
					       'rna'        => $file_rna,  'sequence'   => $file_seq_db,
					       'protein_id' => $protein_id);

		if ( $o_param->Get('FORMAT') eq 'GENETIC')
		{		
			my $file_marker     = $o_sp -> GetMarkerFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
			$h_param1{'marker'} = $file_marker;
		}
		if ( $o_param->Get('FORMAT') ne 'GENETIC')
		{		
			my $file_prot     = $o_sp -> GetProtFiles( -chrom => $chrom, -check => &FALSE, -tmp => &TRUE, -mkdir => &TRUE );
			$h_param1{'prot'} = $file_prot;
		}

		$o_annotfile -> Dump(%h_param1);

		# Keep track of the chromosome length (this is set by BuildChromosomeFile)
		$h_chr_length { $chrom } = $o_annotfile -> GetLength();

		# Keep the summary in memory for later report - If not working with pseudochromomes we don't (too many keys)
		if ( $o_sp -> GetAssemblyPhase() eq 'PSEUDOMOLECULES' )
		{
			$o_summary ->SelectNameSpace ( 'CHROMOSOME'.$chrom.'_' );
			$o_summary->Set('GENE',      $o_annotfile -> GetNbGenes());
			$o_summary->Set('TRANSCRIPT',$o_annotfile -> GetNbTranscripts());
			$o_summary->Set('EXON',      $o_annotfile -> GetNbExons());
	
			$o_summary->Set('RNA',       $o_annotfile -> GetNbRNAs());
			$o_summary->Set('REPEAT',    $o_annotfile -> GetNbRepeats());
	
			$o_summary->Set('GENE_DB',    $o_annotfile -> GetNbGeneSequences());
			$o_summary->Set('TRANSCRIPT_DB', $o_annotfile -> GetNbTranscriptSequences());
			$o_summary->Set('CDS_DB',     $o_annotfile -> GetNbCdsSequences());
			$o_summary->Set('PROTEIN_DB', $o_annotfile -> GetNbProteinSequences());
		}

		$h_total { LENGTH }     += $o_annotfile -> GetLength();
		$h_total { GENE }       += $o_annotfile -> GetNbGenes();
		$h_total { TRANSCRIPT } += $o_annotfile -> GetNbTranscripts();
		$h_total { EXON }       += $o_annotfile -> GetNbExons();
		
		$h_total { RNA }        += $o_annotfile -> GetNbRNAs();
		$h_total { REPEAT }     += $o_annotfile -> GetNbRepeats();
		
		$o_logger->Trace("End chr $chrom\n",'Timer','A');

    } #foreach my $chrom
    
    # Keep the total in summary.cfg
    $o_summary -> SelectNameSpace ( 'TOTAL_' );
    $o_summary -> Set('LENGTH',     $h_total { LENGTH } );
	$o_summary -> Set('GENE',       $h_total { GENE }  );
	$o_summary -> Set('TRANSCRIPT', $h_total { TRANSCRIPT }); 
	$o_summary -> Set('EXON',       $h_total { EXON });

	$o_summary -> Set('RNA',        $h_total { RNA });
	$o_summary -> Set('REPEAT',     $h_total { REPEAT });

	# Dump $o_summary to the dedicated file
	my $summary_file = $o_sp -> GetSummaryFile( -tmp => &TRUE, -check => &FALSE, -mkdir => &TRUE );
	$o_summary -> SelectNameSpace();
	$o_summary -> Dump ($summary_file);

	# Well - It is not obvious knowing which chromosomes/contigs/etc are valid and which are not
	#        We test the transcript files to know this
	#        TODO - Ouh que c'est pas joli ce bordel !
	@a_available_chroms = ();
	foreach my $chrom ( &NARCISSE_utils::SortByChromID ( @a_chromosome_files ) )
	{
		try
		{
			$o_sp->GetTranscriptFiles( -chrom  => $chrom, -tmp => &TRUE );
			push @a_available_chroms, $chrom;
		}
		catch Error with {};
	}
	
	# Please do not use @a_chromosome_files from this point
	@a_chromosome_files = ();
	
	#Extract for each gene the longest peptide and store the corresponding protein in the rep_prot db file
	&ExtractLongestPeptide( $o_sp, $o_param, \@a_available_chroms );

	# Create the karyotype and cytogenetic files
	my $rh_cyto = {};
	if ($o_param -> IsDefined ( 'FILE_IDEOGRAM' ) && -e $o_param -> Get ( 'FILE_IDEOGRAM' ) )
	{
		my $file_ideogram = $o_param -> Get ( 'FILE_IDEOGRAM' );
		$rh_cyto  = &ExtractCytogeneticData($o_sp, $file_ideogram);
	}
	my $ra_files = &DumpCytogeneticData($rh_cyto, \%h_chr_length, $o_sp, $o_param, $o_logger);
	return \@a_available_chroms;
}

=head2 __DiscoverChromosomesFromAnnotation

 Title     : __DiscoverChromosomesFromAnnotation
 Usage     : __DiscoverChromosomesFromAnnotation($o_io,$o_param)
 Procedure : Does nothing unless the ONLY annotation parameter is _DEFAULTS file (all annotations in one file)
			 In this case, reads the file for chromosomes discovery, unless the CHROMOSOMES parameter is specified
			 Modifies the CHROMOSOMES and if necessary creates FILE_ANN_* parameters, also the gff3 files
 Args      : $o_sp, $o_params 
=cut
sub __DiscoverChromosomesFromAnnotation
{
	my ($o_sp, $o_param) = @_;
	my $o_io   = $o_sp -> GetInstanceObserver();

	my $key_ann_regex = '^' . $o_io -> GetCfg ( 'PATTERN_ANNOTATION_FILE' );
	
	# find all annotation files and discover chromosomes
	my @a_annotation_keys = $o_param -> GetKeys( $key_ann_regex );
	
	# If only 1 key, and if this key is  XXXX_DEFAULT, we must have declared the chromosomes somewhere else
	# We build keys for each chromosome annotation, because this means that we have only 1 file for all chromosomes
	# The chromosomes may be declared from the CHROMOSOMES key, OR they may be discovered from the annotation file
	# ATTENTION - CA MARCHE ACTUELLEMENT UNIQUEMENT AVEC GFF3 !!!
	if ( @a_annotation_keys==1 && $a_annotation_keys[0] =~ /^(.+)_DEFAULT$/ )
	{
		die "ERROR - The _DEFAULTS keyword works ONLY with gff3 file formats" if ($o_param->Get('FORMAT') ne 'GFF3');
		die "ERROR - The _DEFAULTS annotation/sequence files are incompatible with assembly process during nc_get_data.pl" if ($o_param -> IsDefined('FILE_ASSEMBLY'));

		my $key_prefix = $1;
		my $ann_file = $o_param -> Get ( $a_annotation_keys[0] );
		my @a_chromosomes;

		# NOTE - This code is specific to gff3 files - not very clean...
		if ( !$o_param->IsDefined('CHROMOSOMES') )
		{
			print "DEFAULT ANNOTATION FILE FOUND - Looking for the " . $o_param -> Get ( 'ASSEMBLY_PHASE' ) . "\n";
			my $fh_ann_file = new IO::File ($ann_file) or throw LipmError::IOException( $ann_file, 'f' );
			my %h_ids;
			my $tmp_download = $o_sp -> GetDownloadPath();

			my $unix_rm = $o_io -> GetUnixPath ( 'UNIX_RM' );
			print "Removing old gff3 files from the download directory\n";
			my $cmd = "$unix_rm -f $tmp_download/*.gff3";
			`$cmd`;

			my $fh_id=undef;
			my $prev_id = '';
			while (my $line = <$fh_ann_file>)
			{
				chomp $line;
				next if ($line =~ /^#/ || $line eq '');
				my @a_line = split "\t", $line;
				if ( $a_line[0] ne $prev_id )
				{
					my $id        = $a_line[0];
					$prev_id      = $id;
					$fh_id -> close() if ( defined $fh_id );
					my $file_id   = File::Spec->catfile ( $tmp_download, $id . '.gff3' );
					if ( -e $file_id )
					{
						$fh_id = new IO::File( $file_id, 'a' ) or throw LipmError::IOException( $file_id, 'f' );
					}
					else
					{
						$fh_id = new IO::File( $file_id, 'w' ) or throw LipmError::IOException( $file_id, 'f' );
						print $fh_id "##gff-version 3\n";
					}
					$h_ids { $id }= 1;
					$o_param -> SetUnlessDefined ( $key_prefix . '_' . $id, $file_id );
				}
				print $fh_id "$line\n";
			}
			$fh_id -> close() if ( defined $fh_id );
			@a_chromosomes = keys %h_ids;
			print "FOUND " . @a_chromosomes . " " . $o_param -> Get ( 'ASSEMBLY_PHASE' ) . " FROM THE ANNOTATION FILE\n";
			$o_param -> Set ( 'CHROMOSOMES', join (',',keys(%h_ids)));
		}
		else
		{
			print "DEFAULT ANNOTATION FILE FOUND - Exploding the CHROMOSOMES: " . $o_param->Get('CHROMOSOMES') . "\n";
			@a_chromosomes = split (/,/,$o_param->Get('CHROMOSOMES'));
			foreach my $chr (@a_chromosomes)
			{
				$o_param -> SetUnlessDefined ( $key_prefix . '_' . $chr, $ann_file );
			}
		}

		$o_param -> Delete ( $a_annotation_keys[0] ); # the _DEFAULTS parameter
		#@a_annotation_keys = $o_param -> GetKeys( $key_ann_regex );
	}
}


=head2 __DiscoverChromosomesFromSequence

 Title     : __DiscoverChromosomesFromSequence
 Usage     : __DiscoverChromosomesFromAnnotation($o_io,$o_param)
 Procedure : Does nothing unless the ONLY sequence parameter is _DEFAULTS file (all annotations in one file)
			 In this case, reads the sequence file, and discover the chromosomes, create the FILE_SEQ_* parameters
			 and the fasta files
 Args      : $o_sp, $o_params 
=cut
sub __DiscoverChromosomesFromSequence
{
	my ($o_sp, $o_param) = @_;
	my $o_io   = $o_sp -> GetInstanceObserver();

	# find all sequence files and discover chromosomes
	my $key_ann_regex = '^' . $o_io -> GetCfg ( 'PATTERN_ANNOTATION_FILE' );
	my $key_seq_regex = '^' . $o_io -> GetCfg ( 'PATTERN_SEQUENCE_FILE' );
	my @a_sequence_keys = $o_param -> GetKeys( $key_seq_regex );
	#my @a_annotation_keys = $o_param -> GetKeys( $key_ann_regex );

	# If only 1 key, and if this key is  XXXX_DEFAULT, we have declared the chromosomes explicitely, using the CHROMOSOMES parameter,
	# or the chromosomes are discovered from the annotation file. A sequence file without any annotation will be simply ignored.
	
	# We build keys for each chromosome sequence, because this means that we have only 1 file for all chromosomes
	# TODO - Faire cela plus proprement, par exemple dans un objet avec derivations dans le style AnnotatedSeqFile
	#        Actuellement les fichiers annotations fixent les chromosomes, et un fichier de sequence sans annotation correspondante sera ignore
	
	# We read the sequence file, which should be a multifasta file, 
	# and for each "chromosome" (or contig, scaffold etc) recognized, we copy the sequence to a single fasta file
	# some sequences do not have a corresponding annotation
	if ( @a_sequence_keys==1 && $a_sequence_keys[0] =~ /^(.+)_DEFAULT$/ )
	{
		die "ERROR - The _DEFAULTS annotation/sequence files are incompatible with assembly process during nc_get_data.pl" if ($o_param -> IsDefined('FILE_ASSEMBLY'));
		my $key_prefix = $1;
		#my @a_chromosomes = split (/,/,$o_param->Get('CHROMOSOMES'));
		my $seq_file = $o_param -> Get ( $a_sequence_keys[0] );
		my $tmp_download = $o_sp -> GetDownloadPath();
		my $o_fasta_r    = New FastaReader( $seq_file );
		my $ra_headers   = $o_fasta_r->GetHeaders();
		foreach my $header ( @$ra_headers )
		{
			my( $chrom ) = $header =~ />([^ ]+).*/;
			my $chrom_seq_file = File::Spec->catfile( $tmp_download, "$chrom.fasta" );

			$o_param->SetUnlessDefined( $key_prefix . '_' . $chrom, $chrom_seq_file );
			my $fh_c_seq_file  = new IO::File( $chrom_seq_file, 'w' )
								or throw LipmError::IOException( $chrom_seq_file, 'f' );
			print $fh_c_seq_file ">$chrom\n";
			print $fh_c_seq_file $o_fasta_r->GetRawSequence( $header ) . "\n";
		}

		$o_param->Delete( $a_sequence_keys[0] );
		@a_sequence_keys = $o_param->GetKeys( $key_seq_regex );
		print "FOUND @a_sequence_keys " . $o_param->Get( 'ASSEMBLY_PHASE' ) . " FROM THE SEQUENCE FILE\n";
	}

	return;
}



=head2 AnnotatedSeqFileFactory

 Title     : AnnotatedSeqFileFactory
 Usage     : my $o_annot = &AnnotatedSeqFileFactory( %h_params );
 Function  : Use $h_param -> informat to choose an AnnotatedSeqfile to build
 Args     : Parameters passed to the AnnotatedSeqFile constructor
 Return   : the built object
=cut
sub AnnotatedSeqFileFactory
{
	my (%h_param) = @_;
	
	my $informat = $h_param{ informat };
	my $o_annotfile;
	if ( $informat eq 'GENBANK' )
	{
		$o_annotfile = New AnnotatedSeqFile::Genbank(%h_param);
	}
	elsif  ( $informat eq 'GFF3' )
	{
		$o_annotfile = New AnnotatedSeqFile::GFF3(%h_param);
	}
	elsif  ( $informat eq 'EMBL' )
	{
		$o_annotfile = New AnnotatedSeqFile::Embl(%h_param);
	}
	elsif  ( $informat eq 'TIGRXML' )
	{
		$o_annotfile = New AnnotatedSeqFile::TIGRXML(%h_param);
	}
	elsif  ( $informat eq 'FASTA' )
	{
		$o_annotfile = New AnnotatedSeqFile::Fasta(%h_param);
	}
	elsif  ( $informat eq 'ENSEMBL' )
	{
		
		$o_annotfile = New AnnotatedSeqFile::Ensembl(%h_param);
	}
	elsif  ( $informat eq 'GENETIC' )
	{
		$o_annotfile = New AnnotatedSeqFile::Genetic(%h_param);
	}
	else
	{
		Carp::croak("Unmanaged input format >$informat<\n");
	}
	
	return $o_annotfile;
}

=head2 procedure Usage

 Title        : Usage
 Usage        : &Usage()
 Prerequisite : none
 Fonction     : Print the program usage
 Returns      :
 Args         :
 Globals      : none

=cut

sub Usage
{
	my $self = shift;

	my $narcisse_cfg = $self->Get('NARCISSE_CFG');
	my $assembly;
	if ($self -> IsDefined('assembly'))
	{
		$assembly = $self->Get('assembly');
	}
	else
	{
		$assembly = 'none specified';
	}

print <<END

NARCISSE_CFG = $narcisse_cfg

Usage : $0

mandatory:
    --cfg  : species configuration file or genome short name

optional:
    --assembly=assembly  :   group_label in seq_contig.md file, default value is the value of ASSEMBLY in cfg file.
    --glintindex_mask    :   The mask for indexation
    --glintindex_pass    :   The nb of passes for indexation (default 1 for bacteria, 5 for eukaryotes)
    --virtual_genome_size: The max size of a virtual genome file (default 300000000, ie 300 Mb)
    --with_random      :   build random chromosomes
    --check_files      :   check if required files are online
    --no_mail          :   No mail will be sent.
    --root_log=dirpath :   specify a directory where logs will be written.
    --debug=1          :   VERY verbose but VERY useful to debug

    --help             :   this message

SELECTING SOME ACTIONS:
    --only_xxx:            If some only_xxx switches are specified, ONLY those specific actions will be made
    --no_xxx:              If some no_xxx actions are reported, ONLY those specific actions will NOT be made
                           It is illegal specifying only_xxx AND no_xxx switches

    --only_check :         Check the data - If the check is positive, we exit from the program
    --no_check
    --check_files:         Only valid with --only_check: check the existence of each data file, thus the validity of the .cfg file
    --only_download        Download files, extract data, mask for low complexity, ...
    --no_download
    --only_suppdata        compute statistics on genomic data (gc%, gene density, gene surface)
    --no_suppdata
    --only_formatdb        Call formatdb for blast indexation
    --no_formatdb
    --only_index           Call glintindex for indexing the data
    --no_index
    --only_clean           Remove downloaded and no-more-useful data
    --no_clean

END
;
exit;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
