#!/usr/bin/perl

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=pod

=head1 NAME

 nc_get_snp_data.pl

=head1 SYNOPSIS

 perl nc_get_snp_data.pl @args @options

 @args:
  --species, short/long species name

 @options:
  --verbose, flag if set log messages will also be printed out.
  --clean,   flag if set delete all previous snp data.
  --raw_output, filepath of the output of the alignment program.
                Specify it if you want to keep it.

=head1 DESCRIPTION

 This script retrieves SNP from various DB described in ?? for the given species
 It has to be launched after nc_get_data.pl

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 05-08, 2009
 $Id: nc_get_snp_data.pl 3675 2011-03-19 16:59:21Z sletort $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;
use FindBin;
use File::Basename;
use File::Path;
use Cwd;
use URI;
use LWP::UserAgent;

use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../lib/snp";
use lib "$FindBin::Bin/../lib/alignement";
use lib "$FindBin::Bin/../API";

# nos modules, nos classes
use NARCISSE_utils;
use LipmError;
use LipmError::ParameterException;
use Sufine;
use General;
use ParamParser;
use OnlyOrNo;
use Logger;
use ConfigurationFactory;
use SequencesLocator::Snp;
use SnpManager;
use NarcisseDataWriter::OnDisk;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3675 $ =~ /\d+/g); $r[0]};

use constant SPECIES => 'species';
use constant VERBOSE => 'verbose';
use constant CLEAN   => 'clean';
use constant RAW_OUTPUT => 'raw_output';

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;
$Error::Debug = &DEBUG;

use constant T_WHOLE => 'WholeProg';
use constant T_DOWNLOAD => 'DownloadProcess';
use constant T_INTEGRATION => 'NarciseIntegration';
use constant CHUNK_SIZE => 100_000;

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{
			warn( $o_err->Explain(), "\n" );
			warn $o_err->stacktrace();
		}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}

		exit( 1 );
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;

		exit( 1 );
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{

	&ParamParser::SetDefaultUsage(\&Usage);
	&ParamParser::SetDefaultBehaviour('use_substitution_table');

	# Init ParamParser from the environment, just to NARCISSE_CFG
	my $o_param = New ParamParser('ENV');
	throw Error ( -text => "Please type chins some_instance before using this script") unless ($o_param->IsDefined('NARCISSE_CFG'));
	$o_param->AssertFileExists('NARCISSE_CFG');

	# Init the instance observer
	my $o_io = ConfigurationFactory::GetConfiguration();
	throw Error ( -text => "ERROR - This script does not work in the instance multi")  if ($o_io->isa('InstanceObserver')==&FALSE);

	# Some parameters that will be used later as substitution values
	my ($root_data) = $o_io->GetDataPaths( 'INSTANCE' );
	my $snp_repository =
		($o_param->IsDefined('NARCISSE_SNP_REPOSITORY')) ? $o_param->Get('NARCISSE_SNP_REPOSITORY') : $o_io->GetCfg('NARCISSE_SNP_REPOSITORY');

	&NARCISSE_utils::Welcome(basename($0), &VERSION, $o_param->Get('NARCISSE_CFG'),$snp_repository);

	# We do not need the ENV any more - A new ParamParser with the parameters
	$o_param = New ParamParser('GETOPTLONG','cfg:s','chromosomes:s','raw_output:s','only_check','no_check','check_files','only_download','no_download','only_install','no_install','only_clean','no_clean','help');
	$o_param->AssertDefined('cfg');
	$o_param->SetUnlessDefined('chromosomes','all');
	if (($o_param->Get('chromosomes') eq 'all' || $o_param->Get('chromosomes') =~ /,/) and $o_param->IsDefined('raw_output'))
	{
		die "ERROR - The switch --raw_output may NOT be specified if there are several chromosomes";
	}

	# The substitutions
	$o_param->SetSubstitution('%s',$snp_repository) if ($snp_repository ne '');
	$o_param->SetSubstitution('%d',$root_data);

	# Only_xxx or no_xxx ?
	my $o_only = New OnlyOrNo($o_param);

	# Read the .cfg species file
	&NARCISSE_utils::AssertCfgParam( $o_io,$o_param,'cfg','INSTANCE' );
	my $file_arg = $o_param->Get('cfg');
	$o_param->Update($file_arg,'A');

	# Minimal parameter set
	$o_param->AssertDefined('DST','ORG_SHORT');
	my $species = $o_param->Get('ORG_SHORT');
	my $o_species = $o_io->GetSpeciesObserver( $species );

	#
	# if only_check is the only switch specified, just check, print a message and leave
	#
	# WARNING The pattern -1- and -0- may be checked by another script, do not modify it
	#

	if ($o_only->Get('check') and $o_only->HowManyOnly()==1)
	{
		print "NO TIMESTAMP MIRROR FOR THE SNP FILES - CANNOT CHECK\n";
        if ($o_param->IsDefined('check_files'))
        {
            &CheckFiles($o_param);
        }
		exit 0;
	}
    else
    {
        if ($o_param->IsDefined('check_files'))
        {
            die ("ERROR - check_files must be specified ONLY with --only_check AS ONLY SWITCH");
        }
    }

	# Compute and print $target
	my $sav_dir     = getcwd();
    my $target_path = &SetTargetLogPath( $o_io, $o_param, $species );

	$o_param->Dump("narcisse.param");

	my $logger_file = "narcisse.log";
	my $o_logger    = New Logger ($logger_file,$o_io->GetCfg('LOG_LEVEL'));
	$o_logger->StartTimer(&T_WHOLE);
	$o_logger->Trace("STARTING nc_get_snp_data.pl: $species\n",&T_WHOLE,'A');

	my $ra_files;
	$ra_files  = &DownloadFiles( $o_species, $o_param, $o_only, $o_logger, &FALSE );

	if ($o_only->Get('install'))
	{
		$o_logger->StartTimer( &T_INTEGRATION );
        $o_logger->Trace("STARTING INTEGRATION to narcisse data\n",&T_INTEGRATION,'A');

        # Create 1 object NarcisseDataWriter / chromosome
        # The object is created in APPEND mode, AND with numerical sort using the position field
		# 2011-01-24: seb - Si on ne met à jour que quelques chroms, on crée plein de fichiers vides. c'est voulu/grave ?
        my %h_writers;
        my $rh_chroms   = $o_species -> GetChromsInfo();
		my @a_headers   = qw( chrom element position strand values extern_id in_cds mutation );
		my %h_options   = ( sort => &TRUE, sort_numerical => &TRUE, sort_uniq => &TRUE, append => &TRUE );
		my @a_sort_keys = qw( position );
        foreach my $chrom (keys %$rh_chroms)
		{
			my $filepath = $o_species->GetSnpFiles(
			-chrom => $chrom,
			-check => &FALSE,
			-where => 'INSTANCE',
			-tmp => &TRUE,
			-mkdir => &TRUE);
			my $o_ndw = New NarcisseDataWriter::OnDisk( $filepath, \@a_headers, \%h_options, \@a_sort_keys, $o_logger);
			$h_writers{$chrom} = $o_ndw;
		}

		# Read and integrate every file
		foreach my $file ( @$ra_files )
		{
			try
			{
				&IntegrateSnpToNarcisse( $file, $o_species, \%h_writers, $o_logger, $o_param->Get(&RAW_OUTPUT) );
			}
			catch LipmError with
			{
				my ( $o_err, $r ) = @_;

				if( &TRUE == &DEBUG )
				{
					warn( $o_err->Explain(), "\n" );
					warn $o_err->stacktrace();
				}
				else
				{
					print '*' . $o_err->GetUserMessage() . "*\n";
				}
			};
		}
		
		# Sorting and closing the destination files
		# 2011-01-24: seb - le code ne fait pas ce que le commentaire dit.
		foreach my $chrom (keys %$rh_chroms)
		{
			$rh_chroms -> { $chrom } = undef;
		}
		
		$o_logger->StopTimer( &T_INTEGRATION );
		$o_logger->Trace( "INTEGRATION finished\n", &T_INTEGRATION, 'A');
	}
	if ($o_only->Get('clean'))
	{
        
		if ($o_param->Get('chromosomes') ne 'all')
		{
			warn "WARNING - No cleaning if --chromosomes ne 'all', the results are in tmp_data";
		}
		else
		{
			&CleanDirectory($o_species);
		}
	}

	$o_logger->StopTimer( &T_WHOLE );
	$o_logger->Trace( "FINISHED\n", &T_WHOLE, 'A');

	return;
}

=head2 CleanDirectory

 Usage     : CleanDirectory($o_species)
 Procedure : Move File from species/Tmp_data/snp directory to species/snp
             (WARNING - The files in snp directory are then cleared
             Remove tmp, tmp_data, tmp_download directories
 Args      : $o_species
 
=cut
sub CleanDirectory
{
	my ($o_species) = @_;

	my $o_io    = $o_species -> GetInstanceObserver();
	my $unix_rm = $o_io -> GetUnixPath( 'UNIX_RM' );
	my $unix_mv = $o_io -> GetUnixPath( 'UNIX_MV' );
	
	# Put new files in place
	my $rh_chroms       = $o_species -> GetChromsInfo();
	foreach my $chr (keys %$rh_chroms)
	{
		my $dst = $o_species -> GetSnpFiles ( -where => 'INSTANCE', -chrom => $chr, -check => &FALSE );
		my $src = $o_species -> GetSnpFiles ( -where => 'INSTANCE', -chrom => $chr, -tmp => &TRUE, -check => &FALSE );
		`$unix_mv $src $dst` if -e $src;
	}
	
	# Remove temporary directories
	my $tmp = $o_species -> GetTmpPath();
	my $download = $o_species -> GetDownloadPath();
	my $tmp_data = $o_species -> GetDataPath( 'INSTANCE', &FALSE, &TRUE );
	my $cmd = $unix_rm . ' -r ';
	$cmd .= "$tmp $download $tmp_data";
	#warn "KOUKOU $cmd";
	############`$cmd`;
}
	
=head2 Procedure IntegrateSnpToNarcisse

 Usage     : IntegrateSnpToNarcisse( $snp_file, $o_species, $rh_writers, $o_logger, $output )
 Procedure : Generate the filesnp files, narcisse formatted
             Split the $snp_file in chunks and call IntegrateChunkToNarcisse for each chunk
 Args      : $snp_file   The file to integrate (may be compressed)
                         Return with a warning if the file is not found
             $o_species  The species observer
             $rh_writers A hash with keys: chromosome names, values: NarcisseDataWriter objects
             $o_logger   A logger object
             $output     opt, path to the glint output raw file

=cut

sub IntegrateSnpToNarcisse
{
	my( $snp_file, $o_species, $rh_writers, $o_logger, $output ) = @_;

	$output = "" unless defined $output;

	&NARCISSE_utils::Unzip($snp_file);
	$snp_file =~ s/\.gz//;

	# We work in chunks, the number of sequences is CHUNK_SIZE for each chunk
	# For each chunk, we create a new fasta file
	# The we call IntegrateSnpChunkToNarcisse
	my( $chunk_path, $chunk_output, $fh_snp_chunk_file );
	my( $chunk_nb, $seq_nb ) = ( 0, 0 );
	$chunk_path  = &__GetChunkPath( $chunk_nb, $snp_file );
	$chunk_output= &__GetChunkPath( $chunk_nb, $output ) if( $output ne "" );
	$fh_snp_chunk_file = new IO::File( $chunk_path, 'w' )
						or throw LipmError::IOException( $chunk_path, 'f' );

	my $o_fasta = New FastaReader( $snp_file );
	for my $header ( $o_fasta->getHeaders() )
	{
		my $seq = $o_fasta->getSequence();
		$seq    = Sequence_utils::FormatSequenceLine( seq => $seq );
		print $fh_snp_chunk_file "$header\n$seq\n";
		++$seq_nb;

		if( 0 == $seq_nb % &CHUNK_SIZE )
		{
			$fh_snp_chunk_file->close();
			&IntegrateChunkToNarcisse( $chunk_path, $o_species, $rh_writers, $o_logger, $chunk_output );

			$chunk_nb++; $seq_nb = 0;
			$chunk_path    = &__GetChunkPath($chunk_nb,$snp_file);
			$chunk_output  = &__GetChunkPath($chunk_nb,$output) if ($output ne "");
			$fh_snp_chunk_file = new IO::File( $chunk_path, 'w' )
							or throw LipmError::IOException( $chunk_path, 'f' );
		}
	}
	# Integrate the last chunk (may be partial)
	$fh_snp_chunk_file->close();
	&IntegrateChunkToNarcisse( $chunk_path, $o_species, $rh_writers, $o_logger, $chunk_output );

	return;
}

# Some little function
# __GetChunkPath(1234,"/here/is/the/path.txt") returns:
# "/here/is/the/01234path.txt"
sub __GetChunkPath
{
	my ($chunk_nb,$path) = @_;
	my $dirname    = dirname  $path;
	my $basename   = basename $path;
	$basename      = sprintf ("%05i%s",$chunk_nb,$basename);
	return File::Spec->catfile( $dirname,$basename);
}

=head2 Procedure IntegrateChunkToNarcisse

 Usage     : IntegrateChunkToNarcisse( $snp_file, $o_species, $o_logger, $output )
 Procedure : Generate a part of the filesnp files, narcisse formatted
             We work in chunk to reduce the memory footprint of the program
 Args      :
	$snp_file   The file to integrate (may be compressed)
                Return with a warning if the file is not found
    $o_species  The species observer
    $rh_writers A hash with keys: chromosome names, values: NarcisseDataWriter objects
    $o_logger   A logger object
    $output     opt, path to the glint output raw file

=cut

sub IntegrateChunkToNarcisse
{
	my( $snp_file, $o_species, $rh_writers, $o_logger, $output ) = @_;

	&NARCISSE_utils::Unzip($snp_file);
	$snp_file =~ s/\.gz//;

	unless ( -r $snp_file )
	{
		my $msg = "WARNING - $snp_file NOT FOUND";
		warn $msg;
		$o_logger->Trace( $msg, '.','A');
		return;
	}
	
	my %h_options = ( -o_logger   => $o_logger,
						-huge_file => &TRUE );
	$h_options{ -raw_output } = $output if( defined $output && $output ne "" );

	# identification des *VRAIS* SNPs (au sens Narcisse)
	my $o_snp_int = New SequencesLocator::Snp(
										%h_options,
										file        => $snp_file,
										o_species   => $o_species );
	$o_snp_int->GenerateData();

	#####&NARCISSE_utils::Zip($snp_file);

	my $rh_data  = $o_snp_int->GetNarcissePrintableData();
	my $tmp_dir  = $o_species->GetTmpPath( 'INSTANCE', &FALSE );

	# recherche d'un effet des SNPs
	my $rh_snps = $o_snp_int->GetData();
	my @ao_snps = values %$rh_snps;

	# locate SNP on transcripts
	my $o_snp_man = New SnpManager( -o_logger => $o_logger,
									o_species => $o_species,
									rao_snps  => \@ao_snps );
	$o_snp_man->CheckForImpact();
# 			$o_snp_man->LocateSnps();

	# Writing data back to snp files -> Note, even if the snp file is related to a chromosome, it appears
	# that we have several snps found on other chromosomes. So that each snp source file may write to every snp destination file
	$rh_data     = $o_snp_man->GetSnpNarcissePrintableData();
	foreach my $chrom ( keys %$rh_data )
	{
		my $rh_snps = $$rh_data{ $chrom };
		my @a_keys = keys (%$rh_snps);
		my $comment = "writing " . scalar(@a_keys) . " data";
		$rh_writers -> { $chrom } -> Update ( $rh_snps, $comment );
	}

# 		# a revoir plus tard
#
# 			# on ecrit le fichier de SnpEffect
# 			my $rh_snps = $o_snp_man->GetSnpEffectNarcissePrintableData();
# 			my $filepath = $o_species->GetSnpEffectFiles( -chrom => $chrom, -check => &FALSE );
# 			&NARCISSE_utils::UpdateNarcisseData( $filepath, \@a_header, $rh_snps, &FALSE );

# 		# write second snp file

	return;
} #IntegrateChunkToNarcisse


=head2 Function DownloadFiles

 Usage        : $ra_files = &DownloadFiles( $o_io, $o_species, $o_logger, $verbose )
 Function     : Download Files from whatever refered to as snp_db in cfg file.
 Args         : $o_species, SpeciesObserver object.
                $o_param, $o_only the parameters
                $o_logger The logger
 return       : $ra_files, ref list of filepaths to the downloaded files
 Error        : none.

=cut

sub DownloadFiles
{
	my( $o_species, $o_param, $o_only, $o_logger, $verbose ) = @_;

	my $o_io = $o_species->GetInstanceObserver();
	my $download_dir = $o_species->GetDownloadPath();
	my $chromosomes = $o_param->Get('chromosomes');
	my @a_keys;
	if ( $chromosomes eq 'all' )
	{
		@a_keys = $o_param->GetKeys( '^FILE_SNP_' );
	}
	else
	{
		@a_keys = split( ',', $chromosomes );
		@a_keys = map 'FILE_SNP_'.$_, @a_keys;
	}
	my @a_files = $o_param->Get(@a_keys);
	my @a_paths;

	# Download the file
	if ($o_only->Get('download'))
	{
        $o_logger->StartTimer(&T_DOWNLOAD);
        $o_logger->Trace("STARTING DOWNLOAD\n",&T_DOWNLOAD,'A');

		for( my $i=0; $i<@a_files; ++$i )
		# foreach my $file (@a_files)
		{
			my $file = $a_files[ $i ];
			if( '' eq $file )
			{
				my $key = $a_keys[ $i ];
				my $msg = "No value found for key $key. Check your cfg.";
				warn $msg; next;
			}

			my $path = &NARCISSE_utils::DownloadData( $o_io, $file, $download_dir, $o_logger );
			push( @a_paths, $path );
			#&NARCISSE_utils::Zip($path);
		}
        $o_logger->StopTimer(&T_DOWNLOAD);
        $o_logger->Trace("DOWNLOAD FINISHED\n",&T_DOWNLOAD,'A');
	}

	# No download, just keep the file name
	else
	{
		foreach my $file (@a_files)
		{
			$file = basename $file;
			my $path = File::Spec->catfile( $download_dir, $file);
			push (@a_paths,$path);
		}
	}
	return \@a_paths;
}

=head2 procedure CheckFiles

 Title        : CheckFiles
 Usage        : &CheckFiles()
 Prerequisite : none
 procedure    : For each file, check if the file could be downloaded and print a message it if can't
 Args         : $o_param

=cut
sub CheckFiles
{
	my ($o_param) = @_;
	my $o_io = ConfigurationFactory::GetConfiguration();

	foreach my $key ($o_param->GetKeys('FILE_SNP_'))
	{
		my $file = $o_param->Get($key);
		my $rvl = &NARCISSE_utils::DownloadData($o_io,$file);
		warn("WARNING - COULD NOT CHECK THIS FILE ($file)") if ($rvl==-1);
		warn("ERROR   - THIS FILE DOES NOT EXIST ($file)")  if ($rvl==0);
		print "OK      - CHECKED $file\n"                   if ($rvl==1);
	}
	return;
}

=head2 function SetTargetLogPath

 Title        : SetTargetLogPath
 Usage        : $log_path = &SetTargetLogPath( $o_io, $o_param, $species );
 Function     : computes the target log path and creates the directory
                It also redirect STDERR and SDTOUT.
 Returns      : $log_path, a string
 Args         : $o_io, $o_param, $org

=cut

sub SetTargetLogPath
{
	my ( $o_io, $o_param, $species ) = @_;

	# Compute and print $target, and lock the target
	my $target = "getsnp_$species";
	print "target $target\n" . '-' x 30 . "\n";

    my $logs_dir = $o_param->IsDefined( 'sessdir') ? $o_io->GetLogsDirpath( 'SESSION') : $o_io->GetLogsDirpath( 'INSTANCE' );

	# change current dir to target dir
	#my $chrom_dir = $o_param->Get('chromosomes');
	#$chrom_dir =~ s/ //g;
	#$chrom_dir =~ s/,/_/g;
	my $target_log_root = File::Spec->catfile( $logs_dir,'getsnp',$species );

	# @a_created_dirs pour loguer plus tard.
	my @a_created_dirs = mkpath( $target_log_root );

	my $target_log_path = Sufine::MkDir(-dir => $target_log_root,-name => $target);
	die unless(defined $target_log_path);
	die unless(defined Sufine::ChDir(-dir => $target_log_root,-name => $target));

	&NARCISSE_utils::RedirectOE($o_io,"narcisse.out","narcisse.err");

	$o_param->Dump("narcisse.param");

	return $target_log_path;
}

sub Usage
{
	my ($o_param) = @_;
	#my $o_param = $o_wb->GetObjectParam();
	my $narcisse_cfg = $ENV{'NARCISSE_CFG'};
	my $exe = basename $0;

	print <<END

NARCISSE_CFG = $narcisse_cfg

Usage : $exe

mandatory:
    --cfg  : species configuration file or genome short name

optional:
    --chromosomes (comma-separated list) restrict the computation to those chromosomes (default=all)
    --raw_output file.out                If file.out does not exist: Keep the raw alignment data to this file
                                         If file.out already exists: Do not glint, consider the glint results are here
    --help             :   this message

SELECTING SOME ACTIONS:
    --only_xxx:            If some only_xxx switches are specified, ONLY those specific actions will be made
    --no_xxx:              If some no_xxx actions are reported, ONLY those specific actions will NOT be made
                           It is illegal specifying only_xxx AND no_xxx switches

    --only_check :         Check the data - If the check is positive, we exit from the program
    --no_check
    --check_files:         Only valid with --only_check: check the existence of each data file, thus the validity of the .cfg file
    --only_download        Download files
    --no_download
    --only_install         Install the snp data to the Narcisse data system
    --no_install
    --only_clean           Remove previous version of files
    --no_clean

END
;
exit;
}
__END__
