#!/usr/bin/perl
#
# $Id: nc_install_narcisse.pl 3855 2014-02-11 08:25:19Z  $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

BEGIN
{
		use File::Basename;
		my $dirprg=dirname($0);
		unshift @INC, "$dirprg/../lib";
		unshift @INC, "$dirprg/../API";
}

use strict;
use warnings;
#use ParamParser;
use NARCISSE_utils;
use Cwd;
use Error qw(:try);

#use constant UNIX_SVN => '/usr/bin/svn';
#use constant URL_SVN  => 'http://lipm-svn.toulouse.inra.fr/svn/narcisse';
use constant UNIX_RM  => '/bin/rm -rf';
use constant UNIX_TAR => '/bin/tar';
use constant UNIX_LN => '/bin/ln';
use constant UNIX_UNAME => '/bin/uname';
use constant UNIX_MAKE  => '/usr/bin/make';

use constant BLAST_RELEASE => '2.2.21';
use constant PARALOOP_RELEASE => '1.3-r247';

# Main program
MAIN:
{
    
    my $cur_dir = basename cwd;
    if ( $cur_dir ne 'Core' )
    {
        die "ERROR - You should start this program from the Core directory\n";
    }
    
    my ($os,$machine) = &GetMachine();
    print "Hello, we are on a $machine machine\n";
    print "===================================\n";
    
    &InstallBlast($os,$machine);
    
    my $unix_make = &UNIX_MAKE;
    
    print "\nNow compiling some programs\n";
    print "===================================\n";
    system($unix_make);
    
    my $paraloop_dir = 'paraloop';
    unless (-e $paraloop_dir)
    {
        print "Now installing paraloop\n";
        print "===================================\n";
        &InstallParaloop($os,$machine);
    }
    
    print "Now configuring paraloop\n";
    print "===================================\n";
    &ConfigureParaloop();
    
    print "Now Checking the existence of required Perl modules\n";
    print "============================\n";
    &CheckRequiredPerlModules();
}

sub GetMachine
{
    my $unix_uname = &UNIX_UNAME;
    my $os      = `$unix_uname -o`;
    chomp $os;
    if ($os ne 'GNU/Linux')
    {
        die "ERROR - system $os unsupported, I cannot install Narcisse automatically";
    }
    
    my $machine = `$unix_uname -m`;
    chomp $machine;
    
    if ($machine =~ /[36]86/)
    {
        die "ERROR - machine $machine unsupported, I cannot install Narcisse automatically";
        return ($os,'ia32');
    };
    if ($machine =~ /ia64/)
    {
        die "ERROR - machine $machine unsupported, I cannot install Narcisse automatically";
        return ($os,'ia64');
    };
    if ($machine =~ /x86_64/ || $machine =~ /amd.*64/)
    {
        return ($os,'x86_64');
    };
    die "ERROR - machine $machine unsupported, I cannot install Narcisse automatically";
}

sub InstallBlast
{
    my ($os, $machine) = (@_);
    
    # Compute the blast tar gz file
    my $m;
    $m = 'ia32' if ($machine eq 'ia32');
    $m = 'x64'  if ($machine eq 'x86_64');
    
    my $system;
    $system = 'linux' if ($os eq 'GNU/Linux');

    my $rel = &BLAST_RELEASE;
    
    my $blast_tgz = "blast-RELEASE-MACHINE-OS.tar.gz";
    $blast_tgz =~ s/RELEASE/$rel/;
    $blast_tgz =~ s/MACHINE/$m/;
    $blast_tgz =~ s/OS/$system/;
    my $blast_dir = "blast-$rel";

    my $unix_tar = &UNIX_TAR;
    my $unix_ln  = &UNIX_LN;
    my $unix_rm  = &UNIX_RM;
    
    if (! -e $blast_dir)
    {
        print "Now installing ncbi-blast from the file $blast_tgz\n";
        `$unix_tar xzf $blast_tgz`;
        
        print "Now creating a link for ncbi-blast\n";
        `$unix_rm ncbi-blast`;
        `$unix_ln -s $blast_dir ncbi-blast`;
    }
    else
    {
        print "Ncbi blast already installed - skipping blast installation\n";
    }
}  

sub InstallParaloop
{
    my ($os, $machine) = (@_);
    
    my $paraloop_dir = 'paraloop-1.3';
    my $paraloop_ln  = 'paraloop';
    my $paraloop_tgz = "paraloop-RELEASE.tgz";

    my $unix_tar = &UNIX_TAR;
    my $unix_ln  = &UNIX_LN;
    my $unix_rm  = &UNIX_RM;

    my $rel = &PARALOOP_RELEASE;
    $paraloop_tgz =~ s/RELEASE/$rel/;
    
    `$unix_tar xzf $paraloop_tgz`;
    `$unix_rm $paraloop_ln`;
    `$unix_ln -s $paraloop_dir $paraloop_ln`;
    
    print "done\n";
}    

sub ConfigureParaloop
{
    print "Do you want to (re) configure paraloop ?        Y/[N] ";
    if ( &NARCISSE_utils::GetAdmAnswer("","Y|N","N") eq 'Y' )
    {
        print "What does your installation look like (answer 1,2,3 or 4):\n";
        print "     -An SMP machine without any queuing system ?                                          (1) \n";
        print "     -A  cluster of computers without any queuing system ?                                 (2) \n";
        print "     -A  multiprocessor machine (SMP, cluster, etc) with the PBS queing system installed ? (3) \n";
        print "     -A  multiprocessor machine (SMP, cluster, etc) with the SGE queing system installed ? (4) \n";
        my $ask = &NARCISSE_utils::GetAdmAnswer("","1|2|3|4","1");
        my $cmd = "";
        if ($ask eq "1")
        {
            $cmd = "cp paraloop/etc/templates/paraloop.root.cfg_smp paraloop/etc/paraloop.root.cfg";
        }
        elsif ($ask eq "2")
        {
            $cmd = "cp paraloop/etc/templates/paraloop.root.cfg_rsh paraloop/etc/paraloop.root.cfg";
        }
        elsif ($ask eq '3')
        {
            $cmd = "cp paraloop/etc/templates/paraloop.root.cfg_PBS paraloop/etc/paraloop.root.cfg";
        }
        elsif ($ask eq '4')
        {
            $cmd = "cp paraloop/etc/templates/paraloop.root.cfg_SGE paraloop/etc/paraloop.root.cfg";
        }
        else
        {
            warn "Nothing configured";
        }
        print "Now executing: $cmd\n";
        `$cmd`;
        
        print "If anything works wrong, please have a look to the files:\n";
        print "   -paraloop/etc/paraloop.root.cfg\n";
        print "   -paraloop/etc/paraloop.cfg\n";
        print "and read the paraloop documentation\n";
        if ($ask eq "2")
        {
            print "YOU HAVE MORE THINGS TO CONFIGURE IN THE FILE paraloop/etc/paraloop.root.cfg\n";
        }
    }
}

sub CheckRequiredPerlModules
{
    my @a_bioperl_modules = qw( Bio::Annotation::DBLink Bio::LocatableSeq Bio::PrimarySeqI Bio::Root::Root
        Bio::Seq Bio::SeqFeature::FeaturePair Bio::SeqFeature::Generic
        Bio::SeqFeatureI Bio::SeqIO Bio::SimpleAlign Bio::Species Bio::Tools::CodonTable );

    my @a_bad_modules;
    foreach my $m (@a_bioperl_modules)
    {
        my $to_eval = "use $m";
        eval $to_eval;
        if ($@ ne "")
        {
            print "$m => $@\n";
            push(@a_bad_modules, $m);
        }
        else
        {
            print "$m => Success\n";
        }
    }
    
    if ($#a_bad_modules != -1)
    {
        print "ERROR - The following modules are NOT installed - Please install bioperl\n";
        print join (' ',@a_bad_modules) . "\n";
    }
    else
    {
        print "\n\nOK - BioPerl seems to be correctly installed\n";
    }
}

sub Usage
{
    print "\nnc_install_narcisse.pl\n";
    print "Install the Narcisse-related applications\n";
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

