#!/usr/bin/perl

=pod

=head1 NAME

 LaunchCircos.pl

=head1 SYNOPSIS

 perl LaunchCircos.pl @args

 @args:
  --reference : string, short species name to use as the reference.
  --targets   : list of string (comma separated), short species names to use as target.

=head1 DESCRIPTION

 # description

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 30-03, 2009
 $Id: LaunchCircos.pl 2556 2009-05-15 14:58:59Z sletort $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;
use Benchmark;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;
use Circos::Narcisse;
use SpeciesObserver;
use Configuration::InstanceObserver;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 2556 $ =~ /\d+/g); $r[0]};

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;

use constant REF => 'reference';
use constant TGT => 'targets';
use constant NB  => 'nb';

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{	warn( $o_err->Explain(), "\n" );	}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $o_log = New Logger( 'LaunchCircos.pl.log', 'IOV' );

	my $rh_params = &GetParameters();
# 	warn Dumper $rh_params;

	my $n = $$rh_params{ &NB };
	print "Building of Circos image $n time(s).\n";

	timethis( $n, sub{ ReBuildCircos( $rh_params, $o_log ) } );

	return;
}


=head2 ReBuildCircos

 Usage        : ReBuildCircos( $rh_params )
 Function     : .
 Args         :
     $rh_params
 Error        : none.

=cut

sub ReBuildCircos
{
	my( $rh_params, $o_log ) = @_;

	`rm -f *.png *.txt`;

	my $rh_orgs = __BuildRHOrgs( $rh_params, $o_log );

	my $o_inst_obs = InstanceObserver::Singleton();
	my $rh_cfg = $o_inst_obs->GetConfigHash();

	my %h_param = (
			circos_dir	=> $$rh_cfg{ ROOT_CIRCOS },
			data_dir	=> $$rh_cfg{ ROOT_DATA },
			image_size  => $$rh_cfg{ CIRCOS_DIM },
			nb_links_max=> $$rh_cfg{ CIRCOS_MAX_LINKS },
			synt        => $$rh_cfg{ SYNT },
			comparative => $$rh_cfg{ COMPARATIVE },
			output_dir	=> '.',
			level		=> 2,
			method		=> 'genomic',
# 			method		=> 'glint',
			rh_orgs		=> $rh_orgs,
			auto_hide	=> 1
# 			chroms_url	=> $args
		);

	my $o_circos = New Circos::Narcisse(%h_param, imagemap => &TRUE );
	$o_circos->BuildImage();

	return;
} #ReBuildCircos


=head2 __BuildRHOrgs

 Usage        : my $rh_orgs = __BuildRHOrgs( $rh_params )
 Function     : .
 Args         :
     $rh_params
 Return       : $rh_orgs
 Error        : none.

=cut

sub __BuildRHOrgs
{
	my( $rh_params, $o_log ) = @_;

	my $ref     = $$rh_params{ &REF };
	my $ref_key = &Circos::Narcisse::GetPrefixRef() . $ref ;

	# pour ne pas relire le fichier de conf plusieurs fois,
	# je cree un objet IO commun a tous les SpeciesObserver.
	my $o_inst_obs = InstanceObserver::Singleton( -o_logger => $o_log );
	my $o_ref   = New SpeciesObserver( species => $ref,
									-o_inst_obs => $o_inst_obs,
									-o_logger => $o_log );

	my %h_orgs;
	$h_orgs{ $ref_key } = { short => $ref,
							long => undef,
							chrs => $o_ref->GetChromsInfo()
						 };

	# la liste des clefs des orgs
	my @a_targets = split( ',', $$rh_params{ &TGT } );
	foreach my $target ( @a_targets )
	{
		my $o_tgt = New SpeciesObserver( species => $target,
										 -o_inst_obs => $o_inst_obs );

		$h_orgs{ $target } = {
            short => $target,
            long => undef,
            chrs => $o_tgt->GetChromsInfo()
        };
	}

	return \%h_orgs;
} #__BuildRHOrgs


=head2 GetParameters

 Usage        : my $rh_params = $o_x->GetParameters()
 Function     : Read parameters from command line and return them formatted as a hash ref.
 Returns      : $rh_params, hash ref with keys ...
 Args         : none

=cut

sub GetParameters
{
	ParamParser::SetDefaultUsage( sub { pod2usage(1) } );

	my @a_keys   = qw(  help );
	push( @a_keys, &REF . '=s', &TGT . '=s', &NB . '=i' );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );

	$o_param->SetUnlessDefined( &NB, 1 );

	my $rh_params = {};
	$o_param->Dump( 'HASH', $rh_params );

	return $rh_params;
}



__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
