#!/usr/bin/perl

=pod

=head1 NAME

 NarcisseDataReader.pl

=head1 SYNOPSIS

 See the Usage function or execute this program with the --help switch

 NarcisseDataReader.pl --help

=head1 DESCRIPTION
 The Narcisse object NarcisseDataReader is used to access the synteny files
 This program is thus a demo of this object, you may copy/paste some code from here to read Narcisse data an integrate them to your project

 TODO - Refaire le code pour benchmarker

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 07-04, 2009
 $Id$

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;
use Benchmark;
use File::Spec;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# our modules
use LipmError;
use General;
use ParamParser;
use Logger;

use Bio::Narcisse::Utils::NarcisseDataReader;
use ConfigurationFactory;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 2714 $ =~ /\d+/g); $r[0]};

MAIN:
{
	try
	{
		&Main();
	}
	catch Error with
	{
		my $o_err = shift;
		print "ERROR \n";
		print "      File: " . $o_err->file() . "\n";
		print "      Line: " . $o_err->line() . "\n";
		print "      Text: " . $o_err->text() . "\n";
	};
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	throw Error ( -text => "Please type chins some_instance before using this script") if (! defined $ENV{'NARCISSE_CFG'});
	my $o_io = ConfigurationFactory::GetConfiguration();
	throw Error ( -text => "ERROR - This script does not work in the instance multi")  if ($o_io->isa('InstanceObserver')==&FALSE);

	&ParamParser::SetDefaultUsage(\&Usage);

    my $o_param = New ParamParser('GETOPTLONG','reference:s','chromosome:s','target:s','level:i','start:s','end:s','columns:s','help');

    $o_param->AssertDefined('reference','chromosome','target','level');

# 	my $o_log = New Logger( 'NDR.pl.log', '' );

	#my $n = $o_param->GetInteger('NB');
	#print "Accessing data $n time(s).\n";

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

	my $rh_data = &GetSynteny( $o_ndr, $o_param );
	&PrintCsvResults($rh_data, $o_param);

	#my $n =1;
	#my $time = timethis( $n, sub{ GetSynteny( $o_io, $o_ndr, $o_param ) } );
# 	timethis( $n, sub{ GetGene( $rh_params ) } );

	return;
}


=head2 Function GetSynteny

 Usage   : GetSynteny( $o_ndr, $o_param )
 Function: Retrieve the wanted syntenies
 Return  : A ref to a hash containing the results
 Args    : $o_ndr a NarcisseDataReader object
           $o_param The parameters
=cut

sub GetSynteny
{
	my( $o_ndr, $o_param ) = @_;

	# The InstanceObserver is a singleton, so we are NOT creating a new object - Just retrieving the already created one
	my $o_io = ConfigurationFactory::GetConfiguration();

	my ($base_dir) = $o_io->GetComparativeDirpaths('INSTANCE');
	my( $ref, $chrom, $tgt, $level ) = $o_param->Get('reference','chromosome','target','level');
	my $o_so = $o_io->GetSpeciesObserver( $ref );
	my $file_path = $o_so->GetSyntFiles( -where => 'INSTANCE', target => $tgt, level => $level, prog => 'genomic', -chrom => $chrom );
	if ($o_param->IsDefined ('start') or $o_param->IsDefined ('end'))
	{
		my %h_where;
		my %h_range;
		if ($o_param->IsDefined ('start'))
		{
			my $start = $o_param->Get('start');
			$start = &ConvertToBase($start);
			$h_range{ 'start' } = $start;
		};
		if ($o_param->IsDefined ('end'))
		{
			my $end = $o_param->Get ('end');
			$end = &ConvertToBase($end);
			$h_range{ 'end' } = $end;
		};
		$h_where{ 'RANGE' } = \%h_range;
		$o_ndr->Prepare( -from => $file_path, -where => \%h_where );
	}
	else
	{
		$o_ndr->Prepare( -from => $file_path );
	}
	my $rh_data = $o_ndr->FetchallHashref( 'element' );
	return $rh_data;
}

=head2 Function ConvertToBase

 Usage   : $pos = &ConvertToBase(10mb);
 Function: Convert the input from kb or mb units to base pairs
 Return  : An integer
 Args    : $position_string: The position expressed as a string, number + (may be) unit
=cut
sub ConvertToBase
{
	my ($position_string) = @_;
	my $position = $position_string;
	my $global_unit = undef;
	if ($position_string =~ /([bmk])[bp]?$/i)
	{
		$global_unit = $1;
	}
    my ($position_nb, $position_unit) = ($position_string =~ /^([.\d]+)([bmk])?/i);
    $position_unit = $global_unit if (!defined $position_unit && defined $global_unit);
    $position = int($position_nb)      if ( defined $position_unit && $position_unit =~ /b/i );
    $position = $position_nb * 1000	   if ( defined $position_unit && $position_unit =~ /k/i );
    $position = $position_nb * 1000000 if ( defined $position_unit && $position_unit =~ /m/i );

	$position += 0;

	return $position;
}

=head2 Procedure PrintCvsResults

 Usage    : &PrintCvsResults( $rh_data, $o_param );
 Procedure: Print the results, cvs formatted (separator = TAB)
 Args     : $rh_data: The results to print as a hash of hashes
                key = the element number
                val = a hash
                    key = the column names
                    val = the values
            $o_param: The parameters (only column is used)
=cut
sub PrintCsvResults
{
	my ($rh_data, $o_param) = @_;
	my @a_keys    = sort { $a <=> $b } keys %$rh_data;

	if (@a_keys==0)
	{
		print "No data \n";
	}
	else
	{
		my @a_headers;
		if ($o_param->IsDefined('columns'))
		{
			# Note - we should check to be sure the wanted columns really exist
			@a_headers = split (',',$o_param->Get('columns'));
		}
		else
		{
			# Note - nothing to check here, but the order of the columns is alphabetic, which is not convenient
			my $rh_tmp    = $rh_data->{$a_keys[0]};
			@a_headers = sort keys %$rh_tmp;
		}
		print join ("\t",@a_headers) . "\n";

		foreach my $key (@a_keys)
		{
			my $rh_row = $rh_data->{$key};
			print join("\t",@$rh_row{@a_headers}) . "\n";
		}
	}
}

sub Usage
{
    my $prgm = basename $0;
    print <<END

Usage : $prgm
  REQUIRED:
	--reference  : string, short species name to use as the reference.
	--chromosome : string, chromosome to display.
	--target     : string, short species name to use as the target
	--level      : 0..4, the synteny level

  OPTIONAL:
    --start   : keep only syntenies starting AFTER this position (1000, 10kb, 4.5mb)
    --end     : keep only syntenies ending   BEFORE this position (1000, 10kb, 4.5mb)
    --columns : strings, list (coma separated) of columns to display (default= all columns)

  EXAMPLE:
    $prgm --reference human --target mouse --chrom 1 --level 4 --columns start,end,s_chrom,s_start,s_end --start 200m --end 210mb

END
;
}
