#!/usr/bin/perl

=pod

=head1 NAME

 NarcisseDataWriter.pl

=head1 SYNOPSIS

 See the Usage function or execute this program with the --help switch

 NarcisseDataWriter.pl --help

=head1 DESCRIPTION
 The NarcisseDataWriter object is used to write Narcisse formatted files
 This program is a demo and a test program
 First, a feature file is read with NarcisseDataReader, then the data are randomized in memory and written back to an output file,
 using chunks of 2 lines


=head1 Author

 emmanuel.courcelletoulouse.inra.fr

=head1 Version

 Created: 1-03-2010
 $Id$

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;
use Benchmark;
use File::Spec;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# our modules
use LipmError;
use General;
use ParamParser;
use Logger;
use NARCISSE_utils;

use Bio::Narcisse::Utils::NarcisseDataReader;
use NarcisseDataWriter;
use NarcisseDataWriter::OnDisk;
use NarcisseDataWriter::OnMemory;

use ConfigurationFactory;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 2714 $ =~ /\d+/g); $r[0]};

my %H_FEATURES = (
	gene       => 'GENE',
	rna        => 'RNA',
	cyto       => 'CYTO',
	exon       => 'EXON',
	ocean      => 'OCEAN',
	repeat     => 'REPEAT',
	transcript => 'TRANSCRIPT',
	private    => 'PERSONAL',
	snp        => 'SNP'
);

MAIN:
{
	try
	{
		&Main();
	}
	catch Error with
	{
		my $o_err = shift;
		print "ERROR \n";
		print "      File: " . $o_err->file() . "\n";
		print "      Line: " . $o_err->line() . "\n";
		print "      Text: " . $o_err->text() . "\n";
	};
	exit( 0 );
}

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	throw Error ( -text => "Please type chins some_instance before using this script") if (! defined $ENV{'NARCISSE_CFG'});
	my $o_io = ConfigurationFactory::GetConfiguration();
	throw Error ( -text => "ERROR - This script does not work in the instance multi")  if ($o_io->isa('InstanceObserver')==&FALSE);

	&ParamParser::SetDefaultUsage(\&Usage);

    my $o_param = New ParamParser('GETOPTLONG','species:s','chromosome:s','feature:s','start:s','end:s','columns:s','sessdir:s','output:s','help');

    $o_param->AssertDefined('feature','output');
    
    my @a_possible_features = keys %H_FEATURES;
    my $feature = $o_param->Get('feature');
    if ($feature eq '')
    {
		print "FEATURES MAY BE: " . join(' ', @a_possible_features ) . "\n";
		exit 0;
	}
	else
	{
		if (grep (/\b$feature\b/, @a_possible_features) == 0)
		{
			print "ERROR - Unknown feature - Features may be: " . join(' ', @a_possible_features ) . "\n";
		}
	}
    
    $o_param->AssertDefined('species','chromosome');
    if ($o_param->IsDefined('sessdir'))
    {
		$o_io->SetSessionDirectory($o_param->Get('sessdir'));
	};

# 	my $o_log = New Logger( 'NDR.pl.log', '' );

	#my $n = $o_param->GetInteger('NB');
	#print "Accessing data $n time(s).\n";

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

	my $ra_data = &GetFeature( $o_ndr, $o_param );
	
	# Shuffle data
	my $ra_data_sh = &ShuffleArray($ra_data);
	
	# Write the data back, inserting regularly some comments
	
	my $output_filename = $o_param->Get('output');
	my $ra_headers = $o_ndr->GetHeaders();
	
	
	# A lot of options to test:
	# Numerically sorting on start, end
	my $o_ndw = New NarcisseDataWriter::OnDisk($output_filename, $ra_headers);
	
	# alphanumeric sort on element
	#my $o_ndw = New NarcisseDataWriter::OnDisk($output_filename, $ra_headers, { sort_numerical => &FALSE ,sort_uniq => &TRUE}, [ 'element' ] );
	
	# no sort
	#my $o_ndw = New NarcisseDataWriter::OnDisk($output_filename, $ra_headers, { sort => &FALSE});
	
	# Same thing in memory
	# Numerically sorting on start, end
	#my $o_ndw = New NarcisseDataWriter::OnMemory($output_filename, $ra_headers);
	
	# alphanumeric sort on element
	# NOTE - dies if sort_unique == &TRUE !
	#my $o_ndw = New NarcisseDataWriter::OnMemory($output_filename, $ra_headers, { sort_numerical => &FALSE ,sort_uniq => &TRUE}, [ 'element' ] );
	#my $o_ndw = New NarcisseDataWriter::OnMemory($output_filename, $ra_headers, { sort_numerical => &FALSE ,sort_uniq => &FALSE}, [ 'element' ] );

	# no sort
	#my $o_ndw = New NarcisseDataWriter::OnMemory($output_filename, $ra_headers, { sort => &FALSE});
	
	for (my $i=0; $i < $#$ra_data_sh; $i += 2)
	{
		my %h_data;
		$h_data{'un'}   = $ra_data_sh->[$i];
		$h_data{'deux'} = $ra_data_sh->[$i+1] if ($i < $#$ra_data_sh);
		my $comment     = "chunk nb " . $i / 2 . "\n\n";
		$o_ndw->Update(\%h_data,$comment);
	}
	
	$o_ndw = undef;

	return;
}

=head2 Function ShuffleArray

 Usage   : my $ra_shuffled = ShuffleArray( $ra_src )
 Function: Shuffle the input array
 Return  : A ref to an array constaining the source, shuffled, array
 Args    : $ra_src Ref to array
 NOTE - Stolen to http://www.rocketaware.com/perl/perlfaq4/How_do_I_shuffle_an_array_random.htm
=cut
sub ShuffleArray
{
	my ($ra_src) = @_;
	my @a_dst;
	srand;
	
    while (@$ra_src) {
        push(@a_dst, splice(@$ra_src, rand @$ra_src, 1));
    }

	return \@a_dst;
}
	
=head2 Function GetFeature

 Usage   : GetFeature( $o_ndr, $o_param )
 Function: Retrieve the wanted features
 Return  : A ref to a hash containing the results
 Args    : $o_ndr a NarcisseDataReader object
           $o_param The parameters
=cut

sub GetFeature
{
	my( $o_ndr, $o_param ) = @_;

	# The InstanceObserver is a singleton, so we are NOT creating a new object - Just retrieving the already created one
	my $o_io = ConfigurationFactory::GetConfiguration();

	my( $species, $chrom, $feature ) = $o_param->Get('species','chromosome','feature');
	$feature = $H_FEATURES{ $feature };
	my $o_so = $o_io->GetSpeciesObserver( $species );
	
	my $file_path = $o_so->GetFeatureFiles( key => $feature, -chrom => $chrom );
	if ($o_param->IsDefined ('start') or $o_param->IsDefined ('end'))
	{
		my %h_where;
		my %h_range;
		if ($o_param->IsDefined ('start'))
		{
			my $start = $o_param->Get('start');
			$start = &ConvertToBase($start);
			$h_range{ 'start' } = $start;
		};
		if ($o_param->IsDefined ('end'))
		{
			my $end = $o_param->Get ('end');
			$end = &ConvertToBase($end);
			$h_range{ 'end' } = $end;
		};
		$h_where{ 'RANGE' } = \%h_range;
		$o_ndr->Prepare( -from => $file_path, -where => \%h_where );
	}
	else
	{
		$o_ndr->Prepare( -from => $file_path );
	}
	my $ra_data = $o_ndr->FetchallArrayref( );
	my $ras_data = &NARCISSE_utils::SortNarcisseDataOnStartEnd($ra_data);
	return $ras_data;
}

=head2 Function ConvertToBase

 Usage   : $pos = &ConvertToBase(10mb);
 Function: Convert the input from kb or mb units to base pairs
 Return  : An integer
 Args    : $position_string: The position expressed as a string, number + (may be) unit
=cut
sub ConvertToBase
{
	my ($position_string) = @_;
	my $position = $position_string;
	my $global_unit = undef;
	if ($position_string =~ /([bmk])[bp]?$/i)
	{
		$global_unit = $1;
	}
    my ($position_nb, $position_unit) = ($position_string =~ /^([.\d]+)([bmk])?/i);
    $position_unit = $global_unit if (!defined $position_unit && defined $global_unit);
    $position = int($position_nb)      if ( defined $position_unit && $position_unit =~ /b/i );
    $position = $position_nb * 1000	   if ( defined $position_unit && $position_unit =~ /k/i );
    $position = $position_nb * 1000000 if ( defined $position_unit && $position_unit =~ /m/i );

	$position += 0;

	return $position;
}

=head2 Procedure PrintCvsResults

 Usage    : &PrintCvsResults( $rh_data, $o_param );
 Procedure: Print the results, cvs formatted (separator = TAB)
 Args     : $rh_data: The results to print as a hash of hashes
                key = the element number
                val = a hash
                    key = the column names
                    val = the values
            $o_param: The parameters (only column is used)
=cut
sub PrintCsvResults
{
	my ($ra_data, $o_param) = @_;
	#my @a_keys    =  keys %$rh_data;

	# The InstanceObserver is a singleton, so we are NOT creating a new object - Just retrieving the already created one
	my $o_io = ConfigurationFactory::GetConfiguration();
	my $species = $o_param->Get('species');
	my $o_so = $o_io->GetSpeciesObserver( $species );
	my $species_short = $o_so->GetShortName();
	my $species_long  = $o_so->GetLongName();
	
	print "SPECIES = $species_long ($species_short)\n";
	
	if (@$ra_data==0)
	{
		print "No data \n";
	}
	else
	{
		my @a_headers;
		if ($o_param->IsDefined('columns'))
		{
			# Note - we should check to be sure the wanted columns really exist
			@a_headers = split (',',$o_param->Get('columns'));
		}
		else
		{
			# We reorder the columns: start,end, and the others, sorted
			my $rh_tmp    = $ra_data->[0];
			my @a_head_tmp = keys %$rh_tmp;
			if ( grep (/\bstart\b/,@a_head_tmp) == 1)
			{
				push(@a_headers,'start');
				@a_head_tmp = grep (!/\bstart\b/,@a_head_tmp)
			}
			if ( grep (/\bend\b/,@a_head_tmp) == 1)
			{
				push(@a_headers,'end');
				@a_head_tmp = grep (!/\bend\b/,@a_head_tmp)
			}
			push(@a_headers, sort(@a_head_tmp));
		}
		print join ("\t",@a_headers) . "\n";

		#foreach my $key (@a_keys)
		#{
		#	my $rh_row = $rh_data->{$key};
		#	print join("\t",@$rh_row{@a_headers}) . "\n";
		#}
		foreach my $rh_row (@$ra_data)
		{
			print join("\t",@$rh_row{@a_headers}) . "\n";
		}

	}
}

sub Usage
{
    my $prgm = basename $0;
    my $f_list = join(' ',keys(%H_FEATURES));
    
    print <<END

Usage : $prgm
  REQUIRED:
	--species    : string, short species name to use as the reference.
	--chromosome : string, chromosome to display.
	--feature    : string, feature to display (ie gene etc.)
	               Recognized features: $f_list
	--output     : the output file name (should be the same as the input file, with some comments added)

  OPTIONAL:
    --start   : keep only syntenies starting AFTER this position (1000, 10kb, 4.5mb)
    --end     : keep only syntenies ending   BEFORE this position (1000, 10kb, 4.5mb)
    --columns : strings, list (coma separated) of columns to display (default= all columns)
    --sessdir : try to search data in the session directory first

  EXAMPLE:
    $prgm --species human --chrom 1 --feature gene --columns start,end,s_chrom,s_start,s_end --start 200m --end 210mb

END
;
}
