#!/usr/bin/perl

=pod

=head1 NAME

 NarcisseService.pl

=head1 SYNOPSIS

 NarcisseService.pl --species <short_or_long_name> --chrom <chromosome> --service <service name> --data <data_for_this_service>


=head1 DESCRIPTION

 A very useful little program to debug and test the NarcisseServices
 See the Usage and the comments above each sub for the details

=head1 Author

 Sebastien.letort@toulouse.inra.fr
 emmanuel.courcelle@toulouse.inra.fr

=head1 Version

 Created: 29-05, 2009
 $Id: NarcisseService.pl 3765 2012-02-07 13:01:34Z manu $

=cut

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../API";

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;

use LipmError;
use General;
use ParamParser;
use Logger;
use NarcisseServiceFactory;
use NarcisseServiceKeywords qw( &NS_FASTA );

# les constantes
use constant VERSION => do {my @r = (q$Rev: 3765 $ =~ /\d+/g); $r[0]};
my @A_SERVICES = qw ( GetSlicesFromRegions GetGenesFromRegions GetCDSsFromRegions GetTranscriptsFromRegions GetProteinsFromRegions
                      GetGenesFromLocusTags GetTranscriptsFromLocusTags GetCDSsFromLocusTags GetGenesFromProteinIds 
                      GetDownstreamsFromLocusTags GetUpstreamsFromLocusTags GetProteinsFromLocusTags 
                      GetGenesFromTranscriptLocusTags GetGenesFromSymbols GetGenesFromAccessionNumbers );

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;
$Error::Debug = &DEBUG;

MAIN:
{
	&TryMain(1);	# WHEN DEBUGGING CODE 
	#&TryMain(0);	# FOR PRODUCTION CODE
};

=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	
	&ParamParser::SetDefaultBehaviour ( 'exit_on_getopt_error' );
	# 	my $o_log = New Logger( 'NarcisseService.log', 'V' );
	my $o_param  = New ParamParser( 'GETOPTLONG',\&Usage,"species=s", "chrom=s", "service=s", "data=s", "fasta", "help" ) ;
	
	$o_param -> AssertDefined ( 'species', 'service', 'data' );
	my $allowed_patterns = '^' . join('|',@A_SERVICES) . '$';
	$o_param -> AssertAllowedPattern ( 'service', $allowed_patterns );
	my $species = $o_param -> Get ( 'species' );
	my $service = $o_param -> Get ( 'service' );
	my $data    = $o_param -> Get ( 'data' );
	my $fasta   = $o_param -> IsDefined ( 'fasta' );
	
	# For some services you have to specify the --chroms parameters, for other services this is not useful
	if ( $service =~ /Regions$/ )
	{
		$o_param -> AssertDefined ( 'chrom' );
		my $chrom   = $o_param -> Get ( 'chrom' );
		&CallService1 ( $service, $species, $chrom, $data, $fasta );
	}
	else
	{
		&CallService2 ( $service, $species, $data, $fasta );
	}
}

=head2 Procedure CallService1

 Title        : CallService1
 Usage        : CallService1( $service, $species, $chrom, $data, $fasta );
 Procedure    : Call the Narcisse Service $service and print the results
 Args         : $service: The name of the service, generally GetSomethingFromRegions
                $species: a species SHORT NAME
                $chrom  : A chromosome label from this species
                $data   : several regions formatted as followes:
                          start:end[,start:end,...]
                $fasta  : If 1, print in fasta, else Dump internal data structures

=cut
sub CallService1
{
	my ( $service, $species, $chrom, $data, $fasta ) = @_;
	my @a_data = split ',', $data;
	my @a_regions;
	foreach my $d (@a_data)
	{
		my @a_d = split ( ':',$d);
		my %h_slice;
		$h_slice { 'start' } = $a_d[0];
		$h_slice { 'end'   } = $a_d[1];
		push @a_regions, \%h_slice;
	}
	my $ra_regions = \@a_regions;
	my $fct_name = 'NarcisseServiceFactory::' . $service;
	
	no strict;
    my $o_ns = &$fct_name ( species => $species, chrom => $chrom, ra_regions => \@a_regions, out_type  => &NS_FASTA );
    use strict;
    my ($ra_headers,$rh_data) = $o_ns -> FetchResults();
    if ( ! $fasta )
    {
	   	print Dumper $ra_headers;
		print Dumper $rh_data;
	}
	else
	{
		&PrintFasta ( $ra_headers, $rh_data );
	}
}

=head2 Procedure CallService2

 Title        : CallService2
 Usage        : CallService1( $service, $species, $data, $fasta );
 Procedure    : Call the Narcisse Service $service and print the results
 Args         : $service: The name of the service, generally GetSomethingFromLocusTags, or ProteinIds, etc.
                $species: a species SHORT NAME
                $data several regions formatted as followes:
                     start:end[,start:end,...]
                $fasta  : If 1, print in fasta, else Dump internal data structures

=cut
sub CallService2
{
	my ( $service, $species, $data, $fasta ) = @_;
	my @a_locus_tags = split ',', $data;

	my $fct_name = 'NarcisseServiceFactory::' . $service;
	
	no strict;
    my $o_ns = &$fct_name ( species => $species, ra_ids => \@a_locus_tags, out_type  => &NS_FASTA );
    use strict;
    my ($ra_headers,$rh_data) = $o_ns -> FetchResults();
    if ( ! $fasta )
    {
	   	print Dumper $ra_headers;
		print Dumper $rh_data;
	}
	else
	{
		&PrintFasta ( $ra_headers, $rh_data );
	}
}
	
=head2 Procedure PrintFasta

 Title        : PrintFasta
 Usage        : &PrintFasta ( $ra_headers, $rh_data );
 Procedure    : Print internal data to a stdout in a fasta format
 Args         : $ra_headers, $rh_data = The internal data structure to print

=cut
sub PrintFasta
{
	my ( $ra_headers, $rh_data ) = @_;
	foreach my $h ( @$ra_headers )
	{
		print "$h\n";
		my $data = $rh_data -> { $h };
		$data =~ s/(.{60})/$1\n/g;
		print "$data\n";
	}
}

sub Usage
{
	my $sep = "\n                               ";
	my $services = join $sep,@A_SERVICES;
    print <<END
    NarcisseService.pl --species <short_or_long_name> --chrom <chromosome> --service <service name> --data <data_for_this_service>\n
             --species A species SHORT NAME known by your Narcisse instance
             --chrom   A chromosome label from this species
             --data    Data to pass to the service
                            => For the "FromRegion" services, a list of regions 10000:11000,12000:13000
                            => For the other services, a list of locus_tags, ProteinIds, Symbols or AccesionNumbers
             --service One of: $services
END
;
}
