#!/usr/bin/perl

=pod

=head1 NAME

 SpeciesObserver

=head1 SYNOPSIS

 perl SpeciesObserver --ref short_name --target short_name --prog method_name
      method_name can be genomic, genetic, auto (or anything else if you have the correct method calculations)
      
=head1 DESCRIPTION

 Return a lot of file paths, testing the SpeciesObserver object

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 04-05, 2009
 $Id: SpeciesObserver.pl 2882 2009-09-21 09:02:49Z manu $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../lib";
	unshift @INC, "$dirprg/../API";
}

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;
use SpeciesObserver;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 2882 $ =~ /\d+/g); $r[0]};

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;
$Error::Debug = &DEBUG;

use constant SPECIES => 'ref';
use constant TARGET  => 'target';
use constant PROG    => 'prog';

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{
			warn( $o_err->Explain(), "\n" );
			warn $o_err->stacktrace(); 
		}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $o_log = New Logger( 'SpeciesObserver.log', 'IV' );

	my $rh_params = &GetParameters();
	my $species = $$rh_params{ &SPECIES } || 'bacc1';
	my $target  = $$rh_params{ &TARGET  } || 'bana';
    my $prog    = $$rh_params{ &PROG  } || 'genomic';

	my $o_inst_obs = New InstanceObserver();
    
    # Required for SpeciesObserver
	$o_inst_obs->SetSessionDirectory( 'WB00001' );
	my $o_spe_obs = New SpeciesObserver( species     => $species,
	                                     -o_inst_obs => $o_inst_obs );

# --  --  --  --  --  --  --
	print "Accessors :\n";
	&TestAccessors( $o_spe_obs );

	print "\nAccess to data files:\n";
	&TestDataFileAccess( $o_spe_obs, $target  );

	print "\nAccess to comparative data files:\n";
	&TestComparativeFileAccess( $o_spe_obs, $target, $prog );

	print "\nAccess to features files:\n";
	&TestFeaturesFileAccess( $o_spe_obs );

	return;
}


sub TestComparativeFileAccess
{
	my( $o_spe_obs, $target, $prog ) = @_;

    if ( $prog eq 'auto' )
    {
        my @a_methods = $o_spe_obs->GetMethods( target => $target );
        print "FOUND METHODS = " . join(' ',@a_methods) . "\n";
        $prog = $a_methods[0];
    }
	#my %h_args = ( target => $target, -chrom => '1', prog => $prog, level => 1, -check => &TRUE );
	my %h_args = ( target => $target, prog => $prog, level => 1, -check => &TRUE );

    my @synt  = $o_spe_obs->GetSyntFiles( %h_args );
	my $ortho = $o_spe_obs->GetOrthologyFiles( %h_args );
	my $synt  = $o_spe_obs->GetSyntFiles( %h_args );
	my $hsp   = $o_spe_obs->GetHspFiles(  %h_args, -check => &FALSE );

	my $res = "\tsynt  : $synt\n";
	$res   .= "\tortho : $ortho\n";
	$res   .= "\thsp   : $hsp\n";
	print $res;

	my $chrom = $o_spe_obs->GetInstanceObserver()->ExtractChromNameFromFilename( $ortho );
	print "chrom label extracted is $chrom\n";

	return;
}


sub TestFeaturesFileAccess
{
	my( $o_spe_obs ) = @_;

	my %h_arg = ( -chrom => '1', -check => &FALSE );

	my $gene   = $o_spe_obs->GetGeneFiles( %h_arg );
	my $exon   = $o_spe_obs->GetExonFiles( %h_arg );
	my $repeat = $o_spe_obs->GetRepeatFiles( %h_arg );
	my $transcript = $o_spe_obs->GetTranscriptFiles( %h_arg );
	my $perso  = $o_spe_obs->GetPersoFiles( %h_arg );

	my $res = "\tgene   : $gene\n";
	$res   .= "\texon   : $exon\n";
	$res   .= "\trepeat : $repeat\n";
	$res   .= "\tperso  : $perso\n";
	$res   .= "\ttranscript :$transcript\n";

	print $res;

	return;
}


sub TestDataFileAccess
{
	my( $o_spe_obs, $target ) = @_;

	my $karyo     = $o_spe_obs->GetKaryotypeFile();
	my $rh_chroms = $o_spe_obs->GetChromsInfo();
	my $root      = $o_spe_obs->GetDataPath();
# 	my $comp      = $o_spe_obs->GetComparativeDirpath( target => $target); no more accessible

	my $res = "\tkaryo : $karyo\n";
	$res   .= "\tles infos sur les chroms :\n";
	$res   .= Dumper $rh_chroms;
	$res   .= "\troots : $root\n";
# 	$res   .= "\tcomp  : $comp\n";

	print $res;

	return;
}


sub TestAccessors
{
	my( $o_spe_obs ) = @_;

	my $short = $o_spe_obs->GetShortName();
	my $long  = $o_spe_obs->GetLongName();
	my $taxid = $o_spe_obs->GetTaxId();
	my $where = $o_spe_obs->GetWhere();
	my $assembly = $o_spe_obs->GetAssembly();

	my $res = "\tshort name : $short\n";
	$res   .= "\tlong name  : $long\n";
	$res   .= "\ttaxid      : $taxid\n";
	$res   .= "\tassembly   : $assembly\n";
	$res   .= "\twhere      : $where\n";

	print $res;
# warn Dumper $o_spe_obs;
	return;
}


=head2 GetParameters

 Usage        : my $rh_params = $o_x->GetParameters()
 Function     : Read parameters from command line and return them formatted as a hash ref.
 Returns      : $rh_params, hash ref with keys ...
 Args         : none

=cut

sub GetParameters
{
	ParamParser::SetDefaultUsage( sub { pod2usage(1) } );

	my @a_keys   = qw(  help );
	push( @a_keys, &SPECIES . '=s', &TARGET . '=s', &PROG . '=s' );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );

	my $rh_params = {};
	$o_param->Dump( 'HASH', $rh_params );

	return $rh_params;
}



__END__

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
