#!/usr/bin/perl

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=pod

=head1 NAME

 ReBuildBacteriaExon -- RUSTINE

=head1 SYNOPSIS

 perl ReBuildBacteriaExon @args

 @args:
  --

=head1 DESCRIPTION

 # description

=head1 Author

 Sebastien.letort@toulouse.inra.fr

=head1 Version

 Created: 18-06, 2009
 $Id: ReBuildBacteriaExon.pl 2707 2009-06-18 14:38:46Z sletort $

=cut


use strict;
use warnings;

use Error qw( :try );
use Data::Dumper;
use Pod::Usage;

BEGIN
{	# pathway to our modules
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../../lib";
	unshift @INC, "$dirprg/../../API";
}

# nos modules, nos classes
use LipmError;
use General;
use ParamParser;
use Logger;

use Configuration::NarcisseObserver;
use Configuration::InstanceObserver;

# les constantes
use constant VERSION => do {my @r = (q$Rev: 2707 $ =~ /\d+/g); $r[0]};

use constant DEBUG => &TRUE;
# use constant DEBUG => &FALSE;
$Error::Debug = &DEBUG;

MAIN:
{
	try
	{
		&Main();
	}
	catch LipmError with
	{
		my ( $o_err, $r ) = @_;

		if( &TRUE == &DEBUG )
		{
			warn( $o_err->Explain(), "\n" );
			warn $o_err->stacktrace(); 
		}
		else
		{	print '*' . $o_err->GetUserMessage() . "*\n";	}
	}
	otherwise
	{
		my ( $o_err, $r ) = @_;

		my $str = $o_err->file;
		$str   .= ' line ' . $o_err->line;
		$str   .= "\n" . $o_err->text . "\n";

		print STDERR $str;
	};

	exit( 0 );
}


=head2 Procedure Main

 Title        : Main
 Usage        : &Main();
 Prerequisite : none
 Procedure    : the main function
 Args         : none

=cut

sub Main
{
	my $o_log = New Logger( 'ReBuildBacteriaExon.log', 'V' );

# 	my $rh_params = &GetParameters();

	my $o_nar_obs = New NarcisseObserver( -o_logger => $o_log );
	my $rao_inst_obs = $o_nar_obs->FindInstanceObservers( ra_instances => [ 'bact_3' ] );

	my $o_inst_obs = $$rao_inst_obs[0];

	# pour toutes les especes
	my $ra_species = $o_inst_obs->ExtractMySpecies( [] );
	foreach my $species ( @$ra_species )
	{
		my $o_species = $o_inst_obs->GetSpeciesObserver( $species );

		my $rh_chroms = $o_species->GetChromsInfo();
		foreach my $chrom ( keys %$rh_chroms )
		{
			my $gene_file = $o_species->GetGeneFiles( -chrom => $chrom );
			my $exon_file = $o_species->GetExonFiles( -chrom => $chrom, -check => &FALSE );

			&ReadGeneAndPrintExon( $gene_file, $exon_file );
		}

	}

	return;
}


=head2 GetParameters

 Usage        : my $rh_params = $o_x->GetParameters()
 Function     : Read parameters from command line and return them formatted as a hash ref.
 Returns      : $rh_params, hash ref with keys ...
 Args         : none

=cut

sub GetParameters
{
	ParamParser::SetDefaultUsage( sub { pod2usage(1) } );

	my @a_keys   = qw(  help );
#	push( @a_keys, );
	my $o_param  = New ParamParser( 'GETOPTLONG', @a_keys );

	my $rh_params = {};
	$o_param->Dump( 'HASH', $rh_params );

	return $rh_params;
}


sub ReadGeneAndPrintExon
{
	my( $gene_file, $exon_file ) = @_;

	my $o_ndr = New Bio::Narcisse::Utils::NarcisseDataReader();

	# lecture des genes
	$o_ndr->Prepare( -from => $gene_file );
	my $rh_genes = $o_ndr->FetchallHashref( 'element' );

	# pour chaque gene, associer un exon
	my %h_exons = ();
	my @a_keys = qw( chrom start end strand );
	foreach my $rh_gene ( values %$rh_genes )
	{
		my %h_exon = ( element => $$rh_gene{ element } . 'e1',
						symbol => $$rh_gene{ locus_tag } . '1',
						parent => $$rh_gene{ element } . 't1' );

		@h_exon{ @a_keys } = @$rh_gene{ @a_keys };

		$h_exons{ $h_exon{ element } } = \%h_exon;
	}

# 	warn Dumper scalar( keys %$rh_data );
# 	warn Dumper \%h_exons;

	# ecriture du fichier exon
	my $ra_headers = [ @a_keys, qw( element symbol parent ) ];
	NARCISSE_utils::PrintNarcisseData( $exon_file, $ra_headers, \%h_exons, { verbose => &DEBUG } );

	# et comme le trie n'est pas fait sur la bonne clef
	NARCISSE_utils::SortFile( $exon_file );
	return;
}

__END__
