#! /usr/bin/perl

use strict;
use warnings;
use JSON;
use Data::Dumper;
use constant START_URL => 'http://narcisse.toulouse.inra.fr/Beta';
use constant CURL_BIN  => 'curl -L -# ';

=pod
  DEMO script to execute a Narcisse request using the REST webservice
  Some DisplayMap images are produced and combined into one single image

  You may run this script on your workstation and interrogate our Narcisse server - or yours
  Please change the constant START_URL

  PREREQUISITE: The program curl SHOULD BE INSTALLED on your workstation
                The suite imagemagick should be installed on your workstation too

  NOTE - you should be able to modify this script to use wget or any other program instead of curl,
         or to use another image processing program to combine and display the downloaded images

=cut

MAIN:
{
    # Detect all the instances
    my $rh_data;
    my ($rh_output) = &ExeRequest(&START_URL);

    #print Dumper($rh_output);

    # For the FIRST instance returned, create a new session and find minimal information about some species
    my @a_inst = keys %$rh_output;
    my $inst   = $a_inst[0];

    # Query the instance base url, creating a session and returning the list of species
    my $url                       = $rh_output->{$inst}->{instance_url};
    my ($rh_inst_output,$session) = &ExeRequest($url);
    my @a_species                 = keys %{$rh_inst_output->{species}};

    # Use the first two species as target/reference
    my $reference    = $a_species[0];
    my @a_chroms_ref = keys %{$rh_inst_output->{species}->{$reference}->{chroms_info}};
    my $chrom_ref    = $a_chroms_ref[0];

    my $target1      = $a_species[1];
    #print "INSTANCE $inst reference $reference chromosome $chrom_ref target1 $target1\n";
    #print Dumper($rh_inst_output);

    # Request a comparison, using a Narcisse map representation
    # We ask two requests: the first for the core features, the second for the supp features
    # NOTE - The default value for 'feature' is 'Core', so if you wan ONLY the Core features you do not need specifying the features parameter
    my ($rh_core) = &ExeRequest($url,$session,'script', 'DISPLAY', 'reference', $reference, 'chroms_reference', $chrom_ref,'target1',$target1,'features', 'Core');
    my ($rh_supp) = &ExeRequest($url,$session,'script', 'DISPLAY', 'reference', $reference, 'chroms_reference', $chrom_ref,'target1',$target1,'features', 'Supp');

    # Merge both hashes
    my %h_features = (%$rh_core,%$rh_supp);
    # warn Dumper \%h_features;

    # Download the images and keep track of the saved file names
    my @a_file_names;
    foreach my $k (keys %h_features)
    {
        my $out_file = $k . '.png';
        push @a_file_names,$out_file;
        &DownloadFile($h_features{$k},$out_file);
    }

    # Create a composite image from the downloaded ones
    my $image_file = 'narcisse.png';
    &CreateGlobalImage(\@a_file_names,$image_file);

    # Display the final image (works ONLY if display from image magisk is installed)
    `display -background white -flatten $image_file`;
}

######################################################
#
#   SUBROUTINES
#
######################################################

# CreateGlobalImage from several images
# Parameters:
#     a ref to an array of file names
#     The created file name
# Return: none
sub CreateGlobalImage
{
    my ($ra_file_names,$global_out_file) = @_;

    # THE FOLLOWING WILL WORK ONLY IF YOU HAVE ImageMagick LOCALLY INSTALLED
    my $flg_1 = 1;
    foreach my $f (@$ra_file_names)
    {
        if ($flg_1==1)
        {
            my $cmd = "cp $f $global_out_file";
            #warn $cmd;
            `$cmd`;
            $flg_1 = 0;
        }
        else
        {
            my $cmd = "composite $f $global_out_file $global_out_file";
            #warn $cmd;
            `$cmd`;
        }
    }

    return $global_out_file;
}

# ExeRequest using curl from the Narcisse server
# Parameters:
#        $url The url to use
#        $session (optional) The session, a new session will be created on the server if not specified
#        @a_params A list of parameters: qw( name value name value name value ... )
#
# Return: An array of 3 elements:
#         - The data (generally a ref to a hash)
#         - The session
#         - The parameters
#
sub ExeRequest
{
    my ($url,$session,@a_params) = @_;
    $session = '' unless defined $session;

    my $parameters = '__wb_rest=1';
    for (my $i =0; $i < $#a_params; $i += 2)
    {
        $parameters .= '&' . $a_params[$i] . '=' . $a_params[$i+1];
    }
    if ($session ne '')
    {
        $parameters .= '&__wb_session='.$session;
    }
    my $cmd = &CURL_BIN . ' ' . '\''. $url . '?' . $parameters . '\'';
    print "$cmd\n";
    my $rvl = `$cmd`;
    #warn $rvl;
    my $json  = new JSON;
    my $rh_data= eval { $json->decode($rvl); };
    if ($@)
    {
        warn "$@\n";
        die  "$rvl\n";
    }

    #warn Dumper $rh_data;
    if (exists $rh_data->{error})
    {
        print Dumper ($rh_data->{error}) . "\n";
        die "ERROR - PROGRAM INTERRUPTED";
    }

    return ($rh_data->{data},$rh_data->{__wb_session},$rh_data->{params});
}

# Downloadfile using curl from the Narcisse server
# Parameters:
#        $url The url of the file to download
#        $out_file The name of the created file
#
# Return: none
#
sub DownloadFile
{
    my ($url,$out_file) = @_;
    my $cmd   = &CURL_BIN . " -o $out_file $url";
    print "$cmd\n";
    `$cmd`;
}
