#! /usr/bin/perl

use strict;
use warnings;
use JSON;
use Data::Dumper;
use constant START_URL => 'http://narcisse.toulouse.inra.fr/Beta';
#use constant START_URL => 'http://localhost/narcisse/2_1';
use constant CURL_BIN  => 'curl -L -# ';

=pod
  DEMO script to execute a Narcisse request using the REST webservice
  For some species, a karyotype image is produced

  You may run this script on your workstation and interrogate our Narcisse server - or yours
  Please change the constant START_URL

  PREREQUISITE: The program curl SHOULD BE INSTALLED on your workstation
                The suite imagemagick should be installed on your workstation too

  NOTE - you should be able to modify this script to use wget or any other program instead of curl,
         or to use another image processing program to combine and display the downloaded images

=cut

MAIN:
{
    # Detect all the instances
    my $rh_data;
    my ($rh_output) = &ExeRequest(&START_URL);

    #print Dumper($rh_output);

    # For the FIRST instance returned, create a new session and find minimal information about the species
    my @a_inst = keys %$rh_output;
    my $inst   = $a_inst[0];

    # Query the instance base url, creating a session and returning the list of species
    my $url                       = $rh_output->{$inst}->{instance_url};
    my ($rh_inst_output,$session) = &ExeRequest($url);
    my @a_species                 = keys %{$rh_inst_output->{species}};

    my $reference = $a_species[0];

    # Ask the use a gene name belonging to this species
    print "Please enter a gene name / locus tag of a gene belonging to $reference\n";
    my $gene = <STDIN>;
    chomp $gene;

    # Request the urls to download
    my ($rh_seq_files) = &ExeRequest($url,$session,'script','SENDSEQUENCES','species',$reference,'GeneSeq',$gene);

    # Download the first
    my $out_file = $gene . '.fasta';
    my $object = "";
    foreach my $obj (keys %$rh_seq_files)
    {
        my $url = $rh_seq_files->{$obj};
        &DownloadFile($url,$out_file);
        $object = $obj;
        last;   #We download only the first file
    }

    print "Done - The file for $object is $out_file\n";
}


######################################################
#
#   SUBROUTINES
#
######################################################


# ExeRequest using curl from the Narcisse server
# Parameters:
#        $url The url to use
#        $session (optional) The session, a new session will be created on the server if not specified
#        @a_params A list of parameters: qw( name value name value name value ... )
#
# Return: An array of two elements:
#         - The data (generally a ref to a hash)
#         - The session
#
sub ExeRequest
{
    my ($url,$session,@a_params) = @_;
    $session = '' unless defined $session;

    my $parameters = '__wb_rest=1';
    for (my $i =0; $i < $#a_params; $i += 2)
    {
        $parameters .= '&' . $a_params[$i] . '=' . $a_params[$i+1];
    }
    if ($session ne '')
    {
        $parameters .= '&__wb_session='.$session;
    }
    my $cmd = &CURL_BIN . ' ' . '\''. $url . '?' . $parameters . '\'';
    print "$cmd\n";
    my $rvl = `$cmd`;
    #warn $rvl;
    my $json  = new JSON;
    my $rh_data= eval { $json->decode($rvl); };
    if ($@)
    {
        warn "$@\n";
        die  "$rvl\n";
    }

    #warn Dumper $rh_data;
    if (exists $rh_data->{error})
    {
        print Dumper ($rh_data->{error}) . "\n";
        die "ERROR - PROGRAM INTERRUPTED";
    }

    return ($rh_data->{data},$rh_data->{__wb_session});
}

# Downloadfile using curl from the Narcisse server
# Parameters:
#        $url The url of the file to download
#        $out_file The name of the created file
#
# Return: none
#
sub DownloadFile
{
    my ($url,$out_file) = @_;
    my $cmd   = &CURL_BIN . " -o $out_file \'$url\'";
    print "$cmd\n";
    `$cmd`;
}
