#! /usr/bin/perl

use strict;
use warnings;
use JSON;
use Data::Dumper;
use constant START_URL => 'http://narcisse.toulouse.inra.fr/Beta';
use constant CURL_BIN  => 'curl -L -# ';

=pod
  DEMO script to execute a Narcisse request using the REST webservice
  For some species, a karyotype image is produced

  You may run this script on your workstation and interrogate our Narcisse server - or yours
  Please change the constant START_URL

  PREREQUISITE: The program curl SHOULD BE INSTALLED on your workstation
                The suite imagemagick should be installed on your workstation too

  NOTE - you should be able to modify this script to use wget or any other program instead of curl,
         or to use another image processing program to combine and display the downloaded images

=cut

MAIN:
{
    # Detect all the instances
    my $rh_data;
    my ($rh_output) = &ExeRequest(&START_URL);

    #print Dumper($rh_output);

    # For the FIRST instance returned, create a new session and find minimal information about the species
    my @a_inst = keys %$rh_output;
    my $inst   = $a_inst[0];

    # Query the instance base url, creating a session and returning the list of species
    my $url                       = $rh_output->{$inst}->{instance_url};
    my ($rh_inst_output,$session) = &ExeRequest($url);
    my @a_species                 = keys %{$rh_inst_output->{species}};

	# Request a karyotype for the FIST species
    my $sp = $a_species[0];
	my ($rh_karyo) = &ExeRequest($url,$session,'script', 'KARYOTYPE', 'reference', $sp);

	# Download the .png
	my $karyo_url = $rh_karyo->{karyotype};
	my $out_file  = $sp . '.png';
	&DownloadFile($karyo_url,$out_file);

    # Display the final image (works ONLY if display from image magick is installed)
    my $cmd = 'display -background white -flatten ' . $out_file;
    `$cmd`;
}

######################################################
#
#   SUBROUTINES
#
######################################################

# CreateGlobalImage from several images
# Parameters:
#     a ref to an array of file names
#     The created file name
# Return: none
sub CreateGlobalImage
{
    my ($ra_file_names,$global_out_file) = @_;

    # THE FOLLOWING WILL WORK ONLY IF YOU HAVE ImageMagick LOCALLY INSTALLED
    my $flg_1 = 1;
    foreach my $f (@$ra_file_names)
    {
        if ($flg_1==1)
        {
            my $cmd = "cp $f $global_out_file";
            #warn $cmd;
            `$cmd`;
            $flg_1 = 0;
        }
        else
        {
            my $cmd = "composite $f $global_out_file $global_out_file";
            #warn $cmd;
            `$cmd`;
        }
    }

    return $global_out_file;
}

# ExeRequest using curl from the Narcisse server
# Parameters:
#        $url The url to use
#        $session (optional) The session, a new session will be created on the server if not specified
#        @a_params A list of parameters: qw( name value name value name value ... )
#
# Return: An array of two elements:
#         - The data (generally a ref to a hash)
#         - The session
#
sub ExeRequest
{
    my ($url,$session,@a_params) = @_;
    $session = '' unless defined $session;

    my $parameters = '__wb_rest=1';
    for (my $i =0; $i < $#a_params; $i += 2)
    {
        $parameters .= '&' . $a_params[$i] . '=' . $a_params[$i+1];
    }
    if ($session ne '')
    {
        $parameters .= '&__wb_session='.$session;
    }
    my $cmd = &CURL_BIN . ' ' . '\''. $url . '?' . $parameters . '\'';
    print "$cmd\n";
    my $rvl = `$cmd`;
    #warn $rvl;
    my $json  = new JSON;
    my $rh_data= eval { $json->decode($rvl); };
    if ($@)
    {
        warn "$@\n";
        die  "$rvl\n";
    }

    #warn Dumper $rh_data;
    if (exists $rh_data->{error})
    {
        print Dumper ($rh_data->{error}) . "\n";
        die "ERROR - PROGRAM INTERRUPTED";
    }

    return ($rh_data->{data},$rh_data->{__wb_session});
}

# Downloadfile using curl from the Narcisse server
# Parameters:
#        $url The url of the file to download
#        $out_file The name of the created file
#
# Return: none
#
sub DownloadFile
{
    my ($url,$out_file) = @_;
    my $cmd   = &CURL_BIN . " -o $out_file $url";
    print "$cmd\n";
    `$cmd`;
}
