/*
  Copyright INRA/CNRS

 Emmanuel.Courcelle@toulouse.inra.fr
 Jerome.Gouzy@toulouse.inra.fr
 Thomas.Faraut@toulouse.inra.fr

 This software is a computer program whose purpose is to provide a
 web-based interface for analyzing the different levels of genome
 conservations.

 This software is governed by the CeCILL license under French law and
 abiding by the rules of distribution of free software.  You can  use,
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 "http://www.cecill.info".

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability.

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or
 data to be ensured and,  more generally, to use and operate it in the
 same conditions as regards security.

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms

*/

/*

=pod

=head1 NAME
 narcisse_area.js - Objects useful for clickable areas in Narcisse

=head1 SYNOPSIS

 DynamicMarkerClass     A class to color in light grey an area
 NarcisseAreaClass      A base class for clickable areas management

=cut
*/

/*
=pod

=head2 DynamicMarkerClass

 Title:  DynamicMarkerClass
 Usage:  var o_marker = new DynamicMarkerClass('ID_DYNAMIC_MARKER');
 Class:  This class defines a tool to color in grey a little area
 Constructor Args:
 elt:    the id of the div to be used as dynamic marker
         WARNING - We keep the Id here, because $(Id) is changed at each ajax request !

=cut
*/
DynamicMarkerClass = Class.create({
	initialize: function(elt)
	{
		this.__elt = elt;
	},

	/*
	Title:     NewPosition
	Usage:     NewPosition(100,200,110,205)
	Procedure: Changes to position/width/height of the div #ID_DYNAMIC_MARKER
	Access:    private
	Args:      The coord of the new position
	Globals:   none
	*/
	NewPosition: function(x1,y1,x2,y2)
	{
		var elt = $(this.__elt);
		if (elt != null)
		{
			var width = x2-x1+1;
			var height= y2-y1+1;

			elt.setStyle({
				left: x1+'px',
				top:y1+'px',
				width:width+'px',
				height:height+'px'
			});
			elt.show();
		}
	},
	/* hide the marker */
	Hide: function() {$(this.__elt).hide();}
});


/*
=pod

=head2 NarcisseAreaClass

 Title:     Narcisse Area class
 Usage:     Base class for some objects
 Class:     Create an area element and declare some default event handlers on this element
 Constructor Arg: a_graphics_objects, an array containing:
               legend:           A div - We keep the Id, because $(legend) may change
               o_control_block   A Controlblock object
               o_popup:          A DynamicPopupClass  object
               o_marker:         A DynamicMarkerClass object
               o_dessus_dessous: A DessusDessousClass object
               o_marker2:        A second DynamicMarkerClass object (facultatif)
               NOTE - It is easier to pass an array than a lot of individual parameters
Globals:   O_CFG, O_PARAM

=cut
*/
NarcisseAreaClass = Class.create({
	initialize: function(a_graphics_objects)
	{
		// this._legend and this._o_marker are protected objects
		this._legend            = a_graphics_objects[0];
		this._o_control_block   = a_graphics_objects[1];
		this._o_popup           = a_graphics_objects[2];
		this.__o_marker         = a_graphics_objects[3];
		this.__o_dessus_dessous = a_graphics_objects[4];
		this._o_marker2         = a_graphics_objects[5];
		this._base_url          = O_CFG['HTTP_NARCISSE_CGI'] + "?__wb_session="+O_PARAM["__wb_session"];
		this._feature           = '';     // Will be filled by the derived classes
	},

	/*
=pod

=head2 MouseOver

 Title:     MouseOver
 Usage:     through a mouseover event handler, called by the MouseOver functions of the derived objects
 Function:  Manage the display property of some elements, marks the selected element and may be the secondary selected element
            If the protected member this._feature is NOT on, return false
            Else, return true
 Args:      x1,y1,x2,y2     required: the coordinates of the area
            xd1,yd1,xd2,yd2 optional: the coordinates of a secondary marker
 Returns:   a boolean - If false, nothing was made
 Globals:   none

=cut
	*/
	MouseOver: function(x1,y1,x2,y2,xd1,yd1,xd2,yd2)
	{
		// Nothing If the feature is unchecked we return false
		if ( this._feature != undefined && this._o_control_block.IsFeatureOn(this._feature) == false )
		{
			return false;
		}
		else
		{

			// If necessary, we hide the active tab of the control box - See DessusDessous for the details
			this.__o_dessus_dessous.LegendIsUp();
	
			// Display the dynamic marker (if the coordinates are provided)
			if (x1 != x2)
			{
				this.__o_marker.NewPosition(x1,y1,x2,y2);
	
				// display the secondary dynamic marker, if necessary
				if ( this._o_marker2 != undefined )
				{
					if ( xd1 != undefined )
					{
						this._o_marker2.NewPosition(xd1,yd1,xd2,yd2);
					}
					else
					{
						this._o_marker2.Hide();
					}
				}
			}

		return true;

		}
	},
	/*
=pod

=head2 MouseDown

 Title:     MouseDown
 Usage:     through a mousedown event handler, called by the MouseDown functions of the derived objects
 Function:  If the protected member this._feature is NOT on, return false, else return true
 Args:      None
 Returns:   a boolean
 Globals:   none

=cut
	*/
	MouseDown: function()
	{
		return this._o_control_block.IsFeatureOn(this._feature);
	},

	/*
=pod

=head2 _GetTitle

 Title:     _GetTitle
 Usage:     var title = o_obj._GetTitle();
 Function:  Return a title for this area element - Used by controlblock when the area is created
 Args:      None
 Returns:   The title
 Globals:   none

=cut
	*/
	_GetTitle: function(){return '';}
});

/*
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license under French law - www.cecill.info

=cut
*/

