
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

/*
#########################################
#
# programme pour tester les locks distribues d'un systeme de fichier
#
# ce programme entre dans une boucle infinie, et pour chaque iteration:
#   - un compteur est incremente dans un fichier (test-lock.txt)
#   - des informations sont ecrites dans test-lock.log
#
# si ca marche, le compteur doit s'incrementer gentiment meme si on fait tourner
# le programme sur plusieurs noeuds - Preuve que le FS gere correctement les locks
#
# sinon, on a des messages d'erreur dans le fichier log, ou alors le compteur ne s'incremente pas correctement, etc.
#
# pour essayer SANS VERROUILLER, mettre le define LOCK_ACTIF a 0
# pour arreter tous les programmes lances, touch stop dans le repertoire de lock
# pour compiler,  gcc -pedantic -o test-lock test-lock.c
#
# ANALYSE DES LOGS:
#      grep WHILE test-lock.log    donne la liste des pids/machines qui ont ete lances, puis arretes
#      grep PROBLEME test-lock.log donne les lignes qui signalent un probleme de lock
#      grep -wc 1234 test-lock.log donne le nombre de lignes concernant un pid, permet de voir si l'acces au repertoire etait equilibre entre les processes
#
###########################################
*/


#define LOCK_ACTIF 1	    /* mettre 1 pour activer les locks */
#define WAIT_MAX   1000     /* mettre a 0 pour desactiver le temps d'attente aleatoire
                               si N (>0), attend entre 0 et N-1 us entre chaque iteration */

							   /* en cas d'erreur de compilation ligne 222, decommenter cette ligne */
/* #define useconds_t unsigned long */

/* Trace = Ecrire une trace dans le fichier log
   arg  : log_file Nom du fichier log
          stamp    le stamp a ecrire en debut de ligne
          msg      le msg a ecrire a droite du stamp
                   On fait un free sur msg avant de revenir, de sorte que le char * n'est PLUS utilisable par la suite
	      flag     Si 1, fait un free de msg
		           ==> si msg est obtenu par un malloc, employer 1
				       si msg est une constante ("un message"), employer 0
*/
void Trace(const char* log_file, const char* stamp, char* msg, int flag)
{
	int log_fd, rvl;
	char* buf;
	
	/* NOTE - from the open man page: 
	The file is opened in append mode.  Before each write(2), the file offset is positioned at the end of the file, as if with lseek(2).  O_APPEND may lead to corrupted files on NFS
    file systems if more than one process appends data to a file at once.  This is because NFS does not support appending to a file, so the client kernel has to simulate  it,  which
    can’t be done without a race condition.
	J'en conclus que si on a corruption du fichier log, c'est une indication que lock n'a pas fonctionne correctement
	*/
	log_fd = open(log_file,O_WRONLY|O_APPEND|O_CREAT,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (log_fd==-1)
	{
		printf ("Fichier log = %s\n",log_file);
		perror("ERREUR lors de l'ouverture du fichier log");
		exit(1);
	}
	
	buf = (char *) malloc(1000);
	strncpy(buf, stamp, 50);
	strcat (buf,"   ");
	strncat(buf,msg,960);
	strcat (buf,"\n");
	
	rvl = write(log_fd,buf,strlen(buf));
	if (rvl==-1)
	{
		perror("ERREUR lors de l'ecriture d'une ligne dans le fichier de log");
		exit(1);
	}
	
	/* fermeture du fichier et liberation de la memoire */
	rvl = close(log_fd);
	if (rvl==-1)
	{
		perror("ERREUR lors de la fermeture du fichier de log");
		exit(1);
	}
	
	if (flag)
		free (msg);
	
	free (buf);
	
	return;
}

/* ReadCpt = Lit et renvoie la valeur du compteur
             Ecrit une ligne dans le fichier log
   arg    : cpt_file le fichier de compteur
            log_file le fichier log
            stamp    le stamp
   return : la valeur du compteur, ou -1 si erreur
*/
int ReadCpt(const char* cpt_file, const char* log_file, const char* stamp)
{
	int cpt_fd, rvl;
	char* buf;
	
	cpt_fd = open(cpt_file,O_RDONLY);
	if (cpt_fd==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de l'ouverture du fichier cpt (O_RDONLY) ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,1);
		return -1;
	}
	
	buf = calloc(50,1);
	rvl = read(cpt_fd, buf, 49);
	if (rvl==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de la lecture du fichier cpt ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,0);
	}
	else
	{
		char *msg = malloc(500);
		rvl = atoi(buf);
		strcat(buf,"");
		strcpy(msg, "LECTURE  DU COMPTEUR: ");
		strncat(msg,buf,20);
		Trace(log_file,stamp,msg,1);
	};
	if (rvl==0)
	{
		char *msg = malloc(500);
		strcpy(msg, "PROBLEME DE LOCK - COMPTEUR REMIS A 0");
		strncat(msg,buf,20);
		Trace(log_file,stamp,msg,1);
	};
	
	free(buf);
	
	if (close(cpt_fd)==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de la fermeture du fichier cpt (O_RDONLY) ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,1);
		return -1;
	}
	
	return rvl;
}
	
/* WriteCpt = Ecrit la valeur du compteur
              Ecrit une ligne dans le fichier log
   arg    : cpt_file le fichier de compteur
            cpt le   compteur
            log_file le fichier log
            stamp    le stamp
   return : 0 si OK ou -1 si erreur
*/
int WriteCpt(const char* cpt_file, int cpt, const char* log_file, const char* stamp)
{
	int cpt_fd, rvl;
	char* buf;
	
	cpt_fd = open(cpt_file,O_WRONLY|O_TRUNC|O_CREAT,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (cpt_fd==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de l'ouverture du fichier cpt (O_WRONLY) ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,1);
		return -1;
	}
	
	buf = malloc(20);
	snprintf(buf,20,"%15i",cpt);
	rvl = write(cpt_fd,buf,strlen(buf));
	if (rvl==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de l'ecriture du fichier cpt ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,1);
	}
	else
	{
		char *msg = malloc(500);
		rvl=0;
		strcat(buf,"");
		strcpy(msg, "ECRITURE DU COMPTEUR: ");
		strncat(msg,buf,20);
		Trace(log_file,stamp,msg,1);
	};
	
	free(buf);
	
	rvl=close(cpt_fd);
	if (rvl==-1)
	{
		char * msg = malloc(500);
		strcpy(msg, "ERREUR lors de la fermeture du fichier cpt (O_WRONLY) ");
		strcat(msg, strerror(errno));
		Trace(log_file, stamp, msg,1);
	}
	
	return rvl;
}
	
/* WaitRand = on attend (par une boucle d'attente) un temps aleatoire */
void WaitRand()
{
	useconds_t sleep_tm;

	if (WAIT_MAX==0)
		return;
	else
		sleep_tm = random() % WAIT_MAX;
	/*printf ("value = %i\n",value);*/
	usleep(sleep_tm);
}

/* LockDir = On verrouille le repertoire 
   arg: lck_file = le nom du fichier de lock
                   on ouvre en ecriture le fichier de lock et on pose un verrou avec fcntl sur ce fichier
   return:    le fichier lock ouvert (file descriptor) (-1 si LOCK_ACTIF est nul)
*/
int LockDir (const char * lck_file)
{
	int   lck_fd;
	struct flock lock;
	int rvl;
	
	/* Si LOCK_ACTIF est null, pas de lock du tout ! */
	if (!LOCK_ACTIF)
		return -1;
	
	/* ouverture du lock en mode write, sortir du programme si probleme */
	lck_fd = open(lck_file,O_RDWR);
	if (lck_fd==-1)
	{
		perror("ERREUR lors de l'ouverture du fichier lock ");
		exit(1);
	}
	
	/* Pose d'un verrou en ecriture sur ce lock, avec attente si le lock est deja verrouille.
	   On verrouille la totalite du fichier (offset 0, longueur 0).
	*/
	lock.l_type   = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start  = 0;
	lock.l_len    = 0;
	rvl = fcntl(lck_fd,F_SETLKW,&lock);
	if (rvl==-1)
	{
		perror("ERREUR lors de la pose du verrou sur le fichier lock ");
		exit(1);
	}
		
	return lck_fd;
}

/* UnLockDir = On deverrouille le repertoire, pour laisser aux autres process l'occasion de travailler 
   arg:  lck_fd = le descripteur de fichier semaphoe ouvert et verrouille
*/ 

void UnLockDir(int lck_fd)
{
	int rvl;
	struct flock lock;
	
	/* Si LOCK_ACTIF est null, pas de lock du tout ! */
	if (!LOCK_ACTIF)
		return;

	/* Suppression du verrou pose sur ce lock, sans attente parce qu'il n'y a aucune raison que ca ne marche pas */
	lock.l_type   = F_UNLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start  = 0;
	lock.l_len    = 0;
	
	rvl = fcntl(lck_fd,F_SETLK,&lock);
	if (rvl==-1)
	{
		perror("ERREUR lors de la suppression du verrou sur le fichier lock ");
		exit(1);
	}
	
	/* Fermeture du descripteur de fichiers */
	rvl = close(lck_fd);
	if (rvl==-1)
	{
		perror("ERREUR lors de la fermeture du fichier lock ");
		exit(1);
	}
	
	return;
}


/* CalcStamp = calcul d'un "stamp" a partir de l'heure courante, du pid et du nom de machine
               retourne une chaine du type: #08#07#14#23#30#15#   14051#  hostname
			   Le premier champ peut etre trie aisement (mais la reolution n'est que de 1 seconde
			   
   return : le stamp calcule
            ATTENTION IL FAUDRA FAIRE UN free POUR EVITER LES FUITES DE MEMOIRE
*/
char* CalcStamp()
{
	int sec,min,hour,mday,mon,year,pid;
    struct tm* now;
	static char* current_stamp=NULL;
	char* stamp_time;
	char* hostname;
	time_t now_s;
	static time_t last_now_s=0;
	static int cpt=0;
	stamp_time=(char *)malloc(100);
	
	/* comparer l'heure qu'il est a l'heure qu'il etait lors du dernier passage */
	/* si egales = incrementer le compteur (modulo 100) */
	now_s = time(NULL);
	if (now_s==last_now_s)
	{
		cpt = (cpt+1) % 100;
	}
	else
	{
		cpt = 0;
		last_now_s = now_s;
	}
	
	now = localtime(&now_s);
	sec = now->tm_sec;
	min = now->tm_min;
	hour= now->tm_hour;
	mday= now->tm_mday;
	mon = now->tm_mon;
	mon += 1;
	year= now->tm_year;
	year= (1900 + year) % 2000;

	pid = getpid();
	
	hostname = (char*) malloc(21);
	gethostname(hostname, 20);
	
	/* calculer le stamp */
	sprintf(stamp_time,"#%02i#%02i#%02i#%02i#%02i#%02i#%02i# %06i %10s", year, mon, mday, hour, min, sec, cpt,pid,hostname);
	
	free(hostname);
	return stamp_time;    
}

/* ************************** M A I N ******************************** */

int main (int argc, char ** argv)
{
	/* Nom du repertoire  et des fichiers utilises */
    char * dir;
	char * log_file = (char * ) malloc (1000);
	char * cpt_file = (char * ) malloc (1000);
	char * stp_file = (char * ) malloc (1000);
	char * lck_file = (char * ) malloc (1000);
	char * stamp;
	struct stat dummy;
	
	/* initialiser le nom de repertoire */
    if (argc>1)
		dir = argv[1];
	else
		dir = "LOCK";

	/* initialiser les noms de fichiers */
	strcpy(log_file,dir);
	strcat(log_file,"/test-lock.log");
	strcpy(cpt_file,dir);
	strcat(cpt_file,"/test-lock.txt");
	strcpy(stp_file,dir);
	strcat(stp_file,"/stop");
	strcpy(lck_file,dir);
	strcat(lck_file,"/lock");
	
	/* si le repertoire n'existe pas, le creer */
	if (stat(dir,&dummy)==-1)
	{
		mkdir(dir,0777);
	};

	/* si le fichier stp existe, le supprimer */
	unlink(stp_file);

	/* si le fichier lock n'existe pas et s'il risque d'etre utile, le creer */
 	if (LOCK_ACTIF)
		creat(lck_file,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	
	stamp = CalcStamp();

	/* si le fichier n'existe pas, le creer et initialiser cpt a 1
	   NOTE - il n'y a pas de lock ici, on suppose que l'un des programmes est lance quelque temps avant les autres */
	if (stat(cpt_file,&dummy)==-1)
	{
		int cpt = 1;
		WriteCpt(cpt_file,cpt,log_file,stamp);
	}
		
	/* boucle infinie - pour en sortir, creer le fichier stop */

	Trace(log_file,stamp,"ENTREE DANS LA BOUCLE WHILE",0);
	while(1)
	{
		char * stamp;
		int cpt;
		int lck_fd;
		
		/* Si le fichier stop existe, on quitte le programme */
		if(stat(stp_file,&dummy) != -1)
			break;
		
		/* Verrouillage du repertoire */
		lck_fd = LockDir(lck_file);
		
		/* Calcul du stamp */
		stamp = CalcStamp();
		
		/* Lecture du compteur */
		cpt = ReadCpt(cpt_file,log_file,stamp);
		
		/* En cas de probleme lors de la lecture, on laisse tomber pour cette iteration */
		if (cpt < 0)
		{
			Trace(log_file,stamp,"ITERATION ABANDONNEE",0);
			free(stamp);
			UnLockDir(lck_fd);
			WaitRand();
			continue;
		}
		
		/* Incrementation du compteur */
		cpt++;
		
		/* Ecriture du compteur dans le fichier */
		WriteCpt(cpt_file,cpt,log_file,stamp);
		
		/* Fin de cette iteration */
		UnLockDir(lck_fd);
		
		free(stamp);
		
		/* on attend un temps aleatoire */
		WaitRand();
		continue;
	}
	
	stamp = CalcStamp();
	Trace(log_file,stamp,"FIN    DE   LA BOUCLE WHILE",0);

	exit(0);
}


