
# $Id: Blast.pm 238 2009-01-06 08:16:20Z manu $

package PARALOOP::PLUGIN::Blast;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Blast - Executing blast (any version)

=head1 Derives from...

PARALOOP::PLUGIN::BpInput

=head1 DESCRIPTION

This plugin reads the next record from the fasta input file, uses this file as a
query for a blastall -p request agains the bank, passed through the --db switch
The tabular output is chosen

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Blast_origin 
     ncbi or wu: both are supported, but default is ncbi
 PARALOOP_Blast_path
     path to the binary program (default blastall if ncbi, blastp if wu)
 PARALOOP_Blast_params
     Parameters for blast (default -p blastp if ncbi, nothing if wu)
 PARALOOP_Blast_chunk
     Group requests in chunks for a better efficiency (default 1, ie not grouped)

=head1 PUBLIC METHODS

 WhatParaloopPlugin
      my $msg=$obj->WhatParaloopPlugin()

 Parameters
      my $msg=$obj->Parameters()

 New
      my $obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);

 Exec
      my ($incr_size,$error,$killed) = $obj->Exec();

 SetOutputFile
      $obj->SetOutputFile('filename');

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Carp;
use File::Basename;

# The NAMESPACE constant are also set in paraloop
use constant NAMESPACE => 'PARALOOP_';

use Logger;
use PARALOOP::PLUGIN::BpInput;
use Runner;
use Fcntl;

our @ISA = qw( PARALOOP::PLUGIN::BpInput );

use constant NCBI_BLAST_PATH_DEFAULT   => 'blastall';
use constant NCBI_BLAST_PARAMS_DEFAULT => "-p blastp";
use constant WU_BLAST_PATH_DEFAULT     => 'blastp';
use constant WU_BLAST_PARAMS_DEFAULT   => "";
use constant BLAST_ORIGIN              => 'ncbi';

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 238 $ =~ /\d+/g); $r[0]};
}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin()
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin
{
    my $rvl = "";
    $rvl =
      "PARALOOP::PLUGIN::Blast v$PARALOOP::PLUGIN::Blast::VERSION - Using Runner v$Runner::VERSION - Derivating from PARALOOP::PLUGIN::BpInput v$PARALOOP::PLUGIN::BpInput::VERSION\n";
    $rvl .= "This plugin reads a query from the input file and does a blast against the databank (--db switch)\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the Blast plugin\n=================\n";
    $rvl .= "PARALOOP_Blast_origin\tncbi or wu ? (default ncbi)\n";
    $rvl .= "PARALOOP_Blast_path\tpath to the binary program (default blastall for ncbi, blastp for wu)\n";
    $rvl .= "PARALOOP_Blast_params\tParameters for blast (default -p blastp for ncbi, nothing for wu)\n";
    $rvl .= "                     \tinout, output, database switches MUST NOT be specified here\n";
    $rvl .= "                     \tExample: -p blastp -M BLOSUM62 -m 8\n";
    $rvl .= "PARALOOP_Blast_chunk\tCall blast for chunk requests grouped together (default 1)\n";
    return $rvl;
}

=head2 procedure _Init

 Usage:  CALLED BY UNIVERSAL::New
         my obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);
 Access: protected (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Logger (already initialized)
 globals:  none

=cut

sub _Init
{
    my $self    = shift;
    #print "Blast.pm"." ".join(' ',@_)."\n";
    my %args    = @_;
    my $o_param = $args{'-params'};
    my $o_log   = $args{'-log'};

    my ($inputfile, $outputfile, $db);

    if ($o_param->IsDefined('input'))
    {    # check if inputfile is correctly defined
        $inputfile = $o_param->Get('input');
        -r $inputfile or croak("ERROR - $inputfile cannot be opened in read mode");
    }
    else
    {
        croak("ERROR - the input file is not defined");
        $o_param->Usage();
    }

    if ($o_param->IsDefined('db'))
    {    
        
        # check if db is correctly defined
        $db = $o_param->Get('db');
        
        # If it is NOT an absolute path, we try (as a last resort) to prepend the env variable BLASTDB
        if ($db !~ /^\// and (exists $ENV{BLASTDB}))
        {
            my $blastdb = $ENV{BLASTDB} .'/' . $db;
            (-r $db) or (-r "$db.pal") or (-r "$db.nal") or (-r $blastdb) or (-r "$blastdb.pal") or (-r "$blastdb.nal") or croak("ERROR - $db not found - may be \$BLASTDB is incorrect ?");
        }
        
        # It it is an absolute path or BLASTDB does not exist
        else
        {
            (-r $db) or (-r "$db.pal") or (-r "$db.nal") or croak("ERROR - $db not found");
        }
    }
    else
    {
        croak("ERROR - the database file is not defined");
        $o_param->Usage();
    }

    if ($o_param->IsDefined('output'))
    {    # check if outputfile is correctly defined
        $outputfile = $o_param->Get('output');
        open(TEST, ">>$outputfile") or croak("ERROR - $outputfile cannot be opened in write mode");
        close(TEST);
    }
    else
    {
        croak("ERROR - Blast: the output file is not defined");
        &$o_param->Usage();
    }

    $self->SUPER::_Init(-params => $o_param, -inputfile => $inputfile, -log => $o_log);    # Init the superclass

    # Build the __bin variable
    my $bin          = "";
    my $blast_params = "";
    $o_param->Set('Blast_origin', 'ncbi') unless $o_param->IsDefined('Blast_origin');
    my $origin = $o_param->Get('Blast_origin');
    if ($origin eq 'ncbi')
    {
        $bin          = $o_param->IsDefined('Blast_path')   ? $o_param->Get('Blast_path')   : NCBI_BLAST_PATH_DEFAULT;
        $blast_params = $o_param->IsDefined('Blast_params') ? $o_param->Get('Blast_params') : NCBI_BLAST_PARAMS_DEFAULT;
    }
    else
    {
        $bin          = $o_param->IsDefined('Blast_path')   ? $o_param->Get('Blast_path')   : WU_BLAST_PATH_DEFAULT;
        $blast_params = $o_param->IsDefined('Blast_params') ? $o_param->Get('Blast_params') : WU_BLAST_PARAMS_DEFAULT;
    }

    system("which $bin >/dev/null 2>&1")==0 or croak "ERROR - $bin cannot be found or executed";

    if ($origin eq 'ncbi')
    {
        $bin .= " -i ITMP -d $db -o OTMP $blast_params";
    }
    else
    {
        $bin .= " $db ITMP -o OTMP $blast_params";
    }

    my $chunk = $o_param->IsDefined('Blast_chunk') ? $o_param->Get('Blast_chunk') + 0 : 1;
    if ($chunk <= 0)
    { 
        croak "ERROR - chunk=$chunk - should be positive";
    }
    if ($chunk > 1)
    {
        if ($o_param->IsDefined('interleaved') and $o_param->Get('interleaved') != 0)
        {
            croak "ERROR - the interleaved mode is NOT supported by this plugin when chunk is higher than 1 (currently $chunk)";
        };
        if ($o_param->IsDefined('step') and $o_param->Get('step') > 1)
        {
            croak "ERROR - step MUST be 1 when chunk is higher than 1 (currently $chunk)";
        };
    };
    
    # Compute job_id
    my $job_id = $self->_JobId($o_param);

    $self->{'__params'}     = $o_param;
    $self->{'__outputfile'} = $outputfile;
    $self->{'__job_id'}     = $job_id;
    $self->{'__db'}         = $db;
    $self->{'__bin'}        = $bin;
    $self->{'__chunk'}      = $chunk;
    $self->{'__end'}        = $o_param->Get('end')

}

=pod

=head2 Procedure SetOutputFile

 Usage:     &$opt->SetOutputFile('filename');
 Procedure: change the name of the output file
 Arg :      the output file name

=cut

sub SetOutputFile
{
    my $self       = shift;
    my $outputfile = shift;
    $self->{'__outputfile'} = $outputfile;
}

=pod

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Execute a blast
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
		                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
		                0: everything OK 
                       -1: killed by timeout
                       -2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec
{
    my $self    = shift;
    my $timeout = shift;
    my $chunk   = $self->{'__chunk'};

    if ($chunk == 1)
    {
        return $self->__SingleSeq($timeout);
    }
    else
    {
        return $self->__SeveralSeq($timeout);
    }
}

=head2 Function __SingleSeq

 Usage:    my ($incr_size, $error,$killed) = &self->__SingleSeq($timeout);
 Function: Called by Exec when blasts are made sequence by sequence
 Arg, Return: see Exec

=cut

sub __SingleSeq
{
    my $self        = shift;
    my $timeout     = shift;
    my $input_file  = $self->{'__inputfile'};
    my $output_file = $self->{'__outputfile'};
    my $o_log       = $self->{'__log'};
    my $blast_bin   = $self->{'__bin'};
    my $tmp         = $self->{'__tmp'};
    my $job_id      = $self->{'__job_id'};
    my $rewrite_input =  $self->{'__rewrite_input'};

    my $o_param=$self->{'__params'};
    my $wc = $o_param->Get('wc');
    my $cat = $o_param->Get('cat');

    my $stamp = $o_log->CalcStamp();
    my $itmp  = $tmp . '/' . &basename($input_file) . '_' . $job_id . '_' . $stamp;
    my $otmp  = $tmp . '/' . &basename($output_file) . '_' . $job_id . '_' . $stamp;
    my ($nb, $nl) = (0, 0);

    my $seq_nb = $self->Tell();
    my $seq    = $self->NextRecord();
    if (!defined($seq))
    {
        my $msg = "NO MORE SEQUENCES TO READ FROM $input_file\n";
        $o_log->Trace("$msg", $stamp, '1');
        return (0, -1, 0);
    }

    my $id = $seq->id();
    my $itmpfh = Bio::SeqIO->new(-file => ">$itmp", -format => 'fasta') or die "Cannot open $itmp for writing";
    $itmpfh->write_seq($seq);
    undef($itmpfh);

    # Call the preatom script 
    # The input of the script is connected to the blast request
    # The output is connected to $output_file
    # WARNING - The preatom script is ran in the paraloop's current directory, NOT in temporary !!!
    $self->CallExternalScript('preatom_script',$itmp,">>$output_file");

    # Execute the blast command
    my $cmd = $blast_bin;
    $cmd =~ s/ITMP/$itmp/;
    $cmd =~ s/OTMP/$otmp/;

    $o_log->Trace("$cmd\n", '.', 2);
    $o_log->StartTimer('Exec_timer');
    my ($sts, $killed) = RunExt(-cmd => $cmd, -timeout => $timeout);
    my $seqlen = $seq->length();
    if ($sts != 0)
    {
        $o_log->Trace("BLAST STATUS $sts - (REQUEST $seq_nb - $id - len $seqlen - 0 MATCHES)\n", 'Exec_timer', '1');
    }
    elsif (!-r $otmp or -z $otmp)
    {
        $o_log->Trace("BLAST NOMATCH - (REQUEST $seq_nb - $id - len $seqlen - 0 MATCHES)\n", 'Exec_timer', '1');
    }
    else
    {
        my $rvl_wc = `$wc -c -l $otmp`;
        chomp($rvl_wc);
        ($nl, $nb) = ($rvl_wc =~ /\A\s*(\d+)\s+(\d+) /);
        $o_log->Trace("BLAST OK - (REQUEST $seq_nb - $id - len $seqlen - $nl MATCHES)\n", 'Exec_timer', '1');
        my $out = new IO::File(">>$output_file") or die "Cannot open $output_file in append mode";
        my $in  = new IO::File($otmp)            or die "Cannot open $otmp for reading";
        while (my $line = <$in>)
        {
            print $out $line;
        }
    }
    
    # Call the postatom script 
    # The input of the script is again connected to the blast request 
    # The output is connected to $output_file
    # WARNING - The postatom script is ran in the paraloop's current directory, NOT in temporary !!!
    $self->CallExternalScript('postatom_script',$itmp,">>$output_file");
    
    # If status != 0 and if __rewrite_input on, write the input line to some file
    if ($sts != 0 and $rewrite_input != 0)
    {
        my $new_input = $output_file.'.in';
        `$cat $itmp >> $new_input`;
        my $msg = "Input data copied back to " . basename($new_input);
        $o_log->Trace("$msg\n", '.', 'R');
    };
        
    unlink($itmp, $otmp);
    return ($nb, 0, $killed);
}

=head2 Function __SingleSeq

 Usage:    my ($incr_size, $error,$killed) = &self->__SeveralSeq($timeout);
 Function: Called by Exec when blasts are made in chunks of sequences
 Arg, Return: see Exec

=cut

sub __SeveralSeq
{
    my $self        = shift;
    my $timeout     = shift;
    my $input_file  = $self->{'__inputfile'};
    my $output_file = $self->{'__outputfile'};
    my $o_log       = $self->{'__log'};
    my $blast_bin   = $self->{'__bin'};
    my $tmp         = $self->{'__tmp'};
    my $job_id      = $self->{'__job_id'};
    my $end         = $self->{'__end'};
    my $chunk       = $self->{'__chunk'};
    my $rewrite_input =  $self->{'__rewrite_input'};

    my $o_param=$self->{'__params'};
    my $wc = $o_param->Get('wc');
    my $cat = $o_param->Get('cat');

    my $stamp = $o_log->CalcStamp();
    my $itmp  = $tmp . '/' . &basename($input_file) . '_' . $job_id . '_' . $stamp;
    my $otmp  = $tmp . '/' . &basename($output_file) . '_' . $job_id . '_' . $stamp;
    my ($nb, $nl) = (0, 0);

    my $seq_nb = $self->Tell();
    my $seq    = $self->NextRecord();
    if (!defined($seq))
    {
        my $msg = "NO MORE SEQUENCES TO READ FROM $input_file\n";
        $o_log->Trace("$msg", $stamp, '1');
        return (0, -1, 0);
    }

    my @id     = ();
    my $itmpfh = Bio::SeqIO->new(-file => ">$itmp", -format => 'fasta') or die "Cannot open $itmp for writing";

    push(@id, $seq->id());    # push/write first sequence
    $itmpfh->write_seq($seq);
    
    for (my $s = 1; $s < $chunk; $s++)    # push/write other sequences
    {
        last if ($self->Tell() > $end);
        $seq = $self->NextRecord();
        push(@id, $seq->id());
        $itmpfh->write_seq($seq);
    }
    undef($itmpfh);

    # Call the preatom script 
    # The input of the script is connected to the blast request
    # The output is connected to $output_file
    # WARNING - The preatom script is ran in the paraloop's current directory, NOT in temporary !!!
    $self->CallExternalScript('preatom_script',$itmp,">>$output_file");

    # Execute the blast command
    my $cmd = $blast_bin;
    $cmd =~ s/ITMP/$itmp/;
    $cmd =~ s/OTMP/$otmp/;

    $o_log->Trace("$cmd\n", '.', 2);
    $o_log->StartTimer('Exec_timer');
    my ($sts, $killed) = RunExt(-cmd => $cmd, -timeout => $timeout);
    if ($sts != 0)
    {
        $o_log->Trace("BLAST STATUS $sts - (REQUEST $seq_nb - " . join(',', @id) . " - 0 MATCHES)\n", 'Exec_timer',
                      '1');
    }
    elsif (!-r $otmp or -z $otmp)
    {
        $o_log->Trace("BLAST NOMATCH - (REQUEST $seq_nb - " . join(',', @id) . " - 0 MATCHES)\n", 'Exec_timer', '1');
    }
    else
    {
        my $rvl_wc = `$wc -c -l $otmp`;
        chomp($rvl_wc);
        ($nl, $nb) = ($rvl_wc =~ /\A\s*(\d+)\s+(\d+) /);
        $o_log->Trace("BLAST OK - (REQUEST $seq_nb - " . join(',', @id) . " - $nl MATCHES)\n", 'Exec_timer', '1');
        my $out = new IO::File(">>$output_file") or die "Cannot open $output_file in append mode";
        my $in  = new IO::File($otmp)            or die "Cannot open $otmp for reading";
        while (my $line = <$in>)
        {
            print $out $line;
        }
    }

    # Call the postatom script 
    # The input of the script is again connected to the blast request 
    # The output is connected to $output_file
    # WARNING - The postatom script is ran in the paraloop's current directory, NOT in temporary !!!
    $self->CallExternalScript('postatom_script',$itmp,">>$output_file");
    
    # If status != 0 and if __rewrite_input on, write the input line to some file
    if ($sts != 0 and $rewrite_input != 0)
    {
        my $new_input = $output_file.'.in';
        `$cat $itmp >> $new_input`;
        my $msg = "Input data copied back to " . basename($new_input);
        $o_log->Trace("$msg\n", '.', 'R');
   };
        
    unlink($itmp, $otmp);
    return ($nb, 0, $killed);
}
=pod

=head2 Function StealWork

 Usage: my $sts = $obj->StealWork($o_scheduler)
 function:  Recompute $threshold and call $obj->SUPER::StealWork($o_scheduler,$threshold)
 precondition: $o_scheduler must exist and be initialized
 Return:   0 if nothing stolen, 1 if something stolen
 Args:     $o_scheduler
 Globals:  none

=cut

sub StealWork
{
    my $self = shift;
    my ($o_sched) = (@_);
    
    my $o_params = $self->{'__params'};
    
    my $threshold = ($o_params->IsDefined('load_balancing_threshold') ? $o_params->Get('load_balancing_threshold') : main::DEF_LOAD_BALANCING_THRESHOLD);
    my $f_chunk   = ($o_params->IsDefined('Blast_chunk')) ? $o_params->Get('Blast_chunk') + 0 : 1;

    #
    # Change the value of threshold, because it MUST be a multiple of chunk
    #

    if ($f_chunk > 1)
    {
        $threshold = $f_chunk * int ($threshold / $f_chunk);
        $threshold = $f_chunk if ($threshold==0);
    };

    # Call StealWork from the super class    
    return $self->SUPER::StealWork($o_sched,$threshold);

}

